/*
** Module   : LRUCache.java
** Abstract : LRU cache implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140925 Created initial version.
** 002 IAS 20160331 Fixed NodeImpl.equals().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.cache;

/**
 * An implementation of a least-recently-used (LRU) cache. A single element (the least recently
 * used) is expired from the cache when capacity for a new element is needed.
 * <p>
 * Registered listeners are notified of each expiry event.
 * <p>
 * <strong>This implementation is not synchronized</strong>. Any access to the methods of this
 * class which modify the internal state of the cache must be synchronized externally.
 * 
 * @param   K
 *          Type of cache element key.
 * @param   V
 *          Type of cache element value.
 */
public class LRUCache<K, V>
extends ExpiryCache<K, V>
{
   /**
    * Constructor.
    * 
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LRUCache(int limit)
   {
      super(limit);
   }
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Cache name, for debug and logging purposes.
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LRUCache(String name, int limit)
   {
      super(name, limit);
   }
   
   /**
    * Get the number of elements which will be expired from the cache each time it becomes full,
    * and capacity is required for a new element.
    * 
    * @return  Number of elements to expire on each clearing pass, in this case 1.
    */
   protected int getCleanLimit()
   {
      return 1;
   }
   
   /**
    * Create a node of the appropriate type for the given key and value.
    * 
    * @param   key
    *          Hashable key.
    * @param   value
    *          Cached value.
    * 
    * @return  A sortable node to store in the cache.
    */
   protected Node<K, V> createNode(K key, V value)
   {
      return new NodeImpl<K, V>(key, value);
   }
   
   /**
    * A sortable node containing the hashable key and value of a cache entry. The node's natural
    * order is determined by the recentness of its last access (creation or retrieval). Less
    * recently accessed nodes sort before more recently accessed nodes.
    * 
    * @param <K>
    *        Hashable key
    * @param <V>
    *        Cached value
    */
   private static class NodeImpl<K, V>
   extends Node<K, V>
   {
      /** Next available node identifier */
      private static long nextAccess = 0;
      
      /** Unique counter assigned each time this element is accessed (created or retrieved) */
      private long accessID;
      
      /**
       * Constructor. Sets initial access time to now, and hit count to 0.
       * 
       * @param   key
       *          Hashable key.
       * @param   value
       *          Cached value.
       */
      NodeImpl(K key, V value)
      {
         super(key, value);
      }
      
      /**
       * Compare this node to another. A node less recently accessed sorts before a more recently
       * accessed node.
       * <p>
       * This algorithm generally leaves the least desirable element first in the sorted result.
       */
      public int compareTo(Node<K, V> n)
      {
         // don't catch ClassCastException, because we control the type of node in the cache
         NodeImpl<K, V> that = (NodeImpl<K, V>) n;
         
         return (int) (this.accessID - that.accessID);
      }
      
      /**
       * An implementation of equivalence checking consistent with <code>compareTo(Node)</code>.
       * 
       * @param   o
       *          Object with which to compare this instance.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this instance, else
       *          <code>false</code>.
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof NodeImpl))
         {
            return false;
         }
         
         // don't catch ClassCastException, because we control the type of node in the cache
         NodeImpl<K, V> that = (NodeImpl<K, V>) o;
         
         return (this.accessID == that.accessID);
      }
      
      /**
       * Assign the next available (higher) access ID to this node.
       */
      protected void access()
      {
         accessID = nextAccess++;
      }
   }
}