/*
** Module   : Emailer.java
** Abstract : helper class to send an email
**                       
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20171020 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.email;

import org.apache.commons.mail.*;
import javax.mail.*;
import java.io.*;
import java.util.*;
import java.net.*;
                    
/**
 * This is a helper class that makes it easy to send an email.
 */
public class Emailer
{
   /**
    * Constructs and sends an email using the given data and configuration.
    *
    * @param    def
    *           The configuration and data needed to send the email.
    */
   public static void send(EmailDefinition def)
   throws EmailException,
          MalformedURLException
   {
      HtmlEmail email = new HtmlEmail();
      
      setupSession(email, def);
      setupEmail(email, def);
      
      email.send();
   }
   
   /**
    * Configure the email instance for the correct connection security type, host, port and
    * authentication needed.
    *
    * @param    def
    *           The configuration and data needed to send the email.
    */
   public static void setupSession(HtmlEmail email, EmailDefinition def)
   throws EmailException
   {
      DefaultAuthenticator auth = null;
      
      String host = (def.host == null) ? "localhost" : def.host;
      
      email.setHostName(host);
        
      if (def.userid != null || def.password != null)
      {
         // include this for password authentication on SSL/STARTTLS/unencrypted connection
         // omit this if there is no authentication
         auth = new DefaultAuthenticator(def.userid   == null ? "" : def.userid,
                                         def.password == null ? "" : def.password);
         email.setAuthenticator(auth);
      }
        
      if (def.debug)
      {
         email.setDebug(true);
      }
                                                               
      // enable the correct connection type security
      if (def.connectType == ConnectionSecurity.STARTTLS_WHEN_AVAILABLE ||
          def.connectType == ConnectionSecurity.STARTTLS_REQUIRED)
      {
         // "TLS" (default port 587) which is really just STARTTLS on 587
         // STARTTLS (default port 25)
         email.setStartTLSEnabled(true);
         
         // add this if it should fail when STARTTLS is not active on the server
         if (def.connectType == ConnectionSecurity.STARTTLS_REQUIRED)
         {
            email.setStartTLSRequired(true);
         }
      }
      else if (def.connectType == ConnectionSecurity.SSL)
      {
         // SSL (default port 465)
         email.setSSLOnConnect(true);
      }
      
      // set the port
      if (def.connectType != ConnectionSecurity.SSL)
      {
         int port = (def.connectType != ConnectionSecurity.UNENCRYPTED) ? 587 : 25;
         
         if (def.port != null)
         {
            port = Integer.parseInt(def.port);
         }
         
         email.setSmtpPort(port);
      }
      else
      {
         email.setSslSmtpPort((def.port == null) ? "465" : def.port);
      }
      
      if (!def.validate)
      {
         // it is not clear what this really does since it does not seem to be honored, but it
         // seems prudent to set it
         email.setSSLCheckServerIdentity(false);
      
         // TODO: this was previously working and now is not, it is not clear why
         //       it is broken
         
         // this disables server cert validation by overriding a specific property used by
         // javamail (the alternative is the Java command line property -Dmail.smtp.ssl.trust=)
         // this must be done after all the other settings are assigned AND we need to re-use
         // the auth instance if it is being used above
         Session    session = email.getMailSession();
         Properties props   = session.getProperties();
         props.put("mail.smtp.ssl.trust", host);
         props.put("mail.smtps.ssl.trust", host);
         email.setMailSession(session.getInstance(props, auth));
      }         
   }
   
   /**
    * Configure the email instance for the FROM, TO, CC, BCC addresses, the subject and the
    * content.
    *
    * @param    def
    *           The configuration and data needed to send the email.
    */
   public static void setupEmail(HtmlEmail email, EmailDefinition def)
   throws EmailException,
          MalformedURLException
   {
      // from address is required
      email.setFrom(def.from);
      
      if (def.replyTo != null)
      {
         email.addReplyTo(def.replyTo);
      }
      
      if (def.subject != null)
      {
         email.setSubject(def.subject);
      }
      
      String[] to = def.getToList();
      
      if (to.length != 0)
      {
         email.addTo(to);
      }
      
      String[] cc = def.getCcList();
      
      if (cc.length != 0)
      {
         email.addCc(cc);
      }
      
      String[] bcc = def.getBccList();
      
      if (bcc.length != 0)
      {
         email.addBcc(bcc);
      }
      
      AttachmentDescriptor[] atts = def.getAttachmentsList();
      
      for (AttachmentDescriptor a : atts)
      {
         if (a.type == ResourceType.PATH)
         {
            File file = new File(a.resource); 
            email.attach(file);
         }
         else
         {
            email.attach(new URL(a.resource), "", "", EmailAttachment.ATTACHMENT);
         }
      }
      
      if (def.text != null)
      {
         email.setTextMsg(def.text);
      }
      
      processHtml(email, def);
   }
   
   /**
    * Process embedded images and modify the 
    */
   public static void processHtml(HtmlEmail email, EmailDefinition def)
   throws EmailException,
          MalformedURLException
   {
      if (def.html == null || def.html.length() == 0)
      {
         return;
      }
      
      String html = def.html;
      
      EmbedDescriptor[] embeds = def.getEmbedsList();
      
      for (EmbedDescriptor e : embeds)
      {
         String old = e.cid;
         
         if (e.type == ResourceType.PATH)
         {
            File file = new File(e.resource);
            e.cid = email.embed(file, e.cid);
         }
         else
         {
            e.cid = email.embed(new URL(e.resource), e.name);
         }
         
         html = html.replace(old, e.cid);
      }
      
      def.html = html;
      email.setHtmlMsg(html);
   }

   
   /**
    * Command line interface for sending emails.
    * <p>
    * Syntax:
    * <p>
    * <pre>
    * java -classpath p2j.jar com.goldencode.email.Emailer <mode> <host> <port> <userid> <password> <from> <to> <subject> <text>
    *
    * mode     = UNENCRYPTED, STARTTLS_WHEN_AVAILABLE, STARTTLS_REQUIRED or SSL
    * host     = IP address or hostname of the SMTP server
    * port     = port of the SMTP server
    * userid   = user for SMTP authentication or the text "null" for no userid
    * password = password for SMTP authentication or the text "null" for no password
    * from     = FROM email address
    * to       = TO email address
    * subject  = SUBJECT line
    * text     = text content of the message
    * </pre>
    *
    * @param    args
    *           Command line arguments.
    */
   public static void main(String[] args)
   {
      try
      {
         if (args.length < 9)
         {
            System.err.println("Missing arguments!\nSyntax: java -classpath p2j.jar " + 
                               "com.goldencode.email.Emailer <mode> <host> <port> " +
                               "<userid> <password> <from> <to> <subject> <text>\n");
            return;
         }
         
         EmailDefinition def = new EmailDefinition();
         
         def.connectType = ConnectionSecurity.valueOf(args[0]);
         def.host        = args[1];
         def.port        = args[2];
         def.userid      = ("null".equalsIgnoreCase(args[3]) ? null : args[3]);
         def.password    = ("null".equalsIgnoreCase(args[4]) ? null : args[4]);
         def.from        = args[5];
         def.addTo(args[6]);
         def.subject     = args[7];
         def.text        = args[8];
         def.debug       = true;
         def.validate    = false;
         
         Emailer.send(def);
      }
      
      catch (EmailException | MalformedURLException ee)
      {
         ee.printStackTrace();
      }
   }
}