/*
** Module   : CodeAttribute.java
** Abstract : Code_attribute structure used within Java class file format
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041130   @21385 Production version from JTA, reworked
**                           into a new package and with minor cleanups
**                           in javadoc.
** 002 ECF 20050527   @21386 Expression engine rewrite. Replaced
**                           references to ExpressionCompiler with
**                           Compiler. Added toString method for debug
**                           purposes. Made class package private.
** 003 ECF 20150715          Replace StringBuffer with StringBuilder.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;
import java.util.*;

/**
 * This class manages information necessary to create the <code>Code</code>
 * attribute structure for a method.  This includes information about the
 * maximum stack depth and maximum local variables, exception table data,
 * and the bytecode array itself.
 */
final class CodeAttribute
extends Attribute
{
   /** Maximum depth, in 32-bit slots, of the operand stack. */
   private int maxStack = 0;
   
   /** Maximum number of local variables used at one time in the method. */
   private int maxLocals = 0;
   
   /** List of all bytecode containers which compose the code array. */
   private List code = new ArrayList();
   
   /** List of all exception tables associated with the code array. */
   private List excTables = new ArrayList();
   
   /** List of all attributes associated with this Code attribute. */
   private List attrs = new ArrayList();
   
   /** Number of associated attributes. */
   private int attrLength = 0;
   
   /**
    * Constructor.  Creates a <code>Code</code> attribute object associated
    * with the specified constant pool.
    *
    * @param   pool
    *          Constant pool with which this code attribute is associated.
    */
   CodeAttribute(ConstantPool pool)
   {
      super(pool, CODE);
   }
   
   /**
    * Get a string representation of this object. This is a line separated
    * listing of the code attributes contents.
    *
    * @return  String representation of this object.
    */
   public String toString()
   {
      String sep = System.getProperty("line.separator");
      StringBuilder buf = new StringBuilder();
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         buf.append(iter.next())
            .append(sep);
      }
      
      return buf.toString();
   }
   
   /**
    * Return the total length in bytes of this code attribute, including
    * the bytecode array and all other data.
    *
    * @return  Total length in bytes of this attribute.
    */
   protected int getLength()
   {
      return (12 + getCodeLength() + (excTables.size() * 8) + attrLength);
   }
   
   /**
    * This method writes the data content of this code attribute to the
    * specified output stream.  This includes all bytecode instructions,
    * exception tables, and other data associated with the attribute.
    * Currently, no associated attributes are written.
    *
    * @param   out
    *          Output stream to which to write data.
    *
    * @exception  IOException
    *             if any error occurs during writing.
    */
   protected void writeContent(DataOutput out)
   throws IOException
   {
      layoutCode();
      
      out.writeShort(maxStack);
      out.writeShort(maxLocals);
      out.writeInt(getCodeLength());
      
      // Write instruction bytecodes.
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         WritableData next = (WritableData) iter.next();
         next.write(out);
      }
      
      out.writeShort(excTables.size());
      // future:  write exception tables here...
      
      out.writeShort(attrs.size());
      // future:  write attributes here...
   }
   
   /**
    * This method adds a single instruction to the end of the code array.
    *
    * @param   instruc
    *          Instruction to add.
    */
   void addInstruction(Instruction instruc)
   {
      code.add(instruc);
   }
   
   /**
    * This method adds a unit of code to the end of the code array.
    *
    * @param   unit
    *          Code unit to add to the end of the array.
    */
   void addCodeUnit(CodeUnit unit)
   {
      code.add(unit);
   }
   
   /**
    * This method iterates over all of the bytecode containers in the code
    * array to determine the maximum depth which the JVM's operand stack
    * may reach in the execution of this bytecode.
    *
    * @param   compiler
    *          Object which manages the maximum stack depth.
    */
   void setMaxStack(Compiler compiler)
   {
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.modifyStackDepth(compiler);
      }
      maxStack = compiler.getMaxStackDepth();
   }
   
   /**
    * Set the maximum number of local variable used at one time in this
    * method.
    *
    * @param   maxLocals
    *          Maximum number of local variables.
    */
   void setMaxLocals(int maxLocals)
   {
      this.maxLocals = maxLocals;
   }
   
   /**
    * This method calls <code>fixupStack</code> on each bytecode container
    * in this attribute.  The purpose is to remove any superfluous stack
    * entries before branching control when checking for null variable
    * references at runtime.
    */
   void fixupStack()
   {
      Stack stack = new Stack();
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.fixupStack(stack);
      }
   }
   
   /**
    * Return the length of the bytecode array in this code attribute,
    * measured in bytes.
    *
    * @return  Number of bytes in the bytecode array.
    */
   private int getCodeLength()
   {
      int length = 0;
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         length += next.getLength();
      }
      
      return length;
   }
   
   /**
    * This method assigns the final address offsets to each unit of code
    * in the code array of this attribute, then fixes up all of the
    * instructions which cause a branch to another instruction.
    */
   private void layoutCode()
   {
      // Layout all instructions with their final offsets.
      int offset = 0;
      Iterator iter = code.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         offset = next.layout(offset);
      }
      
      // Fixup all branch instructions.
      iter = code.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.fixupBranchOffset();
      }
   }
}