/*
** Module   : CodeUnit.java
** Abstract : A container of bytecode instructions and other code units
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041130   @19657 Production version from JTA, reworked
**                           into a new package and with minor cleanups
**                           in javadoc.
** 002 ECF 20050207   @19658 Added support for the expression compiler
**                           overhaul. Added the concept of stack-scope
**                           and the stackMarker flag, which allows a
**                           logical code unit which must branch to code
**                           which prepares interim results to properly
**                           fixup its stack within the proper scope.
** 003 ECF 20050526   @21308 Modifications to support the expression
**                           engine rewrite. Large portions of this class
**                           were left intact; however, the notion of
**                           result type integral constants was replaced
**                           with a return type Class object. The inner
**                           class BranchData was adapted to use ExtraAst
**                           as its logical branch destination placeholder
**                           and to use ExpressionParserTokenTypes instead
**                           of the old CompilerConstants result types.
** 004 ECF 20050615   @21495 Removed NEG_METH_BOOLEAN constant references.
**                           This token type is no longer in use.
** 005 ECF 20150715          Replace StringBuffer with StringBuilder.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;
import java.util.*;

/**
 * Instances of this class organize one or more bytecode instructions into
 * logical units of code.  Each unit of code represents a discrete portion
 * of an assembled expression.  Code units may be nested within one another
 * to an arbitrary depth, but the nesting must never by circular.
 * <p>
 * A code unit generally has a data type associated with it, though this is
 * not strictly necessary in all cases. The data type generally corresponds
 * with the type of data last left on the stack by the code unit's
 * instructions. The bytecode assembler often refers to a code unit's data
 * type to determine how this code unit can be nested or chained with other
 * code units, and whether any additional instructions are necessary to
 * generate another, target result type.
 * <p>
 * A code unit can contain a simple list of instructions and/or nested code
 * units, and may optionally define a single {@link CodeUnit.BranchData}
 * object. Branch data is used when the purpose of the code unit is to jump
 * to another instruction in the method, based upon some test done on the top
 * one or two runtime stack values (or unconditionally, in the case of the
 * <code>GOTO</code> branch).
 * <p>
 * Some branches require obsolete stack values to be popped off the stack
 * before jumping to another instruction. As a rule, the height of the stack
 * just before a jump occurs must be the same as it would be had the
 * destination instruction been reached by any other means (e.g., via linear
 * execution of instructions). To meet this requirement, some code units
 * are defined as <em>stack markers</em>, and others are defined as
 * <em>stack poppers</em>. A combination of stack marker and stack popper
 * code units are used by the bytecode assembler to ensure the proper
 * <code>pop</code> and/or <code>pop2</code> instructions are added before
 * a branching instruction jumps to another location. This fixup is done
 * after the primary assembly of functional bytecode instructions, in a
 * final compilation step using {@link CodeAttribute#fixupStack}. This
 * invocation eventually resolves down to this class' {@link #fixupStack}
 * method, which does the real work of inserting the necessary pops.
 */
class CodeUnit
extends BytecodeContainer
implements ExpressionParserTokenTypes
{
   /** Inner code units and instructions */
   protected List contents = new ArrayList();
   
   /** Data associated with a logic branch */
   private BranchData branchData;
   
   /** Indicates a special stack-management code unit */
   private boolean stackMarker = false;
   
   /** Result type of this code unit */
   private boolean stackPopper = false;
   
   /** The data type associated with this code unit */
   private Class dataType = null;
   
   /**
    * Default constructor. The resulting code unit's data type is
    * <code>null</code>.
    */
   CodeUnit()
   {
   }
   
   /**
    * Constructor which assigns a data type to the resulting code unit. The
    * data type generally implies the purpose of the code unit, and usually
    * corresponds with the type of data left on the runtime stack by this
    * code units instructions.
    *
    * @param   dataType
    *          Data type to be associated with the new code unit.
    */
   CodeUnit(Class dataType)
   {
      setDataType(dataType);
   }
   
   /**
    * This constructor is used to create a code unit containing a {@link
    * CodeUnit.BranchData} object which branches to the specified destination
    * under the specified condition.
    *
    * @param   condition
    *          Condition under which code unit branches to
    *          <code>target</code>'s code unit. This may be ignored,
    *          depending upon the <code>type</code> specified. For
    *          instance, a type <code>GOTO</code> branch is unconditional,
    *          and therefore does not use <code>condition</code> as a
    *          branch trigger.
    * @param   target
    *          An AST which contains information about the destination of
    *          the branch instruction implied by <code>type</code> and
    *          <code>condition</code>. This AST will eventually have a
    *          <code>CodeUnit</code> object stored in its <code>extra</code>
    *          field, which represents the destination of <em>this</em>
    *          code unit's branch instruction.
    * @param   type
    *          Type of branch logic;  indicates which bytecode instruction
    *          is used, depending also on <code>condition</code>.
    * @param   binary
    *          <code>true</code> if this code unit manages a binary operation
    *          (i.e., requires two operands from the runtime stack). If
    *          <code>false</code>, the branch will only consider a single
    *          stack operand (unary), or no operands (unconditional),
    *          depending upon the <code>type</code> parameter.
    *
    * @throws  ExpressionException
    *          if an invalid branching condition is detected.
    */
   CodeUnit(boolean condition, ExtraAst target, int type, boolean binary)
   {
      branchData = new BranchData(condition, target, type, binary);
      setDataType(Boolean.TYPE);
   }
   
   /**
    * This method adds the specified instruction to the end of this code
    * unit.
    *
    * @param   instruc
    *          Instruction to be added.
    */
   void addInstruction(Instruction instruc)
   {
      if (instruc == null)
      {
         throw new NullPointerException();
      }
      contents.add(instruc);
   }
   
   /**
    * Convenience method to add a new instruction. This can only be used
    * if the instruction requires no arguments.
    *
    * @param   opCode
    *          Operation code of the bytecode instruction to add.
    */
   void addInstruction(int opCode)
   {
      contents.add(new Instruction(opCode));
   }
   
   /**
    * This method adds the specified code unit to the end of this code unit.
    * This effectively allows code units to be nested to an arbitrary depth.
    *
    * @param   unit
    *          Code unit to be added.
    */
   void addCodeUnit(CodeUnit unit)
   {
      contents.add(unit);
   }
   
   /**
    * Return the data type associated with this code unit.
    *
    * @return  Data type.
    */
   Class getDataType()
   {
      return dataType;
   }
   
   /**
    * Set the data type associated with this code unit.
    *
    * @param   dataType
    *          Data type.
    */
   void setDataType(Class dataType)
   {
      this.dataType = dataType;
   }
   
   /**
    * Used to indicate that this code unit defines a stack-managing scope,
    * meaning it contains a <code>goto</code> instruction. Before the
    * <code>goto</code> is executed, the runtime stack must be rid of any
    * values which will no longer be relevant at the destination bytecode.
    * bytecode. CodeUnits of type <code>stackPopper</code> will add
    * instructions to pop the appropriate number of values off the stack
    * before the goto. The "appropriate" number of values to be popped is
    * determined by the stack-managing scope of a code unit. Only as many
    * values as have been pushed within this scope at the time of goto will
    * be popped.
    *
    * @param   stackMarker
    *          <code>true</code> to indicate that this code unit defines a
    *          stack-managing scope as described above; else
    *          <code>false</code>. The default value of the stack marker
    *          attribute is <code>false</code>, so there is typically no
    *          need to call this method with a value of <code>false</code>.
    *
    * @see     #setStackPopper
    * @see     #fixupStack
    */
   void setStackMarker(boolean stackMarker)
   {
      this.stackMarker = stackMarker;
   }
   
   /**
    * Used to indicate that this code unit cooperates with code units of
    * type <code>stackMarker</code> to clear the stack of irrelevant data
    * values before a branch instruction. CodeUnits of this type will add
    * instructions to pop the appropriate number of values off the stack
    * before the goto. The "appropriate" number of values to be popped is
    * determined by the stack-managing scope of a code unit. Only as many
    * values as have been pushed within that scope at the time of goto will
    * be popped.
    *
    * @param   stackPopper
    *          <code>true</code> to indicate that this code unit should
    *          add pop/pop2 instructions to clear the stack of irrelevant
    *          values within a stack marker scope. <code>false</code>
    *          indicates a normal code unit.
    *
    * @see     #setStackMarker
    * @see     #fixupStack
    */
   void setStackPopper(boolean stackPopper)
   {
      this.stackPopper = stackPopper;
   }
   
   /**
    * Return the length of this code unit. Its length is comprised of the
    * combined lengths of all of the bytecode containers which this code
    * unit itself contains.
    *
    * @return  Combined length of all instructions inside this code unit,
    *          including those within nested code units.
    */
   public int getLength()
   {
      int length = 0;
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         length += next.getLength();
      }
      
      return length;
   }
   
   /**
    * Each unit of code may have some impact on the maximum depth of the
    * JVM's runtime operand stack.  This method walks through each
    * instruction in this code unit, including those nested within
    * contained code units, and updates the stack depth variables in the
    * current compiler.
    *
    * @param   compiler
    *          Expression compiler which maintains stack depth information.
    */
   public void modifyStackDepth(Compiler compiler)
   {
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.modifyStackDepth(compiler);
      }
   }
   
   /**
    * This method assigns the specified offset address to the beginning of
    * this code unit and assigns the appropriate addresses to each
    * instruction and nested code unit inside this one.
    *
    * @param   offset
    *          Offset address of this code unit within the bytecode array
    *          for the method currently being compiled.
    *
    * @return  The offset address to be assigned to the code unit or
    *          instruction immediately following this one in the bytecode
    *          array.  This address is calculated by taking the offset
    *          address passed into this method as <code>offset</code> and
    *          adding the width of this code unit to that address.
    */
   public int layout(int offset)
   {
      this.offset = offset;
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         offset = next.layout(offset);
      }
      
      return offset;
   }
   
   /**
    * This method checks whether this code unit is designated as a
    * <code>stackPopper</code>. Code units of this type are placed within
    * code units which branch when an object reference is <code>null</code>,
    * immediately prior to a <code>goto</code> instruction. To avoid verify
    * errors at load time, we must ensure that at the time the branch is
    * executed, the stack depth must be exactly what it was when this code
    * unit was entered. If it is not, we insert the appropriate number of
    * <code>pop</code> and/or <code>pop2</code> instructions before the
    * <code>goto</code> to fix up the stack.
    * <p>
    * If this code unit is not a stackPopper, this method is called
    * recursively, on all {@link BytecodeContainer}s found within this
    * code unit.
    *
    * @param   stack
    *          A stack which represents the state of the runtime stack at
    *          the time this code unit will be executed in the final,
    *          compiled method.
    */
   void fixupStack(Stack stack)
   {
      // This code unit was inserted here (i.e., just before a goto) as a
      // placeholder to receive pop instructions. Determine how many (and
      // how wide) we need and insert them as necessary.
      if (stackPopper)
      {
         if (stack.empty())
         {
            return;
         }
         
         // For stack restoration when we're done.
         Stack reverse = new Stack();
         
         // Pop all the elements, pushing them into a temporary reverse order
         // stack for later restoration.
         while (!stack.empty())
         {
            Integer popped = (Integer) stack.pop();
            reverse.push(popped);
            int val = popped.intValue();
            
            if (val == 1)
            {
               addInstruction(new Instruction(_pop));
            }
            else if (val == 2)
            {
               addInstruction(new Instruction(_pop2));
            }
         }
         
         // Set the stack back to its original state.
         while (!reverse.empty())
         {
            stack.push(reverse.pop());
         }
         
         return;
      }
      
      // This code unit defines a stack-managing scope, meaning it contains
      // a goto. Before the goto is executed, the runtime stack must be rid
      // of any values which will no longer be relevant at the destination
      // bytecode. CodeUnits of type TYPE_STACKPOPPER will add instructions
      // to pop the appropriate number of values off the stack before the
      // goto. The "appropriate" number of values to be popped is determined
      // by the stack-managing scope of a code unit. Only as many values as
      // have been pushed within this scope at the time of goto will be
      // popped.
      if (stackMarker)
      {
         // Push a "1" onto the existing fixup stack to represent the boolean
         // return value from a logical subexpression.
         stack.push(new Integer(1));
         
         // Create a new "scope" by creating a new stack for use by child
         // code units.
         stack = new Stack();
      }
      
      // Walk through all child containers and fixup their stacks.
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.fixupStack(stack);
      }
   }
   
   /**
    * Add the branch instruction of a branching code unit to the end of the
    * current contents. This code unit must contain a <code>BranchData</code>
    * object.
    *
    * @throws  NullPointerException
    *          if this code unit has not defined <code>BranchData</code>
    *          object.
    */
   public void addBranchInstruction()
   {
      branchData.addBranchInstruction();
   }
   
   /**
    * This method is called after the final offsets for each bytecode
    * instruction have been assigned.  If this code unit contains any
    * branching instructions, the arguments of those instructions are
    * fixed up now, to ensure the instruction branches to the correct
    * offset in the bytecode array.  The method is then called on each
    * bytecode container contained within this code unit, so that they
    * may fix up their branching instructions as well.
    */
   public void fixupBranchOffset()
   {
      if (branchData != null)
      {
         branchData.fixupOffset();
      }
      
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         BytecodeContainer next = (BytecodeContainer) iter.next();
         next.fixupBranchOffset();
      }
   }
   
   /**
    * This method writes the bytecode contents of this code unit to the
    * specified output stream.
    *
    * @param   out
    *          Output stream to which data is written.
    *
    * @throws  IOException
    *          if any error occurs during writing.
    */
   public void write(DataOutput out)
   throws IOException
   {
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         WritableData next = (WritableData) iter.next();
         next.write(out);
      }
   }
   
   /**
    * Returns a string representation of this object, which is a line
    * separated listing of this unit's contents.
    *
    * @return  String representation of this object.
    */
   public String toString()
   {
      String sep = System.getProperty("line.separator");
      StringBuilder buf = new StringBuilder();
      Iterator iter = contents.iterator();
      while (iter.hasNext())
      {
         buf.append(iter.next())
            .append(sep);
      }
      
      return buf.toString();
   }
   
   /**
    * This class encapsulates information necessary to branch execution
    * of the JVM from one bytecode instruction to another in the bytecode
    * array.
    * <p>
    * A <code>BranchData</code> object provides a bridge between the current
    * code unit and a destination code unit. At the time the code unit
    * containing this object is being assembled, the destination code unit
    * typically has not yet been created. <code>BranchData</code> therefore
    * stores an {@link ExtraAst} instance which represents the logical
    * destination of the branch. This target AST acts as a placeholder for
    * the destination code unit. After the assembly of all code units in a
    * compiled expression's {@link CompiledExpression#execute execute} method
    * has been completed, a fixup is performed to calculate the required
    * offest necessary for this object's branch instruction to jump to the
    * beginning of the destination code unit.
    */
   protected class BranchData
   {
      /** Code unit which is the destination of the branch. */
      private ExtraAst target;
      
      /** Offset to the destination address from the current instruction. */
      private NumericArgument offsetArg;
      
      /** Instruction which will cause the branch to occur. */
      private Instruction instruction;
      
      /**
       * Constructor.  Selects an appropriate opcode for the operation,
       * based on the value of <code>condition</code>, creates a branch
       * instruction and its offset argument.  The final value of the
       * offset cannot be determined until the code array is laid out
       * and all branch instructions are fixed up, so an offset of 0 is
       * used as a temporary placeholder.
       *
       * @param   condition
       *          Condition under which code unit branches to
       *          <code>target</code>'s code unit. This may be ignored,
       *          depending upon the <code>type</code> specified. For
       *          instance, a type <code>GOTO</code> branch is unconditional,
       *          and therefore does not use <code>condition</code> as a
       *          branch trigger.
       * @param   target
       *          An AST which contains information about the destination of
       *          the branch instruction implied by <code>type</code> and
       *          <code>condition</code>. This AST will eventually have a
       *          <code>CodeUnit</code> object stored in its
       *          <code>extra</code> field, which represents the destination
       *          of the branch instruction.
       * @param   type
       *          Type of branch logic;  indicates which bytecode instruction
       *          is used, depending also on <code>condition</code>.
       * @param   binary
       *          <code>true</code> if this code unit manages a binary
       *          operation (i.e., requires two operands from the runtime
       *          stack). If <code>false</code>, the branch will only
       *          consider a single stack operand (unary), or no operands
       *          (unconditional), depending upon the <code>type</code>
       *          parameter.
       *
       * @throws  ExpressionException
       *          if an invalid branching condition is detected.
       */
      protected BranchData(boolean condition,
                           ExtraAst target,
                           int type,
                           boolean binary)
      throws ExpressionException
      {
         int opCode = 0;
         
         switch (type)
         {
            case EQUALS:
               opCode = (binary
                         ? (condition ? _if_icmpeq : _if_icmpne)
                         : (condition ? _ifeq : _ifne));
               break;
               
            case NOT_EQ:
            case METH_BOOLEAN:
               opCode = (binary
                         ? (condition ? _if_icmpne : _if_icmpeq)
                         : (condition ? _ifne : _ifeq));
               break;
               
            case LT:
               opCode = (binary
                         ? (condition ? _if_icmplt : _if_icmpge)
                         : (condition ? _iflt : _ifge));
               break;
               
            case LTE:
               opCode = (binary
                         ? (condition ? _if_icmple : _if_icmpgt)
                         : (condition ? _ifle : _ifgt));
               break;
               
            case GT:
               opCode = (binary
                         ? (condition ? _if_icmpgt : _if_icmple)
                         : (condition ? _ifgt : _ifle));
               break;
               
            case GTE:
               opCode = (binary
                         ? (condition ? _if_icmpge : _if_icmplt)
                         : (condition ? _ifge : _iflt));
               break;
               
            case IS_NULL:
               opCode = (condition ? _ifnull : _ifnonnull);
               break;
               
            case NOT_NULL:
               opCode = (condition ? _ifnonnull : _ifnull);
               break;
               
            case GOTO:
               opCode = _goto;
               break;
               
            default:
               throw new ExpressionException(
                  "Attempted branch instruction from an invalid code unit " +
                  type);
         }
         
         this.target = target;
         instruction = new Instruction(opCode);
         offsetArg = new NumericArgument(0, Argument.SHORT);
         instruction.addArgument(offsetArg);
      }
      
      /**
       * This method returns the instruction which performs the branch.
       */
      protected void addBranchInstruction()
      {
         addInstruction(instruction);
      }
      
      /**
       * This method sets the offset value of the argument to the branch
       * instruction.  It is this offset, when added to the address of the
       * current execution point, which determines the address of the
       * destination instruction.  This method is called after the final
       * addresses have been assigned to each instruction in the bytecode
       * array, so the branch offset is simply determined by subtracting
       * the address of the branch instruction from the address of the
       * destination instruction.
       */
      protected void fixupOffset()
      {
         CodeUnit destination = (CodeUnit) target.getExtra();
         int diff = destination.getOffset() - instruction.getOffset();
         offsetArg.setValue(diff);
      }
   }
}