/*
** Module   : ExtraAst.java
** Abstract : An AST which stores an object reference of the user's choice.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050509 ADD   @21137 Created initial version supporting an
**                               arbitrary object reference and line/column 
**                               numbers.
** 002 ECF 20050604 CHG   @21409 Added unique identifier and access methods.
**                               Used by diagnostic error reporting for
**                               expressions.
** 003 GES 20051003 CHG   @22924 Added new operand type member used by logical
**                               operators to store the common type of the
**                               operands.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import antlr.*;
import antlr.collections.*;

/**
 * A modified abstract syntax tree node which provides an additional service
 * for storing a single object reference of the user's choosing.  In addition,
 * basic support for line and column number is available.
 */
public class ExtraAst
extends CommonAST
{
   /** Extra, user-defined object reference associated with this node. */
   private Object extra = null;
   
   /**
    * Common operand type associated with this node (certain operators 
    * only).
    */
   private Class operandType = null;
   
   /** Source file line number associated with this AST */
   private int line = 0;
   
   /** Source file column number associated with this AST */
   private int column = 0;
   
   /** Identifier which is unique to this node within an expression AST */
   private int id = -1;
   
   /**
    * Default constructor.
    */
   public ExtraAst()
   {
      super();
   }
   
   /**
    * Get this AST node's identifier. This will be unique within an
    * expression's syntax tree.
    *
    * @return  Unique identifier for this node.
    */
   public int getId()
   {
      return id;
   }
   
   /**
    * Set this AST node's identifier. This will be unique within an
    * expression's syntax tree.
    *
    * @param   id
    *          Unique identifier for this node.
    */
   public void setId(int id)
   {
      this.id = id;
   }
   
   /**
    * Constructor which accepts and wrappers token.
    *
    * @param   token
    *          Token upon which this node is based.
    */
   public ExtraAst(Token token)
   {
      super(token);
      
      extractTokenData(token);
   }
   
   /**
    * Retrieve this AST's associated object reference.
    *
    * @return   The associated object reference or <code>null</code> if no
    *           reference has been assigned.
    */
   public Object getExtra()
   {
      return extra;
   }
   
   /**
    * Set this AST's associated object reference.
    *
    * @param    extra
    *           The associated object reference or <code>null</code> if the
    *           reference should be removed.
    */
   public void setExtra(Object extra)
   {
      this.extra = extra;
   }
      
   /**
    * Retrieve this AST's associated operand type.
    *
    * @return   The associated operand type or <code>null</code> if the
    *           reference should be removed.
    */
   public Class getOperandType()
   {
      return operandType;
   }
   
   /**
    * Set this AST's associated operand type.
    *
    * @param    type
    *           The associated operand type or <code>null</code> if the
    *           reference should be removed.
    */
   public void setOperandType(Class type)
   {
      operandType = type;
   }
      
   /**
    * Initialize this node with the specified AST. Delegates to the superclass
    * and copies data from <code>ast</code>.
    *
    * @param   ast
    *          AST upon which this node is based. Expected to be an instance
    *          of {@link ExtraAst}.
    */
   public void initialize(AST ast)
   {
      super.initialize(ast);
      
      if (ast instanceof ExtraAst)
      {
         ExtraAst east = (ExtraAst) ast;
         
         // Copy immutable data.
         this.line        = east.getLine();
         this.column      = east.getColumn();
         this.extra       = east.getExtra();         
         this.operandType = east.getOperandType();
      }
   }
   
   /**
    * Initialize this node with the specified token. Delegates to the
    * superclass and extracts data from <code>token</code> for later use in
    * line number and column queries.
    *
    * @param   token
    *          Token upon which this node is based.
    */
   public void initialize(Token token)
   {
      super.initialize(token);
      
      extractTokenData(token);
   }
   
   /**
    * Get the source code line number which designates where in the source
    * code file the token wrappered by this node was found. This will always
    * return <code>0</code> for an artificially manufactured AST (i.e., one
    * not generated by a parser).
    *
    * @return  Source file line number where the token underlying this AST
    *          was found, or <code>0</code> for an artifical AST.
    *
    * @see     #getColumn
    */
   public int getLine()
   {
      return line;
   }
   
   /**
    * Get the source code column number which designates where in the source
    * code file the token wrappered by this node was found. This will always
    * return <code>0</code> for an artificially manufactured AST (i.e., one
    * not generated by a parser).
    *
    * @return  Column number within a specific source code line, indicating
    *          where the token underlying this AST was found, or
    *          <code>0</code> for an artificial AST.
    *
    * @see     #getLine
    */
   public int getColumn()
   {
      return column;
   }
      
   /**
    * Extract token type, line number, and column information from a token
    * and store this information as member data.
    *
    * @param   token
    *          Token from which information is to be extracted.
    */
   private void extractTokenData(Token token)
   {
      this.line = token.getLine();
      this.column = token.getColumn();
   }
}