/*
** Module   : JanusGraphDB.java
** Abstract : Provides access to a JanusGraph graph DB.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140314 Created initial version.
** 002 GES 20170503 Upgrade to JanusGraph.
** 003 CA  20180327 Fixed mixed indexed support (for prefix searches).
** 004 CA  20180418 Some index related changes, to be able to re-index an existing graph.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.graphdb;

import java.util.concurrent.*;

import org.apache.commons.configuration.*;
import org.apache.tinkerpop.gremlin.structure.*;

import org.janusgraph.core.*;
import org.janusgraph.core.schema.*;
import org.janusgraph.graphdb.database.management.*;

/**
 * An adapter to obtain connections and create indexes in a JanusGraph graph DB
 */
class JanusGraphDB
extends GraphDB
{
   /**
    * Obtain a graph DB with the connection details specified in the given {@link Configuration}.
    * 
    * @param    configuration
    *           The connection details.
    *           
    * @return   A connected graph DB.
    */
   @Override
   protected Graph connectGraphDBImpl(Configuration configuration)
   {
      Graph graph = JanusGraphFactory.open(configuration);

      return graph;
   }

   /**
    * Create a single-property index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    keys
    *           The property key names to be indexed. This will also be the index name.
    * @param    datatypes
    *           The keys' types. 
    * @param    element
    *           {@code true} if the vertex element should be indexed, otherwise the edge element
    *           will be indexed.
    * @param    uniq
    *           {@code true} if the index should enforce uniqueness of the given key.
    * @param    mixed
    *           Flag indicating this index will be created in 'mixed' mode, using an external 
    *           index.
    */
   @Override
   protected void createIndexImpl(Graph      db,
                                  String[]   keys,
                                  Class<?>[] datatypes,
                                  boolean    element,
                                  boolean    uniq,
                                  boolean    mixed)
   {
      JanusGraph graph   = (JanusGraph) db;
      String     name    = "_";
      for (String key : keys)
      {
         name = name + "_" + key;
      }
      name = name + "_";

      boolean    created = true;
      graph.tx().rollback();
      
      JanusGraphManagement mgmt = graph.openManagement();
      if (!uniq)
      {
         StringBuilder sb = new StringBuilder();
         for (char ch : name.toCharArray())
         {
            if (Character.isDigit(ch) || Character.isLetter(ch))
            {
               sb.append(ch);
            }
         }
         name = sb.toString();
      }

      if (!mgmt.containsGraphIndex(name))
      {      
         Class<? extends Element> cls = element ? Vertex.class : Edge.class;
         JanusGraphManagement.IndexBuilder bld = mgmt.buildIndex(name, cls);

         for (int i = 0; i < keys.length; i++)
         {
            String key = keys[i];
            Class<?> datatype = datatypes[i];
            
            PropertyKey prop = null;
            
            if (!mgmt.containsPropertyKey(key))
            {
               prop = mgmt.makePropertyKey(key).dataType(datatype).make();
            }
            else
            {
               created = false;
               prop = mgmt.getPropertyKey(key);
            }
            
            if (mixed && String.class.equals(datatype))
            {
               bld.addKey(prop, Mapping.STRING.asParameter());
            }
            else
            {
               bld.addKey(prop);
            }
         }
         
         if (uniq || !mixed)
         {
            if (uniq)
            {
               bld.unique();
            }
            bld.buildCompositeIndex();
         }
         else 
         {
            bld.buildMixedIndex("search");
         }

         mgmt.commit();
         db.tx().commit();
         
         // if the property was not new, try to re-index
         if (!created)
         {
            reindex(graph, mgmt, name);
         }
      }
      else
      {
         mgmt.rollback();
      }
   }
   
      /**
    * Create a single-property index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed. This will also be the index name.
    * @param    datatype
    *           The key's type. 
    * @param    element
    *           {@code true} if the vertex element should be indexed, otherwise the edge element
    *           will be indexed.
    * @param    uniq
    *           {@code true} if the index should enforce uniqueness of the given key.
    * @param    mixed
    *           Flag indicating this index will be created in 'mixed' mode, using an external 
    *           index.
    */
   @Override
   protected void createIndexImpl(Graph    db,
                                  String   key,
                                  Class<?> datatype,
                                  boolean  element,
                                  boolean  uniq,
                                  boolean  mixed)
   {
      JanusGraph graph   = (JanusGraph) db;
      String     name    = "_" + key + "_";
      boolean    created = true;
      
      graph.tx().rollback();
      
      JanusGraphManagement mgmt = graph.openManagement();
      
      // index name is alphanumeric
      if (!uniq)
      {
         StringBuilder sb = new StringBuilder();
         for (char ch : name.toCharArray())
         {
            if (Character.isDigit(ch) || Character.isLetter(ch))
            {
               sb.append(ch);
            }
         }
         name = sb.toString();
      }
      
      if (!mgmt.containsGraphIndex(name))
      {      
         PropertyKey prop = null;
         
         if (!mgmt.containsPropertyKey(key))
         {
            prop = mgmt.makePropertyKey(key).dataType(datatype).make();
         }
         else
         {
            created = false;
            prop = mgmt.getPropertyKey(key);
         }
         
         Class<? extends Element> cls = element ? Vertex.class : Edge.class;
         JanusGraphManagement.IndexBuilder bld = mgmt.buildIndex(name, cls);
         
         if (mixed && String.class.equals(datatype))
         {
            bld.addKey(prop, Mapping.STRING.asParameter());
         }
         else
         {
            bld.addKey(prop);
         }
         
         if (uniq || !mixed)
         {
            if (uniq)
            {
               bld.unique();
            }
            bld.buildCompositeIndex();
         }
         else
         {
            bld.buildMixedIndex("search");
         }
         
         mgmt.commit();
         db.tx().commit();
         
         // if the property was not new, try to re-index
         if (!created)
         {
            reindex(graph, mgmt, name);
         }
      }
      else
      {
         mgmt.rollback();
      }
   }
   
   /**
    * Attempt to reindex the specified index (which was newly created).
    *
    * @param    graph
    *           The graph database instance.
    * @param    mgmt
    *           The current management object in which the index was created.
    * @param    idx
    *           The name of the index.
    *
    * @return   {@code true} if the reindexing was successful, {@code false} otherwise.
    */
   private boolean reindex(JanusGraph graph, JanusGraphManagement mgmt, String idx)
   {
      boolean result = false;
      
      if (mgmt instanceof ManagementSystem)
      {
         // wait for the index to become available
         try
         {
            ManagementSystem.awaitGraphIndexStatus(graph, idx)
                            .status(SchemaStatus.REGISTERED)
                            .call();

            // force it to reindex
            mgmt = graph.openManagement();
            mgmt.updateIndex(mgmt.getGraphIndex(idx), SchemaAction.REINDEX).get(1, TimeUnit.HOURS);
            mgmt.commit();
            
            ManagementSystem.awaitGraphIndexStatus(graph, idx)
                            .status(SchemaStatus.ENABLED)
                            .call();

            // we probably should check for success or failure, but there doesn't seem to be
            // a foolproof way to do that with the Future<ScanMetrics> that is returned
            result = true;
         }
         
         catch (InterruptedException  |
                CancellationException |
                ExecutionException    |
                TimeoutException e)
         {
            e.printStackTrace();
         }
      }
      
      return result;
   }
}