/*
** Module   : AccountExtension.java
** Abstract : Account data extension point.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20170612 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.client.widget.dialog.*;
import com.goldencode.p2j.admin.client.widget.dialog.InputDialog.*;

import java.io.*;
import java.util.*;
import java.util.function.*;

/**
 * An extension mechanism for the account editing UI. This interface is to be implemented
 * when there are extended editable account fields.
 */
public interface AccountExtension
{
   /** Standard account edit dialog fields */
   enum StandardFields
   {
      ACCT_NAME
   }

   /**
    * Called by the account edit logic to allow the extension to create its own fields to be
    * present in the account edit dialog.
    *
    * @param    edit
    *           If <code>true</code> the account edit dialog edits an existing account object,
    *           <code>false</code> a new account object is being edited.
    * @param    inputDialog
    *           Reference to the input dialog object.
    * @param    standardFields
    *           The map of standard fields present in the input dialog.
    * @param    defaultItems
    *           All input dialog items built by the core account editing logic.
    * @param    doneHandler
    *           The callback the extension must call with the new set of input dialog items.
    */
   void createFields(boolean edit,
                     InputDialog inputDialog,
                     Map<StandardFields, InputDialog.Field> standardFields,
                     InputDialog.Item[] defaultItems,
                     Consumer<List<InputDialog.Item>> doneHandler);

   /**
    * Called by the account editing logic to signal to the extension implementation that the
    * input dialog has been built. At this point the implementation has chance to hookup any
    * event handlers it may have or initialize the fields with backend data.
    *
    * @param    inputDialog
    *           Reference to the input dialog object.
    * @param    doneHandler
    *           Callback signaling success (<code>true</code>) or failure (<code>false</code>).
    *
    */
   void onDialogBuilt(InputDialog inputDialog, Consumer<Boolean> doneHandler);

   /**
    * Returns the object representing the extended account data being edited in the input dialog.
    *
    * @return   A {@link Serializable} reference.
    */
   Serializable getExtData();
   
   /**
    * Builds the Find User Account dialog and returns the dialog action handler.
    * 
    * @param    dialog
    *           The template dialog to create the Find User Account dialog
    * @param    findAccount
    *           The implementation of the find user account use case
    * 
    * @return   The dialog's action handler
    */
   Consumer<Map<Field, Object>> buildUserFindDialog(InputDialog dialog,
                                                    IFindAccountScreen findAccount);
}
