/*
** Module   : AsyncAction.java
** Abstract : AsyncAction is an abstract asynchronous call back action that can chain
**            the next dependent asynchronous call back or consumer.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home;

import java.util.function.Consumer;

import com.google.common.base.Function;
import com.google.gwt.user.client.rpc.AsyncCallback;


/**
 * AsyncAction is an abstract asynchronous call back action that can chain the next dependent
 * asynchronous call back or consumer.
 *
 * @param    <I>
 *           An input type
 * @param    <R>
 *           A result type
 */
public abstract class AsyncAction<I, R> implements AsyncCallback<R>, Consumer<I>
{
   /** The main asynchronous task */
   protected final AsyncCallback<R> first;
   
   /** The chained asynchronous task */
   protected Consumer<R> next;
   
   /** The exposed callback used by subclasses of this asynchronous task */
   protected final DefaultAsyncCallback delegate;
   
   /**
    * Setups the base functionality.
    */
   public AsyncAction()
   {
      delegate = new DefaultAsyncCallback();
      first = this;
   }
   
   /**
    * The default asynchronous callback.
    */
   private class DefaultAsyncCallback
   implements AsyncCallback<R>
   {

      @Override
      public void onFailure(Throwable caught)
      {
         first.onFailure(caught);
      }

      @Override
      public void onSuccess(R result)
      {
         first.onSuccess(result);
         if (next != null)
         {
            next.accept(result);
         }
      }
      
   }
   
   /**
    * Makes the given action to be executed after the main asynchronous action.
    * 
    * @param    action
    *           The chained asynchronous action
    * @param    mapper
    *           Maps the output of the main asynchronous action to the input of the next action
    */
   public <I2, R2> void then(AsyncAction<I2, R2> action, Function<R, I2> mapper)
   {
      next = new Consumer<R>()
      {
         @Override
         public void accept(R result)
         {
            action.accept(mapper.apply(result));
         }
      };
   }
   
   /**
    * Makes the given action to be executed after the main asynchronous action. The output of
    * the main asynchronous action is the input of the given chained asynchronous action.
    * 
    * @param    action
    *           The chained asynchronous action
    */
   public <R2> void then(AsyncAction<R, R2> action)
   {
      then(action, r -> r);
   }
   
   /**
    * Adds a consumer of the output of the main asynchronous action.
    * 
    * @param    consumer
    *           The consumer of the output of the main asynchronous action
    */
   public void then(Consumer<R> consumer)
   {
      next = consumer;
   }

   /**
    * The default on error callback method.
    * 
    * @param    caught
    *           The exception thrown by the main asynchronous action
    */
   @Override
   public void onFailure(Throwable caught)
   {
   }

   /**
    * The default on success callback method.
    * 
    * @param    result
    *           The output of the main asynchronous action
    */
   @Override
   public void onSuccess(R result)
   {
   }
}
