/*
** Module   : BaseViewWithUiHandlers.java
** Abstract : BaseViewWithUiHandlers represents a view that can share its controller with
**            its child dialogs.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home;

import java.util.LinkedList;
import java.util.logging.Logger;

import com.gwtplatform.mvp.client.UiHandlers;
import com.gwtplatform.mvp.client.ViewWithUiHandlers;
import com.gwtplatform.mvp.client.presenter.slots.NestedSlot;


/**
 * BaseViewWithUiHandlers represents a view that can share its controller with its child dialogs.
 *
 * @param    <C>
 *           An interface type that defines all shared business methods.
 */
public abstract class BaseViewWithUiHandlers<C extends UiHandlers> extends ViewWithUiHandlers<C>
{

   /** The logger */
   protected Logger logger = Logger.getLogger(getClass().getSimpleName());
   
   /** The modal content slot */
   private NestedSlot modalContentSlot;
   
   /**
    * Child dialogs must be injected by a concrete view from its constructor.
    */
   private final LinkedList<ViewWithUiHandlers<C>> childDialogs;
   
   /**
    * Creates this view.
    */
   public BaseViewWithUiHandlers()
   {
      this.childDialogs = new LinkedList<ViewWithUiHandlers<C>>();
   }

   /**
    * Method called after the view is attached to the DOM.
    */
   @Override
   protected void onAttach()
   {
      super.onAttach();
      for(ViewWithUiHandlers<C> dialog : childDialogs)
      {
         setupModaDialog(dialog);
      }
   }

   /**
    * Method called after the view is detached to the DOM.
    */
   @Override
   protected void onDetach()
   {
      super.onDetach();
      for(ViewWithUiHandlers<C> dialog : childDialogs)
      {
         removeModaDialog(dialog);
      }
   }
   /**
    * Setups the given modal slot to be used as the child dialog slot.
    * 
    * @param    modalContentSlot
    *           The given modal slot
    * @param    container
    *           The modal slot container
    */
   protected void setupModalSlot(NestedSlot modalContentSlot,  Object container)
   {
      this.modalContentSlot = modalContentSlot;
      bindSlot(modalContentSlot, container);
   }
   
   /**
    * Adds a child dialog to be managed by this view. The dialog must be added before its parent
    * view is attached to the DOM.
    * 
    * @param    dialog
    *           The child dialog view
    */
   protected void addDialog(ViewWithUiHandlers<C> dialog)
   {
      this.childDialogs.add(dialog);
   }

   /**
    * Setups the given child dialog.
    * 
    * @param    dialog
    *           The given child dialog
    */
   protected void setupModaDialog(ViewWithUiHandlers<C> dialog)
   {
      addToSlot(modalContentSlot, dialog);
   }
   
   /**
    * Removes the given child dialog from the managed child dialogs.
    * 
    * @param    dialog
    *           The given child dialog
    */
   protected void removeModaDialog(ViewWithUiHandlers<C> dialog)
   {
      removeFromSlot(modalContentSlot, dialog);
   }

   /**
    * Sets the view's controller and shares it between its child dialogs.
    * 
    * @param    uiHandlers
    *           The shared controller instance
    */
   @Override
   public void setUiHandlers(C uiHandlers)
   {
      super.setUiHandlers(uiHandlers);
      
      for(ViewWithUiHandlers<C> dialog : childDialogs)
      {
         dialog.setUiHandlers(uiHandlers);
      }
   }
}
