/*
** Module   : DatabaseReferenceWorker.java
** Abstract : Helper routines for unreferenced tables/fields processing.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20050724   @21992 Created initial version.
** 002 GES 20080318   @37638 Switched from writer to stream to avoid
**                           unwanted character conversion when the
**                           default character set is not 8-bit ASCII.
** 003 GES 20090515   @42205 Import changes.
** 004 IAS 20160331          Add input stream's close
**/

package com.goldencode.p2j.convert;

import java.io.*;
import java.util.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import com.goldencode.p2j.pattern.*;
import com.goldencode.util.*;

/**
 * Helper routines for the unreferenced tables/fields rule set.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
public final class DatabaseReferenceWorker
extends AbstractPatternWorker
{
   /**
    * Default constructor which initializes libraries.
    */
   public DatabaseReferenceWorker()
   {
      super();
      setLibrary(new Helper());
   }
   
   /**
    * Provides user-level APIs necessary for unreferenced tables/fields
    * processing.
    */
   public class Helper
   {
      /** XML tag for the root element */
      private static final String ELEM_TAG_ROOT  = "temporary-data";
      
      /** XML tag for the table element */
      private static final String ELEM_TAG_TABLE = "table";
      
      /** XML tag for the field element */
      private static final String ELEM_TAG_FIELD = "field";
      
      /** XML attribute for the table and field name */
      private static final String ATTR_NAME = "name";

      /**
       * Store map into the XML file.
       * <p>
       * Each map element represents a table referenced by Progress code
       * and holds another map which contains fields referenced.
       * 
       * @param   map
       *          Map which describes referenced tables and fields.
       * @param   file
       *          XML file name.
       * 
       * @return  <code>true</code> if operation was successful.
       * 
       * @throws  ParserConfigurationException
       *          forwarded from <code>XmlHelper</code> methods.
       */
      public boolean persistMap(Map map, String file)
      throws ParserConfigurationException
      {
         Document dom = XmlHelper.newDocument();

         //Add comments
         Comment comment;

         comment = dom.createComment("DatabaseReferenceWorker temporary file");
         dom.appendChild(comment);

         comment = dom.createComment("Saved " + (new Date()));
         dom.appendChild(comment);

         Element root = dom.createElement(ELEM_TAG_ROOT);
         dom.appendChild(root);

         Iterator iter = new TreeSet(map.keySet()).iterator();
         
         while(iter.hasNext())
         {
            Element elem     = dom.createElement(ELEM_TAG_TABLE);
            String tableName = (String) iter.next();
            Map table        = (Map) map.get(tableName);
            Iterator fields  = (new TreeSet(table.keySet())).iterator();
            
            elem.setAttribute(ATTR_NAME, tableName);
            
            root.appendChild(elem);
            
            while(fields.hasNext())
            {
               String fieldName = (String) fields.next();
               
               Element nodeAttr = dom.createElement(ELEM_TAG_FIELD);
               
               nodeAttr.setAttribute(ATTR_NAME, fieldName);
               
               elem.appendChild(nodeAttr);
            }
         }

         try
         {
            XmlHelper.write(dom, file);
         }
         catch (IOException e)
         {
            return false;
         }

         return true;
      }
      
      /**
       * Restore map which contains information about referenced tables and
       * fields from XML.
       * 
       * @param   file
       *          XML file name.
       * 
       * @return  an instance of <code>Map</code> if operation was successful
       *          or <code>null</code> if not.
       * 
       * @throws  SAXException
       *          forwarded from <code>XmlHelper</code> methods.
       * @throws  IOException
       *          forwarded from <code>XmlHelper</code> methods.
       * @throws  ParserConfigurationException
       *          forwarded from <code>XmlHelper</code> methods.
       * @throws  FileNotFoundException
       *          forwarded from <code>XmlHelper</code> methods.
       */
      public Map loadMap(String file) 
      throws FileNotFoundException, 
             ParserConfigurationException, 
             IOException, 
             SAXException
      {
         FileInputStream in = new FileInputStream(file);
         try 
         {
            Document dom = XmlHelper.parse(in);
            Node node    = dom.getDocumentElement().getFirstChild();
            HashMap res  = new HashMap();

            while (node != null)
            {
               if (node.getNodeName().equals(ELEM_TAG_TABLE))
               {
                  Element elem   = (Element) node;
                  String name    = elem.getAttribute(ATTR_NAME);
                  HashMap fields = new HashMap();
                  
                  NodeList l = node.getChildNodes();

                  for (int i = 0; i < l.getLength(); i++)
                  {
                     if (l.item(i).getNodeName().equals(ELEM_TAG_FIELD))
                     {
                        elem = (Element) l.item(i);

                        String fieldName = elem.getAttribute(ATTR_NAME);

                        if (fieldName == null)
                           return null;
                        
                        fields.put(fieldName, fieldName);
                     }
                  }
                  
                  res.put(name, fields);
               }
               
               node = node.getNextSibling();
            }
            
            return res;
         }
         finally
         {
            in.close();
         }
         
      }
   }
}