/*
** Module   : EmailSender.java
** Abstract : Interface for the SMTP EMAIL FWD extension to 4GL.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 GES 20171026 First commit.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.email;

import com.goldencode.p2j.extension.*;
import com.goldencode.p2j.util.*;

/**
 * Interface of the FWD SMTP EMAIL extension. An {@code EmailSender} object defines the state of
 * an email and allows sending it via SMTP. This is not compatible with the standard 4GL because
 * it is an extension that is specific to FWD.
 */
@FWDResource
@LegacyResource(resource = FWDResource.FWD_SMTP_EMAIL)
public interface EmailSender
{
   /**
    * Gets the connection type for the SMTP communications session.
    *
    * @return   The connection type which is "UNENCRYPTED", "STARTTLS_WHEN_AVAILABLE",
    *           "STARTTLS_REQUIRED" or "SSL".
    */
   @LegacyAttribute(name = "CONNECTION-TYPE")
   public character getConnectionType();
   
   /**
    * Sets the connection type for the SMTP communications session.
    *
    * @param    type
    *           The connection type which must be "UNENCRYPTED", "STARTTLS_WHEN_AVAILABLE",
    *           "STARTTLS_REQUIRED" or "SSL". Must not be <code>null</code> or unknown.
    */
   @LegacyAttribute(name = "CONNECTION-TYPE", setter = true)
   public void setConnectionType(character type);
   
   /**
    * Sets the connection type for the SMTP communications session.
    *
    * @param    type
    *           The connection type which must be "UNENCRYPTED", "STARTTLS_WHEN_AVAILABLE",
    *           "STARTTLS_REQUIRED" or "SSL". Must not be <code>null</code>.
    */
   @LegacyAttribute(name = "CONNECTION-TYPE", setter = true)
   public void setConnectionType(String type);
   
   /**
    * Gets the host for the SMTP server.
    *
    * @return   The name or IP address of the SMTP server.
    */
   @LegacyAttribute(name = "SMTP-HOST")
   public character getSmtpHost();
   
   /**
    * Sets the host for the SMTP server.
    *
    * @param    host
    *           The name or IP address of the SMTP server. Must not be <code>null</code>
    *           or unknown.
    */
   @LegacyAttribute(name = "SMTP-HOST", setter = true)
   public void setSmtpHost(character host);
   
   /**
    * Sets the host for the SMTP server.
    *
    * @param    host
    *           The name or IP address of the SMTP server. Must not be <code>null</code>.
    */
   @LegacyAttribute(name = "SMTP-HOST", setter = true)
   public void setSmtpHost(String host);
   
   /**
    * Gets the port for the SMTP server.
    *
    * @return   The port used by the SMTP server.
    */
   @LegacyAttribute(name = "SMTP-PORT")
   public character getSmtpPort();
   
   /**
    * Sets the port for the SMTP server.
    *
    * @param    port
    *           The port used by the SMTP server.
    */
   @LegacyAttribute(name = "SMTP-PORT", setter = true)
   public void setSmtpPort(character port);
   
   /**
    * Sets the port for the SMTP server.
    *
    * @param    port
    *           The port used by the SMTP server.
    */
   @LegacyAttribute(name = "SMTP-PORT", setter = true)
   public void setSmtpPort(String port);
   
   /**
    * Gets the userid for authenticating with the SMTP server.
    *
    * @return   The userid.
    */
   @LegacyAttribute(name = "SMTP-USER")
   public character getSmtpUser();
   
   /**
    * Sets the userid for authenticating with the SMTP server.
    *
    * @param    userid
    *           The userid for authentication.
    */
   @LegacyAttribute(name = "SMTP-USER", setter = true)
   public void setSmtpUser(character userid);
   
   /**
    * Sets the userid for authenticating with the SMTP server.
    *
    * @param    userid
    *           The userid for authentication.
    */
   @LegacyAttribute(name = "SMTP-USER", setter = true)
   public void setSmtpUser(String userid);
   
   /**
    * Gets the password for authenticating with the SMTP server.
    *
    * @return   The password.
    */
   @LegacyAttribute(name = "SMTP-PASSWORD")
   public character getSmtpPassword();
   
   /**
    * Sets the password for authenticating with the SMTP server.
    *
    * @param    password
    *           The password for authentication.
    */
   @LegacyAttribute(name = "SMTP-PASSWORD", setter = true)
   public void setSmtpPassword(character password);
   
   /**
    * Sets the password for authenticating with the SMTP server.
    *
    * @param    password
    *           The password for authentication.
    */
   @LegacyAttribute(name = "SMTP-PASSWORD", setter = true)
   public void setSmtpPassword(String password);
   
   /**
    * Returns the state of the SMTP-VALIDATE attribute, which determines if the SMTP server
    * certificate is validated during an SSL or STARTTLS connection.
    *
    * @return   Current state of the SMTP-VALIDATE attribute.
    */
   @LegacyAttribute(name = "SMTP-VALIDATE")
   public logical isSmtpValidate();
   
   /**
    * Setter for SMTP-VALIDATE attribute, which determines if the SMTP server certificate is
    * validated during an SSL or STARTTLS connection.
    *
    * @param    validate
    *           <code>true</code> to validate the server certificate.
    */
   @LegacyAttribute(name = "SMTP-VALIDATE", setter = true)
   public void setSmtpValidate(boolean validate);
   
   /**
    * Setter for SMTP-VALIDATE attribute, which determines if the SMTP server certificate is
    * validated during an SSL or STARTTLS connection.
    *
    * @param    validate
    *           <code>true</code> to validate the server certificate.
    */
   @LegacyAttribute(name = "SMTP-VALIDATE", setter = true)
   public void setSmtpValidate(logical validate);
   
   /**
    * Gets the FROM address.
    *
    * @return   The FROM address.
    */
   @LegacyAttribute(name = "SMTP-FROM")
   public character getSmtpFrom();
   
   /**
    * Sets the FROM address.
    *
    * @param    from
    *           The FROM address.
    */
   @LegacyAttribute(name = "SMTP-FROM", setter = true)
   public void setSmtpFrom(character from);
   
   /**
    * Sets the FROM address.
    *
    * @param    from
    *           The FROM address.
    */
   @LegacyAttribute(name = "SMTP-FROM", setter = true)
   public void setSmtpFrom(String from);
   
   /**
    * Gets the REPLY-TO address.
    *
    * @return   The REPLY-TO address.
    */
   @LegacyAttribute(name = "SMTP-REPLY-TO")
   public character getSmtpReplyTo();
   
   /**
    * Sets the REPLY-TO address.
    *
    * @param    replyTo
    *           The REPLY-TO address.
    */
   @LegacyAttribute(name = "SMTP-REPLY-TO", setter = true)
   public void setSmtpReplyTo(character replyTo);
   
   /**
    * Sets the REPLY-TO address.
    *
    * @param    replyTo
    *           The REPLY-TO address.
    */
   @LegacyAttribute(name = "SMTP-REPLY-TO", setter = true)
   public void setSmtpReplyTo(String replyTo);
   
   /**
    * Gets the SUBJECT line.
    *
    * @return   The SUBJECT line.
    */
   @LegacyAttribute(name = "SMTP-SUBJECT")
   public character getSmtpSubject();
   
   /**
    * Sets the SUBJECT line.
    *
    * @param    subject
    *           The SUBJECT line.
    */
   @LegacyAttribute(name = "SMTP-SUBJECT", setter = true)
   public void setSmtpSubject(character subject);
   
   /**
    * Sets the SUBJECT line.
    *
    * @param    subject
    *           The SUBJECT line.
    */
   @LegacyAttribute(name = "SMTP-SUBJECT", setter = true)
   public void setSmtpSubject(String subject);
   
   /**
    * Gets the HTML body contents.
    *
    * @return   The HTML body contents.
    */
   @LegacyAttribute(name = "SMTP-HTML")
   public character getSmtpHtmlBody();
   
   /**
    * Sets the HTML body contents.
    *
    * @param    html
    *           The HTML body contents.
    */
   @LegacyAttribute(name = "SMTP-HTML", setter = true)
   public void setSmtpHtmlBody(character html);
   
   /**
    * Sets the HTML body contents.
    *
    * @param    html
    *           The HTML body contents.
    */
   @LegacyAttribute(name = "SMTP-HTML", setter = true)
   public void setSmtpHtmlBody(String html);
   
   /**
    * Gets the TEXT body contents.
    *
    * @return   The TEXT body contents.
    */
   @LegacyAttribute(name = "SMTP-TEXT")
   public character getSmtpTextBody();
   
   /**
    * Sets the TEXT body contents.
    *
    * @param    text
    *           The TEXT body contents.
    */
   @LegacyAttribute(name = "SMTP-TEXT", setter = true)
   public void setSmtpTextBody(character text);
   
   /**
    * Sets the TEXT body contents.
    *
    * @param    text
    *           The TEXT body contents.
    */
   @LegacyAttribute(name = "SMTP-TEXT", setter = true)
   public void setSmtpTextBody(String text);
   
   /**
    * Add a TO address to the list of recipients.
    *
    * @param    addr
    *           The TO address.
    */
   @LegacyMethod(name = "ADD-TO-ADDRESS")
   public logical addToAddress(character addr);
   
   /**
    * Add a TO address to the list of recipients.
    *
    * @param    addr
    *           The TO address.
    */
   @LegacyMethod(name = "ADD-TO-ADDRESS")
   public logical addToAddress(String addr);
   
   /**
    * Add a CC address to the list of recipients.
    *
    * @param    addr
    *           The CC address.
    */
   @LegacyMethod(name = "ADD-CC-ADDRESS")
   public logical addCcAddress(character addr);
   
   /**
    * Add a CC address to the list of recipients.
    *
    * @param    addr
    *           The CC address.
    */
   @LegacyMethod(name = "ADD-CC-ADDRESS")
   public logical addCcAddress(String addr);
   
   /**
    * Add a BCC address to the list of recipients.
    *
    * @param    addr
    *           The BCC address.
    */
   @LegacyMethod(name = "ADD-BCC-ADDRESS")
   public logical addBccAddress(character addr);
   
   /**
    * Add a BCC address to the list of recipients.
    *
    * @param    addr
    *           The BCC address.
    */
   @LegacyMethod(name = "ADD-BCC-ADDRESS")
   public logical addBccAddress(String addr);
   
   /**
    * Attach the specified file to the email.
    *
    * @param    file
    *           The file to attach.
    */
   @LegacyMethod(name = "ATTACH-FILE")
   public logical attachFile(character file);
   
   /**
    * Attach the specified file to the email.
    *
    * @param    file
    *           The file to attach.
    */
   @LegacyMethod(name = "ATTACH-FILE")
   public logical attachFile(String file);
   
   /**
    * Attach the file specified by the given URL to the email.
    *
    * @param    url
    *           The resource to attach.
    */
   @LegacyMethod(name = "ATTACH-URL")
   public logical attachURL(character url);
   
   /**
    * Attach the file specified by the given URL to the email.
    *
    * @param    url
    *           The resource to attach.
    */
   @LegacyMethod(name = "ATTACH-URL")
   public logical attachURL(String url);
   
   /**
    * Clear the list of attachments.
    */
   @LegacyMethod(name = "CLEAR-ATTACHMENTS-LIST")
   public logical clearAttachmentsList();
   
   /**
    * Clear the list of BCC recipients.
    */
   @LegacyMethod(name = "CLEAR-BCC-LIST")
   public logical clearBccAddressList();
   
   /**
    * Clear the list of CC recipients.
    */
   @LegacyMethod(name = "CLEAR-CC-LIST")
   public logical clearCcAddressList();
   
   /**
    * Clear the list of embedded images.
    */
   @LegacyMethod(name = "CLEAR-EMBEDDED-IMAGE-LIST")
   public logical clearEmbeddedImageList();
   
   /**
    * Clear the list of TO recipients.
    */
   @LegacyMethod(name = "CLEAR-TO-LIST")
   public logical clearToAddressList();
   
   /**
    * Embed the image specified by the given file and return the content ID needed to substitute
    * into the HTML (in an IMG tag). The result will be visible inline in the HTML document as
    * part of the IMG.
    *
    * @param    file
    *           The resource to attach.
    *
    * @return   The content ID for use in an IMG tag.
    */
   @LegacyMethod(name = "EMBED-FILE")
   public character embedImageByFile(character file);
   
   /**
    * Embed the image specified by the given file and return the content ID needed to substitute
    * into the HTML (in an IMG tag). The result will be visible inline in the HTML document as
    * part of the IMG.
    *
    * @param    file
    *           The resource to attach.
    *
    * @return   The content ID for use in an IMG tag.
    */
   @LegacyMethod(name = "EMBED-FILE")
   public character embedImageByFile(String file);
   
   /**
    * Embed the image specified by the given URL and return the content ID needed to substitute
    * into the HTML (in an IMG tag). The result will be visible inline in the HTML document as
    * part of the IMG.
    *
    * @param    url
    *           The resource to attach.
    *
    * @return   The content ID for use in an IMG tag.
    */
   @LegacyMethod(name = "EMBED-URL")
   public character embedImageByURL(character url);
   
   /**
    * Embed the image specified by the given URL and return the content ID needed to substitute
    * into the HTML (in an IMG tag). The result will be visible inline in the HTML document as
    * part of the IMG.
    *
    * @param    url
    *           The resource to attach.
    *
    * @return   The content ID for use in an IMG tag.
    */
   @LegacyMethod(name = "EMBED-URL")
   public character embedImageByURL(String url);
   
   /**
    * Obtain the list of attachments.
    *
    * @return   The array of attachment file and URL names.
    */
   @LegacyMethod(name = "GET-ATTACHMENTS-LIST")
   public character[] getAttachmentsList();
   
   /**
    * Obtain the list of BCC recipients.
    *
    * @return   The array of all BCC recipients.
    */
   @LegacyMethod(name = "GET-BCC-LIST")
   public character[] getBccAddressList();
   
   /**
    * Obtain the list of CC recipients.
    *
    * @return   The array of all CC recipients.
    */
   @LegacyMethod(name = "GET-CC-LIST")
   public character[] getCcAddressList();
   
   /**
    * Obtain the list of embedded image file and URL names.
    *
    * @return   The array of attachment file and URL names.
    */
   @LegacyMethod(name = "GET-EMBEDDED-IMAGE-LIST")
   public character[] getEmbeddedImageList();
   
   /**
    * Obtain the list of TO recipients.
    *
    * @return   The array of all TO recipients.
    */
   @LegacyMethod(name = "GET-TO-LIST")
   public character[] getToAddressList();  
   
   /**
    * Send the email. 
    *
    * @return   <code>true</code> if the send succeeded.
    *
    * @throws   ErrorConditionException
    *           On any failure, unless silent error mode is enabled.
    */
   @LegacyMethod(name = "Send")
   public logical send()
   throws ErrorConditionException;
}
