/*
** Module   : MultiStream.java
** Abstract : Implements a simple switchable input stream.
**            switchTo method is provided to switch the input to any
**            existing stream.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20041112 ADD   @18753 Created.  Needed a class that would multiplex
**                               input streams into a single input stream.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.io;

import java.io.*;

/**
 * Implements a simple switchable input stream based on any number of
 * underlying input streams. Input streams can be
 * switched to at any time. There is no history kept inside the class.
 * That means, there is no switch back capability.
 * <p>
 * When instantiated, the MultiStream object has no connection to any
 * input stream. If <code>read()</code> method is called while in this
 * state, <code>-1</code> is returned signalling EOF condition.
 * <p>
 * Once the <code>switchTo</code> method is called, <code>read()</code>
 * delegates the work to the <code>read()</code> method of the current
 * underlying input stream.
 * <p>
 * This class may be useful in many situations. For instance,
 * PushbackInputStream can provide pushback functionality for multiple
 * underlying streams at once in an efficient way.
 * <p>
 * @author NVS
 * @version 1.0.1
 */
public class MultiStream extends FilterInputStream
{
   /**
    * Constructor.
    * Creates a MultiStream with no associated stream.
    */
   public MultiStream()
   {
      super(null);
   }

   /**
    * Switches input to the specified stream.
    *
    * @param   in
    *          an input stream of any kind to be switched to.
    */
   public void switchTo(InputStream in)
   {
      super.in = in;
   }

   /**
    * Reads the next byte from the current stream. If no stream was
    * switched to, returns -1 as the indication of EOF condition.
    *
    * @return    byte read from the underlying stream or -1 if no
    *            stream was switched to.
    * @throws    IOException
    */
   public int read()
   throws IOException
   {
      if (super.in == null)
         return -1;
      else
         return super.read();
   }
}
