/*
** Module   : NativeInteger.java
** Abstract : descriptor for integer data as an argument or return value to a native function 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131212 First version.
** 002 GES 20140212 Added a flag to enable different rendering between scalar and array elements.
**                  Added a method to properly emit sign-extended values with word-size width.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A descriptor for integer data as an argument or return value to a native function.
 */
public abstract class NativeInteger
extends NativePrimitive
{
   /** The stored value. */
   protected long value;
   
   /** Flag to allow differential rendering between scalar and array elements. */
   private boolean array = false; 
   
   /**
    * Default constructor.
    */
   public NativeInteger()
   {
      super();
   }
   
   /**
    * Access the data element.
    *
    * @return   The integer data element.
    */
   public long getValue()
   {
      return value;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The new integer data element.
    */
   public void setValue(long value)
   {
      this.value = value;
   }
   
   /**
    * Reports if this data element is an array member.
    *
    * @return   <code>true</code> if this is an element of an array.
    */
   public boolean isInArray()
   {
      return array;
   }
   
   /**
    * Defines whether or not the data element is an array member.
    *
    * @param    array
    *           <code>true</code> if this is an element of an array.
    */
   public void setInArray(boolean array)
   {
      this.array = array;
   }
   
   /**
    * This method should be used to render signed scalar elements such that the written data is
    * sign-extended to the word-size.  This is how Progress 4GL handles integer data. The reason
    * this is needed is because scalar values are passed on the stack in a "slot" that is the
    * width of the platform's word-size. Even if the native code is expecting something smaller,
    * the actual memory passed will have a larger, sign-extended form of the data.
    * <p>
    * This doesn't apply to non-integer data.  It also doesn't apply to arrays, since there is
    * no word-size "padding" in array elements. Each native array's element is exactly the size
    * of the element itself and so there is no sign-extending to do.
    *
    * @param    ptr
    *           The buffer in which to write at position 1.
    */
   public void writeNativeWordSize(memptr ptr)
   {
      if (LibraryManager.ptrSize() == 4)
      {
         ptr.setLong(value, 1);
      }
      else
      {
         ptr.setInt64(value, 1);
      }
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      // don't assign the member directly, subclasses may have some behavior hooked to this
      setValue(in.readLong());
      
      array = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeLong(value);
      out.writeBoolean(array);
   }
}