/*
** Module   : NativeString.java
** Abstract : descriptor for text data as an argument or return value to a native function 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20140107 First version.
** 002 GES 20140201 Complete rewrite of the render() method to override the parent's version. It
**                  is important to write strings into their own memory buffer and then write the
**                  pointer into the given buffer. On restore(), the parent's version is fine.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.library;

import java.io.*;
import java.math.*;

import com.goldencode.p2j.util.*;

/**
 * A descriptor for text data as an argument or return value to a native function.
 */
public class NativeString
extends NativePrimitive
{
   /** Amount of extra bytes allocated to reduce the change of buffer overflows. */
   private static int BUFFER_PADDING = 128;
   
   /** The stored value (defaults to empty string). */
   protected String value = ""; 
   
   /**
    * Default constructor.
    */
   public NativeString()
   {
      super();
      type = LibraryManager.TYPE_POINTER;
   }
   
   /**
    * Access the data element.
    *
    * @return   The doubleing point data element.
    */
   public String getValue()
   {
      return value;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The new text data element.  Ignored if <code>null</code>.
    */
   public void setValue(String value)
   {
      if (value != null)
         this.value = value;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      value = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeObject(value);
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   This is always the size of a pointer on the client platform.
    */
   public int size()
   {
      return LibraryManager.ptrSize();
   }
   
   /**
    * Copy our data into a large-enough buffer in native memory and then render the buffer's
    * address into native memory if an explicit address is provided and return the data's
    * address to the caller.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address into which the address should be rendered, if the caller
    *           is supplying a buffer or 0 if the internally created buffer's address should be
    *           returned directly.
    *
    * @return   The memory address of the buffer or 0 if there is no data to output.
    */
   public long render(long addr)
   {
      // make sure there is room for the null terminator and the value itself
      int sz = value.length() + 1 + BUFFER_PADDING;
      
      // save off the buffer for restore later
      ref = new memptr();
      ref.setLength(sz);
      
      // since we are always passed as a pointer, we always copy our data into the
      // native buffer; this will always result in at least the empty string (a null
      // terminator in the first position of the buffer)
      writeNative(ref);
      
      // - return parms (always scalar) do not call through here
      // - non-return scalar parm types will have addr == 0 and we just need to return the
      //   target address of our newly created buffer memptr
      // - array elements will be called with addr != 0 because the address needs to be
      //   written to memory controlled by the caller
      
      // return our buffer's address
      long target = ref.getPointerValue().longValue();
      
      // write it into the caller's buffer if specified
      if (addr != 0)
      {
         memptr local = new memptr();
         local.setPointerValue(addr);
         
         if (size() == 4)
         {
            local.setLong(target, 1);
         }
         else
         {
            local.setInt64(target, 1);
         }
      }
      
      return target;
   }
   
   /**
    * Worker method to read the primitive value from the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void readNative(memptr ptr)
   {
      character txt = ptr.getString(1);
      
      if (!txt.isUnknown())
      {
         value = txt.toStringMessage();
      }
   }
   
   /**
    * Worker method to write the primitive value to the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void writeNative(memptr ptr)
   {
      ptr.setString(value, 1);
   }
}