/*
** Module   : ClientBuilder.java
** Abstract : build and start a platform process.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131118 First version.
** 002 MAG 20140206 Send the referrer address via command line.
** 003 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
** 004 CA  20140224 Redirect STDERR of spawned process to the server STDERR.
** 005 MAG 20140310 Set password.
** 006 MAG 20140710 Implements remote launcher.
** 007 GES 20150211 Removed application arguments support.
** 008 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 GES 20160219 Moved code here from the process builder to share the native secure
**                  connection setup processing.
** 010 OM  20160905 Small optimizations.
** 011 SBI 20171019 Changed to spawn remote clients on dedicated hosts.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.util.*;

/**
 * This class is used to create and spawn a client process using a
 * {@link ProcessBuilder} instance. A special small tools are designed for
 * Linux / Windows OS in order to spawn the process on the OS user account.
 * Both tools are written in C using specific OS API and are build as executable
 * using ant native target. The source code and the make file for this tools are 
 * available in the /src/native process directory. 
 * The password is provided vis stdin redirection on both platforms. 
 * The command line parameters for the spawned process are constructed from default 
 * parameters stored in directory and some specific parameters stored inside a 
 * <code>ClientBuilderParameters</code> structure.   
 */
public abstract class ClientBuilder
{   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ClientBuilder.class.getName());

   /** Pending spawn commands to be retrieved by remote clients. */
   private static final Map<String, String[]> pendingCmds = new HashMap<>();
   
   /** A collection of command line tokens */
   protected List<String> command = new LinkedList<String>();
      
   /** A Java Bean containing specific parameters */
   protected final ClientBuilderParameters  config;
   
   /** Indicate a remote or local spawn */
   protected boolean remote = false;
   
   /** The ProcessBuilder instance */
   private ProcessBuilder pb = new ProcessBuilder();
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The specific parameters for this client.
    */
   public ClientBuilder(ClientBuilderParameters config)
   {
      this.config = config;
   }

   /**
    * Get the spawn command registered for this UUID.
    * 
    * @param    uuid
    *           The UUID.
    *           
    * @return   See above.
    */
   public static String[] getCommand(String uuid)
   {
      synchronized (pendingCmds)
      {
         return pendingCmds.get(uuid);
      }
   }
   
   /**
    * Remove the spawn command registered for this UUID.
    * 
    * @param    uuid
    *           The UUID.
    */
   public static void removeCommand(String uuid)
   {
      synchronized (pendingCmds)
      {
         pendingCmds.remove(uuid);
      }
   }
   
   /**
    * Add custom client options to the {@link #command}.
    */
   protected abstract void addClientOptions();
   
   /**
    * Get the spawn arguments for a specific implementation.
    * 
    * @return   See above.
    */
   protected abstract List<String> getSpawnArguments();

   /**
    * Spawn's new local Java process.
    * 
    * @param   env
    *          A map with environment variables for temporary account.
    *           
    * @return  Process object.
    * 
    * @throws  IOException
    *          If something goes wrong.
    */
   public Process localStart(Map<String, String> env) throws IOException
   {
      // set local
      remote = false;
      
      List<String> spawnCmd = prepareToLaunch();

      pb.environment().putAll(env);
      // we start in the spawner's dir
      File spawnerDir = new File(config.getSpawner());
      pb.directory(spawnerDir.getAbsoluteFile().getParentFile());
      pb.command(spawnCmd);
      pb.redirectError(ProcessBuilder.Redirect.INHERIT);
      
      Process shell = pb.start();

      if (config.isPasswordAuthentication())
      {
         // write password
         OutputStreamWriter os = new OutputStreamWriter(shell.getOutputStream());
         os.write(config.getPassword());
         os.close();
      }
      
      return shell;
   }

   /**
    * Spawn's new remote Java process.
    * 
    * @param    dedicatedUser
    *           The dedicated user associated with the host on which the spawned client
    *           will be run
    * @param    brokers
    *           List of brokers.
    * @param    env
    *           A map with environment variables for temporary account.
    * 
    * @return   Process object.
    * 
    * @throws   IOException
    *           If something goes wrong.
    */
   public BrokerSpawnResult remoteStart(String dedicatedUser,
                                        Brokers brokers,
                                        Map<String, String> env) 
   throws IOException
   {
      // set remote
      remote = true;
      
      // get spawn cmd
      List<String> spawnCmd = prepareToLaunch();
            
      BrokerSpawnParameters args = new BrokerSpawnParameters();
      args.setEnvironment(env);
      args.setCommand(spawnCmd);
      args.setPasswordAuthentication(config.isPasswordAuthentication());
      
      if (args.isPasswordAuthentication())
      {
         args.setPassword(config.getPassword());
      }
   
      return BrokerManager.spawn(dedicatedUser, brokers, args);
   }
   
   /**
    * Initialize the command line parameters for a native secure connection, which is the
    * secure authentication method used for non-interactive clients and those interactive clients
    * which cannot be launched using a password.  These client processes call the {@code spawn}
    * tool using the {@code spawn 0 <secure-port> <server-hostname> <server-alias> <uuid>} 
    * syntax.
    *
    * @param    cmd
    *           The data structure to initialize with the parametes for the native secure
    *           connection.
    * @param    securePort
    *           The server's secure port or -1 if this is a remote launch (the port will be
    *           dynamically filled in later in that case).
    * @param    alias
    *           The alias of the server certificate in the certificate key store which will
    *           be used to authenticate that this server indeed is on the other side of the
    *           secure connection.
    */
   protected void initNativeSecureConnection(List<String> cmd, String securePort, String alias)
   {
      // add the mode that makes this a native secure connection
      cmd.add("0");
      
      // add host and secure port
      if (remote)
      {
         // for remote spawn use placeholders
         cmd.add(BrokerCore.PARAM_PORT);
         cmd.add(BrokerCore.PARAM_HOST);
      }
      else
      {
         // add secure port
         cmd.add(securePort);
         
         // add host; this is hard-coded to localhost because is expected that the in-process JVM
         // will connect to the same P2J server which initiated the request (as this is the P2J
         // server knowing the command to be executed).
         cmd.add("localhost");
      }
      // add server alias
      cmd.add(alias);
      
      // add UUID
      cmd.add(config.getUuid());
   }

   /**
    * Prepare command line arguments for spawned process before to start.
    * 
    * @return  A list with command line arguments.
    * 
    * @throws  IOException
    *          On error.
    */
   private List<String> prepareToLaunch() 
   throws IOException
   {
      SessionManager sessMgr = SessionManager.get();
      BootstrapConfig cfg    = sessMgr.config();
      int securePort         = cfg.getInt("net", "server", "secure_port", -1);
      if (securePort == -1)
      {
         throw new RuntimeException(
                          "Clients can be spawned only if the server is running in secure mode!");
      }
      
      // build the final command (which will be directly started or delayed until after P2J
      // authentication
      {
         buildCommand();
      }
       
      // launch the spawner
      List<String> spawnCmd = new ArrayList<>();
      if (remote)
      {
         // spawner placehoder
         spawnCmd.add(BrokerCore.PARAM_SPAWNER);
      }
      else
      {
         spawnCmd.add(config.getSpawner());
      }
      
      spawnCmd.addAll(getSpawnArguments());
      
      if (LOG.isLoggable(Level.FINE))
      {
         StringBuilder sb = new StringBuilder("Launching a new P2J client using the command: ");
         for (String s : spawnCmd)
         {
            sb.append(s).append(" ");
         }
         
         LOG.logp(Level.FINE, "ClientBuilder", "start", sb.toString().trim());
      }
      
      return spawnCmd;
   }
   
   /**
    * Add server properties.
    * 
    * @param   secure
    *          Indicating to setup a secure <code>true</code> 
    *          or an insecure <code>false</code> connection.
    */
   private void addServerOptions(boolean secure)
   {
      // uuid assigned to this client
      command.add("server:spawner:uuid=" + config.getUuid());
      
      // add target server ports
      BootstrapConfig serverConfig = SessionManager.get().config();

      String host = serverConfig.getString("net", "server", "host", "localhost");
      command.add("net:server:host=" + (remote ? BrokerCore.PARAM_HOST : host));

      if (secure)
      {
         command.add("net:connection:secure=true");
         if (remote)
         {
            command.add("net:server:secure_port=" + BrokerCore.PARAM_PORT);
         }
         else
         {
            getServerConfig(serverConfig, "net", "server", "secure_port", -1);
         }
      }
      else if (getServerConfig(serverConfig, "net", "server", "port", -1) == -1)
      {
         getServerConfig(serverConfig, "net", "server", "insecure_port", -1);
      }
   }

   /**
    * Get a key value form configuration file. 
    * If key exists add the value to builder command.
    * 
    * @param   config
    *          Configuration file.
    * @param   cname
    *          Category name.
    * @param   gname
    *          Group name.
    * @param   kname
    *          Key name.
    * @param   def
    *          Key value if exists default value otherwise.
    * 
    * @return  The integer value for key.
    */
   private int getServerConfig(BootstrapConfig config,
                               String cname, 
                               String gname, 
                               String kname,
                               int def)
   {
      int value = config.getInt(cname, gname, kname, def);
      if (value != def)
      {
         command.add(cname + ":" + gname + ":" + kname + "=" + value);   
      }
      return value;
   }
   
   /**
    * Add Java CLASSPATH and the main class. For remote clients the CLASSPATH and JVM arguments
    * are inherited from broker process.
    */
   private void addClasspath()
   {
      if (remote)
      {
         command.add(BrokerCore.PARAM_CP);
      }
      else
      {
         // add CLASSPATH
         command.add("-classpath");
         command.add(config.getClassPath());
      }
      
      // add main class
      Class<?> mainClass = config.getMainClass();
      command.add(mainClass.getName());
   }

   /**
    * Add client configuration file if exists.
    * 
    * @param   userDir
    *          Client working directory.
    */
   private void addConfigFile(String userDir)
   {
      String configFile = config.getConfigFile();
      if (configFile != null)
      {
         String clientConfig = userDir + File.separator + configFile;
         if (new File(clientConfig).exists())
         {
            command.add(configFile);
         } 
         else
         {
            LOG.logp(Level.WARNING, 
                     "ClientBuilder.addConfigFile()", 
                     "", 
                     LogHelper.generate("%s does not exist.", clientConfig));
         }
      }
   }

   /**
    * Sets this process builder's working directory.
    * 
    * @param   userDir
    *          Server working directory.
    */
   @SuppressWarnings("unused")
   private File setUserDir(String userDir)
   {
      File fUserDir = new File(userDir);
      if (!fUserDir.exists())
      {
         throw new IllegalStateException(
               "The user dir '" + userDir + "' specified does not exist.");
      }
      return fUserDir;
   }
   
   /**
    * Build command for spawned process.
    * 
    * @throws  IOException
    *          If something goes wrong.
    */
   private void buildCommand() 
   throws IOException
   {
      // build the final command (which will be directly started or delayed until after P2J
      // authentication
      String userDir = config.getWorkingDirectory();

      // user name
      command.add(config.getUserName());
      
      // password
      if (!config.isPasswordAuthentication())
      {
         String password = config.getPassword();
         command.add(password == null ? "\"\"" : password);
      }
      
      // working directory
      command.add(userDir);
      
      // Java launcher executable
      command.add(config.getCommand());
      
      // JVM arguments
      if (remote)
      {
         // placeholder, will be replaced by broker with CLASSPATH and JVM arguments
         command.add(BrokerCore.PARAM_JVM);
      }
      else
      {
         // add JVM arguments
         command.addAll(config.getJvmArguments());
         
         // native libraries path
         String libPath = config.getLibPath();
         command.add("-Djava.library.path=" + libPath);
      }
      
      // add CLASSPATH and main class
      addClasspath();
      
      // add client configuration file
      addConfigFile(userDir);
      
      // add server options
      addServerOptions(config.isSecure());
      
      // add other client options
      addClientOptions();
      
      // add custom bootstrap configuration
      command.addAll(config.getBootstrapConfig());

      // ensure there are no empty or null arguments. an empty argument will make the 
      // ClientDriver to assume that is a bootstrap config file name and fail on startup.
      Iterator<String> iter = command.iterator();
      while (iter.hasNext())
      {
         String s = iter.next();
         if (s == null || s.trim().length() == 0)
         {
            iter.remove();
         }
      }
      
      if (!config.isPasswordAuthentication())
      {
         synchronized (pendingCmds)
         {
            // save this for later use
            pendingCmds.put(config.getUuid(), command.toArray(new String[0]));
         }
      }
   }
}