/*
** Module   : ClientCore.java
** Abstract : core client loop which handles init and connects to server
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 GES 20110928 Moved core logic here from ClientDriver so that the same
**                  code could be used from multiple callers.
** 002 GES 20111028 Added support for startup of a client session within the
**                  same process as the serve.
** 003 GES 20111111 Pass single mode flag to thin client.
** 004 CA  20130529 If this is authenticated as an appserver, register the needed server exports.
** 005 CA  20130704 Removed H004, as the appserver launching was reworked.
** 006 CA  20131004 Expose client-side parameters to the server-side application code. 
** 007 EVL 20131104 Batch mode support.
** 008 MAG 20131129 Import server KeyStore.
** 009 EVL 20140117 Moved here the code to load native P2J library. 
** 010 MAG 20140128 Fix CTRL-C process for web clients. 
** 011 MAG 20140131 Temporary accounts should be used only once to import server key store.
** 012 MAG 20140206 On remote application exit notify web client by sending a referrer address
**                  used to redirect the client. 
** 013 CA  20140206 Refactored the temporary client connection to make it unaware of the type of 
**                  P2J client being started.
** 014 MAG 20140217 Refactoring web client redirection flaw.
** 015 EVL 20140327 Changing the call to initErrorWriter() in OutputManager to take into account
**                  the server batch mode flag setting.
** 016 CA  20140805 Changes to isolate ChUI-specific code from common code: added generics, 
**                  classes not related to ChUI behavior import individual ChUI package classes.
** 017 OM  20140909 Emulate the '-ininame' command-line parameter.
** 018 CA  20141123 Renamed TC.killDependentThreads to TC.terminate.
** 019 GES 20141228 Rework to eliminate application arguments and to extend the client parameters
**                  to more use cases.
** 020 GES 20150109 Added native methods to query pid and user name from the OS. Extended the
**                  client parameters with these values and also support for a user-specific
**                  "terminal name" that will be used as the equivalent of the UNIX device name.
**                  Also added a chui/gui flag to the parms so that the server doesn't need to
**                  call down to the client to figure that out.
** 021 CA  20150131 Clean the security caches when client restarts.
** 022 HC  20151124 Improved loading of boolean environment variables from Progress INI files and
**                  Windows Registry.
** 023 IAS 20160328 Client log support.
** 024 OM  20160927 Added batchMode to ClientParameters serialization.
** 025 CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
** 025 SBI 20170628 Changed to shutdown the driver and then the current session, added to register
**                  the web client session on the server and to release the web client data on
**                  the server if the client is about to be closed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * Core looping logic that performs initialization of a client UI and
 * establishes a session with the server.
 * <p>
 * This code provides supports the startup of a client session within the
 * same process as the server.  This can be used to duplicate the traditional
 * Progress 4GL fat client approach (a "single process" mode). It can also be
 * used to implement multiple client sessions within a true server process
 * which would be used to support web-based clients.
 */
public class ClientCore
{
   /** Flag to prevent multiple native library loading.*/
   private static boolean nativeLayerInitialized = false;
   
   /**
    * Start the client loop.  This will not return until the user is exiting
    * the client. The core loop initializes logging, the security manager,
    * the UI and then tries to connect to the server. During the connection
    * attempt the server will send down a login screen or if guest access is
    * enabled, the login will be bypassed. Once login has occurred (or been
    * bypassed), the session is established and the directory manager and the
    * UI client finish initialization.  Finally the main application entry
    * point is called on the server.  The server will not return from that call
    * until the user asks to quit, a CTRL-C occurs (stop condition) or some
    * abnormal end occurs.  Any of these cases cause the loop to terminate and
    * this method will then return.
    * <p>
    * This method implements the full, normal client JVM startup and cannot be
    * used within the server process.
    * 
    * @param    config
    *           The configuration to use for setup of the client and for the
    *           server connection.
    * @param    file
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    */
   public static void start(BootstrapConfig config, boolean file)
   throws Exception
   {
      start(config, file, false);
   }
   
   /**
    * Start the client loop.  This will not return until the user is exiting
    * the client. The core loop initializes logging, the security manager,
    * the UI and then tries to connect to the server. During the connection
    * attempt the server will send down a login screen or if guest access is
    * enabled, the login will be bypassed. Once login has occurred (or been
    * bypassed), the session is established and the directory manager and the
    * UI client finish initialization.  Finally the main application entry
    * point is called on the server.  The server will not return from that call
    * until the user asks to quit, a CTRL-C occurs (stop condition) or some
    * abnormal end occurs.  Any of these cases cause the loop to terminate and
    * this method will then return.
    * 
    * @param    config
    *           The configuration to use for setup of the client and for the
    *           server connection.
    * @param    file
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    */
   public static void start(BootstrapConfig config, boolean file, boolean single)
   throws Exception
   {
      String uuid     = config.getString("server", "spawner", "uuid", null);
      boolean dbg     = config.getBoolean("logging", "debug", "enable", false);
      String referrer = config.getString("web", "referrer", "url", null);
      
      boolean running = true;
      
      // save off the UI driver so the same instance can be used across all
      // iterations of the core loop
      ScreenDriver<?> driver = (uuid == null) ? null : processTemporaryClient(uuid, config);
      
      // Load native library
      loadNativeLibrary();
      
      while (running)
      {
         Object          context = null; 
         Session         session = null;
         MainEntry       main    = null;
         SessionManager  sessMgr = null;
         SecurityManager sm      = null;
         
         try
         {
            if (!single)
            {
               // optionally initialize detailed logging (only use this for
               // debugging sessions!); this must be done before the SM inits
               // to ensure that the logging goes to the right place
               if (dbg)
               {
                  initLogging(file);
               }
               
               // instantiate security manager
               sm = SecurityManager.createInstance(config);
               
               if (dbg)
               {
                  sm.setDebugLevel(SecurityManager.DATA);
               }
               
               // instantiate network session manager
               sessMgr = SessionManagerFactory.createLeafNode(config);
            }
            
            // initialize core UI services
            ThinClient.initializePrep(config, driver, single);
            
            // at this point it is safe to query the driver and save it off
            // in case we need to re-initialize (the same driver must be
            // re-used even though we create new instances of pretty much all
            // other UI classes)
            if (driver == null)
            {
               driver = OutputManager.getDriver();
            }
            
            if (single)
            {
               sm = SecurityManager.getInstance();
               
               // authenticate and create a security context for this thread
               context = sm.authenticateSingle(config);
            }
            else
            {
               // establish or re-establish a session with the P2J server
               // (this handles authentication too)
               session = sessMgr.connectDirect(new InterruptedExceptionHandler(), null);
               
               // some client code must read configuration from the directory
               DirectoryManager.initRemote(session);
               
               if (referrer != null)
               {
                  WebClientRegistrar registrar =
                           (WebClientRegistrar) RemoteObject.obtainNetworkInstance(
                                    WebClientRegistrar.class, session);
                  registrar.registerWebClientSession(uuid, session.getNodeAddress());
               }
            }
            
         }
         catch (Exception exc)
         {
            // STOP conditions don't need to be reported, they are an expected
            // mechanism to terminate
            if (!isCauseStop(exc))
            {
               exc.printStackTrace(System.err);
               System.err.flush();
            }
            
            break;
         }
         
         // finish initializing the thin client
         ThinClient.initializePost(config, session, context);

         // attempt to emulate the '-ininame' command-line parameter: 
         Directory remoteDir = DirectoryManager.getInstance();
         if (remoteDir != null)
         {
            String ininame = remoteDir.getString(Directory.ID_RELATIVE, "ininame", "");
            if (!ininame.isEmpty())
            {
               EnvironmentDaemon daemon = ThinClient.getInstance().getEnvironmentDaemon();
               if (daemon != null)
               {
                  // On WIN32 AVM searches the HKCU then HKLM key for the specified name.
                  // If no registry key of the specified name is found, the AVM searches for
                  // a .INI file that matches name. 
                  // We are only interested to load stanza ini file regardless of the OS, so
                  // we ignore the registry and assume the parameter is a text ini file. 
                  daemon.setDefaultEnvironment(ininame);
               }
            }
         }
         
         ThinClient.loadEnvironment(config);
         
         // After full initialization of the client code we can set up the ErrorWriter for
         // client side.
         if (session != null)
         {
            OutputManager.initErrorWriter(session);
         }
         
         // obtain the entry point
         if (single)
         {
            // client code is "right" side by convention
            main = (MainEntry) RemoteObject.obtainLocalInstance(MainEntry.class, false);
         }
         else
         {
            main = (MainEntry) RemoteObject.obtainNetworkInstance(MainEntry.class, session);
         }
         
         // determine any client-side parameters
         ClientParameters params = processClientParams(driver, config);
         
         // Propagate client-side parameters to the server
         main.setClientParams(params);

         if (params.clientlog != null)
         {
            ThinClient.initClientLog(params.clientlog);
         }
         
         // initialize initial tooltip settings
         ThinClient.setSessionTooltips();

         // above this point, any CTRL-C processing that causes a STOP
         // condition to be thrown will cause an exit (this is very timing
         // sensitive but if you hold down the CTRL-C keys AND the server
         // environment is configured to force a relogin (the return from
         // standardEntry() is true) then at some point you will see such
         // an exception; we can protect against this, BUT this same behavior
         // has been seen in the 4GL, so it has been left here on purpose
            
         // Activates batch mode if exists. If batch mode operation is requested the real work
         // should be postponed until all subsystems to be initialized properly using usual
         // console mode. At this point this preliminary work should be done and we can
         // "activate" the special processing for batch mode. 
         if (OutputManager.inBatchMode())
         {
            OutputManager.activateBatch(true);
         }

         driver.clientReady();
         
         // call the application
         running = main.standardEntry();
         
         // dependent threads need to be shutdown, even if the client restarts on CTRL-C
         ThinClient.terminate();
         
         if (!running)
         {
            driver.shutdown();
         }
         
         // shut down session
         if (single)
         {
            sm.terminateSingle(context);
         }
         else
         {
            if (!running)
            {
               if (referrer != null)
               {
                  WebClientRegistrar registrar =
                           (WebClientRegistrar) RemoteObject.obtainNetworkInstance(
                                    WebClientRegistrar.class, session);
                  registrar.freeWebClientResources(uuid);
               }
            }
            session.terminate();
         }


         // clear after me
         SecurityManager.clearInstance();
      }
   }
   
   /**
    * Inspect the chain of exceptions and report if any of them are STOP
    * conditions.
    *
    * @param    chained
    *           The top of the chain of exceptions.
    *
    * @return   <code>true</code> if any of the exceptions in the chain are
    *           STOP conditions.
    */
   public static boolean isCauseStop(Throwable chained)
   {
      boolean isStop = false;
      
      while (chained != null)
      {
         if (chained instanceof StopConditionException)
         {
            isStop = true;
            break;
         }
         
         chained = chained.getCause();
      }
      
      return isStop;
   }

   /**
    * Loads native library and performs required initial steps for P2J library to working.
    */
   private static synchronized void loadNativeLibrary()
   {
      if (!nativeLayerInitialized)
      {
         System.loadLibrary("p2j");
         processInit();
         nativeLayerInitialized = true;
      }
   }
   
   /**
    * Native call to set up and redefine signal processing handlers.
    */
   private static native void processInit();
   
   /**
    * Native call to read the operating system pid for the current JVM process.
    */
   private static native long getPid();
   
   /**
    * Native call to read the operating system user name for the current JVM process.
    */
   private static native String getUserName();
   
   /**
    * Initialize logging at a detailed level for the infrastructure packages.
    *
    * @param    toFile
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    */
   private static void initLogging(boolean toFile)
   {
      LinkedHashMap<String, Level> loggingLevels =
         new LinkedHashMap<String, Level>();
         
      loggingLevels.put("com.goldencode.p2j.net", Level.FINEST);
      loggingLevels.put("com.goldencode.p2j.main", Level.FINEST);
      
      LogHelper.initialize(null,
                           Level.FINEST,                       // default level
                           !toFile ? Level.FINEST : Level.OFF, // console
                           toFile  ? Level.FINEST : Level.OFF, // file
                           "client_dbg_%g.log",
                           -1,
                           -1,
                           false,
                           loggingLevels);
   }
   
   /**
    * Build a separate BootstrapConfig for temporary clients. 
    * 
    * @param    config
    *           Client bootstrap configuration.
    *
    * @return   A new BootstrapConfig configured for temporary client connections.  
    */
   private static BootstrapConfig buildTemporaryClientConfig(BootstrapConfig config)
   {
      try
      {
         // Create web client bootstrap configuration. 
         BootstrapConfig bc = new BootstrapConfig();
         bc.setServer(false);

         // Get user credentials from environment  
         String subject = System.getenv(TemporaryAccount.P2J_SUBJECT);
         String password = System.getenv(TemporaryAccount.P2J_PASSWORD);
        
         // Add web client configuration.
         bc.setConfigItem("security", "trust_mgr", "disable", "true");
         bc.setConfigItem("net", "queue", "conversation", "true");
         bc.setConfigItem("net", "queue", "start_thread", "false");
         bc.setConfigItem("security", "certificate", "validate", "false");
         bc.setConfigItem("security", "authentication", "type", "program");
         bc.setConfigItem("access", "subject", "id", TemporaryAccount.fromHex(subject));
         bc.setConfigItem("access", "password", "user", TemporaryAccount.fromHex(password));
         
         String[] copy =
         {
            "net:connection:secure",
            "net:server:host",
            "net:server:secure_port",
            "net:server:insecure_port",
            "net:server:port",
            "server:spawner:uuid",
            "client:web:socketTimeout",
            "client:web:watchdogTimeout",
            "client:web:port",
            "client:web:host",
         };
         bc.copyConfigItems(copy, config);

         return bc;
      }
      catch (ConfigurationException e)
      {
         throw new RuntimeException(e);
      }
   }
   
   /**
    * Pull the client parameters from the driver and bootstrap configuration and initialize
    * a parameters object with the resulting data.
    *
    * @param    driver
    *           The primary UI driver for this client session.
    * @param    config
    *           The bootstrap configuration for this session (this must have been previously
    *           merged with any command line overrides).
    *
    * @return   The configured parameters object.
    */
   private static ClientParameters processClientParams(ScreenDriver<?> driver,
                                                       BootstrapConfig config)
   {
      ClientParameters cp  = new ClientParameters();
      String           win = driver.getWindowSystem();
      
      cp.windowSystem = config.getString("client", "driver", "window-system", win);
      cp.param        = config.getString("client", "cmd-line-option", "param", null);
      cp.startupProc  = config.getString("client", "cmd-line-option", "startup-procedure", null);
      cp.clientlog    = config.getString("client", "cmd-line-option", "clientlog", null);
      cp.osUserName   = getUserName();
      cp.terminalName = config.getString("client", "tty", "name", null);
      cp.debugAlert   = config.getBoolean("client", "cmd-line-option", "debugalert", false);
      cp.pid          = getPid();
      cp.chui         = driver.isChui();
      cp.batchMode    = config.getBoolean("client", "mode", "batch", false) ? 2 : 0;
      return cp;
   }
   
   /**
    * Establish a direct connection using the temporary credentials and perform custom work,
    * depending on the {@link TemporaryClient} implementation returned by
    * {@link Spawner#getTemporaryClient}.
    * 
    * @param   uuid
    *          Remote client identifier uuid.
    * @param   config
    *          The configuration to use for setup of the client and for the server connection.
    *          
    * @return  The screen driver returned by {@link TemporaryClient#doWork}.
    *   
    * @throws  Exception
    *          If something goes wrong.
    */
   private static ScreenDriver<?> processTemporaryClient(String uuid, BootstrapConfig config)
   throws Exception
   {
      ScreenDriver<?> driver = null;
      
      // Create a temporary session
      BootstrapConfig bc = buildTemporaryClientConfig(config);
      @SuppressWarnings("unused")
      SecurityManager sm = SecurityManager.createInstance(bc);
      SessionManager sessMgr = SessionManagerFactory.createLeafNode(bc);
      Session session = sessMgr.connectDirect(new InterruptedExceptionHandler(), null);
      DirectoryManager.initRemote(session);
      
      try
      {
         // Obtain Spawner remote access interface
         Spawner spawner = (Spawner) RemoteObject.obtainNetworkInstance(Spawner.class, session);
         if (spawner != null)
         {
            TemporaryClient tempClient = spawner.getTemporaryClient(uuid);
            driver = tempClient.doWork(spawner, uuid, config);
         }
         else
         {
            throw new RuntimeException(String.format("Network server %s not found!",
                                                     Spawner.class.getName()));
         }
      }
      finally
      {
         // Terminate temporary session.
         session.terminate();
      }
      
      return driver;
   }
}