/*
** Module   : UpdateAccountTask.java
** Abstract : a Runnable object used update temporary accounts status.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 MAG 20131216 Created initial version.
** 002 CA  20150909 Fixed a race condition on DirectoryService usage - as this is done in a 
**                  separate thread using the server's context, all threads need to sync on this
**                  object, as the DirectoryService is not thread-safe.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.logging.*;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * Runnable object used to update temporary accounts status.
 * This object is used as parameter to be executed by the temporary accounts pool thread.  
 * After the account is updated the security cache must be refreshed.
 * Otherwise the changes are not visible and the account remains disabled on later calls. 
 * On (re)enable account change also the password.  
 *  
 * @author mag
 *
 */
public class UpdateAccountTask 
implements Runnable
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(UpdateAccountTask.class.getName());
   
   /** Account pool item */
   private TemporaryAccount item;

   /** Enabled parameter */
   private boolean enabled;
     
   /**
    * Constructor.  
    * 
    * @param   item
    *          Account pool item.
    * @param   enabled
    *          Account status to be set.
    */
   public UpdateAccountTask(TemporaryAccount item, boolean enabled)
   {
      this.item = item;
      this.enabled = enabled;
   }
   
   /**
    * Task executor.
    */
   public void run()
   {
      boolean updated = false;
      
      // Change user account status
      UserDef user = AdminServerImpl.getUser(item.getSubject());
      if (user != null)
      {
         user.enabled = enabled;         
         if (enabled)
         {
            // On enable account change also the password.
            String password = RandomWordGenerator.create(TemporaryAccountPool.GEN_LENGTH, 
                                                         TemporaryAccountPool.GEN_DIGITS, 
                                                         TemporaryAccountPool.GEN_SYMBOLS);
            // Store password in pool.
            item.setPassword(password);
            user.password = HashPassword.hashPassword(item.getPassword());
         }
         // Change account
         synchronized (DirectoryService.getInstance())
         {
            updated = AdminServerImpl.setUser(user);
         }
      }            
      
      if (updated)
      {
         try
         {
            // Refresh user cached account.
            if (!SecurityManager.updateCachedUserAccount(user))
            {
               LOG.logp(Level.WARNING,
                        "UpdateAccountTask.run()",
                        "",
                        LogHelper.generate("Account %s not found inside security cache.", 
                                           user.subjectId));               
            }
         }
         catch (RestrictedUseException e)
         {
            LOG.logp(Level.WARNING,
                     "UpdateAccountTask.run()",
                     "",
                     LogHelper.generate("RestrictedUseException"),
                     e);                     
         }
      }
      else
      {
         LOG.logp(Level.WARNING,
                  "UpdateAccountTask.run()",
                  "",
                  LogHelper.generate("Cannot change %s account", item.getSubject()));               
      }
   }
}