/*
** Module   : ChangeBroker.java
** Abstract : Clearinghouse for DMO property change events
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 ECF 20060719   @28114 Created initial version. Brokers notification
**                           of DMO state changes to registered listeners.
** 002 ECF 20060722   @28134 Fixed listener registration defect. Listeners
**                           registered at inner scopes were hiding those
**                           registered for the same DMO types at outer
**                           scopes.
** 003 ECF 20060802   @28348 Make addListener() tolerant of multiple calls
**                           within the same scope. Listeners are now
**                           stored in a HashSet, so duplicates are
**                           prevented.
** 004 ECF 20060918   @29665 Added forcePendingFlush() method. Forces a
**                           pending session flush for a particular DMO
**                           type.
** 005 ECF 20060919   @29696 Fixed handling of pending flush state. The
**                           pending flush state was being tracked
**                           globally, across all databases, rather than
**                           per database. This led to errors in proper
**                           detection of whether a session flush was
**                           needed, because the state was being cleared
**                           too aggressively.
** 006 ECF 20061010   @30308 Made changes to accommodate auto-commit temp
**                           tables. These were primarily side effects of
**                           changes to the Persistence class to support
**                           this feature.
** 007 ECF 20070306   @32286 Added findDirty() implementation to session
**                           interceptor. This allows us to optimize dirty
**                           checking in some cases by avoiding the
**                           default dirty check comparisons which
**                           Hibernate performs.
** 008 ECF 20070322   @32603 Added forcePendingFlush() variant. Allows
**                           forcing a pending flush when a RecordBuffer
**                           instance is not available.
** 009 ECF 20070416   @33019 Integrated user interrupt handling.
** 010 ECF 20080103   @36869 Minor change in stateChanged(). Don't force
**                           a pending flush for a record that is still
**                           transient. Also integrated generics.
** 011 ECF 20080227   @37222 Fixed SessionInterceptor.findDirty(). Was
**                           improperly short-circuiting Hiberate dirty
**                           check in non-legacy mode.
** 012 ECF 20080314   @37471 Added debug logic. Number of DMO dirty checks
**                           performed since last flush are now tracked
**                           and reported when trace level logging is
**                           enabled.
** 013 ECF 20080403   @38138 Use entity name instead of DMO class to test
**                           if flush is required.
** 014 ECF 20080822   @39554 Modified stateChanged() to accept an event
**                           type parameter. Uses inner enum Type from
**                           RecordChangeEvent.
** 015 CA  20080815   @39452 Support API change in DBUtils.
** 016 SVL 20081024   @40217 Added addListener(.., int depth) function.
** 017 ECF 20090212   @41294 Implemented eviction of unused DMOs after a
**                           session flush. Temp table DMOs may be left
**                           behind after a bulk delete, causing more
**                           flushing effort than is necessary. We now
**                           evict these objects after a flush, if they
**                           are not in use.
** 018 ECF 20090610   @42971 Minor change to SessionInterceptor.postFlush().
**                           Log database info in trace mode.
** 019 ECF 20090716   @43214 Minor change to minimize memory used by listener
**                           collections.
** 020 CA  20131013          Added no-op scopeDeleted() method, required by the changes in
**                           Scopeable interface.
** 021 CA  20140513          Added a weight for the context-local var, to ensure predetermined
**                           order during context reset.
** 022 OM  20150112          Added support for storing listeners in the GLOBAl scope for
**                           persistent RUN procedure.
** 023 OM  20150213          Prevented a NPE in removeFromGlobal() when listeners is empty.
** 024 ECF 20150906          Changed SessionInterceptor to match changes in BufferManager API.
** 025 ECF 20160607          Reimplemented listener registration and notification to be more
**                           efficient. Reduced session flushing when processing state changes.
** 026 ECF 20160615          Added support needed for custom dirty entity strategy. Replaced
**                           Apache commons logging with J2SE logging.
** 027 ECF 20160907          Improved session flush management for better performance.
** 028 CA  20161010          Fixed stateChanged - all listeners, from all scopes, must be notified
**                           for a certain DMO interface.
** 029 ECF 20180311          Fixed listener registration. Cleanup interceptor state after
**                           transactions on commit as well rollback.
** 030 ECF 20180709          Refactored pending eviction processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.ref.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * A clearing house which notifies interested parties of changes made to
 * active records within the current client context.  These notifications
 * are used for various purposes:
 * <ul>
 *   <li>to enable {@link RecordBuffer} to defer creating a placeholder
 *       snapshot as long as possible;
 *   <li>to permit {@link Persistence} query methods to disable auto-flushing
 *       of the active Hibernate session when possible;
 *   <li>to allow {@link AdaptiveQuery} to invalidate its current result set
 *       when necessary.
 * </ul>
 * <p>
 * {@link com.goldencode.p2j.persist.event.RecordChangeListener
 * RecordChangeListener} objects register to receive notifications for one or
 * more DMO types, using the {@link #addListener} method.  A listener will
 * only receive notifications for changes to DMOs of the types for which they
 * registered.  Beyond this, no other filtering is performed.  Listeners are
 * maintained in a scoped dictionary, so there is never a need to remove a
 * listener explicitly;  when the scope in which the listener was registered
 * is closed, it is released automatically.
 * <p>
 * Change notifications are driven by the {@link #stateChanged} method.  It
 * is anticipated that this only will be invoked by <code>RecordBuffer</code>
 * objects, as user code calls setter methods on DMO proxies.  A state change
 * causes a {@link com.goldencode.p2j.persist.event.RecordChangeEvent
 * RecordChangeEvent} to be broadcast to all listeners registered against the
 * type of DMO which was modified.  "Type" in this sense corresponds to the
 * DMO interface.
 * <p>
 * This class also enlists a Hibernate <code>Interceptor</code> to listen for
 * Hibernate session events.  The context local <code>Persistence</code>
 * instance installs this interceptor into each Hibernate <code>Session</code>
 * it opens.  Whenever a DMO state change occurs, the implementation class of
 * the modified DMO is stored in a set of pending flush types.  When the
 * interceptor is notified of a session flush event, this set is cleared.
 * As an optimization, the <code>Persistence</code> class consults with the
 * <code>ChangeBroker</code> to determine whether a session flush is required
 * before a query is executed.  If not, it disables auto-flushing for the
 * query, which preempts considerable, unnecessary work.
 * <p>
 * The interceptor has another responsibility associated with temp tables.
 * Temporary DMOs can be removed from the database via a bulk delete, which
 * bypasses the <code>BufferManager</code>'s DMO reference counting mechanism.
 * As a result, a bulk delete can leave records associated with the open
 * Hibernate session, which can result in much more flushing effort than
 * necessary, in certain situations.  During a flush event, the interceptor
 * detects DMOs which are not in use and schedules these for eviction from the
 * session.  The eviction takes place immediately after the flush event is
 * finished.
 * <p>
 * Because it relies upon the <code>RecordBuffer</code> as the source for all
 * state change events, <code>ChangeBroker</code> is not useful in non-legacy
 * client contexts (i.e., where external code accesses the services of the
 * <code>Persistence</code> class directly).  For the most part, this class
 * is transparent for these contexts.  One exception is that the auto-flush
 * optimization technique described above would be harmful if allowed to
 * proceed naturally, because flushing would always be disabled since state
 * changes would never be reported, thus it would always seem flushing is not
 * required.  To prevent this problem, this class detects when it is not in
 * legacy mode and disables this optimization accordingly in such cases.
 */
final class ChangeBroker
implements Scopeable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ChangeBroker.class.getName());
   
   /** Context local instance of this class */
   private static final ContextLocal<ChangeBroker> context =
   new ContextLocal<ChangeBroker>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_6; }
      protected ChangeBroker initialValue() { return (new ChangeBroker()); }
   };
   
   /** Sets of listeners which have registered interest in DMO property changes */
   private final ScopedDictionary<Class<?>, ListenerSet> listeners = new ScopedDictionary<>();
   
   /** Map of saved listeners, per persistent procedure referent. */
   private final Map<Object, Map<Class<?>, ListenerSet>> procListeners = new HashMap<>();
   
   /** Objects which are notified of activity in Hibernate session */
   private final Map<Persistence, SessionInterceptor> interceptors = new IdentityHashMap<>();
   
   /** Is this client context using legacy services? */
   private boolean legacyMode = false;
   
   /**
    * This class can only be constructed privately.
    */
   private ChangeBroker()
   {
   }
   
   /**
    * Retrieve the context-local instance of this class, instantiating it
    * first if necessary.
    *
    * @return  Context-local instance of this class.
    */
   static ChangeBroker get()
   {
      return context.get();
   }
   
   /**
    * Register with the {@link com.goldencode.p2j.util.TransactionManager
    * TransactionManager} a factory object which creates instances of this
    * class, so that they can be registered to receive notifications of
    * runtime scope start and finish events.
    * <p>
    * This method should be invoked once during the server bootstrap phase.
    */
   static void initialize()
   {
      ScopeableFactory factory = new ScopeableFactory()
      {
         public Scopeable createScopeable()
         {
            return ChangeBroker.get();
         }
      };
      TransactionManager.registerScopeableFactory(factory);
   }
   
   /**
    * Add a scope to the listeners dictionary whenever a new runtime scope
    * opens.
    */
   public void scopeStart()
   {
      listeners.addScope(null);
   }
   
   /**
    * Pop the current scope from the listeners dictionary whenever a runtime
    * scope is closed.
    */
   public void scopeFinished()
   {
      Object referent = null;
      boolean saveForProc = false;
      
      if (!TransactionManager.isGlobalBlock())
      {
         referent = ProcedureManager.thisProcedure().get();
         BlockType bt = TransactionManager.getBlockType();
         // if we have a persistent procedure, save the data
         if (bt == BlockType.EXTERNAL_PROC && ProcedureManager._isPersistent(referent))
         {
            saveForProc = true;
         }
      }
      
      Map<Class<?>, ListenerSet> current = listeners.getDictionaryAtScope(0, false);
      if (current != null && !current.isEmpty())
      {
         for (Class<?> key : current.keySet())
         {
            ListenerSet set = current.get(key);
            set.unlink();
         }
         
         if (saveForProc)
         {
            procListeners.put(referent, current);
         }
      }

      listeners.deleteScope();
   }
   
   /**
    * Notification that the instantiating procedure where the scope was opened was deleted.
    * <p>
    * This implementation is a no-op.
    */
   @Override
   public void scopeDeleted()
   {
      Object referent = ProcedureManager.getProcessedProcedure();
      
      if (referent == null || ProcedureManager._isPersistent(referent))
      {
         // if no referent or procedure is still persistent, do not clean up yet
         return;
      }
      
      // this needs to be executed only when the procedure gets deleted.
      procListeners.remove(referent);
   }
   
   /**
    * Register a listener to receive DMO property change notifications.  The
    * listener will receive notifications of any changes to DMOs of the types
    * for which it is registered, as they occur in this scope or in scopes
    * nested within this scope.  The listener is deregistered automatically
    * when the scope ends.
    * <p>
    * This method is tolerant of being called multiple times for the same
    * listener in the same scope;  the listener will only be registered once,
    * and therefore will receive only one notification per distinct event.
    *
    * @param   listener
    *          Listener object to be added.
    */
   void addListener(RecordChangeListener listener)
   {
      addListener(listener, listeners.size());
   }
   
   /**
    * Register a listener to receive DMO property change notifications. The
    * registration is performed in the current scope and the outer scopes,
    * which number is specified by the <code>depth</code> parameter. The
    * listener will receive notifications of any changes to DMOs of the types
    * for which it is registered, as they occur in the specified scopes or in
    * the nested ones.  The listener is deregistered automatically
    * when the scopes end.
    * <p>
    * When we call this function with a <code>depth</code> less than the
    * number of innermost scope, we actually want to register it in some outer
    * scope, but since registrations at inner scopes will hide those at outer
    * scopes during a lookup, we should register it in the several scopes
    * if needed.
    * <p>
    * This method is tolerant of being called multiple times for the same
    * listener in the same scope;  the listener will only be registered once,
    * and therefore will receive only one notification per distinct event.
    * <p>
    * If the buffer is opened in a procedure executed persistent, the buffer has to survive
    * when the procedure code is finished until the prosistent procedure is deleted. In this
    * case the buffer is stored in the GLOBAL scope (depth = 0).
    *
    * @param   listener
    *          Listener object to be added.
    *
    * @param   depth
    *          1-based number of the most outer scope in which we want to register the listener
    *          (starting from the very outermost scope). Use 0 for GLOBAL scope when the buffer
    *          is opened in a persistent procedure.
    */
   void addListener(RecordChangeListener listener, int depth)
   {
      legacyMode = true;
      boolean global = (depth == 0);
      
      Iterator<RecordBuffer> iter = listener.recordBuffers();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         Class<?> dmoIface = buffer.getDMOInterface();
         ListenerSet set;
         
         if (global)
         {
            // insert to global scope
            set = listeners.getValueAtScope(dmoIface, -1);
            boolean createSet = set == null;
            
            if (createSet)
            {
               set = new ListenerSet();
            }
            
            if (set.nextInner == null && listeners.size() > 1)
            {
               ListenerSet nextInner = listeners.reverseLookup(dmoIface, 1);
               if (nextInner != null)
               {
                  nextInner.link(set);
               }
            }
            
            if (createSet)
            {
               listeners.addEntry(true, dmoIface, set);
            }
            
            set.add(listener);
         }
         else
         {
            // add the buffer to the specified scope of listeners
            int scope = listeners.size() - depth;
            set = listeners.getValueAtScope(dmoIface, scope);
            boolean createSet = set == null;
            
            if (createSet)
            {
               set = new ListenerSet();
            }
            
            if (set.nextOuter == null && depth > 1)
            {
               ListenerSet nextOuter = listeners.lookup(dmoIface, depth - 1);
               if (nextOuter != null)
               {
                  set.link(nextOuter);
               }
            }
            
            if (createSet)
            {
               listeners.addEntryAt(depth - 1, dmoIface, set);
            }
            
            set.add(listener);
         }
      }
   }
   
   /**
    * Remove a listener from global scope. This is only called for listeners whose buffers belong
    * to a procedure that was executed as persistent. In this case, opening the buffer registered
    * it to the global scope. This buffer is available as long as the procedure survives. When the
    * procedure is removed / deleted the buffer has to be removed as well.
    * <p>
    * If this buffer was not open in global scope (from a persistent procedure) this is a no-op and
    * the returned value is <code>false</code>.
    * 
    * @param   listener
    *          The listener to be removed.
    *
    * @return  <code>true</code> if the listener was successfully removed from global scope.
    */
   boolean removeFromGlobal(RecordChangeListener listener)
   {
      boolean ret = false;
      Iterator<RecordBuffer> iter = listener.recordBuffers();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         Class<?> dmoIface = buffer.getDMOInterface();
         // get the global scope for the dmoIface:
         ListenerSet set = listeners.size() == 0 ? null : listeners.getValueAtScope(dmoIface, -1);
         if (set != null)
         {
            ret |= set.remove(listener);
            
            // do the cleanup:
            if (set.isEmpty())
            {
               // if set of listeners is empty, remove from global scope
               listeners.removeEntryAtScope(dmoIface, listeners.size() - 1);
               set.unlink();
            }
         }
      }
      return ret;
   }
   
   /**
    * Remove the specified listener from all scopes.
    * 
    * @param    listener
    *           The listener to be removed.
    */
   void removeListener(RecordChangeListener listener)
   {
      Iterator<RecordBuffer> iter = listener.recordBuffers();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         Class<?> dmoIface = buffer.getDMOInterface();
         
         listeners.apply(dmoIface, (dictionary) ->
         {
            ListenerSet set = dictionary.get(dmoIface);
            
            if (set.remove(listener))
            {
               if (set.isEmpty())
               {
                  set.unlink();
                  dictionary.remove(dmoIface);
               }
               
               return true;
            }
            
            return false;
         });
      }
   }
   
   /**
    * Force a Hibernate session flush to occur before the next query against
    * the given buffer is executed.
    *
    * @param   buffer
    *          Buffer whose DMO type must be registered for a pending flush.
    */
   void forcePendingFlush(RecordBuffer buffer)
   {
      forcePendingFlush(buffer.getPersistence(), buffer.getDMOImplementationClass());
   }
   
   /**
    * Force a Hibernate session flush to occur before the next query against
    * the given DMO class is executed.
    *
    * @param   persistence
    *          Persistence service object.
    * @param   dmoClass
    *          DMO implementation class for which flushing must occur.
    */
   void forcePendingFlush(Persistence persistence, Class<?> dmoClass)
   {
      SessionInterceptor interceptor = getInterceptor(persistence);
      interceptor.setFlushPending(dmoClass.getName());
   }
   
   /**
    * Track the DMO specified in the event as dirty until it is flushed.
    * 
    * @param   event
    *          Record change event.
    */
   void markDirtyDMO(RecordChangeEvent event)
   {
      RecordBuffer buffer = event.getSource();
      Persistence persistence = buffer.getPersistence();
      SessionInterceptor interceptor = getInterceptor(persistence);
      RecordIdentifier ident = new RecordIdentifier(event.getEntity(), event.getDMO().getId());
      interceptor.markDirty(ident, event);
   }
   
   /**
    * Broadcast a notification that a DMO's state has changed.  All listeners
    * registered to receive events about DMOs of the type managed by
    * <code>buffer</code> will receive the event.
    *
    * @param   type
    *          Event type:  insert, update, or delete.
    * @param   buffer
    *          Record buffer which originated the event.
    * @param   dmo
    *          <b>Post-modification</b> state of the affected DMO (in the
    *          case where the modification is the deletion of the DMO from
    *          the database, this will be the object instance which was
    *          deleted).
    * @param   snapshot
    *          A deep copy of the DMO which reflects the state of the DMO as
    *          it was first set into the record buffer.  For a newly created
    *          record, this will contain the primary key assigned to the
    *          record.  All other properties will be set to their default
    *          values.
    * @param   properties
    *          A map whose keys are the names of those properties which have
    *          been modified and whose values are lists of the indexes at
    *          which those values are stored (indexed properties only).  For
    *          a simple property, the matching value will be
    *          <code>null</code>.
    *
    * @throws  PersistenceException
    *          if any error occurs within a listener while processing a
    *          change event.  Exceptions will be deferred until all listeners
    *          have had a chance to process the event.  If more than one
    *          listener throws an exception, only the first will be rethrown
    *          by this method;  however, all will be logged.
    */
   void stateChanged(RecordChangeEvent.Type type,
                     RecordBuffer buffer,
                     Persistable dmo,
                     Persistable snapshot,
                     Map<String, List<Integer>> properties,
                     boolean setPendingFlush)
   throws PersistenceException
   {
      // If the buffer is not transient, then the change represents either an
      // insert or an update/delete of an existing record.  In these cases, we
      // want to ensure the database is synchronized at the next opportunity.
      if (!buffer.isTransient() && setPendingFlush)
      {
         forcePendingFlush(buffer);
      }
      
      RecordChangeEvent event = null;
      
      if (dmo != null)
      {
         event = new RecordChangeEvent(type, buffer, dmo, snapshot, properties);
         markDirtyDMO(event);
      }
      
      Class<?> dmoIface = buffer.getDMOInterface();
      ListenerSet set = listeners.lookup(dmoIface);
      if (set == null)
      {
         return;
      }
      
      if (event == null)
      {
         event = new RecordChangeEvent(type, buffer, dmo, snapshot, properties);
      }
      Exception error = null;
      
      // Notify all registered listeners.
      for (RecordChangeListener listener : set)
      {
         try
         {
            listener.stateChanged(event);
         }
         catch (Exception exc)
         {
            DBUtils.handleException(buffer.getDatabase(), exc);
            
            // Can only throw the first exception caught, but we want to
            // give all listeners a chance to process the event.
            if (error == null)
            {
               error = exc;
            }
            
            // Log it regardless.
            LOG.log(Level.SEVERE, exc.getMessage(), exc);
         }
      }
      
      // process all listeners saved for persistent procedures, if they have registered for this 
      // DMO interface
      for (Object referent : procListeners.keySet())
      {
         Map<Class<?>, ListenerSet> procScope = procListeners.get(referent);
         ListenerSet procSet = procScope.get(dmoIface);
         if (procSet != null)
         {
            for (RecordChangeListener listener : procSet)
            {
               try
               {
                  listener.stateChanged(event);
               }
               catch (Exception exc)
               {
                  DBUtils.handleException(buffer.getDatabase(), exc);
                  
                  // Can only throw the first exception caught, but we want to
                  // give all listeners a chance to process the event.
                  if (error == null)
                  {
                     error = exc;
                  }
                  
                  // Log it regardless.
                  LOG.log(Level.SEVERE, exc.getMessage(), exc);
               }
            }
         }
      }
      
      if (error != null)
      {
         throw new PersistenceException(error);
      }
   }
   
   /**
    * Report whether a Hibernate session level flush would be required before
    * processing a query which expects results of type <code>dmoClass</code>.
    * A flush is required if any change has been made to a DMO of this type
    * in the current context, since the last time a session flush occurred.
    * If <code>dmoClass</code> is <code>null</code>, this method determines
    * that a flush is required if a DMO of <i>any</i> type has been modified
    * since the last time a session flush occurred.
    * <p>
    * This method will always recommend a flush when executing in a
    * non-legacy client context.
    *
    * @param   persistence
    *          Persistence object which is the key under which the
    *          interceptor object is stored.
    * @param   entityName
    *          DMO implementation class for which to check for pending flush,
    *          or <code>null</code> to check for any class.
    *
    * @return  <code>true</code> if a flush is required, else
    *          <code>false</code>.
    */
   boolean requiresFlush(Persistence persistence, String entityName)
   {
      if (!legacyMode)
      {
         return true;
      }
      
      SessionInterceptor interceptor = getInterceptor(persistence);
      
      return interceptor.isFlushPending(entityName);
   }
   
   /**
    * Retrieve the <code>Interceptor</code> instance to be installed in the
    * next Hibernate <code>Session</code> to be opened against the specified
    * database.
    *
    * @param   persistence
    *          Persistence object which is the key under which the
    *          interceptor object is stored.
    *
    * @return  This broker's <code>SessionInterceptor</code> instance.
    */
   SessionInterceptor getInterceptor(Persistence persistence)
   {
      SessionInterceptor interceptor = interceptors.get(persistence);
      if (interceptor == null)
      {
         interceptor = new SessionInterceptor(persistence);
         interceptors.put(persistence, interceptor);
      }
      
      return interceptor;
   }
   
   /**
    * An ordered set of listeners (iterates in the order added to the set), which optionally
    * links to other sets of listeners up and down the stack of open scopes.
    * <p>
    * When the set is iterated, it first walks its links to other listener sets to the outermost
    * in scope and iterates those first. It then walks reverse (weak) links back to more deeply
    * nested scopes and iterates listeners in those sets, and so on. Duplicate listeners are
    * only visited once.
    */
   private static class ListenerSet
   implements Iterable<RecordChangeListener>
   {
      /** Entities which have registered interest in DMO property changes */
      private final LinkedHashSet<RecordChangeListener> interested = new LinkedHashSet<>();
      
      /** Next set of listeners registered for the same events in an outer scope */
      private ListenerSet nextOuter = null;
      
      /** Next set of listeners registered for the same events in an inner scope */
      private WeakReference<ListenerSet> nextInner = null;
      
      /**
       * Add a listener.
       * 
       * @param   listener
       *          Listener to be registered.
       * 
       * @return  {@code true} if the listener was added; {@code false} if it was already present.
       */
      boolean add(RecordChangeListener listener)
      {
         return interested.add(listener);
      }
      
      /**
       * Remove a listener.
       * 
       * @param   listener
       *          Listener to be deregistered.
       * 
       * @return  {@code true} if the listener was removed; {@code false} if it could not be
       *          found.
       */
      boolean remove(RecordChangeListener listener)
      {
         return interested.remove(listener);
      }
      
      /**
       * Indicate whether the set of registered listeners is empty.
       * 
       * @return  {@code true} if empty, else {@code false}
       */
      boolean isEmpty()
      {
         return interested.isEmpty();
      }
      
      /**
       * Link to this object the next related listener set in an outer scope. Listeners in that
       * set will receive notifications before listeners in this set.
       * <p>
       * A link from {@code nextOuter} is made back to this listener set, so iteration across
       * these sets can work its way back from the outer set to this one. The link in this
       * direction is made with a weak reference, so this set can be garbage collected if its
       * scope is popped.
       * 
       * @param   nextOuter
       *          Outer listener set to be linked to this one.
       */
      void link(ListenerSet nextOuter)
      {
         this.nextOuter = nextOuter;
         nextOuter.nextInner = new WeakReference<>(this);
      }
      
      /**
       * Unlink this object from its next outer and next inner listener sets, if any, and link
       * those sets to each other instead.
       */
      void unlink()
      {
         ListenerSet nextOuter = this.nextOuter;
         WeakReference<ListenerSet> nextInner = this.nextInner;
         
         if (nextInner != null)
         {
            ListenerSet set = nextInner.get();
            if (set != null)
            {
               // replace nextInner's reference to us with reference to our nextOuter
               set.nextOuter = nextOuter;
            }
         }
         
         if (nextOuter != null)
         {
            // replace nextOuter's reference to us with reference to our nextInner
            nextOuter.nextInner = nextInner;
         }
         
         this.nextOuter = null;
         this.nextInner = null;
      }
      
      /**
       * Get an iterator which will iterate over outermost to innermost listeners in all outer
       * listener sets linked to and including this listener set.
       * 
       * @return  Iterator as described above.
       */
      @Override
      public Iterator<RecordChangeListener> iterator()
      {
         return new Iter();
      }
      
      /**
       * An iterator which walks all record change listeners registered to receive notifications
       * for a DMO type. The walk starts at the outermost scope and works its way to inner scopes,
       * ending with the listeners in the enclosing listener set. If a listener is registered in
       * more than one scope, it is visited only once, in the outermost scope in which it is
       * registered. Listeners are visited in the order in which they were registered. If the
       * enclosing listener set has not been linked to any outer listener sets, only the current
       * set is walked.
       */
      private class Iter
      implements Iterator<RecordChangeListener>
      {
         /** Number of scopes to iterate */
         private int scopes = 0;
         
         /** Listener set currently being iterated */
         private ListenerSet currentSet;
         
         /** Current worker iterator */
         private Iterator<RecordChangeListener> iterator;
         
         /** Set of listeners already visited in this iteration */
         private Set<RecordChangeListener> visited = null;
         
         /** Next listener to be visited */
         private RecordChangeListener next = null;
         
         /**
          * Default constructor.
          */
         Iter()
         {
            // start iterating the listener set at the outermost scope
            ListenerSet set = ListenerSet.this;
            while (set.nextOuter != null)
            {
               set = set.nextOuter;
               scopes++;
            }
            currentSet = set;
            iterator = currentSet.interested.iterator();
            
            // track visited listeners if we have more than one listener set to iterate
            if (scopes > 0)
            {
               visited = new HashSet<>();
            }
         }
         
         /**
          * Indicate whether there is a next listener to be visited.
          * 
          * @return  {@code true} if there is a next listener, else {@code false}.
          */
         @Override
         public boolean hasNext()
         {
            if (next != null)
            {
               return true;
            }
            
            boolean result;
            
            do
            {
               // reset next listener
               next = null;
               
               // if current iterator has no more listeners to visit and currentSet has not been
               // reset, try to advance to next inner listener set
               while (!(result = iterator.hasNext()) && currentSet != null)
               {
                  // only advance as many scopes as we counted during construction
                  if (--scopes >= 0)
                  {
                     WeakReference<ListenerSet> nextInner = currentSet.nextInner;
                     if (nextInner == null)
                     {
                        currentSet = null;
                        
                        return false;
                     }
                     else
                     {
                        currentSet = nextInner.get();
                        if (currentSet != null)
                        {
                           iterator = currentSet.interested.iterator();
                        }
                        else
                        {
                           return false;
                        }
                     }
                  }
                  else
                  {
                     currentSet = null;
                     
                     return false;
                  }
               }
               
               if (result)
               {
                  next = iterator.next();
               }
               
            } while (visited != null && next != null && visited.contains(next));
            
            return result;
         }
         
         /**
          * Get the next listener to be visited.
          * 
          * @return  Next record change listener.
          */
         @Override
         public RecordChangeListener next()
         {
            if (!hasNext())
            {
               throw new NoSuchElementException();
            }
            
            if (visited != null)
            {
               visited.add(next);
            }
            
            RecordChangeListener l = next;
            next = null;
            
            return l;
         }
         
         /**
          * Implementation of listener removal which throws an exception, since removal through
          * this iterator type is not supported.
          * 
          * @throws  UnsupportedOperationException
          *          always.
          */
         @Override
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
      }
   }
   
   /**
    * An implementation of a Hibernate <code>Interceptor</code> which listens
    * for session flush events and clears the enclosing instance's pending
    * flush status whenever a flush occurs.
    */
   class SessionInterceptor
   extends EmptyInterceptor
   {
      /** Persistence service object associated with this interceptor */
      private final Persistence persistence;
      
      /** Indicates an entity is not dirty */
      private final int[] NOT_DIRTY = new int[] {};
      
      /** Set of DMO entity types which require a Hibernate session flush */
      private final Set<String> pendingFlushTypes = new HashSet<>();
      
      /** Map of dirty DMOs which require a Hibernate session flush to record change events */
      private final Map<RecordIdentifier, RecordChangeEvent> dirtyDMOs = new HashMap<>();
      
      /** Flag indicating trace-level debug logging */
      private final boolean trace = LOG.isLoggable(Level.FINEST);
      
      /** Buffer manager for the current context */
      private final BufferManager bufferManager;
      
      /** Number of dirty checks performed by Hibernate since last flush */
      private int dirtyChecks = 0;
      
      /** Set of records detecting during flush as not being in active use */
      private Set<Persistable> pendingEvictions = new HashSet<>();
      
      /** Flag to track presence of any DMO entities which need flushing */
      private boolean empty = true;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence service object associated with this interceptor.
       */
      private SessionInterceptor(Persistence persistence)
      {
         this.persistence = persistence;
         
         if (DatabaseManager.TEMP_TABLE_DB.equals(persistence.getDatabase()))
         {
            // we only need the buffer manager to check whether temp-table DMOs are in use, to
            // account for entities which may be left in the Hibernate session after a bulk
            // delete, so we can evict them after a flush event
            bufferManager = BufferManager.get();
         }
         else
         {
            // we don't need to track in-use DMOs for permanent tables; in the current
            // implementation, there is no bulk delete feature for permanent tables, so we can
            // be pretty sure that the only records in the Hibernate session are those we want
            // there
            bufferManager = null;
         }
      }
      
      /**
       * Indicates whether a data model object entity either:
       * <ul>
       *   <li>might be dirty, possibly requiring a flush to the database;  or
       *   <li>definitely is not dirty.
       * </ul>
       * <p>
       * We do not definitively indicate that a record <i>is</i> dirty, since
       * Hibernate does not honor this indication when it comes to composite
       * elements within first class entities.  Therefore, when we can detect
       * that an entity might be dirty (i.e., its DMO class has been added to
       * the set of pending flush types), we return <code>null</code> to
       * indicate to Hibernate that it should use its default dirty checking
       * strategy to make this determination for sure.
       * <p>
       * We can, however, definitively detect when an entity cannot be dirty
       * (i.e., its DMO class is not present in the set of pending flush
       * types).  In this case, we short-circuit Hibernate's default checking
       * as an optimization.
       *
       * @param   entity
       *          The DMO to be tested for modifications.
       * @param   id
       *          The DMO's primary key ID.
       * @param   currentState
       *          Array of all of the DMO's properties in their current state.
       * @param   previousState
       *          Array of all of the DMO's properties in their previous
       *          state, presumably before they were updated.
       * @param   propertyNames
       *          Property names corresponding with the previous two arrays
       *          parameters.
       * @param   types
       *          Hibernate types corresponding with the previous three
       *          array parameters.
       *
       * @return  An empty array of <code>int</code>s to indicate that the
       *          entity is not dirty, or <code>null</code> to indicate the
       *          entity may be dirty.  Always return <code>null</code> if
       *          not in legacy mode, because we won't have enough information
       *          to second-guess Hibernate in that case.
       */
      @Override
      public int[] findDirty(Object entity,
                             Serializable id,
                             Object[] currentState,
                             Object[] previousState,
                             String[] propertyNames,
                             Type[] types)
      {
         if (trace)
         {
            dirtyChecks++;
         }
         
         // this optimization is only valid in legacy mode, because we don't track changes in
         // non-legacy mode
         if (!legacyMode)
         {
            return null;
         }
         
         Persistable dmo = (Persistable) entity;
         
         maybeEvictAfterFlush(dmo);
         
         if (empty)
         {
            // short circuit dirty check
            return NOT_DIRTY;
         }
         
         String entityName = entity.getClass().getName();
         if (!isFlushPending(entityName))
         {
            // short circuit dirty check
            return NOT_DIRTY;
         }
         
         RecordIdentifier ident = new RecordIdentifier(entityName, dmo.getId());
         if (!dirtyDMOs.containsKey(ident))
         {
            // short circuit dirty check
            return NOT_DIRTY;
         }
         
         // let Hibernate figure it out
         return null;
      }
      
      /**
       * Invoked by Hibernate after the current session has been flushed so
       * as to execute the SQL statements required to synchronize in-memory
       * session state with the database.  Triggers a full reset of the
       * pending flush state for this interceptor's database, such that the
       * broker will not recommend a session flush before the next DMO change
       * occurs.
       *
       * @param   entities
       *          An iterator on the DMO instances whose updates were flushed
       *          to the database.
       */
      public void postFlush(Iterator entities)
      {
         Set<RecordIdentifier> flushed = new HashSet<>();
         
         if (!empty || !dirtyDMOs.isEmpty() || trace)
         {
            while (entities.hasNext())
            {
               Persistable dmo = (Persistable) entities.next();
               RecordIdentifier ident = new RecordIdentifier(dmo.getClass().getName(),
                                                             dmo.getId());
               flushed.add(ident);
            }
         }
         
         if (trace && flushed != null)
         {
            String prefix = "[" + Utils.describeContext() + "] ";
            LOG.finest(prefix +
                       persistence.getDatabase() +
                       " dirty check count:  " +
                       dirtyChecks);
            dirtyChecks = 0;
            
            for (RecordIdentifier ident : flushed)
            {
               LOG.finest(prefix + " flushed: " + ident.getTable() + "#" + ident.getRecordID());
            }
         }
         
         cleanup(flushed);
      }
      
      /**
       * Called after a transaction is committed or rolled back. Causes a full reset of the
       * pending flush state for this interceptor's database, such that the broker will not
       * recommend a session flush before the next DMO change occurs.
       */
      public void afterTransactionCompletion(Transaction tx)
      {
         cleanup(null);
      }
      
      /**
       * Add the given DMO to the set of objects to be evicted from the session after the current
       * flush operation is complete, iff it is not referenced by any record buffer.
       * 
       * @param   dmo
       *          Data model object potentially to be queued for eviction after flush.
       */
      void maybeEvictAfterFlush(Persistable dmo)
      {
         // if the record is not in use, prepare to evict it after the flush
         if (bufferManager != null && bufferManager.getDMOUseCount(dmo) == 0)
         {
            pendingEvictions.add(dmo);
         }
      }
      
      /**
       * Mark a record as dirty.
       * 
       * @param   ident
       *          Primary key.
       * @param   event
       *          Record change event.
       */
      void markDirty(RecordIdentifier ident, RecordChangeEvent event)
      {
         dirtyDMOs.put(ident, event);
      }
      
      /**
       * Get the last record change event for the DMO the given ID, if that DMO currently is
       * dirty.
       * 
       * @param   ident
       *          Unique record identifier.
       * 
       * @return  Last record change event recorded for the DMO with the given ID, or {@code
       *          null} if there is none.
       */
      RecordChangeEvent getLastDirtyEvent(RecordIdentifier ident)
      {
         return dirtyDMOs.get(ident);
      }
      
      /**
       * Stop tracking the DMO with the given ID as dirty; called after the changes have been
       * flushed.
       * 
       * @param   ident
       *          Unique record identifier.
       */
      void resetDirty(RecordIdentifier ident)
      {
         dirtyDMOs.remove(ident);
      }
      
      /**
       * Indicate whether a Hibernate session flush is pending for the given
       * DMO implementation class.
       *
       * @param   entityName
       *          DMO implementation class as stored by the DMO's record
       *          buffer.  Do not use <code>getClass()</code> on a DMO to
       *          retrieve this argument, as Hibernate may have proxied the
       *          DMO and an unexpected class will be returned.
       *
       * @return  <code>true</code> if a flush is pending for the DMO type
       *          indicated by <code>dmoClass</code>.  If the argument is
       *          <code>null</code>, this method conservatively returns
       *          <code>true</code> if <i>any</i> DMO class has a session
       *          flush pending.
       */
      boolean isFlushPending(String entityName)
      {
         if (entityName != null)
         {
            return (!empty && pendingFlushTypes.contains(entityName));
         }
         
         return !empty;
      }
      
      /**
       * Set a Hibernate session flush pending for the given DMO
       * implementation class.
       *
       * @param   entityName
       *          DMO implementation class as stored by the DMO's record
       *          buffer.  Do not use <code>getClass()</code> on a DMO to
       *          retrieve this argument, as Hibernate may have proxied the
       *          DMO and an unexpected class will be returned.
       */
      void setFlushPending(String entityName)
      {
         pendingFlushTypes.add(entityName);
         empty = false;
      }
      
      /**
       * Clean up the interceptor's state after a flush or transaction. This involves removing
       * all flushed records and types from the pending flush types and dirty DMO collections,
       * and evicting any records we detected during the flush as not being in use.
       * <p>
       * A partial flush may result in some entity types being flushed, while others are not.
       * This can happen, for example, if a query only needs to synchronize certain query spaces.
       * 
       * @param   flushed
       *          Set of record identifiers for those records that were flushed during the most
       *          recent session flush, or {@code null} if none. When called at the end of a
       *          transaction, this parameter should be {@code null}.
       */
      private void cleanup(Set<RecordIdentifier> flushed)
      {
         if (flushed == null)
         {
            pendingFlushTypes.clear();
            dirtyDMOs.clear();
            empty = true;
         }
         else if (!flushed.isEmpty())
         {
            for (RecordIdentifier ident : flushed)
            {
               pendingFlushTypes.remove(ident.getTable());
               dirtyDMOs.remove(ident);
            }
            empty = pendingFlushTypes.isEmpty();
         }
         
         if (!pendingEvictions.isEmpty())
         {
            try
            {
               persistence.evict(pendingEvictions);
            }
            catch (ContextLocalCleanupException exc)
            {
               // only occurs during session cleanup, because we can't do context-local lookups
               // at that time; ok to ignore
            }
            finally
            {
               pendingEvictions.clear();
            }
         }
      }
   }
}