/*
** Module   : DMOValidator.java
** Abstract : Responsible for all runtime validation on a DMO
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20051106   @23392 Created initial version. Performs record- and
**                           field-level validation of database records,
**                           including mandatory and unique constraints,
**                           as well as custom validation processing (as
**                           converted from Progress validation
**                           expressions).
** 002 ECF 20060116   @23944 Fix for unique constraint validation. Ensure
**                           record found is not the current record.
** 003 ECF 20060116   @23949 Fix for not-null validation. Skip full record
**                           check if a single property is to be checked,
**                           which is nullable.
** 004 ECF 20060126   @24351 Fix latent bug exposed by temp table support.
**                           NPE was possible checking not-null constraint
**                           on a property with 'field' access (instead of
**                           get/set methods).
** 005 ECF 20060308   @25013 Added natural join validation support. Allow
**                           validation of DMO properties (instead of only
**                           BaseDataType properties). Improved error
**                           message content.
** 006 ECF 20060608   @27004 Changed unique constraint validation. Fixes a
**                           recent regression and uses a slightly more
**                           efficient query.
** 007 ECF 20060627   @27588 Cleanup. Removed unused local variable.
** 008 ECF 20060713   @28044 Use new DBUtils class.
** 009 ECF 20060719   @28102 Use updated signature for Persistence.list().
** 010 ECF 20061207   @31619 Eliminated redundant database access. Unique
**                           constraint validation was executing the same
**                           query once per column participating in the
**                           unique constraint.
** 011 ECF 20070130   @32038 Fixed validation of temp-table records.
**                           Uniqueness constraint validation was not
**                           taking the multiplex ID of a TemporaryBuffer
**                           into account.
** 012 ECF 20070212   @32153 Integrated dirty reads into unique constraint
**                           validation. Uncommitted transaction data from
**                           other client contexts must be used when
**                           performing this validation, in order to mimic
**                           Progress' behavior.
** 013 ECF 20070416   @33021 Integrated user interrupt handling.
** 014 ECF 20070507   @33457 Fixed ClassCastException in
**                           uniqueConstraintViolation() method. Was not
**                           taking multiplex ID and primary key ID into
**                           account when reporting error.
** 015 ECF 20070523   @33726 Added validate() variants to validate a
**                           different DMO than is currently in a record
**                           buffer.
** 016 ECF 20070531   @33912 Removed DMO-based validation method handling.
**                           This was an artifact of early planning to
**                           implement validation methods in DMO classes.
**                           These ultimately were moved into business
**                           logic, so this processing is no longer
**                           necessary or desirable, since it adds
**                           overhead.
** 017 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 018 ECF 20070827   @34999 Optimized certain temp table queries.
**                           Eliminate multiplex ID check from where
**                           where possible and hard code multiplex ID
**                           when this test is necessary, rather than
**                           using a query substitution parameter.
** 019 ECF 20070911   @35286 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections. Cleaned up error
**                           messages.
** 020 ECF 20080330   @37737 Fixed unique constraint check. The HQL being
**                           generated was not able to use indices for
**                           some tables. Multiplex ID was not being
**                           properly integrated for temp tables. Also,
**                           text property references were not being
**                           properly generated to account for trimming
**                           and case sensitivity. As a result, we were
**                           doing table scans for every unique constraint
**                           check on these tables! Major performance
**                           improvement. Also integrated generics.
** 021 ECF 20080402   @38407 Minor improvement to validateUnique().
**                           Removed rightTrim of (character) query
**                           substitution parameters, since this is
**                           redundant with the parameter preprocessing
**                           Persistence does. This also removes a subtle
**                           defect found by CA (this call was always
**                           converting character args to empty strings).
** 022 CA  20080526   @38392 Fixed ClassCastException caused by the method
**                           uniqueConstraintViolation - the multiplex id
**                           should be skipped, for temporary records.
**                           Ref: #38391
** 023 ECF 20080605   @38601 Rewrote unique constraint validation. It now
**                           uses DirtyShareContext instead of
**                           ValidationContext and DirtyReadManager.
** 024 SVL 20080620   @38882 validateUnique() obtains EXCLUSIVE lock 
**                           on unique indexes from the very beginning
**                           (instead of SHARE at first, then EXCLUSIVE).
** 025 CA  20080716   @39123 Modifications due to DirtyShareContext API 
**                           changes.
** 026 ECF 20080827   @39607 Removed unused validate(Persistable, Method)
**                           method. This originally was intended for the
**                           invocation of custom validation logic, but
**                           that is handled directly in business logic
**                           instead.
** 027 SVL 20080905   @39701 Optimization of the case when we meet an
**                           unknown value in the DMO fields fixed into 
**                           validateUnique(..).
** 028 ECF 20090220   @41352 Removed unused data structure. The messages
**                           map was originally intended to support custom
**                           validation logic, but that is handled in
**                           business logic instead.
** 029 ECF 20090312   @41638 Fixed validateUnique(). It was possible for
**                           this method to skip validation for a new,
**                           transient record which was not unique due to
**                           non-unique default values.
** 030 GES 20090424   @41937 Import change.
** 031 ECF 20090508   @42123 Rewrote key parts of validation processing to
**                           allow more incremental validation. Also moved
**                           most dirty property sharing from RecordBuffer to
**                           this class.
** 032 ECF 20090512   @42151 Fixed regressions caused by #031 (@42123).
** 033 ECF 20090513   @42165 Fixed checkUniqueAndShare(). If lone dirty
**                           property was not indexed and represented an
**                           extent field, it would cause a problem sharing
**                           the update via the DirtyShareManager.
** 034 CA  20090709   @43125 Fixed checkUnique(). During unique constraint
**                           check, the new value is handled same as the old
**                           values (unknown is always unique and value is
**                           upper cased for a case-insensitive character
**                           column).
** 035 ECF 20090714   @43216 Optimized unique constraint checks. Check whether
**                           backing table is empty and skip unique check in
**                           this case. Use dialect-optimized queries against
**                           the dirty database.
** 036 ECF 20090724   @43417 Eliminated unnecessary validation. If no indexes
**                           are affected by a pending change, skip unique
**                           constraint validation and dirty share processing.
** 037 AIL 20121206          Update from character to TextOps.
** 038 VMN 20130711          Fixes in messages in checkNotNull().
** 039 ECF 20131010          Added metadata support.
** 040 ECF 20131028          Import change.
** 041 SVL 20131028          Use legacy buffer and field names in uniqueConstraintViolation.
** 042 SVL 20140210          Use HQLExpression instead of HQL string.
** 043 OM  20140508          Added support for validating the index-key size.
** 044 OM  20140617          checkIndexMaxSize(): optimization; use the TableMapper to get the
**                           name of the index that failed size validation.
** 045 ECF 20140821          Changed to work with updated dirty share tracking. Replaced Apache
**                           commons logging with J2SE logging.
** 046 ECF 20150302          Added unique constraint violation logging.
** 047 ECF 20160225          Changes required by PropertyHelper rewrite.
** 048 OM  20160302          Allowed transient records to be shared if at least a full index was
**                           altered. The fields already changed are shared. Including EXTENTS.
** 049 ECF 20160912          Reworked unique constraint validation and flushing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.directory.DirectoryService;
import com.goldencode.p2j.persist.dialect.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * Performs record-level and field-level validation of database records.
 * This class is intended to support the Progress semantic of "real-time"
 * data validation, where assignments of new values to fields of the record
 * held in a record buffer receive immediate validation against database
 * constraints and schema-level validation expressions.  The types of
 * validation are:
 * <ul>
 *   <li>mandatory (non-nullable) constraint checking;
 *   <li>unique constraint checking;
 *   <li>custom validation expression execution.
 * </ul>
 * <p>
 * If validation is triggered for a single field, only the constraints and
 * validation expression (if any) related to that field are validated.  If
 * validation is triggered for the entire record, all constraints are checked
 * and all validation expressions are executed.
 * <p>
 * <strong>Mandatory Field Check</strong><br>
 * This type of check can be done completely in-process, without accessing
 * the database at all.  A single field check is performed within the context
 * of an impending update to the field's current value.  The proposed new
 * value is provided by the caller and the test is simply that this value
 * does not represent the unknown value.  For a full record check, reflection
 * is used to retrieve the current values of all non-nullable properties in a
 * DMO.  If any value (including any element in an array property) represents
 * the unknown value, the check fails.
 * <p>
 * <strong>Unique Constraint Check</strong><br>
 * This type of check requires the execution of (at least) one query against
 * the database per constraint being checked.  This check can be triggered
 * during update of a single field, or arbitrarily against an entire record.
 * To check against a single constraint, reflection is used to collect the
 * current value of each component property (except in the case of an update,
 * in which case the proposed new value is used for the property being
 * updated).  This data is used for the substitution parameters of a
 * pre-defined HQL query.  The query is submitted to the database and if any
 * record returns, the validation fails.  For a full record check, the same
 * process is repeated for all unique constraints defined for the DMO.  In
 * the latter case, <b>all</b> data is collected from the DMO via reflection.
 * Also in the latter case, the modified record must not be flushed to the
 * database before the query is executed (it is the default behavior of
 * Hibernate to flush modifications to the database before executing any
 * query).  Thus, we explicitly disable flushing temporarily in this case,
 * then re-enable it after the query is complete.  In addition, to fully
 * mimic the Progress semantic, we simulate the ability to perform dirty
 * reads across concurrent transactions in other contexts, using the {@link
 * com.goldencode.p2j.persist.dirty.DirtyShareContext DirtyShareContext}.
 * <p>
 * <strong>Custom Validation</strong><br>
 * Not used at this time.  This was originally intended to support field-level
 * validation expressions.  However, these are implemented in converted
 * business logic today because they can invoke arbitrary code paths.
 * <p>
 * <b>Important:</b> while instances of this class contain mapped data to
 * enable lookups of information required for validation, these data
 * structures are all initialized at the time the instance is constructed.
 * Instances of this class are shared across record buffers and across user
 * contexts, so they must remain <i>stateless</i> with respect to the buffers
 * and DMO records being validated.
 */
final class DMOValidator
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(DMOValidator.class.getName());
   
   /** Cache of previously created instances, mapped by DMO interface */
   private static final Map<Class<?>, DMOValidator> cache = new HashMap<>();
   
   /**
    * The maximum permitted size for all combined fields of an index. By default this 
    * is equal to Progress version 10.x limit of 1971 bytes.
    * <p>
    * Set it to 0 to disable P4GL index-key limit check. The dialect specific limit is still 
    * enforced by each dialect because otherwise irrecoverable errors may occur when persisting 
    * records with wider key length. 
    */
   private static final int p4glMaxIndexSize;
   
   /** Map of unique index names to <code>HQLBundles</code> */
   private final Map<String, HQLBundle> uniqueConstraints = new HashMap<>();
   
   /** Map of non-nullable DMO properties to their getter methods */
   private final Map<String, Method> notNullProperties = new HashMap<>();
   
   /** Names of properties which trigger validation mapped to their getters */
   private final Map<String, Method> triggers = new HashMap<>();
   
   /** Setter methods, mapped by property names */
   private final Map<String, Method> setterMap;
   
   /** Property names, mapped by getter methods */
   private final Map<Method, String> propertyNames = new HashMap<>();
   
   /** Ignore case flags, mapped by character getter methods */
   private Map<Method, Boolean> charProps = null;
   
   /* Initialization of static fields performed when the class is loaded.*/
   static 
   {
      // read MAX_INDEX_SIZE from registry configuration anc cache it for later use
      DirectoryService ds = DirectoryService.getInstance();
      p4glMaxIndexSize = Utils.getDirectoryNodeInt(ds,
                                                   "persistence/max-index-size",
                                                   DBUtils.DEFAULT_MAX_INDEX_SIZE,
                                                   false);
   }
   
   /**
    * Construct a validator, based upon the database, and the DMO interface
    * and class represented by <code>buffer</code>.
    * <p>
    * The following data structures are initialized in this constructor:
    * <ul>
    *   <li>{@link #triggers} - A map of DMO properties to their companion
    *       getter methods.  Only those properties whose modification must
    *       trigger a validation check (of any type) are keys in this map.
    *       The getter method values of this map are provided for convenience
    *       during validation processing.
    *   <li>{@link #notNullProperties} - A map of DMO properties to their
    *       companion getter methods.  The contents of this map are a subset
    *       of the contents of the <code>triggers</code> map.  Only those
    *       setters which update non-nullable properties are contained in this
    *       map.
    *   <li>{@link #uniqueConstraints} - A map of unique index names to the
    *       {@link HQLBundle} objects which will be used to issue database
    *       queries to test for the existence of records which should be
    *       unique.
    * </ul>
    *
    * @param   buffer
    *          Record buffer which contains information required to create a
    *          new validator instance.
    */
   private DMOValidator(RecordBuffer buffer)
   {
      Database database = buffer.getDatabase();
      Class<?> dmoIface = buffer.getDMOInterface();
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      
      setterMap = PropertyHelper.legacySettersByProperty(dmoIface);
      Map<String, Method> getterMap = PropertyHelper.legacyGettersByProperty(dmoIface);
      // Map property names by getter methods (for error messages).
      for (Map.Entry<String, Method> next : getterMap.entrySet())
      {
         propertyNames.put(next.getValue(), next.getKey());
      }
      
      // Collect all DMO setter methods which should trigger a not-null test.
      Iterator<String> nnIter = DatabaseManager.notNullProperties(database, dmoClass);
      while (nnIter.hasNext())
      {
         String name = nnIter.next();
         Method getter = getterMap.get(name);
         if (getter != null) // could have "field" access
         {
            notNullProperties.put(name, getter);
            triggers.put(name, getter);
         }
      }
      
      // Map unique index names for the target DMO type to HQLBundles which
      // will be used when performing unique constraint validation.
      IndexHelper idxHlp = buffer.getIndexHelper();
      String entity = buffer.getEntityName();
      try
      {
         Map<String, List<String>> map = idxHlp.getIndexMapForEntity(entity, true);
         for (Map.Entry<String, List<String>> next : map.entrySet())
         {
            String index = next.getKey();
            List<String> properties = next.getValue();
            buildUniqueBundle(index, properties, getterMap, buffer);
         }
      }
      catch (PersistenceException exc)
      {
         throw new StopConditionException(exc);
      }
   }
   
   /**
    * Get an instance of a validator, based upon the DMO interface represented
    * by <code>buffer</code>.  Validator instances are immutable once created,
    * and as such they are shared across multiple buffers and multiple user
    * contexts (they are stateless, so this is thread-safe).
    *
    * @param   buffer
    *          Record buffer which contains information required to create a
    *          new validator instance.
    *
    * @return  A validator instance for use by <code>buffer</code>.
    */
   static DMOValidator getInstance(RecordBuffer buffer)
   {
      synchronized (cache)
      {
         Class<?> dmoIface = buffer.getDMOInterface();
         DMOValidator validator = cache.get(dmoIface);
         if (validator == null)
         {
            validator = new DMOValidator(buffer);
            cache.put(dmoIface, validator);
         }
         
         return validator;
      }
   }
   
   /**
    * Get the set of names of properties which are non-nullable, for the DMO
    * type managed by this validator.
    * 
    * @return  Set of non-nullable property names, or the empty set if there
    *          are no such properties.  If non-empty, the caller can modify
    *          this set, since it is a copy of internal data.
    */
   Set<String> getNonNullables()
   {
      Set<String> set = notNullProperties.keySet();
      if (set.isEmpty())
      {
         return Collections.emptySet();
      }
      
      return new HashSet<>(set);
   }
   
   /**
    * Execute validation on the properties and constraints of the target
    * record specified in the given validation parameters.  Assumes all
    * property values to be tested are already set into the object.
    *
    * @param   buffer
    *          Record buffer.
    * @param   target
    *          Record to be validated.  If <code>null</code>, then
    *          <code>buffer</code>'s current record is validated.
    * @param   vData
    *          Validation parameters.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  StopConditionException
    *          if the size of a index key is found greater than allowable limits. 
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being
    *          validated.
    */
   void validate(RecordBuffer buffer, Persistable target, ValidationData vData)
   throws ValidationException,
          PersistenceException
   {
      validate(buffer, target, vData, null, null);
   }
   
   /**
    * Execute validation on the proposed property value change to the record.
    * Additional property values may be queried from the record to complete
    * the validation.
    * <p>
    * This method is invoked when a setter method is about to be invoked
    * upon a DMO to change the value of one of its properties, or when the
    * entire record must be inspected and validated.  Validation is skipped
    * if <code>property</code> has not been determined to be a validation
    * trigger.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being
    *          validated.
    */
   void validate(RecordBuffer buffer, ValidationData vData, String property, BaseDataType value)
   throws ValidationException,
          PersistenceException
   {
      validate(buffer, buffer.getCurrentRecord(), vData, property, value);
   }
   
   /**
    * Execute validation on the proposed property value change to the record.
    * Additional property values may be queried from the record to complete the validation.
    * <p>
    * This method is invoked when a setter method is about to be invoked upon a DMO to change the
    * value of one of its properties, or when the entire record must be inspected and validated.
    * Validation is skipped if <code>setter</code> would update a property which has not been
    * determined to be a validation trigger.
    *
    * @param   buffer
    *          Record buffer.
    * @param   target
    *          Record to be validated.  If <code>null</code>, then <code>buffer</code>'s current
    *          record is validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation completes
    *          successfully. If <code>null</code>, the validation parameters are consulted to
    *          determine which properties and constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation completes successfully.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  StopConditionException
    *          if the size of a index key is found greater than allowable limits.
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being validated.
    */
   private void validate(RecordBuffer buffer,
                         Persistable target,
                         ValidationData vData,
                         String property,
                         BaseDataType value)
   throws ValidationException,
          PersistenceException
   {
      try
      {
         if (target == null)
         {
            target = buffer.getCurrentRecord();
         }
         
         // the index-key length is checked before all other constraints
         checkIndexMaxSize(buffer, vData, property, value);
         
         checkNotNull(target, vData, property, value);
         checkUniqueAndShare(buffer, target, vData, property, value);
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException("Error inspecting DMO", exc);
      }
      catch (InvocationTargetException exc)
      {
         DBUtils.handleException(buffer.getDatabase(), exc);
         
         throw new RuntimeException("Error inspecting DMO", exc);
      }
   }
   
   /**
    * Validate a single not-null constraint, or all such constraints defined
    * for the <code>target</code> DMO in the validation parameters, depending
    * upon the value of the given <code>property</code>.
    *
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if a constrained property contains the unknown value.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkNotNull(Persistable target,
                             ValidationData vData,
                             String property,
                             BaseDataType value)
   throws ValidationException,
          IllegalAccessException,
          InvocationTargetException
   {
      if (notNullProperties.isEmpty())
      {
         // Nothing to do!
         return;
      }
      
      Set<String> dirtyProps = vData.getDirtyProperties();
      
      // Single property check.
      if (property != null)
      {
         if (dirtyProps.contains(property) &&
             notNullProperties.containsKey(property) &&
             value.isUnknown())
         {
            throw new ValidationException(
               StringHelper.changeCase(property, true) + " is mandatory, but has a value of ?",
               275);
         }
         
         return;
      }
      
      // Walk each non-nullable, dirty property and invoke its getter method;
      // test the result to ensure it is not the unknown value.
      for (String dProp : dirtyProps)
      {
         Method getter = notNullProperties.get(dProp);
         
         if (getter == null)
         {
            // Dirty property is nullable, no need to check.
            continue;
         }
         
         boolean indexed = (getter.getParameterTypes().length == 1);
         try
         {
            // Invoke simple getter once or indexed getter until out of valid
            // index values.
            for (int i = 0; indexed || i == 0; i++)
            {
               Integer index = (indexed ? i : null);
               Object pv = getPropertyValue(target, getter, index);
               if (pv instanceof BaseDataType)  // TODO:  necessary check?
               {
                  BaseDataType bdt = (BaseDataType) pv;
                  if (bdt.isUnknown())
                  {
                     throw new ValidationException(
                        StringHelper.changeCase(dProp, true) +
                        " is mandatory, but has a value of ?",
                        275);
                  }
               }
               else if (pv == null)
               {
                  throw new ValidationException(
                     "Non-nullable property " + dProp + " is null",
                     -1);
               }
            }
         }
         catch (ArrayIndexOutOfBoundsException exc)
         {
            // TODO:  use size API in DMO interface to detect maximum extent
            // for an indexed property, instead of using an out of bounds
            // exception to change control flow.  This try-catch is a hack.
         }
      }
   }
   
   /**
    * Validate the record against the maximum allowable index size for the current configured
    * dialect. If there is at least one index that will pass this limit when the change is
    * applied a <code>ValidationException</code> is thrown.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation completes
    *          successfully. If <code>null</code>, the validation parameters are consulted to
    *          determine which properties and constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation completes successfully.
    *
    * @throws  ValidationException
    *          if any index affected by this record will pass the maximum supported size
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception upon invocation.
    * @throws  StopConditionException
    *          if the size of a index key is found greater than allowable limits.
    */
   void checkIndexMaxSize(RecordBuffer buffer,
                          ValidationData vData,
                          String property,
                          BaseDataType value)
   throws ValidationException,
          PersistenceException,
          IllegalAccessException,
          InvocationTargetException
   {
      final int dialectMaxIndexSize = buffer.getDialect().getIndexLengthLimit();
      if (p4glMaxIndexSize == 0 && dialectMaxIndexSize == 0)
      {
         // quick-out for index-key size validation is disabled
         return;
      }
      
      // cache the size of each computed field size for this record, a field is added to map as 
      // they are processed; fields not part of any index won't appear in the cache
      Map<String, Integer> p4glFieldSizeCache = null;
      Map<String, Integer> dialectFieldSizeCache = null;
      if (p4glMaxIndexSize != 0)
      {
         // enable caching of size of the fields into the p4gl index structure
         p4glFieldSizeCache = new HashMap<>();
      }
      if (dialectMaxIndexSize != 0)
      {
         // enable caching of size of the fields into the current dialect index structure
         dialectFieldSizeCache = new HashMap<>();
      }
      
      if (!vData.getDirtyUnique().isEmpty())
      {
         // look for unique indexes that will be affected by the change
          checkIndexSizeWorker(buffer,
                              p4glFieldSizeCache,
                              dialectFieldSizeCache,
                              vData.getDirtyUnique(),
                              property,
                              value,
                              true);
      }
      
      if (!vData.getDirtyNonUnique().isEmpty())
      {
         // look for non-unique indexes that will be affected by the change
         checkIndexSizeWorker(buffer,
                              p4glFieldSizeCache,
                              dialectFieldSizeCache,
                              vData.getDirtyNonUnique(),
                              property,
                              value,
                              false);
      }
      
      // if we reached this point all index keys are in allowable byte limits 
   }
   
   /**
    * Private worker for checking a the index key size of a record for a set of indexes. If the
    * computed value is greater than the maximum allowable a STOP condition will be raised.
    * The method calculates for each index the sum of sizes of each components using P4GL
    * index storing algorithm described in <code>computeP4GLFieldSize</code>.
    * 
    * TODO: Add a getter in BaseDataType for extracting the size used by an index ? 
    *  
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   p4glFieldSizeCache
    *          The cache with already computed field sizes in the index of Progress from a 
    *          possible previous call. The map will be updated as index field sizes are computed.
    *          If <code>null</code> 4GL validation is disabled.
    * @param   dialectFieldSizeCache
    *          The cache with already computed field sizes in the index of current dialect from a
    *          possible previous call. The map will be updated as index field sizes are computed.
    *          If <code>null</code> dialect validation is disabled because it does not have
    *          index size constraint.
    * @param   property
    *          The property to be checked. If not <code>null</code>, the future value for it can
    *          be found in <code>newValue</code> instead of buffer's DMO. If <code>null</code>
    *          then all the fields/properties are already set in the buffer's record.
    * @param   newValue
    *          The new value for property. If <code>property</code> is <code>null</code>, the
    *          value is ignored.
    * @param   uniqueIndexes
    *          Flag for selecting the correct map of indexes. This method is called twice, once
    *          for unique indexes and once for non-unique, with according parameter.
    *
    * @throws  ValidationException
    *          if any index affected by this record will pass the maximum supported size
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  StopConditionException
    *          if the size of a index key is found greater than allowable limits.
    */
   private void checkIndexSizeWorker(RecordBuffer buffer,
                                     Map<String, Integer> p4glFieldSizeCache,
                                     Map<String, Integer> dialectFieldSizeCache,
                                     Collection<String> dirtyIndexes,
                                     String property,
                                     BaseDataType newValue,
                                     boolean uniqueIndexes)
   throws ValidationException,
          PersistenceException
   {
      // cache the appropriate map of indexes
      final Map<String, List<String>> indexMap =
            buffer.getIndexHelper().getIndexMapForEntity(buffer.getEntityName(), uniqueIndexes);
      
      P2JDialect dialect = buffer.getDialect();
      for (String indexName : dirtyIndexes)
      {
         /* In the case of non-unique indexes, the synthetic [id] field is forcefully added to
            index so the indexed records are always unique */
         int dialectIndexSize = dialect.getSyntheticIndexSize(uniqueIndexes);
         
         int p4glIndexSize = 0;
         List<String> idxComponents = indexMap.get(indexName);
         for (String field : idxComponents)
         {
            // cache this value for a second call, the reflexion is expensive
            BaseDataType value = null;
            
            if (dialectFieldSizeCache != null)
            {
               if (dialectFieldSizeCache.containsKey(field))
               {
                  // already computed, add it to this index-key total length
                  dialectIndexSize += dialectFieldSizeCache.get(field);
               }
               else
               {
                  // If the field is the one to be changed use the new value to compute the size
                  // of the index key, otherwise use the current value from target DMO.
                  // field.equals(property) will hold only when checking unset properties whose
                  // new value is provided, otherwise use the value already set
                  value = field.equals(property) ? newValue : buffer.getValue(field);
                  
                  int dialectFieldSize = dialect.computeIndexKeySize(value);
                  dialectFieldSizeCache.put(field, dialectFieldSize); // add size to cache
                  dialectIndexSize += dialectFieldSize;
               }
            }
            
            if (p4glFieldSizeCache != null)
            {
               if (p4glFieldSizeCache.containsKey(field))
               {
                  // already computed, add it to this index-key total length
                  p4glIndexSize +=  p4glFieldSizeCache.get(field);
               }
               else
               {
                  if (value == null)
                  {
                     // if not already extracted when computing dialectFieldSize, do it now
                     value = field.equals(property) ? newValue : buffer.getValue(field);
                  }
                  int p4glFieldSize = computeP4GLFieldSize(value);
                  p4glFieldSizeCache.put(field, p4glFieldSize); // add computed size to cache
                  p4glIndexSize += p4glFieldSize;
               }
            }
         }
         
         boolean p4glIndexFail = 
            (p4glFieldSizeCache != null) && p4glIndexSize > p4glMaxIndexSize;
         boolean dialectIndexFail =
            (dialectFieldSizeCache != null) && dialectIndexSize > dialect.getIndexLengthLimit();
         
         if (!p4glIndexFail && !dialectIndexFail) 
         {
            // both index check passed, loop to next index
            continue;
         }
         
         // remove optional "idx__" prefix
         String legacyIndexName = 
               TableMapper.getLegacyIndexName((Buffer) buffer.getDMOProxy(), 
                                              DBUtils.extractIndexName(indexName));
         if (legacyIndexName == null)
         {
            // legacy index name could not be detected, let the user know which sql index it was 
            legacyIndexName = indexName; 
         }
         
         String schemaTableName = buffer.isTemporary() ?
               "(TEMP-TABLE)" + buffer.getTable() : 
               buffer.getSchema() + "." + buffer.getTable();
         
         if (p4glIndexFail)
         {
            // the strange thing is, when displayed in 4GL, there are two error codes 129/11353
            // however, there is no DOT at the end of the message!
            final String message =
                  "The total length of the fields in an index exceeds max key size. " +
                  "Index " + legacyIndexName + " of table " + schemaTableName + " (129)";
            
            ErrorManager.displayError(11353, message, false);
            // the message is sometimes displayed multiple times with temp-tables
            throw new StopConditionException(message, 11353);
         }
         
         if (dialectIndexFail) // always true
         {
            // this record is stored successfully on P4GL but the current dialect cannot handle
            // it so a SQL error will be emitted if we continue with the flush to database
            final String message = "The total length of the fields in an index exceeds " +
                  "maximum key size supported by this dialect. " +
                  "Index " + legacyIndexName + " of table " + schemaTableName;
            ErrorManager.displayError(-1, message, false);
            throw new StopConditionException(message, -1);
         }
      }
   }
   
   /**
    * Computes the size (in bytes) of a field value occupied in an index-key in P4GL. Apparently 
    * all datatypes use variable length here so individual computations are needed for each 
    * datatype.
    * 
    * The values are obtained empirically, using ballast technique and altering the value for 
    * each datatype checked.
    * 
    * @param   value
    *          The value of the field to be computed.
    * 
    * @return  the size in bytes taken by the value when stored in a index-key in P4GL.
    */
   private int computeP4GLFieldSize(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         return 3; // apparently all unknown values occupy 3 bytes as index-keys
      }
      
      if (value instanceof Text) 
      {
         // TODO: investigate multibyte texts
         final int len = ((Text) value).getValue().length();
         return (len == 0) ? 2 : 1 + len;
      }
      else if (value instanceof logical)
      {
         // both logical values take 3 bytes
         return 3;
      }
      else if (value instanceof int64)
      {
         long val = ((int64) value).toJavaLongType();
         if (-16 <= val && val <= 15)
         {  // FFFF FF90 ... 0000 000F 
            return 2;
         }
         else if (-1022 <= val && val <= 1022)
         {  // FFFF FC02 ... 0000 03FE
            return 3;
         }
         else if (-65022 <= val && val <= 65278)
         {  // FFFF 0202 ... 0000 FEFE
            //                       65279 (FEFF) uses 5
            // however, I found out that 123456 uses only 4 bytes !
            // 1048318 (0F FEFE) uses 4, BUT 1048319 (0F FEFF) uses 5
            return 4;
         }
         else if (-15999998 <= val && val <= 16711422)
         {  // FF0B DC02 ... 00FE FEFE
            return 5;
         }
         else if (-2130574846 <= val && val <= 2130640638)
         {  // 8102 0202 ... 7EFE FEFE
            return 6;
         }
         
         // other 4 exceptions, detected empirically:
         if (val == 2147483646 || val == -2147483646)
         {
            return 8;
         }
         else if (val == 2147483647 || val == -2147483648 )
         {
            return 9;
         }
         
         if (value instanceof integer)
         {
            // apparently this is not the maximum for 4-byte integer datatype 
            return 7;
         }
         
         // keep processing until all the 8-bytes int64 are covered
         if (-545427160574L <= val && val <= 1095199817470L)
         {  // FFFF FF81 0202 0202 ... 0000 00FE FEFE FEFE
            return 7;
         }
         else if (-139629353106942L <= val && val <= 280371153272574L)
         {  // FFFF 8102 0202 0202 ... 0000 FEFE FEFE FEFE
            return 8;
         }
         else if (-35745114395377150L <= val && val <= 71775015237779198L)
         {  // FF81 0202 0202 0202 ... 00FE FEFE FEFE FEFE
            return 9;
         }
         else if (-9150749285216550398L <= val && val <= 9151031864016699134L)
         {  // 8102 0202 0202 0202 ... 7EFE FEFE FEFE FEFE
            return 10;
         }
         
         // larger index-keys use 11 bytes
         return 11;
      }
      else if (value instanceof decimal)
      {
         return 8; // TODO: replace with actual bytes used 2, 3... 13
      }
      else if (value instanceof datetimetz)
      {
         return 12; // TODO: replace with actual bytes used 2, 3... 13
      }
      else if (value instanceof datetime)
      {
/*
         datetime(today, 0)                  = 12 (08/05/2014 00:00:00.000)
         datetime(today, 16711422)           =  9 (08/05/2014 04:38:31.422)
         now                                 =  8 (08/05/2014 17:51:48.752)
         
         datetime(date(2433404), 16711422)   =  7 (01/05/1950 04:38:31.422)

         datetime(date(2433405), 2130640638) =  7 (26/05/1950 15:50:40.638)
         
         datetime(date(2433406), 0)          = 10 (03/05/1950 00:00:00.000)
         datetime(date(2433405), 86400000)   = 10 (03/05/1950 00:00:00.000)
         datetime(date(2433405), 86399999)   =  6 (02/05/1950 23:59:59.999)
         datetime(date(2433405), 65278)      =  4 (02/05/1950 00:01:05.278)
         datetime(date(2433405), 1022)       =  3 (02/05/1950 00:00:01.022)
         datetime(date(2433405), 0-15)       =  2 (02/05/1950 00:00:00.000 - 015)
         
         datetime(date(2433404), 186400010)  =  7 (03/05/1950 03:46:40.010)
*/
         return 8;
      }
      else if (value instanceof date)
      {
         int jDay = ((date)value).intValue();
         // May 2, 1950 = 2433405 -- this is some kind of epoch for p4gl index-keys
         // recursive call: 
         return computeP4GLFieldSize(new integer(jDay - 2433405));
      }
      
      return -1;
   }
   
   /**
    * Validate a single unique constraint or all unique constraints defined
    * in the validation parameters for the target record.  This method in most
    * cases must access the database to complete the constraint check.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if the constrained properties contains values which are not
    *          unique to this record.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkUniqueAndShare(RecordBuffer buffer,
                                    Persistable target,
                                    ValidationData vData,
                                    String property,
                                    BaseDataType value)
   throws ValidationException,
          PersistenceException,
          IllegalAccessException,
          InvocationTargetException
   {
      // indexes affected by the changes that are being validated
      Collection<String> dirtyUnique = vData.getDirtyUnique();
      Collection<String> dirtyNonUnique = vData.getDirtyNonUnique();
      Collection<String> cumulativeDirtyNonUnique = vData.getCumulativeDirtyNonUnique();
      
      // determine whether any index has been updated; when a record is transient (i.e., not yet
      // flushed to the database), some properties may be dirty/changed, but unless ALL properties
      // in an index are dirty, that index will not have been updated yet
      boolean indexUpdated = !dirtyUnique.isEmpty() || !dirtyNonUnique.isEmpty();
      
      // determine if dirty sharing is necessary, which generally is true if the buffer is backed
      // by a persistent* table and any index has been updated; however, we need to skip sharing
            //TODO: OM: *: or shared temp-table
      // if we've already shared unvalidated updates to a transient record (in which case the
      // ValidationData.deferred flag will be true)
      // NOTE:  currently, it is assumed that ValidationData.isShareDirty() will only return true
      // for the buffer's current record, because the share logic below operates on the buffer's
      // current record and current dirty property state
      boolean share = vData.isShareDirty() && indexUpdated;
      
      // do we need to validate?
      boolean validate = true;
      
      if (!vData.isTransient())
      {
         if (dirtyUnique.isEmpty() && (dirtyNonUnique.isEmpty() || !vData.isShareDirty()))
         {
            return; // nothing more to do
         }
      }
      else // transient record
      {
         Map<String, Object> dirtyIndexedProps = buffer.getDirtyProperties(true);
         if (dirtyUnique.isEmpty()              && 
             cumulativeDirtyNonUnique.isEmpty() && 
             !dirtyIndexedProps.containsKey(property))
         {
            return; // changed fields didn't ever altered the indexes (fully)
         }
         
         if (indexUpdated) // at least one of [dirtyUnique] and [dirtyNonUnique] is NOT empty
         {
            // some index is updating now, validate and share
            validate = true;
            share = vData.isShareDirty();
         } 
         else
         {
            // [dirtyUnique] and [dirtyNonUnique] are both empty so no full indexed were matched.
            // last updated fields are not part of an index or index is not fully matched
            // defer validation until a full index is dirty or the record needs to be flushed
            validate = false;
            
            // if a non-unique index was previously matched we need to share inside context
            // regardless of the field being validated
            share = !cumulativeDirtyNonUnique.isEmpty();
         }
      }
      
      Persistence persistence = buffer.getPersistence();
      Method omit = (property != null ? triggers.get(property) : null);
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      String entity = buffer.getEntityName();
      
      try
      {
         // lock all indexes while sharing (and validating, if needed)
         // TODO:  optimization:  lock only dirty indexes unless we're
         // inserting a new record.  As part of this optimization, we need to
         // consider the remote server case, such that we minimize the data
         // sent across (an array or collections of index name strings may not
         // be efficient).
         dirtyContext.lockAllIndexes(entity, LockType.EXCLUSIVE);
         
         if (validate)
         {
            // Validate each dirty, unique index.
            for (String index : dirtyUnique)
            {
               HQLBundle bundle = uniqueConstraints.get(index);
               if (bundle != null)
               {
                  checkUnique(buffer, target, vData, index, bundle, omit, property, value);
               }
            }
         }
         
         // share dirty updates if required
         if (share)
         {
            // store the update into the dirty database, either as a new insert, or as an update
            // to an existing record
            if (vData.isInsertRequired())
            {
               Persistable dmo = buffer.getCurrentRecord();
               Method setter = (property != null ? setterMap.get(property) : null);
               dirtyContext.insert(buffer, dmo, setter, value, null, false);
            }
            else
            {
               boolean propIndexed = (property != null && buffer.isPropertyIndexed(property));
               Map<String, Object> dirtyProperties = buffer.getDirtyProperties(false);
               
               int size = dirtyProperties.size();
               if (size > 0 || propIndexed)
               {
                  int start = 0;
                  if (propIndexed && !dirtyProperties.containsKey(property))
                  {
                     size++;
                     start++;
                  }
                  
                  // recalculate the size, taking into considerations the arrays
                  int newSize = start;
                  Iterator<Map.Entry<String, Object>> dirtyIter =
                     dirtyProperties.entrySet().iterator();
                  while (dirtyIter.hasNext())
                  {
                     Object change = dirtyIter.next().getValue();
                     if (change instanceof Object[])
                     {
                        newSize++;
                     }
                     else if (change instanceof Map)
                     {
                        Map<Integer, Object> extent = (Map<Integer, Object>) change;
                        newSize += extent.size();
                     }
                     // else: bad dirtyProperties
                  }
                  assert(newSize >= size);
                  
                  String[] dup = new String[newSize]; // property name
                  int[]    dui = new int[newSize];    // extent offset (-1 for scalars)
                  BaseDataType[] duv = new BaseDataType[newSize];
                  if (start > 0)
                  {
                     dup[0] = property;
                     duv[0] = value;
                     dui[0] = -1;
                  }
                  
                  dirtyIter = dirtyProperties.entrySet().iterator();
                  int i = start;
                  for (; dirtyIter.hasNext(); i++)
                  {
                     Map.Entry<String, Object> entry = dirtyIter.next();
                     Object change = entry.getValue();
                     if (change instanceof Object[])
                     {
                        Object[] scalarChange = (Object[]) change;
                        dup[i] = entry.getKey();
                        dui[i] = -1;
                        duv[i] = (BaseDataType) scalarChange[1];
                     }
                     else if (change instanceof Map)
                     {
                        Map<Integer, Object> extent = (Map<Integer, Object>) change;
                        for (Integer idx : extent.keySet())
                        {
                           dup[i] = entry.getKey();
                           dui[i] = idx;
                           Object[] scalarChange = (Object[]) extent.get(idx);
                           duv[i] = (BaseDataType) scalarChange[1];
                           i++;
                        }
                        i--; // decrement back the last one
                     }
                  }
                  assert(i == newSize);
                  
                  dirtyContext.update(buffer,
                                      dup,
                                      dui,
                                      duv,
                                      dirtyUnique,
                                      validate && vData.isTransient());
               }
            }
         }
      }
      finally
      {
         // Unlock unique indexes.  See TODO above.
         dirtyContext.lockAllIndexes(entity, LockType.NONE);
         
         if (!validate)
         {
            // indicate to caller that validation was deferred, but change was shared
            vData.setDeferred();
         }
      }
   }
   
   /**
    * Validate a single unique constraint defined for the target record.
    * This method in most cases must access the database to complete the
    * constraint check.  However, in cases where one of the constrained
    * values represents the unknown value, the validation completes
    * successfully.  In this case it is not necessary to access the database,
    * since the unknown value is always considered unique.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   index
    *          Name of unique index being validated.
    * @param   bundle
    *          An object containing the HQL query statement and all getter
    *          methods necessary to gather substitution parameter values from
    *          <code>target</code>.
    * @param   omit
    *          Getter method associated with a property for which we already
    *          have a value.  This method will <i>not</i> be used to retrieve
    *          a value from <code>target</code> via reflection.  Instead,
    *          the <code>value</code> parameter is used.  If
    *          <code>null</code>, all constrained properties are inspected
    *          for current values via reflection.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if the constrained properties contains values which are not
    *          unique to this record.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkUnique(RecordBuffer buffer,
                            Persistable target,
                            ValidationData vData,
                            String index,
                            HQLBundle bundle,
                            Method omit,
                            String property,
                            BaseDataType value)
   throws ValidationException,
          PersistenceException,
          IllegalAccessException,
          InvocationTargetException
   {
      // An unknown value short-circuits the uniqueness check, since unknown
      // value is by definition unique.
      if (value != null && value.isUnknown())
      {
         return;
      }
      
      // If there are no records in the backing table, it is impossible for
      // the target record to violate a unique constraint, so we can short-
      // circuit the check.  This potentially preempts a lot of reflection
      // calls and a database query.
      if (buffer.isTableDefinitelyEmpty())
      {
         return;
      }
      
      Persistence persistence = buffer.getPersistence();
      
      // Gather substitution parameters.
      Object[] args = null;
      
      // First substitution parameter for a temp table is the multiplex ID.
      boolean temporary = buffer.isTemporary();
      if (temporary)
      {
         args = new Object[bundle.gettersSize() + 1];
         args[0] = buffer.getMultiplexID();
      }
      else
      {
         args = new Object[bundle.gettersSize()];
      }
      
      // Loop through properties in the unique constraint and populate the
      // parameter array.
      boolean forceValid = false;
      Iterator<Method> iter = bundle.getters();
      for (int i = (temporary ? 1 : 0); iter.hasNext(); i++)
      {
         Method getter = iter.next();
         BaseDataType bdt = null;
         
         if (getter.equals(omit))
         {
            // Updated value is not in the DMO yet, so we need to use the
            // value provided to this method.
            bdt = value;
         }
         else
         {
            // Get current value of constraint component in DMO.
            bdt = (BaseDataType) getter.invoke(target);
         }
         
         // An unknown value short-circuits the check, since unknown value
         // is by definition unique.
         if (bdt.isUnknown())
         {
            forceValid = true;
            break;
         }
         
         if (charProps != null && bdt instanceof character)
         {
            // Possibly uppercase text parameter before passing it to query.
            character c = (character) bdt;
            Boolean ignoreCase = charProps.get(getter);
            if (ignoreCase != null && ignoreCase)
            {
               c = TextOps.toUpperCase(c);
            }
            args[i] = c;
         }
         else
         {
            args[i] = bdt;
         }
      }
      
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      String entity = buffer.getEntityName();
      
      if (!forceValid)
      {
         Type[] types = DBUtils.makeTypeArray(args);
         
         // validate against primary data; avoid causing a flush before the query, as this would
         // flush the DMO changes we are checking, which could potentially cause a JDBC exception
         String hql = bundle.getStatement(0, false);
         String[] entities = new String[] { entity };
         List results = persistence.list(entities, hql, args, types, 1, 0, true, false);
         Serializable candidateID = target.getId();
         if (results != null)
         {
            // a record was found; make sure it's not the current record, and that a modified
            // version does not invalidate the violation.
            Persistable found = (Persistable) results.get(0);
            Serializable foundID = found.getId();
            
            if (!candidateID.equals(foundID))
            {
               // a candidate result was found in the primary database which may violate a
               // unique constraint; confirm this finding is not superseded by some unflushed
               // change in our session
               DirtyInfo info = dirtyContext.getDirtyInfo(buffer,
                                                          index,
                                                          QueryConstants.UNIQUE,
                                                          bundle,
                                                          args,
                                                          types,
                                                          found,
                                                          false);
               
               if (info == null)
               {
                  uniqueConstraintViolation(buffer, bundle, args);
               }
               
//               if (info.isLocal())
               {
                  // if a violating record managed by the local session was found in the dirty
                  // database, raise an exception
                  if (info.getDirtyDMO() != null)
                  {
                     uniqueConstraintViolation(buffer, bundle, args);
                  }
                  
                  // if the record we found in the primary database was not deleted subsequently
                  // by the local session, raise an exception
                  if (!info.isDeleted())
                  {
                     uniqueConstraintViolation(buffer, bundle, args);
                  }
               }
            }
         }
         else
         {
            // no conflict was found in primary database, so validate against uncommitted data
            hql = bundle.getStatement(0, true);
            results = dirtyContext.list(entity, hql, args, types, 1, 0, true);
            if (results != null)
            {
               // a record was found; make sure it's not the current record, and that it is not
               // an unvalidated record from another context
               Persistable found = (Persistable) results.get(0);
               Serializable foundID = found.getId();
               if (!candidateID.equals(foundID))
               {
                  // don't fail due to an unvalidated record found in a foreign context; this
                  // can happen due to our implementation of dirty sharing, which may show us
                  // too much in this case
                  RecordIdentifier ident = new RecordIdentifier(buffer.getEntityName(), foundID);
                  if (!dirtyContext.isForeignUnvalidatedIndex(ident, index))
                  {
                     uniqueConstraintViolation(buffer, bundle, args);
                  }
               }
            }
         }
      }
   }
   
   /**
    * Throw a validation exception for a unique constraint violation.  An
    * error message is composed describing the constraint properties and
    * their current values.
    *
    * @param   buffer
    *          Record buffer representing the DMO whose constraint has been
    *          violated.
    * @param   bundle
    *          Object from  which property names will be extracted.
    * @param   args
    *          Current values of the properties which participate in the
    *          unique constraint.
    *
    * @throws  ValidationException
    *          always.
    */
   private void uniqueConstraintViolation(RecordBuffer buffer, HQLBundle bundle, Object[] args)
   throws ValidationException
   {
      BufferImpl dmo = (BufferImpl) buffer.getDMOProxy();
      
      // for dynamic tables table name is used rather than buffer name
      String legacyName = buffer.isDynamic() ? TableMapper.getLegacyName(dmo)
                                             : dmo.name().toStringMessage();
      
      StringBuilder buf = new StringBuilder();
      buf.append(legacyName);
      buf.append(" already exists with");
      
      // Skip the multiplex id for the temporary buffer
      int i = (buffer.isTemporary() ? 1 : 0);
      for (Iterator<Method> iter = bundle.getters(); iter.hasNext(); i++)
      {
         Method getter = iter.next();
         String propName = propertyNames.get(getter);
         if ("id".equals(propName))
         {
            continue;
         }
         buf.append(" ");
         
         // for temporary buffers field name is not displayed
         String propLegacyName = buffer.isTemporary() ? "" :
            TableMapper.getLegacyFieldName(buffer.getDMOInterface(), propName);
         
         buf.append(propLegacyName);
         buf.append(" ");
         buf.append(((BaseDataType) args[i]).toStringMessage());
      }
      
      ValidationException exc = new ValidationException(buf.toString(), 132);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "Unique constraint violation", exc);
         LOG.log(Level.FINE, buffer.toString());
      }
      
      throw exc;
   }
   
   /**
    * Construct and populate an <code>HQLBundle</code> suitable for
    * producing a query to test a unique constraint.  The bundle contains a
    * single HQL statement.  The statement's where clause contains an
    * equality match for each property contained in <code>names</code>.
    * It selects the unique record ID from the DMO currently associated with
    * <code>buffer</code>.  For example:
    * <pre>
    *   select customer.id from CustomerImpl as customer
    *   where customer.prop1 = ? and customer.prop2 = ?
    * </pre>
    * <p>
    * A getter method is added to the bundle for the property associated with
    * each equality match in the where clause.  The bundle is stored in the
    * map of unique constraints, keyed by the setter method associated with
    * each property component of the unique constraint.
    * 
    * @param   index
    *          Name of the unique index which defines the constraint to be
    *          validated.
    * @param   properties
    *          List of property names which comprise a unique constraint for
    *          table represented by the given record buffer.
    * @param   getters
    *          Map of getter methods by property names for the DMO interface
    *          associated with the given record buffer.
    * @param   buffer
    *          Record buffer for which this HQLBundle is to be generated.
    * 
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding a property.
    */
   private void buildUniqueBundle(String index,
                                  List<String> properties,
                                  Map<String, Method> getters,
                                  RecordBuffer buffer)
   throws PersistenceException
   {
      HQLBundle bundle = new HQLBundle();
      
      String alias = buffer.getDMOAlias();
      HQLExpression buf = new HQLExpression();
      /*
      buf.append("select ");
      buf.append(alias);
      buf.append(".id from ");
      */
      buf.append("from ");
      buf.append(buffer.getDMOImplementationName());
      buf.append(" as ");
      buf.append(alias);
      
      boolean temporary = buffer.isTemporary();
      int start = 0;
      if (temporary)
      {
         buf.append(" where ");
         buf.append(alias);
         buf.append(".");
         buf.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         buf.append(" = ", true);
         start++;
      }
      
      HQLExpression bufDirty = (temporary ? null : new HQLExpression(buf));
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      Iterator<String> iter = properties.iterator();
      for (int i = start; iter.hasNext(); i++)
      {
         String property = iter.next();
         
         buf.append(i == 0 ? " where " : " and ");
         StringBuilder propName = new StringBuilder();
         DBUtils.composePropertyName(buffer, property, propName, Database.Type.PRIMARY);
         buf.append(propName.toString());
         buf.append(" = ", true);
         
         if (!temporary)
         {
            bufDirty.append(i == 0 ? " where " : " and ");
            propName = new StringBuilder();
            DBUtils.composePropertyName(buffer, property, propName, Database.Type.DIRTY);
            bufDirty.append(propName.toString());
            bufDirty.append(" = ", true);
         }
         
         Method getter = getters.get(property);
         
         bundle.addGetter(getter);
         
         Boolean ignoreCase = DatabaseManager.getIgnoreCase(dmoClass, property);
         if (ignoreCase != null)
         {
            if (charProps == null)
            {
               charProps = new HashMap<Method, Boolean>(3);
            }
            charProps.put(getter, ignoreCase);
         }
         
         triggers.put(property, getter);
      }
      
      String primaryHQL = buf.toFinalExpression();
      bundle.addStatement(primaryHQL, false);
      
      String dirtyHQL = (temporary ? primaryHQL : bufDirty.toFinalExpression());
      bundle.addStatement(dirtyHQL, true);
      
      uniqueConstraints.put(index, bundle);
      
      if (LOG.isLoggable(Level.FINER))
      {
         StringBuilder sb = new StringBuilder("Unique validation bundle:\n");
         sb.append(bundle);
         LOG.log(Level.FINER, sb.toString());
      }
   }
   
   /**
    * Extract current value of a DMO property, given the DMO, the getter
    * method for the property, and an optional index argument to the getter
    * method.  Uses reflection to obtain the property value.
    *
    * @param   target
    *          Target DMO record.
    * @param   getter
    *          Getter method for the property.
    * @param   index
    *          Index value to pass to the getter (for indexed getters only),
    *          or <code>null</code> if getter takes no arguments.
    *
    * @return  Current value of the property.
    *
    * @throws  IllegalAccessException
    *          if getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private Object getPropertyValue(Object target, Method getter, Object index)
   throws IllegalAccessException,
          InvocationTargetException
   {
      Object value = null;
      if (index != null)
      {
         Object[] args = { index };
         value = getter.invoke(target, args);
      }
      else
      {
         value = getter.invoke(target);
      }
      
      return value;
   }
}