/*
** Module   : DirtyDmoStrategy.java
** Abstract : Custom entity dirtiness strategy for Hibernate
**
** Copyright (c) 2016-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20160615 Created initial version. Hibernate dirty checking optimization.
** 002 ECF 20160912 Signature change in ChangeBroker$SessionInterceptor.getDirtyEvent. Removed
**                  implementation of resetDirty.
** 003 ECF 20180709 Ensure unused entity is evicted from session post-flush, if found non-dirty.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import org.hibernate.*;
import org.hibernate.internal.*;
import org.hibernate.persister.entity.*;
import com.goldencode.p2j.persist.event.*;

/**
 * An implementation of a custom entity dirtiness strategy for Hibernate. The default dirty
 * checking strategy which Hibernate uses is terribly slow. Since we track changes to data model
 * objects as a matter of course already, this class was introduced to short-circuit Hibernate's
 * dirty checking cycle to the degree possible.
 */
final class DirtyDmoStrategy
implements CustomEntityDirtinessStrategy,
           Serializable
{
   /**
    * Indicate whether we can provide custom dirty checking for the given entity.
    * 
    * @param   entity
    *          The DMO to be checked.
    * @param   persister
    *          Not used.
    * @param   session
    *          Not used.
    * 
    * @return  {@code true} always.
    */
   @Override
   public boolean canDirtyCheck(Object entity, EntityPersister persister, Session session)
   {
      // check all types
      return true;
   }
   
   /**
    * Indicate whether we consider the given entity to require synchronization with the database.
    * If we have tracked changes for the particular DMO explicitly, or for the DMO's type in
    * general, indicate that the entity is dirty. Even if it is not in the latter case, Hibernate
    * will determine for sure on a {@code true} return from this method.
    * 
    * @param   entity
    *          The DMO to be checked.
    * @param   persister
    *          Not used.
    * @param   session
    *          Current Hibernate session. This provides a rather tenuous link back to our custom
    *          {@code Interceptor} implementation, in that it relies upon casting to the
    *          Hibernate internal {@code SessionImpl} interface.
    * 
    * @return  {@code true} if the DMO is either definitely dirty or if it may be dirty (based on
    *          our knowledge that a non-specific change (e.g., a bulk delete) has been made to
    *          the backing table. If neither is the case, return {@code false}.
    */
   @Override
   public boolean isDirty(Object entity, EntityPersister persister, Session session)
   {
      ChangeBroker.SessionInterceptor interceptor = getInterceptor(session);
      Persistable dmo = (Persistable) entity;
      String entityName = dmo.getClass().getName();
      RecordIdentifier ident = new RecordIdentifier(entityName, dmo.getId());
      
      RecordChangeEvent event = interceptor.getLastDirtyEvent(ident);
      
      boolean dirty = false;
      
      if (event != null)
      {
         // if the record has been deleted, it can't be dirty and we should not let Hibernate try
         // to update it; otherwise, report a dirty record
         dirty = (!event.isDelete() && !event.isFullBulkDelete());
      }
      else
      {
         // some events have no specific DMO attached, but may affect the current entity; if so,
         // Hibernate will figure it out
         dirty = interceptor.isFlushPending(entityName);
      }
      
      // if we found the entity to be not dirty, notify the interceptor that this record can be
      // evicted from the session if not in use; if we can't determine here whether the entity
      // is dirty or not, the interceptor will queue the entity for eviction in findDirty()
      if (!dirty)
      {
         interceptor.maybeEvictAfterFlush(dmo);
      }
      
      return dirty;
   }
   
   /**
    * Find the properties of the DMO which are dirty.
    * <p>
    * This is currently a no-op, but may be implemented in the future.
    * 
    * @param   entity
    *          The DMO to be checked.
    * @param   persister
    *          Not used.
    * @param   session
    *          Not used.
    * @param   dirtyCheckContext
    *          Not used.
    */
   @Override
   public void findDirty(Object entity,
                         EntityPersister persister,
                         Session session,
                         DirtyCheckContext dirtyCheckContext)
   {
      // no-op; let Hibernate figure it out
      
      // possible future implementation
      /*
      ChangeBroker.SessionInterceptor interceptor = getInterceptor(session);
      RecordChangeEvent event = interceptor.getDirtyEvent(((Persistable) entity).getId());
      dirtyCheckContext.doDirtyChecking(new AttributeCheckerImpl(event));
      */
   }
   
   /**
    * Called by Hibernate when dirty checking of the specified DMO is complete. We perform
    * cleanup after a flush in the session interceptor, so this is a no-op.
    * 
    * @param   entity
    *          The DMO which was checked.
    * @param   persister
    *          Not used.
    * @param   session
    *          Current Hibernate session. This provides a rather tenuous link back to our custom
    *          {@code Interceptor} implementation, in that it relies upon casting to the
    *          Hibernate internal {@code SessionImpl} interface.
    */
   @Override
   public void resetDirty(Object entity, EntityPersister persister, Session session)
   {
   }
   
   /**
    * Convenience method to extract our custom interceptor from the Hibernate session.
    * 
    * @param   session
    *          Current Hibernate session. This provides a rather tenuous link back to our custom
    *          {@code Interceptor} implementation, in that it relies upon casting to the
    *          Hibernate internal {@code SessionImpl} interface.
    * 
    * @return  Interceptor.
    */
   private ChangeBroker.SessionInterceptor getInterceptor(Session session)
   {
      return (ChangeBroker.SessionInterceptor) ((SessionImpl) session).getInterceptor();
   }
   
   // Possible future implementation.
   /*
   private class AttributeCheckerImpl
   implements AttributeChecker
   {
      private final Set<String> properties;
      
      AttributeCheckerImpl(RecordChangeEvent event)
      {
         this.properties = event.getProperties().keySet();
      }
      
      @Override
      public boolean isDirty(AttributeInformation attributeInformation)
      {
         String name = attributeInformation.getName();
         
         // process event to look up name
         // TODO: deal with extent fields properly; we use root names, Hibernate uses inner
         //       composite class names
         // TODO: what about changes that don't have a specific event?
         return properties.contains(name);
      }
   }
   */
}