/*
** Module   : P2JField.java
** Abstract : Container for DMO field definitions.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20130624 Created initial version.
** 002 OM  20140411 Added caseSensitive support.
** 003 ECF 20140914 Added hashCode and equals implementations.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.convert.ParmType;
import com.goldencode.p2j.util.*;

/**
 * Container for DMO field definitions.
 */
public class P2JField
{
   /** Field name. */
   private String name;

   /** Field type. */
   private ParmType type;

   /** Extent size. */
   private long extent;

   /** Field format. */
   private String format;

   /** Initial field value. */
   private BaseDataType initial;

   /** Field label. */
   private String label;

   /** Column label. */
   private String columnLabel;
   
   /** Flag for case-sensitive character fields. Ignored for all other datatypes. */
   private boolean caseSensitive;
   
   /**
    * Convenience constructor.
    * @param name
    *        Field name.
    * @param type
    *        Field type.
    * @param extent
    *        Extent size.
    * @param format
    *        Field format.
    * @param initial
    *        Initial value.
    * @param label
    *        Field label.
    * @param columnLabel
    *        Column label.
    * @param caseSensitive
    *        <code>true</code> for case sensitive character fields.
    */
   public P2JField(String name,
                   ParmType type,
                   long extent,
                   String format,
                   BaseDataType initial,
                   String label,
                   String columnLabel,
                   boolean caseSensitive)
   {
      this.name = name;
      this.type = type;
      this.extent = extent;
      this.format = format;
      this.initial = initial;
      this.label = label;
      this.columnLabel = columnLabel;
      this.caseSensitive = caseSensitive;
   }

   /**
    * Get field name.
    *
    * @return  field name.
    */
   public String getName()
   {
      return name;
   }

   /**
    * Get field type.
    *
    * @return  field type.
    */
   public ParmType getType()
   {
      return type;
   }

   /**
    * Get extent size.
    *
    * @return  extent size.
    */
   public long getExtent()
   {
      return extent;
   }

   /**
    * Get field format.
    *
    * @return  field format.
    */
   public String getFormat()
   {
      return format;
   }

   /**
    * Get initial value.
    *
    * @return  initial value.
    */
   public BaseDataType getInitial()
   {
      return initial;
   }

   /**
    * Get field label.
    *
    * @return  field label.
    */
   public String getLabel()
   {
      return label;
   }

   /**
    * Get column label.
    *
    * @return  column label.
    */
   public String getColumnLabel()
   {
      return columnLabel;
   }
   
   /**
    * Check if this is a case sensitive field. It only make sense for character datatypes.
    *
    * @return  <code>true</code> if this is a case-sensitive character field.
    */
   public boolean isCaseSensitive()
   {
      return caseSensitive;
   }
   
   /**
    * Produce a hash code from the internal data of this object, which is consistent with the
    * implementation of {@link #equals(Object)}.
    * 
    * @return  Hash code.
    */
   @Override
   public int hashCode()
   {
      int result = 17;
      result = 37 * result + name.hashCode();
      result = 37 * result + type.hashCode();
      result = 37 * result + (int) extent;
      result = 37 * result + (format != null ? format.hashCode() : 0);
      result = 37 * result + (initial != null ? initial.hashCode() : 0);
      result = 37 * result + (label != null ? label.hashCode() : 0);
      result = 37 * result + (columnLabel != null ? columnLabel.hashCode() : 0);
      result = 37 * result + (caseSensitive ? 1 : 0);
      
      return result;
   }
   
   /**
    * Test equivalence of this object and the given object in a way that is consistent with the
    * implementation of {@link #hashCode()}.
    * 
    * @return  <code>true</code> if this object is equivalent with <code>o</code>, else
    *          <code>false</code>.
    */
   @Override
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }
      
      if (o instanceof P2JField)
      {
         P2JField that = (P2JField) o;
         
         if (!this.name.equals(that.name) ||
             !this.type.equals(that.type) ||
             this.extent != that.extent   ||
             this.caseSensitive != that.caseSensitive)
         {
            return false;
         }
         
         if ((this.format != null && !this.format.equals(that.format)) ||
             (this.format == null && that.format != null))
         {
            return false;
         }
         
         if ((this.initial != null && !this.initial.equals(that.initial)) ||
             (this.initial == null && that.initial != null))
         {
            return false;
         }
         
         if ((this.label != null && !this.label.equals(that.label)) ||
             (this.label == null && that.label != null))
         {
            return false;
         }
         
         if ((this.columnLabel != null && !this.columnLabel.equals(that.columnLabel)) ||
             (this.columnLabel == null && that.columnLabel != null))
         {
            return false;
         }
         
         return true;
      }
      
      return false;
   }
}