/*
** Module   : Persistable.java
** Abstract : interface required for all persistable objects
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050901 ADD   @22477 Created initial version. Must be implemented
**                               by any data model object which must support
**                               row-level re-use of a single instance for
**                               multiple, subsequent record reads.
** 002 ECF 20050927 ADD   @22889 Added initialize method. This is invoked upon
**                               a DMO created with the default constructor.
**                               It instantiates internal lists and sets all
**                               direct attributes (i.e., those which do not
**                               represent associations with other DMOs) to
**                               their default values.
** 003 ECF 20050929 ADD   @22902 Added uniqueConstraints method. This is
**                               necessary to work around a limitation/defect
**                               in Hibernate, namely, that it does not
**                               guarantee that the order of data members
**                               participating in a unique constraint is
**                               maintained. This order is relevant for
**                               databases, such as PostgreSQL, where an index
**                               is used to implement a unique constraint, and
**                               that index concurrently is used to implement
**                               record selection.
** 004 ECF 20051015 DEL   @23255 Backed out #002 (@22889) and #003 (@22902).
**                               Changes to the DMO architecture and to the
**                               import process and runtime metadata support
**                               eliminated the need for these methods.
** 005 ECF 20080725 CHG   @39321 Added support for 64-bit integer ID. We now
**                               use Long instead of Integer for the ID.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * Defines the API necessary for the persistence framework to manage a Data
 * Model Object (DMO) instance. It must be implemented by all legacy DMO
 * implementation classes, in order to allow the Progress-like persistence
 * semantics to be provided to application code.
 * <p>
 * Although this interface must be public to allow implementation classes to
 * implement it, it is intended that these methods <b>not</b> be invoked
 * directly upon DMO implementation instances from outside the persistence
 * framework.
 */
public interface Persistable
extends Undoable
{
   /**
    * Get the database identifier for this persistable object. This generally
    * equates to the mapped table's primary key. However, it will be
    * <code>null</code> for DMO instances which have not yet been persisted
    * to the database (or more accurately, associated with a Hibernate
    * session).
    *
    * @return  Unique, database identifier associated with this object.
    */
   public Long getId();
   
   /**
    * Set the database identifier for this persistable object. This generally
    * should only be set by Hibernate.
    *
    * @param   id
    *          Unique, database identifier to be associated with this object.
    */
   public void setId(Long id);
}