/*
** Module   : TempTableResultSet.java
** Abstract : A result set used to serialize temp table data and send it to a remote side. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130610 Created initial version.
** 002 CA  20140110 Protected against NPEs, when used from Java side (for a OUTPUT parameter).
** 003 CA  20140124 Added NPE protection.
** 004 CA  20140214 Do not read _multiplex column; use the actual type instead of Hibernate type
**                  names when creating a new PropertyDefinition.
** 005 CA  20140626 Fixed usage from Java code: the default c'tor must not be used, always use
**                  the TempTableResultSet(boolean) c'tor with the argument set to false, as 
**                  from Java code instances of this class can be used only in OUTPUT mode.
** 006 SVL 20140709 Legacy property names are passed in the result set.
** 007 OM  20140801 Initialize iterators after setting the rows from appserver call.
** 008 CA  20140812 Added warnings to APIs which should not be used outside of P2J internal code.
** 009 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 010 ECF 20160225 Changes required by new PropertyHelper implementation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.sql.*;
import java.util.*;

import org.hibernate.mapping.*;
import org.hibernate.mapping.List;
import org.hibernate.property.PropertyAccessor;

/**
 * Implements a method of serializing the metadata and the full data of a temp-table, specified
 * by a {@link Temporary temp-table DMO}.  The data is read fully in memory and sent to the remote
 * side row by row; the DMO metadata is used to extract the properties and their settings (type
 * and extent).  Note that each send row will have the values in the same order as the DMO's
 * properties.
 * <p>
 * The {@link #TempTableResultSet(Temporary, boolean, boolean, boolean)} must not be used outside
 * of the P2J runtime, as it is dependent on the P2J's persistence frame to read the DMO's data.
 * <p>
 * From Java code, use the {@link #TempTableResultSet()} to instantiate a new, empty, result set;
 * this will be used as a container to transport back the result and the metadat.
 */
public class TempTableResultSet
extends TableResultSet
{
   /**
    * The DMO associated with the temp table; this will be used to extract the DMO metadata and 
    * the table data.
    */
   private Temporary dmo;
   
   /** Iterator over the DMO's properties. */
   private Iterator<PropertyDefinition> propIter;
   
   /** Iterator over the temp-table's data. */
   private Iterator<Object[]> rowIter;
   
   /**
    * Default c'tor, explicitly added to allow instances of this class to be created on 
    * deserialization. Not for public use.
    */
   public TempTableResultSet()
   {
   }
   
   /**
    * Create a new instance of this result set, when it acts as a container for an 
    * OUTPUT parameter received from the remote side.
    * 
    * @param    output
    *           Flag indicating this table is sent in OUTPUT mode.  Must always be set to
    *           <code>true</code>.
    * 
    * @throws   IllegalStateException
    *           If the <code>output</code> parameter is not set to <code>true</code>.
    */
   public TempTableResultSet(boolean output)
   {
      super(false, output, false);
      
      if (!output)
      {
         throw new IllegalStateException("This c'tor must be used only in OUTPUT mode!");
      }
   }

   /**
    * Create a new instance of this result set and associate it with the metadata and data for
    * the specified temp DMO.
    * 
    * @param    dmo
    *           DMO instance returned by a previous call to {@link TemporaryBuffer#define} on the 
    *           source temporary buffer.
    * @param    input
    *           Flag indicating this table is sent in INPUT or INPUT-OUTPUT mode.
    * @param    output
    *           Flag indicating this table is sent in OUTPUT or INPUT-OUTPUT mode.
    * @param    append
    *           Flag indicating this table is sent in APPEND mode.
    *           
    * @throws   ClassNotFoundException
    *           If the type of a property can not be resolved to a class.
    */
   public TempTableResultSet(Temporary dmo, boolean input, boolean output, boolean append)
   throws ClassNotFoundException
   {
      super(input, output, append);
      this.dmo = dmo;
      if (dmo != null)
      {
         init();
      }
   }

   /**
    * Set the data sent by a remote side, after the remote appserver call.
    *
    * @param    rows
    *           The data sent by the remote side.
    */
   @Override
   public void setRows(java.util.List<Object[]> rows)
   {
      super.setRows(rows);

      rowIter = rows.iterator();
   }

   /**
    * Set the metadata sent by a remote side, after the remote appserver call.
    * 
    * @param    properties
    *           The metadata sent by the remote side.
    */
   @Override
   public void setProperties(java.util.List<PropertyDefinition> properties)
   {
      super.setProperties(properties);

      propIter = properties.iterator();
   }

   /**
    * Check if there are more properties for this DMO.
    * <p>
    * WARNING: this allows only one iteration over the properties and is for internal usage
    * only.  Use {@link #asResultSet()} and the associated {@link ResultSetMetaData} instead.
    * 
    * @return   See above.
    */
   public boolean hasMoreProperties()
   {
      return (propIter == null) ? false : propIter.hasNext();
   }
   
   /**
    * Returns the next property of this DMO.
    * <p>
    * WARNING: this allows only one iteration over the properties and is for internal usage
    * only.  Use {@link #asResultSet()} and the associated {@link ResultSetMetaData} instead.
    * 
    * @return   See above.
    */
   public PropertyDefinition nextProperty()
   {
      return propIter.next();
   }
   
   /**
    * Check if there are more rows for this table.
    * <p>
    * WARNING: this allows only one iteration over the properties and is for internal usage
    * only.  Use {@link #asResultSet()} instead.
    * 
    * @return   See above.
    */
   public boolean hasMoreRows()
   {
      return (rowIter == null) ? false : rowIter.hasNext();
   }
   
   /**
    * Returns the data for the next row.
    * <p>
    * WARNING: this allows only one iteration over the properties and is for internal usage
    * only.  Use {@link #asResultSet()} instead.
    * 
    * @return    See above.
    */
   public Object[] nextRow()
   {
      return rowIter.next();
   }
   
   /**
    * Initialize this result set, by reading the entire temp-table in memory.
    * 
    * TODO: improve this (i.e. use an open cursor or some other alternative to not read the data
    * in memory)
    * 
    * @throws   ClassNotFoundException
    *           If the type of a property can not be resolved to a class.
    */
   private void init()
   throws ClassNotFoundException
   {
      TemporaryBuffer buffer = (TemporaryBuffer) ((BufferReference) dmo).buffer();
      Class<?> srcClass = buffer.getDMOImplementationClass();
      Class<?> srcIface = buffer.getDMOInterface();
      
      PersistentClass cls = DatabaseManager.getClassMapping(DatabaseManager.TEMP_TABLE_DB,  
                                                            srcClass);
      
      java.util.List<PropertyDefinition> props = new ArrayList<>();
      
      Iterator<Property> iter = cls.getPropertyIterator();
      while (iter.hasNext())
      {
         Property prop = iter.next();
         Value value = prop.getValue();
         
         // skip properties that are not accessible via the DMO's API.
         PropertyAccessor pa = prop.getPropertyAccessor(srcClass);
         if (pa instanceof NopPropertyAccessor || prop.getName().equals("_multiplex"))
         {
            continue;
         }
         
         if (value instanceof List)
         {
            List list = (List) value;
            Value element = list.getElement();
            if (element instanceof Component)
            {
               Component comp = (Component) element;
               
               // compute the number of entries in this component
               String compName = comp.getComponentClassName();
               compName = compName.substring(compName.indexOf("$Composite"));
               compName = compName.substring("$Composite".length());
               int extent = Integer.parseInt(compName);
               
               Iterator<Property> itrc = comp.getPropertyIterator();
               while (itrc.hasNext())
               {
                  Property compProp = (Property) itrc.next();
                  SimpleValue compVal = (SimpleValue) compProp.getValue();
                  Class<?> type = compVal.getType().getReturnedClass();
                  
                  String propName = compProp.getName();
                  String legacyName = TempTableBuilder.getLegacyFieldName(srcIface,
                                                                          propName,
                                                                          buffer.getParentTable());
                  props.add(new PropertyDefinition(propName, type, extent, legacyName));
               }
            }
         }
         else
         {
            String propName = prop.getName();
            String legacyName = TempTableBuilder.getLegacyFieldName(srcIface,
                                                                    propName,
                                                                    buffer.getParentTable());
            SimpleValue val = (SimpleValue) value;
            Class<?> type = val.getType().getReturnedClass();

            props.add(new PropertyDefinition(propName, type, legacyName));
         }
      }

      rowIter = TemporaryBuffer.readAllRows(dmo, props).iterator();
      propIter = props.iterator();
   }
}