/*
** Module   : RecordLockEvent.java
** Abstract : Object which stores information about a record lock event
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20131101 Created initial version.
** 002 ECF 20160822 Added event ID to debug string output.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.lock;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Object which stores information about a record lock event.
 */
public final class RecordLockEvent
{
   /** Sequential, unique event ID */
   private final long eventID;
   
   /** The locker */
   private final SessionToken locker;
   
   /** Unique identifier of the locked record */
   private final RecordIdentifier identifier;
   
   /** Lock type before the event */
   private final LockType oldType;
   
   /** Lock type after the event */
   private final LockType newType;
   
   /** Unique, numeric identifier of user in whose session the event occurred */
   private integer userid = null;
   
   /**
    * Constructor.
    * 
    * @param   eventID
    *          Sequential, unique event ID.
    * @param   locker
    *          The locker.
    * @param   identifier
    *          Unique identifier of the locked record.
    * @param   oldType
    *          Lock type before the event.
    * @param   newType
    *          Lock type after the event.
    */
   RecordLockEvent(long eventID,
                   SessionToken locker,
                   RecordIdentifier identifier,
                   LockType oldType,
                   LockType newType)
   {
      this.eventID = eventID;
      this.locker = locker;
      this.identifier = identifier;
      this.oldType = oldType;
      this.newType = newType;
   }
   
   /**
    * Get the unique ID of this event.
    * 
    * @return  Event ID.
    */
   public long getEventID()
   {
      return eventID;
   }
   
   /**
    * Get the session token of the user which generated this lock event.
    * 
    * @return  Locker's session token.
    */
   public SessionToken getLocker()
   {
      return locker;
   }
   
   /**
    * Get the identifier which uniquely identifies the locked record.
    * 
    * @return  Locked record identifier.
    */
   public RecordIdentifier getIdentifier()
   {
      return identifier;
   }
   
   /**
    * Get the lock type as it was before this event.
    * 
    * @return  Previous lock type.
    */
   public LockType getOldType()
   {
      return oldType;
   }
   
   /**
    * Get the lock type as it was after this event.
    * 
    * @return  New lock type.
    */
   public LockType getNewType()
   {
      return newType;
   }
   
   /**
    * Get the ID of the user who generated this lock event.
    * 
    * @return  User ID.
    */
   public integer getUserid()
   {
      return userid;
   }
   
   /**
    * Set the ID of the user who generated this lock event.
    * @param userid
    */
   public void setUserid(integer userid)
   {
      this.userid = userid;
   }
   
   /**
    * Get a string describing this event.
    * 
    * @return  String representation of this event.
    */
   @Override
   public String toString()
   {
      StringBuilder buf = new StringBuilder("Lock event #");
      buf.append(eventID);
      buf.append(": ");
      buf.append(locker.toString());
      buf.append(" ");
      buf.append(identifier.toString());
      buf.append("; ");
      buf.append(oldType.toString());
      buf.append("->");
      buf.append(newType.toString());
      
      return buf.toString();
   }
}