/*
** Module   : IdentityManagerMultiplexer.java
** Abstract : Defines a remote API for database record identity management
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SVL 20080111 ADD  @36795  Created initial version. Defines a remote API
**                               for database identity management.
** 002 SVL 20080901 CHG  @39648  Added reclaimKeys() function.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.remote;

import java.io.Serializable;
import com.goldencode.p2j.util.RemoteMultiplexer;
import com.goldencode.p2j.persist.PersistenceException;

/**
 * Interface which defines a client-server API for remote database identity
 * management (e.g. primary key generation). {@link RemoteIdentityManager}
 * invokes the methods of this interface on a local proxy for a remote object,
 * in order to access the identity management services of the remote P2J
 * server which is authoritative for a particular database.
 * <p>
 * Since the remote P2J server may manage multiple databases, but only a
 * single instance of an implementor of this interface will be available at
 * any given remote server, identity management service requests must be
 * multiplexed by database.  For this reason, each of the methods defined by
 * this interface accepts an integer ID, which uniquely identifies the target
 * database at the remote server.
 * <p>
 * The database ID to use for a particular physical database is determined by
 * calling {@link IdentityManagerMultiplexer#getMultiplexID(Object)}. This is
 * the responsibility of the <code>RemoteIdentityManager</code> and must be
 * done before any dirty read management service is requested.
 * <p>
 * This interface effectively mirrors that of {@link
 * com.goldencode.p2j.persist.id.IdentityManager the IdentityManager API},
 * except for the additional database multiplex identifier which is the first
 * parameter to each method.
 *
 * @author SVL
 */
public interface IdentityManagerMultiplexer
extends RemoteMultiplexer<String>
{
   /**
    * Return the next primary key for the specified table.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    *
    * @param   table
    *          Table for which the next primary key will be returned.
    *
    * @return  The next primary key for the given table.
    *
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    */
   public Serializable nextPrimaryKey(int database, String table)
   throws PersistenceException;

   /**
    * Inform the identity manager that the given primary keys can be reused.
    *
    * @param database
    *        Value that uniquely identifies the target database on the
    *        remote server.
    *
    * @param keys
    *        Keys for reuse.
    */
   void reclaimKeys(int database, Serializable[] keys);
}