/*
** Module   : SerializeOptions.java
** Abstract : Options related to serialized temp-table data.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20171016 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.serial;

/**
 * Options related to serialized temp-table data.
 */
public class SerializeOptions
{
   /** Modes for reading serialized data into a temp-table */
   public static enum Read
   {
      APPEND,
      EMPTY,
      MERGE,
      REPLACE,
   }
   
   /** Modes for verifying XML data being read into a temp-table */
   public static enum Verify
   {
      LOOSE,
      STRICT,
   }
   
   /** Type of XML nodes for serialized data */
   static enum XmlNodeType
   {
      ELEMENT,
      ATTRIBUTE,
   }
   
   /** Should this field be omitted from serialized output? */
   private final boolean serializeHidden;
   
   /** Name of field in serialized output (overrides field name). */
   private final String serializeName;
   
   /** XML schema data type of field. */
   private final String xmlDataType;
   
   /** Name of element or attribute representing field in XML output. */
   private final String xmlNodeName;
   
   /** Node type of field in XML output. */
   private final XmlNodeType xmlNodeType;
   
   /**
    * Constructor.
    * 
    * @param   serializeHidden
    *          Should this field be omitted from serialized output?
    * @param   serializeName
    *          Name of field in serialized output (overrides field name).
    * @param   xmlDataType
    *          XML schema data type of field
    * @param   xmlNodeName
    *          Name of element or attribute representing field in XML output.
    * @param   xmlNodeType
    *          Node type of field in XML output (overrides {@code serializeName}).
    */
   public SerializeOptions(boolean serializeHidden,
                           String serializeName,
                           String xmlDataType,
                           String xmlNodeName,
                           String xmlNodeType)
   {
      this.serializeHidden = serializeHidden;
      this.serializeName = serializeName;
      this.xmlDataType = xmlDataType;
      this.xmlNodeName = xmlNodeName;
      this.xmlNodeType = XmlNodeType.valueOf(xmlNodeType.toUpperCase());
   }
   
   /**
    * Get the value of the serialize-hidden flag.
    * 
    * @return  {@code true} if the associated field is not included in serialization, else
    *          {@code false}.
    */
   public boolean isSerializeHidden()
   {
      return serializeHidden;
   }
   
   /**
    * Get the name of the field in serialized output (overrides field name in table).
    * 
    * @return  Serialize name.
    */
   public String getSerializeName()
   {
      return serializeName;
   }
   
   /**
    * Get the XML schema data type of the associated field.
    * 
    * @return  XML schema data type.
    */
   public String getXmlDataType()
   {
      return xmlDataType;
   }
   
   /**
    * Get the node name of the field in XML output (overrides serialize-name and field name in
    * table).
    * 
    * @return  XML node name.
    */
   public String getXmlNodeName()
   {
      return xmlNodeName;
   }
   
   /**
    * Get the node type (element or attribute) in XML output.
    * 
    * @return  XML node type.
    */
   public XmlNodeType getXmlNodeType()
   {
      return xmlNodeType;
   }
}
