/*
** Module   : TempTableSchema.java
** Abstract : Temp-table schema used for serialization, based on internal structure of the table.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20171020 Created initial version.
** 002 CA  20180511 Added XML-NODE-NAME table option support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.serial;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * The schema used for serialization of a temp-table, based on the internal definition of the
 * table itself. Names used for serialization are legacy names as defined by the original 4GL
 * code, not derived by conversion. Only legacy fields not marked SERIALIZE-HIDDEN are included
 * in the schema.
 */
class TempTableSchema
{
   /** Map of wrapper types to XML data type names */
   private static Map<Class<? extends BaseDataType>, String> xsdTypes = new HashMap<>();
   
   /** Map of wrapper types to ProData XML data type names */
   private static Map<Class<? extends BaseDataType>, String> proTypes = new HashMap<>();
   
   static
   {
      // xsdTypes.put(blob.class, "base64Binary");
      xsdTypes.put(character.class, "string");
//      xsdTypes.put(clob.class, "string");
//      xsdTypes.put(comhandle.class, "long");
      xsdTypes.put(date.class, "date");
      xsdTypes.put(datetime.class, "dateTime");
      xsdTypes.put(datetimetz.class, "dateTime");
      xsdTypes.put(decimal.class, "decimal");
      xsdTypes.put(int64.class, "long");
      xsdTypes.put(integer.class, "int");
      xsdTypes.put(logical.class, "boolean");
      xsdTypes.put(raw.class, "base64Binary");
      xsdTypes.put(recid.class, "long");
      xsdTypes.put(rowid.class, "base64Binary");
      xsdTypes.put(handle.class, "long");
//      xsdTypes.put(widgethandle.class, "long");
      
      // proTypes.put(blob.class, "blob");
//      proTypes.put(clob.class, "clob");
//      proTypes.put(comhandle.class, "comHandle");
      proTypes.put(datetime.class, "dateTime");
      proTypes.put(recid.class, "recid");
      proTypes.put(rowid.class, "rowid");
      proTypes.put(handle.class, "handle");
//      proTypes.put(widgethandle.class, "handle");
   }
   
   /** Legacy name of the temp-table */
   private final String name;
   
   /** Schema information needed for serialization of all columns/fields */
   private final LinkedHashMap<String, Column> columns = new LinkedHashMap<>();
   
   /** List of column schema objects whose data is to be serialized as XML attributes */
   private final List<Column> attributeColumns;
   
   /** List of column schema objects whose data is to be serialized as XML elements */
   private final List<Column> elementColumns;
   
   /** The serialize options defined at the temp-table. */
   private final TableSerializeOptions serializeOptions;
   
   /**
    * Constructor which gathers table and field data from a temp-table buffer.
    * 
    * @param   buffer
    *          Buffer on the temp-table.
    */
   public TempTableSchema(TemporaryBuffer buffer)
   {
      Buffer buf = (Buffer) buffer.getDMOProxy();
      // the table name is used for serialization, regardless of the individual buffer name
      this.name = TableMapper.getLegacyName(buf);
      this.serializeOptions = TableMapper.getSerializeOptions(buf);
      
      Class<?> dmoIface = buffer.getDMOInterface();
      Map<String, Integer> extentMap = PropertyHelper.extentsByProperty(dmoIface);
      Map<String, Method> getterMap = PropertyHelper.allGettersByProperty(dmoIface);
      Map<String, Method> setterMap = PropertyHelper.allSettersByProperty(dmoIface);
      
      TempTable tt = buffer.getParentTable();
      List<TableMapper.LegacyFieldInfo> fields = TableMapper.getAllLegacyFieldInfo(tt);
      Iterator<TableMapper.LegacyFieldInfo> iter = fields.iterator();
      
      // Construct column schema data object; TableMapper.getAllLegacyFieldInfo returns field
      // info objects in the correct order for serialization
      for (int i = 0; iter.hasNext(); i++)
      {
         TableMapper.LegacyFieldInfo info = iter.next();
         
         // add column to schema if it should be serialized
         if (!info.getSerializeOptions().isSerializeHidden())
         {
            Column col = new Column(buffer, i + 1, getterMap, setterMap, extentMap, info);
            columns.put(col.getName(), col);
         }
      }
      
      this.attributeColumns = filterColumns(columns.values(),
                                            SerializeOptions.XmlNodeType.ATTRIBUTE);
      this.elementColumns = filterColumns(columns.values(),
                                          SerializeOptions.XmlNodeType.ELEMENT);
   }
   
   /**
    * Filter the given list of column schema objects by the given XML node type.
    * 
    * @param   columns
    *          Collection of column schema objects to be filtered.
    * @param   nodeType
    *          Node type which forms the filter criterion.
    * 
    * @return  A list of column schema objects whose node type matches the given node type.
    *          The list may be empty, but will not be {@code null}.
    */
   private static List<Column> filterColumns(Collection<Column> columns,
                                             SerializeOptions.XmlNodeType nodeType)
   {
      List<Column> list = null;
      
      for (Column column : columns)
      {
         if (column.nodeType == nodeType)
         {
            if (list == null)
            {
               list = new ArrayList<>();
            }
            list.add(column);
         }
      }
      
      return list == null ? Collections.emptyList() : list;
   }
   
   /**
    * Get the legacy name of this temp-table.
    * 
    * @return  Legacy table name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Compute the XML node name for this table, fro mthe {@link #serializeOptions}.
    * 
    * @return    See above.
    */
   public String getTableXmlNodeName()
   {
      String tableName = name;

      if (serializeOptions.getXmlNodeName().length() > 0)
      {
         tableName = serializeOptions.getXmlNodeName();
      }
      else
      {
         // TODO: compute from serialize-name, namespace-prefix and namespace-uri
      }
      
      return tableName;
   }
   
   /**
    * Get the serialize options for this temp-table.
    * 
    * @return   The {@link #serializeOptions}.
    */
   public TableSerializeOptions getSerializeOptions()
   {
      return serializeOptions;
   }
   
   /**
    * Get the column schema information associated with the given serialization name.
    * 
    * @param   name
    *          Name used in serialization output for column data.
    * 
    * @return  Column schema information.
    */
   public Column getColumn(String name)
   {
      return columns.get(name);
   }
   
   /**
    * Get an object to iterate over column schema information objects.
    * 
    * @return  An iterable object on column schema info objects.
    */
   public Iterable<Column> columns()
   {
      return columns.values();
   }
   
   /**
    * Get an iterable collection of column schema objects whose data is to be serialized as XML
    * attributes.
    * 
    * @return  see above.
    */
   public Iterable<Column> attributeColumns()
   {
      return attributeColumns;
   }
   
   /**
    * Get an iterable collection of column schema objects whose data is to be serialized as XML
    * elements.
    * 
    * @return  see above.
    */
   public Iterable<Column> elementColumns()
   {
      return elementColumns;
   }
   
   /**
    * Schema information needed to serialize/deserialize a field/column.
    */
   static class Column
   {
      /** Legacy name of field */
      private final String fieldName;
      
      /** Is column nillable (i.e., not mandatory)? */
      private final boolean nillable;
      
      /** Extent of an array field (null for a scalar field) */
      private final Integer extent;
      
      /** Order in which column is serialized */
      private final int order;
      
      /** Type of serialized representation for this column: element or attribute */
      private final SerializeOptions.XmlNodeType nodeType;
      
      /** Name of element or attribute in serialized form (may differ from legacy field name) */
      private final String name;
      
      /** Method to get field data from buffer */
      private final Method getter;
      
      /** Method to set field data into buffer */
      private final Method setter;
      
      /** Data wrapper type of field */
      private final Class<? extends BaseDataType> type;
      
      /**
       * Constructor.
       * 
       * @param   buffer
       *          Temp-table buffer.
       * @param   order
       *          Order of column in serialization output.
       * @param   getterMap
       *          Map of java property names to getter methods.
       * @param   setterMap
       *          Map of java property names to setter methods.
       * @param   extentMap
       *          Map of java property names to field extent values.
       * @param   field
       *          Field info object.
       */
      Column(TemporaryBuffer buffer,
             int order,
             Map<String, Method> getterMap,
             Map<String, Method> setterMap,
             Map<String, Integer> extentMap,
             TableMapper.LegacyFieldInfo field)
      {
         SerializeOptions opts = field.getSerializeOptions();
         
         String serName = opts.getSerializeName();
         if ("".equals(serName))
         {
            serName = null;
         }
         
         String nodeName = opts.getXmlNodeName();
         if ("".equals(nodeName))
         {
            nodeName = null;
         }
         
         this.nillable = !field.isMandatory();
         this.fieldName = field.getLegacyName();
         this.order = order;
         this.nodeType = opts.getXmlNodeType();
         
         // name used in serialized output is the legacy name of the field, unless overridden
         // by the SERIALIZE-NAME attribute, unless overridden by the XML-NODE-NAME attribute
         // TODO: possibly needs rework to differentiate between XML and JSON serialization
         this.name = (nodeName != null ? nodeName : serName != null ? serName : fieldName);
         
         String javaName = field.getJavaName();
         this.getter = getterMap.get(javaName);
         this.setter = setterMap.get(javaName);
         this.extent = extentMap.get(javaName);
         this.type = (Class<? extends BaseDataType>) getter.getReturnType();
      }
      
      /**
       * Get the column's legacy field name.
       * 
       * @return  Legacy field name.
       */
      public String getFieldName()
      {
         return fieldName;
      }
      
      /**
       * Indicate whether column can be null.
       * 
       * @return  {@code true} if nillable, else {@code false}.
       */
      public boolean isNillable()
      {
         return nillable;
      }
      
      /**
       * Get extent of column, if any.
       * 
       * @return  Extent value for an array column, {@code null} for a scalar column.
       */
      public Integer getExtent()
      {
         return extent;
      }
      
      /**
       * Get the order in which this column is serialized.
       * 
       * @return  Order.
       */
      public int getOrder()
      {
         return order;
      }
      
      /**
       * Get the XML node type (element or attribute) for this column.
       * 
       * @return  Node type.
       */
      public SerializeOptions.XmlNodeType getNodeType()
      {
         return nodeType;
      }
      
      /**
       * Get the element or attribute name of the column's serialized representation.
       * 
       * @return  Element/attribute name.
       */
      public String getName()
      {
         return name;
      }
      
      /**
       * Get the DMO getter method for this column's value.
       * 
       * @return  Getter method.
       */
      public Method getGetter()
      {
         return getter;
      }
      
      /**
       * Get the DMO setter method for this column's value.
       * 
       * @return  Setter method.
       */
      public Method getSetter()
      {
         return setter;
      }
      
      /**
       * Get the data type of the field.
       * 
       * @return  Data type.
       */
      public Class<? extends BaseDataType> getType()
      {
         return type;
      }
   }
}
