/*
** Module   : XmlImport.java
** Abstract : Read XML content into a temp-table.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20171016 Created initial version.
** 002 CA  20180511 Added XML-NODE-NAME table option support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.serial;

import java.io.*;
import java.util.*;
import javax.xml.stream.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of the 4GL READ-XML method, which parses an XML stream and reads the content
 * into a temp-table.
 * <p>
 * Runtime support is partial at this time; not all modes and features are supported. PRODATASET
 * is not supported at this time.
 */
public final class XmlImport
{
   /** XSD schema element name */
   private static final String ELEM_SCHEMA = "schema";
   
   /** Nil attribute name */
   private static final String ATTR_NIL = "nil";
   
   /** String constant for 'true' */
   private static final String TEXT_TRUE = "true";
   
   /** Read mode to determine how records are stored and how non-unique records are handled */
   private final SerializeOptions.Read readMode;
   
   /** Verify schema mode */
   private final SerializeOptions.Verify verifySchemaMode;
   
   /** Buffer proxy */
   private final Buffer proxy;
   
   /** Temp-table schema information */
   private final TempTableSchema schema;
   
   /** Helper which tracks current index values for extent fields */
   private final ExtentTracker extentTracker;
   
   /**
    * Constructor which immediately performs the import.
    * 
    * @param   buffer
    *          Temp-table buffer in which data is stored.
    * @param   source
    *          XML data source which normalizes access of various media to an input stream.
    * @param   readMode
    *          Read mode to determine how records are stored and how non-unique records are
    *          handled. Not currently honored, except temp-table records are deleted if mode
    *          is EMPTY.
    * @param   schemaLocation
    *          Path to XML schema (not currently used).
    * @param   overrideDefaultMapping
    *          Not currently used.
    * @param   fieldTypeMapping
    *          Not currently used.
    * @param   verifySchemaMode
    *          Not currently used.
    * 
    * @throws  PersistenceException
    *          if there is any error reading or storing XML data.
    */
   public XmlImport(TemporaryBuffer buffer,
                    DataSource source,
                    SerializeOptions.Read readMode,
                    String schemaLocation,
                    boolean overrideDefaultMapping,
                    String fieldTypeMapping,
                    SerializeOptions.Verify verifySchemaMode)
   throws PersistenceException
   {
      this.readMode = readMode;
      this.verifySchemaMode = verifySchemaMode;
      this.schema = new TempTableSchema(buffer);
      this.proxy = (Buffer) buffer.getDMOProxy();
      this.extentTracker = new ExtentTracker(schema);
      
      if (readMode == SerializeOptions.Read.EMPTY)
      {
         proxy.deleteAll();
      }
      
      readData(source);
   }
   
   /**
    * Read XML data from the input source and store it in the temp-table.
    * 
    * @param   source
    *          Data input source.
    * 
    * @throws  PersistenceException
    *          if there is any error reading or storing XML data.
    */
   private void readData(DataSource source)
   throws PersistenceException
   {
      XMLInputFactory factory = XMLInputFactory.newInstance();
      XMLStreamReader reader = null;
      String tableName = schema.getTableXmlNodeName();
      
      try (InputStream stream = source.getInputStream())
      {
         reader = factory.createXMLStreamReader(stream);
         
         while (reader.hasNext())
         {
            int type = reader.next();
            switch (type)
            {
               case XMLStreamReader.START_ELEMENT:
                  String name = reader.getLocalName();
                  if (tableName.equals(name))
                  {
                     String rowName = name + "Row";
                     readContent(reader, rowName);
                  }
                  else
                  {
                     String msg = "Temp-Table name '" +
                                  tableName +
                                  "' in namespace '' not found in XML Document.";
                     ErrorManager.displayError(13514, msg);
                     
                     throw new PersistenceException(msg);
                  }
                  break;
                  
               case XMLStreamReader.END_ELEMENT:
                  return;
            }
         }
      }
      catch (IOException | XMLStreamException exc)
      {
         throw new PersistenceException(exc);
      }
      finally
      {
         if (reader != null)
         {
            try
            {
               reader.close();
            }
            catch (XMLStreamException exc)
            {
               // TODO: log
            }
         }
      }
   }
   
   /**
    * Read XML content from reader and import records into the temp-table.
    * 
    * @param   reader
    *          XML stream reader.
    * @param   rowName
    *          Element name for temp-table record data in XML content.
    * 
    * @throws  XMLStreamException
    *          if there is an error reading XML content from the stream.
    * @throws  PersistenceException
    *          if there is an error storing data in the temp-table.
    */
   private void readContent(XMLStreamReader reader, String rowName)
   throws XMLStreamException,
          PersistenceException
   {
      while (reader.hasNext())
      {
         int type = reader.next();
         switch (type)
         {
            case XMLStreamReader.START_ELEMENT:
               String name = reader.getLocalName();
               if (rowName.equals(name))
               {
                  readRecord(reader);
               }
               else if (ELEM_SCHEMA.equals(name))
               {
                  readSchema(reader);
               }
               break;
               
            case XMLStreamReader.END_ELEMENT:
               return;
         }
      }
      
      throw new XMLStreamException("Unexpected end of XML stream");
   }
   
   /**
    * Read XSD schema data from the XML content.
    * <p>
    * The current implementation simply skips this section of the content, if present.
    * 
    * @param   reader
    *          XML stream reader.
    * 
    * @throws  XMLStreamException
    *          if there is an error reading XML content from the stream.
    */
   private void readSchema(XMLStreamReader reader)
   throws XMLStreamException
   {
      while (reader.hasNext())
      {
         int type = reader.next();
         switch (type)
         {
            // skip all content inside the 'xsd:schema' element
            // TODO: implement reading XML schema
            case XMLStreamReader.END_ELEMENT:
               String name = reader.getLocalName();
               if (ELEM_SCHEMA.equals(name))
               {
                  return;
               }
               break;
         }
      }
   }
   
   /**
    * Read XML content from the stream corresponding with a single data record and store it in
    * the temp-table.
    * 
    * @param   reader
    *          XML stream reader.
    * 
    * @throws  XMLStreamException
    *          if there is an error reading XML content from the stream.
    * @throws  PersistenceException
    *          if there is an error storing data in the temp-table.
    */
   private void readRecord(XMLStreamReader reader)
   throws XMLStreamException,
          PersistenceException
   {
      proxy.create();
      
      try
      {
         RecordBuffer.startBatch();
         
         // read row attributes
         int attrs = reader.getAttributeCount();
         for (int i = 0; i < attrs; i++)
         {
            String name = reader.getAttributeLocalName(i);
            String value = reader.getAttributeValue(i);
            TempTableSchema.Column column = getColumn(name);
            setField(column, name, value);
         }
         
         // read row's child elements
         while (reader.hasNext())
         {
            int type = reader.next();
            switch (type)
            {
               case XMLStreamReader.START_ELEMENT:
                  String name = reader.getLocalName();
                  TempTableSchema.Column column = getColumn(name);
                  boolean isNil = checkNil(column, reader);
                  String value = readText(reader);
                  setField(column, name, isNil ? null : value);
                  break;
                  
               case XMLStreamReader.END_ELEMENT:
                  return;
            }
         }
      }
      finally
      {
         extentTracker.reset();
         RecordBuffer.endBatch();
      }
   }
   
   /**
    * Read text content from within an element or a CDATA section.
    * 
    * @param   reader
    *          XML stream reader.
    * 
    * @return  Text which was read.
    * 
    * @throws  XMLStreamException
    *          if there is an error reading XML content from the stream.
    */
   private String readText(XMLStreamReader reader)
   throws XMLStreamException
   {
      StringBuilder bld = new StringBuilder();
      while (reader.hasNext())
      {
         int type = reader.next();
         switch (type)
         {
            case XMLStreamReader.CHARACTERS:
            case XMLStreamReader.CDATA:
               bld.append(reader.getText());
               break;
            case XMLStreamReader.END_ELEMENT:
               return bld.toString();
         }
      }
      
      return null;
   }
   
   /**
    * Get schema information for the column associated with the given name.
    * 
    * @param   name
    *          Element or attribute name associated with serialized XML data for a temp-table
    *          column.
    * 
    * @return  Column schema information.
    * 
    * @throws  ErrorConditionException
    *          if {@code name} is not associated with a column.
    */
   private TempTableSchema.Column getColumn(String name)
   {
      TempTableSchema.Column column = schema.getColumn(name);
      if (column == null)
      {
         // TODO: proper error message/number/modality
         ErrorManager.recordOrThrowError(0, "Unknown field " + name, false);
      }
      
      return column;
   }
   
   /**
    * Check whether the current temp-table value read from XML content is null. If so, check
    * whether the associated column allows null values.
    * 
    * @param   column
    *          Schema information for the associated column.
    * @param   reader
    *          XML stream reader.
    * 
    * @return  {@code true} if the XML content is nil; else {@code false}.
    * 
    * @throws  ErrorConditionException
    *          if value is nil and column does not allow null values.
    */
   private boolean checkNil(TempTableSchema.Column column, XMLStreamReader reader)
   {
      // read attributes for current element
      int attrs = reader.getAttributeCount();
      for (int i = 0; i < attrs; i++)
      {
         String name = reader.getAttributeLocalName(i);
         String value = reader.getAttributeValue(i);
         
         if (ATTR_NIL.equals(name))
         {
            if (TEXT_TRUE.equals(value))
            {
               if (!column.isNillable())
               {
                  ErrorManager.recordOrThrowError(0,
                                                  "Column not nillable: " + column.getName(),
                                                  false);
               }
               
               return true;
            }
            
            return false;
         }
      }
      
      return false;
   }
   
   /**
    * Set the value of a single field in the temp-table record currently being read from the XML
    * content.
    *  
    * @param   column
    *          Schema information for the associated column.
    * @param   name
    *          Element or attribute name associated with serialized XML data for a temp-table
    *          column.
    * @param   value
    *          Data value read from XML content. Set to {@code null} to represent unknown value.
    * 
    * @throws  PersistenceException
    *          if there is an error storing data in the temp-table.
    */
   private void setField(TempTableSchema.Column column, String name, String value)
   throws PersistenceException
   {
      try
      {
         BaseDataType datum = column.getType().newInstance();
         if (value == null)
         {
            datum.setUnknown();
         }
         else
         {
            Stream.assignDatum(datum, value);
         }
         
         Integer index = extentTracker.getNext(name);
         Object[] args = (index != null ? new Object[] { index, datum } : new Object[] { datum });
         column.getSetter().invoke(proxy, args);
      }
      catch (Exception exc)
      {
         throw new PersistenceException(exc);
      }
   }
   
   /**
    * Helper object which tracks current and maximum index values for extent fields. This is
    * used to get the next index value needed for a setter method invocation when reading and
    * setting elements into an extent field in the current DMO. A range check is performed to
    * ensure too many array elements are not set.
    */
   private static class ExtentTracker
   {
      /** Extent index trackers mapped to XML content names for columns */
      private final Map<String, Tracker> trackers = new HashMap<>();
      
      /**
       * Constructor. Initializes a map for extent fields, if any, in a temp-table.
       * 
       * @param   schema
       *          Schema information for the temp-table.
       */
      ExtentTracker(TempTableSchema schema)
      {
         for (TempTableSchema.Column column : schema.columns())
         {
            Integer extent = column.getExtent();
            if (extent != null)
            {
               String name = column.getName();
               Tracker tracker = new Tracker(extent);
               trackers.put(name, tracker);
            }
         }
      }
      
      /**
       * Reset the current index values to 0 for all extent fields being tracked.
       */
      void reset()
      {
         for (Tracker tracker : trackers.values())
         {
            tracker.reset();
         }
      }
      
      /**
       * Get the index value to be used to set the next element in the extent field associated
       * with the given name.
       * 
       * @param   name
       *          Name associated with column data in XML content.
       * 
       * @return  Next index value for extent field setter call.
       * 
       * @throws  ArrayIndexOutOfBoundsException
       *          if the current index value reaches the size of the extent field.
       */
      Integer getNext(String name)
      {
         Tracker tracker = trackers.get(name);
         
         return (tracker == null ? null : tracker.getNext());
      }
      
      /**
       * Helper which tracks current and maximum index values for a single extent field.
       */
      private static class Tracker
      {
         /** Size of extent field */
         private final int max;
         
         /** Current index value */
         private int current = 0;
         
         /**
          * Constructor.
          * 
          * @param   max
          *          Size of extent field.
          */
         Tracker(int max)
         {
            this.max = max;
         }
         
         /**
          * Get the index value to be used to set the next element in the extent field.
          * 
          * @return  Next index value for extent field setter call.
          * 
          * @throws  ArrayIndexOutOfBoundsException
          *          if the current index value reaches the size of the extent field.
          */
         Integer getNext()
         {
            if (current >= max)
            {
               throw new ArrayIndexOutOfBoundsException();
            }
            
            return current++;
         }
         
         /**
          * Reset the current value to 0.
          */
         void reset()
         {
            current = 0;
         }
      }
   }
}
