/*
** Module   : DateUserType.java
** Abstract : Custom user type to wrap P2J date for Hibernate access
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050630   @21594 Created initial version. Maps P2J date to SQL date.
** 002 ECF 20050729   @21953 Refactored to move nullSafeSet method implementation up to
**                           superclass. This was necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22073 Refactored to move nullSafeGet method implementation up to
**                           superclass. This was necessary to add centralized logging and null
**                           checking logic.
** 004 ECF 20051123   @23537 Fixed bindNotNull. Now we adjust for timezone and daylight savings
**                           when creating a Date for the prepared statement.
** 009 ECF 20080806   @39296 Moved to type sub-package of persist.
** 010 ECF 20090331   @41712 Reworked bindNotNull() to fix bug introduced with #009 (@39296). We
**                           can't use a static Calendar instance to determine DST offset, as the
**                           offset would be incorrect in a long running server which crosses a
**                           DST boundary. Instead, we rely on date.dateValue(TimeZone) to give us
**                           a Date which is adjusted for the JVM's default time zone settings.
** 011 OM  20130618          Use default constructor for unknown date value.
** 012 ECF 20140624          Refactored SQL_TYPES to use TypeHelper.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link com.goldencode.p2j.util.date
 * P2J date}. This permits Hibernate to convert this data type to and from the SQL DATE data type.
 * The unknown value is mapped to SQL NULL.
 */
public final class DateUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { TypeHelper.getSqlType(date.class) };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a date to/from a database
    * record.
    *
    * @return  Array containing SQL type constant.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.date P2J date} class.
    */
   public Class returnedClass()
   {
      return date.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>date</code>
    * instance.
    *
    * @param   value
    *          The <code>date</code> instance to be copied.
    *
    * @return  A new instance of <code>date</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      return new date((date) value);
   }
   
   /**
    * Extract a <code>Date</code> value from a result set at a particular
    * column.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  Date data extracted from the result set; may be
    *          <code>null</code>.
    *
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      return rs.getDate(column);
   }
   
   /**
    * Construct and return an instance of <code>date</code> which is
    * initialized to the unknown value.
    *
    * @return  A <code>date</code> instance which represents the unknown value.
    */
   protected BaseDataType instantiateUnknown()
   {
      return new date(); // unknown date
   }
   
   /**
    * Construct and return an instance of <code>date</code> which wraps
    * <code>value</code>.
    *
    * @param   value
    *          Date value to be wrapped by a <code>date</code>; will not be <code>null</code>.
    *
    * @return  <code>value</code> wrapped in a <code>date</code>.
    */
   protected BaseDataType instantiateRegular(Object value)
   {
      return new date((Date) value);
   }
   
   /**
    * Set a <code>Date</code> value into the given prepared statement,
    * based upon the <code>date</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          An <code>integer</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps, BaseDataType value, int index)
   throws SQLException
   {
      // Need a java.sql.Date;  date class deals in java.util.Date.  The Date
      // must be constructed using the JVM's default timezone, to ensure there
      // is no drift to the previous date, due to timezone or Daylight Savings
      // offsets.
      date d = (date) value;
      long localMillis = d.dateValue(null).getTime();
      Date sqlDate = new Date(localMillis);
      
      ps.setDate(index, sqlDate);
   }
}