/*
** Module   : DatetimeTzUserType.java
** Abstract : Custom user type to wrap P2J datetime for Hibernate access
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20130122 Created initial version. Maps P2J datetimetz to SQL
**                  timestamp and smallint columns.  2 columns are needed
**                  because the Progress TZ is persisted as an integer offset
**                  which cannot be stored in a SQL timestamp alone.
** 002 OM  20130524 Rewrote entire class to extend CompositeUserType in order to accommodate the 
**                  compound type.
** 003 ECF 20140220 Removed Apache commons logger (it was not being used).
** 004 IAS 20160331 Fixed potential NPE
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.io.Serializable;
import java.sql.*;
import org.hibernate.HibernateException;
import org.hibernate.engine.spi.SessionImplementor;
import org.hibernate.type.*;
import org.hibernate.usertype.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.datetimetz P2J datetimetz}. This permits Hibernate to
 * convert this data type to and from the SQL TIMESTAMP and SMALLINT aggregate
 * data types. The unknown value is mapped to SQL NULL in the TIMESTAMP column.
 */
public final class DatetimeTzUserType
implements CompositeUserType
{
   /** Two element array containing the HIBERNATE type being mapped. */
   private static final Type[] HIBERNATE_TYPES = { TimestampType.INSTANCE, IntegerType.INSTANCE };
   
   /**
    * Get the "property names" that may be used in a query.
    *
    * @return  An array of Strings that contains the names of the fields occupied by this object. 
    */
   @Override
   public String[] getPropertyNames()
   {
      return new String[] { "timestamp", "offset" };
   }

   /**
    * Get the corresponding "property types".
    *
    * @return an array of Hibernate types
    */
   @Override
   public Type[] getPropertyTypes()
   {
      return HIBERNATE_TYPES;
   }

   /**
    * Get the value of a property.
    *
    * @param   component
    *          An instance of class mapped by this "type".
    * @param   property
    *          Index of property to be returned.
    * 
    * @return  The property value
    * 
    * @throws HibernateException
    */
   @Override
   public Object getPropertyValue(Object component, int property)
   throws HibernateException
   {
      if (component == null || !(component instanceof datetimetz))
      {
         return null;
      }
      
      datetimetz dtz = (datetimetz) component;
      if (property == 0)
      {
         return new Timestamp(dtz.dateValue().getTime());
      }
      else if (property == 1)
      {
         return dtz.getOffset();
      }

      return null;
   }

   /**
    * Set the value of a property.
    *
    * @param   component
    *          An instance of datetimetz class mapped by this "type".
    * @param   property
    *          Index of property to be set.
    * @param   value
    *          The value to set.
    * 
    * @throws HibernateException
    */
   @Override
   public void setPropertyValue(Object component, int property, Object value)
   throws HibernateException
   {
      if (value == null || component == null || !(component instanceof datetimetz))
      {
         return;
      }
      
      datetimetz dtz = (datetimetz) component;
      if (property == 0)
      {
         dtz.setTimestamp((Date) value); // a Timestamp in fact
      }
      else if (property == 1)
      {
         dtz.setTimeZoneOffset((Integer) value);
      }
   }

   /**
    * Tests for equality between two data wrapper instances. The objects are considered the same
    * if they meet any of the following conditions:
    * <ul>
    *    <li>they represent the same object instance in memory; OR
    *    <li>they are both <code>null</code>; OR
    *    <li><code>x.equals(y)</code> evaluates to <code>true</code>.
    * </ul>
    * Otherwise, they are considered different.
    * <p>
    * This method is used by Hibernate for dirty checking.
    *
    * @param   x
    *          A data wrapper type instance.
    * @param   y
    *          A data wrapper type instance.
    */
   @Override
   public boolean equals(Object x, Object y)
   throws HibernateException
   {
      if (x == y)
      {
         return true;
      }

      if (x == null || y == null)
      {
         return false;
      }

      return x.equals(y);
   }

   /**
    * Get the hash code of the given object in a manner that is consistent with the equality check
    * performed by {@link #equals}. This implementation simply returns the result of invoking
    * <code>x.hashCode()</code>.
    *
    * @param   x
    *          A data wrapper type instance.
    *
    * @return  The given object's hash code.
    */
   @Override
   public int hashCode(Object x)
   throws HibernateException
   {
      return x.hashCode();
   }

   /**
    * Retrieve an instance of the <code>datetimetz</code> class from a JDBC resultset. 
    *
    * @param   rs 
    *          A JDBC result set.
    * @param   names
    *          The column names.
    * @param   sess
    *          The originating session.
    * @param   owner
    *          The containing entity.
    *          
    * @return  The instantiated <code>datetimetz</code> object.
    *  
    * @throws  HibernateException
    * @throws  SQLException
    *          if a database access error occurs.
    */
   @Override
   public Object nullSafeGet(ResultSet rs, String[] names, SessionImplementor sess, Object owner)
   throws HibernateException, SQLException
   {
      Date timestamp = rs.getDate(names[0]);
      int offset     = rs.getShort(names[1]);
      
      datetimetz ret = new datetimetz();
      ret.setTimeZoneOffset(offset);
      ret.setTimestamp(timestamp);
      return ret;
   }
   
   /**
    * Write an instance of the <code>datetimetz</code> class to a prepared statement. 
    * A multi-column type should be written to parameters starting from <tt>index</tt>.
    *
    * @param   st
    *          A JDBC prepared statement.
    * @param   value
    *          The object to be written.
    * @param   index
    *          The statement parameter index.
    * @param   sess
    * 
    * @throws HibernateException
    * @throws SQLException
    */
   @Override
   public void nullSafeSet(PreparedStatement st, Object value, int index, SessionImplementor sess)
   throws HibernateException, SQLException
   {
      if (value == null)
      {
         st.setObject(index, null);
         st.setObject(index + 1, null);
      }
      
      datetimetz dtz = (datetimetz) value;
      if (dtz == null || dtz.isUnknown())
      {
         st.setObject(index, null);
         st.setObject(index + 1, null);
      }
      else
      {
         st.setTimestamp(index, new Timestamp(dtz.dateValue().getTime()));
         st.setInt(index + 1, dtz.getOffset());
      }
   }

   /**
    * Returns <code>true</code> to indicate that the P2J data wrapper object is mutable.
    *
    * @return  <code>true</code>.
    */
   @Override
   public boolean isMutable()
   {
      return true;
   }

   /**
    * Transform the object into its cacheable representation. This method returns a deep copy 
    * of the passed object. 
    *
    * @param   value
    *          The object to be cached.
    * @param   session
    *          The originating session.
    * @return  A cacheable representation of the object.
    * 
    * @throws HibernateException
    */
   @Override
   public Serializable disassemble(Object value, SessionImplementor session)
   throws HibernateException
   {
      return (Serializable) deepCopy(value);
   }

   /**
    * Reconstruct an object from the cacheable representation. 
    *
    * @param   cached
    *          The object to be cached.
    * @param   session
    *          The originating session.
    * @param   owner 
    *          The owner of the cached object.
    * 
    * @return  a reconstructed object from the cacheable representation.
    * 
    * @throws HibernateException
    */
   @Override
   public Object assemble(Serializable cached, SessionImplementor session, Object owner)
   throws HibernateException
   {
      return deepCopy(cached);
   }

   /**
    * During merge, replace the existing (target) value in the entity we are merging to
    * with a new (original) value from the detached entity we are merging. For immutable
    * objects, or null values, it is safe to simply return the first parameter. For
    * mutable objects, it is safe to return a copy of the first parameter. However, since
    * composite user types often define component values, it might make sense to recursively 
    * replace component values in the target object.
    * 
    * @param   original
    *          The original object to be replaced.
    * @param   target 
    *          The value in the managed entity.
    * @param   session
    *          The originating session.
    * @param   owner 
    *          The owner of the value.
    * 
    * @return  Because <code>datetimetz</code> is mutable, a deep copy of the 
    *          <code>original</code> is returned.
    * 
    * @throws  HibernateException 
    *          Is not thrown.
    */
   @Override
   public Object replace(Object original, Object target, SessionImplementor session, Object owner)
   throws HibernateException
   {
      return deepCopy(original);
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.datetimetz P2J datetimetz} class.
    */
   @Override
   public Class returnedClass()
   {
      return datetimetz.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>datetimetz</code> instance.
    *
    * @param   value
    *          The <code>datetimetz</code> instance to be copied.
    *
    * @return  A new instance of <code>datetimetz</code> which contains the
    *          same internal value as <code>value</code>.
    */
   @Override
   public Object deepCopy(Object value)
   {
      return new datetimetz((datetimetz) value);
   }
}