/*
** Module   : TypeHelper.java
** Abstract : Central, bi-directional mapping of P2J types to SQL types
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 ECF 20140624 Created initial version.
** 002 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 003 ECF 20170917 Added recid.
** 004 CA  20180517 Added comhandle data type.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Central, bi-directional mapping of P2J types to SQL types.
 */
public final class TypeHelper
{
   /** Map of P2J types to SQL types */
   private static final Map<Class<?>, Integer> types = new HashMap<>();
   
   static
   {
      types.put(character.class, Types.VARCHAR);
      types.put(comhandle.class, Types.VARCHAR);
      types.put(date.class     , Types.DATE);
      types.put(datetime.class , Types.TIMESTAMP);
      types.put(decimal.class  , Types.NUMERIC);
      types.put(handle.class   , Types.VARCHAR);
      types.put(int64.class    , Types.BIGINT);
      types.put(integer.class  , Types.INTEGER);
      types.put(logical.class  , Types.BIT);
      types.put(raw.class      , Types.VARBINARY);
      types.put(recid.class    , Types.INTEGER);
      types.put(rowid.class    , Types.BIGINT);
   }
   
   /**
    * Get the SQL type code for the given P2J wrapper type.
    * 
    * @param   wrapperType
    *          P2J type
    * 
    * @return  Corresponding SQL type.
    * 
    * @throws  IllegalArgumentException
    *          if <code>wrapperType</code> is not recognized.
    */
   public static int getSqlType(Class<?> wrapperType)
   {
      Integer sqlType = types.get(wrapperType);
      if (sqlType == null)
      {
         throw new IllegalArgumentException("Unrecognized SQL mapping: " + wrapperType.getName());
      }
      
      return sqlType;
   }
   
   /**
    * Get the P2J wrapper type for the given SQL type code.
    * 
    * <table border="1" summary="">
    * <thead>
    * <tr>
    *    <th>4GL type name</th>
    *    <th>P2J wrapper for 4GL types</th>
    *    <th>SQL type</th>
    * </tr>
    * </thead>
    * <tbody>
    * <tr>
    *    <td>character</td>
    *    <td>{@link character}</td>
    *    <td>{@link Types#VARCHAR}</td>
    * </tr>
    * <tr>
    *    <td>comhandle</td>
    *    <td>{@link handle}</td>
    *    <td>n/a</td>
    * </tr>
    * <tr>
    *    <td>datetime-tz</td>
    *    <td>{@link datetimetz}</td>
    *    <td>n/a</td>
    * </tr>
    * <tr>
    *    <td>datetime</td>
    *    <td>{@link datetime}</td>
    *    <td>{@link Types#TIMESTAMP}</td>
    * </tr>
    * <tr>
    *    <td>date</td>
    *    <td>{@link date}</td>
    *    <td>{@link Types#DATE}</td>
    * </tr>
    * <tr>
    *    <td>decimal</td>
    *    <td>{@link decimal}</td>
    *    <td>{@link Types#NUMERIC}</td>
    * </tr>
    * <tr>
    *    <td>handle</td>
    *    <td>{@link handle}</td>
    *    <td>n/a</td>
    * </tr>
    * <tr>
    *    <td>int64</td>
    *    <td>{@link int64}</td>
    *    <td>{@link Types#BIGINT}</td>
    * </tr>
    * <tr>
    *    <td>integer</td>
    *    <td>{@link integer}</td>
    *    <td>{@link Types#INTEGER}</td>
    * </tr>
    * <tr>
    *    <td>logical</td>
    *    <td>{@link logical}</td>
    *    <td>{@link Types#BIT}</td>
    * </tr>
    * <tr>
    *    <td>raw</td>
    *    <td>{@link raw}</td>
    *    <td>{@link Types#VARBINARY}</td>
    * </tr>
    * <tr>
    *    <td>rowid</td>
    *    <td>{@link rowid}</td>
    *    <td>n/a</td>
    * </tr>
    * <tr>
    *    <td>recid</td>
    *    <td>{@link recid}</td>
    *    <td>n/a</td>
    * </tr>
    * </tbody>
    * </table>
    * 
    * @param   sqlType
    *          SQL type code.
    * 
    * @return  P2J wrapper class, or <code>null</code> if no match.
    */
   public static Class<?> getP2JType(int sqlType)
   {
      switch (sqlType)
      {
         case Types.VARCHAR:  
             // TODO: handle and comhandle uses same SQL mapping; do we need support?
            return character.class;
         case Types.TIMESTAMP:
            return datetime.class;
         case Types.DATE:
            return date.class;
         case Types.NUMERIC:
            return decimal.class;
         case Types.BIGINT:   // TODO: rowid uses same SQL mapping; do we need support?
            return int64.class;
         case Types.INTEGER:  // TODO: recid uses same SQL mapping; do we need support?
            return integer.class;
         case Types.BIT:
            return logical.class;
         case Types.VARBINARY:
            return raw.class;
      }
      
      return null;
   }
}