/*
** Module   : filesview.js
** Abstract : File filter view.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20170609 First version.
**     ECF 20170707 Reworked with different UI controls.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Module which implements file filter profiles. Allows the user to create, edit, save, and
 * delete file filter profiles. The default profile, which includes everything, cannot be
 * deleted.
 */
define(["wsbase",
        "cbtree/store/Hierarchy",
        "cbtree/model/ForestStoreModel",
        "cbtree/Tree",
        "dojo/dom",
        "dojo/dom-construct",
        "dojo/dom-class",
        "dijit/layout/ContentPane",
        "dojox/layout/TableContainer",
        "dijit/form/Button",
        "dijit/form/TextBox",
        "dijit/ConfirmDialog",
        "dojo/store/Memory",
        "dojo/domReady!"],
function(ws, Hierarchy, ForestStoreModel, Tree, Dom, DomConstruct, DomClass, ContentPane,
         TableContainer, Button, TextBox, ConfirmDialog, Memory)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Starting value for report-related message codes. */
   var MSG_BASE = ws.MSG_FILT_BASE;
   
   /** Message types. */
   var msgTypes =
   {
      FILTER_INFO          :  MSG_BASE + 1,
      SET_FILTER           :  MSG_BASE + 2,
      FILTER_SPEC          :  MSG_BASE + 3,
      ALL_SRC_FILES        :  MSG_BASE + 4,
      DEF_PROFILE          :  MSG_BASE + 5,
      DEL_PROFILE          :  MSG_BASE + 6,
   }
   
   /** ID of profile which currently is active */
   var activeProfileId;
   
   /** ID of default filter profile */
   var defaultProfileId;
   
   /** ID of profile currently selected in tabulator or being created, if any */
   var currentProfileId = -1;
   
   /** File/path data which backs the check box tree store */
   var fileData;
   
   /** Model which backs the check box tree */
   var treeModel;
   
   /** Tree widget */
   var tree;
   
   /** Activate selected profile button */
   var activateButton;
   
   /** Create new profile button */
   var createButton;
   
   /** Edit selected profile button */
   var editButton;
   
   /** Delete selected profile button */
   var deleteButton;
   
   /** Save profile edits button */
   var saveButton;
   
   /** Cancel profile edits button */
   var cancelButton;
   
   /** Text box containing name/description of current profile */
   var nameTextBox;
   
   /** Has any tree node been modified since entering edit mode? */
   var treeDirty = false;
   
   /**
    * Up-call to the server to get the list of filter profiles.
    *
    * @param    callback
    *           Worker routine that will be called with the returned data. This routine will
    *           be passed two <code>null</code> values on any failure.  On success, the
    *           first parameter of the callback will be an array of FilterProfile objects
    *           (each with name, id and isDefault members) and the second parameter is the
    *           id of the currently selected filter profile.
    */
   me.readFilterProfileList = function(callback)
   {
      ws.sendMessage(msgTypes.FILTER_INFO, null, function(err, data)
      {
         var len       = data.length;
         var filters   = null;
         var currentId = null;
         
         if (len == 2)
         {
            filters   = data[0];
            currentId = data[1];
         }
         
         callback(filters, currentId);
      });
   }
   
   /**
    * Initialize the source file data and user interface.
    * 
    * @param   {Function} callback
    *          Callback to be invoked if retrieval of source file data from the server completed
    *          successfully.
    */
   function initialize(callback)
   {
      if (fileData !== undefined)
      {
         callback();
         
         return;
      }
      
      // request all source file names from the server to initialize the tree data store
      ws.sendMessage(msgTypes.ALL_SRC_FILES, null, function(err, files)
      {
         if (fileData === undefined)
         {
            if (err)
            {
               fwd.showDialog("Error", err);
               
               return;
            }
            
            fileData = [];
            var map = new Map();
            files.forEach(function(path)
            {
               var datum;
               var parts = path.split("/");
               var parent;
               var parentId = ".";
               var len = parts.length;
               for (var i = 1; i < len; i++)
               {
                  var nextPart = parts[i];
                  
                  if (i < len - 1)
                  {
                     // handle directory parts
                     
                     var array = [ parentId, nextPart ];
                     var id = array.join("/");
                     
                     datum = map.get(id);
                     if (datum === undefined)
                     {
                        var type = i == 1 ? "primary" : "secondary";
                        datum = { id : id, name : nextPart, type : type };
                        if (parent !== undefined)
                        {
                           datum.parent = parentId;
                        }
                        map.set(id, datum);
                        fileData.push(datum);
                     }
                     
                     parent = datum;
                     parentId = id;
                  }
                  else
                  {
                     // handle file part
                     
                     datum = { id : path, name : nextPart, type : "file", parent : parentId };
                     fileData.push(datum);
                  }
               }
            });
         }
         
         // now that we have the data store ready, we can invoke the callback to continue setup
         callback();
      });
   }
   
   /**
    * Get the remaining browser client window height available for this view's UI widgets.
    * 
    * @return  {Number} Number of pixels of height available for this view's widgets.
    */
   function getAvailableHeight()
   {
      var topNavBar = $("#topNavBar");
      var mainView = $("#mainView");
      var mtop = topNavBar.css("margin-top");
      var mbot = topNavBar.css("margin-bottom");
      var tbor = topNavBar.css("border");
      var mbor = mainView.css("border");
      var offsets = (mtop ? parseInt(mtop) : 0)
                  + (mbot ? parseInt(mbot) : 10)
                  + ((tbor ? parseInt(tbor) : 1) * 2)
                  + ((mbor ? parseInt(mbor) : 3) * 2)
                  + 6; // slop
      
      var pixels = $(window).height() - $("#topNavBar").height() - offsets;
      
      return pixels;
   }
   
   /**
    * Set up and display the file filter profile view. This is the entry point into this module
    * which does all the UI setup and server communication necessary to create the view.
    */
   me.setupAndDisplay = function()
   {
      var parent = $("#filesView");
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      var margin = parseInt(containDiv.css("margin-top"))
                 + parseInt(containDiv.css("margin-bottom"));
      containDiv.css("height", (getAvailableHeight() - margin) + "px");
      
      // append attach point for first row
      var row1Div = $(div).addClass("row");
      $(containDiv).append(row1Div);
      
      // retrieve profile info (all profile defs + active profile ID)
      ws.sendMessage(msgTypes.FILTER_INFO, null, function(err, info)
      {
         if (err)
         {
            fwd.showDialog("Error", fwd.errorMessages(err)[0]);
            
            return;
         }
         
         // info is an array of two elements:
         // 0: an array of profile definitions
         // 1: id of the currently active profile
         var rows = info[0];
         activeProfileId = info[1];
         
         // the first time we fetch this information, the active profile will be the default
         // profile; remember this ID
         if (defaultProfileId === undefined)
         {
            defaultProfileId = activeProfileId;
         }
         
         // add "active" column to row data
         rows.forEach(function(row)
         {
            row.active = row.id == activeProfileId;
         });
         
         // add table of filter profiles and button panel to manipulate them on the left side
         // of the screen
         
         // attach point for profile list table
         var r1c1Div = $(div).addClass("col-xs-12 col-sm-6")
                             .css("padding", "3px")
                             .css("height", (containDiv.height() * 0.9) + "px");
         row1Div.append(r1c1Div);
         
         // attach point for profile editing panel
         var r1c2Div = $(div).attr("id", "editProfileView")
                             .addClass("col-xs-12 col-sm-6")
                             .css("padding", "3px")
                             .css("height", (containDiv.height() * 0.9) + "px");
         row1Div.append(r1c2Div);
         
         fwd.showOverlay("Building View", false);
         
         initialize(initializeProfileEditor);
         
         fwd.hideOverlay();
         
         // append title
         var titleDiv = $(div).addClass("toc-heading").text("File List Filter Profiles");
         r1c1Div.append(titleDiv);
         
         var tabRowDiv = $(div).addClass("row");
         r1c1Div.append(tabRowDiv);
         
         var tabColDiv = $(div).addClass("col-xs-12");
         tabRowDiv.append(tabColDiv);
         
         var tabDiv = $(div).attr("id", "profileTable");
         tabColDiv.append(tabDiv);
         
         var btnRowDiv = $(div).addClass("row").css("height", (row1Div.height() * 0.1) + "px");
         r1c1Div.append(btnRowDiv);
         
         var w = dijit.byId("profsBtnPanel");
         if (w)
         {
            w.destroy(true);
         }
         
         btnRowDiv.append($(div).attr("id", "profsBtnPanel")
                                .addClass("col-xs-12")
                                .css("padding", "15px"));
         var profsBtnPanel = new ContentPane({}, "profsBtnPanel");
         
         activateButton = new Button(
         {
            label    : "Activate Selected",
            disabled : true,
            onClick  : function()
                       {
                          var oldRow = tabDiv.tabulator("getRowData", activeProfileId);
                          var newRow = tabDiv.tabulator("getSelectedData")[0];
                          activateProfile(oldRow, newRow);
                       },
         });
         
         createButton = new Button(
         {
            label    : "Create Profile",
            onClick  : createProfile,
         });
         
         editButton = new Button(
         {
            label    : "Edit Profile",
            disabled : true,
            onClick  : function() { loadProfile(currentProfileId, true); },
         });
         
         deleteButton = new Button(
         {
            label    : "Delete Profile",
            disabled : true,
            onClick  : deleteCurrentProfile,
         });
         
         profsBtnPanel.addChild(activateButton);
         profsBtnPanel.addChild(createButton);
         profsBtnPanel.addChild(editButton);
         profsBtnPanel.addChild(deleteButton);
         profsBtnPanel.startup();
         
         var reselect = null;
         
         // initialize the tabulator
         tabDiv.tabulator(
         {
            fitColumns           :  true, // fit columns to width of table
            height               :  r1c1Div.height() - titleDiv.height(),
            selectable           :  1,
            sortable             :  false,
            columns              :
            [
               {
                  title          :  "Profile Name/Description",
                  field          :  "name",
                  tooltip        :  true,
               },
               {
                  title          :  "Active",
                  field          :  "active",
                  align          :  "center",
                  width          :  "15%",
                  formatter      :  "tick",
               },
               {
                  title          :  "Default",
                  field          :  "default",
                  align          :  "center",
                  width          :  "15%",
                  formatter      :  "tick",
               },
            ],
            rowSelectionChanged  :  function(data, rows)
                                    {
                                       if (data.length == 0 && reselect != null)
                                       {
                                          // if reselect is not null, it means a click has
                                          // deselected a selected row, but we want it to remain
                                          // selected, so we manually force selection on it
                                          tabDiv.tabulator("selectRow", reselect);
                                          reselect = null;
                                       }
                                       else if (reselect == null)
                                       {
                                          if (data.length == 0)
                                          {
                                             activateButton.set("disabled", true);
                                             editButton.set("disabled", true);
                                             deleteButton.set("disabled", true);
                                             
                                             return;
                                          }
                                          
                                          var newId = data[0].id;
                                          
                                          // if row is newly selected, enable disabled buttons
                                          // and display the profile details for the selected row
                                          activateButton.set("disabled", data[0].active);
                                          editButton.set("disabled", data[0].default);
                                          deleteButton.set("disabled", data[0].default);
                                          if (treeModel)
                                          {
                                             if (treeDirty)
                                             {
                                                cancelEdit();
                                             }
                                             loadProfile(newId);
                                          }
                                          
                                          currentProfileId = newId;
                                       }
                                    },
            rowClick             :  function(e, id, data, row)
                                    {
                                       // A click will both select and de-select a row, but we
                                       // want a selected row to stay selected on subsequent
                                       // clicks, and we don't want to reload a filter profile
                                       // if it's already loaded. So, if we determine a click
                                       // has deselected a row, remember the row data and let
                                       // the rowSelectionChanged handler re-select it
                                       var selected = tabDiv.tabulator("getSelectedData");
                                       reselect = selected.length > 0 && selected[0] === data
                                                  ? data
                                                  : null;
                                    },
            tableBuilt           :  function()
                                    {
                                       if (rows.length > 0)
                                       {
                                          tabDiv.tabulator("selectRow", activeProfileId);
                                       }
                                    },
            data                 :  rows,
         });
         
         // re-render table on window resize
         $(window).resize(function()
         {
            tabDiv.tabulator("redraw");
         });
      });
      
      $("#filesView").show();
   }
   
   /**
    * Create a new filter profile and leave the user interface for defining it in edit mode.
    */
   function createProfile()
   {
      fwd.showOverlay("Creating Profile", "fast");
      
      // deselect the current row in the tabulator
      $("#profileTable").tabulator("deselectRow", currentProfileId);
      
      // pass a non-existent profile ID to load the file specs of the default profile as a
      // starting point for edits
      currentProfileId = -1;
      loadProfile(currentProfileId, true);
      
      fwd.hideOverlay();
   }
   
   /**
    * Make the filter profile with the specified ID the active profile, which will be used to
    * filter all report results.
    * 
    * @param   {Object} oldProfile
    *          Profile which is about to be deactivated.
    * @param   {Object} newProfile
    *          Profile which is about to be activated.
    */
   function activateProfile(oldProfile, newProfile)
   {
      activateButton.set("disabled", true);
      
      var newPid = newProfile.id;
      var force = oldProfile === newProfile;
      
      fwd.showOverlay("Activating Profile", "fast");
      
      ws.sendMessage(msgTypes.SET_FILTER, [ newPid, force ], function(err)
      {
         fwd.hideOverlay();
         
         if (err)
         {
            fwd.showDialog("Error", fwd.errorMessages(err)[0]);
         }
         
         var tab = $("#profileTable");
         
         if (oldProfile && !force)
         {
            oldProfile.active = false;
            tab.tabulator("updateRow", oldProfile.id, oldProfile);
         }
         
         newProfile.active = true;
         tab.tabulator("updateRow", newProfile.id, newProfile);
         
         tab.tabulator("redraw", true);
         
         activeProfileId = newPid;
      });
   }
   
   /**
    * Manage the enabled state of various UI widgets to reflect the current editing state of the
    * user interface. This prevents certain user operations from being invoked at inappropriate
    * times.
    * 
    * @param   {Boolean} enable
    *          True if the user interface is in profile edit mode, else false.
    * @param   {Number} pid
    *          Profile ID of the currently selected or edited profile (not the same as the active
    *          profile).
    */
   function enableProfileEdit(enable, pid)
   {
      if (!pid || pid < 0)
      {
         pid = currentProfileId;
      }
      
      var isActive = pid == activeProfileId;
      var isDefault = pid == defaultProfileId;
      
      activateButton.set("disabled", enable || isActive);
      editButton.set("disabled", enable || isDefault);
      deleteButton.set("disabled", enable || isDefault);
      
      createButton.set("disabled", enable);
      
      if (saveButton && cancelButton && nameTextBox)
      {
         saveButton.set("disabled", !enable);
         cancelButton.set("disabled", !enable);
         nameTextBox.set("disabled", !enable);
         if (enable)
         {
            nameTextBox.focus();
         }
      }
      
      treeModel.getRoot(function(root)
      {
         var iterate = function(node)
         {
            treeModel.setEnabled(node, enable);
            treeModel.getChildren(node, function(children)
            {
               children.forEach(iterate);
            });
         };
         
         // start the recursive iteration
         iterate(root);
      });
   }
   
   /**
    * Validate and persist the state of the profile currently being edited.
    */
   function saveEdit()
   {
      var filters = [];
      
      // gather filter specs, based on which nodes of the tree are checked; checked nodes
      // represent files/directories to be included by the filter profile
      treeModel.getRoot(function(root)
      {
         var iterate = function(node)
         {
            if (treeModel.getChecked(node) === true)
            {
               var isBranch = treeModel.mayHaveChildren(node);
               var spec = getQualifiedName(node, root, isBranch);
               
               filters.push({ spec : spec, wild : isBranch });
            }
            else
            {
               treeModel.getChildren(node, function(children)
               {
                  children.forEach(iterate);
               });
            }
         };
         
         // start the recursive iteration
         iterate(root);
      });
      
      // confirm the user has entered a name and that it is unique, so we don't run afoul of
      // the database's unique constraint
      
      var name = nameTextBox.get("value");
      var valid = name && name.trim().length > 0;
      if (valid)
      {
         var profiles = $("#profileTable").tabulator("getData");
         
         for (var i = 0; i < profiles.length && valid; i++)
         {
            var profile = profiles[i];
            if (name == profile.name && currentProfileId != profile.id)
            {
               valid = false;
            }
         }
      }
      
      if (!valid)
      {
         fwd.showDialog("Error",
                        "Please give this profile a unique name/description.",
                        4000);
         nameTextBox.focus();
         
         return;
      }
      
      // confirm we have at least one file filter spec defined
      if (filters.length == 0)
      {
         fwd.showDialog("Error",
                        "Please include at least one file or directory in the filter.",
                        4000);
         
         return;
      }
      
      enableProfileEdit(false);
      
      // currentProfileId should be invalid at this point if we are creating a new profile, but
      // valid if we are editing an existing one
      var profile = { id : currentProfileId, name : name, default : false };
      
      // request the server to persist the new filter profile
      ws.sendMessage(msgTypes.DEF_PROFILE, [ profile, filters ], function(err, pid)
      {
         if (err)
         {
            fwd.showDialog("Error", fwd.errorMessages(err)[0]);
            
            return;
         }
         
         profile.id = pid;
         
         // tree is no longer dirty, since the profile was saved
         treeDirty = false;
         
         var tab = $("#profileTable");
         
         if (currentProfileId < 0)
         {
            // if we have created a new profile, add it to the tabulator and select it
            tab.tabulator("addRow", profile);
            tab.tabulator("selectRow", pid);
         }
         else
         {
            // if we have updated an existing profile, update and redraw the tabulator to
            // reflect any changes
            tab.tabulator("updateRow", pid, profile);
            tab.tabulator("redraw", true);
            
            // if existing profile was the active profile, re-activate it to pick up any file
            // filter changes
            if (pid == activeProfileId)
            {
               activateProfile(profile, profile);
            }
         }
      });
   }
   
   /**
    * Delete the currently selected profile after confirming with the user.
    */
   function deleteCurrentProfile()
   {
      var dialog = new ConfirmDialog(
      {
         title   : "Please Confirm",
         content : "Are you sure you want to delete the selected profile?",
      });
      dialog.show();
      
      dialog.on("execute", function()
      {
         ws.sendMessage(msgTypes.DEL_PROFILE, [ currentProfileId ], function(err, result)
         {
            if (err)
            {
               fwd.showDialog("Error", fwd.errorMessages(err)[0]);
               
               return;
            }
            
            if (result)
            {
               var tab = $("#profileTable");
               
               tab.tabulator("deleteRow", currentProfileId);
               
               var wasActive = currentProfileId == activeProfileId;
               
               // select the active profile as the new current profile (will be the default
               // profile if we just deleted the active one)
               tab.tabulator("selectRow", wasActive ? defaultProfileId : activeProfileId);
               
               // if we just deleted the active profile, we have to set the default profile as
               // the new active profile
               if (wasActive)
               {
                  var newProfile = tab.tabulator("getRowData", defaultProfileId);
                  activateProfile(null, newProfile);
               }
            }
            else
            {
               console.log("Current filter profile " + currentProfileId + " was not deleted!");
            }
         });
      });
   }
   
   /**
    * Cancel the current profile edit operation, and optionally reload the profile with the
    * specified ID into the profile viewer.
    * 
    * @param   {Number} reloadPid
    *          Optional parameter indicating the ID of the profile whose definition is to be
    *          reloaded into the profile viewer after cancelling the edit operation.
    */
   function cancelEdit(reloadPid)
   {
      var doIt = function()
      {
         if (reloadPid)
         {
            // disable editing and reset tree state to its pre-dirty condition
            loadProfile(reloadPid);
         }
         else
         {
            // otherwise just disable editing
            enableProfileEdit(false);
         }
         
         treeDirty = false;
      };
      
      if (treeDirty)
      {
         var dialog = new ConfirmDialog(
         {
            title   : "Please Confirm",
            content : "Are you sure you want to discard your changes to this profile?",
         });
         dialog.show();
         
         dialog.on("execute", doIt);
      }
      else
      {
         doIt();
      }
   }
   
   /**
    * Compose a qualified file or directory specification name from the given tree node.
    * 
    * @param   {Object} node
    *          A checkbox tree node.
    * @param   {Object} root
    *          The root node of the checkbox tree.
    * @param   {Boolean} isBranch
    *          True if the given node represents a branch of the tree; false if a leaf.
    * 
    * @return  {String} Qualified specification for the file or directory represented by the
    *          tree node.
    */
   function getQualifiedName(node, root, isBranch)
   {
      var parts = isBranch ? ["*"] : [];
      var next = node;
      while (next != null)
      {
         parts.unshift(next.name);
         var promise = treeModel.getParents(next);
         promise.then(function(parents)
         {
            if (parents)
            {
               next = parents[0] === root ? null : parents[0];
            }
         });
      }
      parts.unshift(".");
      
      var qname = "";
      parts.forEach(function(elem, i, a)
      {
         qname += elem;
         if (i < a.length - 1)
         {
            qname += "/";
         }
      });
      
      return qname;
   };
   
   /**
    * Load the definition of the filter profile associated with the given ID, and optionally
    * enable that definition to be edited.
    * 
    * @param   {Number} pid
    *          Profile ID.
    * @param   {Boolean} edit
    *          True to allow the profile to be edited; false if it should be read-only.
    */
   function loadProfile(pid, edit)
   {
      fwd.showOverlay("Loading Profile", [100, 200]);
      
      if (edit == undefined)
      {
         edit = false;
      }
      
      enableProfileEdit(edit, pid);
      
      var name = "";
      
      if (pid < 0)
      {
         // this will be the case if we are editing a newly created profile which has not yet
         // been persisted
         pid = defaultProfileId;
      }
      else
      {
         var profile = $("#profileTable").tabulator("getRowData", pid);
         name = profile.name;
      }
      
      nameTextBox.set("value", name);
      
      ws.sendMessage(msgTypes.FILTER_SPEC, [ pid ], function(err, data)
      {
         var specSet = new Set();
         
         data.forEach(function(fileSpec)
         {
            var spec = fileSpec.spec;
            if (fileSpec.wild)
            {
               // trim off trailing file separator and wildcard character
               spec = spec.substring(0, spec.length - 2);
            }
            
            specSet.add(spec);
         });
         
         treeModel.getRoot(function(root)
         {
            var iterate = function(node)
            {
               var checkState = specSet.has(node.id);
               treeModel.setChecked(node, checkState);
               tree.getNodesByItem(node).forEach(function(n)
               {
                  if (n && n.isExpandable)
                  {
                     if (checkState)
                     {
                        tree._collapseNode(n);
                     }
                  }
               });
               if (!checkState)
               {
                  treeModel.getChildren(node, function(children)
                  {
                     children.forEach(iterate);
                  });
                  
                  // if we've come back up from a child walk and we're still not selected,
                  // collapse this branch, otherwise expand it
                  tree.getNodesByItem(node).forEach(function(n)
                  {
                     if (n && n.item !== root && n.isExpandable)
                     {
                        if (!treeModel.getChecked(node))
                        {
                           tree._collapseNode(n);
                        }
                        else
                        {
                           tree._expandNode(n);
                        }
                     }
                  });
               }
            };
            
            // start the recursive iteration
            iterate(root);
         });
         
         fwd.hideOverlay();
      });
   }
   
   /**
    * Initialize the UI panel which enables viewing and editing of filter profile definitions.
    */
   function initializeProfileEditor()
   {
      var div = "<div></div>";
      var attach = $("#editProfileView");
      
      // append title
      var titleDiv = $(div).addClass("toc-heading").text("Define File List Filter Profile");
      attach.append(titleDiv);
      
      // append edit panel
      var panelDiv = $(div).addClass("form-panel")
                           .attr("id", "editPanel");
      attach.append(panelDiv);
      
      var w = dijit.byId("editProfNamePanel");
      if (w)
      {
         w.destroy(true);
      }
      
      var nameDiv = panelDiv.append($(div).attr("id", "editProfNamePanel")
                                          .css("width", "100%"));
      var editProfNamePanel = new ContentPane({}, "editProfNamePanel");
      
      // create table container for text box
      var table = new TableContainer(
      {
         cols         : 1,
         labelWidth   : 200,
         customClass  : "customReportTable",
      });
      
      nameTextBox = new TextBox(
      {
         label       : "Name/Description",
         placeHolder : "Enter a unique, descriptive name for the filter profile",
         style       : "width: 30em;",
         disabled    : true,  // initially disabled, until edit mode
      });
      
      table.addChild(nameTextBox);
      editProfNamePanel.addChild(table);
      editProfNamePanel.startup();
      
      var prompt =
         "<p>Files and directories selected in the tree are included in report results:<p>";
      var textDiv = panelDiv.append($(prompt).css("font-weight", "bold")
                                             .css("padding", "0 5px 0 5px"));
      
      var treeDiv = $(div).attr("id", "treePanel");
      panelDiv.append(treeDiv);
      
      tree = dijit.byId("filesSpecTree");
      if (tree)
      {
         tree.destroy(true);
      }
      
      var store = new Hierarchy({ data: fileData });
      
      // use the ForestStoreModel, since we usually will have multiple primary directories
      // (e.g., "abl", "data")
      treeModel = new ForestStoreModel(
      {
         store        : store,
         query        : { type : "primary" },
         enabledAttr    : "writable",
      });
      
      treeModel.ready(function() { loadProfile(activeProfileId); });
      
      tree = new Tree(
      {
         model             : treeModel,
         id                : "filesSpecTree",
         showRoot          : false,
         checkedRoot       : true,
         onCheckBoxClick   : function() { treeDirty = true; },
      }, "treePanel");
      
      tree.startup();
      
      // calculate available height for tree control, so that it scrolls vertically within a
      // fixed height container
      var offset = 0;
      panelDiv.children().each(function(i, child) { offset += child.offsetHeight; });
      var height = attach.outerHeight() - titleDiv.height() - offset;
      $("#filesSpecTree").css("height", height);
      
      // add edit-related buttons
      
      var w = dijit.byId("editProfBtnPanel");
      if (w)
      {
         w.destroy(true);
      }
      
      var btnRowDiv = $(div).addClass("row").css("height", (attach.height() * 0.1) + "px");
      attach.append(btnRowDiv);
      
      btnRowDiv.append($(div).attr("id", "editProfBtnPanel")
                             .addClass("col-xs-12")
                             .css("padding", "15px"));
      var editProfBtnPanel = new ContentPane({}, "editProfBtnPanel");
      
      saveButton = new Button(
      {
         label    : "Save Changes",
         disabled : true,
         onClick  : saveEdit,
      });
      
      cancelButton = new Button(
      {
         label    : "Cancel Editing",
         disabled : true,
         onClick  : function()
                    {
                       if (currentProfileId < 0)
                       {
                          var selected = $("#profileTable").tabulator("getSelectedData");
                          if (selected && selected.length > 0)
                          {
                             currentProfileId = selected[0].id;
                          }
                          else
                          {
                             currentProfileId = activeProfileId;
                          }
                       }
                       cancelEdit(currentProfileId);
                    },
      });
      
      editProfBtnPanel.addChild(saveButton);
      editProfBtnPanel.addChild(cancelButton);
      editProfBtnPanel.startup();
   }
   
   return me;
});