/*
** Module   : listing.js
** Abstract : Source file listing view module.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170727 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Source listing view. Displays a table with the source files and a file viewer.
 */
define(["wsbase"], function(ws)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Starting value for report-related message codes. */
   var MSG_BASE = ws.MSG_LSTG_BASE;
   
   /** Message types, which correspond with those defined in the server. */
   var msgTypes =
   {
      FILE_LISTING         :  MSG_BASE + 1,
   }
   
   /**
    * Show the source listing view. This consists of a tabulator in the left half of the screen
    * and a text prompt for the user to click on the table in the right half. When the user
    * clicks on a cell with a file icon in it, the prompt is replaced with a "raw" source file
    * viewer, if the user clicked on an icon of an unpreprocessed file. If the user clicked on
    * an icon for a schema (*.df) file or a preprocessed file, the entire view is hidden and
    * replaced with a combined source/AST view.
    */
   me.show = function()
   {
      var parent = $("#listingView");
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      
      // append attach point for first row
      var row1Div = $(div).addClass("row");
      $(containDiv).append(row1Div);
      
      fwd.showOverlay("Building Source Listing");
      
      ws.sendMessage(msgTypes.FILE_LISTING, null, function(err, listing)
      {
         fwd.hideOverlay();
         
         if (err)
         {
            fwd.showDialog("Error", err);
            
            return;
         }
         
         var types =
         {
            SCHEMA     : fwd.rpt.srcTypes.SCHEMA.toString(),
            CODE_CACHE : fwd.rpt.srcTypes.CODE_CACHE.toString(),
            CODE_BASE  : fwd.rpt.srcTypes.CODE_BASE.toString(),
         }
         
         // use the information in the types field to compute raw and prep fields for the table
         listing.forEach(function(row)
         {
            if (row.types.indexOf(types.SCHEMA) >= 0 || row.types.indexOf(types.CODE_BASE) >= 0)
            {
               row.raw = "<span class='glyphicon glyphicon-file' style='color: #f70'></span>";
            }
            
            if (row.types.indexOf(types.CODE_CACHE) >= 0)
            {
               row.prep = "<span class='glyphicon glyphicon-file' style='color: #0a0'></span>";
            }
         });
         
         var r1c1Div = $(div).addClass("col-md-6").css("padding", "3px");
         row1Div.append(r1c1Div);
         
         var r1c2Div = $(div).addClass("col-md-6").css("padding", "3px");
         row1Div.append(r1c2Div);
         
         // append table title
         var tocTitleDiv = $(div).addClass("toc-heading").text("Source File Listing");
         r1c1Div.append(tocTitleDiv);
         
         var height = fwd.getAvailableHeight() - 16;
         promptForClick(r1c2Div, height);
         
         var tabDiv = $(div).attr("id", "listingTable");
         r1c1Div.append(tabDiv);
         
         var reselect = null;
         
         // initialize the tabulator
         tabDiv.tabulator(
         {
            fitColumns           :  true, // fit columns to width of table
            height               :  height + "px",
            pagination           :  "local",
            paginationSize       :  250,
            selectable           :  1,
            sortable             :  false,
            tooltipsHeader       :  true,
            columns              :
            [
               {
                  title          :  "Source File Name",
                  field          :  "name",
                  tooltip        :  true,
                  headerFilter   :  true,
               },
               {
                  title          :  "Base File",
                  field          :  "raw",
                  align          :  "center",
                  width          :  "15%",
                  tooltip        :  function(field, value, data)
                                    {
                                       return value ? "Click to view unpreprocessed file" : "";
                                    },
                  formatter      :  function(value, data, cell, row, options, formatterParams)
                                    {
                                       return value;
                                    },
                  onClick        :  function(e, cell, value, data)
                                    {
                                       if (value)
                                       {
                                          if (data.types.indexOf(types.SCHEMA) >= 0)
                                          {
                                             // we want the AST view for DF files
                                             showSourceAstView(r1c2Div,
                                                               height,
                                                               data.id,
                                                               data.name,
                                                               fwd.rpt.srcTypes.SCHEMA);
                                          }
                                          else
                                          {
                                             // just show the raw source, which is not associated
                                             // with an AST
                                             showSourceView(r1c2Div,
                                                            height,
                                                            data.id,
                                                            data.name);
                                          }
                                       }
                                    },
               },
               {
                  title          :  "Preprocessed",
                  field          :  "prep",
                  align          :  "center",
                  width          :  "15%",
                  tooltip        :  function(field, value, data)
                                    {
                                       return value ? "Click to view preprocessed file" : "";
                                    },
                  formatter      :  function(value, data, cell, row, options, formatterParams)
                                    {
                                       return value;
                                    },
                  onClick        :  function(e, cell, value, data)
                                    {
                                       if (value)
                                       {
                                          showSourceAstView(r1c2Div,
                                                            height,
                                                            data.id,
                                                            data.name,
                                                            fwd.rpt.srcTypes.CODE_CACHE);
                                       }
                                    },
               },
            ],
            rowSelectionChanged  :  function(data, rows)
                                    {
                                       if (data.length == 0 && reselect != null)
                                       {
                                          // if reselect is not null, it means a click has
                                          // deselected a selected row, but we want it to remain
                                          // selected, so we manually force selection on it
                                          tabDiv.tabulator("selectRow", reselect);
                                          reselect = null;
                                       }
                                    },
            rowClick             :  function(e, id, data, row)
                                    {
                                       // A click will both select and de-select a row, but we
                                       // want a selected row to stay selected on subsequent
                                       // clicks. So, if we determine a click has deselected a
                                       // row, remember the row data and let the
                                       // rowSelectionChanged handler re-select it
                                       var selected = tabDiv.tabulator("getSelectedData");
                                       reselect = selected.length > 0 && selected[0] === data
                                                  ? data
                                                  : null;
                                    },
            tableBuilt           :  function()
                                    {
                                       // re-render table on window resize
                                       $(window).resize(function()
                                       {
                                          tabDiv.tabulator("redraw");
                                       });
                                    },
            data                 :  listing,
         });
      });
   }
   
   /**
    * Add to the given DOM node parent a prompt for the user to click on a file icon in the
    * table to bring up a source file view. The prompt text is vertically and horizontally
    * centered in its parent element.
    * 
    * @param   {Selection} attachDiv
    *          jQuery selection for the DOM element to which the prompt element will be attached.
    * @param   {Number} height
    *          Height in pixels of the parent element.
    */
   function promptForClick(attachDiv, height)
   {
      attachDiv.empty();
      var promptDiv = $("<div>Click a file icon from the table to view the file</div>")
                      .addClass("toc-heading");
      promptDiv.css(
      {
         "height"          : height + "px",
         "display"         : "flex",
         "align-items"     : "center",
         "justify-content" : "center"
      });
      attachDiv.append(promptDiv);
   }
   
   /**
    * Show a source view without an associated AST viewer. This is meant to view un-preprocessed
    * source files which have no associated AST.
    * 
    * @param   {Selection} parentDiv
    *          jQuery selection for the DOM element to which the prompt element will be attached.
    * @param   {Number} height
    *          Height in pixels of the parent element.
    * @param   {Number} fid
    *          Unique ID of the source file to be loaded into the viewer.
    * @param   {String} file
    *          Name of the source file, used only in the title above the viewer.
    */
   function showSourceView(parentDiv, height, fid, file)
   {
      parentDiv.empty();
      
      var div = "<div></div>";
      
      var tocDiv = $(div).text(file + " (Base File)").addClass("toc-heading");
      parentDiv.append(tocDiv);
      
      // append attach point for row
      var rowDiv = $(div).addClass("row no-padding");
      parentDiv.append(rowDiv);
      
      // append attach point for full-width column (source code view)
      var srcDiv = $(div).attr("id", "rawSrcView").addClass("col-xs-12");
      $(rowDiv).append(srcDiv);
      
      srcDiv.css(
      {
         "height"          : height + "px",
         "display"         : "flex",
         "align-items"     : "center",
         "justify-content" : "center"
      });
      
      var srcType = fwd.rpt.srcTypes.CODE_BASE;
      ws.sendMessage(fwd.rpt.msgTypes.SOURCE_DATA, [ fid, -1, srcType ], function(err, data)
      {
         fwd.srcview.show("#rawSrcView", height, data[0]);
      });
   }
   
   /**
    * Show the source and AST viewers side by side, with the specified file loaded into the
    * source viewer. The AST viewer initially is empty. An AST branch will be loaded when the
    * user clicks on the associated text in the source viewer. When this view is dismissed with
    * the "back" button, the raw (unpreprocessed) version of the same file is loaded into the
    * "raw" source file viewer which appears to the right of the source file listing table.
    * 
    * @param   {Selection} rawParentDiv
    *          jQuery selection of the "raw" source file viewer parent, used when the user
    *          clicks the "back" button after viewing the combined source/AST view.
    * @param   {Number} height
    *          Height allocated to the rawParentDiv element.
    * @param   {Number} fid
    *          Unique ID of the source file to be loaded into the viewer.
    * @param   {String} file
    *          Name of the source file, used only in the title above the viewer.
    * @param   {Number} srcType
    *          The type of source file being loaded into the combined source/AST view (SCHEMA or
    *          CODE_CACHE).
    */
   function showSourceAstView(rawParentDiv, height, fid, file, srcType)
   {
      $("#rawSrcView").empty();
      
      var isSchema = srcType == fwd.rpt.srcTypes.SCHEMA;
      var backFunc = function()
      {
         if (isSchema)
         {
            promptForClick(rawParentDiv, height);
         }
         else
         {
            showSourceView(rawParentDiv, height, fid, file);
         }
      };
      var desc = (srcType == fwd.rpt.srcTypes.SCHEMA ? "" : "Preprocessed");
      fwd.rpt.showSourceAstView("#containerView", fid, file, 0, srcType, 0, 0, desc, backFunc);
   }
   
   /* exports */
   return me;
});
