/*
** Module   : BrowseReportKey.java
** Abstract : Key that is unique across browses with the same structure and design template.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180207 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import java.util.*;

/**
 * Key that is unique across browses with the same structure and design template.
 */
class BrowseReportKey
{
   /** Calculated hash code. */
   private final int hashCode;

   /** Design template file name. */
   private final String designTemplateFile;

   /** Browse column keys. */
   private final BrowseColumnKey[] columnKeys;

   /**
    * Create key for a browse with the given structure and given design template.
    *
    * @param browse
    *        Browse.
    * @param designTemplateFile
    *        Design template file name.
    */
   BrowseReportKey(BrowseWidget browse, String designTemplateFile)
   {
      this.designTemplateFile = designTemplateFile;

      int colNo = browse.getNumColumns().intValue();
      columnKeys = new BrowseColumnKey[colNo];
      for (int i = 0; i < colNo; i++)
      {
         BrowseColumnWidget column =
               (BrowseColumnWidget) browse.getBrowseColumn(i + 1).unwrapBrowseColumn();

         BrowseColumnKey columnKey = new BrowseColumnKey(Browse.isRightAligned(column.config()),
                                                         column.getWidthPixels().intValue(),
                                                         column.getDataType().getValue());
         columnKeys[i] = columnKey;
      }

      //pre-calculate hash code
      int result = this.designTemplateFile == null ? 0 : this.designTemplateFile.hashCode();
      for (BrowseColumnKey columnKey : columnKeys)
      {
         result = 31 * result + columnKey.hashCode();
      }
      hashCode = result;
   }

   /**
    * Override the parent's implementation to return a hash code cached at construction.
    */
   @Override
   public int hashCode()
   {
      return hashCode;
   }

   /**
    * Compares two objects for equality.
    *
    * @param   o
    *          Object to compare with.
    *
    * @return  <code>true</code> if instances are equal and
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }

      if (!(o instanceof BrowseReportKey))
      {
         return false;
      }
      BrowseReportKey that = (BrowseReportKey) o;

      boolean match = Objects.equals(designTemplateFile, that.designTemplateFile) &&
                      columnKeys.length == that.columnKeys.length;
      if (!match)
      {
         return false;
      }

      for (int i = 0; i < columnKeys.length; i++)
      {
         if (!columnKeys[i].equals(that.columnKeys[i]))
         {
            return false;
         }
      }

      return true;
   }

   /**
    * Key that is unique across browse columns with the same structure.
    */
   private class BrowseColumnKey
   {
      /** Calculated hash code. */
      private final int hashCode;

      /** Column data type. */
      private final String dataType;

      /** Width of the column in pixels. */
      private final int widthPixels;

      /** <code>true</code> if the column is right-aligned. */
      private final boolean rightAligned;

      /**
       * Create key for a browse column with the given structure.
       *
       * @param rightAligned
       *        <code>true</code> if the column is right-aligned.
       * @param widthPixels
       *        Width of the column in pixels.
       * @param dataType
       *        Column data type.
       */
      BrowseColumnKey(boolean rightAligned, int widthPixels, String dataType)
      {
         this.rightAligned = rightAligned;
         this.widthPixels = widthPixels;
         this.dataType = dataType;

         //pre-calculate hash code
         int result = dataType.hashCode();
         result = 31 * result + ((Integer) widthPixels).hashCode();
         hashCode = 31 * result + ((Boolean) rightAligned).hashCode();
      }

      /**
       * Override the parent's implementation to return a hash code cached at construction.
       */
      @Override
      public int hashCode()
      {
         return hashCode;
      }

      /**
       * Compares two objects for equality.
       *
       * @param   o
       *          Object to compare with.
       *
       * @return  <code>true</code> if instances are equal and
       *          <code>false</code> otherwise.
       */
      @Override
      public boolean equals(Object o)
      {
         if (this == o)
         {
            return true;
         }

         if (!(o instanceof BrowseColumnKey))
         {
            return false;
         }
         BrowseColumnKey that = (BrowseColumnKey) o;

         return dataType.equals(that.dataType) &&
                rightAligned == that.rightAligned &&
                widthPixels == that.widthPixels;
      }
   }
}
