/*
** Module   : FwdJasperRepositoryService.java
** Abstract : Repository service which allows to get Jasper files using FWD infrastructure.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20171108 Created initial version.
** 002 SVL 20180209 getInputStream throws exception in case of error.
** 003 EVL 20180507 Adding attempt to load resource from application jar first.
** 004 EVL 20180518 Adding ability to work in standalone mode without server and client running.
** 005 EVL 20160525 Replacing abend with XPR error logger.
** 006 EVL 20160625 Restoring the simple mode functionality after resource loading refactoring.
**                  Adding protection for missing files from abend.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;
import net.sf.jasperreports.engine.*;
import net.sf.jasperreports.repo.*;
import java.io.*;
import java.nio.file.*;

/**
 * Repository service which allows to get Jasper files using FWD infrastructure.
 */
class FwdJasperRepositoryService
extends DefaultRepositoryService
{
   /**
    * Default constructor.
    *
    * @param jasperReportsContext
    *        Parent JasperReports context.
    */
   FwdJasperRepositoryService(JasperReportsContext jasperReportsContext)
   {
      super(jasperReportsContext);

      // disable access to server-side file system
      setFileResolver(s -> null);
   }

   /**
    * Get input stream of the specified resource.
    *
    * @param  uri
    *         The name of the resource to load. Resource lookup happens in the following order:
    *         <ol>
    *            <li>Check if resource name is URL.</li>
    *            <li>Check if resource name is a name of the file inside a server-side jar in the
    *            server class path.</li>
    *            <li>Check if resource name is a name of the file in the client-side file
    *            system.</li>
    *         </ol>
    *         Resource can NOT be a file in the server-side file system.
    *
    * @return input stream of the specified resource.
    *
    * @throws JasperReportException
    *         if the resource cannot be found or loaded.
    */
   @Override
   public InputStream getInputStream(String uri)
   throws JasperReportException
   {
      InputStream res = null;
      // detect simple command line mode usage
      boolean inSimpleMode = XprHelper.isSimpleUseMode(); 
      try
      {
         // get file from URL or server jars if not in simple command line mode
         if (!inSimpleMode)
         {
            res = LogicalTerminal.getResourceStreamFromApplication(uri);
         }
         if (res == null)
         {
            res = super.getInputStream(uri);
         }
      }
      catch (JRRuntimeException e)
      {
         XprHelper.displayOrLogError(
            String.format("Error while loading file %s from URL or server jars", uri), e);
         return res;
      }

      if (res == null)
      {
         try
         {
            // simple mode has different route here
            if (inSimpleMode)
            {
               File file = new File(uri);
               res = new ByteArrayInputStream(Files.readAllBytes(file.toPath()));
               if (res != null)
               {
                  return res;
               }
               else
               {
                  XprHelper.displayOrLogError(
                     String.format("File %s was not found by URL", uri), false);
                  return null;
               }
            }
            
            // get file from client file system
            res = new ByteArrayInputStream(FileSystemOps.getFileContents(uri));
            //TODO do we need caching for client-side images?
         }
         catch (IOException e)
         {
            XprHelper.displayOrLogError(
               String.format("File %s was not found by URL, in server jars and client file system",
                             uri), e, false);
         }
      }
      return res;
   }
}
