/*
** Module   : XprObjBase.java
** Abstract : Class to represent base object feature for XPR content elements.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20180510 Created initial version.
** 002 EVL 20180530 Adding left margin field support.
** 003 EVL 20180602 Minor javadoc fix.  Changed default color to full non-transparent. Adding
**                  support for top margin tag.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

/**
 * The single object inside XPR file.  Commom features for all objects.
 */
public class XprObjBase
{
   /** Constant for not yet defined object. */
   public static final int TYPE_UNDEFINED = -1;
   
   /** Constant for object type is line. */
   public static final int TYPE_LINE = 0;
   
   /** Constant for object type is rectangle. */
   public static final int TYPE_RECTANGLE = 1;
   
   /** Constant for object type is filled rectangle. */
   public static final int TYPE_FILLED_RECTANGLE = 2;
   
   /** Constant for object type is image. */
   public static final int TYPE_IMAGE = 3;
   
   /** Constant for object type is text. */
   public static final int TYPE_TEXT = 4;
   
   /** Constant for object type is bookmark. */
   public static final int TYPE_BOOKMARK = 5;
   
   /** Constant for object type is barcode. */
   public static final int TYPE_BARCODE = 6;
   
   /** Initial object row coordinate. */
   private double rowStart = 0.0;
   
   /** Initial object row coordinate. */
   private double colStart = 0.0;
   
   /** Initial object left margin. */
   private double leftMargin = 0.0;
   
   /** Initial object top margin. */
   private double topMargin = 0.0;
   
   /** The type of the current XPR object. */
   private int objType = TYPE_UNDEFINED;
   
   /** Object color. */
   private int color = 0xFF000000;
   
   /**
    * Create a new base XPR object.
    * 
    * @param    type
    *           The object type value.
    * @param    color
    *           Object color.
    * @param    row
    *           The row value for object.
    * @param    column
    *           The column value for object.
    */
   public XprObjBase(int type, int color, double row, double column)
   {
      this(type, row, column);
      this.color = color;
   }
   
   /**
    * Create a new base XPR object.
    * 
    * @param    type
    *           The object type value.
    * @param    row
    *           The row value for object.
    * @param    column
    *           The column value for object.
    */
   public XprObjBase(int type, double row, double column)
   {
      objType = type;
      rowStart = row;
      colStart = column;
   }
   
   /**
    * Returns the object start row value.
    *
    * @return   The row where object begins.
    */
   public double getRow()
   {
      return rowStart;
   }

   /**
    * Returns the object start column value.
    *
    * @return   The column where object begins.
    */
   public double getColumn()
   {
      return colStart;
   }

   /**
    * Returns the object left margin column coordinate.
    *
    * @return   The effective column value for 1-st column.
    */
   public double getLeftMargin()
   {
      return leftMargin;
   }

   /**
    * Returns the object top margin row coordinate.
    *
    * @return   The effective row value for 1-st row.
    */
   public double getTopMargin()
   {
      return topMargin;
   }

   /**
    * Sets the object left margin column and top margin row coordinates.
    *
    * @param    marginLeft
    *           The new value for left margin.
    * @param    marginTop
    *           The new value for top margin.
    */
   public void setMargins(double marginLeft, double marginTop)
   {
      leftMargin = marginLeft;
      topMargin = marginTop;
   }

   /**
    * Returns the object type value.
    *
    * @return   The type of the given object.
    */
   public int getType()
   {
      return objType;
   }

   /**
    * Returns the object color value.
    *
    * @return   The 32-bit color value of the given object.
    */
   public int getColor()
   {
      return color;
   }
}