/*
** Module   : P2OAccessWorker.java
** Abstract : Pattern worker which provides lookup access to information in
**            P2O ASTs
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 ECF 20050921   @22771 Created initial version. Lazily loads P2O
**                           ASTs and enables lookups of Java names for
**                           legacy Progress table and field names.
**                           This first implementation supports fixed
**                           tables only (not temp- or work-tables).
** 002 ECF 20051004   @22933 Added javaPackage user function. Retrieves
**                           java package name of DMO associated with a
**                           legacy table name.
** 003 ECF 20051005   @22985 Shorten P2O filename. Replace ".schema" with
**                           ".p2o" instead of appending ".p2o" extension.
** 004 ECF 20051013   @23035 Refactored to abstract out P2O AST loading
**                           and initialization. These tasks are now
**                           delegated to P2OLookup.
** 005 ECF 20051017   @23053 Added javaPropertyName user function. Returns
**                           name of DMO property associated with a legacy
**                           database field name.
** 006 ECF 20051024   @23096 Added boolean 'qualified' parameter to
**                           javaPropertyName user function. Optionally
**                           qualifies property name with a buffer name to
**                           enable use within HQL where and order by
**                           clauses, where qualification is necessary.
** 007 ECF 20051102   @23182 Added legacyFieldNames user function.
**                           Retrieves set of legacy field names for a
**                           Progress table, in display order.
** 008 ECF 20051121   @23422 Added isValidatable user function. Indicates
**                           whether a schema field must invoke validation
**                           processing before it may be saved to its DMO.
**                           Delegates work to P2OLookup.
** 009 ECF 20060202   @24253 Added support for temp/work-table schema
**                           data lookups. Implemented visitAst method to
**                           load current, temporary schema data into
**                           P2OLookup's state each time a new source file
**                           is visited during conversion.
** 010 ECF 20060408   @25430 Added isDMOInterface() method. Indicates
**                           whether a given class name collides with any
**                           known DMO interface name. Temp table DMOs not
**                           currently handled.
** 011 ECF 20061026   @30796 Added isForeignReference() method. Indicates
**                           whether a given table is referenced by the
**                           foreign key of any other table.
** 012 GES 20090424   @41952 Import change.
** 013 GES 20090518   @42397 Import change.
** 014 OM  20131030          Added support for checking if a schema contains a certain dmo iface.
** 015 VMN 20140328          Added support for denormalized properties.
** 016 VMN 20150128          Fixed denormalized field assignments.
** 017 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 018 ECF 20180305          Added isReadOnly API to check for a read-only table.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.pattern.*;

/**
 * A pattern worker that provides services to support the retrieval of
 * information from previously created, persisted, P2O ASTs.
 * <p>
 * Currently, this worker supports lookups of converted Java schema names,
 * based upon legacy, Progress schema names as input. Specifically, given a
 * fully qualified, Progress database table name, the Java name for the
 * corresponding Data Model Object (DMO) class can be retrieved. The Java
 * name of the implicit (default) buffer for that table is also available.
 * Given a fully qualified, Progress database field name, the Java name for
 * the corresponding DMO getter and setter methods can be retrieved.
 */
public final class P2OAccessWorker
extends AbstractPatternWorker
implements DataModelTokenTypes
{
   /**
    * Default constructor which initializes libraries.
    */
   public P2OAccessWorker()
   {
      super();
      
      setLibrary(new Library());
   }
   
   /**
    * This method is called each time a new AST is visited, just before the
    * pattern engine processes that AST.  We exploit this hook by loading the
    * current temp/work-table schema (P2O) AST into the <code>P2OLookup</code>
    * class' state.  This enables us to resolve references to temp and work
    * tables, which are local to the currently loaded source file AST.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
      P2OLookup.loadTemporarySchema(ast.getFilename());
   }
   
   /**
    * Resolve <code>constant</code> to a literal which can be compiled into
    * expressions.
    *
    * @param   constant
    *          Constant indicating a data model token name.
    *
    * @return  Token type associated with the token name.
    */
   public Object resolveConstant(String constant)
   {
      return DataModelAst.resolveConstant(constant);
   }
   
   /**
    * Helper to lookup Java names for tables and fields, given their
    * historical, Progress names.
    */
   public class Library
   {
      /**
       * Look up the package of a Java DMO interface, given its legacy name.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a table.
       *
       * @return  Package name of the DMO interface.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence.
       */
      public String javaPackage(String historical)
      throws SchemaException
      {
         return P2OLookup.javaPackage(historical);
      }
      
      /**
       * Look up the name of a Java DMO interface, given its legacy name.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a table.
       *
       * @return  Unqualified, Java name of the interface which defines the
       *          DMO API.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the table portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaClassName(String historical)
      throws SchemaException
      {
         return P2OLookup.javaName(historical, EntityName.TABLE);
      }
      
      /**
       * Look up the name of a Java DMO interface, given its legacy name,
       * and return the name of the default, implicit buffer associated with
       * this interface. This is simply the name of the appropriate DMO
       * interface, with the first letter lowercased.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a table.
       *
       * @return  Short DMO interface name with first letter decapitalized.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the table portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaBufferName(String historical)
      throws SchemaException
      {
         return StringHelper.changeCase(javaClassName(historical), false);
      }
      
      /**
       * Look up the name of the name of the DMO property which represents
       * the specified, legacy database field.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       * @param   qualified
       *          <code>true</code> if property name should be qualified with
       *          a buffer name;  <code>false</code> if it should remain
       *          unqualified.
       *
       * @return  Converted Java property name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaPropertyName(String historical, boolean qualified)
      throws SchemaException
      {
         return P2OLookup.javaPropertyName(historical, qualified);
      }

      /**
       * Look up the name of the name of the DMO property which represents
       * the specified, legacy database field.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       * @param   qualified
       *          <code>true</code> if property name should be qualified with
       *          a buffer name;  <code>false</code> if it should remain
       *          unqualified.
       * @param   index
       *          Zero-based index of custom property name for denormalized field with extent,
       *          otherwise null.
       *
       * @return  Converted Java property name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaPropertyName(String historical, boolean qualified, Long index)
      throws SchemaException
      {
         return P2OLookup.javaPropertyName(historical, qualified, index);
      }
      
      /**
       * Retrieve the full set of fully quallified, legacy field names for a
       * legacy Progress table.  The set is guaranteed to iterate in the
       * default display order defined for the original, Progress schema for
       * that table.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a table.
       *
       * @return  Set of fully qualified, legacy field names.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if <code>historical</code> does not represent a known table.
       */
      public Set legacyFieldNames(String historical)
      throws SchemaException
      {
         return P2OLookup.legacyFieldNames(historical);
      }
      
      /**
       * Look up the name of the accessor (getter) method associated with the
       * DMO property which represents the specified, legacy database field.
       * The method name is in the form of a typical Java-bean getter:
       * <pre>
       *    get{PropertyName}
       * </pre>
       * or
       * <pre>
       *    is{PropertyName}
       * </pre>
       * if <i>PropertyName</i> represents a <code>logical</code> data type.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       * @param   isLogical
       *          <code>true</code> if <code>historical</code> represents a
       *          <code>logical</code> field, else <code>false</code>.
       *
       * @return  Java-bean style getter method name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaGetterName(String historical, boolean isLogical)
      throws SchemaException
      {
         String prefix = isLogical ? "is" : "get";
         
         return P2OLookup.beanMethodName(prefix, historical);
      }

      /**
       * Look up the name of the accessor (getter) method associated with the
       * DMO property which represents the specified, legacy database field.
       * The method name is in the form of a typical Java-bean getter:
       * <pre>
       *    get{PropertyName}
       * </pre>
       * or
       * <pre>
       *    is{PropertyName}
       * </pre>
       * if <i>PropertyName</i> represents a <code>logical</code> data type.
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       * @param   index
       *          Zero-based index of custom property name for denormalized field with extent,
       *          otherwise null.
       * @param   isLogical
       *          <code>true</code> if <code>historical</code> represents a
       *          <code>logical</code> field, else <code>false</code>.
       *
       * @return  Java-bean style getter method name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaGetterName(String historical, Long index, boolean isLogical)
      throws SchemaException
      {
         String prefix = isLogical ? "is" : "get";
         if ( index != null)
         {
            index --;
         }
         return P2OLookup.beanMethodName(prefix, historical, index);
      }
      
      /**
       * Look up the name of the mutator (setter) method associated with the
       * DMO property which represents the specified, legacy database field.
       * The method name is in the form of a typical Java-bean setter:
       * <pre>
       *    set{PropertyName}
       * </pre>
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       *
       * @return  Java-bean style setter method name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaSetterName(String historical)
      throws SchemaException
      {
         return P2OLookup.beanMethodName("set", historical);
      }

      /**
       * Look up the name of the mutator (setter) method associated with the
       * DMO property which represents the specified, legacy database field.
       * The method name is in the form of a typical Java-bean setter:
       * <pre>
       *    set{PropertyName}
       * </pre>
       *
       * @param   historical
       *          Fully qualified, legacy, Progress schema name for a
       *          database field.
       * @param   index
       *          Zero-based index of custom property name for denormalized field with extent,
       *          otherwise null.
       *
       * @return  Java-bean style setter method name.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence;
       *          if the field portion of <code>historical</code> cannot be
       *          resolved to a corresponding Java name.
       */
      public String javaSetterName(String historical, Long index)
      throws SchemaException
      {
         if ( index != null)
         {
            index --;
         }
         return P2OLookup.beanMethodName("set", historical, index);
      }

      /**
       * Check if given historical fully qualified property is denormalized.
       *
       * @param    historical
       *           Historical fully qualified property name.
       *
       * @return  <code>true</code> if given property is denormalized.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence.
       */
      public boolean isDenormalizedProperty(String historical)
      throws SchemaException
      {
         return P2OLookup.isDenormalizedProperty(historical);
      }
      
      /**
       * Indicate whether a DMO property can be validated at runtime, given
       * its legacy schema name.  Possible forms of validation include:
       * <ul>
       *   <li>mandatory (non-nullable) constraint;
       *   <li>participation in a unique constraint;
       *   <li>explicit validation method (converted from a Progress
       *       validation expression).
       * </ul>
       *
       * @param   historical
       *          Fully qualified, legacy Progress field name.
       *
       * @return  <code>true</code> if the field should be validated at
       *          runtime, else <code>false</code>.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence.
       */
      public boolean isValidatable(String historical)
      throws SchemaException
      {
         return P2OLookup.isValidatable(historical);
      }
      
      /**
       * Indicate whether a DMO class represents a read-only table, given its schema name.
       * 
       * @param   historical
       *          Fully qualified, legacy Progress table name.
       * 
       * @return  {@code true} if the table is read-only, else {@code false}.
       *
       * @throws  SchemaException
       *          if there is any error loading P2J schema configuration data;
       *          if there is any error loading the AST from persistence.
       */
      public boolean isReadOnly(String historical)
      throws SchemaException
      {
         return P2OLookup.isReadOnly(historical);
      }
      
      /**
       * Report whether the given class name is the unqualified name of any
       * DMO interface, in any schema tracked by the current P2J
       * configuration.
       *
       * @param   name
       *          An unqualified class name.
       *
       * @return  <code>true</code> if <code>name</code> collides with any
       *          DMO interface in any schema, else <code>false</code>.
       *
       * @throws  SchemaException
       *          if there is an error loading schema names from
       *          configuration.
       */
      public boolean isDMOInterface(String name)
      throws SchemaException
      {
         return P2OLookup.isDMOInterface(name);
      }

      /**
       * Report whether the given class name is the unqualified name of a DMO interface, in the
       * <code>schema</code> tracked by the current P2J configuration.
       *
       * @param   name
       *          An unqualified class name.
       * @param   schema
       *          The name of the schema to look into.
       *
       * @return  <code>true</code> if <code>name</code> is present as java name of a table from
       *          the specified <code>schema</code> and <code>false</code> if <code>schema</code>
       *          does not exist or it does not contain such interface.
       *
       * @throws  SchemaException
       *          if there is an error loading schema names from configuration.
       */
      public boolean isDMOInterface(String name, String schema)
      throws SchemaException
      {
         return P2OLookup.isDMOInterface(name, schema);
      }
      
      /**
       * Report whether the given, historical table will be mapped in the
       * converted application to a database table which represents a foreign
       * reference of another table's foreign key.
       *
       * @param   historical
       *          Fully qualified, legacy Progress table name.
       *
       * @return  <code>true</code> if the given table represents the foreign
       *          end of a one-to-one or a one-to-many relation between two
       *          tables.
       *
       * @throws  SchemaException
       *          if there is an error loading schema names from
       *          configuration.
       */
      public boolean isForeignReference(String historical)
      throws SchemaException
      {
         return P2OLookup.isForeignReference(historical);
      }
   }
}