/*
** Module   : RelationNode.java
** Abstract : Tree node for experimental analysis of natural relations
**            between database tables
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050325   @20998 Created initial version.
** 002 GES 20090518   @42400 Import change.
** 003 ECF 20131013          Implemented generics.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.util.*;

/**
 * A node in a tree of table relation dependencies. This class was written
 * for experimental analysis of primary-foreign key relation candidates
 * among Progress tables. It probably is of little value for production use
 * and should be deprecated.
 */
class RelationNode
extends TreeNode
implements SchemaParserTokenTypes
{
   /** Table wrapped by this node */
   private Aast table = null;
   
   /** Names of index fields which comprise the relation */
   private Collection<String> key = null;
   
   /**
    * Constructor which accepts a table and collection of fields which
    * define this node's relation with another node's table.
    *
    * @param   table
    *          Table AST wrapped by this node.
    * @param   key
    *          Collection of field names which define the key between tables.
    */
   RelationNode(Aast table, Collection<String> key)
   {
      this.table = table;
      this.key = key;
   }
   
   /**
    * Set the relation key for this node.
    *
    * @param   key
    *          Collection of field names which define the key between tables.
    */
   void setKey(Collection<String> key)
   {
      this.key = key;
   }
   
   /**
    * Create a string representation of this node for debug output.
    *
    * @return  String representation of this node.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder();
      
      int depth = getDepth();
      for (int i = 0; i < depth; i++)
      {
         buf.append("   ");
      }
      
      buf.append(table.getText());
      
      if (key == null)
      {
         return buf.toString();
      }
      
      buf.append(" [");
      
      Iterator<String> iter = key.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         String name = iter.next();
         
         Aast field = null;
         Aast next = (Aast) table.getFirstChild();
         while (field == null && next != null)
         {
            int type = next.getType();
            if (((type > BEGIN_FIELDTYPES && type < END_FIELDTYPES) ||
                 (type > BEGIN_METATYPES  && type < END_METATYPES)) &&
                 name.equalsIgnoreCase(next.getText()))
            {
               field = next;
            }
            
            next = (Aast) next.getNextSibling();
         }
         
         buf.append(name).append(" (").append(field.getSymbolicTokenType()).append(")");
      }
      
      buf.append(']');
      
      return buf.toString();
   }
}