/*
** Module   : SchemaConfig.java
** Abstract : Schema configuration information
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 ECF 20041217   @19174 Created initial version.
** 002 ECF 20050207   @19639 Added support for schema metadata config
**                           information. Added two public methods:
**                           getMetadataFile and setMetadata.
** 003 ECF 20050411   @20707 Added import filename to NsConfig, along with
**                           getter and setter methods.
** 004 GES 20070402   @32704 Added a preferred flag to the namespace
**                           configuration. This can be used to resolve
**                           some ambiguous name conflicts.
** 005 CA  20101102          Inferred generics.Added support for new namespace
**                           parameters,under the namespace node: ddl-dialects
**                           parameter node and dialect-specific node.
** 006 LMR 20110106          Added toString() method to print the whole object
**                           graph.
** 007 ECF 20130125          Changed and expanded metadata configuration to reflect updated
**                           structure of p2j.cfg.xml. Added Metadata inner class, which reflects
**                           addition of metadata element in the XML.
** 008 ECF 20130804          Added standard set of metadata table names that are always needed at
**                           runtime, so that downstream schema processing will convert these.
**                           Trim and lowercase metadata table names for safety. Added method
**                           getSchemas.
** 009 SVL 20140320          Added getSchemaResource.
** 010 ECF 20141123          Partially backed out 008: removed standard set of metadata table
**                           names.
** 011 ECF 20150328          Fixed bug in getSchemas which did not properly omit metadata schema.
** 012 ECF 20150910          Added isMetadataActive.
** 013 EVL 20160225          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 014 IAS 20160509          Fixed metadata processing in case of more than one 
**                           default namespace.
** 015 OM  20180118          Removed notion of primary database.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.cfg.*;

/**
 * Schema configuration information. This class is instantiated by the master {@link
 * com.goldencode.p2j.cfg.Configuration Configuration} reading input data from an XML
 * configuration file.  The syntax of that XML file is specified external to this class.  The
 * purpose of this class is to store and provide access to the configuration.  How that
 * configuration is read is not of concern here.
 * <p>
 * Each database used by an application is configured in an XML element in the configuration file.
 * That element contains a short name or handle for the database, which matches the logical name
 * of the database. It also contains a path to the DF file exported from the Progress data
 * dictionary (the import file from our perspective), as well as the XML file to which the
 * imported schema information will be written for use in conversion (the schema file). Finally, a
 * database can be a default database (always loaded into the schema dictionary when parsing
 * source code), or non-default (only loaded on demand by the Progress parser). All of this
 * information about a particular database is encoded in an instance of the <code>NsConfig</code>
 * inner class (which is <code>public</code> since the external configuration reading process must
 * create instances of it). Those <code>NsConfig</code> instances are then added using {@link
 * #addNamespace}.
 * <p>
 * One database is designated the "metadata" database. This is the database which contains
 * metadata about the schema structure used to define all databases in Progress (for a specific
 * version of Progress). The tables and fields of this database are referenced in source code as
 * if they were members of every other logical database. Thus, the schema information included in
 * the metadata database schema is merged with each logical database as it is loaded into the
 * schema dictionary. The metadata schema configuration is accessible via the {@link #getMetadata}
 * method.
 *
 * @see  SchemaLoader
 * @see  SchemaDictionary
 */
public final class SchemaConfig
{
   /** Name of metadata definition applied to every loaded schema */
   private Metadata metadata = null;
   
   /** Map of database names to namespace configuration data */
   private Map<String, NsConfig> nsConfigs = new LinkedHashMap<>();
   
   /**
    * Default constructor.
    */
   public SchemaConfig()
   {
   }
   
   /**
    * Determine whether metadata is active for the current application. Metadata is considered
    * active if at least one metadata table is configured.
    * 
    * @return  <code>true</code> if the application uses metadata, else <code>false</code>.
    */
   public boolean isMetadataActive()
   {
      return !metadata.getTables().isEmpty();
   }
   
   /**
    * Get the metadata identifier name associated with all schemas.
    *
    * @return  Common schema metadata identifier.
    */
   public Metadata getMetadata()
   {
      return metadata;
   }
   
   /**
    * Name of metadata definition applied to every loaded schema.
    *
    * @param   metadata
    *          Metadata definition name.
    */
   public void setMetadata(Metadata metadata)
   {
      this.metadata = metadata;
   }
   
   /**
    * Get a list of schema names, optionally omitting the name of the metadata schema.
    * 
    * @param   omitMeta
    *          <code>true</code> to omit the metadata schema, <code>false</code> to include it.
    */
   public List<String> getSchemas(boolean omitMeta)
   {
      Set<String> keys = nsConfigs.keySet();
      ArrayList<String> list = new ArrayList<>(keys.size());
      
      if (omitMeta)
      {
         for (String next : keys)
         {
            if (metadata.getName().equals(next))
            {
               continue;
            }
            
            list.add(next);
         }
      }
      else
      {
         list.addAll(keys);
      }
      
      list.trimToSize();
      
      return list;
   }
   
   /**
    * Get an iterator of configured database names.
    *
    * @return  Object which iterates over configured database names.
    */
   public Iterator<String> databases()
   {
      return nsConfigs.keySet().iterator();
   }
   
   /**
    * Get an iterator of names of those databases which have been configured
    * as default schemas for namespace lookup purposes. These are databases
    * which are loaded automatically by a {@link SchemaDictionary} object at
    * construction.
    *
    * @return  Object which iterates over the names of default databases.
    */
   public Iterator<String> defaultDatabases()
   {
      // Return instance of anonymous nested iterator.
      return new Iterator<String>()
      {
         /** Name of next default database to be returned by iterator */
         private String nextName = null;
         
         /** Iterator of all database names */
         private Iterator<String> iter = databases();
         
         /**
          * Determine whether there is another default database over which
          * to iterate.
          *
          * @return  <code>true</code> if iteration can continue, else
          *          <code>false</code>.
          */
         public boolean hasNext()
         {
            while (iter.hasNext())
            {
               String name = iter.next();
               NsConfig cfg = nsConfigs.get(name);
               if (cfg.isDefault())
               {
                  nextName = cfg.getName();
                  return true;
               }
            }
            
            return false;
         }
         
         /**
          * Return the next default database name in the iterator's sequence.
          *
          * @return  Next name.
          */
         public String next()
         {
            return nextName;
         }
         
         /**
          * Unsupported remove operation.
          *
          * @throws  UnsupportedOperationException
          *          always.
          */
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
      };
   }
   
   /**
    * Add a namespace configuration object to this configuration. Called by
    * the configuration processor when reading the XML schema configuration.
    *
    * @param   cfg
    *          Namespace configuration object.
    */
   public void addNamespace(NsConfig cfg)
   {
      nsConfigs.put(cfg.getName(), cfg);
   }
   
   /**
    * Returns a file object representing the import (DF) file associated with
    * the specified database schema name. The file's path is fully qualified.
    *
    * @param   name
    *          Name of desired namespace configuration
    *
    * @return  Import file.
    *
    * @throws  ConfigurationException
    *          if environment has not been configured correctly.
    */
   public File getImportFile(String name)
   throws ConfigurationException
   {
      File file = null;
      
      NsConfig cfg = nsConfigs.get(name);
      if (cfg != null)
      {
         file = Configuration.toFile(cfg.importFile);
      }
      
      return file;
   }
   
   /**
    * Returns a file object representing the XML document associated with the
    * specified database schema name. The file's path is fully qualified.
    *
    * @param   name
    *          Name of desired namespace configuration
    *
    * @return  File containing namespace schema XML document.
    *
    * @throws  ConfigurationException
    *          if environment has not been configured correctly.
    */
   public File getSchemaFile(String name)
   throws ConfigurationException
   {
      File file = null;
      
      NsConfig cfg = nsConfigs.get(name);
      if (cfg != null)
      {
         file = Configuration.toFile(cfg.xmlFile);
      }
      
      return file;
   }

   /**
    * Returns path to the resource representing the XML document associated with the specified
    * database schema name.
    *
    * @param   name
    *          Name of desired namespace configuration
    *
    * @return  path to the resource containing namespace schema XML document or <code>null</code>
    *          if the resource doesn't exist.
    */
   public String getSchemaResource(String name)
   {
      NsConfig cfg = nsConfigs.get(name);
      if (cfg != null)
      {
         String resourceName = cfg.xmlFile;
         if (ClassLoader.getSystemClassLoader().getResource(resourceName) != null)
         {
            return resourceName;
         }
      }

      return null;
   }
   
   /**
    * Reports whether the database should be preferred in cases where its entities are involved in
    * ambiguous name conflicts.
    *
    * @param   name
    *          Name of desired namespace configuration
    *
    * @return  {@code true} if the database is preferred.
    */
   public boolean isPreferred(String name)
   {
      NsConfig cfg = nsConfigs.get(name);
      return cfg != null && cfg.isPreferred();
   }
   
   /**
    * Get the value for the given namespace parameter.
    * 
    * @param   name
    *          The namespace name.
    * @param   param
    *          The parameter name or path.
    *
    * @return  The parameter's value or {@code null} if {@code name} or {@code param} are
    *          invalid. 
    */
   public String getNamespaceParameter(String name, String param)
   {
      NsConfig cfg = nsConfigs.get(name);
      return cfg == null ? null : cfg.getParameter(param);
   }
   
   /**
    * Returns a string representation of this {@code SchemaConfig}'s object graph.
    * 
    * @return  A string representation of this object.
    * 
    * @see Object#toString()
    */
   @Override
   public String toString()
   {
      StringBuilder sb = new StringBuilder("schema:");
      sb.append("\n  metadata = ");
      sb.append(getMetadata());
      sb.append("\n  namespaces: ");
      for (NsConfig ns : nsConfigs.values())
      {
         sb.append("\n    ");
         sb.append(ns.getName());
         sb.append(":");
         sb.append("\n      importFile = ");
         sb.append(ns.getImportFile());
         sb.append("\n      xmlFile = ");
         sb.append(ns.getXmlFile());
         sb.append("\n      default = ");
         sb.append(ns.isDefault());
         sb.append("\n      preferred = ");
         sb.append(ns.isPreferred());
         sb.append("\n      parameters:");
         for (Map.Entry<String, String> e : ns.parameters.entrySet())
         {
            sb.append("\n        name = ");
            sb.append(e.getKey());
            sb.append(", value = ");
            sb.append(e.getValue());
         }
      }
      return sb.toString();
   }
   
   /**
    * Encapsulates information about a namespace configuration, including
    * name, XML file, and whether namespace should be loaded into the schema
    * dictionary by default.
    * <p>
    * This class is public since it must be accessed as part of the public
    * API for the containing class.
    */
   public static class NsConfig
   {
      /** Identifying name for namespace. */
      private String name;
      
      /** Name of DF import file. */
      private String importFile;
      
      /** XML namespace document filename. */
      private String xmlFile;
      
      /** Flag indicating if namespace is loaded by default. */
      private boolean defDb = false;
      
      /** Namespace is preferred in ambiguous name conflicts. */
      private boolean prefer = false;
      
      /** Namespace parameters. */
      private Map<String, String> parameters = new HashMap<>();
      
      /**
       * Default constructor.
       */
      public NsConfig()
      {
      }
      
      /**
       * Get identifying namespace name.
       *
       * @return  Namespace name.
       */
      public String getName()
      {
         return name;
      }
      
      /**
       * Set name.
       *
       * @param   name
       *          Namespace name.
       */
      public void setName(String name)
      {
         this.name = name;
      }
      
      /**
       * Get DF import filename.
       *
       * @return  Import filename.
       */
      public String getImportFile()
      {
         return importFile;
      }
      
      /**
       * Set DF import filename.
       *
       * @param   importFile
       *          Import filename.
       */
      public void setImportFile(String importFile)
      {
         this.importFile = importFile;
      }
      
      /**
       * Get XML filename associated with the configured namespace.
       *
       * @return  XML filename associated with the namespace.
       */
      public String getXmlFile()
      {
         return xmlFile;
      }
      
      /**
       * Set XML filename associated with the configured namespace.
       *
       * @param   xmlFile
       *          XML document filename.
       */
      public void setXmlFile(String xmlFile)
      {
         this.xmlFile = xmlFile;
      }
      
      /**
       * Is this namespace loaded in the schema dictionary by default?
       *
       * @return  <code>true</code> if namespace is loaded by default, else
       *          <code>false</code>.
       */
      public boolean isDefault()
      {
         return defDb;
      }
      
      /**
       * Set default flag for this namespace configuration.
       *
       * @param   defDb
       *          <code>true</code> if namespace is loaded by default, else
       *          <code>false</code>.
       */
      public void setDefault(boolean defDb)
      {
         this.defDb = defDb;
      }
      
      /**
       * Is this namespace preferred when it's components are involved in
       * ambiguous name conflicts?
       *
       * @return  <code>true</code> if namespace is preferred, else
       *          <code>false</code>.
       */
      public boolean isPreferred()
      {
         return prefer;
      }
      
      /**
       * Set preferred flag for this namespace configuration.  This can 
       * resolve ambiguous name conflicts if this namespace is the only one
       * that is preferred out of all the conflicting names.
       *
       * @param   prefer
       *          <code>true</code> if namespace is preferred, else
       *          <code>false</code>.
       */
      public void setPreferred(boolean prefer)
      {
         this.prefer = prefer;
      }
      
      /**
       * Add the specifid parameter to the namespace.
       * 
       * @param   param
       *          The parameter.
       */
      public void addParameter(Parameter param)
      {
         parameters.put(param.getName(), param.getValue());
      }
      
      /**
       * Add the specifid dialect parameters to the namespace.
       * <p>
       * All the parameter's names will follow the 
       * <code>[dialect]/[parameter]</code> syntax.
       * 
       * @param   dialectParam
       *          The dialect parameter.
       */
      public void addDialectParameter(DialectParam dialectParam)
      {
         // merge the params to the main map
         parameters.putAll(dialectParam.parameters);
      }
      
      /**
       * Get the value for the specified namespace parameter.
       * 
       * @param    param
       *           The parameter name.
       *           
       * @return   The parameter's value.
       */
      public String getParameter(String param)
      {
         return parameters.get(param);
      }
      
      /**
       * Helper class to read and set the dialect parameter. 
       */
      public static class DialectParam
      {
         /** The dialect alias. */
         private String name;
         
         /** The dialect's parameters. */
         private Map<String, String> parameters = new HashMap<>();

         /**
          * Get the dialect alias.
          * 
          * @return   the dialect alias.
          */
         public String getName()
         {
            return name;
         }

         /**
          * Set the dialect alias.
          * 
          * @param   name
          *          the dialect alias.
          */
         public void setName(String name)
         {
            this.name = name;
         }
         
         /**
          * Add a parameter for this dialect alias. The parameter name will be
          * added the <code>[dialect]/</code> prefix.
          * 
          * @param   param
          *          The parameter.
          */
         public void addParameter(Parameter param)
         {
            parameters.put(name + "/" + param.getName(), param.getValue());
         }
      }
   }
   
   /**
    * Encapsulates information about the project's use of database metadata, including the name of
    * the metadata schema and the metadata tables used by the application, if any.
    * <p>
    * This class is public since it must be accessed as part of the public API for the containing
    * class.
    */
   public static class Metadata
   {
      /** Root name of the metadata schema (without any file extension) */
      private String name = null;
      
      /** Names of the tables in use by the application */
      private Set<String> tables = new HashSet<>();
      
      /**
       * Default constructor.
       */
      public Metadata()
      {
      }
      
      /**
       * Get the root name of the metadata schema.
       * 
       * @return  Root name of the metadata schema.
       */
      public String getName()
      {
         return name;
      }
      
      /**
       * Set the root name of the metadata schema.
       * 
       * @param   name
       *          Root name of the metadata schema.
       */
      public void setName(String name)
      {
         this.name = name;
      }
      
      /**
       * Get an unmodifiable set of the names of the metadata tables used by the application.
       * 
       * @return  Metadata table names.
       */
      public Set<String> getTables()
      {
         return Collections.unmodifiableSet(tables);
      }
      
      /**
       * Add a table name to the set of metadata tables used by the application.
       * 
       * @param   table
       *          Metadata table name.
       */
      public void addTable(String table)
      {
         tables.add(table.trim().toLowerCase());
      }
   }
}
