/*
** Module   : AssociatedThread.java
** Abstract : a new thread which inherits the security context of the creating
**            thread
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20111109 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

/**
 * Provides a thread which automatically inherits the security context of the
 * thread that instantiates it. When the thread exits, it will automatically
 * cleanup the context (decrementing the reference counts and so forth). This
 * makes it easy to implement additional threads which all share the same
 * context, which means they have the same rights and have the same
 * context-local data.
 */
public class AssociatedThread
extends Thread
{
   /** The core logic that will be executed. */
   private Runnable core = null;
   
   /** The security context of the instantiating thread. */
   private SecurityContextStack stack = null;
   
   /** Any pending security context switch or <code>null</code>. */
   private SecurityContext context = null;

   /** Security Manager instance. */
   private SecurityManager sm = SecurityManager.getInstance();
   
   /**
    * Constructor which allows the core logic to be specified.
    *
    * @param    core
    *           The logic to execute when the thread is started.  Must not be
    *           <code>null</code>.
    *
    * @throws   NullPointerException
    *           If the instantiating thread has no security context or if the
    *           <code>core</code> is <code>null</code>.
    */
   public AssociatedThread(Runnable core)
   throws NullPointerException
   {
      super();
      
      if (core == null)
         throw new NullPointerException("Core logic must be specified!");
      
      this.core = core;
      
      // duplicate the current context stack (which has both the initial and
      // optionally an effective context assigned)
      if (sm.isServer())
      {
         this.stack = SecurityContextStack.getContext().duplicate();
      }
   }
   
   /**
    * Save the given context as a pending context switch which will be
    * processed later by the key reading thread.
    *
    * @param    context
    *           The security context to switch to (if non-null).
    */
   public synchronized void setPendingContextSwitch(Object context)
   {
      if (context != null && context instanceof SecurityContext)
      {
         this.context = (SecurityContext) context;
      }
   }
   
   /**
    * Switch security contexts on the current thread if a switch is pending.
    */
   public synchronized void morphContextIfNeeded()
   {
      if (context != null)
      {
         if (sm.isServer())
         {
            try
            {
               sm.pushContextWorker(context);
            }
            
            catch (RestrictedUseException rue)
            {
               // should not happen
            }
         }
         
         context = null;
      }
   }
   
   /**
    * This method will be invoked when the thread is started, it sets up the
    * initial security context to match the saved context of the instantiating
    * thread, then it will call the core logic specified by the caller and
    * drop the security context after that logic is done. 
    */
   @Override
   public void run() 
   {
      // server-side processing needs a context to work properly and this
      // is a new thread
      if (sm.isServer())
      {
         sm.forceSecurityContext(stack);
      }
      
      try
      {
         core.run();
      }
      
      finally
      {
         // cleanup processing
         if (sm.isServer())
         {
            try
            {
               sm.popAllSecurityContext();
            }
            
            catch (RestrictedUseException rue)
            {
               // should not happen
            }
         }
      }
   }
}