/*
** Module   : GroupAccount.java
** Abstract : represents a group which has specific rights and to which a
**            list of other subjects belongs
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20050307   @20194 Created the initial version.
** 002 NVS 20050311   @20295 Added getAlias() method implementation.
** 003 GES 20081104   @40339 Refactored to move common code into the base
**                           class.
** 004 NVS 20090603   @42590 Group accounts are created always enabled.
** 005 SVL 20100916          Auth mode and auth plugin fields were added.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

/**
 * Represents a group which has specific rights and to which a list of other
 * subjects belongs.  A group is a subject in the security environment, but
 * it is a subject through which many other subjects can have shared rights.
 * In other words, all subjects that belong to a given group will have the
 * same rights as are assigned to this group.
 */
class GroupAccount
extends Account
{
   /** Authorization mode override for this group. */
   private int mode = SecurityConstants.AUTH_MODE_NONE;

   /** Custom authentication plugin class override for this group. */
   private String authPlugin = null;

   /**
    * Constructor.
    *
    * @param    subjectId
    *           The subject ID (also known as account name).
    * @param    description
    *           The verbose text that describes the purpose of this group, 
    *           which may be <code>null</code>
    * @param    authMode
    *           Authorization mode for this group.
    * @param    authPlugin
    *           Custom authentication plugin class for this group.
    */
   GroupAccount(String subjectId,
                String description,
                int    authMode,
                String authPlugin)
   {
      super(subjectId, true, description, null);
      this.mode       = authMode;
      this.authPlugin = authPlugin;
   }

   /**
    * Gets account type of this account
    *
    * @return account type as <code>ACC_GROUP</code> constant.
    */
   int getAccountType()
   {
      return ACC_GROUP;
   }

   /**
    * Get authorization mode for this group.
    *
    * @return  See above.
    */
   public int getAuthMode()
   {
      return mode;
   }

   /**
    * Get custom authentication plugin class for this group.
    *
    * @return  Authentication plugin class or <code>null</code> if
    *          authorization mode is not
    *          <code>SecurityConstants.AUTH_MODE_CUSTOM</code>.
    */
   public String getAuthPlugin()
   {
      return authPlugin;
   }
}