/*
** Module   : Resource.java
** Abstract : This interface defines the resource plugin registration
**            procedure.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20050208   @19700 Created the initial version.
** 002 NVS 20090707   @43076 Added getRightsEditorName() method.
** 003 NVS 20090824   @43725 Added init() method. When the init() is called,
**                           the directory service is functional and available
**                           for reading directory nodes. 
** 004 HC  20170612          Changes related to implementation of new GWT-based Admin client.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

/**
 * This package private interface defines the resource plugin registration
 * procedure. The only class that implements this interface is
 * <code>AbstractResource</code>. The plugins cannot implement this interface, 
 * so they should extend the <code>AbstractResource</code>.
 */
interface Resource
{
   /**
    * Notifies the plugin about the instance of the Security Manager
    * to communicate with.
    *
    * @param  sm
    *         specifies an instance of <code>SecurityManager</code> this
    *         plugin should communicate with
    */
   public void attach(SecurityManager sm);
    
   /**
    * Gives chance to the plugin to initialize when all accounts are created.
    */
   public void init();
    
   /**
    * Returns the plugin resource type name as a string.
    *
    * @return plugin resource type name
    */
   public String getTypeName();
    
   /**
    * Returns the plugin's exported variables and functions.
    *
    * @return plugin's instance of a library class for symbol resolution or
    *         <code>null</code> if there is nothing to export
    */
   public Object getLibrary();
    
   /**
    * Notifies the plugin that the resource type has been assigned a numeric
    * value.
    *
    * @param  res
    *         specifies the registration number for the resource
    */
   public void registeredAs(int res);
    
   /**
    * Returns an array of descriptions, one object per the plugin's access
    * rights item.
    *
    * @return array of <code>Description</code>s, one per each plugin's access
    *         rights item. 
    */
   public Description[] describeRights();
    
   /**
    * Instantiates a plugin's class that implements the Rights interface,
    * using the array of objects representing a set of access rights fields.
    *
    * @param  rights
    *         an array of objects of proper types representing items in a set
    *         of access rights
    *
    * @return the reference to the interface.
    */
   public Rights getRightsInstance(Object[] rights);
    
   /**
    * Checks whether a given string is a syntactically valid resource name
    * for this resource type.
    *
    * @param  resource
    *         string naming a resource
    *
    * @return <code>true</code> if the name is syntactically correct 
    */
   public boolean isInstanceNameValid(String resource);
    
   /**
    * Checks whether a given array of objects representing a set of access
    * rights fields is acceptable.
    *
    * @param  rights
    *         an array of objects of proper types representing items in a set
    *         of access rights
    *
    * @return <code>true</code> if the array is acceptable for an instance
    *         of access rights 
    */
   public boolean isRightsSetValid(Object[] rights);
}