/*
** Module   : Choose.java
** Abstract : implements CHOOSE statement options
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026 NEW   @23304 Created initial version.
** 002 SIY 20051207 CHG   @24627 Added getData() method, cleanups/formatting.
** 003 SIY 20060220 CHG   @24627 Removed Serializable declaration, other 
**                               changes necessary for choose() implementation.
** 004 SIY 20060719 CHG   @28139 Removed replacing of null with "" in help 
**                               text. Renamed isRowField() to isFieldMode().
** 005 GES 20060807 CHG   @28459 Cleaned up naming of vars and removed use
**                               of setDefined() in ChooseData. 
** 006 SIY 20101122              Inferred generics (partially).
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Supports CHOOSE statement options.
 * <p>
 * It defines:
 * <ul>
 *   <li>ROW, HELP or set of FIELD, HELP
 *   <li>AUTO-RETURN option
 *   <li>GO-ON list
 *   <li>KEYS string
 *   <li>NO-ERROR string
 *   <li>PAUSE value
 * </ul>
 * Methods are defined to set and get all available options.
 *
 * @author   NVS
 */
public class Choose
{
   /** At least one <b>CHOOSE</b> item has been defined. */
   private boolean defined = false;

   /** <code>true</code> if in FIELD mode, <code>false</code> for ROW mode. */
   private boolean fieldMode = false;

   /** Holds the field widget in the <b>ROW</b> mode. */
   private GenericWidget row = null;

   /** Holds the <b>HELP</b> text for the <b>ROW</b> item. */
   private String rowHelp = null;

   /** Holds the field widgets in the <b>FIELD</b> mode. */
   private LinkedList<GenericWidget> fields = new LinkedList<GenericWidget>();

   /** Holds the <b>HELP</b> text for the <b>FIELD</b> items. */
   private LinkedList<String> fieldHelps = new LinkedList<String>();

   /** <b>AUTO-RETURN</b> option.  */
   private boolean autoReturn = false;

   /** Custom color. */
   private ColorSpec color = null;

   /** Holds the <b>GO-ON</b> keys. */
   private List<String> goOn = new LinkedList<String>();

   /** Highlight and navigation variable. */
   private Accessor keysVar = null;

   /** <b>NO-ERROR</b> option. */
   private boolean noError = false;

   /** <b>CHOOSE</b> timeout. */
   private int pause = -1;

   /**
    * Default constructor.
    */
   public Choose()
   {
   }

   /**
    * Sets the <b>ROW </b> mode and stores the row widget in the CHOOSE
    * instance.
    * 
    * @param   row
    *          row widget to use in choose
    */
   public void addRow(GenericWidget row)
   {
      if (defined)
         throw new IllegalArgumentException("multiple items including ROW");

      defined  = true;
      fieldMode = false;
      this.row = row;
   }

   /**
    * Sets the <b>ROW </b> mode and stores the row ID along with its help text
    * in the <b>CHOOSE </b> instance.
    * 
    * @param   row
    *          row widget to use in choose
    * @param   help
    *          associated help text
    */
   public void addRow(GenericWidget row, String help)
   {
      if (defined)
         throw new IllegalArgumentException("multiple items including ROW");

      defined  = true;
      fieldMode = false;
      this.row = row;
      rowHelp  = help;
   }
   
   /**
    * Adds a field to the <b>CHOOSE </b> instance.
    * 
    * @param   field
    *          widget within the frame
    */
   public void addField(GenericWidget field)
   {
      if (defined && !fieldMode)
         throw new IllegalArgumentException("multiple items including ROW");

      defined  = true;
      fieldMode = true;
      fields.addLast(field);
      fieldHelps.addLast(null);
   }

   /**
    * Adds array of fields to the <b>CHOOSE </b> instance.
    * 
    * @param   fields
    *          Array of widgets within the frame
    */
   public void addField(GenericWidget[] fields)
   {
      if (fields != null)
      {
         for (int i = 0; i < fields.length; i++)
         {
            addField(fields[i]);
         }
      }
   }
   
   /**
    * Adds a field with its help text to the <b>CHOOSE </b> instance.
    * 
    * @param   field
    *          widget within the frame
    * @param   help
    *          associated help text
    */
   public void addField(GenericWidget field, String help)
   {
      if (defined && !fieldMode)
         throw new IllegalArgumentException("multiple items including ROW");

      defined  = true;
      fieldMode = true;
      fields.addLast(field);
      fieldHelps.addLast(help);
   }

   /**
    * Add help string to last N items.
    * 
    * @param   help
    *          Help string.
    * @param   num
    *          Number of widgets.
    */
   public void addHelp(String help, int num)
   {
      if (fieldHelps.size() < num)
         throw new RuntimeException("Attempt to set " + num + 
                                    " help strings, while only " + 
                                    fieldHelps.size() + " items available.");
      
      for (int i = 0; i < num; i++)
         fieldHelps.removeLast();
      
      for (int i = 0; i < num; i++)
         fieldHelps.addLast(help);
   }
   
   /**
    * Turns on the <b>AUTO-RETURN </b> option.
    */
   public void setAutoReturn()
   {
      autoReturn = true;
   }

   /**
    * Specifies a color for this <b>CHOOSE </b> instance.
    * 
    * @param   c
    *          color specification instance
    */
   public void setColor(ColorSpec c)
   {
      if (c != null)
         color = new ColorSpec(c);
   }

   /**
    * Adds a key to the list of <b>GO-ON </b> keys.
    * 
    * @param   key
    *          array of key labels
    */
   public void addGoOn(String[] key)
   {
      for (int i = 0; i < key.length; i ++)
         goOn.add(key[i]);
   }

   /**
    * Adds a key to the list of <b>GO-ON </b> keys.
    * 
    * @param   key
    *          key label
    */
   public void addGoOn(String key)
   {
      goOn.add(key);
   }

   /**
    * Highlights the matching item in the <b>CHOOSE </b> and saves the user
    * entered keystrokes.
    * 
    * @param   keysVar
    *          Accessor for variable that contains keystrokes for an item and
    *          receives keystrokes (up to 40) while the user navigates
    */
   public void setKeys(BaseDataType keysVar)
   {
      this.keysVar = new AccessorWrapper(keysVar);
   }

   /**
    * Turns on the <b>NO-ERROR</b> option.
    */
   public void setNoError()
   {
      noError = true;
   }

   /**
    * Specifies the <b>CHOOSE </b> timeout in seconds.
    * 
    * @param   seconds
    *          maximum time between keystrokes
    */
   public void setPause(NumberType seconds)
   {
      pause = seconds.intValue();
   }

   /**
    * Specifies the <b>CHOOSE </b> timeout in seconds.
    * 
    * @param   seconds
    *          maximum time between keystrokes
    */
   public void setPause(double seconds)
   {
      pause = (int)seconds;
   }

   /**
    * Queries the <b>ROW </b> mode of <b>CHOOSE </b>.
    * 
    * @return  <code>true</code> if the <b>CHOOSE </b> specifies rows,
    *          otherwise <code>false</code>
    */
   public boolean isRow()
   {
      return fieldMode == false;
   }

   /**
    * Gets the row widget if the <b>CHOOSE </b> is for rows.
    * 
    * @return  row widget or <code>null</code> if this <b>CHOOSE </b> is for
    *          fields.
    */
   public GenericWidget getRow()
   {
      return (fieldMode == false) ? row : null;
   }

   /**
    * Checks whether the row was specified with a custom help text.
    * 
    * @return  <code>true</code> if there was a custom help text specified
    */
   public boolean isRowHelp()
   {
      return rowHelp != null;
   }

   /**
    * Gets the custom help text for the row.
    * 
    * @return  custom help text or <code>null</code>
    */
   public String getRowHelp()
   {
      return rowHelp;
   }

   /**
    * Gets the count of fields for the <b>FIELD </b> mode <b>CHOOSE </b>.
    * 
    * @return  number of fields defined or 0
    */
   public int getFieldsCount()
   {
      return fields.size();
   }

   /**
    * Gets the array of field widgets for the <b>CHOOSE </b>.
    * 
    * @return  array of field widgets
    */
   public GenericWidget[] getFields()
   {
      return fields.toArray(new GenericWidget[] {});
   }

   /**
    * Checks whether the indexed field was specified with a custom help text.
    * 
    * @param   index
    *          index of the field being checked
    * 
    * @return  <code>true</code> if there was a custom help text specified
    */
   public boolean isFieldHelp(int index)
   {
      if (fieldMode == false || index < 0 || index >= fieldHelps.size())
         return false;

      return fieldHelps.get(index) != null;
   }

   /**
    * Gets the custom help text for the indexed field.
    * 
    * @param   index
    *          index of the field being queried
    * 
    * @return  custom help text or <code>null</code>
    */
   public String getFieldHelp(int index)
   {
      if (fieldMode == false || index < 0 || index >= fieldHelps.size())
         return null;

      return fieldHelps.get(index);
   }

   /**
    * Gets the help texts for all fields.
    * 
    * @return  array of custom help texts or <code>null</code> if in <b>ROW
    *          </b> mode.
    */
   public String[] getFieldsHelp()
   {
      if (fieldMode == false)
         return null;

      return fieldHelps.toArray(new String[] {});
   }

   /**
    * Checks whether the <b>AUTO-RETURN </b> option was specified.
    * 
    * @return  <code>true</code> if <b>AUTO-RETURN </b> option was set
    */
   public boolean getAutoReturn()
   {
      return autoReturn;
   }

   /**
    * Gets the custom color specification.
    * 
    * @return  <code>ColorSpec</code> for the custom color definition or
    *          <code>null</code> if it wasn't specified.
    */
   public ColorSpec getColor()
   {
      return color;
   }

   /**
    * Gets the count of <b>GO-ON </b> keys defined.
    * 
    * @return  number of <b>GO-ON </b> keys or 0
    */
   public int getGoOnCount()
   {
      return goOn.size();
   }

   /**
    * Gets the indexed <b>GO-ON </b> key as a key label or function.
    * 
    * @param   index
    *          index of the field being queried
    * 
    * @return  a string which may be key label or key function, or
    *          <code>null</code> if the index is out of range.
    */
   public String getGoOn(int index)
   {
      if (index < 0 || index >= goOn.size())
         return null;

      return goOn.get(index);
   }

   /**
    * Gets all the <b>GO-ON </b> keys.
    * 
    * @return  array of strings which are key label or key function, possibly
    *          empty.
    */
   public String[] getGoOn()
   {
      return goOn.toArray(new String[] {});
   }

   /**
    * Gets the initial keystrokes if any are defined..
    * 
    * @return  a string of keystrokes or <code>null</code>
    */
   public String getKeys()
   {
      if (keysVar == null || keysVar.get().isUnknown() ||
          !(keysVar.get() instanceof character))
         return null;

      return ((character) keysVar.get()).getValue();
   }

   /**
    * Get keys variable.
    * 
    * @return  keys variable.
    */
   public Accessor getKeysVar()
   {
      return keysVar;
   }
   
   /**
    * Checks whether the <b>NO-ERROR </b> option was specified.
    * 
    * @return  <code>true</code> if <b>NO-ERROR </b> option was set
    */
   public boolean getNoError()
   {
      return noError;
   }

   /**
    * Gets the timeout value.
    * 
    * @return  timeout value in seconds or -1 if not set.
    */
   public int getPause()
   {
      return pause;
   }
   
   /**
    * Get <code>ChooseData</code> instance usable at client side.
    * 
    * @return  instance of <code>ChooseData</code>
    */
   public ChooseData getData()
   {
      ChooseData data = new ChooseData(fields.size(), goOn.size());

      // simple variables
      data.setAutoReturn(autoReturn);
      data.setColor(color);
      
      data.setKeys(getKeys());
      
      data.setNoError(noError);
      data.setPause(pause);
      data.setFieldMode(fieldMode);
      data.setRowHelp(rowHelp);
      
      if (row != null)
         data.setRowId(row.getId());
      
      // arrays
      GenericWidget[] flist = getFields();
      
      for(int i = 0; i < flist.length; i++)
      {
         data.setFields(i, flist[i].getId());
         
         String text = getFieldHelp(i);
         
         data.setFieldHelp(i, text);
      }
      
      String[] list = goOn.toArray(new String[goOn.size()]); 
      
      for(int i = 0; i < list.length; i++)
         data.setGoOn(i, list[i]);

      return data;
   }
}