/*
** Module   : Color.java
** Abstract : A color and attribute container for GUI clients.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20171009 Created initial version
** 002 CA  20171019 Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.util.integer;

/**
 * This represent single logical color as it is seend by the application code.
 * Interpretation of this color and mapping to actual terminal capabilities is
 * up to the low level driver.
 */
public class ColorGui
extends Color
{
   /** Background color index. */
   private int bgcolor;

   /** Foreground color index. */
   private int fgcolor;
   
   /**
    * Construct an instance using specified foreground and background colors.
    *
    * @param   fgcolor
    *          Foreground color.
    * @param   bgcolor
    *          Background color.
    */
   public ColorGui(int fgcolor, int bgcolor)
   {
      this.fgcolor = fgcolor;
      this.bgcolor = bgcolor;
   }

   /**
    * Default constructor.
    */
   public ColorGui()
   {
      bgcolor = -1;
      fgcolor = -1;
   }

   /**
    * Execute the COLOR DISPLAY statement for the specified widget.
    * 
    * @param    widget
    *           The widget instance.
    */
   public void colorDisplay(GenericWidget<?> widget)
   {
      if (bgcolor != -1)
      {
         widget.setBgColor(new integer(bgcolor));
      }
      if (fgcolor != -1)
      {
         widget.setFgColor(new integer(fgcolor));
      }
   }

   /**
    * Execute the COLOR PROMPT statement for the specified widget.  This is a no-op in GUI.
    * 
    * @param    widget
    *           The widget instance.
    */
   public void colorPrompt(GenericWidget<?> widget)
   {
      // no-op
   }

   /**
    * Get bgcolor.
    *
    * @return   bgdcolor.
    */
   public int getBgcolor()
   {
      return bgcolor;
   }

   /**
    * Get fgcolor.
    *
    * @return   fgcolor.
    */
   public int getFgcolor()
   {
      return fgcolor;
   }

   /**
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null || getClass() != obj.getClass())
         return false;

      ColorGui other = (ColorGui) obj;

      return (bgcolor == other.bgcolor) && (fgcolor == other.fgcolor);
   }
   
   /**
    * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
    */
   @Override
   public void writeExternal(ObjectOutput out) throws IOException
   {
      out.writeInt(fgcolor);
      out.writeInt(bgcolor);
   }
   
   /**
    * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
    */
   @Override
   public void readExternal(ObjectInput in) throws IOException
   {
      fgcolor = in.readInt();
      bgcolor = in.readInt();
   }
}