/*
** Module   : ControlFrameWidget.java
** Abstract : An implementation of the server-side component that holds a Java emulated ActiveX.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit. Implemented methods and attributes: LoadControls, COM-HANDLE,
**                  LEFT, TOP, CONTROL-NAME, CONTROLS, HEIGHT, WIDTH and WIDGET-HANDLE.
** 002 CA  20180517 Fixed comhandle FWD-specific resource management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.comauto.*;
import com.goldencode.p2j.util.*;
import java.util.*;

/**
 * A control-frame is a field-level widget that holds a Java emulated ActiveX control. It
 * implements the {@code ControlFrame} interface.
 * <p>
 * The ActiveX that you select for your application from the OpenEdge AppBuilder is replaced by
 * an object completely rewritten in java.
 * <p>
 * These objects are always created dynamically. 
 */
@LegacyResource(resource = LegacyResource.CONTROL_FRAME)
public class ControlFrameWidget
extends ControlEntity<ControlFrameConfig>
implements ControlFrame
{
   /** The COM Object associated with this Control Frame */ 
   private ControlFrameComObjectImpl comObject = new ControlFrameComObjectImpl();
   
   /** The ActiveX-es loaded in this Control Frame, mapped by their names. */ 
   private Map<String, ComObject> ax = null;
   
   /**
    * Default constructor.
    *
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ControlFrameWidget(boolean dynamic)
   {
      super(dynamic, new ControlFrameConfig());
   }
   
   /**
    * Gets the Progress COM-HANDLE attribute. Returns a {@code comhandle} to the internal
    * {@code ControlFrameComObject} that can be used to load and access the COMs/ActiveX-es
    * from this Control Frame.
    *
    * @return   The value for the COM-HANDLE attribute.
    */
   @Override
   public comhandle getComHandle()
   {
      return new comhandle(comObject);
   }
   
   /**
    * Cleanup this dynamic widget before deletion.
    */
   @Override
   public void cleanup()
   {
      if (ax != null)
      {
         for (ComObject comOb : ax.values())
         {
            comOb.release();
            comOb.destroy();
         }
      }
      ax = null;
   }
   
   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!super.resourceDelete())
      {
         return false;
      }
      
      if (comObject != null)
      {
         comObject.destroy();
      }
      
      return true;
   }
      
   /**
    * Each Control Frame has an peer interface that allows to communicate with the ActiveX. 
    */
   private class ControlFrameComObjectImpl
   extends ComObject
   implements ControlFrameComObject
   {
      /**
       * Public default constructor.
       */
      public ControlFrameComObjectImpl()
      {
      }
      
      /**
       * Obtain a (com-)handle to a specific Java emulated ActiveX, if such instance exist.
       * Theoretically a {@code .wrx} file loaded with {@code loadControls()} may contain multiple
       * ActiveX-es. They are identified by their name as top-level properties of the 
       * {@code ControlFrame} widget.
       *
       * @param   axName
       *          The ActiveX name.
       *
       * @return  a {@code comhandle} to the requested activex object if such instance was loaded. 
       */
      public comhandle getActiveX(String axName)
      {
         if (ax == null)
         {
            return new comhandle();
         }
         return new comhandle(ax.get(axName.toUpperCase()));
      }
      
      /**
       * Obtain the name of this COM object.
       *
       * @return  the name of this COM object.
       */
      @Override
      public String getName()
      {
         return ControlFrameWidget.this.name().getValue();
      }
      
      /**
       * Releasing objects of this type is illegal. They need to be deleted using the handle to 
       * its Control Frame handle, that can be obtained using {@code WIDGET-HANDLE}. This method
       * will just throw error 5895.
       */
      @Override
      public void release()
      {
         String msg = "A CONTROL-FRAME must be deleted through its WIDGET-HANDLE," +
                      " not released through its COMPONENT-HANDLE";
         ErrorManager.recordOrThrowError(5895, msg, false);
      }
      
      /**
       * Reports if this object is valid for use.  
       *
       * @return  {@code true} if a COM object was loaded in the Control Frame.  
       */
      @Override
      public boolean valid()
      {
         return ax != null;
      }
      
      /**
       * Loads the named OCX into the associated control frame.
       *
       * @param    wrxFile
       *           .wrx filename describing the OCX content to load.
       * @param    ctrl
       *           Control frame name.
       *
       * @return   {@code true} if operation was successful.
       */
      @Override
      public logical loadControls(character wrxFile, character ctrl)
      {
         if (wrxFile == null || ctrl == null || wrxFile.isUnknown() || ctrl.isUnknown())
         {
            comhandle.showAccessError("LoadControls", "0x80020005", "Type mismatch"); // strange
            return new logical(false);
         }
         
         return loadControls(wrxFile.toStringMessage(), ctrl.toStringMessage());
      }
      
      /**
       * Loads the named OCX into the associated control frame.
       *
       * @param    wrxFile
       *           .wrx filename describing the OCX content to load.
       * @param    ctrl
       *           Control frame name.
       *
       * @return   {@code true} if operation was successful.
       */
      @Override
      public logical loadControls(String wrxFile, character ctrl)
      {
         if (wrxFile == null || ctrl == null || ctrl.isUnknown())
         {
            comhandle.showAccessError("LoadControls", "0x80020005", "Type mismatch"); // strange
            return new logical(false);
         }
         
         return loadControls(wrxFile, ctrl.toStringMessage());
      }
      
      /**
       * Loads the named OCX into the associated control frame.
       *
       * @param    wrxFile
       *           .wrx filename describing the OCX content to load.
       * @param    ctrl
       *           Control frame name. 
       *
       * @return   {@code true} if operation was successful.
       */
      @Override
      public logical loadControls(character wrxFile, String ctrl)
      {
         if (wrxFile == null || ctrl == null || wrxFile.isUnknown())
         {
            comhandle.showAccessError("LoadControls", "0x80020005", "Type mismatch"); // strange
            return new logical(false);
         }
         
         return loadControls(wrxFile.toStringMessage(), ctrl);
      }
      
      /**
       * Loads the named OCX into the associated control frame.
       *
       * @param    wrxFile
       *           .wrx filename describing the OCX content to load.
       * @param    ctrl
       *           Control frame name. 
       *
       * @return   {@code true} if operation was successful.
       */
      @Override
      public logical loadControls(String wrxFile, String ctrl)
      {
         ControlFrameWidget.this.cleanup();
         ax = new HashMap<>(); // drop all existing AX-es, they were unregistered first
         
         // 'open' wrxFile, for each ActiveX, instantiate an ActiveX
         /* foreach (entry : wrxFile) */
         {
            comhandle ch = new comhandle();
            // for the moment [PSTimer] is the only supported COM object
            String axName = "PSTimer".toUpperCase();
            ComServer.create(axName, ch);
            ComObject newComObject = ch.getResource();
            newComObject.setParentControlFrame(ControlFrameWidget.this);
            ax.put(axName, newComObject);
         }
         
         return new logical(true);
      }
      
      /**
       * The component handle to an ActiveX control that has the specified design-time name 
       * (Control-Name) and that is loaded into the control-frame.
       *
       * @return  references to COM object properties and methods extend the syntax used for
       *          referencing widget attributes and methods.
       */
      @Override
      public comhandle getControlName()
      {
         // NOTE: Not enough information to implement this correctly
         //       in ABL the value returned by this method seems to be the ZERO_RESOURCE object
         return new comhandle(handle.fromResourceId(0L));
      }
      
      /**
       * Returns a control collection from this control frame.
       *
       * @return  The collection of controls from this control frame.
       */
      @Override
      public comhandle getControls()
      {
         // NOTE: for the moment [PSTimer] is the only implementation of an ActiveX emulated
         //       Object so it is returned 
         return getActiveX("PSTimer");
      }
      
      /**
       * Returns the handle of the control frame associated with the control-frame COM object.
       *
       * @return  The handle of the control frame associated with the control-frame COM object.
       */
      @Override
      public handle getWidgetHandle()
      {
         return new handle(ControlFrameWidget.this);
      }
      
      /**
       * Obtain the width of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @return  The width of the widget or the screen display used in the current session.
       */
      @Override
      public integer getWidth()
      {
         return getWidthPixels();
      }
      
      /**
       * Sets the width of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @param   width
       *          the new width of the widget or the screen display used in the current session.
       */
      @Override
      public void setWidth(long width)
      {
         setWidthPixels((int) width);
      }
      
      /**
       * Sets the width of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @param   width
       *          the new width of the widget or the screen display used in the current session.
       */
      @Override
      public void setWidth(NumberType width)
      {
         setWidthPixels(width);
      }
      
      /**
       * Obtain the height of the widget or the screen display used in the current session,
       * in pixels.
       *
       * @return  the height of the widget or the screen display used in the current session.
       */
      @Override
      public integer getHeight()
      {
         return getHeightPixels();
      }
      
      /**
       * Sets the new height of the widget or the screen display used in the current session,
       * in pixels.
       *
       * @param   height
       *          the new height of the widget or the screen display used in the current session.
       */
      @Override
      public void setHeight(long height)
      {
         setHeightPixels((int) height);
      }
      
      /**
       * Sets the new height of the widget or the screen display used in the current session,
       * in pixels.
       *
       * @param   height
       *          the new height of the widget or the screen display used in the current session.
       */
      @Override
      public void setHeight(NumberType height)
      {
         setHeightPixels(height);
      }
      
      /**
       * Obtain the left of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @return  The left of the widget or the screen display used in the current session.
       */
      @Override
      public integer getLeft()
      {
         return getX();
      }
      
      /**
       * Sets the left of the widget or the screen display used in the current session, in pixels.
       *
       * @param   left
       *          the new left of the widget or the screen display used in the current session.
       */
      @Override
      public void setLeft(long left)
      {
         setX((int) left);
      }
      
      /**
       * Sets the left of the widget or the screen display used in the current session, in pixels.
       *
       * @param   left
       *          the new left of the widget or the screen display used in the current session.
       */
      @Override
      public void setLeft(NumberType left)
      {
         setX(left);
      }
      
      /**
       * Obtain the top of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @return  the top of the widget or the screen display used in the current session.
       */
      @Override
      public integer getTop()
      {
         return getY();
      }
      
      /**
       * Sets the new top of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @param   top
       *          the new top of the widget or the screen display used in the current session.
       */
      @Override
      public void setTop(long top)
      {
         setY((int) top);
      }
      
      /**
       * Sets the new top of the widget or the screen display used in the current session, in
       * pixels.
       *
       * @param   top
       *          the new top of the widget or the screen display used in the current session.
       */
      @Override
      public void setTop(NumberType top)
      {
         setY(top);
      }
      
      /**
       * Gets the WIDGET-ID writable attribute.
       *
       * @return  the widget id of the parent Control Fame.
       */
      @Override
      public integer getWidgetId()
      {
         return ControlFrameWidget.this.getWidgetId();
      }
      
      /**
       * Gets the FRAME attribute, which is the containing frame instance, as a handle.
       *
       * @return   A handle to the frame that contains this Control Frame widget.
       */
      @Override
      public handle getFrameHandle()
      {
         // NOTE: Not enough information to implement this correctly, like this:
         // return ControlFrameWidget.this.getFrameHandle();
         // in ABL, the value returned by this method seems to be the ZERO_RESOURCE object
         return handle.fromResourceId(0L);
      }
      
      /**
       * Obtains the {@code name} attribute.
       *
       * @return   See above.
       */
      @Override
      public character name()
      {
         return ControlFrameWidget.this.name();
      }
      
      /**
       * Sets the {@code name} attribute.
       *
       * @param   name
       *          The new name to be set.
       */
      @Override
      public void name(String name)
      {
         ControlFrameWidget.this.name(name);
      }
      
      /**
       * Sets the {@code name} attribute.
       *
       * @param   name
       *          The new name to be set.
       */
      @Override
      public void name(character name)
      {
         ControlFrameWidget.this.name(name);
      }
   }
}
