/*
** Module   : EditorWidget.java
** Abstract : server side Editor widget implementation
**
** Copyright (c) 2006-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 SIY 20060427   @26084 Server side part of the EDITOR widget.
** 002 GES 20060803   @28373 Cleanup and refactoring.
** 003 GES 20080320   @37585 Added TYPE attribute.
** 004 SIY 20111101          Updated imports.
** 005 CA  20130129          Added setLarge stub, for EDITOR LARGE.
** 006 CA  20130221          Refactored usage of TYPE attribute.
** 007 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 008 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 009 CA  20131115          Changes related to properly instantiate dynamic resources.
** 010 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 011 SVL 20141102          Make this class implement ScrollbarVerticalElement.
** 012 GES 20141124          Added EditorInterface support (with all associated attrs/meths).
** 013 GES 20150103          Added runtime support for new attrs/methods.
** 014 EVK 20150209          Added runtime support for save-file and read-file methods.
** 015 IAS 20150318          Added runtime support for new attrs/methods (
**                           BOX, RETURN-INSERTED. CURSOR-XXX, REPLACE...)
** 016 HC  20150323          Removed business logic from config classes.
** 017 CA  20150509          Moved logic from afterConfigUpdate to the client-side, as it's 
**                           dependent on the client's font and mode (GUI/ChUI).
** 018 IAS 20150604          Refactored to the ReplaceInterface
** 019 CA  20150615          Added support for MAX-CHARS and WORD-WRAP.
** 020 CA  20150910          If WORD-WRAP is on, horizontal scrollbar must be off.
** 021 IAS 20151214          LOAD-MOUSE-POINTER support.
** 022 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 023 IAS 20160331          Fixed potential NPE
** 024 HC  20160922          Fixed an NPE.
** 025 CA  20161005          Fixed issues related to SCREEN-VALUE attribute (initial value or 
**                           explicitly set, when the value is empty string or unknown, or the 
**                           widget is uninitialized).
** 026 OM  20170504          Added support for MOVE-TO-EOF, SEARCH and INSERT-FILE. Added
**                           FIND_ public constants.
** 027 EVL 20170728          Adding javadoc fixes.
** 028 SVL 20171025          Fixed widget name in error messages.
** 029 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
** 030 SBI 20180202          Changed to update the default wordWrap value.
** 031 OM  20180525          The [modified] attribute is not changed by UI statements.
** 032 CA  20180502          Added LARGE attribute - stubs.
** 033 OM  20180613          Added valid widget guard for setScreenValue.
** 034 EVL 20180618          Removing warning log from setLarge() method, just do nothing there.
** 035 EVL 20180814          Adding method to check if the widget has valid format string.  Editor
**                           widget does not have the FORMAT attribute to use.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side implementation of the EDITOR widget.
 */
@LegacyResource(resource = LegacyResource.EDITOR)
public class EditorWidget
extends ControlTextWidget<EditorConfig>
implements EditorInterface,
           ReplaceInterface,
           MaxCharsElement,
           BoxInterface,
           ScrollbarVerticalElement,
           ScrollbarHorizontalElement,
           InnerLines,
           SelectableText,
           Hoverable
{
   /** Find &amp; replace flags: find next occurrence from the cursor. */
   public static final int FIND_NEXT_OCCURRENCE = 1;
   
   /** Find &amp; replace flags: find previous occurrence from the cursor. */
   public static final int FIND_PREV_OCCURRENCE = 2;
   
   /** Find &amp; replace flags: the search is performed in case sensitive. */
   public static final int FIND_CASE_SENSITIVE = 4;
   
   /** Find &amp; replace flags. */
   public static final int FIND_GLOBAL = 8;
   
   /**
    * Find &amp; replace flags: search loops from the beginning if the pattern is not found from the
    * current cursor position up to the end of the text. 
    */
   public static final int FIND_WRAP_AROUND = 16;
   
   /**
    * Find &amp; replace flags: the occurrence is selected, otherwise the cursor is moved at its
    * beginning.
    */
   public static final int FIND_SELECT = 32;
   
   /**
    * Default constructor.
    */
   public EditorWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public EditorWidget(boolean dynamic)
   {
      super(dynamic, new EditorConfig());
   }

   /**
    * Mark this EDITOR widget as EDITOR LARGE, which can be used to display a value of type 
    * <code>LONGCHAR</code>.
    * 
    * @param    large
    *           When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @Override
   public void setLarge(boolean large)
   {
      // nothing to do here, notification removed
   }

   /**
    * Mark this EDITOR widget as EDITOR LARGE, which can be used to display a value of type 
    * <code>LONGCHAR</code>.
    * 
    * @param    large
    *           When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @Override
   public void setLarge(logical large)
   {
      setLarge(large.booleanValue());
   }
   
   /**
    * Get the EDITOR's LARGE attribute.
    * 
    * @return   When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @Override
   public logical isLarge()
   {
      UnimplementedFeature.missing("isLarge() method");
      return new logical();
   }
   
   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   public integer getInnerLines()
   {
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget. For the INNER-LINES clause in the VIEW-AS phrase
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInitialInnerLines(double lines)
   {
      setInnerLines(lines);
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(double lines)
   {
      // WARNING: even if the lines are the same, this can't be a no-op.
      // also, if the initial assign value is zero, it defaults to 4
      
      config.innerLines = (int) lines;
      pushWidgetAttr("innerLines", config.innerLines);
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         UnimplementedFeature.todo("INNER-LINES attribute unknown behavior is not known.");
         return;
      }
   
      setInnerLines(lines.intValue());
   }

   /**
    * Gets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @return   The number of character columns.
    */
   public integer getInnerChars()
   {
      return new integer(config.innerChars);
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(double chars)
   {
      config.innerChars = (int) chars;
      pushWidgetAttr("innerChars", config.innerChars);
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(NumberType chars)
   {
      if (chars == null || chars.isUnknown())
         return;
      
      setInnerChars(chars.intValue());
   }

   /**
    * Set value of the MAX-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   @Override
   public void setMaxChars(double chars)
   {
      setMaxChars(new decimal(chars));
   }

   /**
    * Set value of the MAX-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   @Override
   public void setMaxChars(NumberType chars)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set RETURN-INSERTED because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      int mc = (chars == null || chars.isUnknown() ? null : chars.intValue());
      
      if (config.maxChars != null && config.maxChars == mc)
      {
         return;
      }
      
      config.maxChars = mc;
      pushWidgetAttr("maxChars", config.maxChars);
   }
   
   /**
    * Get value of the MAX-CHARS attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public integer getMaxChars()
   {
      return config.maxChars == null ? new integer() : new integer(config.maxChars);
   }
   
   /**
    * Set value of the BUFFER-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setBufferChars(final int chars)
   {
      config.bufferChars = chars;
      pushWidgetAttr("bufferChars", config.bufferChars);
   }
   
   /**
    * Set value of the BUFFER-CHARS attribute.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setBufferChars(NumberType chars)
   {
      if (chars.isUnknown())
         return;
      
      setBufferChars(chars.intValue());
   }
   
   /**
    * Set value of the BUFFER-LINES attribute.
    * 
    * @param   lines
    *          The new value for the attribute.
    */
   public void setBufferLines(final int lines)
   {
      config.bufferLines = lines;
      pushWidgetAttr("bufferLines", config.bufferLines);
   }
   
   /**
    * Set value of the BUFFER-LINES attribute.
    * 
    * @param   lines
    *          The new value for the attribute.
    */
   public void setBufferLines(NumberType lines)
   {
      if (lines.isUnknown())
         return;
      
      setBufferLines(lines.intValue());
   }
   
   /**
    * Set value of the BOX attribute as opposite to passed value.
    * 
    * @param   box
    *          The new value for the attribute.
    */
   public void setNoBox(boolean box)
   {
      if (config.box == !box)
      {
         return;
      }
      
      config.box = !box;
      pushWidgetAttr("box", config.box);
   }

   /**
    * Set value of the BOX attribute as opposite to passed value.
    * 
    * @param   box
    *          The new value for the attribute.
    */
   public void setNoBox(logical box)
   {
      if (box.isUnknown())
         return;
      
      setNoBox(box.booleanValue());
   }
   
   /**
    * Set value of the WORD-WRAP attribute as opposite to passed value.
    * 
    * @param   wordwrap
    *          The new value for the attribute.
    */
   public void setNoWordWrap(boolean wordwrap)
   {
      if (config.wordWrap == !wordwrap)
      {
         return;
      }
      
      config.wordWrap = !wordwrap;
      pushWidgetAttr("wordWrap", config.wordWrap);
   }

   /**
    * Set value of the WORD-WRAP attribute as opposite to passed value.
    * 
    * @param   wordwrap
    *          The new value for the attribute.
    */
   public void setNoWordWrap(logical wordwrap)
   {
      if (wordwrap.isUnknown())
         return;
      
      setNoWordWrap(wordwrap.booleanValue());
   }
   
   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param   vertical
    *          The new value for the attribute.
    */
   public void setScrollbarVertical(final boolean vertical)
   {
      if (config.scrollbarVertical == vertical)
      {
         return;
      }
      
      config.scrollbarVertical = vertical;
      pushWidgetAttr("scrollbarVertical", config.scrollbarVertical);
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param   vertical
    *          The new value for the attribute.
    */
   public void setScrollbarVertical(logical vertical)
   {
      if (vertical == null || vertical.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-VERTICAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarVertical(vertical.booleanValue());
   }

   /**
    * Get value of the SCROLLBAR-VERTICAL attribute.
    *
    * @return  value of the attribute.
    */
   public logical isScrollbarVertical()
   {
      return new logical(config.scrollbarVertical);
   }
   
   /**
    * Get the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    * 
    * @return   The value of the SCROLLBAR-HORIZONTAL attribute.
    */
   public logical isScrollbarHorizontal()
   {
      return new logical(config.scrollbarHorizontal);
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(logical horiz)
   {
      if (horiz == null || horiz.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-HORIZONTAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarHorizontal(horiz.booleanValue());
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(final boolean horiz)
   {
      if (config.scrollbarHorizontal == horiz)
      {
         return;
      }
      
      if (config.scrollbarHorizontal)
      {
         config.wordWrap = false;
      }      
      config.scrollbarHorizontal = horiz;
      
      pushWidgetAttr(new String[] {"wordWrap", "scrollbarHorizontal"}, 
                     new Object[] {config.wordWrap, config.scrollbarHorizontal});
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int64 end)
   {
      if (start == null || start.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int end)
   {
      if (start == null || start.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int64 end)
   {
      if (end == null || end.isUnknown())
      {
         config.selectionActive = false;
         return new logical(true);
      }
      return setSelection(start, end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int end)
   {
      return new logical(LogicalTerminal.setEditorSelection(getId(), start, end));
   }
   
   /**
    * Clears text selection.
    * TODO: not working, EditorGuiImpl is not able to 'see' the [selectionActive] altered
    */
   @Override
   public logical clearSelection()
   {
      boolean wasSelection = config.selectionActive;
      config.selectionActive = false;
      pushWidgetAttr("selectionActive", config.selectionActive);
      return new logical(wasSelection);
   }
   
   /**
    * Removes the currently selected text.
    * 
    * @return always {@code yes}.
    */
   @Override
   public logical removeSelection()
   {
      LogicalTerminal.removeEditSelection(getId());
      return new logical(true);
   }
   
   /**
    * Implements the DELETE-LINE() method which deletes the line of text in which the cursor is
    * positioned.
    *
    * @return   <code>true</code> if the line was deleted.
    */
   public logical deleteLine()
   {
      return new logical(LogicalTerminal.deleteEditorLine(getId()));
   }
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   public logical insertString(character txt)
   {
      return insertString(txt == null || txt.isUnknown() ? "" : txt.toStringMessage());
   }
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   public logical insertString(String txt)
   {
      return new logical(LogicalTerminal.insertEditorString(getId(), txt == null ? "" : txt));
   }
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   public logical readFile(character name)
   {
      if (name == null || name.isUnknown())
      {
         return new logical(false);
      }
      
      boolean res = LogicalTerminal.editorReadFile(config.id.asInt(), name.toStringMessage());
      
      return new logical(res);
   }
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   public logical readFile(String name)
   {
      return readFile(new character(name));
   }
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replaceSelectionText(character txt)
   {
      return replaceSelectionText(txt == null || txt.isUnknown() ? "" : txt.toStringMessage());
   }
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replaceSelectionText(String txt)
   {
      if (!config.selectionActive)
      {
         return new logical(false);
      }
      return new logical(
               LogicalTerminal.replaceEditorSelectionText(getId(), txt == null ? "" : txt)
            );
   }

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @Override
   public logical saveFile(String name)
   {
      return saveFile(new character(name));
   }

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @Override
   public logical saveFile(character name)
   {
      if (name == null || name.isUnknown())
      {
         return new logical(false);
      }

      boolean res = LogicalTerminal.editorSaveFile(config.id.asInt(), name.toStringMessage());

      return new logical(res);
   }

   /**
    * Get the value of the CURSOR-CHAR attribute which is the current cursor position as a
    * 1-based index on the current line.
    *
    * @return   The current position of the cursor on the current line.
    */
   public integer getCursorChar()
   {
      return new integer(config.cursorChar);
   }
   
   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The 1-based position to which to move.
    */
   public void setCursorChar(double idx)
   {
      if (idx < 1)
      {
         invalidPosition("CURSOR-CHAR", 0);
         return;
      }

      Boolean result = LogicalTerminal.setEditorCursorChar(getId(), (int)idx);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-CHAR return false compatibility testing is needed");
      }
      else if (!result.booleanValue())
      {
         invalidPosition("CURSOR-CHAR", (int)idx);
      }
   }

   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The 1-based position to which to move.
    */
   public void setCursorChar(NumberType idx)
   {
      if (idx == null || idx.isUnknown())
      {
         invalidPosition("CURSOR-CHAR", 0);
         return;
      }
      setCursorChar(idx.doubleValue()); 
   }

   /**
    * Get the value of the CURSOR-LINE attribute which is the current cursor position as a
    * 1-based line number.
    *
    * @return   The current position of the cursor on the current line.
    */
   public integer getCursorLine()
   {
      return new integer(config.cursorY + config.scrollY + 1);
   }
   
   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   public void setCursorLine(double line)
   {
      if (line < 1)
      {
         invalidPosition("CURSOR-LINE", 0);
         return;
      }
      
      Boolean result = LogicalTerminal.setEditorCursorLine(getId(), (int)line);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-LINE return false compatibility testing is needed");
      }
      else if (!result.booleanValue())
      {
         invalidPosition("CURSOR-LINE", (int)line);
      }
   }

   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   public void setCursorLine(NumberType line)
   {
      if (line == null || line.isUnknown())
      {
         invalidPosition("CURSOR-LINE", 0);
         return;
      }
      
      setCursorLine(line.doubleValue()); 
   }
   
   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in the  EDITOR widget.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @Override
   public integer getCursorOffset()
   {
      return new integer(config.cursorOffset);
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in the EDITOR widget.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(double offset)
   {
      if (offset < 1)
      {
         invalidPosition("CURSOR-OFFSET", 0);
         return;
      }

      Boolean result = LogicalTerminal.setEditorCursorOffset(getId(), (int) offset);
      if (result == null)
      {
         UnimplementedFeature.todo("CURSOR-OFFSET return false compatibility testing is needed");
      }
      else if (!result)
      {
         invalidPosition("CURSOR-OFFSET", (int) offset);
      }
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in the EDITOR widget.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(NumberType offset)
   {
      if (offset == null || offset.isUnknown())
      {
         invalidPosition("CURSOR-OFFSET", 0);
         return;
      }
      
      setCursorOffset(offset.doubleValue()); 
   }

   /**
    * Get the value of the NUM-LINES attribute which is the number of lines of text in the
    * editor.
    *
    * @return   The number of lines in the editor.
    */
   public integer getNumLines()
   {
      return new integer(config.numLines);
   }

   /**
    * Get the value of the RETURN-INSERTED attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isReturnInserted()
   {
      return new logical(config.returnInserted);
   }

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setReturnInserted(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute RETURN-INSERTED on %s %s", 
                  type(), widgetName()
               ), 
               false);
      }
      setReturnInserted(value.booleanValue());
   }

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setReturnInserted(boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set RETURN-INSERTED because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      if (config.returnInserted == value)
      {
         return;
      }
      
      config.returnInserted = value;
      pushWidgetAttr("returnInserted", config.returnInserted);
   }

   /**
    * Get the value of the WORD-WRAP attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isWordWrap()
   {
      return new logical(config.wordWrap);
   }

   /**
    * Set the value of the WORD-WRAP attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setWordWrap(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute WORD-WRAP on %s %s", 
                  type(), widgetName()
               ), 
               false);
      }
      setWordWrap(value.booleanValue());
   }

   /**
    * Set the value of the WORD-WRAP attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setWordWrap(final boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set WORD-WRAO because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      
      if (config.wordWrap == value && !value)
      {
         return;
      }

      if (value)
      {
         config.scrollbarHorizontal = false;
      }

      config.wordWrap = value;

      pushWidgetAttr(new String[] {"scrollbarHorizontal", "wordWrap"},
                     new Object[] {config.scrollbarHorizontal, config.wordWrap});
   }

   /**
    * Get the value of the BOX attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isBox()
   {
      return new logical(config.box);
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute BOX on %s %s", 
                  type(), widgetName()
               ), 
               false);
         
      }
      setBox(value.booleanValue());
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set BOX because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      if (config.box == value)
      {
         return;
      }
      
      config.box = value;
      pushWidgetAttr("box", config.box);
  }

   /**
    * Get the value of the LENGTH attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public integer getLength()
   {
      return new integer(config.length);
   }

   /**
    * Get SELECTION-TEXT attribute.
    *
    * @return  SELECTION-TEXT value   
    */
   @Override
   public character getSelectionText()
   {
      if (!config.selectionActive)
      {
         return new character("");
      }
      return new character(config.selectionText);
   }

   /**
    * Get SELECTION-END attribute.
    *
    * @return  SELECTION-END value   
    */
   @Override
   public integer getSelectionEnd()
   {
      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionEnd);
   }

   /**
    * Get SELECTION-START attribute.
    *
    * @return  SELECTION-START value   
    */
   @Override
   public integer getSelectionStart()
   {
      if (!config.selectionActive)
      {
         return new integer();
      }
      
      return new integer(config.selectionStart);
   }

   /**
    * Get TEXT-SELECTED attribute.
    *
    * @return  TEXT-SELECTED value   
    */
   @Override
   public logical isTextSelected()
   {
      return new logical(config.selectionActive);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, character to, integer flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, character to, integer flag)
   {
      return replace(from, to == null || to.isUnknown() ? "" : to.toStringMessage(), flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, String to, integer flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, character to, int flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, character to, int flag)
   {
      return replace(from, to == null || to.isUnknown() ? "" : to.toStringMessage(), flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, String to, int flag)
   {
      if (from == null || from.isUnknown())
      {
         // 4GL hangs in this case 
         return new logical(false);
      }
      return replace(from.toStringMessage(), to, flag);
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replace(String from, String to, integer flag)
   {
      if (flag == null || flag.isUnknown())
      {
         return new logical(false);
      }
      return replace(from, to, flag.intValue());
   }

   /**
    * Implements the REPLACE() method.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    * @param    flag
    *           The type of search to be performed..
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   public logical replace(String from, String to, int flag)
   {
      if (from == null || "".equals(from))
      {
         // 4GL crashes in this case 
         return new logical(false);
      }

      Boolean result = LogicalTerminal.replaceEditorString(getId(), from, to, flag);
      if (result == null)
         UnimplementedFeature.todo("REPLACE() return false compatibility testing is needed");
      return new logical(result);
   }

   /**
    * Implements the REPLACE() method with default FIND-NEXT-OCCURRENCE mode.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, character to)
   {
      return replace(from, to, 1);
   }

   /**
    * Implements the REPLACE() method with default FIND-NEXT-OCCURRENCE mode.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(character from, String to)
   {
      return replace(from, to, 1);
   }


   /**
    * Implements the REPLACE() method with default FIND-NEXT-OCCURRENCE mode.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, character to)
   {
      return replace(from, to, 1);
   }

   /**
    * Implements the REPLACE() method with default FIND-NEXT-OCCURRENCE mode.
    *
    * @param    from
    *           The source text.
    * @param    to
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @Override
   public logical replace(String from, String to)
   {
      return replace(from, to, 1);
   }

   /**
    * REPLACE(character, integer) method and its variations
    * not applicable to the EDITOR widget
    */
   @Override
   public logical replace(character newItemList, integer listIndex)
   {
      ErrorManager.recordOrShowError(4067, 
            String.format(
               "The second argument for REPLACE on the %s %s should not be of type INTEGER", 
               type(), widgetName()), 
            false);
      return new logical(false);
   }

   /**
    * REPLACE(character, integer) method and its variations
    * not applicable to the EDITOR widget
    */
   @Override
   public logical replace(character newItemList, int listIndex)
   {
      ErrorManager.recordOrShowError(4067, 
            String.format(
               "The second argument for REPLACE on the %s %s should not be of type INTEGER", 
               type(), widgetName()), 
            false);
      return new logical(false);
   }

   /**
    * REPLACE(character, integer) method and its variations
    * not applicable to the EDITOR widget
    */
   @Override
   public logical replace(String newItemList, integer listIndex)
   {
      ErrorManager.recordOrShowError(4067, 
            String.format(
               "The second argument for REPLACE on the %s %s should not be of type INTEGER", 
               type(), widgetName()), 
            false);
      return new logical(false);
   }

   /**
    * REPLACE(character, integer) method and its variations
    * not applicable to the EDITOR widget
    */
   @Override
   public logical replace(String newItemList, int listIndex)
   {
      ErrorManager.recordOrShowError(4067, 
            String.format(
               "The second argument for REPLACE on the %s %s should not be of type INTEGER", 
               type(), widgetName()), 
            false);
      return new logical(false);
   }

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, character oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(String newLabel, BaseDataType newValue, character oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(character newLabel, String  newValue, character oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }   
    
   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(String newLabel, String  newValue, character oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(character newLabel, BaseDataType  newValue, String oldItem)
   {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The third argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), "CHARACTER"), 
               false);
         return new logical(false);
   }

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(String newLabel, BaseDataType  newValue, String oldItem)
   {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The third argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), "CHARACTER"), 
               false);
         return new logical(false);
   }   
   
   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(character newLabel, String  newValue, String oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }   

   /**
   * REPLACE(character, BaseDataType, character) method and its variations
   * not applicable to the EDITOR widget
   */
   @Override
   public logical replace(String newLabel, String  newValue, String oldItem)
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "CHARACTER"), 
            false);
      return new logical(false);
   }   

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, integer listIndex)
   {
      if (newValue != null && !(newValue instanceof character))
      {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The second argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), newValue.getTypeName()), 
               false);
         return new logical(false);
      }
      return replace(newLabel, (character)newValue, listIndex);
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(String newLabel, BaseDataType  newValue, integer listIndex)
   {
      if (newValue != null && !(newValue instanceof character))
      {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The second argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), newValue.getTypeName()), 
               false);
         return new logical(false);
      }
      return replace(newLabel, (character)newValue, listIndex);
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType  newValue, int listIndex)
   {
      if (newValue != null && !(newValue instanceof character))
      {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The second argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), newValue.getTypeName()), 
               false);
         return new logical(false);
      }
      return replace(newLabel, (character)newValue, listIndex);
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "REPLACE")
   public logical replace(String newLabel, BaseDataType  newValue, int listIndex)
   {
      if (newValue != null && !(newValue instanceof character))
      {
         ErrorManager.recordOrShowError(4068, 
               String.format(
                  "The second argument for REPLACE on the %s %s should not be of type %s", 
                  type(), widgetName(), newValue.getTypeName()), 
               false);
         return new logical(false);
      }
      return replace(newLabel, (character)newValue, listIndex);
   }

   /**
    * Report invalid cursor positioning attribute 
    *
    * @param attr
    *          attribute name
    * @param value
    *          attribute value
    */
   private void invalidPosition(String attr, int value)
   {
      ErrorManager.recordOrShowError(4057, 
            String.format(
               "Attribute %s for the %s %s has an invalid value of %d", 
               attr, type(), widgetName(), value
            ), 
            false);
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
      pushWidgetAttr("mousePointer", config.mousePointer);
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Checks if the given widget has valid format string or even has the format attribute.
    * 
    * @return   <code>FALSE</code>, editor does not have the format option.
    */
   @Override
   public boolean hasFormat()
   {
      return false;
   }

   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   @Override
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized)
      {
         return new character();
      }
      
      return super.getScreenValue(initialized, bdt, fmt, ignoreFormat);
   }
   
   /**
    * Internal worker for setting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   frameBuf
    *          The frame buffer where to save the value.
    * @param   value
    *          The value to be set via SCREEN-VALUE attribute.
    * @param   inUIStmt
    *          Flag indicating this call originates from a UI statement.
    *          
    * @return  <code>true</code> if the caller can proceed, as the screen-value can be set.
    */
   @Override
   boolean setScreenValue(ScreenBuffer frameBuf, Object value, boolean inUIStmt)
   {
      if (value instanceof BaseDataType && ((BaseDataType) value).isUnknown())
      {
         ((character) value).assign("");
      }
      
      int widgetID = getId();
      if (frameBuf.isIdValid(widgetID))
      {
         // do not call super method, because [modified] should not be set, instead put the value
         // directly:
         frameBuf.putWidgetValue(widgetID, value);
         // update the [modified] flag. The UI stmts will reset it back to [false]
         config.modified = !inUIStmt;
      }
      
      return true;
   }
   
   /**
    * Attempts to insert the content of a text file at the current position of the caret.
    *
    * @param   file
    *          The file whose content should be inserted. Full or relative path may be used. In
    *          latter case, the {@code PROPATH} is used to find the file.
    *
    * @return  if the text insertion succeeded, the method returns {@code yes}.
    */
   @Override
   public logical insertFile(String file)
   {
      return new logical(LogicalTerminal.insertEditorFile(getId(), file));
   }
   
   /**
    * Attempts to insert the content of a text file at the current position of the caret.
    *
    * @param   file
    *          The file whose content should be inserted. Full or relative path may be used. In
    *          latter case, the {@code PROPATH} is used to find the file.
    *
    * @return  if the text insertion succeeded, the method returns {@code yes}.
    */
   @Override
   public logical insertFile(character file)
   {
      if (file == null || file.isUnknown())
      {
         return new logical(false);
      }
      return insertFile(file.toStringMessage());
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @Override
   public logical search(String pattern, long mode)
   {
      if (pattern == null || pattern.isEmpty() || mode < 0 || mode >= 2 * FIND_SELECT)
      {
         return new logical(false);
      }
      return new logical(LogicalTerminal.searchInEditor(getId(), pattern, (int) mode));
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @Override
   public logical search(character pattern, long mode)
   {
      if (pattern == null || pattern.isUnknown())
      {
         return new logical(false);
      }
      return search(pattern.toStringMessage(), mode);
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @Override
   public logical search(String pattern, int64 mode)
   {
      if (mode == null || mode.isUnknown())
      {
         return new logical(false);
      }
      return search(pattern, mode.longValue());
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @Override
   public logical search(character pattern, int64 mode)
   {
      if (pattern == null || pattern.isUnknown() || mode == null || mode.isUnknown())
      {
         return new logical(false);
      }
      return search(pattern.toStringMessage(), mode.longValue());
   }
   
   /**
    * Moves the caret position at the end of the current text
    *
    * @return  if the operation is successful, the method returns {@code yes}.
    */
   @Override
   public logical moveCaretToEof()
   {
      return new logical(LogicalTerminal.setEditorCursorOffset(getId(), -1));
   }
}
