/*
** Module   : EnhancedBrowseConfig.java
** Abstract : Enhanced browse configuration.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180529 Created initial version.
** 002 OM  20180627 Added missing field in copy constructor.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import java.io.*;
import java.util.*;

/**
 * Enhanced browse configuration. Allows to specify:
 * <ul>
 *    <li>custom colors and fonts for browse;</li>
 *    <li>custom colors and fonts for columns;</li>
 *    <li>separators visibility and color;</li>
 *    <li>row height;</li>
 *    <li>width, order and visibility of columns.</li>
 * </ul>
 */
public class EnhancedBrowseConfig
implements Serializable
{
   /*
    * Part of the key: the name of the parent external procedure which contains the browse.
    * <code>null</code> if config is applied to all browses.
    */
   public String ehKeyProcName = null;

   /*
   * Part of the key: browse name.
   */
   public String ehKeyBrowseName = null;

   /**
    * ID of the corresponding entry in the directory which stores this configuration.
    */
   public Integer directoryNodeId = null;

   /** Foreground color. */
   public ColorRgb ehFgColor = null;

   /** Background color. */
   public ColorRgb ehBgColor = null;

   /** Labels foreground color. */
   public ColorRgb ehLabelFgColor = null;

   /** Labels background color. */
   public ColorRgb ehLabelBgColor = null;

   /** Browse font. */
   public FontDetails ehFont = null;

   /** Labels font. */
   public FontDetails ehLabelFont = null;

   /** Separators on/off. */
   public Boolean separators = null;

   /** Separators color. */
   public ColorRgb ehSeparatorsColor = null;

   /** Row height, in pixels. */
   public Integer rowHeightPixels = null;

   /** Labels area height, in pixels. */
   public Integer labelsHeightPixels = null;

   /** Enhanced column configs keyed by column keys. */
   public Map<String, EnhancedColumnConfig> columnConfigs = null;

   /**
    * Default constructor.
    */
   public EnhancedBrowseConfig()
   {
   }

   /**
    * Creates enhanced config from the configuration of the specified browse.
    *
    * @param browse
    *        Browse which enhanced configuration is copied into this config.
    */
   public EnhancedBrowseConfig(BrowseGuiImpl browse)
   {
      int numColumns = browse.getColumnCount();
      copyBrowseConfig(browse.config());

      columnConfigs = new HashMap<>(numColumns);
      for (int i = 0; i < numColumns; i++)
      {
         BrowseColumnConfig config = browse.getColumn(i).config();
         columnConfigs.put(config.ehColumnKey, new EnhancedColumnConfig(config));
      }
   }

   /**
    * Creates enhanced config from the configuration of the specified browse.
    *
    * @param browseWidget
    *        Browse which enhanced configuration is copied into this config.
    */
   public EnhancedBrowseConfig(BrowseWidget browseWidget)
   {
      int numColumns = browseWidget.getNumColumns().intValue();
      copyBrowseConfig(browseWidget.config());

      columnConfigs = new HashMap<>(numColumns);
      for (int i = 0; i < numColumns; i++)
      {
         BrowseColumnWidget widget =
               ((BrowseColumnWidget) browseWidget.getBrowseColumn(i + 1).unwrapBrowseColumn());
         BrowseColumnConfig config  = widget.config();
         columnConfigs.put(config.ehColumnKey, new EnhancedColumnConfig(config));
      }
   }

   /**
    * Copy constructor.
    *
    * @param cfg
    *        Configuration to copy.
    */
   public EnhancedBrowseConfig(EnhancedBrowseConfig cfg)
   {
      ehFgColor = cfg.ehFgColor;
      ehBgColor = cfg.ehBgColor;
      ehLabelFgColor = cfg.ehLabelFgColor;
      ehLabelBgColor = cfg.ehLabelBgColor;
      ehFont = cfg.ehFont;
      ehLabelFont = cfg.ehLabelFont;
      separators = cfg.separators;
      ehSeparatorsColor = cfg.ehSeparatorsColor;
      rowHeightPixels = cfg.rowHeightPixels;
      labelsHeightPixels = cfg.labelsHeightPixels;
      ehKeyProcName = cfg.ehKeyProcName;
      ehKeyBrowseName = cfg.ehKeyBrowseName;
      directoryNodeId = cfg.directoryNodeId; 
      
      if (cfg.columnConfigs != null)
      {
         columnConfigs = new HashMap<>(cfg.columnConfigs);
      }
   }

   /**
    * Get the key of this configuration.
    *
    * @return the key of this configuration.
    */
   public EnhancedBrowseConfigKey getKey()
   {
      return new EnhancedBrowseConfigKey(ehKeyProcName,
                                         ehKeyBrowseName,
                                         columnConfigs != null ? columnConfigs.keySet() : null);
   }

   /**
    * Copy enhanced configuration from the given browse config into this enhanced config.
    *
    * @param cfg
    *        Browse config to copy.
    */
   private void copyBrowseConfig(BrowseConfig cfg)
   {
      ehFgColor = cfg.ehFgColor;
      ehBgColor = cfg.ehBgColor;
      ehLabelFgColor = cfg.ehLabelFgColor;
      ehLabelBgColor = cfg.ehLabelBgColor;
      ehSeparatorsColor = cfg.ehSeparatorsColor;
      ehFont = cfg.ehFont;
      ehLabelFont = cfg.ehLabelFont;
      separators = cfg.separators;
      rowHeightPixels = cfg.rowHeightPixels;
      labelsHeightPixels = cfg.labelsHeightPixels;
      ehKeyProcName = cfg.ehKeyProcName;
      ehKeyBrowseName = cfg.ehKeyBrowseName;
   }
}
