/*
** Module   : FileDialog.java
** Abstract : Server-side component for file selection dialogs.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170719 Initial commit. Gather information using primitive. Not sent to client yet.
** 002 EVL 20180207 Fix for invalid char in comment issue.
** 003 OM  20180321 Updated javadoc.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

/**
 * A class that collects option information about a file selection statement using chained
 * methods.
 * <p>
 * Each public method emulates an option of the statement. They do a quick validation. If
 * successful, they return a reference to this object to allow other method to be chained.
 * <p>
 * The {@code execute()} method does the final checks, send the request with gathered options
 * to client, awaits the user to interact and finally sets up the response and confirmation
 * variable {@code update}.
 */
public class FileDialog
{
   /**
    * Flag to mark a dialog for selecting a directory. If {@code true} the user can select only
    * directories. If {@code false} the user can select only files.
    */
   private final boolean findDirectories;
   
   /**
    * As input parameter, when {@code useFilename == true} this is the default filename presented
    * to user. When the dialog finished with success it contains the filename the user has
    * selected - as OUTPUT parameter.
    */
   private character path;
   
   /**
    * A logical variable used as output parameter that will contain the status of the user's
    * dialog interaction: if the user validates it by selecting the {@code OK} button, the
    * value of the variable is set to {@code TRUE} on return. If the user rejects the dialog by
    * pressing the {@code Cancel} button, the variable's value on return will be {@code FALSE}.
    */
   private logical toUpdate = null;
   
   /**
    * Specifies the parent window for the dialog box to be displayed. The value must be a
    * handle to a window.
    */
   private handle inWindow = null;
   
   /**
    *  The title of the dialog. This is optional. If not provided default titles for {@code Open}
    * and {@code Save As} dialogs will be used.
    */
   private String title = null;
   
   /**
    * If {@code true},  a dialog is displayed to prompt for confirmation if the user selects the
    * a file that already exists. This option is used only if {@code SAVE-AS} option is
    * also active.
    */
   private boolean askOverwrite = false;
   
   /**
    * The dialog will attempt to create a temporary file with selected name and location before it
    * completes in order to verify whether the user has write access to respective directory.
    * On error, the user is notified by a message and user is prompted to select another filename.
    * The dialog will complete when the user enters a filename associated with a writable
    * directory or completely rejects the dialog. The intermediary file is deleted.
    * <p>
    * Note: this option is not supported yet.
    */
   private boolean createTestFile = false;
   
   /**
    * The default extension to be appended to the user's selection if it does not have one.
    */
   private String defaultExtension = null;
   
   /**
    * The initial directory where dialog opens for the first time.
    */
   private String initialDir = null;
   
   /**
    * Enforces that the user must select and existing file. If (s)he chooses otherwise, an error
    * message is displayed and he is requested to select another filename. The dialog does not
    * complete until the user selected an existing file or uses {@code Cancel} button to reject
    * the dialog.
    */
   private boolean mustExist = false;
   
   /**
    * By default, the last directory where the user has selected a file remains will be used as
    * the default initial directory for subsequent invocations of this method. When the option
    * is specified, the selected directory is not saved. It has no effect if subsequent
    * invocations that specify the INITIAL-DIR option.
    */
   private boolean returnToStartDir = false;
   
   /** Converts the dialog to a "Save As..." dialog box. */
   private boolean saveAs = false;
   
   /**
    * Specifies the contents of {@code path} as the default filename entry for the dialog. The
    * user is free to change it if needed.
    */
   private boolean useFilename = false;
   
   /**
    * If {@code filters} are specified, then the default one is designed by this value.
    */
   private int initialFilter = -1;
   
   /**
    * Defines one or more filters for the filename dialog. Each filter selects a subset of the
    * available files in the common dialog directory to build the dialog file selection-list.
    * A filter consists of two parts: a label (name) and file specification (filespec). Must be
    * an even size array.
    */
   private String[] filters = null;
   
   /**
    * Creates a configurable {@code FileDialog} object that will prompt the user to select a file
    * when executed.
    *
    * @param   filePath
    *          The output parameter that will contain result (full absolute path).
    *
    * @return  A {@link FileDialog} that can be configured with various options using chained
    *          methods before calling {@link FileDialog#execute()} to be displayed.
    */
   public static FileDialog createFileSearchDialog(character filePath)
   {
      return new FileDialog(filePath, false);
   }
   
   /**
    * Creates a configurable {@code FileDialog} object that will prompt the user to select a
    * directory when executed.
    *
    * @param   dirPath
    *          The output parameter that will contain result (full absolute path).
    *
    * @return  A {@link FileDialog} that can be configured with various options using chained
    *          methods before calling {@link FileDialog#execute()} to be displayed.
    */
   public static FileDialog createDirSearchDialog(character dirPath)
   {
      return new FileDialog(dirPath, true);
   }
   
   /**
    * The constructor. Configures the dialog in the simplest form when only the path reference is
    * passed. To activate it, just call {@code execute}. Depending on the {@code findDirs} flag,
    * some of the options to be set may not be valid (FILTERS, INITIAL-FILTER, DEFAULT-EXTENSION,
    * MUST-EXIST, CREATE-TEST-FILE, SAVE-AS, ASK-OVERWRITE, USE-FILENAME and UPDATE are not valid
    * when {@code findDirs == false}). 
    * 
    * @param   path
    *          A reference to a {@code character} variable to be updated if the dialog is
    *          accepted.
    * @param   findDirs
    *          Sets the type of file dialog. If {@code true} this will select folders only (behave
    *          like the SYSTEM-DIALOG-GET-DIR), otherwise, only files are allowed to be returned
    *          (behave like the SYSTEM-DIALOG-GET-FILE).
    */
   private FileDialog(character path, boolean findDirs)
   {
      this.findDirectories = findDirs;
      this.path = path;
   }
   
   /**
    * Configures the file dialog to be created as child of a certain windows.
    * 
    * @param   window
    *          A handle to a window that will be the parent of the dialog. The handle must
    *          refer a window or an error condition will be issued.
    * 
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    * 
    * @see     #inWindow
    */
   public FileDialog setWindow(handle window)
   {
      if (window.isUnknown())
      {
         String msg = "Could not find widget handle for window from expression";
         ErrorManager.recordOrThrowError(3138, msg, false);
      }
      else
      {
         WrappedResource r = window.getResource();
         if (r instanceof WindowWidget)
         {
            this.inWindow = window;
         }
         else
         {
            String msg = "Wrong type of object handle (%s)  -- must be a window handle";
            ErrorManager.recordOrThrowError(3139,
                                            String.format(msg, r.getClass().toString()),
                                            false);
         }
      }
      return this;
   }
   
   /**
    * Configures the title of the file selection dialog.
    *  
    * @param   title
    *          The title of the file selection dialog.
    * 
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #title
    */
   public FileDialog setTitle(character title)
   {
      validateParam(title);
      this.title = title.toJavaType();
      return this;
   }
   
   /**
    * Configures the title of the file selection dialog.
    * 
    * @param   title
    *          The title of the file selection dialog.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #title
    */
   public FileDialog setTitle(String title)
   {
      this.title = title;
      return this;
   }
   
   /**
    * Configures the file selection dialog to warn the user that the selected file already exists
    * and prompt for overwrite it.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #askOverwrite
    */
   public FileDialog askOverwrite()
   {
      askOverwrite = true;
      return this;
   }
   
   /**
    * Configures the file selection dialog to check write access.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #createTestFile
    */
   public FileDialog createTestFile()
   {
      createTestFile = true;
      return this;
   }
   
   /**
    * Configures the default extension of the file selection dialog.
    * 
    * @param   defaultExtension
    *          The default extension for selected file.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #defaultExtension
    */
   public FileDialog setDefaultExtension(character defaultExtension)
   {
      validateParam(defaultExtension);
      this.defaultExtension = defaultExtension.toJavaType();
      return this;
   }
   
   /**
    * Configures the default extension of the file selection dialog.
    *
    * @param   defaultExtension
    *          The default extension for selected file.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #defaultExtension
    */
   public FileDialog setDefaultExtension(String defaultExtension)
   {
      this.defaultExtension = defaultExtension;
      return this;
   }
   
   /**
    * Configures the initial directory where the file selection dialog is open.
    * 
    * @param   initialDir
    *          the initial directory where the file selection dialog is open.
    * 
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #initialDir
    */
   public FileDialog setInitialDir(character initialDir)
   {
      validateParam(initialDir);
      this.initialDir = initialDir.toJavaType();
      return this;
   }
   
   /**
    * Configures the initial directory where the file selection dialog is open.
    *
    * @param   initialDir
    *          the initial directory where the file selection dialog is open.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #initialDir
    */
   public FileDialog setInitialDir(String initialDir)
   {
      this.initialDir = initialDir;
      return this;
   }
   
   /**
    * Configures the file selection dialog not to select new files.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #mustExist
    */
   public FileDialog mustExist()
   {
      this.mustExist = true;
      return this;
   }
   
   /**
    * Configures the file selection dialog to discard any used navigation and not to save the
    * destination directory as default location for subsequent dialogs.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #returnToStartDir
    */
   public FileDialog returnToStartDir()
   {
      this.returnToStartDir = true;
      return this;
   }
   
   /**
    * Configures the file selection dialog as a Save-as dialog.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #saveAs
    */
   public FileDialog saveAs()
   {
      this.saveAs = true;
      return this;
   }
   
   /**
    * Configures the file selection dialog to use the current value of the {@code path} as default
    * filename and location.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #useFilename
    */
   public FileDialog useFilename()
   {
      this.useFilename = true;
      return this;
   }
   
   /**
    * Configures the variable to be set when the file selection dialog is closed.
    * 
    * @param   update
       *          A variable that will hold the return status of the dialog. That is {@code true}
       *          if the dialog was accepted and {@code false} otherwise.
    * 
    * @return  The reference to current {@code FileDialog}. Used for chaining. 
    *
    * @see     #toUpdate
    */
   public FileDialog update(logical update)
   {
      this.toUpdate = update;
      return this;
   }
   
   /**
    * Configures the initial filter of the file selection dialog.
    * 
    * @param   initialFilter
    *          The index of the initial filter.
    * 
    * @return  The reference to current {@code FileDialog}. Used for chaining. 
    *
    * @see     #initialFilter
    */
   public FileDialog setInitialFilter(long initialFilter)
   {
      this.initialFilter = (int) initialFilter;
      return this;
   }
   
   /**
    * Configures the initial filter of the file selection dialog.
    *
    * @param   initialFilter
    *          The index of the initially active filter.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #initialFilter
    */
   public FileDialog setInitialFilter(NumberType initialFilter)
   {
      validateParam(initialFilter);
      this.initialFilter = initialFilter.intValue();
      return this;
   }
   
   /**
    * Configures the filters of the file selection dialog. Only files of the active filter
    * type will be visible for selection. User is free to choose any of them as active.
    * 
    * @param   filters
    *          The list of filters for the file selection dialog.
    *
    * @return  The reference to current {@code FileDialog}. Used for chaining.
    *
    * @see     #filters
    */
   public FileDialog setFilters(Object... filters)
   {
      this.filters = new String[filters.length];
      for (int k = 0; k < filters.length; k++)
      {
         if (filters[k] instanceof String)
         {
            this.filters[k] = (String) filters[k];
         }
         else if (filters[k] instanceof Text)
         {
            Text aFilter = (Text) filters[k];
            validateParam(aFilter);
            this.filters[k] = aFilter.toJavaType();
         }
         else
         {
            ErrorManager.recordOrThrowError(223,
                                            "Incompatible data types in expression or assignment",
                                            true);
         }
      }
      if (filters.length % 2 == 1)
      {
         ErrorManager.recordOrThrowError(247,
                                         "Unable to understand after -- \"INITIAL-FILTER\"",
                                         false);
      }
      return this;
   }
   
   /**
    * Displays the file selection dialog that was already configured. The dialog box allows the 
    * user to enter a filename that is assigned to a character variable. The method provides a
    * dialog box appropriate to the environment in which it runs.
    */
   public void execute()
   {
      if (inWindow == null)
      {
         inWindow = LogicalTerminal.defaultWindow();
      }
      
      WindowWidget window = (WindowWidget) inWindow.getResource();
      ClientExports lc = LogicalTerminal.getClient();
      String res = lc.chooseFile(window.getId(), 
                                 path.toJavaType(),
                                 title,
                                 findDirectories,
                                 saveAs,
                                 askOverwrite, 
                                 mustExist,
                                 createTestFile,
                                 returnToStartDir, 
                                 useFilename,
                                 defaultExtension,
                                 initialDir,
                                 filters,
                                 initialFilter);
      
      // on return check whether the user accepted or rejected the dialog 
      if (toUpdate != null)
      {
         toUpdate.assign(res != null);
      }
      if (res != null)
      {
         path.assign(res);
      }
   }
   
   /**
    * Validates a parameter. The BDT parameters for all options must not be {@code null} or 
    * {@code unknown}.
    * 
    * @param   param
    *          The parameter to check.
    * 
    * @throws  ErrorConditionException
    *          if the parameter does not meet the requirements.
    */
   private static void validateParam(BaseDataType param)
   {
      if (param == null || param.isUnknown())
      {
         ErrorManager.recordOrThrowError(3189,
                                         "Unable to extract expression for system-dialog",
                                         false);
      }
   }
   
}
