/*
** Module   : FrameElement.java
** Abstract : abstract base class for passing data/widget lists to UI helpers
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060204 NEW   @24267 Created initial version as abstract base
**                               class for passing data/widget lists to UI 
**                               helpers.
** 002 GES 20060208 CHG   @24382 Added a bypass flag.
** 003 GES 20060404 CHG   @25337 Added a word wrap flag.
** 004 GES 20060428 CHG   @25825 Added statistics accumulation support.
** 005 GES 20060915 CHG   @29625 Added format string support.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This class provides abstract base for passing data/widget lists to UI
 * helpers. 
 */
public abstract class FrameElement
implements Accessor
{
   /**
    * Determines if the element is only for the purpose of assignment or if
    * the element must be honored as something that can affect the state of
    * the user interface.
    *
    * @return   Always <code>false</code>;
    */
   public boolean assignOnly()
   {
      return false;
   }
   
   /**
    * Determines if the element stores a format string in addition to a data
    * value.
    *
    * @return   Always <code>false</code>;
    */
   public boolean hasFormat()
   {
      return false;
   }
   
   /**
    * Determines if the element represents a widget with no data (like a
    * button).
    *
    * @return   Always <code>false</code>;
    */
   public boolean widgetOnly()
   {
      return false;
   }
   
   /**
    * Determines if the element should cause a bypass (ignore) of a stream
    * field read operation.
    *
    * @return   Always <code>false</code>;
    */
   public boolean bypass()
   {
      return false;
   }
   
   /**
    * Determines if the element should cause word wrap to be activated during
    * editing operations.
    *
    * @return   Always <code>false</code>;
    */
   public boolean wordWrap()
   {
      return false;
   }
   
   /**
    * Determines if the element should cause statistics aggregation to be
    * activated during display operations.
    *
    * @return   Always <code>false</code>;
    */
   public boolean statsColumn()
   {
      return false;
   }
   
   /**
    * Accesses the formatted value associated with this frame element.
    *
    * @return   The formatted value.
    */
   public FormattedValue getFormattedValue()
   {
      throw new UnsupportedOperationException("This element has no format " +
                                              "string."); 
   }
   
   /**
    * Accesses the format string associated with this frame element.
    *
    * @return   Always <code>null</code>.
    */
   public String getFormat()
   {
      return null; 
   }
   
   /**
    * Accesses the widget associated with this frame element.
    *
    * @return   The widget.
    */
   public GenericWidget getWidget()
   {
      throw new UnsupportedOperationException("This element contains no " +
                                              "widget."); 
   }
   
   /**
    * Accesses the statistics aggregator associated with this frame element.
    *
    * @return   The accumulator.
    */
   public Accumulator getAccumulator()
   {
      throw new UnsupportedOperationException("This element contains no " +
                                              "accumulator"); 
   }   
}