/*
** Module   : HeaderElement.java
** Abstract : encodes a header expression such that evaluation can be delegated
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20060317   @25093 Created initial version, which encodes a header expression such 
**                           that its evaluation can be delegated to the user interface
**                           processing at a specific point in execution.
** 002 GES 20080910   @39770 Added constructors to allow literals and variables to be passed
**                           directly instead of using a separate inner class.
** 003 GES 20160607          Added lambda support and switched internal implementation to use it.
**     GES 20160615          Added lambda Function as an option which provides the ability to
**                           pass arguments (and thus avoid capturing). Also refined the
**                           support to bypass lambda usage if a BDT or Resolvable is passed
**                           in directly.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import java.util.function.*;

/**
 * This class which encodes a header expression such that evaluation can be delegated to the
 * user interface processing at a specific point in execution.  This expression is linked to
 * a specified widget and will generate its display value as needed by the UI runtime. 
 * <p>
 * Typically, this element is part of a larger array of other instance of this same class.  The 
 * expression is only used for in the {@link #get} and the {@link #set} is unsupported since
 * this is a one-way transfer to the client.
 */
public class HeaderElement
extends FrameElement
{
   /** For literals and simple variables, we can avoid inner classes and lambdas. */
   private BaseDataType cached = null;
   
   /** For resolvables, we don't use a lambda but cache the instance and access it directly. */
   private Resolvable resolver = null;
   
   /** The (possibly capturing) expression to be executed at {@link #get} time. */
   private Supplier<BaseDataType> source = null;
   
   /** Non-capturing form of complex expression. */
   private Function<Object[], BaseDataType> func = null;
   
   /** Arguments to be passed to the non-capturing lambda form. */
   private Object[] args = null;
   
   /** The widget associated with this element. */
   private GenericWidget widget = null;
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The data used for the header.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(int source, GenericWidget widget)
   {
      this(new integer(source), widget);
   }   
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The data used for the header.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(double source, GenericWidget widget)
   {
      this(new decimal(source), widget);
   }   
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The data used for the header.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(boolean source, GenericWidget widget)
   {
      this(new logical(source), widget);
   }   
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The data used for the header.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(String source, GenericWidget widget)
   {
      this(new character(source), widget);
   }   
   
   /**
    * Create an instance.
    *
    * @param    val
    *           The variable to be de-referenced for its value as header data.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(final BaseDataType val, GenericWidget widget)
   {
      this.cached = val;
      this.widget = widget;
   }
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The expression to be evaluated for its value as header data.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(final Resolvable source, GenericWidget widget)
   {
      this.resolver = source;
      this.widget   = widget;
   }
   
   /**
    * Create an instance.
    *
    * @param    source
    *           The expression to be evaluated for its value as header data.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    */
   public HeaderElement(Supplier<BaseDataType> source, GenericWidget widget)
   {
      this.source = source;
      this.widget = widget;
   }
   
   /**
    * Create an instance.
    *
    * @param    func
    *           The expression to be evaluated for its value as header data.
    * @param    widget
    *           The widget instance representing the exact frame field and state of that field.
    *           Must NOT be <code>null</code>.
    * @param    args
    *           The arguments to be passed to the function when it is executed.
    */
   public HeaderElement(Function<Object[], BaseDataType> func, GenericWidget widget, Object... args)
   {
      this.func = func;
      this.args = args;
      this.widget = widget;
   }
   
   /**
    * Accesses the widget associated with this frame element.
    *
    * @return   The widget.
    */
   public GenericWidget getWidget() 
   {
      return widget;
   }
   
   /**
    * Get a copy of the referenced object's current runtime value.
    *
    * @return   The copy of the current value.
    */
   public BaseDataType get()
   {
      BaseDataType bdt = cached;
      
      if (bdt == null)
      {
         if (resolver != null)
         {
            bdt = resolver.resolve();
         }
         else if (func == null)
         {
            bdt = source.get();
         }
         else
         {
            bdt = func.apply(args);
         }
      }
      
      return bdt;
   }
   
   /**
    * Set the referenced object's current runtime value to the given value.
    *
    * @param    val
    *           The current value to be assigned into the referenced object.
    *           This value will be ignored for embedded assignments and can
    *           be passed as <code>null</code>.
    */
   public void set(BaseDataType val)
   {
      throw new UnsupportedOperationException("Header expressions are read-only.");
   }
}