/* 
** Module   : ImageConfig.java
** Abstract : stores IMAGE specific state
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 EVL 20141208 Adding attributes for image loading and color conversion.
** 003 EVL 20150202 Adding image offset related members.
** 004 EVL 20150216 Adding TRANSPARENT, STRETCH-TO-FIT and RETAIN-SHAPE options support.
** 005 HC  20161018 Fixed abends in image processing and drawing when image offset was involved.
** 006 HC  20171102 Implemented image load size and offset, plus related changes.
** 007 SBI 20180716 Changed to set z-order widget class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * Contains IMAGE specific widget state.
 */
public class ImageConfig
extends BaseConfig
{
   /** Image file name. */
   public String image = null;

   /** Image CONVERT-3D-COLORS attribute flag. */
   public boolean convert3D = false;
   
   /** Image TRANSPARENT attribute flag. */
   public boolean transparent = false;
   
   /** Image STRETCH-TO-FIT attribute flag. */
   public boolean stretchToFit = false;
   
   /** Image RETAIN-SHAPE attribute flag. */
   public boolean retainShape = false;
   
   /** X offset in pixel units inside the image data to start drawing image. */
   public int offsetX = 0;
   
   /** Y offset in pixel units inside the image data to start drawing image. */
   public int offsetY = 0;
   
   /** Row offset in character units inside the image data to start drawing image. */
   public double offsetRow = 0.0;
   
   /** Column offset in character units inside the image data to start drawing image. */
   public double offsetCol = 0.0;

   /**
    * Width of the image to load in pixels. This field is needed in addition to the standard
    * size fields because the load size (in LOAD-IMAGE*()) may not affect the widget size.
    */
   public int loadWidth = -1;

   /**
    * Height of the image to load in pixels. This field is needed in addition to the standard
    * size fields because the load size (in LOAD-IMAGE*()) may not affect the widget size.
    */
   public int loadHeight = -1;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public ImageConfig()
   {
      zorder = ZOrderClass.DECORATION;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ImageConfig(int id)
   {
      super(id);
      
      zorder = ZOrderClass.DECORATION;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ImageConfig(WidgetId id)
   {
      super(id);
      
      zorder = ZOrderClass.DECORATION;
   }
   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      ImageConfig cfg = (ImageConfig) config;
      
      image        = cfg.image;
      convert3D    = cfg.convert3D;
      transparent  = cfg.transparent;
      stretchToFit = cfg.stretchToFit;
      retainShape  = cfg.retainShape;
      offsetX      = cfg.offsetX;
      offsetY      = cfg.offsetY;
      offsetRow    = cfg.offsetRow;
      offsetCol    = cfg.offsetCol;
      loadWidth    = cfg.loadWidth;
      loadHeight   = cfg.loadHeight;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      convert3D    = in.readBoolean();
      transparent  = in.readBoolean();
      stretchToFit = in.readBoolean();
      retainShape  = in.readBoolean();
      offsetX      = in.readInt();
      offsetY      = in.readInt();
      offsetRow    = in.readDouble();
      offsetCol    = in.readDouble();
      image        = (String) in.readObject();
      loadWidth    = in.readInt();
      loadHeight   = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(convert3D);
      out.writeBoolean(transparent);
      out.writeBoolean(stretchToFit);
      out.writeBoolean(retainShape);
      out.writeInt(offsetX);
      out.writeInt(offsetY);
      out.writeDouble(offsetRow);
      out.writeDouble(offsetCol);
      out.writeObject(image);
      out.writeInt(loadWidth);
      out.writeInt(loadHeight);
   }
}