/*
** Module   : ImageSupport.java
** Abstract : Defines image loading widget-related APIs available via a widget handle instance.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150128 Created initial version.
** 002 EVL 20150223 Adding CONVERT-3D-COLORS attribute.
** 003 HC  20170829 Added runtime stubs overloads for LOAD-IMAGE* methods with image offset and
**                  size parameters.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Provide access to image loading widget-related APIs which can be accessed via handle instance.
 */
public interface ImageSupport
{
   /**
    * Load widget(button or image) image from specified file name.
    *
    * @param   name
    *          Image file name.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public boolean loadImage(String name);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public boolean loadImage(String name, long xOffset);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public boolean loadImage(String name, long xOffset, long yOffset);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public boolean loadImage(String name, long xOffset, long yOffset, long width);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    * @param   height
    *          Specifies the number of pixels along the y-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public boolean loadImage(String name, long xOffset, long yOffset, long width, long height);

   /**
    * Load widget(button or image) image from specified file name.
    *
    * @param   name
    *          Image file name.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public logical loadImage(character name);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public logical loadImage(character name, int64 xOffset);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public logical loadImage(character name, int64 xOffset, int64 yOffset);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public logical loadImage(character name, int64 xOffset, int64 yOffset, int64 width);

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    * @param   height
    *          Specifies the number of pixels along the y-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-IMAGE")
   public logical loadImage(character name,
                            int64 xOffset,
                            int64 yOffset,
                            int64 width,
                            int64 height);

   /**
    * Gets CONVERT-3D-COLORS attribute.
    * 
    * @return  CONVERT-3D-COLORS current value.
    */
   @LegacyAttribute(name = "CONVERT-3D-COLORS")
   public logical getConvert3D();

   /**
    * Set CONVERT-3D-COLORS attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @LegacyAttribute(name = "CONVERT-3D-COLORS", setter = true)
   public void setConvert3D(boolean flag);

   /**
    * Set CONVERT-3D-COLORS attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @LegacyAttribute(name = "CONVERT-3D-COLORS", setter = true)
   public void setConvert3D(logical flag);
}