/*
** Module   : ImageWidget.java
** Abstract : server side image widget implementation
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141208 Created initial version.
** 002 EVL 20150115 Rework for image support implementation.
** 003 EVL 20150128 Image widget now implements image loading interface.
** 004 EVL 20150130 Adding image phrase conversion and runtime support.
** 005 EVL 20150216 Adding TRANSPARENT, STRETCH-TO-FIT and RETAIN-SHAPE options support.
** 006 EVL 20150222 Adding boolean option getters for CONVERT-3D-COLORS, TRANSPARENT,
**                  STRETCH-TO-FIT and RETAIN-SHAPE.
** 007 HC  20150323 WidgetId.getId() renamed to asInt().
** 008 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 009 HC  20170829 Added runtime stubs overloads for LOAD-IMAGE* methods with image offset and
**                  size parameters.
**     CA  20170831 Moved the image-related attributes/methods to CommonWidget, as they
**                  can be accessed from any widget type (4GL reports an error, but compiles).
** 010 HC  20171102 Implemented image load size and offset, plus related changes.
** 011 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
** 012 CA  20180313 Added getters for BUTTON image related attributes.
** 013 SBI 20180705 Changed loadImageInt in order to clear the current image if an unknown image
**                  is loaded via LOAD-IMAGE(?).
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side image widget implementation.
 */
@LegacyResource(resource = LegacyResource.IMAGE)
public class ImageWidget
extends BaseEntity<ImageConfig>
{
   /**
    * Default constructor.
    */
   public ImageWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ImageWidget(boolean dynamic)
   {
      super(dynamic, new ImageConfig());
   }
   
   /**
    * Returns the name of the image loaded by LOAD-IMAGE. Corresponds Progress
    * IMAGE attribute
    *
    * @return the name of the image loaded by LOAD-IMAGE. 
    */
   @Override
   public character getImage()
   {
      return new character(config.image);
   }

   /**
    * Set button image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImage(String name)
   {
      loadImage(new character(name));
   }
   
   /**
    * Set image tooltip.
    * 
    * @param   text
    *          Tooltip text
    */
   public void setTooltip(String text)
   {
      if (Utils.equals(config.tooltip, text))
      {
         return;
      }
               
      config.tooltip = text;
      pushWidgetAttr("tooltip", config.tooltip);
   }

   /**
    * Get image tooltip.
    * 
    * @return  image tooltip.
    */
   public character getTooltip()
   {
      return new character(config.tooltip);
   }
   
   /**
    * Set CONVERT-3D-COLORS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(boolean flag)
   {
      if (config.convert3D == flag)
      {
         return;
      }
      
      config.convert3D = flag;
      pushWidgetAttr("convert3D", config.convert3D);
   }

   /**
    * Set CONVERT-3D-COLORS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setConvert3D(flag.booleanValue());
      }
   }

   /**
    * Gets CONVERT-3D-COLORS option.
    * 
    * @return  CONVERT-3D-COLORS current value.
    */
   @Override
   public logical getConvert3D()
   {
      return new logical(config.convert3D);
   }

   /**
    * Set TRANSPARENT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setTransparent(boolean flag)
   {
      if (config.transparent == flag)
      {
         return;
      }
      
      config.transparent = flag;
      pushWidgetAttr("transparent", config.transparent);
   }

   /**
    * Set TRANSPARENT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setTransparent(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setTransparent(flag.booleanValue());
      }
   }

   /**
    * Gets TRANSPARENT option.
    * 
    * @return  TRANSPARENT current value.
    */
   @Override
   public logical getTransparent()
   {
      return new logical(config.transparent);
   }

   /**
    * Set STRETCH-TO-FIT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setStretchToFit(boolean flag)
   {
      if (config.stretchToFit == flag)
      {
         return;
      }
      
      config.stretchToFit = flag;
      pushWidgetAttr("stretchToFit", config.stretchToFit);
   }

   /**
    * Set STRETCH-TO-FIT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setStretchToFit(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setStretchToFit(flag.booleanValue());
      }
   }

   /**
    * Gets STRETCH-TO-FIT option.
    * 
    * @return  STRETCH-TO-FIT current value.
    */
   @Override
   public logical getStretchToFit()
   {
      return new logical(config.stretchToFit);
   }

   /**
    * Set RETAIN-SHAPE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setRetainShape(boolean flag)
   {
      if (config.retainShape == flag)
      {
         return;
      }
      
      config.retainShape = flag;
      pushWidgetAttr("retainShape", config.retainShape);
   }

   /**
    * Set RETAIN-SHAPE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setRetainShape(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setRetainShape(flag.booleanValue());
      }
   }

   /**
    * Gets RETAIN-SHAPE option.
    * 
    * @return  RETAIN-SHAPE current value.
    */
   @Override
   public logical getRetainShape()
   {
      return new logical(config.retainShape);
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of X-offset in pixels.
    */
   public void setImageOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of X-offset in pixels.
    */
   public void setImageOffsetX(int offset)
   {
      config.offsetX = offset;
      pushWidgetAttr("offsetX", config.offsetX);
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of Y-offset in pixels.
    */
   public void setImageOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of Y-offset in pixels.
    */
   public void setImageOffsetY(int offset)
   {
      config.offsetY = offset;
      pushWidgetAttr("offsetY", config.offsetY);
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param offset
    *        new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param offset
    *        new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(double offset)
   {
      config.offsetRow = offset;
      pushWidgetAttr("offsetRow", config.offsetRow);
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param offset
    *        new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param offset
    *        new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(double offset)
   {
      config.offsetCol = offset;
      pushWidgetAttr("offsetCol", config.offsetCol);
   }

   /**
    * Load image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean loadImage(String name)
   {
      return loadImageInt(new character(name), null, null, null, null).booleanValue();
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean loadImage(String name, long xOffset)
   {
      return loadImageInt(new character(name), new int64(xOffset), null, null, null).booleanValue();
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean loadImage(String name, long xOffset, long yOffset)
   {
      return loadImageInt(new character(name),
                          new int64(xOffset),
                          new int64(yOffset),
                          null,
                          null).booleanValue();
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean loadImage(String name, long xOffset, long yOffset, long width)
   {
      return loadImageInt(new character(name),
                          new int64(xOffset),
                          new int64(yOffset),
                          new int64(width),
                          null).booleanValue();
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    * @param   height
    *          Specifies the number of pixels along the y-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean loadImage(String name, long xOffset, long yOffset, long width, long height)
   {
      return loadImageInt(new character(name),
                          new int64(xOffset),
                          new int64(yOffset),
                          new int64(width),
                          new int64(height)).booleanValue();
   }

   /**
    * Load image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name)
   {
      return loadImageInt(name, null, null, null, null);
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name, int64 xOffset)
   {
      return loadImageInt(name, xOffset, null, null, null);
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name, int64 xOffset, int64 yOffset)
   {
      return loadImageInt(name, xOffset, yOffset, null, null);
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name, int64 xOffset, int64 yOffset, int64 width)
   {
      return loadImageInt(name, xOffset, yOffset, width,null);
   }

   /**
    * Load widget(button or image) image from specified file name. This method overload allows
    * to load only part of the image.
    *
    * @param   name
    *          Image file name.
    * @param   xOffset
    *          An integer expression that specifies the pixel along the x-axis at which to begin
    *          reading from the image file.
    * @param   yOffset
    *          Specifies the pixel along the y-axis at which to begin reading from the image file.
    * @param   width
    *          Specifies the number of pixels along the x-axis to read from the image file.
    * @param   height
    *          Specifies the number of pixels along the y-axis to read from the image file.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name, int64 xOffset, int64 yOffset, int64 width, int64 height)
   {
      return loadImageInt(name, xOffset, yOffset, width, height);
   }


   /**
    * This widget can't be a tab item.
    * 
    * @return   False
    */
   @Override
   public boolean isTabWidget()
   {
      return false;
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    widthChars
    *           The new size value.
    */
   public void setImageWidthChars(NumberType widthChars)
   {
      if (widthChars.isUnknown())
      {
         return;
      }

      setImageWidthChars(widthChars.intValue());
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    widthChars
    *           The new size value.
    */
   public void setImageWidthChars(int widthChars)
   {
      if (widthChars > 0)
      {
         CoordinatesConversion cc = LogicalTerminal.coordinates();
         setImageWidthPixels(cc.columnToPixels(widthChars));
      }
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    heightChars
    *           The new size value.
    */
   public void setImageHeightChars(NumberType heightChars)
   {
      if (heightChars.isUnknown())
      {
         return;
      }

      setImageHeightChars(heightChars.intValue());
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    heightChars
    *           The new size value.
    */
   public void setImageHeightChars(int heightChars)
   {
      if (heightChars > 0)
      {
         CoordinatesConversion cc = LogicalTerminal.coordinates();
         setImageHeightPixels(cc.rowToPixels(heightChars));
      }
   }
   
   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    widthPixels
    *           The new size value.
    */
   public void setImageWidthPixels(NumberType widthPixels)
   {
      if (widthPixels.isUnknown())
      {
         return;
      }

      setImageWidthPixels(widthPixels.intValue());
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    widthPixels
    *           The new size value.
    */
   public void setImageWidthPixels(int widthPixels)
   {
      if (widthPixels > 0)
      {
         config.loadWidth = widthPixels;
         pushWidgetAttr("loadWidth", config.loadWidth);
      }
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    heightPixels
    *           The new size value.
    */
   public void setImageHeightPixels(NumberType heightPixels)
   {
      if (heightPixels.isUnknown())
      {
         return;
      }

      setImageHeightPixels(heightPixels.intValue());
   }

   /**
    * Sets the image load size as defined in the DEFINE IMAGE ... IMAGE-SIZE ... statement.
    *
    * @param    heightPixels
    *           The new size value.
    */
   public void setImageHeightPixels(int heightPixels)
   {
      if (heightPixels > 0)
      {
         config.loadHeight = heightPixels;
         pushWidgetAttr("loadHeight", config.loadHeight);
      }
   }

   /**
    * Load image for widget.  Internal version.
    * 
    * @param    imageSrc
    *           Image file name to load.
    * @param    xOffset
    *           Image load x-offset.
    * @param    yOffset
    *           Image load y-offset.
    * @param    width
    *           Image load width.
    * @param    height
    *           Image load height.
    *
    * @return  <code>true</code> if operation was successful.
    */
   private logical loadImageInt(character imageSrc,
                                int64 xOffset,
                                int64 yOffset,
                                int64 width,
                                int64 height)
   {
      long xoff = xOffset == null || xOffset.isUnknown() ? 0 : xOffset.longValue();
      long yoff = yOffset == null || yOffset.isUnknown() ? 0 : yOffset.longValue();
      long w = width == null || width.isUnknown() ? 0 : width.longValue();
      long h = height == null || height.isUnknown() ? 0 : height.longValue();

      if (w < 0 || h < 0 || w > Integer.MAX_VALUE || h > Integer.MAX_VALUE ||
          xoff < 0 || xoff > Integer.MAX_VALUE || yoff < 0 || yoff > Integer.MAX_VALUE)
      {
         String msg = "**The LOAD-IMAGE attribute on the IMAGE "+
                      widgetName() +
                      " has invalid arguments. (4065)";
         ErrorManager.recordOrShowWarning(4065, msg, true, true, true, true);
         return new logical(false);
      }

      int origConfigOffsetX = config.offsetX;
      int origConfigOffsetY = config.offsetY;
      int origConfigLoadWidth = config.loadWidth;
      int origConfigLoadHeight = config.loadHeight;

      boolean doPush = false;

      // save current values
      String imageNameSaved = config.image;

      String newImageSrc = (imageSrc == null || imageSrc.isUnknown()) ? "" : imageSrc.toStringMessage();

      // this must be done before sending image bytes because the image should be on client side
      if (frame != null)
      {
         config.image = newImageSrc;
         doPush = true;
      }

      config.offsetX = (int) xoff;
      config.offsetY = (int) yoff;

      // load size does affect the widget size, this seems to be when the widget size has not been explicitly
      // set. Note that the widget realization state also seems to play a role here.
      config.loadWidth = width == null || width.isUnknown() ? -1 : (int) w;
      config.loadHeight = height == null || height.isUnknown() ? -1 : (int) h;

      doPush |= config.offsetX != origConfigOffsetX ||
                config.offsetY != origConfigOffsetY ||
                config.loadWidth != origConfigLoadWidth ||
                config.loadHeight != origConfigLoadHeight;

      if (doPush)
      {
         pushWidgetAttr(new String[] {"image", "offsetX", "offsetY",
                                      "loadWidth", "loadHeight"},
                        new Object[] {config.image, config.offsetX, config.offsetY,
                                      config.loadWidth, config.loadHeight});
      }

      // logical terminal will try to load image from application jar, then ask the
      // client to load from local file system in cas of failure
      boolean result = LogicalTerminal.setImage(
                          config.id.asInt(), ButtonConfig.IMAGE_IMG, newImageSrc);
      
      // if we are here - there is no image to load we need to undo image name change
      if (!result && frame != null)
      {
         config.image = imageNameSaved;
         pushWidgetAttr("image", config.image);
      }
      
      return new logical(result);
   }
}