/*
** Module   : MenuItemConfig.java
** Abstract : Menu item widget attribute container.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20141118 Created initial version.
** 002 VIG 20150123 Changed base class, added constructors and changed 
**                  externalization order of fields, copied color fields.
** 003 VIG 20150227 Added checked, initDisabled attributes. Fixed mnemonic.
**                  Removed mnemonic field. 
** 004 CA  20170929 Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 * MENU-ITEM widget configuration container.
 */
public class MenuItemConfig
extends MnemonicConfig
{
   /**
    * Default SUBTYPE value. Means that this menu-item is normal element of menu
    * with label, sub-menu or some action, that called when this menu-item is chosen.
    */
   public static final String SBTP_NORMAL = "NORMAL";
   
   /**
    * This SUBTYPE denotes, that a blank line is inserted at this point in the menu.
    */
   public static final String SBTP_SKIP = "SKIP";
   
   /**
    * This SUBTYPE denotes, that a rule or line is inserted at this point in the menu.
    */
   public static final String SBTP_RULE = "RULE";
   
   /**
    * READ-ONLY attr. Specifies that this menu item is read-only text. 
    * The user cannot choose this item.
    */
   public boolean readOnly = false;
   
   /**
    * TOGGLE-BOX attr. Specifies that the menu item is displayed as a checkbox 
    * that the user can toggle on or off. In environments that do not support 
    * this option, it is ignored.
    */
   public boolean toggleBox = false;
   
   /**
    * ACCELERATOR attr. Specifies a keyboard accelerator for this menu item. 
    * The value keylabel must be character-string expression that
    * evaluates to a valid key label recognized by Progress, such as a, F1, or ALT-SHIFT-F1.
    */
   public String accelerator;
   
   /**
    * SUBTYPE attr. Can have one of three following values: {@link MenuItemConfig#SBTP_NORMAL},  
    * {@link MenuItemConfig#SBTP_SKIP} or {@link MenuItemConfig#SBTP_RULE}. Specifies an appearance
    * and behavior of this menu-item. Can be specified only before realization. 
    */
   public String subtype = SBTP_NORMAL;

   /** Progress CHECKED attribute. */
   public boolean checked = false;
   
   /** Initial state of DISABLED flag, set during construction in menu definition */
   public Boolean initDisabled = null;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public MenuItemConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public MenuItemConfig(WidgetId id)
   {
      super(id);
   }
   
   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      MenuItemConfig mcfg = (MenuItemConfig) config;
      subtype = mcfg.subtype;
      readOnly = mcfg.readOnly;
      toggleBox = mcfg.toggleBox;
      accelerator = mcfg.accelerator;
      checked = mcfg.checked;
      initDisabled = mcfg.initDisabled;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If pay-load can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      readOnly     = in.readBoolean();
      toggleBox    = in.readBoolean();
      checked      = in.readBoolean();
      subtype      = (String) in.readObject();
      accelerator  = (String) in.readObject();
      initDisabled = (Boolean) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      
      out.writeBoolean(readOnly);
      out.writeBoolean(toggleBox);
      out.writeBoolean(checked);
      out.writeObject(subtype);
      out.writeObject(accelerator);
      out.writeObject(initDisabled);
   }

}