/*
/*
** Module   : PrintingService.java
** Abstract : Printing management class.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 OM  20170721 Initial commit with stubbed methods.
** 002 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 003 EVL 20180207 Fix for invalid char in comment issue.
** 004 OM  20180321 Updated javadoc.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import java.util.*;
import java.util.logging.*;

/**
 * This class is a factory for {@code PrinterSetupDialog}s. It exposes a single factory method and
 * some constants used for setting options in the chained methods of the returned dialog class.
 * 
 * @see PrinterSetupDialog
 */
public class PrintingService
{
   /** Constant for page orientation. NONE means that the value is not initialized. */
   public static final int NONE = 0;
   
   /** Constant for page orientation. This value represents a portrait orientation of the page. */
   public static final int PORTRAIT = 1;
   
   /** Constant for page orientation. It represents the landscape orientation of the page. */
   public static final int LANDSCAPE = 2;
   
   /** Constant for page orientation. Both portrait and landscape orientations are acceptable. */
   public static final int ANY = 3;

   /**
    * Name of the built-in PDF printer, this is the name the legacy app can use
    * in the OUTPUT TO PRINTER to address the built-in PDF printer directly.
    */
   public static final String PDF_PRINTER_NAME = "FWD_PDF_Printer";

   /** Displayable name of the built-in PDF printer visible to the legacy app */
   public static final String PDF_PRINTER_DESCR = "FWD built-in PDF printer";

   /** Control handle of the built-in PDF printer visible to the legacy app */
   public static final int PDF_PRINTER_CONTROL_HANDLE = 0xFFFFFF;

   /** The root node name of the directory print settings */
   private static final String PRINT_ROOT_NODE = "print";

   /** Logger for this class. */
   private static final Logger LOG = LogHelper.getLogger(PrintingService.class.getName());

   /** The print options as set in the current session */
   private static final ContextLocal<WorkArea> workArea = new ContextLocal<WorkArea>()
   {
      @Override
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         wa.printOptions = loadPrintOptionsFromDirectory();
         return wa;
      }
   };
   
   /**
    * Factory method for a printer setup dialog. 
    * 
    * @return  a new {@link PrinterSetupDialog} that can chain eventual option and and finally
    *          execute on client side.
    */
   public static PrinterSetupDialog createSetupDialog()
   {
      return new PrinterSetupDialog();
   }

   /**
    * The value of SESSION:PRINTER-CONTROL-HANDLE.
    *
    * @return  See above.
    */
   public static int getControlHandle()
   {
      return workArea.get().printerControlHandle;
   }

   /**
    * The value of SESSION:PRINTER-HDC.
    *
    * @return  See above.
    */
   public static int getHDC()
   {
      return workArea.get().printerHDC;
   }

   /**
    * The method sets SESSION:PRINTER-CONTROL-HANDLE and SESSION:PRINTER-HDC to their initial values.
    */
   public static void resetControlHandle()
   {
      WorkArea wa = workArea.get();
      wa.printerControlHandle = 0;
      wa.printerHDC = 0;
   }

   /**
    * Returns the print options set in the current context.
    *
    * @return  See above, never <code>null</code>.
    */
   public static PrintOptions getPrintOptions()
   {
      return workArea.get().printOptions;
   }

   /**
    * Validates the supplied name. Currently only the internal FWD PDF printer is supported. The method
    * compares the supplied name against the formal name of the printer (PDF_PRINTER_NAME). If it doesn't
    * match 4110 error is displayed.
    *
    * @param  name
    *         A printer name.
    */
   public static void validatePrinterName(String name)
   {
      if (name != null && !PDF_PRINTER_NAME.equalsIgnoreCase(name))
      {
         ErrorManager.displayError(4110, "Unable to create Printer Device Context: " + name + ".");
      }
   }

   /**
    * Loads the following attributes from server directory.
    * <ul>
    *    <li>
    *       page_size: see  {@link com.goldencode.p2j.util.PrintOptions.PageSize} for all possible values
    *    </li>
    *    <li>
    *       page_orientation: see  {@link com.goldencode.p2j.util.PrintOptions.PageOrientation} for all
    *       possible values
    *    </li>
    *    <li>
    *       margin_top, margin_left, margin_bottom, margin_right: margin value is a float and denotes
    *       millimeters
    *    </li>
    * </ul>
    *
    * The directory structure that captures the attributes above is as follows:
    * <code>
    *    &lt;node class="container" name="print"&gt;
    *       &lt;node class="string" name="page_size"&gt;
    *          &lt;node-attribute name="value" value="Letter"/&gt;
    *       &lt;/node&gt;
    *       &lt;node class="string" name="page_orientation"&gt;
    *          &lt;node-attribute name="value" value="Portrait"/&gt;
    *       &lt;/node&gt;
    *       &lt;node class="string" name="page_margin_top"&gt;
    *          &lt;node-attribute name="value" value="0"/&gt;
    *       &lt;/node&gt;
    *       &lt;node class="string" name="page_margin_left"&gt;
    *          &lt;node-attribute name="value" value="10"/&gt;
    *       &lt;/node&gt;
    *       &lt;node class="string" name="page_margin_bottom"&gt;
    *          &lt;node-attribute name="value" value="0"/&gt;
    *       &lt;/node&gt;
    *       &lt;node class="string" name="page_margin_right"&gt;
    *          &lt;node-attribute name="value" value="10"/&gt;
    *       &lt;/node&gt;
    *    &lt;/node&gt;
    * </code>
    * 
    * The method searches for the above structure in the following directory nodes in the order given:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    *
    * @return  An object representing the print options.
    */
   private static PrintOptions loadPrintOptionsFromDirectory()
   {
      PrintOptions result = new PrintOptions();

      String path = Utils.findDirectoryNodePath(null, PRINT_ROOT_NODE, "container",
                                                Utils.DirScope.BOTH, null);
      if (path == null)
      {
         return result;
      }

      String pageSize = Utils.getDirectoryNodeString(null,
                                                     PRINT_ROOT_NODE + "/page_size",
                                                     null,
                                                     Utils.DirScope.BOTH, null);
      if (pageSize != null)
      {
         try
         {
            result.pageSize = PrintOptions.PageSize.valueOf(pageSize.trim().toUpperCase());
         }
         catch (IllegalArgumentException e)
         {
            // the directory value is not valid
            LOG.warning(PRINT_ROOT_NODE + "/page_size directory attribute value of '" + pageSize + "' " +
                        "is invalid. Valid values are " + Arrays.toString(PrintOptions.PageSize.values()) +
                        ".");
         }
      }

      String pageOrientation = Utils.getDirectoryNodeString(null,
                                                            PRINT_ROOT_NODE + "/page_orientation",
                                                            null,
                                                            Utils.DirScope.BOTH,
                                                            null);
      if (pageOrientation != null)
      {
         try
         {
            pageOrientation = pageOrientation.trim().toUpperCase();
            result.pageOrientation = PrintOptions.PageOrientation.valueOf(pageOrientation);
         }
         catch (IllegalArgumentException e)
         {
            // the directory value is not valid
            LOG.warning(PRINT_ROOT_NODE + "/page_orientation directory attribute value of '" +
                        pageOrientation + "' is invalid. " + "Valid values are " +
                        Arrays.toString(PrintOptions.PageOrientation.values()) + ".");
         }
      }

      result.marginTop = loadMargin("page_margin_top");
      result.marginLeft = loadMargin("page_margin_left");
      result.marginBottom = loadMargin("page_margin_bottom");
      result.marginRight = loadMargin("page_margin_right");

      return result;
   }

   /**
    * Loads margin value from server directory.
    *
    * @param   attributeName
    *          The margin attribute name.
    *
    * @return  The margin value or 0 if the value can not be found in the directory or the value stored in
    *          the directory is not a valid float.
    */
   private static float loadMargin(String attributeName)
   {
      String marginStr = Utils.getDirectoryNodeString(null,
                                                      PRINT_ROOT_NODE + "/" + attributeName,
                                                      null,
                                                      Utils.DirScope.BOTH, null);
      if (marginStr != null)
      {
         try
         {
            return Float.parseFloat(marginStr);
         }
         catch (NumberFormatException ex)
         {
            LOG.warning("print/" + attributeName + " directory attribute value of '" + marginStr +
                        "' is invalid. " + "A float number is expected.");
         }
      }

      return 0;
   }

   /**
    * A class that collects option information about a printer selection statement using chained
    * methods.
    * <p>
    * Each public method emulates an option of the statement. They do a quick validation. If
    * successful they return a reference to this object to allow other method to be chained.
    * <p>
    * The {@code execute()} method does the final checks, send the request with gathered options
    * to client, awaits the user to interact and finally sets up the response and confirmation
    * variable {@code update}.
    */
   public static class PrinterSetupDialog
   {
      /**
       * The initial number of copies in the {@code Print} dialog box. The user can directly
       * change this value in the dialog box. This option is supported only if the printer drivers
       * support multi-copy printing. In rare cases when they don't, the field is disabled.
       */
      private int copyCount = -1;
      
      /** The orientation of the paper. */
      private int orientation = NONE;
      
      /**
       * A logical variable used as output parameter that will contain the status of the user's
       * dialog interaction: if the user validates it by selecting the {@code OK} button, the
       * value of the variable is set to {@code TRUE} on return. If the user rejects the dialog by
       * pressing the {@code Cancel} button, the variable's value on return will be {@code FALSE}.
       */
      private logical toUpdate;
      
      /**
       * Specifies the parent window for the dialog box to be displayed. The value must be a
       * handle to a window.
       */
      private handle inWindow;
      
      /**
       * Configures initial number of copies in the {@code Print} dialog box.
       * 
       * @param   count
       *          The initial number of copies in the {@code Print} dialog box.
       * 
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       * 
       * @see     #copyCount
       */
      public PrinterSetupDialog setCopyCount(int64 count)
      {
         this.copyCount = count.intValue();
         return this;
      }
      
      /**
       * Configures the initial number of copies in the {@code Print} dialog box.
       *
       * @param   count
       *          The initial number of copies in the {@code Print} dialog box.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #copyCount
       */
      public PrinterSetupDialog setCopyCount(long count)
      {
         this.copyCount = (int) count;
         return this;
      }
      
      /**
       * Configures the orientation of the page in dialog.
       *
       * @param   orientation
       *          The new default page orientation. 
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #orientation
       */
      public PrinterSetupDialog setOrientation(int orientation)
      {
         if (this.orientation != NONE) 
         {
            // this is a compile-time error in fact (even duplication of same orientation)
            ErrorManager.displayAbend(
                  277, "Phrase or option conflicts with previous phrase or option");
            return null;
         }
         
         this.orientation = orientation;
         return this;
      }
      
      /**
       * Configures the output variable to be set when the selection dialog is closed.
       *
       * @param   update
       *          A variable that will hold the return status of the dialog. That is {@code true}
       *          if the dialog was accepted and {@code false} otherwise.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #toUpdate
       */
      public PrinterSetupDialog update(logical update)
      {
         this.toUpdate = update;
         return this;
      }
      
      /**
       * Configures the dialog to be created as child of a certain window.
       *
       * @param   window
       *          A handle to a window that will be the parent of the dialog. The handle must
       *          refer a window or an error condition will be issued.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #inWindow
       */
      public PrinterSetupDialog setWindow(handle window)
      {
         if (window.isUnknown())
         {
            String msg = "Could not find widget handle for window from expression";
            ErrorManager.recordOrThrowError(3138, msg, false);
         }
         else
         {
            WrappedResource r = window.getResource();
            if (r instanceof WindowWidget)
            {
               this.inWindow = window;
            }
            else
            {
               String msg = "Wrong type of object handle (%s)  -- must be a window handle";
               ErrorManager.recordOrThrowError(3139,
                                               String.format(msg, r.getClass().toString()),
                                               false);
            }
         }
         return this;
      }
      
      /**
       * Instruct the client to show the {@code Printer Setup} dialog using the configured options
       * using previously chained methods. When the client returns, update the {@link #toUpdate}
       * variable, if set, to let the programmer whether the client accepted the dialog or
       * rejected it.
       */
      public void execute()
      {
         if (inWindow == null)
         {
            inWindow = LogicalTerminal.currentWindow();
         }
         
         WindowWidget window = (WindowWidget) inWindow.getResource();
         ClientExports lc = LogicalTerminal.getClient();
         
         WorkArea wa = workArea.get();
         PrintOptions printOptions = wa.printOptions;
         PrintOptions changedPrintOptions = lc.setupPrinter(window.getId(),
                                                            copyCount,
                                                            new PrintOptions(printOptions));
         
         Random rnd = new Random();
         
         if (changedPrintOptions != null)
         {
            printOptions.apply(changedPrintOptions);
            
            // PRINTER-HDC is assigned to a value only when the printer setup dialog is OKed
            wa.printerHDC = rnd.nextInt();
         }
         
         // on return check whether the user accepted or rejected the dialog
         if (toUpdate != null)
         {
            toUpdate.assign(changedPrintOptions != null);
         }
         
         // PRINTER-CONTROL-HANDLE is kept until the attribute is assigned
         if (wa.printerControlHandle == 0)
         {
            // PRINTER-CONTROL-HANDLE is assigned to a value even when the printer setup dialog is
            // canceled
            wa.printerControlHandle = rnd.nextInt();
         }
      }
   }

   /**
    * Contextual data.
    */
   private static class WorkArea
   {
      /** Print options */
      public PrintOptions printOptions;

      /** Holds the SESSION:PRINTER-CONTROL-HANDLE value */
      public int printerControlHandle = 0;

      /** Holds the SESSION:PRINTER-HDC value */
      public int printerHDC = 0;
   }
}
