/*
** Module   : SkipEntity.java
** Abstract : server side skip and space pseudo-widgets
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23496 Created initial version.
** 002 SIY 20060316   @25140 Additional constructor.
** 003 GES 20060803   @28384 Some cleanup and refactoring.
** 004 SIY 20060805   @28446 Added required SkipEntity(int) constructor.
** 005 GES 20061101   @30862 Allow decimal values for skip and space
**                           sizes (and properly round them).
** 006 SIY 20070706   @34396 Fixed setWidth().
** 007 NVS 20070706   @34398 setWidth(double) should round up, not ceil()
**                           nor floor() the parameter. Confirmed with 
**                           skip-round.p testcase
** 008 SIY 20111101          Updated imports.
** 009 CA  20130930          Added LegacyResource annotation, to determine resource type.
** 010 CA  20131115          Changes related to properly instantiate dynamic resources.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 012 EVL 20160420          Changing the skip length internal to double to have enough presision
**                           for GUI mode.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Implements the server side skip and space pseudo-widgets.
 */
@LegacyResource(resource = LegacyResource.SKIP_OR_SPACE)
public class SkipEntity
extends GenericWidget<SkipConfig>
{
   /**
    * Default constructor (creates a space, not a skip).
    */
   public SkipEntity()
   {
      super(false, new SkipConfig());
   }

   /**
    * Construct SPACE or SKIP depending on the argument.
    * 
    * @param   vertical
    *          If <code>true</code> then SKIP is created, otherwise created
    *          SPACE.
    */
   public SkipEntity(boolean vertical)
   {
      this();
      config.vertical = vertical;
   }

   /**
    * Create an instance of SKIP with specified size.
    * 
    * @param   size
    *          The number of line feeds.
    */
   public SkipEntity(int size)
   {
      this(true);
      setHeight(size);
   }

   /**
    * Gets the number of line feeds.
    *
    * @return   The number of line feeds to insert (0 if this represents a
    *           space instead of a skip).
    */
    public int getSkip()
    {
       return config.vertical ? (int) Math.round(config.length) : 0;
    }
    
   /**
    * Sets the type of the widget (skip or space).
    *
    * @param    vertical
    *           <code>true</code> for skip, <code>false</code> for space.
    */
    public void setVertical(boolean vertical)
    {
       config.vertical = vertical;
    }

   /**
    * Sets the width of the space or skip object. It is equal to the number of
    * skipped columns for space object. Operation is ignored for the skip
    * object.
    * 
    * @param    width
    *           The width of the object.
    */
   public void setWidth(double width)
   {
      if (!config.vertical)
      {
         if (LogicalTerminal.isChui())
         {
            config.length = Math.round(width);
         }
         else
         {
            config.length = width;
         }
      }
   }
   
   /**
    * Sets the height of the space or skip object. It is equal to the number
    * of skipped rows for skip object. Operation is ignored for the space
    * object.
    * 
    * @param    height
    *           The height of the skip object.
    */
   public void setHeight(double height)
   {
      if (config.vertical)
      {
         if (LogicalTerminal.isChui())
         {
            config.length = Math.floor(height);
         }
         else
         {
            config.length = height;
         }
      }
   }
   
   /**
    * This widget can't be a tab item.
    * 
    * @return   False
    */
   @Override
   public boolean isTabWidget()
   {
      return false;
   }
}