/*
** Module   : TextWidget.java
** Abstract : server-side text widget representation.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130927 Created initial version.
** 002 CA  20131115 Changes related to properly instantiate dynamic resources.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 CA  20141220 Track if an explicit VIEW-AS TEXT phrase is in use.  GUI has code dependent on
**                  this fact.
** 005 IAS 20150801 Added support for the BLANK interface
** 006 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 007 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Represents a TEXT widget on server-side. This widget is used in following cases:
 * <ul>
 *   <li>CREATE TEXT statement</li>
 *   <li>DISPLAY "string-literal" VIEW-AS TEXT</li>
 *   <li>DISPLAY ... WITH ... USE-TEXT</LI>
 * </ul>
 */
@LegacyResource(resource = LegacyResource.TEXT)
public class TextWidget
extends ControlTextWidget<TextConfig>
implements BlankInterface
{
   /**
    * Default constructor.
    */
   public TextWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public TextWidget(boolean dynamic)
   {
      super(dynamic, new TextConfig());

      setStatic(false);
   }

   /**
    * Mark this widget as using an explicit VIEW-AS TEXT phrase.
    * 
    * @param    explicit
    *           <code>true</code> when there is a VIEW-AS TEXT phrase.
    */
   public void explicitViewAs(boolean explicit)
   {
      config.explicitViewAs = explicit;
   }

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setBlank(boolean value)
   {
      if (config.blank == value)
      {
         return;
      }
      
      config.blank = value;
      pushWidgetAttr("blank", config.blank);
   }

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setBlank(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute BLANK on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setBlank(value.booleanValue());
   }

   /**
    * Get value of the BLANK attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isBlank()
   {
      return new logical(config.blank);
   }

   /**
    * This widget can't be a tab item.
    * 
    * @return   False
    */
   @Override
   public boolean isTabWidget()
   {
      return false;
   }
}