/*
** Module   : WidgetId.java
** Abstract : represents widget identifiers and their relationships
**
** Copyright (c) 2007-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 NVS 20070116   @31884 Created initial version.
** 002 EVL 20140731          Rework to centralize all widget ID related code in single place.
** 003 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 004 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 005 CA  20150114          Added the createId API, so that a new ID can be copied the location  
**                           in the down frame of an existing ID.
** 006 HC  20150323          WidgetId.getId() renamed to asInt().
** 007 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 008 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 HC  20160406          Added method virtualWidget(int).
** 010 ECF 20171228          Minor performance improvement.
** 011 HC  20180116          Made WIDGET_ID_RANGE_BEGIN publicly accessible.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.net.SessionManager;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.chui.ThinClient;

/**
 * This class represents widget identifiers.  The widget ID address space is stored and managed on
 * a context local basis.  This means that each user's P2J session maintains an independent set of
 * widget ID allocations.
 * <p>
 * Approach currently used reserves the 1 number for default window, the number 2 and 3 reserved
 * for tiny editor used to implement message that allows value change so it has one frame and one
 * fill-in widget.  Every new widget static or dynamic will consume numbers from 4 up to maximum
 * value of the <code>int</code> variable type.  The widget ID is not normally reused when the
 * widget is deleted.  Instead this freed number can be used again when integer number wrapping
 * happen - integer become negative in attempt to increase after maximum value reached.  Two
 * sequential pass of full integer range without getting free ID means there is no free ID and
 * this generates runtime exception to avoid infinite loop.
 */
public class WidgetId
implements Externalizable,
           Comparable<WidgetId>
{
   /** The ID for default window widget. */
   public static final WidgetId DEFAULT_WINDOW_ID = new WidgetId(1);
   
   /** The ID for default window widget. */
   public static final int _DEFAULT_WINDOW_ID = DEFAULT_WINDOW_ID.asInt();
   
   /** The ID for tiny input editor frame used on client side for editable message. */
   public static final WidgetId INPUT_FRAME_ID = new WidgetId(DEFAULT_WINDOW_ID.id + 1);
   
   /** Starting number for the first widget ID.  Tiny input consumes two widget ID. */
   public static final int WIDGET_ID_RANGE_BEGIN = INPUT_FRAME_ID.id + 2;
   
   /** Stores session context-local state variables. */
   private static ContextContainer session = new ContextContainer(); 
   
   /** The widget's ID. */
   private int id = - 1;
   
   /** Cached hash code. */
   private int hashCode;
   
   /**
    * Default c'tor.  Not for public use, for de-serialization only.
    */
   public WidgetId()
   {
      // no-op
   }
   
   /**
    * Create a new instance and associate the given ID.
    * <p>
    * Can be used on both client and server-side.
    * 
    * @param    id
    *           The widget ID.
    */
   public WidgetId(int id)
   {
      this.id = id;
      this.hashCode = computeHashCode();
   }
   
   /**
    * Check if the widget ID is virtual (null or negative).
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   See above.
    */
   public static boolean virtualWidget(WidgetId id)
   {
      return id == null || id.asInt() < 0;
   }
   
   /**
    * Check if the widget ID is virtual (negative).
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   See above.
    */
   public static boolean virtualWidget(int id)
   {
      return id < 0;
   }
   
   /**
    * Retrieve the unique per session integer id number for next new widget to create.
    *
    * @return The integer Id number for next dynamic widget to create. 
    *
    * @throws   RuntimeException
    *           If all possible IDs are in use. This is very unlikely but could happen!
    */
   public static WidgetId nextID()
   {
      return new WidgetId(session.obtain().allocateID());
   }
   
   /**
    * Returns <code>true</code> when <code>objId != null &amp;&amp; objId.asInt() == id</code>.
    * 
    * @param    objId
    *           The object id.
    * @param    id
    *           The numeric id.
    * 
    * @return   See above.
    */
   public static boolean equals(WidgetId objId, int id)
   {
      return objId != null && objId.asInt() == id; 
   }
   
   /**
    * Create a new {@link WidgetDownId} instance which is assigned the given <code>widgetId</code>.
    * 
    * @param    widgetId
    *           The new widgetId.
    * 
    * @return   See above.
    */
   public WidgetId createId(WidgetId widgetId)
   {
      return createId(widgetId.id);
   }
   
   /**
    * Create a new {@link WidgetDownId} instance which is assigned the given <code>widgetId</code>.
    * 
    * @param    widgetId
    *           The new widgetId.
    * 
    * @return   See above.
    */
   public WidgetId createId(int widgetId)
   {
      return new WidgetId(widgetId);
   }
   
   /**
    * Get the numeric representation of this widget id.
    * 
    * @return   The {@link #id}.
    */
   public int asInt()
   {
      return id;
   }
   
   /**
    * Replacement for the default object writing method. The latest state is written to the output
    * destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out) throws IOException
   {
      out.writeInt(id);
   }
   
   /**
    * Replacement for the default object reading method. The latest state is read from the input 
    * source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      id = in.readInt();
      hashCode = computeHashCode();
   }
   
   /**
    * Check if the given object is the same as this one.
    * 
    * @param    obj
    *           The candidate.
    * 
    * @return   <code>true</code> if this instance represents the same widget ID as the given one.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (!(obj instanceof WidgetId))
      {
         return false;
      }
      
      WidgetId id = (WidgetId) obj;
      
      return id.id == this.id;
   }
   
   /**
    * Return the cached hash code for this widget ID.
    * 
    * @return   See above.
    */
   @Override
   public int hashCode()
   {
      return hashCode;
   }
   
   /**
    * Get a string representation of this ID.
    * 
    * @return   See above.
    */
   @Override
   public String toString()
   {
      return Integer.toString(id);
   }
   
   /**
    * Compare this instance with the given object.  Return value matches the 
    * {@link Comparable#compareTo}.
    *
    * @param    wid
    *           The widget ID to compare against.
    * 
    * @return   See {@link Comparable#compareTo}.
    */
   @Override
   public int compareTo(WidgetId wid)
   {
      return this.id - wid.id;
   }
   
   /**
    * Compute a hash code for this widget ID.
    * 
    * @return   See above.
    */
   private int computeHashCode()
   {
      int hash = 17;
      hash = hash * 37 + id;
      
      return hash;
   }
   
   /**
    * The base class for widget ID generation.
    */
   private abstract static class IDGenerator
   {
      /** The initial value for next widget ID. */
      protected int nextId;
      
      /** The value from which IDs can be allocated for created widgets. */
      private final int initialValue;
      
      /** Flag indicating the integer wrapping happen while incrementing by 1. */
      private boolean wrapped = false;
      
      /**
       * Create a new ID generator.
       * 
       * @param    initialValue
       *           The value from which IDs can be allocated for created widgets.
       */
      public IDGenerator(int initialValue)
      {
         this.initialValue = initialValue;
         this.nextId = initialValue;
      }
      
      /**
       * Retrieve the unique per session integer id number for next new widget to create.
       *
       * @return   The integer Id number for next dynamic widget to create. 
       *
       * @throws   RuntimeException
       *           If all possible IDs are in use. This is very unlikely but could happen!
       */
      public int allocateID()
      {
         // per call wrapping indicator
         int wrapCount = 0;
         int candidate;
         
         while (true)
         {
            // we are deliberately assigning the value before getting the next value
            candidate = nextID();
            
            if (wrapped)
            {
               // if the id is already in use
               if (usedId(candidate))
               {
                  continue;
               }
               
               // don't return directly from here
            }
            
            // protection from infinite loop when there is no free ID to allocate
            // we are here on the second wrapping
            if (wrapped(candidate) && wrapCount > 1)
            {
               throw new RuntimeException("No free ID to allocate!");
            }
               
            if (wrapped(candidate))
            {
               wrapped = true;
               nextId  = initialValue;
               wrapCount++;
            }
            
            // at this point we know that candidate is OK so we exit our loop
            break;
         }
         
         return candidate;
      }
      
      /**
       * Get the next ID, in  sequential order.
       * 
       * @return   The next ID.
       */
      protected abstract int nextID();
      
      /**
       * Check if the given ID is already used.
       * 
       * @param    id
       *           The ID to check.
       * @return   <code>true</code> if the ID is already used by some other widget.
       */
      protected abstract boolean usedId(int id);
      
      /**
       * Check if the ID has wrapped.
       * 
       * @param    id
       *           The ID to check.
       *           
       * @return   <code>true</code> if the ID has wrapped.
       */
      protected abstract boolean wrapped(int id);
   }
   
   /** 
    * Stores global data relating to the state of the current session context and provides
    * server-side ID generator.
    */
   private static class SessionArea
   extends IDGenerator
   {
      /**
       * Create the server-side  ID generator.
       */
      public SessionArea()
      {
         super(WIDGET_ID_RANGE_BEGIN);
      }
      
      /**
       * Get the next ID, in  sequential order.
       * 
       * @return   The next ID.
       */
      @Override
      protected int nextID()
      {
         return nextId++;
      }
      
      /**
       * Check if the given ID is already used.
       * 
       * @param    id
       *           The ID to check.
       * @return   <code>true</code> if the ID is part of the server-side registry.
       */
      @Override
      protected boolean usedId(int id)
      {
         // lookup candidate in the widget registry to see if a widget of that ID already
         // exists, don't keep a new widget registry, use the one that already exists
         
         return LogicalTerminal.isWidgetIdInUse(id);
      }
      
      /**
       * Check if the ID has wrapped.  On server-side, this is when the ID becomes negative.
       * 
       * @param    id
       *           The ID to check.
       *           
       * @return   <code>true</code> if the ID has wrapped.
       */
      @Override
      protected boolean wrapped(int id)
      {
         return id < 0;
      }
   }
   
   /** 
    * Stores global data relating to the state of the current session context, and provides 
    * client-side ID generator.
    */
   private static class ClientArea
   extends IDGenerator
   {
      /** Cached {@link ThinClient} instance. */
      private final ThinClient tc;
      
      /**
       * Create the client-side  ID generator.
       */
      public ClientArea()
      {
         super(-1);
         
         tc = ThinClient.getInstance();
      }
      
      /**
       * Get the next ID, in  sequential order.
       * 
       * @return   The next ID.
       */
      @Override
      protected int nextID()
      {
         return nextId--;
      }
      
      /**
       * Check if the given ID is already used.
       * 
       * @param    id
       *           The ID to check.
       * @return   <code>true</code> if the ID is part of the client-side registry.
       */
      @Override
      protected boolean usedId(int id)
      {
         return tc.getWidget(id) != null;
      }
      
      /**
       * Check if the ID has wrapped.  On client-side, this is when the ID becomes positive.
       * 
       * @param    id
       *           The ID to check.
       *           
       * @return   <code>true</code> if the ID has wrapped.
       */
      @Override
      protected boolean wrapped(int id)
      {
         return id >= 0;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<IDGenerator>
   {
      /**
       * Obtains the context-local instance of the contained global work area for the session.
       *
       * @return   The work area associated with this context.
       */
      public IDGenerator obtain()
      {
         return this.get();
      }
      
      /**
       * Initializes the work area, the first time it is requested within a new session context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized IDGenerator initialValue()
      {
         IDGenerator gen = (SessionManager.get().isLeaf() ? new ClientArea() : new SessionArea());
         
         return gen;
      }
   }
}