/*
** Module   : WindowConfig.java
** Abstract : window widget attribute container.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20050831   @22524 Created initial version. Part of refactoring
**                           (see @@22394-22409). 
** 002 NVS 20051128   @23543 Added fields and methods to serve window
**                           attributes max-height-chars, max-width-chars.
** 003 SIY 20060201   @24213 Refactoring and cleanups.      
** 004 NVS 20060608   @27012 Assigned a permanent ID to the session window
** 005 GES 20060802   @28421 Made externalizable for better performance.
**                           Many code standard cleanups.
** 006 GES 20061120   @31334 Removed dead code.
** 007 SIY 20100619          Switching to own implementation of character UI.
** 008 SIY 20100804          Changes in imports, some refactorings.
** 009 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 010 MAG 20140512          Implemented new attributes and support for CREATE WINDOW.
** 011 CA  20140711          Enhanced font support, for GUI clients.
** 012 CA  20140805          Do not set the window ID upfront, to allow multi-window support.  
**                           Added GUI-only hack to allow the window title to be pushed from the 
**                           server-side to client-side, without overriding other config.
** 013 MAG 20140802          Add environment name associated with window. 
** 014 MAG 20140909          Fixed NPE on serialization of envName. 
** 015 EVL 20140812          In new widget ID allocation approach ID for default window is taking
**                           from WidgetId class static constant.
** 016 HC  20140911          GUI coordinate support - data types refactoring.
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 VIG 20141024          Added RESOLVE_WINDOW constant.
** 019 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 020 CA  20141123          No special processing is required when setting the window's COLUMN,
**                           ROW, WIDTH-CHARS or HEIGHT-CHARS attributes.  Also, the WIDTH-/HEIGHT-
**                           CHARS attributes need to reflect the workspace's dimension.
** 021 EVL 20150108          Differentiate small icon from large one flag.  Rework for icon/image
**                           implementation.
** 022 VIG 20150120          Added menubarId.
** 023 CA  20150415          On client-side, track the origin of window iconification.
**                           Added support for window's TITLE attribute.
** 024 HC  20150323          Removed business logic from config classes.
** 025 CA  20150507          Message and status-area fonts need to be initialized to the 
**                           DEFAULT-FONT.
** 026 HC  20150508          Improvements to GUI window scrolling support.
** 027 HC  20150803          Improved runtime support for WINDOW size attributes.
** 028 IAS 20151007          Added support for the TOP-ONLY attribute.
** 029 IAS 20151214          MOUSE-POINTER attribute support.
** 030 IAS 20160303          Removed sessDataEntryReturn field.
** 031 ECF 20170130          Replaced default gclogo.ico with fwd.ico.
**     HC  20170202          Implemented runtime support for WINDOW-STATE. The implementation is
**                           missing support for WINDOW-DELAYED-MINIMIZE.
** 032 OM  20170623          Replaced iconifiedBy with hiddenByIconificationOf because of
**                           different semantics.
** 033 HC  20170629          Implemented support of WINDOW:DISABLE-REDRAW, the FWD extension to
**                           4GL.
** 034 EVL 20170926          Javadoc fix.
** 035 HC  20171027          Extended DISABLE-REDRAW to all widget types. Previously only WINDOW
**                           widget was supported.
** 035 SBI 20180716          Added z-order widget class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;

/**
 * Window widget configuration container.
 */
public class WindowConfig
extends PaneConfig
{
   /** 
    * Window id which used in calls of client and means, that target window in the client-side
    * must be resolved automatically.
    */
   public static final int RESOLVE_WINDOW = -1;

   /** Normal window icon size. */   
   public static final int ICON_SIZE = 32;
   
   /** Small sized window icon size. */   
   public static final int SMALL_ICON_SIZE = 16;
   
   /** Default application icon directory section. */
   public static final String APP_ICON_SECTION = "window-icon";
   
   /** Default application icon. */
   public static final String DEFAULT_ICON = "fwd.ico";
   
   /** The Progress FULL-HEIGHT-CHARS attribute. */
   public double fullHeightChars = BaseConfig.INV_COORD;
   
   /** The Progress FULL-HEIGHT-PIXELS attribute. */
   public int fullHeightPixels = BaseConfig.INV_COORD;
   
   /** The Progress SCREEN-LINES attribute. */
   public double screenLines = 0;
   
   /** The Progress KEEP-FRAME-Z-ORDER attribute. */
   public boolean keepFrameZOrder = false;
   
   /** The Progress NUM-SELECTED-WIDGET attribute. */
   public int numSelectedWidget = 0;

   /** The Progress FULL-WIDTH-CHARS attribute. */
   public double fullWidthChars = BaseConfig.INV_COORD;
   
   /** The Progress FULL-WIDTH-PIXELS attribute. */
   public int fullWidthPixels = BaseConfig.INV_COORD;
   
   /** MAX-HEIGHT-CHARS attribute, use {@link ConfigHelper} to get the effective value. */
   public double maxHeightChars = BaseConfig.INV_COORD;

   /** MAX-WIDTH-CHARS attribute, use {@link ConfigHelper} to get the effective value. */
   public double maxWidthChars = BaseConfig.INV_COORD;

   /** MAX-HEIGHT-PIXELS attribute, use {@link ConfigHelper} to get the effective value. */
   public int maxHeightPixels = BaseConfig.INV_COORD;

   /** MAX-WIDTH-PIXELS attribute, use {@link ConfigHelper} to get the effective value. */
   public int maxWidthPixels = BaseConfig.INV_COORD;
   
   /** MIN-HEIGHT-CHARS attribute, use {@link ConfigHelper} to get the effective value. */
   public double minHeightChars = BaseConfig.INV_COORD;

   /** MIN-WIDTH-CHARS attribute, use {@link ConfigHelper} to get the effective value. */
   public double minWidthChars = BaseConfig.INV_COORD;

   /** MIN-HEIGHT-PIXELS attribute, use {@link ConfigHelper} to get the effective value. */
   public int minHeightPixels = BaseConfig.INV_COORD;

   /** MIN-WIDTH-PIXELS attribute, use {@link ConfigHelper} to get the effective value. */
   public int minWidthPixels = BaseConfig.INV_COORD;
   
   /** The value of MAX-HEIGHT-PIXELS assigned to unrealized WINDOW widget. */
   public int unrealizedMaxHeightPixels = BaseConfig.INV_COORD;
   
   /** Work area truncation condition maintained by the client. */ 
   public boolean waTruncCondition;
   
   /** The flag indicates whether window height has been assigned. */
   public boolean heightAssigned;
   
   /** HELP attribute. */
   public String help = null;

   /** The Progress RESIZE attribute. */
   public boolean resize = true;
   
   /** The Progress SCROLL-BARS attribute. */
   public boolean scrollBars = false;
   
   /** The Progress STATUS-AREA attribute. */
   public boolean statusArea = true;

   /** The Progress MESAGE-AREA attribute. */
   public boolean messageArea = true;
   
   /** The Progress ICON attribute used by LOAD-ICON(). */
   public String iconFileName;
   
   /** Flag indicating the icon was loaded by LOAD-SMALL-ICON(). */
   public boolean useSmallIcon = false;

   /** The font for the window's message area. */
   public int messageAreaFont = FontManager.DEFAULT_FONT;
   
   /** The font for the window's status area. */
   public int statusAreaFont = FontManager.DEFAULT_FONT;
   
   /** Environment name */
   public String envName;
   
   /** ID of menubar */
   public int menubarId = 0;
   
   /** Whether this container is over all other non-top only containers. */
   public boolean topOnly = false;

   /** Value of ALWAYS-ON-TOP attribute. */
   public boolean alwaysOnTop = false;

   /** MOUSE-POINTER attribute */
   public String mousePointer = null;

   /**
    * The ID of the closer ancestor window that was iconified and triggered this window not to be
    * visible in taskbar. When the ancestor is restored, this flag is reset to -1.  
    */
   public transient int hiddenByIconificationOf = -1;

   /**
    * WINDOW-STATE attribute; to get the effective value use
    * {@link WindowWidget#getWindowState}
    */
   public byte windowState = CommonWindow.WINDOW_STATE_NORMAL;

   /** Set when window should exhibit the delayed minimize behavior */
   public boolean delayedMinimize;

   /** Changes the visibility of the window minimize button. */
   public boolean minButton = true;

   /** Changes the visibility of the window maximize button. */
   public boolean maxButton = true;

   /** Value of the SHOW-IN-TASKBAR attribute */
   public boolean showInTaskbar = true;

   /** Value of CONTROL-BOX attribute. */
   public boolean controlBox = true;

   /** Value of SMALL-TITLE attribute. */
   public boolean smallTitle = false;

   /**
    * Default constructor (only used in deserialization).
    */
   public WindowConfig()
   {
      this.visible = false;
      this.title = null;
      
      zorder = ZOrderClass.WINDOW;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected WindowConfig(int id)
   {
      super(id);

      this.visible = false;
      this.title = null;
      
      zorder = ZOrderClass.WINDOW;
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected WindowConfig(WidgetId id)
   {
      super(id);
      
      this.visible = false;
      this.title = null;
      
      zorder = ZOrderClass.WINDOW;
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      WindowConfig cfg = (WindowConfig) config;
      
      fullHeightChars = cfg.fullHeightChars;
      fullHeightPixels = cfg.fullHeightPixels;
      screenLines = cfg.screenLines;
      keepFrameZOrder = cfg.keepFrameZOrder;
      numSelectedWidget = cfg.numSelectedWidget;
      fullWidthChars = cfg.fullWidthChars;
      fullWidthPixels = cfg.fullWidthPixels;
      maxHeightChars = cfg.maxHeightChars;
      maxWidthChars = cfg.maxWidthChars;
      maxHeightPixels = cfg.maxHeightPixels;
      maxWidthPixels = cfg.maxWidthPixels;
      minHeightChars = cfg.minHeightChars;
      minWidthChars = cfg.minWidthChars;
      minHeightPixels = cfg.minHeightPixels;
      minWidthPixels = cfg.minWidthPixels;
      unrealizedMaxHeightPixels = cfg.unrealizedMaxHeightPixels;
      waTruncCondition = cfg.waTruncCondition;
      heightAssigned = cfg.heightAssigned;
      help = cfg.help;
      resize = cfg.resize;
      scrollBars = cfg.scrollBars;
      statusArea = cfg.statusArea;
      messageArea = cfg.messageArea;
      iconFileName = cfg.iconFileName;
      useSmallIcon = cfg.useSmallIcon;
      messageAreaFont = cfg.messageAreaFont;
      statusAreaFont = cfg.statusAreaFont;
      envName = cfg.envName;
      menubarId = cfg.menubarId;
      topOnly = cfg.topOnly;
      mousePointer = cfg.mousePointer;
      windowState = cfg.windowState;
      delayedMinimize = cfg.delayedMinimize;
      minButton = cfg.minButton;
      maxButton = cfg.maxButton;
      showInTaskbar = cfg.showInTaskbar;
      alwaysOnTop = cfg.alwaysOnTop;
      controlBox = cfg.controlBox;
      smallTitle = cfg.smallTitle;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      fullHeightChars     = in.readDouble();
      fullHeightPixels    = in.readInt();
      fullWidthChars      = in.readDouble();
      fullWidthPixels     = in.readInt();
      screenLines         = in.readDouble();
      numSelectedWidget   = in.readInt();
      keepFrameZOrder     = in.readBoolean();
      maxHeightChars      = in.readDouble();
      maxWidthChars       = in.readDouble();
      maxHeightPixels     = in.readInt();
      maxWidthPixels      = in.readInt();
      minHeightChars      = in.readDouble();
      minWidthChars       = in.readDouble();
      minHeightPixels     = in.readInt();
      minWidthPixels      = in.readInt();
      unrealizedMaxHeightPixels = in.readInt();
      waTruncCondition    = in.readBoolean();
      heightAssigned      = in.readBoolean();
      resize              = in.readBoolean();
      scrollBars          = in.readBoolean();
      statusArea          = in.readBoolean();
      messageArea         = in.readBoolean();
      iconFileName        = (String) in.readObject();
      useSmallIcon        = in.readBoolean();
      help                = (String) in.readObject();
      messageAreaFont     = in.readInt();
      statusAreaFont      = in.readInt();
      envName             = (String) in.readObject();
      menubarId           = in.readInt();
      topOnly             = in.readBoolean();
      mousePointer        = (String) in.readObject();
      windowState         = in.readByte();
      delayedMinimize     = in.readBoolean();
      minButton           = in.readBoolean();
      maxButton           = in.readBoolean();
      showInTaskbar       = in.readBoolean();
      alwaysOnTop         = in.readBoolean();
      controlBox          = in.readBoolean();
      smallTitle          = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeDouble(fullHeightChars);
      out.writeInt(fullHeightPixels);
      out.writeDouble(fullWidthChars);
      out.writeInt(fullWidthPixels);
      out.writeDouble(screenLines);
      out.writeInt(numSelectedWidget);
      out.writeBoolean(keepFrameZOrder);
      out.writeDouble(maxHeightChars);
      out.writeDouble(maxWidthChars);
      out.writeInt(maxHeightPixels);
      out.writeInt(maxWidthPixels);
      out.writeDouble(minHeightChars);
      out.writeDouble(minWidthChars);
      out.writeInt(minHeightPixels);
      out.writeInt(minWidthPixels);
      out.writeInt(unrealizedMaxHeightPixels);
      out.writeBoolean(waTruncCondition);
      out.writeBoolean(heightAssigned);
      out.writeBoolean(resize);
      out.writeBoolean(scrollBars);
      out.writeBoolean(statusArea);
      out.writeBoolean(messageArea);
      out.writeObject(iconFileName);
      out.writeBoolean(useSmallIcon);
      out.writeObject(help);
      out.writeInt(messageAreaFont);
      out.writeInt(statusAreaFont);
      out.writeObject(envName);
      out.writeInt(menubarId);
      out.writeBoolean(topOnly);
      out.writeObject(mousePointer);
      out.writeByte(windowState);
      out.writeBoolean(delayedMinimize);
      out.writeBoolean(minButton);
      out.writeBoolean(maxButton);
      out.writeBoolean(showInTaskbar);
      out.writeBoolean(alwaysOnTop);
      out.writeBoolean(controlBox);
      out.writeBoolean(smallTitle);
   }
}
