/*
** Module   : ButtonImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111122 Extracted from Button.
** 002 SIY 20111221 Fixed constructor visibility.
** 003 SVL 20120214 Fixed mnemonic processing.
** 004 GES 20140627 Changed color attribute method names to match naming standards.
** 005 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 IAS 20150924 Added AUTO-RESIZE attribute support
** 009 VIG 20160315 Fixed drawing: added mnemonic information instead of underscore flag.
** 010 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * Specific implementation of a "push button" with a text label for CHUI.
 */
@SuppressWarnings("unchecked")
public class ButtonImpl
extends Button<ChuiOutputManager>
implements AutoResizable<ButtonConfig>
{
   /** 
    * The flag indicating whether this button supports mnemonic or not. 
    * Default - doesn't support, but can be supported in ALERT-BOX. 
    */
   private boolean supportsMnemonic = false;
   
   /**
    * This constructor creates a button with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   ButtonImpl(WidgetId id, Container<ChuiOutputManager> owner, String text) 
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ButtonImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ButtonConfig) cfg);
   }
   
   /**
    * Set the mnemonics support for this button.
    */
   @Override
   public void setSupportsMnemonic()
   {
      supportsMnemonic = true;
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      // get the absolute origin of this component
      Point origin = screenLocation();

      screen().at(origin);
      Color dColor = UiUtils.getDColor(this);

      //  "highlighted" flag tells whether this button needs emphasis
      if (super.isEnabled() && highlighted())
      {
         // Progress behaves somewhat strange here:
         // 1. Focused blinking button does not blink
         // 2. Focused underlined button blinks(!) and does not use reverse
         //    attribute.
         if (dColor.hasUnderline())
         {
            dColor = dColor.withoutUnderline().withBlink();
         }
         else
         {
            dColor = dColor.withReverse().withoutBlink();
         }
      }
      
      // obtain the full text of the button's label
      String ct = textLabel();
      
      // calculate the total size of the button
      // in chui char coordinates are always whole numbers
      int bSize = (int) width();
      
      // calculate the max size of the text part
      int tSize = bSize - 2;
      
      // calculate the actual text length (may be shorter or longer than the
      // text size)
      int tActual = ct.length(); 
      
      StringBuffer sb = new StringBuffer();
      
      // there is always the left angle bracket since the size must be at
      // least 1 (but it may only be 1... and this would be the only output)
      sb.append("<");
      
      if (tSize > 0)
      {
         if (tSize < tActual)
         {
            // right truncate the text
            sb.append(ct.substring(0, tSize));
         }
         else if (tSize == tActual)
         {
            // we are just big enough to hold the full text
            sb.append(ct);
         }
         else
         {
            // we must pad with spaces (if there an even number of spaces
            // then the left and right sides will have the same number, if
            // odd then the right side will have the extra space)
            
            int spaces = tSize - tActual;
            int left   = spaces / 2; 
            int right  = (spaces % 2) > 0 ? left + 1 : left;
            
            for (int i = 0; i < left; i++)
            {
               sb.append(" ");
            }
            
            sb.append(ct);
            
            for (int i = 0; i < right; i++)
            {
               sb.append(" ");
            }
         }
      }
      
      // any button size of 2 or more will always have a right angle
      // bracket, even if it has no text
      if (bSize > 1)
      {
         sb.append(">");
      }

      // note that this draws in the non-reversed form whether selected
      // or not (this is the equivalent of a "highlight" in the selected
      // mode because the rest of the button was drawn in reverse video

      // mnemonic processing
      MnemonicInfo label = getMnemonic();
      String title = sb.toString();

      if (hasFocus())
      {
         dColor = dColor.withReverse();
      }

      Point tp = new Point(origin);

      if (label == null || label.getMnemonic() == null)
      {
         // actually draw the core button
         screen().append(title, dColor);
      }
      else
      {
         Color mnemonicColor = dColor.withUnderline().withoutReverse();
         
         // include prepending "<"
         Integer ampIdx = label.getMnemIdx() + 1; 

         String leftPart = title.substring(0, ampIdx);
         screen().at(tp).append(leftPart, dColor);

         tp.x += ampIdx; 

         if (hasFocus())
         {
            mnemonicColor = mnemonicColor.withoutUnderline();
         }

         screen().at(tp).append(title.substring(ampIdx, ampIdx + 1), mnemonicColor);

         tp.x += 1;
         String rightPart = title.substring(ampIdx + 1);
         screen().at(tp).append(rightPart, dColor);
         tp.x += rightPart.length();
      }
   }
   
   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();

      screen().at(screenLocation().offset(1,0));
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

    /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ButtonConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.label, beforeUpdate.label))
      {
         config.widthChars = width();
         config.heightChars = height();
      }
   }

   /**
    * Check whether this button supports mnemonics. Intended to be overridden.
    * CHUI: buttons do not support mnemonics except the ones in ALERT-BOX.
    * GUI: all buttons support mnemonics. 
    * 
    * @return   <code>true</code> if this button supports mnemonic.
    */
   public boolean isMnemonicSupported()
   {
      return supportsMnemonic;
   }
}