/* 
** Module   : ContentResolver.java
** Abstract : Worker class which resolves the content for TEXT/FILL-IN/LITERAL widgets in ChUI.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.  Extracted and adjusted from FillIn.getText(boolean).
** 002 HC  20150323 Removed business logic from config classes.
** 003 CA  20150805 Fixed FILL-IN cursor problems when editable width is smaller than content
**                  length.
** 004 SVL 20160419 Minor refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.util.*;

/**
 * Worker class which resolves the content for TEXT/FILL-IN/LITERAL widgets in ChUI.
 */
class ContentResolver
{
   /** Stores amount of pad on left side for right aligned widgets. */
   protected int leftPad = 0;
   
   /** Marks a blank (all spaces) right aligned display value. */
   protected boolean cleared = false;

   /** Reference to the widget which performed the request to resolve its contents. */
   private final FillIn<ChuiOutputManager,?> fillIn;
   
   /**
    * Create a new instance and associate it with the given widget.
    * 
    * @param    fillIn
    *           The source widget.
    */
   ContentResolver(FillIn<ChuiOutputManager,?> fillIn)
   {
      this.fillIn = fillIn;
   }

   /**
    * Get screen representation of the data.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative variable value is used 
    *          when available.
    * 
    * @return  Text representation of the variable value according to all current settings, 
    *          flags, format, etc.
    */
   public String getText(boolean isInternal)
   {
      FillInConfig cfg = fillIn.config();
      
      DisplayFormat.Presentation curPres = cfg.appDataPres;
      DisplayFormat              curFmt  = cfg.appFormat;
      
      if (isInternal && cfg.displayVar != null && cfg.displayDataPres != null)
      {
         curPres = cfg.displayDataPres;
         curFmt  = cfg.displayFormat;
      }

      String screenValue = fillIn.getScreenValue(isInternal);
      int contentWidth = (int) fillIn.width();
      
      String display = null;
      
      if (fillIn.config().blank)
      {
         display = DisplayFormat.generateSpaces(contentWidth);
      }
      else
      {
         boolean leftTrim = !cfg.widgetFocused && curFmt.isRightPositioning();
         boolean rightAlign = fillIn.isRightAlignedInt();
         
         // this section of code implements some fixups and saves the result 
         // of some calculations to properly support the display and cursor
         // positioning of NUMBERS (other data types will all bypass this
         // code below until the formatScreenValue call)
         
         // reset left padding
         leftPad = -1;
         
         // check for special case where the left sign is optional
         if (curPres != null && curPres.isOptionalLeftSign())
         {
            // special case where the optional sign space must be put back
            // because the rendering process removes it (this affects formats
            // such as "-999" which will display 1 char too far left in the
            // non-negative case
            if (!curPres.isUnknown()      &&
                !rightAlign               &&
                !curPres.isNegative()     && 
                curPres.isFixedNumeric())
            {
               StringBuilder sb = new StringBuilder(" ").append(screenValue);
               screenValue = sb.toString();
               
               // we aren't right aligned, but in this case we do need to
               // shift the cursor
               leftPad = 1;
            }
         }
         
         cleared = false;
         
         // calculate the amount of left padding will be present in the case
         // of a right aligned widget
         if (rightAlign)
         {
            int len = screenValue.length();
            DisplayFormat fmt = fillIn.getCurrentFormat(); 
            
            if (fmt instanceof NumberFormat &&
                screenValue.equals("?")     &&
                fmt.isRightPositioning())
            {
               // initial number formatting case, suppress padding
               len = contentWidth;
            }
            
            // calc and save the amount of padding on the left that will 
            // occur (it would be better to get this from DisplayFormat 
            // since we essentially are hard coding the algorithm in 2
            // places, but that would require a more significant change to
            // the interface, so we cheap it and do it here)
            leftPad = (contentWidth <= len) ? 0 : (contentWidth - len);
            
            cleared = true;
            
            // detect all-blank output and reset the cleared flag in that
            // case
            for (int j = 0; j < len; j++)
            {
               // exit at first non-blank
               if (screenValue.charAt(j) != ' ')
               {
                  cleared = false;
                  break;
               }
            }
         }
         
         if (curFmt instanceof NumberFormat)
         {
            NumberFormat nf = (NumberFormat) curFmt;
            
            if (!rightAlign &&
                !nf.isRightPositioning() &&
                nf.isSplitFormat())
            {
               // handle the case when an left-aligned numeric field with a
               // side-labeled frame gets right-aligned
               int savedLen = screenValue.length();
               
               screenValue = screenValue.trim();
               
               int len = screenValue.length();
               int spaceCount = savedLen - len;
               
               StringBuilder sb = new StringBuilder(screenValue);

               // add spaceCount spaces to the end of the screen
               for (int j = 0; j < spaceCount; j++)
               {
                  sb.append(' ');
               }
               screenValue = sb.toString();
            }
         }
         
         int offsScroll = fillIn.getEditScroller().getOffset();
         
         display = DisplayFormat.formatScreenValue(screenValue,
                                                   contentWidth,
                                                   offsScroll,
                                                   leftTrim,
                                                   rightAlign);
         
         if (display.length() != contentWidth)
         {
            // ensure the "display" part fully fils the width
            if (rightAlign)
            {
               display = StringHelper.rightAlignText(display, contentWidth);
            }
            else
            {
               display = StringHelper.leftAlignText(display, contentWidth);
            }
         }
      }
      
      return display;
   }
}