/*
** Module   : DoubleBufferedTerminal.java
** Abstract : terminal output driver that outputs diffs only
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20090209   @41282 A plugin which provides a compatible back-end
**                           for interactive terminal, but uses double
**                           buffering to output only the absolute minimum
**                           which is calculated as the difference between
**                           the current and the previous buffers.
** 002 SIY 20090212   @41292 Minor spelling and javadoc cleanups.
** 003 NVS 20090213   @41313 Minor code cleanup
** 004 NVS 20090225   @41361 Added debugging aids for trapping specific
**                           output.
** 005 SIY 20090312   @41544 Added support for conditional widget 
**                           refreshing.
** 006 SIY 20090319   @41646 Changed addArray() signature. Added method 
**                           for conditional refreshing of given rectangle.
** 007 SIY 20090320   @41661 Added bounds checking in conditionalRefresh().
** 008 SIY 20090911   @43895 Now class is inherited from BasePrimitives.
** 009 SIY 20100115   @44529 Added conditionalLineRefresh() method. 
** 010 EVL 20100715          New method has been added to be able to refresh
**                           the mask cells state. Required if we need to paint
**                           in the reserved lines at the bottom of screen.
** 011 SIY 20100619          Switching to own implementation of character UI.
** 012 SIY 20100804          Changes in imports, some refactorings.
** 013 SIY 20101202          Constants moved to OutputManager.
** 014 SIY 20110112          Changes related to new color handling scheme.
** 015 CA  20140805          Allow direct access to all ChUI specific APIs provided by the 
**                           BasePrimitives, as this is the base class for all ChUI primitives.
** 016 HC  20140911          GUI coordinate support - data types refactoring.
** 017 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 018 IAS 20160331          Fixed potential NPE
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Provides a plugin as a compatible back-end for interactive terminal, which
 * uses a double buffering technique to output only the absolute minimum
 * of information to the terminal, which is calculated as the difference 
 * between the current and the previous buffers.
 * <p>
 * The master buffer holds the most recent rendering of the terminal screen
 * image. The current buffer is where the new rendering is performed until
 * a sync() request comes.
 * <p>
 * Syncing means the current rendering is over. The two buffers are compared
 * position by position and only the differing positions are transmitted to
 * the terminal screen using bulk update. The same positions are copied to
 * the master buffer as well. When the syncing is over, the buffers look
 * identical.
 */
public class DoubleBufferedTerminal
extends AbstractTerminalBuffer
{
   /** The OutputManager instance. */
   private ChuiOutputManager screen = (ChuiOutputManager) OutputManager.instance();

   /** The direct terminal output driver. */
   private BasePrimitives directOutput = screen.getDirectOutput();

   /** Master buffer dirty flag */
   private boolean dirty = false;

   /** Master buffer. */
   private Cell[][] master = new Cell[rows][cols];

   /** Low level call parameters array. */
   private Cell[][] matrix = new Cell[rows][cols];

   /** Mask for actually used cells. */
   private ScreenBitmap mask = new ScreenBitmap(rows, cols);
   
   /** Outer rectangle for all used cells. */
   private NativeRectangle outer = null;
   
   /** (Debug) trapping in rendering buffer is active. */
   private boolean trapR = false;

   /** (Debug) trapping in master buffer is active. */
   private boolean trapM = false;

   /** (Debug) trapping of refreshing is active. */
   private boolean trapF = false;

   /** (Debug) trapping rectangle for rendering buffer. */
   private NativeRectangle trapRend = new NativeRectangle(0, 0, 0, 0);
   
   /** (Debug) trapping rectangle for master buffer. */
   private NativeRectangle trapMast = new NativeRectangle(0, 0, 0, 0);
   
   /** (Debug) trapping rectangle for refreshing the master buffer. */
   private NativeRectangle trapRefr = new NativeRectangle(0, 0, 0, 0);

   /**
    * Constructor.
    * Creates both buffers and does the initial sync to clear screen.
    * 
    * @param    height
    *           The terminal's height.
    * @param    width
    *           The terminal's width.
    */
   public DoubleBufferedTerminal(int height, int width)
   {
      super(height, width);

      // do initial sync to clear the screen
      dirty = true;
      sync();
      mask.resetScreen(false);
   }

   /**
    * Add a character to the virtual terminal at the current cursor
    * position.
    * <p>
    * This routine is NOT safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           The number of the color-pair (foreground + background) to
    *           use for blanking the rectangle.
    */
   @Override
   public void append(int chr, Color color)
   {
      // put the data and video attributes
      render()[cy][cx] = new Cell((char)chr, color);
      mask.setBit(cy, cx, true);

      NativeRectangle charCell = new NativeRectangle(cy, cx, cy, cx);

      outer = outer == null ? charCell : outer.union(charCell);

      // debug support
      if (trapR)
      {
         trap(trapRend, cy, cx);
      }
      
      advanceCursor();

      // mark the buffer dirty
      dirty = true;
   }

   /**
    * Callback method when cursor is actually relocated at given position.
    *
    * @param   csx
    *          Cursor X.
    * @param   csy
    *          Cursor Y.
    */
   @Override
   protected void cursorRelocated(int csx, int csy)
   {
      directOutput.cursorAt(csx, csy);
      directOutput.sync();
   }

   /**
    * Mark a rectangle as needing repainting.
    *
    * @param    rect
    *           rectangle to mark for repainting in screen coordinates
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
      refreshWorker(rect, false);
   }

   /**
    * Mark a rectangle as needing repainting within the mask array.
    *
    * @param    rect
    *           rectangle to mark for repainting in screen coordinates
    */
   public void refreshMask(NativeRectangle rect)
   {
      refreshWorker(rect, true);
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()      
   {
      // even though the buffer may be dirty, it may still produce no diff
      boolean needSync = false;

      // the cursor may need syncing alone
      boolean needCursorSync = false;

      if (dirty)
      {
         // reset the flag
         dirty = false;

         needSync = flush(true);
   
         // sync the direct drawing if any
         if (needSync)
         {
            directOutput.updateScreen(new ScreenData(matrix, csx, csy));
         }
         else
         {
            needCursorSync = true;
         }
      }
      else
      {
         needCursorSync = true;
      }

      // sync the cursor if moved and not synced yet
      if (needCursorSync)
      {
         directOutput.cursorAt(csx, csy);
         directOutput.sync();
      }
   }
   
   /**
    * Mark a rectangle as needing repainting in either the main bitmap
    * array or within the mask array.
    *
    * @param    rect
    *           Rectangle to mark for repainting in screen coordinates.
    * @param    isMask
    *           <code>true</code> to edit the mask array, <code>false</code>
    *           to edit the main bitmap array.
    */
   private void refreshWorker(NativeRectangle rect, boolean isMask)
   {
      NativeRectangle bounds = new NativeRectangle(Math.max(rect.top(), 0),
                                                   Math.max(rect.left(), 0),
                                                   Math.min(rect.bottom(), rows - 1),
                                                   Math.min(rect.right(), cols - 1));

      for (int i = bounds.top(); i <= bounds.bottom(); i++)
      {
         for (int j = bounds.left(); j <= bounds.right(); j++)
         {
            if (isMask)
            {
               mask.setBit(i, j, true);
            }
            else
            {
               master[i][j] = null;
            }

            // debug support
            if (trapF)
            {
               trap(trapRefr, i, j);
            }
         }
      }

      if (!isMask)
         dirty = true;
   }
   
   /**
    * Compare buffers and fill the matrix for update.
    * 
    * @param    merge
    *           If <code>true</code> both buffers are synchronized.
    * 
    * @return   <code>true</code> if sync is required.
    */
   private boolean flush(boolean merge)
   {
      boolean needSync = false;
      
      // scan the buffers linewise and fill the low level output matrix
      for (int i = 0; i < rows; i++)
      {
         for (int j = 0; j < cols; j++)
         {
            // change detection
            if (Utils.equals(master[i][j], render()[i][j]))
            {
               // this position is unchanged
               matrix[i][j] = null;
            }
            else
            {
               needSync = true;
  
               if (merge)
               {
                  // save the changes first
                  master[i][j] = render()[i][j];
               }

               matrix[i][j] = render()[i][j];

               // debug support
               if (trapM && merge)
               {
                  trap(trapMast, i, j);
               }
            }
         }
      }
      
      return needSync;
   }

   /**
    * Debugging aid that helps in trapping the drawing operation touching
    * specific rectangle of the screen.
    * 
    * @param    rect
    *           Rectangle to watch
    *
    * @param    row
    *           screen position being changed
    * 
    * @param    col
    *           screen position being changed
    */
   private void trap(NativeRectangle rect, int row, int col)
   {
      if (rect.contains(col, row))
      {
         // trap here
         @SuppressWarnings("unused")
         int i = 0;
      }
   }

   /**
    * Conditionally refresh output.
    */
   public void conditionalRefresh()
   {
      if (outer == null)
         return;
      
      BasePrimitives shadow1 = screen.getDirectOutput();
      
      boolean needSync = false;
      
      for (int r = outer.top(); r <= outer.bottom(); r++)
      {
         boolean needRefresh = true;
         boolean spaces      = false;
         boolean lastBit     = false; 
         
         for (int c = outer.left(); c <= outer.right(); c++)
         {
            if (mask.canDraw(r, c) && lastBit == false)
            {
               spaces = false;
            }
            
            lastBit = mask.canDraw(r, c);

            Cell cell = shadow1.cellAt(r, c);

            if (Cell.isSpace(cell))
               spaces = true;
            
            if (spaces)
            {
               if (!Cell.isSpace(cell))
               {
                  needRefresh = false;
                  break;
               }
            }
            else
            {
               if (isSame(r, c, cell))
               {
                  needRefresh = false;
                  break;
               }
            }
         }
         
         if (needRefresh)
         {
            for (int c = outer.left(); c <= outer.right(); c++)
            {
               master[r][c] = null;
            }
            needSync = true;
         }
      }
      
      mask.resetScreen(false);
      outer = null;
      
      if (needSync)
         sync();
   }

   /**
    * Conditionally refresh specified rectangle. First character is assumed an
    * right edge of the frame. If leftmost character of each row in shadow
    * buffer and in current rendering buffer is the same, then we assume that
    * frame is placed over PUT SCREEN and refresh row. Otherwise appropriate
    * row is skipped.
    * 
    * @param    rect
    *           Rectangle to refresh.
    */
   public void conditionalRefresh(Rectangle rect)
   {
      BasePrimitives shadow1 = screen.getDirectOutput();
      
      NativeRectangle nrect = screen.coordinates().rectangleToNative(rect);
      nrect = nrect.intersection(new NativeRectangle(0, 0, rows, cols));
      
      if (nrect == null)
         return;
      
      for (int r = nrect.top(); r <= nrect.bottom(); r++)
      {
         if (isSame(r, nrect.left(), shadow1.cellAt(r, nrect.left())))
         {
            // found PUT SCREEN over frame
            continue;
         }
         
         for (int c = nrect.left() + 1; c <= nrect.right(); c++)
         {
            master[r][c] = null;
         }
      }
   }

   /**
    * Special refreshing algorithm which is used for refreshing of horizontal
    * lines of frame border.
    * 
    * @param    rect
    *           Rectangle where horizontal lines should be refreshed.
    */
   public void conditionalLineRefresh(Rectangle rect)
   {
      if (rect == null)
         return;
      
      BasePrimitives shadow1 = screen.getDirectOutput();
      
      NativeRectangle nrect = screen.coordinates().rectangleToNative(rect);
      nrect = nrect.intersection(new NativeRectangle(0, 0, rows, cols));
      
      for (int r = nrect.top(); r <= nrect.bottom(); r++)
      {
         for (int c = nrect.left(); c <= nrect.right(); c++)
         {
            Cell cell = shadow1.cellAt(r, c);

            if (!isSame(r, nrect.left(), cell) && !Cell.isSpace(cell))
            {
               master[r][c] = null;
            }
         }
      }
   }

   /**
    * Reset outer rectangle.
    */
   public void reset()
   {
      outer = null;
      mask.resetScreen(false);
   }

   /**
    * Compare given cell with rendering buffer and return <code>true</code> if
    * cells are equal.
    *
    * @param   r
    *          Row.
    * @param   c
    *          Column.
    * @param   cell
    *          Cell to compare.
    *
    * @return  <code>true</code> if cells are equal.
    */
   private boolean isSame(int r, int c, Cell cell)
   {
      return Utils.equals(render()[r][c], cell);
   }
}