/*
** Module   : ScrollableListImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111127 Extracted from ScrollableList
** 002 SIY 20111221 Fixed constructor visibility.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20150603 Removed requestSync from drawCaret.
** 007 EVL 20150517 Adding api to set up the top row index.
** 008 HC  20150913 Scrolling support overhaul.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.util.*;

/**
 * CHUI-specific portion of ScrollableList.
 */
public class ScrollableListImpl
extends ScrollableList<ChuiOutputManager>
{
   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   ScrollableListImpl(ComboBox<ChuiOutputManager> box, ListModel<String> model)
   {
      super(box, model);
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();

      screen().at(screenLocation().translate(0, currentRow()));
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // get the absolute origin of this component
      Point origin = screenLocation();
      Color dc  = box.getDColor();
      
      String blanks = StringHelper.repeatChar(' ', (int) width());
      
      // in chui char coordinates are always whole numbers
      int y = (int) location().y;
      
      for (int i = 0; i < model().size() &&
                      y < visibleRows(); i++, y++, origin.y++)
      {
         if (y < 0)
            continue;
         
         screen().at(origin);

         Color color = (i == currentRow()) ? dc.withReverse() : dc;

         StringBuilder buf = new StringBuilder(getItemText(i));
         
         buf.append(blanks).setLength((int) width());

         screen().append(buf.toString(), color);
      }

      for (; y < visibleRows(); y++, origin.y++)
      {
         Color color = (y == 0 && currentRow() == -1) ? dc.withReverse() : dc;

         screen().at(origin);
         screen().append(blanks, color);
      }

      if (hasFocus() && isEnabled())
         drawCaret();
   }

   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      Point origin = screenLocation();

      if (currentRow() > 0)
      {
          origin.y += currentRow();
      }

      screen().at(origin);
   }
   
   /**
    * Directs the widget to scroll to the specified X and Y positions.
    * It is the widget's responsibility to update its visuals accordingly.
    * Note that the position values are in the same units as the returned 
    * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
    * and {@link #getScrollStep()}. 
    * 
    * @param    newXPosition
    *           New X scroll position.
    * @param    newYPosition
    *           New Y scroll position.
    */
   @Override
   public void scroll(Integer newXPosition, Integer newYPosition)
   {
      Point loc = location();
      loc.y = -newYPosition;
      setLocation(loc);
      repaint();
   }
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   @Override
   protected void setTopRow(int index)
   {
      // in ChUI just ensure the item is visible for now
      ensureRowVisible(index);
   }

   /**
    * Calculate the scroll Y position based on the current state. 
    * 
    * @return   the new scroll Y position
    */
   @Override
   protected int calculateScrollYPosition()
   {
      int currentRow = currentRow();
      int topRow = (int) -location().y;
      
      if (currentRow < topRow)
      {
         return currentRow;
      }
      
      int visibleRows = visibleRows();
      int bottomRow = visibleRows + topRow - 1;
      int maxTopRow = model().size() - visibleRows;
      
      // current row is higher than the bottom visible and less
      // than the max top row
      if (currentRow > bottomRow)
      {
         return topRow + currentRow - bottomRow;
      }
      
      return topRow;
   }
}