/*
** Module   : AbstractButton.java
** Abstract : common button like widgets implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140610 Initial implementation.
** 002 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 CA  20150212 Allow events to be captured and processed by the widgets without server-side
**                  or main event-loop notification.  This fixes problems in the EDITING and
**                  WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR widgets.
** 005 HC  20150323 Removed business logic from config classes.
** 006 IAS 20150801 TAB-STOP attribute support
** 007 HC  20151023 Fixed GUI ALERT-BOX keyboard input processing.
** 008 EVL 20160411 Adding accelerator symbol mark('&') handling for label text.
** 009 EVL 20160425 Adding NPE protection for textLabel() method.
** 010 OM  20170320 ThinClient.setHelp() signature change.
** 011 CA  20171126 repaint if button's label has changed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * An implementation of a "abstract button" with a text label. This is the place to handle
 * features common for all button like objects, e.g. buttons, toggle-boxes.  This is one line
 * height button like widget.
 */
public abstract class AbstractButton<O extends OutputManager<?>, C extends ControlConfig>
extends AbstractWidget<O>
implements WidgetWithConfig<C>,
           KeyListener,
           FocusListener
{
   /** The config instance. */
   protected C config;
   
   /** Button label. */
   private String textLabel;

   /** Widget in focus flag. */
   protected boolean widgetFocused = false;
   
   /** Previous cursor visibility flag to restore on focus lost. */
   private boolean curvis = false;
   
   /**
    * Default constructor.
    */
   public AbstractButton()
   {
      super();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public C config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(C beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Check if the given event needs to be ignored, as it can't be treated by the widget.
    * 
    * @param    action
    *           The event code.
    *           
    * @return   See above.
    */
   @Override
   public boolean ignoreAction(int action)
   {
      return super.ignoreAction(action)               ||
             action == Keyboard.KA_CURSOR_DOWN        ||
             action == Keyboard.KA_CURSOR_UP          ||
             action == Keyboard.KA_CURSOR_LEFT        ||
             action == Keyboard.KA_CURSOR_RIGHT;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * Get the default event associated for this widget and the given key.
    * 
    * @param    key
    *           The current key.
    * @param    inEditingBlock
    *           <code>true</code> if EDITING mode is active.
    *           
    * @return   The action code for the widget's default event, {@link Keyboard#SE_CHOOSE}.
    */
   @Override
   public int getDefaultEvent(KeyInput key, boolean inEditingBlock)
   {
      return inEditingBlock && (key.keyChar() == ' ' || key.keyCode() == Key.VK_ENTER) 
               ? Keyboard.SE_CHOOSE
               : -1;
   }

   // AbstractWidget overrides start
   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Draw the button implementation. Called by this AbstarctButton's parent container.
    */
   @Override
   public abstract void draw();

   /**
    * Check if widget needs in cursor to be on when having the input focus.  For toggle-box widget
    * this method should always return <code>false</code> because toggle-box never needs cursor
    * visibility, even when it is required to force the cursor to be on.
    * 
    * @return  <code>false</code> unconditionally.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   // AbstractWidget overrides end
   
   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String textLabel()
   {
      // remove possible accelerator symbol from on-screen label
      return textLabel != null ? textLabel.replaceFirst("\\&", "") : textLabel;
   }

   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   public abstract void setText(String text);

   /**
    * Set button label.
    * 
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      if ((this.textLabel == null && label != null) || 
          (this.textLabel != null && !this.textLabel.equals(label)))
      {
         repaint();
      }
      
      this.textLabel = label;
   }

   // KeyListener interface implementation start
   /**
    * This method is invoked when key is pressed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
   }

   /**
    * This method is invoked when key is typed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
   }
   // KeyListener interface implementation end
   
   // FocusListener interface implementation start
   /**
    * This is called when the toggle-box takes the input focus. When the help string is defined
    * for the toggle-box it should be displayed at the window status line.
    *
    * @param   event
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // no need in code below if widget already focused
      if (!widgetFocused)
      {
         // force the cursor off since it should not display while we have focus
         curvis = screen().setCursorStatus(false);
         
         // mark widget as gained input focus
         widgetFocused = true;
      }
      
      // even for consecutive FOCUS_GAINED events, the HELP text must be set
      // for this abstract button, because the HELP message-text may have been changed
      // for another widget (i.e. on a nested UPDATE)
      
      // draw the help string for focused abstract button
      ThinClient.setHelp(config().getDynamicHelp(), this);
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      // no need in code below if widget already lost the focus
      if (!widgetFocused)
      {
         return;
      }
      
      widgetFocused = false;
      
      // restore the cursor visibility that existed when we gained focus 
      screen().setCursorStatus(curvis);
   }
   // FocusListener interface implementation end

   // Common button-like widgets public and private code
   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @SuppressWarnings("unchecked")
   protected void processKeyEventCommon(KeyInput keyEvent)
   {
      // use arrow keys to move focus between widgets (only left/right work
      // in the alert box case)
      if (parent(AlertBox.class) != null)
      {
         setActionCode(keyEvent);
      }
      else
      {
         if (handleCursorKeys(keyEvent.actionCode(), (Frame<O>) UiUtils.locateFrame(this)))
            return;
      }

      handleBasicEvents(keyEvent);
   }

   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   protected abstract void handleBasicEvents(KeyInput event);
   
   /**
    * Refresh widget at the screen.
    */
   protected void refresh()
   {
      draw();
      requestSync();
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   protected boolean handleCursorKeys(int key, Frame<O> frame)
   {
      boolean returnNow = false;
      
      if (key == Keyboard.KA_TAB)
      {
         returnNow = nextTabStop();
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         returnNow = prevTabStop();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // the left most widget ignores this condition
         if (!frame.isLeftMostWidget(this))
         {
            returnNow = prevFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         returnNow = moveAboveWidget(frame);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // the right most widget ignores this condition
         if (!frame.isRightMostWidget(this))
         {
            returnNow = nextFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         returnNow = moveBelowWidget(frame);
      }
      return returnNow;
   }

   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   protected void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();
      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);         
      }
   }
   
   /**
    * Move focus to widget below in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveBelowWidget(Frame<O> frame)
   {
      frame.moveBelowWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to widget above in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveAboveWidget(Frame<O> frame)
   {
      frame.moveAboveWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean prevFocus()
   {
      ThinClient.previousFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean nextFocus()
   {
      ThinClient.nextFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order if initiated by BACK-TAB.
    * 
    * @return  <code>true</code>
    */
   private boolean prevTabStop()
   {
      ThinClient.previousTabStop(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order if initiated by NAB.
    * 
    * @return  <code>true</code>
    */
   private boolean nextTabStop()
   {
      ThinClient.nextTabStop(this);
      refresh();
      return true;
   }
}