/*
** Module   : ActivexFrame.java
** Abstract : An implementation of the client-side widget that holds a client-side Java emulated
**            ActiveX.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit. Implemented methods and attributes: LoadControls, COM-HANDLE,
**                  LEFT, TOP, CONTROL-NAME, CONTROLS, HEIGHT, WIDTH and WIDGET-HANDLE.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.widget.Container;
import java.awt.Color;

/**
 * An implementation of the client-side widget that holds a client-side Java emulated ActiveX.
 */
public class ActivexFrame<O extends OutputManager<?>>
extends AbstractWidget<O>
implements WidgetWithConfig<ControlFrameConfig>
{
   /** Screen driver */
   private GuiDriver gd;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** ActivexFrame attributes. */
   protected ControlFrameConfig config = null;
   
   /**
    * This constructor creates an ActivexFrame.
    *
    * @param   owner
    *          The parent container, frame which will include this ActivexFrame.
    * @param   id
    *          The widget's ID.
    */
   public ActivexFrame(WidgetId id, Container<O> owner)
   {
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, ControlFrameConfig.class);
   }
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ActivexFrame(WidgetId id, ControlFrameConfig cfg)
   {
      // TODO: TBA
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    *
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ControlFrameConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // screen coordinates conversion support 
      cc = screen().coordinates();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    *
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public ControlFrameConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with the implementor changes. The
    * parameter points to a reference representing the original configuration state. This
    * reference can be used to detect what configuration fields changed and so optimize any
    * processing related to the configuration change.
    *
    * @param   beforeUpdate
    *          Config reference capturing the config state before the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(ControlFrameConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }
   
   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }
   
   /**
    * Draw widget.
    * <p>
    * This method should have been empty, abstract or delegate the draw event to contained
    * client-side Java emulated ActiveX. For debugging purpose, a crossed rectangle is drawn, at
    * this moment.
    */
   @Override
   public void draw()
   {
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
   
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
   
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () ->
      {
         gd.setColor(Color.YELLOW);
         gd.fillRect(0, 0, d.width, d.height);
         gd.setColor(Color.GREEN);
         gd.drawRect(0,           0, d.width - 1, d.height - 1);
         gd.drawLine(0,           0, d.width - 1, d.height - 1);
         gd.drawLine(d.width - 1, 0, 0,           d.height - 1);
      });
   }
   
   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public double height()
   {
      return config.heightChars;
   }
   
   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   
   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public double width()
   {
      return config.widthChars;
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is pressed.
    *
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
}
