/*
** Module   : BrowseModel.java
** Abstract : data model for Browse widget
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------------------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22527 Formatting and spelling cleanups. 
** 003 SIY 20050926   @22825 Added isQueryOffEnd() and isNewRow() methods.
** 004 SIY 20051201   @23591 Removed isQueryOffEnd() method.
** 005 SIY 20060315   @25124 Changes for complete BROWSE implementation.
** 006 GES 20060808   @28494 Use int instead of long in the APIs.
** 007 SIY 20061105   @30986 Added getCurrentRow() method.
** 008 SIY 20070504   @33419 Updated getRowCount().
** 009 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 010 SVL 20150212          Modified getRows signature.
** 011 SVL 20150411          Added isRowAvailable.
** 012 SVL 20150503          Added resetColumnsCache.
** 013 SVL 20150608          Added getEditingFillIns.
** 014 SVL 20150625          Added getRowCount().
** 015 CA  20150722          Browse columns have client-side widgets which need to have their own
**                           IDs and configurations.
** 016 SVL 20150722          Updated updateRow signature.
** 017 SVL 20160419          Added moveColumn.
** 018 SVL 20170417          Added getRows() for search by key.
** 019 SVL 20170813          Changed signature of getRows.
** 020 SVL 20170831          getEditingFillIns refactored to getEditorsConfigs.
** 021 SVL 20171024          Added resetEditorsConfigsCache.
** 022 SVL 20180209          Added runReport and setEnhancedParameter.
** 023 SVL 20180321          Improved javadoc.
** 024 HC  20180410          Renamed ReportOutputFormat to MediaType.
** 025 SVL 20180529          Added functions for managing enhanced browse configuration.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Data model for BROWSE widget.
 */
public interface BrowseModel
{
   /**
    * Get configurations of editors (fill-ins, toggle-boxes or combo-boxes) used for in-browse
    * editing.
    *
    * @return   Array of configurations of editor widgets (fill-ins, toggle-boxes or combo-boxes)
    *           which are used for in-browse editing. An array element is configuration of the
    *           editor widget for the corresponding editable column or <code>null</code> if the
    *           column is not editable. Return value is <code>null</code> if the browse is
    *           read-only.
    */
   public WidgetConfig[] getEditorsConfigs();

   /**
    * Reset cache of editors configurations.
    */
   public void resetEditorsConfigsCache();

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @return  number of iterated rows in the result set.
    */
   public int getRowCount();

   /**
    * Get number of available rows starting from the given one.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows.
    */
   public int getRowCount(int startRow);

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(int startRowIndex,
                              int count,
                              boolean forward,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb,
                              RowDisplayStrategy displayStrategy);

   /**
    * This function is used when search by key is performed in a browse. Retrieves the set of
    * rows which should be in view after search is completed. There are complex rules which
    * determine the top row of the view depending on how the <i>current</i> row is positioned
    * relatively to the current view and how the <i>found</i> row is positioned relatively to the
    * current view and to the current row.
    *
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(char searchChar,
                              int searchColumn,
                              int viewportTopRow,
                              int viewportBottomRow,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb);
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * @param   onlyNotifyRowReleased
    *          Only notify server that the row has been released. Data is not updated.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues,
                         boolean        onlyNotifyRowReleased)
   throws DataValidationException;

   /**
    * Inform server about current row in the browse.
    * 
    * @param   rowIndex
    *          0-based index of the current row.
    *          
    * @return  New current row.
    */
   public int setCurrentRow(int rowIndex);

   /**
    * Get current row in the model.
    * 
    * @return  The current row index in the model.
    */
   public int getCurrentRow();

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int rowIndex);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in the browse widget.
    *
    * @param  src
    *         0-based position of the column to be moved (column ordinal).
    * @param  dest
    *         0-based position to which the column is moved (column ordinal).
    */
   public void moveColumn(int src, int dest);

   /**
    * Run browse report.
    *
    * @param  outputFileName
    *         Name of the output file.
    * @param  outputFormat
    *         Format of the output format.
    *
    * @return <code>true</code> on success.
    */
   public boolean runReport(String outputFileName, MediaType outputFormat);

   /**
    * Get save targets for an enhanced browse configuration, which are available for the current
    * user.
    *
    * @return see above.
    */
   EnhancedBrowseSaveTarget[] getEnhancedBrowseSaveTargets();

   /**
    * Apply default 4GL configuration (non-enhanced) for this browse.
    *
    * @return <code>true</code> on success.
    */
   boolean applyDefault4glConfig();

   /**
    * Apply the given enhanced browse configuration to this browse.
    *
    * @param  config
    *         Enhanced browse configuration to apply.
    *
    * @return <code>true</code> on success.
    */
   boolean applyEnhancedConfig(EnhancedBrowseConfig config);

   /**
    * Save the current enhanced browse configuration using the specified save target.
    *
    * @param  saveTarget
    *         Save target (target server, target account, applied to all browses or not).
    *
    * @return <code>true</code> on success.
    */
   boolean saveEnhancedConfig(EnhancedBrowseSaveTarget saveTarget);
}