/*
** Module   : FontManager.java
** Abstract : client-side font management class
**
** Copyright (c) 2011-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20111115 Created initial version
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Fixed warnings related to widget generics.
** 003 CA  20140805 Fixed warnings related to widget generic's.
** 004 MAG 20140910 Added unload notification.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20141123 Allow the GUI client to be reinitialized.
** 008 CA  20141125 USE-3D-SIZE property is set by the ThinClient, on client initialization (to 
**                  allow configuration via the client's bootstrap config).
** 009 CA  20141220 Fixed constant for DEFAULT-FIXED-FONT; empty text length in GUI is always 0.
** 010 CA  20150317 Added tracking of fixed-size fonts and font's max width.
** 011 CA  20150406 Cache the text metrics on client-side, when they are resolved first time.
** 012 CA  20150509 Added a NPE protection in getFontKey.
** 013 GES 20150428 Rewrite to eliminate Swing Font class dependencies. Now it is the FontDetails
**                  instance which is always used by common code.
** 014 CA  20150519 Added implicit font size support.  Refactored to load the font and text 
**                  metrics at client startup, when the FontManager gets initialized.
** 015 CA  20150910 If the text is null, getTextWidth will return 0.
** 016 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 017 SVL 20151013 Added getBrowseCellBottomTextOffset.
** 018 CA  20151123 When dialog-frames are used, their owning window (from where the font-table is
**                  determined) is different than the drawing window.  Thus release/restore the
**                  drawing state when calling GuiDriver APIs.
** 019 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 020 SBI 20160110 Changed to install/upload binary fonts only if they are not detected.
** 021 EVL 20160415 Fix for system font index calculation.  Adding separate font for stack trace
**                  window.
** 022 VIG 20160410 Fixed ALERT-BOX font and added system MENU_FONT.
** 023 SVL 20170126 Added SORT_NUMBER_FONT.
** 024 OM  20170320 Added support for Themes.
** 025 EVL 20170512 Init procedure changed to pass the default font aliases to the server too.
**                  Changing the font initialization sequence fo implement new schema.
** 026 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 027 OM  20171027 Added setFont() method.
** 028 ECF 20180604 Faster getFontKey implementation.
** 029 CA  20180523 Fixed environment name, when used in maps: registries are case-insensitive, 
**                  stanza INI are case-sensitive.
** 030 SBI 20180710 Added registerFontResolver, removeFontResolver and registerFontDetails.
** 031 SVL 20180529 Refactoring due to support of enhanced fonts.
** 032 CA  20180815 Fixed bold style if the font definition is present with inherited bold style.
**     SBI 20180815 Changed according to GuiDriver.isFontInstalled.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.driver.ScreenDriver;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Font configuration manager, for a specific client session.
 */
public class FontManager
{
   /**
    * Reference for the undefined font.  This instructs the runtime to compute the font from the
    * parent's specification (i.e. from the frame/window/etc).
    */
   public static final int UNDEFINED_FONT = -3;

   /**
    * Reference for the default font.  0 is a valid entry in the font-table, so it can't be used.
    */
   public static final int DEFAULT_FONT = -1;

   /**
    * Reference for the default fixed font.  0 is a valid entry in the font-table, so it can't be 
    * used.
    */
   public static final int DEFAULT_FIXED_FONT = -2;

   /** The base for the system fonts. */
   public static final int SYSTEM_FONT_BASE = -1000;
   
   /** Reference to the window title font, a system font. */
   public static final int WINDOW_TITLE_FONT = SYSTEM_FONT_BASE - 0;
   
   /** Reference to the alert box font, a system font. */
   public static final int ALERT_BOX_FONT = SYSTEM_FONT_BASE - 1;
   
   /** Reference to the stack trace window font, a system font. */
   public static final int STACK_TRACE_WINDOW_FONT = SYSTEM_FONT_BASE - 2;
   
   /** Reference to the menu font, a system font. */
   public static final int MENU_FONT = SYSTEM_FONT_BASE - 3;

   /** Reference to the sort number font (on browse column labels), a system font. */
   public static final int SORT_NUMBER_FONT = SYSTEM_FONT_BASE - 4;

   /** Reference to the print font. */
   public static final int PRINT_FONT = SYSTEM_FONT_BASE - 5;
   
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(FontManager.class.getName());

   /**
    * Map "font key" -&gt; "offset of a text in a browse cell, from the cell bottom" (in pixels).
    */
   private static Map<String, Integer> browseCellBottomTextOffset = new HashMap<>();

   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         return new WorkArea();
      };
   };
   
   /**
    * Get the height of the given font.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           Font index.
    * @param    driver
    *           The reference screen driver.
    *          
    * @return   Height of the font.
    */
   public static int getFontHeight(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      String key = getFontKey(window, font, driver);
      return getFontHeight(window, key, font, driver);
   }

   /**
    * Get the height of the given font.
    *
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The font details.
    * @param    fontIndex
    *           Font index, for logging purposes. {@code null} if the font has no index.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   Height of the font.
    */
   public static int getFontHeight(TitledWindow<?> window,
                                   FontDetails font,
                                   Integer fontIndex,
                                   ScreenDriver<?> driver)
   {
      return getFontHeight(window, font.key, fontIndex, driver);
   }

   /**
    * Get the height of the given font.
    *
    * @param    window
    *           The window from which the font-table is used.
    * @param    key
    *           Font key.
    * @param    fontIndex
    *           Font index, for logging purposes. {@code null} if the font has no index.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   Height of the font.
    */
   private static int getFontHeight(TitledWindow<?> window,
                                   String key,
                                   Integer fontIndex,
                                   ScreenDriver<?> driver)
   {
      WorkArea wa = locate();

      FontDetails<?> details = wa.fontCache.get(key);
      int height = details.legacyHeight;

      if (height == -1)
      {
         // log warning that a font with no legacy height is used.
         if (fontIndex != null && LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy height.";
            LOG.log(Level.WARNING, String.format(msg, fontIndex, key));
         }

         // let the driver measure the font
         FontDetails<?> fd = wa.fontCache.get(key);

         GuiDriver gd = (GuiDriver) driver;
         gd.selectWindow(UiUtils.getWidgetIdAsInt(window));
         height = gd.getFontHeight(fd);
         gd.releaseWindow();
      }

      return height;
   }
   
   /**
    * Check if the system font from the given font table entry is a fixed-size font.
    * 
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    font
    *           The font number in the window's font table.
    * @param    driver
    *           The active screen driver.
    *           
    * @return   <code>true</code> if the associated font is fixed-size.
    */
   public static boolean isFixedSize(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return true;
      }
      
      WorkArea wa = locate();
      
      String key = getFontKey(window, font, driver);
      FontDetails<?> details = wa.fontCache.get(key);
      
      return details.fixedSize;
   }

   /**
    * Get the width of the given font.
    *
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           Font details.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   Width of the font.
    */
   public static int getFontWidth(TitledWindow<?> window,
                                  FontDetails font,
                                  ScreenDriver<?> driver)
   {
      return getFontWidth(window, font.key, null, driver);
   }

   /**
    * Get the width of the given font.
    *
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           Font index.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   Width of the font.
    */
   public static int getFontWidth(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      String key = getFontKey(window, font, driver);
      return getFontWidth(window, key, font, driver);
   }

   /**
    * Get the width of the given font.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    key
    *           Font key.
    * @param    fontIndex
    *           Font index, for logging purposes. {@code null} if the font has no index.
    * @param    driver
    *           The reference screen driver.
    *          
    * @return   Width of the font.
    */
   private static int getFontWidth(TitledWindow<?> window,
                                   String key,
                                   Integer fontIndex,
                                   ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      WorkArea wa = locate();
      
      FontDetails<?> details = wa.fontCache.get(key);
      
      int width = details.legacyWidth;
      
      if (width == -1)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy width.";
            LOG.log(Level.WARNING, String.format(msg, fontIndex, key));
         }

         // let the driver measure the font
         FontDetails<?> fd = wa.fontCache.get(key);
         
         GuiDriver gd = (GuiDriver) driver;
         gd.selectWindow(UiUtils.getWidgetIdAsInt(window));
         width = gd.getFontWidth(fd);
         gd.releaseWindow();
      }
      
      return width;
   }

   /**
    * Registers and fills legacy font metrics if it is not in the font cache.
    *
    * @param    windowId
    *           ID of the current window.
    * @param    font
    *           The new selected font
    * 
    * @return   The selected font with filled legacy font metrics or <code>null</code> if
    *           the <code>font</code> is <code>null</code>.
    */
   public static FontDetails<?> registerFontDetails(int windowId, FontDetails<?> font)
   {
      if (font == null)
      {
         return null;
      }

      WorkArea wa = locate();
      FontDetails<?> details = wa.fontCache.get(font.key);
      
      if (details == null)
      {
         // register and fill legacy font metrics
         OutputManager mgr = ThinClient.getInstance().getOutputManager();
         try
         {
            mgr.selectWindow(windowId);
            details = wa.createFont(font);
         }
         finally
         {
            mgr.releaseWindow();
         }
      }
      
      return details;
   }
   
   /**
    * Get the maximum font width for the given font.
    *
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    font
    *           The font number in the window's font table.
    * @param    driver
    *           The active screen driver.
    *
    * @return   The font's maximum width.
    */
   public static int getMaxFontWidth(TitledWindow<?> window,
                                     FontDetails font,
                                     ScreenDriver<?> driver)
   {
      return getMaxFontWidth(window, font.key, null, driver);
   }

   /**
    * Get the maximum font width for the given font.
    *
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    font
    *           The font number in the window's font table.
    * @param    driver
    *           The active screen driver.
    *
    * @return   The font's maximum width.
    */
   public static int getMaxFontWidth(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      String key = getFontKey(window, font, driver);
      return getMaxFontWidth(window, key, font, driver);
   }
   
   /**
    * Get the maximum font width for the given font.
    * 
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    key
    *           Font key.
    * @param    fontIndex
    *           Font index, for logging purposes. {@code null} if the font has no index.
    * @param    driver
    *           The active screen driver.
    * 
    * @return   The font's maximum width.
    */
   public static int getMaxFontWidth(TitledWindow<?> window,
                                     String key,
                                     Integer fontIndex,
                                     ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      WorkArea wa = locate();
      
      FontDetails<?> details = wa.fontCache.get(key);
      int width = details.legacyMaxWidth;
      
      if (width == -1)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy max width.";
            LOG.log(Level.WARNING, String.format(msg, fontIndex, key));
         }
         
         // let the driver measure the font
         GuiDriver gd = (GuiDriver) driver;
         gd.selectWindow(UiUtils.getWidgetIdAsInt(window));
         width = gd.getFontMaxWidth(details);
         gd.releaseWindow();
      }
      
      return width;
   }

   /**
    * Registers the given font resolver for the given widget.
    * 
    * @param    window
    *           The given window
    * @param    baseConfig
    *           The widget config
    * @param    fontResolver
    *           The given font resolver
    */
   public static void registerFontResolver(Window<?>       window,
                                           BaseConfig      baseConfig,
                                           GuiFontResolver fontResolver)
   {
      WorkArea wa = locate();
      
      Map<BaseConfig, List<GuiFontResolver>> fontResolversMap = wa.fontResolvers.get(window);
      
      if (fontResolversMap == null)
      {
         fontResolversMap = new HashMap<>();
         
         wa.fontResolvers.put(window, fontResolversMap);
      }
      
      List<GuiFontResolver> widgetFontResolversList = fontResolversMap.get(baseConfig);
      
      if (widgetFontResolversList == null)
      {
         widgetFontResolversList = new LinkedList<GuiFontResolver>();
         
         fontResolversMap.put(baseConfig, widgetFontResolversList);
      }
      
      widgetFontResolversList.add(fontResolver);
   }

   /**
    * Removes the font cache associated with the given window.
    * 
    * @param    window
    *           The given window
    */
   public static void removeFontResolver(Window<?> window)
   {
      WorkArea wa = locate();
      
      wa.fontResolvers.remove(window);
   }

   /**
    * Removes the registered font resolver for the given widget.
    * 
    * @param    window
    *           The given window
    * @param    widget
    *           The given widget
    */
   public static void removeFontResolver(Window<?> window, AbstractWidget<?> widget)
   {
      WorkArea wa = locate();
      
      Map<BaseConfig, List<GuiFontResolver>> widgetsFontResolvers = wa.fontResolvers.get(window);
      
      if (widgetsFontResolvers != null)
      {
         widgetsFontResolvers.remove(widget.config());
      }
   }

   /**
    * Resolve a font instance associated with the specified font entry in the window's 
    * font table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   public static FontDetails<?> resolveFont(Window<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return null;
      }
      
      WorkArea wa = locate();
      
      String key = getFontKey(window, font, driver);
      
      FontDetails<?> details = wa.fontCache.get(key);
      
      return details;
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.
    *
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The font details.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   public static int getTextWidth(String          text,
                                  Window<?>       window,
                                  FontDetails     font,
                                  ScreenDriver<?> driver)
   {
      return getTextWidth(text, window, font.key, driver);
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getTextWidth(String          text,
                                  Window<?>       window,
                                  int             font,
                                  ScreenDriver<?> driver)
   {
      String key = getFontKey(window, font, driver);
      return getTextWidth(text, window, key, driver);
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.
    *
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    key
    *           The font key.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   private static int getTextWidth(String         text,
                                  Window<?>       window,
                                  String          key,
                                  ScreenDriver<?> driver)
   {
      if (text == null)
      {
         return 0;
      }

      if (driver.isChui() || text.length() == 0)
      {
         return text.length();
      }

      int[] metrics = getTextMetrics(text, window, key, driver);

      return metrics[0];
   }

   /**
    * Get the height of the specified text, in pixels, using the specified font.  
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getTextHeight(String          text,
                                   Window<?>       window,
                                   int             font,
                                   ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      int[] metrics = getTextMetrics(text, window, font, driver);
      
      return metrics[1];
   }
   
   /**
    * Get the PIXELS-PER-COLUMN value for this client session.
    *
    * @param    driver
    *           The reference screen driver.
    * 
    * @return   See above.
    */
   public static int getPixelsPerColumn(ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         return CoordinatesConversionParams.CHUI_PIXELS_PER_COLUMN;
      }

      return wa.defaultFont.legacyWidth;
   }
   
   /**
    * Get the PIXELS-PER-ROW value for this client session.
    * 
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getPixelsPerRow(ScreenDriver<?> driver)
   {
      WorkArea wa = locate();

      if (driver.isChui())
      {
         return CoordinatesConversionParams.CHUI_PIXELS_PER_ROW;
      }

      int legacyHeight = wa.defaultFont.legacyHeight;
      
      return legacyHeight + (int) (legacyHeight / 2.0) + (wa.isThreeD ? 2 : 0);
   }

   /**
    * Compute and register the window's font-table.  This may be from the window's environment
    * (if the window is using a custom environment), or from the default environment (i.e. the
    * in-directory font-table).
    * <p>
    * The same font-table may be shared by one or more windows.
    * 
    * @param    window
    *           The window.
    * @param    driver
    *           The reference screen driver.
    */
   public static void registerFontTable(Window<?> window, ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("The font-table can be resolved only in GUI mode.");
         }

         // no-op for chui
         return;
      }

      // determine the environments's name
      String env = (wa.ed == null ? null : wa.ed.getEnvironmentName());

      EnvironmentFontTable envft = wa.envFontTables.get(env);
      if (envft == null)
      {
         // read the current font-table from the current environment, if not already built
         FontDetails<?>[] fonts = wa.server.readFontTable();
         
         envft = new EnvironmentFontTable(env, fonts);
         wa.envFontTables.put(env, envft);
      }

      // the font-table for a certain environment is common for all windows created with that
      // environment.
      wa.windowFontTables.put(window, envft);
   }
   
   /**
    * Deregister the font-table associated with this window.
    * 
    * @param    window
    *           The window.
    * @param    driver
    *           The reference screen driver.
    */
   public static void deregisterFontTable(Window<?> window, ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("The font-table can be resolved only in GUI mode.");
         }

         // no-op for chui
         return;
      }

      wa.windowFontTables.remove(window);
   }
   
   /**
    * Set the number of entries in the font-table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    numEntries
    *           The new size of the font-table.
    */
   public static void adjustFontTableSize(Window<?> window, int numEntries)
   {
      // TODO: what happens if the font-table is shrunk and there are widgets referring those
      // fonts?
      
      WorkArea wa = locate();
      EnvironmentFontTable  eft = wa.windowFontTables.get(window);
      
      while (eft.fonts.size() != numEntries)
      {
         if (numEntries < eft.fonts.size())
         {
            // shrink it
            eft.fonts.remove(eft.fonts.size() - 1);
         }
         else
         {
            // enlarge it with the default font
            eft.fonts.add(wa.defaultFont);
         }
      }
   }

   /**
    * Get the number of entries in the font-table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * 
    * @return   See above.
    */
   public static int fontTableSize(Window<?> window)
   {
      WorkArea wa = locate();
      EnvironmentFontTable  eft = wa.windowFontTables.get(window);
      
      return eft.fonts.size();
   }

   /**
    * Get a key describing the given font.
    * 
    * @param    name
    *           The font name.
    * @param    size
    *           The font point size.
    * @param    bold
    *           Flag indicating if the font style is bold.
    * @param    italic
    *           Flag indicating if the font style is italic.
    * @param    underline
    *           Flag indicating if the font style is underline.
    *           
    * @return   See above.
    */
   public static String getFontKey(String  name,
                                   int     size,
                                   boolean bold,
                                   boolean italic,
                                   boolean underline)
   {
      StringBuilder buf = new StringBuilder(name.toLowerCase());
      buf.append(',')
         .append(size)
         .append(',')
         .append(bold)
         .append(',')
         .append(italic)
         .append(',')
         .append(underline);
      
      return buf.toString();
   }

   /**
    * Resolve the font key associated with the given font, in the current window's font-table.
    *
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   public static String getFontKey(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return null;
      }
      
      FontDetails fd = getFontDetails(window, font);
      return fd != null ? fd.key : null;
   }

   /**
    * Resolve the font details associated with the given font.
    * 
    * @param    window
    *           The window from which the font-table is used, may be <code>null</code>.
    * @param    font
    *           The entry in the font-table.
    *
    * @return   See above.
    */
   public static FontDetails getFontDetails(TitledWindow<?> window, int font)
   {
      WorkArea wa = locate();
      EnvironmentFontTable ft = wa.windowFontTables.get(window);
      
      FontDetails<?> fd = null;
      
      if (font == UNDEFINED_FONT)
      {
         // the font should have been computed before this call.
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Undefined font is not allowed - the caller needs to specify a resolved " +
                       "font.");
         }
         
         font = DEFAULT_FONT;
      }
      
      if (font == DEFAULT_FONT)
      {
         fd = wa.defaultFont;
      }
      else if (font == DEFAULT_FIXED_FONT)
      {
         fd = wa.defaultFixedFont;
      }
      else if (font <= SYSTEM_FONT_BASE)
      {
         fd = wa.systemFonts[SYSTEM_FONT_BASE - font];
      }
      else if (ft == null || font < 0 || font >= ft.fonts.size())
      {
         if (LOG.isLoggable(Level.FINE))
         {
            int sz = ft.fonts.size() - 1;
            
            LOG.fine(String.format("Font %d is not in the font-table (max-size is %d).", fd, sz));
         }

         // if the font is not referenced in the font-table, then the DefaultFont is used.
         fd = wa.defaultFont;
      }
      else
      {
         fd = ft.fonts.get(font);
      }
      
      return fd;
   }

   /**
    * Unload environment. A notification from environment API whenever an 
    * <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void unload(String envName)
   {
      WorkArea wa = local.get(false);
      if (wa != null && envName != null)
      {
         wa.envFontTables.remove(envName);
      }
   }

   /**
    * Initialize the font manager's helpers. This is called every time the ThinClient is
    * initialized/re-initialized.
    * 
    * @param    server
    *           The server exports.
    */
   public static void init(ServerExports server)
   {
      // do not use locate() here
      WorkArea wa = local.get();
      
      wa.initialize(server);
   }

   /**
    * Get the state of the <code>use-3d-size</code> flag.
    * 
    * @return   See above.
    */
   public static boolean isThreeD()
   {
      WorkArea wa = locate();
      
      return wa.isThreeD;
   }
   
   /**
    * Set the state of the <code>use-3d-size</code> flag.
    * 
    * @param    isThreeD
    *           The new state of this flag.
    */
   public static void setThreeD(boolean isThreeD)
   {
      WorkArea wa = locate();
      wa.isThreeD = isThreeD;
   }

   /**
    * If the given font is undefined, default to either the {@link #DEFAULT_FONT} or 
    * {@value #DEFAULT_FIXED_FONT}, depending on whether a fixed font is required or not.
    * 
    * @param    font
    *           The font number.
    * @param    fixedSize
    *           Flag indicating if a fixed-size font is needed, if the font is 
    *           {@link #UNDEFINED_FONT undefined}.
    * 
    * @return   The received font, if an explicit font number is given, or a resolved font,
    *           otherwise.
    */
   public static int resolveFont(int font, boolean fixedSize)
   {
      boolean explicit = (font != FontManager.UNDEFINED_FONT);
      return (explicit ? font : (fixedSize ? FontManager.DEFAULT_FIXED_FONT 
                                           : FontManager.DEFAULT_FONT));
   }

   /**
    * Get offset of a text drawn with the specific font in a browse cell, from the cell bottom
    * (in pixels).
    *
    * @param  font
    *         Font used in the browse cell.
    *
    * @return see above.
    */
   public static int getBrowseCellBottomTextOffset(FontDetails font)
   {
      Integer res = browseCellBottomTextOffset.get(font.key);
      if (res != null)
      {
         return res;
      }

      switch (font.fontName)
      {
         case "Courier New":
         case "Segoe UI":
            res = (int) (font.pointSize * 0.4 + 2);
            break;
         case "Tahoma":
            res = (int) (font.pointSize * 0.3 + 2);
            break;
         default:
            res = (int) (font.pointSize * 0.4 + 2);
            LOG.warning(String.format("There is no formula for browse cell bottom text offset " +
                  "for the font '%s', using default formula.", font.fontName));
      }
      
      browseCellBottomTextOffset.put(font.key, res);
      return res;
   }
   
   /**
    * Sets a specific font to the font-table at a specific index.
    *
    * @param   window
    *          The window from which the font-table is used.
    * @param   fontRef
    *          The index of the font.
    * @param   fd
    *          The new font details to replace th old one at {@code fontRef} for the specified 
    *          window.
    */
   public static void setFont(Window<?> window, int fontRef, FontDetails<?> fd)
   {
      WorkArea wa = locate();
      EnvironmentFontTable eft = wa.windowFontTables.get(window);
      
      if (eft == null || fontRef < 0 || fontRef >= eft.fonts.size())
      {
         return;
      }
      
      eft.fonts.set(fontRef, fd);
      
      Map<BaseConfig, List<GuiFontResolver>> fontResolvers = wa.fontResolvers.get(window);
      
      if (fontResolvers != null)
      {
         for(Map.Entry<BaseConfig, List<GuiFontResolver>> entry : fontResolvers.entrySet())
         {
            boolean fontChanged = false;
            
            for (GuiFontResolver fontResolver : entry.getValue())
            {
               fontChanged = fontChanged || fontResolver.updateFontCache(fontRef, fd);
            }
            
            if (fontChanged)
            {
               Widget affectedWidget = ThinClient.getInstance().getWidget(entry.getKey().id);
               
               if (affectedWidget != null)
               {
                  affectedWidget.repaint();
               }
            }
         }
      }
   }

   /**
    * Set metrics for the text drawn with the specific font.
    *
    * @param fontKey
    *        The key for the font used to draw the text.
    * @param text
    *        The text.
    * @param width
    *        Text width.
    * @param height
    *        Text height.
    */
   public static void setTextMetrics(String fontKey, String text, int width, int height)
   {
      WorkArea wa = locate();
      String tkey = getTextKey(fontKey, text);
      int[] metrics = new int[] { width, height };
      wa.textMetrics.put(tkey, metrics);
   }
   
   /**
    * Locate the context-local data of this class.
    * 
    * @return   A {@link WorkArea} instance.
    * 
    * @throws   IllegalStateException
    *           If the {@link WorkArea} has not been {@link WorkArea#initialized}.
    */
   private static WorkArea locate()
   {
      WorkArea wa = local.get();
      if (!wa.initialized)
      {
         throw new IllegalStateException("The font manager has not been initialized!");
      }
      
      return wa;
   }

   /**
    * Get the metrics for the specified text and font from the
    * {@link WorkArea#textMetrics cache}.
    * <p>
    * If not present, ask the server-side for the metrics.
    *
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   private static int[] getTextMetrics(String          text,
                                       Window<?>       window,
                                       int             font,
                                       ScreenDriver<?> driver)
   {
      String key = getFontKey(window, font, driver);
      return getTextMetrics(text, window, key, driver);
   }

   /**
    * Get the metrics for the specified text and font from the
    * {@link WorkArea#textMetrics cache}.
    * <p>
    * If not present, ask the server-side for the metrics.
    *
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The font details.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   private static int[] getTextMetrics(String          text,
                                       Window<?>       window,
                                       FontDetails     font,
                                       ScreenDriver<?> driver)
   {
      return getTextMetrics(text, window, font.key, driver);
   }
   
   /**
    * Get the metrics for the specified text and font from the
    * {@link WorkArea#textMetrics cache}.
    * <p>
    * If not present, ask the server-side for the metrics.
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    key
    *           Font key.
    * @param    driver
    *           The reference screen driver.
    * 
    * @return   See above.
    */
   private static int[] getTextMetrics(String          text,
                                       Window<?>       window,
                                       String          key,
                                       ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      String tkey = getTextKey(key, text);
      int[] metrics = wa.textMetrics.get(tkey);
      
      if (metrics != null)
      {
         // found from cache, fast exit
         return metrics;
      }

      // check server-side
      metrics = wa.server.getLegacyTextMetrics(text, key);
      
      if (metrics == null)
      {
         // let the driver measure the string
         FontDetails<?> fd = wa.fontCache.get(key);
         
         GuiDriver gd = (GuiDriver) OutputManager.getDriver();
         
         Object state = gd.saveDrawingState();
         
         gd.selectWindow(UiUtils.getWidgetIdAsInt(window));
         
         try
         {
            int width = gd.getTextWidth(text, fd);
            int height = gd.getTextHeight(text, fd);
            
            metrics = new int[] { width, height };
         }
         finally
         {
            gd.releaseWindow();
            
            gd.restoreDrawingState(state);
         }
      }
      
      // cache it
      wa.textMetrics.put(tkey, metrics);
      
      return metrics;
   }

   /**
    * Get the key for the text drawn using the specific font.
    *
    * @param  fontKey
    *         The key for the font used to draw the text.
    * @param  text
    *         The text.
    *
    * @return combined key for the text drawn using the specific font.
    */
   private static String getTextKey(String fontKey, String text)
   {
      return fontKey + "#" + text;
   }

   /**
    * Context-local data for this font manager.
    */
   private static class WorkArea
   {
      /** A font-key to font details map, containing all the fonts loaded by this client. */
      private final Map<String, FontDetails<?>> fontCache = new HashMap<String, FontDetails<?>>();
      
      /** A map of window to its in-use environment, as when it was created. */
      private final Map<Window<?>, EnvironmentFontTable> windowFontTables = new HashMap<>();
      
      /** A map of environment names to their associated font-table. */
      private final Map<String, EnvironmentFontTable> envFontTables = new HashMap<>();
      
      /** The registered font resolvers */
      private final Map<Window<?>, Map<BaseConfig, List<GuiFontResolver>>> fontResolvers;
      
      /** The default font for this session. */
      private FontDetails<?> defaultFont;
      
      /** The default fixed font for this session. */
      private FontDetails<?> defaultFixedFont;

      /** The system fonts for this session. */
      private FontDetails<?>[] systemFonts;
      
      /** 
       * Flag identifying if this session was started in 3d.  Default value is <code>true</code>.
       */
      private boolean isThreeD = true;
      
      /** The server exports. */
      private ServerExports server;
      
      /** Access to the client's environment.  May be <code>null</code> if not on a Windows OS. */
      private EnvironmentDaemon ed;
      
      /** A cache of text metrics.  0-index is the width, 1-index is the height. */
      private final Map<String, int[]> textMetrics = new HashMap<String, int[]>();
      
      /** Flag to track if the work area has been initialized. */
      private boolean initialized = false;
      
      /**
       * Create a new work-area.
       */
      public WorkArea()
      {
         initialized = false;
         
         fontResolvers = new HashMap<>();
      }

      /**
       * Initialize this work area.
       * 
       * @param    server
       *           The server exports.
       */
      public void initialize(ServerExports server)
      {
         initialized = true;
         if (OutputManager.getDriver().isChui())
         {
            this.server           = null;
            this.ed               = null;
            this.isThreeD         = false;
            this.defaultFont      = null;
            this.defaultFixedFont = null;
            this.systemFonts      = null;
         }
         else
         {
            // clear the maps
            fontCache.clear();
            windowFontTables.clear();
            envFontTables.clear();
            textMetrics.clear();
            
            // initialize only in GUI mode
            this.server = server;
            
            // initialize the default fonts
            server.setThemePreferredFonts(ThemeManager.getCurrentTheme().getStandardFontTable(),
                                          ThemeManager.getCurrentTheme().getDefaultFontAliases());
            FontDetails<?>[] defaults = server.readDefaultFonts();
            
            // select the default window first
            GuiDriver driver = (GuiDriver) OutputManager.getDriver();
            driver.selectWindow(WidgetId.DEFAULT_WINDOW_ID.asInt());
            
            this.defaultFont      = createFont(defaults[0]);
            this.defaultFixedFont = createFont(defaults[1]);
            
            FontDetails<?>[] sysFonts = server.readSystemFontTable();
            this.systemFonts = new FontDetails<?>[sysFonts.length];
            for (int i = 0; i < sysFonts.length; i++)
            {
               this.systemFonts[i] = createFont(sysFonts[i]);
            }
            
            // pre-populate the legacy text metrics for all the fonts in the font-table
            Map<String, int[]> legacyTextMetrics = server.readLegacyTextMetrics();
            textMetrics.putAll(legacyTextMetrics);

            // the rest of the fonts are installing here
            driver.releaseWindow();
         }
      }

      /**
       * Create a new font, with the specified details.
       * <p>
       * It will access the server-side for details about the legacy metrics and/or custom 
       * definition.
       * 
       * @param    details
       *           The font specifications.
       *           
       * @return   A new {@link FontDetails} instance, with this font. 
       */
      public FontDetails<?> createFont(FontDetails<?> details)
      {
         FontDetails<?> fd  = fontCache.get(details.key);
         
         if (fd == null)
         {
            GuiDriver driver = (GuiDriver) OutputManager.getDriver();
            
            // check if the font has alias name
            boolean hasAlias = details.fontAlias != null;
            
            // font details for possible alias
            FontDetails<?> fontAlias = hasAlias ? details.fontAlias : details;
            
            // copy common font info
            FontDetails<?> detailsCopy = details.copy();

            // Clear the font definition if the target font is installed and restore its style
            // if the font has a special font definition for the bold style.
            //
            // check if the legacy font is system wide, the "system" font is considering
            // as always installed
            if (driver.isFontInstalled(details.fontName, details.boldFontDefinition))
            {
               // the legacy font is system wide,
               // if the font was loaded from jar or file the bold attribute was reset
               // to false so we need to set up it again from boldFontDefinition copy
               boolean isBold = details.fontDefinition == null ? details.style.isBold()
                                                               : (details.boldFontDefinition ||
                                                                  details.style.isBold());
               FontStyle fontStyle = FontStyle.createStyle(isBold, details.style.isItalic(),
                                                           details.style.isUnderline());

               // loading from OS means the local copy should be cleared
               detailsCopy.fontDefinition = null;
               detailsCopy.style = fontStyle;

               // ignore font name alias
               detailsCopy.fontAlias = null;
            }
            // check if the server has ability to load legacy font from jar or file system 
            else if (details.fontDefinition != null)
            {
               // the legacy font is not system wide but loaded from deploy/server/fonts
               // ignore font name alias
               detailsCopy.fontAlias = null;
            }
            // check if the alias font is system wide
            else if (hasAlias && driver.isFontInstalled(fontAlias.fontName,
                                                        fontAlias.boldFontDefinition))
            {
               // the replacement font is system wide,
               // if the font was loaded from jar or file the bold attribute was reset
               // to false so we need to set up it again from boldFontDefinition copy
               boolean isBold = fontAlias.fontDefinition == null ? fontAlias.style.isBold()
                                                                 : fontAlias.boldFontDefinition;
               FontStyle fontStyle = FontStyle.createStyle(isBold, fontAlias.style.isItalic(),
                                                           fontAlias.style.isUnderline());
               
               // loading from OS means the local copy should be cleared
               detailsCopy.fontAlias.fontDefinition = null;
               detailsCopy.fontAlias.style = fontStyle;
            }
            
            fd = driver.createFont(detailsCopy);
            
            // this is here to maximize common code
            driver.scaleFont(fd);
            
            // the driver methods for doing this require a font details instance and the
            // logic is complex enough that it doesn't make sense to duplicate it such that
            // the driver can efficiently do this itself so it is best to put this into
            // common code
            if (fd.legacyWidth < 0 || fd.legacyMaxWidth < 0 || fd.legacyHeight < 0)
            {
               fd.legacyWidth    = driver.getFontWidth(fd);
               fd.legacyMaxWidth = driver.getFontMaxWidth(fd);
               fd.legacyHeight   = driver.getFontHeight(fd);
            }
            else
            {
               fd.legacyMaxWidth = details.legacyMaxWidth;
            }
            
            // cache newly created font
            fontCache.put(fd.key, fd);
         }
         
         return fd;
      }
   }
   
   /**
    * The font-table associated with a certain environment.
    */
   private static class EnvironmentFontTable
   {
      /** The enviroment's name. */
      @SuppressWarnings("unused")
      private final String envName;
      
      /** The enviroment's font-table. */
      private final List<FontDetails<?>> fonts;
      
      /**
       * Create a new font-table, for the specified environment.
       * 
       * @param    name
       *           The environment's name.
       * @param    fontTable
       *           The fonts part of this font-table.
       */
      public EnvironmentFontTable(String name, FontDetails<?>[] fontTable)
      {
         this.envName = name;
         this.fonts = new ArrayList<>(fontTable.length);
         WorkArea wa = locate();
         
         for (FontDetails<?> aFont : fontTable)
         {
            this.fonts.add(wa.createFont(aFont));
         }
      }
   }
}
