/*
** Module   : MenuItem.java
** Abstract : MenuItem abstract client implementation
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150401 Added doLayout method and is* sub-type methods.
**                  Added focusTraversable method and event processing methods.
** 004 VIG 20150724 Just removed unnecessary doLayout method.
** 005 VIG 20150905 Removed prepTitle method and add Label class and field.
** 006 GES 20150923 Javadoc fix.
** 007 VIG 20151226 focusTraversable() is removed. 
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 VIG 20160413 added setTitle() method.
** 010 VIG 20160222 Extracted Label class and renamed to MnemonicInfo.
** 011 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 012 VIG 20160504 Changed getPreparedTitle method for null case.
** 013 VIG 20160627 MnemonicInfo.init() call is renamed to MnemonicInfo.prepare().
** 014 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 015 VIG 20160831 ThinClient.nextFocus/prevFocus is changed to parent().nextFocus/prevFocus
**                  in key processing that simplified focusing for menu-items.
** 016 CA  20170929 Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
** 017 CA  20180202 Minor performance improvement for parent() (avoid multiple calls).
** 018 HC  20180509 Improved MENU focus management. Fixed drawing of menu items in the root of a
**                  MENU-BAR. Fixed menu item subtype checks.
**     CA  20180530 Fixed menubar components sensitivity when menubar is disabled.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>MenuItem</code> client abstract implementation class. 
 */
public abstract class MenuItem<O extends OutputManager<?>>
extends AbstractWidget<O>
implements WidgetWithConfig<MenuItemConfig>,
           MenuElement<O>
{
   /** Frame attribute container. */
   protected MenuItemConfig config = null;
   
   /** Label with mnemonic info */
   private MnemonicInfo label;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuItem(WidgetId id, MenuItemConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, MenuItemConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public MenuItemConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(MenuItemConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      
      if (!Objects.equals(config().title, beforeUpdate.title) ||
          (config().title != null && label == null))
      {
         if (label == null)
            label = createLabelInfo();
         
         label.prepare(config().title);
      }
      
      if (beforeUpdate.title != null && config().title == null)
      {
         label = null;
      }
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Get MENU-ITEM prepared label.
    * 
    * @return  MnemonicInfo.
    */
   @Override
   public MnemonicInfo getPreparedLabel()
   {
      return label;
   }
   
   /**
    * Get MENU-ITEM prepared label.
    * 
    * @return  Label.
    */
   @Override
   public String getPreparedTitle()
   {
      return label == null ? config().title :
                             (label.getPrepLabel() == null ? config().title :
                                                             label.getPrepLabel());
   }
   
   /**
    * Set item title and actualize mnemonic info.
    * 
    * @param   title
    *          New item title.
    */
   public void setTitle(String title)
   {
      config().title = title;

      if (title != null)
      {
         if (label == null)
            label = createLabelInfo();

         label.prepare(config().title);
      }
      else
      {
         label = null;
      }
   }
   
   /**
    * Check sub-type attribute if it is NORMAL or not (RULE or SKIP)
    * 
    * @return   <code>true</code> if NORMAL <code>false</code> otherwise.
    */
   public boolean isNormal()
   {
      return MenuItemConfig.SBTP_NORMAL.equalsIgnoreCase(config.subtype);
   }

   /**
    * Check sub-type attribute if it is RULE or not.
    * 
    * @return   <code>true</code> if RULE <code>false</code> otherwise.
    */
   public boolean isRule()
   {
      return MenuItemConfig.SBTP_RULE.equalsIgnoreCase(config.subtype);
   }
   
   /**
    * Check sub-type attribute if it is SKIP or not.
    * 
    * @return   <code>true</code> if SKIP <code>false</code> otherwise.
    */
   public boolean isSkip()
   {
      return MenuItemConfig.SBTP_SKIP.equalsIgnoreCase(config.subtype);
   }
 
   /**
    * Check if menu-item was defined in menu definition class as disabled initially.
    * 
    * @return   <code>true</code> if menu-item was defined as disabled.
    */
   public boolean isInitDisabled()
   {
      return config().initDisabled != null && config().initDisabled.booleanValue(); 
   }
   
   /**
    * Check if this menu-item is enabled.
    */
   @Override
   public boolean isEnabled()
   {
      return super.isEnabled() && (!Menu.isMenubarElement(this) || parent().isEnabled());
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      int actionCode = event.actionCode();
      
      // menu-item choosing: cursor-down and cursor-up do not raise value-changed/choose events, 
      // they only switch focus!
      if (key == Key.VK_ENTER)
      {
         selectMenuItem();

         return;
      }

      Container<O> parent = parent();
      boolean menuBar = Menu.isMenubarElement(this);
      Menu menu = getParentMenu();
      
      // next focus from menu-item
      if ((!menuBar && actionCode == Keyboard.KA_CURSOR_DOWN) ||
          (menuBar && actionCode == Keyboard.KA_CURSOR_RIGHT))
      {
         parent.nextFocus();
         parent.repaint();
         return;
      }

      // previous focus from menu-item
      if ((!menuBar && actionCode == Keyboard.KA_CURSOR_UP) ||
          (menuBar && actionCode == Keyboard.KA_CURSOR_LEFT))
      {
         parent.prevFocus();
         parent.repaint();
         return;
      }

      if (menu.config().popupOnly && actionCode == Keyboard.KA_CURSOR_LEFT)
      {
         SubMenu sub = getParentSubMenu();
         if (sub != null)
         {
            sub.hideBody(false);
         }
         else
         {
            menu.leaveMenu();
         }
         return;
      }

      // hiding sub-menu from menu-item: only current if LEFT or all if RIGHT 
      if (!menuBar && (actionCode == Keyboard.KA_CURSOR_RIGHT || actionCode == Keyboard.KA_CURSOR_LEFT))
      {
         SubMenu<O> parentSub = getParentSubMenu();
         SubMenu<O> rootSub = getRootSubMenu();

         if (actionCode == Keyboard.KA_CURSOR_RIGHT)
         {
            ThinClient.nextFocus(rootSub);
         }
         else
         {
            if (parentSub == rootSub)
            {
               ThinClient.previousFocus(parentSub);
            }
            else
            {
               parentSub.hideBody(false);
            }
         }

         return;
      }
         
      // mnemonic
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED || 
          event.isRealKey()                          ||
          actionCode == Keyboard.SE_VALUE_CHANGED    ||
          actionCode == Keyboard.SE_CHOOSE)
      {
         getParentMenuContainer().processKeyEvent(event);
      }
      else
      {
         super.processKeyEvent(event);
      }
   }

   /**
    * Creates new MnemonicInfo instance.
    * 
    * @return   MnemonicInfo or its descendant.
    */
   protected MnemonicInfo createLabelInfo()
   {
      MnemonicInfo minfo = new MnemonicInfo();
      minfo.init(this.config);
      return minfo;
   }
   
   /**
    * Process an event which trigger selection of this menu-item, depending if the item is a 
    * toggle-box or not.
    */
   protected void selectMenuItem()
   {
      MenuItemConfig cfg = config();
      if (cfg.toggleBox)
      {
         cfg.checked = !cfg.checked;
         ConfigHelper.setModified(cfg, true);
         EventManager.postEvent(new KeyInput(this, 
                                             EventType.ACTION_PERFORMED, 
                                             Keyboard.SE_VALUE_CHANGED));
      }
      else
      {
         EventManager.postEvent(new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_CHOOSE));
      }

      Menu m = Menu.getMenu(this);
      m.leaveMenu();
   }
}