/*
** Module   : Point.java
** Abstract : Screen point primitive.
**
** Copyright (c) 2010-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100629          Created initial version
** 002 SIY 20100804          Changes in imports, some refactorings.
** 003 HC  20140911          GUI coordinate support - data types refactoring.
** 004 HC  20141021          Implemented window move/position support.
** 005 HC  20150913          Added human-friendly toString implementation.
** 006 HC  20180813          Implemented Comparable iface.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.Coordinate;

/**
 * Point primitive.
 * The coordinate values are expressed in character units.
 */
public class Point
implements Comparable<Point>
{
   /** Point X coordinate. */
   public double x;

   /** Point Y coordinate. */
   public double y;

   /**
    * Constructor.
    * 
    * @param   x
    *          Point X.
    * @param   y
    *          Point Y.
    */
   public Point(double x, double y)
   {
      this.x = Coordinate.scale(x);
      this.y = Coordinate.scale(y);
   }

   /**
    * Copy constructor.
    * 
    * @param   point
    *          Source point.
    */
   public Point(Point point)
   {
      x = point.x;
      y = point.y;
   }

   /**
    * Shift point to specified offset.
    * 
    * @param   insets
    *          Source insets.
    *          
    * @return  new shifted point. 
    */
   public Point offset(Insets insets)
   {
      return offset(insets.left, insets.top);
   }

   /**
    * Shift point to specified offset.
    * 
    * @param   x
    *          X offset.
    * @param   y
    *          Y offset.
    * 
    * @return  new shifted point. 
    */
   public Point offset(double x, double y)
   {
      return new Point(this.x + x, this.y + y);
   }

   /**
    * Translate point object by specified offset without creating new instance.
    *  
    * @param   insets
    *          Source insets.
    *          
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public Point translate(Insets insets)
   {
      return translate(insets.left, insets.top);
   }

   /**
    * Translate point object by specified offset without creating new instance.
    * 
    * @param   dx
    *          Delta X. 
    * @param   dy 
    *          Delta Y.
    * 
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public Point translate(double dx, double dy)
   {
      x = Coordinate.scale(x + dx);
      y = Coordinate.scale(y + dy);
      return this;
   }

   /**
    * Translate point object by specified offset without creating new instance.
    *  
    * @param   point
    *          Source point.
    *          
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public Point translate(Point point)
   {
      return translate(point.x, point.y);
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return  See above.
    */
   @Override
   public String toString()
   {
      return "Point[" + x + "," + y + "]";  
   }

   /**
    * Returns a hash code value for the object.
    * 
    * @return  See above.
    */
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      long temp;
      temp = Double.doubleToLongBits(x);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      temp = Double.doubleToLongBits(y);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      return result;
   }

   /**
    * Indicates whether some other object is "equal to" this one.
    * 
    * @param   obj
    *          The other reference to compare to.
    *          
    * @return  Returns true if this and the passed-in object are equal.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (getClass() != obj.getClass())
         return false;
      Point other = (Point) obj;
      if (Double.doubleToLongBits(x) != Double.doubleToLongBits(other.x))
         return false;
      if (Double.doubleToLongBits(y) != Double.doubleToLongBits(other.y))
         return false;
      return true;
   }

   /**
    * Compares this object with the specified object for order.  Returns a
    * negative integer, zero, or a positive integer as this object is less
    * than, equal to, or greater than the specified object.
    *
    * @param    o
    *           The object to be compared.
    *
    * @return   a negative integer, zero, or a positive integer as this object
    *           is less than, equal to, or greater than the specified object.
    *
    * @throws   NullPointerException
    *           if the specified object is null
    */
   @Override
   public int compareTo(Point o)
   {
      if (o == null)
      {
         throw new NullPointerException();
      }

      int res = Double.compare(x, o.x);
      if (res != 0)
      {
         return res;
      }

      return Double.compare(y, o.y);
   }
}