/*
** Module   : ScrollBar.java
** Abstract : generic scroll bar. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141015 Notify scroll listeners only if position is in range.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150509 Allow the scrollbar to use double values for step/scrolling.
** 005 EVL 20150512 Need to introduce one more parameter to correctly calculate thumb size in
**                  GUI mode.
** 006 HC  20150831 Scrolling support overhaul.
** 007 SVL 20151019 Added ability to get non-trimmed events.
** 008 SVL 20151101 Added support for scroll bar controller.
** 009 VIG 20160307 Added mouse event to onButtonClick(), changed visibility of 
**                  notifyScrollListeners(), added ScrollStep enum, added stype type
**                  parameter to onButtonClick(), getBaseStep(), positionUpdated().
** 010 OM  20170426 Remove posting UI update on no-ops.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A generic scroll bar. 
 */
public abstract class ScrollBar<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<BaseConfig>,
           ScrollPublisher,
           ScrollListener
           
{   
   /**
    * Container bars. Containers having scroll bars could have:
    * 
    * NONE       - No scroll bars.
    * HORIZONTAL - Only horizontal scroll bar.
    * VERTICAL   - Only vertical scroll bar.
    * BOTH       - Both horizontal and vertical scroll bars.
    */
   public enum Bars
   {
      NONE,
      HORIZONTAL,
      VERTICAL,
      BOTH
   }
   
   /**
    * Scroll bar orientation enumeration. Scroll bars are of two kinds horizontal or vertical. 
    */
   public enum Orientation
   {
      HORIZONTAL,
      VERTICAL
   }
   
   /**
    * Type of scroll step. 
    */
   public static enum ScrollStep
   {
      MIN,
      SHORT,
      LONG,
      MAX,
      CUSTOM;
   }

   /**
    * Scroll bar buttons position. For a vertical scroll bar buttons positions are as follow:
    * LEFT  button is at the TOP. 
    * RIGHT button is at the BOTTOM.
    */
   public enum Position
   {
      LEFT,
      RIGHT
   }
   
   /** Scroll bar orientation */
   protected Orientation orientation;

   /** Scroll maximum value */
   protected int max = 0;
            
   /** Current position */
   protected int position = 0;
   
   /** Short scroll step; this is used when scroll buttons are clicked. */
   protected int shortStep = 1;

   /** 
    * Long scroll step; this is used when the space between scroll buttons 
    * and scroll thumb is clicked or for mouse wheel. 
    */
   protected int longStep = 4;

   /** Storage for listeners. */
   protected ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /** Widget configuration */
   protected BaseConfig config;

   /** Custom scroll bar controller. */
   protected ScrollBarController controller;

   /**
    * Default constructor. Build a horizontal scroll bar.
    */
   public ScrollBar(WidgetId id)      
   {
      this(id, Orientation.HORIZONTAL);
   }

   /**
    * Constructor.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    */
   public ScrollBar(WidgetId id, Orientation orientation)
   {
      super();
      this.orientation = orientation;
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
   }

   /**
    * Initialize scroll bar parameters.
    */
   public void clear()
   {
      max = 0;
      position = 0;
   }
   
   /**
    * Return scroll bar range.
    */
   public double range()
   {
      return max;
   }

   /**
    * Add scroll listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   public void addScrollListener(ScrollListener listener)
   {
      scrollListeners.add(listener);
   }

   /**
    * Clear the set of scroll listeners.
    */
   public void clearScrollListeners()
   {
      scrollListeners.clear();
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Return maximum value.
    * 
    * @return  see above.
    */
   public int getMax()
   {
      return max;
   }

   /**
    * Set maximum value.
    * 
    * @param   max
    *          New maximum value.
    */
   public void setMax(int max)
   {
      this.max = max;
   }

   /**
    * Get current position.
    * 
    * @return  see above.
    */
   public int getPosition()
   {
      return position;
   }

   /**
    * Set current position. Position MUST be in range.
    * 
    * @param   position
    *          New position.
    */
   public void setPosition(int position)
   {
      // do not adjust the position according the min/max range, 
      // this allows to first set position and only later adjust 
      // min/max values
      this.position = position;
   }

   /**
    * Get scroll bar orientation.
    * 
    * @return  see above.
    */
   public Orientation orientation()
   {
      return orientation;
   }
   
   /**
    * This method is invoked when scroll pane is scrolled.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void handleScrollEvent(ScrollEvent event)
   {
      boolean dirty = false;
      // update scroll bars if necessary
      if (orientation == Orientation.HORIZONTAL)
      {
         Integer newX = event.newX();
         if (newX != null && this.position != newX)
         {
            setPosition(newX);
            dirty = true;
         }
      }
      else // orientation == Orientation.VERTICAL
      {
         Integer newY = event.newY();
         if (newY != null && this.position != newY)
         {
            setPosition(newY);
            dirty = true;
         }
      }
      
      if (dirty)
      {
         repaint();
      }
   }
   
   /**
    * Notification on scroll bar button click.
    *  
    * @param   button
    *          Clicked button.
    * @param   scrollStep
    *          Scroll step type.
    * @param   e
    *          The mouse event. <code>null</code> means KeyEvent
    */
   public void onButtonClick(Position button, ScrollStep scrollStep, MouseEvt e)
   {    
      int pos = position;
      
      int step = 0;
      if (ScrollStep.SHORT.equals(scrollStep))
      {
         step = shortStep;
      }
      if (ScrollStep.LONG.equals(scrollStep))
      {
         step = longStep;
      }
      
      if (button == Position.LEFT)
      {
         pos -= step;
      }
      else
      {
         pos += step;         
      }
      
      if (pos > max || ScrollStep.MAX.equals(scrollStep))
      {
         pos = max;
      }

      if (pos < 0 || ScrollStep.MIN.equals(scrollStep))
      {
         pos = 0;
      }

      if (isControlled())
      {
         // WAS: non-trimmed into [0, max] values.
         // TODO: Why it was not-trimmed?
         controller.positionUpdated(pos, scrollStep, false);
      }

      notifyScrollListeners(pos);
   }
      
   /**
    * Notification on scroll bar drag
    *  
    * @param   pos
    *          Drag scroll position.
    */
   public void onDrag(int pos)
   {    
      notifyScrollListeners(pos);
   }
   
   /**
    * Widget drawing method.
    */
   @Override
   public abstract void draw();

   /**
    * Set the step for this scroll-bar.
    * 
    * @param    step
    *           The step.
    */
   public void setStep(int step)
   {
      shortStep = step;
      longStep  = step * 4;
   }
   
   /**
    * Get the base step of this scroll-bar.
    * 
    * @return   The step value.
    */
   public int getBaseStep()
   {
      return shortStep;
   }

   /**
    * Set custom scroll bar controller.
    *
    * @param controller
    *        Custom scroll bar controller.
    */
   public void setController(ScrollBarController controller)
   {
      this.controller = controller;
   }

   /**
    * Returns <code>true</code> if this scrollbar has custom controller assigned.
    *
    * @return see above.
    */
   public boolean isControlled()
   {
      return controller != null;
   }

   /**
    * Notify controller that position has been updated.
    *
    * @param position
    *        New position of the scroll bar.
    * @param scrollStep 
    * @param drag
    *        <code>true</code> if position was changed because of thumb dragging.
    *        <code>false</code> if position was changed because a scroll button was pressed.
    */
   public void positionUpdated(int position, ScrollStep scrollStep, boolean drag)
   {
      if (isControlled())
      {
         controller.positionUpdated(position, scrollStep, drag);
      }
   }
   
   /**
    * Adjust the following values:
    * <ul>
    *    <li>visibility</li>
    *    <li>enable state</li>
    *    <li>position</li>
    *    <li>max</li>
    *    <li>step</li>
    * </ul>
    * according to the values returned by controller assigned to this scroll bar.
    */
   public void adjustControlledScrollbar()
   {
      setVisible(controller.isVisible());
      setEnabled(controller.isEnabled());
      setMax(controller.getMax());
      setPosition(controller.getPosition());
      setStep(controller.getStep());
   }
   
   /**
    * Notify scroll listeners about event.
    * 
    * @param   newPosition
    *          New scroll bar position.
    */
   public void notifyScrollListeners(int newPosition)
   {
      // notify listeners only if is on range
      if (newPosition >= 0 && newPosition <= max)
      {
         ScrollEvent event = createScrollEvent(newPosition);
         for (ScrollListener listener : scrollListeners)
         {
            listener.handleScrollEvent(event);
         }
      }            
   }

   /**
    * Create UI specific scroll event.
    * 
    * @param   newPosition
    *          New scroll position.
    */
   protected ScrollEvent createScrollEvent(int newPosition)
   {
      if (orientation == Orientation.HORIZONTAL)
      {
         return new ScrollEvent(this, newPosition, null);
      }
      else
      {
         return new ScrollEvent(this, null, newPosition);
      }
   }
}
