/*
** Module   : SigintWaiter.java
** Abstract : External INT signal processing.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140120 Initial version.
** 002 EVL 20180320 Need to disable call to signal watcher in Windows mode to avoid second copy
**                  of interruption event.  The first one is coming from TypeAhead reader. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.util.*;

public class SigintWaiter
extends Thread
{
   /** The consumer for external interrupt. */
   private Watcher watcher = null;
   
   /** For safely terminate this thread. */
   private volatile boolean running = true;
   
   /**
    * The native method to wait for incoming SIGINT event.
    * 
    * @return   0 in case of success -1 in case of failure.
    */
   private native int waitForSigint();

   /**
    * The native method to resume SIGING interception in library.
    */
   private native void unlockSigint();
   
   /**
    * The native method to unblock SIGING waiting library.
    */
   private native void stopWaitingSigint();
   
   /**
    * The constructor that accepts watcher instance as signal consumer.
    * 
    * @param    watcher
    *           Target of the SIGINT interest.
    */
   public SigintWaiter(Watcher watcher)
   {
      super();
      this.watcher = watcher;
      setDaemon(true);
      start();
   }
   
   /**
    * The main worker method for this thread.
    */
   public void run()
   {
      while (running)
      {
         // Check the native layer for incoming SIGINT or terminate if shutdown was called
         if (waitForSigint() != -1 && running)
         {
            // Got signal, temporary ignore further SIGINT until unlockSigint()
            if (!PlatformHelper.isUnderWindowsFamily())
            {
               watcher.watchCtrlC();
            }
            // Resume signal waiting in native code
            unlockSigint();
         }
      }
   }
   
   /**
    * The method to provide safe cleanup and exit.
    */
   public void shutdown()
   {
      running = false;
      stopWaitingSigint();
   }
}