/* 
** Module   : StreamableManager.java
** Abstract : Defines APIs which can be used in redirected mode, common for both ChUI and GUI.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20171020 Initial version.
** 002 OM  20171029 Small fix in javadoc.
** 003 EVL 20171025 Javadoc fix.
** 004 EVL 20180314 Fix for no key assignment to be -1.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Defines APIs which can be used in redirected mode, common for both ChUI and GUI.
 */
public interface StreamableManager
{
   /**
    * Create a new {@link OutputManager} instance which can be used in redirected mode.
    * 
    * @return   See above.
    */
   public static OutputManager<?> newInstance()
   {
      DriverPrimitives noOpPrimitives = new DriverPrimitives()
      {
         @Override
         public void sync()
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
         
         @Override
         public int screenWidth()
         {
            return 512;
         }
         
         @Override
         public int screenHeight()
         {
            return 128;
         }
         
         @Override
         public int getY()
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
         
         @Override
         public int getX()
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
         
         @Override
         public boolean cursorStay(int x, int y)
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
         
         @Override
         public boolean cursorAt(int x, int y)
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
         
         @Override
         public void append(String str, Color color)
         {
            throw new IllegalStateException("This API can't be used in GUI!");
         }
      };
      
      AbstractChuiDriver chuiDriver = new AbstractChuiDriver(noOpPrimitives, null)
      {
         @Override
         public void clear() { } // no-op
         
         @Override
         public void beep() { } // no-op
         
         @Override
         public int readKey() { return -1; } // no-op

         @Override
         public void resetMode() { } // no-op

         @Override
         public void suspend() { } // no-op

         @Override
         public void resume() { } // no-op

         @Override
         public ChildProcessFactory getChildProcessFactory()
         {
            AbstractChuiDriver thisDriver = this;
            return new ChildProcessFactory()
            {
               public InteractiveChildProcess getChildProcess()
               {
                  return new NativeChildProcess(thisDriver);
               }
            };
         }
      };
      chuiDriver.init();

      OutputManager<?> output = new ChuiOutputManager(chuiDriver)
      {
         {
            init();
         }
         
         @Override
         public void init()
         {
            output        = chuiDriver.getPrimitives();
            defaultOutput = output;
            directOutput  = output;
         }
         
         @Override
         protected TitledWindow<?> overrideWindow(TopLevelWindow<?> window)
         {
            return window;
         }
      };
      
      return output;
   }
   
   /**
    * Get the {@link Window} instance which can be used in redirected mode.
    * 
    * @return   See above.
    */
   public Window<?> getStreamableWindow();

   /**
    * Append string with specified color at current cursor location and advance the cursor.
    *  
    * @param   message
    *          Character to append.
    * @param   color
    *          Color and attribute.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager<?> append(String message, Color color);

   /**
    * Determine the screen bitmap associated with the default window.
    * 
    * @return   The {@link ScreenBitmap} associated with this window.
    */
   public ScreenBitmap screenBitmap();
}
