/*
** Module   : ToggleBox.java
** Abstract : toggle box implementation
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140610 Initial implementation.
** 002 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 HC  20141021 Implemented window move/position support.
** 006 HC  20150323 Removed business logic from config classes.
** 007 VIG 20150506 Extracted check method.
** 008 VIG 20151005 Added isToggleKey method, defining keys, able to change the value.
**                  Added VALUE-CHANGED suppression for low-level mouse events.
** 009 EVL 20151020 Fix for NPE in check() method for evList dereferencing.
** 010 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 011 CA  20171127 Removed an unnecessary repaint.
** 012 CA  20180502 Fixed initialize, where disable (and other) logic needs to be called after 
**                  config field is assigned.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a "toggle box" with a text label.
 */
public abstract class ToggleBox<O extends OutputManager<?>>
extends AbstractButton<O,ToggleBoxConfig>
implements DataContainer,
           MnemonicTrait<O>
{
   /** Mnemonic information */
   private MnemonicInfo mnemonic = null;

   /**
    * This constructor creates a toggle-box with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle-box text.
    */
   public ToggleBox(WidgetId id, Container<O> owner, String text)
   {
      disable(); // disabled by default
      setTextLabel(text);
      
      config = ConfigManager.getInstance().resolveWidgetConfig(id, ToggleBoxConfig.class);
      
      config.name = text; // direct access
      config.widthChars = width(); // direct access
      config.help = null;
      addFocusListener(this);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ToggleBox(WidgetId id, ToggleBoxConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ToggleBoxConfig cfg)
   {
      // this constant is defined during testing.
      config = (ToggleBoxConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);

      disable(); // disabled by default
      setTextLabel("");

      if (cfg.label != null)
         setText(cfg.label);
      
      addFocusListener(this);
   }

   /**
    * The method is called after the configuration associated with
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    *
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(ToggleBoxConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   // DataContainer methods implementations start
   /**
    * Get label text.
    * 
    * @return  label text.
    */
   @Override
   public String getText()
   {
      return textLabel();
   }

   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      return new logical(config.checked);
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.modified;
   }

   /**
    * Sets the value of the toggle-box (the checked state) based on a
    * BaseDataType value.
    * 
    * @param    value
    *           The BaseDataType value of the logical value for toggle-box.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (value instanceof logical)
      {
         config.checked = ((logical)value).booleanValue();
      }
   }
   // DataContainer methods implementations end
   
   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public double width()
   {
      // safe, comparison of scaled doubles
      return config.widthChars > 0 ? config.widthChars : textLabel().length() + 3;
   }
   
   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public double height()
   {
      // safe, comparison of scaled doubles
      return config.heightChars > 0 ? config.heightChars : 1;
   }
   
   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      super.processKeyEventCommon(keyEvent);
   }

   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   @Override
   public void setText(String text)
   {
      setTextLabel(text);

      if (config.autoResize)
      {
         double w = textLabel().length() + 3;
         boolean repaint = Double.compare(w, config.widthChars) != 0;
         if (repaint)
         {
            repaint();
         }
         
         config.widthChars = w; // direct access

         if (repaint)
         {
            repaint();
         }
      }
   }

   /**
    * MnemonicInfo instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic is advised.
    *
    * @return   See above.
    */
   public MnemonicInfo getMnemonic()
   {
      return mnemonic;
   }

   /**
    * MnemonicInfo instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic is advised.
    */
   public void setMnemonic(MnemonicInfo mnemonic)
   {
      this.mnemonic = mnemonic;
   }

   /**
    * Set button label.
    *
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      super.setTextLabel(label);
      updateMnemonic(null, label, false);
   }

   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   protected void handleBasicEvents(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      if (key == '\t')
      {
         parent().nextFocus();
         return;
      }
      else if (key == Key.VK_BACK_TAB)
      {
         parent().prevFocus();
         return;
      }

      if (super.isEnabled() && isToggleKey(key))
      {
         check(null);
         // eat the original event
         event.consume();
      }
   }

   /**
    * Inverse checked state and post VALUE-CHANGED event.
    * 
    * @param  e
    *         Mouse event or null, that means, that the source event is a key press.
    */
   protected void check(MouseEvent e)
   {
      config.checked = !config.checked;
      ConfigHelper.setModified(config, true);

      refresh();
      
      EventList evList = ThinClient.getInstance().getCurrentEventList();
      // need to check to prevent NPE
      if (evList != null)
      {
         evList = evList.getNonExitEventsForWidget(this.getId().asInt());
      }

      boolean suppressValueChanged = false;
      
      // Check trigger existence for low level mouse events (portable or three-button):
      // except *-DBLCLICK. If a trigger is found - suppress VALUE-CHANGED event. 
      // e is null when check is called by key press.
      if (e != null && evList != null)
      {
         int frameId = UiUtils.locateFrame(this).getId().asInt();

         // The codes of events, that suppress VALUE-CHANGED event 
         int[] mouseEvCodes = new int[] {Keyboard.keyCode(Keyboard.MOUSE_SELECT_CLICK),
                                         Keyboard.keyCode(Keyboard.MOUSE_SELECT_UP),
                                         Keyboard.keyCode(Keyboard.MOUSE_SELECT_DOWN)};
         
         for (int eventCode : mouseEvCodes)
         {
            TriggerMatch result = new TriggerMatch(eventCode);
            evList.lookup(frameId, getId().asInt(), true, false, result);
            if (result.triggerId != -1)
            {
               suppressValueChanged = true;
            }
         }
      }
      
      // Suppress VALUE-CHANGED event if needed
      if (suppressValueChanged || (e != null && e.getClickCount() == 2))
      {
         return;
      }
      
      EventManager.postEvent(new KeyInput(this, EventType.KEY_TYPED, Keyboard.SE_VALUE_CHANGED));
   }

   /**
    * Return <code>true</code> if specified key can change toggle-box value.
    * 
    * @param   key
    *          Key code being processed.
    * 
    * @return  <code>true</code> if specified key is a toggle key. 
    */
   protected boolean isToggleKey(int key)
   {
      return key == Key.VK_ENTER || key == Key.VK_SPACE;
   }
}