/*
** Module   : WidgetEntry.java
** Abstract : container to store a mapping between an int and a value
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20060802   @28419 First version which supports a primitive
**                           widget ID to object mapping.
** 002 SIY 20100818          Added @Override annotations.
** 003 SIY 20100822          Suppressed warnings.
** 004 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.util.*;

/**
 * Simple container for a widget map key and value pair.
 */
public class WidgetEntry
implements Externalizable
{
   /** The map entry's key. */
   public int key = -1;
   
   /** The map entry's value. */
   public Object value = null;

   /**
    * Add all elements of the given array of objects to the given map, using 
    * the base ID (and the index position in the array) to properly 
    * reconstitute the each associated widget ID (which is the key in the
    * map).
    *
    * @param    array
    *           The array of objects to add.
    * @param    target
    *           The map to fill.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   public static void arrayToMap(WidgetEntry[] array, Map target)
   {
      if (array == null)
         return;
      
      for (int i = 0; i < array.length; i++)
      {
         target.put(new Integer(array[i].key), array[i].value);
      }
   }
   
   /**
    * Add all elements of the given map to a new array of objects that 
    * properly encode both the key and the value.
    *
    * @param    source
    *           The map to convert to an array.
    *           
    * @return   Array of {@link WidgetEntry} instances suitable for 
    *           externalization. 
    */
   @SuppressWarnings({ "rawtypes", "null" })
   public static WidgetEntry[] mapToArray(Map source)
   {
      int sz = (source == null) ? 0 : source.size();
      
      WidgetEntry[] we = new WidgetEntry[sz];
      
      if (sz != 0)
      {
         int idx = 0;
         
         Iterator iter = source.entrySet().iterator();
         
         while (iter.hasNext())
         {
            Map.Entry   me  = (Map.Entry) iter.next();
            Integer     key = (Integer) me.getKey();
            
            we[idx]       = new WidgetEntry();
            we[idx].key   = key.intValue();
            we[idx].value = me.getValue();
            idx++;
         }
      }
      
      return we;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      key   = in.readInt();
      value = in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(key);
      out.writeObject(value);
   }   
}