/*
** Module   : ConsoleHelper.java
** Abstract : The class provides JNI calls to text console based C code. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130723 Created initial version.
** 002 EVL 20140117 The native library call is performing now in dedicated place and only once.
**                  See the ClientCore class.
** 003 HC  20140119 Implemented runtime support for the COLOR 
**                  protermcap-attribute.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.console;

/**
 * The ConsoleHelper class provides the interface to the "ncurses" library functions or other.
 * implementation of the character based terminal functions.
 * <p>
 * The default colors are white foreground and black background or other currently used by OS
 * console terminal re-definitions.
 */
public class ConsoleHelper 
{
   private static ConsoleHelper instance;
   
   /**
    * This flag is true if the system property "console.color" has been set.
    */
   public static final boolean isColorEnabled =
      (System.getProperty("console.color") != null);

   /**
    * Default empty constructor.
    */
   private ConsoleHelper()
   {
   }

   /**
    * Initialize the native layer.
    */
   public static void initChui()
   {
      initConsole(); // call native function to initialize console.
   }

   /**
    * Access the instance of the console helper.
    *
    * @return   The static console helper instance.
    */
   public synchronized static ConsoleHelper getDefaultConsoleHelper()
   {
      if (instance == null)
      {
         instance = new ConsoleHelper();
      }

      return instance;
   }
   
   /**
    * This method is used for getting the line-drawing special characters from underlying native
    * layer.
    *
    * @param index 
    *        The index of the character to return.
    *
    * @return   The character at the given index from internal array.
    */
   public static native int getACSchar(int index);

   /**
    * This method is used for getting the attributes from underlying native layer.
    *
    * @param index 
    *        The index of the attribute to return.
    *
    * @return   The attribute at the given index from internal array.
    */
   public static native int getAttribute(int index);

   /**
    * This method performs console initialization on the native side.
    */
   private static native void initConsole();

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The specified array safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param    chars
    *           The array of characters to output. The array is supposed 
    *           to provide an item per every screen position.
    *           This item is either -1 if this position is masked or a code
    *           which goes directly to ncurses.
    * @param    attrs
    *           The array of attributes to output. The array is supposed 
    *           to provide an item per every screen position.
    *           The array must be of the same length as the chars array.
    * @param    colors
    *           The array of colors to output. The array is supposed 
    *           to provide an item per every screen position.
    *           The item value of -1 denotes a default terminal color.
    *           The array must be of the same length as the chars array.           
    * @param    cursorX
    *           final cursor position, column
    * @param    cursorY
    *           final cursor position, row
    */
   public native void addArrayNative(int chars[], int attrs[], int colors[], int cursorX, int cursorY);

   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The array of characters to output.
    * @param    len
    *           The lenght of the string to draw in characters.
    * @param    attrib
    *           The attribute to be used while drawing.
    * @param    colorpair
    *           The color value to be used while drawing.
    */
   public native void addStringNative(byte[] str, int len, int attrib, int colorpair);

   /**
    * Beeps the tone on the terminal.
    */
   public native void beep();

   /**
    * Clears the current console screen.
    */
   public native void clear();

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   public native int getScreenColumnsNative();

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   public native int getScreenRowsNative();
   
   /**
    * Check if the terminal is able to display colors or not.
    *
    * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
    */
   public native boolean hasColors();

   /**
    * Set absolute cursor position
    * 
    * @param    x
    *           New X coordinate of the cursor.
    * @param    y
    *           New Y coordinate of the cursor.
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   public native boolean setCursorNative(int x, int y);

   /**
    * Synchronizes the state of the console buffers with native terminal window. Meaning the kind
    * of refresh calls depending on the backend used.
    */
   public native void syncNative();
   
   /**
    * Sets the terminal type used by console backend.
    *
    * @param    name
    *           The new terminal type (must not be <code>null</code>).
    */
   public static native void setTermType(String name);
   
   /**
    * Gets the terminal type reported by console native backend.
    *
    * @return   The terminal type or an empty string on any failure.
    */
   public static native String getTermType();
   
   /**
    * Set the cursor visibility on or off. Note that can invisible cursor
    * can still be moved (with {@link #setCursorNative} and used normally.  This
    * method only affects the visibility of the cursor.
    *
    * @param    on
    *           When <code>true</code> the cursor will become visible, when
    *           <code>false</code> the cursor will become invisible.
    *
    * @return   The previous state of the cursor (<code>true</code> means visible).
    */
   public native boolean setCursorStatus(boolean on)
   throws IllegalStateException;

   /**
    * This method reads the key from keyboard by the special reading thread.
    *
    * @return   The integer code of the key has been read or -1 if there is no character to read.
    */
   public native int readKey();

   /**
    * Get X coordinate of the cursor.
    *
    * @return   The integer value of the X coordinate.
    */
   public native int getx();

   /**
    * Get Y coordinate of the cursor.
    *
    * @return   The integer value of the Y coordinate.
    */
   public native int gety();

   /**
    * Re-activates the console mode after ending of the previous suspension state.
    */
   public native void resume();

   /**
    * Save the current state of the console mode and switch to TTY mode temporarily.
    */
   public native void suspend();

   /**
    * Reset terminal settings.  May be called only if suspend() was previously called. 
    */
   public native void resetMode();
}