/*
** Module   : Chunk.java
** Abstract : A structure to hold screen chunks.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140227 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

/**
 * This structures holds screen chunks. A chunk is a list of contiguous screen cells 
 * within a line having the same color and attributes that should be send down to JS
 * code and redraw.
 */
public class Chunk
{
   /** Header size */
   public static final int HEADER_SIZE = 2;
   
   /** Start row  */
   private final int row;
   
   /** Start column */
   private final int column;
   
   /** Chunk size */
   private int size;
   
   /** Chunk payload */
   private byte[] chunk;

   /**
    * Constructor.
    * 
    * @param   row
    *          Start row;
    * @param   column
    *          Start column;
    * @param   size
    *          Chunk size.
    */
   public Chunk(int row, int column, int size)
   {
      this.row = row;
      this.column = column;
      this.size = size;
   }

   /**
    * Constructor.
    * 
    * @param   chunk
    *          Chunk payload.
    */
   public Chunk(byte[] chunk)
   {
      this.chunk = chunk;
      this.row = chunk[0];
      this.column = chunk[1];
      this.size = chunk.length;
   }
   
   /**
    * Get chunk payload.
    * 
    * @return  Chunk payload
    */
   public byte[] getChunk()
   {
      return chunk;
   }

   /**
    * Get start row.
    * 
    * @return  row number;
    */
   public int getRow()
   {
      return row;
   }

   /**
    * Get start column.
    * 
    * @return  column number.
    */
   public int getColumn()
   {
      return column;
   }

   /**
    * Get chunk size.
    * 
    * @return  chunk size.
    */
   public int getSize()
   {
      return size;
   }

   /**
    * Get text size.
    * 
    * @return  Text size.
    */
   public int getTextSize()
   {
      return size - Chunk.HEADER_SIZE;
   }
   
   /**
    * Return the unique hash code for this object.
    *
    * @return  Unique hash code for this object.
    */   
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      result = prime * result + column;
      result = prime * result + row;
      result = prime * result + size;
      return result;
   }

   /**
    * This method returns true if this object is equal to the specified
    * object.  Equality is defined as both objects being instances of this
    * class and the class name being the same.
    *
    * @param   obj
    *          Another Chunk object to check for equality.
    *
    * @return  True if objects are equal, else false.
    */   
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (getClass() != obj.getClass())
         return false;
      Chunk other = (Chunk) obj;
      if (column != other.column)
         return false;
      if (row != other.row)
         return false;
      if (size != other.size)
         return false;
      return true;
   }
}