/*
** Module   : WebPageHandler.java
** Abstract : delivers the main page for a web client
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150312 Created initial version by moving code from the ChuiWebPageHandler and making
**                  it more generic (to handle GUI too).
** 002 GES 20150717 Moved clipboard processing to ChUI-specific location.
** 003 SBI 20160225 Added "maxBinaryMessage", "watchdogTimeout" and "maxIdleTime" parameters to
**                  the template file in order to deliver their values to the JS client.
**     GES 20160203 Matched change in the class from which to read a constant.
**     SBI 20160321 Added the web client embedded mode "embedded".
** 004 SBI 20160628 Changed the context path to the web root path, "webRoot" is used to locate
**                  all client's resources under this virtual folder.
** 005 SBI 20180405 Changed to use template key and value provider.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.commons.codec.digest.*;
import org.eclipse.jetty.server.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;
import com.google.common.net.*;

/**
 * Jetty Handler which delivers HTML resource files. This class provide a method to parse the
 * HTML text and substitute place holders. 
 */
public class WebPageHandler
extends HtmlResourceHandler
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WebPageHandler.class.getName());
      
   /** Configuration. */
   protected BootstrapConfig config;
   
   /** Cookie token */
   private String cookieToken;   
   
   /** Authorization token */
   private String authorizationToken;
   
   /**
    * Constructor.
    *
    * @param    config
    *           Configuration parameters.
    * @param    root
    *           Prefix for all URLs relating to this page.
    * @param    skel
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public WebPageHandler(BootstrapConfig config, String root, String skel)
   {
      super(root, skel);
      this.config = config;
      
      LOG.logp(Level.INFO,
               "WebPageHandler",
               "",
               LogHelper.generate("target=%s page=%s", root, skel));
   }
   
   /**
    * Return the main page for AJAX clients. If authorization required and not authorized
    * do not load the main page client will get a HTTP 404 error, page not found.
    *    
    * @param   target 
    *          The target of the request - either a URI or a name.
    * @param   base
    *          The base request.
    * @param   request
    *          The object or a wrapper of the request.
    * @param   response 
    *          The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      if (isAuthorized(base, response))
      {
         super.handle(target, base, request, response);
      }    
   }
   
   /**
    * Set authorization token.
    * 
    * @param   token
    *          Authorization token;
    */
   public void setAuthorizationToken(String token)
   {
      this.authorizationToken = token;
   }
   
   /**
    * Convert to HTML color format.
    * 
    * @param   color
    *          Color text string.
    * @return  HTML format color
    */
   public String htmlColor(String color)
   {
      return color.startsWith("0x") ? color.replace("0x", "#") : color;
   }

   /**
    * Check authorization. First check the query parameter PARAM_TOKEN. If found and equals with
    * stored token set a new authorization cookie if not already set. This should happen on first
    * connection when the user is redirected.
    * <p>  
    * If query parameter is missing check the browser cookies list for a cookie having an 
    * authorization token that match the expected cookie. If found the request is authorized 
    * otherwise the request is denied. This should happen after initial redirection whenever the
    * page is reloaded.
    * 
    * @param   request
    *          HttpServletRequest instance.
    * @param   response
    *          HttpServletResponse instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   private boolean isAuthorized(HttpServletRequest request, HttpServletResponse response)
   {
      if (authorizationToken == null)
      {
         return false;
      }

      return (cookieToken == null) ? checkQueryParam(request, response) : checkCookies(request);
   }
   
   /**
    * Check authorization token PARAM_TOKEN. If found and equals with stored token set a new
    * authorization cookie if not already was set. This should happen on first connection when
    * the user is redirected. After new cookie is set the authorization token is no longer valid.
    * 
    * @param   request
    *          HttpServletRequest instance.
    * @param   response
    *          HttpServletResponse instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   @SuppressWarnings("deprecation")
   private boolean checkQueryParam(HttpServletRequest request, HttpServletResponse response)
   {
      // check query parameter
      String token = request.getParameter(WebClientSpawner.PARAM_TOKEN);
      
      // compare tokens
      boolean match = authorizationToken.equals(token);
      
      if (match)
      {
         // set a cookie using a new generated token
         cookieToken = DigestUtils.shaHex(UUID.randomUUID().toString());
         Cookie cookie = new Cookie(HttpHeaders.AUTHORIZATION, cookieToken);
         response.addCookie(cookie); 
      }
      
      return match;
   }
   
   /**
    * Check cookies list for a cookie that match the expected value. If found the request is
    * authorized otherwise the request is denied.
    * 
    * @param   request
    *          HttpServletRequest instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   private boolean checkCookies(HttpServletRequest request)
   {
      // check browser cookies
      Cookie[] cookies = request.getCookies();
   
      if (cookies != null)
      {
         // search for authorization cookie
         for (Cookie cookie : cookies)
         {
            String name = cookie.getName();
            String value = cookie.getValue();
            
            if (HttpHeaders.AUTHORIZATION.equals(name) && cookieToken.equals(value))
            {
               return true;
            }
         }
      }
      
      return false;      
   }

   /**
    * Generates a map of template keys with their value suppliers in order to fill gaps in
    * the html skeleton template page.
    * 
    * @param    base
    *           The request
    * 
    * @return   The target map of template keys with their value suppliers
    */
   @Override
   protected Map<String, Supplier<String>> getTemplateKeys(Request base)
   {
      WebPageKeysProvider provider = new WebPageKeysProvider(base);
      
      return provider.getTemplateKeys();
   }

   /**
    * Builds the target map of keys and their value providers for the associated template
    * given by the skeleton page.
    */
   protected class WebPageKeysProvider
   extends HtmlTemplateKeysProvider
   {
      /**
       * Creates this key and value provider.
       * 
       * @param    base
       *           The request
       */
      public WebPageKeysProvider(Request base)
      {
         super(base);
         
         add("webRoot", () -> { return config.getString("client", "web", "webRoot", "");});
         
         add("referrer",() ->
         {
            String referrer = config.getString("web", "referrer", "url", null);
            
            return referrer != null ? referrer : "" ;
         });
         
         add("maxBinaryMessage", () ->
         {
            // configuration must provide the valid value of client:web:maxBinaryMessage
            int maxBinaryMessage = config.getInt("client", "web", "maxBinaryMessage", 1024);
            
            return String.valueOf(maxBinaryMessage);
         });
         
         add("maxIdleTime", () ->
         {
            int socketTimeout = config.getInt("client", "web", "socketTimeout", -1);
            
            int maxIdleTime;
            // if socketTimeout > 0, then it overrides maxIdleTime
            // at the start of the websocket session.
            if (socketTimeout > 0)
            {
               maxIdleTime = socketTimeout;
            }
            else
            {
               // configuration must provide the valid value of client:web:maxIdleTime
               maxIdleTime = config.getInt("client", "web", "maxIdleTime", -1);
            }
            
            return String.valueOf(maxIdleTime);
         });
         
         add("watchdogTimeout", () ->
         {
            // a watch dog checks the elapsed time from its start one per a minute
            int watchdogTimeout = config.getInt("client", "web", "watchdogTimeout", 60000);
            
            return String.valueOf(watchdogTimeout);
         });
         
         // sets the web client embedded mode
         add("embedded", () ->
         {
            boolean embedded = config.getBoolean("client", "web", "embedded", false);
            
            return embedded ? "true" : "false";
         });
      }
   }
}