/*
** Module   : p2j.fonts.js
** Abstract : Font and text metrics helpers.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 CA  20150812 First version.
** 002 CA  20150818 Fixed param/return data types in the function javadoc. Added font maxWidth.
**                  If a font width the same specifications already exists, re-use it from the 
**                  local font-table by sending its index to the caller. Fixed font scaling
**                  (use window.devicePixelRatio).
** 003 CA  20151007 Fix for underline font support.
** 004 SBI 20160107 Added the fonts checker to detect if the target font family is available.
**     CA  20160323 Added getFontMaxWidth and getTextWidths.
** 005 CA  20171109 Fixed a problem with font naming in Firefox (never use double-quotes, just 
*                   single quotes).
** 006 SBI 20180130 Changed getTextWidth/getTextHeight to check validity of font parameters.
** 007 SBI 20180329 Changed derivedFont due to window.devicePixelRatio reflects zooming in browser.
** 008 SVL 20180624 Proper handling of underlined fonts: font key string and font specifications
**                  string for 2d canvas are now different: the latter shouldn't contain
**                  the "underline" keyword.
** 009 SBI 20180815 Changed to detect if its bold font is available, changed isFontInstalled in
**                  order to check the bold font is installed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.fonts = (function()
{
   /** Font object which will be returned from this initialization function. */
   var me = {}; 

   /* 
    * A js-side font-table of known fonts.  Provides a 1-to-1 mappings with the fonts known
    * known by the java side.
    *
    * This data structure contains:
    * 1. keys: a numeric value starting from 0  
    * 2. each value referenced by the key contains the font details, via an associative array:
    *    - "name"
    *    - "size"
    *    - "bold"
    *    - "italic"
    *    - "underline"
    *    - "font" - the full font name, to be used by the context
    *    - "height": font height
    *    - "widths": font widths
    *    - "maxWidth": font max width
    */
   var fontTable = {};

   /** The canvas used as a worker for measuring texts. */
   var fontCanvas;

   /** The canvas context. */
   var fontCtx;

   /** The fonts checker tests if the target font is installed. */
   var fontsChecker;
   
   /** The bold fonts checker tests if the target font is installed. */
   var boldFontsChecker;

   /** The map of embedded fonts. */
   var embeddedFontFamilies = new Map();
   
   /**
    * Initialize module.
    */
   me.init = function()
   {
      fontCanvas = document.createElement('canvas');
      fontCtx = fontCanvas.getContext('2d');
      fontsChecker = new FontsChecker(false);
      boldFontsChecker = new FontsChecker(true);
   };

   /**
    * Create a font with the given specs and registered it in the font table.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} name
    *           The font name.
    * @param    {Number} size
    *           The font size, in points.
    * @param    {Number} style
    *           The serialized font style.
    * @param    {Boolean} boldFontDef
    *           Indicates that the font definition is for the bold font.
    * @param    {String} b64Font
    *           The base64-encoded font definition.  Empty string if not available.
    */
   me.createFont = function(font, name, size, style, boldFontDef, b64Font)
   {

      var bold      = ((style & (1 << 0)) === (1 << 0));
      var italic    = ((style & (1 << 1)) === (1 << 1));
      var underline = ((style & (1 << 2)) === (1 << 2));
      var fontKey = getFontString(bold, italic, underline, name, size, true);

      // check if there is already a font with the same details
      var fcandidate = fontTable[fontKey];
      if (fcandidate != undefined)
      {
         return fcandidate;
      }

      var theFont = getFontString(bold, italic, underline, name, size, false);
      var fheight = textHeight(theFont, '\u2503Eghy', 128);
      var fwidths = fontWidths(theFont);
      var maxWidth = 0;
      
      for (var i = 0; i < fwidths.length; i++)
      {
         if (fwidths[i] > maxWidth)
         {
            maxWidth = fwidths[i];
         }
      }
      
      var fontDef = 
      {
         "name"      : name,
         "size"      : size,
         "bold"      : bold,
         "italic"    : italic,
         "underline" : underline,
         "font"      : theFont,
         "height"    : fheight,
         "widths"    : fwidths,
         "maxWidth"  : maxWidth
      };
      
      fontTable[font] = fontDef;
      fontTable[fontKey] = font;
      
      var fontFamilyKey = name + "#" + boldFontDef;
      if (b64Font != "" && !embeddedFontFamilies.has(fontFamilyKey))
      {
         embeddedFontFamilies.set(fontFamilyKey, name);
         
         var cssFont = "@font-face {  font-family: '" + name + "';";
         cssFont = cssFont + " src: url('data:application/x-font-truetype;base64," + b64Font + "')";
         cssFont = cssFont + " format('truetype');";
         if (boldFontDef)
         {
            cssFont = cssFont + "  font-weight: bold;";
         }
         cssFont = cssFont + " }";

         var newStyle = document.createElement('style');
         newStyle.appendChild(document.createTextNode(cssFont));
         document.head.insertBefore(newStyle, document.head.childNodes[0]);
         
         // added in order to load the embedded fonts
         var span = document.createElement('span');
         span.appendChild(document.createTextNode("text"));
         p2j.setStyleForElement(span,
               {
                  "font-family" : name,
                  "font-size"   : "8px",
                  "position"    : "absolute",
                  "width"       : "auto",
                  "left"        : "-9999px",
               });
         document.body.appendChild(span);
      }
      
      return font;
   };

   /**
    * Tests if the target font provided by its family name is installed.
    * 
    * @param    {String} fontFamily
    *           The font family name.
    * @param    {Boolean} boldFont
    *           Indicates that the bold font is queried.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   me.isFontInstalled = function(fontFamily, boldFont)
   {
      if (boldFont)
      {
         return boldFontsChecker.isFontAvailable(fontFamily);
      }
      
      return fontsChecker.isFontAvailable(fontFamily);
   }

   /**
    * Derive the font with the given ID to have the new size.
    * 
    * @param    {Number} font
    *           The font identifier.
    */
   me.deriveFont = function(font)
   {
      var fd = fontTable[font];
      
      // we can't use window.devicePixelRatio since it reflects browser's zoom factor and
      // window.devicePixelRatio = 1 for desktop
      var size = Math.round(/*window.devicePixelRatio * */fd.size);
      
      if (fd.size != size)
      {
         fd.size = size;

         var fontKey = getFontString(fd.bold, fd.italic, fd.underline, fd.name, fd.size, true);
         var theFont = getFontString(fd.bold, fd.italic, fd.underline, fd.name, fd.size, false);

         // set the key for the new name
         fontTable[fontKey] = font;
         fd.font = theFont;
      }
   };
   
   /**
    * For the specified font in the font table, set its style.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {Number} style
    *           The serialized font style.
    * 
    * @return   {boolean} true if the font style has changed.
    */
   me.setFontStyle = function(font, style)
   {
      var bold      = ((style & (1 << 0)) == (1 << 0));
      var italic    = ((style & (1 << 1)) == (1 << 1));
      var underline = ((style & (1 << 2)) == (1 << 2));

      var fd = fontTable[font];
      var changed = false;
      
      if (fd.bold != bold)
      {
         fd.bold = bold;
         changed = true;
      }
      
      if (fd.italic != italic)
      {
         fd.italic = italic;
         changed = true;
      }
      
      if (fd.underline != underline)
      {
         fd.underline = underline;
         changed = true;
      }
      
      if (changed)
      {
         var theFont = getFontString(fd.bold, fd.italic, fd.underline, fd.name, fd.size, false);
         fd["font"] = theFont;
      }
      
      return changed;
   };
   
   /*
    * Determine if the specified font is underlined.
    *
    * @return   <code>true</code> if the font is underlined.
    */
   me.isUnderlined = function(font)
   {
      return font != undefined && fontTable[font].underline;
   };
   
   /**
    * Get the name of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {String} The font name.
    */
   me.getFontName = function(font)
   {
      return fontTable[font].font;
   };
   
   /**
    * Get the widths of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {Number[]} The font widths.
    */
   me.getFontWidths = function(font)
   {
      return fontTable[font].widths;
   };
   
   /**
    * Get the height of the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {Number} The font height.
    */
   me.getFontHeight = function(font)
   {
      return fontTable[font].height;
   };
   
   /**
    * Get the font's max width.
    *
    * @param    {Number} font
    *           The font identifier.
    *
    * @return   {Number} The font's maximum height.
    */
   me.getFontMaxWidth = function(font)
   {
      return fontTable[font].maxWidth;
   };
   
   /**
    * Get the text width using the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} text
    *           The text.
    *
    * @return   {Number} The text width.
    */
   me.getTextWidth = function(font, text)
   {
      if (text.length == 0)
      {
         return 0;
      }

      if (!font)
      {
         return 0;
      }

      var f = fontTable[font];

      if (!f || !f.font)
      {
         return 0;
      }

      return stringWidth(f.font, text);
   };
   
   /**
    * Get the text widths using the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} text
    *           The text.
    *
    * @return   {Number} The text widths.
    */
   me.getTextWidths = function(font, text)
   {
      if (text.length == 0)
      {
         return 0;
      }
   
      var f = fontTable[font];

      var widths = [];
      
      for (var i = 1; i <= text.length; i++)
      {
         widths.push(stringWidth(f.font, text.substring(0, i)));
      }
      
      return widths;
   };
   
   /**
    * Get the text height using the font associated with the font ID.
    * 
    * @param    {Number} font
    *           The font identifier.
    * @param    {String} text
    *           The text.
    *
    * @return   {Number} The text height.
    */
   me.getTextHeight = function(font, text)
   {
      if (text.length == 0)
      {
         return 0;
      }
   
      if (!font)
      {
         return 0;
      }
      
      var f = fontTable[font];

      if (!f || !f.maxWidth || !f.font)
      {
         return 0;
      }
      
      text = '\u2503' + text;

      // ensure the text will fit the canvas
      var pxWidth = f.maxWidth * text.length;
      
      var theight = textHeight(f.font, text, pxWidth);
      
      return theight;
   };
 
   /**
    * Compute the height of the text, using the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * @param    {String} text
    *           The text.
    * @param    {Number} width
    *           The canvas width, so that the text will fit.
    *
    * @return   {Number} The text height.
    */
   function textHeight(font, text, width)
   {
      var height = 128;

      fontCtx.save();

      fontCanvas.setAttribute('height', height);
      fontCanvas.setAttribute('width', width);
      fontCtx.fillStyle = 'black';
      fontCtx.fillRect(0, 0, width, height);
      fontCtx.textBaseline = 'top';
      fontCtx.fillStyle = 'white';
      font = font.replace(/'/g , '"');
      fontCtx.font = font;
      fontCtx.fillText(text, 0, 0);

      var pixels = fontCtx.getImageData(0, 0, width, height).data;

      // row numbers where we first find letter end where it ends 
      var beg = -1;
      var end = -1;

      for (var row = 0; row < height; row++)
      {
         for (var col = 0; col < width; col++)
         {
            var index = (row * width + col) * 4;
            
            // if pixel is not white (background color)
            if (pixels[index] === 0)
            {
               // we haven't met white (font color) pixel
               // on the row and the letters was detected
               if (col === (width - 1) && beg !== -1)
               {
                  end = row;
                  row = height;
                  break;
               }
            }
            else
            {
               // we find top of letter
               if (beg === -1)
               {
                  beg = row;
               }
               // ..letters body
               break;
            }
         }
      }

      fontCtx.restore();
      
      return end - beg;
   }
   
   /**
    * Compute the font widths, for the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * 
    * @return   {Number[]} The font widths.
    */
   function fontWidths(font)
   {
      var widths = [];
      for (var ch = 0 ; ch < 256 ; ch++)
      {
          widths.push(stringWidth(font, String.fromCharCode(ch)));
      }
      return widths;
   }
   
   /**
    * Measure the width of the text, using the given font.
    *
    * @param    {Number} font
    *           The font, as in the fontTable[x].font.
    * @param    {String} text
    *           The text.
    * 
    * @return   {Number} The text width.
    */
   function stringWidth(font, text)
   {
      fontCtx.save();

      fontCtx.textBaseline = 'top';
      font = font.replace(/'/g , '"');
      fontCtx.font = font;
      
      var width = fontCtx.measureText(text).width;
      
      fontCtx.restore();

      return Math.round(width);  // Chrome returns float values which should be properly rounded.
   }

    /**
     * Compose the font key used in the fontTable or the font specifications as in
     * fontTable[x].font.
     *
     * @param   {boolean} bold
     *          True if the font is bold.
     * @param   {boolean} italic
     *          True if the font is italic.
     * @param   {boolean} underline
     *          True if the font is underlined.
     * @param   {String} name
     *          Font name.
     * @param   {Number} size
     *          Font size.
     * @param   {boolean} keyString
     *          If True, the font key used in the fontTable is returned. If False, then the font
     *          specifications as in fontTable[x].font is returned.
     *
     * @returns {string}
     *          the font key used in the fontTable or the font specifications as in
     *          fontTable[x].font.
     */
   function getFontString(bold, italic, underline, name, size, keyString)
   {
      var res = (bold ? "bold " : "") +
                (italic ? "italic " : "") +
                (underline && keyString ? "underline " : "") +
                size + "pt" +
                " '" + name + "'";

      if (keyString)
      {
         res = res.toLowerCase();
      }

      return res;
   }

   /**
    * Checks if a font is available to be used on a web page. The bold font parameter indicates
    * that the tested font has the bold font weight.
    * 
    * @param   {Boolean} boldFont
    *          Indicates that the tested font has the bold font weight.
    */
   function FontsChecker(boldFont) {
      
      /** Create a span element to contain the predefined sample text. */
      var spanEl  = document.createElement('span');
      
      /** Create a sample text. */
      var sampleText = "wwwwwwwwwwmmmmmmmmmmmiiiiiiiiiillllllllll";
      
      // Added the predefined sample text.
      spanEl.appendChild(document.createTextNode(sampleText));
      
      /** The style properties */
      var spanElStyle =
      {
         "font-size"   : "64px",
         "position"    : "absolute",
         "width"       : "auto",
         "left"        : "-9999px"
      };
      
      if (boldFont)
      {
         spanElStyle["font-weight"] = "bold";
      }
      
      /** Store the sample text with against the target font family using it as a key field. */
      var sampleWidths = {};
      
      // Save the sample text width against monospace font.
      sampleWidths["monospace"]  = getSampleWidth("monospace");
      
      // Save the sample text width against serif font.
      sampleWidths["serif"]      = getSampleWidth("serif");
      
      // Save the sample text width against sans-serif font.
      sampleWidths["sans-serif"] = getSampleWidth("sans-serif");
      
      
      /**
       * Calculates the predefined sample text width for the given font family.
       * 
       * @param    {String} fontFamily
       *           The font family name
       * 
       * @return   The with of the predefined sample text against the given font family.
       */
      function getSampleWidth(fontFamily) {
          spanElStyle["font-family"] = fontFamily;
          p2j.setStyleForElement(spanEl, spanElStyle);
          document.body.appendChild(spanEl);
          var width = spanEl.clientWidth;
          document.body.removeChild(spanEl);
          
          return width;
      };
      
      /**
       * Tests the given font against the predefined fall back font. 
       * 
       * @param    {String} fontFamily
       *           The font family name
       * @param    {String} fallbackFontFamily
       *           The fall back font family name
       * 
       * @return   True if the tested font is available and the sample text has a different width
       *           than the fall back font, otherwise false.
       */
      function testAgainst(fontFamily, fallbackFontFamily)
      {
         var width = sampleWidths[fallbackFontFamily];
         if (width === undefined)
         {
            width = getSampleWidth(fallbackFontFamily);
            sampleWidths[fallbackFontFamily] = width;
         }
         return width !== getSampleWidth(fontFamily + "," + fallbackFontFamily);
      }
      
      /**
       * Check if the font is available.
       * 
       * @param    {String} fontFamily
       *           The font family name
       * 
       * @return   True if the font is available, otherwise false.
       */
      this.isFontAvailable = function (fontFamily)
      {
          return testAgainst(fontFamily, "monospace") || testAgainst(fontFamily, "sans-serif") ||
              testAgainst(fontFamily, "serif");
      };
   }
   
   return me;
}());
