/*
** Module   : p2j.remote.js
** Abstract : Define APIs for handling messages into the converted code or from the converted code
**            into a customer-specific application.
**
** Copyright (c) 2016-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20160418 Created initial version.
** 002 CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 003 HC  20171203 Added notification of Web embedded client when program execution enters modal
**                  or wait-for event loop.
** 004 CA  20180727 Added APPLY extension.  Added notifications to show a window, and to send the
**                  window's status and message texts.
** 005 CA  20180801 Removed SHOW_WINDOW, as it is not enough for ADM window show.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.remote = (function()
{   
   var me = {};
   
   me.REQUEST         = "request";
   me.RESPONSE        = "response";
   me.REMOTE_CALL     = "remote-call";
   me.INVOKE          = "invoke";
   me.DELETE_HANDLE   = "delete-handle";
   me.SUBSCRIBE       = "subscribe";
   me.UNSUBSCRIBE     = "unsubscribe";
   me.PUBLISH         = "publish";
   me.CREATE_RESOURCE = "create-resource";
   me.APPLY_EVENT     = "apply-event";
   me.SET_STATUS_TEXT = "set-status";
   me.CLEAR_MESSAGES  = "clear-messages";
   me.ADD_MESSAGES    = "add-messages";
   
   me.EXTERNAL_PROGRAM = 0;
   me.PROCEDURE        = 1;
   me.FUNCTION         = 2;

   me.PAGE_LOADED         = -1;
   me.PAGE_CLOSED         = -2;
   me.SOCKET_DISCONNECTED = -3;
   me.SOCKET_RECONNECTED  = -4;
   me.WAIT_USER_INPUT     = -5;

   me.handleMessages = false;

   /**
    * Initialize module.
    * 
    * @param   {object} win
    *          The window where to send the messages.
    */
   me.init = function(win)
   {
      if (p2j.embedded)
      {
         this.targetWindow = win;
         this.targetOrigin = "*";
      
         var handler = receiveP2JMessage;
   
         if (window.addEventListener)
         {
            window.addEventListener("message", handler, false)
         }
         else
         {
            window.attachEvent("onmessage", handler)
         }
      }
   };

   /**
    * Send the specified message to the <code>targetWindow</code>.
    *
    * @param    {object} msg
    *           An associative array message which will be sent to the <code>targetWindow</code>.
    */
   me.send = function(msg)
   {
      if (p2j.embedded && this.targetWindow)
      {
         this.targetWindow.postMessage(msg, this.targetOrigin);
      }
      else
      {
         var smsg = JSON.stringify(msg);
         
         if (!p2j.embedded)
         {
            console.log("Could not send message " + smsg + 
                        " - P2J client is not running in embedded mode!");
         }
         else
         {
            console.log("Could not send message " + smsg + " - target window was not set!");
         }
      }
   };
   
   /**
    * Send an error message to the <code>targetWindow</code>.
    *
    * @param    {number} msgId
    *           The message ID from the request.
    * @param    {number} msgType
    *           The message type.
    * @param    {string} errMsg
    *           The error message.
    * @param    {string} errNum
    *           The error number.
    */
   me.sendError = function(msgId, msgType, errorMsg, errorNum)
   {
      var payload = {};
      payload["error"] = true;
      payload["error-msg"] = errorMsg;
      payload["error-num"] = errorNum || 0;
   
      var msg = {};
      msg["type"] = msgType;
      msg["id"] = msgId;
      msg["mode"] = this.RESPONSE;
      msg["payload"] = payload;
      
      this.send(msg);
   };
   
   /**
    * Notify the parent window that the P2J client has loaded.
    */
   me.notifyPageLoaded = function()
   {
      if (!p2j.embedded)
      {
         return;
      }
      
      this.secret = Math.floor((Math.random() * 1000000) + 1);
   
      // no message ID is set

      var msg = {};
      msg["type"] = this.PAGE_LOADED;
      msg["mode"] = this.REQUEST;
      msg["payload"] = this.secret;
      
      this.send(msg);
   };
   
   /**
    * Notify the parent window that the P2J client's window has closed.
    */
   me.notifyPageClosed = function()
   {
      if (!p2j.embedded)
      {
         return;
      }

      // no message ID is set; no response is expected for this API

      var msg = {};
      msg["type"] = this.PAGE_CLOSED;
      msg["mode"] = this.REQUEST;
      
      this.send(msg);

      this.targetOrigin = "*";
      this.handleMessages = false;
      
      delete this.secret;
   };
   
   /**
    * Notify the parent window that the P2J web-socket has disconnected.
    */
   me.notifySocketDisconnected = function()
   {
      if (!p2j.embedded)
      {
         return;
      }

      // no message ID is set; no response is expected for this API

      var msg = {};
      msg["type"] = this.SOCKET_DISCONNECTED;
      msg["mode"] = this.REQUEST;
      
      this.send(msg);
   };
   
   /**
    * Notify the parent window that the P2J web-socket has reconnected.
    */
   me.notifySocketReconnected = function()
   {
      if (!p2j.embedded)
      {
         return;
      }

      // no message ID is set; no response is expected for this API

      var msg = {};
      msg["type"] = this.SOCKET_RECONNECTED;
      msg["mode"] = this.REQUEST;
      
      this.send(msg);
   };

   /**
    * Callback executed when a message is received by the P2J client window.
    *
    * @param    {event} event
    *           The message event details.
    */
   function receiveP2JMessage(event)
   {
      // WARNING! here "this" will be the Window object (not "me"!)

      var msg = event.data;

      if (!p2j.embedded)
      {
         var smsg = JSON.stringify(msg);
         console.log("Received message " + smsg + " in non-embedded mode!");
         return;
      }
      
      if (msg.target === "Frame" && msg.bgcolor)
      {
         p2j.setDefaultDesktopBgcolor(msg.bgcolor);
         return;
      }

      // For Chrome, the origin property is in the event.originalEvent object.
      var origin = event.origin || event.originalEvent.origin;

      // if the message is a response for PAGE_LOADED, then check that the source is the same as
      // the targetWindow and the secret
      if (msg.type == p2j.remote.PAGE_LOADED)
      {
         if (p2j.remote.handleMessages)
         {
            var smsg = JSON.stringify(msg);
            console.log("p2j.remote received a PAGE_LOADED message after it was initialized: " +
                        smsg);
            return;
         }
      
         if (msg.mode == p2j.remote.RESPONSE)
         {
            if (event.source == p2j.remote.targetWindow)
            {
               var secret = msg.payload;
               if (secret == p2j.remote.secret)
               {
                  p2j.remote.handleMessages = true;
                  p2j.remote.targetOrigin = origin;
                  
                  delete p2j.remote.secret;
               }
               else
               {
                  console.log("Received a PAGE_LOADED message with incorrect secret " + 
                              secret + " from window: " + origin);
               }
            }
            else
            {
               console.log("Received a PAGE_LOADED message from incorrect window: " + origin);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a PAGE_LOADED message with " + msgMode + 
                        " mode - P2J can handle only RESPONSE modes!");
         }
         
         return;
      }

      if (!p2j.remote.handleMessages)
      {
         var smsg = JSON.stringify(msg);
         console.log("p2j.remote received a message before it was initialized: " + smsg);
         return;
      }

      if (origin !== p2j.remote.targetOrigin)
      {
         var smsg = JSON.stringify(msg);

         console.log("Received message " + smsg + " from unexpected window origin: " + origin +
                     "(expected origin is " + p2j.remote.targetOrigin + ")");
         return;
      }

      var types =
      [
         p2j.remote.REMOTE_CALL,     // 0
         p2j.remote.INVOKE,          // 1
         p2j.remote.DELETE_HANDLE,   // 2
         p2j.remote.SUBSCRIBE,       // 3
         p2j.remote.UNSUBSCRIBE,     // 4
         p2j.remote.PUBLISH,         // 5
         p2j.remote.CREATE_RESOURCE, // 6
         p2j.remote.APPLY_EVENT      // 7
      ];
      var modes = 
      [
         p2j.remote.RESPONSE, // 0
         p2j.remote.REQUEST,  // 1
         p2j.remote.REQUEST,  // 2
         p2j.remote.REQUEST,  // 3
         p2j.remote.REQUEST,  // 4
         p2j.remote.REQUEST,  // 5
         p2j.remote.REQUEST,  // 6
         p2j.remote.REQUEST   // 7
      ];
      var validators = 
      [
         undefined,                  // 0
         validInvokeRequest,         // 1
         validDeleteHandleRequest,   // 2
         validSubscribeRequest,      // 3
         validSubscribeRequest,      // 4
         validPublishRequest,        // 5
         validCreateResourceRequest, // 6
         undefined                   // 7
      ];
      var executors = 
      [
         p2j.socket.sendRemoteCallResponse,    // 0
         p2j.socket.sendInvokeRequest,         // 1
         p2j.socket.sendDeleteHandleRequest,   // 2
         p2j.socket.sendSubscribeRequest,      // 3
         p2j.socket.sendUnsubscribeRequest,    // 4
         p2j.socket.sendPublishRequest,        // 5
         p2j.socket.sendCreateResourceRequest, // 6
         p2j.socket.sendApplyEvent             // 7
      ];
      
      var found = false;
      
      for (var i = 0; i < types.length; i++)
      {
         if (msg.type == types[i])
         {
            if (msg.mode == modes[i])
            {
               if (!socketConnected(msg))
               {
                  return;
               }
   
               if (validators[i] == undefined || validators[i](msg))
               {
                  // execute the request
                  executors[i](msg);
               }
            }
            else
            {
               var msgMode = (msg.mode ? msg.mode : "undefined");
               console.log("Received a " + types[i] + " message with " + msgMode + 
                           " mode - P2J can handle only " + modes[i] + " modes!");
            }
               
            found = true;
            break;
         }
      }

      if (!found)
      {
         // unknown payload, discard
         var data = JSON.stringify(msg);
         console.log("P2J has received an unknown message, and will be discarded: " + data);
      }
   };
   
   /**
    * Check if the client socket is connected.  If it is not connected, bounce back the message,
    * with the payload holding an error.
    *
    * @param   {object} request
    *          The request message.
    */
   function socketConnected(request)
   {
      if (!p2j.socket.isConnected())
      {
         // send back the response, with a "not connected" error
         var payload = {};
         payload["error"] = true;
         payload["error-msg"] = "Client socket is not connected";
         payload["error-num"] = -1000;
      
         var msg = {};
         msg["type"] = request.type;
         msg["id"] = request.id;
         msg["mode"] = p2j.remote.RESPONSE;
         msg["payload"] = payload;

         msg["request"] = request;
         
         p2j.remote.send(msg);
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Validate the INVOKE request data.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    *
    * @return   <code>true</code> if the INVOKE request contains valid fields.
    */
   function validInvokeRequest(msgReq)
   {
      var msgId   = msgReq["id"];
      var payload = msgReq["payload"];
      var name    = payload["name"];
      var type    = payload["type"];
      var noError = payload["noError"];
      
      var errorMsg;
      
      // msgId must be a positive number
      if (typeof msgId != "number")
      {
         errorMsg = "The message ID must be a number!";
      }
      else if (msgId <= 0)
      {
         errorMsg = "The message ID must be a positive number!";
      }
      else
      {
         // validate info from the payload
         
         if (typeof name != "string")
         {
            // name must be a string
            errorMsg = "The payload.name field must be string value!";
         }
         else if (typeof type != "number")
         { 
            // type must be numeric
            errorMsg = "The payload.type field must be numeric value!";
         }
         else if (!(type == p2j.remote.EXTERNAL_PROGRAM || 
                    type == p2j.remote.PROCEDURE || 
                    type == p2j.remote.FUNCTION))
         {
            // type must be a known type (0, 1, 2)
            errorMsg = "The payload.type field [" + type +" ] must be a known value (0, 1 or 2)!";
         }
         else if (payload["noError"] && (typeof noError != "boolean"))
         {
            // noError must be boolean
            errorMsg = "The payload.noError field must be a boolean value!";
         }
         else if (type == p2j.remote.EXTERNAL_PROGRAM)
         {
            var persistent = payload["persistent"];
            
            if (payload["persistent"] && (typeof persistent != "boolean"))
            {
               // noError must be boolean
               errorMsg = "The payload.persistent field must be a boolean value!";
            }
         }
         else
         {
            var handleId = payload["handle"];
            
            if (payload["handle"] && (typeof handleId != "string"))
            {
               // handleId must be string
               errorMsg = "The payload.handle [" + handleId + "] " + 
                           "must be a hex-encoded value, as string!";
            }
         }
      }
      
      if (errorMsg)
      {
         p2j.remote.sendError(msgId, p2j.remote.INVOKE, errorMsg);
         return false;
      }
      
      return true;
   };
   
   /**
    * Validate the DELETE_HANDLE request data.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    *
    * @return   <code>true</code> if the DELETE_HANDLE request contains valid fields.
    */
   function validDeleteHandleRequest(msgReq)
   {
      var msgId = msgReq["id"];
      var handleId = msgReq["payload"];
      var noError = msgReq["noError"];
      
      var errorMsg;
      
      // msgId must be a positive number
      if (typeof msgId != "number")
      {
         errorMsg = "The message ID must be a number!";
      }
      else if (msgId <= 0)
      {
         errorMsg = "The message ID must be a positive number!";
      }
      else if (typeof handleId != "string")
      {
         // handleId must be string
         errorMsg = "The handle ID [" + handleId + "] " + 
                    "passed as the payload must be a hex-encoded value, as string!";
      }
      else if (msgReq["noError"] && (typeof noError != "boolean"))
      {
         // noError must be boolean
         errorMsg = "The noError field must be a boolean value!";
      }

      if (errorMsg)
      {
         p2j.remote.sendError(msgId, p2j.remote.DELETE_HANDLE, errorMsg);
         return false;
      }

      return true;
   };
   
   /**
    * Validate the SUBSCRIBE or UNSUBSCRIBE request data.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    *
    * @return   <code>true</code> if the request contains valid fields.
    */
   function validSubscribeRequest(msgReq)
   {
      var msgId = msgReq["id"];
      var payload = msgReq["payload"];
      var noError = msgReq["noError"];
      
      var global     = payload["global"];
      var subscriber = payload["subscriber"];
      var event      = payload["event"];
      var publisher  = payload["publisher"];
      
      var errorMsg;
      
      // msgId must be a positive number
      if (typeof msgId != "number")
      {
         errorMsg = "The message ID must be a number!";
      }
      else if (msgId <= 0)
      {
         errorMsg = "The message ID must be a positive number!";
      }
      else if (event && (typeof event != "string"))
      {
         // event must be string
         errorMsg = "The event [" + event + "] in the payload must be a string!";
      }
      else if (publisher && (typeof publisher != "string"))
      {
         // publisher must be string
         errorMsg = "The publisher ID [" + publisher + "] " + 
                    " must be a hex-encoded value, as string!";
      }
      else if (subscriber && (typeof subscriber != "string"))
      {
         // subscriber must be string
         errorMsg = "The subscriber ID [" + subscriber + "] " + 
                    " must be a hex-encoded value, as string!";
      }
      else if (msgReq["noError"] && (typeof noError != "boolean"))
      {
         // noError must be boolean
         errorMsg = "The noError field must be a boolean value!";
      }
      else if (payload["global"] && (typeof global != "boolean"))
      {
         // global must be boolean
         errorMsg = "The global flag must be a boolean value!";
      }

      if (errorMsg)
      {
         p2j.remote.sendError(msgId, msgReq["type"], errorMsg);
         return false;
      }

      return true;
   };

   /**
    * Validate the PUBLISH request data.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    *
    * @return   <code>true</code> if the request contains valid fields.
    */
   function validPublishRequest(msgReq)
   {
      var msgId = msgReq["id"];
      var payload = msgReq["payload"];
      
      var event      = payload["event"];
      var publisher  = payload["publisher"];
      var args       = payload["args"];
      
      var errorMsg;
      
      // msgId must be a positive number
      if (typeof msgId != "number")
      {
         errorMsg = "The message ID must be a number!";
      }
      else if (msgId <= 0)
      {
         errorMsg = "The message ID must be a positive number!";
      }
      else if (event && (typeof event != "string"))
      {
         // event must be string
         errorMsg = "The event [" + event + "] in the payload must be a string!";
      }
      else if (typeof publisher != "string")
      {
         // publisher must be string
         errorMsg = "The publisher ID [" + publisher + "] " + 
                    " must be a hex-encoded value, as string!";
      }
      else if (payload["args"] && (typeof args != "object"))
      {
         // arguments must be object
         errorMsg = "The arguments parameter must be an object!";
      }

      if (errorMsg)
      {
         p2j.remote.sendError(msgId, msgReq["type"], errorMsg);
         return false;
      }

      return true;
   };

   /**
    * Validate the CREATE_RESOURCE request data.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    *
    * @return   <code>true</code> if the request contains valid fields.
    */
   function validCreateResourceRequest(msgReq)
   {
      var msgId = msgReq["id"];

      var errorMsg;
      
      // msgId must be a positive number
      if (typeof msgId != "number")
      {
         errorMsg = "The message ID must be a number!";
      }
      else if (msgId <= 0)
      {
         errorMsg = "The message ID must be a positive number!";
      }

      if (errorMsg)
      {
         p2j.remote.sendError(msgId, msgReq["type"], errorMsg);
         return false;
      }

      return true;
   };

   return me;
}());