/*
** Module   : ServerEvent.java
** Abstract : defines an event which originates from the server-side 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
** 002 CA  20130908 Added API to remove certain server events from the queue.
** 003 HC  20131031 Changed resource ID data type to Long, see issue #2183.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Instances of this class represent a definition of an event which originates from the server 
 * side. Currently, the following server-side events are supported:
 * <ul>
 *    <li>{@link Keyboard#PROCEDURE_COMPLETE}: raised when an async appserver request is finished
 *    </li> 
 * </ul>
 * 
 * This class also keeps a registry of server-side events posted to the client-side; when the
 * client-side processes the event, the associated {@link ServerEvent} instance will be resolved
 * and the event will be executed, using the registered {@link #task}.
 * <p>
 * In 4GL, is possible to add user-defined triggers or use the server-events in APPLY statements;
 * testing showed that doing so results in abnormal behavior (as the i.e. parameters, procedure 
 * context or the SELF handle can't be resolved). 
 * <p>
 * In current P2J implementation, any server-event names used when defining triggers or in APPLY
 * statements will be treated and will behave as any other UI event; their server-event "flavor" 
 * will be hidden and is exposed only when the {@link ServerEvent} instances are posted to the
 * event queue via {@link LogicalTerminal#postServerEvent}.
 */
public class ServerEvent
extends Event
implements Externalizable
{
   /** A registry of server-side events which were posted to the client. */
   private static ContextLocal<Map<Integer, ServerEvent>> events =
   new ContextLocal<Map<Integer, ServerEvent>>()
   {
      protected Map<Integer, ServerEvent> initialValue()
      {
         return new HashMap<Integer, ServerEvent>();
      };
   };
   
   /** The id of the event. */
   private int id;
   
   /** The resource ID which created the event. */
   private long resourceId;
   
   /** The name of this event. */
   private String name;
   
   /**
    * The associated task which will be executed when the event is processed by the client-side.
    * May be <code>null</code> if the instance is created on the server-side. 
    */
   private Runnable task = null;

   /**
    * Public c'tor added for deserialization purposes only. Not for public use. This will be
    * invoked only on client-side, when a new server event instance is received. In this case, it
    * is OK for the {@link #task} field to be null.
    */
   public ServerEvent()
   {
      super(null, -1);

      // no-op
   }
   
   /**
    * Create a new server-side event.
    * 
    * @param    resourceId
    *           The id of the resource which created the event.
    * @param    name
    *           The name of this event.
    * @param    task
    *           The task associated with this event.
    */
   public ServerEvent(long resourceId, String name, Runnable task)
   {
      super(null, Utils.uniqueId());

      // inherit the overwritten id from the super implementation
      this.id = super.id();
      this.name = name;
      this.task = task;
      this.resourceId = resourceId;
      
      Map<Integer, ServerEvent> evts = events.get();
      synchronized (evts)
      {
         evts.put(id(), this);
      }
   }

   /**
    * Resolve and invoke the event with the given ID.
    * 
    * @param    id
    *           The event ID.
    */
   public static void invoke(int id)
   {
      Map<Integer, ServerEvent> evts = events.get();
      ServerEvent event = null;
      
      synchronized (evts)
      {
         event = events.get().remove(id);
      }
      
      // invoke only if the event was not already purged
      if (event != null)
      {
         event.task.run();
      }
   }
   
   /**
    * Remove from the server event queue all the events in the given set, identified by their id.
    * 
    * @param    eventIds
    *           The set of event IDs to be removed.
    */
   public static void removeServerEvents(Set<Integer> eventIds)
   {
      Map<Integer, ServerEvent> evts = events.get();
      synchronized (evts)
      {
         evts.keySet().removeAll(eventIds);

         LogicalTerminal.removeServerEvents(eventIds);
      }
   }
   
   /**
    * Get the ID of this server-side event.
    * 
    * @return   The {@link #id} of this server-side event.
    */
   @Override
   public int id()
   {
      return id;
   }
   
   /**
    * Get the name of this server-event.
    * 
    * @return   See above.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Get the resource ID of this server-event.
    * 
    * @return   See above.
    */
   public long resourceId()
   {
      return resourceId;
   }

   /**
    * Send this server-event to the destination. Only the {@link #id}, {@link #resourceId} and
    * {@link #name} fields will be sent.
    * 
    * @param    out
    *           The output destination to which the server event will be sent.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(id);
      out.writeLong(resourceId);
      out.writeUTF(name);
   }

   /**
    * Initialize this server event instance with the data received from the input source.
    * 
    * @param    in
    *           The input source from which server event will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      id = in.readInt();
      resourceId = in.readLong();
      name = in.readUTF();
   };
}