/* 
** Module   : DialogBoxWindow.java
** Abstract : Provides window for DIALOG-BOX.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20150929 Initial version.
** 002 SBI 20151022 Calls doLayout on show events due to the resize issue inside drawing circles
**                  in the case of the web client.
** 003 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 004 CA  20160327 Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 005 VIG 20160313 Import cleanup and window title bar support moved to TopLevelWindow.
** 006 EVL 20160512 Dialog-box needs to have additional 4 pixels to be added to workspace size.
**                  Refreshing the color set to use with the dialog-box.
** 007 SBI 20160811 The dialog box owner should be CURRENT-WINDOW.
** 008 CA  20161005 If the dialog box frame has fixed height/width, do not set the client width or
**                  height fields in the frame config.
** 009 HC  20161227 Close title bar button must trigger WINDOW-CLOSE event for the dialog's
**                  frame.
** 010 HC  20170913 Added runtime support of WINDOW:ALWAYS-ON-TOP.
** 011 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 012 CA  20180604 Fixed height attributes for dialog frames (they must include the  title
**                  height, when explicitly set).
**                  Fixed centering of dialog-box, when business logic explicitly set one or both
**                  location coordinates.
**     HC  20180611 Fixed wrong frame size displayed in dialog-box.
** 013 EVL 20180713 No difference between explicit and imlicit size change, title height must be
**                  added to the content pane size.
** 014 EVL 20180724 Moving the title height adding to the separate method to unify the approach
**                  to all dialog boxed including system dialogs.
** 015 HC  20180813 Implemented initial version of dynamic layout.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The class implements a window container for a FRAME displayed in DIALOG-BOX. 
 */
public class DialogBoxWindow
extends ModalWindow
implements HasDynamicLayout
{
   /** Color resolver. */
   private GuiColorResolver gc;

   /** Enabler of dynamic layout. */
   private WindowDynamicLayoutIntegration dynamicLayout;

   /**
    * The constructor.
    * 
    * @param    title
    *           The frame title to be displayed in the window title bar.
    */
   public DialogBoxWindow(String title)
   {
      super(title, WindowManager.getCurrentWindow());
      
      gc = new GuiColorResolver(gd, new BaseConfig(), 
                                SysColor.COLOR_3DFACE, 
                                SysColor.COLOR_WINDOWTEXT);

      ((BorderedPanelGuiImpl) windowPane).setColorResolver(gc);
      
      ((BorderedPanelGuiImpl) contentPane).setColorResolver(gc);
      contentPane.setLayout(new BorderLayout<GuiOutputManager>());

      // the titlebar can be null when the web client is in embedded mode
      WindowTitleBar tbar = getTitleBar();
      if (tbar != null)
      {
         tbar.setClickHandler(CaptionButtonType.CLOSE, () ->
         {
            Event evt = new WindowEvent(getFrame(), Keyboard.SE_WINDOW_CLOSE);
            EventManager.postOSEvent(evt);
         });
      }

      if (ThinClient.isDynamicLayoutEnabled())
      {
         dynamicLayout = new WindowDynamicLayoutIntegration(this);
      }
   }
   
   /**
    * Adds the specified frame to the window's content pane.
    * 
    * @param    frame
    *           The frame instance to process.
    */
   @Override
   public void addFrame(Widget<GuiOutputManager> frame, double column, double row)
   {
      // TOOD: fix the drop down abuse of this method 
      if (getFrame() != null && !(frame instanceof DropDownGuiImpl))
      {
         throw new RuntimeException("Cannot host multipe frames!");
      }
      
      contentPane.add(frame);
      
      FrameConfig c = ((FrameGuiImpl) frame).config();
      c.parentId = c.windowID = getOwner().getId().asInt();
   }
   
   /**
    * Returns the managing frame instance.
    * 
    * @return   See above.
    */
   public FrameGuiImpl getFrame()
   {
      return (FrameGuiImpl) contentPane.get(0);
   }
   
   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For a dialog-box,
    * the call is delegated to {@link Frame#getPortableMouseEventCoordinates(NativePoint)}.
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   A point representing the coordinates of a mouse-portable event or 
    *           <code>null</code> if such an event can't be raised.
    */
   @Override
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      return getFrame().getPortableMouseEventCoordinates(point);
   }
   
   /**
    * Changes the dialog-box color set.
    * 
    * @param    bc
    *           New config to apply.
    */
   public void refreshColors(BaseConfig bc)
   {
      GuiColorResolver gc = new GuiColorResolver(gd, bc, SysColor.COLOR_3DFACE,
                                                 SysColor.COLOR_WINDOWTEXT);

      ((BorderedPanelGuiImpl) contentPane).setColorResolver(gc);
   }

   /**
    * This method must be called when the associated frame instance has 
    * changed its dimension.
    */
   public void onFrameDimensionChanged()
   {
      if (!isRealized())
      {
         return;
      }

      FrameGuiImpl frame = getFrame();
      FrameConfig fc = frame.config();
      Dimension dim = frame.dimension();

      getContentPane().setSize(dim.width, dim.height);
      doLayout();
   }
   
   /**
    * This method must be called when the associated frame instance has 
    * changed its location.
    */
   public void onFrameLocationChanged()
   {
      TopLevelWindow<?> owner = getOwner();
      if (owner == null)
      {
         return;
      }
      
      gd.selectWindow(owner.getId().asInt());
      NativePoint ownerLocation = gd.getWindowLocation();
      gd.releaseWindow();

      Point location = ConfigHelper.getAssignedLocation(getFrame().config());
      NativePoint nloc = screen().coordinates().pointToNative(location);
      
      // dialog-box location is the same as the frame assigned location
      // relative to the ower window
      if (nloc.x != BaseConfig.INV_COORD)
      {
         nloc.x += ownerLocation.x;
      }
      if (nloc.y != BaseConfig.INV_COORD)
      {
         nloc.y += ownerLocation.y;
      }
      
      setPhysicalLocation(nloc.x, nloc.y);

      if (isRealized())
      {
         gd.selectWindow(getId().asInt());
         gd.setWindowLocation(nloc.x, nloc.y);
         gd.releaseWindow();
      }
   }

   /**
    * Show window.
    */
   @Override
   public void show()
   {
      if (isVisible())
      {
         return;
      }

      realize();
      
      FrameGuiImpl frame = getFrame();
      FrameConfig c = frame.config();
      int parentId = c.parentId;
      
      // update owner if changed
      if (parentId != getOwner().getId().asInt())
      {
         // the parent has changed since the frame has been created
         TopLevelWindow<GuiOutputManager> owner = WindowManager.findWindow(parentId);
         if (owner == null || !owner.isVisible())
         {
            // if frame's parent is not visible, use the OS-active window or DEFAULT-WINDOW
            owner = WindowManager.topLevelWindow();
            if (owner == null || !owner.isVisible())
            {
               owner = WindowManager.getDefaultWindow();
            }
         }
         
         if (owner == null)
         {
            throw new IllegalArgumentException("No valid owner window was resolved.");
         }
         
         setOwner(owner);
         c.windowID = parentId;
      }
      
      TopLevelWindow<?> owner = getOwner();
      gd.selectWindow(owner.getId().asInt());
      NativePoint ownerLocation = gd.getWindowLocation();
      gd.releaseWindow();

      // set location, if location not assigned, (BaseConfig.INV_COORD) the
      // parent class will take care of centering the window
      Point point = ConfigHelper.getAssignedLocation(frame.config());
      NativePoint nloc = screen().coordinates().pointToNative(point);
      if (nloc.x != BaseConfig.INV_COORD)
      {
         nloc.x += ownerLocation.x;
      }
      if (nloc.y != BaseConfig.INV_COORD)
      {
         nloc.y += ownerLocation.y;
      }
      setPhysicalLocation(nloc.x, nloc.y);
      
      getContentPane().setSize(frame.dimension());
      getContentPane().doLayout();
      doLayout();

      super.show();
   }
   
   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   event
    *          Event to process.
    *          
    * @throws  IllegalStateException
    *          If window MENUBAR contains wrong widget.
    */
   @Override
   public void processEvent(Event event)
   {
      // filter the default button related events
      if (event instanceof KeyInput && !(event.source() instanceof Frame))
      {
         Widget<?> src = event.source();
         Frame<?> frame = (Frame<?>) UiUtils.locateFrame(src);
         if (frame != null && frame.processDefaultButton((KeyInput) event))
         {
            return;
         }
      }
      
      super.processEvent(event);
   }

   /**
    * The method returns the window's relative z-order. The value is used to compare the expected
    * z-order between two windows and serves as the primary key for the final windows z-ordering.
    *
    * @return  The ordinal value that represents the window's relative z-order.
    */
   @Override
   public int getZOrderScore()
   {
      GuiWindow owner = (GuiWindow) getOwner();
      if (owner == null)
      {
         owner = (GuiWindow) window();
      }

      int res = owner.getZOrderScore();

      if (owner instanceof DialogBoxWindow)
      {
         // the case when multiple dialog boxes are stacked
         return res;
      }
      else
      {
         // one dialog box with non-modal owner
         return res + 1;
      }
   }

   /**
    * Get the native size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWorkspaceNativeSize()
   {
      NativeDimension wsSizeRes = super.getWorkspaceNativeSize();
      // the dialog box height takes into account the theme dependent title height
      wsSizeRes.height += ThemeManager.getCurrentTheme().getTitleBarHeight(false);
      return wsSizeRes;
   }

   /**
    * Returns the dynamic layout widget integration implementation.
    *
    * @return  see above
    */
   @Override
   public DynamicLayoutIntegration<GuiOutputManager> getDynamicLayout()
   {
      return dynamicLayout;
   }

   /**
    * Get the color resolver.
    * 
    * @return   <code>null</code> by default.
    */
   @Override
   protected GuiColorResolver getColorResolver()
   {
      return gc;
   }
}