/*
** Module   : EditorGuiImpl.java
** Abstract : EDITOR widget implementation. GUI-specific version.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20150509 Created initial version.
** 002 HC  20150612 Improved widget registration/deregistration.
** 003 HC  20150626 Javadoc fixes.
** 004 CA  20150616 Finished support for the missing functionalities (selection, scrolling, keys).
** 005 CA  20150730 The caret can be drawn outside of window's draw call - so ensure the widget's
**                  window is selected before drawing it.
** 006 CA  20150818 Fixed refresh() - it must draw from origin, as it can be called from either 
**                  the event processing queue or outside of it (i.e. mouse selection, 4GL editor 
**                  method call, etc). 
** 007 HC  20150831 Scrolling support overhaul.
** 008 CA  20150910 Added mouse actions. Scroll fixes.
** 009 IAS 20150924 AUTO-RESIZE and TOOLTIP attributes support
** 010 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 011 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 012 SBI 20151014 Added to notify the client the current selection is changed.
** 013 EVL 20151104 Fix for '\r' line separators in incoming character data.  All these characters
**                  are considering the same way as '\n' in GUI mode.
**     HC  20151112 Implemented THREE-D runtime support.
** 014 IAS 20151214 LOAD-MOUSE-POINTER support.
** 015 CA  20160216 Cache the intermediate widget states.  Improved content parsing.
** 016 VIG 20160127 Fixed abend during double-click inside empty Editor (HC change).
**                  Added isRMBEntrySupported() method, added system popup support.
**                  Added double call check and changed showPopupMenu call in mouseClick().
** 017 EVL 20160419 Adding more effects handling for editor in 3d mode.
** 018 SVL 20160412 Javadoc fix.
** 019 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 020 SBI 20160614 Changed to notify the client the current selection is invalid.
** 021 HC  20160628 Fixed calculation of EDITOR size, when inner-lines and inner-chars were
**                  specified during widget definition the editor's size came up as zero.
** 023 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 024 CA  20161005 When EDITOR:SCREEN-VALUE is set, the window must be selected to properly 
**                  compute the text metrics.
** 025 CA  20161101 Added support for mnemonic processing.
** 026 OM  20170426 Added theme support. Fixed rendering and behaviour.
** 027 OM  20170505 Added support for MOVE-TO-EOF, SEARCH and INSERT-FILE.
** 028 EVL 20170824 Adding file list drag-and-drop support.  Fix for unicode aware string
**                  processing on getting width.
** 029 CA  20171111 Added WIDGET:BGCOLOR-RGB and WIDGET:FGCOLOR-RGB FWD extension attributes
**     OM  20171121 Forced window selection in computeLineWidths().
** 030 OM  20171201 Fixed regression from H029.
** 031 SBI 20180130 Fixed the mouse text selection for the web client to invalidate selection on 
**                  the mouse pressed event only.
** 032 SBI 20180209 Fixed double click word selection.
** 033 EVL 20180326 Fix for window request for given widget.  We need the actual window the
**                  current widget belongs too.  It is very important for dialog box window.
** 034 EVL 20180413 Fix for usage negative inner lines value in mouse click handler for detecting
**                  the current line.
** 035 HC  20180420 Implemented popup menu resolution in the tree of widgets.
** 036 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 037 HC  20180530 Fixed the widget stops responding to mouse clicks after its popup menu
**                  displayed. This is second attempt to resolve this issue.
** 038 CA  20180615 Do not draw the caret if we are still in invalidation mode (the root drawing
**                  bracket is not finished yet).
** 039 SBI 20180725 Fixed IndexOutOfBoundsException if the current cursor becomes hidden due to
**                  the editor content has been scrolled, fixed adjustCursorX.
**     SBI 20180726 Fixed EditorContent.draw to use a correct horizontal offset, fixed the cursor
**                  position to be unchanged if the text is scrolled by the thumb button,
**                  changed repaint() to clear the editor lines model if the font has been changed.
**     SBI 20180731 Fixed  computeLineWidths(int from, int to) in order to get native metrics for
**                  the editor's text lines.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

//this usage of AWT events is intentional, they are being used separately from the
//AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.*;
import com.goldencode.p2j.ui.client.gui.ScrollPaneGuiImpl.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implementation of Progress EDITOR component for GUI.
 */
public class EditorGuiImpl
extends Editor<GuiOutputManager>
implements AutoResizable<EditorConfig>,
           LightweightTextWidget
{
   /** Screen driver */
   private final GuiDriver gd;

   /** Output manager */
   private final GuiOutputManager screen;
   
   /** Coordinate converter */
   private final CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;

   /** Font resolver */
   private GuiFontResolver gf;

   /** The scrollable container to place inside the ScrollPane. */
   private EditorScrollContainer contentPane = null;
   
   /** Scroll pane which is actual container. */
   private ScrollPaneGuiImpl editorScroll = null;

   /** Editor border. */
   private BorderedPanel<GuiOutputManager> editorBorder = new BorderedPanelGuiImpl();
   
   /** The editor content, where the text is actually drawn. */
   private final EditorContent editor = new EditorContent();
   
   /** Offset when scrolling horizontally via the scrollbar (the cursor does not move). */
   private int barScrollX = 0;
   
   /** Offset when scrolling vertically via the scrollbar (the cursor does not move). */
   private int barScrollY = 0;
   
   /** The location of first selected character (regardless of where the cursor is). */
   private NativePoint selectionStart = null;
   
   /** The location of last selected character (regardless of where the cursor is). */
   private NativePoint selectionEnd = null;
   
   /** Instance which manages the editor's UNDO/CTRL-Z. */
   private final UndoableEditor undo;
   
   /** Cached border instance. */
   private Border<GuiOutputManager> border;
   
   /**
    * Indexes where the content must be split.  These may be only for a sub-set of lines, if the
    * content is not parsed fully.
    */
   private int[] lineSplitIndexes = null;
   
   /**
    * It holds indexes from {@link #lineSplitIndexes}, for which the split was caused by a 
    * new-line character.
    */
   private Set<Integer> newLineIndexes = null;
   
   /** Total size of the editor border 1 pixel on each side. */
   private static final int BORDER_SIZE = 2;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public EditorGuiImpl(WidgetId id, EditorConfig cfg)
   {
      super(id, cfg);

      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set default colors
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      undo = new UndoableEditor(this);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, EditorConfig cfg)
   {
      super.initialize(id, cfg);

      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_FILLIN_BG, SysColor.COLOR_FILLIN_FG);
      // box and title colors
      gc.boxColor = gc.sysFgColor;

      gf = new GuiFontResolver(resolveWindow(), gd, config(), false);
      
      adjustSize(true);
      
      contentPane = new EditorScrollContainer();
      
      WidgetFactory<GuiOutputManager> factory = screen().getFactory();

      editorScroll = (ScrollPaneGuiImpl) factory.createScrollPane(contentPane);
      editorScroll.setBorder(null);
      ((BorderedPanelGuiImpl) editorScroll.getRootPanel()).setColorResolver(gc);
      add(editorScroll);
      
      // register scroll pane and its widgets for mouse events
      WidgetRegistry<GuiOutputManager> reg = screen().getRegistry();
      reg.addWidget(editorScroll);
      reg.addWidgetsRecursive(editorScroll.widgets());
      
      editorBorder.add(editor);
      ((BorderedPanelGuiImpl) editorBorder).setColorResolver(gc);
      
      contentPane.add(editorBorder);
      
      processScrollPane();
   }
   
   /**
    * Set focus to first focusable widget.  For editors, to not allow the scrollbar buttons to
    * receive focus, will always return this widget.
    */
   @Override
   public void firstFocus()
   {
      setFocusInt(this);
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return processMnemonicImpl(key, this);
   }
   
   /**
    * Process KeyEvents for this widget.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      int key = ke.actionCode();
      
      boolean invalidateSelection = true;
      NativePoint crtCaret = new NativePoint(absoluteX(), absoluteY());
      boolean setSelEnd = false;
      boolean setSelStart = false;
      
      try
      {
         // no matter the key, reset the pseudo-scrolling via bars
         barScrollX = 0;
         barScrollY = 0;
         
         // check for selection first
         switch (key)
         {
            case Key.VK_CTRL_SHIFT_CURSOR_RIGHT:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionStart = crtCaret;
                  setSelEnd = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Key.VK_CTRL_CURSOR_RIGHT);
               ke.setActionCode(Key.VK_CTRL_CURSOR_RIGHT);
               break;
            
            case Key.VK_CTRL_SHIFT_CURSOR_LEFT:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionEnd = crtCaret;
                  setSelStart = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Key.VK_CTRL_CURSOR_LEFT);
               ke.setActionCode(Key.VK_CTRL_CURSOR_LEFT);
               break;
            
            case Key.VK_SHIFT_CURSOR_RIGHT:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionStart = crtCaret;
                  setSelEnd = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_CURSOR_RIGHT);
               ke.setActionCode(Keyboard.KA_CURSOR_RIGHT);
               break;
            
            case Key.VK_SHIFT_CURSOR_LEFT:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionEnd = crtCaret;
                  setSelStart = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_CURSOR_LEFT);
               ke.setActionCode(Keyboard.KA_CURSOR_LEFT);
               break;
            
            case Key.VK_SHIFT_CURSOR_UP:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionEnd = crtCaret;
                  setSelStart = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_CURSOR_UP);
               ke.setActionCode(Keyboard.KA_CURSOR_UP);
               break;
            
            case Key.VK_SHIFT_CURSOR_DOWN:
               if (isSelectionValid())
               {
                  // selection gets expanded or shrunk (depending on where the caret is positioned)
                  setSelEnd = selectionEnd.equals(crtCaret);
                  setSelStart = selectionStart.equals(crtCaret);
               }
               else
               {
                  selectionStart = crtCaret;
                  setSelEnd = true;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_CURSOR_DOWN);
               ke.setActionCode(Keyboard.KA_CURSOR_DOWN);
               break;
            
            case Key.VK_SHIFT_HOME:
               setSelStart = true;
               if (isSelectionValid())
               {
                  selectionEnd =
                     (selectionStart.equals(crtCaret) ? selectionEnd : selectionStart);
               }
               else
               {
                  selectionEnd = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_HOME);
               ke.setActionCode(Keyboard.KA_HOME);
               break;
            
            case Key.VK_SHIFT_END:
               setSelEnd = true;
               if (isSelectionValid())
               {
                  selectionStart =
                     (selectionEnd.equals(crtCaret) ? selectionStart : selectionEnd);
               }
               else
               {
                  selectionStart = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_END);
               ke.setActionCode(Keyboard.KA_END);
               break;
            
            case Key.VK_CTRL_SHIFT_HOME:
               setSelStart = true;
               if (isSelectionValid())
               {
                  selectionEnd =
                     (selectionStart.equals(crtCaret) ? selectionEnd : selectionStart);
               }
               else
               {
                  selectionEnd = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Key.VK_CTRL_HOME);
               ke.setActionCode(Key.VK_CTRL_HOME);
               break;
            
            case Key.VK_CTRL_SHIFT_END:
               setSelEnd = true;
               if (isSelectionValid())
               {
                  selectionStart =
                     (selectionEnd.equals(crtCaret) ? selectionStart : selectionEnd);
               }
               else
               {
                  selectionStart = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Key.VK_CTRL_END);
               ke.setActionCode(Key.VK_CTRL_END);
               break;
            
            case Key.VK_SHIFT_PAGE_UP:
               setSelStart = true;
               if (isSelectionValid())
               {
                  selectionEnd =
                     (selectionStart.equals(crtCaret) ? selectionEnd : selectionStart);
               }
               else
               {
                  selectionEnd = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_PAGE_UP);
               ke.setActionCode(Keyboard.KA_PAGE_UP);
               break;
            
            case Key.VK_SHIFT_PAGE_DOWN:
               setSelEnd = true;
               if (isSelectionValid())
               {
                  selectionStart =
                     (selectionEnd.equals(crtCaret) ? selectionStart : selectionEnd);
               }
               else
               {
                  selectionStart = crtCaret;
               }
               
               invalidateSelection = false;
               ke.setKeyCode(Keyboard.KA_PAGE_DOWN);
               ke.setActionCode(Keyboard.KA_PAGE_DOWN);
               break;
         }
         
         if (key == Keyboard.KA_COPY || key == Keyboard.KA_CUT)
         {
            if (key == Keyboard.KA_CUT && config.readOnly)
            {
               return; // cannot CUT in a readonly editor
            }
            
            if (isSelectionValid())
            {
               boolean remove = (key == Keyboard.KA_CUT);
               NativePoint saveSelStart = selectionStart;
               String selectedText = getSelectedText(remove);

               if (remove)
               {
                  undo.recordDeletedText(key, saveSelStart, selectedText);
               }
               
               // copy selection
               gd.copyToClipboard(selectedText);
            }
            
            if (key == Keyboard.KA_CUT)
            {
               refresh(); // refresh only if text has changed, not on COPY action
            }
            return;
         }
         else if (key == Keyboard.KA_PASTE)
         {
            if (config.readOnly)
            {
               return; // cannot paste in a readonly editor
            }
            
            if (isSelectionValid())
            {
               // clear everything first
               getSelectedText(true); // TODO: OM: from my tests, if some text is already selected,
                                      // on PASTE, the new new text is inserted after teh selection,
                                      // without removing selected text first 
            }
            
            // insert the text
            String txt = gd.clipboardContents();
            int offset = getCursorOffset() - 1;
            int nextOffset = offset + txt.length() + 1;
            
            int currentLine = absoluteY();
            String content = resolveContent();
            StringBuilder newText = new StringBuilder();
            if (offset > 0 && content.length() > 0)
            {
               newText.append(content.substring(0, offset));
            }
            newText.append(txt);
            if (content.length() > 0)
            {
               newText.append(content.substring(offset));
            }
            
            parseContent(newText.toString(), currentLine, false);
            
            setCursorOffset(nextOffset);
            
            return;
         }
         else if (key == Keyboard.SE_LEAVE || key == Keyboard.SE_ENTRY)
         {
            // do not invalidate selection on focus gain or lost
            invalidateSelection = false;
         }
         
         // get this again, so we can re-apply the event
         key = ke.actionCode();
         
         boolean processUndo = true;

         // anything from down here invalidates the selection
         if (invalidateSelection)
         {
            if (isSelectionValid())
            {
               boolean explicitRemove = (key == Key.VK_CTRL_DEL              ||
                                         key == Keyboard.KA_BACKSPACE        ||
                                         key == Keyboard.KA_DELETE_CHARACTER);
               boolean printableKey = Keyboard.isPrintableKey(key);

               // any printable + backspace + delete removes the selection
               if ((explicitRemove || printableKey)  && !config.readOnly)
               {
                  NativePoint saveSelStart = selectionStart;
                  String selectedText = getSelectedText(true);
                  
                  // when saving selection, if is deleted via a printable key, then assume we are
                  // using a DELETE key
                  undo.recordDeletedText((printableKey ? Keyboard.KA_DELETE_CHARACTER : key),
                                         saveSelStart,
                                         selectedText);
                  
                  // let undo processing execute only if selection is invalidated by a printable
                  // key - so that the printable key gets recorded; otherwise, avoid processing
                  // the "remove text" key twice
                  processUndo = printableKey;
               }
               
               invalidateSelection();
               
               if (explicitRemove)
               {
                  // nothing else to do - the key is consumed
                  if (!config.readOnly)
                  {
                     refresh();
                  }
                  return;
               }
            }
         }
         
         if (processUndo && !config.readOnly)
         {
            // check for undo processing
            undo.processKey(key);
         }
         
         // text navigation events
         if (key == Keyboard.KA_HOME)
         {
            cursorBeginOfLine();
            
            refresh();
            return;
         }
         else if (key == Keyboard.KA_END)
         {
            cursorEndOfLine();
            
            refresh();
            return;
         }
         else if (key == Key.VK_CTRL_HOME)
         {
            config.scrollX = 0;
            config.scrollY = 0;
            config.cursorY = 0;
            config.cursorX = 0;
            
            refresh();
            return;
         }
         else if (key == Key.VK_CTRL_END)
         {
            if (parsedText.size() > config.innerLines)
            {
               config.scrollY = parsedText.size() - config.innerLines;
               config.cursorY = config.innerLines - 1;
            }
            else
            {
               config.scrollY = 0;
               config.cursorY = lastRow();
            }
            
            cursorEndOfLine();
            
            refresh();
            return;
         }
         else if (key == Key.VK_CTRL_CURSOR_RIGHT)
         {
            wordRight();
            refresh();
            return;
         }
         else if (key == Key.VK_CTRL_CURSOR_LEFT)
         {
            wordLeft();
            refresh();
            return;
         }
         else if (key == Key.VK_CTRL_DEL)
         {
            if (!config.readOnly)
            {
               deleteToEndOfLine();
               refresh();
            }
            return;
         }
         else if (key == Key.VK_CTRL_ENTER)
         {
            ke.setActionCode(Keyboard.KA_RETURN);
         }
         
         super.processKeyEvent(ke);
      }
      finally
      {
         // if we haven't invalidated the selection, check the new position - everything
         // between selectionStart and current location must be selected
         if (!invalidateSelection)
         {
            NativePoint newCaretLocation = new NativePoint(absoluteX(), absoluteY());
            
            // re-adjust the selection start/end
            if (setSelStart)
            {
               selectionStart = newCaretLocation;
            }
            if (setSelEnd)
            {
               selectionEnd = newCaretLocation;
            }
            
            adjustSelection();
            
            // do not call refresh here, just post a repaint
            repaint();
         }
         
         setScrollPosition();
      }
   }
   
   /**
    * Sets the new variable value in the editor.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (editorScroll != null)
      {
         processScrollPane();
      }

      TopLevelWindow<GuiOutputManager> wnd = parentOrSelf(TopLevelWindow.class);
      if (wnd == null)
      {
         // TODO: this is incorrect: during initialize, we can't parse the default content, as
         // the widget is not yet attached to a frame (and the frame may not have been attached
         // to a window)
         wnd = WindowManager.getDefaultWindow();
      }
      
      int wid = wnd.getId().asInt();
      screen.selectWindow(wid);
      super.setValue(value);
      screen.releaseWindow();
   }
   
   /**
    * Sets position of terminal's cursor unconditionally (even if this editor is not in focus).
    */
   @Override
   public void drawCaretUnconditionally()
   {
      if (!isEnabled())
      {
         return;
      }

      if (ThinClient.getInstance().getOutputManager().isInvalidate())
      {
         // if this is called from outside of the actual driver drawing mechanism, ignore it.
         return;
      }

      // check if the cursor is visible
      String currentTextLine = getScreenText();
      
      int endOfLineCursor = currentTextLine.length();
      
      int scrollColumn = getScrollColumnOffset();
      
      int cursorColumn = getCursorColumnOffset();
      
      if (endOfLineCursor < scrollColumn || cursorColumn < scrollColumn)
      {
         // the current cursor is not visible
         return;
      }

      // because of scrolling, we need first to establish the editor's clipping rectangle, and
      // after this draw the caret
      
      NativePoint origin = new NativePoint(editor.screenPhysicalLocation());
      NativeDimension d = editor.physicalDimension();
      NativeRectangle clip = clipRectangle(origin, d);

      // ensure we are on the correct window, as this can be called outside of window's draw
      TopLevelWindow<?> window = topLevelWindow();
      gd.selectWindow(window.getId().asInt());
      gd.setGuiFont(gf.font().font);
      
      gd.drawFromOrigin(origin, clip, () -> gd.drawCursor(this));
      
      gd.releaseWindow();
   }

   /**
    * Render this editor.
    * 
    * @param   color
    *          Color for drawing. Ignored.
    */
   @Override
   public void draw(Color color)
   {
      if (!isVisible())
      {
         return;
      }
      
      // setup the border, do this only once when the widget has been realized
      if (border == null)
      {
         if (config.box)
         {
            border = isThreeD() ? new ThreeDBorderGuiImpl() : new LineBorderGuiImpl();
         }
         else
         {
            border = new EmptyBorder<>();
         }
         
         editorBorder.setBorder(border);

         // editor widget can have more 3d effects to draw than other widgets
         // especially when the scrollbars are defined
         if (config.box && isThreeD())
         {
            ((BorderedPanelGuiImpl) editorScroll.getRootPanel()).set3DBorder(true);
         }
         
         int bgColor = isThreeD() && !(isEnabled() && !config.readOnly) ? 
                       SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
         gc.sysBgColor = gd.getSysColor(bgColor);
      }

      // TODO: explicit fg/bg colors do not work!
      
      // determine frame colors
      gc.colors();
      
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(getId(), p, clip, () ->
      {
         gd.setGuiFont(gf.font().font);
         
         // clear the area
         gd.setColor(gc.bgColor);
         
         editorScroll.draw();
      });
      
      if (widgetFocused && isEnabled())
      {
         drawCaret();
      }
   }

   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      super.destroy();
      
      if (contentPane != null)
      {
         contentPane.destroy();
      }
      
      if (editorScroll != null)
      {
         editorScroll.destroy();
      }
   }
   
   /**
    * Obtain any selected text in the widget for copying to the clipboard.
    *
    * @return   The selected text or <code>null</code> if there is no selected text or if the
    *           widget type does not support clipboard operations.
    */
   @Override
   public String getSelectedText()
   {
      return getSelectedText(false);
   }
   
   /**
    * Enable/disable the widget for editing.  During this operation, any data values are saved off 
    * for subsequent restore (during editing) using the <code>RECALL</code> key.
    * <p>
    * In GUI, it registers this widget with the driver, to allow responsive actions when the
    * mouse is hovered over the widget.
    * 
    * @param    enable
    *           <code>true</code> to enable editing in this widget.
    */
   @Override
   public void setEnabled(boolean enable)
   {
      // WARNING! Do not rely on this API to post PaintEvents - when ENABLE stmt is used, the
      // "enabling" is done in a separate event processing loop, which consumes the PaintEvents.
      super.setEnabled(enable);

      // propagate this to all widgets
      if (contentPane != null)
      {
         editorScroll.setEnabled(enable);
         editor.setEnabled(enable);
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If the editor is enabled, it will position the caret.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      if (e.getButton() != getMenuMouse())
      {
         invalidateSelection();
      }
      
      if (!hasFocus())
      {
         // process the focus here, so we can continue with the special processing bellow.
         ThinClient.getInstance().independentEventDrawingBracket(this, this::requestFocus);
      }
      
      NativePoint editorLoc = editor.screenPhysicalLocation();
      NativeDimension editorDim = editor.physicalDimension();
      int nrow = e.getY() - editorLoc.y;
      int ncol = e.getX() - editorLoc.x;
      if (ncol < 0)
      {
         ncol = 0;
      }
      else if (ncol > editorDim.width)
      {
         ncol = editorDim.width;
      }
      
      if (nrow < 0)
      {
         nrow = 0;
      }
      else if (nrow > editorDim.height)
      {
         nrow = editorDim.height;
      }
      
      int nfontHeight = gf.font().fontHeight;
      int editRow = nrow / nfontHeight;
      
      int prevY = absoluteY();
      int prevX = absoluteX();
      
      config.scrollY = config.scrollY + barScrollY;
      config.cursorY = Math.min(editRow, parsedText.size() - 1);
      if (config.innerLines > 1)
      {
         config.cursorY = Math.min(config.cursorY, config.innerLines - 1);
      }
      
      Line line = getLine();
      if (line != null)
      {
         adjustCursorX(prevY, prevX, ncol + computeHorizontalScrollDelta(barScrollX));
      }
      
      barScrollX = 0;
      barScrollY = 0;
      
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
      
      setScrollPosition();
      repaint();
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If the editor is enabled and has focus, it will re-adjust the selected text depending on
    * where the mouse is moving. 
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (!isEnabled() || !hasFocus())
      {
         return;
      }

      NativePoint prevCaret = new NativePoint(absoluteX(), absoluteY());
      NativePoint prevStart = selectionStart;
      NativePoint prevEnd = selectionEnd;

      // check for scrolling first; if mouse is outside N/S/E/W border, then scroll in that
      // direction
      NativeRectangle bounds = new NativeRectangle(editor.screenPhysicalLocation(), 
                                                   editor.physicalDimension());
      boolean outside = false;
      if (e.getX() < bounds.left())
      {
         outside = true;

         // scroll left
         if (config.scrollX > 0)
         {
            config.scrollX--;
            config.cursorX = 0;
         }
      }
      if (e.getX() > bounds.right())
      {
         outside = true;
         if (absoluteX() < getLine().length() - 1)
         {
            cursorRight();
         }
      }
      if (e.getY() < bounds.top())
      {
         outside = true;

         // scroll up
         if (config.scrollY > 0)
         {
            config.scrollY--;
            config.cursorY = 0;
         }
      }
      if (e.getY() > bounds.bottom())
      {
         outside = true;

         // scroll down
         int lastLine = parsedText.size() - config.innerLines;
         if (config.scrollY < lastLine)
         {
            config.scrollY++;
            config.cursorY = config.innerLines - 1;
         }
      }
      
      if (!outside)
      {
         // assume a click and just reposition the cursor
         mouseClicked(e);
      }
      else
      {
         // get rid of pseudo-scrolling and move the cursor to the new position
         barScrollX = 0;
         barScrollY = 0;
         
         config.cursorChar = getCursorChar();
         config.cursorOffset = getCursorOffset();

         refresh();
      }
      
      NativePoint currentCaret = new NativePoint(absoluteX(), absoluteY());
      
      selectionStart = prevStart;
      selectionEnd = prevEnd;
      
      if (selectionStart == null)
      {
         selectionStart = prevCaret;
      }
      else if (prevCaret.equals(selectionStart))
      {
         selectionStart = currentCaret;
      }
      
      if (selectionEnd == null)
      {
         selectionEnd = currentCaret;
      }
      else if (prevCaret.equals(selectionEnd))
      {
         selectionEnd = currentCaret;
      }
      
      adjustSelection();
      setScrollPosition();
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (e.getButton() == getMenuMouse())
      {
         // prevent double popup processing.
         if (e.getID() != MouseEvent.MOUSE_CLICKED)
         {
            return;
         }
      }

      super.mouseClicked(e);

      // when popup-menu is opened due to this event, do not proceed further
      if (e.isConsumed())
      {
         return;
      }

      if (!isEnabled())
      {
         return;
      }

      NativePoint editorLoc = editor.screenPhysicalLocation();
      NativeDimension editorDim = editor.physicalDimension();
      int nrow = e.getY() - editorLoc.y;
      int ncol = e.getX() - editorLoc.x;
      if (ncol < 0)
      {
         ncol = 0;
      }
      else if (ncol > editorDim.width)
      {
         ncol = editorDim.width;
      }
      
      if (nrow < 0)
      {
         nrow = 0;
      }
      else if (nrow > editorDim.height)
      {
         nrow = editorDim.height;
      }
      
      int nfontHeight = gf.font().fontHeight;
      int editRow = nrow / nfontHeight;
      
      int prevY = absoluteY();
      int prevX = absoluteX();
      
      config.scrollY = config.scrollY + barScrollY;
      config.cursorY = Math.min(editRow, parsedText.size() - 1);
      config.cursorY = Math.min(config.cursorY, config.innerLines - 1);

      Line line = getLine();
      if (line != null)
      {
         adjustCursorX(prevY, prevX, ncol + computeHorizontalScrollDelta(barScrollX));
      }

      barScrollX = 0;
      barScrollY = 0;
      
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();

      setScrollPosition();
      
      int contentLength = contentLength();
      
      // if this is a double-click, select the current word and its right-side spaces
      if (e.getClickCount() == 2 && contentLength > 0)
      {
         int cursorOffset = getCursorOffset() - 1;
         int[] limits = StringHelper.getWordWithRightSpacesAt(resolveContent(), cursorOffset);
         setSelection(limits[0] + 1, limits[1] + 1);
         // the cursor must follow text selection
         config.cursorX = selectionEnd.x;
      }
      
      repaint();
   }

   /**
    * Notification of a mouse wheel event occurred for this widget.
    * <p>
    * If the editor is vertically scrollable, enabled and has focus, it will scroll the content
    * vertically or horizontally.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (!config.scrollbarVertical || !isEnabled() || !hasFocus())
      {
         return;
      }

      int scroll = e.getWheelRotation();
      if (scroll == 0)
      {
         return;
      }
      
      if (scroll < 0)
      {
         if (config.scrollY + barScrollY + scroll >= 0)
         {
            barScrollY = barScrollY + scroll;
         }
      }
      else
      {
         int lastFirstRow = parsedText.size() - config.innerLines;
         if (config.scrollY + barScrollY + 1 <= lastFirstRow)
         {
            barScrollY = barScrollY + 1;
         }
      }

      contentPane.postScrollEvent(null, config.scrollY + barScrollY);
   }
   
   /**
    * Returns the size of the scrollable area of this widget in character
    * units. 
    * <p>
    * The scrollable area is the total area including the hidden parts 
    * that must be scrolled in order to be visible. 

    * @return   See above.
    */
   @Override
   public Dimension getScrollableSize()
   {
      return editor.dimension();
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           The previous configuration.
    */
   @Override
   public void afterConfigUpdate(EditorConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);

      // buffer-chars must be greater or equal to visible area width
      if (config.bufferChars < config.innerChars)
      {
         config.bufferChars = config.innerChars;
      }
      
      // buffer-lines, if specified, must be at least equal to visible area height 
      if (config.bufferLines > 0 && config.bufferLines < config.innerLines)
      {
         config.bufferLines = config.innerLines;
      }

      if (editorScroll == null)
      {
         return;
      }
      
      processScrollPane();
   }

   /**
    * Sets the cursor position in the current line
    *
    * @param    pos
    *           The new cursor position value.
    * @param    redraw
    *           request redraw/sync
    *           
    * @return   <code>true</code> if the cursor position is valid and set 
    */
   @Override
   public boolean setCursorChar(int pos, boolean redraw)
   {
      // get current line
      Line line = getLine();
      int lineLength = line.length();
      
      if (pos < 0 || pos > lineLength)
      {
         // go to end of line
         cursorEndOfLine();
         return false;
      }
      
      if (pos < config.scrollX)
      {
         String txt = line.getText();
         if (pos < txt.length())
         {
            txt = txt.substring(0, pos);
         }
   
         // scroll to make this position visible
         int maxWidth = editor.physicalDimension().width;
         int foundPos = -1;
         for (int i = 0; i < txt.length(); i++)
         {
            int txtWidth = getTextWidthNative(line, i, txt.length(), gf.resolveFontNum());
            
            if (txtWidth <= maxWidth)
            {
               break;
            }

            foundPos = i + 1;
         }
         
         if (foundPos >= 0)
         {
            txt = txt.substring(foundPos);
         }
         
         config.scrollX = pos - txt.length() + 1; // scroll what is made hidden
         config.cursorX = txt.length();
      }
      else
      {
         // do not scroll, just adjust the X
         config.cursorX = pos - config.scrollX;
      }
      
      return super.setCursorChar(pos, redraw);
   }
   
   /**
    * Removes (deletes) the selected text.
    */
   public void removeSelectedText()
   {
      // replacing the selected test with empty string has the same effect like removing it
      replaceSelectionText("");
   }
   
   /**
    * Clears all cached line widths.
    */
   private void clearLinesWidths()
   {
      if (parsedText == null || parsedText.isEmpty())
      {
         return;
      }
      
      for(Line line : parsedText)
      {
         line.setWidths(null);
      }
   }
   
   /**
    * Updates the line widths if the current font has been changed and repaint.
    */
   @Override
   public void repaint()
   {
      if (gf != null && gf.fontChanged)
      {
         clearLinesWidths();
      }
      
      super.repaint();
      
      if (gf != null && gf.fontChanged)
      {
         gf.fontChanged = false;
      }
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public boolean setSelection(int start, int end)
   {
      int contentLength = contentLength();
      if (start <= 0 || start > contentLength)
      {
         start = contentLength + 1;
      }
      if (end <= 0 || end > contentLength)
      {
         end = contentLength + 1;
      }
      
      //fixed the cursor line position on text selection
      config.selectionEnd = (end > start) ? end : start;
      
      super.setSelection(start, end);
   
      if (start == end)
      {
         invalidateSelection();
         return true;
      }

      int istart = Math.min(start, end);
      int iend = Math.max(start, end);

      start = istart;
      end = iend;
      
      // adjust to be 0-based instead of 1-based
      start--;
      end--;
   
      selectionStart = null;
      selectionEnd = null;
      for (int y = 0; 
           y < parsedText.size() && !(selectionStart != null && selectionEnd != null); 
           y++)
      {
         Line line = parsedText.get(y);
         int rightOffset = line.getOffset() + line.length();
         
         if (selectionStart == null && rightOffset > start)
         {
            selectionStart = new NativePoint(start - line.getOffset(), y);
         }
         
         if (selectionEnd == null && rightOffset >= end)
         {
            selectionEnd = new NativePoint(end - line.getOffset(), y);
         }
      }
      
      config.selectionText = resolveContent().substring(start, end);
      
      // notify the selection is changed
      gd.setCurrentSelection(config.selectionText);
      
      refresh();
      return true;
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    * The search is performed on each editor's line.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>EditorWidget.FIND_NEXT_OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>EditorWidget.FIND_PREV_OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>EditorWidget.FIND_CASE_SENSITIVE</td><td>4</td></tr>
    *              <tr><td>EditorWidget.FIND_WRAP_AROUND</td><td>16</td></tr>
    *              <tr><td>EditorWidget.FIND_SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @Override
   public Boolean search(String pattern, int mode)
   {
      int selStart = config.selectionStart;
      int selEnd = config.selectionEnd;
      invalidateSelection();  // the current selection gets invalidated 
      
      if (pattern == null || pattern.isEmpty() ||
          pattern.contains("\n") || pattern.contains("\r")||
          parsedText == null || parsedText.size() == 0)
      {
         // if pattern is null or empty or has multiple lines, P4GL will fail to find it
         return false;
      }
      
      // FIND_NEXT_OCCURRENCE is default. It also takes precedence if both are set 
      int directionMask = EditorWidget.FIND_PREV_OCCURRENCE | EditorWidget.FIND_NEXT_OCCURRENCE;
      boolean forward = (mode & directionMask) != EditorWidget.FIND_PREV_OCCURRENCE;
      boolean ciSearch = (mode & EditorWidget.FIND_CASE_SENSITIVE) == 0;   // (case insensitive)
      boolean wrap = (mode & EditorWidget.FIND_WRAP_AROUND) == EditorWidget.FIND_WRAP_AROUND;
      boolean select = (mode & EditorWidget.FIND_SELECT) == EditorWidget.FIND_SELECT;
      
      if (ciSearch)
      {
         pattern = pattern.toLowerCase();
      }
      int crtCaretOffset = -1; // used for wrap stop, otherwise ignored!
      
      // start from cursor's position on the current line
      int crtLineNo = absoluteY();   // use cursor's position by default
      int lineOffset = absoluteX();
      int gOffset = getCursorOffset() - 1;
      if (selStart != -1 && selEnd != -1)
      {
         // if we have a selection go right after or right before it
         gOffset = (forward ? selEnd : selStart) - 1;
         for (int k = 0; k < parsedText.size(); k++)
         {
            Line aLine = parsedText.get(k);
            if (aLine.getOffset() <= gOffset && 
                aLine.getOffset() + aLine.getText().length() >= gOffset)
            {
               // found it!
               crtLineNo = k;
               lineOffset = gOffset - aLine.getOffset();
               break;
            }
         }
      }
      
      // prepare the first line to be search
      Line crtLine = parsedText.get(crtLineNo);
      String crtText = crtLine.getText();
      if (!forward)
      {
         ++ lineOffset;
         if (lineOffset > crtText.length())
         {
            lineOffset = crtText.length();
         }
         crtText = crtText.substring(0, lineOffset);
      }
      if (ciSearch)
      {
         crtText = crtText.toLowerCase();
      }
      
      // now we can do the actual search:
      int offset = -1;
      while (crtText != null)
      {
         offset = forward ? crtText.indexOf(pattern, lineOffset)
                          : crtText.lastIndexOf(pattern);
         lineOffset = 0; // reset after 1st usage
         
         if (offset != -1)
         {
            break; // found occurrence, stop looping
         }
         
         if (crtCaretOffset != -1                  &&
             crtLine.getOffset() <= crtCaretOffset &&
             crtLine.getOffset() + crtText.length() >= crtCaretOffset)
         {
            // we wrapped and reached the initial offset. abort
            return false;
         }
         
         // go to next line using the known direction
         if (forward)
         {
            ++crtLineNo;
            if (crtLineNo >= parsedText.size())
            {
               if (wrap)
               {
                  crtCaretOffset = getCursorOffset();
                  crtLineNo = 0;    // jump to first line
                  wrap = false;     // block multiple loops
               }
               else
               {
                  setCursorOffset(gOffset);
                  return false;
               }
            }
         }
         else // backwards
         {
            --crtLineNo;
            if (crtLineNo < 0)
            {
               if (wrap)
               {
                  crtCaretOffset = getCursorOffset();
                  crtLineNo = parsedText.size() - 1;   // jump to last line
                  wrap = false;                        // block multiple loops
               }
               else
               {
                  setCursorOffset(gOffset);
                  return false;
               }
            }
         }
         
         crtLine = parsedText.get(crtLineNo);
         crtText = crtLine.getText();
         if (ciSearch)
         {
            crtText = crtText.toLowerCase();
         }
      }
      
      if (offset == -1)
      {
         return false; // string not found (should not happen)
      }
      
      int newCursorPos = offset + crtLine.getOffset() + 1;
      setCursorOffset(newCursorPos);
      if (select)
      {
         setSelection(newCursorPos, newCursorPos + pattern.length());
      }
      
      return true;
   }
   
   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(EditorConfig beforeUpdate)
   {
      if (config.font != beforeUpdate.font)
      {
         config.widthChars = 0;
         config.heightChars = 0;
         adjustSize(false);
      }
      
      // explicit attribute manipulation
      if (config.innerLines != beforeUpdate.innerLines ||
          config.innerChars != beforeUpdate.innerChars)
      {
         adjustSize(false);
      }
   }

   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * <p>
    * For editors, if the sources is either {@link #editor} or {@link #contentPane}, then return
    * this widget.
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      Widget<GuiOutputManager> src = super.findMouseSource(p);
      
      return (src == editor || src == contentPane) ? this : src;
   }

   /**
    * Get the widget's actual bounds
    * 
    * @return widget's actual bounds
    */
   @Override
   public NativeRectangle getActualBounds()
   {

      Widget scrollPane = WidgetRegistry.findDescendant(this, ScrollPaneGuiImpl.class);
      NativeDimension dim = ((ScrollPaneGuiImpl) scrollPane).viewport().physicalDimension();
      return new NativeRectangle(screenPhysicalLocation(), dim);
   }

   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   public boolean isRMBEntrySupported()
   {
      return false;
   }
   
   /**
    * Checks if the given widget is able to be the target of the drop operation.
    * 
    * @return   <code>TRUE</code> if this widget can be droppable at this time,
    *           <code>FALSE</code> otherwise.
    */
   @Override
   public boolean isDroppable()
   {
      return config.dropTarget; 
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_WHEEL,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvt.MOUSE_HOVERABLE,
      };
   }

   /**
    * Adjust Editor.
    * 
    * @param    init
    *           <code>true</code> if this is called from {@link #initialize}.
    */
   protected void adjustSize(boolean init)
   {
      int border = (config.box ? BORDER_SIZE : 0);
      int nativeWidth = 0;
      int nativeHeight = 0;

      // do we need to update inner chars and lines flag(no need by default)
      boolean explicitSizeSetup = config.widthChars != 0 && config.heightChars != 0;

      if (explicitSizeSetup)
      {
         // for compatibility reasons the insets size preserved here however looks like
         // the insets are always 0 for GUI selection list
         nativeWidth = cc.widthToNative(config.widthChars) -
                       cc.widthToNative(insets().left + insets().right);
         nativeHeight = cc.heightToNative(config.heightChars) -
                        cc.heightToNative(insets().top + insets().bottom);
      }
      else
      {
         int font = fontResolver().resolveFontNum();
         int rowHeight = FontManager.getFontHeight(resolveWindow(), font, gd);
         int rowWidth = FontManager.getFontWidth(resolveWindow(), font, gd);

         // TODO: shouldn't this use the DEFAULT-FONT?
         // in this case we use inner chars and lines option as the base for size calculation
         nativeHeight = config.innerLines * rowHeight + border;
         nativeWidth = config.innerChars * rowWidth + border;
      }
      
      // inner size in pixels (without the border)
      int innerCharsInPixels = nativeWidth - border;
      int innerLinesInPixels = nativeHeight - border;

      // the following two doubles are not the same innerLines >= inCharHeight
      double inCharHeight = screen().coordinates().heightFromNative(nativeHeight);
      double innerLines = cc.heightFromNative(innerLinesInPixels);

      // the following two doubles are not the same innerChars >= inCharWidth
      double inCharWidth = screen().coordinates().widthFromNative(nativeWidth);
      double innerChars = cc.widthFromNative(innerCharsInPixels);
      
      // set new size for GUI selection-list
      setSize(new Dimension(inCharWidth, inCharHeight));

      // update widget config size
      if (config.widthChars == 0.0)
      {
         config.widthChars = inCharWidth;
      }
      if (config.heightChars == 0.0)
      {
         config.heightChars = inCharHeight;
      }
      
      // update widget config inner chars and lines if necessary 
      if (explicitSizeSetup)
      {
         // using floor() guarantees all lines/cells will be completely visible
         if (innerChars <= 0)
            config.innerChars = (int) Math.floor(innerChars);
         if (innerLines <= 0)
            config.innerLines = (int) Math.floor(innerLines);
      }
   }

   /**
    * Refresh the editor.
    */
   @Override
   protected void refresh()
   {
      // this needs to start drawing relative to parent container, as draw() will translate to
      // editor's location
      NativePoint origin = parent().screenPhysicalLocation();
      gd.selectWindow(topLevelWindow().getId().asInt());
      
      gd.drawFromOrigin(origin, null, () ->
      {
         gd.startBatch(); // remove flickering caused by painting individual components
         try
         {
            draw();
            requestSync();
         }
         finally
         {
            gd.endBatch();
         }
      });
      
      repaint();
      
      gd.releaseWindow();
   }
   
   /**
    * Check if the given text needs to output on a new line.
    * 
    * @param    subtext
    *           The text.
    * @param    placeChar
    *           flag indicating if this call is from {@link #placeChar}.
    *
    * @return   <code>true</code> if the text will output on a new line.
    */
   @Override
   protected boolean goToNewLine(String subtext, boolean placeChar)
   {
      if (config.scrollbarHorizontal || !config.wordWrap)
      {
         return false;
      }
      
      final int[] subtextWidth = new int[1]; // the way to change final var inside lambda
      int mw = gf.font().font.legacyMaxWidth * 2;
      int editWidth = editor.physicalDimension().width;

      // getting subtext width into subtextWidth[0]
      ThinClient.getInstance().independentEventDrawingBracket(this, 
            () -> subtextWidth[0] = gd.getTextWidth(subtext, gf.font().font) );
      // summing max width
      mw += subtextWidth[0];
      
      // TODO: in word-wrap mode, this optimization may lead to false-negatives, as the 
      // formula above does not adjust for the kerning space between the text's letters...
      
      if (mw < editWidth)
      {
         // do not go to next line if the text fits in the line
         return false;
      }
      
      // get text width using the current font
      int txtWidth = getTextWidthNative(subtext, gf.resolveFontNum());
      
      return txtWidth > editWidth;
   }
   
   /** 
    * Determine if a blank line needs to be inserted when RETURN key is pressed on a empty line.
    * 
    * @return   Always <code>true</code> for GUI mode.
    */
   @Override
   protected boolean insertBlankLineOnLF()
   {
      return true;
   }

   /**
    * Split absolute horizontal cursor position into two components - position
    * of the first visible character and relative cursor position in the
    * visible region.
    * 
    * @param   pos
    *          The new absolute cursor position.
    */
   @Override
   protected void cursorPos(int pos)
   {
      Line line = getLine();
      config.cursorX = (line == null || line.length() == 0 ? 0 : pos % line.length());

      super.cursorPos(pos);
   }

   /**
    * Merge line with subsequent one(s) if possible
    * 
    * @param   row
    *          line's row
    */
   @Override
   protected void mergeLines(int row)
   {
      /*
      if (true)
      {
         // this is left behind in case debugging is needed in a future point - it can help seeing
         // what the content really is, not just what is drawn on screen.
         String content = resolveContent();
         // parsing the content forces the editor to show what it really holds - so any errors from
         // processing a key are immediately viewed
         parseContent(content, -1, false);
         return;
      }
      */

      int editorWidth = (config.scrollbarHorizontal || !config.wordWrap 
                           ? Integer.MAX_VALUE 
                           : editor.physicalDimension().width);
      int font = gf.resolveFontNum();
      
      computeLineWidths(row, parsedText.size() - 1);
      
      for(; row < parsedText.size() - 1; row++)
      {
         Line current = parsedText.get(row);
         if (current.isNewLine())
         {
            break;
         }
         int currentWidth = getTextWidthNative(current, 0, current.length(), font);

         Line next = parsedText.get(row + 1);
         String s = next.getText();
         int pos = 0;
         while (pos < s.length())
         {
            // try to append text from "s" until current line fits the screen
            
            int len = currentWidth + getTextWidthNative(next, 0, pos + 1, font);
            if (len > editorWidth)
            {
               // previous position was correct.
               pos--;
               break;
            }
            
            pos++;
         }
         
         current = new Line(current.getText() + s.substring(0, Math.min(s.length(), pos + 1)), 
                            current.getOffset(),
                            false);
         parsedText.set(row, current);
         if (pos == s.length())
         {
            if (next.isNewLine())
            {
               current.setNewLine(true);
            }
            
            parsedText.remove(row + 1);
         }
         else
         {
            parsedText.set(row + 1, new Line(s.substring(pos + 1), 
                                             next.getOffset() + pos,
                                             next.isNewLine()));
         }
      }
   }
   
   /**
    * Get the next split position of this line, so that it fits the editor.
    * 
    * @param    line
    *           The line to be split.
    * @param    splitAfter
    *            minimal split position.
    *           
    * @return   An index where the split ends.
    */
   @Override
   protected int getNextSplitPosition(String line, int splitAfter)
   {
      // TODO: optimize
      
      // remove trailing space
      line = StringHelper.safeTrimTrailing(line);
      
      int editorWidth = editor.physicalDimension().width;
      int font = gf.resolveFontNum();

      // split by word first
      int pos = line.lastIndexOf(SPACE, line.length());
      while (pos > 0)
      {
         int width = getTextWidthNative(line.substring(0, pos), font);
         if (width <= editorWidth)
         {
            return pos + 1;
         }
         
         pos = line.lastIndexOf(SPACE, pos - 1);
      }
      
      // no space, determine text
      pos = line.length();
      while (pos >= 0)
      {
         int width = getTextWidthNative(line.substring(0, pos), font);
         if (width <= editorWidth)
         {
            return pos;
         }
         
         pos = pos - 1;
      }
      
      return 0;
   }
   
   /**
    * Replace line, splitting it if necessary
    * 
    * @param    line
    *           line's content
    * @param    row
    *           line's row
    * @param    splitAfter
    *           minimal split position 
    */
   @Override
   protected void splitLine(String line, int row, int splitAfter)
   {
      computeLineWidths(row, parsedText.size() - 1);

      super.splitLine(line, row, splitAfter);

      // let all the new lines have their widths set
      computeLineWidths(row, parsedText.size() - 1);
   }
   
   /**
    * Worker to parse content of the editor and prepare array of displayable values.
    * <p>
    * It uses {@link #computeSplitIndexes(String, int, int, boolean)} to determine the text's
    * split indexes.
    * 
    * @param   source
    *          Source string to parse.
    * @param   start
    *          Row index to start parsing from.
    * @param   offset
    *          Text offset to start parsing from.
    * @param   placeChar
    *          Set to {@code true} if called from {@link #placeChar(char)} method in insert mode.
    */
   @Override
   protected List<Line> parseContentImpl(String source, int start, int offset, boolean placeChar)
   {
      computeSplitIndexes(source, start, offset, placeChar);
      
      if (lineSplitIndexes == null)
      {
         // fallback to the old approach...
         return super.parseContentImpl(source, start, offset, placeChar);
      }
      
      char[] text = source.toCharArray();
      
      List<Line> lines = new ArrayList<>();
      
      int prevSplit = -1;
      for (int splitIdx : lineSplitIndexes)
      {
         if (splitIdx < offset || prevSplit == -1)
         {
            prevSplit = splitIdx;
            continue;
         }
         
         boolean newLine = newLineIndexes.contains(splitIdx);
         lines.add(new Line(text, prevSplit, splitIdx - prevSplit, newLine));

         prevSplit = splitIdx;
      }

      if (prevSplit != -1)
      {
         boolean newLine = newLineIndexes.contains(prevSplit);
         lines.add(new Line(text, prevSplit, text.length - prevSplit, newLine));
      }

      return lines;
   }
   
   /**
    * Move cursor one char right.
    */
   @Override
   protected void cursorRight()
   {
      Line line = getLine();
      int absX = absoluteX();
      final String txt = (line == null ? "" : line.getText());
      int ntxtLength = txt.length() == 0 
                          ? 0 
                          : getTextWidthNative(line, 
                                               config.scrollX, 
                                               Math.min(absX + 1, txt.length()),
                                               gf.resolveFontNum());
      int edWidth = editor.physicalDimension().width;
      
      if (config.scrollX + config.cursorX + 1 > txt.length())
      {
         // just go to next row, nothing else to do
         if (absoluteY() < parsedText.size() - 1)
         {
            // go to next line
            cursorDown();
            cursorBeginOfLine();
         }
         
         return;
      }
         
      if (txt.length() > 0)
      {
         // if the current text is still visible, just advance the cursor
         if (ntxtLength < edWidth)
         {
            // just advance the cursor; scrollX remains where it is
            config.cursorX++;
         }
         else
         {
            // the width no longer matches. we need to right-anchor the text on the current 
            // absX + 1 and find a new text which fits the screen
            int len = absX;
            String newVisTxt = null;
            int font = gf.resolveFontNum();
            while (len >= 0)
            {
               newVisTxt = txt.substring(len, absX + 1);

               int visLength = newVisTxt.length() == 0 
                                 ? 0 
                                 : getTextWidthNative(line, len, absX + 1, font);
               if (visLength > edWidth)
               {
                  // found one
                  newVisTxt = newVisTxt.substring(1);
                  break;
               }
               
               len = len - 1;
            }
            
            config.scrollX = absX - newVisTxt.length() + 1;
            config.cursorX = newVisTxt.length();
         }
      }
      
      super.cursorRight();
   }
   
   /**
    * Move cursor one row down, if possible.
    */
   @Override
   protected void cursorDown()
   {
      int prevY = absoluteY();
      int prevX = absoluteX();
      
      super.cursorDown();

      adjustCursorX(prevY, prevX, Integer.MIN_VALUE);
   }

   /**
    * Move cursor one row up, if possible.
    */
   @Override
   protected void cursorUp()
   {
      int prevY = absoluteY();
      int prevX = absoluteX();
      
      super.cursorUp();
      
      adjustCursorX(prevY, prevX, Integer.MIN_VALUE);
   }
   
   /**
    * Move cursor one char left, if possible.
    */
   @Override
   protected void cursorLeft()
   {
      if (config.cursorX + config.scrollX == 0 && absoluteY() > 0)
      {
         cursorUp();
         cursorEndOfLine();
      }
      else if (config.cursorX > 0)
      {
         config.cursorX--;
      }
      else
      {
         if (config.scrollX > 0)
            config.scrollX--;
      }

      super.cursorLeft();
   }

   /**
    * Move cursor to the end of line.
    */
   @Override
   protected void cursorEndOfLine()
   {
      Line l = getLine();

      int len = l.length() - 1;
      String line = l.getText();
      int font = gf.resolveFontNum();
      
      String txt = "";
      while (len >= 0)
      {
         txt = line.substring(len);
         
         int nwidth = getTextWidthNative(l, len, l.length(), font);
         if (nwidth > editor.physicalDimension().width)
         {
            txt = txt.substring(1);
            break;
         }
         
         len = len - 1;
      }

      if (txt.length() == 0)
      {
         config.scrollX = 0;
         config.cursorX = 0;
      }
      else if (txt.length() == line.length())
      {
         config.scrollX = 0;
         config.cursorX = txt.length();
      }
      else
      {
         config.scrollX = l.length() - txt.length();
         config.cursorX = txt.length();
      }
      
      super.cursorEndOfLine();
   }
   
   /**
    * Sets the cursor line
    *
    * @param    line
    *           the new cursor line value
    *
    * @return   <code>true</code> if the cursor line is valid and set 
    */
   @Override
   public boolean setCursorLine(int line)
   {
      int oldCursorY = config.cursorY;
      boolean ret = super.setCursorLine(line);
      
      if (ret && oldCursorY != config.cursorY)
      {
         invalidateSelection();
      }
      return ret;
   }
   
   /**
    * Handle BACKSPACE key.
    */
   @Override
   protected void backspace()
   {
      Line line = getLine();
      boolean merge = (contentLength() > 0 && line != null && absoluteX() == 0 && absoluteY() > 0);
      
      if (merge)
      {
         // simulate going up, so we can save the cursorX/scrollX
         cursorEndOfPrevLine();
         int cursorX = config.cursorX;
         int scrollX = config.scrollX;
         
         // go back to our line
         cursorDown();
         cursorBeginOfLine();
         
         // delete the NL for the previous line
         int prevY = absoluteY() - 1;
         Line prev = prevY < 0 ? null : parsedText.get(prevY);
         if (prev != null)
         {
            if (prev.isNewLine())
            {
               prev.setNewLine(false);
            }
            else
            {
               // delete the last char in prev line
               parsedText.set(prevY, 
                              new Line(prev.getText().substring(0, prev.getText().length() - 1),
                                       prev.getOffset(),
                                       prev.isNewLine()));
               cursorX--;
            }
         }
         updateOffset();
         
         mergeLines(absoluteY() - 1);
         
         // move the cursor
         if (config.cursorY == 0)
         {
            config.scrollY--;
         }
         else
         {
            config.cursorY--;
         }
         
         config.cursorX = cursorX;
         config.scrollX = scrollX;
      }
      else
      {
         super.backspace();
      }
      
      adjustScrollY();
   }
   
   /**
    * Delete character at current cursor position - called by {@link #processKeyEvent}.
    */
   @Override
   protected void deleteCharacterKey()
   {
      Line line = getLine();
      int contentLength = contentLength();
      boolean merge = (contentLength > 0            && 
                       line != null                 && 
                       absoluteX() == line.length() && 
                       absoluteY() < parsedText.size());
      
      if (merge)
      {
         if (contentLength > line.getOffset() + line.length() && 
             (line.length() != 0 || absoluteY() != parsedText.size() - 1))
         {
            // delete the NL for the current line
            if (line.isNewLine())
            {
               line.setNewLine(false);
            }
            else
            {
               // delete the first char from the next line
               int nextY = absoluteY() + 1;
               Line next = parsedText.get(nextY);
               parsedText.set(nextY,
                              new Line(next.getText().substring(1), 
                                       next.getOffset(), 
                                       next.isNewLine()));
            }
            updateOffset();
   
            mergeLines(absoluteY());
         }
      }
      else
      {
         super.deleteCharacterKey();
      }

      adjustScrollY();
   }

   /**
    * Remove trailing whitespaces. In GUI, this is a no-op.
    * 
    * @param   value
    *          the string to be truncated
    *    
    * @return  the truncated argument
    */
   protected String rtrim(String value)
   {
      return value;
   }

   /**
    * Get absolute horizontal position in text.
    * 
    * @return  The horizontal position in the text.
    */
   protected int absoluteX()
   {
      // must not exceed line length
      Line line = getLine();
      return Math.min(config.cursorX + config.scrollX, line == null ? 0 : line.length()); 
   }

   /**
    * Determine if a character can be placed in the editor.
    * 
    * @return    <code>true</code> if the editor is not read-only and its content is less than
    *            the {@link EditorConfig#maxChars} value (or MAX-CHARS attribute is not set).
    */
   @Override
   protected boolean canPlaceChar()
   {
      return super.canPlaceChar() && 
             (config.maxChars == null ||
              config.maxChars <= 0    ||
              contentLength() < config.maxChars);
   }
   
   /**
    * Processing for unusual EOL characters in GUI mode.  Currently replaces '\r' and '\r'+'\n'
    * with '\n'.
    * 
    * @param   src
    *          the string to be processed.
    *    
    * @return  the string with only '\n' characters as line seperator(EOL).
    */
   @Override
   protected String processEOLChars(String src)
   {
      // this sequence is a single EOL in GUI
      final String CRLF = "\r\n";
      // the EOL the P2J work with
      final String STRING_LF = "\n";
      
      // simple check to void NPE
      if (src == null || src.isEmpty())
      {
         return src;
      }
      
      // start with replacing CR+LF
      String result = src.replace(CRLF, STRING_LF);
      // finally replace all the rest single CR chars
      result = result.replace(CR, LF);
      
      return result;
   }

   /**
    * The method shows the popup menu for this widget. The menu is resolved with the help of
    * {@link #resolvePopupMenuId()}.
    *
    * @param   e
    *          The related mouse event that caused a popup to be shown.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    *
    * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
    */
   protected Menu<GuiOutputManager> showPopupMenu(MouseEvent e, Runnable dismissHandler)
   {
      Menu<GuiOutputManager> menu = super.showPopupMenu(e, dismissHandler);
      if (menu != null || !isEnabled())
      {
         return menu;
      }

      NativePoint origin = editor.screenPhysicalLocation();
      if (editor.physicalBounds().contains(e.getX() - origin.x, e.getY() - origin.y))
      {
         String selection = getSelectedText();
         boolean selectionEmpty = selection == null || selection.isEmpty();

         menu = EditorPopupGuiImpl.create(this, selectionEmpty);
         menu.showPopup(this, new NativePoint(e.getX(), e.getY()), dismissHandler);
         return menu;
      }

      return null;
   }
 
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return {@code true} if the widget supports LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }
   
   /**
    * Create handler for processing mouse hover events
    * 
    * @return handler for processing mouse hover events
    */
   @Override
   protected MouseHoverAction createMouseHoverAction()
   {
      return editorScroll == null ? null : new MouseHoverAction(this);
   }

   /**
    * Compute the split indexes for the given text.
    * 
    * @param   source
    *          Source string to parse.
    * @param   start
    *          Row index to start parsing from.
    * @param   offset
    *          Text offset to start parsing from.
    * @param   placeChar
    *          Set to {@code true} if called from {@link #placeChar(char)} method in insert mode.
    */
   private void computeSplitIndexes(String source, int start, int offset, boolean placeChar)
   {
      // invalidate them.
      lineSplitIndexes = null;
      newLineIndexes = null;

      if (config.scrollbarHorizontal || !config.wordWrap)
      {
         // no split indexes if in word-wrap mode
         return;
      }
      
      if (placeChar)
      {
         // can not track this from the JS side and back, currently
         return;
      }
      
      if (offset != 0)
      {
         source = source.substring(offset);
      }
      
      @SuppressWarnings("unchecked")
      TopLevelWindow<GuiOutputManager> wnd = parentOrSelf(TopLevelWindow.class);
      if (wnd == null)
         return;
      
      Object res = gd.parseEditorContent(getId().asInt(),
                                         wnd.getId().asInt(),
                                         editor.physicalBounds().width(),
                                         source, 
                                         offset,
                                         fontResolver().font().font);
      
      if (res != null)
      {
         Object[] o = (Object[]) res;
         lineSplitIndexes = (int[]) o[0];
         newLineIndexes = new HashSet<>(Utils.primitiveToIntegerList((int[]) o[1]));
      }
   }
   
   /**
    * Adjust the horizontal cursor so that it preservers the pixel-column, when moving up or down.
    * <p>
    * This is also used on mouse click, when the content is scrolled via the scroll bar 
    * ({@link #barScrollX} is non zero).  In this case, the <code>setCaretTo</code> parameter will
    * be set to a value other than {@link Integer#MIN_VALUE}, and will represent the relative 
    * pixel column (to the editor) where the mouse was clicked.  Using this, it will preserve the
    * mouse scroll (so that the left-edge character will be referenced by 
    * {@link EditorConfig#scrollX}) and the cursor X will be adjusted to represent the clicked
    * character.
    * 
    * @param    prevY
    *           Absolute line before the up/down key was processed.
    * @param    prevX
    *           Absolute column before the up/down key was processed.
    * @param    setCaretTo
    *           The pixels position relative to the view port.
    */
   private void adjustCursorX(int prevY, int prevX, int setCaretTo)
   {
      if (prevX == 0 && setCaretTo == Integer.MIN_VALUE)
      {
         // nothing to do
         return;
      }
      int font = gf.resolveFontNum();
      
      // adjust the cursor so that when moving up/down, it stays on the same character
      
      // fixed out of bounds on text cut
      if (prevY >= parsedText.size())
      {
         prevY = parsedText.size() - 1;
      }
      
      Line prevLine = parsedText.get(prevY);

      int absPixelX = getTextWidthNative(prevLine, 0, prevX, font) + 1;
      
      if (absPixelX == 0 && setCaretTo == Integer.MIN_VALUE)
      {
         // nothing to do
         return;
      }
      
      // scrollX must be the left-side of the currently visible text
      int absScrollX = computeHorizontalDelta(prevY, barScrollX);
      
      boolean scrollXSet = false;
      
      if (setCaretTo != Integer.MIN_VALUE)
      {
         // override the caret so it lands on what we need (i.e. on mouse click)
         absPixelX = absScrollX + setCaretTo;
      }
      
      // get the current line and determine where the caret will land; scrollX must remain 
      // "fixed", too, unless we are going to end of line
      Line line = getLine();

      int idx = -1;

      for (int i = 1; i < line.length(); i++)
      {
         int len = getTextWidthNative(line, 0, i, font);
         if (!scrollXSet && len > absScrollX)
         {
            config.scrollX = i - 1;
            scrollXSet = true;
         }
         
         if (len > absPixelX)
         {
            idx = i;
            break;
         }
      }
      
      if (idx == -1)
      {
         cursorEndOfLine();
      }
      else
      {
         config.cursorX = idx - config.scrollX - 1;
      }
   }
   
   /**
    * Move cursor one word left.
    */
   private void wordLeft()
   {
      // move left until cursor offset reaches beginning of text or:
      // a new line is found
      // non-whitespace char is found

      Line lastLine = getLine();
      cursorLeft();
      
      while (getCursorOffset() > 1)
      {
         Line current = getLine();
         
         if (lastLine != current)
         {
            // reached a new line, all search stops here
            if (current.isNewLine())
            {
               return;
            }
            
            // otherwise, one more left to end up on the last char
            if (current.length() > 0)
            {
               cursorLeft();
            }
         }
         
         char c = current.getText().charAt(absoluteX());
         if (c != ' ')
         {
            // reached a non-space char, we can find the beginning of the word
            break;
         }
         
         lastLine = current;
         cursorLeft();
      }

      lastLine = getLine();
      while (getCursorOffset() > 1)
      {
         Line current = getLine();
         
         if (lastLine != current)
         {
            // reached a new line
            if (current.isNewLine())
            {
               break;
            }
            
            // otherwise, one more left to end up on the last char
            if (current.length() > 0)
            {
               cursorLeft();
            }
         }
         
         char c = current.getText().charAt(absoluteX());
         if (c == ' ')
         {
            // reached a non-space char, we can find the beginning of the word
            break;
         }

         lastLine = current;
         cursorLeft();
      }
      
      // one cursor-right to position it on the beginning of the text
      if (getCursorOffset() != 1)
      {
         cursorRight();
      }
   }

   /**
    * Move cursor one word right.
    */
   private void wordRight()
   {
      // move right until cursor offset reaches end of text or:
      //      1. a new line is found
      //      2. non-whitespace char is found

      Line lastLine = getLine();
      boolean nonSpace = false;
      if (lastLine.length() != 0 && absoluteX() < lastLine.length())
      {
         char currentChar = lastLine.getText().charAt(absoluteX());
         nonSpace = (currentChar != ' ');
      }

      if (!nonSpace)
      {
         cursorRight();
         int curLength = getLine().length();
         if (curLength != 0 && absoluteX() >= curLength)
         {
            // one more right to get to the next line
            cursorRight();
         }
      }

      int contentLength = contentLength();
      
      while (!nonSpace && getCursorOffset() < contentLength)
      {
         Line current = getLine();
         
         if (current.isNewLine() && absoluteX() >= current.length())
         {
            // reached end of line, all search stops here
            return;
         }
         
         char c = current.getText().charAt(absoluteX());
         if (c != ' ')
         {
            // reached a non-space char, we can find the beginning of the word
            nonSpace = true;

            if (absoluteX() == 0)
            {
               // if we ended up at the beginning of the line, end here
               return;
            }

            continue;
         }
         
         lastLine = current;
         cursorRight();
      }
      
      lastLine = getLine();
      while (getCursorOffset() <= contentLength)
      {
         Line current = getLine();
         
         if (lastLine != current)
         {
            // reached end of line
            if (lastLine.isNewLine())
            {
               // move back one pos to remain on the last line
               cursorLeft();
               return;
            }
         }
         
         if (current.length() == 0 && current.isNewLine())
         {
            break;
         }
         
         char c = current.getText().charAt(absoluteX());
         if (c == ' ')
         {
            // reached a non-space char, we can find the beginning of the word
            cursorRight(); // pass the space(s), too  (TODO: xN ?)
            break;
         }

         lastLine = current;
         cursorRight();
         if (absoluteX() >= getLine().length())
         {
            // one more right to get to the next line
            cursorRight();
         }
      }
   }
   
   /**
    * Resize editor scroll pane, and other scroll-related processing (colors for the child
    * widgets, scrollbars, etc). 
    */
   private void processScrollPane()
   {
      int bgcolor = (config.bgcolor == -1 ? SysColor.COLOR_FILLIN_BG : config.bgcolor);
      int fgcolor = (config.fgcolor == -1 ? SysColor.COLOR_FILLIN_FG : config.fgcolor);
      
      contentPane.config().bgcolor = bgcolor;
      contentPane.config().fgcolor = fgcolor;
      editorScroll.config().bgcolor = bgcolor;
      editorScroll.config().fgcolor = fgcolor;

      contentPane.config().bgcolorRgb = config.bgcolorRgb;
      contentPane.config().fgcolorRgb = config.fgcolorRgb;
      editorScroll.config().bgcolorRgb = config.bgcolorRgb;
      editorScroll.config().fgcolorRgb = config.fgcolorRgb;

      // process the scroll bars
      boolean vbar = config.scrollbarVertical;
      boolean hbar = config.scrollbarHorizontal;
      boolean both = (vbar && hbar);
      editorScroll.setShowBars(hbar || vbar ? ShowBars.ALWAYS : ShowBars.NEVER);
      editorScroll.setScrollBars(both ? ScrollBar.Bars.BOTH 
                                      : hbar ? ScrollBar.Bars.HORIZONTAL 
                                             : vbar ? ScrollBar.Bars.VERTICAL 
                                                    : ScrollBar.Bars.NONE);
   
      WidgetRegistry<GuiOutputManager> reg = screen().getRegistry();
      reg.addWidgetsRecursive(editorScroll.widgets());
      
      if (gf == null)
      {
         // next, we initialize values after the widget has been realized
         return;
      }
      
      // set the size of the viewport to have the same size as the editor; this is needed because
      // we don't use an actual viewport, but a form of pseudo-scrolling, where only the visible
      // text is drawn
      boolean box = config().box;
      
      // location
      NativePoint np = new NativePoint(0, 0);
   
      editorScroll.setPhysicalLocation(np.x, np.y);
      editorScroll.setLocation(cc.pointFromNative(np));

      np.translate(box ? 1 : 0, box ? 1 : 0);
      editor.setPhysicalLocation(np.x, np.y);
      editor.setLocation(cc.pointFromNative(np));

      // size
      NativeDimension nd = physicalDimension();
      editorScroll.setSize(cc.dimensionFromNative(nd));

      editorScroll.adjustScrollLayout();
      
      setScrollPosition();
   }
   
   /**
    * Adjust the scroll position.
    */
   private void setScrollPosition()
   {
      if (ancestor() == null)
      {
         return;
      }

      ScreenDriver<?> driver = screen().getInstanceDriver();
      Window<GuiOutputManager> window = window();
      
      int fontNum = gf.resolveFontNum();
      int nfontHeight = gf.font().fontHeight;
      int nfontWidth = FontManager.getFontWidth(window, fontNum, driver);

      // compute the inner-lines/-chars
      NativeDimension dim = editor.physicalDimension();
      if (config.innerLines <= 0)
      {
         config.innerLines = dim.height / nfontHeight;
      }
      if (config.innerChars <= 0)
      {
         config.innerChars = dim.width / nfontWidth;
      }
      
      Integer scrollX = null;
      Integer scrollY = null;
      
      if (config.scrollbarVertical)
      {
         scrollY = config.scrollY + barScrollY;
      }
      
      if (config.scrollbarHorizontal)
      {
         scrollX = computeHorizontalDelta(absoluteY(), 0);
      }
      
      contentPane.postScrollEvent(scrollX, scrollY);
   }
   
   /**
    * Compute the full horizontal delta (i.e. left-side text which is scrolled by caret or by 
    * mouse).
    * <p>
    * The <code>absY</code> represents the line index to which the {@link EditorConfig#scrollX}
    * will be applied.  The length of the this prefix text (from 0 to scroll X) will be computed
    * for this line, and to this value the {@link #computeHorizontalScrollDelta mouse scroll
    * delta}.
    * 
    * @param    absY
    *           The reference line from where to take the caret-scrolled prefix.
    * @param    barX
    *           The physical delta scrolled by mouse, in number of clicks.
    * 
    * @return   See above.
    */
   private int computeHorizontalDelta(int absY, int barX)
   {
      // fixed out of bounds on text cut
      if (absY >= parsedText.size())
      {
         absY = parsedText.size() - 1;
      }
      
      Line l = parsedText.get(absY);
      
      // determine the length of the scrolled text
      int scrollDelta = (config.scrollX == 0 || l.length() == 0 
                           ? 0
                           : getTextWidthNative(l, 0, config.scrollX, gf.resolveFontNum()));
      
      // determine the barX adjustment
      int hdelta = computeHorizontalScrollDelta(barX);
      
      return scrollDelta + hdelta;
   }
   
   /**
    * Compute the horizontal scroll delta, based on how far the scroll bar was moved via mouse to
    * the left or the right, relative to the current {@link EditorConfig#scrollX} value.  This is
    * always computed using the font's width multiplied by the given delta.
    * 
    * @param    barX
    *           The physical delta scrolled by mouse, in number of clicks.
    * 
    * @return   See above.
    */
   private int computeHorizontalScrollDelta(int barX)
   {
      int fontWidth = FontManager.getFontWidth(window(), 
                                               gf.resolveFontNum(), 
                                               screen().getInstanceDriver());
      
      return fontWidth * barX;
   }

   /**
    * Compute the maximum line width in this editor, in native units.
    * 
    * @return   See above.
    */
   private int maxLineWidthNative()
   {
      int font = gf.resolveFontNum();
      int maxLineWidth = 0;
      
      computeLineWidths(0, parsedText.size() - 1);
      
      for (Line line : parsedText)
      {
         int nlineWidth = getTextWidthNative(line, 0, line.length(), font);
         
         maxLineWidth = Math.max(nlineWidth, maxLineWidth);
      }

      return maxLineWidth;
   }
   
   /**
    * Gets the first maximal line in this editor.
    * 
    * @return   The maximal line.
    */
   private Line getMaximalLine()
   {
      int font = gf.resolveFontNum();
      int maxLineWidth = 0;
      Line maximalLine = null;
      
      computeLineWidths(0, parsedText.size() - 1);
      
      for (Line line : parsedText)
      {
         int nlineWidth = getTextWidthNative(line, 0, line.length(), font);
         
         if (nlineWidth > maxLineWidth)
         {
            maximalLine = line;
            maxLineWidth = nlineWidth;
         }
      }
      
      return maximalLine;
   }
   
   /**
    * Find a scroll offset to set the {@link EditorConfig#scrollX} value.
    * 
    * @param    width
    *           The pixel width for the text to be scrolled.
    *           
    * @return   See above.
    */
   private int findScrollOffset(int width)
   {
      if (width == 0)
      {
         return 0;
      }
      
      int font = gf.resolveFontNum();
      
      Line l = getMaximalLine();
      
      for (int i = 1; i < l.length(); i++)
      {
         int nlineWidth = getTextWidthNative(l, 0, i, font);
         
         if (nlineWidth >= width)
         {
            return i;
         }
      }
      
      return l.length();
   }
   
   /**
    * If we are removing the last line in the editor, then all content needs to "scroll up" one
    * row so that the editor's last line is on the bottom.
    */
   private void adjustScrollY()
   {
      // there is a catch here: adjust the scrollY if we are currently showing the last line
      if ((lastRow() < config.innerLines && config.scrollY != 0) || 
          (lastRow() >= config.innerLines && config.scrollY + config.innerLines - 1 > lastRow()))
      {
         config.scrollY = Math.max(0, lastRow() - config.innerLines + 1);
         config.cursorY++;
      }
   }
   
   /**
    * Invalidate the editor's selection.
    */
   @Override
   public void invalidateSelection()
   {
      selectionStart = null;
      selectionEnd = null;
      
      config.selectionActive = false;
      config.selectionStart = -1;
      config.selectionEnd = -1;
      config.selectionText = null;
      // notify the client the selection is invalid
      gd.setCurrentSelection(null);
   }
   
   /**
    * Determine if the editor has selected text.
    * 
    * @return  {@code true} if the {@link #selectionStart} and {@link #selectionEnd} are set.
    */
   public boolean isSelectionValid()
   {
      if (selectionStart != null && selectionEnd != null)
      {
         if (selectionStart.y < 0 || selectionStart.y >= parsedText.size() ||
             selectionEnd.y < 0   || selectionEnd.y >= parsedText.size() + 1)
         {
            return false;
         }
         
         // check if the selection is within the text
         Line lstart = parsedText.get(selectionStart.y);
         Line lend   = parsedText.get(selectionEnd.y);
         
         return selectionStart.x >= 0 && selectionStart.x <= lstart.length() &&
                selectionEnd.x >= 0   && selectionEnd.x < lend.length() + 1;
      }
      
      return false;
   }
   
   /**
    * Adjust the selection interval so that the SELECTION-START and SELECTION-END attributes
    * reflect the current selection.
    */
   private void adjustSelection()
   {
      if (selectionStart == null || selectionEnd == null)
      {
         return; // invalid selection
      }
      
      // if we have gone before the selection start after the key was applied, then switch
      // the points
      if ((selectionStart.y > selectionEnd.y) || 
          (selectionStart.y == selectionEnd.y && selectionStart.x > selectionEnd.x))
      {
         NativePoint aux = selectionStart;
         selectionStart = selectionEnd;
         selectionEnd = aux;
      }
      
      // if there is no selection, invalidate it
      if (selectionStart.y == selectionEnd.y && selectionStart.x == selectionEnd.x)
      {
         invalidateSelection();
      }
      else
      {
         // selection-start/selection-end attributes
         config.selectionActive = true;
         config.selectionStart = parsedText.get(selectionStart.y).getOffset() + 
                                 selectionStart.x + 1;
         config.selectionEnd = parsedText.get(selectionEnd.y).getOffset() + 
                               selectionEnd.x + 1;
         config.selectionText = getSelectedText(false);
         gd.setCurrentSelection(config.selectionText);
      }
   }
   
   /**
    * Get the currently selected sub-text.  If the <code>remove</code> parameter is set to 
    * <code>true</code>, this text is also removed.
    * 
    * @param    remove
    *           Flag determining if the selected text needs to be removed.
    * 
    * @return   The selected text or empty string if there is no selection.
    */
   private String getSelectedText(boolean remove)
   {
      if (!isSelectionValid())
      {
         return "";
      }

      // determine the position of the text between selectionStart and selectionEnd
      String txt = resolveContent();

      Line lstart = parsedText.get(selectionStart.y);
      Line lend   = parsedText.get(selectionEnd.y);
      
      int offsetStart = lstart.getOffset() + selectionStart.x;
      int offsetEnd   = lend.getOffset() + selectionEnd.x;
      
      String selected = txt.substring(offsetStart, offsetEnd);
      
      if (remove)
      {
         txt = txt.substring(0, offsetStart) + txt.substring(offsetEnd);
         
         NativePoint selStart = selectionStart;
         NativePoint selEnd = selectionEnd;
         
         invalidateSelection();
         
         parseContent(txt, 0, false);
      
         setCursorLine(selStart.y + 1);
         setCursorChar(selStart.x);
      }
      
      return selected;
   }

   /**
    * Get the parent window
    * 
    * @return parent window
    */
   private Window<?> resolveWindow()
   {
      return WindowManager.resolveWindow(this);
   }
   
   /**
    * Get the font resolver
    * 
    * @return font resolver
    */
   private GuiFontResolver fontResolver()
   {
      if (gf == null)
      {
         gf = new GuiFontResolver(resolveWindow(), gd, this.config(), false);
      }
      return gf;
   }

   /**
    * Compute the line widths (if not already computed) in the given interval.
    * 
    * @param    from
    *           Start index in the parsed content.
    * @param    to
    *           End index in the parsed content (inclusive).
    */
   private void computeLineWidths(int from, int to)
   {
      Widget tlw = parentOrSelf(TopLevelWindow.class);
      if (tlw == null)
      {
         return;
      }
      List<String> texts = new ArrayList<>();
      List<Integer> indices = new ArrayList<>();
      for (int i = from; i <= to; i++)
      {
         Line l = parsedText.get(i);
         if (l.getWidths() != null || l.getText().length() == 0)
         {
            continue;
         }
         
         texts.add(l.getText());
         indices.add(i);
      }
      
      if (texts.isEmpty())
      {
         return;
      }
      
      Map<FontDetails<?>, Set<String>> map = Collections.singletonMap(gf.font().font,
                                                                      new LinkedHashSet<String>(texts));
      
      gd.cacheTextMetrics(map, tlw.getId().asInt());
      
      gd.withSelectedWindow(tlw.getId().asInt(), () -> {
         int[][] widths = gd.getTextWidths(texts.toArray(new String[0]), gf.font().font);
         for (int i = 0; i < indices.size(); i++)
         {
            parsedText.get(indices.get(i)).setWidths(widths[i]);
         }
      });
   }
   
   /**
    * Compute the widths for this line (if not already computed).
    * 
    * @param    l
    *           The line for which the widths need to be computed.
    */
   private void computeLineWidths(Line l)
   {
      if (parent() == null)
      {
         return;
      }
      
      gd.withSelectedWindow(topLevelWindow().getId().asInt(), () -> {
         if (l == null || l.getWidths() != null || l.getText().length() == 0)
         {
            return;
         }
         
         l.setWidths(gd.getTextWidths(l.getText(), gf.font().font));
      });
   }
   
   /**
    * Get the width of the specified substring, in the given line.
    * <p>
    * If the line's intermediate widths were computed, use these to compute the substring's width.
    * Otherwise, default to the {@link AbstractWidget#getTextWidthNative}.
    * 
    * @param    lineNo
    *           The line number (0-base).
    * @param    start
    *           The start index in the line.
    * @param    end
    *           The end index in the line (exclusive).
    * @param    font
    *           The font.
    *           
    * @return   The substring's width, in pixels.
    */
   public int getTextWidthNative(int lineNo, int start, int end, int font)
   {
      if (parsedText == null || lineNo < 0 || lineNo >= parsedText.size())
      {
         return 0;
      }
      
      return getTextWidthNative(parsedText.get(lineNo), start, end, font); 
   }
   
   /**
    * Get the width of the specified substring, in the given line.
    * <p>
    * If the line's intermediate widths were computed, use these to compute the substring's width.
    * Otherwise, default to the {@link AbstractWidget#getTextWidthNative}.
    * 
    * @param    line
    *           The line instance.
    * @param    start
    *           The start index in the line.
    * @param    end
    *           The end index in the line (exclusive).
    * @param    font
    *           The font.
    *           
    * @return   The substring's width, in pixels.
    */
   public int getTextWidthNative(Line line, int start, int end, int font)
   {
      if (end <= start)
      {
         return 0;
      }

      computeLineWidths(line);
      
      int[] widths = line.getWidths();
      
      if (widths == null)
      {
         return getTextWidthNative(line.getText().substring(start, end), font);
      }
      
      if (end >= widths.length)
      {
         end = widths.length;
      }
      
      if (start >= widths.length)
      {
         start = widths.length;
      }
      
      return (start == 0 ? widths[end - 1] : widths[end - 1] - widths[start - 1]);
   }
   
   /**
    * Worker class which is responsible of drawing the editor's text.
    */
   private class EditorContent 
   extends AbstractContainer<GuiOutputManager>
   {
      @Override
      public Dimension dimension()
      {
         return minimumSize();
      }

      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         NativeRectangle clip = clipRectangle(p, d);
         
         Line maxLine = getMaximalLine();
         
         int deltaScrollX = (config.scrollX == 0 || maxLine.length() == 0)
                                    ? 0
                                    : EditorGuiImpl.this.getTextWidthNative(maxLine,
                                                                            0,
                                                                            config.scrollX,
                                                                            gf.resolveFontNum());
         
         gd.draw(p, clip, () -> ThemeManager.getCurrentTheme().drawEditorContent(
               EditorGuiImpl.this, gd, gf, gc, 0, 0, d.width, d.height, 
               selectionStart, selectionEnd, deltaScrollX, barScrollY));
      }

      @Override
      public double width()
      {
         Point baseUnits = screen().coordinates().baseUnits();
         boolean box = EditorGuiImpl.this.config.box;
         
         return parent().width() - baseUnits.x * (box ? 4 : 2);
      }

      @Override
      public double height()
      {
         Point baseUnits = screen().coordinates().baseUnits();
         boolean box = EditorGuiImpl.this.config.box;

         return parent().height() - baseUnits.y * (box ? 4 : 2);
      }
      
      @Override
      public Dimension minimumSize()
      {
         return new Dimension(width(), height());
      }
   }
   
   /**
    * This class implements the editor's undo (CTRL-Z) behavior.
    * <p>
    * Two states need to be tracked:
    * <ol>
    *    <li>The deleted text, when one of the following keys are pressed: 
    *        {@link Key#VK_CTRL_DEL}, {@link Keyboard#KA_DELETE_CHARACTER}, 
    *        {@link Keyboard#KA_BACKSPACE} or {@link Keyboard#KA_CUT}.
    *        When a selected text is deleted by a printable key, 
    *        {@link UndoableEditor#recordDeletedText} is called as if DELETE was pressed.
    *     </li>
    *     <li>The inserted text, when a {@link Keyboard#isPrintableKey printable key} is pressed.
    *     </li>
    * </ol>
    * <p>
    * The undo data is reset if the inserted or deleted text is not contiguous.
    * <p>
    * On CTRL-Z, the {@link UndoableEditor#deletedText} and {@link UndoableEditor#insertedText} 
    * are being switched, with the shown text being already selected.
    */
   private static class UndoableEditor
   {
      /** The editor instance managed by this class. */
      private final EditorGuiImpl editor;
      
      /** The previous caret/cursor offset. */
      private int prevCaretOffset = -1;
      
      /**
       * The caret/cursor offset when the last reset was performed (this is the offset where
       * CTRL-Z will append the deleted text). */
      private int lastResetOffset = -1;
      
      /** The last processed key by this class (not pressed). */
      private int lastKey = -1;
      
      /** The deleted text (all contiguous). */
      private String deletedText = "";
      
      /** The inserted text (all contiguous). */
      private String insertedText = "";
      
      /**
       * Flag indicating which text to show on CTRL-Z ({@link #deletedText} or 
       * {@link #insertedText}).
       */
      private boolean showDelete = true;

      /** Flag indicating if previous operation was a character removal key. */
      private boolean wasDelete = true;
      
      /**
       * Create a new instance to process undo for the given editor.
       * 
       * @param    editor
       *           The editor.
       */
      public UndoableEditor(EditorGuiImpl editor)
      {
         this.editor = editor;
         reset(-1);
      }
      
      /**
       * Process the given key. Only the following are processed: CTRL-Z, printable keys and
       * character removal keys.
       * 
       * @param    key
       *           The key to process.
       */
      public void processKey(int key)
      {
         boolean printableKey = Keyboard.isPrintableKey(key);
         int cursorOffset = editor.getCursorOffset();
         
         if (key == Key.VK_CTRL_Z)
         {
            processUndo();
         }
         else if (printableKey)
         {
            recordInsertedText(key, cursorOffset);
         }
         else
         {
            // check if this is a "remove text" key
            recordDeletedText(key, cursorOffset);
         }
      }

      /**
       * Process CTRL-Z by showing the {@link #deletedText} or {@link #insertedText}.
       */
      private void processUndo()
      {
         String show = (showDelete ? deletedText : insertedText);
         String delete = (showDelete ? insertedText : deletedText);
         
         editor.invalidateSelection();
            
         // 1. delete text in [lastResetOffset, lastResetOffset + delete.length)
         editor.setCursorOffset(lastResetOffset);
         for (int i = 0; i < delete.length(); i++)
         {
            editor.deleteCharacterKey();
         }

         // 2. show text by placing chars start from lastResetOffset
         editor.setCursorOffset(lastResetOffset);
         for (int i = 0; i < show.length(); i++)
         {
            editor.placeChar(show.charAt(i));
         }

         // 3. select the shown text
         int selStart = lastResetOffset;
         int selEnd = lastResetOffset + show.length();
         if (selStart != selEnd)
         {
            editor.setSelection(selStart, selEnd);
         }
         
         showDelete = !showDelete;
      }
      
      /**
       * Get the editor chars from the given interval.
       * 
       * @param    startOffset
       *           The 1-based start offset.
       * @param    endOffset
       *           The 1-based end offset.
       *           
       * @return   A string with the chars in the given interval.
       */
      private String getChars(int startOffset, int endOffset)
      {
         // transform them to 0-based indexes
         startOffset = Math.max(startOffset - 1, 0);
         endOffset   = Math.min(endOffset - 1, editor.contentLength());
         
         String content = editor.resolveContent();
         String chars = content.substring(startOffset, endOffset);
         
         return chars;
      }
      
      /**
       * Reset the undo data and save the processed key.
       * 
       * @param    key
       *           The key code which triggered the reset.
       */
      private void reset(int key)
      {
         prevCaretOffset = editor.getCursorOffset();
         lastResetOffset = editor.getCursorOffset();
         lastKey = key;
         deletedText = "";
         insertedText = "";
         showDelete = true;
         wasDelete = false;
      }
      
      /**
       * Record the inserted text by saving it into {@link #insertedText}.
       * <p>
       * This is called BEFORE the key is processed by the editor.
       * 
       * @param    key
       *           The key code.
       * @param    cursorOffset
       *           The cursor offset where the key will be inserted.
       */
      private void recordInsertedText(int key, int cursorOffset)
      {
         if (cursorOffset != prevCaretOffset + (wasDelete ? 0 : 1))
         {
            reset(key);
         }
         
         // record the char
         insertedText += (char) key;
         lastKey = key;
         prevCaretOffset = cursorOffset;
         wasDelete = false;
      }
      
      /**
       * Record the deleted text by saving it into {@link #deletedText}.
       * <p>
       * This is called BEFORE the key is processed by the editor, in cases when a selected text
       * needs to be removed.
       * 
       * @param    key
       *           The key code.
       * @param    start
       *           The position where the text is located in the editor.
       * @param    text
       *           The text to be removed.
       */
      private void recordDeletedText(int key, NativePoint start, String text)
      {
         if (!showDelete)
         {
            // this selection is from a ctrl-z: thus invalidate currently deleted text
            deletedText = "";
            showDelete = true;
         }
         
         Line line = editor.parsedText.get(start.y);
         int cursorOffset = line.getOffset() + start.x + 1;
         
         recordDeletedText(key, cursorOffset, text);
      }
      
      /**
       * Record the deleted text by saving it into {@link #deletedText}.
       * <p>
       * This is called BEFORE the key is processed by the editor.
       * 
       * @param    key
       *           The key code.
       * @param    cursorOffset
       *           The cursor offset from where the text is deleted.
       */
      private void recordDeletedText(int key, int cursorOffset)
      {
         recordDeletedText(key, cursorOffset, null);
      }

      /**
       * Record the deleted text by saving it into {@link #deletedText}.
       * <p>
       * This is called BEFORE the key is processed by the editor.
       * 
       * @param    key
       *           The key code.
       * @param    cursorOffset
       *           The cursor offset from where the character is removed.
       * @param    text
       *           If non-null, this overrides the deleted character (is assumed the entire text
       *           is being removed).
       */
      private void recordDeletedText(int key, int cursorOffset, String text)
      {
         // this is called always BEFORE the key is processed

         // if the previous key was a printable key, then reset
         boolean prevPrintable = Keyboard.isPrintableKey(lastKey);
         
         if (key == Keyboard.KA_DELETE_CHARACTER || key == Keyboard.KA_CUT)
         {
            // CUT/DELETE must remain on the same char
            if (prevPrintable || cursorOffset != prevCaretOffset)
            {
               reset(key);
            }
            
            deletedText += (text != null ? text : getChars(cursorOffset, cursorOffset + 1));
            lastKey = key;
            prevCaretOffset = cursorOffset;
            wasDelete = true;
         }
         else if (key == Keyboard.KA_BACKSPACE)
         {
            // BACKSPACE must have the caret on the left-side of the previous caret
            if (prevPrintable || cursorOffset != prevCaretOffset - 1)
            {
               reset(key);
            }

            // backspace goes to the left side, so the text needs to be prefixed to the previous
            // content
            deletedText = (text != null ? text : getChars(cursorOffset - 1, cursorOffset)) + 
                          deletedText;
            lastKey = key;
            prevCaretOffset = cursorOffset;
            // for backspace, this needs to follow the current offset 
            // (as the origin is on the right-side, not left-side, as for normal delete).
            lastResetOffset = Math.max(cursorOffset - 1, 1);
            wasDelete = true;
         }
         else if (key == Key.VK_CTRL_DEL)
         {
            // CTRL-DEL (delete to end of line)  must remain on the same char
            if (prevPrintable || cursorOffset != prevCaretOffset)
            {
               reset(key);
            }
            
            Line current = editor.getLine();
            int start = current.getOffset() + editor.absoluteX();
            int end   = current.getOffset() + editor.getLength();
            
            // getChars accepts 1-based indexes
            deletedText += (text != null ? text : getChars(start + 1, end + 1));
            lastKey = key;
            prevCaretOffset = cursorOffset;
            wasDelete = true;
         }
      }
   }
   
   /** 
    * The editor's scroll container. 
    */
   private class EditorScrollContainer
   extends ScrollContainer<GuiOutputManager>
   {
      /**
       * Directs the widget to scroll to the specified X and Y positions.
       * It is the widget's responsibility to update its visuals accordingly.
       * Note that the position values are in the same units as the returned 
       * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
       * and {@link #getScrollStep()}. 
       * 
       * @param    newXPosition
       *           New X scroll position.
       * @param    newYPosition
       *           New Y scroll position.
       */
      @Override
      public void scroll(Integer newXPosition, Integer newYPosition)
      {
         boolean dirty = false;
         
         // any horizontal scroll? 
         if (newXPosition != null)
         {
            // WARNING! config.scrollX is in characters, but the posted value is in pixels (as
            // this is how the horizontal scroll works...).  This needs to translate the X from
            // pixels to an offset in the editor
            int oldX = EditorGuiImpl.this.config.scrollX;
            EditorGuiImpl.this.config.scrollX = findScrollOffset(newXPosition);
            barScrollX = 0;
            int d = EditorGuiImpl.this.config.scrollX - oldX;
            EditorGuiImpl.this.config.cursorX = EditorGuiImpl.this.config.cursorX - d;
            dirty |= (d != 0);
         }
         
         if (newYPosition != null)
         {
            int oldY = EditorGuiImpl.this.config.scrollY;
            EditorGuiImpl.this.config.scrollY = newYPosition;
            barScrollY = 0;
            dirty |= EditorGuiImpl.this.config.scrollY != oldY;
         }
         
         if (dirty)
         {
            repaint();
         }
      }

      /**
       * Returns the scroll dimension of this widget. The scroll dimension is
       * the full size of the widget's content, it may be bigger than widget's
       * visible size.
       * <p>
       * The returned dimension doesn't need to be in the same units as the 
       * widget's physical size but it must be compatible with the result
       * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
       * 
       * @return   widget's scroll dimension
       */
      @Override
      public NativeDimension getScrollDimension()
      {
         int nmaxLineWidth = maxLineWidthNative();
         return new NativeDimension(nmaxLineWidth, parsedText.size());
      }

      /**
       * Returns the visible dimension of this widget. The visible dimension is
       * the visible portion of the widget's content (including scrollbars area).
       * <p>
       * The returned dimension doesn't need to be in the same units as the 
       * widget's physical size but it must be compatible with the result
       * of {@link #getScrollDimension()} and {@link #getScrollStep()}.
       *
       * @return   widget's visible dimension
       */
      @Override
      public NativeDimension getVisibleDimension()
      {
         double width  = 0;
         
         if (editorScroll != null)
         {
            Dimension dimViewPort = editorScroll.getViewPortSize();
            
            width  = dimViewPort.width;
         }
         
         return new NativeDimension(screen().coordinates().widthToNative(width),
                                    EditorGuiImpl.this.config.innerLines);
      }

      /**
       * Returns the dimension of the base scroll step desired by the widget
       * in the X and Y directions.
       * <p>
       * The scroll step must be in the same units as the dimensions returned
       * by {@link #getScrollDimension()} and {@link #getVisibleDimension()}. 
       * 
       * @return   The dimension of the base scroll step.
       */
      @Override
      public NativeDimension getScrollStep()
      {
         return new NativeDimension(1, 1);
      }

      /**
       * Post scroll event with the specified scroll positions.
       * 
       * @param   scrollX
       *          New X position.
       * @param   scrollY
       *          New Y position.
       */
      public void postScrollEvent(Integer scrollX, Integer scrollY)
      {
         super.postScrollEvent(scrollX, scrollY);
      }
   }

   /**
    * Gets the screen text content.
    * 
    * @param    line
    *           The given line number
    * 
    * @return   The screen text content
    */
   @Override
   public String getScreenText(int line)
   {
      String text = getLineText(line);
      
      if (text != null)
      {
         return text;
      }

      return "";
   }

   /**
    * Gets the screen text content at the current line.
    * 
    * @return   The screen text content
    */
   @Override
   public String getScreenText()
   {
      Line line = getLine();
      
      if (line != null)
      {
         return line.getText();
      }
      
      return "";
   }
   
   /**
    * Gets the current cursor column within its text content.
    * 
    * @return   The current cursor column
    */
   @Override
   public int getCursorColumnOffset()
   {
      int absX = absoluteX();
      
      if (absX > 0)
      {
         return absX;
      }
      
      return 0;
   }

   /**
    * Gets the current cursor row within its text content.
    * 
    * @return   The current cursor row
    */
   @Override
   public int getCursorRowOffset()
   {
      int absY = absoluteY();
      
      if (absY > 0)
      {
         return absY;
      }
      
      return 0;
   }

   /**
    * Draws a special cursor relative to the current origin.
    */
   @Override
   public void drawCursor()
   {
      int nfontHeight = getFontCache().fontHeight;
      int deltaScrollX = computeHorizontalScrollDelta(barScrollX);
      int deltaScrollY = nfontHeight * barScrollY;
      
      ThemeManager.getCurrentTheme().drawEditorCaret(this, gd,
               -deltaScrollX, -deltaScrollY, -deltaScrollX, -deltaScrollY + nfontHeight - 1);
   }

   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll column
    */
   @Override
   public int getScrollColumnOffset()
   {
      return config.scrollX;
   }

   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll row
    */
   @Override
   public int getScrollRowOffset()
   {
      return config.scrollY;
   }

   /**
    * Returns the current font cache.
    * 
    * @return   The current font cache
    */
   @Override
   public FontCache getFontCache()
   {
      return gf.font();
   }

   /**
    * Gets the current text selection object for the given line. If the given line doesn't have
    * selected text, then this method returns null.
    * 
    * @param    line
    *           The given line number
    * 
    * @return   The current text selection object if there is selected text in the given line,
    *           otherwise null.
    */
   @Override
   public TextLineSelection getTextSelection(int line)
   {
      // draw the text, keeping in mind if there is selection on this line or not
      boolean select = isSelectionValid() && line >= selectionStart.y && line <= selectionEnd.y;
      
      if (!select)
      {
         return null;
      }
      
      String txt = getScreenText(line);
      
      String[] texts = new String[3];
      
      int start = (line == selectionStart.y ? selectionStart.x : 0);
      
      int end = (line == selectionEnd.y ? selectionEnd.x : txt.length());
      
      if (end > txt.length())
      {
         end = txt.length();
      }
      
      // prefix not-selected text
      texts[0] = (start == 0 ? "" : txt.substring(0, start));
      // selected text
      texts[1] = txt.substring(start, end);
      // suffix not-selected text
      texts[2] = (end == txt.length() ? "" : txt.substring(end));
      
      int[][] splits = {
         { 0, start },
         { start, end },
         { end, txt.length() },
      };

      int[] hintWidths = new int[3];
      
      for(int i = 0; i < 3; i++)
      {
         hintWidths[i] = getTextWidthNative(line, splits[i][0], splits[i][1], gf.resolveFontNum());
      }
      
      /**
       * Defines the text line selection object.
       */
      return new TextLineSelection()
      {
         /**
          * Returns the column and row, where the current text selection gets the its origin.
          * 
          * @return   The beginning of the current text selection
          */
         @Override
         public NativePoint getSelectionStart()
         {
            return selectionStart;
         }

         /**
          * Returns the most closest point, column and row, that breaks the continuous selection text
          * part.
          * 
          * @return   The break point, column and row, of the current text selection
          */
         @Override
         public NativePoint getSelectionEnd()
         {
            return selectionEnd;
         }

         /**
          * Returns the 3-tuple of string widths used as hint widths for this line split into outer
          * and selected text.
          * 
          * @return   The 3-tuple of string hint widths in pixels
          */
         @Override
         public int[] getTextWidthsHint()
         {
            return hintWidths;
         }

         /**
          * Returns the 3-tuple of text parts into which this line has been split by text selection.
          * 
          * @return   The 3-tuple of text parts
          */
         @Override
         public String[] getLineParts()
         {
            return texts;
         }
      };
   }
}
