/*
** Module   : GuiButton.java
** Abstract : add mouse sensitivity to push button implementation
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170321 First revision. Click and hover mouse event supported.
** 002 OM  20170407 Pulled mouseActions() from subclasses.
** 003 OM  20180730 Fixed moment when focus is requested on the button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.Container;

import java.awt.event.*;
import java.util.logging.*;

/**
 * This abstract class add mouse sensitiveness to {@link Button} objects. Instances of this
 * class are notified when mouse hovers over them, and when they are pressed down using the mouse.
 * The implementation will track mouse events on this button and provide the current state using
 * getters. It also supports the any combination of pressed and hover, if pressed and mouse cursor
 * is dragged outside the bounds, the state is kept as pressed, so in an eventual re-enter event
 * the state is correct. It's up to called to caller renderer to draw the state of the button and
 * the listener to act on the state provided by this class.
 */
public abstract class GuiButton
extends Button<GuiOutputManager>
{
   /** Whether the mouse is over this button. Some themes paints this differently. */
   private boolean mouseOver = false;
   
   /** Pressed button state flag. */
   private java.awt.Point pressedLoc = null;
   
   /**
    * This constructor creates a button with a given text label.
    *
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   public GuiButton(WidgetId id, Container<GuiOutputManager> owner, String text)
   {
      super(id, owner, text);
   }
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public GuiButton(WidgetId id, ButtonConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Checks if the mouse hovers over this widget. 
    *
    * @return  {@code true} if the mouse pointer is inside the visible bounds of this widget
    */
   public boolean isMouseOver()
   {
      return mouseOver;
   }
   
   /**
    * Process the notification of a mouse enter event occurred for this widget. Store the new 
    * state and call repaint to update the button's visual.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      if (isEnabled())
      {
         this.mouseOver = true;
         if (ThemeManager.getCurrentTheme().isMouseHoverSensitive())
         {
            repaint();
         }
      }
   }
   
   /**
    * Process the notification of a mouse leave event occurred for this widget. Store the new 
    * state and call repaint to update the button's visual.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (isEnabled())
      {
         this.mouseOver = false;
         if (ThemeManager.getCurrentTheme().isMouseHoverSensitive())
         {
            repaint();
         }
      }
   }
   
   /**
    * Checks pressed mouse state on this widget. 
    * 
    * @return  {@code true} if user pressed the left mouse button on this widget and has not
    *          released it yet.
    */
   public boolean isPressed()
   {
      return pressedLoc != null;
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw button in pressed
    * state and request focus if focusable.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && pressedLoc == null && e.getButton() == MouseEvent.BUTTON1)
      {
         super.mousePressed(e);
         pressedLoc = e.getLocationOnScreen();
         if (!mouseOver)
         {
            mouseOver = true;
            LOG.log(Level.WARNING, "Missing mouseEnter() event on " + this);
         }
         repaint();
      }
   }
   
   /**
    * Notification of a mouse moved event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (isEnabled())
      {
         if (!mouseOver)
         {
            mouseOver = true;
            LOG.log(Level.WARNING, "Missing mouseEnter() event on " + this);
            
            if (ThemeManager.getCurrentTheme().isMouseHoverSensitive())
            {
               repaint();
            }
         }
      }
   }   
   
   /**
    * Notification of a mouse dragged event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (isEnabled())
      {
         boolean wasOver = mouseOver;
         
         // get coords of the drag event related to button
         NativePoint mp = translate(e);
         // get button location and size
         NativeDimension d = physicalDimension();
         
         // recalculate the mouseOver value (whether the event is inside the bound of this widget)
         mouseOver = (0 <= mp.x && mp.x <= d.width && 0 <= mp.y && mp.y <= d.height);
         if (wasOver != mouseOver)
         {
            repaint();
         }
      }
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.  Redraw button in released
    * state.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressedLoc != null && e.getButton() == MouseEvent.BUTTON1)
      {
         java.awt.Point oldLoc = pressedLoc;
         pressedLoc = null;
         
         if (isFocusable())
         {
            requestFocus();
         }
         
         repaint();
         
         if (mouseOver && !e.getLocationOnScreen().equals(oldLoc))
         {
            // Java will not trigger click event if the mouse has moved while button was pressed
            // (drag inside the button surface). 
            // We need to emulate the click as ABL does: do the click if at the end of drag, the
            // mouse cursor is inside button bounds
            mouseClicked(e);
            
            // TODO: maybe we should ignore completely the click event and use only release event,
            //       normally, without e.getLocationOnScreen().equals(oldLoc) condition
         }
      }
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_MOVED,
         MouseEvent.MOUSE_DRAGGED,
      };
   }
}
