/*
** Module   : GuiFontResolver.java
** Abstract : GUI widget font resolver
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20141022 Created initial version.
** 002 CA  20141220 When working with fonts, these need to be scaled to match their legacy width
**                  and height.
** 003 CA  20150317 If a font is not specified, resolve it from the default-font or 
**                  default-fixed-font, depending if the widget requires a fixed font or not. 
** 004 CA  20150406 If undefined font, get it from the frame (if exists).
** 005 CA  20150507 No need to re-scale the font, it gets scaled when it gets created - so all 
**                  fonts are assumed to match the legacy metrics.
** 006 GES 20150425 Eliminated AWT dependencies.
** 007 HC  20160628 Fixed an unhandled exception when resolving font of an destroyed/unregistered
**                  frame.
** 008 SBI 20180710 Changed to register this font resolver within FontManager.
** 009 SVL 20180529 Added support for explicitly specified font.
** 010 SBI 20180726 Changed to notify widgets to update their models.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Resolve widget font. 
 */
public class GuiFontResolver
{
   /** <code>true</code> when there is a newly added font. */
   public boolean fontChanged;
   
   /** Widget font cache. */
   private Map<Integer, FontCache> fontCache;

   /** Window instance. */
   private Window<?> window;
      
   /** Screen driver. */
   private GuiDriver gd;
   
   /** Widget config. */
   private BaseConfig config;

   /**
    * Flag identifying that the referenced font must default to the default fixed font, if is 
    * not specified.
    */
   private boolean fixedFont;

   /**
    * Explicitly specified font which takes precedence over the font specified by the font number
    * in the widget config.
    */
   private FontCache explicitFontCache;

   /**
    * Constructor.
    * <p>
    * If the referenced {@link BaseConfig#font} is undefined, it defaults to the 
    * {@link FontManager#DEFAULT_FONT}.
    * 
    * @param    window
    *           Window instance.
    * @param    gd
    *           Screen driver.
    * @param    config
    *           Widget configuration.
    */
   public GuiFontResolver(Window<?>  window, GuiDriver gd, BaseConfig config)
   {
      this(window, gd, config, false);
   }

   /**
    * Constructor.
    * 
    * @param    window
    *           Window instance.
    * @param    gd
    *           Screen driver.
    * @param    config
    *           Widget configuration.
    * @param    fixedFont
    *           Flag identifying that the referenced font must default to the default fixed font, 
    *           if is not specified.
    */
   public GuiFontResolver(Window<?> window, GuiDriver gd, BaseConfig config, boolean fixedFont)
   {
      this(window, gd, config, fixedFont, null);
   }

   /**
    * Constructor.
    *
    * @param    window
    *           Window instance.
    * @param    gd
    *           Screen driver.
    * @param    config
    *           Widget configuration.
    * @param    fixedFont
    *           Flag identifying that the referenced font must default to the default fixed font,
    *           if is not specified.
    * @param    explicitFont
    *           Explicitly specified font which takes precedence over the font specified by the
    *           font number in the widget config.
    */
   public GuiFontResolver(Window<?> window,
                          GuiDriver gd,
                          BaseConfig config,
                          boolean fixedFont,
                          FontDetails explicitFont)
   {
      this.window = window;
      this.gd = gd;
      this.config = config;
      this.fixedFont = fixedFont;
      this.fontCache = new HashMap<>();
      
      if (explicitFont != null)
      {
         int fontHeight = FontManager.getFontHeight(window, explicitFont, null, gd);
         explicitFontCache = new FontCache(fontHeight, explicitFont);
      }

      FontManager.registerFontResolver(window, config, this);
   }
   
   /**
    * Resolve font.
    */
   public FontCache font()
   {
      if (explicitFontCache != null)
      {
         return explicitFontCache;
      }

      int font = resolveFontNum();

      FontCache cached = fontCache.get(font);
      
      if (cached == null)
      {
         fontChanged = true;
         FontDetails<?> guiFont = FontManager.resolveFont(window, font, gd);
         
         int legacyHeight = FontManager.getFontHeight(window, font, gd);
         int legacyWidth = FontManager.getFontWidth(window, font, gd);
         
         // font is already scaled when is created, no reason to scale it again
         cached = new FontCache(legacyHeight, guiFont);
         fontCache.put(font, cached);
      }
      else
      {
         fontChanged = false;
      }
      
      return cached;
   }
   
   /**
    * Resolve the font number, in case the configuration has an undefined font.
    * 
    * @return   If the {@link #config configured} font is undefined, then either the 
    *           {@link FontManager#DEFAULT_FONT} or {@link FontManager#DEFAULT_FIXED_FONT} is 
    *           returned, depending on the {@link #fixedFont} flag.
    */
   public int resolveFontNum()
   {
      if (explicitFontCache != null)
      {
         throw new IllegalStateException("Widgets with an explicit font shouldn't operate with " +
                                         "font number.");
      }

      int font = config.font;
      
      if (config.font == FontManager.UNDEFINED_FONT && config.frameId != -1)
      {
         // get it from the frame
         Frame<?> w = (Frame<?>) ThinClient.getInstance().getWidget(config.frameId);
         
         if (w != null && w.config() != null)
         {
            font = w.config().font;
         }
      }
      
      return FontManager.resolveFont(font, fixedFont);
   }

   /**
    * Clears the given font from this font cache in order to resolve it again as this font has
    * been changed, then it repaints the affected widget.
    * 
    * @param    font
    *           The given font number
    * @param    newFont
    *           The new font description
    * 
    * @return   True if the assigned font is a new one, otherwise it returns false.
    */
   public boolean updateFontCache(int font, FontDetails<?> newFont)
   {
      boolean result = false;
      
      FontCache fontWrapper = fontCache.remove(font);
      
      if (fontWrapper != null)
      {
         if (!fontWrapper.font.key.equals(newFont.key))
         {
            result = true;
            
            // notify widget to update its model on repaint
            fontChanged = true;
         }
      }
      
      return result;
   }

   /**
    * Font cache item.
    */
   public static class FontCache
   {
      /** Font height. */
      public final int fontHeight;
      
      /** Font. */
      public final FontDetails<?> font;
      
      /**
       * Constructor.
       * 
       * @param    fontSize
       *           Font size.
       * @param    font
       *           Font instance.
       */
      public FontCache(int fontSize, FontDetails<?> font)
      {
         this.fontHeight = fontSize;
         this.font = font;
      }
   }
}
