/*
** Module   : Hyperlink.java
** Abstract : Hyperlink GUI widget.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180614 Created initial version for in-browse use.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import java.awt.event.*;
import java.util.*;

/**
 * Hyperlink GUI widget.
 */
public class Hyperlink
extends AbstractWidget<GuiOutputManager>
{
   /** Screen driver. */
   private final GuiDriver gd;

   /** Hyperlink area width. */
   private double width;

   /** Hyperlink area height. */
   private double height;

   /** Hyperlink font. */
   private final FontDetails font;

   /** Hyperlink text. */
   private final String text;

   /** Hyperlink widget ID. */
   private final WidgetId widgetId;

   /** Top offset of the hyperlink text inside the clip rectangle. */
   private int textTopOffset;

   /**
    * Left offset of the hyperlink text inside the clip rectangle. <code>0</code> or a negative
    * number.
    */
   private int textLeftOffset;

   /** 0-based viewport index of the browse row which contains the hyperlink. */
   private final int viewportRowIndex;

   /** Mouse hover handler. */
   private MouseHoverAction mouseHoverAction;

   /** Parent column which contains the hyperlink. */
   private final BrowseColumnGuiImpl column;

   /**
    * Creates the hyperlink for the specific cell of the browse. Does NOT perform actual widget
    * layout.
    *
    * @param column
    *        Parent browse column.
    * @param text
    *        Hyperlink text.
    * @param viewportRowIndex
    *        0-based viewport index of the browse row which contains the hyperlink.
    */
   public Hyperlink(BrowseColumnGuiImpl column,
                    String text,
                    int viewportRowIndex)
   {
      widgetId = WidgetId.nextID();
      gd = (GuiDriver) OutputManager.getDriver();

      Theme theme = ThemeManager.getCurrentTheme();
      font = theme.getHyperlinkFont(column.window().getId(),
                                    column.getCellsFont());
      this.text = text;
      this.column = column;
      this.viewportRowIndex = viewportRowIndex;
   }

   /**
    * Do layout of this hyperlink (calculate widget size, position etc.).
    */
   public void doLayout()
   {
      CoordinatesConversion cc = screen().coordinates();
      NativePoint clipPoint = column.getCellTextClipPoint(viewportRowIndex);
      NativeRectangle clipRect = column.getCellTextClipRectangle(clipPoint);

      int clipWidth = clipRect.width();
      int x = clipPoint.x;
      int textWidth = column.getTextWidth(text, font);
      int widthPx = Math.min(clipWidth, textWidth);
      if (Browse.isRightAligned(column))
      {
         int diff = clipWidth - textWidth;
         if (diff < 0)
         {
            textLeftOffset = diff;
         }
         else if (diff > 0)
         {
            x += diff;
         }
      }

      height = cc.heightFromNative(clipRect.height());
      this.width = cc.widthFromNative(widthPx);
      setPhysicalLocation(x, clipPoint.y);

      textTopOffset = column.getCellTextTopOffset();

      column.add(this);
      screen().getRegistry().addWidget(this);

      mouseHoverAction = createMouseHoverAction();
      mouseHoverAction.gd().selectWindow(topLevelWindow().getId().asInt());
      mouseHoverAction.register();
      mouseHoverAction.gd().releaseWindow();
   }

   /**
    * Get font used to render this hyperlink.
    *
    * @return font used to render this hyperlink.
    */
   public FontDetails getFont()
   {
      return font;
   }

   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      boolean selectedRow = column.getBrowse().drawRowAsSelected(viewportRowIndex);
      NativeRectangle clipRect = clipRectangle(physicalLocation(), physicalDimension());
      gd.draw(physicalLocation(), clipRect, ()->
      {
         //gd.setColor(new ColorRgb(255, 204, 153));
         //gd.fillRect(0, 0, physicalDimension().width, physicalDimension().height);

         gd.setColor(ThemeManager.getCurrentTheme().getHyperlinkColor(selectedRow));
         gd.setGuiFont(font);
         gd.drawString(text, textLeftOffset, textTopOffset);
      });
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return height;
   }

   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return width;
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      BrowseGuiImpl browse = column.getBrowse();
      if (browse.isEditPossible() && !column.config().readOnly && !browse.isCtrlModifier(e))
      {
         column.mouseClicked(e);
      }
      else
      {
         ThinClient.getInstance().openHyperlink(column, text);
      }
   }

   /**
    * Determine the mouse actions processed by this widget.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_MOVED,
         MouseEvt.MOUSE_HOVERABLE
      };
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      column.mouseWheelMoved(e);
   }

   /**
    * Returns widget ID.
    *
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      return widgetId;
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method
    *
    * @return <code>true</code> if the widget supports
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Destroy this widget.  Notify all listeners about the widget being destroyed.
    */
   @Override
   public void destroy()
   {
      mouseHoverAction.deregister();
      screen().getRegistry().removeWidget(getId());
      super.destroy();
   }
   
   /**
    * Get default mouse pointer for this widget
    * 
    * @return   MousePtr or <code>Optional.empty()</code> if one is inherited from the enclosing 
    *           frame
    *    
    */
   @Override
   public Optional<MousePtrWrapper> defaultMousePtr()
   {
      return MousePtrWrapper.HAND_PTR;
   }
}
