/* 
** Module   : LabelGuiImpl.java
** Abstract : Implementation for LABEL in GUI clients.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Second pass at LABEL in GUI mode (side-labels currently).  The label's width 
**                  is not yet computed properly.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed label drawing and size in GUI.
** 005 CA  20150507 Fixed label drawing - centered vertically.
** 006 CA  20150510 Set the label's font, before drawing the text.
** 007 GES 20150425 Eliminated AWT dependencies.
** 008 CA  20150728 Do not allow the label text to be drawn outside of screen.
** 009 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 010 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 011 CA  20151019 Fixed drawing of top-labels in frame.
** 012 HC  20151112 Implemented THREE-D runtime support.
** 013 CA  20160216 Cache the intermediate widget states.
** 014 HC  20160519 Fixed label to be drawn starting at horizontal position 0.
** 015 EVL 20160823 Fix for column calculation with side-labeled widgets.
** 016 EVL 20160921 Minor fix initial end column calculation.
** 017 HC  20161018 Frame layout improvements.
** 018 CA  20161101 Added support for mnemonic drawing.
** 019 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Label;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation for a widget's LABEL, in a GUI client.
 */
public class LabelGuiImpl
extends Label<GuiOutputManager>
{
   /** Font resolver */
   private GuiFontResolver gf;

   /** Screen driver */
   private GuiDriver gd;

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Label to copy parameters from.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public LabelGuiImpl(WidgetId id, Label<GuiOutputManager> labelInstance, int frameId)
   {
      super(id, labelInstance, frameId);

      initialize(id, labelInstance.config());
   }

   /**
    * Constructs label with given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Initial text.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public LabelGuiImpl(WidgetId id, String text, int frameId)
   {
      super(id, text, frameId);
      
      initialize(id, config());
   }
   
   /**
    * Retrieve the widget's width in native units.
    *
    * TODO: this is duplicated in {@link FrameGuiImpl#getLabelWidth}, fix this.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (gf == null)
      {
         return 0;
      }
      
      int font = gf.resolveFontNum();
      
      int nativeWidth = getTextWidth(name(), font);
      
      return nativeWidth;
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * TODO: this is duplicated in {@link FrameGuiImpl#getLabelHeight}, fix this.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      Window<?> window = WindowManager.resolveWindow(this);
      int legacyHeight = 0;
      boolean text = true;
      // top-label case: the height is always the default-fixed-font's height!
      int font = FontManager.DEFAULT_FIXED_FONT;
      if (sideWidget != null)
      {
         // side-label case: use the same height as the widget's

         font = gf.resolveFontNum();
         text = (sideWidget instanceof Text);
      }
      
      legacyHeight = FontManager.getFontHeight(window, font, screen().getInstanceDriver());
      if (!text)
      {
         boolean threeD = FontManager.isThreeD();
         legacyHeight = legacyHeight + (int) (legacyHeight / 2) + (threeD ? 2 : 0);
      }

      return legacyHeight;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, LabelConfig cfg)
   {
      super.initialize(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      
      // the label's window is resolved from the frame... the label is not necessarily attached
      // to the frame at the time of this call, just the frame ID is known

      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent.window();
      
      gf = new GuiFontResolver(window, gd, config(), false);

      updateSize();
   }

   /**
    * Compute the label's width.
    * 
    * @return   See above.
    */
   @Override
   public double width()
   {
      if (config().fixedWidth)
      {
         return config().widthChars;
      }
      
      return super.width();
   }
   
   /**
    * Compute the label's height.
    * 
    * @return   See above.
    */
   @Override
   public double height()
   {
      if (config().fixedHeight)
      {
         return config().heightChars;
      }
      
      return super.height();
   }

   /**
    * Draws text.
    */
   @Override
   public void draw()
   {
      if (parent() == null)
      {
         return;
      }
      
      int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
      GuiColorResolver gc = new GuiColorResolver(gd,
                                                 config(),
                                                 bgColor,
                                                 SysColor.COLOR_FILLIN_FG);
      gc.colors();
      
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      Theme theme = ThemeManager.getCurrentTheme();
      
      gd.draw(getId(), p, clip, () ->
            theme.drawLabel(this, gd, gc, gf, 0, 0, dim.width, dim.height, emptyMode));
      
      if (isUnderlined())
      {
         // if underlined, draw a horizontal bar.  The width of this horizontal bar is
         // the label's width
         final int ppr = ThinClient.getInstance().getPixelsPerRow();
         
         final NativeDimension sepDim = new NativeDimension(dim.width, ppr);
         NativePoint sepLoc = new NativePoint(p.x, (isDelimiter() ? p.y : dim.height));
         NativeRectangle sepClip = clipRectangle(sepLoc, sepDim);
         
         gd.draw(getId(), sepLoc, sepClip, () ->
               theme.drawLabelUnderline(gd, 0, ppr / 2, sepDim.width, 1, gc.fgColor));
      }
      
      if (disableEmptyMode)
      {
         emptyMode = false;
      }
      
      disableEmptyMode = false;
   }
   
   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      NativeDimension dim = super.physicalDimension();
      
      if (sideWidget != null)
      {
         NativePoint swLoc = sideWidget.physicalLocation();
         NativePoint thisLoc = physicalLocation();
         if (swLoc.y == thisLoc.y)
         {
            dim.width = swLoc.x - thisLoc.x;   
         }
      }
      
      return dim;
   }
}