/*
** Module   : PrintSetupDialog.java
** Abstract : Print setup dialog.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20171024 Initial version.
** 002 HC  20180611 Fixed several issues in runtime support of FRAME:VIRTUAL-* attributes: wrong
**                  size reported by the VIRTUAL* attributes, wrong frame outer size calculated,
**                  wrong inner size calculated. And related changes.
** 003 OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

import java.util.*;

/**
 * GUI print setup dialog.
 */
public class PrintSetupDialog
extends DialogBoxWindow
{
   /** Width of buttons in characters. */
   private static double BUTTONS_CHAR_WIDTH = 15;

   /** Height of buttons in characters. */
   private static double BUTTONS_CHAR_HEIGHT = 1.04;

   /** Widget factory */
   private final GuiWidgetFactory widgetFactory;

   /** Print options */
   private PrintOptions printOptions;

   /** Cached driver. */
   private GuiDriver gd;

   /** Buttons panel. */
   private ButtonsPanel buttonsPanel;

   /** Screen buffer */
   private ScreenBuffer sb;

   /** Radio set for page orientation */
   private RadioSet orientationWidget;

   /** Combo for page size */
   private ComboBox sizeWidget;

   /**
    * Default constructor.
    */
   public PrintSetupDialog(PrintOptions printOptions)
   {
      super("Print");
      widgetFactory = screen().getFactory();
      this.printOptions = new PrintOptions(printOptions);
   }
   
   /**
    * Layout.
    */
   @Override
   public void doLayout()
   {
      Frame frame = getFrame();
      Container<GuiOutputManager> content = getContentPane();

      double frameWidth = frame.width();

      buttonsPanel.setLocation(0, frame.height());
      buttonsPanel.setWidth(frameWidth);
      buttonsPanel.doLayout();
      
      double contentWidth = Math.max(frame.width(), buttonsPanel.width());
      content.setSize(new Dimension(contentWidth, frame.height() + buttonsPanel.height()));
      
      super.doLayout();
      
      // place the window at the display center
      NativeRectangle bounds = ThinClient.getInstance().getDisplayBounds(getId().asInt());
      NativeDimension dim = physicalDimension();
      setPhysicalLocation(bounds.left() + (bounds.width()  - dim.width) / 2, 
                          bounds.top()  + (bounds.height() - dim.height) / 2);
   }

   /**
    * This method must be called when the associated frame instance has
    * changed its dimension.
    */
   @Override
   public void onFrameDimensionChanged()
   {
      if (!isRealized())
      {
         return;
      }

      FrameGuiImpl frame = getFrame();
      Dimension dim = frame.dimension();
      FrameConfig fcfg = frame.config();
      getContentPane().setSize(new Dimension(dim.width, dim.height + buttonsPanel.height()));
      doLayout();

      // properly set up width and height for dialog to get by the server side
      // direct access to frame config, pixel values will be updated automatically
      Dimension decorDim = frame.getDialogDecorationCharSize();
      if (!fcfg.fixedHeight && fcfg.clientHeightChars != ConfigHelper.getEffectiveVirtualHeight(fcfg))
      {
         fcfg.clientHeightChars = fcfg.heightChars + decorDim.height;
      }
      if (!fcfg.fixedWidth && fcfg.clientWidthChars != ConfigHelper.getEffectiveVirtualWidth(fcfg))
      {
         fcfg.clientWidthChars = fcfg.widthChars + decorDim.width;
      }
   }

   /**
    * Show widget.
    */
   @Override
   public void realize()
   {
      super.realize();
      build();
   }
   
   /**
    * Close this window. This should be only called when window close caption
    * button is cliced.
    */
   @Override
   public void close()
   {
      dismiss();
   }

   /**
    * Let the modal UI element now that it should finish its modal input processing.
    */
   @Override
   public void dismiss()
   {
      destroy();
   }

   /**
    * Returns <code>true</code> when the modal UI element has dismissed (finished its modal input).
    *
    * @return  see above.
    */
   @Override
   public boolean isDismissed()
   {
      return !isVisible();
   }

   /**
    * Helper method to get the edited result when finished.
    * 
    * @return   The result of the editing operation which was set depending
    *           on the pressed button.
    */
   public PrintOptions getOptions()
   {
      return printOptions;
   }

   /**
    * Destroy this window.
    */
   @Override
   public void destroy()
   {
      super.destroy();

      if (gd == null)
      {
         return;
      }

      GuiOutputManager screen = screen();
      screen.getRegistry().removeFrame(getFrame());

      // frame removal from registry will trigger frame destroy,
      // that hides and destroys the modal window

      gd = null;
   }

   /**
    * Returns the frame that is on top of the Z-order or <code>null</code>
    * if no such frame present. If requested method also can filter out all
    * invisible, disabled and out of scope frames.
    *
    * @param ignoreState if <code>true</code> then ignore frame state during search.
    * @return The topmost active frame or <code>null</code>.
    */
   @Override
   public Frame<?> getTopFrame(boolean ignoreState)
   {
      // printer setup dialog doesn't host a legacy frame that should participate in the legacy frame
      // processing
      return null;
   }

   /**
    * Initializes this instance. 
    */
   private void build()
   {
      if (gd != null)
      {
         // already initialized
         return;
      }
      
      // the titlebar can be null when the web client is in embedded mode
      WindowTitleBar tbar = getTitleBar();
      if (tbar == null)
      {
         return;
      }
      tbar.setClickHandler(CaptionButtonType.CLOSE, this::dismiss);
      
      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      
      int wndId = getId().asInt();
      
      BorderedPanelGuiImpl content = (BorderedPanelGuiImpl) getContentPane();
      content.setLayout(new StackLayout<>(StackLayout.Direction.VERTICAL));

      buttonsPanel = new ButtonsPanel();

      FrameConfig frameConfig = new FrameConfig(WidgetId.nextID());
      frameConfig.box         = false;
      frameConfig.overlay     = true;
      frameConfig.labels      = false;
      frameConfig.sideLabels  = false;
      frameConfig.windowID    = wndId;
      frameConfig.dialog      = true;
      frameConfig.sideLabels  = true;
      frameConfig.visible     = true;

      int frameId = frameConfig.id.asInt();

      FillInConfig printerCfg = new FillInConfig(WidgetId.nextID());
      printerCfg.frameId    = frameId;
      printerCfg.label      = "Printer";
      printerCfg.forceLabel = true;
      printerCfg.readOnly   = true;
      printerCfg.widthChars = 30;

      RadioSetConfig orientationCfg = new RadioSetConfig(WidgetId.nextID());
      orientationCfg.frameId    = frameId;
      orientationCfg.label      = "Page Orientation";
      orientationCfg.forceLabel = true;
      orientationCfg.horizontal = true;
      orientationCfg.items      = new ControlSetItem[2];
      orientationCfg.items[0]   = new ControlSetItem(1, new character("Portrait"), new logical(true));
      orientationCfg.items[1]   = new ControlSetItem(2, new character("Landscape"), new logical(false));

      ComboBoxConfig sizeCfg = new ComboBoxConfig(WidgetId.nextID());
      sizeCfg.frameId    = frameId;
      sizeCfg.label      = "Page Size";
      sizeCfg.forceLabel = true;
      sizeCfg.items      = buildPageSizeComboItems();

      SkipConfig skip1 = new SkipConfig(WidgetId.nextID());
      skip1.frameId    = frameId;
      skip1.vertical   = true;
      SkipConfig skip2 = new SkipConfig(WidgetId.nextID());
      skip2.frameId    = frameId;
      skip2.vertical   = true;

      GuiOutputManager screen = screen();
      WidgetRegistry<GuiOutputManager> wr = screen.getRegistry();

      wr.reconstructWidget(frameConfig);
      wr.reconstructWidget(printerCfg);
      wr.reconstructWidget(orientationCfg);
      wr.reconstructWidget(sizeCfg);
      wr.reconstructWidget(skip1);
      wr.reconstructWidget(skip2);

      FrameGuiImpl frame = (FrameGuiImpl) screen.getRegistry().getComponent(frameConfig.id);
      FillIn printerFin = (FillIn) screen.getRegistry().getComponent(printerCfg.id);
      printerFin.setForceLeft(true);
      orientationWidget = (RadioSet) screen.getRegistry().getComponent(orientationCfg.id);
      sizeWidget = (ComboBox) screen.getRegistry().getComponent(sizeCfg.id);
      sizeWidget.setEnabled(true);
      Skip skip1Widget = (Skip) screen.getRegistry().getComponent(skip1.id);
      Skip skip2Widget = (Skip) screen.getRegistry().getComponent(skip2.id);

      frame.addWidget(printerFin);
      frame.addWidget(skip1Widget);
      frame.addWidget(orientationWidget);
      frame.addWidget(skip2Widget);
      frame.addWidget(sizeWidget);

      // Frame adding/removal is performed manually,
      // so we must protect if from automated processing.
      frame.setProtected(true);
      frame.setLocation(1, 0.5);

      sb = new ScreenBuffer(frameId, 3);
      sb.putWidgetValue(printerCfg.id.asInt(), new character(PrintingService.PDF_PRINTER_DESCR));
      sb.putWidgetValue(orientationCfg.id.asInt(),
                        new logical(printOptions.pageOrientation == PrintOptions.PageOrientation.PORTRAIT));
      sb.putWidgetValue(sizeCfg.id.asInt(), new integer(printOptions.pageSize.ordinal()));

      ThinClient tc = ThinClient.getInstance();
      tc.enable(frameId, null, new ScreenBuffer[] { sb }, true, wndId, null);
      sb.resetChanged();

      content.add(frame);
      content.add(buttonsPanel);
      buttonsPanel.init();

      frame.doLayout();

      setColorResolver(frame.getColorResolver());
      buttonsPanel.setColorResolver(new GuiColorResolver(gd,
                                                         new BaseConfig(),
                                                         SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT));
   }

   /**
    * Get the maximum size of the window workspace area.
    *
    * @return See above.
    */
   @Override
   public Dimension getMaxWorkspaceSize()
   {
      NativeRectangle nrect = gd.getDisplayWorkArea(this.getId().asInt());
      return screen().coordinates().dimensionFromNative(new NativeDimension(nrect.width(), nrect.height()));
   }

   /**
    * Returns the items of the page size combo.
    *
    * @return  See above.
    */
   private ControlSetItem[] buildPageSizeComboItems()
   {
      PrintOptions.PageSize[] sizes = PrintOptions.PageSize.values();
      ControlSetItem[] result = new ControlSetItem[sizes.length];
      for (int i = 0; i < sizes.length; i++)
      {
         result[i] = new ControlSetItem(i + 1, new character(sizes[i].name()), new integer(i));
      }

      return result;
   }

   /**
    * Saves the screen values of the frame widgets in {@link #printOptions}.
    */
   private void saveOptions()
   {
      boolean portrait = ((logical) orientationWidget.getValue()).booleanValue();
      printOptions.pageOrientation = portrait ?
                                     PrintOptions.PageOrientation.PORTRAIT :
                                     PrintOptions.PageOrientation.LANDSCAPE;

      int i = sizeWidget.getSelection()[0];
      printOptions.pageSize = PrintOptions.PageSize.values()[i];
   }

   /**
    * The class implements the buttons-panel part of the alert box widget.
    */
   private class ButtonsPanel
   extends BorderedPanelGuiImpl
   implements ActionListener
   {
      /** Panel padding in pixels. */
      private static final int PADDING = 10;
      
      /** Button spacing in characters. */
      private static final double SPACING = 10;
      
      /** Extra left border padding. */
      private static final int LEFT_BORDER_PADDING = 40;
      
      /** Extra right border padding. */
      private static final int RIGHT_BORDER_PADDING = 40;

      /**
       * Initializes the panel.
       */
      public void init()
      {
         createButton("OK");
         createButton("Cancel");
      }

      /**
       * Layout.
       */
      @Override
      public void doLayout()
      {
         CoordinatesConversion cc = screen().coordinates();
         List<Button<GuiOutputManager>> buttons = new ArrayList<>(2);
         for (Widget w : widgets())
         {
            if (w instanceof Button)
            {
               buttons.add((Button<GuiOutputManager>) w);
            }
         }

         int buttonNativeWidth  = cc.widthToNative(BUTTONS_CHAR_WIDTH);
         int buttonNativeHeight = cc.heightToNative(BUTTONS_CHAR_HEIGHT);
         int buttonSpacing      = cc.widthToNative(SPACING);

         // it is expected that panel width has been set
         NativeDimension dim = physicalDimension();
         
         // padding
         int minWidth = buttons.size() * buttonNativeWidth + 
                        (buttons.size() - 1) * buttonSpacing +
                        RIGHT_BORDER_PADDING + LEFT_BORDER_PADDING;
         
         dim.width  = Math.max(dim.width, minWidth);
         dim.height = buttonNativeHeight + 2*PADDING;
         
         setSize(cc.dimensionFromNative(dim));
         
         int buttonX = dim.width - RIGHT_BORDER_PADDING - buttonNativeWidth;
         int buttonY = dim.height / 2 - buttonNativeHeight / 2; 
         
         // iterate buttons in reverse order
         ListIterator<Button<GuiOutputManager>> it = buttons.listIterator(buttons.size());
         while (it.hasPrevious())
         {
            Button<GuiOutputManager> button = it.previous();
            button.config().autoResize = false;
            button.config().widthChars = BUTTONS_CHAR_WIDTH;
            button.config().heightChars = BUTTONS_CHAR_HEIGHT;
            button.setPhysicalLocation(buttonX, buttonY);
            
            buttonX -= buttonNativeWidth + buttonSpacing; 
         }

         if (!buttons.isEmpty())
         {
            // make OK the default button
            Button<GuiOutputManager> defaultButton = buttons.get(0);
            defaultButton.parent().setFocus(defaultButton);
            defaultButton.setHighlighted(true);
         }
      }
      
      /**
       * Action handler to process the AlertBox buttons event. Each event
       * changes the return code appropriately.
       *
       * @param    action
       *           The action event to handle is coming from defined buttons.
       */
      public void onAction(ActionEvent action)
      {
         if (action.command().equals("OK"))
         {
            saveOptions();
         }
         else
         {
            // Cancel chosen
            printOptions = null;
         }

         dismiss();
      }

      /**
       * Creates and initializes a button widget to be used in the alert box.
       *
       * @return   A reference to {@link Button} instance.
       */
      private ButtonGuiImpl createButton(String text)
      {
         ButtonGuiImpl b = widgetFactory.createButton(WidgetId.nextID(), buttonsPanel, text);
         b.initialize(b.getId(), null);
         b.setTextLabel(text);
         b.setCommand(text);
         b.setEnabled(true);

         b.addListener(this);
         return b;
      }
   }
}
