/*
** Module   : RadioSetGuiImpl.java
** Abstract : RadioSet implementation for GUI Radio-set
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150923 Initial version.
** 002 EVL 20151009 Changing approach to get the native text width.  GuiDriver cal is no more
**                  safe to use.  Shifted to AbstractWidget.getText(Width|Height)Native.
** 003 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 004 HC  20151112 Implemented THREE-D runtime support.
** 005 CA  20151125 Added mouseActions() API (to allow web client mouse interactions).
** 006 IAS 20151214 LOAD-MOUSE-POINTER support.
** 007 CA  20160216 Cache the intermediate widget states.
**                  Disabled radio-set must not respond to mouse press/click.
** 008 SVL 20160719 Updated translate() signature.
** 009 EVL 20160926 Fix for frame buffer out of sync with radio-set value.  When widget become
**                  realized the state is CHANGED for server side to grab the initial value.
** 010 CA  20160930 Fixed radio-button position in horizontal and expanded mode.
** 011 CA  20161101 Added support for mnemonic processing.
** 012 OM  20170320 Added support for Themes.
** 013 EVL 20170824 Adding file list drag-and-drop support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.util.*;

/**
 * RadioSet is a implementation of radio-set widget, a mutually exclusive set
 * of {@link RadioButton} objects.
 * <p>
 * This is GUI specific implementation of the widget.
 */
@SuppressWarnings("unchecked")
public class RadioSetGuiImpl
extends RadioSet<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Selection in progress flag. */
   private boolean selecting = false;
   
   /** Flag to keep track first time state check when widget become realized. */
   private boolean firstStateCheck = true;
   
   /** The padding of radio-buttons to the right-side. */
   private int rightPadding = 0;

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RadioSetGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, cfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RadioSetConfig cfg)
   {
      super.initialize(id, cfg);

      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
   }
   
   /**
    * Refresh widget look using new set of buttons, after the concrete implementation has done any
    * other needed calculations.
    * <p>
    * In GUI, this computes the right-padding to be added to each radio-button.
    * 
    * @param    selectedId
    *           The selected button ID; -1 if none selected.
    * @param    selectedDisabled
    *           Flag indicating if the selected button is disabled.
    */
   @Override
   public void refreshItemsImpl(int selectedId, boolean selectedDisabled)
   {
      rightPadding = 0;
      
      if (config.horizontal && !config.expand && config.fixedWidth && !buttons.isEmpty())
      {
         int totalButtonWidth = 0;
         
         for (RadioButton<GuiOutputManager> button : buttons)
         {
            RadioButtonGuiImpl b = (RadioButtonGuiImpl) button;
            
            totalButtonWidth += b.widthNative();
         }
      
         if (totalButtonWidth < config.widthPixels)
         {
            // in horizontal, non-expanded mode, the location of the radio buttons is computed 
            // following these rules:
            // - compute the sum of all the button sizes (including the 'knob' image)
            // - a fix-sized padding will be added to each button, on its right side, computed by
            // dividing (by the number of buttons) the difference between the radio-set's width
            // and the the sum of the native button widths
            
            int extraPadding = config.widthPixels - totalButtonWidth;
            rightPadding = extraPadding / buttons.size();
         }
      }
      
      super.refreshItemsImpl(selectedId, selectedDisabled);
   }
   
   /**
    * Gets the current state of this component's data.
    * 
    * @return   The state of this widget's data  (for example 
    *           <code>ScreenBuffer.UNINITIALIZED</code>).
    */
   @Override
   public byte getState()
   {
      // initial state for radio-set on the client side is changed just after realization
      if (firstStateCheck && isRealized())
      {
         firstStateCheck = false;
         return ScreenBuffer.CHANGED;
      }
      else
      {
         return super.getState();
      }
   }

   /**
    * Returns the width of the radio-set widget.
    * 
    * @return  The radio-set width value.
    */
   @Override
   public double width()
   {
      double wRes = super.width();
      
      if (wRes != 0.0)
      {
         return wRes;
      }

      double wResBtn = 0.0;
      
      // compare to the longest button
      for (RadioButton<GuiOutputManager> button : buttons)
      {
         double btnWidth = button.width();
         if (config.horizontal)
         {
            wResBtn += btnWidth + 1;
         }
         else
         {
            if (btnWidth > wResBtn)
            {
               wResBtn = btnWidth; 
            }
         }
      }
      
      return wResBtn > wRes ? wResBtn : wRes;
   }

   /**
    * Returns the height of the radio-set widget.
    * 
    * @return  The radio-set height value.
    */
   @Override
   public double height()
   {
      double hRes = super.height();

      if (hRes != 0.0)
      {
         return hRes;
      }

      // calculate height dynamically
      for (RadioButton<GuiOutputManager> button : buttons)
      {
         double btnHeight = button.height();
         if (config.horizontal)
         {
            if (btnHeight > hRes)
            {
               hRes = btnHeight;
            }
         }
         else
         {
            hRes += btnHeight;
         }
      }
      
      return hRes;
   }

   /**
    * Draws RadioSet widget in GUI mode.
    */
   @Override
   public void draw()
   {
      // do nothing if not visible
      if (!isVisible())
      {
         return;
      }
      
      // init color resolver
      final RadioSetConfig rscfg = config();
      
      int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_WINDOW;
      gc = new GuiColorResolver(gd, rscfg, bgColor, SysColor.COLOR_BTNTEXT);
      // determine colors to use
      gc.colors();
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(getId(), p, clip, () ->
      {
         // fill the background
         gd.setColor(gc.bgColor);
         gd.fillRect(0, 0, d.width, d.height);
         
         // call superclass to draw the buttons
         drawInt();
      });
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      if (!isEnabled() || !isVisible())
      {
         return false;
      }

      if (!processMnemonicImpl(key, this))
      {
         // check the buttons
         int btnPressIndex = 0;
         for (RadioButton<GuiOutputManager> button : buttons)
         {
            if (button.processMnemonic(key))
            {
               if (!button.isSelected())
               {
                  selectButton(button);

                  // refresh selection dotted frame position
                  if (current != btnPressIndex)
                  {
                     buttons.get(current).setFocused(false);
                     current = btnPressIndex;
                     buttons.get(current).setFocused(true);
                  }
               
                  ConfigHelper.setModified(config, true);
                  EventManager.postEvent(new KeyInput(this,
                          EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true));

                  repaint();
               }

               return true;
            }

            btnPressIndex++;
         }
         
         return false;
      }
      else
      {
         return true;
      }
   }

   /**
    * This overriding is used to move the focus from this widget to the
    * next/previous one, rather than moving between radio buttons inside the
    * current radio-set.
    * 
    * @param   keyEvent
    *          The key event to be processed.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      // call the superclass first
      super.processKeyEvent(keyEvent);
      
      // arrow keys generate VALUE-CHANGED so emits spacebar key
      int action = keyEvent.actionCode();
      if (action == Keyboard.KA_CURSOR_UP || action == Keyboard.KA_CURSOR_LEFT ||
          action == Keyboard.KA_CURSOR_DOWN || action == Keyboard.KA_CURSOR_RIGHT )
      {
         KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_CHOOSE);
         ki.setKeyCode(' ');
         EventManager.postEvent(ki);
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled() || e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }
      
      synchronized (this)
      {
         // try to find the pressed item index
         int btnPressIndex = getRadioButtonIndex(e);
         // found the button
         if (btnPressIndex != -1 && !buttons.get(btnPressIndex).isSelected())
         {
            selecting = true;
            selectButton(buttons.get(btnPressIndex));

            // refresh selection dotted frame position
            if (current != btnPressIndex)
            {
               buttons.get(current).setFocused(false);
               current = btnPressIndex;
               buttons.get(current).setFocused(true);
               repaint();
            }
         }
      }
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled() || e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }

      synchronized (this)
      {
         // try to find the clicked item index
         int btnPressIndex = getRadioButtonIndex(e);
         // found the button
         if (btnPressIndex != -1 && selecting)
         {
            // inform client the value has been changed
            // yes, this must be done on click while the selection work must be done on press
            ConfigHelper.setModified(config, true);
            EventManager.postEvent(new KeyInput(this,
                    EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true));
            // prevent double message generation
            selecting = false;
            repaint();
         }
      }
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED 
      };
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Checks if the given widget is able to be the target of the drop operation.
    * 
    * @return   <code>TRUE</code> if this widget can be droppable at this time,
    *           <code>FALSE</code> otherwise.
    */
   @Override
   public boolean isDroppable()
   {
      return config.dropTarget; 
   }

   /**
    * Gets button width for expand mode.
    * 
    * @param    items
    *           Radio-button set to check.
    * 
    * @return   The GUI dependent button width for expand mode. 
    */
   @Override
   protected double getButtonsExpandWidth(ControlSetItem[] items)
   {
      if (items.length == 0)
      {
         return 0;
      }
      
      if (config.fixedWidth)
      {
         return Coordinate.scale(config.widthChars / items.length);
      }

      double wRet = -1.0;
      
      int wPixels = 0;
      
      // ensure we have font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
      }
      
      // find longest button in pixels
      for (ControlSetItem item: items)
      {
         character label = item.getLabel();
         if (label != null)
         {
            wPixels = Math.max(wPixels, getTextWidthNative(label.toStringMessage(),
                                                           gf.resolveFontNum()));
         }
      }
      
      wPixels += ThemeManager.getCurrentTheme().getRadioButtonImageSpace();
      // convert to character units
      wRet = screen().coordinates().widthFromNative(wPixels);
      
      return wRet;
   }
   
   /**
    * Gets extra distance between radio-buttons.
    * 
    * @return   The GUI dependent distance between buttons. 
    */
   @Override
   protected double getButtonsExtraDistance()
   {
      return 0.0;
   }
   
   /**
    * Gets the list height in character units for radio-set.
    * 
    * @return   The GUI dependent height for item list. 
    */
   @Override
   protected double getListHeightChars()
   {
      double hRet = config.listHeightChars;

      // sometimes on this step the buttons are not yet ready(dynamic widget)
      if (buttons.size() == 0)
      {
         return hRet; 
      }
      
      // horizontal and vertical orientation has different ways
      if (config.horizontal)
      {
         hRet = buttons.get(0).height();
      }
      else
      {
         hRet *= buttons.get(0).height();
      }
      
      return hRet;
   }

   /**
    * Gets the list width in character units for radio-set.
    * 
    * @return   The GUI dependent width for item list. 
    */
   @Override
   protected double getListWidthChars()
   {
      double wRet = config.listWidthChars;
      Theme theme = ThemeManager.getCurrentTheme();
      
      // horizontal and vertical orientation has different ways
      if (config.horizontal)
      {
         int listSize = config.items.length;
         // explicitly defined width has priority
         if (config.widthChars > 0)
         {
            wRet = config.widthChars;
         }
         // remove ChUI extra characters(3.0 + 1.0)
         else if (config.expand)
         {
            wRet = getButtonsExpandWidth(config.items) * listSize; 
         }
         else
         {
            wRet += (screen().coordinates().widthFromNative(theme.getRadioButtonImageSpace()) - 2.0) * listSize;
            if (listSize > 1)
            {
               wRet += (listSize - 1) * screen().coordinates().widthFromNative(
                                                         theme.getRadioButtonExtraDist() / 2);
            }
         }
      }
      else
      {
         wRet += screen().coordinates().widthFromNative(theme.getRadioButtonImageSpace() + 3) + 1.0;
      }
      
      return wRet;
   }
   
   /**
    * Get the right-padding width to be added to each button, in horizontal mode.
    * 
    * @return   The {@link #rightPadding} value.
    */
   int getRightPadding()
   {
      return rightPadding;
   }

   /**
    * Draws radio-set widget.  Internal wodker.
    */
   private void drawInt()
   {
      super.draw();
   }
   
   /**
    * Locates the radio-button index for the given mouse event coordinates inside radio-set.
    * 
    * @param   mEvt
    *          The mouse event to be checked.
    * 
    * @return  The radio-button that contains the point with given X,Y or <code>null</code>
    *          if none.
    */
   private int getRadioButtonIndex(MouseEvent mEvt)
   {
      int ndxRes = -1;
      
      // convert coordinates to radio-set related
      NativePoint mp = translate(mEvt);
      
      // check the buttons visible area 
      for (int i = 0; i < buttons.size(); i++)
      {
         RadioButton<GuiOutputManager> button = buttons.get(i);
         NativePoint p = button.physicalLocation();
         NativeDimension d = button.physicalDimension();
         // check ix X,Y is inside the button
         if (mp.x >= p.x && mp.x <= p.x + d.width && mp.y >= p.y && mp.y <= p.y + d.height)
         {
            ndxRes = i;
            break;
         }
      }
      
      return ndxRes;
   }
}