/*
** Module   : SelectionListBodyGuiImpl.java
** Abstract : Internal object of the selection-list widget for GUI interface.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150818 Initial version.
** 002 HC  20150831 Scrolling support overhaul
** 003 HC  20150917 Scroll fixes.
** 004 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 EVL 20151014 Fix for selection-list auto resize issue.
** 007 HC  20151112 Implemented THREE-D runtime support.
**     HC  20151117 Fixed broken SELECTION-LIST scrolling.
** 008 SVL 20160412 Javadoc fix.
** 009 SVL 20160719 Updated translate() signature.
** 010 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D). 
**     CA  20161101 SELECTION-LIST items can't be deselected once selected. 
**                  Fixed mouse processing for selection-list.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.model.*;

/**
 * GUI-specific part of the SelectionListBody.
 */
@SuppressWarnings("unchecked")
public class SelectionListBodyGuiImpl
extends SelectionListBody<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Color resolver */
   protected GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected item text color. */
   private ColorRgb selectionFore;

   /** Disabled state text color. */
   private ColorRgb disabledFore;
   
   /** Selected dotted line color. */
   private ColorRgb selectionDots;

   /** Recently selected row by mouse. */
   private int lastRow = -1;
   
   /** The current left column in the item list to display. */
   private int colLeft = 0;
   
   /** Maximum width of the item list in character units. */
   private int maxWidth;
   
   /** The width in pixels of the list visible area. */
   private int nativeWidth;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SelectionListBodyGuiImpl(WidgetId id, SelectionListConfig cfg)
   {
      super(id, cfg);
      
      allowDeselect(false);
      
      // initialize screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // and font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, cfg, false);
      // color initializing
      selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);

      // the color below is negative to selectionBack
      selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                    255 - selectionBack.getGreen(),
                                    255 - selectionBack.getBlue());

      // top row number will be initialized later in setTopRow()
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SelectionListConfig cfg)
   {
      // do superclass first
      super.initialize(id, cfg);
      
      // GUI specifics
      maxWidth = getMaxWidth();

      final SelectionListConfig slcfg = config();
      
      int bgColor = isThreeD() && !config.enabled ? 
                    SysColor.COLOR_3DFACE : SysColor.COLOR_WINDOW;
      
      gc = new GuiColorResolver(gd, slcfg, bgColor, SysColor.COLOR_BTNTEXT);
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();
   }

   /**
    * Process key event.
    * 
    * @param key
    *          key to process
    */
   @Override
   public void processKey(int key)
   {
      // check to not process direction keys when it is not possible
      if (model().size() <= 0 ||
          ((key == Key.VK_SHIFT_CURSOR_UP || key == Key.VK_CURSOR_UP) && currentRow() == 0) ||
          ((key == Key.VK_SHIFT_CURSOR_DOWN || key == Key.VK_CURSOR_DOWN) &&
            currentRow() == model().size() - 1))
      {
         // do nothing if we try to scroll outside the list model
         return;
      }
      
      // in GUI mode we need to process some keys differently
      switch (key)
      {
         case Key.VK_SHIFT_CURSOR_DOWN :
            // in multiple selection mode this must accumulate/subtract selections
            // in single selection mode it works as VK_CURSOR_DOWN 
            if (selectionMode() != ListSelectionModel.SINGLE_SELECTION &&
                selected(currentRow() + 1))
            {
               deselect(currentRow(), currentRow() + 1);
            }
            super.processKey(Key.VK_CURSOR_DOWN);
            break;

         case Key.VK_SHIFT_CURSOR_UP :
            // in multiple selection mode this must accumulate/subtract selections
            // in single selection mode it works as VK_CURSOR_UP 
            if (selectionMode() != ListSelectionModel.SINGLE_SELECTION &&
                selected(currentRow() - 1))
            {
               deselect(currentRow() - 1, currentRow());
            }
            super.processKey(Key.VK_CURSOR_UP);
            break;
            
         case Key.VK_CURSOR_LEFT :
         case Key.VK_CURSOR_RIGHT :
            {
               // this is the hook for horizontal scrolling
               // TODO: calculate whether scroll is needed and initiate scroll
               boolean needRefresh = false;
               if (key == Key.VK_CURSOR_LEFT && colLeft > 0)
               {
                  colLeft -= 1;
                  needRefresh = true;
               }
               else if (key == Key.VK_CURSOR_RIGHT && (colLeft + config().innerChars) < maxWidth)
               {
                  colLeft += 1;
                  needRefresh = true;
               }
      
               // do repaint if look was changed
              if (needRefresh)
              {
                 ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
                 {
                    @Override
                    public void run()
                    {
                       repaint();
                    }
                 });
               }
            }
            break;
            
         case Key.VK_CURSOR_DOWN :
         case Key.VK_CURSOR_UP :
            // multiple selection mode should clear the current selection if single
            // arrow key was pressed
            if (selectionMode() != ListSelectionModel.SINGLE_SELECTION)
            {
               clear();
            }

         default:
            super.processKey(key);
      }
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      int rowHeight = getRowHeight();
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(e);
            
            // calculate clicked line
            int row = n.y / rowHeight + rowTop;
            
            // calculate modifiers mask
            int modMask = e.getModifiersEx();

            // go to selected item
            super.ensureRowVisible(row, false);

            // check the key modifiers for multiple selection mode
            if (selectionMode() != ListSelectionModel.SINGLE_SELECTION)
            {
               // SHIFT key was down when mouse press, SHIFT has priority over CRTL
               if ((modMask & MouseEvent.SHIFT_DOWN_MASK) == MouseEvent.SHIFT_DOWN_MASK)
               {
                  // set new selection range depending on current selection
                  int[] currSelected = selection();
                  // by default in this case start == end == current
                  int selStart = row;
                  int selEnd = row;
                  
                  // take into account the current selection if it exists
                  if (currSelected != null && currSelected.length > 0)
                  {
                     // first find the min and max values
                     int min = currSelected[0];
                     int max = currSelected[currSelected.length - 1];
                     // check where the current row
                     if (row < min) // abobe current selection
                     {
                        // the new last selected item will be current selection start
                        selEnd = lastRow != -1 ? lastRow : min;
                     }
                     else if (row > max) // below current selection
                     {
                        // the new first selected item will be current selection end
                        selStart = lastRow != -1 ? lastRow : max;
                     }
                     else // inside current selection
                     {
                        // the new last selected item will be current selection end
                        selEnd = max;
                     }
                  }
                  
                  // clear previous selection first
                  clear();
                  // select new range
                  select(selStart, selEnd);
               }
               // CTRL key was down when mouse press
               else if ((modMask & MouseEvent.CTRL_DOWN_MASK) == MouseEvent.CTRL_DOWN_MASK)
               {
                  // add new item to the selection range
                  // select new item
                  selectItem();
               }
               // neither CTRL nor SHITFT was pressed during mouse presss 
               else
               {
                  // clear previous selection first
                  clear();
                  // select new item
                  selectItem();
               }
            }
            else
            {
               // select new item
               selectItem();
            }
            
            repaint();
            
            // remember recently selected row
            lastRow = row;
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      synchronized (this)
      {
         if (model().size() > 0)
         {
            // inform client the value has been changed
            // yes, this must be done on click while the selection work must be done on press
            EventManager.postEvent(new KeyInput(this,
                    EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true));
         }
      }
   }
   
   /**
    * Implements SCROLL-TO-ITEM() method for selection-list widget in GUI interface.
    *
    * @param    index
    *           The 0-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public boolean scrollToItem(int index)
   {
      // GUI case
      if (index < 0 || index >= model().size())
      {
         // index is out of range
         return false;
      }
      else
      {
         // recalc top row index
         if ((index + visibleRows()) < model().size())
         {
            rowTop = index;
         }
         else
         {
            rowTop = model().size() - visibleRows();
         }
         
         postScrollEvent(colLeft, rowTop);
         
         repaint();
         return true;
      }
   }
   
   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // determine colors to use
      gc.colors();

      // resolve font
      final FontCache fntCache = gf.font();

      // get actual rows to view
      final int rowsVisible = visibleRows();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      // add one pixel for panel border
      d.width++;
      final int rowHeight = getRowHeight();
      
      d.height++;

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);

      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is (0,1) to consider internal border
            NativePoint po = new NativePoint(0, 1);
            
            // fill the background
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);
            
            // draw entry field text
            gd.setGuiFont(fntCache.font);
            gd.setFontStyle(fntCache.font.style);
            
            int x = po.x;
            int y = po.y;
            int fontSize = fntCache.font.pointSize;
            int modelSize = model().size();
            
            // check if the list is enabled
            boolean enabled = isEnabled();
            // and focused
            boolean focused = highlighted();
            
            // draw items from the top to bottom
            for (int i = rowTop; i < modelSize && i < rowsVisible + rowTop;
                            i++, y+= rowHeight)
            {
               if (y < 0)
                  continue;
               
               // draw next visible item in the list          
               drawItem(i, x, y, d.width, rowHeight, fontSize, enabled, focused);
            }
         }
         
      });
   }

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      NativeDimension dimRes = super.physicalDimension();

      // in the case native dimension was changes we need to take this into account
      if (nativeWidth > 0)
      {
         dimRes.width = nativeWidth; 
      }
      
      return dimRes;
   }

   /**
    * Gets the current row height depending on choosen font.
    * 
    * @return  The row height value in pixels.
    */
   public int getRowHeight()
   {
      return FontManager.getFontHeight(WindowManager.resolveWindow(this), resolveCurrentFontNum(), gd);
   }
   
   /**
    * Gets the current char width depending on choosen font.
    * 
    * @return  The row height value in pixels.
    */
   public int getCharWidth()
   {
      return FontManager.getFontWidth(WindowManager.resolveWindow(this), resolveCurrentFontNum(), gd);
   }
   
   /**
    * Sets the visible width in pixels for list are.  Depends on the exiplicit sizing and
    * vertical scrollbar existence.  Used in drawing of the item selection dotted frame.
    * 
    * @param    widthInPixels
    *           New pixel width.
    */
   public void setInternalWidth(int widthInPixels)
   {
      nativeWidth = widthInPixels;
   }
   
   /**
    * Returns the scroll dimension of this widget. The scroll dimension is
    * the full size of the widget's content, it may be bigger than widget's
    * visible size.
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
    * 
    * @return   widget's scroll dimension
    */
   @Override
   public NativeDimension getScrollDimension()
   {
      return new NativeDimension(maxWidth, model().size());
   }
   
   /**
    * Returns the visible dimension of this widget. The visible dimension is
    * the visible portion of the widget's content (including scrollbars area).
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getScrollDimension()} and {@link #getScrollStep()}.
    *
    * @return   widget's visible dimension
    */
   @Override
   public NativeDimension getVisibleDimension()
   {
      return new NativeDimension(config().innerChars, visibleRows());
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvt.MOUSE_HOVERABLE
      };
   }

   /**
    * Directs the widget to scroll to the specified X and Y positions.
    * It is the widget's responsibility to update its visuals accordingly.
    * Note that the position values are in the same units as the returned 
    * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
    * and {@link #getScrollStep()}. 
    * 
    * @param    newXPosition
    *           New X scroll position.
    * @param    newYPosition
    *           New Y scroll position.
    *           
    * @return   <code>true</code> to indicate that the scroll operation actually 
    *           invalidated this list. 
    */
   @Override
   protected boolean scrollWorker(Integer newXPosition, Integer newYPosition)
   {
      boolean dirty = false;
      
      if (newXPosition != null && newXPosition != colLeft)
      {
         colLeft = newXPosition;
         dirty = true;
      }
      
      return super.scrollWorker(newXPosition, newYPosition) || dirty;
   }
   
   /**
    * Checks if we need to post VALUE-CHANGED event.  GUI specific part.
    *
    * @return  Always <code>TRUE</code> for GUI mode.
    */
   @Override
   protected boolean needValueChanged()
   {
      return true;
   }
   
   /**
    * Calculate the scroll Y position based on the current state. 
    * 
    * @return   the new scroll Y position
    */
   @Override
   protected int calculateScrollYPosition()
   {
      // GUI SELECTION-LIST doesn't scroll up when current is at the last visible row,
      // but when current is at the last visible row - 1.
      // Temporary decrease the number of visible rows to simulate this behavior.
      
      int visRows = visibleRows();
      setVisibleRows(visRows - 1);
      
      try 
      {
         int result = super.calculateScrollYPosition();
         return result;
      }
      finally
      {
         setVisibleRows(visRows);
      }
   }
   
   /**
    * Draws the next text item in the list inside visible part of the drop-down.
    * 
    * @param    number
    *           The number of the item to draw.
    * @param    x
    *           The X coordinate of the enclosing rectangle to draw item.
    * @param    y
    *           The Y coordinate of the enclosing rectangle to draw item.
    * @param    rw
    *           The pixel width of the enclosing rectangle to draw item.
    * @param    rh
    *           The pixel height of the enclosing rectangle to draw item.
    * @param    fontSize
    *           The vertical size of font to draw item text.
    * @param    enabled
    *           Flag indicating the enabled state of the list.
    * @param    focused
    *           Flag indicating the focused state of the list.
    */
   private void drawItem(int number, int x, int y, int rw, int rh,
                         int fontSize, boolean enabled, boolean focused)
   {
      // selected item needs special background and foreground
      boolean isItemSelected = selected(number); 
      if (isItemSelected)
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, rw, rh);
      }

      // draw item text with respective color
      if (enabled && isItemSelected)
      {
         gd.setColor(selectionFore);
      }
      else if (enabled)
      {
         gd.setColor(gc.fgColor);
      }
      else
      {
         gd.setColor(disabledFore);
      }

      // this X shift is the horizontal scrolling key point
      int xShift = colLeft * getCharWidth(); 
      gd.drawString(getItemText(number),
                    x + SelectionListGuiImpl.TEXT_SHIFT - xShift, y + (rh + fontSize)/2 + 1);

      // draw dotted rectangle around current row
      if (number == currentRow() && enabled && focused)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         // left edges of the dotted rect become invisible in srolling right
         gd.drawRect(x + 1 - xShift, y, nativeWidth - 1 + xShift, rh - 1);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Gets the maximum possible width of the current selection-list in character units.
    *
    * @return  Max width of the item in selection-list.
    */
   private int getMaxWidth()
   {
      double wResult = 0.0;
      int fnt = resolveCurrentFontNum();
      
      for (int i = 0; i < config().items.length; i++)
      {
         String c = config().items[i].getValue().toStringMessage();

         // calculate effective width of the item in characters         
         int txtWidthPix = getTextWidth(c, fnt);
         double txtWidthChar = screen().coordinates().widthFromNative(txtWidthPix);

         // compare with the current max value
         if (txtWidthChar > wResult)
         {
            wResult = txtWidthChar;
         }
      }

      // to the biggest double without fractions
      return (int) Math.ceil(wResult);
   }
   
   /**
    * Gets the font number currently using with this selection-list.
    *
    * @return  The font number in using.
    */
   private int resolveCurrentFontNum()
   {
      // get GUI driver for current combo-box window
      if (gd == null)
      {
         gd = (GuiDriver) OutputManager.getDriver();
      }
      // get font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
      }
      
      // get font to use in selection-list
      return gf.resolveFontNum();
   }
}