/*
** Module   : SubMenuGuiImpl.java
** Abstract : SubMenuGuiImpl client implementation
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150525 Created initial version.
** 002 VIG 20150906 Changed mnemonic drawing, added accelerators support and naming using
**                  keyboard layout, fixed drawing bugs.
** 003 CA  20150918 Added mouse actions.
** 004 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 CA  20151026 Enclose driver API calls in select/releaseWindow().
**     VIG 20151027 Added "pressed" state transferring to neighbor MENUBAR element.
**                  Added mouseEntered=false to hideBody(..) and fixed javadoc.
** 007 VIG 20151226 Added setting of keyNavigation mode flag, added setInitialFocus method.
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 VIG 20160222 Label is renamed to MnemonicInfo.
** 011 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 012 VIG 20160410 Added hardcoded font setting. Tuned sizes, and arrow position. 
*                   Fixed height/width/bodyDimension methods and sizes for menubar and popup.
** 013 SVL 20160719 Updated translate() signature.
** 014 VIG 20160628 window() call is changed to topLevelWindow() in mouse* methods.
**                  Changed drawTitleRectangle: draw pressed state independently of mouseEntered.
**                  "pressed" state transferring is refactored, FOCUS_GAINED is moved from
**                  mouseExit() to mouseEntered(), added menubar flags setting, fixed hideBodies().
** 015 HC  20170201 Prevented unexpected exceptions during SUB-MENU rendering and input
**                  processing.
**     HC  20170223 Fixed sub menu body generating paint events for the wrong top-level window
**                  during hideBody(). This resolves draw artifacts in the owner window when sub-
**                  menu body closed.
** 016 OM  20170320 Added support for Themes.
** 017 CA  20170929 Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
** 018 EVL 20171211 Adding fix for pop-up menu width.  Fix for menu and submenu hiding issues.
**                  Adding fix for submenu items hide/show on mouse leave events.  Changed
**                  approach to select menu item criteria.
** 019 EVL 20180103 Fixes for remaining sub-menu drawing issues.
** 020 CA  20180208 Notify the driver when it switches control to server-side (and vice-versa).
** 021 SVL 20180304 Added createDynamicSubmenu.
** 022 HC  20180309 Fixed multiple menu-originated abends.
** 023 HC  20180419 Fixed an NPE condition. Fixed popup menu highlight state being remembered between
**                  menu invocations. Fixed menu-bar won't display after menu dismissed due to the submenu
**                  body overlay window losing focus. Fixed key input processing in popup menus. Fixed menu
**                  input events sent to wrong widgets in web GUI.
** 024 HC  20180509 Improved MENU focus management. Fixed a race condition during opening and
**                  closing of submenu bodies.
** 025 HC  20180528 Fixed focus-related menu abend, MENU widget in some cases didn't have its
**                  Container.current set, which caused an NPE in focus management logic. Also
**                  fixed focus-related draw artifacts.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.FocusEvent;
import com.goldencode.p2j.ui.client.event.listener.FocusListener;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

import java.awt.event.*;
import java.util.*;
import java.util.concurrent.*;

/**
 * <code>SubMenu</code> client abstract implementation class. 
 */
@SuppressWarnings("unchecked")
public class SubMenuGuiImpl
extends SubMenu<GuiOutputManager>
implements FocusListener
{
   /** This menu-item top/bottom padding in a MENUBAR, in pixels. */
   static final int HEIGHT_PAD = 3;
   
   /** This menu-item minimum left/righ padding in a MENUBAR, in pixels. */
   static final int WIDTH_PAD = 8;

   /** Default delay time to show/hide submenu body in milliseconds if not specified in config. */
   private static final int HIDE_SHOW_MS_DELAY = 500;

   /** Default delay time to show/hide submenu body for menubar items in milliseconds. */
   private static final int HIDE_SHOW_MENUBAR_MS_DELAY = 10;

   /** Executor service for show/hide workers. */
   private static ScheduledExecutorService srvPopUp;

   /** Scheduled task for show/hide workers. */
   private static ScheduledFuture<?> sf = null;

   /** Configurable menu delay time in milliseconds. */
   private static int menuDelayMS;

   /** Screen driver */
   @SuppressWarnings("rawtypes")
   private GuiDriver gd;

   /** Font resolver */
   private GuiFontResolver gf;
   
   /** The sub-menu body container. */
   private SubMenuBody body;
   
   /** Cached width of this item. If 0 or negative, value is invalidated and need re-computed. */
   private double cachedWidth = -1;
   
   /** Cached height of this item. If 0 or negative, value is invalidated and need re-computed. */
   private double cachedHeight = -1;

   /**
    * Request delay value onw time for every menu for current JVM.
    */
   static
   {
      menuDelayMS = ThinClient.getInstance().getServer().getGuiSubMenuDelay(HIDE_SHOW_MS_DELAY);

      if (menuDelayMS > 0)
      {
         srvPopUp = Executors.newScheduledThreadPool(1);
      }
   }
   
   /**
    * Recursively hides all the sub-menu bodies starting at the passed in container reference
    * down.
    *
    * This method doesn't work for popup menus, use {@link #hideBody(boolean)} instead.
    *
    * @param   container
    *          The reference container.
    * @param   exclude
    *          The SubMenu item to exclude.
    */
   static void hideBodies(Container<GuiOutputManager> container, SubMenuGuiImpl exclude)
   {
      for (Widget<GuiOutputManager> w : container.widgets())
      {
         if (w instanceof SubMenuGuiImpl)
         {
            if (exclude != null && exclude.equals(w))
            {
               continue;
            }

            // call with keepMenuActive = false, the bracketing already in place
            hideBodies((Container<GuiOutputManager>) w, null);
            ((SubMenuGuiImpl) w).hideBody(false);
         }
      }
   }

   /**
    * Recursively hides all the sub-menu bodies starting at this sub-menu. The bodies are hidden with the
    * sub-menu close/open delay.
    *
    * @param   container
    *          The reference container.
    * @param   exclude
    *          The SubMenu item to exclude.
    */
   static void hideBodiesWithDelay(Container<GuiOutputManager> container, SubMenuGuiImpl exclude)
   {
      if (menuDelayMS > 0)
      {
         sf = srvPopUp.schedule(() ->
         {
            if (sf != null && !sf.isCancelled())
            {
               ThinClient.getInstance().invokeLaterOS(() -> 
               {
                  hideBodies(container, exclude);
               });
            }
         },
         menuDelayMS, TimeUnit.MILLISECONDS);
      }
      else
      {
         hideBodies(container, exclude);
      }
   }

   /**
    * Clears highlighted flag for all child menu elements.
    *
    * @param   container
    *          The reference container.
    */
   static void clearHighlights(Container<GuiOutputManager> container)
   {
      if (container == null)
      {
         return;
      }

      for (Widget w : container.widgets())
      {
         w.setHighlighted(false);
      }

      container.repaint();
   }

   /**
    * Create a dynamic submenu with the specified title.
    *
    * @param  title
    *         The title of the submenu.
    *
    * @return a submenu.
    */
   static SubMenuGuiImpl createDynamicSubmenu(final String title)
   {
      final SubMenuGuiImpl submenu = new SubMenuGuiImpl(null, null);
      final SubMenuConfig config = new SubMenuConfig();
      submenu.initialize(WidgetId.nextID(), config);
      ConfigManager.getInstance().addWidgetConfig(submenu.config());
      submenu.config().title = title;
      submenu.afterConfigUpdate(config);
      submenu.setEnabled(true);
      return submenu;
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuGuiImpl(WidgetId id, SubMenuConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Calculates max menu/sub-menu item accelerator width.
    * 
    * @param    menu
    *           Parent container: MENU or SUB-MENU.
    * @param    gf
    *           Font resolver.
    * @param    gd
    *           Gui driver.
    *            
    * @return   max accelerator width.
    */
   public static int getLongestItemAccelWidth(Container<?> menu, 
                                              GuiFontResolver gf, 
                                              GuiDriver<?, ?> gd)
   {
      int width = 0;
      if (menu != null)
      {
         Window<?> wnd = WindowManager.resolveWindow((AbstractWidget<?>) menu);
         for(Widget<?> w : menu.widgets())
         {
            // ignore SKIP or RULE or SUB-MENU
            if (!(w instanceof MenuItem) || !((MenuItem<?>)w).isNormal())
            {
               continue;
            }
         
            String accel = ((MenuItem<?>)w).config().accelerator;
            int itemWidth = 0;
            if (accel != null)
            {
               accel = Keyboard.keyName(accel);
               itemWidth = FontManager.getTextWidth(accel, wnd, gf.resolveFontNum(), gd);
            }
         
            if (width < itemWidth)
            {
               width = itemWidth; 
            }
         }
      }
      
      return width;
   }
   
   /**
    * Calculates max menu/sub-menu item title (config's) width.
    * 
    * @param    menu
    *           Parent container: MENU or SUB-MENU.
    * @param    exclude
    *           Current menu item, to exclude from processing.
    * @param    gf
    *           Font resolver.
    * @param    gd
    *           Gui driver. 
    * 
    * @return   max title width in pixels.
    */
   public static int getLongestItemTitleWidth(Container<?> menu, 
                                              MenuElement<?> exclude,
                                              GuiFontResolver gf,
                                              GuiDriver<?, ?> gd)
   {
      if (exclude != null && !exclude.getParentMenuContainer().equals(menu))
      {
         throw new IllegalArgumentException("Exclude item must be a child of menu.");
      }
      
      int width = 0;
      Window<?> wnd = WindowManager.resolveWindow((AbstractWidget<?>) menu);
      for(Widget<?> w : menu.widgets())
      {
         // prevent possible infinite loop during width calculation.
         if (w.equals(exclude))
         {
            continue;
         }
         
         // ignore SKIP or RULE 
         if (w instanceof MenuItem && !((MenuItem<?>)w).isNormal())
         {
            continue;
         }

         String title = w instanceof MenuItem ? 
                        ((MenuItem<?>) w).getPreparedTitle() : 
                        ((SubMenu<?>) w).getPreparedTitle();
          
         int itemWidth = FontManager.getTextWidth(title, wnd, gf.resolveFontNum(), gd);
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      
      return width;
   }
   
   /**
    * Calculates a max height of specified Menu or SubMenu including borders.
    * Analyzes nested dropped-down sub-menus.
    * 
    * @param   menu
    *          Menu or SubMenu.
    *          
    * @return  a height of menu.
    */
   @SuppressWarnings("rawtypes")
   public static double getMaxSubMenuHeight(AbstractContainer<?> menu)
   {
      Theme theme = ThemeManager.getCurrentTheme();
      Window<?> wnd = WindowManager.resolveWindow(menu);
      CoordinatesConversion c = wnd.screen().coordinates();
      double menuTitleHeight = titleHeightWithBorders(menu);
      
      Widget<?>[] widgets = menu.widgets();
      NativeInsets insets = theme.getMenuInsets();
      double height = c.heightFromNative(insets.top + insets.bottom);
      
      // in case of menubar, the height is equals to title-height, otherwise we sum up height of 
      // all child items
      if (menu instanceof Menu && !((Menu) menu).config().popupOnly)
      {
         height = menuTitleHeight;
         // the shadow height is 2px lower and 1px higher => 2+1=3
      }
      else 
      {
         for (Widget<?> widget : widgets)
         {
            if (widget instanceof MenuItem)
            {
               height += widget.height();
            }
            else if (widget instanceof SubMenu)
            {
               SubMenu subMenu = (SubMenu) widget;
               height += Math.max(titleHeight(subMenu), 
                                  c.heightFromNative(theme.getMenuItemHeight(subMenu)));
            }
         }
      }
      
      return height;
   }
   
   /**
    * Calculates max menu/sub-menu item width with borders
    * 
    * @param    menu
    *           MENU or SUB-MENU
    * @param    includeNested
    *           <code>true</code> means, that we should process nested sub-menus, 
    *           <code>false</code> otherwise.
    *
    * @return   max width.
    */
   public static double getMaxSubMenuItemWidth(Container<?> menu, boolean includeNested)
   {
      Window<?> wnd = WindowManager.resolveWindow((AbstractWidget<?>) menu);
      CoordinatesConversion c = wnd.screen().coordinates();
      
      double width = 0;
      for (Widget<?> w : menu.widgets())
      {
         // ignore SKIP or RULE
         if (w instanceof MenuItem && !((MenuItem<?>)w).isNormal())
         {
            continue;
         }
         
         AbstractWidget<?> aw = (AbstractWidget<?>) w;
         double itemWidth;
         if (aw instanceof SubMenu)
         {
            if (includeNested)
            {
               itemWidth = aw.width();
            }
            else
            {
               // increase body width itself along with arrow position
               itemWidth = ((SubMenu<?>) aw).titleWidth();
            }
         }
         else
         {
            itemWidth = aw.width();
         }
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      
      return width;
   }
   
   /**
    * Gets sub-menu title height
    * 
    * @param    menu
    *           The parent MENU or SUB-MENU widget.
    * 
    * @return   height of sub-menu title
    */
   @SuppressWarnings("rawtypes")
   public static double titleHeight(AbstractContainer<?> menu)
   {
      Window<?> wnd = WindowManager.resolveWindow(menu);
      GuiDriver gd = (GuiDriver) OutputManager.getDriver();
      GuiFontResolver gf = new GuiFontResolver(wnd, gd, menu.config(), false);
      int h = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
      return wnd.screen().coordinates().heightFromNative(h);
   }
   
   /**
    * Gets sub-menu title height, considering border padding. 
    * 
    * @param    menu
    *           The parent MENU or SUB-MENU widget.
    * 
    * @return   height of sub-menu title
    */
   public static double titleHeightWithBorders(AbstractContainer<?> menu)
   {
      Window<?> wnd = WindowManager.resolveWindow(menu);
      CoordinatesConversion c = wnd.screen().coordinates();
      int textHeight  = c.heightToNative(titleHeight(menu));
      Theme theme = ThemeManager.getCurrentTheme();
      
      if (menu instanceof MenuGuiImpl || Menu.isMenubarElement(menu))
      {
         textHeight = Math.max(textHeight + 2 * HEIGHT_PAD,
                               theme.getMenuBarElementMinSize(false));
      }
      else if (menu instanceof MenuElement) 
      {
         textHeight = Math.max(textHeight,
                               theme.getMenuItemHeight((MenuElement) menu));
      }
      
      return c.heightFromNative(textHeight);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   @SuppressWarnings("rawtypes")
   public void initialize(WidgetId id, SubMenuConfig cfg)
   {
      super.initialize(id, cfg);
      
      config().font = FontManager.MENU_FONT;
      
      gd = (GuiDriver) OutputManager.getDriver();
      
      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent != null ? parent.window() : WindowManager.resolveWindow(this);
      
      gf = new GuiFontResolver(window, gd, config(), false);
      
      addFocusListener(this);
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_DRAGGED
      };
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<GuiOutputManager> defaultFocus()
   {
      return this;
   }
   
   /**
    * Lay out container. Set children locations.
    */
   @Override
   public void doLayout()
   {
      TitledWindow<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      NativeInsets menuInsets = ThemeManager.getCurrentTheme().getMenuInsets();
      int x = menuInsets.left;
      int y = menuInsets.top;
      
      for (Widget<GuiOutputManager> mi : widgets())
      {
         // we redo the layout, invalidate the existing sizes
         if (mi instanceof MenuItemGuiImpl)
         {
            ((MenuItemGuiImpl) mi).invalidateCachedSize();
         }
         else if (mi instanceof SubMenuGuiImpl)
         {
            ((SubMenuGuiImpl) mi).invalidateCachedSize();
         }
         
         int itemHeight = c.heightToNative(mi.height());
         
         mi.setPhysicalLocation(x, y);
         mi.setLocation(c.widthFromNative(x), c.heightFromNative(y));
         y += itemHeight;
         
         if (mi instanceof SubMenu)
         {
            ((SubMenu<?>) mi).doLayout();
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      // this implementation only draws the menu label, for the body drawing logic
      // see SubMenuBody class
      
      TopLevelWindow<GuiOutputManager> topWindow = topLevelWindow();
      final CoordinatesConversion c = topWindow.screen().coordinates();
      
      final Window<?> wnd = WindowManager.resolveWindow(this);      

      int titleWidth = c.widthToNative(titleWidthWithBorders());
      int titleHeight = c.heightToNative(titleHeightWithBorders(this));
      final NativeDimension tDim = new NativeDimension(titleWidth, titleHeight);
      
      NativePoint p = physicalLocation();
      NativeRectangle clip = clipRectangle(p, tDim);

      ScreenBitmap bitmap = topWindow.screen().getBitmapCopy(topWindow.getId().asInt());
      NativeRectangle nr = bitmap.getOuterRectangle();
      NativePoint screenLoc = screenPhysicalLocation();

      // the clipping rectangle is always with a (0,0) origin: thus we need to translate before
      // intersection
      if (bitmap.getOuterRectangle() != null && nr.intersects(clip.translate(screenLoc)))
      {
         // the resulting rectangle will have the same relative origin as your "clip" rectangle,
         // so you can intersect safely
         nr = nr.translate(-screenLoc.x, -screenLoc.y); 
         clip = clip.intersection(nr);
      }
      
      // draw title
      gd.draw(getId(), p, clip, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
         
         if (Menu.isMenubarElement(this))
         {
            theme.drawMenuBarElement(this, gd, gf, 0, 0, tDim.width, tDim.height, highlighted());
         }
         else 
         {
            boolean selected = highlighted();
            int fh = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
            theme.drawMenuItemBackground(this, gd, 0, 0, tDim.width, tDim.height, selected);
            theme.drawMenuItemContent(this, gd, gf, 0, 0, tDim.width, tDim.height, fh, selected);
         }
      });
   }
   
   /**
    * Invalidate the cached size of this item. This way, the next time the height/width is
    * requested it is automatically re calculated using new environment.   
    */
   public void invalidateCachedSize()
   {
      if (cachedHeight > 0)
      {
         cachedHeight = -cachedHeight;
      }
      
      if (cachedWidth > 0)
      {
         cachedWidth = -cachedWidth;
      }
   }
   
   /**
    * Get widget width with borders.
    * 
    * @return  widget width.  
    */
   @Override
   public double width()
   {
      if (cachedWidth > 0)
      {
         return cachedWidth;
      }
      
      // cache the computed value before returning for fast access next time 
      cachedWidth = titleWidthWithBorders();
      return cachedWidth;
   }

   /**
    * Get sub-menu height.
    * 
    * @return  sub-menu height.
    */
   @Override
   public double height()
   {
      if (cachedHeight > 0)
      {
         return cachedHeight;
      }
   
      // cache the computed value before returning for fast access next time 
      cachedHeight = titleHeightWithBorders(this);
      return cachedHeight;
   }

   /**
    * Hides sub-menu body.

    * @param   recursiveUp
    *          <code>true</code> means hide body for all parent sub-menus.
    */
   @Override
   public void hideBody(boolean recursiveUp)
   {
      if (body == null)
      {
         return;
      }

      // force the driver to be in control, as this affects UI
      gd.setControl(true);
      
      int bodyId = UiUtils.getWidgetIdAsInt(body);
      
      TopLevelWindow<GuiOutputManager> window = body.topLevelWindow();
      detachFromPopupWindow();

      window.close();
      body = null;
      
      // reset selection for sub-menu body (just for it: don't call setFocus)
      setFocusInt(null);
      
      config.displayed = false;
      
      for (Widget<?> w : widgets())
      {
         if (w instanceof SubMenuGuiImpl)
         {
            SubMenuGuiImpl childSm = (SubMenuGuiImpl) w;

            childSm.config.displayed = false;
            childSm.setHighlighted(false);
         }
      }
      
      // discard paint events generated by super class, those events would be generated
      // for the body's owner window and not the top-level window of the menu body resulting
      // in unexpected paint behavior
      ThinClient.getInstance().eventBracket(true, () -> super.hideBody(recursiveUp));
      
      int winID = window.getId().asInt();
      if (gd.hasWindowEmulator(winID))
      {
         gd.selectWindow(winID);
         gd.deregisterWidget(bodyId);
         gd.releaseWindow();
      }

      repaint();
   }
   
   /**
    * Gets sub-menu title width (with '-&gt;' if sub-menu is in other sub-menu or in popup menu)
    * Borders are not included.
    * 
    * @return   width of sub-menu title
    */
   @Override
   public double titleWidth()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      Window<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion coords = wnd.screen().coordinates();
      
      String title = getPreparedTitle();
      int w = getTextWidth(title, gf.resolveFontNum());
      
      if (!Menu.isMenubarElement(this))
      {
         // an arrow
         w += theme.getMenuSmArrowWidth(); // 4 + 5;
         
         // toggle-box space
         w += theme.getMenuCheckmarkWidth();
         
         // mandatory additional spacing
         w += theme.getMenuLeftPad() + theme.getMenuRightPad();
         
         // menu-items accelerators
         int accelWidth = SubMenuGuiImpl.getLongestItemAccelWidth(parent(), gf, gd); 
         w += accelWidth == 0 ? 0 : accelWidth;
      }
      
      return coords.widthFromNative(w);
   }
   
   /**
    * Gets sub-menu title width (with '-&gt;' if sub-menu is in other sub-menu or in popup menu)
    * If SUBMENU is not MENUBAR element, then calculate longest neighbor. 
    * 
    * @return   width of sub-menu title
    */
   public double titleWidthWithBorders()
   {
      // in the case of menubar 
      if (Menu.isMenubarElement(this))
      {
         CoordinatesConversion c = screen().coordinates();
         int textWidth = c.widthToNative(titleWidth()) + 2 * WIDTH_PAD - 1;
         textWidth = Math.max(textWidth,
                              ThemeManager.getCurrentTheme().getMenuBarElementMinSize(true));
         return c.widthFromNative(textWidth);
      }
      
      return SubMenuGuiImpl.getMaxSubMenuItemWidth(parent(), false);
   }
   
   /**
    * Returns <code>true</code> only if this widget is navigable using mouse.
    *
    * @return  <code>true</code> if widget is navigable using mouse.
    */
   @Override
   public boolean navigable()
   {
      return focusTraversable();
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int actionCode = event.actionCode();

      // the ESC key processing for menu when it focused
      if (actionCode == Keyboard.KA_END_ERROR)
      {
         if (isBodyDisplayed())
         {
            hideBody(false);
         }
         else
         {
            SubMenu sm = getParentSubMenu();
            if (sm != null)
            {
               sm.hideBody(false);
            }
            else
            {
               Menu m = getParentMenu();
               if (m != null)
               {
                  m.leaveMenu();
               }
            }
         }

         event.consume();
      }
      else if (event.keyCode() == Key.VK_ALT)
      {
         Menu m = getParentMenu();
         if (m != null)
         {
            m.leaveMenu();
         }

         event.consume();
      }
      
      super.processKeyEvent(event);
   } 
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }

      if (Menu.isMenubarElement(this))
      {
         if (hasFocus())
         {
            hideBody(true);
            getParentMenu().leaveMenu();
         }
      }
   }
   
   /**
    * Notification of a mouse entered event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      // cancel possible previously scheduled tasks
      stopScheduledShowHide();

      clearHighlights(parent());
      SubMenuGuiImpl psm = (SubMenuGuiImpl) getParentSubMenu();
      if (psm != null)
      {
         psm.setHighlighted(true);
         psm.repaint();
      }

      if (!isEnabled())
      {
         if (Menu.isMenubarElement(this))
         {
            SubMenuGuiImpl.hideBodies(parent(), null);
         }
         else
         {
            hideBodiesWithDelay(parent(), null);
         }
         return;
      }

      setHighlighted(true);

      if (Menu.hasMenuFocus(this))
      {
         openSubmenuBody();
      }
      repaint();
   }

   /**
    * Notification of a mouse exited event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      stopScheduledShowHide();

      boolean isMenuBarElement = Menu.isMenubarElement(this);
      boolean highlight = false;

      // keep the highlight in case mouse is exiting towards the submenu body
      if (!isMenuBarElement && isBodyDisplayed())
      {
         final NativePoint[] wndLocation = {null};
         TopLevelWindow wnd = body.topLevelWindow();
         gd.withSelectedWindow(wnd.getId().asInt(), () ->
         {
            wndLocation[0] = gd.getWindowLocation();
         });
         NativePoint loc = wndLocation[0];
         if (loc != null)
         {
            NativeRectangle rec = new NativeRectangle(loc, wnd.physicalDimension());
            highlight = rec.contains(new NativePoint(e.getXOnScreen(), e.getYOnScreen()));
         }
      }

      setHighlighted(highlight);
      repaint();
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }

      // Force focus here, we always want the clicked submenu item to receive focus,
      // although focus is also set to this in ThinClient's event logic, there are some
      // (legacy) corner cases where the focus is not switched. Requesting focus when it is already
      // owned by this is NOP anyway.
      requestFocus(true);

      if (!isBodyDisplayed())
      {
         stopScheduledShowHide();
         ThinClient.getInstance().invokeLaterOS(this::showBody);
      }
   }

   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      if (body == null || !isBodyDisplayed())
      {
         return null;
      }
      
      // if the body is displayed, look into each of the items.  for each item, use its "title"
      // dimension - the width() and height() return the smallest rectangle which includes the
      // entire displayed sub-menu tree, thus it can't be used
      
      CoordinatesConversion cc = screen().coordinates();
      NativePoint bodyLoc = cc.pointToNative(bodyLocation());
      NativeDimension bodyDim = cc.dimensionToNative(bodyDimension());
      NativeRectangle bodyBounds = new NativeRectangle(bodyLoc, bodyDim);
      
      return MenuGuiImpl.findMenuMouseSource(this, p, bodyBounds);
   }
   
   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   @Override
   public Point childrenLocation(Widget<GuiOutputManager> child)
   {
      return Coordinate.scale(((MenuElement<?>) child).getParentMenuContainer().screenLocation());
   }
   
   /**
    * This method is invoked when focus is gained.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      SubMenuGuiImpl.clearHighlights(getParentMenuContainer());
      setHighlighted(true);
      parent().repaint();

      if (!isBodyDisplayed())
      {
         if (Menu.isMenubarElement(this))
         {
            // post the MENU-DROP event to hook up the legacy logic, the body will be opened
            // in the handler of the event
            menuDropEvents();
         }
      }
   }
   
   /**
    * This method is invoked when focus is lost. If current focus is on in subtree 
    * of this sub-menu, than hide body.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onFocusLost(FocusEvent event)
   {
      stopScheduledShowHide();
   }

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);
      
      // subitems are only made visible in showBody
      for (Widget<GuiOutputManager> w : widgets())
      {
         w.setVisible(false);
      }
   }
   
   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      return super.toString() + " title=" + config.title;
   }

   /**
    * Check if widget is displayed. Changed parent() to getParent()
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      return parent() != null && getParentMenuContainer().isDisplayed();
   }

   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      if (body != null)
      {
         body.destroy();
         body = null;
      }

      super.destroy();
   }

   /**
    * Creates new MnemonicInfo instance
    * 
    * @return   MnemonicInfo or its descendant.
    */
   @Override
   protected MnemonicInfo createLabelInfo()
   {
      MnemonicInfo minfo = new MnemonicInfoGuiImpl();
      minfo.init(this.config);
      return minfo;
   }
   
   /**
    * Raise the MENU-DROP 4GL events.
    */
   @Override
   protected void menuDropEvents()
   {
      KeyInput e = new KeyInput(this, EventType.ACTION_PERFORMED, Keyboard.SE_MENU_DROP);
      EventManager.postEvent(e);
   }

   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component - must be a sub-menu.
    *           
    * @return   corrected parent location.
    */
   @Override
   protected Point prepParentLocation(Widget<?> parent)
   {
      if (!(parent instanceof SubMenu))
      {
         return super.prepParentLocation(parent);
      }
      
      Window<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      NativePoint p = c.pointToNative(parent.location());
      
      if (!Menu.isMenubarElement(this))
      {
         p.x += c.widthToNative(this.titleWidthWithBorders());
      }
      else
      {
         // under the title
         p.y += c.heightToNative(titleHeightWithBorders((AbstractContainer<?>) parent));  
      }
      
      return c.pointFromNative(p); 
   }
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   @Override
   protected Point bodyLocation()
   {
      Window<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion coord = wnd.screen().coordinates();
      NativePoint p = new NativePoint(0, 0);
      
      if (Menu.isMenubarElement(this))
      {
         p.y += coord.heightToNative(titleHeightWithBorders(this)) - 1;  // under the title
         NativePoint screenLoc = screenPhysicalLocation();
         p.translate(screenLoc);
      }
      else
      {
         NativeInsets insets = ThemeManager.getCurrentTheme().getMenuInsets();
         p.x = coord.widthToNative(this.titleWidthWithBorders());
         // the first item aligned with menu title item in parent menu 
         p.y = physicalLocation().y - insets.top;
      }
      
      return coord.pointFromNative(p); 
   }
   
   /**
    * Calculates body dimension
    * 
    * @return  Dimension
    */
   @Override
   protected Dimension bodyDimension()
   {
      NativeInsets insets = ThemeManager.getCurrentTheme().getMenuInsets();
      double height = SubMenuGuiImpl.getMaxSubMenuHeight(this);
      
      Window<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      
      double width = SubMenuGuiImpl.getMaxSubMenuItemWidth(this, false);
      width += c.widthFromNative(insets.left + insets.right);
      
      return new Dimension(width, height);
   }

   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   @Override
   protected List<Widget<GuiOutputManager>> prepareWidgets(List<Widget<GuiOutputManager>> widgets)
   {
      List<Widget<GuiOutputManager>> result = new ArrayList<>();
      Widget<GuiOutputManager> displayedSubmenu = null;
      
      for (Widget<GuiOutputManager> w : widgets)
      {
         if (w instanceof SubMenu && ((SubMenu<?>)w).config().displayed)
         {
            displayedSubmenu = w;
         }
         else
         {
            result.add(w);
         }
      }
      
      if (displayedSubmenu != null)
      {
         result.add(displayedSubmenu);
      }
      
      return Collections.unmodifiableList(result);
   }
   
   
   /**
    * Shows this sub-menu body.
    * <p>
    * In GUI, it also registers it for "any" actions, so that it is notified when a click
    * occurs outside of its boundaries.
    */
   @Override
   protected void showBody() 
   {
      if (isBodyDisplayed() || parent() == null)
      {
         return;
      }

      // force the driver to be in control, as this affects UI
      gd.setControl(true);
      
      super.showBody();
      
      body = new SubMenuBody();
      for (Widget<GuiOutputManager> w : widgets())
      {
         // clean up highlighted state before body items show
         w.setHighlighted(false);
         body.add(w);
      }
      
      OverlayWindow wndBody = new OverlayWindow(body, topLevelWindow());
      wndBody.setDestroyingListener(() ->
      {
         detachFromPopupWindow();
         body = null;
      });

      // disable overlay window activation change
      body.setLocation(bodyLocation());
      body.setSize(bodyDimension());
      
      wndBody.show();

      // this requests focus for submenu, but not highlighting first item by default
      body.requestFocus(true);
      setHighlighted(true);
   }
   
   /**
    * Returns the boolean, denoting is body displayed or not.
    * 
    * @return   <code>true</code> if body is displayed, <code>false</code> - otherwise. 
    */
   protected boolean isBodyDisplayed()
   {
      return body != null && body.isDisplayed();
   }
 
   /**
    * Cancels the scheduled task if one exists.
    */
   private void stopScheduledShowHide()
   {
      if (sf != null)
      {
         sf.cancel(false);
         sf = null;
      }
   }
   
   /**
    * Opens submenu body immediately or delayed.
    */
   private void openSubmenuBody()
   {
      // Enqueue the show on the event queue, a potentially opened
      // submenu body window is closed() on the event queue, too.
      // If showBody() called directly the enqueued close() and resulting
      // OverlayWindow.hide() would activate the owner window, causing 
      // this body to be closed right away.
      if (menuDelayMS > 0)
      {
         sf = srvPopUp.schedule(() ->
         {
            if (sf != null && !sf.isCancelled())
            {
               ThinClient tc = ThinClient.getInstance();
               tc.invokeLaterOS(() ->
               {
                  tc.eventDrawingBracket(this, () ->
                  {
                     requestFocus(true);
                     if (!Menu.isMenubarElement(this))
                     {
                        showBody();
                     }
                  });
               });
            }
         },
         Menu.isMenubarElement(this) ? HIDE_SHOW_MENUBAR_MS_DELAY : menuDelayMS,
         TimeUnit.MILLISECONDS);
      }
      else
      {
         ThinClient.getInstance().invokeLater(this::showBody);
      }
   }
   
   /**
    * Hard reset sub menu internal to initial state.  Use with caution bevause it makes
    * recursive calls to walk through all sub menus in the current item.
    */
   public void resetBody()
   {
      // the body is not always become null when body is closing
      body = null;
      setHighlighted(false);
      for (Widget<GuiOutputManager> w : widgets())
      {
         // need to find out all submenu items with current one as root
         if (w instanceof SubMenuGuiImpl)
         {
            ((SubMenuGuiImpl)w).resetBody();
         }
      }
   }

   /**
    * Detaches this widget from the parent body widget.
    */
   private void detachFromPopupWindow()
   {
      if (body == null)
      {
         return;
      }

      Widget[] items = body.widgets();
      for (Widget<GuiOutputManager> w : items)
      {
         body.detach(w);
         w.setParent(this);
      }
   }

   /**
    * Sub-menu body. The class serves the purpose of a temporary container
    * to contain sub-menu body widgets and bridge the events to the actual
    * sub-menu class.
    */
   private class SubMenuBody
   extends AbstractContainer<GuiOutputManager>
   implements MenuStructure<GuiOutputManager>,
              MouseWidgetsEnumerator
   {
      /** Widget ID. */
      private WidgetId id = WidgetId.nextID();
      
      /**
       * Widget ID.
       * 
       * @return  see above.
       */
      @Override
      public WidgetId getId()
      {
         return id;
      }

      /**
       * Check if widget is visible.
       * 
       * @return  <code>true</code> if widget is visible.
       */
      @Override
      public boolean isVisible()
      {
         return SubMenuGuiImpl.this.isVisible();
      }

       /** 
        * Draw widget.
        */
      @Override
      public void draw()
      {
         Theme theme = ThemeManager.getCurrentTheme();
         
         TopLevelWindow<GuiOutputManager> topWindow = topLevelWindow();
         CoordinatesConversion cc = topWindow.screen().coordinates();
         NativeDimension bodyDim = cc.dimensionToNative(bodyDimension());
         
         theme.drawMenuBackground(gd, 0, 0, bodyDim.width, bodyDim.height, true);
         
         Widget<GuiOutputManager> activeSubmenu = null;
         Container<GuiOutputManager> origParent = null;
         
         // do this for submenu with active body
         for (Widget<GuiOutputManager> c : widgets())
         {
            Container<GuiOutputManager> cparent = c.parent();
            if (cparent != this)
            {
               activeSubmenu = c;
               origParent = cparent;
               activeSubmenu.setParent(this);
               
               // there can be only one active submenu body among the child widgets
               break;
            }
         }
         
         super.draw();
         
         if (activeSubmenu != null)
         {
            activeSubmenu.setParent(origParent);
         }
         
         ThemeManager.debug(() -> {
            gd.setColor(java.awt.Color.CYAN);
            gd.drawLine(0, 0, bodyDim.width - 1, bodyDim.height - 1);
            gd.drawLine(0, bodyDim.height - 1, bodyDim.width - 1, 0);
         });
      }

      /**
       * Process key event.
       *
       * @param   event
       *          Key event to process.
       */
      @Override
      public void processKeyEvent(KeyInput event)
      {
         int actionCode = event.actionCode();
         Widget focus = null;

         if (actionCode == Keyboard.KA_CURSOR_DOWN)
         {
            focus = getFirstFocusableItem();
         }
         else if (actionCode == Keyboard.KA_CURSOR_UP)
         {
            focus = getLastFocusableItem();
         }
         else
         {
            SubMenuGuiImpl.this.processKeyEvent(event);
         }

         if (focus != null)
         {
            focus.requestFocus(true);
         }
      }

      /**
       * Returns a string representation of the object.
       * 
       * @return   A string representation of the object.
       */
      @Override
      public String toString()
      {
         return super.toString() + "[SM " + getParentSubMenu() + "]";
      }
      
      /**
       * Returns the tab item list for this container. If the tab item list is not given, then
       * an empty list is returned.
       * 
       * @return   The tab item list.
       */
      @Override
      public List<Widget<GuiOutputManager>> getTabItemList()
      {
         // for sub-menu get the menu item list
         return widgetsAsList();
      }

      /**
       * Returns the list of widgets that will be considered when registering mouse widgets by GUI driver.
       *
       * @return  list of widgets, may be {@code null}.
       */
      @Override
      public List<Widget> getMouseWidgets()
      {
         return Arrays.asList(widgets());
      }

      /**
       * Returns the parent menu.
       *
       * @return   The parent menu reference.
       */
      @Override
      public Menu<GuiOutputManager> getParentMenu()
      {
         return SubMenuGuiImpl.this.getParentMenu();
      }

      /**
       * Returns the parent menu. The method may return <code>null</code>,
       * in this case the menu element is not located under a sub-menu.
       *
       * @return   The parent sub-menu reference.
       */
      public SubMenu<GuiOutputManager> getParentSubMenu()
      {
         return SubMenuGuiImpl.this;
      }

      /**
       * Get widget which is in focus.
       * <p>
       * The method allows to choose whether the focused widget is searched in the nested
       * containers (<code>recurse == true</code>) or whether only this container is searched
       * (<code>recurse == false</code>).
       *
       * The override allows the submenu widget to become focused even without any of the child widgets
       * in focus. This allows the widget to process key events.
       *
       * @param   recurse
       *          The parameter controls whether the method should recurse to the nested
       *          containers when searching for the focused widget.
       *
       * @return  focused widget.
       */
      @Override
      public Widget<GuiOutputManager> currentFocus(boolean recurse)
      {
         Widget<GuiOutputManager> current = getFocusInt();
         return current == null ? this : current;
      }

      /**
       * Search for the next or previous focusable widget in the container and
       * set focus on it.
       *
       * @param   direction
       *          Widget search direction.
       */
      protected void focusWorker(boolean direction)
      {
         cycleFocus(direction);
      }

      /**
       * Returns the outer instance.
       *
       * @return   See above.
       */
      private SubMenuGuiImpl getSubMenu()
      {
         return SubMenuGuiImpl.this;
      }
   }
}