/*
** Module   : ToolTip.java
** Abstract : GUI tooltip feature implementation.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20141201 Initial version.
** 002 EVL 20141209 Fix for tooltip offset.  Separating tooltip into own thread.
** 003 HC  20150323 Removed business logic from config classes.
** 004 GES 20150425 Eliminated AWT dependencies.
** 005 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 006 EVL 20151012 Fix for tooltip painting approach.
** 007 IAS 20161017 More generic type for the constructor argument.
** 008 OM  20170320 Added support for Themes.
** 009 OM  20170412 Rewritten to use a Timer instead of multiple threads.
** 010 CA  20171206 Do not draw tooltip if the widget has been destroyed.
** 011 OM  20180321 Added multiline support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import java.util.*;

/**
 * The help string showing when mouse become over the GUI widget.  There is no counterpart in
 * character mode interface.
 */
@SuppressWarnings("unchecked")
public class ToolTip
{
   /** The distance from mouse pointer coordinate to draw tooltip. */
   private static final int ORIGIN_SHIFT = 16;
   
   /** The delay between mouse enter widget and draw tooltip, ms. */
   private static final int DELAY_MS = 500;
   
   /** The time the tooltip is shown without any action, ms. */
   private static final int DURATION_MS = 5000;
   
   /** Tooltips enabled by default for GUI. */
   private static boolean sessionTooltips = false;
   
   /** The ToolTip singleton. */
   private static ToolTip instance = null;
   
   /** Screen driver. */
   private GuiDriver gd = null;
   
   /** Clock that times tooltip events (show and hide). */
   private TooltipWorker clock = null;
   
   /**
    * This constructor creates a tooltip object for given screen driver. It is private since the
    * class is a singleton.
    */
   private ToolTip()
   {
      // initialize tooltip internals
      gd = (GuiDriver) OutputManager.getDriver();
   }
   
   /**
    * Obtain the singleton for this class.
    *
    * @return  the {@code ToolTip} singleton.
    */
   public static ToolTip getInstance()
   {
      if (instance == null)
      {
         instance = new ToolTip();
      }
      return instance;
   }
   
   /**
    * Checks if the given tooltip engine should be activated in a current conditions. This depends
    * on the session wide TOOLTIPS boolean attribute and on the current text value for given
    * widget.  If TOOLTIPS attribute is {@code false} or text does not exist, empty or unknown
    * value the tooltip will not be drawn. 
    *
    * @param   tooltipTxt
    *          Tooltip string to check.
    *
    * @return  {@code true} if the tooltip needs to be used, {@code false} otherwise.
    */
   public static boolean isActive(String tooltipTxt)
   {
      // first check the SESSION:TOOLTIPS attribute value
      // second check the tooltip string itself
      // null string or empty string both disable the tooltip engine
      return sessionTooltips && tooltipTxt != null && tooltipTxt.length() > 0;
   }
   
   /**
    * Sets the SESSION:TOOLTIPS attribute for this session.
    *
    * @param    tooltips
    *           The new value for session related TOOLTIPS attribute.
    */
   public static void setSessionTooltips(boolean tooltips)
   {
      // if nothing changed - do nothing
      if (sessionTooltips == tooltips)
      {
         return;
      }
      
      // accept new value
      sessionTooltips = tooltips;
   }
   
   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   {@code true} if tooltips are available for this session {@code false} otherwise.
    */
   public static boolean getSessionTooltips()
   {
      return sessionTooltips;
   }
   
   /**
    * Activates a tooltip. It will not be immediately shown.
    * 
    * @param   widget
    *          The widget's reference for tooltip.
    * @param   fnt
    *          The font to use to draw the tooltip.
    * @param   orig
    *          The origin of the mouse event that fired this activation.
    */
   public void activate(Widget<?> widget, int fnt, NativePoint orig)
   {
      // deactivate timer/hide old tooltip if any visible
      deactivate();
      
      // simple validity check
      if (widget == null || orig.x == -1 || orig.y == -1)
      {
         return;
      }
      
      // widget owning the tooltip
      AbstractWidget owner = (AbstractWidget) widget;
      
      // get the current tooltip text, it can be different from call to call
      String toolTip = UiUtils.getConfig(owner, BaseConfig.class).tooltip;
      
      if (!isActive(toolTip))
      {
         return;
      }
      
      // calculate required rectangle to include the text and mark we have something to be
      // hidden further at the same time
      int width = 0;
      int height = 0;
      
      String lines[] = toolTip.trim().split("\\r?\\n");
      for (String line : lines)
      {
         height += owner.getTextHeightNative(line, fnt);
         int w = owner.getTextWidthNative(line, fnt);
         if (w > width)
         {
            width = w;
         }
      }
      
      NativeDimension size = new NativeDimension(width + 5, height + 3);
      NativePoint position = new NativePoint(orig.x, orig.y + ORIGIN_SHIFT);
      
      clock = new TooltipWorker(owner, toolTip, new NativeRectangle(position, size), fnt);
   }
   
   /**
    * Stop the lifetime of the current tooltip, if any. Runs the tooltip hiding code if the last
    * shown tooltip is still visible.
    */
   public void deactivate()
   {
      if (clock != null)
      {
         clock.cancel();
         clock = null;
      }
   }
   
   /**
    * Worker class for tooltips. It is responsible with scheduling and the execution of the show 
    * and hide events of the tooltip.
    */
   class TooltipWorker
   extends Timer
   {
      /** Flags an aborted state. If {@code true}, the tooltip will not be shown any more. */
      private volatile boolean cancelled = false;
      
      /** {@code true} if the tooltip is visible and need to be erased. */
      private volatile boolean displayed = false;
      
      /** The widget that owns the tooltip. */
      private final AbstractWidget owner;
      
      /* The text to be displayed. */
      private final NativeRectangle rect;
      
      /** The font used to display the tooltip. */
      private final int fnt;
      
      /**
       * Creates a new object and schedule tasks for show and hide events for the tooltip.
       * 
       * @param   owner
       *          The widget that owns the tooltip.
       * @param   toolTip
       *          The text to be displayed.
       * @param   rect
       *          The surface occupied by tooltip, in native points.
       * @param   fnt
       *          The font used to display the tooltip.
       */
      TooltipWorker(AbstractWidget<?> owner, String toolTip, NativeRectangle rect, int fnt)
      {
         super("Tooltip Worker");
         this.owner = owner;
         this.rect = rect;
         this.fnt = fnt;
         
         // schedule painting
         schedule(new TimerTask()
         {
            @Override
            public void run()
            {
               if (cancelled)
               {
                  // the mouse left the widget by the time DELAY_MS expired, so it was cancelled
                  return;
               }
               ThinClient tc = ThinClient.getInstance();
               
               tc.invokeLaterOS(
                  () -> tc.eventDrawingBracket(owner, false, true, () -> {
                     if (cancelled || !owner.isDisplayed())
                     {
                        return;
                     }
                     // set up tooltip font
                     FontDetails fd = FontManager.resolveFont(owner.window(),
                                                              fnt,
                                                              owner.screen().getInstanceDriver());

                     ThemeManager.getCurrentTheme().drawToolTip(
                           gd, rect.left(), rect.top(), rect.width(), rect.height(),
                           fd, toolTip);
                     displayed = true;
                     
                     // schedule clearing
                     schedule(new TimerTask()
                     {
                        @Override
                        public void run()
                        {
                           clear();
                        }
                     }, DELAY_MS + DURATION_MS);
                  }));
            }
         }, DELAY_MS);
      }
      
      /**
       * Clears the area occupied by tooltip by posting a refresh event for respective rectangle.
       * Updates the status as clean to prevent multiple calls.
       */
      private void clear()
      {
         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() -> tc.eventDrawingBracket(owner, () ->
            {
               TopLevelWindow<?> aWindow = WindowManager.resolveTopLevelWindow(owner);
               CoordinatesConversion coords = owner.screen().coordinates();
               aWindow.repaint(coords.rectangleFromNative(rect));
               displayed = false;
               cancelled = true;
            }));
      }
      
      /**
       * Terminates this timer, discarding any currently scheduled tasks. If the tooltip is
       * already shown the area occupied is cleaned up / redrawn. 
       * <p>
       * This method may be called repeatedly; the second and subsequent calls have no effect.
       */
      @Override
      public void cancel()
      {
         if (cancelled)
         {
            return;
         }
         cancelled = true;
         super.cancel();
         
         if (displayed)
         {
            clear();
         }
      }
   }
}
