/*
** Module   : ColorChooserGuiImpl.java
** Abstract : The dialog that allows to pick a color directly or based on its RGB or HSL.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20171016 Created initial version.
** 002 OM  20180321 Added direct input of RGB/HSL component values.
** 003 EVL 20180721 The color chooser dialog must include theme dependent title height while
**                  doing layout.  Every FWD dialog should match this rule.  In a case the dialog
**                  is inherited directly from FrameGuiImpl instead of DialogBoxWindow it must
**                  return true from isSystemDialog() method for WindowLayout to do respective
**                  calculation.
**     OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.colorchooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.filechooser.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.util.*;
import java.util.logging.*;

/**
 * This is a frame viewed as dialog that allows to pick a color directly or based on its RGB or 
 * HSL codes. It mimics the color picker dialog from Windows.
 * <p>
 * Usage:
 *    <ol>
 *       <li>call {@link #createInstance} to obtain an instance of the frame;</li>
 *       <li>make it {@code visible} and {@code enabled} and set {@code focus};</li>
 *       <li>install an event handing routine for events of its components (see 
 *          {@link ThinClient#modalEventLoop}) ;</li>
 *       <li>when dialog is closed retrieve the selected value with {@link #getSelectedColor}
 *          method if the dialog was accepted.</li>
 *    </ol>
 * <p>
 */
public class ColorChooserGuiImpl
extends FrameGuiImpl
implements Container<GuiOutputManager>,
           IsModal,
           ActionListener
{
   /** The native width of the dialog. */
   private final int NATIVE_WIDTH = 447;
   
   /** The native height of the dialog. */
   private final int NATIVE_HEIGHT = 299;
   
   /** Flag that turns to true only if the user closed the dialog by accepting it. */
   private boolean accepted = false;
   
   /** Flag which tracks user input. */
   private boolean inputDone = false;
   
   /** Cached driver. */
   private GuiDriver gd = null;
   
   /** Cached coordinates. */
   private CoordinatesConversion cc = null;
   
   /** A simple label that reads "Basic colors:". */
   private SimpleLabel basicColorsLabel = null;
   
   /** A simple label that reads "Custom colors:". */
   private SimpleLabel customColorsLabel = null;

   /** A simple label that reads "Color|Solid". */
   private SimpleLabel colorSolidLabel = null;
   
   /** Label for red RGB component editor. */
   private SimpleLabel redLabel = null;
   
   /** Label for green RGB component editor. */
   private SimpleLabel greenLabel = null;
   
   /** Label for blue RGB component editor. */
   private SimpleLabel blueLabel = null;
   
   /** Label for hue HSL component editor. */
   private SimpleLabel hueLabel = null;
   
   /** Label for saturation HSL component editor. */
   private SimpleLabel satLabel = null;
   
   /** Label for luminosity HSL component editor. */
   private SimpleLabel lumLabel = null;
   
   /** The button for defining custom colors. */
   private ButtonGuiImpl defCustom = null;
   
   /** The button for adding a custom color. */
   private ButtonGuiImpl addCustom = null;
   
   /** The button for accepting the current selection. */
   private ButtonGuiImpl accept = null;
   
   /** The button for rejecting hte dialog. */
   private ButtonGuiImpl reject = null;
   
   /** Editor for hue HSL component. */
   private LineEditor hueE = null;
   
   /** Editor for saturation HSL component. */
   private LineEditor satE = null;
   
   /** Editor for luminosity HSL component. */
   private LineEditor lumE = null;
   
   /** Editor for red RGB component. */
   private LineEditor redE = null;
   
   /** Editor for green RGB component. */
   private LineEditor greenE = null;
   
   /** Editor for blue RGB component. */
   private LineEditor blueE = null;
   
   /** 
    * The gamut picker. Allows to select hue and saturation and preserving the current luminosity
    * component.
    */
   private GamutPicker gamutPicker = null;
   
   /**
    * The gamut picker. Allows to select luminosity and preserving the current hue and saturation
    * components.
    */
   private LuminosityPicker luminosityPicker = null;
   
   /** 
    * Basic color buttons. Allows quick access to a standard palette of basic colors. 
    */
   private ColorSampleButton[][] basicColors = new ColorSampleButton[6][8];
   
   /**
    * Custom color buttons. Allows quick access to a user-defined palette of basic colors. 
    */
   private ColorSampleButton[][] customColors = new ColorSampleButton[2][8];
   
   /** Preview of the selected color. */
   private ColorSampleButton solidColor = null;
   
   /** 
    * If not {@code null} this stores the last selected custom color button. This will be the one 
    * to be customized with the currently selected color when {@code addCustom} button is pressed.
    */
   private ColorSampleButton activeCustom = null;
   
   /** The active basic color. Only used for focus management. */
   private ColorSampleButton activeBasic = null;
   
   /** Logger for all related classes. */
   static final Logger LOG = LogHelper.getLogger(ColorChooserGuiImpl.class.getName());
   
   /** The RGB codes of the basic colors. */
   private static final int[][] BASIC_COLOR_VALUES = {
         {0xff8080, 0xffff80, 0x80ff80, 0x00ff80, 0x80ffff, 0x0080ff, 0xff80c0, 0xff80ff},
         {0xff0000, 0xffff00, 0x80ff00, 0x00ff40, 0x00ffff, 0x0080c0, 0x8080c0, 0xff00ff},
         {0x804040, 0xff8040, 0x00ff00, 0x008080, 0x004080, 0x8080ff, 0x800040, 0xff0080},
         {0x800000, 0xff8000, 0x008000, 0x008040, 0x0000ff, 0x0000a0, 0x800080, 0x8000ff},
         {0x400000, 0x804000, 0x004000, 0x004040, 0x000080, 0x000040, 0x400040, 0x400080},
         {0x000000, 0x808000, 0x808040, 0x808080, 0x408080, 0xc0c0c0, 0x400040, 0xffffff},
   };
   
   /**
    * The default RGB codes of the custom colors. In a future revision, the colors will be loaded
    * form a dedicated file.
    */
   private static final int[][] CUSTOM_COLOR_VALUES = {
         {0x0000ff, 0x000080, 0x008000, 0x008080, 0x800000, 0x800080, 0x808000, 0x808080},
         {0xc0c0c0, 0x0000ff, 0x00ff00, 0x00ffff, 0xff0000, 0xff00ff, 0xffff00, 0xffffff},
   };
   
   /**
    * Factory method. Creates and configures a {@code ColorChooserGuiImpl} dialog. 
    * 
    * @param   rgbColor
    *          The color to be selected in the dialog when it is first displayed.
    * @param   windowId
    *          The parent window of the newly created dialog.
    *          
    * @return  A new instance of {@code ColorChooserGuiImpl} configured as requested. 
    */
   public static ColorChooserGuiImpl createInstance(int rgbColor, int windowId)
   {
      WidgetId fid = WidgetId.nextID();
      FrameConfig fCfg = new FrameConfig(fid);
      fCfg.dialog = true;
      fCfg.font = FontManager.STACK_TRACE_WINDOW_FONT;
      fCfg.widthPixels = 447 - 4;
      fCfg.heightPixels = 299 - 4;
      fCfg.windowID = -1; // parent windowId;
      fCfg.scrollable = false;
      fCfg.scrollbarVertical = false;
      fCfg.scroll = 0; 
      fCfg.overlay = false;
      fCfg.title = "Color";
      fCfg.sideLabels = false;
      fCfg.labels = false;
      fCfg.threeD = true;
      
      ColorChooserGuiImpl ret = new ColorChooserGuiImpl(fid, fCfg);
      ret.initialize(fid, fCfg);
      ret.init(rgbColor & 0x00FFFFFF);
      ret.selectColor(rgbColor);
      return ret;
   }
   
   /**
    * Private constructor. To create this kind of object use the {@link #createInstance} factory
    * method.
    * 
    * @param   id
    *          The {@link WidgetId} of the dialog. 
    * @param   cfg
    *          The {@link FrameConfig} of the dialog.
    */
   private ColorChooserGuiImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
      setLayout(new ColorChooserLayout());
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    * 
    * @param   startupRgb
    *          The color preselected at startup. Will be used to select the a matching basic
    *          color button. 
    */
   protected void init(int startupRgb)
   {
      if (gd != null)
      {
         return; // already initialized
      }
      
      // install system close handler.
      // the titlebar can be null when the web client is in embedded mode
      WindowTitleBar tbar = topLevelWindow().getTitleBar();
      if (tbar == null)
      {
         return;
      }
      tbar.setClickHandler(CaptionButtonType.CLOSE, this::dismiss);
      
      gd = (GuiDriver<?, ?>) OutputManager.getDriver();
      cc = screen().coordinates();
      
      basicColorsLabel = createSimpleLabel("&Basic colors:", 212, 13, false);
      addWidget(basicColorsLabel);
      
      customColorsLabel = createSimpleLabel("&Custom colors:", 212, 13, false);
      addWidget(customColorsLabel);
      
      colorSolidLabel = createSimpleLabel("Color|S&olid", 60, 13, false);
      addWidget(colorSolidLabel);
      
      redLabel = createSimpleLabel("&Red:", 40, 20, true);
      addWidget(redLabel);
      
      greenLabel = createSimpleLabel("&Green:", 40, 20, true);
      addWidget(greenLabel);
      
      blueLabel = createSimpleLabel("Bl&ue:", 40, 20, true);
      addWidget(blueLabel);
      
      hueLabel = createSimpleLabel("Hu&e:", 31, 20, true);
      addWidget(hueLabel);
      
      satLabel = createSimpleLabel("&Sat", 31, 20, true);
      addWidget(satLabel);
      
      lumLabel = createSimpleLabel("&Lum", 31, 20, true);
      addWidget(lumLabel);
      
      hueE = createComponentEditor(324, 202);
      addWidget(hueE);
      satE = createComponentEditor(324, 224);
      addWidget(satE);
      lumE = createComponentEditor(324, 247);
      addWidget(lumE);
      
      redE = createComponentEditor(404, 202);
      addWidget(redE);
      greenE = createComponentEditor(404, 224);
      addWidget(greenE);
      blueE = createComponentEditor(404, 247);
      addWidget(blueE);
      
      defCustom = createButton("&Define Custom Colors >>", null, 207, 23, false);
      defCustom.setEnabled(false);
      addWidget(defCustom);
      
      addCustom = createButton("&Add to CustomColors", "SET-CUSTOM-COLOR", 213, 23, false);
      addWidget(addCustom);
      
      accept = createButton("OK", "OK", 66, 23, true);
      config().defaultButtonId = accept.getId().asInt();
      addWidget(accept);
      
      reject = createButton("Cancel", "CANCEL", 66, 23, false);
      config().cancelButtonId = reject.getId().asInt();
      addWidget(reject);
      
      gamutPicker = new GamutPicker();
      gamutPicker.addListener(this);
      addWidget(gamutPicker);
      
      luminosityPicker = new LuminosityPicker();
      luminosityPicker.addListener(this);
      addWidget(luminosityPicker);
      
      ColorSampleButton csb;
      for (int i = 0; i < 6; i++)
      {
         for (int j = 0; j < 8; j++)
         {
            csb = new ColorSampleButton(BASIC_COLOR_VALUES[i][j]);
            csb.setSelectable(true);
            csb.setSettable(false);
            csb.addListener(this);
            basicColors[i][j] = csb;
            addWidget(csb);
            if ((BASIC_COLOR_VALUES[i][j] & 0x00FFFFFF) == startupRgb)
            {
               startupRgb = -1;
               activeBasic = csb;
            }
         }
      }
      if (activeBasic == null)
      {
         activeBasic = basicColors[0][0];
      }
      
      for (int i = 0; i < 2; i++)
      {
         for (int j = 0; j < 8; j++)
         {
            csb = new ColorSampleButton(CUSTOM_COLOR_VALUES[i][j]);
            csb.setSelectable(true);
            csb.setSettable(true);
            csb.addListener(this);
            customColors[i][j] = csb;
            addWidget(csb);
         }
      }
      activeCustom = customColors[0][0];
      
      solidColor = new ColorSampleButton(-1);
      solidColor.setSelectable(false);
      solidColor.setSettable(true);
      addWidget(solidColor);
      
      screen().getRegistry().addFrame(this);
      setSize(cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH, NATIVE_HEIGHT)));
      
      activeBasic.requestFocus();
      updateTabbingList();
   }
   
   /**
    * Checks whether the user accepted the selection from the dialog.
    *
    * @return  {@code true} if the dialog was accepted, and {@code false} if the dialog was not
    *          yet displayed or it was rejected by user.
    */
   public boolean isAccepted()
   {
      return accepted;
   }
   
   /**
    * Get status of the input processing.
    *
    * @return  current status of the input processing.
    */
   @Override
   public boolean isDismissed()
   {
      return inputDone;
   }
   
   /**
    * Let the modal UI element now that it should finish its modal input processing.
    */
   @Override
   public void dismiss()
   {
      inputDone = true;
      destroy();
   }
   
   /**
    * This method is invoked when action event is generated. Only processes events generated by 
    * member buttons and pickers.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
      String cmd = event.command();
      if (cmd == null)
      {
         return; // unknown command
      }
      
      if (event.source() instanceof ColorSampleButton)
      {
         ColorSampleButton trigg = (ColorSampleButton) event.source();
         if (trigg.isCustom())
         {
            activeCustom = trigg;
            updateTabbingList();
         }
         else if (trigg != solidColor)
         {
            activeBasic = trigg;
            updateTabbingList();
         }
      }
      
      switch (cmd)
      {
         case "SET-CUSTOM-COLOR":
            activeCustom.setRgbColor(solidColor.getRgbColor());
            // TODO: save custom colors somewhere (like registry or local (hidden) file) 
            //       maybe when the dialog is closed
            return;
            
         case "OK":
            accepted = true;
            dismiss();
            return;
         
         case "CANCEL":
            accepted = false;
            dismiss();
            return;
            
         case LineEditor.AC_TEXT_VALIDATED:
            if (event.source() == redE)
            {
               int red = validateComponentVal(redE, true); 
               
               ColorRgb rgb = new ColorRgb(solidColor.getRgbColor());
               rgb.setRed(red);
               selectColor(rgb.getGuiRgb());
            }
            else if (event.source() == greenE)
            {
               int green = validateComponentVal(greenE, true);
               ColorRgb rgb = new ColorRgb(solidColor.getRgbColor());
               rgb.setGreen(green);
               selectColor(rgb.getGuiRgb());
            }
            else if (event.source() == blueE)
            {
               int blue = validateComponentVal(blueE, true);
               ColorRgb rgb = new ColorRgb(solidColor.getRgbColor());
               rgb.setBlue(blue);
               selectColor(rgb.getGuiRgb());
            }
            else if (event.source() == hueE)
            {
               int hue = validateComponentVal(hueE, false);
               ColorHSL hsl = new ColorRgb(solidColor.getRgbColor()).toHSL();
               hsl.setHue(hue);
               selectColor(hsl.toRGB().getGuiRgb());
            }
            else if (event.source() == satE)
            {
               int sat = validateComponentVal(satE, false);
               ColorHSL hsl = new ColorRgb(solidColor.getRgbColor()).toHSL();
               hsl.setSaturation(sat);
               selectColor(hsl.toRGB().getGuiRgb());
            }
            else if (event.source() == lumE)
            {
               int lum = validateComponentVal(lumE, false);
               ColorHSL hsl = new ColorRgb(solidColor.getRgbColor()).toHSL();
               hsl.setLuminosity(lum);
               selectColor(hsl.toRGB().getGuiRgb());
            }
            return;
            
         default:
            if (cmd.startsWith("SET-COLOR "))
            {
               // a ColorSampleButton was pressed: set the color directly
               int newRgbColor = Integer.parseInt(cmd.substring(10), 16);
               selectColor(newRgbColor);
            }
            else if (cmd.startsWith("SET-LUM "))
            {
               // generated by the gamut picker: use current HUE/SAT and new LUM to create a 
               // new color to be displayed and saved
               int lum = Integer.parseInt(cmd.substring(8));
               // obtain current color in HSL color space
               ColorHSL newHSL = new ColorRgb(solidColor.getRgbColor()).toHSL();
               // do the adjustment:
               newHSL.setLuminosity(lum);
               // update to this color 
               selectColor(newHSL.toRGB().getGuiRgb());
            }
            else if (cmd.startsWith("SET-HUE-SAT "))
            {
               // generated by the luminosity picker: use current LUM and new HUE/SAT to create a 
               // new color to be displayed and saved
               StringTokenizer st = new StringTokenizer(cmd);
               st.nextToken(); // drop command
               int hue = Integer.parseInt(st.nextToken());
               int sat = Integer.parseInt(st.nextToken());
               // obtain current color in HSL color space
               ColorHSL newHSL = new ColorRgb(solidColor.getRgbColor()).toHSL();
               // do the adjustments:
               newHSL.setHue(hue);
               newHSL.setSaturation(sat);
               // update to this color 
               selectColor(newHSL.toRGB().getGuiRgb());
            }
      }
   }
   
   /**
    * Validates the a color component. If needed the value is adjusted to it matches the range
    * for respective color space. 
    * 
    * @param   componentEditor
    *          The component editor. 
    * @param   rgb
    *          {@code true} for RGB color space, {@code false} for HSL color space.
    *
    * @return  The closest valid value for the component.
    */
   private int validateComponentVal(LineEditor componentEditor, boolean rgb)
   {
      int val = 0;
      try
      {
         val = Integer.parseInt(componentEditor.getText());
      }
      catch (NumberFormatException e)
      {
         return 0;
      }
      
      if (val < 0)
      {
         val = 0;
      }
      int max = rgb ? 255 : 240; // in RGB color space max is 256, in HSL is only 240
      if (val > max)
      {
         val = max;
      }
      return val;
   }
   
   /**
    * Adding new widget to the container. This method does not require the coordinates to be
    * defined. The location will be selected automatically. Makes sure this frame is also set on
    * target {@code widget}.
    *
    * @param   widget
    *          The new widget to be added to the container.
    */
   @Override
   public void addWidget(Widget<GuiOutputManager> widget)
   {
      super.addWidget(widget);
      widget.config().frameId = getId().asInt();
   }
   
   /**
    * Gets the currently selected color.
    * 
    * @return  the currently selected color.
    */
   public int getSelectedColor()
   {
      return solidColor.getRgbColor();
   }
   
   /**
    * Updates the list that dictates the focusing order using TAB/BACK-TAB keys.
    */
   private void updateTabbingList()
   {
      getContentPane().setTabItemList(Arrays.asList(new Widget[] {
               activeBasic, activeCustom, accept, reject, 
               hueE, satE, lumE, redE, greenE, blueE, addCustom}));
   }
   
   /**
    * Sets the currently selected color. Sets the color on solid preview and update all other
    * widgets to reflect the component of the two color spaces. 
    *  
    * @param   rgbColor
    *          The color in RGB encoding.
    */
   private void selectColor(int rgbColor)
   {
      solidColor.setRgbColor(rgbColor);
      ColorRgb rgb = new ColorRgb(rgbColor);
      ColorHSL hsv = rgb.toHSL();
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, 
                 "SELECTED #" + ColorSampleButton.toHexString(rgbColor) + 
                 " RGB(" + rgb.getRed() + ", " + rgb.getGreen() + ", " + rgb.getBlue() + ")" + 
                 " HSL(" + hsv.getHue() + ", " + hsv.getSaturation() + ", " + 
                           hsv.getLuminosity() + ")");
      }
      
      redE.setValue(new integer(rgb.getRed()));
      greenE.setValue(new integer(rgb.getGreen()));
      blueE.setValue(new integer(rgb.getBlue()));
      hueE.setValue(new integer(hsv.getHue()));
      satE.setValue(new integer(hsv.getSaturation()));
      lumE.setValue(new integer(hsv.getLuminosity()));
      gamutPicker.selectValue(hsv.getHue(), hsv.getSaturation());
      luminosityPicker.selectValue(hsv.getHue(),
                                   hsv.getSaturation(),
                                   hsv.getLuminosity());
      
      repaint();
   }
   
   /**
    * Creates and initializes a customized button widget.
    * 
    * @param   text
    *          The text (label) displayed on the button.  
    * @param   action
    *          The action to trigger when button is pressed.
    * @param   width
    *          The width of the button (native units).
    * @param   height
    *          The height of the button (native units). 
    * @param   def
    *          {@code true} if this is the default button for the frame.
    * 
    * @return  A reference to {@link ButtonGuiImpl} instance.
    */
   private ButtonGuiImpl createButton(String text,
                                      String action,
                                      int width,
                                      int height,
                                      boolean def)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = width;
      bcfg.heightPixels = height;
      bcfg.isDefault = def;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      
      ButtonGuiImpl b = new ButtonGuiImpl(bid, bcfg);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(action);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      return b;
   }
   
   /**
    * Creates an editor that allows to display and direct key in a component of the currently
    * selected color in one of the color spaces.
    * 
    * @param   x
    *          Physical location where the editor will be displayed on the frame.
    * @param   y
    *          Physical location where the editor will be displayed on the frame.
    *          
    * @return  The new editor.
    */
   private LineEditor createComponentEditor(int x, int y)
   {
      WidgetId fid = WidgetId.nextID();
      
      FillInConfig fcfg = new FillInConfig(fid);
      fcfg.clientX = x;
      fcfg.clientY = y;
      fcfg.heightPixels = 20;
      fcfg.widthPixels = 27;
      fcfg.state = ScreenBuffer.UNCHANGED;
      fcfg.autoResize = false;
      try
      {
         fcfg.appFormat = new NumberFormat(">>9");
      }
      catch (DisplayFormatParsingException e)
      {
         // just in case, no expected exceptions for this format
         e.printStackTrace();
      }
      
      LineEditor fill = new LineEditor(fid, fcfg);
      fill.initialize(fid, fcfg);
      fill.setDataType("integer");
      fill.setSize(cc.dimensionFromNative(new NativeDimension(27, 20)));
      fill.setEnabled(true);
      fill.setForceLeft(true);
      fill.addListener(this);
      
      return fill;
   }
   
   /**
    * Creates a customized {@code SimpleLabel} to be used on this dialog. 
    * 
    * @param   text
    *          The text to be displayed.
    * @param   w
    *          The physical width of the label.
    * @param   h
    *          The physical height of the label.
    * @param   rightAlign
    *          Whether th text is right aligned.
    *          
    * @return  A customized {@code SimpleLabel}.
    */
   private SimpleLabel createSimpleLabel(String text, int w, int h, boolean rightAlign)
   {
      SimpleLabel label = new SimpleLabel(text);
      label.setSize(cc.dimensionFromNative(new NativeDimension(w, h)));
      LabelConfig lcfg = (LabelConfig) label.config();
      lcfg.align = rightAlign ? ControlEntity.ALIGN_RIGHT : ControlEntity.ALIGN_LEFT;
      return label;
   }
   
   /**
    * The manager that does the layout of the widgets of this frame. 
    */
   private class ColorChooserLayout
   implements LayoutManager<GuiOutputManager>
   {
      /**
       * Calculate minimum size required for widgets stored in passed container.
       *
       * @param   container
       *          Container to process.
       *
       * @return  A constant dimension for this unresizable dialog.
       */
      @Override
      public Dimension minimumSize(Container<GuiOutputManager> container)
      {
         // constant minim size:
         return cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH - 5, NATIVE_HEIGHT + 5));
      }
      
      /**
       * Place components in specified container.
       *
       * @param   container
       *          Container to process.
       */
      @Override
      public void doLayout(Container<GuiOutputManager> container)
      {
         basicColorsLabel.setPhysicalLocation(6, 8);
         customColorsLabel.setPhysicalLocation(6, 173);
         colorSolidLabel.setPhysicalLocation(228, 246);
         hueLabel.setPhysicalLocation(293, 202);
         satLabel.setPhysicalLocation(293, 224);
         lumLabel.setPhysicalLocation(293, 247);
         redLabel.setPhysicalLocation(364, 202);
         greenLabel.setPhysicalLocation(364, 224);
         blueLabel.setPhysicalLocation(364, 247);
         
         defCustom.setPhysicalLocation(6, 244);
         addCustom.setPhysicalLocation(228, 270);
         accept.setPhysicalLocation(6, 270);
         reject.setPhysicalLocation(78, 270);
         
         hueE.setPhysicalLocation(324, 202);
         satE.setPhysicalLocation(324, 224);
         lumE.setPhysicalLocation(324, 247);
         redE.setPhysicalLocation(404, 202);
         greenE.setPhysicalLocation(404, 224);
         blueE.setPhysicalLocation(404, 247);
         
         for (int i = 0; i < 6; i++)
         {
            for (int j = 0; j < 8; j++)
            {
               basicColors[i][j].setPhysicalLocation(7 + j * 26, 23 + i * 22);
            }
         }
         for (int i = 0; i < 2; i++)
         {
            for (int j = 0; j < 8; j++)
            {
               customColors[i][j].setPhysicalLocation(7 + j * 26, 189 + i * 22);
            }
         }
         solidColor.setPhysicalLocation(228, 202);
         
         gamutPicker.setPhysicalLocation(228, 7);
         luminosityPicker.setPhysicalLocation(420, 7);
      }
   }
}
