/*
** Module   : ColorSampleButton.java
** Abstract : A button in ColorChooserGuiImpl dialog that allows to quickly pick a color.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20171016 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.colorchooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * This is a button in a {@link ColorChooserGuiImpl} dialog that displays a certain color allows 
 * to quickly select it. There are three kind of instances of this class:
 * <ul>
 *    <li><strong>basic colors</strong>. Can be selected. The color is immutable;</li>
 *    <li><strong>custom colors</strong>. Can be selected. The color can be changed if the user
 *       assigns another color for this button;</li>
 *    <li><strong>selection</strong>. Works as preview of the currently selected color. Can be
 *       changed but not selected. Also rendering is a bit different.</li>
 * </ul>
 * 
 * If the button is a basic or custom color picker, it will fire an action that will bear the 
 * RGB code of the color assigned to button. The preview button is not an action button.
 */
public class ColorSampleButton
extends ButtonGuiImpl
{
   /** The color currently assigned to this button, in RGB color space. */
   private int rgbColor = -1;
   
   /** The {@code selectable} property. Action buttons are selectable. The preview is not. */
   private boolean selectable = true;
   
   /** Whether this color can be changed. Only basic color buttons cannot be changed. */
   private boolean settable = false;
   
   /**
    * This constructor creates a button configured with a color. By default this is a basic color
    * button (selectable but not settable).
    *
    * @param   rgbColor
    *          The color assigned to this button, in RGB color space.  
    */
   public ColorSampleButton(int rgbColor)
   {
      super(null/* ignored */, null/* ignored */);
      
      WidgetId bid = WidgetId.nextID();
      String text = "SET-COLOR " + toHexString(rgbColor);
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.isDefault = false;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.name = text; // direct access
      bcfg.help = null;
      
      initialize(bid, bcfg);
      initButtonText(text);
      
      addFocusListener(this);
      setRgbColor(rgbColor);
      setEnabled(true);
   }
   
   /**
    * Utility method for obtaining the 6 digits string of a RGB encoding value. Only the 6 lower
    * significant bytes are used (RGB), the other (A) is ignored.
    * 
    * @param   rgbColor
    *          The color, in RGB encoding.
    * 
    * @return A 6 hex-digit string representing the color in RGB color space. 
    */
   public static String toHexString(int rgbColor)
   {
      String s = "000000" + Integer.toHexString(rgbColor);
      return s.substring(s.length() - 6);
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () -> {
         if (!selectable)
         {
            // this is the 'preview' button, It has a one-pixel line width sunken border.
            gd.setColor(new ColorRgb(160, 160, 160)); // grey 1
            gd.drawLine(0, 0, d.width - 2, 0);
            gd.drawLine(0, 0, 0,           d.height - 2);
            
            gd.setColor(new ColorRgb(255, 255, 255)); // white
            gd.drawLine(0,            d.height - 1, d.width - 1, d.height - 1);
            gd.drawLine(d.width - 1, 0,             d.width - 1, d.height - 1);
            
            gd.setColor(new ColorRgb(rgbColor));
            gd.fillRect(1, 1, d.width - 2, d.height - 2);
         }
         else
         {
            // this is a custom or basic button, It has a two-pixel line width sunken border.
            // N & W sides
            gd.setColor(new ColorRgb(160, 160, 160)); // grey 1
            gd.drawLine(3, 3, d.width - 5, 3);
            gd.drawLine(3, 3, 3, d.height - 5);
            gd.setColor(new ColorRgb(105, 105, 105)); // grey 2
            gd.drawLine(4, 4, d.width - 6, 4);
            gd.drawLine(4, 4, 4, d.height - 6);
            // E & S sides
            gd.setColor(new ColorRgb(255, 255, 255)); // white
            gd.drawLine(d.width - 4, 3, d.width - 4, d.height - 4);
            gd.drawLine(3, d.height - 4, d.width - 4, d.height - 4);
            gd.setColor(new ColorRgb(227, 227, 227)); // grey 3
            gd.drawLine(d.width - 5, 4, d.width - 5, d.height - 5);
            gd.drawLine(4, d.height - 5, d.width - 5, d.height - 5);
            
            gd.setColor(new ColorRgb(rgbColor));
            gd.fillRect(5, 5, d.width - 10, d.height - 10);
            
            if (hasFocus() && selectable)
            {
               gd.setColor(new ColorRgb(0)); // black
               
               gd.drawRect(2, 2, d.width - 5, d.height - 5);
               
               gd.setLineStroke(LineStroke.DOTTED_SMALL);
               gd.drawRect(0, 0, d.width - 1, d.height - 1);
               // restore line style
               gd.setLineStroke(LineStroke.DEFAULT);
            }
         }
      });
   }
   
   /**
    * Get widget height. Action buttons are 23 pixel height. The preview button is 42 pixels in
    * height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return cc.heightFromNative(selectable ? 23 : 42); 
   }
   
   /**
    * Get widget width. Action buttons are 25 pixel width. The preview button is 60 pixels width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return cc.widthFromNative(selectable ? 25 : 60);
   }
   
   /**
    * Sets the color assigned to this button. The color is displayed on the button and contained
    * within the action text when the button is clicked. 
    * 
    * @param   newColor
    *          The new color to be assigned to this button. 
    */
   public void setRgbColor(int newColor)
   {
      if (rgbColor != -1 && !settable)
      {
         throw new RuntimeException("Color already set");
      }
      rgbColor = newColor;
      setCommand("SET-COLOR " + toHexString(rgbColor));
      
      repaint();
   }
   
   /**
    * Sets the {@link #selectable} property. Action buttons (basic and custom colors) are
    * selectable. The preview is not.
    * 
    * @param   selectable
    *          The new value for {@code selectable} property of this button.
    */
   public void setSelectable(boolean selectable)
   {
      this.selectable = selectable;
      config.noFocus = !selectable; 
   }
   
   /**
    * Sets the {@link #settable} property. Only custom colors and preview are settable. The 
    * basic colors buttons are not.
    *
    * @param   settable
    *          The new value for {@code settable} property of this button.
    */
   public void setSettable(boolean settable)
   {
      this.settable = settable;
   }
   
   /**
    * Returns a string representation of the object.
    *
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      return "ColorSampleButton(#" + toHexString(rgbColor) + ')';
   }
   
   /**
    * Obtain the currently selected color.
    * 
    * @return  The currently selected color in RGB color space.
    */
   public int getRgbColor()
   {
      return rgbColor;
   }
   
   /**
    * Checks whether this is a custom color button. A custom color button is both settable and 
    * selectable. 
    * 
    * @return  {@code true} only for custom color button.
    */
   public boolean isCustom() 
   {
      return settable && selectable;
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  Only action buttons are focus traversable; preview is not.
    */
   @Override
   public boolean focusTraversable()
   {
      return selectable;
   }
}
