/*
** Module   : GamutPicker.java
** Abstract : A widget in ColorChooserGuiImpl dialog that allows to quickly select the HUE/SAT of
**            a color. 
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20171016 Created initial version.
** 002 OM  20180503 Used Theme.getBinaryResourceLocation() to locate image resources.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.colorchooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import javax.imageio.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.util.logging.*;

/**
 * A fixed size widget in {@link ColorChooserGuiImpl} dialog that allows to view and select with
 * the mouse the HUE and SATURATION of a color. The LUMINOSITY/VALUE is completely ignored.
 */
public class GamutPicker
extends ButtonGuiImpl
{
   /** Constant representing the height (including one-pixel width border) of the widget. */
   private final int NATIVE_HEIGHT = 189;
   
   /** Constant representing the width (including one-pixel width border) of the widget. */
   private final int NATIVE_WIDTH = 177;
   
   /** Flags whether we have already tried to load the bundle images. */
   private boolean imgLoaded = false;
   
   /**
    * The colored background of the picker. If {@code null}, the widget was not yet displayed or
    * load of the image has failed.
    */
   private ImageWrapper<BufferedImage> background = null;
   
   /**
    * The pointer of the selected color on the picker. If {@code null}, the widget was not yet
    * displayed or load of the image has failed.
    */
   private ImageWrapper<BufferedImage> pointer = null;
   
   /**
    * The currently selected HUE. The value uses the UI units (0 to {@code NATIVE_WIDTH} - 2) to
    * represent the location on the screen of the selected color (as opposed 0 - 240 interval used
    * by the {@code ColorChooserGuiImpl} and {@link ColorRgb}). 
    */
   private int hue = 0;
   
   /**
    * The currently selected SATURATION.  The value uses the UI units (0 to {@code NATIVE_HEIGHT}
    * - 2) to represent the location on the screen of the selected color (as opposed 0 - 240
    * interval used by the {@code ColorChooserGuiImpl} and {@link ColorRgb}).
    */
   private int sat = 0;
   
   /**
    * This constructor instantiate and initialize the object. 
    */
   public GamutPicker()
   {
      super(null/* ignored */, null/* ignored */);
      
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.isDefault = false;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.name = "gamut picker"; // direct access
      bcfg.help = null;
      
      initialize(bid, bcfg);
      initButtonText("gamut picker");
      
      addFocusListener(this);
      setEnabled(true);
   }
   
   /**
    * Draw the picker widget. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      if (!imgLoaded)
      {
         // if image resources are loaded yet, attempt to load them now
         background = loadImage(ThemeManager.getCurrentTheme().getBinaryResourceLocation() + 
                                "/colordialog/gamut.png");
         pointer = loadImage(ThemeManager.getCurrentTheme().getBinaryResourceLocation() + 
                             "/colordialog/gamut-ptr.png");
         imgLoaded = true;
      }
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () -> {
         if (background != null)
         {
            // draw the background
            gd.drawImage(background, 1, 1, background.getWidth(), background.getHeight(), 0, 0,
                         true, false, false, false);
         }
         
         if (pointer != null)
         {
            // draw the pointer to present the currently selected HUE/SAT
            gd.drawImage(pointer, hue - pointer.getWidth() / 2, sat - pointer.getHeight() / 2,
                         pointer.getWidth(), pointer.getHeight(), 0, 0,
                         true, false, false, false);
         }
         
         // draw the sunken border 
         gd.setColor(new ColorRgb(160, 160, 160)); // grey 1
         gd.drawLine(0, 0, d.width - 2, 0);
         gd.drawLine(0, 0, 0,           d.height - 2);
         
         gd.setColor(new ColorRgb(255, 255, 255)); // white
         gd.drawLine(0,           d.height - 1, d.width - 1, d.height - 1);
         gd.drawLine(d.width - 1, 0,            d.width - 1, d.height - 1);
      });
   }
   
   /**
    * Utility method for loading a resource bundle. Because the resources are usually theme
    * dependent this method uses the {@link Theme} class to open the resource input stream.
    *
    * @param   imagePath
    *          The path of the image to be loaded.
    * 
    * @return  The resource image wrapped in a {@link ImageWrapper} container object or
    *          {@code null} if loading or decoding of image failed.
    */
   public static ImageWrapper<BufferedImage> loadImage(String imagePath)
   {
      InputStream res = Theme.class.getResourceAsStream(imagePath);
      if (res != null)
      {
         try
         {
            BufferedImage image = ImageIO.read(res);
            return new ImageWrapper<>(image, image.getHeight(), image.getWidth());
         }
         catch (IOException e)
         {
            // TODO: create a dummy image ?
         }
      }
      
      ColorChooserGuiImpl.LOG.log(Level.WARNING, "Failed to load resource: [" + imagePath + "].");
      return null;
   }
   
   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return cc.heightFromNative(NATIVE_HEIGHT);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return cc.widthFromNative(NATIVE_WIDTH);
   }
   
   /**
    * Sets a combination of HUE and SATURATION as the selected value of the picker. The new values
    * are validated and converted to internal drawing coordinates. 
    * 
    * @param   hue
    *          New HUE value. Must be in 0-240 interval.
    * @param   sat
    *          New SATURATION value. Must be in 0-240 interval.
    */
   public void selectValue(int hue, int sat)
   {
      if (hue < 0 || hue > 240 || sat < 0 || sat > 240)
      {
         ColorChooserGuiImpl.LOG.log(Level.WARNING, "Invalid Hue/Sat combination");
         return;
      }
      
      // convert to our drawing coordinate
      this.hue = (hue * (NATIVE_WIDTH - 2)) / 240 + 1;
      this.sat = ((240 - sat) * (NATIVE_HEIGHT - 2)) / 240 + 1;
      
      repaint();
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget. The method analyses the
    * location of the click and selects the new values of HUE/SAT for internal and external use.
    * Then the {@link #command} is updated so that the action event emitted by super class will
    * bear the new selected HUE/SAT combination.  
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      NativePoint wOff = screenPhysicalLocation();
      
      // computes the new HUE/SAT internal representation
      this.hue = e.getX() - wOff.x - 1;
      this.sat = e.getY() - wOff.y - 1;
      
      // convert to 0-240 interval used standard by ColorChooserGuiImpl
      int h = hue * 240 / (NATIVE_WIDTH - 2);
      int s = 240 - sat * 240 / (NATIVE_HEIGHT - 2);
      if (h < 0 || s < 0 || h > 240 || s > 240)
      {
         return; // drop the action, compute error ?
      }
      
      setCommand("SET-HUE-SAT " + h + " " + s);
      super.mouseClicked(e);
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  Always {@code true} because this widget is NOT focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
}
