/*
** Module   : LuminosityPicker.java
** Abstract : A widget in ColorChooserGuiImpl dialog that allows to quickly select the LUMINOSITY 
**            component of a color. 
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20171016 Created initial version.
** 002 OM  20180503 Used Theme.getBinaryResourceLocation() to locate image resources.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.colorchooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.logging.*;

/**
 * A fixed size widget in {@link ColorChooserGuiImpl} dialog that allows to view and select with
 * the mouse the LUMINOSITY of a color. The HUE and SATURATION are used to display the active 
 * selected color but otherwise ignored.
 */
public class LuminosityPicker
extends ButtonGuiImpl
{
   /** Constant representing the height (including one-pixel width border) of the widget. */
   private final int NATIVE_HEIGHT = 189;
   
   /** Constant representing the total width of the widget. */
   private final int NATIVE_WIDTH = 20;
   
   /** Constant representing the width (including one-pixel width border) the gradient rule. */
   private final int GRADIENT_WIDTH = 12;
   
   /**
    * Constant representing the height of the banding. The bigger this number is the most evident
    * the banding is drawn (improves performance especially on web UI). When the constant is 
    * lowered the better the gradient is rendered but with performance penalty of increased
    * primitive number generated to GUI driver.
    */
   private final int BANDING_HEIGHT = 6;
   
   /** Flags whether we have already tried to load the bundle images. */
   private boolean imgLoaded = false;
   
   /**
    * The pointer of the selected color on the picker. If {@code null}, the widget was not yet
    * displayed or load of the image has failed.
    */
   private ImageWrapper<BufferedImage> pointer = null;
   
   /**
    * The currently selected HUE. Uses 0 - 240 interval used by default by all widgets in
    * {@code ColorChooserGuiImpl} and {@link ColorRgb}). Only used for preview.
    */
   private int hue = 0;
   
   /**
    * The currently selected SATURATION. Uses 0 - 240 interval used by default by all widgets in
    * {@code ColorChooserGuiImpl} and {@link ColorRgb}). Only used for preview.
    */
   private int sat = 0;
   
   /**
    * The currently selected LUMINOSITY. Uses 0 - 240 interval used by default by all widgets in
    * {@code ColorChooserGuiImpl} and {@link ColorRgb}).
    */
   private int lum = 0;
   
   /**
    * This constructor instantiate and initialize the object.
    */
   public LuminosityPicker()
   {
      super(null/* ignored */, null/* ignored */);
      
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.isDefault = false;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.name = "luminosity picker"; // direct access
      bcfg.help = null;
      
      initialize(bid, bcfg);
      initButtonText("luminosity picker");
      
      addFocusListener(this);
      setEnabled(true);
   }
   
   /**
    * Draw the picker widget. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      if (!imgLoaded)
      {
         // if image resource is loaded yet, attempt to load it now
         pointer = GamutPicker.loadImage(
               ThemeManager.getCurrentTheme().getBinaryResourceLocation() +
               "/colordialog/lum-ptr.png");
         imgLoaded = true;
      }
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () -> {
         // draw the rectangle pointer bundled as image
         if (pointer != null)
         {
            int grLum = (240 - lum) * (NATIVE_HEIGHT - 2) / 240 + 1; 
            gd.drawImage(pointer, NATIVE_WIDTH - pointer.getWidth(), grLum - pointer.getHeight() / 2,
                         pointer.getWidth(), pointer.getHeight(), 0, 0,
                         true, false, false, false);
         }
         
         // draw the gradient having the full spectrum of luminosity for same HUE/SAT
         for (int k = 0; k < d.height + BANDING_HEIGHT; k += BANDING_HEIGHT)
         {
            ColorHSL band = new ColorHSL(hue, sat, 240 - k * 240 / (d.height + BANDING_HEIGHT));
            gd.setColor(band.toRGB());
            gd.fillRect(1, k - BANDING_HEIGHT / 2, GRADIENT_WIDTH - 2, BANDING_HEIGHT);
         }
         
         // then draw the sunken 1-pixel border
         gd.setColor(new ColorRgb(160, 160, 160)); // grey 1
         gd.drawLine(0, 0, GRADIENT_WIDTH - 2, 0);
         gd.drawLine(0, 0, 0,                  d.height - 2);
         
         gd.setColor(new ColorRgb(255, 255, 255)); // white
         gd.drawLine(0,                  d.height - 1, GRADIENT_WIDTH - 1, d.height - 1);
         gd.drawLine(GRADIENT_WIDTH - 1, 0,            GRADIENT_WIDTH - 1, d.height - 1);
      });
   }
   
   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return cc.heightFromNative(NATIVE_HEIGHT);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return cc.widthFromNative(NATIVE_WIDTH);
   }
   
   /**
    * Sets a combination of HUE, SATURATION and LUMINOSITY as the value of the picker. The new
    * values are validated the new gradient is drawn.
    *
    * @param   hue
    *          New HUE value. Must be in 0-240 interval.
    * @param   sat
    *          New SATURATION value. Must be in 0-240 interval.
    * @param   lum
    *          New LUMINOSITY value. Must be in 0-240 interval.
    */
   public void selectValue(int hue, int sat, int lum)
   {
      if (hue < 0 || hue > 240 || sat < 0 || sat > 240 || lum < 0 || lum > 240)
      {
         ColorChooserGuiImpl.LOG.log(Level.WARNING, "Invalid Hue/Sat/Lum combination");
         return;
      }
      
      this.hue = hue;
      this.sat = sat;
      this.lum = lum;
      
      repaint();
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget. The method analyses the
    * location of the click and computes the LUMINOSITY of the gradient at the clicked point.
    * Then the {@link #command} is updated so that the action event emitted by super class will
    * bear the new computed LUMINOSITY.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // use the mouse click position to compute the LUMINOSITY of the gradient in that point.
      NativePoint wOff = screenPhysicalLocation();
      this.lum = 240 - (e.getY() - wOff.y - 1) * 240 / (NATIVE_HEIGHT - 2);
      
      if (lum < 0 || lum > 240)
      {
         return; // drop the action
      }
      
      setCommand("SET-LUM " + lum);
      super.mouseClicked(e);
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  Always {@code true} because this widget is NOT focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
}
