/*
** Module   : EmbeddedImageStructure.java
** Abstract : A structure used to store images drawing parameters.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SBI 20151011 First version.
**/

package com.goldencode.p2j.ui.client.gui.driver;

import java.util.*;


/**
 * Encapsulates the image drawing parameters.
 * 
 * @param    <I>
 *           The source image type.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
public class EmbeddedImageStructure<I>
{
   /** convert 3d image bit mask */
   static final int CONVERT_3D_MASK  = 1;

   /** transparent image bit mask */
   static final int TRANSPARENT_MASK = 2;

   /** stretch to fit image bit mask */
   static final int STRETCH_TO_FIT_MASK = 4;

   /** retain shape image bit mask */
   static final int RETAIN_SHAPE_MASK   = 8;


   /** Width parameter */
   private final int width;
   
   /** Height parameter */
   private final int height;
   
   /** X offset to start drawing image. */
   private final int xOffset;
   
   /** Y offset to start drawing image. */
   private final int yOffset;

   /** Image to draw */
   private final ImageWrapper<I> img;

   /** The bits mask that encodes image conversions. */
   private final int imageConversionMask;

   /**
    * Creates the embedded image structure filled with the required parameters taken from
    * the provided paint structure. 
    * 
    * @param paintStructure
    *        The paint structure parameterized by F and I types.
    */
   private <F> EmbeddedImageStructure(PaintStructure<F, I> paintStructure)
   {
      this.width   = paintStructure.width;
      this.height  = paintStructure.height;
      this.xOffset = paintStructure.xOffset;
      this.yOffset = paintStructure.yOffset;
      this.img     = paintStructure.img;
      this.imageConversionMask = encodeImageConversionMask(paintStructure);
   }

   /**
    * Creates the embedded image structure filled with the required parameters to draw the target.
    * 
    * @param    ps
    *           The paint structure.
    * 
    * @return   The embedded image structure.
    */
   public static <F, I> EmbeddedImageStructure<I> create(PaintStructure<F, I> ps)
   {
      return new EmbeddedImageStructure<I>(ps);
   }

   /**
    * Encodes image conversions: 3d, transparent, stretch, retain.
    * 
    * @param    paintStructure
    *           The paint structure parameterized by F and I types. 
    * 
    * @return   The bits mask that encodes image conversions.
    */
   private static <F, I> byte encodeImageConversionMask(PaintStructure<F, I> paintStructure)
   {
      byte mask = 0;
      if (paintStructure.convert3D)
      {
         mask |= CONVERT_3D_MASK;
      }
      if (paintStructure.transparent)
      {
         mask |= TRANSPARENT_MASK;
      }
      if (paintStructure.stretchToFit)
      {
         mask |= STRETCH_TO_FIT_MASK;
      }
      if (paintStructure.retainShape)
      {
         mask |= RETAIN_SHAPE_MASK;
      }

      return mask;
   }

   /**
    * Returns the image canvas width.
    * 
    * @return   The width.
    */
   public int getWidth()
   {
      return this.width;
   }

   /**
    * Returns the image canvas height.
    * 
    * @return The canvas height.
    */
   public int getHeight()
   {
      return this.height;
   }

   /**
    * The image drawing offset along x-axis.
    * 
    * @return   The xOffset
    */
   public int getXOffset()
   {
      return this.xOffset;
   }

   /**
    * The image drawing offset along y-axis.
    * 
    * @return   The yOffset.
    */
   public int getYOffset()
   {
      return this.yOffset;
   }

   /**
    * Returns the wrapped image.
    * 
    * @return   The wrapped image.
    */
   public ImageWrapper<I> getImage()
   {
      return this.img;
   }

   /**
    * Returns the target image to draw.
    * 
    * @return   The target image to draw.
    */
   public I getWrappedImage()
   {
      return (this.img != null) ? this.img.getImage() : null;
   }

   /**
    * Encodes image conversions: 3d, transparent, stretch, retain.
    * 
    * @return   The bits mask that encodes image conversions.
    */
   public int getImageConversionMask()
   {
      return this.imageConversionMask;
   }

   /**
    * Returns the embedded image identity 6-tuple.
    * 
    * @return   The array of 6 integer to represent the embedded image identity.
    */
   public List<Integer> getObjectSeal()
   {
      return Arrays.<Integer>asList(
               new Integer [] {
                           this.xOffset, this.yOffset, this.width, this.height,
                           this.imageConversionMask, this.img.getUniqueId()
                        });
   }

   /**
    * Indicates the image should retain shape.
    * 
    * @return   true iff the image should retain shape. 
    */
   public boolean isRetainShape()
   {
      return hasConversion(RETAIN_SHAPE_MASK);
   }

   /**
    * Indicates the image should be 3D style.
    * 
    * @return   true iff the image should have 3D style. 
    */
   public boolean is3D()
   {
      return hasConversion(CONVERT_3D_MASK);
   }

   /**
    * Indicates the image should stretch to fit the embedded space.
    * 
    * @return   true iff the image should stretch to fit the embedded space.
    */
   public boolean isStretchToFit()
   {
      return hasConversion(STRETCH_TO_FIT_MASK);
   }

   /**
    * Indicates the image should be transparent.
    * 
    * @return   true iff the image should be transparent.
    */
   public boolean isTransparent()
   {
      return hasConversion(TRANSPARENT_MASK);
   }

   /**
    * Helper method to get the conversion mask is supported or not.
    * 
    * @param    conversion
    *           The conversion mask that can be one of CONVERT_3D_MASK, RETAIN_SHAPE_MASK,
    *           STRETCH_TO_FIT_MASK, TRANSPARENT_MASK.
    * 
    * @return   true iff the target conversion is supported.
    */
   private boolean hasConversion(int conversion)
   {
      return (this.imageConversionMask & conversion) == conversion;
   }
}