/*
** Module   : GuiDriver.java
** Abstract : top-level GUI driver API
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support; some other new APIs were added.
** 003 CA  20140805 Added mouse event support and misc primitives for window actions/support.
** 004 MAG 20140902 Integrate COLOR-TABLE. Added support for BGCOLOR and FGCOLOR color conversion.
**                  Added system colors used solely for widget drawing.  
** 005 MAG 20140912 Implement GUI status bar. 
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Added new APIs to allow:
**                  - drawing of nested containers (translatedOrigin)
**                  - notification of any mouse event (listenAnyMouseEvent)
**                  - batch drawing (start/endBatch).
** 009 HC  20141031 Implemented support for drawing highlighted areas.
** 010 MAG 20141029 Added clip rectangle parameter to draw method.
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 EVL 20141116 Adding round rectangles draw and fill.  Adding api to get the UI theme name.
** 013 CA  20141123 Added APIs to allow enforcement of window stacking at OS level.
** 014 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 015 EVL 20141125 Adding dotted line draw capability.
** 016 CA  20141205 Added API to draw relative to window's real origin (top-left corner), 
**                  regardless of the current translated origin.
**                  Added API to register editable widgets (thus the driver can adjust the mouse 
**                  cursor accordingly).
** 017 CA  20141220 Added APIs for clipboard access.  Added APIs to allow text drawing and font 
**                  selection matching legacy sizes.
** 018 HC  20141207 Implemented runtime support for several SESSION attributes.
** 019 EVL 20141211 Adding tooltip worker registration.
** 020 EVL 20150203 Adding image offset drawing capability.
** 021 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 022 CA  20150317 Added API to determine the font's max width.
** 023 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 024 CA  20150507 Added APIs to draw text centered vertically.
** 025 GES 20150331 Reworked API to remove Swing-specific implementation features (so that the
**                  GUI web client can be cleanly implemented.
** 026 CA  20150520 Changed signatures for scaleFont and createFont.
** 027 HC  20150514 Initial support for modal windows.
** 028 HC  20150624 Improved Widget Browser GUI highlight. Restoring the highlighted area was
**                  causing artifacts. Also the highlighted area was frozen for draw updates.
** 029 HC  20150702 Implemented window modality independently on the actual UI driver. Improved
**                  positioning of top-level windows on systems with multiple displays.
** 030 GES 20150730 Removed translatedOrigin() since it isn't actually used.translate() was
**                  converted to translatePush()/translatePop() to make it clear that they need
**                  to have matched pairs.
** 031 HC  20150831 Scrolling support overhaul.
** 032 GES 20150817 Removed isOnTop() since it is unnecessary, added setIcon().
** 033 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 034 SVL 20151013 Added setXORComposite and resetComposite.
** 035 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
** 036 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     EVL 20151106 Adding grey image creation for the given original.
** 037 HC  20160104 Added the capability to identify active window in gui drivers.
** 038 EVL 20160107 Adding overlay window positioning.
** 039 EVL 20160123 Removing isWindowActive() method.  Logic is now in WindowManager. 
** 040 IAS 20160130 Renamed registerEditableWidget to registerHoverableWidget,
**                  LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 041 SBI 20160110 Added new method to query if the target font is installed or not.
**     CA  20160216 Added optional widget ID parameter to draw, which will let the driver record
**                  and cache the state of this widget.  Added font and text-widths APIs
**     SBI 20160324 Added new methods isWindowDecorationSupported() and
**                  setDesktopBgColor(ColorRgb color).
** 042 HC  20160407 Removed setOverlayWindowLocation(), the method was unnecessary and was
**                  causing the COMBO-BOX drop-down location incorrect.
** 043 HC  20160607 Added the method withSelectedWindow().
** 044 CA  20170104 Added flag to track window's modal state.
** 045 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
**     CA  20161101 Added API for drawing text with mnemonic.
**     HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
** 046 OM  20170320 Added support for Themes.
** 047 OM  20170502 Drooped dead code (getUiTheme() method).
** 048 OM  20170518 Small javadoc fixes.
** 049 EVL 20170608 Adding function to register direct manipulation handler and method to draw
**                  or clean selection box.
** 050 EVL 20170831 Adding method to set up file size uploading options.  Adding method to do
**                  client specific setups for different client drivers.
** 051 HC  20170819 Added extension widget method GET-MOUSE-POSITION. Added runtime support of
**                  WINDOW:SHOW-IN-TASKBAR.
** 052 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 053 HC  20171102 Implemented image load size and offset, plus related changes.
** 054 HC  20171123 Fixed a stack overlow condition in GUI redirected mode.
** 055 HC  20171203 Added notification of Web client when program execution enters modal or wait-for loop.
** 056 SBI 20180130 Added drawTextSelection, fillRectAroundText and drawCursor to implement
**                  native text selection and cursor positioning.
**     CA  20180207 flush() is handled internally by the driver's window implementation; this is
**                  required because flush can be done only after batch, translate and clip 
**                  brackets reach 0 (solves an issue in web mode draw cache, which requires to
**                  not have any open brackets).
**                  Warning: in web mode, unbalanced translate push/pop or CLIP/NO_CLIP will cause
**                  an infinite loop.  In Swing driver, it will not flush until all batch, 
**                  translate and clip brackets reach 0.
** 057 CA  20180210 Renamed deliverPDFPrintOutput to deliverPrintOutput and added report format 
**                  parameter.
** 058 SBI 20180410 Renamed deliverPrintOutput to deliverDocumentOutput in order to be used in
**                  OPEN-URL url. (OPEN-MIME-RESOURCE mimeType url [embedded].)
** 059 CA  20180313 Added FWD extension SESSION:DISABLE-REDRAW.
** 060 HC  20180525 Fixed mouse pointer loading.
** 061 HC  20180629 Added a check to independentEventDrawingBracket for window existence.
** 062 SVL 20180718 Added cacheTextMetrics and needsExplicitTextMetricsCaching.
** 063 CA  20180727 Removed setStatusText APIs.
** 064 SBI 20180815 Changed isFontInstalled in order to query if the tested bold font is installed.
**     OM  20180815 Added support for drawing ellipse primitives.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.LightweightTextWidget.*;
import com.goldencode.p2j.ui.client.gui.theme.Theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

import java.io.*;
import java.util.*;
import java.util.function.*;

/**
 * GUI screen driver contract.
 */
public interface GuiDriver<F, I>
extends ScreenDriver<GuiOutputManager>
{
   /**
    * Translates the origin of the graphics context to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * current coordinate system.  The modified context will have its new origin corresponding to
    * the point (<i>x</i>,&nbsp;<i>y</i>) in the context's former coordinate system.  All  
    * coordinates used in subsequent rendering operations on this graphics context are relative
    * to this new origin.
    * <p>
    * All usage of this method MUST be paired with a subsequent {@link #translatePop}. A stack
    * of translations is maintained, so that translate calls can be nested.
    * <p>
    * Warning: unbalanced translate push/pop will cause an infinite loop in web driver, as it will
    * keep waiting for the translate to reach 0.  In Swing driver, it will not flush until all
    * batch, translate and clip brackets reach 0.
    *
    * @param    x 
    *           the specified x coordinate
    * @param    y 
    *           the specified y coordinate
    */
   public void translatePush(int x, int y);
   
   /**
    * Reverses the translation at the top of the stack.
    * <p>
    * Warning: unbalanced translate push/pop will cause an infinite loop in web driver, as it will
    * keep waiting for the translate to reach 0.  In Swing driver, it will not flush until all
    * batch, translate and clip brackets reach 0.
    */
   public void translatePop();
   
   /**
    * Draw a string with a mnemonic.
    * 
    * @param    text
    *           The full text, which needs to be drawn.
    * @param    tp
    *           The text's coordinates.   If <code>centered</code> parameter is set to 
    *           <code>true</code>, then the {@link NativePoint#y} will represent the height on 
    *           which centerin will be done.
    * @param    info
    *           The mnemonic details.
    * @param    gf
    *           The widget's font details.
    * @param    widget
    *           The widget instance to which the text belongs.
    * @param    centered
    *           Flag indicating if the text needs to be centered vertically.
    */
   public void drawStringWithMnemonic(String                            text,
                                      NativePoint                       tp,
                                      MnemonicInfo                      info,
                                      GuiFontResolver                   gf,
                                      AbstractWidget<GuiOutputManager>  widget,
                                      boolean                           centered);
   
   /**
    * Draw a string at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   public void drawString(String text, int x, int y);
   
   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   height
    *          the height to center on.
    */
   public void drawStringCentered(String text, int x, int height);

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    legacyWidth
    *           The legacy width, which needs to be matched when drawing.
    * @param    legacyHeight
    *           The legacy width, which needs to be matched when drawing.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y);

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   height
    *          the height to center on.
    */
   public void drawStringCentered(String text, int legacyWidth, int legacyHeight, int x, int height);
   
   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link #drawParagraph(String, FontDetails, int, int, int)} and so
    * can be used in conjunction with this method. 
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   public int getParagraphHeight(String text, FontDetails<F> font, int maxWidth);
   
   /**
    * Draws a multiline text (paragraph) at the supplied position and 
    * with the specified font. The paragraph width will not exceed the
    * supplied maximum width. The text will be properly broken into
    * multiple lines.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link #getParagraphHeight(String, FontDetails, int)} and so
    * can be used in conjunction with this method.
    *
    * @param    text
    *           The text to draw.
    * @param    font
    *           The drawing font.
    * @param    x
    *           The x position.
    * @param    y
    *           The y position.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   public void drawParagraph(String text, FontDetails<F> font, int x, int y, int maxWidth);

   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param    x1  
    *           the first point's <i>x</i> coordinate.
    * @param    y1  
    *           the first point's <i>y</i> coordinate.
    * @param    x2  
    *           the second point's <i>x</i> coordinate.
    * @param    y2  
    *           the second point's <i>y</i> coordinate.
    */
   public void drawLine(int x1, int y1, int x2, int y2);

   /**
    * Draw a rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    */
   public void drawRect(int x, int y, int width, int height);
   
   /**
    * Draw a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter);

   /**
    * Fill a rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    */   
   public void fillRect(int x, int y, int width, int height);

   /**
    * Fill a rectangle around the given text.
    * 
    * @param   x
    *          The text left position
    * @param   y
    *          The text top position
    * @param   textWidthHint
    *          The text hint width
    * @param   textHeightHint
    *          The text hint height
    * @param   text
    *          The given text
    */
   public void fillRectAroundText(int x, int y, int textWidthHint, int textHeightHint, String text);
   
   /**
    * Fill a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */   
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter);
   
   /**
    * Draws an ellipse.
    *
    * @param    x
    *           Left position.
    * @param    y
    *           Top position.
    * @param    width
    *           Ellipse width.
    * @param    height
    *           Ellipse height.
    */
   public void drawEllipse(int x, int y, int width, int height);
   
   /**
    * Fills an ellipse.
    *
    * @param    x
    *           Left position.
    * @param    y
    *           Top position.
    * @param    width
    *           Ellipse width.
    * @param    height
    *           Ellipse height.
    */
   public void fillEllipse(int x, int y, int width, int height);
   
   /**
    * Draw a selection box rectangular area with current color.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    clear
    *           <code>TRUE</code> means clean up area instead of drawing selections.
    */
   public void drawSelection(int x, int y, int width, int height, boolean clear);
   
   /**
    * Draw a text selection.
    * 
    * @param    textLine
    *           The text selection object
    * @param    textLookAndFeel
    *           The text look and feel
    * @param    x
    *           The left position
    * @param    y
    *           The top position
    * @param    lineHeight
    *           The line height
    * @param    fontDescent
    *           The font descent
    */
   public void drawTextSelection(TextLineSelection textLine,
                                 TextLookAndFeel textLookAndFeel,
                                 int x,
                                 int y,
                                 int lineHeight,
                                 int fontDescent);

   /**
    * Reset selection boxes painting mode and return to normal drawing mode.
    */
   public default void resetSelection()
   {
      // do nothing by default
   }
   
   /**
    * Initializes several client specific internal variables used with drag-n-drop.
    * 
    * @param    srv
    *           Server to make config related calls.
    */
   public default void initDragAndDrop(ServerExports srv)
   {
      // do nothing by default in regular GUI driver
   }
   
   /**
    * Setting up file size limits for cases when drag and drop needs file uploads from web client
    * to Java side.
    * 
    * @param    single
    *           Single file size limit.
    * @param    total
    *           Total file size limit for single operation. 
    */
   public default void setUploadFileSizeLimits(int single, int total)
   {
      // do nothing by default
   }
   
   /**
    * Fill polygon.
    * 
    * @param    xPoints
    *           X coordinates.
    * @param    yPoints
    *           Y coordinates.
    */   
   public void fillPolygon(int[] xPoints, int[] yPoints);
   
   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>ColorRgb</code>.  This method uses
    * the current <code>ColorRgb</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be filled.
    * @param    y 
    *           the y coordinate of the rectangle to be filled.
    * @param    width
    *           the width of the rectangle to be filled.
    * @param    height 
    *           the height of the rectangle to be filled.
    * @param    raised 
    *           a boolean value that determines whether the rectangle appears 
    *           to be raised above the surface or etched into the surface.
    */
   public void fill3DRect(int x, int y, int width, int height, boolean raised);
   
   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>ColorRgb</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be drawn.
    * @param    y 
    *           the y coordinate of the rectangle to be drawn.
    * @param    width 
    *           the width of the rectangle to be drawn.
    * @param    height 
    *           the height of the rectangle to be drawn.
    * @param    raised 
    *           A boolean that determines whether the rectangle appears to be raised 
    *           above the surface or sunk into the surface.
    */
   public void draw3DRect(int x, int y, int width, int height, boolean raised);
   
   /**
    * Draw an image at the specified location.
    * 
    * @param    img
    *           Image to draw.
    * @param    x
    *           left location.
    * @param    y
    *           top location.
    * @param    w
    *           the image width.
    * @param    h
    *           the image height.
    * @param    xoff
    *           x offset inside image to start drawing.
    * @param    yoff
    *           y offset inside image to start drawing.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   public void drawImage(ImageWrapper<I> img,
                         int             x,
                         int             y,
                         int             w,
                         int             h,
                         int             xoff,
                         int             yoff,
                         boolean         transparent,
                         boolean         stretch,
                         boolean         retain,
                         boolean         convert3D);

   /**
    * Creates a sub image from the supplied source image, offset coordinates and size. The method will throw
    * a runtime exception when the offset or size is outside of the image bounds.
    *
    * @param   img
    *          The source image to cut, the instance is not modified, instead new image wrapper is returned.
    * @param   x
    *          The offset x.
    * @param   y
    *          The offset y.
    * @param   width
    *          The width relative to the offset.
    * @param   height
    *          The height relative to the offset/.
    *
    * @return  New image wrapper instance representing the sub image.
    */
   public ImageWrapper<I> getSubImage(ImageWrapper<I> img, int x, int y, int width, int height);

   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           area width
    * @param    height
    *           area height
    * @param    color
    *           the highlight color
    */
   public void drawHighlight(int x, int y, int width, int height, ColorRgb color);

   /**
    * Set current color used on draw operations.
    * 
    * @param    color
    *           Drawing color.
    */
   public void setColor(ColorRgb color);
   
   /**
    * Set current color used on draw operations.
    * 
    * @param    color
    *           Drawing color.
    */
   public void setColor(java.awt.Color color);
   
   /**
    * Set the desktop background color.
    * 
    * @param    color
    *           The new desktop background color.
    */
   public void setDesktopBgColor(ColorRgb color);
   
   /**
    * Set "XOR composite" mode: color of the drawn elements is calculated as XOR (for each RGB
    * component) of their color with color of underlying pixels.
    */
   public void setXORComposite();

   /**
    * Set XOR mode painting allowing to invert the destination color with current drawing
    * primitive color.  Double painting of the same figure at the same location should restore
    * original picture.
    * 
    * @param    color
    *           XOR alternation color.
    */
   public void setXORMode(ColorRgb color);

   /**
    * Reset composite mode and return to normal drawing mode.
    */
   public void resetComposite();
   
   /**
    * Set the window title at the OS level.
    * 
    * @param    title
    *           The window's title.
    */
   public void setTitle(String title);
   
   /**
    * Set current font used on draw operations.
    * 
    * @param    font
    *           Drawing font.
    */
   public void setGuiFont(FontDetails<F> font);
   
   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.  This will be modified on return.
    */
   public void scaleFont(FontDetails<F> font);
   
   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param    style
    *           Font style.
    */
   public void setFontStyle(FontStyle style);

   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param    stroke
    *           Line style.
    */
   public void setLineStroke(LineStroke stroke);

   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param    stroke
    *           Line style.
    * @param    width
    *           Width value for the line.
    */
   public void setLineStroke(LineStroke stroke, float width);

   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   public void resizeWindow(int width, int height);
   
   /**
    * Set window bounds.
    * 
    * @param    x
    *           new x
    * @param    y
    *           new y
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   public void setWindowBounds(int x, int y, int width, int height);

   /** 
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>false</code>
    *           otherwise.
    */
   public boolean hasWindowEmulator(int windowId);

   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    showInTaskbar
    *           If set to {@code true} the window will be shown in the system taskbar.
    */
   public void registerWindow(int windowId, boolean showInTaskbar);
   
   /**
    * Register new child window with the driver.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           The window ID.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    */
   public void registerChildWindow(int ownerId, int windowId, String windowTitle, 
                                   boolean resizable, boolean isShareActivationWithOwner,
                                   boolean modal);
   
   /**
    * Deregister a given window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   public void deregisterWindow(int windowId);
   
   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * <p>
    * The GUI driver primitives can be accessed by only a single thread at a time.  This requires
    * all GUI driver API calls to be bracketed in {@link #selectWindow} and {@link #releaseWindow}
    * calls.
    * <p>
    * If the thread drawing owner is <code>null</code>, the current thread will gain exclusive
    * access immediately and will hold it until has finished work.
    * 
    * @param   windowId
    *          Window ID.
    */
   public void selectWindow(int windowId);
   
   /**
    * The method works as an envelope for {@link #selectWindow(int)} and 
    * {@link #releaseWindow()} methods.
    * 
    * @param   windowId
    *          Window ID.
    * @param   run
    *          The code to run after the window is selected in the driver.
    */
   public default void withSelectedWindow(int windowId, Runnable run)
   {
      selectWindow(windowId);
      
      try
      {
         run.run();
      }
      finally
      {
         releaseWindow();
      }
   }
   
   /**
    * Release the last window used by the current thread.  This can be called only by the same
    * thread as the drawing owner thread which acquired the exclusive drawing access.
    * <p>
    * If no more windows are in use by the current thread, then the exclusive drawing access is
    * released and other other threads waiting in {@link #selectWindow} can acquire it.
    * 
    */
   public void releaseWindow();
   
   /**
    * Save the current drawing state and release the drawing lock, so that other threads can 
    * acquire it.  This is useful in cases when the current thread needs to wait an indefinite
    * amount of time (i.e. for keyboard/mouse events).
    * 
    * @return   An object encoding the drawing state.
    */
   public Object saveDrawingState();
   
   /**
    * Restore the drawing state to the specified state.  If the current thread is not currently 
    * the drawing oOwner, then wait for it to relinquish the exclusive lock, before 
    * restoring the state.
    * 
    * @param    state
    *           The drawing state.
    */
   public void restoreDrawingState(Object state);
   
   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top);
   
   /**
    * Set the window details for resize; this includes the minimum and maximun dimensions, and 
    * also the window's resizeable state.
    * 
    * @param    resize
    *           Flag indicating if the window can be resized or not.
    * @param    minWidth
    *           The window's minimum width.
    * @param    minHeight
    *           The window's minimum height.
    * @param    maxWidth
    *           The window's maximum width.
    * @param    maxHeight
    *           The window's maximum height.
    */
   public void setResizeableWindow(boolean resize,
                                   int     minWidth, 
                                   int     minHeight, 
                                   int     maxWidth, 
                                   int     maxHeight);
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   public NativePoint getWindowLocation();
   
   /**
    * Create a native grey image from the original initial image to be used for disabled control
    * states.
    *
    * @param    imageIni
    *           The image to be used as source for grey transformation.
    *
    * @return   The generic wrapper representing the given image or <code>null</code> if the
    *           image cannot be created (e.g. if the bytes are not a valid image type).
    */
   public ImageWrapper<I> createGreyImage(I imageIni);
   
   /**
    * Create a native icon from the raw bytes and return a generic wrapper that can be used
    * for operations.
    *
    * @param    byts
    *           The image as persisted in the file system or a JAR file.
    * @param    small
    *           <code>true</code> if the small icon size is to be loaded.
    *
    * @return   The generic wrapper representing the given icon or <code>null</code> if the
    *           icon cannot be created (e.g. if the bytes are not a valid icon type).
    */
   public ImageWrapper<I> createIcon(byte[] byts, boolean small);
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    details
    *           The font specifications.
    *
    * @return   A font instance.
    */
   public FontDetails<F> createFont(FontDetails<F> details);
   
   /**
    * Queries if the target font is installed.
    * 
    * @param    fontName
    *           The target font name.
    * @param    boldFont
    *           Indicates that the bold font is queried.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   public boolean isFontInstalled(String fontName, boolean boldFont);
   
   /**
    * Get the width of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's width.
    */
   public int getFontWidth(FontDetails<F> font);

   /**
    * Get the maximum font width, for the given font.
    * 
    * @return   The font's maximum width.
    */
   public int getFontMaxWidth(FontDetails<F> font);
   
   /**
    * Report the advance widths of the first 256 characters in the font.  Advance width measures 
    * the length of the character's baseline.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font widths.
    */
   public int[] getFontWidths(FontDetails<F> fd);
   
   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   public int getFontHeight(FontDetails<F> font);
   
   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   public int getTextWidth(String text, FontDetails<F> font);

   /**
    * Get and cache metrics for the given texts.
    *
    * @param  texts
    *         Map of texts for which metrics should be calculated, keyed by fonts which are used
    *         for drawing specific texts.
    * @param  windowId
    *         ID of the parent window. Can be <code>null</code> if there is no need to select a
    *         window.
    */
   public void cacheTextMetrics(Map<FontDetails<F>, Set<String>> texts, Integer windowId);

   /**
    * Returns <code>true</code> if getting actual text metrics is an expensive operation for this
    * driver and you want to manually manage metrics cache using
    * {@link #cacheTextMetrics(Map, Integer)}.
    *
    * @return see above.
    */
   public boolean needsExplicitTextMetricsCaching();

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public int getTextHeight(String text, FontDetails<F> font);
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible);
   
   /**
    * Set the window's sensitive state, depending on the {@code enable} flag.
    * 
    * @param    enable
    *           The window sensitive state.
    */
   public void setWindowEnabled(boolean enable);

   /**
    * Iconify the window.
    *
    * @param   windowId
    *          Target window.
    */
   public void iconifyWindow(int windowId);

   /**
    * De-iconify the window.
    *
    * @param   windowId
    *          Target window.
    */
   public void deiconifyWindow(int windowId);
   
   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   public NativeDimension getWindowDimension();

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    *
    * @param    id
    *           When not-null, it caches the drawing result of this widget.
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   public void draw(WidgetId id, NativePoint origin, NativeRectangle clip, Runnable core);

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * <p>
    * The method intersects the supplied <code>clip</code> with the screen bitmap rectangles
    * retrieved by {@link ScreenBitmap#getClippings(NativeRectangle)}.
    *
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   public void draw(NativePoint origin, NativeRectangle clip, Runnable core);

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the window's real (0,0) origin (top-left corner) and then translating to the
    * specified location, before executing the drawing.
    * <p>
    * The translated origin (as before this was called) needs to be restored, before this call
    * ends.
    * 
    * @param    origin
    *           The drawing origin (relative to window's real origin).
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   public void drawFromOrigin(NativePoint origin, NativeRectangle clip, Runnable core);

   /**
    * Limit the available drawing area to the specified boundaries provided in the
    * <code>bounds</code> parameter. The individual rectangular areas may or may not overlap.
    * <p>
    * Warning: unbalanced CLIP/NO_CLIP will cause an infinite loop in web driver, as it will
    * keep waiting for the clip brackets to reach 0.  In Swing driver, it will not flush until all
    * batch, translate and clip brackets reach 0.
    *
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingAreas(Set<NativeRectangle> bounds);

   /**
    * Limit the available drawing area to the specified boundaries.
    * <p>
    * Warning: unbalanced CLIP/NO_CLIP will cause an infinite loop in web driver, as it will
    * keep waiting for the clip brackets to reach 0.  In Swing driver, it will not flush until all
    * batch, translate and clip brackets reach 0.
    *
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(NativeRectangle bounds);
   
   /**
    * Sets the global SESSION:DISABLE-REDRAW attribute.
    * When this mode is active, all output is disabled.
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   public void setDisableRedraw(boolean value);
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void registerMoveableWidget(int wid);

   /**
    * Register the specified widget as a resizeable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void registerResizeableWidget(int wid);
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void deregisterWidget(int wid);

   /**
    * Register the specified widget for popup events, via mouse.
    * <p>
    * The popup with the given ID will be shown when the specified mouse button is pressed inside
    * the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   public void registerWidgetPopup(int wid, int btn, int pid);

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * 
    * @param    action
    *           The widget mouse action.
    */
   public void registerHoverableWidget(MouseHoverAction action);

   /**
    * Register this widget as having tooltip functionality reacting on mouse entering and exiting
    * events.  The registration is pending if tooltip engine is not available for session or
    * tooltip text is empty.
    * 
    * @param    widget
    *           The widget object to register.
    * @param    font
    *           The font to draw tooltip text.
    * @param    now
    *           The flag indicating whether the action be registered immediately or later.
    */
   public void registerTooltipWorker(Widget<GuiOutputManager> widget,
                                     int                      font,
                                     boolean                  now); 

   /**
    * Register the tooltip workers if this process was dalayed with regular registration.
    * 
    * @param    wid
    *           The widget id. 0 means register all pending workers. 
    */
   public void registerPendingTooltipWorkers(int wid);
   
   /**
    * Deregister the tooltip workers already registered with mouse handler.
    * 
    * @param    wid
    *           The widget id. 0 means deregister all registered workers. 
    */
   public void deregisterTooltipWorkers(int wid);
   
   /**
    * Register the specified widget as a container able to use direct manipulation via mouse.
    * 
    * @param    widget
    *           The widget object to register.
    */
   public void registerDirectManipulation(Widget<GuiOutputManager> widget);
   
   /**
    * Get BGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of ColorRgb
    */
   public ColorRgb getBgColor(int index);

   /**
    * Get FGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of ColorRgb
    */
   public ColorRgb getFgColor(int index);
 
   /**
    * Get a system color.
    * 
    * @param   index
    *          Color index.
    *          
    * @return  see above.
    */
   public ColorRgb getSysColor(int index);
   
   /**
    * Get foreground color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   public ColorRgb getFgStdColor(String stdColor);

   /**
    * Get background color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   public ColorRgb getBgStdColor(String stdColor);
   
   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    id
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   public void listenAnyMouseEvent(int id, long when);
   
   /** 
    * Start a batch drawing routine.
    */
   public void startBatch();
   
   /**
    * End a batch drawing routine.
    */
   public void endBatch();

   /**
    * Check if the currently selected window is visible.
    * 
    * @return   See above.
    */
   public boolean isWindowVisible();
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   public void stackWindows(int[] windowIds);
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window 
    *           and its state changed to active. 
    */
   public void moveToTop(int windowId, boolean focus);
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    */
   public void moveToBottom(int windowId);
   
   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    process
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public void enableEvents(int[] windowIds, boolean process);
   
   /**
    * Returns bounds of a physical screen the passed in window is positioned on.
    * The bounds are in virtual device coordinate system. For more information
    * about the virtual device coordinate system see {@link TopLevelWindow}.
    * <p>
    * For single-display setups the top-left corner of the returned rectangle is always 
    * 0,0. For multiple-display setups it can be 0,0 or positive.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The physical screen bounds.
    */
   public NativeRectangle getDisplayBounds(int windowId);
   
   /**
    * Returns an application-usable display area. For example the task bar is
    * not counted as the usable area.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   The dimensions in pixels.
    */
   public NativeRectangle getDisplayWorkArea(int windowId);
   
   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   public int getScreenResolution();

   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   public void copyToClipboard(String txt);

   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   See above.
    */
   public String clipboardContents();

   /**
    * Sets the current selection.
    * 
    * @param txt
    *        The current selection.
    */
   public void setCurrentSelection(String txt);

   /**
    * Register all the mouse-aware widgets (in all Windows) with the driver.
    */
   public void registerMouseWidgets();

   /**
    * Sets the image to the window to be displayed as the window icon representation
    * for the system task bar, the title window panel or the system tray icons panel. 
    * 
    * @param    icon
    *           The wrapped icon.
    * @param    x
    *           The icon left location.
    * @param    y
    *           The icon top location.
    * @param    width
    *           The icon width.
    * @param    height
    *           The icon height.
    * @param    xoff
    *           x offset inside image of the icon to select drawing region.
    * @param    yoff
    *           y offset inside image of the icon to select drawing region.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   public void setIcon(ImageWrapper<I> icon,
                         int             x,
                         int             y,
                         int             width,
                         int             height,
                         int             xoff,
                         int             yoff,
                         boolean         transparent,
                         boolean         stretch,
                         boolean         retain,
                         boolean         convert3D);
   /**
    * Load mouse pointer from the InputStream
    * 
    * @param data
    *        cursor data
    * @param ptrName
    *        pointer name
    *        
    * @return
    *        mouse pointer or <code>null</code>
    */
   public MousePtrWrapper createMousePointer(InputStream data, String ptrName);

   /**
    * Propagate wait state
    * 
    * @param isWaitState
    *        wait state
    */
   public default void propagateWaitState(boolean isWaitState)
   {
      // do nothing
   }

   /**
    * Set the mouse cursor
    * 
    * @param ptr
    *        mouse cursor
    */
   public default void setCursor(MousePtrWrapper ptr)
   {
      // do nothing
   }

   /**
    * Returns true if a window can be decorated in this graphics environment, otherwise false.
    * 
    * @return   true if a window can have its title, caption buttons and border,
    *           otherwise false.
    */
   public boolean isWindowDecorationSupported();

   /**
    * Split the text for a editor widget, so that it fits the specified maximum width.  The 
    * <code>delta</code> parameter represents the start position of this text, in the editor's 
    * content; when the full content is parsed, this will be zero.
    *  
    * @param    widgetId
    *           The editor's ID.
    * @param    windowId
    *           The editor's window ID.
    * @param    maxWidth
    *           The width for the content to fit (the editor's editable area).
    * @param    text
    *           The text to split into lines.
    * @param    delta
    *           The start position of this text in the editor's full content.
    * @param    font
    *           The editor's font.
    *           
    * @return   Two arrays: one with the split indexes and one with the split indexes which were
    *           determined to be caused by a new line.
    */
   public default int[][] parseEditorContent(int            widgetId,
                                             int            windowId,
                                             int            maxWidth,
                                             String         text,
                                             int            delta,
                                             FontDetails<F> font)
   {
      // do nothing
      return null;
   }
   
   /**
    * Get the widths of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's widths.
    */
   public int[] getTextWidths(String text, FontDetails<F> font);
   
   /**
    * Get the widths of the specified texts, using the given font.
    * 
    * @param    texts
    *           The texts.
    * @param    font
    *           The font details used to draw and measure the text.
    *
    * @return   The texts' widths.
    */
   public int[][] getTextWidths(String[] texts, FontDetails<F> font);

   /**
    * Returns current mouse cursor display position relative to the display origin.
    *
    * @return  mouse cursor position, never <code>null</code>
    */
   public NativePoint getMousePosition();

   /**
    * The method delivers document output to the client. The supplied consumer is invoked by
    * the driver providing a suitable output stream. The consumer implementation is responsible
    * for sending the data to the provided stream. The consumer implementation must not close
    * the provided stream. When the finalizer is supplied, it will be called by the driver when
    * the document output is no longer needed. The finalizer should free up any additional
    * resources the print output may need.
    *
    * @param   consumer
    *          Output consumer, must not be null. See above.
    * @param   finalizer
    *          Finalizer callback, may be null. See above.
    * @param   format
    *          Document output format.
    */
   public void deliverDocumentOutput(Consumer<OutputStream> consumer, 
                                     Runnable               finalizer, 
                                     MediaType format);

   /**
    * Draws the cursor for the given text widget.
    * 
    * @param    widget
    *           The text widget
    */
   public void drawCursor(LightweightTextWidget widget);

   /**
    * Makes the supplied window to be on top of other windows. This method is used to
    * implement the top-only functionality directly in the driver when
    * {@link #stackWindows(int[])} is not fully supported by the driver.
    *
    * @param   windowId
    *          The window to be changed.
    * @param   topOnly
    *          Corresponds to the WINDOW:TOP-ONLY attribute.
    * @param   alwaysOnTop
    *          Corresponds to the WINDOW:ALWAYS-ON-TOP attribute.
    */
   default void setAlwaysOnTop(int windowId, boolean topOnly, boolean alwaysOnTop)
   {
      // no-op - this method is not required to be implemented by the driver
   }

   /**
    * The method is called when the 4GL program execution enters a modal or wait-for event loop, in other
    * words when user input is expected.
    */
   default void notifyWait()
   {
      // no-op by default
   }
}
