/*
** Module   : GuiPrimitives.java
** Abstract : GUI version of the output primitives
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Added mouse event support.  More misc changes related to window behavior.
** 004 MAG 20140905 Added fillPoly method.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Changed drawImage, to specify the image's width and height.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 EVL 20141116 Adding round rectangles draw and fill.
** 011 EVL 20141125 Adding dotted line draw capability.
** 012 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 013 HC  20141207 Implemented runtime support for several SESSION attributes.
** 014 EVL 20150203 Adding image offset drawing capability.
** 015 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 016 CA  20150317 Added API to determine the font's max width.
** 017 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 018 CA  20150507 Added APIs to draw text centered vertically.
** 019 GES 20150501 Minimized the APIs needed here to only those relaced to managing or accessing
**                  the registered windows and/or the current window. This class is now a kind of
**                  router that keeps the list of emulated window instances and a current
**                  instance which can be swapped.
** 020 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.ui.client.*;

/** 
 * This is a low level GUI-specific window emulator management interface.  Real drawing does not
 * occur in this class. This is a kind of router that keeps the list of emulated window instances
 * and a current instance which can be swapped.
 * <p>
 * This interface is for GUI-specific features. It is intended to be specific to a given
 * UI-modality (GUI in this case). No driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public interface GuiPrimitives
extends OutputPrimitives
{   
   /**
    * Get the currently selected window emulator.
    * 
    * @return    The current window emulator.
    */
   public EmulatedWindowState getWindowEmulator();
   
   /**
    * Returns the window emulator registered for the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   The identified window emulator.
    * 
    * @throws   RuntimeException
    *           When no window emulator is registered for the supplied window id.
    */
   public EmulatedWindowState getWindowEmulator(int windowId);
   
   /**
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>true</code>
    *           otherwise.
    */
   public boolean hasWindowEmulator(int windowId);
   
   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    ews
    *           New emulated window state to register.  
    */
   public void registerWindow(int windowId, EmulatedWindowState ews);

   /**
    * Deregister the given window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   public void deregisterWindow(int windowId);
   
   /**
    * On a GUI interface multiple windows can be created. Each window has it own emulator.
    * When a window is first drawn, this method is used to set it's emulator as the current 
    * target for output operations.  
    * 
    * @param    windowId
    *           Window ID.
    */
   public void selectWindow(int windowId);
   
   /**
    * On a device which allows multiple windows to be created, this allows releasing the window to
    * allow drawing from other threads.
    */
   public void releaseWindow();
   
   /**
    * Perform shutdown.
    */
   public void shutdown();
}