/*
** Module   : LightweightTextWidget.java
** Abstract : Defines a text widget facade to get cursor character position, and to draw a cursor.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SBI  20180116 Added getScreenText(), getCursorColumn[Row]Offset(), drawCursor().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.LightweightTextWidget
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.*;


/**
 * Defines a text widget facade to get cursor character position within the text content
 * represented by this text widget, and to draw a special cursor relative to the current
 * origin.
 */
public interface LightweightTextWidget
{
   /**
    * Gets the screen text content at the given line.
    * 
    * @param    line
    *           The given line number
    *           
    * @return   The screen text content
    */
   String getScreenText(int line);
   
   /**
    * Gets the screen text content at the current line.
    * 
    * @return   The screen text content
    */
   String getScreenText();
   
   /**
    * Gets the current cursor column within its text content.
    * 
    * @return   The current cursor column
    */
   int getCursorColumnOffset();
   
   /**
    * Gets the current cursor row within its text content.
    * 
    * @return   The current cursor row
    */
   int getCursorRowOffset();
   
   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll column
    */
   int getScrollColumnOffset();
   
   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll row
    */
   int getScrollRowOffset();
   
   /**
    * Returns the current font cache.
    * 
    * @return   The current font cache
    */
   FontCache getFontCache();
   
   /**
    * Draws a special cursor relative to the current origin.
    */
   void drawCursor();
   
   /**
    * Gets the current text selection object for the given line. If the given line doesn't have
    * selected text, then this method returns null.
    * 
    * @param    line
    *           The given line number
    * 
    * @return   The current text selection object if there is selected text in the given line,
    *           otherwise null.
    */
   TextLineSelection getTextSelection(int line);
   
   /**
    * Defines text selection objects in text lines.
    */
   public interface TextLineSelection
   {
      /**
       * Returns the column and row, where the current text selection gets the its origin.
       * 
       * @return   The beginning of the current text selection
       */
      NativePoint getSelectionStart();
      
      /**
       * Returns the most closest point, column and row, that breaks the continuous selection text
       * part.
       * 
       * @return   The break point, column and row, of the current text selection
       */
      NativePoint getSelectionEnd();
      
      /**
       * Returns the 3-tuple of string widths used as hint widths for this line split into outer
       * and selected text.
       * 
       * @return   The 3-tuple of string hint widths in pixels
       */
      int[] getTextWidthsHint();
      
      /**
       * Returns the 3-tuple of text parts into which this line has been split by text selection.
       * 
       * @return   The 3-tuple of text parts
       */
      String[] getLineParts();
   }
}
