/*
** Module   : MousePopupable.java
** Abstract : Implementation of a mouse action which shows a popup.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141021 Better implementation of the mouse popupable action.
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 HC  20150323 WidgetId.getId() renamed to asInt().
** 005 GES 20150501 Moved to a more generic package.
** 006 CA  20150910 Reworked to use the MenuGuiImpl widget.
** 007 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 008 VIG 20160307 Cleanup, fixed popup physical location, changed showPopupMenu() call.
** 009 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 010 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the AWT/Swing
// event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * Listener for mouse actions which can trigger a popup to be shown for a specified widget.
 */
public class MousePopupable
extends MouseWidgetAction
{
   /** The button on which the popup is associated. */
   private final int btn;

   /** The popup associated with this widget. */
   private Menu popup;

   /**
    * Create a new mouse popupable action, for the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   public MousePopupable(int wid, int btn, int pid)
   {
      super(wid, 
            btn == MouseEvent.BUTTON1 ? MouseEvt.MOUSE_POPUPABLE_LEFT 
                                      : MouseEvt.MOUSE_POPUPABLE_RIGHT);
      
      this.popup = (Menu) OutputManager.resolveWidget(pid);

      this.btn = btn;
   }

   /**
    * Notification of a mouse clicked event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      drawPopup(e);
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      drawPopup(e);
   }
   
   /**
    * Draw the {@link #popup}, based on current popup state and mouse event.
    * 
    * @param    evt
    *           The current mouse event.
    */
   @SuppressWarnings("unchecked")
   private void drawPopup(MouseEvent evt)
   {
      TopLevelWindow<?> window = popup.topLevelWindow();
      
      gd.selectWindow(window.getId().asInt());

      if (popup.isVisible())
      {
         ThinClient.getInstance().eventDrawingBracket(window, false, true, new Runnable()
         {
            @Override
            public void run()
            {
               popup.hide();
            }
         });
         
         window.getContentPane().moveToBottom(popup);
      }

      int evtID = evt.getID();
      if (evt.getButton() == btn && 
          (evtID == MouseEvent.MOUSE_CLICKED || evtID == MouseEvent.MOUSE_PRESSED))
      {
         // TODO: embedded mode?
         popup.topLevelWindow().getTitleBar().parent().moveToTop(popup);
         popup.showPopup(widget, new NativePoint(evt.getX(), evt.getY()), null);
      }

      gd.releaseWindow();
   }
}