/*
** Module   : MouseWidgetAction.java
** Abstract : Definition of a generic mouse action, to be applied to a widget.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 GES 20150501 Moved to a more generic package.
** 003 CA  20150826 Added explicit mouse events on which an instance can operate.
** 004 IAS 20160217 Added new constructor.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the AWT/Swing
// event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.MouseEvt;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * A generic action which can be applied to a specified widget, via mouse.
 */
public abstract class MouseWidgetAction
extends MouseAdapter
{
   /** The widget associated with this action. */
   protected final Widget<?> widget;
   
   /** The active GUI driver. */
   protected final GuiDriver gd;

   /** The low-level mouse actions on which this instance can operate. */
   protected Set<Integer> mouseActions;

   /** The {@link MouseEvt} event ID for this high-level action. */
   private final int eventId;
   
   /**
    * Create a new mouse action, for the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    * @param    eventId
    *           The event ID.
    */
   @SuppressWarnings("unchecked")
   public MouseWidgetAction(int wid, int eventId)
   {
      this(wid, null, eventId);
   }

   /**
    * Create a new mouse action, for the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    * @param    mouseActions
    *           The low-level mouse events on which this instance operates.  May be 
    *           <code>null</code> if it can act on any mouse event.
    * @param    eventId
    *           The event ID.
    */
   @SuppressWarnings("unchecked")
   public MouseWidgetAction(int wid, int[] mouseActions, int eventId)
   {
      this((Widget<GuiOutputManager>) OutputManager.resolveWidget(wid), mouseActions, eventId);
   }


   /**
    * Create a new mouse action, for the specified widget.
    * 
    * @param    widget
    *           The widget.
    * @param    mouseActions
    *           The low-level mouse events on which this instance operates.  May be 
    *           <code>null</code> if it can act on any mouse event.
    * @param    eventId
    *           The event ID.
    */
   @SuppressWarnings("unchecked")
   public MouseWidgetAction(Widget<?> widget, int[] mouseActions, int eventId)
   {
      this.widget = widget;
      
      gd = (GuiDriver) OutputManager.getDriver();
      
      if (mouseActions != null)
      {
         this.mouseActions = new HashSet<>(Utils.primitiveToIntegerList(mouseActions));
         this.mouseActions = Collections.unmodifiableSet(this.mouseActions);
      }
      else
      {
         this.mouseActions = null;
      }
      
      this.eventId = eventId;
   }

   /**
    * Get the {@link MouseEvt event ID} for this high-level action.
    * 
    * @return   See above.
    */
   public int getEventId()
   {
      return eventId;
   }

   /**
    * Determine if this complex action requires to process the given mouse event.
    * 
    * @param    e
    *           The mouse event.
    *           
    * @return   If the event's code is part of the {@link #mouseActions}.
    */
   public boolean hasEvent(MouseEvent e)
   {
      return mouseActions == null || mouseActions.contains(e.getID());
   }
}