/*
** Module   : GuiWebSocket.java
** Abstract : web-sockets implementation which provides a protocol for GUI
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 GES 20150331 First version which is a placeholder using the common parent code as ChUI.
** 002 GES 20150701 Added messages for a first pass at GUI support,
**                  implemented an images drawing operation.
** 003 CA  20150810 Added operations for font, metrics and string drawing.
** 004 CA  20150821 Added z-order, line strokes, mouse support. Moved the message receiving APIs 
**                  to parent class. Fixed font registration: if the JS font-table already has a
**                  font with the same specification, reuse the entry. Improvements for mouse
**                  support; added window activated.  Added setIconImage().
**                  deriveFont() now is driver specific - it scales the font depending on the DPI.
** 005 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
**                  z-order fixes: the full z-order widget list must always be sent if there are
**                  changes in the widget tree (adjusting the new/dead widget list can not provide
**                  correct z-order information).
** 006 SVL 20151013 Added setXORComposite and resetComposite.
** 007 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
** 008 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151026 Fixed mouse processing for modal windows.
** 009 CA  20151126 When waiting for a message result to be received from the JS side, each message
**                  must be uniquely identified; otherwise, if the same message type is sent in
**                  the same batch more than once, it will collide and it will not preserve the
**                  message result.
**     SBI 20151201 Changed to provide processMouseWidgets with calculated widgets bounds.
** 010 EVL 20151223 Protection from NPE in child window creation.  The child window can have no
**                  title, like the overlay window case.
** 011 IAS 20160217 LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 012 SBI 20151208 Added new parameters, scroll amount and scroll unit for wheel events and
**                  a mouse modifiers mask for all mouse events.
**     SBI 20160107 Supported new message type MSG_IS_FONT_INSTALLED to request if the target font
**                  family is installed on the client side.
**     CA  20160629 Added a MD5 hash for each drawing batch sent to the JS side, via an expiry 
**                  cache.  Added MSG_PARSE_EDITOR_CONTENT support.
**     SBI 20160324 Added setDesktopBgColor(int red, int green, int blue).
** 013 HC  20160406 Overhaul of window-activation logic.
** 014 SBI 20160414 Fixed the window location coordinates to have 32 bits size if they are send
**                  with MSG_SET_WINDOW_LOC and MSG_WINDOW_RESIZED.
**     CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
**     SBI 20160422 The resize drawing operations must not be cached.
** 015 CA  20160429 Fix for H014/CA: do not consume QUIT, STOP or ERROR conditions, as the 
**                  business logic might rely on them, if explicitly raised.
** 016 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
**     SBI 20160523 Fixed to handle clipboard messages without contents.
**     SBI 20160615 Added MSG_INVALIDATE_SELECTION.
** 017 CA  20161018 Any client-to-server code invocations must maintain the trigger registries, 
**                  not just explicit trigger invocation.
** 018 CA  20170104 Added flag to track window's modal state.
** 019 HC  20170202 Fixed potential race condition between GUI and WindowManager.
**     HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
**     CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
**     HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
**     CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 020 OM  20170320 Caught and reported a ClassCastException.
**     CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
** 021 EVL 20170706 Adding XOR mode and selection box handling for web client.
** 022 EVL 20170825 Adding handling of web client file list dropping message.  Setting up file
**                  size limits.
**     SBI 20170903 Used file channels for uploading large files.
**     EVL 20170905 Changing signature of the hook function to use list of strings.
** 023 HC  20170819 Added extension widget method GET-MOUSE-POSITION.
**                  Implemented runtime support for WINDOW:SHOW-IN-TASKBAR. Made web js client to
**                  notify the java client when desktop (browser window) resized. This makes the
**                  Web GUI driver to report correct desktop dimensions.
** 024 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 025 HC  20171203 Added notification of Web embedded client when program execution enters modal or wait-for
**                  event loop.
** 026 GES 20171207 Silent error mode cannot be used from the client in the new design. Instead,
**                  silent error mode is delegated to the server export when needed. Refactored
**                  the server up-calls to use common code for processing the error handling.
** 027 CA  20171127 Batch the MSG_REGISTER_HOVERABLE_WIDGET and MSG_DEREGISTER_WIDGET messages.
**     CA  20171204 Allow the Web driver to measure text on the client-side, via a AWT font; this
**                  avoids expensive JS calls to measure the text.  For this reason, the text
**                  related APIs were deprecated.
** 028 GES 20180102 Clarified an error handling comment.
** 029 SBI 20180130 Added native implementation for text selection and cursor positioning within
**                  the given text.
** 030 SBI 20180403 Changed message MSG_OPEN_PRINT_OUTPUT to handle document mime type, added
**                  MSG_OPEN_MIME_RESOURCE.
** 031 SBI 20180606 Changed registerHoverableWidget as its "bound" parameter that represents
**                  a registered widget actual bound permits null values.
** 032 SVL 20180718 Getting text metrics in batch now support multiple fonts per request,
**                  sub-widths for a text are no longer returned.
** 033 CA  20180727 Added support for MSG_P2J_APPLY_EVENT, MSG_SET_WINDOW_STATUS_TEXT, 
**                  MSG_ADD_WINDOW_MESSAGE_TEXT and MSG_CLEAR_WINDOW_MESSAGES.
** 034 CA  20180809 If the JS side sends a desktop dimension with either height or width 0, then
**                  do not use it, as downstream code expects a valid dimension.
**     SBI 20180815 Changed MSG_IS_FONT_INSTALLED and MSG_CREATE_FONT.
**     OM  20180815 Added support for drawing ellipse primitives.
**     ECF 20180816 Minor javadoc cleanup.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.security.*;
import java.util.*;

import com.goldencode.p2j.ui.*;

import org.eclipse.jetty.websocket.api.annotations.WebSocket;

import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.InvocationRequestPayload.InvokeType;
import com.goldencode.util.*;

/**
 * This is a WebSocket implementation for GUI which uses a binary protocol to communicate with
 * the javascript (browser-based) desktop simulator.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application is blocked.
 */
@WebSocket
public class GuiWebSocket
extends WebClientProtocol
{
   /** The length of the MD5 hash. */
   private static final int HASH_LENGTH = 16;
   
   /** The maximum cache size. */
   private static final int MAX_DRAW_CACHE_SIZE = 1000;

   /** The magic hash sum indicates that the batch of drawing operations mustn't be cached. */
   private static byte[] MAGIC = new byte[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

   /** The provider for MD5 hashes. */
   private MessageDigest MD5 = null; 
   
   /** Describes the currently pending clipboard request. */
   private ClipboardRequest pending = ClipboardRequest.NONE;
   
   /** Lock for clipboard operations. */
   private Object clipboardLock = new Object();
   
   /** The most recently read contents of the clipboard from the web client. */
   private volatile String clipboardContents = null;
   
   /** 
    * Pending drawing operations (will be cleared and sent as a batch).
    * In cases when a recorded drawing is already cached, it will contain only a 
    * {@link PaintPrimitives#DRAW_CACHED} operation, to let the Web driver draw from cache.
    */
   private List<byte[]> drawingOps = new ArrayList<>();
   
   /** Accumulator to track the total number of pending bytes on the drawing ops list. */
   private int pendingBytes = 0;
   
   /** Pending resize operations counter */
   private int resizeOps = 0;

   /**
    * Accumulator to track the total number of pending bytes on the drawing ops list, including  
    * the operations which were replaced with a hash code (as the resulted image is already 
    * cached).
    */
   private int pendingNoCacheBytes = 0;
   
   /**
    * Pending drawing operations (will be cleared and sent as a batch); this includes all 
    * operations, unaltered. 
    */
   private List<byte[]> drawingNoCacheOps = new ArrayList<>();

   /**
    * The next message ID, used to track message request which require an immediate response
    * from the JS side.
    */
   private int nextMessageId = 0;
   
   /** The MD5 of the last drawing message sent to the JS side. */
   private byte[] lastHash = null;
   
   /** The drawing cache, for an entire window state. */
   private Map<Integer, ExpiryCache<String, byte[]>> fullWindowCache = new HashMap<>();
   
   /**
    * Drawing cache for intermediate widget states.  This is not kept in a per-window map to 
    * allow states to be re-used between windows.
    */
   private ExpiryCache<String, byte[]> perWidgetCache = new LFUAgingCache<>("per-widget", 
                                                                            MAX_DRAW_CACHE_SIZE);

   /** This stack will contain the drawing operation recordings. */
   private Deque<DrawRecording> recordings = new ArrayDeque<>();
   
   /** The true value indicates that batches of drawing commands are cached.*/
   private boolean graphicsCached;
   
   /** The hoverable widget-related messages (register or deregister), to be sent in batch. */
   private List<BinaryMessage> hoverableWidgets = new LinkedList<>();
   
   /**
    * Create a new GUI web socket instance.
    * 
    * @param    lock
    *           Object to use for locking. By passing this in, the caller can share the
    *           locking instance with this code.  Use <code>null</code> if an internal
    *           lock object should be created.
    * @param    callbacks
    *           An instance with hooks to call at specific events.
    * @param    timeout
    *           Web socket timeout.
    * @param    wdtimeout
    *           Watchdog thread timeout.
    * @param    graphicsCached
    *           True if the drawing batches are cached, otherwise false.
    */
   public GuiWebSocket(Object              lock,
                       ClientProtocolHooks callbacks,
                       long                timeout,
                       long                wdtimeout,
                       boolean             graphicsCached)
   {
      super(lock, callbacks, timeout, wdtimeout);
      
      // window ID = 0 means the global cache
      this.perWidgetCache.addCacheExpiryListener(new DrawCacheExpiryListener(0));
      this.graphicsCached = graphicsCached;
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    request
    *           The request data incoming from 4GL.
    * 
    * @return   The response from the customer-specific application, which will be sent back to 
    *           the converted 4GL code.
    */
   public UnformattedPayload remoteCall(UnformattedPayload request)
   {
      String payload = request.payload;
      int msgId = nextMessageId();
      
      // type + msgId + request length + request
      int len = 1 + 4 + 4 + 2 * payload.length();
      BinaryMessage msg = new BinaryMessage(MSG_REMOTE_CALL, len);
      msg.writeInt32(msgId);
      msg.writeInt32(payload.length());
      msg.writeText(payload);
      
      sendBinaryMessage(msg.message);
      
      Object res = waitForResult(msgId);
      
      return (UnformattedPayload) res;
   }
   
   /**
    * Add message texts of specified window.
    * 
    * @param   windowId
    *          Window ID.
    * @param   texts
    *          Text message.
    */
   public void addWindowMessageText(int windowId, String[] texts)
   {
      int len = 1 + 4 + 4;
      for (String msg : texts)
      {
         len = len + 4 + 2 * (msg == null ? 0 : msg.length());
      }
      
      BinaryMessage msg =  new BinaryMessage(MSG_ADD_WINDOW_MESSAGE_TEXT, len);
      msg.writeInt32(windowId);
      msg.writeInt32(texts.length);
      for (String txt : texts)
      {
         msg.writeInt32(txt == null ? 0 : txt.length());
         msg.writeText(txt);
      }

      sendBinaryMessage(msg.message);
   }
   
   /**
    * Clear all message texts of specified window.
    * 
    * @param   windowId
    *          Window ID.
    */
   public void clearWindowMessages(int windowId)
   {
      int len = 1 + 4;
      BinaryMessage msg =  new BinaryMessage(MSG_CLEAR_WINDOW_MESSAGES, len);
      msg.writeInt32(windowId);

      sendBinaryMessage(msg.message);
   }
   
   /**
    * Set status text message of specified window.
    * 
    * @param   windowId
    *          Window ID.
    * @param   text
    *          Text message.
    */
   public void setWindowStatusText(int windowId, String text)
   {
      int len = 1 + 4 + 4 + 2 * (text == null ? 0 : text.length());
      BinaryMessage msg =  new BinaryMessage(MSG_SET_WINDOW_STATUS_TEXT, len);
      msg.writeInt32(windowId);
      msg.writeInt32(text == null ? 0 : text.length());
      msg.writeText(text);

      sendBinaryMessage(msg.message);
   }

   /**
    * Send a PUBLISH notification to a remote application. 
    * 
    * @param    subId
    *           The subscriber's ID, known by the remote application.
    * @param    event
    *           The published event name.
    * @param    param
    *           The parameters for this event.
    */
   public void publish(long subId, String event, String[] param)
   {
      int msgId = nextMessageId();
      String hexRes = Long.toHexString(subId);
      
      // type + msgId + (hexRes.len + hexres.len * 2) + (event.length + 2 * event.length) +
      // param.length + sum(param[i].length + 2 * param[i].length)
      int len = 1 + 4 + (1 + hexRes.length() * 2) + (1 + 2 * event.length()) + 1;
      for (String p : param)
      {
         len = len + 2 + 2 * (p == null ? 0 : p.length());
      }
      
      BinaryMessage msg = new BinaryMessage(MSG_P2J_PUBLISH, len);
      
      msg.writeInt32(msgId);
      msg.writeByte((byte) hexRes.length());
      msg.writeText(hexRes);
      msg.writeByte((byte) event.length());
      msg.writeText(event);
      msg.writeByte((byte) param.length);
      for (String p : param)
      {
         msg.writeInt16(p == null ? 0 : p.length());
         if (p != null)
         {
            msg.writeText(p);
         }
      }
      
      sendBinaryMessage(msg.message);
   }
   
   /**
    * Create a new top-level window (which will be backed by a canvas instance in the client).
    *
    * @param    id
    *           The window id to associate with the newly created window.
    * @param    showInTaskbar
    *           The flag indicates whether the window should be shown in task bar.
    */
   public void createWindow(int id, boolean showInTaskbar)
   {
      int len = 1 + 4 + 1;
      BinaryMessage msg = new BinaryMessage(MSG_CREATE_WINDOW, len);
      msg.writeInt32(id);
      msg.writeBoolean(showInTaskbar);
      msg.send();
   }
   
   /**
    * Create a new child window (which will be backed by a canvas instance in the client).
    *
    * @param    id
    *           The window id to associate with the newly created window.
    * @param    owner
    *           The window id to associate with the owning window.
    * @param    title
    *           Taskbar title for the child window.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    */
   public void createChildWindow(int id, int owner, String title, boolean resizable, 
                                 boolean isShareActivationWithOwner, boolean modal)
   {
      // the title can be null for overlay windows
      // type + modal + resizable + overlay + wid + owner + title  
      int len = 1 + 1 + 1 + 1 + 4 + 4 + 2 * (title != null ? title.length() : 0);
      BinaryMessage msg = new BinaryMessage(MSG_CREATE_CHILD_WINDOW, len);

      msg.writeByte(modal ? (byte) 1 : 0);
      msg.writeByte(resizable ? (byte) 1 : 0);
      msg.writeByte(isShareActivationWithOwner ? (byte) 1 : 0);
      msg.writeInt32(id);
      msg.writeInt32(owner);
      msg.writeText(title);
      msg.send();
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    *
    * @param    id
    *           The window to modify.
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(int id, boolean visible)
   {
      byte[] message = new byte[6];
      
      message[0] = MSG_WINDOW_VISIBILITY;
      writeMessageInt32(message, 1, id);
      message[5] = (byte)(visible ? 1 : 0);
      
      sendBinaryMessage(message);
   }
   
   /**
    * Set the window's sensitive state, depending on the {@code enable} flag.
    *
    * @param    id
    *           The window to modify.
    * @param    enable
    *           The window sensitive state.
    */
   public void setWindowEnabled(int id, boolean enable)
   {
      byte[] message = new byte[6];
      
      message[0] = MSG_WINDOW_SENSITIVITY;
      writeMessageInt32(message, 1, id);
      message[5] = (byte)(enable ? 1 : 0);
      
      sendBinaryMessage(message);
   }

   /**
    * Draw the caret that is located within the given text content by its column and row offsets.
    * 
    * @param    text
    *           The editing text
    * @param    columnOffset
    *           The horizontal cursor offset within the editing text content in character units
    * @param    rowOffset
    *           The vertical cursor offset within the editing text content in character units
    * @param    columnScrollOffset
    *           The horizontal scroll offset within the editing text content in character units
    * @param    rowScrollOffset
    *           The vertical scroll offset within the editing text content in character units
    * @param    rowHeight
    *           The given row height
    */
   public void startDrawCursor(String text,
                               int columnOffset,
                               int rowOffset,
                               int columnScrollOffset,
                               int rowScrollOffset,
                               int rowHeight)
   {
      int length = text.length() * 2 + 25; 
      byte[] message = allocateDrawingOp(PaintPrimitives.START_DRAW_CURSOR, length);
      
      int offset = 1;
      writeMessageInt32(message, offset, text.length());
      offset += 4;
      
      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, columnOffset);
      offset += 4;
      
      writeMessageInt32(message, offset, rowOffset);
      offset += 4;
      
      writeMessageInt32(message, offset, columnScrollOffset);
      offset += 4;
      
      writeMessageInt32(message, offset, rowScrollOffset);
      offset += 4;
      
      writeMessageInt32(message, offset, rowHeight);
   }

   /**
    * Complete the caret drawing.
    */
   public void endDrawCursor()
   {
      allocateDrawingOp(PaintPrimitives.END_DRAW_CURSOR, 1);
   }
   
   /**
    * Shift the current origin along X axis to the native length of the given text.
    * 
    * @param    text
    *           The given text
    */
   public void textShift(String text)
   {
      int textLength = text.length(); 
      byte[] message = allocateDrawingOp(PaintPrimitives.TEXT_SHIFT, textLength * 2 + 5);
      
      writeMessageInt32(message, 1, textLength);
      
      writeMessageText(message, 5, text);
   }

   /**
    * Undo all previous text shifts.
    */
   public void undoTextShift()
   {
      allocateDrawingOp(PaintPrimitives.UNDO_TEXT_SHIFT, 1);
   }
   
   /**
    * Destroy the specified window at the client side.  All resources will be cleaned up after
    * this call.
    *
    * @param    id
    *           The window id to destroy.
    * @param    cachedImages
    *           The array of client cached images to remove.
    */
   public void destroyWindow(int id, Integer[] cachedImages)
   {
      byte[] message = new byte[9 + cachedImages.length * 4];
      
      message[0] = MSG_DESTROY_WINDOW;
      writeMessageInt32(message, 1, id);
      writeMessageInt32(message, 5, cachedImages.length);
      int idx = 9;
      for(Integer imageHash : cachedImages)
      {
         writeMessageInt32(message, idx, imageHash);
         idx += 4;
      }
      sendBinaryMessage(message);
      
      // remove the cache for this window
      fullWindowCache.remove(id);
   }
   
   /**
    * Draw a line in the current color/stroke from the given starting and ending points.
    * 
    * @param    x1
    *           The starting X coordinate in pixels.
    * @param    y1
    *           The starting Y coordinate in pixels.
    * @param    x2
    *           The ending X coordinate in pixels.
    * @param    y2
    *           The ending Y coordinate in pixels.
    */
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      allocateAndSend(PaintPrimitives.DRAW_LINE, x1, y1, x2, y2);
   }
   
   /**
    * Draw a rectangle in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    */
   public void drawRect(int x, int y, int width, int height)
   {
      allocateAndSend(PaintPrimitives.DRAW_RECT, x, y, width, height);
   }
   
   /**
    * Draw a rectangle with the area filled by the current color.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    */
   public void fillRect(int x, int y, int width, int height)
   {
      allocateAndSend(PaintPrimitives.FILL_RECT, x, y, width, height);
   }
   
   /**
    * Draw an ellipse in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The ellipse width in pixels.
    * @param    height
    *           The ellipse height in pixels.
    */
   public void drawEllipse(int x, int y, int width, int height)
   {
      allocateAndSend(PaintPrimitives.DRAW_ELLIPSE, x, y, width, height);
   }
   
   /**
    * Draw an ellipse with the area filled by the current color.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The ellipse width in pixels.
    * @param    height
    *           The ellipse height in pixels.
    */
   public void fillEllipse(int x, int y, int width, int height)
   {
      allocateAndSend(PaintPrimitives.FILL_ELLIPSE, x, y, width, height);
   }
   
   /**
    * Draw a rectangle around the given text filled with the current color.
    * 
    * @param    x
    *           The top left X coordinate in pixel.
    * @param    y
    *           The top left Y coordinate in pixels
    * @param    textWidthHint
    *           The text width hint in pixels
    * @param    textHeightHint
    *           The text height hint in pixels
    * @param    text
    *           The given text
    */
   public void fillRectAroundText(int x, int y, int textWidthHint, int textHeightHint, String text)
   {
      int textLength = text.length();
      byte[] message = allocateDrawingOp(PaintPrimitives.FILL_RECT_AROUND_TEXT, 2 * textLength + 21);
      
      writeMessageInt32(message, 1, x);
      writeMessageInt32(message, 5, y);
      writeMessageInt32(message, 9, textHeightHint);
      writeMessageInt32(message, 13, textHeightHint);
      writeMessageInt32(message, 17, textLength);
      writeMessageText(message, 21, text);
   }
   
   /**
    * Draw a rounded stroked rectangle in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    * @param    diameter
    *           Diameter of the arc that defines the rounded corner.
    */
   public void drawRoundRect(int x, int y, int width, int height, int diameter)
   {
      allocateAndSend(PaintPrimitives.DRAW_ROUND_RECT, x, y, width, height, diameter);
   }
   
   /**
    * Draw a rounded filled rectangle in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    * @param    diameter
    *           Diameter of the arc that defines the rounded corner.
    */
   public void fillRoundRect(int x, int y, int width, int height, int diameter)
   {
      allocateAndSend(PaintPrimitives.FILL_ROUND_RECT, x, y, width, height, diameter);
   }
   
   /**
    * Draw a 3D rectangle in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    * @param    raised
    *           <code>true</code> if the rendering should appear raised.
    */
   public void draw3DRect(int x, int y, int width, int height, boolean raised)
   {
      allocateAndSend(PaintPrimitives.DRAW_3D_RECT, x, y, width, height, raised);
   }
   
   /**
    * Draw a 3D filled rectangle in the current color/stroke.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    * @param    raised
    *           <code>true</code> if the rendering should appear raised.
    */
   public void fill3DRect(int x, int y, int width, int height, boolean raised)
   {
      allocateAndSend(PaintPrimitives.FILL_3D_RECT, x, y, width, height, raised);
   }

   /**
    * Draw a selection box rectangular area with current color.
    * 
    * @param    x
    *           The top left X coordinate in pixels.
    * @param    y
    *           The top left Y coordinate in pixels.
    * @param    width
    *           The rectangle width in pixels.
    * @param    height
    *           The rectangle height in pixels.
    * @param    clear
    *           <code>TRUE</code> means clean up area instead of drawing selections.
    */
   public void drawSelection(int x, int y, int width, int height, boolean clear)
   {
      allocateAndSend(PaintPrimitives.DRAW_SELECTION, x, y, width, height, clear);
   }
   
   /**
    * Reset selection boxes painting mode and return to normal drawing mode.
    */
   public void resetSelection()
   {
      allocateAndSend(PaintPrimitives.RESET_SELECTION);
   }
   
   /**
    * Draw a given image on the virtual screen and prepare it to transfer.
    * 
    * @param    x
    *           The x-coordinate of the image position.
    * @param    y
    *           The y-coordinate of the image position.
    * @param    width
    *           The width of the image.
    * @param    height
    *           The height of the image.
    * @param    encoding
    *           The encoding id.
    * @param    imageHash
    *           The image hash code, its own id.
    * @param    encodedImage
    *           The encoded image represented by bytes.
    */
   public void drawImage(int x, int y, int width, int height, ImageEncoding encoding,
            int imageHash, byte[] encodedImage)
   {
      int msgLen = 22;
      if (encoding == ImageEncoding.RAW)
      {
         msgLen += encodedImage.length;
      }
      byte[] drawImageMsg = new byte[msgLen];
      drawImageMsg[0] = (byte) (PaintPrimitives.DRAW_IMAGE.ordinal());
      writeMessageInt32(drawImageMsg, 1, x);
      writeMessageInt32(drawImageMsg, 5, y);
      writeMessageInt32(drawImageMsg, 9, width);
      writeMessageInt32(drawImageMsg, 13, height);
      drawImageMsg[17] = (byte) encoding.getValue();
      writeMessageInt32(drawImageMsg, 18, imageHash);
      if (encoding == ImageEncoding.RAW)
      {
         System.arraycopy(encodedImage, 0, drawImageMsg, 22, encodedImage.length);
      }

      addDrawingOp(drawImageMsg);
   }

   /**
    * Draw a filled, closed polygon in the current color/stroke.
    * 
    * @param    xPoints
    *           Array of X coordinates.
    * @param    yPoints
    *           Array of Y coordinates.
    * @param    num
    *           The number of points defining vertices.
    */
   public void fillPolygon(int[] xPoints, int[] yPoints, int num)
   {
      byte[] message = allocateDrawingOp(PaintPrimitives.FILL_POLYGON, (num * 4 * 2) + 5);
      
      // write out the number of points
      writeMessageInt32(message, 1, num);
      
      int idx = 5;
      
      for (int i = 0; i < num; i++)
      {
         writeMessageInt32(message, idx, xPoints[i]);
         idx += 4;
         writeMessageInt32(message, idx, yPoints[i]);
         idx += 4;
      }
   }
   
   /**
    * Translate the graphics context by the given values. 
    * 
    * @param   push
    *          <code>true</code> for a PUSH operation, <code>false</code> for POP.
    * @param   x
    *          Change in X coordinate.
    * @param   y
    *          Change in Y coordinate.
    */
   public void translate(boolean push, int x, int y)
   {
      PaintPrimitives type = push ? PaintPrimitives.TRANSLATE_PUSH
                                  : PaintPrimitives.TRANSLATE_POP;
                                  
      allocateAndSend(type, x, y);
   }
   
   /**
    * Add the given rectangles to the current clipping region.
    *
    * @param   clip
    *          The clipping rectangles.
    */
   public void clip(Set<NativeRectangle> clip)
   {
      if (clip == null || clip.isEmpty())
      {
         return;
      }

      int clipCount = clip.size();
      byte[] message = allocateDrawingOp(PaintPrimitives.CLIP, 5 + clipCount * 16);

      // clip count
      writeMessageInt32(message, 1, clipCount);

      int i = 0;
      for (NativeRectangle nr : clip)
      {
         int off = i * 16 + 5;
         writeMessageInt32(message, off, nr.left());
         writeMessageInt32(message, off + 4, nr.top());
         writeMessageInt32(message, off + 8, nr.width());
         writeMessageInt32(message, off + 12, nr.height());
         i++;
      }
   }
   
   /**
    * Remove the most recently added rectangle from the current clipping region.
    * The javascript code maintains a stack of state that allows the matching
    * rectangle to be determined.
    */
   public void unclip()
   {
      allocateAndSend(PaintPrimitives.NO_CLIP);
   }
   
   /**
    * Set the given color as the current color for graphics operations.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    */
   public void setColor(int red, int green, int blue)
   {
      allocateAndSend(PaintPrimitives.SET_COLOR, red, green, blue);
   }
   
   /**
    * Set the desktop background color.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    */
   public void setDesktopBgColor(int red, int green, int blue)
   {
      allocateAndSend(PaintPrimitives.SET_DESKTOP_BGCOLOR, red, green, blue);
   }

   /**
    * Set "XOR composite" mode: color of the drawn elements is calculated as XOR (for each RGB
    * component) of their color with color of underlying pixels.
    */
   public void setXORComposite()
   {
      allocateAndSend(PaintPrimitives.SET_XOR_COMPOSITE);
   }

   /**
    * Reset composite mode and return to normal drawing mode.
    */
   public void resetComposite()
   {
      allocateAndSend(PaintPrimitives.RESET_COMPOSITE);
   }
   
   /**
    * Set XOR mode painting allowing to invert the destination color with current drawing
    * primitive color.  Double painting of the same figure at the same location should restore
    * original picture.
    * 
    * @param   red
    *          Red XOR alternation color value.
    * @param   green
    *          Green XOR alternation color value.
    * @param   blue
    *          Blue XOR alternation color value.
    */
   public void setXORMode(int red, int green, int blue)
   {
      allocateAndSend(PaintPrimitives.SET_XOR_MODE, red, green, blue);
   }
   
   /**
    * Resize the window.
    * 
    * @param    width
    *           The window width in pixels.
    * @param    height
    *           The window height in pixels.
    */
   public void resizeWindow(int width, int height)
   {
      allocateAndSend(PaintPrimitives.RESIZE_WINDOW, width, height);
      
      resizeOps++;
   }
   
   /**
    * Set the window location in X and Y coordinates.
    * 
    * @param    x
    *           The X coordinate in pixels.
    * @param    y
    *           The Y coordinate in pixels.
    */
   public void setWindowLocation(int x, int y)
   {
      allocateAndSend(PaintPrimitives.SET_LOCATION, x, y);
      
      resizeOps++;
   }
   
   /**
    * Set the window bounds.
    * 
    * @param    x
    *           The X coordinate in pixels.
    * @param    y
    *           The Y coordinate in pixels.
    * @param    width
    *           The window width in pixels.
    * @param    height
    *           The window height in pixels.
    */
   public void setWindowBounds(int x, int y, int width, int height)
   {
      allocateAndSend(PaintPrimitives.SET_WINDOW_BOUNDS, x, y, width, height);
      
      resizeOps++;
   }
   
   /**
    * Set the window details for resize; this includes the minimum and maximun dimensions, and 
    * also the window's resizeable state.
    * 
    * @param    wid
    *           The window ID.
    * @param    resize
    *           Flag indicating if the window can be resized or not.
    * @param    minWidth
    *           The window's minimum width.
    * @param    minHeight
    *           The window's minimum height.
    * @param    maxWidth
    *           The window's maximum width.
    * @param    maxHeight
    *           The window's maximum height.
    */
   public void setResizeableWindow(int     wid,
                                   boolean resize,
                                   int     minWidth, 
                                   int     minHeight, 
                                   int     maxWidth, 
                                   int     maxHeight)
   {
      byte[] message = new byte[1 + 4 + 1 + (resize ? 4 * 2 : 0)];
      
      int offset = 0;
      message[0] = SET_RESIZEABLE_WINDOW;
      offset += 1;
      
      writeMessageInt32(message, offset, wid);
      offset += 4;
      
      message[offset] = (byte) (resize ? 1 : 0);
      offset += 1;

      if (resize)
      {
         writeMessageInt16(message, offset, minWidth);
         offset += 2;
         writeMessageInt16(message, offset, minHeight);
         offset += 2;
   
         writeMessageInt16(message, offset, maxWidth);
         offset += 2;
         writeMessageInt16(message, offset, maxHeight);
         offset += 2;
      }
      
      sendBinaryMessage(message);
   }
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   public void restackWindows(int[] windowIds)
   {
      byte[] message = new byte[5 + (windowIds.length * 4)];
      
      message[0] = MSG_RESTACK_WINDOWS;
      
      // write out the number of windows
      writeMessageInt32(message, 1, windowIds.length);
      
      int idx = 5;
      
      for (int i = 0; i < windowIds.length; i++)
      {
         writeMessageInt32(message, idx, windowIds[i]);
         idx += 4;
      }
      
      sendBinaryMessage(message);
   }
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window
    *           and its state changed to active.
    */
   public void moveToTop(int windowId, boolean focus)
   {
      byte[] message = new byte[6];

      message[0] = MSG_MOVE_TO_TOP;
      writeMessageInt32(message, 1, windowId);
      message[5] = (byte) (focus ? 1 : 0);

      sendBinaryMessage(message);

   }
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    */
   public void moveToBottom(int windowId)
   {
      sendBinaryMessage(MSG_MOVE_TO_BOTTOM, windowId);
   }
   
   /**
    * Enable or disable capture of mouse events.
    *  
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    */
   public void captureMouseEvents(boolean state)
   {
      sendBinaryMessage(CAPTURE_MOUSE, (byte) (state ? 1 : 0));
   }

   /**
    * Enable or disable processing of OS events.
    *  
    * @param    windowIds
    *           The window IDs.
    * @param    state
    *           {@code true} to process OS events, {@code false} to ignore them.
    */
   public void enableOsEvents(int windowIds[], boolean state)
   {
      byte[] message = new byte[1 + 4 + 4 * windowIds.length + 1];
      int offset = 0;
      
      message[0] = ENABLE_OS_EVENTS;
      offset++;
      
      writeMessageInt32(message, offset, windowIds.length);
      offset += 4;
      
      for (int wid : windowIds)
      {
         writeMessageInt32(message, offset, wid);
         offset += 4;
      }
      
      message[offset] = (byte) (state ? 1 : 0);
      offset++;
      
      sendBinaryMessage(message);
   }

   /**
    * Set the specified window's iconification state.
    *  
    * @param    windowId
    *           The window ID.
    * @param    state
    *           {@code true} to iconify the window, {@code false} to de-iconify it.
    */
   public void setIconificationState(int windowId, boolean state)
   {
      byte[] message = new byte[1 + 4 + 1];
      int offset = 0;
      
      message[offset] = SET_ICONIFICATION_STATE;
      offset = offset + 1;
      
      writeMessageInt32(message, offset, windowId);
      offset = offset + 4;
      
      message[offset] = (byte) (state ? 1 : 0);
      offset = offset + 1;
      
      sendBinaryMessage(message);
   }
   
   /**
    * Allocate and send message, with no subsequent data
    *
    * @param    type
    *           Message type.
    */
   public void allocateAndSend(PaintPrimitives type)
   {
      allocateDrawingOp(type, 1);
   }
   
   /**
    * Allocate and send message with 1 integer.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           Only integer to send.
    */
   public void allocateAndSend(PaintPrimitives type, int one)
   {
      byte[] message = allocateDrawingOp(type, 5);
      
      writeMessageInt32(message, 1, one);
   }
   
   /**
    * Allocate and send message with 2 integers.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           First integer to send.
    * @param    two
    *           Second integer to send.
    */
   public void allocateAndSend(PaintPrimitives type, int one, int two)
   {
      byte[] message = allocateDrawingOp(type, 9);
      
      writeMessageInt32(message, 1, one);
      writeMessageInt32(message, 5, two);
   }
   
   /**
    * Allocate and send message with 3 integers.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           First integer to send.
    * @param    two
    *           Second integer to send.
    * @param    three
    *           Third integer to send.
    */
   public void allocateAndSend(PaintPrimitives type, int one, int two, int three)
   {
      byte[] message = allocateDrawingOp(type, 13);
      
      writeMessageInt32(message, 1, one);
      writeMessageInt32(message, 5, two);
      writeMessageInt32(message, 9, three);
   }
   
   /**
    * Allocate and send message with 4 integers.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           First integer to send.
    * @param    two
    *           Second integer to send.
    * @param    three
    *           Third integer to send.
    * @param    four
    *           Fourth integer to send.
    */
   public void allocateAndSend(PaintPrimitives type, int one, int two, int three, int four)
   {
      byte[] message = allocateDrawingOp(type, 17);
      
      writeMessageInt32(message, 1, one);
      writeMessageInt32(message, 5, two);
      writeMessageInt32(message, 9, three);
      writeMessageInt32(message, 13, four);
   }
   
   /**
    * Allocate and send message with 5 integers.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           First integer to send.
    * @param    two
    *           Second integer to send.
    * @param    three
    *           Third integer to send.
    * @param    four
    *           Fourth integer to send.
    * @param    five
    *           Fifth integer to send.
    */
   public void allocateAndSend(PaintPrimitives type,
                               int             one,
                               int             two,
                               int             three,
                               int             four,
                               int             five)
   {
      byte[] message = allocateDrawingOp(type, 21);
      
      writeMessageInt32(message, 1, one);
      writeMessageInt32(message, 5, two);
      writeMessageInt32(message, 9, three);
      writeMessageInt32(message, 13, four);
      writeMessageInt32(message, 17, five);
   }
   
   /**
    * Allocate and send message with 4 integers and a boolean.
    *
    * @param    type
    *           Message type.
    * @param    one
    *           First integer to send.
    * @param    two
    *           Second integer to send.
    * @param    three
    *           Third integer to send.
    * @param    four
    *           Fourth integer to send.
    * @param    flag
    *           Boolean value to send.
    */
   public void allocateAndSend(PaintPrimitives type,
                               int             one,
                               int             two,
                               int             three,
                               int             four,
                               boolean         flag)
   {
      byte[] message = allocateDrawingOp(type, 18);
      
      writeMessageInt32(message, 1, one);
      writeMessageInt32(message, 5, two);
      writeMessageInt32(message, 9, three);
      writeMessageInt32(message, 13, four);
      
      message[17] = flag ? (byte) 1 : (byte) 0;
   }
   
   /**
    * Set the window title at the OS level.
    * 
    * @param    title
    *           The window's title.
    */
   public void setWindowTitle(String title)
   {
      byte[] message = allocateDrawingOp(PaintPrimitives.SET_TITLE, title.length() * 2 + 5);
      writeMessageInt32(message, 1, title.length());
      writeMessageText(message, 5, title);
   }
   
   /**
    * Sets the window icon. If an image encoding is HASH, then the binary image isn't sent to
    * the client.
    * 
    * @param    width
    *           The width of the image.
    * @param    height
    *           The height of the image.
    * @param    encoding
    *           The encoding id.
    * @param    imageHash
    *           The image hash code, its own id.
    * @param    encodedImage
    *           The encoded image represented by bytes.
    */
   public void setIconImage(int width, int height, ImageEncoding encoding, int imageHash, byte[] encodedImage)
   {
      int msgLen = 14;
      if (encoding == ImageEncoding.RAW)
      {
         msgLen += encodedImage.length;
      }

      byte[] setIconMsg = new byte[msgLen];
      setIconMsg[0] = (byte) (PaintPrimitives.SET_ICON.ordinal());
      writeMessageInt32(setIconMsg, 1, width);
      writeMessageInt32(setIconMsg, 5, height);
      setIconMsg[9] = (byte) encoding.getValue();
      writeMessageInt32(setIconMsg, 10, imageHash);
      if (encoding == ImageEncoding.RAW)
      {
         System.arraycopy(encodedImage, 0, setIconMsg, 14, encodedImage.length);
      }

      addDrawingOp(setIconMsg);
   }

   /**
    * Send all pending drawing operations to the javascript side and clear the list of pending
    * operations.
    *
    * @param    windowId
    *           The window which is being drawn.
    */
   public void sendPendingDrawingOps(int windowId)
   {
      if (drawingOps.isEmpty() && hoverableWidgets.isEmpty())
      {
         // nothing to send
         return;
      }

      // send the hoverableWidgets
      if (!hoverableWidgets.isEmpty())
      {
         ByteArrayOutputStream baos = new ByteArrayOutputStream(64 * hoverableWidgets.size());
         for (BinaryMessage msg : hoverableWidgets)
         {
            try
            {
               baos.write(msg.message);
            }
            catch (IOException e)
            {
               // ignore
            }
         }
         
         sendBinaryMessage(baos.toByteArray());
      }

      byte[] total = buildDrawingMessage(windowId, drawingOps, pendingBytes);

      // before sending it, determine if is the same as the last drawing operation sent; if so, 
      // skip it
      boolean doSend = true;

      if (graphicsCached)
      {
         byte[] drawHash = md5Hash(buildDrawingMessage(windowId, 
                                                       drawingNoCacheOps,
                                                       pendingNoCacheBytes));
         if (drawHash != null)
         {
            if (resizeOps > 0)
            {
               System.arraycopy(MAGIC, 0, total, 1, HASH_LENGTH);
            }
            else if (Arrays.equals(drawHash, lastHash))
            {
               // do not send if duplicated message
               doSend = false;
            }
            else
            {
               // set the hash
               System.arraycopy(drawHash, 0, total, 1, drawHash.length);
   
               ExpiryCache<String, byte[]> windowCache = fullWindowCache.get(windowId);
               if (windowCache == null)
               {
                  windowCache = new LFUAgingCache<>("window " + windowId, MAX_DRAW_CACHE_SIZE);
                  fullWindowCache.put(windowId, windowCache);
                  windowCache.addCacheExpiryListener(new DrawCacheExpiryListener(windowId));
               }
   
               String hash = XmlHelper.convertBytesToHex(drawHash);
               if (windowCache.get(hash) != null)
               {
                  // the hash is already known by the JS side; send only the:
                  // type (DRAW) (1 byte)
                  // hash (16 bytes)
                  // window ID (4 bytes)
                  
                  total = Arrays.copyOf(total, 1 + HASH_LENGTH + 4);
               }
               else
               {
                  // cache it and let the JS side cache the draw, too
                  windowCache.put(hash, drawHash);
               }
            }
            lastHash = drawHash;
         }
      }
      
      if (doSend)
      {
         // send it to the javascript side
         sendBinaryMessage(total);
      }
      
      // clear the batch
      hoverableWidgets.clear();
      drawingOps.clear();
      pendingBytes = 0;
      drawingNoCacheOps.clear();
      pendingNoCacheBytes = 0;
      resizeOps = 0;
   }
   
   /**
    * Send a request to the web client to read the current contents of the clipboard and send
    * that back as text.  This request will block until the client sends back the text.
    *
    * @return   The text read from the web client's clipboard.
    */
   public String readClipboard()
   {
      try
      {
         String contents = "";
         
         synchronized (clipboardLock)
         {
            if (pending == ClipboardRequest.SERVER_GENERATED)
            {
               throw new IllegalStateException("readClipboard() is not reentrant!");
            }
            
            if (pending == ClipboardRequest.CLIENT_GENERATED)
            {
               contents = clipboardContents != null ? clipboardContents : "";
            }
            
            // clear old contents
            clipboardContents = null;
            
            if (pending == ClipboardRequest.CLIENT_GENERATED)
            {
               return contents;
            }
            else
            {
               pending = ClipboardRequest.SERVER_GENERATED;
            }
         }
         
         // request the contents be refreshed from the client
         sendBinaryMessage(MSG_READ_CLIPBOARD);
         
         synchronized (clipboardLock)
         {
            // TODO: check a termination flag here
            while (clipboardContents == null)
            {
               try
               {
                  // TODO: should we wait with a timeout here? if not then we may need a
                  //       mechanism to notify waiters when we have disconnected or quit
                  //       asynchronously
                  clipboardLock.wait();
               }
               catch (InterruptedException ie)
               {
                  // ignore spurious interrupts
               }
            }
            
            if (clipboardContents != null)
               contents = clipboardContents;
         }
         
         return contents;
      }
      
      finally
      {
         synchronized (clipboardLock)
         {
            pending = ClipboardRequest.NONE;
            clipboardContents = null;
         }
      }
   }
   
   /**
    * Handles messages of MSG_FILE_UPLOADING type for large messages. The messages that don't
    * exceed the maximal buffer size are still handled by processBinaryMessage directly.
    * 
    * @param    channel
    *           The given channel
    * 
    * @return   The true value if this message is handled here, otherwise false.
    */
   @Override
   protected boolean processChannel(FileChannel channel)
   {
      try
      {
         ByteBuffer dst = ByteBuffer.allocate(1);
         //sets the position before reading its content
         channel.position(0);
         channel.read(dst);
         dst.flip();
         
         byte messageType = dst.get();
         
         if (messageType == MSG_FILE_UPLOADING)
         {
            ByteBuffer nameLengthBuffer = ByteBuffer.allocate(2);
            channel.read(nameLengthBuffer);
            nameLengthBuffer.flip();
            // file name length
            int nameLength = readMessageInt16(nameLengthBuffer.array(), 0);
            
            ByteBuffer fileNameBuffer = ByteBuffer.allocate(2 * nameLength);
            channel.read(fileNameBuffer);
            fileNameBuffer.flip();
            // file name
            String fileName = readMessageText(fileNameBuffer.array(), 0, 2 * nameLength);
            
            ByteBuffer fileSizeBuffer = ByteBuffer.allocate(4);
            channel.read(fileSizeBuffer);
            fileSizeBuffer.flip();
            
            // file size
            int fileSize = readMessageInt32(fileSizeBuffer.array(), 0);
            
            // ask java to store the file locally in temp directory
            this.callbacks.saveDropTargetFile(fileName, channel, fileSize);
            
            return true;
         }
      }
      catch(IOException ex)
      {
      }
      
      return false;
   }
   
   /**
    * Copies the given text to the application clipboard and notifies the JS client
    * about the current clipboards content.
    * 
    * @param    text
    *           The current clipboard content.
    */
   public void writeClipboard(String text)
   {
      synchronized (clipboardLock)
      {
         clipboardContents = text;
         clipboardLock.notify();
      }
      sendBinaryMessage(MSG_WRITE_CLIPBOARD, (text == null) ? "" : text);
   }

   /**
    * This is the worker method which is called when a binary message is received.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Message offset..
    * @param    length
    *           Message size.
    *
    * @return   <code>true</code> if the message has been handled, <code>false</code> otherwise.
    */
   public boolean processBinaryMessage(byte[] message, int offset, int length)
   {
      // try common (shared with ChUI) processing first
      boolean handled = super.processBinaryMessage(message, offset, length);
      
      if (handled)
      {
         return true;
      }
      
      // now try GUI processing if the message was not handled by common processing
      if (message[offset] == MSG_CLIPBOARD_CONTENTS)
      {
         synchronized (clipboardLock)
         {
            clipboardContents = readMessageText(message, offset + 1, length);
            clipboardLock.notify();
         }
         
         handled = true;
      }
      else if (length == 1 && message[offset] == MSG_CLIPBOARD_PREPARE)
      {
         String text = UiUtils.getCurrentFocus().getSelectedText();
         sendBinaryMessage(MSG_WRITE_CLIPBOARD, (text == null) ? "" : text);
         handled = true;
      }
      else if (length == 7 && message[offset] == MSG_SENT_PAR_HEIGHT)
      {
         int msgId = readMessageInt32(message, offset + 1);
         int msg = readMessageInt16(message, offset + 5);
         
         receiveResult(msgId, msg);
         handled = true;
      }
      else if (length == 6 && message[offset] == MSG_SENT_TEXT_HEIGHT)
      {
         int msgId = readMessageInt32(message, offset + 1);
         int msg = message[offset + 5];
         
         receiveResult(msgId, msg);
         handled = true;
      }
      else if (length == 7 && message[offset] == MSG_SENT_TEXT_WIDTH)
      {
         int msgId = readMessageInt32(message, offset + 1);
         int msg = readMessageInt16(message, offset + 5);
         
         receiveResult(msgId, msg);
         handled = true;
      }
      else if (length >= 9 && message[offset] == MSG_SENT_TEXT_WIDTHS)
      {
         offset = offset + 1;

         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int no = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int[] widths = new int[no];
         for (int i = 0; i < widths.length; i++)
         {
            widths[i] = readMessageInt16(message, offset);
            offset = offset + 2;
         }
         
         receiveResult(msgId, widths);
         handled = true;
      }
      else if (length >= 7 && message[offset] == MSG_SENT_TEXT_WIDTHS_BATCH)
      {
         offset = offset + 1;

         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int arraySize = readMessageInt16(message, offset);
         offset = offset + 2;
         
         int[][] res = new int[arraySize][];
         
         for (int i = 0; i < arraySize; i++)
         {
            int subArraySize = readMessageInt16(message, offset);
            offset = offset + 2;

            res[i] = new int[subArraySize];
            for (int j = 0; j < subArraySize; j++)
            {
               res[i][j] = readMessageInt16(message, offset);
               offset = offset + 2;
            }
         }
         
         receiveResult(msgId, res);
         handled = true;
      }
      else if (length == 6 && message[offset] == MSG_SENT_FONT_HEIGHT)
      {
         int msgId = readMessageInt32(message, offset + 1);
         int msg = message[offset + 5];
         
         receiveResult(msgId, msg);
         handled = true;
      }
      else if (length > 5 && message[offset] == MSG_SENT_FONT_WIDTHS)
      {
         offset = offset + 1;
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // the length fits 2 bytes
         int wlength = message[offset];
         offset = offset + 1;
         wlength = (message[offset] << 8) | wlength;
         offset = offset + 1;

         int[] widths = new int[wlength];
         for (int i = 0; i < wlength; i++)
         {
            // each width can fit in a byte, but the API expects a int[] array
            widths[i] = message[offset];
            offset = offset + 1;
         }
         
         receiveResult(msgId, widths);
         handled = true;
      }
      else if (length == 9 && message[offset] == MSG_DONE_CREATE_FONT)
      {
         int msgId = readMessageInt32(message, offset + 1);

         int msg = readMessageInt32(message, offset + 5);

         receiveResult(msgId, msg);
         handled = true;
      }
      else if (length == 6 && message[offset] == MSG_DONE_DERIVE_FONT)
      {
         int msgId = readMessageInt32(message, offset + 1);

         receiveResult(msgId, new Object());
         handled = true;
      }
      else if ((length == 21 || length == 25 || length == 23 || length == 27) &&
               message[offset] == MSG_MOUSE_EVENT)
      {
         int idx = offset + 1;
         
         int windowId = readMessageInt32(message, idx);
         idx = idx + 4;

         int mouseOpCode = message[idx];
         idx = idx + 1;
         
         long tstamp = readMessageInt64(message, idx);
         idx = idx + 8;
         
         int mouseX = readMessageInt16(message, idx);
         idx = idx + 2;
         
         int mouseY = readMessageInt16(message, idx);
         idx = idx + 2;
         
         int clickCount = message[idx];
         idx = idx + 1;
         
         int button = message[idx];
         idx = idx + 1;
         
         int modifiersMask = message[idx];
         idx = idx + 1;

         int[] wheelParameters = null;
         if (length == 23 || length == 27)
         {
            int delta = message[idx];
            idx = idx + 1;
            int unit  = message[idx];
            idx = idx + 1;
            wheelParameters = new int[] { delta, unit};
         }
         
         int wid = -1;
         if (length == 25 || length == 27)
         {
            wid = readMessageInt32(message, idx);
            idx = idx + 4;
         }
         
         this.callbacks.raiseMouseEvent(windowId,
                                        mouseOpCode, 
                                        tstamp, 
                                        mouseX, 
                                        mouseY,
                                        button,
                                        clickCount, 
                                        modifiersMask,
                                        wheelParameters,
                                        wid);
         
         handled = true;
      }
      else if (length == 13 && message[offset] == MSG_SET_WINDOW_LOC)
      {
         int idx = offset + 1;
         
         int windowId = readMessageInt32(message, idx);
         idx = idx + 4;

         int x = readMessageInt32(message, idx);
         idx = idx + 4;
         
         int y = readMessageInt32(message, idx);
         idx = idx + 4;
         
         this.callbacks.setWindowLocation(windowId, x, y);
         
         handled = true;
      }
      else if (length == 9 && message[offset] == MSG_WINDOW_ACTIVATED)
      {
         int idx = offset + 1;
         
         int deactivatingWindowId = readMessageInt32(message, idx);
         idx = idx + 4;
         
         int activatingWindowId = readMessageInt32(message, idx);
         idx = idx + 4;

         ThinClient.getInstance().invokeLaterOS(() ->
         {
            callbacks.windowActivated(deactivatingWindowId, activatingWindowId);
         });
      }
      else if (length == 6 && message[offset] == MSG_WINDOW_ICONIFY)
      {
         int idx = offset + 1;
         
         int windowId = readMessageInt32(message, idx);
         idx = idx + 4;
         
         boolean state = message[idx] == 1;

         this.callbacks.windowIconified(windowId, state);
      }
      else if (length == 17 && message[offset] == MSG_WINDOW_RESIZED)
      {
         int idx = offset + 1;
         
         int windowId = readMessageInt32(message, idx);
         idx = idx + 4;
         
         int y = readMessageInt32(message, idx);
         idx = idx + 4;
         
         int x = readMessageInt32(message, idx);
         idx = idx + 4;
         
         int width = readMessageInt16(message, idx);
         idx = idx + 2;
         
         int height = readMessageInt16(message, idx);
         idx = idx + 2;

         this.callbacks.setWindowLocation(windowId, x, y);
         this.callbacks.setWindowDimension(windowId, width, height);
      }
      else if (length == 6 && message[offset] == MSG_IS_FONT_INSTALLED)
      {
         int msgId = readMessageInt32(message, offset + 1);
         byte msg = message[offset + 5];
         receiveResult(msgId, Boolean.valueOf(msg == 1));
         handled = true;
      }
      else if (length == 6 && message[offset] == MSG_DRAW_HASH_REMOVED)
      {
         int msgId = readMessageInt32(message, offset + 1);

         receiveResult(msgId, new Object());
         handled = true;
      }
      else if (length == 9 && message[offset] == MSG_DESKTOP_RESIZED)
      {
         int idx = offset + 1;

         int width = readMessageInt32(message, idx);
         idx = idx + 4;

         int height = readMessageInt32(message, idx);

         this.callbacks.desktopResized(width, height);
      }
      else if (message[offset] == MSG_PARSE_EDITOR_CONTENT)
      {
         int[][] result = new int[2][];
         
         offset = offset + 1;
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int splitNo = readMessageInt32(message, offset);
         offset = offset + 4;
         
         result[0] = new int[splitNo];
         List<Integer> newLines = new ArrayList<>();
         for (int i = 0; i < splitNo; i++)
         {
            result[0][i] = readMessageInt32(message, offset);
            offset += 4;
            
            boolean newLine = message[offset] == 1;
            offset = offset + 1;
            if (newLine)
            {
               newLines.add(result[0][i]);
            }
         }
         
         result[1] = Utils.integerCollectionToPrimitive(newLines);
         
         receiveResult(msgId, result);
         handled = true;
      }
      else if (message[offset] == MSG_P2J_INVOKE)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // the call type: 0-extprog, 1-proc, 2-funct
         InvokeType type = InvocationRequestPayload.InvokeType.fromByte(message[offset]);
         offset = offset + 1;

         // NO-ERROR mode
         boolean noError = (message[offset] == 1);
         offset = offset + 1;

         // notification on wait
         boolean needsNotificationOnWait = (message[offset] == 1);
         offset = offset + 1;

         // read the target's name
         int nameLength = readMessageInt16(message, offset);
         offset = offset + 2;
         String name = readMessageText(message, offset, offset + 2 * nameLength);
         offset = offset + 2 * name.length();

         final boolean persistent;
         String inHandle;

         if (type == InvocationRequestPayload.InvokeType.EXTERNAL_PROGRAM)
         {
            // persistent - only for external programs
            persistent = message[offset] == 1;
            offset = offset + 1;
            
            inHandle = null;
         }
         else
         {
            persistent = false;
            
            // read the handle - only for function or procedure
            // handle = 0 means is not specified

            // read its hex representation (16 bits)
            int handleStringLen = message[offset];
            offset = offset + 1;

            inHandle = readMessageText(message, offset, offset + 2 * handleStringLen);
            offset = offset + 2 * handleStringLen;
         }
         
         Map<String, String> arguments = new HashMap<>();
         // read the arguments
         int argNo = readMessageInt16(message, offset);
         offset = offset + 2;

         for (int i = 0; i < argNo; i++)
         {
            // read the arg name
            int argNameLength = message[offset];
            offset = offset + 1;
            String argName = readMessageText(message, offset, offset + 2 * argNameLength);
            offset = offset + 2 * argName.length();
            
            // read the arg value
            int argValueLength = readMessageInt16(message, offset);
            offset = offset + 2;
            String argValue = readMessageText(message, offset, offset + 2 * argValueLength);
            offset = offset + 2 * argValue.length();
            
            arguments.put(argName, argValue);
         }
         
         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            if (!tc.isEmbedded())
            {
               String msg =
                 "Attempted to execute a remote invoke request in a non-embedded client.";
               throw new SilentUnwindException(new IllegalStateException(msg));
            }

            // if not positive, abend!
            if (msgId <= 0)
            {
               String msg = 
                "Attempted to execute a remote invoke request with a non-positive message ID: %d";
               
               msg = String.format(msg, msgId);
               
               throw new SilentUnwindException(new IllegalStateException(msg));
            }

            boolean error = false;
            String errorMessage = null;
            
            long handle = 0;
            try
            {
               handle = (inHandle == null || inHandle.length() == 0 
                           ? 0 
                           : Long.parseLong(inHandle, 16));
            }
            catch (NumberFormatException e)
            {
               handle = 0;
               error = true;
               errorMessage = "Error resolving the handle ID " + inHandle + ": " + e.getMessage();
            }
            
            InvocationResponsePayload resp;
            
            if (error)
            {
               resp = new InvocationResponsePayload();
               resp.error = true;
               resp.errorMessage = errorMessage;
            }
            else
            {
               InvocationRequestPayload payload = new InvocationRequestPayload();
               payload.name = name;
               payload.handle = handle;
               payload.type = type;
               payload.arguments = arguments;
               payload.persistent = persistent;
               payload.noError = noError;

               resp = (InvocationResponsePayload) tc.invokeOnServer(() ->
               {
                  if (needsNotificationOnWait)
                  {
                     callbacks.setNeedsNotificationOnWait(true);
                  }

                  try
                  {
                     ServerReturnValue ret = tc.getServer().invoke(payload);
                     return ret;
                  }
                  finally
                  {
                     if (needsNotificationOnWait)
                     {
                        callbacks.setNeedsNotificationOnWait(false);
                     }
                  }
               });

               if (resp == null)
               {
                  resp = new InvocationResponsePayload();
                  resp.error = true;
                  resp.errorMessage = "An unknown error occurred while executing the request";
               }
            }
            
            String shandle = (resp.handle == 0 ? "" : Long.toHexString(resp.handle));
            
            Map<String, String> output = resp.outputArguments;
            if (output == null)
            {
               output = Collections.emptyMap();
            }

            int responseLength = 1 + 4; // type + msgId
            // handle (this must always be sent, as on the JS side we won't check if the call was
            // ran persistent or not); the number is sent as a 64-bit value, encoded in hex
            responseLength += 1 + 2 * shandle.length();
            // return-value
            // length
            responseLength += 4;
            // content
            responseLength += 2 * (resp.returnValue == null ? 0 : resp.returnValue.length()); 
            // ERROR
            responseLength += 1;
            if (resp.error)
            {
               // error+number + error-message-length + error-message
               responseLength += 4 + 2 + 2 * resp.errorMessage.length();
            }
            // quit
            responseLength += 1;
            // stop
            responseLength += 1;
            
            // output arguments
            responseLength += 2;
            for (String key : output.keySet())
            {
               // key length + key
               responseLength += 1;
               responseLength += 2 * key.length();
               
               // value length + value
               String val = output.get(key);
               responseLength += 2;
               responseLength += 2 * val.length(); 
            }
            
            BinaryMessage msg = new BinaryMessage(MSG_P2J_INVOKE, responseLength);
            msg.writeInt32(msgId);
            msg.writeByte((byte) shandle.length());
            msg.writeText(shandle);
            msg.writeInt32(resp.returnValue == null ? 0 : resp.returnValue.length());
            msg.writeText(resp.returnValue);
            msg.writeBoolean(resp.error);
            if (resp.error)
            {
               msg.writeInt32(resp.errorNumber);
               msg.writeInt16(resp.errorMessage == null ? 0 : resp.errorMessage.length());
               msg.writeText(resp.errorMessage);
            }
            msg.writeBoolean(resp.quit);
            msg.writeBoolean(resp.stop);
            
            // write the output arguments
            msg.writeInt16(output.size());
            for (String key : output.keySet())
            {
               msg.writeByte((byte) key.length());
               msg.writeText(key);
               
               String val = output.get(key);
               msg.writeInt16(val.length());
               msg.writeText(val);
            }
            
            sendBinaryMessage(msg.message);
            
            // if a condition was raised by the converted code, then re-throw it; the business
            // logic might have called an explicit i.e. QUIT, and this needs to be handled by
            // down-stream logic
            String conditionMsg = "Re-throwing a condition raised by a remote invoke API.";
            if (resp.error)
            {
               // for MSG_P2J_INVOKE, the ControlFlowOps.invoke(InvocationRequestPayload) will
               // eat any conditions; the reason is because we want to send the response to the
               // JS side BEFORE re-raising the conditions to allow the business logic to
               // intercept/treat them; for the ERROR (without NO-ERROR mode) case, we need to
               // throw it unconditionally; since this is client-side processing AND there is no
               // 4GL compatibility requires here, we don't need to use recordOrThrowError()
               if (!noError)
               {
                  throw new ErrorConditionException(resp.errorNumber, resp.errorMessage);
               }
            }
            else if (resp.quit)
            {
               throw new QuitConditionException(conditionMsg);
            }
            else if (resp.stop)
            {
               throw new StopConditionException(conditionMsg);
            }
         });
         
         handled = true;
      }
      else if (message[offset] == MSG_REMOTE_CALL)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int payloadLength = readMessageInt32(message, offset);
         offset = offset + 4;
         String payload = readMessageText(message, offset, offset + 2 * payloadLength);
         
         UnformattedPayload result = new UnformattedPayload();
         result.payload = payload;
         
         receiveResult(msgId, result);
         
         handled = true;
      }
      else if (message[offset] == MSG_P2J_DELETE_HANDLE)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         boolean noError = (message[offset] == 1);
         offset = offset + 1;

         // read its hex representation (16 bits)
         int handleStringLen = message[offset];
         offset = offset + 1;

         String shandle = readMessageText(message, offset, offset + 2 * handleStringLen);
         offset = offset + 2 * handleStringLen;

         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            if (!tc.isEmbedded())
            {
               String msg = 
                  "Attempted to execute a delete handle request in a non-embedded client.";
               throw new SilentUnwindException(new IllegalStateException(msg));
            }
            
            // if not positive, abend!
            if (msgId <= 0)
            {
               String msg = 
                "Attempted to execute a delete handle request with a non-positive message ID: %d";
               
               msg = String.format(msg, msgId);
               
               throw new SilentUnwindException(new IllegalStateException(msg));
            }

            BlockRunner block = new BlockRunner((BlockRunner br) ->
            { 
               long handleId = 0;
               
               try
               {
                  handleId = (handleStringLen == 0 ? 0 : Long.parseLong(shandle, 16));
               }
               catch (NumberFormatException e)
               {
                  br.error = true;
                  br.errorNumber = 0;
                  br.errorMessage = "Error resolving the handle id " + shandle + ": " +
                                       e.getMessage();
               }
               
               if (!br.error)
               {
                  tc.getServer().deleteHandle(handleId, noError);
               }
               
               return (BaseDataType) null;
            }); 

            block.execute();
            
            int msgLength = 1 + 4 + 1; // msgId + error
            if (block.error)
            {
               msgLength += 4 + 2 + 2 * block.errorMessage.length();
            }
            
            BinaryMessage msg = new BinaryMessage(MSG_P2J_DELETE_HANDLE, msgLength);
            msg.writeInt32(msgId);
            msg.writeBoolean(block.error);
            if (block.error)
            {
               msg.writeInt32(block.errorNumber);
               msg.writeInt16(block.errorMessage.length());
               msg.writeText(block.errorMessage);
            }
            
            sendBinaryMessage(msg.message);
         });
         
         handled = true;
      }
      else if (message[offset] == MSG_P2J_SUBSCRIBE)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;

         boolean noError = (message[offset] == 1);
         offset = offset + 1;
         
         boolean global = (message[offset] == 1);
         offset = offset + 1;
         
         // read subscriber hex representation (16 bits)
         // this is a procedure/external resource handle ID
         int subidStringLen = message[offset];
         offset = offset + 1;
         String subscriberId = readMessageText(message, offset, offset + 2 * subidStringLen);
         offset = offset + 2 * subidStringLen;
         Long subscriber = (subidStringLen == 0 ? null : Long.parseLong(subscriberId, 16));
         
         int eventStringLen = message[offset];
         offset = offset + 1;
         String event = readMessageText(message, offset, offset + 2 * eventStringLen);
         offset = offset + 2 * eventStringLen;
         
         // read publisher hex representation (16 bits)
         // this is a procedure/external resource handle ID
         int phandleStringLen = message[offset];
         offset = offset + 1;
         String phandle = readMessageText(message, offset, offset + 2 * phandleStringLen);
         offset = offset + 2 * phandleStringLen;
         Long publisher = (phandleStringLen == 0 ? null : Long.parseLong(phandle, 16));

         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            long sub = 0;
            
            BlockRunner block = new BlockRunner((BlockRunner br) ->
            {
               long l = tc.getServer().subscribeExternal(global,
                                                         subscriber,
                                                         event,
                                                         publisher,
                                                         noError);
               return new int64(l);
            }); 

            block.execute();
            
            if (block.result != null)
            {
               sub = ((int64) block.result).longValue();
            }

            // WARNING: do not use handle.resourceId(), as we are on the client-side!
            String hexsub = Long.toHexString(sub);

            // msgId + computed subscriber's handle ID
            int msgLength = 1 + 4 + 2 + 2 * hexsub.length() + 1;
            
            if (block.error)
            {
               msgLength += 4 + 2 + 2 * block.errorMessage.length();
            }
            
            BinaryMessage msg = new BinaryMessage(MSG_P2J_SUBSCRIBE, msgLength);
            msg.writeInt32(msgId);
            msg.writeByte((byte) hexsub.length());
            msg.writeText(hexsub);
            msg.writeBoolean(block.error);
            if (block.error)
            {
               msg.writeInt32(block.errorNumber);
               msg.writeInt16(block.errorMessage.length());
               msg.writeText(block.errorMessage);
            }
            
            sendBinaryMessage(msg.message);
         });
         
         handled = true;
      }
      else if (message[offset] == MSG_P2J_UNSUBSCRIBE)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;

         boolean noError = (message[offset] == 1);
         offset = offset + 1;
         
         boolean global = (message[offset] == 1);
         offset = offset + 1;

         // read subscriber hex representation (16 bits)
         // this is a procedure/external resource handle ID
         int subidStringLen = message[offset];
         offset = offset + 1;
         String subscriberId = readMessageText(message, offset, offset + 2 * subidStringLen);
         offset = offset + 2 * subidStringLen;
         Long subscriber = (subidStringLen == 0 ? null : Long.parseLong(subscriberId, 16));
         
         // event may be null (when empty text)
         int eventStringLen = message[offset];
         offset = offset + 1;
         String event;
         if (eventStringLen > 0)
         {
            event = readMessageText(message, offset, offset + 2 * eventStringLen);
            offset = offset + 2 * eventStringLen;
         }
         else
         {
            event = null;
         }
         
         // read publisher hex representation (16 bits)
         // this is a procedure/external resource handle ID
         int phandleStringLen = message[offset];
         offset = offset + 1;
         String phandle = readMessageText(message, offset, offset + 2 * phandleStringLen);
         offset = offset + 2 * phandleStringLen;
         Long publisher = phandleStringLen == 0 ? null : Long.parseLong(phandle, 16);

         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            BlockRunner block = new BlockRunner((BlockRunner br) ->
            {
               // noError can be ignored here because this is always implicitly noError
               tc.getServer().unsubscribeExternal(global, subscriber, event, publisher);
               return (BaseDataType) null;
            });
            
            block.execute();

            // msgId
            int msgLength = 1 + 4 + 1;
            if (block.error)
            {
               msgLength += 4 + 2 + 2 * block.errorMessage.length();
            }
            
            BinaryMessage msg = new BinaryMessage(MSG_P2J_UNSUBSCRIBE, msgLength);
            msg.writeInt32(msgId);
            msg.writeBoolean(block.error);
            if (block.error)
            {
               msg.writeInt32(block.errorNumber);
               msg.writeInt16(block.errorMessage.length());
               msg.writeText(block.errorMessage);
            }

            sendBinaryMessage(msg.message);
         });
         
         handled = true;
      }
      else if (message[offset] == MSG_P2J_PUBLISH)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;

         // event may be null (when empty text)
         int eventStringLen = message[offset];
         offset = offset + 1;
         String event = readMessageText(message, offset, offset + 2 * eventStringLen);
         offset = offset + 2 * eventStringLen;
         
         // read publisher hex representation (16 bits)
         // this is a procedure/external resource handle ID
         int phandleStringLen = message[offset];
         offset = offset + 1;
         String phandle = readMessageText(message, offset, offset + 2 * phandleStringLen);
         offset = offset + 2 * phandleStringLen;
         Long publisher = Long.parseLong(phandle, 16);

         // read arg num
         int argNum = message[offset];
         offset = offset + 1;
         String[] args = new String[argNum];
         for (int i = 0; i < argNum; i++)
         {
            // read arg
            int argLen = message[offset];
            offset = offset + 1;
            String arg = readMessageText(message, offset, offset + 2 * argLen);
            offset = offset + 2 * argLen;
            
            args[i] = arg;
         }
         
         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            try
            {
               // no return value or output arguments are supported
               tc.getServer().publishExternal(event, publisher, args);
            }
            catch (Throwable t)
            {
               t.printStackTrace(System.err);
            }

            // no response is sent for remote PUBLISH 
         });
         
         handled = true;
      }
      else if (message[offset] == MSG_P2J_CREATE_RESOURCE)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            long resId = tc.getServer().createExternal();
            String hexId = Long.toHexString(resId);

            // type + msgId + hexId.length + 2 * hexId.length
            int msgLength = 1 + 4 + 1 + 2 * hexId.length();
            
            BinaryMessage msg = new BinaryMessage(MSG_P2J_CREATE_RESOURCE, msgLength);
            msg.writeInt32(msgId);
            msg.writeByte((byte) hexId.length());
            msg.writeText(hexId);

            sendBinaryMessage(msg.message);
         });
         
         handled = true;
      }
      else if (length == 9 && message[offset] == MSG_SENT_DESKTOP_DIMENSION)
      {
         offset = offset + 1;
         
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;
         
         int width = readMessageInt16(message, offset);
         offset = offset + 2;
         
         int height = readMessageInt16(message, offset);
         offset = offset + 2;
         
         receiveResult(msgId, new int[] { width, height });
         handled = true;
      }
      else if (message[offset] == MSG_DROP_FILE_NOTIFY)
      {
         offset = offset + 1;
         
         // getting the target window
         int winID = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // getting the target widget candidate
         int widTargetID = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // getting the file list size
         int listSize = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // getting the file names and building the Java list
         List<String> fileNameList = new ArrayList<String>();
         
         for (int i = 0; i < listSize; i++)
         {
            int nameLength = readMessageInt16(message, offset);
            offset = offset + 2;
            String fileName = readMessageText(message, offset, offset + 2 * nameLength);
            offset = offset + 2 * nameLength;
            // new File object
            fileNameList.add(fileName);
         }
         
         // process dropping if list is not empty
         if (!fileNameList.isEmpty())
         {
            // calling gui driver to continue processing
            this.callbacks.startDropTarget(winID, widTargetID, fileNameList);
         }
         
         handled = true;
      }
      else if (message[offset] == MSG_FILE_UPLOADING)
      {
         offset = offset + 1;
         
         // file name length
         int nameLength = readMessageInt16(message, offset);
         offset = offset + 2;
         
         // file name
         String fileName = readMessageText(message, offset, offset + 2 * nameLength);
         offset = offset + 2 * nameLength;
         
         // file size
         int fileSize = readMessageInt32(message, offset);
         offset = offset + 4;
         
         // file content itself
         byte[] fileData = readByteArray(message, offset, fileSize);
         
         // content can be empty
         if (fileData != null)
         {
            // ask java to store the file locally in temp directory
            this.callbacks.saveDropTargetFile(fileName, fileData, fileSize);
         }
         
         handled = true;
      }
      else if (length == 9 && message[offset] == MSG_GET_MOUSE_POSITION)
      {
         offset = offset + 1;

         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;

         int x = readMessageInt16(message, offset);
         offset = offset + 2;

         int y = readMessageInt16(message, offset);

         receiveResult(msgId, new int[] { x, y });
         handled = true;
      }
      else if (message[offset] == MSG_P2J_APPLY_EVENT)
      {
         offset = offset + 1;

         // read from socket
         int msgId = readMessageInt32(message, offset);
         offset = offset + 4;

         // event may be null (when empty text)
         int eventStringLen = message[offset];
         offset = offset + 1;
         String event = readMessageText(message, offset, offset + 2 * eventStringLen);
         offset = offset + 2 * eventStringLen;
         
         // read the target's HWND id (i.e. the widget ID)
         int widgetId = readMessageInt32(message, offset);
         offset = offset + 4;

         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
         {
            try
            {
               tc.apply(widgetId, Keyboard.eventCode(event), null, new EventList(), true);
            }
            catch (Throwable t)
            {
               t.printStackTrace(System.err);
            }

            // no response is sent for remote PUBLISH 
         });
         
         handled = true;
      }
      
      return handled;
   }
   
   /**
    * Process the incoming PASTE message from the client.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the text begins.
    * @param    length
    *           Location in the message where the text ends.
    */
   public void paste(byte[] message, int offset, int length)
   {
      if (isVT100)
      {
         super.paste(message, offset, length);
         return;
      }
      
      // save off the clipboard contents
      synchronized (clipboardLock)
      {
         pending = ClipboardRequest.CLIENT_GENERATED;
         
         // handle the special case of an empty clipboard text
         if (length == 3 && message[offset + 1] == 0 && message[offset + 2] == 22)
         {
            clipboardContents = "";
         }
         else
         {
            clipboardContents = readMessageText(message, offset + 1, length);
         }
         
         clipboardLock.notify();
      }
      
      // now generate the CTRL-V, this should generate a PASTE 4GL event which will
      // result in the clipboard contents being accessed via readClipboard()
      injectKey(22);
   }

   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param    fontStyle
    *           Font style.
    */
   public void setFontStyle(int fontStyle)
   {
      allocateAndSend(PaintPrimitives.SET_FONT_STYLE, fontStyle);
   }

   /**
    * Set current font used on draw operations.
    * 
    * @param    font
    *           Drawing font.
    */
   public void setFont(int font)
   {
      allocateAndSend(PaintPrimitives.SET_FONT, font);
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    font
    *           The font ID to register it on the js-side.
    * @param    fontName
    *           The font name.
    * @param    pointSize
    *           The font's size, in points.
    * @param    fontStyle
    *           The font style as packaged via {@link FontStyle#toInt()}.
    * @param    boldFontDefinition
    *           Indicates that the font definition is for the bold font
    * @param    fontDefinition
    *           The font definition.  May be <code>null</code>.
    *           
    * @return   The font index on the JS side.
    */
   public int createFont(int     font, 
                         String  fontName,
                         int     pointSize, 
                         int     fontStyle, 
                         boolean boldFontDefinition,
                         byte[]  fontDefinition)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // font
      len = len + 2; // fontName.length()
      len = len + fontName.length() * 2; // font name
      len = len + 1; // point size
      len = len + 1; // font-style
      len = len + 1; // boldFontDefinition
      len = len + 4; // fontDefinition.length
      len = len + (fontDefinition == null ? 0 : fontDefinition.length); // fontDefinition
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_CREATE_FONT;
      
      writeMessageInt32(message, offset, msgId);
      offset += 4;
      
      writeMessageInt32(message, offset, font);
      offset += 4;
      
      writeMessageInt16(message, offset, fontName.length());
      offset += 2;

      writeMessageText(message, offset, fontName);
      offset += fontName.length() * 2;

      message[offset++] = (byte) pointSize;
      
      message[offset++] = (byte) fontStyle;
      
      message[offset++] = boldFontDefinition ? (byte) 1 : (byte) 0;
      
      int length = (fontDefinition == null ? 0 : fontDefinition.length);
      writeMessageInt32(message, offset, length);
      offset += 4;
      
      if (fontDefinition != null)
      {
         System.arraycopy(fontDefinition, 0, message, offset, length);
      }
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);

      if (res == null)
      {
         // TODO: log me
         return 0;
      }
      else
      {
         return ((Integer) res).intValue();
      }
   }

   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    font
    *           Font to scale.  This instance's contained font will be modified. 
    */
   public void deriveFont(int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_DERIVE_FONT;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, font);
      offset += 4;

      sendBinaryMessage(message);
      
      waitForResult(msgId);
   }

   /**
    * Get the desktop's dimension.
    * 
    * @return   See above.
    */
   public NativeDimension getDesktopDimension()
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_GET_DESKTOP_DIMENSION;
      writeMessageInt32(message, offset, msgId);

      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return null;
      }
      else
      {
         try
         {
            int[] dim = (int[]) res;
            if (dim[0] == 0 || dim[1] == 0)
            {
               // something happened, return null
               return null;
            }
            
            return new NativeDimension(dim[0], dim[1]);
         }
         catch (Exception e)
         {
            e.printStackTrace();
            
            return null;
         }
      }
   }
   
   /**
    * The method performs a layout operation on the supplied text and * returns the resulting 
    * paragraph height while maintaining the supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with {@link #drawParagraph} and so can
    * be used in conjunction with this method. 
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   public int getParagraphHeight(String text, int font, int maxWidth)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // text.length()
      len = len + text.length() * 2; // text
      len = len + 4; // font
      len = len + 2; // maxWidth
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_GET_PAR_HEIGHT;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, text.length());
      offset += 4;

      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, font);
      offset += 4;
      
      writeMessageInt16(message, offset, maxWidth);
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return 0;
      }
      else
      {
         return ((Integer) res).intValue();
      }
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Deprecated
   public int getTextHeight(String text, int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // text.length()
      len = len + text.length() * 2; // text
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_TEXT_HEIGHT;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, text.length());
      offset += 4;

      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, font);
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return 0;
      }
      else
      {
         return ((Integer) res).intValue();
      }
   }

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Deprecated
   public int getTextWidth(String text, int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // text.length()
      len = len + text.length() * 2; // text
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_TEXT_WIDTH;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, text.length());
      offset += 4;

      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, font);
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return 0;
      }
      else
      {
         return ((Integer) res).intValue();
      }
   }

   /**
    * Get the widths of the specified texts, using the given fonts.
    *
    * @param    fonts
    *           The fonts used to draw and measure the texts.
    * @param    texts
    *           The texts. The main array contains sub-arrays, every sub-array is drawn with a
    *           separate font. Corresponding font can be found in the <code>fonts</code> array:
    *           it has the same index as the texts sub-array in the <code>texts</code> array.
    *
    * @return   The text's widths. Resulting array of integers has the same configuration as the
    *           <code>texts</code> array.
    */
   public int[][] getTextWidths(int[] fonts, String[][] texts)
   {
      if (fonts.length != texts.length)
      {
         throw new IllegalArgumentException("fonts.length should equal texts.length!");
      }

      if (texts.length == 0)
      {
         return new int[0][];
      }

      int len = 1;   // type
      len += 4; // msg ID
      len += 2; // fonts.length


      len += 4 * fonts.length; // fonts contents
      for (int i = 0; i < texts.length; i++)
      {
         len += 2; // sub-array length

         for (int j = 0; j < texts[i].length; j++)
         {
            len += 2 + texts[i][j].length() * 2; // text.length + text
         }
      }
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_TEXT_WIDTHS_BATCH;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt16(message, offset, fonts.length); // fonts and texts size
      offset += 2;

      for (int font : fonts)
      {
         writeMessageInt32(message, offset, font);
         offset += 4;
      }

      for (String[] fontTexts : texts)
      {
         writeMessageInt16(message, offset, fontTexts.length);
         offset += 2;

         for (String text : fontTexts)
         {
            writeMessageInt16(message, offset, text.length());
            offset += 2;

            writeMessageText(message, offset, text);
            offset += text.length() * 2;
         }
      }

      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return null;
      }
      else
      {
         return (int[][]) res;
      }
   }

   /**
    * Get the widths of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Deprecated
   public int[] getTextWidths(String text, int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // text.length()
      len = len + text.length() * 2; // text
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_TEXT_WIDTHS;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, text.length());
      offset += 4;

      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, font);
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return null;
      }
      else
      {
         return (int[]) res;
      }
   }

   /**
    * Requests the client if the target font is installed.
    * 
    * @param    fontName
    *           The target font name.
    * @param    boldFont
    *           Indicates that the bold font is queried.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   public boolean isFontInstalled(String fontName, boolean boldFont)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 1; // boldFont
      len = len + 4; // fontName.length()
      len = len + fontName.length() * 2; // fontName
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_IS_FONT_INSTALLED;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      message[offset++] = boldFont ? (byte) 1 : (byte) 0;

      writeMessageInt32(message, offset, fontName.length());
      offset += 4;

      writeMessageText(message, offset, fontName);
      
      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         return false;
      }
      else
      {
         if (res instanceof Boolean)
         {
            return (Boolean) res;
         }
         else
         {
            // log exception and return negative 
            new RuntimeException("Not a Boolean: " + String.valueOf(res)).printStackTrace();
            return false;
         }
      }
   }
   
   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   public int getFontHeight(int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_GET_FONT_HEIGHT;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, font);
      offset += 4;

      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return 0;
      }
      else
      {
         return ((Integer) res).intValue();
      }
   }

   /**
    * Report the advance widths of the first 256 characters in the font.  Advance width measures 
    * the length of the character's baseline.
    *
    * @return   The array of widths.
    */
   public int[] resolveWidths(int font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // font
      
      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_FONT_WIDTHS;

      writeMessageInt32(message, offset, msgId);
      offset += 4;

      writeMessageInt32(message, offset, font);
      offset += 4;

      sendBinaryMessage(message);
      
      Object res = waitForResult(msgId);
      
      if (res == null)
      {
         // TODO: log me
         return null;
      }
      else
      {
         return (int[]) res;
      }
   }

   /**
    * Sends MSG_SET_CURSOR_STYLE that the cursor style must be changed on the client side.
    * 
    * @param    cursorStyle
    *           Denotes the cursor style id.
    * @param    wid
    *           The window id.
    */
   public void setCursorStyle(int cursorStyle, int wid)
   {
      byte[] message = new byte[9];

      message[0] = MSG_SET_CURSOR_STYLE;

      writeMessageInt32(message, 1, cursorStyle);
      writeMessageInt32(message, 5, wid);

      sendBinaryMessage(message);
   }

   /**
    * Sends MSG_SET_STROKE_STYLE to order the current window stroke to be changed
    * on the client side. It sends a window its stroke style and line width. All possible stroke
    * styles are listed by LineStroke enumeration. 
    * 
    * @param    strokeStyle
    *           The known stroke style id.
    * @param    width
    *           The stroke line width.
    */
   public void setLineStroke(int strokeStyle, int width)
   {
      allocateAndSend(PaintPrimitives.SET_LINE_STROKE, strokeStyle, width);
   }

   /**
    * Draw a string at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    centered
    *           Flag indicating if the text is centered vertically.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   public void drawString(String text, boolean centered, int x, int y)
   {
      int length = 1 + 4 + text.length() * 2 + 4 + 4 + 4; 
      byte[] message = allocateDrawingOp(PaintPrimitives.DRAW_STRING, length);
      
      int offset = 1;
      writeMessageInt32(message, offset, text.length());
      offset += 4;
      
      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, (centered ? 1 : 0));
      offset += 4;
      
      writeMessageInt32(message, offset, x);
      offset += 4;
      
      writeMessageInt32(message, offset, y);
      offset += 4;
   }

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    centered
    *           Flag indicating if the text is centered vertically.
    * @param    width
    *           The legacy width, which needs to be matched when drawing.
    * @param    height
    *           The legacy width, which needs to be matched when drawing.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   public void drawStringScaled(String text, boolean centered, int x, int y, int width, int height)
   {
      int length = 1 + 4 + text.length() * 2 + 4 + 4 + 4 + 4 + 4; 
      byte[] message = allocateDrawingOp(PaintPrimitives.DRAW_STRING_SCALED, length);
      
      int offset = 1;
      writeMessageInt32(message, offset, text.length());
      offset += 4;
      
      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, (centered ? 1 : 0));
      offset += 4;
      
      writeMessageInt32(message, offset, x);
      offset += 4;
      
      writeMessageInt32(message, offset, y);
      offset += 4;

      writeMessageInt32(message, offset, width);
      offset += 4;
      
      writeMessageInt32(message, offset, height);
      offset += 4;
   }

   /**
    * Draws a multiline text (paragraph) at the supplied position and with the specified font. 
    * The paragraph width will not exceed the supplied maximum width. The text will be properly
    * broken into multiple lines.
    * <p>
    * Note that the algorithm for text layout is compatible with {@link #getParagraphHeight} and 
    * so can be used in conjunction with this method.
    *
    * @param    text
    *           The text to draw.
    * @param    font
    *           The drawing font.
    * @param    x
    *           The x position.
    * @param    y
    *           The y position.
    * @param    width
    *           The maximum paragraph width.
    */
   public void drawParagraph(String text, int font, int x, int y, int width)
   {
      int length = 1 + 4 + text.length() * 2 + 4 + 4 + 4 + 4; 
      byte[] message = allocateDrawingOp(PaintPrimitives.DRAW_PARAGRAPH, length);
      
      int offset = 1;
      writeMessageInt32(message, offset, text.length());
      offset += 4;
      
      writeMessageText(message, offset, text);
      offset += text.length() * 2;
      
      writeMessageInt32(message, offset, font);
      offset += 4;
      
      writeMessageInt32(message, offset, x);
      offset += 4;
      
      writeMessageInt32(message, offset, y);
      offset += 4;

      writeMessageInt32(message, offset, width);
      offset += 4;
   }

   /**
    * Process the details for the specified widgets and send them to the JS side.
    * <p>
    * All <code>newWidgets</code> will be registered with their JS window.  
    * The <code>deadWidgets</code>  will be removed.
    * <p>
    * Each map has a key a window ID and as value a map with the window's widgets: for each widget,
    * its supported mouse events are specified.
    * 
    * @param    allWidgets
    *           The entire widget tree, in z-order.  This will always be sent to the Web side, if
    *           there are changes in the new/dead widgets.
    * @param    newWidgets
    *           Details about the new widgets.
    * @param    newWidgetsBounds
    *           Maps a window to its widgets bounds: screen locations and physical dimensions.
    * @param    deadWidgets
    *           Details about the widgets which no longer process mouse events (i.e. they were
    *           hidden/disabled/etc).
    * @param    newAnyWidgets
    *           Details about "any mouse" widgets that process all mouse events.
    * @param    deadAnyWidgets
    *           Details about the "any mouse" widgets which no longer process mouse events
    *           (i.e. they were hidden/disabled/etc).
    */
   void processMouseWidgets(Map<Integer, Map<Integer, Set<Integer>>>    allWidgets,
                            Map<Integer, Map<Integer, Set<Integer>>>    newWidgets,
                            Map<Integer, Map<Integer, NativeRectangle>> newWidgetsBounds,
                            Map<Integer, Map<Integer, Set<Integer>>>    deadWidgets,
                            Map<Integer, Set<Integer>>                  newAnyWidgets,
                            Map<Integer, Set<Integer>>                  deadAnyWidgets)
   {
      if (newWidgets.isEmpty()    &&
          deadWidgets.isEmpty()   &&
          newAnyWidgets.isEmpty() &&
          deadAnyWidgets.isEmpty())
      {
         // nothing to do
         return;
      }
      
      int length = 1;
      
      // 2 bytes for the number of windows with new widgets
      length = length + 2;
      for (Integer windowId : newWidgets.keySet())
      {
         Map<Integer, Set<Integer>> widgets = newWidgets.get(windowId);
         
         // the window ID
         length = length + 4;
         
         // the number of new widgets
         length = length + 2;
         
         // the widgets: widget ID + x + y + width + height + actions 
         length = length + widgets.size() * (4 + 2 + 2 + 2 + 2 + 4);
         
         // the z-order widgets: number of all widgets + their IDs in a list
         length = 2 + length + allWidgets.get(windowId).size() * 4;
      }
      
      // 2 bytes for the number of windows with dead widgets
      length = length + 2;
      for (Integer windowId : deadWidgets.keySet())
      {
         Map<Integer, Set<Integer>> widgets = deadWidgets.get(windowId);
         
         // the window ID
         length = length + 4;
         
         // the number of dead widgets
         length = length + 2;
         
         // the widgets: widget ID
         length = length + widgets.size() * 4;
      }

      // 2 bytes for the number of windows with new "any widgets"
      length = length + 2;
      for (Integer windowId : newAnyWidgets.keySet())
      {
         Set<Integer> widgets = newAnyWidgets.get(windowId);
         
         // the window ID
         length = length + 4;
         
         // the number of widgets
         length = length + 2;
         
         // the widgets: widget ID
         length = length + widgets.size() * 4;
      }

      // 2 bytes for the number of windows with dead "any widgets"
      length = length + 2;
      for (Integer windowId : deadAnyWidgets.keySet())
      {
         Set<Integer> widgets = deadAnyWidgets.get(windowId);
         
         // the window ID
         length = length + 4;
         
         // the number of widgets
         length = length + 2;
         
         // the widgets: widget ID
         length = length + widgets.size() * 4;
      }
      
      byte[] message = new byte[length];
      
      int offset = 0;
      
      message[offset] = PROCESS_MOUSE_WIDGETS;
      offset = offset + 1;
      
      // the number of windows with new widgets
      writeMessageInt16(message, offset, newWidgets.size());
      offset = offset + 2;
      
      for (Integer windowId : newWidgets.keySet())
      {
         Map<Integer, Set<Integer>> widgets = newWidgets.get(windowId);
         
         // the window ID
         writeMessageInt32(message, offset, windowId);
         offset = offset + 4;
         
         // the number of new widgets in this window
         writeMessageInt16(message, offset, widgets.size());
         offset = offset + 2;
         Map<Integer, NativeRectangle> bounds = newWidgetsBounds.get(windowId);
         // the new widgets
         for (Integer wid : widgets.keySet())
         {
            // the widget ID
            writeMessageInt32(message, offset, wid);
            offset = offset + 4;
            
            NativeRectangle nr = bounds.get(wid);
 
            // the widget coordinates
            writeMessageInt16(message, offset, nr.left());
            offset = offset + 2;
            writeMessageInt16(message, offset, nr.top());
            offset = offset + 2;
            // important: + 1 ! see constructor NativeRectangle(NativePoint, NativeDimension)
            int width =  nr.right() - nr.left() + 1;
            // the widget size
            writeMessageInt16(message, offset, width);
            offset = offset + 2;
            // important: + 1 ! see constructor NativeRectangle(NativePoint, NativeDimension)
            int height = nr.bottom() - nr.top() + 1;
            writeMessageInt16(message, offset, height);
            offset = offset + 2;

            // bit-encode the actions
            int actions = 0;
            for (Integer i : widgets.get(wid))
            {
               actions = actions | (1 << i);
            }
            
            writeMessageInt32(message, offset, actions);
            offset = offset + 4;
         }
         
         // the z-order widgets

         // the number of all widgets in this window
         Map<Integer, Set<Integer>> all = allWidgets.get(windowId);
         writeMessageInt16(message, offset, all.size());
         offset = offset + 2;
         
         for (Integer wid : all.keySet())
         {
            writeMessageInt32(message, offset, wid);
            offset = offset + 4;
         }
      }

      // the number of windows with dead widgets
      writeMessageInt16(message, offset, deadWidgets.size());
      offset = offset + 2;
      
      for (Integer windowId : deadWidgets.keySet())
      {
         Map<Integer, Set<Integer>> widgets = deadWidgets.get(windowId);
         
         // the window ID
         writeMessageInt32(message, offset, windowId);
         offset = offset + 4;
         
         // the number of dead widgets in this window
         writeMessageInt16(message, offset, widgets.size());
         offset = offset + 2;
         
         // the dead widgets
         for (Integer wid : widgets.keySet())
         {
            // the widget ID
            writeMessageInt32(message, offset, wid);
            offset = offset + 4;
         }
      }

      // the number of windows with new "any widgets"
      writeMessageInt16(message, offset, newAnyWidgets.size());
      offset = offset + 2;
      
      for (Integer windowId : newAnyWidgets.keySet())
      {
         Set<Integer> widgets = newAnyWidgets.get(windowId);
         
         // the window ID
         writeMessageInt32(message, offset, windowId);
         offset = offset + 4;
         
         // the number of new "any widgets" in this window
         writeMessageInt16(message, offset, widgets.size());
         offset = offset + 2;
         
         // the new "any widgets"
         for (Integer wid : widgets)
         {
            // the widget ID
            writeMessageInt32(message, offset, wid);
            offset = offset + 4;
         }
      }
      
      // the number of windows with dead "any widgets"
      writeMessageInt16(message, offset, deadAnyWidgets.size());
      offset = offset + 2;
      
      for (Integer windowId : deadAnyWidgets.keySet())
      {
         Set<Integer> widgets = deadAnyWidgets.get(windowId);
         
         // the window ID
         writeMessageInt32(message, offset, windowId);
         offset = offset + 4;
         
         // the number of dead widgets in this window
         writeMessageInt16(message, offset, widgets.size());
         offset = offset + 2;
         
         // the dead widgets
         for (Integer wid : widgets)
         {
            // the widget ID
            writeMessageInt32(message, offset, wid);
            offset = offset + 4;
         }
      }

      sendBinaryMessage(message);
   }

   /**
    * Notifies the JS client that the current editors selection is changed.
    * 
    * @param    text
    *           The current editors selection.
    */
   public void setSelection(String text)
   {
      if (text != null)
      {
         sendBinaryMessage(MSG_CURRENT_SELECTION, text);
      }
      else
      {
         sendBinaryMessage(MSG_INVALIDATE_SELECTION);
      }
   }

   /**
    * Register hoverable widget.
    * 
    * @param    wid
    *           The windows id.
    * @param    id
    *           The widget id.
    * @param    pid
    *           The parent widget id.
    * @param    bound
    *           The widget boundaries
    * @param    ptr
    *           The widget custom mouse pointer.
    * @param    defaultPtr
    *           The widget default mouse pointer.
    * @param    enforceDefault
    *           <code>true</code> id default pointer should be always used.
    */
   public void registerHoverableWidget(int wid, int id, int pid,
         NativeRectangle bound,
         MousePtrWrapper ptr, MousePtrWrapper defaultPtr, boolean enforceDefault)
   {
      if (!defaultPtr.isPredefined())
      {
         throw new UnsupportedOperationException("Only pre-defined mouse cursors can be default");
      }
      // type + wid + id + pid + 2*dim + defaultPtr + enforce + ptr  
      int len = 1 + 4 + 4 + 4 + 4*2 + 2 + 1 + serializedSize(ptr);
      BinaryMessage msg = new BinaryMessage(MSG_REGISTER_HOVERABLE_WIDGET, len);
      msg.writeInt32(wid);
      msg.writeInt32(id);
      msg.writeInt32(pid);
      // the widget coordinates
      msg.writeInt16(bound != null ? bound.left() : 0);
      msg.writeInt16(bound != null ? bound.top() : 0);
      // the widget size
      msg.writeInt16(bound != null ? bound.width() : 0);
      msg.writeInt16(bound != null ? bound.height() : 0);
      
      msg.writeInt16(defaultPtr.style());
      msg.writeByte(enforceDefault ? (byte)1 : 0);
      msg.writeMousePtr(ptr);
      
      hoverableWidgets.add(msg);
   }

   /**
    * Deregister widget.
    * 
    * @param    wid
    *           The windows id.
    * @param    id
    *           The widget id.
    */
   public void deregisterWidget(int wid, int id)
   {
      // type + wid + id  
      int len = 1 + 4 + 4;
      BinaryMessage msg = new BinaryMessage(MSG_DEREGISTER_WIDGET, len);
      msg.writeInt32(wid);
      msg.writeInt32(id);
      
      hoverableWidgets.add(msg);
   }

   /**
    * Lock/unlock mouse pointer.
    * 
    * @param    ptr
    *           The custom mouse pointer or <code>null</code> for unlock.
    */
   public void lockMousePointer(int widgetId, MousePtrWrapper ptr)
   {
      // type + wid + ptr  
      int len = 1 + 4 + serializedSize(ptr);
      BinaryMessage msg = new BinaryMessage(MSG_LOCK_MOUSE_PTR, len);
      msg.writeInt32(widgetId);
      msg.writeMousePtr(ptr);
      msg.send();
   }

   /**
    * Set custom mouse pointer for a widget.
    * 
    * @param    wid
    *           The windows id.
    * @param    id
    *           The widget id.
    * @param    ptr
    *           The widget custom mouse pointer.
    */
   public void setMousePtr(int wid, int id, MousePtrWrapper ptr)
   {
      // type + wid + id + ptr  
      int len = 1 + 4 + 4 + serializedSize(ptr);
      BinaryMessage msg = new BinaryMessage(MSG_SET_MOUSE_PTR, len);
      msg.writeInt32(wid);
      msg.writeInt32(id);
      msg.writeMousePtr(ptr);
      msg.send();
   }

   /**
    * Setting up file size limits for cases when drag and drop needs file uploads from web client
    * to Java side.
    * 
    * @param    single
    *           Single file size limit.
    * @param    total
    *           Total file size limit for single operation. 
    */
   public void setUploadFileSizeLimits(int single, int total)
   {
      // type + wid + id + ptr  
      int len = 1 + 4 + 4;
      BinaryMessage msg = new BinaryMessage(MSG_SET_UPLOAD_FILE_SIZE_LIMITS, len);
      msg.writeInt32(single);
      msg.writeInt32(total);
      msg.send();
   }
   
   /**
    * Propagate wait state
    * 
    * @param isWaitState
    *        wait state
    */
   public void propagateWaitState(boolean isWaitState)
   {
      BinaryMessage msg = new BinaryMessage(MSG_SET_WAIT_STATE, 2);
      msg.writeByte(isWaitState ? (byte)1 : 0);
      msg.send();
   }

   /**
    * Returns current mouse cursor display position relative to the display origin.
    *
    * @return  mouse cursor position, never <code>null</code>
    */
   public NativePoint getMousePosition()
   {
      int len = 1;   // type
      len = len + 4; // msg ID

      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();

      message[offset++] = MSG_GET_MOUSE_POSITION;

      writeMessageInt32(message, offset, msgId);
      sendBinaryMessage(message);
      Object res = waitForResult(msgId);

      if (res == null)
      {
         throw new RuntimeException("Unexpected result!");
      }
      else
      {
         try
         {
            int[] pos = (int[]) res;

            return new NativePoint(pos[0], pos[1]);
         }
         catch (Exception e)
         {
            throw new RuntimeException(e);
         }
      }
   }

   /**
    * Requests JS client to open the PDF print output identified with the supplied identifier.
    *
    * @param    id
    *           The id of the print output.
    * @param    mimeType
    *           The document mime type.
    */
   public void openPrintOutput(String id, String mimeType)
   {
      openMimeResource(id, mimeType, true, true);
   }

   /**
    * Requests JS client to open the specified resource.
    *
    * @param    id
    *           The resource id.
    * @param    mimeType
    *           The resource mime type.
    * @param    embedded
    *           The document is embedded in the template document.
    * @param    local
    *           Indicates that the target resource is local.
    */
   public void openMimeResource(String id, String mimeType, boolean embedded, boolean local)
   {
      int len = 1 + 4 + 2 * id.length() + 4 + 2 * mimeType.length() + 2;
      BinaryMessage msg = new BinaryMessage(MSG_OPEN_MIME_RESOURCE, len);
      msg.writeInt32(id.length());
      msg.writeText(id);
      msg.writeInt32(mimeType.length());
      msg.writeText(mimeType);
      msg.writeByte((byte) (embedded ? 1 : 0));
      msg.writeByte((byte) (local ? 1 : 0));
      msg.send();
   }

   /**
    * The method is called when the 4GL program execution enters a modal or wait-for event loop, in other
    * words when user input is expected.
    */
   public void notifyWait()
   {
      BinaryMessage msg = new BinaryMessage(MSG_WAIT_USER_INPUT, 1);
      msg.send();
   }

   /**
    * Notify the driver that the client is ready (useful for embedded mode, to notify the embedded
    * app the client can handle messages).
    */
   void clientReady()
   {
      BinaryMessage msg = new BinaryMessage(MSG_CLIENT_READY, 1);
      msg.send();
   }
   
   /**
    * Split the text for a editor widget, so that it fits the specified maximum width.  The 
    * <code>delta</code> parameter represents the start position of this text, in the editor's 
    * content; when the full content is parsed, this will be zero.
    *  
    * @param    widgetId
    *           The editor's ID.
    * @param    windowId
    *           The editor's window ID.
    * @param    maxWidth
    *           The width for the content to fit (the editor's editable area).
    * @param    text
    *           The text to split into lines.
    * @param    delta
    *           The start position of this text in the editor's full content.
    * @param    font
    *           The editor's font.
    *           
    * @return   Two arrays: one with the split indexes and one with the split indexes which were
    *           determined to be caused by a new line.
    */
   int[][] parseEditorContent(int     widgetId,
                              int     windowId,
                              int     maxWidth,
                              String  text,
                              int     delta,
                              int     font)
   {
      int len = 1;   // type
      len = len + 4; // msg ID
      len = len + 4; // window ID
      len = len + 4; // widget ID
      len = len + 4; // max width
      len = len + 4; // start 
      len = len + 4; // font num
      len = len + 1; // placeChar
      len = len + text.length() * 2; // text

      int offset = 0;
      byte[] message = new byte[len];
      int msgId = nextMessageId();
      
      message[offset++] = MSG_PARSE_EDITOR_CONTENT;
      
      writeMessageInt32(message, offset, msgId);
      offset += 4;
      
      writeMessageInt32(message, offset, windowId);
      offset += 4;
      
      writeMessageInt32(message, offset, widgetId);
      offset += 4;
      
      writeMessageInt32(message, offset, maxWidth);
      offset += 4;
      
      writeMessageInt32(message, offset, delta);
      offset += 4;
      
      writeMessageInt32(message, offset, font);
      offset += 4;

      offset = writeMessageText(message, offset, text);
      
      sendBinaryMessage(message);

      return (int[][]) waitForResult(msgId);
   }

   /**
    * Calculate the serialized size of the mouse pointer
    * 
    * @param ptr
    *        the mouse pointer
    * @return the serialized size of the mouse pointer
    */
   private static int serializedSize(MousePtrWrapper ptr)
   {
      if (ptr == null)
      {
         return 1;
      }
      if (ptr.isPredefined())
      {
         return 1 + 2;
      }
      return 1 + 2 + 2 * ptr.mimeType().length() + 2 + ptr.data().length;
   }

   /**
    * Mark the start of recording the drawing operations.
    * 
    * @param    x
    *           The X origin at the time recording was started. 
    * @param    y
    *           The Y origin at the time recording was started.
    * @param    width
    *           The width of the resulted image, which will be captured and cached.
    * @param    height
    *           The height of the resulted image, which will be captured and cached.
    */
   void startRecording(int x, int y, int width, int height)
   {
      if (!graphicsCached)
      {
         return;
      }
      DrawRecording dr = new DrawRecording(x,
                                           y, 
                                           width, 
                                           height, 
                                           drawingOps.size(), 
                                           drawingNoCacheOps.size());
      this.recordings.push(dr);
   }
   
   /**
    * Mark the stop of the drawing operations recording.  This will compute an MD5 hash using the
    * {@link #drawingNoCacheOps pristine drawing operation list}.  If there is already a cached
    * image, it will replace the drawing operations with a {@link PaintPrimitives#DRAW_CACHED}
    * operation, which will tell the Web driver to draw the cached image.
    */
   void stopRecording()
   {
      if (!graphicsCached)
      {
         return;
      }
      DrawRecording recording = this.recordings.pop();

      try
      {
         ByteArrayOutputStream ops = new ByteArrayOutputStream();
         
         // write the drawing operations, as they were posted, without intermediate hashes
         for (int i = recording.startNoCacheOffset; i < drawingNoCacheOps.size(); i++)
         {
            byte[] op = drawingNoCacheOps.get(i);
            ops.write(op);
         }
         
         byte[] drawHash = md5Hash(ops.toByteArray());
         
         if (drawHash == null)
         {
            // the hash could not be computed, do nothing.
            return;
         }
         
         String hash = XmlHelper.convertBytesToHex(drawHash);

         if (perWidgetCache.get(hash) != null)
         {
            // the cache exists, remove the drawing ops (only from the drawingOps list, as this
            // is where the ops which will be sent to the JS side are collected)
            while (recording.startOffset != drawingOps.size())
            {
               pendingBytes -= drawingOps.remove(drawingOps.size() - 1).length;
            }
            
            // and replace them with DRAW_CACHED(x,y,md5)
            byte[] message = allocateDrawingOp(PaintPrimitives.DRAW_CACHED, 
                                               1 + 4 + 4 + HASH_LENGTH);
            int offset = 1;
            writeMessageInt32(message, offset, recording.x);
            offset += 4;
            writeMessageInt32(message, offset, recording.y);
            offset += 4;
            System.arraycopy(drawHash, 0, message, offset, HASH_LENGTH);
            offset += HASH_LENGTH;
         }
         else
         {
            byte[] message = new byte[1 + 4 + 4 + 4 + 4 + HASH_LENGTH];
            pendingBytes += message.length;

            message[0] = (byte) PaintPrimitives.START_RECORDING.ordinal();
            int offset = 1;
            writeMessageInt32(message, offset, recording.x);
            offset += 4;
            writeMessageInt32(message, offset, recording.y);
            offset += 4;
            writeMessageInt32(message, offset, recording.width);
            offset += 4;
            writeMessageInt32(message, offset, recording.height);
            offset += 4;
            System.arraycopy(drawHash, 0, message, offset, HASH_LENGTH);
            offset += HASH_LENGTH;

            // bracket this into START_RECORDING(x,y,width,height,md5)/STOP_RECORDING 
            drawingOps.add(recording.startOffset, message);
            allocateDrawingOp(PaintPrimitives.STOP_RECORDING, 1);
            
            perWidgetCache.put(hash, drawHash);
         }
      }
      catch (IOException e)
      {
         // ignore
      }
   }
   
   /**
    * Allocate a new byte array of the proper size, add it to the pending operations list and
    * set the message byte.
    *
    * @param    type
    *           The drawing operation type.
    * @param    len
    *           The message size.
    *
    * @return   The newly allocated message.
    */
   private byte[] allocateDrawingOp(PaintPrimitives type, int len)
   {
      byte[] message = new byte[len];
      
      // WARNING 1: if the PaintPrimitive ever has more than 256 values, this will break
      // WARNING 2: the javascript code will be dependent upon the ordinal values matching,
      //            if new values are inserted in the middle of the enum definition, then
      //            that ordinal and all following ones will need to be fixed up on the
      //            javascript side
      message[0] = (byte) type.ordinal();
      
      addDrawingOp(message);
      
      return message;
   }
   
   /**
    * Get the next message ID.
    * 
    * @return   See above.
    */
   private synchronized int nextMessageId()
   {
      int next = --nextMessageId;
      if (next >= 0)
      {
         nextMessageId = 0;
         next = --nextMessageId;
      }
      
      return next;
   }
   
   /**
    * Add a prepared drawing to the pending operations list.
    * 
    * @param    message
    *           The prepared drawing to transfer.
    */
   private void addDrawingOp(byte[] message)
   {
      drawingOps.add(message);
      pendingBytes += message.length;
      drawingNoCacheOps.add(message);
      pendingNoCacheBytes += message.length;
   }
   
   /**
    * Compute an MD5 hash for the given message.
    * 
    * @param    message
    *           The message with the drawing ops.
    * 
    * @return   A 16-byte array with the MD5 hash.  If the hash could not be computed, return
    *           <code>null</code>.
    */
   private byte[] md5Hash(byte[] message)
   {
      try
      {
         if (MD5 == null)
         {
            MD5 = MessageDigest.getInstance("MD5");
         }

         byte[] drawHash = MD5.digest(message);
         for (int i = 0; i < drawHash.length; i++)
         {
            drawHash[i] = (byte) (drawHash[i] & 0xFF);
         }
         return drawHash;
      }
      catch (NoSuchAlgorithmException e)
      {
         e.printStackTrace();
      }
      
      return null;
   }

   /**
    * Build a byte array with all the drawing operations from the given list.
    * 
    * @param    windowId
    *           The window ID targeted by the drawing ops.
    * @param    ops
    *           The drawing ops.
    * @param    length
    *           The length (in bytes) of all drawing ops.
    *           
    * @return   A byte array containing all the drawing operations, with space reserved for the
    *           MD5 hash.
    */
   private byte[] buildDrawingMessage(int windowId, List<byte[]> ops, int length)
   {
      int numOps = ops.size();
      
      byte[] total = new byte[(numOps * 4) + length + 9 + HASH_LENGTH];
      
      int idx = 0;
      
      total[idx] = MSG_DRAW;
      idx += 1;
      
      // reserve space for the 16-byte MD5 hash 
      idx += HASH_LENGTH;
      
      // write the window id first
      writeMessageInt32(total, idx, windowId);
      idx += 4;
      
      // write the total number of operations in this batch
      writeMessageInt32(total, idx, numOps);
      idx += 4;
      
      for (byte[] message : ops)
      {
         // write the length of the message portion
         writeMessageInt32(total, idx, message.length);
         
         idx += 4;
         
         // copy the message portion
         System.arraycopy(message, 0, total, idx, message.length);
         idx += message.length;
      }
      
      return total;
   }

   /**
    * Defines the possible ways that the clipboard contents can be requested.
    */
   private enum ClipboardRequest
   {
      CLIENT_GENERATED,
      SERVER_GENERATED,
      NONE
   }

   /**
    * Auxiliary class for creating messages
    */
   private class BinaryMessage
   {
      /** current offset */
      private int offset = 0;

      /** message buffer */
      private final byte[] message;
      
      /**
       * Constructor
       * @param type
       *        message type
       * @param len
       *        message length
       */
      public BinaryMessage(byte type, int len)
      {
         message = new byte[len];
         message[offset++] = type;
      }
      
      /**
       * Write a byte 
       * 
       * @param b 
       *        byte to be written
       */
      public void writeByte(byte b)
      {
         message[offset++] = b;
      }
      
      /**
       * Write a boolean value (1 for true, 0 for false).
       * 
       * @param b 
       *        boolean to be written
       */
      public void writeBoolean(boolean b)
      {
         message[offset++] = (byte) (b ? 1 : 0);
      }

      /**
       * Write a 32-bit integer 
       * 
       * @param value 
       *        value to be written
       */
      public void writeInt32(int value)
      {
         offset = writeMessageInt32(message, offset, value); 
      }
      
      /**
       * Write a 16=bit integer 
       * 
       * @param value 
       *        value to be written
       */
      public void writeInt16(int value)
      {
         offset = writeMessageInt16(message, offset, value); 
      }

      /**
       * Write a string 
       * 
       * @param text 
       *        value to be written
       */
      public void writeText(String text)
      {
         offset = writeMessageText(message, offset, text); 
      }

      /**
       * Write a mouse pointer 
       * 
       * @param ptr 
       *        value to be written
       */
      public void writeMousePtr(MousePtrWrapper ptr)
      {
         if (ptr == null)
         {
            message[offset++] = 0;
            return;
         }

         if (ptr.isPredefined())
         {
            message[offset++] = 1;
            writeInt16(ptr.style());
            return;
         }
         message[offset++] = 2;
         String mimeType = ptr.mimeType();
         writeInt16(mimeType.length());
         writeText(mimeType);
         byte[] data = ptr.data();
         writeInt16(data.length);
         System.arraycopy(data, 0, message, offset, data.length);
      }
      
      /**
       * Send the message
       */
      public void send()
      {
         sendBinaryMessage(message);
      }
   }
   
   /**
    * A listener for notification when draw cache entries expire.
    * <p>
    * This will inform the JS-side to remove the drawing cache, from its storage.
    */
   private class DrawCacheExpiryListener
   implements CacheExpiryListener<String, byte[]>
   {
      /** The window ID to which the cache belongs. */
      private final int windowId;
      
      /**
       * Create a new listener, for the specified window.
       * 
       * @param    windowId
       *           The window ID.
       */
      public DrawCacheExpiryListener(int windowId)
      {
         this.windowId = windowId;
      }
      
      /**
       * Notification for expired hashes.
       * <p>
       * It will call the JS side to remove the give hashes, and wait for it to finish.
       */
      @Override
      public void cacheExpired(CacheExpiryEvent<String, byte[]> event)
      {
         LinkedHashMap<String, byte[]> expired = event.getExpiredEntries();
 
         int len = 1;   // type
         len = len + 4; // msg ID
         len = len + 4; // window ID
         len = len + 4; // number of expired caches
         len = len + expired.size() * HASH_LENGTH; // the hashes
         
         int offset = 0;
         byte[] message = new byte[len];
         int msgId = nextMessageId();
         
         message[offset++] = MSG_REMOVE_EXPIRED_HASH;

         writeMessageInt32(message, offset, msgId);
         offset += 4;

         writeMessageInt32(message, offset, windowId);
         offset += 4;

         writeMessageInt32(message, offset, expired.size());
         offset += 4;

         // write the hashes
         for (byte[] hash : expired.values())
         {
            System.arraycopy(hash, 0, message, offset, hash.length);
            offset += hash.length;
         }
         
         sendBinaryMessage(message);
         
         // wait for JS side to process it
         waitForResult(msgId);
      }
   }
   
   /**
    * Internal class containing details about an intermediate recording of the drawing operations.
    */
   private class DrawRecording
   {
      /** The X origin where the drawing will be performed. */
      private final int x;
      
      /** The Y origin where the drawing will be performed. */
      private final int y;
      
      /** The width of the resulted image, which will be captured and cached. */
      private final int width;
      
      /** The height of the resulted image, which will be captured and cached. */
      private final int height;

      /**
       * The start offset in the {@link GuiWebSocket#drawingOps} array, at the time recording was 
       * started.
       */
      private final int startOffset;

      /**
       * The start offset in the {@link GuiWebSocket#drawingNoCacheOps} array, at the time  
       * recording was started.
       */
      private final int startNoCacheOffset;
      
      /**
       * Constructor.
       * 
       * @param    x
       *           The X origin at the time recording was started. 
       * @param    y
       *           The Y origin at the time recording was started.
       * @param    width
       *           The width of the resulted image, which will be captured and cached.
       * @param    height
       *           The height of the resulted image, which will be captured and cached.
       * @param    offset
       *           The start offset in the {@link GuiWebSocket#drawingOps} array, at the time  
       *           recording was started.
       * @param    noCacheOffset
       *           The start offset in the {@link GuiWebSocket#drawingNoCacheOps} array, at   
       *           the time recording was started.
       */
      public DrawRecording(int x, int y, int width, int height, int offset, int noCacheOffset)
      {
         this.x = x;
         this.y = y;
         this.width = width;
         this.height = height;
         this.startOffset = offset;
         this.startNoCacheOffset = noCacheOffset;
      }
   }
}
