/*
** Module   : MouseOps.java
** Abstract : Mapping of DOM mouse events to AWT events.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20150821 First version.
** 002 SBI 20160612 Added mouse system actions: copy, cut, paste, minimize, maximize and close. 
** 003 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 004 HC  20170823 Added runtime support of WINDOW:SHOW-IN-TASKBAR and removed dead code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.client.event.*;

/**
 * A mapping of DOM mouse event IDs (as specified in p2j.screen.js mouseOps) to their corresponding
 * AWT {@link MouseEvent} IDs.
 */
enum MouseOps
 {
   // this can hold a max of 32 entries - as they are bit-encoded on 4 bytes
   CLICK            ( 1, MouseEvent.MOUSE_CLICKED),
   CONTEXTMENU      ( 2, MouseEvent.MOUSE_CLICKED),
   DBLCLICK         ( 3, MouseEvent.MOUSE_CLICKED),
   MOUSEDOWN        ( 4, MouseEvent.MOUSE_PRESSED),
   MOUSEENTER       ( 5, MouseEvent.MOUSE_ENTERED),
   MOUSELEAVE       ( 6, MouseEvent.MOUSE_EXITED),
   MOUSEMOVE        ( 7, MouseEvent.MOUSE_MOVED),
   MOUSEOVER        ( 8, MouseEvent.MOUSE_MOVED),
   MOUSEOUT         ( 9, MouseEvent.MOUSE_EXITED),
   MOUSEUP          (10, MouseEvent.MOUSE_RELEASED),
   WHEEL            (11, MouseEvent.MOUSE_WHEEL),
   DRAG             (12, MouseEvent.MOUSE_DRAGGED),
   HOVERABLE        (13, MouseEvt.MOUSE_HOVERABLE),
   MOVABLE          (14, MouseEvt.MOUSE_MOVABLE),
   POPUPABLE_LEFT   (18, MouseEvt.MOUSE_POPUPABLE_LEFT),
   POPUPABLE_RIGHT  (19, MouseEvt.MOUSE_POPUPABLE_RIGHT),
   COPY             (20, MouseEvt.COPY),
   CUT              (21, MouseEvt.CUT),
   PASTE            (22, MouseEvt.PASTE),
   MINIMIZE         (23, MouseEvt.MINIMIZE),
   MAXIMIZE         (24, MouseEvt.MAXIMIZE),
   CLOSE            (25, MouseEvt.CLOSE);

   /** The event code, as known by the JS code. */
   public final int code;

   /** The corresponding AWT event code. */
   public final int awtEventCode;

   /**
    * Associate the JS and AWT codes for this instance.
    * 
    * @param    code
    *           The event code, as known by the JS code.
    * @param    awtEventCode
    *           The corresponding AWT event code.
    */
   private MouseOps(int code, int awtEventCode)
   {
      this.code = code;
      this.awtEventCode = awtEventCode;
   }
   
   /**
    * Determine which {@link MouseOps} is associated with the given code.
    * 
    * @param    code
    *           The event code, as sent by the JS code.
    *           
    * @return   The associated {@link MouseOps}.
    * 
    * @throws   IllegalArgumentException
    *           If the <code>code</code> does not map to any of the known codes.
    */
   public static MouseOps fromCode(int code)
   {
      for (MouseOps op : MouseOps.values())
      {
         if (op.code == code)
         {
            return op;
         }
      }
      
      throw new IllegalArgumentException("Unknown mouse operation type " + code + "!");
   }
   
   /**
    * Determine if this is a wheel event.
    * 
    * @return    <code>true</code> if this is a mouse wheel event.
    */
   public boolean isWheel()
   {
      return this == WHEEL;
   }
}