/*
** Module   : WebMouseHandler.java
** Abstract : Provides mouse event support for the web GUI driver.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140826 Created initial version.
** 002 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
** 003 SBI 20151201 Fixed to deliver mouse events to their target widgets (JS client has
**                  no information about child-parent relations, only widgets z-order.)
** 004 IAS 20160229 Changed first argument for the  WidgetRegistry.findAncestor() call
** 005 SBI 20160104 Fixed the cursor positioning within the editor.
** 006 SBI 20160523 Changed to execute raiseMouseEvent in the main thread.
** 007 HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
** 008 OM  20170419 Cleanup code. Upgrade to Java8.
** 009 HC  20180422 Fixed menu input events sent to wrong widgets in web GUI.
** 010 CA  20180530 Avoid using widgets not attached to a window, when processing mouse events.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

import java.awt.event.MouseEvent;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provide mouse event support for the web GUI driver.
 */
public class WebMouseHandler
extends MouseHandler
{
   /**
    * Raise a mouse event explicitly.
    * 
    * @param    event
    *           The mouse event to be raised.
    * @param    windowId
    *           Target window id.
    * @param    widgetId
    *           Target widget id.
    */
   public void raiseMouseEvent(MouseEvent event, int windowId, int widgetId)
   {
      if (!canProcessMouse())
      {
         return;
      }
      tc.invokeLaterOS(() ->
      {
         int effectiveWidgetId = widgetId;
         if (widgetId == -1)
         {
            tc.postMouseEvent(event, windowId);
         }
         else
         {
            Widget w = tc.getWidget(widgetId);
            if (w == null || w.ancestor() == null)
            {
               return;
            }

            if (w instanceof AbstractContainer)
            {
               // it needs to get the widget's location with respect to the parent window
               NativePoint origin = w.screenPhysicalLocation();
               Widget target = w.findMouseSource(
                        new NativePoint(event.getX() - origin.x, event.getY() - origin.y));
               if (target != null && target.getId() != null)
               {
                  effectiveWidgetId = target.getId().asInt();
               }
               else
               {
                  // TODO: move this widget-specific logic into methods that are implemented in
                  // the widget hierarchy so that we do not have direct references to widget
                  // classes in this common code
                  
                  Widget editor = WidgetRegistry.findAncestor(target,
                                                t -> EditorGuiImpl.class.equals(t.getClass()));
                  if (editor != null && editor.getId() != null)
                  {
                     effectiveWidgetId = editor.getId().asInt();
                  }
               }
            }
            tc.postMouseEvent(event, windowId, effectiveWidgetId);
         }
      });
   }
}
