/*
** Module   : ContentThumbView.java
** Abstract : This view shows complex file details in a vertical flow (single column of thick
**            thumbnails).
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.
** 002 OM  20180321 Fixed frameId of thumbs. Added Theme and Edit support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import java.nio.file.*;
import java.util.*;

/**
 * This view shows complex file details in a vertical flow (single column). This is the most
 * complex view. Each file thumbnail is rendered as a single line and contains information about
 * file type, file size and the date of last modification.
 */
public class ContentThumbView
extends EmptyThumbView
{
   /** The constant height of the thumbnails. */
   private final int ICON_HEIGHT = 49;
   
   /** The vertical offset where the thumbnails start. */
   private final int VERTICAL_OFFSET = 6;

   /** The horizontal offset where the thumbnails start. */
   private final int HORIZONTAL_OFFSET = 17;
   
   /**
    * Query the preferred scroll types. Each type of view uses a different way of scrolling.
    *
    * @return the preferred scroll type for this thumbnail view.  
    */
   @Override
   public ScrollPaneGuiImpl.ShowBars getScrollType()
   {
      return ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS;
   }
   
   /**
    * Query the preferred scroll bars. Each type of view uses a different way of scrolling.
    *
    * @return the preferred scroll type for this thumbnail view.  
    */
   @Override
   public ScrollBar.Bars getScrollBars()
   {
      return ScrollBar.Bars.VERTICAL;
   }
   
   /**
    * Factory method for creating file thumbnails depending on the current view.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          The widget's configuration.
    * @param   path
    *          The path this thumbnail represents.
    * @param   isDir
    *          Flags the thumbnails that represents directories.
    * @param   click2Command
    *          The action to be performed when the thumbnail is double clicked.
    */
   @Override
   FileThumb createThumb(WidgetId id,
                         ButtonConfig cfg,
                         Path path,
                         boolean isDir,
                         String click2Command)
   {
      cfg.heightPixels = ICON_HEIGHT;
      return new ContentIcon(id, cfg, path, isDir, click2Command);
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      NativeDimension dim = physicalDimension();
      int xOffset = HORIZONTAL_OFFSET;
      int yOffset = VERTICAL_OFFSET;
      
      Widget[] widgets = widgets();
      for (Widget widget : widgets)
      {
         if (widget instanceof ContentIcon)
         {
            ContentIcon mi = (ContentIcon) widget;
            mi.setPhysicalLocation(xOffset, yOffset);
            mi.config().widthPixels = dim.width - xOffset * 2 - 3;
            yOffset += ICON_HEIGHT - 1; // they are overlapped
         }
      }
   }
   
   /**
    * Draw the widget.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      else if (files.isEmpty())
      {
         super.draw();
         return;
      }
      
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      gd.draw(getId(), p, clip, () ->
      {
         // draw background
         gd.setColor(new ColorRgb(0xFFFFFF)); // white
         gd.fillRect(0, 0, dim.width, dim.height);
         
         for (int k = 1; k < widgets().length; k++)
         {
            gd.setColor(new ColorRgb(0xDEDEDE));
            gd.drawLine(HORIZONTAL_OFFSET + 6, VERTICAL_OFFSET + (ICON_HEIGHT - 1) * k, 
                        dim.width - HORIZONTAL_OFFSET - 6, VERTICAL_OFFSET + (ICON_HEIGHT - 1) * k);
         }
         
         // then mimic the super.super.draw() implementation:
         List<Widget<GuiOutputManager>> widgets = prepareWidgets(Arrays.asList(widgets()));
         NativeRectangle nouter = OutputManager.getDriver().getPrimitives()
                                               .getScreenBitmap().getOuterRectangle();
         Rectangle outer = (nouter == null) ? null : 
               screen().coordinates().rectangleFromNative(nouter);
         
         // old method for nested containers
         widgets.forEach(c -> {
            if (c.isVisible())
            {
               Rectangle bounds = c.bounds().translate(childrenLocation(c));
               if (outer == null || bounds.intersects(outer))
               {
                  c.draw();
               }
            }
         });
      });
   }
   
   /**
    * Obtain the name of this view.
    *
    * @return  the name of this view.
    */
   @Override
   public String getViewName()
   {
      return "Content";
   }
   
   /**
    * The thumbnail that shows a file in Content view. The thumbnail has a fixed thick height and
    * will take all available space horizontally. The design combines:
    * <ul>
    *    <li>a custom icon (for the moment we use just two, one for files and one for folders) in
    *       the left side;</li>
    *    <li>the file name (this font is a bit bigger) in the right of icon;</li>
    *    <li>under the file name the file type is shown. We use the simple extension instead of 
    *       the string the extension is registered in Windows. Only for files;</li>
    *    <li>align to the right is the date of last modification and</li>
    *    <li>under it, and only for files, the size on the disk.</li>
    * </ul>
    */
   private class ContentIcon
   extends FileThumb
   {
      /**
       * Constructor used to create component using given configuration.
       *
       * @param   id
       *          The widget's ID.
       * @param   cfg
       *          The widget's configuration.
       * @param   path
       *          The path this thumbnail represents.
       * @param   isDir
       *          Flags the thumbnails that represents directories.
       * @param   click2Command
       *          The action to be performed when the thumbnail is double clicked.
       */
      ContentIcon(WidgetId id, ButtonConfig cfg, Path path, boolean isDir, String click2Command)
      {
         super(id, cfg, path, isDir ? "content-folder" : "content-file", isDir, click2Command);
      }
      
      /**
       * Draws the text and icon on this thumbnail.
       *
       * @param   width
       *          The width of the widget
       * @param   height
       *          The height of the widget.
       * @param   theme
       *          The current theme.
       */
      @Override
      protected void drawForeground(int width, int height, Theme theme)
      {
         int tX = 4;
         if (img != null)
         {
            FileChooserUtils.drawImage(gd, img, 2, (height - img.getHeight()) / 2);
            tX += img.getWidth() + 4;
         }
         
         boolean focused = hasFocus();
         boolean selected = isSelected();
         boolean mouseOver = isMouseOver();
         gd.setColor(theme.getFileChooserForegroundColor(
               gd, focused, selected, mouseOver, true, false, false));
         FontDetails<?> fd = gf.font().font;
         gd.setGuiFont(fd);
         gd.setFontStyle(fd.style);
         int fontNum = gf.resolveFontNum();
         String txt = textLabel();
         
         gd.drawStringCentered(txt, tX, height * 1 / 4);
         
         if (!isDir())
         {
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, focused, selected, mouseOver, false, true, false));
            gd.drawStringCentered("Type:", tX, height * 3 / 4);
            tX += getTextWidth("Type:", fontNum) + 3;
            
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, focused, selected, mouseOver, false, false, true));
            int k = txt.lastIndexOf(".");
            String type = ((k > 0) ? txt.substring(k + 1) : "").toUpperCase() + " File";
            gd.drawStringCentered(type, tX, height * 3 / 4);
         }
         
         tX = width * 4 / 7;
         gd.setColor(theme.getFileChooserForegroundColor(
               gd, focused, selected, mouseOver, false, true, false));
         gd.drawStringCentered("Date modified:", tX, height * 1 / 4);
         tX += getTextWidth("Date modified:", fontNum) + 3;
         gd.setColor(theme.getFileChooserForegroundColor(
               gd, focused, selected, mouseOver, false, false, true));
         gd.drawStringCentered(FileChooserUtils.getDateAsString(created), tX, height * 1 / 4);
         
         if (!isDir())
         {
            tX = width * 4 / 7;
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, focused, selected, mouseOver, false, true, false));
            gd.drawStringCentered("Size: ", tX, height * 3 / 4);
            tX += getTextWidth("Size:", fontNum) + 3;
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, focused, selected, mouseOver, false, false, true));
            gd.drawStringCentered(
                  FileChooserUtils.getSizeAsString(fileSize, false, 2), tX, height * 3 / 4);
         }
      }
      
      /**
       * Obtain the dimension of the editor that when renaming this thumbnail.  
       *
       * @return  the dimension of the editor that when renaming this thumbnail.
       */
      @Override
      protected NativeDimension getEditorDimension()
      {
         NativeDimension d = physicalDimension();
         int tX = 4 + (img != null ? img.getWidth() + 4 : 0);
         return new NativeDimension(d.width * 4 / 7 - tX, d.height / 2);
      }
      
      /**
       * Obtain the physical location of the editor that when renaming this thumbnail. The
       * location is relative to this widget.
       *
       * @return  the physical location of the editor that when renaming this thumbnail.
       */
      @Override
      protected NativePoint getEditorLocation()
      {
         int tX = 4 + (img != null ? img.getWidth() + 4 : 0);
         return new NativePoint(tX, 0);
      }
   }
}

