/*
** Module   : FileDialogGuiImpl.java
** Abstract : a dialog that allows the user to select a file form local disk in GUI mode.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170815 Initial commit. Just a stub with mandatory methods. Work in progress.
** 002 OM  20171026 Added user interface. User can navigate and select files.
** 003 OM  20180321 Various fixes.  Added Theme, new folder, and key navigation support.
** 004 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed.
** 005 OM  20180529 Fixed NPE. Fixed default dir to user's home.
** 006 EVL 20180723 Fixed NPE with not defined initial directory.
**     OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.ActionEvent;
import com.goldencode.p2j.ui.client.event.ItemEvent;
import com.goldencode.p2j.ui.client.event.listener.ItemListener;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.nio.file.*;
import java.util.*;
import java.util.logging.*;
import java.util.stream.Stream;

import static com.goldencode.p2j.ui.client.SysColor.*;
import static com.goldencode.p2j.ui.client.gui.filechooser.FileChooserUtils.*;

/**
 * This class represents a dialog that allows the user to select a file form local disk, in GUI
 * client. The dialog is composed from six rows. Although the dialog should allow the user to
 * resize it, for the moment its dimensions are fixed (yet parametrised).
 * <p>
 * All dimensions are expressed in pixels, rather than character units.
 */
public class FileDialogGuiImpl
extends FileDialogImpl
implements ItemListener
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(FileDialogGuiImpl.class.getName());
   
   /** Cached icon that marks a folder to be expanded, usually a down arrow or [-]. */
   private ImageWrapper<BufferedImage> expImg = null;
   
   /** Cached icon that marks a folder to be collapsed, usually a right arrow or [+]. */
   private ImageWrapper<BufferedImage> colImg = null;
   
   /** The native width of the dialog. */
   private final int NATIVE_WIDTH = 676;
   
   /** The native height of the dialog. */
   private final int NATIVE_HEIGHT = 432;
   
   /** Constant height of the first (upper) pane (navigation). */
   private final int NAV_PANE_HEIGHT = 37;
   
   /** Constant height of the second pane (tools). */
   private final int TOOLS_PANE_HEIGHT = 31;
   
   /** Constant height of the fourth pane (filename direct input) in SaveAs mode. */
   private final int FILENAME_PANE_HEIGHT_SA = 34;
   
   /** Constant height of the fourth pane (filename direct input) in Open mode. */
   private final int FILENAME_PANE_HEIGHT_OP = 37;
   
   /** Constant height of the fourth pane (file type selection). */
   private final int SAVEAS_PANE_HEIGHT = 31;
   
   /** Constant height of the last pane (dialog closing pane) in SaveAs mode. */
   private final int BUTTONS_PANE_HEIGHT_SA = 41;
   
   /** Constant height of the last pane (dialog closing pane) in Open mode. */
   private final int BUTTONS_PANE_HEIGHT_OP = 28;
   
   /** The bottom inset. */
   private final int BOTTOM_INSET = 11;
   
   /** The alignment for file and type labels (on last two rows). */
   private final int FILE_TYPE_ALIGN_SA = 125;
   
   /** The height of the folders in tree. */
   private static final int FOLDER_HEIGHT = 21;
   
   /** The width of the shortcuts panel in center of the dialog in SaveAs mode. */
   private static final int FOLDER_TREE_WIDTH = 175;
   
   /** The width of the shortcuts panel in center of the dialog in Open mode. */
   private static final int FILE_TYPE_WIDTH_OP = 188;
   
   /** Cached driver. */
   protected GuiDriver gd = null;
   
   /** Cached coordinates. */
   private CoordinatesConversion cc = null;
   
   /** The upper panel that allows basic navigation primitives. */
   private NavigationPane navigationPane = null;
   
   /** Button that allows the user to go back to previous location. */
   private SimpleImageButton navBack = null;
   
   /** Button that allows the user to advances from a 'go back' action. */
   private SimpleImageButton navForward = null;
   
   /**
    * Button that allows the user select the most recent location from a popup.
    */
   private SimpleImageButton recentLocations = null;
   
   /**
    * Button that allows the user to go up the file tree from the current location.
    */
   private ButtonGuiImpl folderUp = null;
   
   /**
    * The {@code addressCbx} combobox allows the user to select one of previously visited locations
    * or to manually enter it.
    */
   private ComboBoxGuiImpl addressCbx = null;
   
   /**
    * Refreshes the content of the displayed folder (if the content was altered externally).
    */
   private ButtonGuiImpl refresh = null;
   
   /**
    * Allows to enter a pattern or select a previous one for search inside current location.
    */
   private ComboBoxGuiImpl search = null;
   
   /** Is search in progress ? */
   private volatile SearchOperation searchOp = null; 
   
   /**
    * Groups a set of tool buttons. This can be noticed by reddish background.
    */
   private ToolsPane toolsPane = null;
   
   /**
    * The button opens a contextual popup menu with low-level OS operations: Cut/Copy/Paste/Undo/
    * Redo/Delete/Rename.
    * Note: Not implemented yet.
    */
   private ButtonGuiImpl organize = null;
   
   /**
    * The button allows creation of a new empty folder and prompts the user to rename it.
    */
   private ButtonGuiImpl newFolder = null;
   
   /**
    * Loops over the existing view modes: Medium Icons, Small Icons, List, Details, Tiles and
    * Content. The Extra large icons and Large Icons are not implemented.
    */
   private ButtonGuiImpl view = null;
   
   /**
    * Allows selection of same view modes like {@code view} button but using a popup menu instead
    * of looping.
    */
   private ButtonGuiImpl more = null;
   
   /**
    * Opens native contextual help.
    * Note: Not implemented yet.
    */
   private ButtonGuiImpl help = null;
   
   /**
    * The container that holds the tree shortcuts accessible from the left-side of the dialog.
    */
   private FolderTree treePane = null;
   
   /**
    * The scroll pane that allows the {@code treePane} to be scrolled up and down.
    */
   private ScrollPaneGuiImpl treeScroll = null;
   
   /** Root node that groups the {@code Favorite} shortcuts. */
   private Folder scFavorites;
   
   /** Root node that groups the shortcuts to locations from {@code this computer}. */
   private Folder scThisPC;
   
   /** Root node that groups the {@code Network} shortcuts. */
   private Folder scNetwork;
   
   /** The folder that is currently selected. */
   private Folder crtFolder = null;
   
   /** The set of view modes for currently selected directory. */
   private EmptyThumbView[] views = new EmptyThumbView[6];
   
   /**
    * The set of view scroll panes for each view container. The set of scrollbar is different for
    * each view.
    */
   private ScrollPaneGuiImpl[] viewScrolls = new ScrollPaneGuiImpl[6];
   
   /** The currently selected view mode. */
   private int currentView = 0;
   
   /** A label for filename editor. */
   private SimpleLabel filenameLabel = null;
   
   /**
    * The filename editor should allow the user to manually enter the name of the file.
    */
   private LineEditor filename = null;
   
   /** A label for file type chooser. */
   private SimpleLabel saveasLabel = null;
   
   /**
    * The combobox allows the user to select a predefined pattern for the files displayed.
    * The content of the current directory is automatically filtered when a pattern is chosen.
    */
   private ComboBoxGuiImpl fileTypeCbx = null;
   
   /**
    * Reduce the size of the dialog by hiding the central widgets (shortcuts and directory
    * content).
    * Note: Not implemented yet.
    */
   private ButtonGuiImpl hideFolders = null;
   
   /** The {@code OK} button. Validates the current selection and closes the dialog. */
   private ButtonGuiImpl accept = null;
   
   /** The {@code Cancel} button. Rejects the dialog. */
   private ButtonGuiImpl reject = null;
   
   /** The list of visited locations accessible through {@code Back} button. */
   private Deque<String> history = new LinkedList<>();
   
   /**
    * The list of visited locations that have been reverted using {@code Back} button. To access
    * them use the {@code Forward} navigation button.
    */
   private Deque<String> future = new LinkedList<>();
   
   /** The default file filter in case none is specified. */
   private String activeFilter = "*.*";
   
   /** The cached list of files at the current location. */
   private List<Path> currentDirContent = new ArrayList<>(23);
   
   /** The absolute path of the directory displayed in dialog. */
   private String currentDir;
   
   /**
    * The Editor widget that allows renaming of directories just created.
    * (TODO: and other files, when the Rename (F2) action will be enabled). 
    */
   private LineEditor rename = null; 
   
   /**
    * The constructor build an object by forwarding the required parameters to super class.
    *
    * @param   id
    *          The {@link WidgetId} of the dialog. 
    * @param   cfg
    *          The {@link FrameConfig} of the dialog.
    */
   public FileDialogGuiImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
      setLayout(new FileDialogLayoutManager());
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    */
   protected void init()
   {
      if (gd != null)
      {
         return; // already initialized
      }
      
      // install system close handler
      super.init();
      
      gd = (GuiDriver<?, ?>) OutputManager.getDriver();
      cc = screen().coordinates();
      
      // build navigation row
      {
         navigationPane = new NavigationPane();
         addWidget(navigationPane);
         /* 
          * Navigation row contains the following widgets:
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    HistoryBack       button      round          Loads the previous folder
          *    HistoryForward    button      round          Loads the forward folder 
          *    RecentLocations   button                     Opens a popup-menu with recent locations
          *    FolderUp          button                     Loads the parent folder
          *    Address           editor      expandable     Allows user to type in a location
          *    Refresh           button                     Reload the current location
          *                      /splitter/                 Changes ratio between Address and Search
          *    Search            editor      expandable     Searches sub-folders for specific pattern
          */
         navBack = createImageButton("BACK", "Back", 28, 27);
         addWidgetToPane(navBack, navigationPane);
         
         navForward = createImageButton("FORWARD", "Forward", 28, 27);
         addWidgetToPane(navForward, navigationPane);
         
         recentLocations = createImageButton("HISTORY", "Recent locations", 13, 19);
         addWidgetToPane(recentLocations, navigationPane);
         
         folderUp = createButton("", "UP", 24, 24, true, false, "Up one level");
         folderUp.setImage(ButtonConfig.IMAGE_UP, getBinaryResource("up-en"));
         addWidgetToPane(folderUp, navigationPane);
         
         addressCbx = createCombobox(ComboBoxConfig.Mode.DROP_DOWN, 107, 6, 277, 24,
                                     new ControlSetItem[0], null, "x(255)");
         addressCbx.addItemListener(this);
         addWidgetToPane(addressCbx, navigationPane);
         addressCbx.setVisible(true); // must be called AFTER addWidget() 
         
         refresh = createButton("", "REFRESH", 25, 22, true, false, "Refresh");
         refresh.setImage(ButtonConfig.IMAGE_UP, getBinaryResource("refresh"));
         addWidgetToPane(refresh, navigationPane);
         
         search = createCombobox(ComboBoxConfig.Mode.DROP_DOWN, 394, 6, 204, 23,
                                 new ControlSetItem[0], null, "x(50)");
         search.addItemListener(this);
         addWidgetToPane(search, navigationPane);
         search.setVisible(true);
      }
      
      // build tools row
      {
         toolsPane = new ToolsPane();
         addWidget(toolsPane);
         /*
          * Tools row contains the following widgets:
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    Organize          button                     Opens a contextual popup-menu
          *    NewFolder         button                     Creates a new folder at current location
          *                      /spacer/    expandable     Nothing (fills in empty space)
          *    ChangeView        button                     Selects the next view mode
          *    MoreOptions       button                     Opens a popup-menu with different view modes
          *    Help              button                     Opens Help page
          */
         
         organize = createButton("Organize", "ORGANIZE", 91, 26, true, false,
                                 "Organize this folder's content.");
         organize.setEnabled(false);
         addWidgetToPane(organize, toolsPane);
         
         newFolder = createButton("New Folder", "NEW-FOLDER", 88, 26, true, false,
                                  "Create a new folder.");
         addWidgetToPane(newFolder, toolsPane);
         
         view = createButton("", "CHANGE-VIEW", 27, 26, true, false, "Change your view.");
         view.setImage(ButtonConfig.IMAGE_UP, getBinaryResource("view-en"));
         addWidgetToPane(view, toolsPane);
         
         more = createButton("", "MORE-OPTIONS", 19, 26, true, false, "More options");
         more.setImage(ButtonConfig.IMAGE_UP, getBinaryResource("more-en"));
         addWidgetToPane(more, toolsPane);
         
         help = createButton("", "HELP", 26, 26, true, false, "Get help.");
         help.setImage(ButtonConfig.IMAGE_UP, getBinaryResource("help"));
         addWidgetToPane(help, toolsPane);
      }
      
      // build central content row
      {
         /*
          * The central container row contains the following widgets:
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    FoldersTree       treeview                   Contains 3 trees: Favorites/This PC/Network
          *                      /splitter/                 Changes ratio between Folders and Contents
          *    FolderContent     browse/list                List the files and folder at current location
          */
         
         treePane = new FolderTree();
         treePane.setScrollable(true);
         treeScroll = new ScrollPaneGuiImpl(treePane);
         treeScroll.setShowBars(ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS);
         treeScroll.setScrollBars(ScrollBar.Bars.VERTICAL);
         addWidget(treeScroll);
         
         expImg = loadImage("expanded");
         colImg = loadImage("collapsed");
         String userHome = System.getProperty("user.home") + File.separator;
         
         scFavorites = createFolder(
               null, getDisplayName(VF_FAVORITES), VF_FAVORITES, "favorites");
         scFavorites.content = new Folder[3];
         scFavorites.content[0] = createFolder(
               scFavorites, "Desktop", userHome + "Desktop", "desktop");
         scFavorites.content[1] = createFolder(
               scFavorites, "Downloads", userHome + "Downloads", "downloads");
         scFavorites.content[2] = createFolder(
               scFavorites, getDisplayName(VF_RECENTPLACES), VF_RECENTPLACES, "recent");
         scFavorites.content[2].content = new Folder[0];
         
         scThisPC = createFolder(null, getDisplayName(VF_THISPC), VF_THISPC, "thispc");
         File[] roots = File.listRoots();
         scThisPC.content = new Folder[6 + roots.length];
         scThisPC.content[0] = createFolder(scThisPC, "Desktop", userHome + "Desktop", "desktop");
         scThisPC.content[1] = createFolder(
               scThisPC, "Documents", userHome + "Documents", "documents");
         scThisPC.content[2] = createFolder(
               scThisPC, "Downloads", userHome + "Downloads", "downloads");
         scThisPC.content[3] = createFolder(scThisPC, "Music", userHome + "Music", "music");
         scThisPC.content[4] = createFolder(
               scThisPC, "Pictures", userHome + "Pictures", "pictures");
         scThisPC.content[5] = createFolder(scThisPC, "Videos", userHome + "Videos", "videos");
         /*
           The set of shortcuts for each root of file system. On Linux there is only one 
           ({@code /}) while on Windows there is a letter entry for each drive.
          */
         for (int k = 0; k < roots.length; k++)
         {
            scThisPC.content[6 + k] = createFolder(
                  scThisPC, "Local Disk (" + roots[k] + ")", roots[k].getAbsolutePath(), "drive");
         }
         
         scNetwork = createFolder(null, getDisplayName(VF_NETWORK), VF_NETWORK, "network");
         scNetwork.content = new Folder[0];
         
         treePane.doLayout();
         
         views[0] = new MediumIconsThumbView();
         views[1] = new SmallIconsThumbView();
         views[2] = new ListThumbView();
         views[3] = new DetailsThumbView();
         views[4] = new TilesThumbView();
         views[5] = new ContentThumbView();
         
         for (int k = 0; k < viewScrolls.length; k++)
         {
            views[k].setListener(this);
            views[k].config().frameId = getId().asInt();
            
            viewScrolls[k] = new ScrollPaneGuiImpl(views[k]);
            viewScrolls[k].setShowBars(views[k].getScrollType());
            viewScrolls[k].setScrollBars(views[k].getScrollBars());
            viewScrolls[k].setVisible(false);
            addWidget(viewScrolls[k]);
         }
         viewScrolls[0].setVisible(true);
      }
      
      // build filename row:
      {
         /*
          * The filename row contains the following widgets:
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    FileName          label       fixed          Nothing (right-aligned)
          *    FileNameCb        combobox    expandable     Allows user to type in a filename 
          */
         
         filenameLabel = createSimpleLabel("File &name:", FOLDER_TREE_WIDTH, 22, true);
         addWidget(filenameLabel);
         
         WidgetId fid = WidgetId.nextID();
         FillInConfig fcfg = new FillInConfig(fid);
         fcfg.format = "x(255)";
         fcfg.state = ScreenBuffer.CHANGED;
         fcfg.autoResize = false;
         fcfg.realized = true;
         fcfg.readOnly = false;
         try
         {
            fcfg.appFormat = new StringFormat("x(255)");
         }
         catch (DisplayFormatParsingException e)
         {
            // just in case, no expected exceptions for this format
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING, "Unexpected StringFormat", e);
            }
         }
         
         filename = new LineEditor(fid, fcfg);
         filename.initialize(fid, fcfg);
         filename.setDataType("character");
         filename.setValidateOnFocusLost(false); // firing that will close the dialog
         filename.setEnabled(true);
         filename.setVisible(true);
         filename.setForceLeft(true);
         filename.addListener(this);
         addWidget(filename);
      }
      
      // build save-as row:
      {
         /*
          * Contains the following widgets:  
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    SaveType          label       fixed          Nothing (right-aligned)
          *    SaveAsCb          combobox    expandable     Allows user to select a filter/type
          */
         if (saveAs)
         {
            saveasLabel = createSimpleLabel("Save as &type:", 122, 22, true);
            addWidget(saveasLabel);
         }
         
         ControlSetItem[] filtersCSI = null;
         if (filters != null && filters.length > 1)
         {
            filtersCSI = new ControlSetItem[filters.length / 2];
            for (int i = 0; i < filtersCSI.length; i++)
            {
               filtersCSI[i] = new ControlSetItem(i,
                                                  new character(filters[2 * i]),
                                                  new character(filters[2 * i + 1]));
            }
         }
         else
         {
            // use generic filter:
            filtersCSI = new ControlSetItem[1];
            filtersCSI[0] = new ControlSetItem(0,
                                               new character("All files (*.*)"),
                                               new character("*.*"));
         }
         BaseDataType defaultSelected = filtersCSI[initialFilter >= 0 &&
               initialFilter < filtersCSI.length ? initialFilter : 0].getValue();
         activeFilter = defaultSelected.toStringMessage();
         fileTypeCbx = createCombobox(ComboBoxConfig.Mode.DROP_DOWN_LIST,
                                      0, 0, 188, 23, filtersCSI, activeFilter, "x(99)");
         fileTypeCbx.setValue(new character(defaultSelected));
         fileTypeCbx.valueChangedCommon();
         fileTypeCbx.addItemListener(this);
         fileTypeCbx.setEnabled(true);
         
         addWidget(fileTypeCbx);
         OutputManager.instance().getRegistry().addWidget((Widget) fileTypeCbx);
      }
      
      // build bottom buttons row:
      {
         /*
          * The bottom buttons row contains the following widgets:
          *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
          *    HideFolders       button                     Toggles visibility of Tools and Content panes
          *                      /spacer/    expandable     Nothing (fills in empty space)
          *    OK                button                     Accepts the current selection
          *    Cancel            button                     Rejects the dialog
          */
         if (saveAs)
         {
            hideFolders = createButton("Hide Folders", "TOGGLE-FOLDERS", 94, 23, true, false, null);
            hideFolders.setEnabled(false);
            addWidget(hideFolders);
         }
         
         accept = createButton(saveAs ? "Save" : "Open", "OK", 88, 26, false, true, null);
         config().defaultButtonId = accept.getId().asInt();
         addWidget(accept);
         
         reject = createButton("Cancel", "CANCEL", 88, 26, false, false, null);
         config().cancelButtonId = reject.getId().asInt();
         addWidget(reject);
      }
      
      screen().getRegistry().addFrame(this);
      setSize(cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH, NATIVE_HEIGHT)));
      getContentPane().setSize(dimension());
      
      // for directory search backup selection
      String userHome = System.getProperty("user.home");
      
      // to avoid NPE set up backup initial directory if required
      if (initialDir == null || initialDir.isEmpty())
      {
         setInitialDir(userHome + File.separator);
      }
      
      // everything done, go to initial dir (select current file):
      File initFile = new File(initialDir);
      if (initFile.exists())
      {
         currentDir = initFile.isFile() ? initFile.getParent() : initialDir;
      }
      else
      {
         // check whether the parent exists
         if (initFile.getParent() == null)
         {
            currentDir = userHome;
         }
         else
         {
            File parentFile = new File(initFile.getParent());
            if (parentFile.exists() && parentFile.isDirectory())
            {
               currentDir = parentFile.getAbsolutePath();
            }
            else
            {
               // defaulting to user's home
               currentDir = System.getProperty("user.home");
            }
         }
      }
      
      {
         // prepare the renaming widget
         WidgetId fid = WidgetId.nextID();
         FillInConfig fcfg = new FillInConfig(fid);
         fcfg.format = "x(255)";
         fcfg.state = ScreenBuffer.CHANGED;
         fcfg.autoResize = false;
         fcfg.realized = true;
         fcfg.readOnly = false;
         fcfg.frameId = getId().asInt();
         try
         {
            fcfg.appFormat = new StringFormat("x(255)");
         }
         catch (DisplayFormatParsingException e)
         {
            // just in case, no expected exceptions for this format
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING, "Unexpected StringFormat", e);
            }
         }
         
         rename = new LineEditor(fid, fcfg);
         rename.initialize(fid, fcfg);
         rename.setDataType("character");
         rename.setEnabled(true);
         rename.setForceLeft(true);
         rename.addListener(this);
         rename.setVisible(false);
         GuiOutputManager.instance().getRegistry().addWidget(rename);
      }
      
      doLayout(); // make sure the view have their size before populating them in navigate() 
      navigate(currentDir);
      updateNavButtons();
      updateTabbingList();
   }
   
   /**
    * Updates the list that dictates the focusing order using TAB/BACK-TAB keys.
    */
   private void updateTabbingList()
   {
      List<Widget<GuiOutputManager>> tabList = new ArrayList<>(12);
      
      tabList.add(addressCbx);
      tabList.add(search);
      if (organize != null)
      {
         tabList.add(organize);
      }
      if (crtFolder != null)
      {
         tabList.add(crtFolder);                // last selected folder from treePane
      }
      if (FileThumb.getSelected() != null)
      {
         tabList.add(FileThumb.getSelected());  // last selected file-thumb
      }
      tabList.add(rename);                      // automatically skipped if not visible/enabled
      tabList.add(filename);
      tabList.add(fileTypeCbx);
      if (hideFolders != null)
      {
         tabList.add(hideFolders);
      }
      tabList.add(accept);
      tabList.add(reject);
      
      getContentPane().setTabItemList(tabList);
      setTabItemList(tabList);
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  Always {@code true}.
    */
   @Override
   public boolean focusTraversable()
   {
      return true;
   }
   
   /**
    * This method is invoked when action event is generated.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
      String cmd = event.command();
      Widget source = event.source();
      if (cmd == null)
      {
         return; // unknown command
      }
      if (source instanceof FileThumb)
      {
         FileThumb thumb = (FileThumb) source;
         if (thumb.isDoubleClick())
         {
            cmd = thumb.getDoubleClickCommand();
         }
      }
      
      Folder oldFolder = crtFolder;
      switch (cmd)
      {
         case "BACK":
            goBack();
            return;
         case "FORWARD":
            goForward();
            return;
         case "UP":
            goUp();
            return;
         case "HISTORY":
            showHistory();
            return;
         case "REFRESH":
            refresh();
            return;
         case "MORE-OPTIONS":
            moreOptions();
            return;
         case "NEW-FOLDER":
            createNewFolder();
            return;
         case "RENAME":
            if (source instanceof FileThumb)
            {
               FileThumb.select((FileThumb) source); // probably a no-op
               FileThumb.startRename(rename);
               config().defaultButtonId = -1; // disable default button while in rename mode
               updateTabbingList();
            }
            return;
         case "ESCAPE":
            if (searchOp != null)
            {
               // stop the search operation
               searchOp = null;
            }
            else
            {
               // morph the action into [CANCEL] 
               onAction(new ActionEvent(source, "CANCEL"));
            }
            return;
            
         case LineEditor.AC_TEXT_VALIDATED:
            if (source == filename)
            {
               validateSelection();
            }
            else if (source == rename)
            {
               rename.setVisible(false);
               
               // restore default button while in rename mode
               config().defaultButtonId = accept.getId().asInt();
               String newName = rename.getText();
               FileThumb.getSelected().requestFocus();
               FileThumb.getSelected().setTextLabel(newName);
               renameCurrentFile(newName);
            }
            return;
            
         case LineEditor.AC_CANCEL_EDIT:
            rename.setVisible(false);
            
            // restore default button while in rename mode
            config().defaultButtonId = accept.getId().asInt();
            FileThumb.getSelected().requestFocus();
            return;
         
         case "OK":
            // cancel search background job
            searchOp = null;
            
            if (source == accept)
            {
               FileThumb selThumb = FileThumb.getSelected();
               if (selThumb != null && selThumb.isDir())
               {
                  // recursive call with adjusted command
                  onAction(new ActionEvent(source, "REL-NAVIGATE " + selThumb.textLabel()));
                  return;
               }
            }
            String file = filename.getText();
            if (file == null || file.isEmpty())
            {
               // cannot accept empty file
               return;
            }
            
            validateSelection();
            return;
         
         case "CANCEL":
            setSelectedPath(null);
            accepted = false;
            dismiss();
            return;
         
         case "CHANGE-VIEW":
            int newView = currentView + 1;
            changeView(newView >= views.length ? 0 : newView);
            return;
         
         default:
            if (cmd.startsWith("REL-NAVIGATE "))
            {
               // cancel search background job
               searchOp = null;
               
               String relativeLocation = cmd.substring(13);
               if (isVirtualFolder(relativeLocation))
               {
                  openFolder(relativeLocation, true);
                  return;
               }
               // relative location to [currentDir] 
               String newPath = (isVirtualFolder(currentDir) && source instanceof FileThumb) 
                     ? ((FileThumb) source).getPath().toAbsolutePath().toString()
                     : currentDir + File.separator + relativeLocation;
               
               File newFile = new File(newPath);
               if (newFile.isDirectory())
               {
                  future.clear();
                  addToHistory(currentDir);
                  navigate(newPath);
                  assert (Objects.equals(currentDir, newPath));
                  updateNavButtons();
               }
               else
               {
                  validateSelection();
               }
               
               crtFolder = null;
               updateTabbingList();
               if (oldFolder != null)
               {
                  oldFolder.repaint();
               }
            }
            else if (cmd.startsWith("SELDIR "))
            {
               updateTabbingList();
            }
            else if (cmd.startsWith("SELECT "))
            {
               filename.setValue(new character(cmd.substring(7)));
               updateTabbingList();
            }
            else if (cmd.startsWith("OPEN "))
            {
               openFolder(cmd.substring(5), true);
               crtFolder = (source instanceof Folder) ? (Folder) source : null;
               updateTabbingList();
               if (oldFolder != null)
               {
                  oldFolder.repaint();
               }
            }
      }
   }
   
   /**
    * Renames the file whose thumbnail is currently selected. This is the final step of the make
    * folder or rename action, after the rename editor has returned control.
    * 
    * @param   newName
    *          The new name for the selected file. 
    */
   private void renameCurrentFile(String newName)
   {
      Path oldPath = FileThumb.getSelected().getPath();
      String oldAbsName = oldPath.toAbsolutePath().toString();
      File oldFile = oldPath.toFile();
      String newAbsName = oldPath.getParent()
                                 .toAbsolutePath()
                                 .toString() + File.separator + newName;
      if (Objects.equals(oldAbsName, newAbsName))
      {
         return; // no change
      }
      
      File newFile = new File(newAbsName);
      if (newFile.exists())
      {
         ThinClient tc = ThinClient.getInstance();
         logical res = null;
         if (FileThumb.getSelected().isDir())
         {
            Object[] msgs = {
                  "The destination already contains a folder named '" + newName + "'.\n\n",
                  "If any files have the same names, you will be asked if you want to " +
                        "replace those files.\n\n",
                  "Do you still want to replace this folder '" + 
                        FileThumb.getSelected().textLabel() + "' with this one '" + 
                        newName + "'?"};
            res = tc.messageBox(msgs,
                                null,
                                LogicalTerminal.ALERT_QUESTION,
                                LogicalTerminal.BTN_YES_NO,
                                "Confirm Folder Replace",
                                null,
                                WindowConfig.RESOLVE_WINDOW,
                                null);
         }
         else
         {
            Object[] msgs = {
                  "Do you want to rename \"" + FileThumb.getSelected().textLabel() + "\" " +
                  "to \"" + newName + "\"?",
                  "There is already a file with the same name in this location."
            };
            res = tc.messageBox(msgs,
                                null,
                                LogicalTerminal.ALERT_QUESTION,
                                LogicalTerminal.BTN_YES_NO,
                                "Rename File",
                                null,
                                WindowConfig.RESOLVE_WINDOW,
                                null);
         }
         
         if (res.isUnknown() || !res.booleanValue())
         {
            return; // no-apply
         }
      }
      
      oldFile.renameTo(newFile);
      refresh();
   }
   
   /**
    * Validates the current selection. Checks constraints of file that were set when dialog was
    * created and accepts the dialog if all are met.
    */
   private void validateSelection()
   {
      String selFile = filename.getText();
      File f = new File(currentDir + File.separator + selFile);
      ThinClient tc = ThinClient.getInstance();
      if (!saveAs && mustExist && !f.exists())
      {
         Object[] msgs = {
                  selFile, 
                  "\nFile not found.", 
                  "\nCheck the file name and try again."};
         tc.messageBox(msgs,
                       null,
                       LogicalTerminal.ALERT_WARNING,
                       LogicalTerminal.BTN_OK,
                       config().title,
                       null,
                       WindowConfig.RESOLVE_WINDOW,
                       null);
         
         return;
      }
      
      if (saveAs && askOverwrite && f.exists())
      {
         Object[] msgs = {selFile + " already exist.", "\nDo you want to replace it?"};
         logical res = tc.messageBox(msgs,
                                     null,
                                     LogicalTerminal.ALERT_WARNING,
                                     LogicalTerminal.BTN_YES_NO,
                                     "Confirm Save As",
                                     null,
                                     WindowConfig.RESOLVE_WINDOW,
                                     null);
         if (res.isUnknown() || !res.booleanValue())
         {
            return;
         }
         // otherwise the user accepted file overwrite so we can safely return the selected file
      }
      
      // all checks passed, we can save the selected file and close the dialog
      setSelectedPath(f.getAbsolutePath());
      accepted = true;
      dismiss();
   }
   
   /**
    * Adds a folder to location history. If the last entry is {@code location} this is a NO-OP.
    * 
    * @param   location
    *          The location to add.
    */
   private void addToHistory(String location)
   {
      if (history.isEmpty() || !Objects.equals(history.getLast(), location))
      {
         history.addLast(location);
      }
   }
   
   /**
    * Let the modal UI element now that it should finish its modal input processing.
    */
   @Override
   public void dismiss()
   {
      searchOp = null; // cancel search background job
      super.dismiss();
   }
   
   /**
    * Opens a folder.
    * 
    * @param   path
    *          The absolute path of the folder or a name of a virtual folder (starts with 
    *          {@code $} sign).
    * @param   clearForward
    *          If {@code true} clears the forward stack (only in case of virtual folders).
    */
   private void openFolder(String path, boolean clearForward)
   {
      addToHistory(currentDir);
   
      if (isVirtualFolder(path))
      {
         if (clearForward)
         {
            future.clear();
         }
         
         searchOp = null;
         currentDirContent.clear();
         // populate view with virtual content
         switch (path)
         {
            case VF_FAVORITES:
               Arrays.spliterator(scFavorites.content).forEachRemaining(
                     f -> currentDirContent.add(Paths.get(f.path)));
               break;
               
            case VF_RECENTPLACES:
               //TODO: use addressCbx.config().items
               ComboBoxModel<String> model = addressCbx.model();
               model.iterator().forEachRemaining(s ->
                   {
                      // avoid RECENT_PLACES in RECENT_PLACES folder 
                      if (!VF_RECENTPLACES.equals(s))
                      {
                         currentDirContent.add(Paths.get(s));
                      }
                   });
               break;
               
            case VF_THISPC:
               Arrays.spliterator(scThisPC.content).forEachRemaining(
                     f -> currentDirContent.add(Paths.get(f.path)));
               break;
               
            case VF_NETWORK:
               break;
         }
         currentDir = path;
         views[currentView].populate(currentDirContent);
         viewScrolls[currentView].repaint();
      }
      else
      {
         future.clear();
         navigate(path);
         assert (Objects.equals(currentDir, path));
      }
      updateNavButtons();
   }
   
   /**
    * Expands a folder.
    * 
    * @param   folder
    *          The {@code Folder} to be expanded.
    */
   private void expandFolder(Folder folder)
   {
      if (folder == null)
      {
         return;
      }
   
      boolean doLayout = false;
      if (folder.content == null)
      {
         if (!isVirtualFolder(folder.path))
         {
            Path newPath = Paths.get(folder.path);
            try
            {
               Object[] dirs = Files.list(newPath)
                                    .filter(FileChooserUtils::isPathVisible)
                                    .filter(Files::isDirectory)
                                    .toArray();
               folder.content = new Folder[dirs.length];
               for (int k = 0; k < folder.content.length; k++)
               {
                  Path path = (Path) dirs[k];
                  folder.content[k] = createFolder(folder,
                                                   getFileName(path),
                                                   path.toAbsolutePath().toString(),
                                                   "small-folder");
               }
               doLayout = true;
            }
            catch (IOException e)
            {
               if (LOG.isLoggable(Level.WARNING))
               {
                  LOG.log(Level.WARNING, "Failed to explore [" + folder.path + "]", e);
               }
               folder.content = new Folder[0];
            }
         }
      }
   
      if (!folder.expanded)
      {
         doLayout = true;
      }
      folder.expanded = true;
      
      if (doLayout)
      {
         treePane.doLayout();
         treeScroll.repaint();
      }
   }
   
   /**
    * Creates a new folder at the current location.
    */
   private void createNewFolder()
   {
      File newFolder = new File(currentDir + File.separator + "New Folder");
      boolean res = false;
      if (!newFolder.exists())
      {
         try
         {
            res = newFolder.mkdir();
         }
         catch (Exception e)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING, "Cannot create new folder in [" + currentDir + "]", e);
            }
            return;
         }
      }
      else
      {
         int k = 1;
         do
         {
            newFolder = new File(currentDir + File.separator + "New Folder (" + (++k) + ")");
            if (!newFolder.exists())
            {
               try
               {
                  res = newFolder.mkdir();
               }
               catch (Exception e)
               {
                  if (LOG.isLoggable(Level.WARNING))
                  {
                     LOG.log(Level.WARNING, "Cannot create new folder in [" + currentDir + "]", 
                             e);
                  }
                  return;
               }
            }
         }
         while (!res && k < 999);
      }
      
      if (res)
      {
         // folder successfully created: refresh and enter editing mode:
         refresh();
         FileThumb ft = views[currentView].locate(newFolder.getName());
         if (ft == null)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING,
                       "Oops, I created[" + newFolder.getName() + "] but I lost it.");
            }
            return;
         }
         FileThumb.select(ft);
         FileThumb.startRename(rename);
         updateTabbingList();
      }
   }
   
   /**
    * Changes the current view mode. Drops the content of the current view, load the new view and
    * populates it.
    *
    * @param newView The index of the view in {@code views}.
    */
   private void changeView(int newView)
   {
      if (currentView == newView || newView >= views.length)
      {
         return; // no-op
      }
      
      // drop content from current view:
      FileThumb.clearSelection();
      views[currentView].populate(null);
      viewScrolls[currentView].setVisible(false);
      
      // skip to next view:
      currentView = newView;
      
      // repopulate the new view:
      views[currentView].populate(currentDirContent);
      viewScrolls[currentView].setVisible(true);
      repaint();
      // view changed to views[currentView].getViewName())
      
      updateTabbingList();
   }
   
   /**
    * This method is invoked on each item selection change.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onSelect(ItemEvent event)
   {
      Widget source = event.source();
      if (source == addressCbx)
      {
         String newLocation = addressCbx.getScreenValue().toStringMessage();
         if (Objects.equals(newLocation, currentDir))
         {
            // this is the recoil from setting the [currentDir] to [addressCbx]. This should not
            // happen, only user-interaction should trigger this event.
            return;
         }
         future.clear();
         addToHistory(currentDir);
         navigate(newLocation);
         assert (Objects.equals(currentDir, newLocation));
         updateNavButtons();
      }
      else if (source == search)
      {
         String pattern = search.getScreenValue().toStringMessage().toLowerCase();
         if (pattern.isEmpty())
         {
            navigate(currentDir); // restore view content
            return;
         }
         
         currentDirContent = new ArrayList<>(); 
         views[currentView].populate(currentDirContent);
         viewScrolls[currentView].repaint();
         
         // cancels search background job automatically
         searchOp = new SearchOperation(Paths.get(currentDir), pattern);
      }
      else if (source == filename)
      {
         // TODO [filename.getText())] selected
      }
      else if (source == fileTypeCbx)
      {
         ComboBoxModel<String> model = fileTypeCbx.model();
         String newFileType = model.selected().trim();
         int idx = model.indexOf(newFileType);
         String newFilter = (filters != null && filters.length > 2 * idx + 1)
               ? filters[2 * idx + 1] : "*.*";
         if (!Objects.equals(newFilter, activeFilter))
         {
            activeFilter = newFilter;
            refresh();
         }
      }
   }
   
   /**
    * The method called when {@code history} button was pressed.
    */
   private void showHistory()
   {
      WidgetId mid = WidgetId.nextID();
      MenuConfig mc = new MenuConfig();
      mc.popupOnly = true;
      mc.ownerId = getId().asInt();
      
      MenuGuiImpl menu = new MenuGuiImpl(mid, mc);
      menu.initialize(mid, mc);
      final int[] k = new int[] { -future.size() }; // well, avoid final attribute on scalar int
      future.descendingIterator().forEachRemaining(s -> menu.add(createHistoryMenu(s, k[0]++)));
      menu.add(createHistoryMenu(currentDir, k[0]++));
      history.descendingIterator().forEachRemaining(s -> menu.add(createHistoryMenu(s, k[0]++)));
      
      NativePoint p = recentLocations.screenPhysicalLocation();
      p.x = 0;
      p.y += recentLocations.physicalDimension().height;
      menu.showPopup(this, p, null);
   }
   
   /**
    * Worker method for creating a history entry in recent locations popup menu. Creates a menu
    * item that has name of the directory as menu item label and on selection will trigger a
    * navigation action to its full path.  
    * 
    * @param   strPath
    *          The full path of the directory history entry.
    * 
    * @return  An {@code MenuItemGuiImpl} with requested constraints.
    */
   private MenuItemGuiImpl createHistoryMenu(String strPath, int idx)
   {
      return createMenuItem(getDisplayName(strPath), idx == 0, () -> {
         int count = idx;
         if (count == 0)
         {
            return; // no change 
         }
         if (count > 0)
         {
            // fast forward from back to future for [count] times
            future.addFirst(currentDir);
            String p = null;
            while (count > 0)
            {
               p = history.removeLast();
               count--;
               if (count != 0)
               {
                  future.addFirst(p);
               }
            }
            currentDir = p;
            assert Objects.equals(strPath, currentDir);
         }
         else 
         {
            // fast rewind from future to back for [count] times
            history.addFirst(currentDir);
            String p = null;
            while (count < 0)
            {
               p = future.removeFirst();
               count++;
               if (count != 0)
               {
                  addToHistory(p);
               }
            }
            currentDir = p;
            assert Objects.equals(strPath, currentDir);
         }
         
         if (!isVirtualFolder(strPath))
         {
            navigate(strPath);
         }
         else 
         {
            openFolder(strPath, false);
         }
         updateNavButtons();
      });
   }
   
   /**
    * Helper methods for creating items in a (popup) menu.
    *
    * @param   menuLabel
    *          The text on the new item.
    * @param   checked
    *          {@code true} if the item is checked.
    * @param   code
    *          The code to be executed when the menu item is clicked.
    *          
    *  @return A {@code MenuItemGuiImpl} with requested constraints.
    */
   private static MenuItemGuiImpl createMenuItem(String menuLabel, boolean checked, Runnable code)
   {
      MenuItemGuiImpl mi = MenuItemGuiImpl.createDynamicItem(menuLabel, true, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            code.run();
         }
      });
      mi.config().toggleBox = true;
      mi.config().checked = checked;
      return mi;
   }
   
   /**
    * This is the main method of the dialog. It sets a new location, populate the container with
    * new content loaded from file system and updates the status of the other widgets.
    *
    * @param strPath The new location to be presented to user, (absolute path).
    */
   private void navigate(String strPath)
   {
      if (isVirtualFolder(strPath))
      {
         openFolder(strPath, false);
         return;
      }
      
      // be sure we force the search to stop when opening a new location
      searchOp = null;
      
      Path newPath = Paths.get(strPath);
      if (!Files.exists(newPath))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Dir does not exist: [" + strPath + "]");
         }
         return;
      }
      else if (!Files.isDirectory(newPath))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Path is not a directory: [" + strPath + "]");
         }
         return;
      }
      
      String asRegex = getPatternAsRegexp(activeFilter);
      currentDirContent.clear();
      
      // load path from filesystem and populate content panel:
      try
      {
         Stream<Path> paths = Files.list(newPath);
         paths.filter(FileChooserUtils::isPathVisible)
              .filter(aPath -> Files.isDirectory(aPath) || 
                               getFileName(aPath).matches(asRegex))
              .sorted((o1, o2) -> {
                 boolean d1 = Files.isDirectory(o1);
                 boolean d2 = Files.isDirectory(o2);
                 if (d1 == d2)
                 {
                    // same type? use lexical compare
                    return getFileName(o1).compareTo(getFileName(o2));
                 }
                 
                 // not same type?
                 return d1 ? -1 : 1;
              }).forEach(aPath -> {
                   currentDirContent.add(aPath);
                   //   System.out.println(Files.isDirectory(aPath)
                   //           ? "\t[" + getFileName(aPath) + "]"
                   //           : "\t" + getFileName(aPath));
              });
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Access denied to directory: [" + strPath + "]", e);
         }
      }
      
      views[currentView].populate(currentDirContent);
      currentDir = newPath.toAbsolutePath().toString();
      
      repaint(); // update the UI
   }
   
   /**
    * Converts a set of file filters (which uses {@code *} and {@code ?} as wildcards) to a
    * combined regex expression. Individual filters are separated with {@code ;}. Some symbols
    * that have special meanings in regexp are escaped.
    *
    * @param   activeFilter
    *          A set of file selection patterns separated by {@code ;}.
    *
    * @return  A regexp string that matches the same set of files as the entry pattern.
    */
   private static String getPatternAsRegexp(String activeFilter)
   {
      StringBuilder sb = new StringBuilder();
      StringTokenizer st = new StringTokenizer(activeFilter, ";");
      while (st.hasMoreTokens())
      {
         String pattern = st.nextToken().trim();
         sb.append("(");
         for (int k = 0; k < pattern.length(); k++)
         {
            char c = pattern.charAt(k);
            switch (c)
            {
               case '?': sb.append("."); break; // single char 
               case '*': sb.append(".*"); break; // many chars 
               case '(': sb.append("\\("); break; // escape it, just a simple '(' 
               case ')': sb.append("\\)"); break; // escape it, just a simple ')' 
               case '+': sb.append("\\+"); break; // escape it, just a simple '+' 
               case '-': sb.append("\\-"); break; // escape it, just a simple '-' 
               case '.': sb.append("\\."); break; // escape it, just a simple '.' 
               case '\\': sb.append("\\\\"); break; // escape it, just a simple '\' 
               default: sb.append(c); break;
            }
         }
         sb.append(")");
         if (st.hasMoreTokens())
         {
            sb.append("|");
         }
      }
      return sb.toString();
   }
   
   /**
    * Updates the status of the navigation buttons, disabling the ones that have inactive or
    * incorrect actions in the current moment. Also updated the tooltips for the enabled buttons.
    */
   private void updateNavButtons()
   {
      // update the [Back] button state
      navBack.setEnabled(!history.isEmpty());
      if (!history.isEmpty())
      {
         navBack.config().tooltip = "Back to " + getDisplayName(history.getLast());
         ThinClient.setupTooltip(navBack);
      }
      
      // update the [Forward] button state
      navForward.setEnabled(!future.isEmpty());
      if (!future.isEmpty())
      {
         navForward.config().tooltip = "Forward to " + getDisplayName(future.getFirst());
         ThinClient.setupTooltip(navForward);
      }
      
      // update the recent location button state
      recentLocations.setEnabled(!history.isEmpty() || !future.isEmpty());
      ThinClient.setupTooltip(recentLocations);
      
      // update the [Up] button state
      Path newPath = Paths.get(currentDir);
      Path parentPath = newPath.getParent();
      folderUp.setEnabled(parentPath != null);
      if (parentPath != null)
      {
         folderUp.config().tooltip = "Up to \"" + getDisplayName(parentPath) + "\"";
         ThinClient.setupTooltip(folderUp);
      }
      
      // update the [addressCbx] combobox
      ComboBoxModel<String> addrModel = addressCbx.model();
      String strOldPath = addrModel.selected();
      if (!Objects.equals(strOldPath, currentDir))
      {
         int idx = addrModel.indexOf(currentDir);
         if (idx == -1)
         {
            addrModel.add(currentDir);
         }
         idx = addrModel.indexOf(currentDir);
         addrModel.select(currentDir);
         addrModel.setSelectedId(idx);
         
         addressCbx.setValue(new character(currentDir));
         addressCbx.valueChangedCommon();
      }
   }
   
   /**
    * Goes back to previous location. Extracts the location from {@code history} stack and
    * {@code navigate}s back to it.
    */
   private void goBack()
   {
      if (!history.isEmpty())
      {
         String path = history.removeLast();
         future.addFirst(currentDir);
         navigate(path);
         assert (Objects.equals(currentDir, path));
      }
      
      updateNavButtons();
   }
   
   /**
    * Goes forward to a previous location. Extract the location from {@code future} stack and
    * {@code navigate}s back to it.
    */
   private void goForward()
   {
      if (!future.isEmpty())
      {
         String path = future.removeFirst();
         addToHistory(currentDir);
         navigate(path);
         assert (Objects.equals(currentDir, path));
      }
      
      updateNavButtons();
   }
   
   /**
    * The {@code UP} action. Computes the parent directory of the current location and, if
    * available navigates to it.
    */
   private void goUp()
   {
      Path parent = Paths.get(currentDir).getParent();
      if (parent != null)
      {
         String newStrPath = parent.toAbsolutePath().toString();
         addToHistory(currentDir);
         future.clear();
         navigate(newStrPath);
         assert (Objects.equals(currentDir, newStrPath));
      }
      
      updateNavButtons();
      
      // move focus to first thumb in the view:
      FileThumb thumb = views[currentView].findFirst();
      if (thumb != null)
      {
         FileThumb.select(thumb);
         updateTabbingList();
      }
   }
   
   /**
    * Refreshes the content of the current location (in the event of an external change of file
    * system in the current location).
    */
   private void refresh()
   {
      // remember what file/folder was selected before
      FileThumb selFT = FileThumb.getSelected();
      String crtSelThumb = (selFT == null) ? null : selFT.textLabel();
      
      navigate(currentDir);
      
      selFT = views[currentView].locate(crtSelThumb);
      // if found, set focus back to it
      FileThumb.select(selFT);
   }
   
   /**
    * Shows the more options popup menu. Allow user to select a new View Mode.
    */
   private void moreOptions()
   {
      WidgetId mid = WidgetId.nextID();
      MenuConfig mc = new MenuConfig();
      mc.popupOnly = true;
      mc.ownerId = getId().asInt();
      
      MenuGuiImpl menu = new MenuGuiImpl(mid, mc);
      menu.initialize(mid, mc);
      menu.add(createMenuItem("Medium icons", currentView == 0, () -> changeView(0)));
      menu.add(createMenuItem("Small icons", currentView == 1, () -> changeView(1)));
      menu.add(MenuItemGuiImpl.createDynamicRule());
      menu.add(createMenuItem("List", currentView == 2, () -> changeView(2)));
      menu.add(MenuItemGuiImpl.createDynamicRule());
      menu.add(createMenuItem("Details", currentView == 3, () -> changeView(3)));
      menu.add(MenuItemGuiImpl.createDynamicRule());
      menu.add(createMenuItem("Tiles", currentView == 4, () -> changeView(4)));
      menu.add(MenuItemGuiImpl.createDynamicRule());
      menu.add(createMenuItem("Content", currentView == 5, () -> changeView(5)));
      
      menu.showPopup(this, more.screenPhysicalLocation(), null);
   }
   
   /**
    * Creates a customized {@code SimpleLabel} to be used on this dialog.
    *
    * @param   text
    *          The text to be displayed.
    * @param   w
    *          The physical width of the label.
    * @param   h
    *          The physical height of the label.
    * @param   rightAlign
    *          Whether the text is right aligned.
    *
    * @return  A customized {@code SimpleLabel}.
    */
   private SimpleLabel createSimpleLabel(String text, int w, int h, boolean rightAlign)
   {
      SimpleLabel label = new SimpleLabel(text);
      label.setSize(cc.dimensionFromNative(new NativeDimension(w, h)));
      LabelConfig lcfg = (LabelConfig) label.config();
      lcfg.align = rightAlign ? ControlEntity.ALIGN_RIGHT : ControlEntity.ALIGN_LEFT;
      lcfg.frameId = getId().asInt();
      return label;
   }
   
   /**
    * Creates and initializes a button widget.
    *
    * @return   A reference to {@link Button} instance.
    */
   private ButtonGuiImpl createButton(String text,
                                      String action,
                                      int width,
                                      int height,
                                      boolean flat,
                                      boolean def,
                                      String tooltip)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = width;
      bcfg.heightPixels = height;
      bcfg.isDefault = def;
      bcfg.flatButton = flat;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.frameId = getId().asInt();
      
      ButtonGuiImpl b = new ButtonGuiImpl(bid, bcfg);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(action);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      if (tooltip != null)
      {
         bcfg.tooltip = tooltip;
         ThinClient.setupTooltip(b);
      }
      
      return b;
   }
   
   /**
    * Creates a customized {@code SimpleImageButton} to be used on this dialog.
    *
    * @param   action
    *          The action to perform when button is clicked. The button used this to load the
    *          appropriate image it will display.
    * @param   tooltip
    *          The text to be displayed when mouse hovers the button. Can be {@code null}.
    * @param   w
    *          The physical width of the label.
    * @param   h
    *          The physical height of the label.
    *
    * @return  A customized {@code SimpleImageButton}.
    */
   private SimpleImageButton createImageButton(String action, String tooltip, int w, int h)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = w;
      bcfg.heightPixels = h;
      bcfg.isDefault = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.frameId = getId().asInt();
      
      SimpleImageButton b = new SimpleImageButton(bid, action);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(action);
      b.setCommand(action);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      if (tooltip != null)
      {
         bcfg.tooltip = tooltip;
         ThinClient.setupTooltip(b);
      }
      
      return b;
   }
   
   /**
    * Creates a customized {@code Folder} to be used on this dialog. The new object is
    * automatically added to {@code dirTree} container. 
    *
    * @param   parent
    *          The parent folder. {@code null} if node is the root.
    * @param   text
    *          The text to be displayed on the button.
    * @param   path
    *          The absolute file path.
    * @param   img
    *          The image to be displayed on the button.
    *
    * @return  A customized {@code SimpleImageButton}.
    */
   private Folder createFolder(Folder parent, String text, String path, String img)
   {
      boolean root = (parent == null);
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = FOLDER_TREE_WIDTH - 1 - 17; // NOTE: theme dependent (scroll-width)
      bcfg.heightPixels = FOLDER_HEIGHT;
      bcfg.isDefault = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.frameId = getId().asInt();
      
      Folder f = new Folder(bid, parent, text, path, img, root ? 0 : parent.indent + 1);
      f.initialize(bid, bcfg);
      f.setSupportsMnemonic();
      f.setTextLabel(text);
      f.setCommand("OPEN " + path);
      f.setEnabled(true);
      f.setVisible(true);
      f.addListener(this);
      f.expanded = root;
      
      treePane.add(f);
      
      return f;
   }
   
   /**
    * Creates a customized {@code ComboBoxGuiImpl} to be used on this dialog.
    *
    * @param   mode
    *          The type of combobox to be returned.
    * @param   x
    *          The physical X coordinate of the widget. 
    * @param   y
    *          The physical Y coordinate of the widget.
    * @param   w
    *          The physical width of the widget.
    * @param   h
    *          The physical height of the widget.
    * @param   items
    *          An array of items to be used in the combobox.
    * @param   selVal
    *          The value to be selected bu default.
    * @param   format
    *          The format of the displayed data. 
    *
    * @return  A customized {@code ComboBoxGuiImpl}.
    */
   private ComboBoxGuiImpl createCombobox(ComboBoxConfig.Mode mode,
                                          int x,
                                          int y,
                                          int w,
                                          int h,
                                          ControlSetItem[] items,
                                          String selVal,
                                          String format)
   {
      WidgetId rid = WidgetId.nextID();
      ComboBoxConfig ccfg = new ComboBoxConfig();
      ccfg.id = rid;
      ccfg.clientX = x;
      ccfg.clientY = y;
      ccfg.widthPixels = h;
      ccfg.heightPixels = w;
      ccfg.widthChars = cc.widthFromNative(w);
      ccfg.heightChars = cc.heightFromNative(h);
      ccfg.clientWidthChars = cc.widthFromNative(w);
      ccfg.clientHeightChars = cc.heightFromNative(h);
      ccfg.autoResize = false;
      ccfg.fixedWidth = true;
      ccfg.fixedHeight = true;
      ccfg.mode = mode;
      ccfg.items = items;
      ccfg.format = format;
      ccfg.frameId = getId().asInt();
      
      ComboBoxGuiImpl combo = new ComboBoxGuiImpl(rid, ccfg);
      combo.initialize(rid, ccfg);
      combo.setEnabled(true);
      combo.setPhysicalLocation(x, y);
      ComboBoxModel<String> comboModel = combo.model();
      for (ControlSetItem it : items)
      {
         comboModel.add(it.getLabel().toStringMessage());
      }
      combo.setValue(new character(selVal));
      combo.valueChangedCommon();
      comboModel.select(selVal);
      
      GuiOutputManager.instance().getRegistry().addWidget(combo); // TODO: why is not this done automatically?
      return combo;
   }
   
   /**
    * Sets the physical location of a widget related to its parent physical location.
    *
    * @param   widget
    *          The widget to be 'moved'.
    * @param   x
    *          The new location on X axis (in pixels).
    * @param   y
    *          The new location on Y axis (in pixels).
    * @param   rowOffset
    *          The location of {@code widget}'s parent.
    */
   private void setWidgetLocation(AbstractWidget widget,
                                  int x,
                                  int y,
                                  NativePoint rowOffset)
   {
      if (rowOffset != null)
      {
         x += rowOffset.x;
         y += rowOffset.y;
      }
      widget.setPhysicalLocation(x, y);
   }
   
   /**
    * Sets the size of a combobox widget.
    * @param   cbx
    *          The combobox widget to be altered. 
    * @param   width
    *          The new width for the combobox (in native units / pixels).
    * @param   height
    *          The new height for the combobox (in native units / pixels).
    */
   private void setComboboxSize(ComboBoxGuiImpl cbx, int width, int height)
   {
      ComboBoxConfig cbxCfg = cbx.config();
      cbxCfg.widthPixels = width;
      cbxCfg.heightPixels = height;
      cbxCfg.widthChars = cc.widthFromNative(width);
      cbxCfg.heightChars = cc.heightFromNative(height);
      cbxCfg.clientWidthChars = cc.widthFromNative(width);
      cbxCfg.clientHeightChars = cc.heightFromNative(height);
   }
   
   /**
    * Adding new widget to the container. This method does not require the coordinates to be
    * defined. The location will be selected automatically. Makes sure this frame is also set on
    * target {@code widget}.
    *
    * @param   widget
    *          The new widget to be added to the container.
    */
   @Override
   public void addWidget(Widget<GuiOutputManager> widget)
   {
      super.addWidget(widget);
      widget.config().frameId = getId().asInt();
   }
   
   /**
    * Adding new widget to a container. This method does not require the coordinates to be
    * defined. It makes sure this frame is also set on target {@code widget}.
    *
    * @param   widget
    *          The new widget to be added to the container.
    */
   public void addWidgetToPane(Widget<GuiOutputManager> widget, AbstractContainer container)
   {
      container.add(widget);
      widget.config().frameId = getId().asInt();
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    HistoryBack       button      round          Loads the previous folder
    *    HistoryForward    button      round          Loads the forward folder 
    *    RecentLocations   button                     Opens a popup-menu with recent locations
    *    FolderUp          button                     Loads the parent folder
    *    Address           editor      expandable     Allows user to type in a location
    *    Refresh           button                     Reload the current location
    *                      /splitter/                 Changes ratio between Address and Search
    *    Search            editor      expandable     Searches sub-folders for specific pattern
    */
   private class NavigationPane
   extends AbstractContainer<GuiOutputManager>
   {
      /** The {@code BaseConfig} configuration for this widget. */
      private BaseConfig config = new BaseConfig();
      
      /**
       * Obtain the {@code WidgetConfig} configuration for this widget. 
       * 
       * @return  the {@code WidgetConfig} configuration for this widget.
       */
      @Override
      public WidgetConfig config()
      {
         return config;
      }
      
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(NAV_PANE_HEIGHT);
      }
      
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
         NativePoint rowOffset = new NativePoint(0, 0); //physicalLocation();
         
         setWidgetLocation(navBack,           9, 8, rowOffset);
         setWidgetLocation(navForward,       39, 8, rowOffset);
         setWidgetLocation(recentLocations,  66, 8, rowOffset);
         setWidgetLocation(folderUp,         83, 6, rowOffset);
         setWidgetLocation(addressCbx,      107, 6, rowOffset);
         setWidgetLocation(refresh, natWidth - 218 - 24, 6, rowOffset);
         setWidgetLocation(search,  natWidth - 218,      6, rowOffset);
         
         setComboboxSize(addressCbx, refresh.physicalLocation().x - addressCbx.physicalLocation().x, 22);
         setComboboxSize(search, natWidth - search.physicalLocation().x - 19, 23);
      }
      
      /**
       * Draw widget.
       */
      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeDimension dim = physicalDimension();
         NativeRectangle clip = clipRectangle(p, dim);
         gd.draw(getId(), p, clip, () ->
         {
            if (ThemeManager.getCurrentTheme().isClassic())
            {
               gd.setColor(gd.getSysColor(COLOR_3DFACE));
            }
            else
            {
               gd.setColor(ColorRgb.WHITE);
            }
            gd.fillRect(0, 0, dim.width, dim.height);
            
            super.draw();
         });
      }
   }
   
   /**
    * Tools row contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    Organize          button                     Opens a contextual popup-menu
    *    NewFolder         button                     Creates a new folder at current location
    *                      /spacer/    expandable     Nothing (fills in empty space)
    *    ChangeView        button                     Selects the next view mode
    *    MoreOptions       button                     Opens a popup-menu with different view modes
    *    Help              button                     Opens Help page
    */
   private class ToolsPane
   extends AbstractContainer<GuiOutputManager>
   {
      /** The {@code BaseConfig} configuration for this widget. */
      BaseConfig config = new BaseConfig();
      
      /**
       * Obtain the {@code WidgetConfig} configuration for this widget.
       *
       * @return  the {@code WidgetConfig} configuration for this widget.
       */
      @Override
      public WidgetConfig config()
      {
         return config;
      }
      
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(TOOLS_PANE_HEIGHT);
      }
      
      /**
       * Get widget minimum size.
       *
       * @return  widget minimum size.
       */
      @Override
      public Dimension minimumSize()
      {
         return new Dimension(0D, 0D);
      }
      
      /**
       * Get widget width.
       *
       * @return  widget width.
       */
      @Override
      public double width()
      {
         return cc.widthFromNative(NATIVE_WIDTH);
      }
      
      /**
       * Get widget dimension.
       *
       * @return  widget dimension.
       */
      @Override
      public Dimension dimension()
      {
         return new Dimension(width(), height());
      }
      
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
         NativePoint rowOffset = new NativePoint(0, 0); //physicalLocation();
         
         setWidgetLocation(organize,          3, 2, rowOffset);
         setWidgetLocation(newFolder,        94, 2, rowOffset);
         setWidgetLocation(view, natWidth - 100, 2, rowOffset);
         setWidgetLocation(more, natWidth -  74, 2, rowOffset);
         setWidgetLocation(help, natWidth -  36, 2, rowOffset);
      }
      
      /**
       * Draw widget.
       */
      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeDimension dim = physicalDimension();
         NativeRectangle clip = clipRectangle(p, dim);
         gd.draw(getId(), p, clip, () ->
         {
            if (ThemeManager.getCurrentTheme().isClassic())
            {
               gd.setColor(gd.getSysColor(COLOR_3DFACE));
            }
            else
            {
               gd.setColor(new ColorRgb(0xEFEFFF)); // light magenta
            }
            gd.fillRect(0, 0, dim.width, dim.height);
            
            super.draw();
         });
      }
   }
   
   /**
    * The {@code LayoutManager} that handles the placement and size of the widgets on this dialog.
    */
   private class FileDialogLayoutManager
   implements LayoutManager<GuiOutputManager>
   {
      /**
       * Calculate minimum size required for widgets stored in passed container.
       *
       * @param   container
       *          Container to process.
       *
       * @return  calculated space.
       */
      @Override
      public Dimension minimumSize(Container<GuiOutputManager> container)
      {
         return cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH - 5, NATIVE_HEIGHT - 5));
      }
      
      /**
       * Place components in specified container.
       *
       * @param   container
       *          Container to process.
       */
      @Override
      public void doLayout(Container<GuiOutputManager> container)
      {
         NativeDimension dim = container.physicalDimension();
         
         {
            // set location (and size, optionally) of navigation widgets:
            setWidgetLocation(navigationPane, 0, 0, new NativePoint(0, 0));
            navigationPane.doLayout();
         }
         
         // set location (and size, optionally) of tools widgets:
         {
            setWidgetLocation(toolsPane, 0, NAV_PANE_HEIGHT, new NativePoint(0, 0));
            toolsPane.doLayout();
         }
         
         if (saveAs)
         {
            // set location (and size, optionally) of contents widgets:
            {
               NativePoint rowOffset = new NativePoint(0, NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT);
               int rowHeight = dim.height - (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_SA +
                     SAVEAS_PANE_HEIGHT + FILENAME_PANE_HEIGHT_SA + NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT);
               
               // [contentPane] expands so it occupy all remaining vertical space:
               setWidgetLocation(treeScroll, 0, 0, rowOffset);
               treeScroll.setSize(
                     cc.dimensionFromNative(new NativeDimension(FOLDER_TREE_WIDTH, rowHeight)));
               
               // all scrolls are overlapped, but only one is visible
               for (ScrollPaneGuiImpl scroll : viewScrolls)
               {
                  setWidgetLocation(scroll, FOLDER_TREE_WIDTH, 0, rowOffset);
                  scroll.setSize(cc.dimensionFromNative(
                        new NativeDimension(dim.width - FOLDER_TREE_WIDTH, rowHeight)));
               }
            }
            
            // set location (and size, optionally) of filename widgets:
            {
               NativePoint rowOffset = new NativePoint(0, dim.height -
                     (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_SA + SAVEAS_PANE_HEIGHT + FILENAME_PANE_HEIGHT_SA));
               
               setWidgetLocation(filenameLabel, 10, 10, rowOffset);
               setWidgetLocation(filename, FILE_TYPE_ALIGN_SA, 10, rowOffset);
               
               filenameLabel.setSize(cc.dimensionFromNative(new NativeDimension(115, 23)));
               int width = dim.width - filename.physicalLocation().x - 5;
               filename.setSize(cc.widthFromNative(width), cc.heightFromNative(23));
            }
            
            // set location (and size, optionally) of save-as widgets: 
            {
               NativePoint rowOffset = new NativePoint(0, dim.height -
                     (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_SA + SAVEAS_PANE_HEIGHT));
               
               setWidgetLocation(saveasLabel, 10, 1, rowOffset);
               setWidgetLocation(fileTypeCbx, FILE_TYPE_ALIGN_SA, 1, rowOffset);
               
               saveasLabel.setSize(cc.dimensionFromNative(new NativeDimension(115, 23)));
               setComboboxSize(fileTypeCbx, dim.width - fileTypeCbx.physicalLocation().x - 5, 23);
            }
            
            // set location (and size, optionally) of bottom buttons widgets: 
            {
               NativePoint rowOffset =
                     new NativePoint(0, dim.height - (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_SA));
               
               setWidgetLocation(hideFolders, 9, 18, rowOffset);
               setWidgetLocation(accept, dim.width - 205, 14, rowOffset);
               setWidgetLocation(reject, dim.width - 105, 14, rowOffset);
            }
         }
         else // simpler open file layout
         {
            // set location (and size, optionally) of contents widgets:
            {
               NativePoint rowOffset = new NativePoint(0, NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT);
               int rowHeight = dim.height - (NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT +
                     BOTTOM_INSET + BUTTONS_PANE_HEIGHT_OP + FILENAME_PANE_HEIGHT_OP);
               
               // [contentPane] expands so it occupy all remaining vertical space:
               setWidgetLocation(treeScroll, 0, 0, rowOffset);
               treeScroll.setSize(
                     cc.dimensionFromNative(new NativeDimension(FOLDER_TREE_WIDTH, rowHeight)));
               
               // all scrolls are overlapped, but only one is visible
               for (ScrollPaneGuiImpl scroll : viewScrolls)
               {
                  setWidgetLocation(scroll, FOLDER_TREE_WIDTH, 0, rowOffset);
                  scroll.setSize(cc.dimensionFromNative(
                        new NativeDimension(dim.width - FOLDER_TREE_WIDTH, rowHeight)));
               }
            }
            
            // set location (and size, optionally) of filename widgets:
            {
               NativePoint rowOffset = new NativePoint(0, dim.height -
                     (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_OP + FILENAME_PANE_HEIGHT_OP));
               
               setWidgetLocation(filenameLabel, 0, 10, rowOffset);
               filenameLabel.setSize(
                     cc.dimensionFromNative(new NativeDimension(FOLDER_TREE_WIDTH, 22)));
               
               setWidgetLocation(filename, FOLDER_TREE_WIDTH, 9, rowOffset);
               int width = dim.width - (FOLDER_TREE_WIDTH + FILE_TYPE_WIDTH_OP + 30);
               filename.setSize(cc.widthFromNative(width), cc.heightFromNative(23));
               
               setWidgetLocation(fileTypeCbx, dim.width - FILE_TYPE_WIDTH_OP - 14, 9, rowOffset);
               setComboboxSize(fileTypeCbx, FILE_TYPE_WIDTH_OP, 23);
            }
            
            // set location (and size, optionally) of bottom buttons widgets: 
            {
               NativePoint rowOffset =
                     new NativePoint(0, dim.height - (BOTTOM_INSET + BUTTONS_PANE_HEIGHT_OP));
               
               setWidgetLocation(accept, dim.width - 205, 1, rowOffset);
               setWidgetLocation(reject, dim.width - 105, 1, rowOffset);
            }
         }
         // 11 pixels in bottom inset (paints the 'resize' grip in the right corner)
      }
   }
   
   /**
    * A dedicated container for holding the shortcuts in a vertical flow.
    */
   private class FolderTree
   extends SensitiveScrollContainer<GuiOutputManager>
   {
      /** X inset for the tree. */
      private static final int Y_OFFSET = 9;
      
      /** Indexes the shown Folder. Used for cursor key navigation. */
      private Folder[] shown = null;
      
      /**
       * Draw widget (white background).
       */
      @Override
      public void draw()
      {
         if (!isVisible())
         {
            return;
         }
         
         NativePoint p = physicalLocation();
         NativeDimension dim = physicalDimension();
         NativeRectangle clip = clipRectangle(p, dim);
         gd.draw(getId(), p, clip, () ->
         {
            gd.setColor(new ColorRgb(0xFFFFFF)); // white
            gd.fillRect(0, 0, dim.width, dim.height);
            
            // make sure the draw origin is properly translated
            super.draw();
         });
      }
      
      /**
       * Lay out container (in a vertical flow).
       */
      @Override
      public void doLayout()
      {
         // invalidate all:
         Widget[] allWidgets = widgets();
         for (Widget w : allWidgets)
         {
            w.setVisible(false);
         }
         shown = new Folder[allWidgets.length + 2];
         // all [shown] elements are [null]. 
         // We added two extra elements for the two gaps between the roots (see below).  
         
         int[] yOffset = { 0 };
         doLayout(scFavorites, yOffset);
         ++yOffset[0];
         doLayout(scThisPC, yOffset);
         ++yOffset[0];
         doLayout(scNetwork, yOffset);
      }
      
      /**
       * Check if widget is focus traversable.
       *
       * @return  Always {@code true}.
       */
      @Override
      public boolean focusTraversable()
      {
         return true;
      }
      
      /**
       * Recursive worker that do the layout of the widgets in a tree fashion.
       *
       * @param   node
       *          The current node to be processed.
       * @param   yOffset
       *          The vertical position of the node. To avoid using a class field, the procedure
       *          is made re-entrant by using a single element array (to pass value by reference). 
       */
      private void doLayout(Folder node, int[] yOffset)
      {
         ButtonConfig bcfg = node.config();
         if (bcfg.widthPixels == -1)
         {
            bcfg.widthPixels = node.computePreferredWidth();
         }
         
         node.setVisible(true);
         node.setPhysicalLocation(0, Y_OFFSET + yOffset[0] * FOLDER_HEIGHT);
         assert (shown.length > yOffset[0]);
         shown[yOffset[0]] = node;
         ++yOffset[0];
         
         if (node.expanded && node.content != null && node.content.length != 0)
         {
            for (Folder folder : node.content)
            {
               doLayout(folder, yOffset);
            }
         }
      }
      
      /**
       * Find the folder above {@code dir}, in tree structure.
       *
       * @param   dir
       *          A reference folder.
       *
       * @return  The folder above {@code dir}, or {@code null} if there is none.
       */
      Folder getPreviousFolder(Folder dir)
      {
         NativePoint p = dir.physicalLocation();
         int idx = (p.y - Y_OFFSET) / FOLDER_HEIGHT;
         assert (idx >= 0 && shown[idx] == dir);
         
         if (idx >= 1 && shown[idx - 1] != null)
         {
            return shown[idx - 1];
         }
         
         // in case of the gap between the trees we may need to go up 2 rows, if possible
         if (idx >= 2 && shown[idx - 2] != null)
         {
            return shown[idx - 2];
         }
         
         return null;
      }
      
      /**
       * Find the next folder under {@code dir}, in tree structure.
       *
       * @param   dir
       *          A reference folder.
       *
       * @return  The folder under {@code dir}, or {@code null} if there is none.
       */
      Folder getNextFolder(Folder dir)
      {
         NativePoint p = dir.physicalLocation();
         int idx = (p.y - Y_OFFSET) / FOLDER_HEIGHT;
         assert (idx < shown.length && shown[idx] == dir);
         
         if (idx + 1 < shown.length && shown[idx + 1] != null)
         {
            return shown[idx + 1];
         }
         
         // in case of the gap between the trees we may need to go down 2 rows, if possible
         if (idx + 2 < shown.length && shown[idx + 2] != null)
         {
            return shown[idx + 2];
         }
         
         // this is the last folder, cannot go down
         return null;
      }
   }
   
   /**
    * A thumbnail for a directory.
    */
   private class Folder
   extends ButtonGuiImpl
   {
      /** Indentation step (in pixels). */
      static final int X_INDENT = 10;
      
      /** The parent folder or {@code null} if this si a root. */
      private final Folder parent;
      
      /**
       * The list of sub-folders or {@code null} if it was never expanded.
       */
      private Folder[] content = null;
      
      /**
       * The absolute path of this folder.
       */
      private final String path;
      
      /**
       * The image to be displayed in the lef-side of the button.
       */
      private ImageWrapper<BufferedImage> img = null;
      
      /**
       * Indentation of this folder.
       */
      private final int indent;
      
      /** {@code true} only if this folder is expanded, showing its children. */
      private boolean expanded = false;
      
      /**
       * The constructor creates a button with given properties.
       *
       * @param   id
       *          The widget's ID.
       * @param   parent
       *          The parent folder. {@code null} if node is the root.
       * @param   text
       *          The string to be used as the button text.
       * @param   path
       *          The absolute path of the folder.
       * @param   icon
       *          The image to be displayed.
       * @param   indent
       *          The indentation of image and text.
       */
      public Folder(WidgetId id, Folder parent, String text, String path, String icon, int indent)
      {
         super(id, null/* ignored */);
         setTextLabel(text);
         this.img = FileChooserUtils.loadImage(icon);
         this.path = path;
         this.indent = indent;
         this.parent = parent;
      }
      
      /**
       * Draw the button.
       */
      @Override
      public void draw()
      {
         if (!isVisible())
         {
            return;
         }
         
         // get location and size
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         
         // set up clipping rectangle
         NativeRectangle clip = clipRectangle(p, d);
         
         // clippable drawing start 
         gd.draw(this.getId(), p, clip, () -> {
            Theme theme = ThemeManager.getCurrentTheme();
            theme.drawFileChooserThumbnailBackground(
                  gd, 0, 0, d.width, d.height, this == crtFolder, isMouseOver(), hasFocus());
            
            int x = X_INDENT * indent + 3;
            if (content == null || content.length != 0)
            {
               drawImage(gd, expanded ? expImg : colImg, x, 6);
            }
            
            drawImage(gd, img, x + X_INDENT, 0);
            
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, hasFocus(), this == crtFolder, isMouseOver(), true, false, false));
            
            FontDetails<?> fd = gf.font().font;
            gd.setGuiFont(fd);
            gd.setFontStyle(fd.style);
            int tX = x + 2 * X_INDENT + (img == null ? 0 : img.getWidth()) - 3;
            gd.drawStringCentered(textLabel(), tX, d.height / 2);
         });
      }
      
      /**
       * Checks whether this is the root of the tree. 
       *
       * @return  {@code true} if this is the root of the tree.
       */
      boolean isRoot()
      {
         return parent == null; // also indent == 0
      }
      
      /**
       * Computes the preferred width for this thumbnail.
       *
       * @return the preferred width for this thumbnail.
       */
      int computePreferredWidth()
      {
         int tW = getTextWidthNative(textLabel(), gf.resolveFontNum());
         int ret = 2 + (img != null ? img.getWidth() + 2 : 0) + tW + 4;
         if (!isRoot())
         {
            ret += X_INDENT;
         }
         return ret;
      }
      
      /**
       * Searches for a child among all direct children.
       *
       * @param   child
       *          The {@code Folder} to search.
       *
       * @return  The requested {@code Folder} or {@code null} if not found. 
       */
      Folder lookup(String child)
      {
         if (content == null)
         {
            return null;
         }
         
         for (Folder folder : content)
         {
            if (Objects.equals(folder.textLabel(), child))
            {
               return folder;
            }
         }
         
         // not found?
         return null;
      }
      
      /**
       * Process basic events - key navigation event.
       *
       * @param event Key event to process.
       */
      @Override
      protected void handleBasicEvents(KeyInput event)
      {
         boolean handled = false;
         if (this == crtFolder)
         {
            Folder newSelFolder = null;
            switch (event.keyCode())
            {
               case Key.VK_ENTER:
               case Key.VK_SPACE:
                  onAction(new ActionEvent(this, getCommand()));
                  handled = true;
                  break;
               
               case Key.VK_ESCAPE:
                  EventManager.postEvent(new ActionEvent(this, "ESCAPE"));
                  handled = true;
                  break;
               
               case Key.VK_CURSOR_RIGHT:
                  // expand this node
                  expandFolder(crtFolder);
                  handled = true;
                  break;
               
               case Key.VK_CURSOR_LEFT:
                  if (this.expanded && (content == null || content.length != 0))
                  {
                     // collapse this node if expanded
                     collapse();
                  }
                  else if (!isRoot())
                  {
                     // if already collapsed move to parent
                     newSelFolder = parent;
                  }
                  handled = true;
                  break;
               
               case Key.VK_CURSOR_UP:
                  // move focus to above Folder
                  newSelFolder = treePane.getPreviousFolder(this);
                  handled = true;
                  break;
               
               case Key.VK_CURSOR_DOWN:
                  // move focus to Folder below
                  newSelFolder = treePane.getNextFolder(this);
                  handled = true;
                  break;
            }
            
            if (newSelFolder != null)
            {
               crtFolder = newSelFolder;
               crtFolder.requestFocus();
               treePane.ensureVisibility(crtFolder, true);
               updateTabbingList();
            }
         }
         
         if (!handled)
         {
            super.handleBasicEvents(event);
         }
      }
      
      /**
       * Collapse this folder.
       */
      private void collapse()
      {
         this.expanded = false;
         treePane.doLayout();
         treeScroll.repaint();
      }
      
      /**
       * Notification of a mouse clicked event occurred for this widget.
       * The main reason it is overwritten here is to set {@code doubleClick} if double click is
       * detected and expand the folder in this case. If the user clicks on expand area 
       * {@code [+]}, the folder will be expanded or collapsed.
       *
       * @param    e
       *           The mouse event.
       */
      @Override
      public void mouseClicked(MouseEvent e)
      {
         if (e.getClickCount() > 1 && !expanded)
         {
            expandFolder(this);
         }
         if (e.getClickCount() == 1 && content == null || content.length != 0)
         {
            int x = X_INDENT * indent + 3;
            if (x < e.getX() && e.getX() < x + X_INDENT)
            {
               if (expanded)
               {
                  collapse();
               }
               else
               {
                  expandFolder(this);
               }
               return;
            }
         }
      
         super.mouseClicked(e);
      }
   
      /**
       * Returns a string representation of the object.
       *
       * @return   A string representation of the object.
       */
      @Override
      public String toString()
      {
         return "Folder{'" + path + File.separator + '}';
      }
   }
   
   /**
    * A dedicated thread that handles a recursive search operation in background.
    */
   private class SearchOperation
   extends Thread
   {
      /** The start directory. */
      private final Path startDir;
      
      /** The pattern to be search for. */
      private final String pattern;
      
      /** 
       * The rate at which recursive searches are performed expressed as interval between two
       * searches (in milliseconds).
       */
      private final int SEARCH_RATE = 200;
      
      /**
       * Do the recursive search. The recursive calls are a bit delayed on a separate thread in
       * order to let the UI be responsive. After the random delay, the search is processed back
       * on main UI thread.
       * 
       * @param   startDir
       *          The start directory. 
       * @param   pattern
       *          The pattern to be search for.
       */
      SearchOperation(Path startDir, String pattern)
      {
         this.startDir = startDir;
         this.pattern = pattern;
         setPriority(MIN_PRIORITY);
         start();
      }
      
      /**
       * The worker method.
       */
      @Override
      public void run()
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Started search for [" + pattern + "].");
         }
         ThinClient tc = ThinClient.getInstance();
         Deque<Path> toCall = new LinkedList<>();
         toCall.addLast(startDir);
         
         while (searchOp == this && isVisible() && !toCall.isEmpty())
         {
            Path dir = toCall.pop();
            if (LOG.isLoggable(Level.FINEST))
            {
               LOG.log(Level.FINEST, "searching inside [" + dir.toString() + "].");
            }
            List<Path> toAdd = new ArrayList<>();
            
            try
            {
               Stream<Path> paths = Files.list(dir);
               paths.filter(FileChooserUtils::isPathVisible)
                    .forEach(aPath -> {
                          if (Files.isDirectory(aPath))
                          {
                             toCall.add(aPath);
                          }
                          if (getFileName(aPath).toLowerCase().contains(pattern))
                          {
                             toAdd.add(aPath);
                          }
                       });
            }
            catch (IOException e)
            {
               if (LOG.isLoggable(Level.WARNING))
               {
                  LOG.log(Level.WARNING, "Access denied to directory: [" + dir.toString() + "]",
                          e);
               }
            }
            
            if (searchOp != this)
            {
               break;
            }
            
            if (!toAdd.isEmpty())
            {
               tc.invokeLaterOS(() -> tc.eventDrawingBracket(
                  FileDialogGuiImpl.this, false, true, () -> {
                     if (searchOp == this)
                     {
                        views[currentView].appendSearchResult(toAdd);
                        viewScrolls[currentView].repaint();
                     }
                  }));
            }
            
            // allow user to interact (stop the search) in case of complex tree which generates a
            // long-running operation 
            try
            {
               Thread.sleep(SEARCH_RATE);
            }
            catch (InterruptedException e)
            {
               // ignore 
            }
         }
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Search for [" + pattern + "] finished.");
         }
         if (searchOp == this)
         {
            searchOp = null;
         }
      }
   }
}
