/*
** Module   : SimpleLabel.java
** Abstract : a label that is not attached to any other widget. 
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.  
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This label just displays a text, in the sense of {@code javax.swing.JLabel} does. It is not
 * linked to any other widget.
 */
public class SimpleLabel
extends AbstractWidget
implements ConfigOwner
{
   /** Screen driver */
   private GuiDriver gd = null;
   
   /** Font resolver */
   private GuiFontResolver gf = null;
   
   /** The text to be displayed on the label*/
   private String text = null;
   
   /** The configuration. */
   private LabelConfig config = null;
   
   /** 
    * Opacity flag. If opaque, it will paint the whole rectangle, when transparent, only the text
    * is drawn.
    */
   private boolean opaque = false;
   
   /**
    * Simple constructor.
    * 
    * @param   text
    *          The text to be displayed.
    */
   public SimpleLabel(String text)
   {
      this.text = text;
      config = new LabelConfig();
      config.id = WidgetId.nextID();
      
      setEnabled(true);
   }
   
   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(config.clientWidthChars, config.clientHeightChars);
   }
   
   /**
    * Sets a fixed size for the label.
    * 
    * @param   dim
    *          The new dimension of the label.
    */
   public void setSize(Dimension dim)
   {
      config.clientWidthChars = dim.width;
      config.clientHeightChars = dim.height;
   }
   
   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      if (text == null || text.isEmpty())
      {
         return; // no-op
      }
      
      if (gd == null)
      {
         gd = (GuiDriver) OutputManager.getDriver();
      }
      
      if (gf == null)
      {
         gf = new GuiFontResolver(window(), gd, config, false);
      }
      GuiFontResolver.FontCache fc = gf.font();
      FontDetails<?> font = fc.font;
      
      int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
      
      GuiColorResolver gc = new GuiColorResolver(gd, config, bgColor, SysColor.COLOR_FILLIN_FG);
      gc.colors();
      
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      gd.draw(getId(), p, clip, () ->
      {
         if (opaque)
         {
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, dim.width, dim.height);
         }
         
         MnemonicInfoGuiImpl info = new MnemonicInfoGuiImpl(false);
         String txt = info.prepare(text);
         gd.setGuiFont(font);
         gd.setColor(gc.fgColor); // new ColorRgb(0x000000)); // black
         int tx = 0;
         if (config.align == ControlEntity.ALIGN_RIGHT)
         {
            tx = dim.width - getTextWidthNative(txt, gf.resolveFontNum()) - 5;
         }
         NativePoint tp = new NativePoint(tx, dim.height / 2);
         if (info.getMnemIdx() != null)
         {
            gd.drawStringWithMnemonic(txt, tp, info, gf, this, true);
         }
         else 
         {
            gd.drawStringCentered(txt, tx, dim.height / 2);
         }
      });
   }
   
   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return config.clientHeightChars;
   }
   
   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return new Dimension(config.clientWidthChars, config.clientHeightChars);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return config.clientWidthChars;
   }
   
   /**
    * Obtain the text this label displays.
    * 
    * @return  the text.
    */
   public String getText()
   {
      return text;
   }
   
   /**
    * Mark the label as being opaque. By default is is transparent (ie. background is not filled).
    */
   public void setOpaque()
   {
      this.opaque = true;
   }
   
   /**
    * Configures the text to be displayed on the label.
    * 
    * @param   text
    *          The new text to be displayed.
    */
   public void setText(String text)
   {
      this.text = text;
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is pressed.
    *
    * @return  Always {@code false}.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
   
   /**
    * Returns config instance associated with this instance.
    *
    * @return  Config reference.
    */
   @Override
   public WidgetConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with the owner changes. This method
    * is never called. These object do not have a correspondent on server-side. 
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(WidgetConfig beforeUpdate)
   {
      // NOP, this widget does not expect config updates 
   }
}
