/*
** Module   : Windows8Theme.java
** Abstract : Windows 18 theme manager. 
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170420 First version.
** 002 OM  20170707 Small patches.
** 003 SVL 20170901 Added getRowToggleOffset, getRowToggleImage.
** 004 HC  20170823 Added runtime support of WINDOW:SMALL-TITLE.
** 005 OM  20170920 Fixed inactive window border. Updated name property.
** 006 EVL 20180123 Fixed browser widget colors for selection and row separator.
** 007 SVL 20180321 Refactoring related to getCellAttr signature change.
** 008 OM  20180321 Added multiline support for tooltips. Added support for file/dir choosers.
** 009 OM  20180503 Added getBinaryResourceLocation().
** 010 SVL 20180529 Support for enhanced browse coloring.
** 011 OM  20180808 Fixed browse sub-components drawing on Windows8.
** 012 EVL 20180817 Fix for button feedback for SPACEBAR key.  The pressed state is possible even
**                  when the mouse is not over widget.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.theme;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Label;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.Color;
import java.awt.image.*;
import java.util.*;

import static com.goldencode.p2j.ui.client.SysColor.*;

/**
 * This theme emulates the Windows 8 look. image and color resource were captured from a 
 * Windows Server 2012 R2, with default palette set to color #12 (light blue, 0x63B6FF).
 * <p>
 * The theme is flatter than previous Windows UI, but with more finery than Windows 10. Because,
 * with the exception of windows decorations, the sizes and general aspects of the widgets are
 * similar to Windows 10, we extend this class. Like its super class, this theme also very
 * sensitive to mouse hovering movements, almost all components will highlight when the mouse
 * pointer hovers over them. 
 */
public class Windows8Theme
extends Windows10Theme
{
   /** The background color of the titlebar when window is active. */
   private static final Color WINDOW_TITLEBAR_COLOR_ACTIVE = new Color(0x63B6FF);
   
   /** The background color of the titlebar when window is NOT active. */
   private static final Color WINDOW_TITLEBAR_COLOR_INACTIVE = new Color(0xC6D3E7);
   
   /** The foreground color of the titlebar when window is active. */
   private static final Color WINDOW_TITLE_COLOR_ACTIVE = new Color(0x000000);
   
   /** The foreground color of the titlebar when window is NOT active. */
   private static final Color WINDOW_TITLE_COLOR_INACTIVE = new Color(0x000000);
   
   /** Background color for tooltips. */
   private static final Color TOOLTIP_BACKGROUND_COLOR = new Color(0xF7F3F7);
   
   /** Background color for drop-down and popup menus. */
   private static final Color MENU_POPUP_BACKGROUND_COLOR = new Color(0xF7F3F7);
   
   /** Background color for highlighted menu items. */
   private static final Color MENU_BACKGROUND_SELECTED_COLOR = new Color(0xDEEFFF); 
   
   /** Background color for menu-menus. */
   private static final Color MENU_BAR_BACKGROUND_COLOR = new Color(0xF7F3F7);
   
   /** Foreground color for (all) menus. */
   private static final Color MENU_FOREGROUND_COLOR = new Color(0x6B6D6B);
   
   /** Background color of the workspace */
   private static final Color WORKSPACE_BACKGROUND_COLOR = new Color(0xa5a2a5);
   
   /** Background color for highlighted text (Eg. selected browse cell).*/
   private static final Color HIGHLIGHT_BACKGROUND_COLOR = new Color(0x319AFF);
   
   /** Main background color for 3D objects. */
   private static final Color BACKGROUND_3DFACE_COLOR = new Color(0xF7F3F7);
   
   /** The test colors on enabled buttons. */
   private static final Color BUTTON_FOREGROUND_COLOR = new Color(0x000000);
   
   /** The background color of a caption button when windows is not active (same a border). */
   private static final Color CAPTION_BUTTON_COLOR_INACTIVE = WINDOW_TITLEBAR_COLOR_INACTIVE;
   
   /** The background color of a caption button (except close) when down/pressed. */
   private static final Color CAPTION_BUTTON_COLOR_DOWN = new Color(0x219ECE);
   
   /** The background color of the Close caption button when down/pressed. */
   private static final Color CAPTION_BUTTON2_COLOR_DOWN = new Color(0xb5514a);
   
   /** The background color of a caption button (except close) when highlighted/hovered. */
   private static final Color CAPTION_BUTTON_COLOR_HIGHLIGHT = new Color(0x7BD3F7);
   
   /** The background color of the Close caption button when highlighted/hovered. */
   private static final Color CAPTION_BUTTON2_COLOR_HIGHLIGHT = new Color(0xEF6D63);
   
   /** The background color of a caption button when windows is not active (same a border). */
   private static final Color CAPTION_BUTTON_COLOR_ACTIVE = WINDOW_TITLEBAR_COLOR_ACTIVE;
   
   /** The background color of the Close caption button when windows is not active. */
   private static final Color CAPTION_BUTTON2_COLOR_ACTIVE = new Color(0xC65952);
   
   /** Cached map with default system colors. */
   private Map<Integer, ColorRgb> defaultColors = null;
   
   /** The width of the border in this Theme. */
   private static final int WINDOW_BORDER_WIDTH = 8;
   
   /** The color of the background of a scrollbar. */
   private static final Color SCROLL_GAUGE_COLOR = new Color(0xFFFFFF);
   
   /** The color of the corner of a scrollbar. */
   private static final Color SCROLL_CORNER_COLOR = new Color(0xFFFFFF);
   
   /** The color of the border of a scrollbar. */
   private static final Color SCROLL_BORDER_COLOR = new Color(0xBDBEBD);
   
   /** The background color of the of a scrollbar button when pressed. */
   private static final Color SCROLL_BUTTON_COLOR_PRESSED = new Color(0xBDE7FF);
   
   /** The background color of the of a scrollbar button in normal state. */
   private static final Color SCROLL_BUTTON_COLOR_NORMAL = new Color(0xFFFFFF);
   
   /** The background color of the of a scrollbar button when disabled. */
   private static final Color SCROLL_BUTTON_COLOR_DISABLED = new Color(0xFFFFFF);
   
   /** The background color of the of a scrollbar button when mouse is hovering it. */
   private static final Color SCROLL_BUTTON_COLOR_HIGHLIGHT = new Color(0xBDE7FF);
   
   /** The color of the of a scrollbar thumb/slider when pressed. */
   private static final Color SCROLL_SLIDER_COLOR_PRESSED = new Color(0xBDE7FF);
   
   /** The color of the of a scrollbar thumb/slider in normal state. */
   private static final Color SCROLL_SLIDER_COLOR_NORMAL = new Color(0xDEDFDE);
   
   /** The color of the of a scrollbar thumb/slider when mouse is hovering it. */
   private static final Color SCROLL_SLIDER_COLOR_HIGHLIGHT = new Color(0xBDE7FF);
   
   /**
    * Get the name of this {@code Theme}.
    *
    * @return  the name of this theme.
    */
   @Override
   public String getName()
   {
      return "windows8";
   }
   
   /**
    * Obtain the default system palette for this theme.
    *
    * @return  the default system palette for this theme.
    */
   @Override
   public Map<Integer, ColorRgb> getDefaultSystemPalette()
   {
      if (defaultColors == null)
      {
         Map<Integer, ColorRgb> pal = new HashMap<>();
         pal.put(COLOR_SCROLLBAR, new ColorRgb(new Color(212, 208, 200)));
         pal.put(COLOR_BACKGROUND, new ColorRgb(new Color(58, 110, 165)));
         pal.put(COLOR_ACTIVECAPTION, new ColorRgb(WINDOW_TITLEBAR_COLOR_ACTIVE));
         pal.put(COLOR_INACTIVECAPTION, new ColorRgb(WINDOW_TITLEBAR_COLOR_INACTIVE));
         pal.put(COLOR_MENU, new ColorRgb(MENU_POPUP_BACKGROUND_COLOR));
         pal.put(COLOR_WINDOW, new ColorRgb(new Color(0xFFFFFF)));
         pal.put(COLOR_WINDOWFRAME, new ColorRgb(new Color(100, 100, 100)));
         pal.put(COLOR_MENUTEXT, new ColorRgb(MENU_FOREGROUND_COLOR));
         pal.put(COLOR_WINDOWTEXT, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_CAPTIONTEXT, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_ACTIVEBORDER, new ColorRgb(WINDOW_TITLEBAR_COLOR_ACTIVE));
         pal.put(COLOR_INACTIVEBORDER, new ColorRgb(WINDOW_TITLEBAR_COLOR_INACTIVE));
         pal.put(COLOR_APPWORKSPACE, new ColorRgb(WORKSPACE_BACKGROUND_COLOR));
         pal.put(COLOR_HIGHLIGHT, new ColorRgb(HIGHLIGHT_BACKGROUND_COLOR));
         pal.put(COLOR_HIGHLIGHTTEXT, new ColorRgb(new Color(0xFFFFFF)));
         pal.put(COLOR_3DFACE, new ColorRgb(BACKGROUND_3DFACE_COLOR));
         pal.put(COLOR_3DSHADOW, new ColorRgb(new Color(0xA5A2A5)));
         pal.put(COLOR_GRAYTEXT, new ColorRgb(new Color(128, 128, 128)));
         pal.put(COLOR_BTNTEXT, new ColorRgb(BUTTON_FOREGROUND_COLOR));
         pal.put(COLOR_INACTIVECAPTIONTEXT, new ColorRgb(new Color(212, 208, 200)));
         pal.put(COLOR_BTNHIGHLIGHT, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_3DDKSHADOW, new ColorRgb(new Color(64, 64, 64)));
         pal.put(COLOR_3DLIGHT, new ColorRgb(new Color(212, 208, 200)));
         pal.put(COLOR_INFOTEXT, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_INFOBK, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_BUTTONALTERNATEFACE, new ColorRgb(new Color(181, 181, 181)));
         pal.put(COLOR_HOTLIGHT, new ColorRgb(new Color(0, 0, 128)));
         pal.put(COLOR_GRADIENTACTIVECAPTION, new ColorRgb(new Color(166, 202, 240)));
         pal.put(COLOR_GRADIENTINACTIVECAPTION, new ColorRgb(new Color(192, 192, 192)));
         pal.put(COLOR_MENUHILIGHT, new ColorRgb(MENU_BACKGROUND_SELECTED_COLOR));
         pal.put(COLOR_MENUBAR, new ColorRgb(MENU_BAR_BACKGROUND_COLOR));
         pal.put(COLOR_TOOLTIPBK, new ColorRgb(TOOLTIP_BACKGROUND_COLOR));
         pal.put(COLOR_FILLIN_BG, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_FILLIN_FG, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_CARET, new ColorRgb(new Color(0, 0, 0)));
         
         defaultColors = Collections.unmodifiableMap(pal);
      }
      
      return defaultColors;
   }
   
   /**
    * Obtain the map of standard fonts specific to current {@code Theme}, to which a key defaults,
    * if not overridden in the directory.
    *
    * @return  the map of standard fonts specific to this {@code Theme}. 
    */
   @Override
   public Map<String, String> getStandardFontTable()
   {
      Map<String, String> fontTable = new HashMap<>();
      
      fontTable.put("default-font",        /* "Segoe UI, size=9"); */ "MS Sans Serif, size=8");
      fontTable.put("default-fixed-font",  "Courier New, size=8");
      
      // standard fonts
      fontTable.put("font0", "Courier New, size=8");
      fontTable.put("font1", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      fontTable.put("font2", "Courier New, size=8");
      fontTable.put("font3", "Courier New, size=8");
      fontTable.put("font4", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      fontTable.put("font5", "Segoe UI, size=10");        //"MS Sans Serif, size=10");
      fontTable.put("font6", "Segoe UI, size=9, bold");   //"MS Sans Serif, size=8, bold");
      fontTable.put("font7", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      
      // system fonts
      fontTable.put(FontTable.WINDOW_TITLE_FONT,        "Segoe UI, size=10");   //"Tahoma, size 8, bold");
      fontTable.put(FontTable.ALERT_BOX_FONT,           "Segoe UI, size=9");    //"Tahoma, size=8");
      fontTable.put(FontTable.STACK_TRACE_WINDOW_FONT,  "Segoe UI, size=8");    //"MS Sans Serif, size=8");
      fontTable.put(FontTable.MENU_FONT,                "Segoe UI, size=9");    //"Tahoma, size 8");
      fontTable.put(FontTable.SORT_NUMBER_FONT,         "Segoe UI, size=7");    //"MS Sans Serif, size=7");
      
      // freeze the map
      return Collections.unmodifiableMap(fontTable);
   }
   
   /**
    * Obtain the location of binary resources (like pixmaps and sprites) used by this theme. The 
    * location is relative to @Theme.class@ location so that the resources could be loaded easily
    * using @getResourceAsStream()@.
    *
    * @return  the location of binary resources used by this theme.
    */
   @Override
   public String getBinaryResourceLocation()
   {
      return "windows8";
   }
   
   /**
    * Returns the window's border insets. In Windows8 theme, all insets are NOT equals. The top
    * is 0 (zero), leaving space for taller titlebar, whose caption buttons are top aligned.
    *
    * @param   modal
    *          Usually the modal dialogs/windows have a different width if modal. 
    *
    * @return  the insets of window's border. 
    */
   @Override
   public NativeInsets getWindowBorderInsets(boolean modal)
   {
      return new NativeInsets(0, WINDOW_BORDER_WIDTH, WINDOW_BORDER_WIDTH, WINDOW_BORDER_WIDTH);
   }
   
   @Override
   public int getTitleBarHeight(boolean maximized)
   {
      return 24 + 8;
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border
    *          The width of the border on each side.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowBorder(GuiDriver gd,
                                int x,
                                int y,
                                int width,
                                int height,
                                NativeInsets border,
                                boolean active)
   {
      gd.setColor(active ? WINDOW_TITLEBAR_COLOR_ACTIVE :
                           WINDOW_TITLEBAR_COLOR_INACTIVE);
      gd.fillRect(x, y + height - border.bottom, width, border.bottom); // S
      gd.fillRect(x, y, border.left, height - border.bottom); // W
      gd.fillRect(x + width - border.right, y, border.right, height - border.bottom); // E
      
      if (active)
      {
         gd.setColor(new Color(0x73AAD6));
         gd.drawLine(x + border.left - 1,         y + height - border.bottom,
                     x + width - border.left + 1, y + height - border.bottom);
      }
      
      gd.setColor(active ? new Color(0x293031)
                         : new Color(0x5a6163));
      gd.drawRect(x, y, x + width - 1, y + height - 1);
   }
   
   /**
    * Get the alignment of caption button for this theme. The returned value combines both
    * horizontal and vertical alignment of the buttons.
    *
    * @param   type
    *          On some specific themes, the alignment of buttons is not the same.
    **
    * @return  the alignment of caption button.
    */
   @Override
   public int getCaptionButtonsAlignment(CaptionButtonType type)
   {
      switch (type)
      {
         case CLOSE_SMALL:
            return ALIGN_RIGHT | ALIGN_MIDDLE;
         default:
            return ALIGN_RIGHT | ALIGN_TOP;
      }
   }
   
   /**
    * Get the title alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of title.
    */
   @Override
   public int getWindowTitleAlignment()
   {
      return ALIGN_CENTER | ALIGN_MIDDLE;
   }
   
   /**
    * Returns the height of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the height of buttons is not the same.
    * @param   maximized
    *          On some themes, the maximized window have a different button height when 
    *          maximized. 
    *
    * @return  the height of the caption button. 
    */
   @Override
   public int getCaptionButtonHeight(CaptionButtonType type, boolean maximized)
   {
      switch (type)
      {
         case CLOSE_SMALL:
            return 21;
         default:
            return maximized ? 20 : 23;
      }
   }
   
   /**
    * Returns the width of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the width of buttons is not the same.
    *
    * @param   dialog
    *          {@code true} for dialogs. The dialogs may have specific width for caption buttons.  
    *
    * @return  the width of the requested caption button. 
    */
   @Override
   public int getCaptionButtonWidth(CaptionButtonType type, boolean dialog)
   {
      if (dialog)
      {
         return 36;
      }
      
      switch (type)
      {
         case ICONIFY:
            return 30;
         case MAXIMIZE:
            return 28;
         case CLOSE:
            // 2 px less. At the moment we cannot paint this button overlapped to East border so
            // we cut it without moving the other caption buttons 
            return 47;
         case CLOSE_SMALL:
            return 21;
         default:
            return 30;
      }
   }
   
   /**
    * Do the layout of a titlebar. Each {@code Theme} can define how its components are arranged. 
    *
    * @param   window
    *          The window whose titlebar is laid out. 
    * @param   icon
    *          The icon widget of the window (if any).
    * @param   title
    *          The {@link Label} that displays the window's title. 
    * @param   buttons
    *          An array with the {@link CaptionButton}-s present on the titlebar of this window. 
    * @param   width
    *          The width of the window (in character units).
    * @param   c
    *          the {@link CoordinatesConversion} used for possible units conversions.
    *
    * @return  the computed titlebar height.
    */
   @Override
   public double layoutTitleBar(TopLevelWindow window,
                                Widget icon,
                                Label title,
                                CaptionButton[] buttons,
                                double width,
                                CoordinatesConversion c)
   {
      int ICON_INSETS = 2; //px
      boolean isWindowMaximized = window.isMaximized();
      boolean isDialog = window.ancestor() != window;
      int font = window.resolveTitleFont();
      // compute the window title's font height, in pixels.  this will determine the titlebar's
      // height.
      int fHeight = c.heightToNative(1, window.getId(), font);
      int barNativeHeight = Math.max(fHeight, getTitleBarHeight(isWindowMaximized));
      double barHeight = c.heightFromNative(barNativeHeight); // to return !!
      
      if (icon != null)
      {
         // - icon on left-side; to ensure the character unit location is properly in sync with
         // the native location, compute the character location first and re-transform to native
         // for the physical location.
         NativeDimension id = icon.physicalDimension();
         int vAlign = getWindowIconAlignment() & Theme.VERTICAL_ALIGN_MASK;
         int dy = 0;
         switch (vAlign)
         {
            case Theme.ALIGN_BOTTOM:
               dy = barNativeHeight - id.height;
               break;
            case Theme.ALIGN_MIDDLE:
               dy = (barNativeHeight - id.height) / 2;
               break;
         }
         icon.setLocation(c.widthFromNative(ICON_INSETS), c.heightFromNative(dy));
         NativePoint iconLoc = c.pointToNative(icon.location());
         icon.setPhysicalLocation(iconLoc.x, iconLoc.y);
      }
      
      if (buttons.length > 0)
      {
         // add the buttons (minimize, maximize, close buttons)
         // all coordinates are relative to this container
         double bx = width - c.widthFromNative(1);
         
         // determine the vertical location so that the button is centered vertically
         double by = 0;
         
         CaptionButtonType fstButtonType = buttons[0].getType();
         int btnHeight = getCaptionButtonHeight(fstButtonType, isWindowMaximized);
         int alignment = getCaptionButtonsAlignment(fstButtonType);
         
         switch (alignment & Theme.VERTICAL_ALIGN_MASK)
         {
            case Theme.ALIGN_MIDDLE:
               by = c.heightFromNative((barNativeHeight - btnHeight) / 2);
               break;
            case Theme.ALIGN_BOTTOM:
               by = c.heightFromNative(barNativeHeight - btnHeight);
               break;
         }
         double bh = c.heightFromNative(btnHeight);
         
         for (int i = buttons.length - 1; i >= 0; i--)
         {
            CaptionButton btn = buttons[i];
            double bw = c.widthFromNative(getCaptionButtonWidth(btn.getType(), isDialog));
            btn.setDimension(bw, bh);
            
            int buttonSpace = getCaptionButtonWidth(btn.getType(), isDialog) +
            getSpaceBeforeButton(btn.getType());
            bx -= c.widthFromNative(buttonSpace);
            
            btn.setLocation(bx, by);
            NativePoint btnLoc = c.pointToNative(btn.location());
            btn.setPhysicalLocation(btnLoc.x, btnLoc.y);
         }
      }
      
      double iconColumn = 0;
      if (icon != null && !icon.hidden())
      {
         iconColumn = icon.bounds().right();
      }
      
      // title in the middle, between the icon and the minimize button
      int tAlign = getWindowTitleAlignment();
      NativePoint titleLoc = new NativePoint(c.columnToNative(iconColumn) + ICON_INSETS, 0);
      
      // compute title width
      double btnsCol = buttons.length == 0 ? width : buttons[0].bounds().left();
      double twidth = btnsCol - iconColumn; // extract 3 * ICON_INSETS
      switch (tAlign & Theme.VERTICAL_ALIGN_MASK)
      {
         case Theme.ALIGN_BOTTOM:
            titleLoc.y = barNativeHeight - fHeight;
            break;
         case Theme.ALIGN_MIDDLE:
            // compute including the window upper border
            fHeight += 4;
            titleLoc.y = (barNativeHeight - fHeight) / 2 - 2;
            break;
      }
      title.setHeight(c.heightFromNative(fHeight));
      
      // Windows8 draws titlebar centered! use tAlign & Theme.HORIZONTAL_ALIGN_MASK:
      String vTitle = window.title();
      if (vTitle == null)
      {
         vTitle = WindowGuiImpl.DEFAULT_TITLE;
      }
      int fNatWidth = 4 + window.getTextWidthNative(vTitle, font);
      int tNatWidth = c.columnToNative(twidth);
      int btnNatLeft = buttons.length == 0 ?
                       c.columnToNative(width) :
                       c.columnToNative(buttons[0].bounds().left());
      if (fNatWidth < tNatWidth)
      {
         // W8 prefer to center the title:
         int prefX = (c.columnToNative(width) - fNatWidth) / 2;
         
         // check if this does not overlap the caption buttons:
         if (prefX + fNatWidth > btnNatLeft)
         {
            // if they overlap, we center the text in available space
            int availableSpace = btnNatLeft - titleLoc.x;
            prefX = titleLoc.x + (availableSpace - fNatWidth) / 2;
         }
         
         // set the offset to preferred location and adjust the label width:
         titleLoc.x = prefX;
         twidth = c.columnFromNative(fNatWidth); 
      }
      
      title.setLocation(c.pointFromNative(titleLoc));
      title.setPhysicalLocation(titleLoc.x, titleLoc.y);
      
      // set title width
      title.setWidth(twidth);
      
      return barHeight;
   }
   
   /**
    * Return the amount of space to the right of this button. On some themes, different spaces can
    * be found between the caption buttons. Although unusual, negative values can be returned,
    * when buttons overlap for a few pixels. 
    *
    * @param   type
    *          The caption button type.
    *
    * @return  the amount of space to the right of this button.
    */
   @Override
   public int getSpaceBeforeButton(CaptionButtonType type)
   {
      return -1; // they are 1 px overlapped, in fact
   }
   
   /**
    * Renders a {@link CaptionButton}.
    *
    * @param   button
    *          The {@code CaptionButton} to be drawn. 
    * @param   gd
    *          The {@link GuiDriver} used for drawing operations. Note that internal state of 
    *          this object can affect the resulting output.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawCaptionButton(CaptionButton button,
                                 GuiDriver gd,
                                 int x,
                                 int y,
                                 int width,
                                 int height)
   {
      TopLevelWindow<GuiOutputManager> window = button.topLevelWindow();
      boolean active = WindowManager.isWindowActive(window);
      boolean enabled = button.isEnabled();  // TODO: testcase
      boolean mouseOver = button.isMouseOver();
      boolean pressed = button.isPressed() && mouseOver; // even is pressed, mouse exited bounds
      
      // close button is handled specially because it uses a different palette
      CaptionButtonType type = button.getType();
      boolean close = type == CaptionButtonType.CLOSE || type == CaptionButtonType.CLOSE_SMALL;
      gd.setColor(!active ? CAPTION_BUTTON_COLOR_INACTIVE :
                  pressed ? (close ? CAPTION_BUTTON2_COLOR_DOWN :
                                     CAPTION_BUTTON_COLOR_DOWN) :
                  mouseOver ? (close ? CAPTION_BUTTON2_COLOR_HIGHLIGHT :
                                       CAPTION_BUTTON_COLOR_HIGHLIGHT) :
                  /* normal ? */ (close ? CAPTION_BUTTON2_COLOR_ACTIVE :
                                          CAPTION_BUTTON_COLOR_ACTIVE));
      gd.fillRect(x + 1, y + 1, width - 2, height - 2);
      
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/wnd-");
      switch (button.getType())
      {
         case ICONIFY:
            sb.append("min");
            break;
         case MAXIMIZE:
            sb.append("max");
            break;
         case CLOSE:
            sb.append("close");
            break;
         case CLOSE_SMALL:
            sb.append("close-small");
            break;
         case HELP:
            sb.append("help");
            break;
         case RESTORE_DOWN:
         case RESTORE_UP:
            sb.append("restore");
            break;
      }
      sb.append(".png");
   
      ImageWrapper img = getPngImage(sb.toString());
      if (img != null)
      {
         gd.drawImage(img,
               x + (width - img.getWidth()) / 2, y + (height - img.getHeight()) / 2,
               img.getWidth(), img.getHeight(),
               0, 0, false, false, true, false);
      }
   
      gd.setColor(!active ? new Color(0x5A6163) :
            pressed ? (close ? new Color(0x5A1410) :
                               new Color(0x00455A)) :
            mouseOver ? (close ? new Color(0x7B1C10) :
                                 new Color(0x005573)) :
            /* normal ? */ (close ? new Color(0x5A3831) :
                                    new Color(0x313031)));
      gd.drawRect(x, y, width - 1, height - 1);
   }
   
   /**
    * Renders the text on the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@link GuiFontResolver} for selecting the required font.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   title
    *          The title of the window.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitle(GuiDriver gd,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               String title,
                               boolean active)
   {
      // remove old title
      fillTitlebar(gd, x, y, width, height, active, true);
      
      if (title.length() > 0)
      {
         FontDetails<?> font = gf.font().font;
         gd.setGuiFont(font);
         
         // draw text
         gd.setColor(active ? WINDOW_TITLE_COLOR_ACTIVE :
                              WINDOW_TITLE_COLOR_INACTIVE);
         gd.drawStringCentered(title, x, height / 2);
      }
   }
   
   /**
    * Renders the background of the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitleBar(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean active)
   {
      fillTitlebar(gd, x, y, width, height, active, false);
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   button
    *          The {@link ButtonGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawButton(ButtonGuiImpl button,
                          GuiDriver gd,
                          GuiColorResolver gc,
                          GuiFontResolver gf,
                          int x,
                          int y,
                          int width,
                          int height)
   {
      ButtonConfig bcfg =   button.config();
      boolean flat =        bcfg.flatButton;
      boolean highlighted = button.highlighted() && !flat;
      boolean isDefault =   button.isDefaultButton();
      boolean mouseOver =   button.isMouseOver();
      boolean pressed =     button.isPressed() || mouseOver; // even is pressed, mouse exited bounds
      
      // draw button text
      FontDetails fnt = gf.font().font;
      gd.setGuiFont(fnt);
      int fontNum = gf.resolveFontNum();
      MnemonicInfo labelInfo = button.getMnemonic();
      String label = (labelInfo == null) ? button.textLabel() : labelInfo.getPrepLabel();
   
      // normal buttons are drawn smaller than actually laid out, the border is WHITE,
      // the FLAT buttons instead, cover the entire area
      if (!flat)
      {
         gd.setColor(new Color(0xFFFFFF));
         gd.drawRect(x, y, width - 1, height - 1);
         ++x;
         ++y;
         width -= 2;
         height -= 2;
      }
      
      // do not simplify the line below, this makes the font height to be up nearest even number 
      int tY = y + (height - 1) / 2;
      int tX = x + (width - button.getTextWidthNative(label, fontNum)) / 2;
      gd.setFontStyle(fnt.style);
      
      if (!button.isEnabled()) // disabled button
      {
         // for flat buttons only the text is displayed
         if (!flat)
         {
            gd.setColor(new Color(0xF7F7F7)); // button fill
            gd.fillRect(x + 1, y + 1, width - 2, height - 2);
            
            gd.setColor(new Color(0xADB2B5)); // button border
            gd.drawRect(x, y, width - 1, height - 1);
         }
      }
      else if (pressed)
      {
         gd.setColor(new Color(0xBDE7FF)); // button fill
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
         gd.setColor(new Color(0x397DB5)); // button border
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else if (!flat || mouseOver)
      {
         // button fill
         gd.setColor(mouseOver ? new Color(0xBDE7FF) : // mouse-on button color 
                                 new Color(0xDEDFDE)); // normal button color 
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
         // button border
         gd.setColor(isDefault || mouseOver ? new Color(0x397DB5) : // default/hover button border
                     highlighted            ? new Color(0x21A2DE) : // mouse hover
                                              new Color(0x737173)); // normal button border
         gd.drawRect(x, y, width - 1, height - 1);
      }
      
      // draw button content
      if (button.hasImages())
      {
         ImageGuiImpl img = null;
         
         // draw button body images
         if (pressed)
         {
            // determine what image to draw
            ImageGuiImpl imgDown = button.getImgDown();
            if (imgDown != null)
            {
               img = imgDown;
            }
            else
            {
               ImageGuiImpl imgUp = button.getImgUp();
               if (imgUp != null)
               {
                  img = imgUp;
               }
            }
         }
         else
         {
            // determine what image to draw
            ImageGuiImpl imgUp = button.getImgUp();
            if (button.isEnabled() && imgUp != null)
            {
               img = imgUp;
            }
            else
            {
               ImageGuiImpl imgDisabled = button.getImgDisabled();
               if (!button.isEnabled() && imgDisabled != null)
               {
                  img = imgDisabled;
               }
            }
         }
         
         // if we found it, draw!
         if (img != null)
         {
            img.draw();
         }
      }
      else
      {
         // draw text:
         if (button.isEnabled())
         {
            gd.setColor(gd.getSysColor(COLOR_BTNTEXT)); // normal text
         }
         else
         {
            gd.setColor(new Color(0x848284)); // disabled text, not in palette ?
         }
         if (labelInfo != null && labelInfo.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(tX, tY);
            gd.drawStringWithMnemonic(label, tp, labelInfo, gf, button, true);
         }
         else
         {
            // no mnemonic available
            gd.drawStringCentered(label, tX, tY);
         }
      }
      
      // focused button needs special highlight
      if (button.isEnabled() && highlighted && !button.config().noFocus)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         
         gd.setColor(pressed ? new Color(0x3C1407)   // pressed button selection dotted line
                             : new Color(0x292C29)); // normal button selection dotted line
         gd.drawRect(x + 2, y + 2, width - 5, height - 5);
         
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Draw a {@link ToolTip}.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   fnt
    *          The {@code FontDetails} to be used when drawing the text.
    * @param   text
    *          The actual text to be drawn.
    */
   @Override
   public void drawToolTip(GuiDriver gd,
                           int x,
                           int y,
                           int width,
                           int height,
                           FontDetails fnt,
                           String text)
   {
      // start batch drawing
      gd.startBatch();
      try
      {
         // another version of clipping
         gd.setDrawingArea(new NativeRectangle(y, x, y + height - 1, x + width - 1));
         
         // tooltip background
         gd.setColor(gd.getSysColor(COLOR_TOOLTIPBK));
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
         // tooltip frame
         gd.setColor(new Color(0x737573));
         gd.drawRect(x, y, width - 1, height - 1);
         
         // tooltip text
         gd.setGuiFont(fnt);
         gd.setFontStyle(fnt.style);
         gd.setColor(new Color(0x525552));
         
         String lines[] = text.trim().split("\\r?\\n");
         int tH = (height - 4) / lines.length;
         int tY = y + tH / 2 + 2;
         for (String line : lines)
         {
            gd.drawStringCentered(line, x + 2, tY);
            tY += tH;
         }
      }
      finally
      {
         // draw immediately
         gd.setDrawingArea(null);
         gd.endBatch();
      }
   }
   
   /**
    * Obtain the size of the scroll button on the direction perpendicular to scroll direction.
    *
    * @return  size of the scroll button on the direction perpendicular to scroll direction.
    */
   @Override
   public int getScrollBarButtonWidth()
   {
      return 17;
   }
   
   /**
    * Obtain the thumb/slider width in a scrollbar widget.
    *
    * @return  the thumb/slider width in a scrollbar widget.
    */
   @Override
   public int getScrollBarThumbWidth()
   {
      return 17;
   }
   
   /**
    * Renders the gauge (background) of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarGauge(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean horizontal)
   {
      gd.setColor(SCROLL_GAUGE_COLOR);
      gd.fillRect(x, y, width, height);
      
      gd.setColor(SCROLL_BORDER_COLOR);
      if (horizontal)
      {
         gd.drawLine(x, y,              x + width - 1, y);
         gd.drawLine(x, y + height - 1, x + width,     y + height - 1);
      }
      else
      {
         gd.drawLine(x,             y, x,              y + height - 1);
         gd.drawLine(x + width - 1, y, x + width - 1 , y + height - 1);
      }
   }
   
   /**
    * Renders the thumb/slider of the scrollbar widget.
    * 
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          If the mouse is dragging the thumb of the scrollbar.
    * @param   hover
    *          Is mouse hovering this component ?
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarThumb(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean hover,
                                  boolean horizontal)
   {
      gd.setColor(pressed || hover ? SCROLL_SLIDER_COLOR_PRESSED : SCROLL_SLIDER_COLOR_NORMAL);
      gd.fillRect(x + 1, y + 1, width - 2, height - 2);
      
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/scroll-thumb-");
      sb.append(horizontal ? "hor-" : "vert-");
      sb.append(pressed ? "down" :
                hover ? "hi" :
                /*normal ?*/ "en");
      sb.append(".png");
      
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      if (img != null && img.getImage() != null)
      {
         // if the button is too small, don't draw the 'groves'
         if (horizontal && width > img.getWidth() + 10 ||
             !horizontal && height > img.getHeight() + 10)
         {
            gd.drawImage(img, x + (width - img.getWidth()) / 2, 
                              y + (height - img.getHeight()) / 2,
                              img.getWidth(), img.getHeight(), 0, 0, false, false, true, false);
         }
      }
      
      // draw the border now:
      gd.setColor(pressed ? new Color(0x007194) :
                  hover ? new Color(0x397DB5) :
                  /*default ?*/ new Color(0xBDBEBD));
      gd.drawRect(x, y, width - 1 , height - 1);
   }
   
   /**
    * Renders a button of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pos
    *          The position of the button (UP/DOWN, LEFT/RIGHT).
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    * @param   enabled
    *          If the button is enabled.
    * @param   pressed
    *          If the button is down (pressed by mouse).
    * @param   highlighted
    *          If the mouse is hovering over the button.
    */
   @Override
   public void drawScrollBarButton(GuiDriver gd,
                                   GuiColorResolver gc,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   ScrollBar.Position pos,
                                   boolean horizontal,
                                   boolean enabled,
                                   boolean pressed,
                                   boolean highlighted)
   {
      gd.setColor(!enabled ? SCROLL_BUTTON_COLOR_DISABLED :
                  pressed ? SCROLL_BUTTON_COLOR_PRESSED :
                  highlighted ? SCROLL_BUTTON_COLOR_HIGHLIGHT :
                  /*default ?*/ SCROLL_BUTTON_COLOR_NORMAL);
      gd.fillRect(x + 1, y + 1, width - 2, height - 2);
      
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/scroll-btn-");
      if (horizontal)
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "left-" : "right-");
      }
      else
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "up-" : "down-");
      }
      sb.append(!enabled ? "dis" :
                pressed ? "down" :
                highlighted ? "hi" : 
                /*normal ?*/ "en");
      sb.append(".png");
      
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      if (img != null && img.getImage() != null)
      {
         gd.drawImage(img, x + (width - img.getWidth()) / 2, y + (height - img.getHeight()) / 2,
               img.getWidth(), img.getHeight(), 0, 0, false, false, true, false);
      }
   
      // draw the border now:
      gd.setColor(!enabled ? new Color(0xBDBEBD) :
                  pressed ? new Color(0x007194) :
                  highlighted ? new Color(0x397DB5) :
                  /*default ?*/ new Color(0xBDBEBD));
      gd.drawRect(x, y, width - 1, height - 1);      
   }
   
   /**
    * Renders a GUI toggle box widget.
    * @param   cb
    *          The {@link ToggleBoxGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the checkbox is pressed (the mouse button is down on the button).
    * @param   hover
    *          {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   @Override
   public void drawToggleBox(ToggleBoxGuiImpl cb,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             GuiFontResolver gf,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean pressed,
                             boolean hover)
   {
      NativePoint po = new NativePoint(x, y);
      String text = cb.getMnemonicLabel(false).orElse(cb.config().label);
      GuiFontResolver.FontCache fntCache = gf.font();
      boolean enabled = cb.isEnabled();
      boolean highlighted = cb.highlighted(); // toggle-box state flags
      
      // clear the area
      gd.setColor(gc.bgColor);
      gd.fillRect(po.x, po.y, width, height);
      
      // draw the checkbox as an image 
      ToggleBoxConfig config = cb.config();
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/checkbox-");
      sb.append(config.checked ? "on-" : "off-"); // TODO: add support for 3-state checkbutton
      if (!enabled)
      {
         sb.append("dis");
      }
      else if (pressed)
      {
         sb.append("down");
      }
      else if (hover) // highlighted / mouse over
      {
         sb.append("hi");
      }
      else
      {
         sb.append("en");
      }
      sb.append(".png");
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      int imgWidth = 0;
      if (img != null)
      {
         int imgHeight = img.getHeight();
         imgWidth = img.getWidth();
         int y2 = y + (height - imgHeight) / 2;
         // TODO: enlarge the images and use automatic transparency --v
         gd.drawImage(img, x, y2, imgWidth, imgHeight, 0, 0, false, false, true, false);
      }
      
      // draw the text to the right of the checkbox
      if (text != null && text.length() > 0)
      {
         int xt = po.x + imgWidth + getCheckBoxSpace(); // space between checkbox and label
         int yt = po.y + height / 2;
         
         gd.setGuiFont(fntCache.font);
         gd.setFontStyle(fntCache.font.style);
         
         // draw the light text first!
         gd.setColor(enabled ? new Color(0x000000) :
                               new Color(0x6B6D6B));
         
         MnemonicInfo info = cb.getMnemonic();
         if (info != null && info.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(xt, yt);
            gd.drawStringWithMnemonic(text, tp, info, gf, cb, true);
         }
         else
         {
            gd.drawStringCentered(text, xt, yt);
         }
         
         // focused toggle-box needs special highlight
         if (highlighted)
         {
            int font = gf.resolveFontNum();
            int textWidth = cb.getTextWidthNative(text, font) + 1;
            int textHeight = cb.getTextHeightNative(text, font) + 2;
            yt = po.y + (height - textHeight) / 2;
            
            gd.setColor(new Color(0x000000));
            gd.setLineStroke(LineStroke.DOTTED_SMALL_THIN);
            gd.drawRect(xt - 1, yt, textWidth, textHeight);
            gd.setLineStroke(LineStroke.DEFAULT);
         }
      }
   }
   
   /**
    * Obtain the image from a radio button with specified attributes. 
    *
    * @param   on
    *          The button state. {@code true} if button is on.
    * @param   enabled
    *          {@code true} is button enabled. 
    * @param   hover
    *          {@code true} if mouse is hovering over the button. Some styles are mouse hovering
    *          sensitive. 
    * @param   pressed
    *          {@code true} if the radio button is pressed (the mouse button is down on the 
    *          button).
    *
    * @return  The {@code ImageWrapper} containing the image with requested properties. 
    */
   @Override
   protected ImageWrapper getRadioButtonImage(boolean on,
                                              boolean enabled,
                                              boolean hover,
                                              boolean pressed)
   {
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/radio-");
      sb.append(on ? "on-" : "off-");
      if (!enabled)
      {
         sb.append("dis");
      }
      else if (pressed)
      {
         sb.append("down");
      }
      else if (hover) // highlighted / mouse over
      {
         sb.append("hi");
      }
      else
      {
         sb.append("en");
      }
      sb.append(".png");
      
      return getPngImage(sb.toString());
   }
   
   /**
    * Renders the opening button of the combo box widget. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxButton(ComboBoxGuiImpl combo,
                                  GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean enabled,
                                  boolean hover)
   {
      boolean isDropDown = combo.config().mode == ComboBoxConfig.Mode.DROP_DOWN;
      // otherwise it's a DROP_DOWN_LIST because SIMPLE does not have a button
      
      if (enabled && isDropDown && (hover || pressed))
      {
         gd.setColor(pressed ? new Color(0xCCE4F7) :
                     /*hover*/ new Color(0xE5F1FB));
         gd.fillRect(x + width - COMBOBOX_BUTTON_WIDTH, y, x + width - 1, y + height - 1);
      }
      
      // just draw the down arrow for NON-editable cbx
      ImageWrapper<BufferedImage> img = getPngImage(getBinaryResourceLocation() + 
                                                    "/combo-arrow.png");
      if (img != null && img.getImage() != null)
      {
         // this is the location of the button: 
         int x1 = x + width - COMBOBOX_BUTTON_WIDTH;
         int y1 = y;
         // centering image:
         x1 += (COMBOBOX_BUTTON_WIDTH - img.getWidth()) / 2;
         y1 += (height - img.getHeight()) / 2;
         
         gd.drawImage(img, x1, y1, img.getWidth(), img.getHeight(), 0, 0, true, false, true, false);
      }
      
      if (enabled && isDropDown && (hover || pressed))
      {
         gd.setColor(pressed ? new Color(0x005499) :
                     /*hover*/ new Color(0x0078D7));
         gd.drawRect(x + width - COMBOBOX_BUTTON_WIDTH - 1, y, 
                     COMBOBOX_BUTTON_WIDTH, y + height - 1);
      }
   }
   
   /**
    * Renders the border (and possibly the background) of the combo box widget. Draw 3D border 
    * around different parts of the combo-box widget.
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxFieldBorder(ComboBoxGuiImpl combo,
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height,
                                       boolean pressed,
                                       boolean enabled,
                                       boolean hover)
   {
      boolean simple = combo.isSimpleMode();
      gd.setColor(!enabled ? new Color(0xBFBFBF) :
                   pressed ? new Color(0x29618C) :
                     hover ? new Color(0x21A2DE) :
                    simple ? new Color(0x29618C) :
                  /*normal*/ new Color(0x737173));
      gd.drawRect(x, y, width - 1, height - 1);
   }
   
   /**
    * Renders the entry field of a combo box. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose entry field is painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   font
    *          A {@code FontDetails} used for text rendering.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxEntryField(ComboBoxGuiImpl combo,
                                      GuiDriver gd,
                                      GuiColorResolver gc,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      FontDetails<?> font,
                                      boolean pressed,
                                      boolean enabled,
                                      boolean hover)
   {
      boolean isDropDown = combo.config().mode == ComboBoxConfig.Mode.DROP_DOWN;
      boolean isDDL = combo.config().mode == ComboBoxConfig.Mode.DROP_DOWN_LIST;
      
      if (!enabled)
      {
         gd.setColor(isDDL ? new Color(0xFFFFFF) :
                             new Color(0xCCCCCC));
      }
      else if (isDropDown)
      {
         gd.setColor(new Color(0xFFFFFF));
      }
      else
      {
         // SIMPLE & DROP_DOWN_LIST cases
         gd.setColor(pressed ? new Color(0xC6E7F7) :
                       hover ? new Color(0xBDE7FF) :
                    /*normal*/ new Color(0xFFFFFF));
      }
      gd.fillRect(x, y, width, height);
      
      // text only in DROP-DOWN-LIST mode. In SIMPLE or DROP-DOWN modes the entry field drawing
      // is happening from the frame the entry field belongs to.
      if (!isDDL)
      {
         return;
      }
      
      // get the text to draw with field
      String display = null;
      if (combo.isSelected())
      {
         display = combo.model().selected();
      }
      else
      {
         BaseDataType initial = combo.getValue();
         if (initial != null && !initial.isUnknown())
         {
            display = combo.getInitialScreenText(initial);
         }
      }
      if (display == null)
      {
         display = "";
      }
   
      // limit the data portion of the field to the maximum allowed size
      int limit = combo.getFormatLength();
   
      if (display.length() > limit && combo.isFormatHonored())
      {
         display = display.substring(0, limit);
      }
   
      // draw entry field text
      gd.setGuiFont(font);
      // left-center alignment
      int xt = x + ComboBoxGuiImpl.TEXT_SHIFT;
      int yt = y + height / 2;
      gd.setFontStyle(font.style);
      // enabled state controls text drawing
      gd.setColor(combo.isEnabled() ? new Color(0x000000) :
                                      new Color(0x6D6D6D));
      int textWidth = width - COMBOBOX_BUTTON_WIDTH - ComboBoxGuiImpl.TEXT_SHIFT;
      final String finalDisplay = display;
      gd.draw(new NativePoint(0, 0), new NativeRectangle(0, 0, height, textWidth),
              () -> gd.drawStringCentered(finalDisplay, xt, yt));
   
      // draw dotted rectangle around focused item text
      if (combo.hasFocus() && combo.isEnabled())
      {
         // black background for dotted line
         gd.setLineStroke(LineStroke.DOTTED_SMALL); // set dotted line style
         gd.setColor(new Color(0x1E1E1E));
      
         gd.drawRect(x + 3, y + 3, textWidth - 3, height - 7);
         gd.setLineStroke(LineStroke.DEFAULT); // restore default line style
      }
   }
   
   /**
    * Renders the border (and possibly the background) of the fill-in widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border3D
    *          If fill-in widget has a 3d border. Some flat {@code Theme}s may ignore this.
    * @param   hasFocus
    *          The fill-in widget is focused.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawFillInBorder(GuiDriver gd,
                                GuiColorResolver gc,
                                int x,
                                int y,
                                int width,
                                int height,
                                boolean border3D,
                                boolean hasFocus,
                                boolean hover)
   {
      if (!border3D)
      {
         gd.setColor(new Color(0x636563)); // non 3D always same color
      }
      else
      {
         gd.setColor(hasFocus ? new Color(0x29618C) : // has focus 
                        hover ? new Color(0x21A2DE) : // mouse hover
                       /*else*/ new Color(0x737173)); // non-focused
                       /* disabled: 0xADB2B5 */
      }
      gd.drawRect(x, y, width - 1, height - 1);
   }
   
   /**
    * Renders the background of an item in a menu widget.
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuItemBackground(AbstractWidget item,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      boolean hover)
   {
      if (Menu.isMenubarElement(item))
      {
         System.err.println(
            "WARNING: Use Theme.drawMenuBarElement() instead to render a Menu Bar element");
         return;
      }
      
      boolean enabled = item.isEnabled();
      if (hover)
      {
         gd.setColor(enabled ? new Color(0xDEEFFF)
                             : new Color(0xEFEFEF));
         gd.fillRect(1, 1, width - 2, height - 2);
         gd.setColor(enabled ? new Color(0x21A2DE)
                             : new Color(0x848284));
         gd.drawRect(0, 0, width - 1, height - 1);
      }
      else 
      {
         // normal menu, just draw the vertical grey line
         gd.setColor(new Color(0xBDBEBD));
         int vLineXoffset = x + MENU_CHECKMARK_WIDTH + MENU_LEFT_PAD / 2;
         gd.drawLine(vLineXoffset, 0, vLineXoffset, height - 1);
      }
      
      ThemeManager.debug(() -> {
         gd.setColor(Color.GREEN);
         gd.drawRect(0, 0, width - 1, height - 1);
      });
   }
   
   /**
    * Obtain the fixed height of a separator in a popup menu widget.
    *
    * @return  the height of a separator in a popup menu.
    */
   @Override
   public int getMenuSeparatorHeight()
   {
      return 4;
   }
   
   /**
    * Renders the separator line in a menu widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawMenuSeparator(GuiDriver gd, int x, int y, int width, int height)
   {
      gd.setColor(new Color(0xBDBEBD));
      int vLineXoffset = x + MENU_CHECKMARK_WIDTH + MENU_LEFT_PAD / 2;
      gd.drawLine(vLineXoffset, 0, vLineXoffset, height - 1);
      gd.drawLine(vLineXoffset, y + height / 2 - 1, x + width, y + height / 2 - 1);
   }
   
   /**
    * Renders the background of a menu widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    * @param   popup
    *          {@code true} for popup menus.
    */
   @Override
   public void drawMenuBackground(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean popup)
   {
      if (!popup)
      {
         // this is a menu bar
         gd.setColor(gd.getSysColor(COLOR_MENUBAR));
         gd.fillRect(x, y, width, height);
         return;
      }
      
      gd.setColor(gd.getSysColor(COLOR_MENU));
      gd.fillRect(x, y, width, height);
      
      gd.setColor(new Color(0x949694));
      gd.drawRect(x, y, x + width - 1, y + height - 1);
   }
   
   /**
    * Get the offset of the in-browse toggle-box relatively to the containing cell.
    *
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the offset of the in-browse toggle-box relatively to the containing cell, in native
    *         units.
    */
   public NativePoint getRowToggleOffset(BrowseColumnGuiImpl column, boolean editor)
   {
      return getRowToggleOffsetCentered(column, editor);
   }

   /**
    * Get the cell separator color.
    *
    * @param   browse
    *          The browse widget.
    *
    * @return  cell separator color.
    */
   @Override
   public ColorRgb getBrowseCellSeparatorColor(BrowseGuiImpl browse)
   {
      return browse.config().ehSeparatorsColor != null
                ? browse.config().ehSeparatorsColor
                : browse.getSeparatorColor();
      // TODO: the default cell separator color depends on another color I did not identified yet
      //       if the cells have white background the separator was seen both WHITE or 0xC0C0C0
      //       0xC0C0C0 seems to be the default color, when this is the cell background's color
      //       then 0x808080 is used, instead
   }

   /**
    * Get foreground color of the selected cell type.
    *
    * @param   column
    *          The parent {@code BrowseColumnGuiImpl} of the cell.
    * @param   label
    *          {@code true} for label cell, {@code false} for normal cell.
    * @param   selected
    *          {@code true} if the row is selected.
    * @param   viewportRowIndex
    *          0-based number of the target row in the viewport (for cells only). <code>-1</code>
    *          for general cell foreground color.
    *
    * @return  foreground color of the selected cell type.
    */
   @Override
   public ColorRgb getBrowseCellForegroundColor(BrowseColumnGuiImpl column,
                                                boolean label,
                                                boolean selected,
                                                int viewportRowIndex)
   {
      return super.getBrowseCellForegroundColor(column, label, false, viewportRowIndex);
   }

   /**
    * Set the selected cell background color into the {@code gd} {@link GuiDriver}.
    *
    * @param   gd
    *          The {@code GuiDriver} to set the color for separators.
    * @param   unselColor
    *          The color for unselected cells. If {@code null} then the unselected cells' 
    *          background is not drawn. 
    */
   @Override
   protected void setBrowseCellBackgroundSelectionColor(GuiDriver gd, ColorRgb unselColor)
   {
      if (unselColor == null)
      {
         gd.setColor(new Color(0xE5E5E5));
      }
      else
      {
         // the 30% of Color.darker() is too strong, we only need about 18% (1/6)
         gd.setColor(new ColorRgb(
            unselColor.getRed() * 5 / 6,
            unselColor.getGreen() * 5 / 6,
            unselColor.getBlue() * 5 / 6
         ));
      }
   }
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    * 
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseTitlePanel(BrowseGuiImpl browse,
                                    GuiDriver gd,
                                    GuiColorResolver gc,
                                    GuiFontResolver gf,
                                    int x,
                                    int y,
                                    int width,
                                    int height)
   {
      boolean focused = browse.isDrawnAsFocused();
      String title = browse.config().title;
      
      gd.setColor(focused ? new Color(0x9CB6D6)
                          : new Color(0xBDCFDE));
      gd.fillRect(x + 1, y + 1, width - 1, height - 1);
      // add a 3d effect
      gd.setColor(new Color(0xFFFFFF));
      gd.drawLine(x, y, x + width - 1, y);
      gd.drawLine(x, y, x,             y + height - 1);
      
      // calculate text position
      int textWidth = browse.getTextWidth(title, gf.resolveFontNum());
      x += (width - textWidth) / 2;
      
      // draw title string
      gd.setColor(focused ? new Color(0x000000)
                          : new Color(0x000000));
      gd.setGuiFont(gf.font().font);
      gd.drawStringCentered(title,
                            x,
                            browse.getTitleHeight() / 2 - BrowseGuiImpl.BORDER_THICKNESS);
   }
   
   /**
    * Draw browse row marker column resctangle.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   selected
    *          Is the current row selected ?
    * @param   topOffset
    *          Top offset, in pixels.
    * @param   w
    *          The border width in pixels.
    * @param   h
    *          The border height in pixels.
    */
   @Override
   protected void drawBrowseColumnMarker(GuiDriver gd,
                                         boolean selected,
                                         int topOffset,
                                         int w,
                                         int h)
   {
      if (selected)
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.fillRect(1, topOffset, w - 2, h - 1);
         
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(w - 2, topOffset + 2, w - 2, topOffset + h - 2);
         gd.drawLine(1, topOffset + h - 2, w - 3, topOffset + h - 2);
      }
      else
      {
         gd.setColor(gd.getSysColor(COLOR_3DFACE));
         gd.fillRect(1, topOffset, w - 2, h - 1);
         
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(1, topOffset + 1, 1, topOffset + h - 3);
         gd.drawLine(2, topOffset + 1, w - 3, topOffset + 1);
         
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.drawLine(w - 2, topOffset + 2, w - 2, topOffset + h - 2);
         gd.drawLine(1, topOffset + h - 2, w - 2, topOffset + h - 2);
      }
   }
   
   /**
    * Draw label-style rectangle.
    *
    * @param   column
    *          The {@code BrowseColumnGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   xOffset
    *          Left offset, in pixels.
    * @param   yOffset
    *          Top offset, in pixels.
    * @param   w
    *          The border width in pixels.
    * @param   h
    *          The border height in pixels.
    */
   @Override
   protected void drawBrowseColumnLabelBorder(BrowseColumnGuiImpl column,
                                              GuiDriver gd,
                                              int xOffset,
                                              int yOffset,
                                              int w,
                                              int h)
   {
      boolean pressed = column.isDrawnAsPressed() || column.isLabelPressed();
      xOffset = 0;
      if (column.isFirstColumn())
      {
         ++xOffset;
         --w;
      }
      
      // and draw the border itself depending on pressed flag
      if (pressed)
      {
         gd.setColor(new Color(0x26A0DA));
         gd.drawRect(xOffset, yOffset, w, h - 1);
      }
      else
      {
         gd.setColor(new Color(0xE1E1E1));
         gd.drawRect(xOffset, yOffset, w, h - 1);
      }
   }
   
   /**
    * Fills the background of the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   active
    *          {@code true} if the drawn windows is active.
    * @param   cleanup
    *          if {@code true} do not draw the border since this is a text cleanup call.
    */
   private void fillTitlebar(GuiDriver gd,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean active,
                             boolean cleanup)
   {
      int color = active ? SysColor.COLOR_ACTIVECAPTION : SysColor.COLOR_INACTIVECAPTION;
      
      gd.setColor(gd.getSysColor(color));
      gd.fillRect(x, y, width, height);
      
      if (!cleanup)
      {
         gd.setColor(active ? new Color(0x293031)
                            : new Color(0x5a6163));
         gd.drawLine(x, y, x + width - 1, y);
         
         if (active)
         {
            gd.setColor(new Color(0x7BA6CE));
            gd.drawLine(x,         y + height - 1,
                        x + width, y + height - 1);
         }
      }
   }
   
   /**
    * Draws a File or Folder Thumbnail of from a file/dir chooser dialog.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    * @param   selected
    *          {@code true} if the thumbnail is currently selected.
    * @param   hover
    *          {@code true} if the mouse is currently hovering the thumbnail.
    * @param   focus
    *          {@code true} if the thumbnail has focus.
    */
   @Override
   public void drawFileChooserThumbnailBackground(GuiDriver gd,
                                                  int x,
                                                  int y,
                                                  int width,
                                                  int height,
                                                  boolean selected,
                                                  boolean hover,
                                                  boolean focus)
   {
      if (selected)
      {
         gd.setColor(new ColorRgb(new Color(0xCEE3FF)));
         gd.fillRect(x + 1, y + 1, width - 1, height - 1);
         gd.setColor(new ColorRgb(new Color(0x7BA2CE)));
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else if (hover)
      {
         gd.setColor(new ColorRgb(new Color(0xF4F9FF)));
         gd.fillRect(x + 1, y + 1, width - 1, height - 1);
         gd.setColor(new ColorRgb(new Color(0xBDD7FF)));
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else
      {
         // leave it transparent
         // gd.setColor(new ColorRgb(new Color(0xFFFFFF)));
         // gd.fillRect(x, y, width, height);
      }
   }
   
   /**
    * Obtain color information needed to render the foreground of a thumbnail. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   focused
    *          {@code true} if thumbnail is focused.
    * @param   selected
    *          {@code true} if thumbnail is selected.
    * @param   hover
    *          {@code true} if mouse hovers the thumbnail.
    * @param   filename
    *          {@code true} to obtain the color of the filename.
    * @param   infoLabel
    *          {@code true} to obtain the info label (for details view).
    * @param   infodata
    *          {@code true} to obtain the info value (for details view).
    *
    * @return  The foreground color that matches the requested state.
    */
   @Override
   public ColorRgb getFileChooserForegroundColor(GuiDriver gd,
                                                 boolean focused,
                                                 boolean selected,
                                                 boolean hover,
                                                 boolean filename,
                                                 boolean infoLabel,
                                                 boolean infodata)
   {
      return filename || infodata 
            ? new ColorRgb(new Color(0x000000))
            : new ColorRgb(new Color(0x6B6B6B));
   }
}
