/*
** Module   : AbstractDynamicLayout.java
** Abstract : Abstract implementation of the core dynamic layout algorithm.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20180813 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.google.common.collect.*;

import java.util.*;
import java.util.function.*;

/**
 * Abstract implementation of the core dynamic layout algorithm. The class provides base logic for all the
 * dynamic layout variations.
 */
public abstract class AbstractDynamicLayout<O extends OutputManager<?>>
{
   /** Dynamic layout configuration */
   protected DynamicLayoutConfig config;

   /** The container this layout instance operates on */
   protected Container<O> container;

   /** Horizontal cached data */
   private TreeMultimap<Double, Widget> xWidgets;

   /** Vertical cached data */
   private TreeMultimap<Double, Widget> yWidgets;

   /** Cached widget initial sizes */
   private HashMap<Widget, Dimension> initialDims = new HashMap<>();

   /**
    * Constructor.
    *
    * @param   container
    *          The container this instance will layout.
    */
   public AbstractDynamicLayout(Container<O> container)
   {
      this.container = container;
      config = DynamicLayoutConfig.getInstance();
   }

   /**
    * Performs the actual layout.
    */
   public abstract void doLayout();

   /**
    * The method returns the container's widgets grouped into rows.
    *
    * @return  widgets grouped in rows
    */
   protected List<RowInfo> parseRows()
   {
      return parseRows(false);
   }

   /**
    * The method returns the container's widgets grouped into rows.
    *
    * @param   useOptimalSize
    *          The parameter instructs the parse logic to either use the the widgets' actual size or
    *          optimal size on those widgets that do support optimal size calculation.
    *
    * @return  widgets grouped in rows
    */
   protected List<RowInfo> parseRows(boolean useOptimalSize)
   {
      buildIndexes();
      return parseRowsImpl(yWidgets,
                           w -> w.location().x,
                           d -> d.width,
                           d -> d.height,
                           useOptimalSize);
   }

   /**
    * The method returns the container's widgets grouped into columns.
    *
    * @return  widgets grouped in columns
    */
   protected List<RowInfo> parseColumns()
   {
      buildIndexes();
      return parseRowsImpl(xWidgets,
                           w -> w.location().y,
                           d -> d.height,
                           d -> d.width,
                           false);
   }

   /**
    * The method determines whether the supplied widget is considered "dynamic" for the purpose of the
    * layout algorithm.
    *
    * @param   w
    *          The widget being checked.
    *
    * @return  <code>true</code> if the widget is considered "dynamic", <code>false</code> otherwise.
    */
   protected boolean isDynamicWidget(Widget w)
   {
      if (w instanceof BrowseGuiImpl)
      {
         return true;
      }

      if (w instanceof Container)
      {
         for (Object widgetObj : ((Container) w).widgetsAsList())
         {
            Widget widget = (Widget) widgetObj;

            if (isDynamicWidget(widget))
            {
               return true;
            }
         }
      }

      return false;
   }

   /**
    * Builds intermediate data for the layout algorithm.
    */
   private void buildIndexes()
   {
      xWidgets = TreeMultimap.create(Ordering.natural(), Ordering.arbitrary());
      yWidgets = TreeMultimap.create(Ordering.natural(), Ordering.arbitrary());

      double minX =  Double.MAX_VALUE;
      double minY =  Double.MAX_VALUE;
      double maxX = -Double.MAX_VALUE;
      double maxY = -Double.MAX_VALUE;

      for (Object wObj : container.widgetsAsList())
      {
         Widget w = (Widget) wObj;

         Point wloc = w.location();
         Dimension wdim = w.dimension();

         if (wloc.x < minX)
         {
            minX = wloc.x;
         }

         if (wloc.y < minY)
         {
            minY = wloc.y;
         }

         double wmaxX = wloc.x + wdim.width;
         if (wmaxX > maxX)
         {
            maxX = wmaxX;
         }

         double wmaxY = wloc.y + wdim.height;
         if (wmaxY > maxY)
         {
            maxY = wmaxY;
         }

         xWidgets.put(wloc.x, w);
         yWidgets.put(wloc.y, w);
      }
   }

   /**
    * Sets size of the supplied widget.
    *
    * @param   w
    *          The subjected widget.
    * @param   width
    *          The new width.
    * @param   height
    *          The new height.
    */
   protected void setWidgetSize(Widget w, double width, double height)
   {
      if (w instanceof Frame)
      {
         ((Frame) w).setOuterSize(width, height);
         ((Frame) w).updateInnerSize();
      }
      else if (w instanceof AbstractContainer)
      {
         ((AbstractContainer) w).setSize(width, height);
      }
      else
      {
         throw new RuntimeException("Unsupported widget type");
      }
   }

   /**
    * Groups widgets into rows or columns based on the input parameters.
    *
    * @param   yWidgets
    *          The widgets sorted based on the vertical locations.
    * @param   xLocationF
    *          Function for getting widget's horizontal position.
    * @param   widthF
    *          Function for getting widget's horizontal size.
    * @param   heightF
    *          Function for getting widget's vertical size.
    * @param   useOptimalSize
    *          The parameter instructs the parse logic to either use the the widgets' actual size or
    *          optimal size on those widgets that do support optimal size calculation.
    *
    * @return  container's widgets grouped into rows
    */
   private List<RowInfo> parseRowsImpl(TreeMultimap<Double, Widget> yWidgets,
                                       Function<Widget, Double> xLocationF,
                                       Function<Dimension, Double> widthF,
                                       Function<Dimension, Double> heightF,
                                       boolean useOptimalSize)
   {
      List<RowInfo> rows = new ArrayList<>();
      RowInfo emptyRow = null;

      for (Map.Entry<Double, Collection<Widget>> we : yWidgets.asMap().entrySet())
      {
         double y = we.getKey();

         for (Widget w : we.getValue())
         {
            if (emptyRow == null || y >= emptyRow.bottom)
            {
               // this is new row
               emptyRow = new RowInfo();
               emptyRow.top = y;
               emptyRow.bottom = -Double.MAX_VALUE;
               rows.add(emptyRow);
            }

            double bottom = Coordinate.scale(y + heightF.apply(w.dimension()));

            if (emptyRow.bottom < bottom)
            {
               emptyRow.bottom = bottom;
            }

            emptyRow.xWidgets.put(xLocationF.apply(w), w);
         }
      }

      HashMap<Widget, Dimension> prevInitialDims = initialDims;
      initialDims = new HashMap<>();

      for (RowInfo row : rows)
      {
         CellInfo cell = null;
         CellInfo prevCell = null;

         for (Map.Entry<Double, Collection<Widget>> we : row.xWidgets.asMap().entrySet())
         {
            double x = we.getKey();

            for (Widget w : we.getValue())
            {
               if (cell == null || x >= cell.right)
               {
                  if (cell != null)
                  {
                     row.totalInitialWidth += cell.initialWidth;

                     if (prevCell != null && !cell.dynamic && !prevCell.dynamic)
                     {
                        mergeCells(cell, prevCell);
                        row.cells.remove(cell);
                        cell = prevCell;
                     }
                  }

                  prevCell = cell;

                  // this is new cell
                  cell = new CellInfo();
                  cell.left = x;
                  cell.right = -Double.MAX_VALUE;
                  row.cells.add(cell);
               }

               boolean doOptimalSize = useOptimalSize && w instanceof HasOptimalSize;

               Dimension wDim = w.dimension();
               double wWidth = doOptimalSize ? ((HasOptimalSize) w).getOptimalWidth() : widthF.apply(wDim);
               double right = Coordinate.scale(x + wWidth);

               if (cell.right < right)
               {
                  cell.right = right;
               }

               cell.widgets.add(w);
               cell.dynamic = cell.dynamic || isDynamicWidget(w);

               if (cell.dynamic)
               {
                  Dimension initialDim = prevInitialDims.get(w);
                  if (initialDim == null)
                  {
                     initialDim = wDim;
                  }
                  initialDims.put(w, initialDim);

                  double initialDimWidth = widthF.apply(initialDim);
                  if (initialDimWidth > cell.initialWidth)
                  {
                     cell.initialWidth = initialDimWidth;
                  }
               }
            }
         }

         if (cell != null)
         {
            row.totalInitialWidth += cell.initialWidth;

            if (prevCell != null && !cell.dynamic && !prevCell.dynamic)
            {
               mergeCells(cell, prevCell);
               row.cells.remove(cell);
            }
         }

         // calculate total static width after all cells are merged
         for (CellInfo cellInfo : row.cells)
         {
            if (!cellInfo.dynamic)
            {
               row.totalStaticWidth += cellInfo.right - cellInfo.left;
            }

         }
      }

      return rows;
   }

   /**
    * Merges two adjacent cells.
    *
    * @param   source
    *          The cell that will be merged to target.
    * @param   target
    *          The cell where the source cell will be merged.
    */
   private void mergeCells(CellInfo source, CellInfo target)
   {
      target.widgets.addAll(source.widgets);
      target.right = source.right;
   }

   /**
    * Data class for widgets grouped into rows.
    */
   protected static class RowInfo
   {
      /* row top location */
      public double top;

      /* row bottom location */
      public double bottom;

      /* Widgets sorted according to the x-axis */
      public TreeMultimap<Double, Widget> xWidgets = TreeMultimap.create(Ordering.natural(),
                                                                         Ordering.arbitrary());

      /* individual cells with */
      public List<CellInfo> cells = new ArrayList<>();

      /* the total initial width of all dynamic cells */
      public double totalInitialWidth;

      /* sum of all static widths - widgets that can't be resized and paddings */
      public double totalStaticWidth;

      /**
       * Returns <code>true</code> is the row contains at least one dynamic widget.
       *
       * @return  see above
       */
      public boolean isDynamic()
      {
         return totalInitialWidth > 0;
      }
   }

   /**
    * Data class for widgets grouped into columns in the individual rows.
    */
   protected static class CellInfo
   {
      /* cell left */
      public double left;

      /* cell right */
      public double right;

      /** Is set to <code>true</code> when the cell contains at least one dynamic widget */
      public boolean dynamic;

      /** The max initial width of all the initial widths of the cell widgets */
      public double initialWidth;

      /* widgets in the cell */
      public Set<Widget> widgets = new HashSet<>();
   }
}
