/*
** Module   : DynamicLayoutIntegration.java
** Abstract : Widget tree dynamic layout enabler.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20180813 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The dynamic layout algorithm is not directly attached to the widget tree. Instead there is another
 * indirection layer that provides the integration. This allows to clearly formalize the integration
 * points between the widget tree and the layout algorithm.
 */
public class DynamicLayoutIntegration<O extends OutputManager<?>>
implements ContainerStateListener<O>
{
   /** The linked container */
   protected Container<O> container;

   /** The layout implementation */
   private DynamicLayout<O> layout;

   /** A flag indicating whether dynamic layout has been requested */
   private boolean pendingDynamicLayout;

   /**
    * Constructor.
    *
    * @param   container
    *          The container that will be laid out.
    */
   public DynamicLayoutIntegration(Container<O> container)
   {
      this.container = container;
      layout = new DynamicLayout<>(container);
   }

   /**
    * Called by the frame widget when it has been realized.
    */
   public void onFrameRealized()
   {
      requestDynamicLayout();
   }

   /**
    * Invoked by widget container when new widget is added to it.
    *
    * @param   widget
    *          The added widget.
    */
   @Override
   public void widgetAdded(Widget<O> widget)
   {
      // no-op
   }

   /**
    * Invoked by widget container when a widget is removed from it.
    *
    * @param   widget
    *          The removed widget.
    */
   @Override
   public void widgetRemoved(Widget<O> widget)
   {
      // no-op
   }

   /**
    * Invoked by widget container when DISABLE-REDRAW is assigned FALSE on a widget in the container.
    *
    * @param   widget
    *          The subjected widget.
    */
   @Override
   public void redrawEnabled(Widget<O> widget)
   {
      DynamicLayoutIntegration<O> layout = getRootLayout();
      if (layout.pendingDynamicLayout)
      {
         layout.triggerDynamicLayout();
      }
   }

   /**
    * The method is called to request dynamic layout. The dynamic layout logic is not executed directly,
    * instead all the multiple requests are gathered during a UI transformation so that the actual layout
    * is not executed excessively.
    */
   protected void requestDynamicLayout()
   {
      ThinClient tc = ThinClient.getInstance();

      if (tc.isDisableRedraw() || container.isRepaintDisabled())
      {
         getRootLayout().pendingDynamicLayout = true;
         return;
      }

      if (pendingDynamicLayout)
      {
         return;
      }

      pendingDynamicLayout = true;

      // trigger the layout on the event queue to synchronize it with the other
      // UI events
      tc.invokeLater(this::triggerDynamicLayout);
   }

   /**
    * Performs the actual layout with the help of the contained layout implementations, recursively down the
    * widget tree.
    */
   protected void doDynamicLayout()
   {
      layout.doLayout();

      for (Widget<O> w : container.widgetsAsList())
      {
         if (!(w instanceof HasDynamicLayout) || !w.isVisible(true))
         {
            continue;
         }

         DynamicLayoutIntegration childLayout = ((HasDynamicLayout) w).getDynamicLayout();
         if (childLayout != null)
         {
            childLayout.doDynamicLayout();
         }
      }

      pendingDynamicLayout = false;
   }

   /**
    * Triggers dynamic layout when the linked container is the top container that was requested dynamic
    * layout.
    */
   private void triggerDynamicLayout()
   {
      DynamicLayoutIntegration layout = getParentLayout(this);
      while (layout != null)
      {
         if (layout.pendingDynamicLayout)
         {
            // this container will be laid out by the parent
            return;
         }

         layout = getParentLayout(layout);
      }

      doDynamicLayout();
   }

   /**
    * Returns the parent dynamic layout integration.
    *
    * @param   relativeTo
    *          Where the search should start.
    *
    * @return  the dynamic layout integration reference or <code>null</code>
    */
   private DynamicLayoutIntegration getParentLayout(DynamicLayoutIntegration relativeTo)
   {
      Container<O> parent = relativeTo.container.parent();
      while (parent != null)
      {
         if (parent instanceof HasDynamicLayout)
         {
            DynamicLayoutIntegration layout = ((HasDynamicLayout) parent).getDynamicLayout();
            if (layout != null && layout != relativeTo)
            {
               return layout;
            }
         }

         parent = parent.parent();
      }

      return null;
   }

   /**
    * Returns the root dynamic layout integration.
    */
   private DynamicLayoutIntegration getRootLayout()
   {
      DynamicLayoutIntegration rootLayout = this;
      while (rootLayout != null)
      {
         DynamicLayoutIntegration temp = getParentLayout(rootLayout);
         if (temp == null)
         {
            break;
         }

         rootLayout = temp;
      }

      return rootLayout;
   }
}
