/*
** Module   : Container.java
** Abstract : Common interface for widget container classes.
**
** Copyright (c) 2010-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100000 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100831 Changes related to refactoring of FrameHelper.
** 004 VIG 20140515 Adding moveBeforeTabItem method.
** 005 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 HC  20150221 Improvements to frame tab order.
** 009 VIG 20150330 Added childrenLocation method.
** 010 IAS 2050926  MOVE-TO-TOP()/MOVE-TO-BOTTOM() support
** 011 CA  20151027 Added support for MOVE-AFTER-TAB-ITEM.
** 012 CA  20151123 Added APIs to place a widget on a specified location.
** 013 HC  20160922 Added currentFocus(boolean) method.
** 014 SBI 20161101 Added getTabItemList(), renamed getWidgetsList to widgetsAsList().
** 015 HC  20180813 Implemented initial version of dynamic layout.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.layout.*;

/**
 * Common interface implemented by all widget containers.
 */
public interface Container<O extends OutputManager<?>>
extends Widget<O>
{
   /**
    * Add widget to container.
    * 
    * @param   widget
    *          Widget to add.
    *          
    * @return  <code>this</code> for fluent syntax support.
    */
   public Container<O> add(Widget<O> widget);

   /**
    * Add the new widget before the given widget's position.
    * 
    * @param    before
    *           The widget to insert before.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   public Container<O> add(Widget<O> before, Widget<O> widget);

   /**
    * Add the widget on the specified position.
    * 
    * @param    pos
    *           The widget's position.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   public Container<O> add(int pos, Widget<O> widget);
   
   /**
    * This method should be overridden in a container that allows clipping.
    * 
    * @return  <code>true</code> if container clips its children. 
    */
   public boolean clipping();

   /**
    * Get widget which is in focus.
    * 
    * @return  focused widget.
    */
   public Widget<O> currentFocus();

   /**
    * Get widget which is in focus.
    * <p>
    * The method allows to choose whether the focused widget is searched in the nested
    * containers (<code>recurse == true</code>) or whether only this container is searched
    * (<code>recurse == false</code>).
    *
    * @param   recurse
    *          The parameter controls whether the method should recurse to the nested
    *          containers when searching for the focused widget.
    *
    * @return  focused widget.
    */
   public Widget<O> currentFocus(boolean recurse);

   /**
    * Get default focusable widget.
    * 
    * @return  default focusable widget.
    */
   public Widget<O> defaultFocus();

   /**
    * Lay out container.
    */
   public void doLayout();

   /**
    * Indicates the ending of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    * 
    * @param    repaint
    *           If <code>true</code> then the current bounding rectangle and
    *           the rectangle saved during <code>startSizeChange</code> call
    *           are repainted.
    */
   public void endSizeChange(boolean repaint);

   /**
    * Returns the focused direct child Component.
    * 
    * @return   The focused direct child Component.
    */
   public Widget<O> focus();

   /**
    * Get widget for provided index or <code>null</code> if no such component
    * exists.
    * 
    * @param   index
    *          Widget index.
    *           
    * @return  widget at specified index or <code>null</code> if index is out
    *          of range.
    */
   public Widget<O> get(int index);

   /**
    * Get clipping rectangle.
    * 
    * @return   clipping rectangle.
    */
   public Rectangle getClipRect();

   /**
    * Reports if there is at least 1 focusable widget that is not the given
    * widget.
    * 
    * @param    exclude
    *           A widget to ignore or <code>null</code> if any focusable
    *           widget will do.
    * 
    * @return   <code>true</code> if there is at least one other widget that
    *           can be focused.
    */
   public boolean hasFocusable(Widget<O> exclude);

   /**
    * Get container insets.
    *  
    * @return  container insets.
    */
   public Insets insets();
   
   /**
    * Mode specified widget before the widget "before".
    * 
    * @param   widget
    *          Widget to move.
    * @param   before
    *          The widget before to insert mover.
    */
   public void moveBeforeTabItem(Widget<O> widget, Widget<O> before);
   
   /**
    * Mode specified after before the widget "after".
    * 
    * @param   widget
    *          Widget to move.
    * @param   after
    *          The widget before to insert mover.
    */
   public void moveAfterTabItem(Widget<O> widget, Widget<O> after);

   /**
    * Mode specified widget to the bottom of the widget list.
    * 
    * @param   widget
    *          Widget to mover.
    */
   public void moveToBottom(Widget<O> widget);

   /**
    * Move specified widget to the top of the widget list.
    * 
    * @param   widget
    *          Widget to mover.
    */
   public void moveToTop(Widget<O> widget);

   /**
    * Mode specified widget to the bottom of the widget list considering Z-order class.
    * 
    * @param   widget
    *          Widget to mover.
    * @return <code>true</code> if the widget was moved
    */
   public boolean moveToBottomInClass(Widget<O> widget);

   /**
    * Move specified widget to the top of the widget list considering Z-order class.
    * 
    * @param   widget
    *          Widget to mover.
    * @return <code>true</code> if the widget was moved
    */
   public boolean moveToTopInClass(Widget<O> widget);

   /**
    * Move widget to specified position in the widget list. For overlaying 
    * widgets this is basically position in widget Z-order.
    *  
    * @param   widget
    *          Widget to mover.
    * @param   pos
    *          Position in Z-order.
    */
   public void moveToZpos(Widget<O> widget, int pos);

   /**
    * Select next focusable widget. 
    */
   public void nextFocus();

   /**
    * Select previous focusable widget.
    */
   public void prevFocus();
   
   /**
    * Remove specified widget from the widget list.
    * 
    * @param   widget
    *          Widget to remove.
    */
   public void remove(Widget<O> widget);

   /**
    * Replace specified widget with another with preserving widget Z-order.
    * 
    * @param   oldWidget
    *          Old widget.
    * @param   newWidget
    *          New widget.
    */
   public void replace(Widget<O> oldWidget, Widget<O> newWidget);

   /**
    * Set focus to specified widget.
    * 
    * @param   focus
    *          Widget to set focus on.
    */
   public void setFocus(Widget<O> focus);

   /**
    * Set container height.
    * 
    * @param   height
    *          New container height.
    */
   public void setHeight(double height);

   /**
    * Set container insets.
    * 
    * @param   insets
    *          Insets to set.
    */
   public void setInsets(Insets insets);

   /**
    * Set layout manager for the container.
    * 
    * @param   manager
    *          New layout manager.
    */
   public void setLayout(LayoutManager<O> manager);

   /**
    * Get active layout manager for the container.
    * 
    * @return  Active layout manager.
    */
   LayoutManager<O> getLayout();
   
   /**
    * Set container size.
    * 
    * @param   size
    *          Size to set.
    */
   public void setSize(Dimension size);

   /**
    * Set container size.
    * 
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   public void setSize(double width, double height);

   /**
    * Set container width.
    * 
    * @param   width
    *          Width to set.
    */
   public void setWidth(double width);

   /**
    * Indicates the beginning of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    * 
    * @param    saveOldRect
    *           If <code>true</code> then the current bounding rectangle is
    *           saved for future repainting.
    */
   public void startSizeChange(boolean saveOldRect);

   /**
    * Get all widgets stored in the container.
    * 
    * @return  array of stored widgets.
    */
   public Widget<O>[] widgets();

   /**
    * Returns the tab item list for this container.
    * 
    * @return   The tab item list.
    */
   public List<Widget<O>> getTabItemList();
   
   /**
    * Get all widgets stored in the container.
    *
    * @return  the new list with stored widgets.
    */
   public List<Widget<O>> widgetsAsList();
   
   /**
    * Set focus to first focusable widget.
    */
   public void firstFocus();

   /**
    * Set focus to last focusable widget.
    */
   void lastFocus();

   /**
    * Detach this widget from this container.  The widget will not be destroyed.
    * 
    * @param   widget
    *          Widget to detach.
    */
   public void detach(Widget<O> widget);
   
   /** 
    * Returns the set of widgets that are never selected for focus.
    * 
    * @return  set of widgets or <code>null</code>. 
    */
   public Set<Widget<?>> getIgnoreFocus();
   
   /**
    * Sets the set of widgets that are never selected for focus.
    * <code>null</code> may be used to clear the set.
    * 
    * @param   widgets
    *          Set of widgets or <code>null</code>.
    */
   public void setIgnoreFocus(Collection<Widget<?>> widgets);
   
   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   public Point childrenLocation(Widget<O> child);

   /**
    * Adds container state listener.
    *
    * @param   listener
    *          A container state listener.
    */
   void addContainerStateListener(ContainerStateListener<O> listener);

   /**
    * Removes container state listener.
    *
    * @param   listener
    *          A container state listener.
    */
   void removeContainerStateListener(ContainerStateListener<O> listener);
}