/*
** Module   : PropertyTable.java
** Abstract : The tree Widget Browser component.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141130 Added the capability to display widget config fields.
** 003 HC  20150216 Widget properties can be copied to clipboard.
** 004 HC  20160210 Minor layout improvements.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.*;

import javax.swing.*;
import javax.swing.table.*;

import org.apache.commons.lang.StringEscapeUtils;

/**
 * The tree Widget Browser component.
 */
@SuppressWarnings("serial")
public class PropertyTable 
extends JTable 
{
   /**
    * Column names.
    */
   private String[] tableColumns = { "Property", "Value" };

   /**
    * Table model.
    */
   private DefaultTableModel tableModel;

   /**
    * Event listener.
    */
   private ActionListener actionListener;

   /**
    * Constructor.
    */
   public PropertyTable() 
   {
      // set the tooltip delay
      ToolTipManager.sharedInstance().setDismissDelay(120000);
      
      // create table model
      tableModel = new DefaultTableModel();
      tableModel.setColumnIdentifiers(tableColumns);
      setModel(tableModel);
      
      getColumnModel().getColumn(0).setPreferredWidth(20);

      addMouseListener(new MouseAdapter() 
      {
         public void mouseClicked(MouseEvent e) 
         {
            // check double click
            if (e.getClickCount() != 2) 
            {
               return;
            }

            // if clicked row is widget property, redirect to the widget node
            if (actionListener != null)
            {
               int row = PropertyTable.this.rowAtPoint(e.getPoint());
               if (row >= 0)
               {
                  actionListener.actionPerformed(
                        new Event(PropertyTable.this, Event.PROPERTY_VALUE_NAVIGATED, 
                                  PropertyTable.this.dataModel.getValueAt(row, 1)));
               }
            }
         }
      });
   }

   /**
    * The override disables cell editing.
    */
   @Override
   public boolean isCellEditable(int row, int column) 
   {
      return false;
   };
   
   /**
    * The override adds tooltips to the value cells.
    * 
    * @param   renderer  
    *          the <code>TableCellRenderer</code> to prepare
    * @param   row       
    *          the row of the cell to render, where 0 is the first row
    * @param   column    
    *          the column of the cell to render,
    *          where 0 is the first column
    *        
    * @return  the <code>Component</code> under the event location
    */
   @Override
   public Component prepareRenderer(TableCellRenderer renderer, int row, int column)
   {
      Component c = super.prepareRenderer(renderer, row, column);
      
      // tooltips only for property values
      if (column != 1)
      {
         return c;
      }
      
      if (c instanceof JComponent)
      {
         JComponent jc = (JComponent) c;
         
         String text = getValueAt(row, column).toString();
         
         // make the value multi-line
         text = StringEscapeUtils.escapeHtml(text);
         String[] parts = text.split("\\s");
         StringBuffer buf = new StringBuffer();
         buf.append("<html>");
         int lineLength = 0;
         for (String part : parts)
         {
            if (lineLength + part.length() > 80 && lineLength > 0)
            {
               lineLength = 0;
               buf.append("<br>");
            }
            
            buf.append(part);
            buf.append(" ");
            lineLength += part.length();
         }
         buf.append("</html>");     
         jc.setToolTipText(buf.toString());
      }
      return c;
   }

   /**
    * Sets the event handler.
    * 
    * @param   listener
    *          Reference to {@link ActionListener} or <code>null</code>.
    */
   public void setActionListener(ActionListener listener) 
   {
      actionListener = listener;
   }

   /**
    * Clears all property rows.
    */
   public void resetProperties() 
   {
      tableModel.setRowCount(0);
   }
   
   /**
    * Assign the properties to be shown in the table.
    * 
    * @param   properties
    *          The sorted property map.
    */
   public void setTableContent(SortedMap<String, Object> properties) 
   {
      // clear old content
      tableModel.setRowCount(0);
      List<TableRow> rows = new ArrayList<>();

      for (Map.Entry<String, Object> entry : properties.entrySet()) 
      {
         // substitute null with "<null>"
         Object value = entry.getValue();
         if (value == null)
         {
            value = "<null>";
         }
         
         // add object, toString() is called automatically
         rows.add(new TableRow(entry.getKey(), value));
      }

      // add rows to the table
      for (TableRow tableRow : rows) 
      {
         Object[] row = {tableRow.getName(), tableRow.getValue()};
         tableModel.addRow(row);
      }
   }
   
   /**
    * The instance representing one row in the table.
    */
   private static class TableRow 
   implements Comparable<TableRow> 
   {
      /**
       * The name of the property. 
       */
      private String name;

      /**
       * The value of the property.
       */
      private Object value;

      /**
       * The constructor.
       * 
       * @param   name
       *          Property name.
       * @param   value
       *          Property value.
       */
      public TableRow(String name, Object value) 
      {
         super();
         this.name = name;
         this.value = value;
      }

      /**
       * Name getter.
       * 
       * @return  Property name.
       */
      public String getName() 
      {
         return name;
      }

      /**
       * Value getter.
       * 
       * @return  Property value.
       */
      public Object getValue() 
      {
         return value;
      }

      /**
       * Comparator to compare by the property names.
       */
      @Override
      public int compareTo(TableRow o) 
      {
         return getName().compareTo(o.getName());
      }
      
   }
}