/*
** Module   : WidgetBrowserAspect.java
** Abstract : The integration of Widget Browser into the widget drawing infrastructure.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141118 Fixed widget highlight - the coordinates are now relative to the screen.
** 003 HC  20141130 Added support for other window types, like alert boxes.
** 004 GES 20150508 Matched driver API change.
** 005 HC  20150624 Improved Widget Browser GUI highlight. Restoring the highlighted area was
**                  causing artifacts. Also the highlighted area was frozen for draw updates.
** 006 HC  20151013 Added support of modal windows.
** 007 HC  20160203 Do not create widget-browser window for overlay windows.
** 008 HC  20160406 Overhaul of window-activation logic.
** 009 HC  20160510 Fixed Widget Browser to be activated for ChUI P2J windows.
** 010 OM  20170330 Dropped batch bracketing on no-ops in highlightWidget().
** 011 HC  20180611 Do not show Widget browser for alert boxes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.event.*;
import java.util.*;

import org.aspectj.lang.annotation.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.GuiDriver;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This aspect is responsible for lifetime management of the Widget Browser component
 * and the integration with the widget drawing facility.
 * <p>
 * To activate the Widget Browser feature the Java property "widgetbrowser" must be defined,
 * for example using the -D option for the Java application launcher. 
 * The class displays the Widget Browser main window for each drawn {@link Window}.  
 */
@Aspect
public class WidgetBrowserAspect
{
   /**
    * Map of active Widget Browsers.
    */
   private Map<TopLevelWindow<?>, WidgetBrowser> activeBrowsers = new HashMap<>();
   
   /**
    * This method is called on every {@link Widget#draw()} method and serves as the main
    * integration entry-point into the widget drawing facility.
    * <p>
    * For production-quality code, this class can be omitted from AdpectJ compilation to
    * remove the Widget Browser overhead.
    * <p>
    * Note that the advice is intentionally defined with "call" join point. The reason is
    * for this is so that the function is called only once for the whole {@link Widget#draw()} 
    * hierarchy and not for every override. 
    *   
    * @param   widget
    *          The target on which the {@link Widget#draw()} method is called.
    */
   @After("call(void com.goldencode.p2j.ui.client.widget.Widget+.draw()) && target(widget)")
   public void onWidgetDrawn(@SuppressWarnings("rawtypes") Widget widget)
   {
      // is widget browser active?
      if (System.getProperty("widgetbrowser") == null && 
          System.getProperty("widget.browser") == null)
      {
         return;
      }
      
      // start widget browser only for main gui windows 
      if (widget instanceof Window ||
          (!(widget instanceof AlertBox) && widget instanceof GuiWindow &&
           !((GuiWindow) widget).isShareActivationWithOwner()))
      {
         handleStartWidgetBrowser((TopLevelWindow<?>) widget);
      }
      
      // the widget may need a highlight
      handleHighlightWidget(widget);
   }
   
   /**
    * The method creates and shows new Widget Browser main window.
    * If one has been created for the supplied window, this method is a no-op.
    * 
    * @param   window
    *          The {@link TopLevelWindow} instance for which new Widget Browser is being started.
    */
   private synchronized void handleStartWidgetBrowser(TopLevelWindow<?> window)
   {
      if (activeBrowsers.containsKey(window))
      {
         // browser for this window already exists, nothing to do
         return;
      }
      
      WidgetBrowser browserWindow = new WidgetBrowser(window, new ActionListener() 
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            switch (e.getID())
            {
            case Event.WIDGET_SELECTION_CHANGED:
               // be paranoid and redraw the whole window
               TopLevelWindow<?> wnd = null;
               try
               {
                  wnd = ((Event) e).widget.topLevelWindow();
               } 
               catch (RuntimeException ex)
               {
                  // an exception is thrown when widget gets detached from its top-level
                  // window. catch this exception to prevent ill behavior 
                  // when navigating the widget browser UI
               }
               
               if (wnd != null)
               {
                  wnd.clipWith(wnd.bounds());
                  OutputManager.instance().resetInvalidate(wnd);
               }
               break;
            }
         }
      });

      browserWindow.show();
      activeBrowsers.put(window, browserWindow);
   }
   
   /**
    * Highlights the supplied widget when a related active Widget Browser exists, 
    * the widget is visible and it is selected to be highlighted in the Widget Browser.
    * 
    * @param   widget
    *          The widget to be highlighted.
    */
   private synchronized void handleHighlightWidget(Widget<?> widget)
   {
      // search for the ancestor Window,
      // do not use ancestor() since that doesn't return Window reference in 
      // case the widget is placed in an alert box or a drop down for example
      Widget<?> parent = widget.parent();
      while (parent != null && !(parent instanceof TopLevelWindow<?>))
      {
         parent = parent.parent();
      }
      
      TopLevelWindow<?> window = (TopLevelWindow<?>) parent;
      
      WidgetBrowser browser = activeBrowsers.get(window);
      if (browser == null || !browser.isHighlighted(widget))
      {
         return;
      }
      
      highlightWidget(widget);
   }
   
   /**
    * Highlights the supplied widget when the widget is visible.
    * 
    * @param   widget
    *          The widget to be highlighted.
    */
   private void highlightWidget(Widget<?> widget)
   {
      if (widget == null)
      {
         return;
      }
      
      OutputManager<?> om = widget.screen();
      if (om.getInstanceDriver() instanceof GuiDriver)
      {
         GuiDriver<?,?> driver = ((GuiDriver<?, ?>) om.getInstanceDriver());
         
         if (widget.isVisible())
         {
            // draw in a batch so that the widget is redrawn
            driver.startBatch();
            try
            {
               NativePoint loc = widget.screenPhysicalLocation();
               NativeDimension dim = widget.physicalDimension();
               driver.drawHighlight(loc.x, loc.y, dim.width, dim.height, new ColorRgb(255, 0, 0));
            }
            finally
            {
               driver.endBatch();
            }
         }
      }
   }
}