/*
** Module   : AccessorWrapper.java
** Abstract : wraps a BaseDataType object in an Accessor interface
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 GES 20060204   @24275 Created initial version.  This is an API
**                           for an object that can provide access
**                           (get/set) to a BaseDataType.
** 002 CA  20130112          When the wrapped data is a handle, get() will
**                           convert it to its numeric representation and
**                           set() will change it back to a handle, using the
**                           handle.fromString call. This is needed to be able
**                           to display the handle's hashCode to terminal, as
**                           handle instances can't be serialized and sent to
**                           the client.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Defines methods an object that can provide access (get/set) to a
 * {@link BaseDataType}.  This provides a means of delegating the
 * access to an object, such that it can be invoked at an appropriate time
 * by code that does not have direct access or knowledge of the original
 * <code>BaseDataType</code> instance.
 */
public class AccessorWrapper
implements Accessor
{
   /** The wrappered data. */
   private BaseDataType wrappee = null;
   
   /**
    * Construct an instance that references the given wrappee.
    *
    * @param    wrappee
    *           The object to wrapper.  Must NOT be <code>null</code>.
    */
   public AccessorWrapper(BaseDataType wrappee)
   {
      this.wrappee = wrappee;
   }
   
   /**
    * Get a copy of the referenced object's current runtime value (this does
    * not return the original instance itself).
    * <p>
    * In case of handles, an {@link integer} value is returned, which holds
    * the handle's numeric representation.
    *
    * @return   The copy of the current value.
    */
   public BaseDataType get()
   {
      if (wrappee instanceof handle)
      {
         return new integer(wrappee.hashCode());
      }
      return wrappee.duplicate();
   }
   
   /**
    * Set the referenced object's current runtime value to the given value.
    * <p>
    * In case of handles, the <code>val</code> must be its numeric 
    * representation, which is used to create the handle back using the 
    * {@link handle#fromString(String)} call.
    *
    * @param    val
    *           The current value to be assigned into the referenced object.
    */
   public void set(BaseDataType val)
   {
      if (wrappee instanceof handle)
      {
         wrappee = handle.fromString(val.toStringMessage());
      }
      else
      {
         wrappee.assign((Undoable) val);
      }
   }
}