/*
** Module   : AgentPool.java
** Abstract : Definition and implementation of an appserver agent.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130628 Fixed context-local data reset for State-reset operating mode.
** 003 CA  20130701 Although official docs mention that State-reset doesn't call the shutdown or
**                  startup procedure, testing showed startup/shutdown procedures are called for  
**                  this operating mode too.
**                  Reworked appserver launching - implemented by the AppServerLauncher class.
** 004 CA  20130823 Allow each request to be identified via an unique ID. Changes related to 
**                  support async requests.
** 005 OM  20131018 Updated to conform the new SessionListener API.
** 006 SVL 20131203 Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 007 CA  20140218 Always use the full program name when interrogating SourceNameMapper.
** 008 CA  20140220 Argument postProcessing needs to be done before the procedure scope is popped.
**                  The propath needs to be re-initialized after context reset.
** 009 CA  20140223 Startup and Connect (State-aware and State-reset) procedures need to be ran 
**                  persistent.
** 010 CA  20140326 The remote persistent procedures need to be deleted by the Agent which created
**                  them.
** 011 CA  20140509 Fixed passing the parameter to the startup procedure.
** 012 CA  20140407 prepare() needs to force loading of the context-local DatabaseTriggerManager.
** 013 SVL 20140414 Implemented processing for TableParameters, added "appserver-agent" scope.
** 014 SVL 20140709 Added fall-through codes for table parameters errors.
** 015 CA  20150827 Release the memory for all memptr arguments, before returning to the caller.
** 016 CA  20151222 Any ERROR conditions raised during instantiation in obtainProxy are sent back
**                  to the caller (also, the pushed scope must be removed).
** 017 CA  20160404 Enhanced the appserver invocations, to allow implicit or explicit termination 
**                  of requests being executing on a certain connection.
** 018 CA  20160422 When resetting the Agent's context, make a bogus call to the P2J client-side,
**                  to clean up the frames (via the StateSynchronizer).
**                  Fixed deadlock when the Agent's java process is killed while it has control.
** 019 CA  20160626 Let the Agent recover from an UnstoppableExitException.
** 020 GES 20171220 Added TODO comment.
** 021 GES 20180102 Clarified an error handling comment.
** 022 CA  20180503 Added SESSION:SERVER-CONNECTION-ID support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.Array;
import java.util.*;

import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.BlockManager.Condition;

/**
 * An apserver agent definition.
 */
class Agent
{
   /**
    * List of possible invocation modes: 
    * <ul>
    * <li>{@link #IMMEDIATELY} - in this thread (the thread must be an agent thread)</li>
    * <li>{@link #AGENT_WAIT} - in a different agent, wait for response</li>
    * <li>{@link #AGENT_NO_WAIT} in a different agent, do not wait for response</li>
    * </ul>
    */
   enum InvocationModes
   {
      IMMEDIATELY,
      AGENT_WAIT,
      AGENT_NO_WAIT
   }
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerManager.class);

   /** The command queue for this agent. */
   private final Queue<AgentCommand> commands = new LinkedList<AgentCommand>();
   
   /** The agent's pool. */
   private final AgentPool pool;

   /** The agent's stopper worker. */
   private final AgentStopper stopperWorker = new AgentStopper();
   
   /** 
    * Dedicated thread to stop any job currently being executed by this Agent.  
    * The thread is running under the Agent's context.
    */
   private final AssociatedThread stopper = new AssociatedThread(stopperWorker);
   
   /** Flag indicating if the agent has been terminated. */
   private boolean terminated = false;
   
   /** Flag indicating the agent has been bound to the connection. */
   private boolean bound = false;
   
   /** An instance of an external program to which this agent is dedicated. */
   private ExternalProgramWrapper extProg;
   
   /**
    * Build a new agent and associate it with the given pool.
    * 
    * @param    pool
    *           The agent's pool.
    */
   public Agent(AgentPool pool)
   {
      this.pool = pool;
   }

   /**
    * Invoke an external procedure with no parameters. 
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    */
   private static void invokeScoped(String id, character name)
   {
      invokeScoped(id, name, null, false, false, false, false, null, null, null);
   }
   
   /**
    * Invoke an external procedure, internal procedure or user-defined function.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    proc
    *           The external procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    exports
    *           The appserver's exports.
    * @param    popTask
    *           A task to be execued before the global scope is popped.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    The result returned by this procedure or user-defined function call. 
    */
   private static BaseDataType invokeScoped(String    id,
                                            character name,
                                            handle    proc,
                                            boolean   function,
                                            boolean   dynamicFunction,
                                            boolean   superCall,
                                            boolean   transactionDistinct,
                                            character exports,
                                            Runnable  popTask,
                                            String    modes,
                                            Object... args)
   {
      // add the topmost scope to the TransactionManager
      TransactionManager.pushScope("startup",
                                   TransactionManager.NO_TRANSACTION,
                                   true,
                                   true,
                                   false,
                                   false);
      try
      {
         if (proc == null)
         {
            // force invocation of an external procedure
            ControlFlowOps.invokeExternalProcedure(name, false, null, transactionDistinct,
                                                   exports, modes, args);
            return ControlFlowOps.getReturnValue();
         }
         else
         {
            // this is an internal-entry call
            return ControlFlowOps.invoke(proc, name, function,
                                         dynamicFunction, superCall,
                                         transactionDistinct, modes, args);
         }
      }
      catch (StopConditionException e)
      {
         // check if this can be converted to error
         if (exports != null  && invokeFailure(e))
         {
            return new unknown();
         }
         
         throw e;
      }
      finally
      {
         // before popping the scope, process any custom task
         if (popTask != null)
         {
            popTask.run();
         }

         // remove the topmost scope from the TransactionManager
         TransactionManager.popScope();
      }
   }

   /**
    * Check if the given STOP condition is from an invoke failure.
    * 
    * @param    e
    *           The STOP condition.
    * 
    * @return   <code>true</code> if the cause if this STOP is a {@link NumberedException} known
    *           to be produced by invoke failures.
    */
   private static boolean invokeFailure(StopConditionException e)
   {
      Throwable cause = e.getCause();
      if (cause != null && cause instanceof NumberedException)
      {
         // this specific case treats the error raised by ControlFlowOps.invokeFailure, which will
         // always set the topmost cause as a NumberedException
         NumberedException ne = (NumberedException) cause;
         if (ne.getNumber() == 293)
         {
            // throws a ControlFlowOps.invokeFailure incoming error to allow the current command
            // to abort (which is a RUN of an external program) and will just 'eat' the exception
            // and copy it to the RemoteInvocationResult instance; it should be safe to directly 
            // instantiate the error because we are outside of 4GL normal behavior; anyway, the
            // idea is also 'this needs to be thrown'
            
            // raise ERROR condition if the external procedure was not found
            throw new ErrorConditionException(ne);
         }
      }
      
      return false;
   }
   
   /**
    * Set the external procedure associated with this agent.
    * 
    * @param    extProg
    *           The external procedure. May be <code>null</code>.
    */
   public synchronized void setProcedure(ExternalProgramWrapper extProg)
   {
      this.extProg = extProg;
   }
   
   /**
    * Get the external procedure associated with this agent.
    * 
    * @return   The {@link #extProg procedure} associated with this agent or <code>null</code> if
    *           no procedure is associated.
    */
   public synchronized ExternalProgramWrapper getProcedure()
   {
      return this.extProg;
   }
   
   /**
    * Start the agent and listen for commands.
    */
   public void listen()
   {
      // register a listener to terminate the agent, if its session terminates
      Session sess = SessionManager.get().getSession();
      sess.addSessionListener(new SessionListener()
      {
         /**
          * This callback will be called when the P2J client session is ending.
          * 
          * @param    session
          *           The session that is ending.
          */
         @Override
         public void terminate(Session session)
         {
            // terminate the stopper worker explicitly - otherwise it will deadlock when the queue
            // tears down
            stopperWorker.terminate();
            
            Agent.this.terminate();
         }

         /**
          * This callback will be called after the security context has been set up.
          *
          * @param    session
          *           The session that is starting.
          */
         @Override
         public void initialize(Session session)
         {
            // no-ops
         }
      });
      
      prepare();

      String startup = pool.getStartup();
      // although official docs mention that state-reset doesn't call the startup procedure, 
      // testing showed that startup procedure is called
      handle hStartup = new handle();
      if (startup != null && startup.length() > 0)
      {
         String startupParameter = pool.getStartupParameter();
         if (!invoke(null, null, InvocationModes.IMMEDIATELY, startup, hStartup, "Startup",
                     "I", new character(startupParameter)))
         {
            // if there were problems invoking the startup procedure, terminate the agent
            return;
         }
      }
   
      // notify the launch
      AppServerLauncher.notifyLaunch(pool.getName(), true);
      
      // start the agent's stopper thread
      stopper.setName("Agent Stopper for " + Thread.currentThread().getName());
      stopper.setDaemon(true);
      stopper.start();
      
      while (!terminated())
      {
         AgentCommand command = null;
         
         // this loop will continue until someone sends the terminate command:
         // - in case the session is ended (P2J client gets killed)
         // - in case the manager is ended
         
         boolean interrupted = false;
         synchronized (this)
         {
            if (commands.size() == 0)
            {
               try
               {
                  wait();
               }
               catch (InterruptedException e)
               {
                  interrupted = true;
                  if (LOG.isWarnEnabled())
                  {
                     LOG.warn("Agent was interrupted for appserver " + pool.getName(), e);
                  }
               }
            }
   
            // get the last command (or null if none)
            command = commands.poll();
         }

         // someone has interrupted my waiting, check if the session is still alive
         if (interrupted && !sess.isRunning())
         {
            // mark the agent as terminated
            synchronized (this)
            {
               terminated = true;
            }

            stopper.interrupt();
            return;
         }
         
         if (command == null)
         {
            // no command yet, continue
            continue;
         }
         
         // execute only if an explicit command is not already stopped
         int reqId = command.getRequestId();
         String connId = command.getConnectionId();
         
         if (reqId == 0 || AppServerManager.registerRequest(reqId, this))
         {
            AppServerManager.addActiveAgent(connId, this);

            try
            {
               command.execute();
            }
            catch (Throwable t)
            {
               if (LOG.isWarnEnabled())
               {
                  LOG.warn("Agent encountered an error while executing a command for appserver " + 
                           pool.getName(), t);
               }
               command.setError(t);
            }
            finally
            {
               AppServerManager.removeActiveAgent(connId, this);

               if (reqId != 0)
               {
                  AppServerManager.deregisterRequest(reqId);
               }
               command.finished();
            }
         }
      }
      
      String shutdown = pool.getShutdown();
      // although official docs mention that state-reset doesn't call the shutdown procedure, 
      // testing showed that shutdown procedure is called
      if (shutdown != null && shutdown.length() > 0)
      {
         // ignore errors here
         invoke(null, null, InvocationModes.IMMEDIATELY, shutdown, null, "Shutdown", null);
      }
   
      // the startup procedure handle is deleted automatically, as the agent context is gone
      // we can't force an explicit delete, because we are outside of a P2J external block.

      // notify the agent's stopper thread
      stopper.interrupt();
   }

   /**
    * Reset the context of this agent.  Called only if the appserver is in State-reset operating
    * mode.
    * 
    * @param    mode
    *           How to execute the reset context command - immediately or send it to the agent.
    */
   public void resetContext(InvocationModes mode)
   {
      AgentCommand command = new ResetContextCommand();
      switch (mode)
      {
         case IMMEDIATELY:
            command.execute();
            break;
         default:
            waitForCommand(null, command);
            break;
      }
   }

   /**
    * Invoke the specified external procedure as persistent and wait (or not) for it to complete.
    * This is used to invoke any connect/disconnect and startup/shutdown procedures.
    * <p>
    * The "kind" parameter will always receive one of the "Startup", "Shutdown", "Connect" or 
    * "Disconnect" values.
    * 
    * @param    res
    *           The instance where to save the invocation result. May be <code>null</code>.
    * @param    id
    *           The appserver connection id. 
    * @param    mode
    *           Flag indicating how to invoke the external procedure.
    * @param    proc
    *           The legacy 4GL name for the external procedure.
    * @param    hProc
    *           If not-null, the handle where to save this procedure.
    * @param    kind
    *           A helper text about this procedure.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   <code>true</code> if no error was raised during the procedure invocation.
    */
   public boolean invoke(AppServerInvocationResult res,
                         final String              id,
                         final InvocationModes     mode,
                         final String              proc,
                         final handle              hProc,
                         final String              kind,
                         final String              modes,
                         final Object...           args)
   {
      final AppServerInvocationResult result = (res == null ? new AppServerInvocationResult() 
                                                            : res);

      AgentCommand command = new AgentCommand(id) 
      {
         @Override
         public void execute()
         {
            // the procedure is ran persistent only in certain cases:
            // 1. startup procedure 
            //   - automatically deleted when agent terminates
            //   - can be explicitly deleted via "delete object this-procedure".
            // 2. connect procedure
            //   - only for State-reset and State-aware modes
            //   - can be explicitly deleted via "delete object this-procedure".
            //   - automatically deleted on context reset, if State-reset
            
            // in both cases, the explicit delete is postponed until the program (startup or 
            // connect) terminates

            AppServerOperatingMode mode = pool.getOperatingMode();
            boolean persistentConnect = ("Connect".equals(kind) && 
                                         (mode == AppServerOperatingMode.STATE_RESET ||
                                          mode == AppServerOperatingMode.STATE_AWARE));
            boolean startup = ("Startup".equals(kind));
            
            boolean persistent = (persistentConnect || startup);
            
            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            Throwable error = null;
            try
            {
               if (persistent)
               {
                  ControlFlowOps.invokePersistentSetWithMode(proc, hProc, modes, args);
                  
               }
               else
               {
                  ControlFlowOps.invokeExternalProcedure(new character(proc), false, null, false,  
                                                         null, modes, args);
               }
            }
            catch (Throwable t)
            {
               boolean ignore = false;

               Throwable cause = t;
               while (cause != null)
               {
                  if (cause instanceof UnstoppableExitException)
                  {
                     LogicalTerminal.activateBatchMode(true);
                     ignore = true;
                     break;
                  }
                  
                  cause = cause.getCause();
               }
               
               if (!ignore)
               {
                  error = t;
               }
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
   
            if (error != null && LOG.isWarnEnabled())
            {
               // procedure has failed
               LOG.warn(String.format("%s procedure has failed for appserver %s:",
                                      kind, pool.getName()), 
                        error);
            }
            
            if (result != null)
            {
               result.setError(error);
               result.setResult(ControlFlowOps.getReturnValue());
            }
            
            // check if we are calling the disconnect procedure
            if ("Disconnect".equals(kind))
            {
               pool.terminateConnection(id, true);
            }
         }
      };
      
      switch (mode)
      {
         case IMMEDIATELY:
            try
            {
               command.execute();
            }
            catch (Throwable t)
            {
               result.setError(t);
            }
            break;

         case AGENT_NO_WAIT:
            sendCommand(command);
            break;

         case AGENT_WAIT:
            waitForCommand(result, command);
            break;
      }
      
      return (result.getError() == null);
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    resId
    *           The code of a remote procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invoke(final String    id,
                                           final int       requestId,
                                           final long      timeout,
                                           final character name,
                                           final String    resId,
                                           final boolean   function,
                                           final boolean   dynamicFunction,
                                           final boolean   superCall,
                                           final boolean   transactionDistinct,
                                           final String    modes,
                                           final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            String activate = pool.getActivate();
            if (!isBound() && activate != null && activate.length() > 0)
            {
               // call the activate procedure (only if the agent is not already bound)
               try
               {
                  invokeScoped(id, new character(activate));
               }
               catch (ConditionException ce)
               {
                  Condition c = BlockManager.exceptionToCondition(ce);
                  String type = c.toString();
                  int code = 0;
                  switch (c)
                  {
                     case ERROR:
                        code = 8025;
                        break;
                     case STOP:
                        code = 8026;
                        break;
                     case QUIT:
                        code = 8027;
                        break;
                  }

                  final String msg = "%s  activation procedure ended with an %s condition";
                  invocationResult.setError(
                     new NumberedException(String.format(msg, activate, type), code));
               }
               catch (Throwable t)
               {
                  boolean ignore = false;

                  Throwable cause = t;
                  while (cause != null)
                  {
                     if (cause instanceof UnstoppableExitException)
                     {
                        LogicalTerminal.activateBatchMode(true);
                        ignore = true;
                        break;
                     }
                     
                     cause = cause.getCause();
                  }
                  
                  if (!ignore)
                  {
                     invocationResult.setError(t);
                  }
               }
               finally
               {
                  // pass return value
                  invocationResult.setResult(ControlFlowOps.getReturnValue());
                  
                  // if an error was encountered during activate, call deactivate and disconnect
                  if (invocationResult.getError() != null)
                  {
                     String deactivate = pool.getDeactivate();
                     if (!isBound() && deactivate != null && deactivate.length() > 0)
                     {
                        // call the deactivate procedure (only if the agent is not bound)
                        try
                        {
                           invokeScoped(id, new character(deactivate));
                        }
                        catch (Throwable t)
                        {
                           // ignore all errors
                        }
                     }

                     AppServerManager.disconnect(id);
                     
                     return;
                  }
               }
            }
            
            // resolve the procedure
            handle proc = (resId == null ? null : new handle(pool.getProcedure(id, resId)));
            character exports = pool.getExports();
            boolean disconnect = false;
            final String localModes;
            if (args.length > 0)
            {
               if (proc == null)
               {
                  localModes = SourceNameMapper.getParameterModes(name.toStringMessage());
               }
               else if (ProcedureManager.isProcedure(proc))
               {
                  String pname = ProcedureManager.getAbsoluteName(proc.get());
                  String iename = name.toStringMessage();
                  localModes = SourceNameMapper.getParameterModes(pname, iename, function);
               }
               else
               {
                  localModes = null;
               }
            }
            else
            {
               localModes = null;
            }

            // make a copy of the requester's args
            final Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            Runnable argProcess = new Runnable()
            {
               @Override
               public void run()
               {
                  Object[] finalArgs = postProcessArguments(localModes, modes, args, requesterArgs);
                  invocationResult.setArguments(finalArgs);
                  invocationResult.setModes(localModes);
               }
            };
            
            try
            {
               preProcessArguments(localModes, modes, args);

               Object result = invokeScoped(id, name, proc,
                                            function, dynamicFunction, superCall,
                                            transactionDistinct, exports, 
                                            argProcess, modes, args);
               if (!function)
               {
                  result = ControlFlowOps.getReturnValue();
               }
               
               if (result instanceof handle)
               {
                  // handle results are converted to a container holding only the remote resource
                  handle h = (handle) result;
                  result = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
               }
               invocationResult.setResult(result);
            }
            catch (ErrorConditionException e)
            {
               Throwable cause = e.getCause();
               if (cause instanceof NumberedException)
               {
                  // only the following codes will be sent to the remote side
                  // TODO: this needs to be done only if the errors are caused by problems in
                  // resolving the appserver call target, and not by nested RUN statements
                  final Set<Integer> notifyErrs = new HashSet<Integer>();
                  final Integer[] codes = 
                     { 6349, 3230, 1005, 3234, 5729, 2570, 11428, 12290, 15472, 5639, 6445, 8030,
                       12324 };
                  notifyErrs.addAll(Arrays.asList(codes));

                  NumberedException ne = (NumberedException) cause;
                  int num = ne.getNumber();
                  if (notifyErrs.contains(num))
                  {
                     invocationResult.setError(e);
                  }
               }
            }
            catch (QuitConditionException e)
            {
               invocationResult.setError(e);
               disconnect = true;
            }
            catch (Throwable t)
            {
               boolean ignore = false;

               Throwable cause = t;
               while (cause != null)
               {
                  if (cause instanceof UnstoppableExitException)
                  {
                     LogicalTerminal.activateBatchMode(true);
                     ignore = true;
                     break;
                  }
                  
                  cause = cause.getCause();
               }
               
               if (!ignore)
               {
                  invocationResult.setError(t);
               }
            }
            finally
            {
               String deactivate = pool.getDeactivate();
               if (!isBound() && deactivate != null && deactivate.length() > 0)
               {
                  // call the deactivate procedure (only if the agent is not bound)
                  try
                  {
                     invokeScoped(id, new character(deactivate));
                  }
                  catch (Throwable t)
                  {
                     // ignore all errors
                  }
               }

               if (disconnect)
               {
                  AppServerManager.disconnect(id);
               }
            }
         }
      });
      
      return invocationResult;
   }

   /**
    * Obtain a proxy for the given external procedure. This will also reserve an agent which will
    * be used for all subsequent invocation (including external procedure initialization).
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           A name of a remote external procedure.
    *           
    * @return   A proxy to this procedure. May be invalid if the remote side can't resolve the
    *           procedure.
    */
   ExternalProgramWrapper obtainProxy(final String id, final String name)
   {
      final handle hProc = new handle();
      
      waitForCommand(null, new AgentCommand(id)
      {
         @Override
         public void execute()
         {
            // add the topmost scope to the TransactionManager; this will be popped only when:
            // a. the ext prog can't be resolved
            // b. the ext prog has been initialized
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);

            ExternalProgramWrapper extProg = null;
            
            try
            {
               extProg = ControlFlowOps.resolveExternalProcedure(name, pool.getExports());
               
               if (extProg != null)
               {
                  hProc.assign(extProg);
                  
                  // save this handle globally, as it needs to be exposed to the conversation thread
                  // which handles the AppServerEntry calls too
                  if (id != null)
                  {
                     pool.saveProcedure(id, Agent.this, hProc.toString(), extProg);
                  }
   
                  // make it bound only if the invocation was successful
                  bind();
                  
                  // associate the agent with this ext prog
                  setProcedure(extProg);
               }
            }
            
            finally
            {
               if (extProg == null)
               {
                  // remove the topmost scope from the TransactionManager, as the ext prog could
                  // not be resolved
                  TransactionManager.popScope();
               }
            }
         }
      });

      return (ExternalProgramWrapper) hProc.getResource();
   }

   /**
    * Initialize the specified proxy procedure, by calling the <code>execute</code> method.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *           
    * @return   The result of the external program initialization.
    */
   public AppServerInvocationResult initializeProxy(final String   id,
                                                    final int      requestId,
                                                    final boolean  transactionDistinct,
                                                    final String   modes,
                                                    final Object[] args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(id, requestId)
      {
         @Override
         public void execute()
         {
            String name = ProcedureManager.getAbsoluteName(extProg.referent);
            String localModes = null;
            if (args.length > 0)
            {
               localModes = SourceNameMapper.getParameterModes(name);
            }

            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            try
            {
               preProcessArguments(localModes, modes, args);
               
               ControlFlowOps.initializeExternalProcedure(extProg,
                                                          transactionDistinct, 
                                                          modes,
                                                          args);
            }
            catch (StopConditionException e)
            {
               if (invokeFailure(e))
               {
                  return;
               }
               
               throw e;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(localModes, modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);
               invocationResult.setModes(localModes);

               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }

            invocationResult.setResult(ControlFlowOps.getReturnValue());
            invocationResult.setProcedureId(new handle(extProg).toString());
         }
      });
      
      return invocationResult;
   }

   
   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *
    * @return    A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *            procedure, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    */
   public AppServerInvocationResult invokePersistent(final String    id,
                                                     final int       requestId,
                                                     final long      timeout,
                                                     final character name,
                                                     final boolean   transactionDistinct,
                                                     final String    modes,
                                                     final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            final handle hProc = new handle();

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            String localModes = null;
            if (args.length > 0)
            {
               localModes = SourceNameMapper.getParameterModes(name.toStringMessage());
            }

            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            try
            {
               preProcessArguments(localModes, modes, args);
               
               character exports = pool.getExports();
               ControlFlowOps.invokeExternalProcedure(name, true, hProc, transactionDistinct,
                                                      exports, modes, args);
            }
            catch (StopConditionException e)
            {
               if (invokeFailure(e))
               {
                  return;
               }
               
               throw e;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(localModes, modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);
               invocationResult.setModes(localModes);

               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }

            // save this handle globally, as it needs to be exposed to the conversation thread
            // which handles the AppServerEntry calls too
            pool.saveProcedure(id,
                               Agent.this,
                               hProc.toString(),
                               (PersistentProcedure) hProc.getResource());

            // make it bound only if the invocation was successful
            bind();
            
            // set the agent's procedure
            setProcedure((ExternalProgramWrapper) hProc.getResource());
            
            invocationResult.setResult(ControlFlowOps.getReturnValue());
            invocationResult.setProcedureId(hProc.toString());
         }
      });

      return invocationResult;
   }

   /**
    * Send the terminate command to this agent.
    */
   public void terminate()
   {
      // interrupt current work
      stop();
      
      // send the terminate command
      sendCommand(new AgentCommand(null)
      {
         @Override
         public void execute()
         {
            synchronized (Agent.this)
            {
               terminated = true;
            }
         }
      });
   }

   /**
    * Check if this agent has been terminated.
    * 
    * @return   See above.
    */
   public synchronized boolean terminated()
   {
      return terminated;
   }
   
   /**
    * Bind the agent to the connection.
    */
   public synchronized void bind()
   {
      this.bound = true;
   }
   
   /**
    * Unbind the agent from the connection.
    */
   public synchronized void unbind()
   {
      this.bound = false;
   }
   
   /**
    * Check if the agent is bound to the connection.
    * 
    * @return   See above.
    */
   public synchronized boolean isBound()
   {
      return bound;
   }
   
   /**
    * Delete the specified procedure, under the Agent's context.
    * 
    * @param    proc
    *           A persistent procedure.
    */
   public void deleteProcedure(final PersistentProcedure proc)
   {
      waitForCommand(null, new AgentCommand(null)
      {
         @Override
         public void execute()
         {
            if (handle.knownResource(proc))
            {
               proc.delete();
            }
         }
      });
   }
   
   /**
    * Set the current Agent's connection ID.
    * 
    * @param    id
    *           The connection ID.
    */
   void setConnectionId(String id)
   {
      waitForCommand(null, new AgentCommand(id)
      {
         @Override
         public void execute()
         {
            AppServerManager.setConnectionId(id);
         }
      });
   }

   /**
    * Send a STOP condition to the request currently being executed on this connection.
    * 
    * @throws   IllegalStateException
    *           If the appserver is not connected.
    */
   void stop()
   {
      // the STOP condition needs to be raised from a thread with the agent's context.  assuming
      // the agent's thread is busy executing another command, we can't use the agent's thread
      // to send this stop.  so an AssociatedThread with the same context as the Agent is used 
      // instead to wait for STOP commands
      synchronized (stopper)
      {
         stopper.notify();
      }
   }
   
   /**
    * Send the given command to this agent and return immediately.
    * 
    * @param    command
    *           The command to be executed.
    */
   private synchronized void sendCommand(AgentCommand command)
   {
      commands.add(command);
      
      this.notify();
   }
   
   /**
    * Send the given command to the agent and wait for it to complete.
    * 
    * @param    result
    *           The instance where to place the command result.
    * @param    command
    *           The command to be executed.
    */
   private void waitForCommand(AppServerInvocationResult result, AgentCommand command)
   {
      sendCommand(command);

      // we need to synchronize on the loop, else the command might finish between the test and
      // the wait and the notification will never be picked up
      synchronized (command)
      {
         long timeout = command.getTimeout();
         boolean waitForFinish = (timeout == 0);
         long start = (!waitForFinish ? 0 : System.currentTimeMillis());
         
         while (!command.isFinished())
         {
            try
            {
               command.wait(timeout);
            }
            catch (InterruptedException e)
            {
               if (LOG.isWarnEnabled())
               {
                  LOG.warn("Command was interrupted for appserver " + pool.getName(), e);
               }
            }
            
            if (!waitForFinish)
            {
               long finish = System.currentTimeMillis();
               long elapsed = finish - start;
               
               if (!command.isFinished() && elapsed >= timeout)
               {
                  // terminate the agent's command, and wait for it to be marked as "finished"
                  stop();
                  timeout = 0;
                  waitForFinish = true;
               }
               else
               {
                  // wait again, for the remainder time
                  timeout = timeout - elapsed;
               }
            }
         }
      }

      if (result != null && command.getError() != null)
      {
         result.setError(command.getError());
      }
   }
   
   /**
    * Preprocess the arguments sent by the requester so that they are converted to instances
    * compatible with the P2J runtime.  This includes:
    * <ul>
    * <li>Creating the handles for all received resources</li>
    * <li>Creating the temp-tables for the received {@link TableWrapper} instances</li>
    * <li>Creating {@link memptr} instances from all {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    localModes
    *           The parameter modes as defined at the external procedure or internal entry.
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    args
    *           The argument list sent by the requester.  The elements in this array will be
    *           changed as needed.
    */
   private void preProcessArguments(String localModes, String requesterModes, Object[] args)
   {
      if (localModes != null && requesterModes != null && !localModes.equals(requesterModes))
      {
         // invocation will fail, so nothing to prepare
         return;
      }
      
      String modes = (localModes == null ? requesterModes : localModes);
      if (modes == null || modes.length() != args.length)
      {
         // no modes specified or incompatible modes
         return;
      }
      
      for (int i = 0; i < modes.length(); i++)
      {
         Object arg = args[i];

         if (arg instanceof TableWrapper)
         {
            // a result set will be sent only if the requester and the appserver are not in the  
            // same JVM
            TableWrapper rs = (TableWrapper) arg;
            arg = new TableParameter(rs, rs.isAppend());
            ((TableParameter) arg).setParameterIndex(i + 1);
         }
         else
         {
            boolean isArray = arg.getClass().isArray();
            Object[] data = null;
            
            if (isArray)
            {
               // check if this argument is an array
               data = new Object[Array.getLength(arg)];
               
               for (int j = 0; j < data.length; j++)
               {
                  data[j] = Array.get(arg, j);
               }
            }
            else
            {
               data = new Object[] { arg };
            }
            
            for (int j = 0; j < data.length; j++)
            {
               Object darg = data[j];
               
               if (darg instanceof RemoteResource)
               {
                  String resid = ((RemoteResource) darg).getResourceId();
                  darg = (resid == null ? new handle() : handle.fromString(resid));
               }
               else if (darg instanceof MemoryBuffer)
               {
                  // convert MemoryBuffer instances to memptr instances
                  darg = new memptr(((MemoryBuffer) darg).getValue());
               }
               
               data[j] = darg;
            }
            
            arg = (isArray ? data : data[0]);
         }
         
         args[i] = arg;
      }
   }

   /**
    * Post-process the given arguments so that all OUTPUT and INPUT-OUTPUT arguments will be sent
    * back to the requester. This includes:
    * <ul>
    *    <li>Serializing all table parameters via {@link TableWrapper} instances</li>
    *    <li>Sending the resource IDs for all handle parameters</li>
    *    <li>Converting all {@link memptr} instances to {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    localModes
    *           The parameter modes as defined at the external procedure or internal entry.
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    localArgs
    *           The argument list after the call has been performed.
    * @param    requesterArgs
    *           The argument list sent by the requester.
    * 
    * @return   The postprocessed arguments, to be sent back to the requester.
    */
   private Object[] postProcessArguments(String   localModes,
                                         String   requesterModes,
                                         Object[] localArgs,
                                         Object[] requesterArgs)
   {
      try
      {
         if (localModes != null && requesterModes != null && !localModes.equals(requesterModes))
         {
            // invocation will fail, so nothing to prepare
            return new Object[0];
         }
         
         String modes = (localModes == null ? requesterModes : localModes);
         if (modes == null || requesterArgs.length != modes.length())
         {
            // no modes specified, so no arguments can be returned - assume all are input
            return new Object[0];
         }
         
         Object[] arguments = new Object[localArgs.length];
         for (int i = 0; i < modes.length(); i++)
         {
            char mode = modes.charAt(i);
   
            // in input mode, the argument is not sent back
            Object localArg = (mode == 'O' || mode == 'U' ? localArgs[i] : null);
            
            if (localArg == null)
            {
               // nothing to post-process if the local arg was determined to be null
               arguments[i] = null;
               continue;
            }
            Object requesterArg = requesterArgs[i];
   
            // if output/input-output mode, send back the data. 
   
            // a table parameter is serialized to a result set only if the requester has sent a 
            // result set
            if (localArg instanceof TableParameter && requesterArg instanceof TableWrapper)
            {
               // we have a DMO on the remote side, use the P2J persistence layer to get the data
               TableParameter tableParam = (TableParameter) localArg;
               localArg = tableParam.getResultSet();
            }
            else
            {
               boolean isArray = localArg.getClass().isArray();
               Object[] data = null;
               
               if (isArray)
               {
                  // check if this argument is an array
                  data = new Object[Array.getLength(localArg)];
                  
                  for (int j = 0; j < data.length; j++)
                  {
                     data[j] = Array.get(localArg, j);
                  }
               }
               else
               {
                  data = new Object[] { localArg };
               }
               
               for (int j = 0; j < data.length; j++)
               {
                  Object dlocalArg = data[j];
                  
                  if (dlocalArg instanceof handle)
                  {
                     handle h = (handle) dlocalArg;
                     
                     dlocalArg = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
                  }
                  else if (dlocalArg instanceof memptr)
                  {
                     // convert memptr instances to MemoryBuffer instances
                     dlocalArg = new MemoryBuffer(((memptr) dlocalArg).asByteArray());
                  }
                  
                  data[j] = dlocalArg;
               }
               
               localArg = (isArray ? data : data[0]);
            }
            
            arguments[i] = localArg;
         }
         
         return arguments;
      }
      finally
      {
         for (int i = 0; i < localArgs.length; i++)
         {
            Object arg = localArgs[i];
            
            if (arg instanceof memptr)
            {
               // release the memory for all memptr arguments
               ((memptr) arg).setLength(0);
            }
         }
      }
   }

   /**
    * Prepare this agent by setting some batch-processing related configuration.
    */
   private void prepare()
   {
      // make sure the TM is loaded before LT, as the LT context-local instance is dependent on TM
      // when creating the ScopeableFactory instances, LT will be loaded by TM and this instance
      // will be used as the factory. if we don't do this and load LT first, TM will end up with
      // the wrong LT instance as ScopeableFactory.
      TransactionManager.isHeadless();

      // avoid the end-of-program message
      LogicalTerminal.setRelogin(true);
      
      // set this as batch mode
      EnvironmentOps.setBatchMode(true);
      LogicalTerminal.activateBatchMode(true);

      // set the propath
      EnvironmentOps.setSearchPath(pool.getPropath());
      
      // force DatabaseTriggerManager context-local instantiation
      DatabaseTriggerManager.get();

      // Global scope for keeping buffers states (like BufferManager.allBuffers). This scope is
      // not explicitly popped, removed when the agent context is reset.
      TransactionManager.pushScope("appserver-agent",
                                   TransactionManager.NO_TRANSACTION,
                                   true,
                                   true,
                                   false,
                                   false);
   }
   
   /**
    * Implements the reset context command.
    */
   private class ResetContextCommand
   extends AgentCommand
   {
      public ResetContextCommand()
      {
         super(null);
      }

      @Override
      public void execute()
      {
         // before cleaning up, cleanup all still-alive dynamic frames
         LogicalTerminal.cleanupDynamicFrames();

         // reset the context
         try
         {
            SecurityManager.getInstance().resetContext();
         }
         catch (RestrictedUseException e)
         {
            // ignore
         }
         finally
         {
            // set some state again
            prepare();
         }
      }
   }
   
   /**
    * Class used to stop a command currently being executed by this Agent, by sendint an interrupt
    * request via {@link Control#interrupt()}.
    */
   private class AgentStopper
   implements Runnable
   {
      /** Flag indicating if the stopper was explicitly terminated. */
      private volatile boolean terminated = false;
      
      @Override
      public void run()
      {
         boolean doInterrupt = false;
         
         while (!terminated && !Agent.this.terminated())
         {
            synchronized (stopper)
            {
               try
               {
                  stopper.wait();
                  doInterrupt = !terminated;
               }
               catch (InterruptedException e)
               {
                  // ignore - if the thread gets interrupted, the loop will terminate only when
                  // the agent is terminated
               }
            }
            
            if (doInterrupt)
            {
               doInterrupt = false;
               
               // received a notification, if agent is still running, stop its command
               if (!Agent.this.terminated())
               {
                  Control.interrupt();
               }
               else
               {
                  // terminated, finish the loop
                  break;
               }
            }
         }
      }

      /**
       * Terminate the stopper thread.
       */
      public void terminate()
      {
         this.terminated = true;
         
         synchronized (stopper)
         {
            stopper.notify();
         }
      }
   }
   
   /**
    * Defines a command to be sent to an agent.
    */
   private static abstract class AgentCommand
   {
      /** The ID of the request. */
      private final int requestId;
      
      /**
       * The connection ID on which this command is executed.  May be <code>null</code> if this
       * command must terminate execution always (i.e. a procedure proxy initialization, procedure
       * deletion, etc).
       */
      private final String connectionId;
      
      /**
       * The maximum allowed time for this invocation to complete.  In milliseconds.
       * Use <code>0</code> to disable the timeout.
       */
      private final long timeout;
      
      /** Any error caught during the command execution. */
      private Throwable error = null;
      
      /** Flag indicating if this command has finished or not. */
      private boolean finished = false;
      
      /**
       * Initialize this command for no explicit request (an internal request).
       * 
       * @param    id
       *           The connection ID on which this command is executed.
       */
      public AgentCommand(String id)
      {
         this(id, -1);
      }
      
      /**
       * Initialize this command for an explicit request.
       *  
       * @param    id
       *           The connection ID on which this command is executed.
       * @param    requestId
       *           The ID of the request.
       */
      public AgentCommand(String id, int requestId)
      {
         this(id, requestId, 0);
      }
      
      /**
       * Initialize this command for an explicit request.
       *  
       * @param    id
       *           The connection ID on which this command is executed.
       * @param    requestId
       *           The ID of the request.
       * @param    timeout
       *           The maximum allowed time for this invocation to complete.  In milliseconds.
       *           Use <code>0</code> to disable the timeout.
       */
      public AgentCommand(String id, int requestId, long timeout)
      {
         this.requestId = requestId;
         this.connectionId = id;
         this.timeout = timeout;
      }
      
      /** Override this to define the command body. */
      public abstract void execute();
      
      /**
       * Get the request ID of this command.
       * 
       * @return   The request ID.
       */
      public int getRequestId()
      {
         return requestId;
      }
      
      /**
       * Get the {@link #connectionId connection ID} where this command is executed.
       * 
       * @return   The connection ID.
       */
      public String getConnectionId()
      {
         return connectionId;
      }
      
      /**
       * Get the {@link #timeout} of execution for this command.
       * 
       * @return   The connection ID.
       */
      public long getTimeout()
      {
         return timeout;
      }
      
      /**
       * Check the {@link #finished} flag.
       * 
       * @return   See above.
       */
      public synchronized boolean isFinished()
      {
         return finished;
      }

      /**
       * Notify that this command has finished.
       */
      public synchronized void finished()
      {
         this.finished = true;
         this.notify();
      }
      
      /**
       * Set the error caught during this command execution.
       * 
       * @param    t
       *           The caught error.
       */
      public synchronized void setError(Throwable t)
      {
         this.error = t;
      }
      
      /**
       * Get the error caught during this command execution.
       * 
       * @return   See above.
       */
      public synchronized Throwable getError()
      {
         return this.error;
      }
   }
}