/*
** Module   : AppServerManager.java
** Abstract : Defines APIs to manage the appservers defined by this P2J server and any incoming
**            remote API calls for these appservers.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130628 Context-local data for this class must not be reset by the agent.  Added a 
**                  listener to automatically disconnect if the remote side has closed its session
**                  while the connection is still active.
** 003 CA  20130705 Reworked the appserver launching.
** 004 CA  20130822 Fixed server disconnection when the P2J client ends abruptly, while still
**                  connected to a local appserver.
** 005 OM  20131018 Updated to conform the new SessionListener API.
** 006 CA  20140326 The remote persistent procedures need to be deleted by the Agent which created
**                  them.
** 007 CA  20140826 Fixed agent context cleanup if the P2J session is terminated before the 
**                  appserver connection is establised.
** 008 CA  20141124 Fixed uniqueId() - the nextId field was not being incremented properly, making
**                  execution of async requests impossible.
** 009 IAS 20160331 Fixed appServers.remove
** 010 CA  20160404 Enhanced the appserver invocations, to allow implicit or explicit termination 
**                  of requests being executing on a certain connection.
** 011 CA  20160405 Fixed H010 - during CONNECT procedure call, activeAgents can not be tracked.
** 012 CA  20180503 Added SESSION:SERVER-CONNECTION-ID support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;

import org.apache.commons.logging.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * Manage the appservers defined by this P2J server by:
 * <ul>
 * <li>starting the agents when the appserver is started</li>
 * <li>listening and handling all incoming requests for the managed appservers</li>
 * </ul>
 * Also, it manages the appservers known to this P2J server. These known appservers are 
 * disambiguated using the data configured in directory,  under the "server/default/app_services"
 * node.
 */
public class AppServerManager
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerManager.class);

   /** Provides a map of default appService names, for each name server. */
   private static Map<String, String> nameServerDefaults = null;
   
   /** The list of the known, remote, appservers. */
   private static List<RemoteAppServer> remoteAppServers = null;
   
   /** Map of subject IDs to appserver names. */
   private static final Map<String, String> subjectToAppserver = new HashMap<>();

   /** Map of appservers to their associated pool, managed by this P2J server. */
   private static final Map<String, AgentPool> appServers = new HashMap<>();
   
   /** A map of authorized sessions which can access the given connection ID. */
   private static final Map<String, Set<Integer>> authorizedSessions = new HashMap<>();

   /** Map of executing requests to their associated Agents. */
   private static final Map<Integer, Agent> requests = new HashMap<>();
   
   /** Set of stopped requests. */
   private static final Set<Integer> stoppedRequests = new HashSet<>();
   
   /**
    * For each connection ID, this map includes the Agents which are currently executing requests.
    */
   private static final Map<String, Set<Agent>> activeAgents = new HashMap<>();
   
   /** Lock used to sync access to {@link #requests} and {@link #stoppedRequests}. */
   private static final Object lock = new Object();
   
   /** The next request ID. */
   private static int nextId = 0;
   
   /**
    * This context-local data is used only to determine which appserver is authenticated and its
    * mode (manager or agent). 
    */
   private static ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
      
      protected boolean isResetAllowed()
      {
         return false;
      };
   };

   /**
    * Obtain an unique ID from the remote side.
    * 
    * @return   An unique ID.
    */
   public synchronized static int uniqueId()
   {
      do
      {
         nextId = nextId + 1;
         if (nextId == Integer.MAX_VALUE)
         {
            nextId = 0;
         }
        
         synchronized (lock)
         {
            if (!requests.containsKey(nextId) && !stoppedRequests.contains(nextId))
            {
               break;
            }
         }
      }
      while (true);
     
      return nextId;
   }

   /**
    * Obtain a proxy for the given external procedure. This will also reserve an agent which will
    * be used for all subsequent invocation (including external procedure initialization).
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           A name of a remote external procedure.
    *           
    * @return   A proxy code for this procedure. May be invalid if the remote side can't resolve 
    *           the procedure.
    */
   public static String obtainProxy(String id, String name)
   {
      AgentPool pool = getAgentPool(id);
      
      // acquire a free agent
      Agent agent = null;
      String code = null;
      
      try
      {
         agent = pool.acquireAgent(id, null);
         ExternalProgramWrapper extProg = agent.obtainProxy(id, name);

         code = new handle(extProg).toString();
         
         return code;
      }
      finally
      {
         pool.releaseAgent(id, agent);
      }
   }
   
   /**
    * Initialize the specified proxy procedure, by calling the <code>execute</code> method.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    code
    *           The code of a persistent external procedure to which the agent is associated. May
    *           be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *           
    * @return   The result of the external program initialization.
    */
   public static AppServerInvocationResult initializeProxy(String   id,
                                                           int      requestId,
                                                           String   code, 
                                                           boolean  transactionDistinct,
                                                           String   modes,
                                                           Object[] args)
   {
      AgentPool pool = getAgentPool(id);
      
      Agent agent = pool.acquireAgent(id, code);
      
      try
      {
         // all internal-entries must use the same agent, if this is async and session-free!!!
         return agent.initializeProxy(id, requestId, transactionDistinct, modes, args);
      }
      finally
      {
         pool.releaseAgent(id, agent);
      }
   }
   
   /**
    * Send the STOP condition to the request running with the specified ID.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    */
   public static void sendStop(String id, int requestId)
   {
      // security check
      getAgentPool(id);
      
      synchronized (lock)
      {
         Agent a = requests.get(requestId);
         if (a == null)
         {
            // this means the agent didn't get a chance to start, collect this as a stopped 
            // request
            stoppedRequests.add(requestId);
         }
         else
         {
            // terminate the agent
            a.stop();
         }
      }
   }
   
   /**
    * Send a STOP condition to the request currently being executed on the specified connection.
    * 
    * @param    id
    *           The appserver connection id. 
    */
   public static void sendStop(String id)
   {
      // security check
      getAgentPool(id);

      // get all requests currently processed for this connection ID and terminate them
      synchronized (lock)
      {
         Set<Agent> active = activeAgents.get(id);
         
         for (Agent a : active)
         {
            a.stop();
         }
      }
   }
   
   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and name-patterns separated by comma.
    *         
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(String procList)
   {
      return export(new character(procList));
   }

   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and name-patterns separated by comma.
    *
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(character procList)
   {
      AgentPool pool = getAgentPoolBySubject(SecurityManager.getInstance().getUserId());
      if (pool == null)
      {
         // not an AppServer
         return new logical(false);
      }
      pool.setExports(procList);
      
      return new logical(true);
   }

   /**
    * Method implementation for getting the SERVER-OPERATING-MODE attribute of the session handle.
    * <p>
    * Possible values:
    * <ul>
    *    <li>State-reset
    *    <li>State-aware
    *    <li>Stateless
    *    <li>State-free
    * </ul>
    * 
    * @return   The corresponding SERVER-OPERATING-MODE attribute of the session handle.
    */
   public static character getServerOperatingMode()
   {
      AgentPool pool = getAgentPoolBySubject(SecurityManager.getInstance().getUserId());
      return new character(pool == null ? "" : pool.getOperatingMode().toString());
   }

   /**
    * Indicates if the session is an AppServer session or not.
    * 
    * @return   <code>true</code> if the current session is an AppServer session or false 
    *           otherwise.
    */
   public static boolean isRemote()
   {
      return local.get().appServer != null;
   }
   
   /**
    * Get the value of SESSION:SERVER-CONNECTION-ID attribute. 
    * 
    * @return  the value of SESSION:SERVER-CONNECTION-ID attribute.
    */
   public static String getServerConnectionId()
   {
      WorkArea wa = local.get();
      
      return wa.appServer == null ? "" : wa.connectionId;
   }

   /**
    * Get the name of the remote procedure, specified by the resource ID.
    *  
    * @param    id
    *           The appserver connection id. 
    * @param    resId
    *           The ID of a remote procedure.
    *
    * @return   The procedure's name.
    *
    * @see      #getAgentPool(String)
    */
   public static character getProcedureName(String id, String resId)
   {
      AgentPool pool = getAgentPool(id);
      
      PersistentProcedure pp = pool.getProcedure(id, resId);
      
      return pp == null ? new character() : pp.getFileName();
   }

   /**
    * Delete the remote procedure, specified by the resource id. This must be called at the same
    * time when the local proxy procedure is deleted.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    resId
    *           The ID of a remote procedure which needs to be deleted.
    *
    * @return   <code>true</code> if the remote procedure was deleted.
    * 
    * @see      #getAgentPool(String)
    */
   public static boolean deleteProcedure(String id, String resId)
   {
      AgentPool pool = getAgentPool(id);
      PersistentProcedure pp = pool.removeProcedure(id, resId);
      boolean knownProc = (pp != null);
      
      if (knownProc)
      {
         // this needs to be done using the agent's context
         pool.deleteProcedure(pp);
      }

      return knownProc;
   }
   
   /**
    * Establish a new appserver connection. This will invoke any connect procedure for this 
    * appserver, to which the given <code>user</code>, <code>pwd</code> and <code>serverInfo</code>
    * parameters will be passed.
    * <p>
    * The appserver connection must be configured to start under the given P2J-level subject ID.
    *  
    * @param    appServer
    *           The name of the target appserver.
    * @param    sessionFree
    *           Flag indicating if the connection request is for a session-free operating mode.
    * @param    user
    *           The first parameter to pass to the connect procedure, if any.
    * @param    pwd
    *           The second parameter to pass to the connect procedure, if any.
    * @param    serverInfo
    *           The third parameter to pass to the connect procedure, if any.
    *           
    * @return   An {@link AppServerInvocationResult} instance with any result or caught errors.
    */
   public static AppServerInvocationResult connect(String  appServer,
                                                   boolean sessionFree,
                                                   String  user,
                                                   String  pwd,
                                                   String  serverInfo)
   {
      final AgentPool pool;
      synchronized (appServers)
      {
         pool = appServers.get(appServer);
      }

      AppServerInvocationResult result = null;

      if (pool == null)
      {
         result = new AppServerInvocationResult();
         
         result.setError(new RuntimeException("No appserver named " + appServer + " exists!"));
         
         return result;
      }
      
      // latch to mark when the session listener can process
      final CountDownLatch connectDone = new CountDownLatch(1);
      
      // special var to track session disconnected state
      final AtomicBoolean sessionTerminated = new AtomicBoolean(false);

      // the connect ID - use an array to expose it to the session listener
      final String[] connectId = new String[1];
      
      // the session listener needs to be registered early, in case the the session is terminated
      // while establishing the appserver connection
      SessionListener listener = new SessionListener()
      {
         /**
          * This callback will be called when the P2J client session is ending.
          * 
          * @param    session
          *           The session that is ending.
          */
         @Override
         public void terminate(Session session)
         {
            // notify the session has been terminated
            sessionTerminated.set(true);

            // wait for the connect procedure to process, before going forward
            try
            {
               connectDone.await();
            }
            catch (InterruptedException e)
            {
               // no-op
            }
            
            // 1. the connection was not established: thus nothing to do, as the agent should be
            //    properly disconnect at this time.
            if (connectId[0] == null)
            {
               return;
            }

            // 2. the connection was established, thus disconnect might be needed
            SecurityManager sm = SecurityManager.getInstance();
            synchronized (authorizedSessions)
            {
               Set<Integer> sessions = authorizedSessions.get(connectId[0]);
               
               // only if we are the last authorized session on this connection
               if (sessions != null                     && 
                   sessions.contains(sm.getSessionId()) &&
                   sessions.size() == 1)
               {
                  disconnect(connectId[0]);
               }
            }
            
            synchronized (lock)
            {
               activeAgents.remove(connectId[0]);
            }
         }

         /**
          * This callback will be called after the security context has been set up.
          *
          * @param    session
          *           The session that is starting.
          */
         @Override
         public void initialize(Session session)
         {
            // no ops
         }
      };
      
      final Session session = SessionManager.get().getSession();
      
      // the session listener registration needs to protected, so that the queue can't be stopped
      // while adding it.  This can be done only by obtaining a lock on the Queue object associated
      // with the session, and is done by the specialized API in the Session class.
      
      if (!session.addSessionListener(listener, true))
      {
         result = new AppServerInvocationResult();
         
         result.setError(new RuntimeException("Client session disconnected before the connect" +
                                              "procedure was started."));
         
         return result;
      }

      // WARNING! From this point forward, no synchronized Queue APIs can be called, else a
      // deadlock might occur.  Also, no other listeners can be added for the same reason.
      try
      {
         result = pool.connect(sessionTerminated, appServer, sessionFree, user, pwd, serverInfo);
      }
      finally
      {
         if (result != null && result.getError() == null)
         {
            connectId[0] = result.getConnectionID();
            
            if (connectId[0] != null)
            {
               // only if the client socket is still alive
               
               SecurityManager sm = SecurityManager.getInstance();
               
               // add this context to the authorized connection IDs
               synchronized (authorizedSessions)
               {
                  Set<Integer> sessions = authorizedSessions.get(connectId[0]);
                  if (sessions == null)
                  {
                     authorizedSessions.put(connectId[0], sessions = new HashSet<Integer>());
                  }
                  
                  sessions.add(sm.getSessionId());
               }
               
               synchronized (lock)
               {
                  activeAgents.put(connectId[0], new HashSet<>());
               }
            }
         }
         
         // notify the session listener that now is OK to process
         connectDone.countDown();
      }

      return result;
   }

   /**
    * Disconnect the appserver connection, with the given ID.  It will invoke any configured 
    * disconnect procedure, but will not return any errors or values to the caller.
    * 
    * @param    id
    *           The connection ID to disconnect.
    *
    * @see      #getAgentPool(String)
    */
   public static void disconnect(String id)
   {
      if (!connected(id))
      {
         return;
      }

      AgentPool pool = getAgentPool(id);
      
      try
      {
         pool.disconnect(id);
      }
      finally
      {
         SecurityManager sm = SecurityManager.getInstance();
         synchronized (authorizedSessions)
         {
            Set<Integer> sessions = authorizedSessions.get(id);
            sessions.remove(sm.getSessionId());
            
            if (sessions.isEmpty())
            {
               authorizedSessions.remove(id);
            }
         }
      }
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    procID
    *           The code of a remote procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    *
    * @see      #getAgentPool(String)
    */
   public static AppServerInvocationResult invoke(String    id,
                                                  int       requestId,
                                                  long      timeout,
                                                  character name,
                                                  String    procID,
                                                  boolean   function,
                                                  boolean   dynamicFunction,
                                                  boolean   superCall,
                                                  boolean   transactionDistinct,
                                                  String    modes,
                                                  Object... args)
   {
      AgentPool pool = getAgentPool(id);
      
      // get the agent from the pool, depending on the operating mode
      // ensure all internal-entries must use the same agent, if this is async and session-free
      Agent agent = pool.acquireAgent(id, procID);
      
      try
      {
         // send the invoke command to the agent
         return agent.invoke(id, requestId, timeout, name, procID,
                             function, dynamicFunction, superCall,
                             transactionDistinct, modes, args);
      }
      finally
      {
         pool.releaseAgent(id, agent);
      }
   }
   
   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *
    * @return    A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *            procedure, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    *
    * @see      #getAgentPool(String)
    */
   public static AppServerInvocationResult invokePersistent(String    id,
                                                            int       requestId,
                                                            long      timeout,
                                                            character name,
                                                            boolean   transactionDistinct,
                                                            String    modes,
                                                            Object... args)
   {
      AgentPool pool = getAgentPool(id);
      
      // get the agent from the pool, depending on the operating mode
      // ensure all internal-entries must use the same agent, if this is async and session-free
      Agent agent = pool.acquireAgent(id, null);
      
      try
      {
         // send the invoke command to the agent
         return agent.invokePersistent(id, requestId, timeout, name, transactionDistinct, modes, args);
      }
      finally
      {
         pool.releaseAgent(id, agent);
      }
   }

   /**
    * Search through the known remote appservers and determine the address of the P2J server on
    * which the target appserver is running.
    * 
    * @param    appService
    *           The appserver's service name (the -AppService connect parameter).
    * @param    host
    *           The -H connect parameter.
    * @param    serviceName
    *           The network service name, specified by the -S parameter.
    * @param    port
    *           The network port, specified by the -S parameter.
    * @param    directConnect
    *           Flag indicating this is a direct connect (the -DirectConnect parameter).
    *
    * @return   The address of the P2J server which handles this appserver.
    */
   public static InetSocketAddress lookupP2JServer(String  appService,
                                                   String  host,
                                                   String  serviceName,
                                                   int     port,
                                                   boolean directConnect)
   {
      // used on the client-side of the appServer connection
      RemoteAppServer data = lookupRemoteAppServer(appService,
                                                   host,
                                                   serviceName,
                                                   port,
                                                   directConnect);
      InetSocketAddress address = null;
      if (data != null)
      {
         address = new InetSocketAddress(data.p2jHost, data.p2jPort);
      }
      
      return address;
   }
   
   /**
    * Search through the known remote appservers and determine the P2J account configured for this
    * appserver connection request.
    * 
    * @param    appService
    *           The appserver's service name (the -AppService connect parameter).
    * @param    host
    *           The -H connect parameter.
    * @param    serviceName
    *           The network service name, specified by the -S parameter.
    * @param    port
    *           The network port, specified by the -S parameter.
    * @param    directConnect
    *           Flag indicating this is a direct connect (the -DirectConnect parameter).
    *
    * @return   The P2J account to which the appservice belongs.
    */
   public static String lookupP2JAccount(String  appService,
                                         String  host,
                                         String  serviceName,
                                         int     port,
                                         boolean directConnect)
   {
      // used on the client-side of the appServer connection
      RemoteAppServer data = lookupRemoteAppServer(appService,
                                                   host,
                                                   serviceName,
                                                   port,
                                                   directConnect);

      return data == null ? null : data.p2jAccount;
   }
   
   /**
    * Search through the known remote appservers and determine the name of the appserver 
    * configured for this connection request.
    * 
    * @param    appService
    *           The appserver's service name (the -AppService connect parameter).
    * @param    host
    *           The -H connect parameter.
    * @param    serviceName
    *           The network service name, specified by the -S parameter.
    * @param    port
    *           The network port, specified by the -S parameter.
    * @param    directConnect
    *           Flag indicating this is a direct connect (the -DirectConnect parameter).
    *
    * @return   The name of the resolved appserver. 
    */
   public static String lookupTargetAppserver(String  appService,
                                              String  host,
                                              String  serviceName,
                                              int     port,
                                              boolean directConnect)
   {
      // used on the client-side of the appServer connection
      RemoteAppServer data = lookupRemoteAppServer(appService,
                                                   host,
                                                   serviceName,
                                                   port,
                                                   directConnect);

      return data == null ? null : data.name;
   }

   /**
    * Start this P2J client as an appserver agent or manager. The first established connection
    * will act as the manager and will start the initial agents.
    * 
    * @return   <code>true</code> if this was an appserver. On <code>false</code>, normal P2J
    *           conversation mode will be started.
    */
   public static boolean startAppServer()
   {
      // called on the server-side of the appServer connection

      SecurityManager sm = SecurityManager.getInstance();
      SessionManager sessMgr = SessionManager.get();
      Session session = sessMgr.getSession();
      // if this is not-null, we are on the client-side, on the client acting as the manager for
      // the appserver
      String appServer = sessMgr.getAppServer(session);
      
      if (appServer == null)
      {
         // here we are on the server-side, where the appserver is defined
         appServer = sm.getAppServer();
      }
      
      if (appServer == null)
      {
         // if no appServer defined for this account, then we are a normal client, so fast exit
         return false; 
      }
      
      String subject = SecurityManager.getInstance().getUserId();

      // if this is the first connection of this appServer, it will register as the manager
      // all subsequent connections authenticated as this appServer will act as agents
      AgentPool pool = null;
      synchronized (appServers)
      {
         pool = appServers.get(appServer);

         if (pool == null)
         {
            // initialize the pool, if this is the first agent to connect
            pool = AgentPool.newPool(appServer);
            appServers.put(appServer, pool);

            synchronized (subjectToAppserver)
            {
               subjectToAppserver.put(subject, appServer);
            }
         }
      }

      WorkArea wa = local.get();
      try
      {
         wa.appServer = appServer;
         
         pool.start();
      }
      finally
      {
         if (pool.isEmpty())
         {
            // the app server was terminated, clean up after it
            synchronized (appServers)
            {
               appServers.remove(pool.getName());
            }
            
            synchronized (subjectToAppserver)
            {
               subjectToAppserver.remove(subject);
            }
            
            AppServerLauncher.terminated(appServer);
         }

         wa.appServer = null;
      }

      return true;
   }
   
   /**
    * Set the current Agent's connection ID.
    * 
    * @param    id
    *           The connection ID.
    */
   static void setConnectionId(String id)
   {
      local.get().connectionId = id;
   }
   
   /**
    * Check if current running appserver knows the given procedure.
    * 
    * @param    pp
    *           A persistent procedure instance.
    *           
    * @return   <code>true</code> if the persistent procedure is known to the current running
    *           appserver.
    */
   static boolean hasProcedure(PersistentProcedure pp)
   {
      AgentPool pool = getAgentPoolBySubject(SecurityManager.getInstance().getUserId());
      return pool.hasProcedure(pp);
   }

   /**
    * Get the agent pool for the given appserver.
    * 
    * @param    appServer
    *           The appserver name.
    *           
    * @return   See above.
    */
   static AgentPool getAgentPoolForAppserver(String appServer)
   {
      synchronized (appServers)
      {
         return appServers.get(appServer);
      }
   }
   
   /**
    * Collect an {@link Agent} as is currently executing a task on the specified connection ID.
    * 
    * @param    connectionId
    *           The connection ID.
    * @param    agent
    *           The agent instance.
    */
   static void addActiveAgent(String connectionId, Agent agent)
   {
      if (connectionId != null)
      {
         synchronized (lock)
         {
            if (activeAgents.containsKey(connectionId))
            {
               activeAgents.get(connectionId).add(agent);
            }
         }
      }

      setConnectionId(connectionId);
   }
   
   /**
    * Remove the {@link Agent} as it finished executing a task on the specified connection ID.
    * 
    * @param    connectionId
    *           The connection ID.
    * @param    agent
    *           The agent instance.
    */
   static void removeActiveAgent(String connectionId, Agent agent)
   {
      if (connectionId != null)
      {
         synchronized (lock)
         {
            if (activeAgents.containsKey(connectionId))
            {
               activeAgents.get(connectionId).remove(agent);
            }
         }
      }
   }
   
   /**
    * Associate the specified thread with the specified request, in the {@link #requests} map.
    * 
    * @param    requestId
    *           The request ID.
    * @param    a
    *           The agent executing the request.
    * 
    * @return   <code>true</code> if a STOP was not already received for this request.
    */
   static boolean registerRequest(int requestId, Agent a)
   {
      synchronized (lock)
      {
         // if a STOP was received for this request, then we don't process it
         if (stoppedRequests.remove(requestId))
         {
            return false;
         }
         
         // else, register the agent thread for this request
         requests.put(requestId, a);
         
         return true;
      }
   }

   /**
    * Deregister the specified request from the {@link #requests} map.
    *  
    * @param    requestId
    *           The request ID.
    */
   static void deregisterRequest(int requestId)
   {
      synchronized (lock)
      {
         requests.remove(requestId);
      }
   }

   /**
    * Get the agent pool for the given connection ID.
    * 
    * @param    connectionId
    *           The connection ID.
    *           
    * @return   See above.
    */
   private static AgentPool getAgentPool(String connectionId)
   {
      SecurityManager sm = SecurityManager.getInstance();
      boolean allowed = false;
      synchronized (authorizedSessions)
      {
         Set<Integer> sessions = authorizedSessions.get(connectionId);
         allowed = (sessions != null && sessions.contains(sm.getSessionId()));
      }
      
      if (!allowed)
      {
         final String msg = 
            String.format("Unauthorized usage of connection ID %s from " +
                          "session ID %d by subject %s.",
                          connectionId,
                          sm.getSessionId(),
                          sm.getSessionToken().getSubject());
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg);
         }
         
         throw new RuntimeException(msg);
      }
      
      String appServer = AgentPool.getAppserver(connectionId);
      synchronized (appServers)
      {
         return appServers.get(appServer);
      }
   }

   /**
    * Check if the given connection ID is still connected.
    * 
    * @param    connectionId
    *           The connection ID.
    * 
    * @return   <code>true</code> if the connection is still alive.
    */
   private static boolean connected(String connectionId)
   {
      String appServer = AgentPool.getAppserver(connectionId);
      return appServer != null;
   }

   /**
    * Search through and determine the remote appserver associated with the given options.
    * 
    * @param    appService
    *           The appserver's service name (the -AppService connect parameter).
    * @param    host
    *           The -H connect parameter.
    * @param    serviceName
    *           The network service name, specified by the -S parameter.
    * @param    port
    *           The network port, specified by the -S parameter.
    * @param    directConnect
    *           Flag indicating this is a direct connect (the -DirectConnect parameter).
    *
    * @return   The resolved appserver configuration.
    */
   private static RemoteAppServer lookupRemoteAppServer(String  appService,
                                                        String  host,
                                                        String  serviceName,
                                                        int     port,
                                                        boolean directConnect)
   {
      initRemoteAppServers();
      
      // handle the defaults
      if (port == 0 && serviceName == null)
      {
         port = 5162;
      }
      if (host == null)
      {
         host = "localhost";
      }
      if (!directConnect && appService == null)
      {
         appService = nameServerDefaults.get(host);
      }
      
      RemoteAppServer appServer = null;
      for (RemoteAppServer data : remoteAppServers)
      {
         // check only host and serviceName/port
         if ((directConnect || data.aliases.contains(appService)) &&
             data.hosts.contains(host)                            &&
             (port == 0 || data.ports.contains(port))             &&
             (serviceName == null || data.services.contains(serviceName)))
         {
            appServer = data;
            break;
         }
      }
      
      return appServer;
   }
   
   /**
    * Init the data related to remote appservers, from the directory. It reads all the appservers
    * configured under the "server/default/app_services" node.
    */
   private static synchronized void initRemoteAppServers()
   {
      // do this only once
      if (nameServerDefaults != null)
      {
         return;
      }

      nameServerDefaults = new HashMap<String, String>();
      remoteAppServers = new ArrayList<RemoteAppServer>();
      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
         throw new RuntimeException("directory bind failed");

      String appSrvPath = Utils.findDirectoryNodePath(ds, "app_services", "container", false);
      String[] serviceNodes = ds.enumerateNodes(appSrvPath);
      for (int i = 0; i < serviceNodes.length; i++)
      {
         String name = serviceNodes[i];
         String path = appSrvPath + "/" + name;
         
         RemoteAppServer data = new RemoteAppServer(ds, appSrvPath, name);
         
         String[] defaults = ds.getNodeStrings(path + "/defaults", "values");
         if (defaults != null)
         {
            for (int j = 0; j < defaults.length; j++)
            {
               String ns = defaults[j];

               if (nameServerDefaults.containsKey(ns))
               {
                  throw new RuntimeException("More than one default service specified for " +
                                             "nameserver " + ns + "!");
               }
               
               nameServerDefaults.put(ns, name);
            }
         }
         
         remoteAppServers.add(data);
      }
      
      // last use of the directory, end our session with it
      ds.unbind();
   }
   
   /**
    * Get the agent pool for the given subject ID.
    * 
    * @param    subjectId
    *           The subject ID.
    *           
    * @return   See above.
    */
   private static AgentPool getAgentPoolBySubject(String subjectId)
   {
      String appServer = getAppserverBySubject(subjectId);
      synchronized (appServers)
      {
         return appServers.get(appServer);
      }
   }
   
   /**
    * Get the appserver name started by the given subject ID.
    * 
    * @param    subjectId
    *           The subject ID used to start an appserver.
    *           
    * @return   See above.
    */
   private static String getAppserverBySubject(String subjectId)
   {
      synchronized (subjectToAppserver)
      {
         return subjectToAppserver.get(subjectId);
      }
   }

   /**
    * Container for agent-related info.
    */
   private static class WorkArea
   {
      /**
       * When the P2J client is connected as an appserver, this indicates the appserver's name. 
       */
      private String appServer = null;
      
      /** The Agent's SERVER-CONNECTION-ID value. */
      private String connectionId = null;
   }

   /**
    * Container for the information needed to identify an appserver, by the data configured at the
    * CONNECT statement's options.
    */
   private static class RemoteAppServer
   {
      /** The name of this known appserver. */
      private final String name;
      
      /** The P2J host identifying this appserver. */
      private final String p2jHost;
      
      /** The P2J port (secure or insecure). */
      private final int p2jPort;
      
      /** The account (configured as an appserver) to authenticate on the remote side.*/
      private final String p2jAccount;
      
      /** The set of aliases defined for this appserver. */
      private final Set<String> aliases = new HashSet<String>();

      /** The list of name servers with which the appserver is registered. */
      private final Set<String> nameServers = new HashSet<String>();
      
      /** The hosts on which this appserver is running. */
      private final Set<String> hosts = new HashSet<String>();
      
      /**
       * The ports used by CONNECT statement to target this appserver.
       * May be nameserver or broker ports.
       */
      private final Set<Integer> ports = new HashSet<Integer>();
      
      /**
       * The network service names used by CONNECT statement to target this appserver.
       */
      private final Set<String> services = new HashSet<String>();

      /**
       * Initialize this remote appserver by reading its configuration from the directory.
       * 
       * @param    ds
       *           The {@link DirectoryService} instance.
       * @param    appSrvPath
       *           The directory path where remote appservers are configured.
       * @param    name
       *           The remote appserver's name.
       */
      public RemoteAppServer(DirectoryService ds, String appSrvPath, String name)
      {
         this.name = name;
         
         String path = appSrvPath + "/" + name;
         this.p2jHost = ds.getNodeString(path + "/p2j_host", "value");
         this.p2jPort = ds.getNodeInteger(path + "/p2j_port", "value");
         this.p2jAccount = ds.getNodeString(path + "/p2j_account", "value");

         String[] aliases = ds.getNodeStrings(path + "/aliases", "values");
         if (aliases != null)
         {
            this.aliases.addAll(Arrays.asList(aliases));
         }

         String[] nameServers = ds.getNodeStrings(path +"/nameservers", "values");
         if (nameServers != null)
         {
            this.nameServers.addAll(Arrays.asList(nameServers));
         }

         String[] hosts = ds.getNodeStrings(path +"/hosts", "values");
         if (hosts != null)
         {
            this.hosts.addAll(Arrays.asList(hosts));
         }

         Integer[] ports = ds.getNodeIntegers(path +"/ports", "values");
         if (ports != null)
         {
            this.ports.addAll(Arrays.asList(ports));
         }

         String[] services = ds.getNodeStrings(path +"/services", "values");
         if (services != null)
         {
            this.services.addAll(Arrays.asList(services));
         }
      }
   }   
}