/*
** Module   : AppServerOperatingMode.java
** Abstract : Define the available operating-modes for an appserver.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Defines the appserver's operating modes:
 * <ul>
 *    <li>{@link #STATE_AWARE} for the "State-aware" mode</li>
 *    <li>{@link #STATE_RESET} for the "State-reset" mode</li>
 *    <li>{@link #STATELESS} for the "Stateless" mode</li>
 *    <li>{@link #STATE_FREE} for the "State-free" mode</li>
 * </ul>
 *
 */
enum AppServerOperatingMode
{
   STATE_RESET(false) { public String toString() { return "State-reset"; } },
   STATE_AWARE(false) { public String toString() { return "State-aware"; } },
   STATELESS(false)   { public String toString() { return "Stateless"; } },
   STATE_FREE(true)   { public String toString() { return "State-free"; } };
   
   /**
    * Flag indicating this is a session-free operating mode.
    */
   private final boolean sessionFree;
   
   /**
    * Build a new value and set its {@link #sessionFree} flag.
    * 
    * @param    sessionFree
    *           <code>true</code> if this is a session-free operating mode.
    */
   private AppServerOperatingMode(boolean sessionFree)
   {
      this.sessionFree = sessionFree;
   }
   
   /**
    * Create the appropriate enum constant, from the given string. It must match the constant's
    * {@link #toString()} text.
    * 
    * @param    txt
    *           The string to search, case-insensitively.
    *           
    * @return   The found constant.
    * 
    * @throws   IllegalArgumentException
    *           If the passed text can't be matched with a valid operating mode.
    */
   public static AppServerOperatingMode fromString(String txt)
   {
      for (AppServerOperatingMode mode : AppServerOperatingMode.values())
      {
         if (mode.toString().equalsIgnoreCase(txt))
         {
            return mode;
         }
      }
      
      throw new IllegalArgumentException("Unknown operating mode: '" + txt + "'");
   }

   /**
    * Getter for the {@link #sessionFree} flag.
    * 
    * @return   See above.
    */
   public boolean isSessionFree()
   {
      return sessionFree;
   }
   
   /**
    * Check if this is a session-managed operating mode.
    * 
    * @return   <code>true</code> if this is not a session-free operating mode.
    */
   public boolean isSessionManaged()
   {
      return !sessionFree;
   }
}