/*
** Module   : CommonSession.java
** Abstract : Define all APIs available via the SESSION system handle.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CS  20130211 Moved first/last procedure attribute to a separate interface. 
** 003 CA  20130221 Added Sizeable support.
** 004 CA  20130222 Added more SESSION-related attributes and methods.
** 005 OM  20130607 Added SESSION:YEAR-OFFSET attribute support.
** 006 CA  20130927 Added LegacyResource annotation, to determine resource type. Implemented the 
**                  following attributes: PARAMETER, FIRST-QUERY, FIRST-/LAST-SERVER, 
**                  FIRST-/LAST-SOCKET, FIRST-/LAST-SERVER-SOCKET.
** 007 CA  20131004 WINDOW-SYSTEM is converted to getWindowSystem instead of getWindowsVersion.
** 008 EVL 20131008 Adding legacy attributes and methods annotation. Also adding the setters for
**                  the following attributes: FIRST-BUFFER, TIME-SOURCE, TIMEZONE, HANDLE,
**                  DISPLAY-TIMEZONE.
** 009 CA  20140701 Added PIXELS-PER-ROW and PIXELS-PER-COLUMN. 
** 010 OM  20141111 Refactored get/setDateOrder() to get/setDateFormat().
** 011 HC  20141207 Implemented runtime support for several SESSION attributes.
** 012 GES 20150227 Added FIRST-CHILD and LAST-CHILD attribute support.
** 013 CA  20150416 Moved FIRST/LAST-CHILD APIs to separate interface (CommonHandleTree).
** 014 CA  20151017 Added SESSION:DEBUG-ALERT, DATA-ENTRY-RETURN, APPL-ALERT-BOXES, LOAD-ICON.
** 015 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 016 HC  20170811 Added conversion support for SESSION attributes PRINTER-NAME, PRINTER-PORT,
**                  PRINTER-CONTROL-HANDLE, PRINTER-HDC and the method GET-PRINTERS.
**     HC  20170817 Added conversion support and runtime stubs for SESSION:STARTUP-PARAMETERS and
**                  SESSION:SUPPRESS-WARNINGS.
** 017 HC  20171003 Added new overload method setSuppressWarnings.
** 018 CA  20180503 Added MULTITASKING-INTERVAL, SYSTEM-ALERT-BOXES and SERVER-CONNECTION-ID.
**                  Renamed/moved some SESSION attribute getters/setters to fix issues with using
**                  the SESSION via a handle.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Exposes the following methods or attributes from the SYSTEM handle:
 * <table summary="">
 * <tr>
 *    <th>SESSION attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>BATCH-MODE</td>
 *    <td>{@link #isBatchMode()}</td>
 * </tr>
 * <tr>
 *    <td>CHARSET</td>
 *    <td>{@link #getCharset()}</td>
 * </tr>
 * <tr>
 *    <td>DATE-FORMAT</td>
 *    <td>{@link #getDateFormat()}, {@link #setDateFormat(character)} and 
 *        {@link #setDateFormat(String)}
 *    </td>
 * </tr>
 * <tr>
 *    <td>DISPLAY-TYPE</td>
 *    <td>{@link #getDisplayType()}</td>
 * </tr>
 * <tr>
 *    <td>HEIGHT-CHARS</td>
 *    <td>emitted as getHeightChars and setHeightChars, not implemented</td>
 * </tr>
 * <tr>
 *    <td>HEIGHT-PIXELS</td>
 *    <td>{@link #getHeightPixels()} and setHeightPixels</td>
 * </tr>
 * <tr>
 *    <td>TEMP-DIRECTORY</td>
 *    <td>{@link #getTempDirectory()}</td>
 * </tr>
 * <tr>
 *    <td>WIDTH-CHARS</td>
 *    <td>emitted as getWidthChars and setWidthChars, not implemented</td>
 * </tr>
 * <tr>
 *    <td>WIDTH-PIXELS</td>
 *    <td>{@link #getWidthPixels()} and setWidthPixels</td>
 * </tr>
 * <tr>
 *    <td>LAST-PROCEDURE</td>
 *    <td>{@link #lastProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>FIRST-PROCEDURE</td>
 *    <td>{@link #firstProcedure()}</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link com.goldencode.p2j.util.handle#unwrapSession()} API, which, depending on the
 * handle's state, it either returns an {@link SessionUtils#asHandle()}
 * instance or a dynamic proxy using {@link handle#invalidAttrAccessProxy}.
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public interface CommonSession
extends CommonProcedure,
        CommonHandle,
        CommonHandleTree,
        FirstLastProcedureAttribute,
        Sizeable
{
   /**
    * Get a the instance for the SESSION system handle.
    * 
    * @return   The handle to the object.
    */
   @LegacyAttribute(name = "HANDLE")
   public handle asHandle();
   
   /**
    * Reports if the current context is running in batch mode.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/batchMode
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/batchMode
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>false</code> will be returned.
    *
    * @return   <code>true</code> if the current context is in batch mode.
    */
   @LegacyAttribute(name = "BATCH-MODE")
   public logical isBatchMode();
   
   /**
    * Reports the default character set name. Since the Progress default
    * charset is ISO-8859-1, when the JVM reports that UTF-8 is the default,
    * this method will return the string "ISO-8859-1" instead.
    * <p>
    * The assumption (possibly not always correct) is that UTF-8 is not a
    * valid Progress charset and instead, to provide a proper default, we
    * must override with a compatible charset. Note that if the default
    * charset name is not UTF-8, it is assumed that the charset has been
    * overriden already (at JVM startup).  In that case, the default is
    * returned unchanged.
    *
    * @return   A default charset name.
    */
   @LegacyAttribute(name = "CHARSET")
   public character getCharset();
   
   /** 
    * Gets the order of date components for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components is a 3 character string using
    * "M", "D" and "Y" once each.  The order from index 0 to index 2
    * represents the left to right ordering of the components. So the leftmost
    * date component will be defined by the character at index position 0.
    * <p>
    * The value accessed is specific to the current user's context.
    *
    * @return   Defines the order of date components for conversion to/from strings.
    */
   @LegacyAttribute(name = "DATE-FORMAT")
   public character getDateFormat();
   
   /** 
    * Sets the order of date components for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using
    * "M", "D" and "Y" once each.  The order from index 0 to index 2
    * represents the left to right ordering of the components. So the leftmost
    * date component will be defined by the character at index position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   @LegacyAttribute(name = "DATE-FORMAT", setter = true)
   public void setDateFormat(String dateFormat);
   
   /** 
    * Sets the order of date components for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using
    * "M", "D" and "Y" once each.  The order from index 0 to index 2
    * represents the left to right ordering of the components. So the leftmost
    * date component will be defined by the character at index position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   @LegacyAttribute(name = "DATE-FORMAT", setter = true)
   public void setDateFormat(character dateFormat);
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute getter. Returns the 
    * currently configured start year for the two-digit Y2K windowing range.
    * This is the equivalent to the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @return  The start year for the two-digit windowing range.
    */
   @LegacyAttribute(name = "YEAR-OFFSET")
   public integer getYearOffset();
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute setter. Sets the 
    * configured start year for the two-digit Y2K windowing range. This is
    * the equivalent of the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @param   year
    *          The new start year for the two-digit windowing range.
    */
   @LegacyAttribute(name = "YEAR-OFFSET", setter = true)
   public void setYearOffset(NumberType year);
   
   /**
    * Gets the <code>DISPLAY-TIMEZONE</code> attribute for the user's context.
    * This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @return   The time zone offset, in minutes, used to display DATETIME-TZ data.
    */
   @LegacyAttribute(name = "DISPLAY-TIMEZONE")
   public integer getSessionDisplayTimeZone();

   /**
    * Gets the display type (GUI or TTY).  Corresponds with the
    * <code>SESSION:DISPLAY-TYPE</code> system handle attribute.
    * 
    * @return  Display type.  Currently, only TTY is supported.
    */
   @LegacyAttribute(name = "DISPLAY-TYPE")
   public character getDisplayType();

   /**
    * Get the width in chars of the display.  Corresponds with the
    * <code>SESSION:WIDTH-CHARS</code> system handle attribute.
    * 
    * @return  Display width in chars.
    */
   @LegacyAttribute(name = "WIDTH-CHARS")
   public decimal getWidthChars();
   
   /**
    * Setter for WIDTH-CHARS attribute on the SESSION handle.
    * 
    * @param   widthChars
    *          Attribute value.
    */
   @LegacyAttribute(name = "WIDTH-CHARS", setter = true, ignore = true)
   public void setWidthChars(NumberType widthChars);
   
   /**
    * Get the width in pixels of the display.  Corresponds with the
    * <code>SESSION:WIDTH-PIXELS</code> system handle attribute.
    * 
    * @return  Display width in pixels.
    */
   @LegacyAttribute(name = "WIDTH-PIXELS")
   public integer getWidthPixels();
   
   /**
    * Setter for WIDTH-PIXELS attribute on the SESSION handle.
    * 
    * @param   widthPixels
    *          Attribute value.
    */
   @LegacyAttribute(name = "WIDTH-PIXELS", setter = true, ignore = true)
   public void setWidthPixels(NumberType widthPixels);
   
   /**
    * Get the width in chars of the display.  Corresponds with the
    * <code>SESSION:HEIGHT-CHARS</code> system handle attribute.
    * 
    * @return  Display height in chars.
    */
   @LegacyAttribute(name = "HEIGHT-CHARS")
   public decimal getHeightChars();
   
   /**
    * Setter for HEIGHT-CHARS attribute on the SESSION handle.
    * 
    * @param   heightChars
    *          Attribute value.
    */
   @LegacyAttribute(name = "HEIGHT-CHARS", setter = true, ignore = true)
   public void setHeightChars(NumberType heightChars);
   
   /**
    * Get the height in pixels of the display.  Corresponds with the
    * <code>SESSION:HEIGHT-PIXELS</code> system handle attribute.
    * 
    * @return  Display height in pixels.
    */
   @LegacyAttribute(name = "HEIGHT-PIXELS")
   public integer getHeightPixels();
   
   /**
    * Setter for HEIGHT-PIXELS attribute on the SESSION handle.
    * 
    * @param   heightPixels
    *          Attribute value.
    */
   @LegacyAttribute(name = "HEIGHT-PIXELS", setter = true)
   public void setHeightPixels(NumberType heightPixels);
   
   /**
    * Gets the name of the directory where temporary files are stored.
    * Corresponds with with the <code>SESSION:TEMP-DIRECTORY</code> system
    * handle attribute.
    * 
    * @return  Name of temporary directory.
    */
   @LegacyMethod(name = "TEMP-DIRECTORY")
   public character getTempDirectory();
   
   /**
    * Gets the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @return   The offset in minutes for current timezone.
    */
   @LegacyAttribute(name = "TIMEZONE")
   public integer getSessionTimeZone();
   
   /**
    * Gets the <code>TIME-SOURCE</code> attribute for the user's context.
    * This attribute is can be read and written but it is not honored for date/time processing in
    * P2J.  The intention of the attribute is to define the system from which time values will be
    * obtained. In P2J, time values are always obtained from the application server.
    *
    * @return   The value of the attribute for current session.
    */
   @LegacyAttribute(name = "TIME-SOURCE")
   public character getSessionTimeSource();
   
   /**
    * Method that specifies the remote procedures the client can execute 
    * from the AppServer session
    * 
    * @param    attrList
    *           The list of procedure  and  name-patterns separated by comma to be
    *           exported.
    *         
    * @return   <code>true</code> if REMOTE attribute is true and operation succeeded or
    *           <code>false</code> otherwise
    */
   @LegacyMethod(name = "EXPORT")
   public logical export(String attrList);
   
   /**
    * Method that specifies the remote procedures the client can execute 
    * from the AppServer session
    * 
    * @param    attrList
    *           The list of procedure  and  name-patterns separated by comma to be
    *           exported.
    *         
    * @return   <code>true</code> if REMOTE attribute is true and operation succeeded or
    *           <code>false</code> otherwise
    */
   @LegacyMethod(name = "EXPORT")
   public logical export(character attrList);
   
   /**
    * Method implementation for getting the SERVER-OPERATING-MODE attribute of
    * the session handle.
    * <p>
    * Possible values:
    * <ul>
    *    <li>State-reset
    *    <li>State-aware
    *    <li>Stateless
    *    <li>State-free
    * </ul>
    * 
    * @return   the corresponding SERVER-OPERATING-MODE attribute of the session 
    *           handle.
    */
   @LegacyAttribute(name = "SERVER-OPERATING-MODE")
   public character getServerOperatingMode();
   
   /**
    * Indicates if the session is an AppServer session or not.
    * 
    * @return   <code>true</code> if the current session is an AppServer session
    *           or false otherwise.
    */
   @LegacyAttribute(name = "REMOTE")
   public logical isRemote();

   /**
    * Get the value of CLIENT-TYPE attribute, one of the following values:
    * <ul>
    *    <li>Standard ABL client 4GL: CLIENT</li>
    *    <li>WebClient: WEBCLIENT</li>
    *    <li>AppServer agent: APPSERVER</li>
    *    <li>WebSpeed agent: WEBSPEED</li>
    *    <li>Other special-purpose clients: Unknown value (?)</li>
    * </ul>
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "CLIENT-TYPE")
   public character getClientType();

   /**
    * Returns the handle for the first created dynamic query.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "FIRST-QUERY")
   public handle firstQuery();

   /**
    * Get the FIRST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "FIRST-SERVER")
   public handle firstServer();
   
   /**
    * Get the FIRST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "FIRST-SERVER-SOCKET")
   public handle firstServerSocket();
   
   /**
    * Get the FIRST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "FIRST-SOCKET")
   public handle firstSocket();

   /**
    * Returns the handle for the first dynamic buffer in the first table containing a dynamic 
    * buffer.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "FIRST-BUFFER")
   public handle firstBuffer();
   
   /**
    * Get the NUMERIC-SEPARATOR attribute, which is the character that represents a number's 
    * thousands separator, in its formatted representation.
    * 
    * @return   See above.
    */
   @LegacyMethod(name = "NUMERIC-SEPARATOR")
   public character getNumericSeparator();

   /**
    * Get the NUMERIC-DECIMAL-POINT attribute, which is the character that represents a number's
    * decimal point, in its formatted representation.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "NUMERIC-DECIMAL-POINT")
   public character getNumericDecimalPoint();

   /**
    * Get the NUMERIC-FORMAT attribute, which provides the way of interpreting commas and periods
    * within numeric values.
    * <p>
    * The possible values are "American", "European" or a character string consisting of the 
    * thousands separator followed by the decimal point. This attribute provides the same 
    * functionality as the European Numeric Format (-E) parameter.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "NUMERIC-FORMAT")
   public character getNumericFormat();

   /**
    * Set the NUMERIC-FORMAT attribute. See {@link #getNumericFormat()} for its representation.
    * 
    * @param    format
    *           The new value.
    */
   @LegacyAttribute(name = "NUMERIC-FORMAT", setter = true)
   public void setNumericFormat(String format);

   /**
    * Set the NUMERIC-FORMAT attribute. See {@link #getNumericFormat()} for its representation.
    * 
    * @param    format
    *           The new value.
    */
   @LegacyAttribute(name = "NUMERIC-FORMAT", setter = true)
   public void setNumericFormat(Text format);

   /**
    * Get the value of the PARAMETER attribute, which is set to the value of the Parameter 
    * (-param) startup parameter specified for the current session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PARAMETER")
   public character getParameter();

   /**
    * Get the value of the WINDOW-SYSTEM attribute, which is the version on which the client was 
    * started.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "WINDOW-SYSTEM")
   public character getWindowSystem();

   /**
    * Returns a string indicating the current wait-state; valid values are:
    * <ul>
    *    <li>GENERAL</li>
    *    <li>COMPILER
    *    <li>CUSTOM - if SET-WAIT-STATE was not called with <code>GENERAL</code> or 
    *        <code>COMPILER</code> constants.</li>
    *    <li>empty string, if none of the above.</li>
    * </ul>
    * 
    * @return   See above.
    */
   @LegacyMethod(name = "GET-WAIT-STATE")
   public character getWaitState();

   /**
    * Sets the NUMERIC-SEPARATOR and NUMERIC-DECIMAL-POINT attributes simultaneously.
    * 
    * @param    separator
    *           The new NUMERIC-SEPARATOR value.
    * @param    decPoint
    *           The new NUMERIC-DECIMAL-POINT value.
    *           
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-NUMERIC-FORMAT")
   public logical setNumericFormat(String separator, String decPoint);

   /**
    * Sets the NUMERIC-SEPARATOR and NUMERIC-DECIMAL-POINT attributes simultaneously.
    * 
    * @param    separator
    *           The new NUMERIC-SEPARATOR value.
    * @param    decPoint
    *           The new NUMERIC-DECIMAL-POINT value.
    *           
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-NUMERIC-FORMAT")
   public logical setNumericFormat(character separator, String decPoint);

   /**
    * Sets the NUMERIC-SEPARATOR and NUMERIC-DECIMAL-POINT attributes simultaneously.
    * 
    * @param    separator
    *           The new NUMERIC-SEPARATOR value.
    * @param    decPoint
    *           The new NUMERIC-DECIMAL-POINT value.
    *           
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-NUMERIC-FORMAT")
   public logical setNumericFormat(String separator, character decPoint);

   /**
    * Sets the NUMERIC-SEPARATOR and NUMERIC-DECIMAL-POINT attributes simultaneously.
    * 
    * @param    separator
    *           The new NUMERIC-SEPARATOR value.
    * @param    decPoint
    *           The new NUMERIC-DECIMAL-POINT value.
    *           
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-NUMERIC-FORMAT")
   public logical setNumericFormat(character separator, character decPoint);

   /**
    * The value of state-string determines the wait state and the type of wait cursor displayed. 
    * The valid values are: 
    * <ul>
    *    <li>GENERAL - Displays the system-busy cursor</li>
    *    <li>COMPILER - Displays the ABL compiler cursor</li>
    * </ul>
    * 
    * @param    state
    *           The new wait-state value.
    * 
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-WAIT-STATE")
   public logical setWaitState(String state);

   /**
    * The value of state-string determines the wait state and the type of wait cursor displayed. 
    * The valid values are: 
    * <ul>
    *    <li>GENERAL - Displays the system-busy cursor</li>
    *    <li>COMPILER - Displays the ABL compiler cursor</li>
    * </ul>
    * 
    * @param    state
    *           The new wait-state value.
    * 
    * @return   <code>true</code> if the operation was successful.
    */
   @LegacyMethod(name = "SET-WAIT-STATE")
   public logical setWaitState(character state);
   
   /**
    * Get the LAST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "LAST-SERVER")
   public handle lastServer();

   /**
    * Get the LAST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "LAST-SERVER-SOCKET")
   public handle lastServerSocket();

   /**
    * Get the LAST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "LAST-SOCKET")
   public handle lastSocket();
   
   /**
    * Get the PIXELS-PER-ROW attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PIXELS-PER-ROW")
   public integer getPixelsPerRow();
   
   /**
    * Get the PIXELS-PER-COLUMN attribute of this session.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PIXELS-PER-COLUMN")
   public integer getPixelsPerColumn();
   
   /**
    * WORK-AREA-X attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   @LegacyAttribute(name = "WORK-AREA-X")
   public integer getWorkAreaX();
   
   /**
    * Setter for WORK-AREA-X attribute on the SESSION handle.
    * 
    * @param   x
    *          Attribute value.
    */
   @LegacyAttribute(name = "WORK-AREA-X", setter = true)
   public void setWorkAreaX(NumberType x);

   /**
    * WORK-AREA-Y attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   @LegacyAttribute(name = "WORK-AREA-Y")
   public integer getWorkAreaY();
   
   /**
    * Setter for WORK-AREA-Y attribute on the SESSION handle.
    * 
    * @param   y
    *          Attribute value.
    */
   @LegacyAttribute(name = "WORK-AREA-Y", setter = true)
   public void setWorkAreaY(NumberType y);
   
   /**
    * WORK-AREA-WIDTH-PIXELS attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   @LegacyAttribute(name = "WORK-AREA-WIDTH-PIXELS")
   public integer getWorkAreaWidthPixels();
   
   /**
    * Setter for WORK-AREA-WIDTH-PIXELS attribute on the SESSION handle.
    * 
    * @param   widthPixels
    *          Attribute value.
    */
   @LegacyAttribute(name = "WORK-AREA-WIDTH-PIXELS", setter = true)
   public void setWorkAreaWidthPixels(NumberType widthPixels);
   
   /**
    * WORK-AREA-WIDTH-PIXELS attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   @LegacyAttribute(name = "WORK-AREA-HEIGHT-PIXELS")
   public integer getWorkAreaHeightPixels();
   
   /**
    * Setter for WORK-AREA-WIDTH-PIXELS attribute on the SESSION handle.
    * 
    * @param   heightPixels
    *          Attribute value.
    */
   @LegacyAttribute(name = "WORK-AREA-HEIGHT-PIXELS", setter = true)
   public void setWorkAreaHeightPixels(NumberType heightPixels);
   
   /**
    * Get the state of SESSION:DEBUG-ALERT attribute. 
    * 
    * @return  the state of SESSION:DEBUG-ALERT attribute.
    */
   @LegacyAttribute(name = "DEBUG-ALERT")
   public logical isDebugAlert();
   
   /**
    * Set the state of SESSION:DEBUG-ALERT attribute.
    * 
    * @param   on
    *          The new state for SESSION:DEBUG-ALERT attribute.
    */
   @LegacyAttribute(name = "DEBUG-ALERT", setter = true)
   public void setDebugAlert(logical on);
   
   /**
    * Get the state of SESSION:DATA-ENTRY-RETURN attribute. 
    * 
    * @return  the state of SESSION:DATA-ENTRY-RETURN attribute.
    */
   @LegacyAttribute(name = "DATA-ENTRY-RETURN")
   public logical isDataEntryReturn();
   
   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   @LegacyAttribute(name = "DATA-ENTRY-RETURN", setter = true)
   public void setDataEntryReturn(logical on);
   
   /**
    * Get the state of SESSION:APPL-ALERT-BOXES attribute. 
    *
    * @return  the state of SESSION:APPL-ALERT-BOXES attribute.
    */
   @LegacyAttribute(name = "APPL-ALERT-BOXES")
   public logical isUsingAlertBoxes();
   
   /**
    * Set the state of SESSION:APPL-ALERT-BOXES attribute.
    *
    * @param   on
    *          The new state for SESSION:APPL-ALERT-BOXES attribute.
    */
   @LegacyAttribute(name = "APPL-ALERT-BOXES", setter = true)
   public void setUsesAlertBoxes(logical on);
   
   /**
    * Set the state of SESSION:LOAD-ICON attribute.
    *
    * @param   iconName
    *          The file name of the icon to be loaded
    */
   @LegacyMethod(name = "LOAD-ICON")
   public logical loadIcon(String iconName);

   /**
    * Get the state of SESSION:PRINTER-NAME attribute.
    *
    * @return  the state of SESSION:PRINTER-NAME attribute.
    */
   @LegacyAttribute(name = "PRINTER-NAME")
   public character getPrinterName();

   /**
    * Set the state of SESSION:PRINTER-NAME attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-NAME attribute.
    */
   @LegacyAttribute(name = "PRINTER-NAME", setter = true)
   public void setPrinterName(character value);

   /**
    * Get the state of SESSION:PRINTER-PORT attribute.
    *
    * @return  the state of SESSION:PRINTER-PORT attribute.
    */
   @LegacyAttribute(name = "PRINTER-PORT")
   public character getPrinterPort();

   /**
    * Set the state of SESSION:PRINTER-PORT attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-PORT attribute.
    */
   @LegacyAttribute(name = "PRINTER-PORT", setter = true)
   public void setPrinterPort(character value);

   /**
    * Get the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    *
    * @return  the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    */
   @LegacyAttribute(name = "PRINTER-CONTROL-HANLDE")
   public integer getPrinterControlHandle();

   /**
    * Set the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-CONTROL-HANLDE attribute.
    */
   @LegacyAttribute(name = "PRINTER-CONTROL-HANLDE", setter = true)
   public void setPrinterControlHandle(integer value);

   /**
    * Get the state of SESSION:PRINTER-HDC attribute.
    *
    * @return  the state of SESSION:PRINTER-HDC attribute.
    */
   @LegacyAttribute(name = "PRINTER-HDC")
   public integer getPrinterHDC();

   /**
    * The SESSION:GET-PRINTERS method.
    *
    * @return  the result of SESSION:GET-PRINTERS method.
    */
   @LegacyMethod(name = "GET-PRINTERS")
   public character getPrinters();

   /**
    * Get the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @return  the state of SESSION:SUPPRESS-WARNINGS attribute.
    */
   @LegacyAttribute(name = "SUPPRESS-WARNINGS")
   public logical isSuppressWarnings();

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   @LegacyAttribute(name = "SUPPRESS-WARNINGS", setter = true)
   public void setSuppressWarnings(logical value);

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   @LegacyAttribute(name = "SUPPRESS-WARNINGS", setter = true)
   public void setSuppressWarnings(boolean value);

   /**
    * Get the state of SESSION:STARTUP-PARAMETERS attribute.
    *
    * @return  the state of SESSION:STARTUP-PARAMETERS attribute.
    */
   @LegacyAttribute(name = "STARTUP-PARAMETERS")
   public character getStartupParameters();
   
   /**
    * Sets the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @param    offset
    *           The new timezone offset which must be between -840 and 840 inclusive. 
    *           The value may also be <code>unknown</code> which is the equivalent of resetting
    *           this attribute to the default.
    */
   @LegacyAttribute(name = "TIMEZONE", setter = true)
   public void setSessionTimeZone(NumberType offset);
   
   /**
    * Sets the <code>TIME-SOURCE</code> attribute for the user's context.
    * This attribute is can be read and written but it is not honored for date/time processing in
    * P2J.  The intention of the attribute is to define the system from which time values will be
    * obtained. In P2J, time values are always obtained from the application server.
    *
    * @param    timeSource
    *           The new value to set. If <code>null</code>, then the empty string will be used,
    *           which in the 4GL has the same affect as "local".
    */
   @LegacyAttribute(name = "TIME-SOURCE", setter = true)
   public void setSessionTimeSource(character timeSource);

   /**
    * Sets the <code>DISPLAY-TIMEZONE</code> attribute for the user's context.
    * This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @param    displayTimezone
    *           The new value to set.
    */
   @LegacyAttribute(name = "DISPLAY-TIMEZONE", setter = true)
   public void setSessionDisplayTimeZone(NumberType displayTimezone);
   
   /**
    * Get the state of SESSION:IMMEDIATE-DISPLAY attribute. 
    * 
    * @return  the state of SESSION:IMMEDIATE-DISPLAY attribute.
    */
   @LegacyAttribute(name = "IMMEDIATE-DISPLAY")
   public logical isImmediateDisplay();

   /**
    * Set the state of SESSION:IMMEDIATE-DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:IMMEDIATE-DISPLAY attribute.
    */
   @LegacyAttribute(name = "IMMEDIATE-DISPLAY", setter = true)
   public void setImmediateDisplay(logical on);

   /**
    * Set the state of SESSION:IMMEDIATE-DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:IMMEDIATE-DISPLAY attribute.
    */
   @LegacyAttribute(name = "IMMEDIATE-DISPLAY", setter = true)
   public void setImmediateDisplay(boolean on);

   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   @LegacyAttribute(name = "SYSTEM-ALERT-BOXES", setter = true)
   public void setSystemAlertBoxes(boolean on);
   
   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   @LegacyAttribute(name = "SYSTEM-ALERT-BOXES", setter = true)
   public void setSystemAlertBoxes(logical on);
   
   /**
    * Get the state of SESSION:SYSTEM-ALERT-BOXES attribute. 
    * 
    * @return  the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   @LegacyAttribute(name = "SYSTEM-ALERT-BOXES")
   public logical isSystemAlertBoxes();
   
   /**
    * Set the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @param   interval 
    *          The new value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   @LegacyAttribute(name = "MULTITASKING-INTERVAL", setter = true)
   public void setMultitaskingInterval(NumberType interval);
   
   /**
    * Get the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @return  the value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   @LegacyAttribute(name = "MULTITASKING-INTERVAL")
   public integer getMultitaskingInterval();
   
   /**
    * Get the value of SESSION:SERVER-CONNECTION-ID attribute. 
    * 
    * @return  the value of SESSION:SERVER-CONNECTION-ID attribute.
    */
   @LegacyAttribute(name = "SERVER-CONNECTION-ID")
   public character getServerConnectionId();
}