/*
** Module   : ControlFrameComObject.java
** Abstract : An interface for the server-side component that describe the interface of a 
**            Control-frame COM object.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit. Exposed methods and attributes: LoadControls, CONTROL-NAME,
**                  LEFT, TOP, CONTROLS, HEIGHT, WIDTH and WIDGET-HANDLE.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * This is the interface for the server-side component that describe the interface of a 
 * Control-frame COM object. Objects of this type can only be obtained using 
 * {@link ControlFrame#getComHandle()}. 
 */
public interface ControlFrameComObject
extends Nameable
{
   /**
    * Loads the named OCX into the associated control frame.
    *
    * @param   ocx
    *          Name of the OCX to load.
    * @param   ctrl
    *          Control frame name.
    *
    * @return  {@code yes} if operation was successful.
    */
   @LegacyMethod(name = "LOADCONTROLS")
   public logical loadControls(character ocx, character ctrl);
   
   /**
    * Loads the named OCX into the associated control frame.
    *
    * @param   ocx
    *          Name of the OCX to load.
    * @param   ctrl
    *          Control frame name.
    *
    * @return  {@code yes} if operation was successful.
    */
   @LegacyMethod(name = "LOADCONTROLS")
   public logical loadControls(String ocx, character ctrl);
   
   /**
    * Loads the named OCX into the associated control frame.
    *
    * @param   ocx
    *          Name of the OCX to load.
    * @param   ctrl
    *          Control frame name. 
    *
    * @return  {@code yes} if operation was successful.
    */
   @LegacyMethod(name = "LOADCONTROLS")
   public logical loadControls(character ocx, String ctrl);
   
   /**
    * Loads the named OCX into the associated control frame.
    *
    * @param   ocx
    *          Name of the OCX to load.
    * @param   ctrl
    *          Control frame name. 
    *
    * @return  {@code yes} if operation was successful.
    */
   @LegacyMethod(name = "LOADCONTROLS")
   public logical loadControls(String ocx, String ctrl);
   
   /**
    * The component handle to an ActiveX control that has the specified design-time name 
    * (Control-Name) and that is loaded into the control-frame.
    *
    * @return  references to COM object properties and methods extend the syntax used for
    *          referencing widget attributes and methods.
    */
   @LegacyAttribute(name = "CONTROL-NAME")
   public comhandle getControlName();
   
   /**
    * Returns a control collection (with only one control in OE 10.2B) from this control frame.
    *
    * @return  The collection of controls from this control frame.
    */
   @LegacyAttribute(name = "CONTROLS")
   public comhandle getControls();
   
   /**
    * Returns the handle of the control frame associated with the control-frame COM object.
    *
    * @return  The handle of the control frame associated with the control-frame COM object.
    */
   @LegacyAttribute(name = "WIDGET-HANDLE")
   public handle getWidgetHandle();

   /**
    * Obtain the width of the widget or the screen display used in the current session, in pixels.
    *
    * @return  The width of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "WIDTH")
   public integer getWidth();
   
   /**
    * Set the width of the widget or the screen display used in the current session, in pixels.
    *
    * @param   width
    *          The new width of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "WIDTH", setter = true)
   public void setWidth(long width);
   
   /**
    * Set the width of the widget or the screen display used in the current session, in pixels.
    *
    * @param   width
    *          The new width of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "WIDTH", setter = true)
   public void setWidth(NumberType width);
   
   /**
    * Obtain the height of the widget or the screen display used in the current session,
    * in pixels.
    *
    * @return  The height of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "HEIGHT")
   public integer getHeight();
   
   /**
    * Sets the new height of the widget or the screen display used in the current session,
    * in pixels.
    *
    * @param   height
    *          The new height of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "HEIGHT", setter = true)
   public void setHeight(long height);
   
   /**
    * Sets the new height of the widget or the screen display used in the current session,
    * in pixels.
    *
    * @param   height
    *          The new height of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "HEIGHT", setter = true)
   public void setHeight(NumberType height);
   
   /**
    * Obtain the left of the widget or the screen display used in the current session, in pixels.
    *
    * @return  The left of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "LEFT")
   public integer getLeft();
   
   /**
    * Set the left of the widget or the screen display used in the current session, in pixels.
    *
    * @param   left
    *          The new left of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "LEFT", setter = true)
   public void setLeft(long left);
   
   /**
    * Set the left of the widget or the screen display used in the current session, in pixels.
    *
    * @param   left
    *          The new left of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "LEFT", setter = true)
   public void setLeft(NumberType left);
   
   /**
    * Obtain the top of the widget or the screen display used in the current session, in pixels.
    *
    * @return  the top of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "TOP")
   public integer getTop();
   
   /**
    * Sets the new top of the widget or the screen display used in the current session, in pixels.
    *
    * @param   top
    *          The new top of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "TOP", setter = true)
   public void setTop(long top);
   
   /**
    * Sets the new top of the widget or the screen display used in the current session, in pixels.
    *
    * @param   top
    *          The new top of the widget or the screen display used in the current session.
    */
   @LegacyAttribute(name = "TOP", setter = true)
   public void setTop(NumberType top);
   
   /**
    * Gets the WIDGET-ID writable attribute.
    *
    * @return
    *          the widget id
    */
   @LegacyAttribute(name = "WIDGET-ID")
   public integer getWidgetId();
   
   /**
    * Gets the FRAME attribute, which is the containing frame instance, as a handle.
    *
    * @return   The frame that contains this widget.
    */
   @LegacyAttribute(name = "FRAME")
   public handle getFrameHandle();
   
   /**
    * Obtain a (com-)handle to a specific Java emulated ActiveX, if such instance exist. This is
    * an internal method, not exposed by ABL interface. 
    *
    * @param   axName
    *          The ActiveX name.
    *
    * @return  a {@code comhandle} to the requested activex object if such instance was loaded. 
    */
   public comhandle getActiveX(String axName);
}

