/*
** Module   : DatabaseTrigger.java
** Abstract : Abstract class to be extended by session database triggers.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 SVL 20130213 Created initial version.
** 002 OM  20131107 Added declaration for assign trigger methods and implementation of trigger() 
**                  method. Added replication support. Stubs for ON ASSIGN NEW VALUE form.
** 003 GES 20160907 Changes to match BlockManager API change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;

/**
 * Abstract class to be extended by a session database trigger implementation. The trigger class
 * should implement a single function (<code>create</code>, <code>delete</code>, <code>find</code>
 * or some form of  <code>write</code>) which corresponds the database event type handled by the
 * trigger. Default implementations are no-ops.
 *
 * @param <Buf>
 *        The class of the buffer to which the database event served by this trigger is applied.
 */
public abstract class DatabaseTrigger<Buf extends Buffer>
{
   /**
    * Should be implemented by an ASSIGN trigger.
    *
    * @param   buf
    *          Buffer which contains the record that triggered the event.
    * @param   oldValueCopy
    *          A copy of the old value that was changed by assign statement. Even if this value
    *          can be overwritten, the change will be lost when exiting the trigger body block.
    */
   public void assign(Buf buf, BaseDataType oldValueCopy)
   {
   }

   /**
    * Should be implemented by an ASSIGN trigger. This overloaded method does not pass in the
    * old value.
    *
    * @param   buf
    *          Buffer which contains the record that triggered the event.
    */
   public void assign(Buf buf)
   {
   }

   /**
    * Should be implemented by a SCHEMA ASSIGN trigger with NEW form (no field reference).
    *
    * @param   newVal
    *          The value that triggered the event.
    * @param   oldValueCopy
    *          A copy of the old value that was changed by assign statement. Even if this value
    *          can be overwritten, the change will be lost when exiting the trigger body block.
    */
   public void assign(Undoable newVal, BaseDataType oldValueCopy)
   {
   }

   /**
    * Should be implemented by a SCHEMA ASSIGN trigger with NEW form (no field reference).
    * This overloaded method does not pass in the old value.
    *
    * @param   newVal
    *          The value that triggered the event.
    */
   public void assign(Undoable newVal)
   {
   }

   /**
    * Should be implemented by a CREATE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void create(Buf buf)
   {
   }

   /**
    * Should be implemented by a FIND trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void find(Buf buf)
   {
   }

   /**
    * Should be implemented by a DELETE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void delete(Buf buf)
   {
   }

   /**
    * Should be implemented by a WRITE trigger without OLD buffer specified.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void write(Buf buf)
   {
   }

   /**
    * Should be implemented by a WRITE trigger with OLD buffer specified.
    *
    * @param   newBuf
    *          Buffer which contains the new value of the record which triggered the event.
    * @param   oldBuf
    *          Buffer which contains the old value of the record which triggered the event.
    */
   public void write(Buf newBuf, Buf oldBuf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-CREATE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationCreate(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-DELETE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationDelete(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-WRITE trigger without OLD buffer specified.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationWrite(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-WRITE trigger with OLD buffer specified.
    *
    * @param   newBuf
    *          Buffer which contains the new value of the record which triggered the event.
    * @param   oldBuf
    *          Buffer which contains the old value of the record which triggered the event.
    */
   public void replicationWrite(Buf newBuf, Buf oldBuf)
   {
   }

   /**
    * Execute the given trigger block.
    *
    * @param   blk
    *          The code for the trigger block body.
    */
   protected void trigger(Block blk)
   {
      // TODO: check if TransactionType.FULL is not the appropriate type
      BlockManager.topLevelBlock(null, BlockManager.TransactionType.SUB, BlockType.TRIGGER, blk);
   }
}