/*
** Module   : EnvironmentDaemon.java
** Abstract : Client side daemon to handle Progress 4GL compatible Winodws environment handling
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130604 Created initial version.
** 002 EVL 20140327 Changing constructor parameter list to be able to request the server for
**                  properties.
** 003 CA  20140702 Added isCustomEnvironment and getEnvironmentName APIs.
** 004 MAG 20140908 Added getEnvironmentType APIs. Added return value to load() method.
**                  Change the unload() implementation. Implements read only environments loading.
**                  Added setDefaultEnvironment() which override default environment.              
** 005 HC  20151124 Improved loading of boolean environment variables from Progress INI files and
**                  Windows Registry.
**     EVL 20151130 Fix for image clipping regression.  The server side received wrong pixels per
**                  row value because none started value was not replaced with default one.
** 006 CA  20180523 Fixed environment name, when used in maps: registries are case-insensitive,
**                  stanza INI are case-sensitive. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A client side class to manipulate with environments. The environment can be the INI file or
 * registry sections containing the keys.
 *
 * @author    EVL
 */
public class EnvironmentDaemon
implements EnvironmentAccessor
{
   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;
   
   /** The default environment to use when load() and use() was not specified. */
   private EnvironmentReader defaultEnv = null;
   
   /** The currently loaded environment to use instead of default. */
   private EnvironmentReader currentEnv = null;
   
   /** Map to store registry environments for possible reuse. */
   private Map<String, EnvironmentReader> envRegMap = new HashMap<String, EnvironmentReader>();
   
   /** Map to store INI environments for possible reuse. */
   private Map<String, EnvironmentReader> envIniMap = new HashMap<String, EnvironmentReader>();

   /** Client type */
   private boolean isGui = false;
   
   /** INI files read only property */
   private boolean isReadOnly = false;

   /**
    * Create an instance and export its API to the network.
    *
    * @param   session
    *          The server session to get the properties.
    * @param   single
    *          <code>true</code> to startup within the server process which must bypass the
    *          shared infrastructure initialization. Use <code>false</code> for the normal
    *          client JVM startup.
    * @param   isWin
    *          <code>true</code> if this is an WIN32 environment backed up by a default. 
    */
   public EnvironmentDaemon(Session session, boolean single, boolean isWin)
   {
      // get driver type and read only property.
      ThinClient tc = ThinClient.getInstance();
      isGui = !tc.isChui();      
      
      // get environment INI files read only flag
      isReadOnly = getReadOnly();   

      defaultEnv = (isWin) ? new Registry(session) : null;
            
      synchronized (EnvironmentDaemon.class)
      {
         modToken = RemoteObject.registerServer(EnvironmentAccessor.class,
                                                this,
                                                modToken,
                                                single);
      }
   }
      
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key, initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be created overwriting the
    *        possibly existed one.
    * @param baseKey
    *        When the parameter is specified the searching is performed under the particular key
    *        value. It is possible to specify "INI" as parameter. In this case search or creation
    *        will be done inside initialization file specified in env parameter.
    *        
    * @return  <code>true</code> if loaded false otherwise      
    */
   public boolean load(String env, String directory, boolean flagNew, String baseKey)
   {      
      EnvironmentReader envAcc = null;
      Map<String, EnvironmentReader> envs = null;
      
      // GUI clients are using INI files always.
      if (EnvironmentReader.TYPE_INI.equalsIgnoreCase(baseKey) || isGui)
      {
         envs = envIniMap;
         envAcc = new StanzaIni(env, directory, flagNew, isReadOnly);
      }
      else
      {
         envs = envRegMap;
         envAcc = new Registry(env, flagNew, baseKey);
      }
      
      boolean loaded = envAcc.load();
      
      if (loaded)
      {
         envs.put(envAcc.getEnvironmentName(), envAcc);
      }
      
      return loaded;
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param    env 
    *           The environment to load. If the parameter is empty string ("") the default
    *           environment becomes the current one.
    *           
    * @return   The loaded environment's name, as it was saved on LOAD.  <code>null</code> if the
    *           environment can't be found.
    */
   public String use(String env)
   {
      // Empty string as environment makes the default environment the current one
      if (env == null || env.isEmpty())
      {
         currentEnv = null;
         env = null;
      }
      else
      {
         // registry first
         currentEnv = envRegMap.get(env.toLowerCase());
         if (currentEnv != null)
         {
            return currentEnv.getEnvironmentName();
         }
         
         // stanza second
         if (env.indexOf('.') < 0)
         {
            env = env + ".ini";
         }
         currentEnv = envIniMap.get(env);
         if (currentEnv != null)
         {
            return currentEnv.getEnvironmentName();
         }
         
         // not found
         env = null;
      }
      
      return env;
   }
   
   /**
    * Check if a custom environment is in use.
    * 
    * @return   See above.
    */
   public boolean isCustomEnvironment()
   {
      return currentEnv != null;
   }
   
   /**
    * Get the current environment's name.
    * 
    * @return   See above.
    */
   public String getEnvironmentName()
   {
      EnvironmentReader env = (currentEnv == null ? defaultEnv : currentEnv);
      
      return (env == null ? null : env.getEnvironmentName());
   }

   /**
    * Get the current environment's type.
    * 
    * @return  environment type.
    */
   public String getEnvironmentType()
   {
      EnvironmentReader env = (currentEnv == null ? defaultEnv : currentEnv);
      
      return (env == null ? null : env.getEnvironmentType());
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public void setKeyValue(String section, String key, String value)
   {
      if (section == null || value == null)
      {
         return;
      }
      if (currentEnv != null)
      {
         currentEnv.setKeyValue(section, key, value);
      }
      else if (defaultEnv != null)
      {
         defaultEnv.setKeyValue(section, key, value);
      }
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public String getKeyValue(String section, String key)
   {
      if (section == null)
      {
         return new String("");
      }
      if (currentEnv != null)
      {
         return currentEnv.getKeyValue(section, key);
      }
      else if (defaultEnv != null)
      {
         return defaultEnv.getKeyValue(section, key);
      }
      return "";
   }
   
   /**
    * Getting the boolean key value from the current environment.
    * 
    * @param    section
    *           The name of the section containing the key to get.
    * @param    key
    *           The name of the key key to get or default key if not specified.
    * @param    defValue
    *           The default value, to be returned when the value does not exist
    *           in the environment.
    * 
    * @return   The boolean key value, <code>defValue</code> if the value
    *           does not exist in the environment.
    */
   public Boolean getBooleanKeyValue(String section, String key, Boolean defValue)
   {
      String strValue = getKeyValue(section, key);
      // if no value for the key - return default one
      if (strValue == null || strValue.isEmpty())
      {
         return defValue;
      }
      
      // the only valid boolean true for Progress INI file is the case insensitive 'yes' 
      // TODO: is this compatible with the allowed boolean values in Windows registry?
      return strValue.trim().equalsIgnoreCase("yes"); 
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param    env 
    *           The environment to unload. Can be the registry or initialization file. The
    *           environment to unload should be previously loaded by load() method.
    *           
    * @return   The loaded environment's name, as it was saved on LOAD.  <code>null</code> if the
    *           environment can't be found.
    */
   public String unload(String env)
   {
      if (env != null)
      {
         Map<String, EnvironmentReader> envs = null;
         
         if (envRegMap.containsKey(env.toLowerCase()))
         {
            env = env.toLowerCase();
            envs = envRegMap;
         }
         else
         {
            if (env.indexOf('.') < 0)
            {
               env = env + ".ini";
            }
            
            if (envIniMap.containsKey(env))
            {
               envs = envIniMap;
            }
         }

         for (TitledWindow<?> window : WindowManager.windowList())
         {
            WindowConfig cfg = (WindowConfig) window.config();
            if (cfg != null && env.equals(cfg.envName))
            {
               return null;
            }
         }

         EnvironmentReader envReader = envs == null ? null : envs.get(env);
         if (envReader != null)
         {
            // unload and remove
            envReader.unload();
            envs.remove(env);
   
            // set current environment to default if unloaded environment was current environment
            if (env.equals(currentEnv.getEnvironmentName()))
            {
               currentEnv = null;
            }

            // notify about unload
            ColorManager.unload(env);
            FontManager.unload(env);
            
            return envReader.getEnvironmentName();
         }
      }
      
      return null;
   }

   /**
    * Sets the default environment to a stanza ini file, overwriting any default registry for
    * WIN32.
    * 
    * @param   iniFile
    *          The name of the ini file backing the stanzas.
    */
   public void setDefaultEnvironment(String iniFile)
   {
      File file = new File(iniFile);

      // check if file exists when the file is not loaded from server.
      if (!isReadOnly && !file.exists()) 
      {
         // client aborts if environment file is not found
         throw new RuntimeException(iniFile + " file not found. (5643)");
      }
      
      // directory
      String dir = file.getParent();
      // file name
      String fileName = file.getName();
      
      // load the file and store it as the default environment 
      EnvironmentReader reader = new StanzaIni(fileName, dir, false, isReadOnly);
      if (reader.load())
      {
         defaultEnv = reader;
      }
      else
      {
         // client aborts if environment file is not loaded
         throw new RuntimeException(iniFile + " file not found. (5643)");         
      }
   }   
   
   /**
    * Get environments read only flag.  When read only flag is set to <code>true</code> the 
    * environments INI files are loaded as resources from remote server jar files otherwise
    * will follow default behavior and INI files are loaded from client file system.  
    * 
    * @return  Read only flag (see above).
    */
   private boolean getReadOnly()
   {
      boolean result = false;
      
      Directory remoteDir = DirectoryManager.getInstance();
      if (remoteDir != null)
      {
         result = remoteDir.getBoolean(Directory.ID_RELATIVE, "readOnlyEnvironments", false);
      }      
      
      return result;
   }
}