/*
** Module   : EnvironmentOps.java
** Abstract : Progress 4GL compatible environment related methods
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050628   @21564 Created initial version, supporting common
**                           global variables that are called built-in
**                           functions in Progress but which take no
**                           parameters, no empty parens. These also
**                           are read-only (they cannot be assigned).
** 002 GES 20050729   @21921 Added getSourceName() and modified the
**                           default propath.
** 003 ECF 20051214   @23753 Removed getCurrentDatabaseName(). This was a
**                           placeholder for P2J's DBNAME replacement
**                           implementation. This is now provided in the
**                           ConnectionManager class instead.
** 004 GES 20060123   @24034 Move to a directory based implementation for
**                           the Progress compatibility global vars.
** 005 GES 20060125   @24071 Shift to SourceNameMapper backed stack
**                           processing.
** 006 EVL 20060713   @28035 Changing the approach to get the source file
**                           name from index of the Progress procedure on
**                           the stack.
** 007 GES 20070321   @32506 Added support for assigning the PROPATH.
** 008 GES 20071214   @36369 Fix for NPE in getSourceName().
** 009 GES 20080123   @36902 Better propath and separator defaults.
** 010 GES 20080321   @37628 Added SESSION:BATCH-MODE support.
** 011 GES 20080414   @38014 Added SESSION:CHARSET support.
** 012 GES 20080924   @39940 Mask off inner classes from program-name
**                           processing.
** 013 GES 20090512   @42147 Minor optimization for program-name.
** 014 ECF 20101005          Added getTempDirectory method.
** 015 GES 20110713          Changed default 4GL version string to "10.2B".
** 016 CS  20121104          Implement PROPATH assignment.
** 017 EVL 20130108          Fixing the PROPATH handling. We have to substitute
**                           system path separator with comma when getting
**                           PROPATH value. Also the default path and file
**                           separators value is chaged to return the current
**                           OS dependent one from File.pathSeparator and
**                           File.separator. Also the getLegacyCaseSensitive()
**                           method has been changed to set the default value
**                           depending on the current OS.
** 018 EVL 20130128          Adding support for GET-KEY-VALUE, PUT-KEY-VALUE,
**                           USE, LOAD, UNLOAD statements. The COLOR and FONT
**                           options for PUT-KEY-VALUE is not yet supported.
**                           The corresponting Java methods are just stubs to
**                           be implemented later.
** 019 VMN 20130219          Added support for attributes
**                           SESSION:FIRST-BUFFER (KW_FIRST_BU)
** 020 CA  20130222          Removed getFirstBuffer (stub is in SessionUtils).
** 021 EVL 20130529          Changing the setSearchPath method to take into account the current
**                           PROPATH value while setting up the new one. The new item is adding
**                           in the head of the string replacing the variable part of the PROPATH
**                           value. Also the methods splitPath and fixupPropath were modified to
**                           consider the propath separator is not the same as OS dependent file
**                           separator. The PROPATH_SEPARATOR string constant has been added to
**                           centralize this value. The fixed part of the PROPATH value has been
**                           initialized only once from server ini file and used in every propath
**                           modification as additional appendix after comma separator. The new
**                           change is adding the one more member to the WorkArea to separate the
**                           fixed and changeable parts. The member propath has been removed from
**                           main class, this value is getting now dynamically.
** 022 CA  20130604          Added setter for the batch mode state.
** 023 EVL 20130625          Changing the handling of incoming load() call to work with
**                           environment. In case of only two character or String parameters in
**                           this call we have environment and base-key input options. Because
**                           the directory option is useless in this case - this option is valid
**                           only for INI file based environment but without base-key "INI" the
**                           load() call is referred to the registry related case. For registry
**                           on the other hand the directory option can not be applied. So two
**                           input parameters means only environment and base-key variant.
** 024 CA  20140218          PROGRAM-NAME is now implemented by ProcedureManager.getStackEntry,
**                           which is called by getSourceName().
** 025 EVL 20140325          Changing request the OS name, type and path separator to the client
**                           side detection.
** 026 CA  20140702          Added isCustomEnvironment and getEnvironmentName APIs.
**                           Added generic's.  Removed unused WorkArea.pkgroot field.
** 027 CA  20140715          Fixed getTempDirectory - it needs to interrogate the P2J client.
** 028 MAG 20140822          Added getLegacyPlatform APIs.
** 029 MAG 20140908          Added COLOR-TABLE and FONT-TABLE notification.
** 030 MAG 20140912          INI files are available on all platforms regarding client type.
** 031 GES 20141030          Added CPINTERNAL.
** 032 SVL 20150114          For admin client use server-side OS version.
** 033 ECF 20150130          Added CPSTREAM, currently hard-coded to return UTF-8.
** 034 GES 20150615          PROGRAM-NAME() is updated to accept 64-bit integer values (this is
**                           valid in the 4GL.
** 035 ECF 20150715          Replace StringBuffer with StringBuilder.
** 036 IAS 20160331          Fixed potential NPE.
** 037 GES 20170228          Added javadoc.
** 038 GES 20171204          Improved comments and code formatting.
** 039 ECF 20180104          Added runtime support for converted RT-OPSYS global variable (FWD
**                           extension).
** 040 EVL 20180220          Now the current printer name is getting from client OS base. 
** 041 CA  20180221          Added listeners to be notified when PROPATH has been set.
** 042 GES 20180322          Added getProcessArchitecture().
** 043 CA  20180313          Javadoc fix.
** 044 SBI 20180522          Renamed file-system container name into legacy-system and mapped
**                           legacyPlatform to legacy-system/opsys and removed legacyPlatform,
**                           added getLegacyAbsolutePathPrefixMap(DirectoryService).
** 045 SBI 20180531          Changed to use common Utils.getDirectoryNodeMap, added
**                           getLegacyEnvironmentVariables.
** 046 ECF 20180614          Performance improvement: cached frequently accessed configuration
**                           values to avoid expensive directory lookups.
** 047 CA  20180514          Added support for CURRENT-LANGUAGE setter.
**     CA  20180523          Fixed environment name, when used in maps: registries are 
**                           case-insensitive, stanza INI are case-sensitive.
**                           Fixed APIs when LOAD is used with mix of DIR, NEW and BASE-KEY. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;

/**
 * Progress compatible environment-related static utility methods.
 * <p>
 * This supports some common global variables that are called built-in
 * functions in Progress but which take no parameters, not even empty
 * parenthesis. These variables are read-only (they cannot be assigned),
 * except for <code>PROPATH</code> which can be modified as it can be in
 * Progress as well.
 * <p>
 * The following is the list of supported variables:
 * <p>
 * <pre>
 * Progress Variable                    Java Equivalent
 * ---------------------                ----------------------
 * current-language                     {@link #getCurrentLanguage}
 * dataservers                          {@link #getDataServerList}
 * gateways                             {@link #getDataServerList}
 * opsys                                {@link #getOperatingSystem}
 * progress                             {@link #getRuntimeType}
 * promsgs                              {@link #getMessageSource}
 * propath                              {@link #getSearchPath}
 * proversion                           {@link #getVersion}
 * </pre>
 * <p>
 * <pre>
 * Progress Language Statement          Java Equivalent
 * ---------------------------          ----------------------
 * get-key-value                        {@link #getKeyValue}
 * load                                 {@link #load}
 * propath                              {@link #setSearchPath}
 * put-key-value                        {@link #setKeyValue}
 * unload                               {@link #unload}
 * use                                  {@link #use}
 * </pre>
 * <p>
 * <pre>
 * Progress Function                    Java Equivalent
 * ---------------------                ----------------------
 * program-name                         {@link #getSourceName}
 * </pre>
 * <p>
* <pre>
 * System Handle:Attribute              Java Equivalent
 * -------------------------            ----------------------
 * session:batch-mode                   {@link #isBatchMode}       (get)
 * session:charset                      {@link #getCharset}        (get)
 * session:cpinternal                   {@link #getCPInternal}     (get)
 * session:cpstream                     {@link #getCPStream}       (get)
 * session:temp-directory               {@link #getTempDirectory}  (get)
 * </pre>
 */
public class EnvironmentOps
{
   /** Progress OPSYS string for Unix systems. */
   public static final String OS_UNIX = "UNIX";
   
   /** Progress OPSYS string for Windows systems. */
   public static final String OS_WIN = "WIN32";
   
   /** Progress PROPATH separator string for all systems. */
   public static final String PROPATH_SEPARATOR = ",";
   
   /** 
    * Maps known Java os.name system property entries to a given Progress 
    * OPSYS string.  Entries should be inserted in lowercase and all keys
    * should be lowercased before reading from the map.
    */
   public static final Map<String, String> osMap = new HashMap<>();
   
   /** The fixed default part of the PROPATH value. */
   private static String propathFixed = null;
   
   static
   {
      // unix variants
      osMap.put("linux",    OS_UNIX);
      osMap.put("aix",      OS_UNIX);
      osMap.put("solaris",  OS_UNIX);
      osMap.put("sunos",    OS_UNIX);
      osMap.put("hp-ux",    OS_UNIX);
      osMap.put("freebsd",  OS_UNIX);
      osMap.put("mac os x", OS_UNIX);
      
      // windows variants
      osMap.put("windows xp",   OS_WIN);
      osMap.put("windows 2000", OS_WIN);
      osMap.put("windows nt",   OS_WIN);
      osMap.put("windows me",   OS_WIN);
      osMap.put("windows 98",   OS_WIN);
      osMap.put("windows 95",   OS_WIN);
      
      // Initialize default PROPATH value
      propathFixed = fixupPropath(Utils.getDirectoryNodeString(null, "searchpath-fixed", ""));
   }
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /**
    * Gets the name of the current language being used for the Progress
    * compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/currentLanguage
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/currentLanguage
    * <p>
    * If no value is found via this lookup, then the default value of
    * "?" will be returned.
    *
    * @return   The name of the current language.
    */
   public static character getCurrentLanguage()
   {
      String lang = Utils.getDirectoryNodeString(null,
                                                 "currentLanguage",
                                                 "?");
      return new character(lang);
   }
   
   /**
    * Sets the name of the current language being used for the Progress compatible environment.
    * <p>
    * This value will be saved in the directory in the 
    * /server/&lt;serverID&gt;/runtime/&lt;account&gt;/currentLanguage node.
    * <p>
    * All intermediary parents will be created.
    * 
    * @param    val
    *           The new CURRENT-LANGUAGE value.
    */
   public static void setCurrentLanguage(String val)
   {
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         return;
      }
      
      if (val == null)
      {
         val = "?";
      }

      boolean commit = false;
      try
      {
         SecurityManager sm = SecurityManager.getInstance();
         String userId = sm.getUserId();
         String serverId = sm.getServerId();
         
         String nodeId = String.format("/server/%s/runtime/%s/currentLanguage", serverId, userId);
         String oldVal = ds.getNodeString(nodeId, "value");

         if (val.equals(oldVal))
         {
            return;
         }

         if (oldVal == null && val != null)
         {
            String[] nodes =
            {
               "server",
               serverId,
               "runtime",
               userId,
               "currentLanguage"
            };
            
            // create the parents, if old value can't be read
            
            String pnode = "/";
            for (String nid : nodes)
            {
               String inode = pnode + nid + "/";
               if (!ds.openBatch(inode))
               {
                  ds.openBatch(pnode);
                  boolean added;
                  if (nid.equals("currentLanguage"))
                  {
                     NodeAttribute nodeAttr = ds.getClassNodeAttribute("string", "value");
                     Attribute attr = new Attribute(nodeAttr, new Object[] { "?" });
                     added = ds.addNode(inode, "string", new Attribute[] { attr });
                  }
                  else
                  {
                     added = ds.addNode(inode, "container", null);
                  }
                  ds.closeBatch(added);
               }
               else
               {
                  ds.closeBatch(false);
               }
               pnode = inode;
            }
         }
         
         ds.openBatch(nodeId);
         ds.setNodeString(nodeId, "value", val);
         commit = true;
      }
      finally
      {
         ds.closeBatch(commit);
         ds.unbind();
      }
   }
   
   /**
    * Sets the name of the current language being used for the Progress compatible environment.
    * <p>
    * This value will be saved in the directory in the 
    * /server/&lt;serverID&gt;/runtime/&lt;account&gt;/currentLanguage node.
    * <p>
    * All intermediary parents will be created.
    * 
    * @param    val
    *           The new CURRENT-LANGUAGE value.
    */
   public static void setCurrentLanguage(character val)
   {
      setCurrentLanguage(val.isUnknown() ? null : val.toStringMessage());
   }
   
   /**
    * Gets the list of the data servers available in Progress compatible
    * environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/dataServers
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/dataServers
    * <p>
    * If no value is found via this lookup, then the default value of
    * "PROGRESS,ORACLE,AS400,ODBC,MSS" will be returned.
    *
    * @return   The list of data servers.
    */
   public static character getDataServerList()
   {
      String def     = "PROGRESS,ORACLE,AS400,ODBC,MSS";
      String dserver = Utils.getDirectoryNodeString(null,
                                                    "dataServers",
                                                    def);
      return new character(dserver);
   }

   /**
    * Gets the name of the operating system upon which the Progress compatible
    * client environment is running.  The first attempt to resolve this value
    * will occur via a lookup in the directory.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/opsys/override
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/opsys/override
    * <p>
    * If no value is found via this lookup, then the Java system property
    * "os.name" is queried.  This is then translated via an optional
    * mapping table in the directory (that same paths and precedence as
    * above but ending in "opsys/mapping/&lt;lowercased_system_propert_name&gt;.
    * <p>
    * If no mapping is found, then an internal mapping table in this class
    * is consulted.  This table maps the following:
    * <p>
    * <pre>
    * os.name          Progress OPSYS
    * ---------------- ---------------
    * "linux"          "UNIX"
    * "aix"            "UNIX"
    * "solaris"        "UNIX"
    * "sunos"          "UNIX"
    * "hp-ux"          "UNIX"
    * "freebsd"        "UNIX"
    * "mac os x"       "UNIX"
    * "windows xp"     "WIN32"
    * "windows 2000"   "WIN32"
    * "windows nt"     "WIN32"
    * "windows me"     "WIN32"
    * "windows 98"     "WIN32"
    * "windows 95"     "WIN32"
    * </pre>
    * <p>
    * If there is still no match, then any <code>os.name</code> with the
    * lowercase text "unix" will return "UNIX" and any <code>os.name</code>
    * with the lowercase text "windows" will return "WIN32". 
    * <p>
    * If there is still no match, the <code>os.name</code> returned by
    * Java will be returned.
    *
    * @return   The name of the current operating system.
    */
   public static character getOperatingSystem()
   {
      WorkArea wa = work.obtain();
      
      if (wa.clientOsName == null)
      {
         // look for a hard coded value
         String osname = Utils.getDirectoryNodeString(null, "opsys/override", "");
         
         // if there is a hard coded operating system override, use it
         if (osname.length() == 0)
         {
            osname = getClientOpSysName(wa);
         }
         
         // Store per session OS name string to avoid every time call down to client
         wa.clientOsName = osname;
         
         // Calculate the cache for boolean Windows flag
         wa.isWindowsClient = osname.toLowerCase().indexOf("win") != -1; 
      }
      
      return new character(wa.clientOsName);
   }
   
   /**
    * Gets the name of the operating system being upon which the Progress compatible client
    * environment is running.  This method differs from {@link #getOperatingSystem()} in that
    * the first attempt to resolve this value will NOT occur via a lookup in the directory under
    * the {@code opsys/override} key. However, the rest of the resolution of this value is the
    * same as is described in {@link #getOperatingSystem()}.
    *
    * @return   The name of the current operating system upon which the client is running.
    */
   public static character getRuntimeOperatingSystem()
   {
      WorkArea wa = work.obtain();
      
      if (wa.clientRuntimeOsName == null)
      {
         wa.clientRuntimeOsName = getClientOpSysName(wa);
      }
      
      return new character(wa.clientRuntimeOsName);
   }
   
   /**
    * Checks is the current OS the P2J is running on belongs to the Windows
    * family.
    *
    * @return   <code>true</code> if the current OS is from Windows family or 
    *           <code>false</code> otherwise.
    */
   public static boolean isUnderWindowsFamily()
   {
      WorkArea wa = work.obtain();
      
      if (wa.clientOsName == null)
      {
         getOperatingSystem();
      }
      
      return wa.isWindowsClient;
   }
   
   /**
    * Gets the current runtime type being provided by the Progress
    * compatible environment.  The P2J environment only provides a
    * "Full" environment.
    *
    * @return   Always returns the string "Full".
    */
   public static character getRuntimeType()
   {
      return new character("Full");
   }
   
   /**
    * Gets the name of the directory where temporary files are stored.
    * Corresponds with with the <code>SESSION:TEMP-DIRECTORY</code> system
    * handle attribute.
    * 
    * @return  Name of temporary directory.
    */
   public static character getTempDirectory()
   {
      return new character(work.get().osInsp.getTempDirectory());
   }
   
   /**
    * Gets the name of the current message source being used for the Progress
    * compatible environment.
    *
    * @return   Always returns the string "promsgs".
    */
   public static character getMessageSource()
   {
      return new character("promsgs");
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    * @param    cvt
    *           This marks if the function is being called at conversion time or not.
    */
   public static void setSearchPath(String path, boolean cvt)
   {
      if (path != null)
      {
         WorkArea wa = work.obtain();
         
         // the passed in value is the variable PROPATH part; since this was potentially
         // built up at runtime using converted 4GL, the code might be relying upon
         // the legacy system-specific propath separator character instead of the comma
         // (which is platform neutral); convert the legacy path-separator to the comma
         // here
         String override = fixupPropath(path);

         String newpath = override;
         
         // combine the final value from variable and fixed parts, if needed
         if (!propathFixed.equals(""))
         {
            newpath = newpath + PROPATH_SEPARATOR + propathFixed;
         }

         // save the new effective propath 
         wa.propath = newpath;
         
         // save the variable part
         wa.propathOverride = override;

         if (!cvt)
         {
            wa.notifySourcePath(wa.propath);
         }
      }
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(String path)
   {
      setSearchPath(path, false);
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(character path)
   {
      setSearchPath(path.toStringMessage());
   }
   
   /**
    * Gets any context-specific override for the Progress search path which
    * was previously set via {@link #setSearchPath}.
    *
    * @return   The path override or <code>null</code> if no override was
    *           ever set.
    */
   public static String getSearchPathOverride()
   {
      return work.obtain().propath;
   }
   
   /**
    * Gets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * use of the "PROPATH" variable in Progress and is also used in the
    * implementation of {@link FileSystemOps#searchPath}.  The actual I/O
    * operations when executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    * <p>
    * If a specific PROPATH has been assigned in this user's context, then
    * that "override" will be returned.  Otherwise, the directory will be
    * consulted for the stored version.
    * <p>
    * The value returned from the directory may have been found via a search
    * algorithm that is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/searchpath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/searchpath
    * <p>
    * If no value is found via this lookup, then the default value of
    * ".," will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static character getSearchPath()
   {
      WorkArea wa      = work.obtain();
      String   propath = wa.propath;
      
      // is there a context-specific override?
      if (propath == null)
      {
         // support the override
         propath = wa.propathOverride;
         // There is system default, add to the end
         if (!propathFixed.equals(""))
         {
            propath = propath + PROPATH_SEPARATOR + propathFixed;  
         }
         // Save the local copy of the effective propath
         wa.propath = propath; 
         
         wa.notifySourcePath(propath);
      }
      
      return new character(propath);
   }
   
   /**
    * Substitutes the path separator values with the comma character in
    * the given string and returns the 4GL compatible PROPATH value that is
    * the combination of the directory entries separated by comma.
    *
    * @param    propathInitial
    *           The PROPATH that possibly needs to be fixed due to containing
    *           of the path separators instead of the expected comma values.
    *
    * @return   The PROPATH value with comma separation instead of possible
    *           path separators.
    */
   public static String fixupPropath(String propathInitial)
   {
      // First select the right path separator
      String pathSep = Utils.getDirectoryNodeString(null, "path-separator", File.pathSeparator);
      StringBuilder sb = new StringBuilder();
      int pathSepLength = pathSep.length();
      
      // The starting position of the next PROPATH element
      int nextSegment = 0;
      // Scan input parameter for the next occurence of the path separator.
      // The next PROPATH element starting point is the sum of the current
      // PROPATH element end point and length of the path sepatator.
      for (int i = 0;
           (i = propathInitial.indexOf(pathSep, nextSegment)) >= 0;
           nextSegment = i + pathSepLength)
      {
         // Extract the PROPATH entry up to the nearest path separator 
         sb.append(propathInitial.substring(nextSegment, i));
         // Replace path separator with comma
         sb.append(PROPATH_SEPARATOR);
      }
      // We have to handle the case when new PROPATH sting is not finished
      // with path separator.
      if (nextSegment < propathInitial.length())
         sb.append(propathInitial.substring(nextSegment));
      
      return sb.toString();
   }
   
   /**
    * Gets the original Progress SOURCE system's PROPATH which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/legacy-system/propath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/legacy-system/propath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/legacy-system/propath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/legacy-system/propath
    * <p>
    * If no value is found via this lookup, then the default value of
    * "" (empty string) will be returned.
    * <p>
    * This must be distinguished from Progress source file/search path on the
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getSearchPath}.
    *
    * @return   The path to Progress source files.
    */
   public static String getLegacySearchPath()
   {
      return Utils.getDirectoryNodeString(null, "legacy-system/propath", "");
   }
   
   /**
    * Gets the path separator character used to parse the Progress search
    * path (see {@link #getSearchPath}).
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ':' will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static String getPathSeparator()
   {
      if (work == null)
      {
         return Utils.getDirectoryNodeString(null, "path-separator", File.pathSeparator);
      }
      
      WorkArea wa = work.obtain();
      
      if (wa.clientPathSeparator == null)
      {
         // check the directory for an override value
         String pathSep = Utils.getDirectoryNodeString(null, "path-separator", "");
         
         // read client side OS value as fallback for path separator value
         if (pathSep.length() == 0 && wa.osInsp != null)
         {
            pathSep = wa.osInsp.getPathSeparator();
         }
         
         // if we still don't have a value, just use the server's OS value
         if (pathSep == null || pathSep.length() == 0)
         {
            pathSep = File.pathSeparator;
         }
         
         // save the calculated value
         wa.clientPathSeparator = pathSep;
      }
      
      return wa.clientPathSeparator;
   }
   
   /**
    * Gets the original Progress SOURCE system's path separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/legacy-system/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/legacy-system/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/legacy-system/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/legacy-system/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ":" will be returned.
    * <p>
    * This must be distinguished from the path separator on the (runtime)
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getPathSeparator}.
    *
    * @return   The path separator used on the SOURCE system.
    */
   public static String getLegacyPathSeparator()
   {
      return Utils.getDirectoryNodeString(null,
                                          "legacy-system/path-separator",
                                          File.pathSeparator);
   }
   
   /**
    * Gets the list of directories in the Progress source file path.  Each
    * element of the array is a directory in the source path.  This source
    * path is split based on the path separator used in this path
    * specification which may be different than that reported by the
    * system (<code>File.pathSeparator</code>.  The resulting array is
    * returned to the caller.
    * <p>
    * The {@link #getSearchPath} is used for the path specification and
    * the {@link #getPathSeparator} is used to split the path.
    *
    * @return   The array of directory names.
    */
   public static String[] getSearchPathList()
   {
      return splitPath(getSearchPath().toStringMessage(), PROPATH_SEPARATOR);
   }
   
   /**
    * Gets the original Progress SOURCE system's file separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/legacy-system/file-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/legacy-system/file-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/legacy-system/file-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/legacy-system/file-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * "/" will be returned.
    *
    * @return   The legacy system's file separator.
    */
   public static String getLegacyFileSeparator()
   {
      return work.obtain().legacyFileSeparator;
   }
   
   /**
    * Gets the absolute path prefix map from the current directory settings.
    * 
    * @return   The absolute path prefix map
    */
   public static Map<String, String> getLegacyAbsolutePathPrefixMap()
   {
      WorkArea wa = work.obtain();
      
      if (wa.pathPrefixMap == null)
      {
         wa.pathPrefixMap = Utils.getDirectoryNodeMap(DirectoryService.getInstance(),
                                                      "legacy-system/absolute-path-prefix-map",
                                                      null,
                                                      true,
                                                      false);
      }
      
      return wa.pathPrefixMap;
   }
   
   /**
    * Gets the environment variables map from the current directory settings.
    * 
    * @param    caseInsensitiveKeys
    *           {@code true} to use case-insensitive keys for the environment variables map.
    * 
    * @return   The environment variables map
    */
   public static Map<String, String> getLegacyEnvironmentVariables(boolean caseInsensitiveKeys)
   {
      return Utils.getDirectoryNodeMap(DirectoryService.getInstance(),
                                       "legacy-system/environment-variables",
                                       null,
                                       true,
                                       caseInsensitiveKeys);
   }
   
   /**
    * Gets the original Progress SOURCE file system's case-sensitive setting
    * which must be maintained for certain runtime translations of dynamically
    * generated names (the algorithms of which were designed with that path
    * assumed) but which is not necessarily a valid path on the system on
    * which this is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/legacy-system/case-sensitive
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/legacy-system/case-sensitive
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/legacy-system/case-sensitive
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/legacy-system/case-sensitive
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>true</code> will be returned.
    *
    * @return   <code>true</code> if the legacy file system was 
    *           case-sensitive.
    */
   public static boolean getLegacyCaseSensitive()
   {
      WorkArea wa = work.obtain();
      
      if (wa.legacyCaseSensitive == null)
      {
         wa.legacyCaseSensitive = Utils.getDirectoryNodeBoolean(null,
                                                                "legacy-system/case-sensitive",
                                                                !isUnderWindowsFamily());
      }
      
      return wa.legacyCaseSensitive;
   }
   
   /**
    * Gets the current version string of the Progress compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/version
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/version
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/version
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/version
    * <p>
    * If no value is found via this lookup, then the default value of
    * "10.2B" will be returned.
    *
    * @return   The current version string.
    */
   public static character getVersion()
   {
      String version = Utils.getDirectoryNodeString(null,
                                                    "version",
                                                    "10.2B");
      return new character(version);
   }
   
   /**
    * Returns the Progress source file name that matches the program at the given index in the 
    * stack trace if the given index position exists. The Progress source file name will be 
    * returned if possible using the {@link ProcedureManager#getStackEntry}.
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned. Decimal values are
    *           not allowed in the 4GL (compile-time error), but any 64-bit integer value is
    *           allowed.
    *
    * @return   The original Progress 4GL source file name (if the class is a converted 4GL
    *           program).  If the index is out of bounds or is unknown, then the unknown value
    *           is returned.
    */
   public static character getSourceName(int64 index)
   {
      if (index.isUnknown())
      {
         return new character();
      }
      
      return getSourceName(index.longValue());
   }
   
   /**
    * Returns the Progress source file name that matches the program at the given index in the 
    * stack trace if the given index position exists. The Progress source file name will be 
    * returned if possible using the {@link ProcedureManager#getStackEntry}.
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned. Decimal values are
    *           not allowed in the 4GL (compile-time error), but any 64-bit integer value is
    *           allowed.
    *
    * @return   The original Progress 4GL source file name (if the class is a converted 4GL
    *           program).  If the index is out of bounds, then the unknown value is returned.
    */
   public static character getSourceName(long index)
   {
      return new character(ProcedureManager.getStackEntry((int) (index - 1)));
   }
   
   /**
    * Splits a path using the alternative separator string
    * 
    * @param path 
    *        The <code>String</code> path to split into an array of directory
    *        names.
    * @param separator 
    *        The <code>String</code> separator to split into an array of directory
    *        names.
    * 
    * @return The array of directory names.
    */
   private static String[] splitPath(String path, String separator)
   {
      if (separator == null)
      {
         separator = getPathSeparator();
      }
      if (path != null)
      {
         return path.split(separator, -1);
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Reports if the current context is running in batch mode.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/batchMode
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/batchMode
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>false</code> will be returned.
    *
    * @return   <code>true</code> if the current context is in batch mode.
    */
   public static logical isBatchMode()
   {
      WorkArea wa = work.obtain();
      if (wa.batchMode == null)
      {
         wa.batchMode = Utils.getDirectoryNodeBoolean(null, "batchMode", false);
      }
      
      return new logical(wa.batchMode);
   }
   
   /**
    * Force the batch mode of the current context to the specified state.
    *
    * @param   batchMode
    *          <code>true</code> if this context must be in batch mode.
    */
   public static void setBatchMode(boolean batchMode)
   {
      WorkArea wa = work.obtain();
      wa.batchMode = batchMode;
   }
   
   /**
    * Reports the default charset used by the JVM. Since the Progress default charset is
    * ISO-8859-1, when the JVM reports that UTF-8 is the default, this method will return
    * the string "ISO-8859-1" instead.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   The default charset name.
    */
   public static character getCharset()
   {
      // TODO: is this right?
      return getCPInternal();
   }

   /**
    * Reports the default codepage used by the JVM. Since the Progress default charset is
    * ISO8859-1, when the JVM reports that UTF-8 is the default, this method will return
    * the string "ISO8859-1" instead.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   The codepage used for internal operations.
    */
   public static character getCPInternal()
   {
      String cp = Utils.getCharsetName();
      
      if ("ISO-8859-1".equals(cp))
      {
         // TODO: the 4GL doesn't insert the extra hyphen, this really needs to be explored
         cp = "ISO8859-1";
      }
      
      return new character(cp);
   }
   
   /**
    * Reports the codepage used for streams (the equivalent of the CPSTREAM 4GL attribute, which
    * reports the value set using the <code>-cpstream</code> startup parameter.
    * 
    * @return   The codepage used for streams (TODO: proper implementation).
    */
   public static character getCPStream()
   {
      UnimplementedFeature.todo("Always returns 'UTF-8'; needs proper implementation");
      
      return new character("UTF-8");
   }
   
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    */
   public static void load(character env)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(false), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    */
   public static void load(String env)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(false), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, boolean flagNew, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, boolean flagNew, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, boolean flagNew, character baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(flagNew), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    */
   public static void load(character env, character directory)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, directory, new logical(false), null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    */
   public static void load(String env, character directory)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), directory, new logical(false), null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    */
   public static void load(character env, String directory)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, new character(directory), new logical(false), null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    */
   public static void load(String env, String directory)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), new character(directory), new logical(false), null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, character directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, directory, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, character directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), directory, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, String directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, new character(directory), new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, String directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), new character(directory), new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, String baseKey)
   {
      load(env, directory, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, character baseKey)
   {
      load(env, directory, new logical(false), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, String directory, String baseKey)
   {
      load(env, new character(directory), new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, String baseKey)
   {
      load(new character(env), directory, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, character baseKey)
   {
      load(new character(env), directory, new logical(false), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, String directory, String baseKey)
   {
      load(new character(env), new character(directory),
           new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, boolean flagNew, String baseKey)
   {
      load(env, directory, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, String directory, boolean flagNew, String baseKey)
   {
      load(env, new character(directory), new logical(flagNew), new character(baseKey));
   }
   
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, boolean flagNew, String baseKey)
   {
      load(new character(env), directory, new logical(flagNew), new character(baseKey));
   }
   
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, boolean flagNew, character baseKey)
   {
      load(new character(env), directory, new logical(flagNew), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, String directory, boolean flagNew, String baseKey)
   {
      load(new character(env), new character(directory),
           new logical(flagNew), new character(baseKey));
   }
   
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, logical flagNew, character baseKey)
   {
      // check if INI file
      boolean isIni = baseKey != null && 
            EnvironmentReader.TYPE_INI.equalsIgnoreCase(baseKey.toStringMessage());
            
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIni)
      {
         onLoadError(env, directory);
         return;
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String envAsString = null;
      String directoryAsString = null;
      String baseKeyAsString = null;
      boolean flagNewAsBoolean = false;
      
      if (env != null)
      {
         envAsString = env.toStringMessage();
      }
      if (directory != null)
      {
         directoryAsString = directory.toStringMessage();
      }
      if (flagNew != null)
      {
         flagNewAsBoolean = flagNew.booleanValue();
      }
      if (baseKey != null)
      {
         baseKeyAsString = baseKey.toStringMessage();
      }
      
      if (!Environments.load(envAsString, directoryAsString, flagNewAsBoolean, baseKeyAsString))
      {
         onLoadError(env, directory);
      }
   }
   
   /**
    * On load error message.
    * 
    * @param    env 
    *           The environment to create or load. Can be the registry key,
    *           initialization file.
    * @param    directory 
    *           The directory path for existing initialization file or place
    *           to create initialization file.
    */
   private static void onLoadError(character env, character directory)
   {
      // This is support for NO-ERROR option
      if (directory == null || directory.isUnknown())
      {
         ErrorManager.recordOrThrowError(4450, "The LOAD of " + env.toStringMessage() + " failed");
      }
      else
      {
         ErrorManager.recordOrThrowError(4449, "The LOAD of " + env.toStringMessage() +
                                               " in path " +  directory.toStringMessage() +
                                               " failed");
      }
   }
   
   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is empty string ("") the default
    *        environment becomes the current one.
    */
   public static void use(String env)
   {
      use(new character(env));
   }
   
   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is empty string ("") the default
    *        environment becomes the current one.
    */
   public static void use(character env)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         ErrorManager.recordOrThrowError(4452, "The USE of " + env.toStringMessage() + " failed");
         return;
      }
      
      String envName = Environments.use(env.toStringMessage());

      if (envName == null)
      {
         ErrorManager.conditionalShowErrorAndAbend(4452, "The USE of " + env.toStringMessage() +
                                                         " failed");
         return;
      }
      
      // notify
      onUse(envName);
   }
   
   /**
    * Check if a custom environment is in use.
    * 
    * @return   See above.
    */
   public static boolean isCustomEnvironment()
   {
      return (isUnderWindowsFamily() || isIniType()) 
               && Environments.isCustomEnvironment();
   }
   
   /**
    * Get the current environment's name.
    * 
    * @return   See above.
    */
   public static String getEnvironmentName()
   {
      return Environments.getEnvironmentName();
   }
   
   /**
    * Get the current environment type.
    * 
    * @return   See above.
    */
   public static String getEnvironmentType()
   {
      return Environments.getEnvironmentType();
   }
   
   /**
    * Get the name of the current printer.
    *
    * @return  the printer name.
    */
   public static character getCurrentPrinterName()
   {
      WorkArea wa = work.obtain();
      
      // getting current printer from system
      String prnName = wa.osInsp != null ? wa.osInsp.getCurrentPrinterName() : null;
      // return it if not null
      return new character(prnName == null || prnName.isEmpty()
                              ? PrintingService.PDF_PRINTER_DESCR : prnName);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(section, (character)null, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(new character(section), (character)null, value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(new character(section), (character)null, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(section, (character)null, value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character key, String value)
   {
      setKeyValue(section, key, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String key, character value)
   {
      setKeyValue(section, new character(key), value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character key, character value)
   {
      setKeyValue(new character(section), key, value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String key, String value)
   {
      setKeyValue(section, new character(key), new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String key, character value)
   {
      setKeyValue(new character(section), new character(key), value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character key, String value)
   {
      setKeyValue(new character(section), key, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String key, String value)
   {
      setKeyValue(new character(section), new character(key), new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character key, character value)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         return;
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String sectionAsString = null;
      String keyAsString = null;
      String valueAsString = null;
      
      if (section != null)
      {
         // Unknown section value should be handled the same way as empty string
         if (section.isUnknown())
         {
            sectionAsString = "";
         }
         else
         {
            sectionAsString = section.toStringMessage();
         }
      }
      if (key != null)
      {
         // Unknown key value should be handled the same way as empty string
         if (key.isUnknown())
         {
            keyAsString = "";
         }
         else
         {
            keyAsString = key.toStringMessage();
         }
      }
      if (value != null)
      {
         // Unknown value should be handled the same way as empty string
         if (value.isUnknown())
         {
            valueAsString = "";
         }
         else
         {
            valueAsString = value.toStringMessage();
            if (valueAsString.equals("?"))
            {
               valueAsString = "";
            }
         }
      }
      
      Environments.setKeyValue(sectionAsString, keyAsString, valueAsString);
   }
   
   /**
    * Getting the default key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      return getKeyValue(section, (character)null);
   }
   
   /**
    * Getting the default key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      return getKeyValue(new character(section), (character)null);
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section, String key)
   {
      return getKeyValue(new character(section), new character(key));
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section, String key)
   {
      return getKeyValue(section, new character(key));
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section, character key)
   {
      return getKeyValue(new character(section), key);
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section, character key)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         return new character();
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String sectionAsString = null;
      String keyAsString = null;
      
      if (section != null)
      {
         // Unknown section value should be handled the same way as empty string
         if (section.isUnknown())
         {
            sectionAsString = "";
         }
         else
         {
            sectionAsString = section.toStringMessage();
         }
      }
      if (key != null)
      {
         // Unknown key value should be handled the same way as empty string
         if (key.isUnknown())
         {
            keyAsString = "";
         }
         else
         {
            keyAsString = key.toStringMessage();
         }
      }
      
      return new character(Environments.getKeyValue(sectionAsString, keyAsString));
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param env 
    *        The environment to unload. Can be the registry or initialization
    *        file. The environment to unload should be previously loaded by
    *        load() method.
    */
   public static void unload(String env)
   {
      unload(new character(env));
   }
   
   /**
    * Unloads specified environment from the current one if possible.
    * 
    * @param env 
    *        The environment to unload. Can be the registry or initialization
    *        file. The environment to unload should be previously loaded by
    *        load() method.
    */
   public static void unload(character env)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         String msg = "The UNLOAD of " + env.toStringMessage() + " failed";
         ErrorManager.recordOrThrowError(4451, msg);
         return;
      }
      
      String envName = Environments.unload(env.toStringMessage());
      
      if (envName != null)
      {
         onUnload(envName);
      }
      else
      {
         String msg = "The UNLOAD of " + env.toStringMessage() + " failed";
         ErrorManager.recordOrThrowError(4451, msg);
      }
   }
   
   /**
    * Get current environment name. If no custom environment is set return default environment
    * name which is a <code>null</code> value.
    * 
    * @return  see above.
    */
   public static String getCurrentEnvironmentName()
   {
      return isCustomEnvironment() ? getEnvironmentName() : null;
   }
   
   /**
    * Get client legacy platform name.
    * <p> 
    * Sometimes we need to specify a legacy platform regarding the real platform on which the 
    * clients actually runs. @link ColorSpec#convert. By default the client legacy platform is 
    * the same as OPSYS built-in returned value in Progress.
    * <p> 
    * The value is read from server directory. If no value is set for <code>legacy-system/opsys</code>
    * then the returned value is the same as returned by the <code>getOperatingSystem</code> call.
    * If set in the directory, the value of opsys MUST be UNIX or WIN32.
    * 
    * @return  Client legacy platform name:
    *          UNIX  - Unix/Linux OS familly.
    *          WIN32 - Windows OS familly.          
    */
   public static String getLegacyPlatform()
   {
      WorkArea wa = work.obtain();
      
      if (wa.legacyPlatform == null)
      {
         String os = getOperatingSystem().toStringMessage();
         wa.legacyPlatform = Utils.getDirectoryNodeString(null, "legacy-system/opsys", os);
      }
      
      return wa.legacyPlatform;
   }
   
   /**
    * Returns the word size of the client's compiled native code, which may be different
    * from the "bitness" of the client's OS. For example, the client may be compiled as
    * 32-bit but executed on a 64-bit OS.
    *
    * @return   32 or 64 depending on the compiled word size of the client's native code. If
    *           there is an error in calling down to the client, unknown value will be 
    *           returned.
    */
   public static integer getProcessArchitecture()
   {
      int arch = memptr.compiledWordSize();
      
      // TODO: is there any case where this fails (and in 4GL it would not) so
      //       we introduce unknown unexpectedly? 
      
      return (arch == -1) ? new integer() : new integer(arch);
   }
   
   /**
    * Add a new listener to be notified when the PROPATH has been set.
    * 
    * @param    listener
    *           The listener to be notified.
    */
   static void addSourcePathListener(Consumer<String> listener)
   {
      WorkArea wa = work.obtain();
      
      wa.listeners.add(listener);
   }
   
   /**
    * Gets the name of the operating system upon which the Progress compatible client environment
    * is running.  The Java system property "os.name" of the client JVM is queried (unless running
    * in headless mode, in which case that of the server JVM is used).  This is then translated
    * via an optional mapping table in the directory (using the same paths and precedence as
    * described for the {@link #getOperatingSystem()} method, but ending in
    * {@code opsys/mapping/&lt;lowercased_system_propert_name&gt;}.
    * <p>
    * If no mapping is found, then an internal mapping table in this class is consulted.  This
    * table maps the following:
    * <p>
    * <pre>
    * os.name          Progress OPSYS
    * ---------------- ---------------
    * "linux"          "UNIX"
    * "aix"            "UNIX"
    * "solaris"        "UNIX"
    * "sunos"          "UNIX"
    * "hp-ux"          "UNIX"
    * "freebsd"        "UNIX"
    * "mac os x"       "UNIX"
    * "windows xp"     "WIN32"
    * "windows 2000"   "WIN32"
    * "windows nt"     "WIN32"
    * "windows me"     "WIN32"
    * "windows 98"     "WIN32"
    * "windows 95"     "WIN32"
    * </pre>
    * <p>
    * If there is no match, then any {@code os.name} with the lowercase text "unix" will return
    * "UNIX" and any {@code os.name} with the lowercase text "windows" will return "WIN32".
    * <p>
    * If there is still no match, the {@code os.name} returned by Java will be returned.
    * 
    * @param    wa
    *           Context-local work area.
    * 
    * @return   The name of the current operating system upon which the client is running.
    */
   private static String getClientOpSysName(WorkArea wa)
   {
      String osname = null;
      
      // query the JVM value (it will always be present)
      String fallback = null;
      
      // We have client running.
      // If the client is in the headless mode (e.g. admin client), there is no physical
      // terminal, so the OS can't be accessed.
      if (!TransactionManager._isHeadless() && wa.osInsp != null)
      {
         fallback = wa.osInsp.getOperatingSystem();
      }
      // We make conversion, take standalone/server value
      else
      {
         fallback = System.getProperty("os.name");
      }
      // New candidate               
      String sysprop  = fallback.toLowerCase();
      
      // check if there is a mapping in the directory for this exact
      // value
      osname = "opsys/mapping/" + sysprop;
      osname = Utils.getDirectoryNodeString(null, osname, "");
                                             
      // use the mapping if there was one
      if (osname.length() == 0)
      {
         // attempt to translate it by using our predefined tables
         osname = (String) osMap.get(sysprop);
         
         // no match?
         if (osname == null)
         {
            // last try
            if (sysprop.indexOf("unix") != -1)
            {
               osname = OS_UNIX;
            }
            else if (sysprop.indexOf("windows") != -1)
            {
               osname = OS_WIN;
            }
            else
            {
               // all else fails, return what Java specified
               osname = fallback;
            }
         }
      }
      
      return osname;
   }

   /**
    * A notification when <code>USE environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   private static void onUse(String envName)
   {
      BaseSysTable.use(envName);
   }
   
   /**
    * A notification when <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   private static void onUnload(String envName)
   {
      BaseSysTable.unload(envName);
   }
   
   /**
    * Check if it's a INI file environment type.
    *  
    * @return  <code>true</code> if INI file environment type.
    */
   private static boolean isIniType()
   {
      return EnvironmentReader.TYPE_INI.equals(getEnvironmentType());
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Stores any context-specific full effective override to the <code>PROPATH</code>. */
      private String propath = null;
      
      /** Stores any context-specific override to the <code>PROPATH</code>, changeable part. */
      private String propathOverride =
         fixupPropath(Utils.getDirectoryNodeString(null, "searchpath", ""));
      
      /** The batch-mode state of this context. */
      private Boolean batchMode = null;
      
      /** The name of the OS on the client side or OPSYS override specified in the directory. */
      private String clientOsName = null;
      
      /** The name of the OS on the client side. */
      private String clientRuntimeOsName = null;
      
      /** The flag indicating Windows on client side. */
      private boolean isWindowsClient;
      
      /** The string representing delimiter between path entries on the client side. */
      private String clientPathSeparator = null;
      
      /** Object to check OS properties on client side. */
      private OperatingSystemInspector osInsp = null;
      
      /** Client legacy plaform name */
      private String legacyPlatform = null;
      
      /** Set of listeners to be notified when the {@link #propath} has been set. */
      private Set<Consumer<String>> listeners = new HashSet<>();
      
      /** Mapping of legacy absolute file path prefixes to new environment path prefixes */
      private Map<String, String> pathPrefixMap = null;
      
      /** File name case-sensitivity of the legacy platform */
      private Boolean legacyCaseSensitive = null;
      
      /** File name separator of the legacy platform */
      private String legacyFileSeparator =
         Utils.getDirectoryNodeString(null, "legacy-system/file-separator", File.separator);
      
      /**
       * Send notifications that the PROPATH has been set.
       * 
       * @param    sourcePath
       *           The new PROPATH value.
       */
      protected void notifySourcePath(String sourcePath)
      {
         for (Consumer<String> pl : listeners)
         {
            pl.accept(sourcePath);
         }
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();

         // This will work for run-time only
         if (SessionManager.get() != null)
         {
            wa.osInsp = (OperatingSystemInspector) RemoteObject.obtainInstance(
                                                       OperatingSystemInspector.class, true);
         }
         
         return wa;
      }   
   }   
}
