/*
** Module   : Environments.java
** Abstract : Progress 4GL compatible Winodws registry environments related methods
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130522 Created initial version.
** 002 CA  20140702 Added isCustomEnvironment and getEnvironmentName APIs.
** 003 MAG 20140908 Added getEnvironmentType APIs. Added return value to load() method.
** 004 CA  20180523 Fixed environment name, when used in maps: registries are case-insensitive,
**                  stanza INI are case-sensitive. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * A server side class to manipulate with environments. Contains static calls to get the client
 * side part and call respective class methods on the client side. The environment can be the INI
 * file or registry sections containing the keys.
 *
 * @author    EVL
 */
public class Environments
{
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();
      
   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key, initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be created overwriting the
    *        possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under the particular key
    *        value. It is possible to specify "INI" as parameter. In this case search or creation
    *        will be done inside initialization file specified in env parameter.
    *        
    * @return  <code>true</code> if loaded false otherwise      
    */
   public static boolean load(String env, String directory, boolean flagNew, String baseKey)
   {
      return work.obtain().envAcc.load(env, directory, flagNew, baseKey);
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param    env 
    *           The environment to load. If the parameter is empty string ("") the default
    *           environment becomes the current one.
    *           
    * @return   The loaded environment's name, as it was saved on LOAD.  <code>null</code> if the
    *           environment can't be found.
    */
   public static String use(String env)
   {
      return work.obtain().envAcc.use(env);
   }
   
   /**
    * Check if a custom environment is in use.
    * 
    * @return   See above.
    */
   public static boolean isCustomEnvironment()
   {
      return work.obtain().envAcc.isCustomEnvironment();
   }
   
   /**
    * Get the current environment's name.
    * 
    * @return   See above.
    */
   public static String getEnvironmentName()
   {
      return work.obtain().envAcc.getEnvironmentName();
   }

   /**
    * Get the current environment's type.
    * 
    * @return   See above.
    */
   public static String getEnvironmentType()
   {
      return work.obtain().envAcc.getEnvironmentType();
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String key, String value)
   {
      work.obtain().envAcc.setKeyValue(section, key, value);
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static String getKeyValue(String section, String key)
   {
      return work.obtain().envAcc.getKeyValue(section, key);
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param    env 
    *           The environment to unload. Can be the registry or initialization file.
    *           The environment to unload should be previously loaded by load() method.
    *           
    * @return   The loaded environment's name, as it was saved on LOAD.  <code>null</code> if the
    *           environment can't be found.
    */
   public static String unload(String env)
   {
      return work.obtain().envAcc.unload(env);
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Remote instance of our file system worker. */ 
      private EnvironmentAccessor envAcc = null; 
   }
   
   /**
    * Simple container that stores and returns a context-local instance of the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   { 
      /**
       * Obtains the context-local instance of the contained global work area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return (WorkArea) this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         // obtain the our proxy, server-side code is left-side by convention
         wa.envAcc = (EnvironmentAccessor) RemoteObject.obtainInstance(EnvironmentAccessor.class,
                                                                       true);
         
         return wa;
      }   
   }   
}