/*
** Module   : ErrorWriter.java
** Abstract : Interface for writer objects that display the error messages.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131119 Created initial version.
** 002 IAS 20160328          Client log support
** 003 HC  20171003          Implemented SESSION:SUPPRESS-WARNINGS runtime support and the
**                           ability to handle (display or output depending on the runtime state)
**                           legacy warnings.
** 004 SVL 20171024          Added displayError(errmsg, noMessageBox).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Interface decribing the error messages outputs.
 */
public interface ErrorWriter
{
   /**
    * Displays error message using currently installed writer implementation.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayError(String errmsg);

   /**
    * Displays error message using currently installed writer implementation.
    *
    * @param    errmsg
    *           Text of the message to display.
    * @param    noMessageBox
    *           <code>true</code> if error is always displayed in message area or logged rather
    *           than displayed as a message box.
    */
   public void displayError(String errmsg, boolean noMessageBox);

   /**
    * Displays warning message using currently installed writer implementation.
    *
    * @param    wrnmsg
    *           Text of the message to display.
    */
   default public void displayWarning(String wrnmsg)
   {
      // by default the semantics is the same as in case of an error
      displayError(wrnmsg);
   }

   /**
    * Displays error message using currently installed writer implementation.  This is the
    * redirected version of the displayError.  If the output is redirected to file the message
    * is writing to the file.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayErrorRedirected(String errmsg);
   
   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean isInBatchMode();
   
   /**
    * Displays error message using currently installed writer implementation.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void message(String errmsg);
   
   /**
    * Displays error message box with title using currently installed writer implementation.
    *
    * @param    masterMsg
    *           Text of the message to display.
    * @param    title
    *           The title for the message box to display.
    */
   public void messageBox(String masterMsg, String title);

   /**
    * Implements pause processing using currently installed writer implementation.
    */
   public void pause();

   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public void stopBatchSession();

   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    */
   public void logStackTrace(String message);
}