/*
** Module   : ExplicitFileList.java
** Abstract : creates a file list based on a user-defined explicit list
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050125 NEW   @22975 Created initial version supporting a 
**                               user-defined list of filenames (absolute or
**                               relative) to convert into an array of
**                               File objects.
** 002 GES 20060129 CHG   @24129 Reimplemented with a new base class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * Creates a sorted list of files based on a user-defined list of filenames
 * (absolute or relative).  The list is sorted lexicographically.
 *
 * @author    GES
 */
public class ExplicitFileList
extends FileList
{
   /** The list of explicit filenames. */
   private String[] filenames = null;
   
   /**
    * Base constructor to handle all possible cases.
    *
    * @param    filenames
    *           The array of absolute or relative file names.
    * @param    caseSens
    *           Sets the case-sensitivity of the sorting algorithm.
    */
   public ExplicitFileList(String[] filenames, boolean caseSens)
   {
      super(caseSens);
      this.filenames = filenames;
   }
   
   /**
    * Common constructor which uses a case-insensitive sorting algorithm.
    *
    * @param    filenames
    *           The array of absolute or relative file names.
    */
   public ExplicitFileList(String[] filenames)
   {
      super(false);
      this.filenames = filenames;
   }
   
   /**
    * Core implementation of the listing algorithm which uses the
    * <code>File</code> class to create and return the list of files based
    * on user input specified during construction.  No directories will be
    * returned in the list and only files that actually exist will be
    * returned. The resulting list will be sorted lexicographically.
    *
    * @param    results
    *           The list to which all matching files must be appended.
    */
   protected void listImpl(ArrayList results)
   {
      // separate file and directory results
      for (int i = 0; i < filenames.length; i++)
      {
         File next = new File(filenames[i]);
         
         if (next.exists() && next.isFile())
         {
            // add the file to the results list
            results.add(next);
         }
      }
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the ExplicitFileList class.  The list of resulting filenames is
    * displayed to the user on <code>stdout</code>.
    * <p>
    * Syntax:
    * <pre>
    *    java ExplcitFileList &lt;filename&gt; ...
    * </pre>
    * Where:
    * <ul>
    *    <li> &lt;filename&gt; is one or more relative or absolute file names 
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)
   {
      String syntax = "Syntax: java ExplicitFileList <filename> ... ";
      
      if (args.length < 1)
      {
         System.err.println(syntax);
         System.exit(-1);
      }
      
      try
      {
         FileList flist  = new ExplicitFileList(args);
         String   list[] = flist.listFilenames();
         
         for ( int i = 0; i < list.length; i++ )
         {
            System.out.println(list[i]);
         }
      }
      
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }
}