/*
** Module   : FieldAssigner.java
** Abstract : Assigns [input-]output parameter values to database fields upon exit from
**            procedure/function
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20090816   @43673 Created initial version. Assigns [input-]output parameter values to 
**                           database fields upon exit from procedure/function.
** 002 CA  20131013          Added no-op scopeDeleted() method, required by the changes in 
**                           Scopeable interface.
** 003 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 004 OM  20150626          Most logic extracted OutputParameterAssigner in order to support the
**                           updating the wrapper variable (of a different type).
**                           Moved to com.goldencode.p2j.util package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;
import java.lang.reflect.*;
import java.util.logging.*;

/**
 * Specialization of {@link OutputParameterAssigner} that handles various {@code *Field} data-type
 * wrapper implementations, which assigns an INPUT-OUTPUT or OUTPUT parameter's value back to the
 * database field with which it is associated, upon exit from a procedure or function.
 * <p>
 * An individual instance of this class is created by each {@code *Field} instance, and that
 * variable is stored within this object. At construction, it initializes the parameter (if 
 * necessary) to the appropriate current value from the database. Upon exit from the procedure or
 * function with which it is associated, it attempts to assign the current value of its contained
 * variable back into the associated DMO, optionally applying the provided conversion constructor.
 * This may trigger normal validation processing, which may produce error conditions.
 */
public final class FieldAssigner
extends AbstractSimpleParameter
{
   /** Field reference which manages the association with the DMO property. */
   private final FieldReference field;
   
   /**
    * Constructor.
    * 
    * @param   variable
    *          Parameter value to and from which we will assign.
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   index
    *          Index to access, if property represents an extent field.
    * @param   initialize
    *          Should this variable be initialized with the DMO's current value for the target
    *          field?
    */
   public FieldAssigner(BaseDataType variable,
                        DataModelObject dmo,
                        String property,
                        NumberType index,
                        boolean initialize)
   {
      super(variable);
      this.field = new FieldReference(dmo, property, index);
      
      if (initialize)
      {
         variable.assign(field.get());
      }
   }
   
   /**
    * Assign the current value of the parameter variable back to the associated DMO property. This
    * is expected to occur upon exit from the function or procedure with which this object is
    * associated.
    */
   @Override
   protected void assign()
   {
      if (convCtor != null)
      {
         try
         {
            field.set((BaseDataType) convCtor.newInstance(outputValue));
         }
         catch (InstantiationException | IllegalAccessException | InvocationTargetException e)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               String msg = "Failed to convert OUTPUT value from " + outputValue.getClass() +
                            " to " + field.getType() + " for " + field.getProperty() + 
                            " using " + convCtor + " conversion constructor.";
               LOG.log(Level.WARNING, msg);
            }
         }
      }
      else 
      {
         field.set(outputValue);
      }
   }
}