/*
** Module   : FileSystemOps.java
** Abstract : file system utility methods
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050706   @21665 Created initial version, supporting common
**                           language statements and built-in functions
**                           for manipulating and querying the file
**                           system.
** 002 GES 20050728   @21922 Added searchPath to support the built-in
**                           search() function.
** 003 GES 20050831   @22425 Added FILE-INFO system handle support.
** 004 GES 20060123   @24033 Moved to a context-local implementation.
** 005 GES 20060211   @24540 Split the implementation from the Progress
**                           interface/semantic. The FileSystem interface
**                           was designed to allow this. This enables
**                           a remote implementation of the interface.
** 006 GES 20060322   @25202 Search path must be aware of the mapping
**                           between Progress source filenames and the
**                           Java equivalent.
** 007 GES 20080123   @36900 Changed getValue to toStringMessage which is
**                           safer. Added null checks everywhere to make
**                           the code safer.
** 008 GES 20080321   @37631 Added full support for nearly all remaining
**                           FILE-INFO features.  The remaining features
**                           are stubbed out.
** 009 GES 20110712          Initialize the case-sensitivity setting of the
**                           legacy system to allow searching to honor it.
** 010 GES 20111020          Modified per-session initialization to detect
**                           when the client code is running inside the server
**                           process AND to initialize for local redirection
**                           instead of using a remote object.
** 011 CS  20121104          Implement PROPATH assignment.
** 012 CA  20130112          Added asHandle, isValid, getType and isUnknown  
**                           methods, accessed by the FILE-INFO handle. 
**                           FILE-NAME attribute is converted to getFileName, 
**                           as it needs to be consistent with the procedure's
**                           FILE-NAME attribute.
** 013 CA  20130221          getType getter for TYPE attribute was renamed to getResourceType.
** 014 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 015 CS  20130325          Task 1613: finish FILE-INFO system handle support
** 016 OM  20130618          Used default constructors for unknown values.
** 017 CA  20130927          Static proxies need the unknown() API. Resource type is determined
**                           from LegacyResource constants.
** 018 EVL 20150119          Add init server call to push the PROPATH related values to client.
** 019 CA  20150502          The resource ID needs to be kept at the resource, so it can be
**                           accessed after the resource is deleted.
** 020 OM  20160527          searchPath() returns legacy procedure names instead of java class.
** 021 OM  20160927          Fixed getProperty() to return unknown value when property not set.
** 022 SVL 20170126          Added MEMBER function.
**     GES 20170228          Added javadoc.
** 023 SVL 20171109          Added getFileContents.
** 024 CA  20180221          Use listeners to be notified when PROPATH has been set.
** 025 SBI 20180522          Changed to set legacy file system parameters.
** 026 SBI 20180531          Added environment variables map.
**     ECF 20180609          Added {get|set}WorkingDirectory methods.
** 027 ECF 20180614          Adapt to changes in EnvironmentOps API signatures.
**         20180615          Refactored and moved code from LogicalTerminal to search for
**                           resources in jar files.
** 028 SBI 20180704          Changed searchPath to resolve a legacy path name into a valid system
**                           path name on the client system.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.concurrent.*;
import java.util.function.Function;
import java.io.*;

import com.goldencode.p2j.classloader.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.proxy.StaticProxy;
import com.goldencode.util.PlatformHelper;

/**
 * Progress compatible static file-system utility methods.
 * <p>
 * This supports some common global variables that are called built-in
 * functions in Progress but which take no parameters, not even empty
 * parenthesis. These variables are read-only (they cannot be assigned).
 * <p>
 * A partial implementation of the <code>FILE-INFO</code> system handle is
 * available but only the <code>FILE-NAME</code> and <code>FILE-SIZE</code>
 * attributes are supported.
 * <p>
 * The following is the list of supported Progress 4GL features:
 * <p>
 * <pre>
 * Progress Variable                    Java Equivalent
 * ---------------------                ----------------------
 * os-error                             {@link #getLastError}
 * os-drives                            {@link #getRootList}
 * os-getenv                            {@link #getProperty}
 * </pre>
 * <p>
 * <pre>
 * Language Statement                   Java Equivalent
 * ---------------------                ----------------------
 * os-append                            {@link #append}
 * os-copy                              {@link #copy}
 * os-create-dir                        {@link #mkdir}
 * os-delete                            {@link #delete}
 * os-rename                            {@link #rename}
 * </pre>
 * <p>
 * <pre>
 * Function                             Java Equivalent
 * ---------------------                ----------------------
 * member                               {@link #member}
 * search                               {@link #searchPath}
 * </pre>
 * <p>
* <pre>
 * System Handle:Attribute              Java Equivalent
 * -------------------------            ----------------------
 * file-info:file-create-date           {@link #fileInfoGetCreationDate} (get)
 * file-info:file-create-time           {@link #fileInfoGetCreationTime} (get)
 * file-info:file-mod-date              {@link #fileInfoGetModDate}      (get)
 * file-info:file-mod-time              {@link #fileInfoGetModTime}      (get)
 * file-info:file-name                  {@link #initFileInfo}            (set)
 * file-info:file-name                  {@link #getFileName}             (get)
 * file-info:file-size                  {@link #fileInfoGetSize}         (get)
 * file-info:file-type                  {@link #fileInfoGetType}         (get)
 * file-info:full-pathname              {@link #fileInfoGetFullPath}     (get)
 * file-info:pathname                   {@link #fileInfoGetPathName}     (get)
 * </pre>
 * <p>
 * This class implements the Progress interface/semantic. However the actual
 * implementation of this function is provided by a network object that
 * implements the {@link FileSystem} interface.  This means that the actual
 * operations occur on the remote node associated with this user's session.
 * Please see {@link FileSystemDaemon} and the <code>RemoteObject</code> in
 * the <code>net</code> package.
 */
public class FileSystemOps
{
   /** File name separator used in jar file entries */
   private static final char JAR_SEP = '/';
   
   /** Application package root path */
   private static final String pkgRootPath =
      Utils.getDirectoryNodeString(null, "legacy-system/pkgroot", "")
           .replace('.', JAR_SEP) + JAR_SEP;
   
   /** Jar classloader cache. */
   private static Map<String, JarClassLoader> jclCache = new ConcurrentHashMap<>();
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();
   
   /**
    * Get a the instance for the FILE-INFO system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link CommonFileInfo} interface and its methods implemented by these
    * classes: {@link FileSystemOps}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = work.obtain();
      
      Class[] classes = { FileSystemOps.class };
      CommonFileInfo proxy = StaticProxy.obtain(CommonFileInfo.class, classes);
      
      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }
      
      return new handle(proxy);
   }   
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = work.obtain();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * Push the PROPATH values to the client side.
    */
   public static void init()
   {
      work.obtain();
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Get the error code of the last operating system operation.
    *
    * @return   The Progress-compatible error code.
    */
   public static integer getLastError()
   {
      return new integer(work.obtain().lastError);
   }
   
   /**
    * Sets the error code of the last operating system operation.
    *
    * @param    err
    *           The Progress-compatible error code.
    */
   public static void setLastError(int err)
   {
      work.obtain().lastError = err;
   }
   
   /**
    * Search for the specified file and initialize the related data
    * structures if found.  Relative filenames are searched for via the 
    * PROPATH.  If <code>name</code> is <code>unknown</code> or not found,
    * the current data structures are cleared.
    *
    * @param    name
    *           Relative or absolute filename to obtain information on.
    */
   public static void initFileInfo(character name)
   {
      if (name == null || name.isUnknown())
      {
         work.obtain().info = null;
      }
      
      initFileInfo(name.toStringMessage());
   }
   
   /**
    * Search for the specified file and initialize the related data
    * structures if found.  Relative filenames are searched for via the 
    * {@link EnvironmentOps#getSearchPath}.  If <code>name</code> is
    * <code>unknown</code> or not found, the current data structures are
    * cleared.
    *
    * @param    name
    *           Relative or absolute filename to obtain information on.
    */
   public static void initFileInfo(String name)
   {
      WorkArea wa = work.obtain();
      wa.info = (name == null) ? null : wa.fs.accessFileInfo(name);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file name, otherwise return the <code>unknown value</code>.
    *
    * @return   The file name of the current file being inspected as input
    *           by the caller (even if the file could not be found) or the
    *           <code>unknown value</code> if the data structures were never
    *           initialized.
    */
   public static character getFileName()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.name == null)
      {
         return new character();
      }
      
      return new character(wa.info.name);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the absolute or relative path name, otherwise return the
    * <code>unknown value</code>. If the specified file name was absolute,
    * then an absolute pathname will be returned.  Otherwise the relative
    * path that was found via {@link #searchPath} will be returned.
    *
    * @return   The path name of the current file being inspected or the
    *           <code>unknown value</code> if the data structures were never
    *           initialized or if initialization failed.
    */
   public static character fileInfoGetPathName()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.pathname == null)
      {
         return new character();
      }
      
      return new character(wa.info.pathname);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the absolute (canonicalized) path name, otherwise return the
    * <code>unknown value</code>.
    *
    * @return   The absolute path name of the current file being inspected or
    *           the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   public static character fileInfoGetFullPath()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.fullname == null)
      {
         return new character();
      }
      
      return new character(wa.info.fullname);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file's size, otherwise return the <code>unknown value</code>.
    *
    * @return   The size in bytes of the current file being inspected or the
    *           <code>unknown value</code> if the data structures were never
    *           initialized or if initialization failed.
    */
   public static integer fileInfoGetSize()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.size == -1)
      {
         return new integer();
      }
      
      return new integer(wa.info.size);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * a string with characters that specify attributes about the file,
    * otherwise return the <code>unknown value</code>.
    * <p>
    * One of the following mutually exclusive characters will be present:
    * <p>
    * <pre>
    * Character   Meaning
    * ---------   -----------------------------
    * D           directory
    * F           file
    * X           unknown
    * </pre>
    * <p>
    * <b>There is no support for "S" (UNIX special devices) or "M" (procedure
    * library members) at this time.</b>
    * <p>
    * One or more of the following characters will be present:
    * <p>
    * <pre>
    * Character   Meaning
    * ---------   -----------------------------
    * H           hidden
    * R           readable
    * W           writable
    * </pre>
    * <p>
    * <b>There is no support for "L" (UNIX symbolic links) or "P" (pipes)
    * at this time.</b>
    *
    * @return   The file attributes of the current file being inspected or
    *           the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   public static character fileInfoGetType()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.type == null)
      {
         return new character();
      }
      
      return new character(wa.info.type);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file's last modification date, otherwise return the 
    * <code>unknown value</code>.
    *
    * @return   The last modification date of the current file being inspected
    *           or the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   public static date fileInfoGetModDate()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.modification == -1)
      {
         return new date();
      }
      
      return new date(date.neutralMillisToDate(wa.info.modification, null));
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file's last modification time (in seconds from midnight), otherwise 
    * return the <code>unknown value</code>.
    *
    * @return   The last modification time in seconds from midnight of the
    *           current file being inspected or the <code>unknown value</code>
    *           if the data structures were never initialized or if
    *           initialization failed.
    */
   public static integer fileInfoGetModTime()
   {
      WorkArea wa = work.obtain();
      
      if (wa.info == null || wa.info.modification == -1)
      {
         return new integer();
      }
      
      // turn the modification millis into a specific date (this drops the
      // time value and leaves just the date as a julian day number)
      Date cd     = date.neutralMillisToDate(wa.info.modification, null);
      date create = new date(cd);
      
      // convert the julian day number into a base millis value for that day
      long base = date.julianDayToMillis(create.longValue());
      
      // get the difference and convert to seconds
      return new integer((wa.info.modification - base) / 1000);
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file's creation date, otherwise return the 
    * <code>unknown value</code>.
    *
    * @return   Always <code>unknown value</code> at this time.
    */
   public static date fileInfoGetCreationDate()
   {
      // the creation date is only given on windows, on Unix unknown value 
      // will be returned.
      if (PlatformHelper.isUnderWindowsFamily())
      {
         WorkArea wa = work.obtain();
         
         // if creation is -1 error occurred
         if (wa.info == null || wa.info.creation == -1)
         {
            return new date();
         }
         Date createDate = new Date(wa.info.creation);
         
         return new date(createDate);
      }
      else
      {
         return new date();
      }
   }
   
   /**
    * If the current file information data structures are initialized, return
    * the file's creation time (in seconds from midnight), otherwise 
    * return the <code>unknown value</code>.
    *
    * @return   Always returns <code>unknown value</code> at this time.
    */
   public static integer fileInfoGetCreationTime()
   {
      // the creation time is only given on windows, on Unix no value
      // will be returned.
      if (PlatformHelper.isUnderWindowsFamily())
      {
         WorkArea wa = work.obtain();
         // if creation is -1 error occurred
         if (wa.info == null || wa.info.creation == -1)
         {
            return new integer();
         }
         Date createDate = new Date(wa.info.creation);
         GregorianCalendar calendar = new GregorianCalendar();
         calendar.setTime(createDate);
         return new integer(date.secondsSinceMidnight(calendar));
      }
      else
      {
         return new integer();
      }
   }
   
   /**
    * Get the comma-separated list of filesystem roots (drive letters on
    * platforms that support the concept, otherwise returns the equivalent
    * of the empty string).
    *
    * @return   The comma-separated list of drive letters.
    */
   public static character getRootList()
   {
      String roots = null;
      
      try
      {
         roots = work.obtain().fs.getRootList();
      }
      catch (IOException ioe)
      {
         setLastError(FileSystem.ERR_IO);
         roots = "";
      }
      
      return new character(roots); 
   }
   
   /**
    * Get the system property (user-defined properties can be specified at the Java command line)
    * associated with the passed key.  If the specified system property does not exist,
    * {@code unknown} value is returned.
    *
    * @return   The system property found or {@code unknown} if the property does not exist.
    */
   public static character getProperty(character key)
   {
      if (key == null || key.isUnknown())
      {
         return new character();
      }
      
      return getProperty(key.toStringMessage());
   }
   
   /**
    * Get the system property (user-defined properties can be specified at the Java command line)
    * associated with the passed key.  If the specified system property does not exist,
    * {@code unknown} value is returned.
    *
    * @return   The system property found or {@code unknown} if the property does not exist.
    */
   public static character getProperty(String key)
   {
      if (key == null)
      {
         return new character();
      }
      
      WorkArea wa = work.obtain();
      
      try
      {
         String transformedKey = key;
         
         if (!wa.legacyCaseSensitive)
         {
            transformedKey = key.toUpperCase(Locale.ROOT);
         }
         
         String value = wa.environmentVariables.get(transformedKey);
         
         if (value != null)
         {
            return new character(value);
         }
         
         //try interceptors
         Function<String, String> interceptor = wa.keysInterceptors.get(transformedKey);
         
         if (interceptor != null)
         {
            key = interceptor.apply(transformedKey);
         }
         
         // fall back to the client system properties
         return new character(wa.fs.getProperty(key));
      }
      catch (Exception exc)
      {
         // return the unknown value if a null pointer, security or illegal argument occurs
         return new character();
      }
   }
   
   /**
    * Appends the specified source file to the end of the target file. If
    * any error occurs, no exception will be thrown. The {@link #getLastError}
    * can be used to retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name or
    *           if this is a non-existent filename, this method is converted
    *           into a copy operation. 
    */
   public static void append(character src, character tar)
   {
      if (src == null || src.isUnknown() || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         append(src.toStringMessage(), tar.toStringMessage());
      }
   }
   
   /**
    * Appends the specified source file to the end of the target file. If
    * any error occurs, no exception will be thrown. The {@link #getLastError}
    * can be used to retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name or
    *           if this is a non-existent filename, this method is converted
    *           into a copy operation. 
    */
   public static void append(character src, String tar)
   {
      if (src == null || src.isUnknown() || tar == null)
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         append(src.toStringMessage(), tar);
      }
   }
   
   /**
    * Appends the specified source file to the end of the target file. If
    * any error occurs, no exception will be thrown. The {@link #getLastError}
    * can be used to retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name or
    *           if this is a non-existent filename, this method is converted
    *           into a copy operation. 
    */
   public static void append(String src, character tar)
   {
      if (src == null || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         append(src, tar.toStringMessage());
      }
   }
   
   /**
    * Appends the specified source file to the end of the target file. If
    * any error occurs, no exception will be thrown. The {@link #getLastError}
    * can be used to retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name or
    *           if this is a non-existent filename, this method is converted
    *           into a copy operation. 
    */
   public static void append(String src, String tar)
   {
      WorkArea wa = work.obtain();
      
      if (src == null || tar == null)
      {
         wa.lastError = FileSystem.ERR_NO_ERROR;
      }
      else
      {
         wa.lastError = wa.fs.copy(src, tar, false);
      }
   }
   
   /**
    * Copies the specified source file to or over the target file. This
    * method always overwrites the target file.  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    */
   public static void copy(character src, character tar)
   {
      if (src == null || src.isUnknown() || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         copy(src.toStringMessage(), tar.toStringMessage());
      }
   }
   
   /**
    * Copies the specified source file to or over the target file. This
    * method always overwrites the target file.  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    */
   public static void copy(character src, String tar)
   {
      if (src == null || src.isUnknown() || tar == null)
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         copy(src.toStringMessage(), tar);
      }
   }
   
   /**
    * Copies the specified source file to or over the target file. This
    * method always overwrites the target file.  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    */
   public static void copy(String src, character tar)
   {
      if (src == null || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         copy(src, tar.toStringMessage());
      }
   }
   
   /**
    * Copies the specified source file to or over the target file. This
    * method always overwrites the target file.  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file.  Must NOT be a directory name.
    * @param    tar
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    */
   public static void copy(String src, String tar)
   {
      WorkArea wa = work.obtain();
      
      if (src == null || tar == null)
      {
         wa.lastError = FileSystem.ERR_NO_ERROR;
      }
      else
      {
         wa.lastError = wa.fs.copy(src, tar, true);
      }
   }
   
   /**
    * Creates each of the directories requested, if it doesn't exist and all
    * path segments up to the last segment do exist.  If the directory name
    * already exists in the filesystem (even if it is a file), this method
    * just returns (this is the Progress behavior).  If any error occurs, no
    * exception will be thrown, but no subsequent directories in the list
    * will be created. The {@link #getLastError} can be used to retrieve
    * the error code.
    *
    * @param    dirs
    *           The list of directories to create.
    */
   public static void mkdir(String[] dirs)
   {
      if (dirs == null)
      {
         return;
      }
      
      WorkArea wa = work.obtain();
      wa.lastError = wa.fs.mkdir(dirs);
   }
   
   /**
    * Deletes each of the directories or files specified.  If a non-empty
    * directory name is specified, an error will occur unless the
    * <code>recursive</code> flag is <code>true</code>.  If any error occurs,
    * no exception will be thrown, the deletion process just continues. The
    * {@link #getLastError} can be used to retrieve the error code.
    *
    * @param    list
    *           The list of files and directories to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting the entire subtree.
    */
   public static void delete(String[] list, boolean recursive)
   {
      if (list == null)
      {
         return;
      }
      
      WorkArea wa = work.obtain();
      wa.lastError = wa.fs.delete(list, recursive);
   }
   
   /**
    * Get the file contents from the client-side file system.
    *
    * @param  filename
    *         Name of the file in the client-side file system (relatively to the current
    *         directory on the client side).
    *
    * @return file contents as array of bytes.
    *
    * @throws java.io.IOException
    *         If file cannot not be found or read, or is too large.
    */
   public static byte[] getFileContents(String filename)
   throws IOException
   {
      WorkArea wa = work.obtain();
      return wa.fs.getFileContents(filename);
   }
   
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file or directory.
    * @param    tar
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    */
   public static void rename(character src, character tar)
   {
      if (src == null || src.isUnknown() || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         rename(src.toStringMessage(), tar.toStringMessage());
      }
   }
   
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file or directory.
    * @param    tar
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    */
   public static void rename(character src, String tar)
   {
      if (src == null || src.isUnknown() || tar == null)
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         rename(src.toStringMessage(), tar);
      }
   }
   
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file or directory.
    * @param    tar
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    */
   public static void rename(String src, character tar)
   {
      if (src == null || tar == null || tar.isUnknown())
      {
         setLastError(FileSystem.ERR_NO_ERROR);
      }
      else
      {
         rename(src, tar.toStringMessage());
      }
   }
   
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown. The {@link #getLastError} can be used to
    * retrieve the error code.
    *
    * @param    src
    *           The name of the source file or directory.
    * @param    tar
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    */
   public static void rename(String src, String tar)
   {
      WorkArea wa = work.obtain();
      
      if (src == null || tar == null)
      {
         wa.lastError = FileSystem.ERR_NO_ERROR;
      }
      else
      {
         wa.lastError = wa.fs.rename(src, tar);
      }
   }
   
   /**
    * Search each directory listed in the {@link EnvironmentOps#getSearchPath} for the given file.
    * If the file system object exists and is a file, then the filename is returned. If the path
    * contains an empty string or a '.', then the current directory is searched, otherwise only
    * the listed directories are searched.  
    * <p>
    * Special processing is provided in case the original Progress code was checking for the
    * availability of an external procedure.  In this case the filename is checked using the
    * {@link SourceNameMapper#convertName} method. If such the Java class exists, its legacy
    * filename is returned.  This will "fake" the converted code into thinking that the external
    * procedure this still exists.
    *
    * @param   filename
    *          The absolute or relative filename for which to search.
    *
    * @return  If the file is found in the current directory, the relative filename is returned.
    *          If the file is absolute or found in any other path, the full absolute filename is
    *          returned.  If the file is not found or the referenced file system object is not a
    *          file, the {@code unknown value} is returned.
    */
   public static character searchPath(character filename)
   {
      if (filename == null || filename.isUnknown())
      {
         return new character();
      }
      
      return searchPath(filename.toStringMessage());
   }
   
   /**
    * Search each directory listed in the {@link EnvironmentOps#getSearchPath} for the given file.
    * If the file system object exists and is a file, then the filename is returned. If the path
    * contains an empty string or a '.', then the current directory is searched, otherwise only
    * the listed directories are searched.  
    * <p>
    * Special processing is provided in case the original Progress code was checking for the
    * availability of an external procedure.  In this case the filename is checked using the
    * {@link SourceNameMapper#convertName} method. If such the Java class exists, its legacy
    * filename is returned.  This will "fake" the converted code into thinking that the external
    * procedure this still exists.
    * <p>
    * Application resource jars are searched before the remote file system.
    *
    * @param   filename
    *          The absolute or relative filename for which to search.
    *
    * @return  If the file is found in the current directory, the relative filename is returned.
    *          If the file is absolute or found in any other path, the full absolute filename is
    *          returned.  If the file is not found or the referenced file system object is not a
    *          file, the {@code unknown value} is returned.
    */
   public static character searchPath(String filename)
   {
      if (filename == null)
      {
         return null;
      }
      
      // if the Progress code was checking for the availability of an external procedure, we
      // have to "fake" it out into thinking that this still exists, try to match this given
      // filename with a known program name, optionally honoring any absolute prefix passed in
      String legacy = SourceNameMapper.lookupLegacyName(filename);
      if (legacy != null)
      {
         // this is a program name, return the Java class name instead
         return new character(resolveLegacyPath(legacy));
      }
      
      // check resource jar files first, before checking remote file system
      JarResource jarRes = searchResourceJars(filename, null);
      
      String result = (jarRes != null ? jarRes.getName() : work.obtain().fs.searchPath(filename));
      
      return (result == null ? new character() : new character(result));
   }
   
   /**
    * Search the resource jar files for the resource with the given filename, appending the
    * extensions provided in the optional list, if present.
    * <p>
    * The algorithm first attempts to find an exact match for the filename. If not found, it
    * applies the PROPATH to the filename, first prepending the package root to the name.
    * 
    * @param   filename
    *          Filename of the resource, optionally including a path.
    * @param   extensions
    *          Optional list of extensions to apply to the name. If this list is non-null,
    *          {@code filename} is assumed to be the file's base name, sans extension.
    * 
    * @return  A {@code JarResult} object, if the search succeeds, else {@code null}.
    */
   public static JarResource searchResourceJars(String filename, List<String> extensions)
   {
      final char WIN_SEP = '\\';
      final String JAR_SEP_STR = "/";
      boolean caseSens = EnvironmentOps.getLegacyCaseSensitive();
      
      // the first thing to do working with jar base filename is replace all Windows separators
      filename = filename.replace(WIN_SEP, JAR_SEP);
      
      // resource name as found in jar file
      String resName = null;
      
      // class loader for jar file in which resource is found
      JarClassLoader jarClassLoader = null;
      
      // get class loader for application jar
      ClassLoader cl = ClassLoader.getSystemClassLoader();
      if (cl != null)
      {
         // first clean up initial string
         
         // possible current directory symbol
         int len = filename.length();
         if (len > 0 && filename.charAt(0) == '.')
         {
            filename = filename.substring(1);
         }
         
         // try to consider filenames starting with / as current from PROPATH entries
         if (len > 0 && filename.charAt(0) == JAR_SEP)
         {
            filename = filename.substring(1);
         }
         
         String clientOS = EnvironmentOps.getRuntimeOperatingSystem().toJavaType();
         String legacyOS = EnvironmentOps.getLegacyPlatform();
         
         PathResolver pathResolver = null;
         
         if (!legacyOS.equalsIgnoreCase(clientOS))
         {
            String fileSeparator = EnvironmentOps.getLegacyFileSeparator();
            
            Map<String, String> pathMap = EnvironmentOps.getLegacyAbsolutePathPrefixMap();
            
            pathResolver = Utils.getPathResolver(fileSeparator, caseSens, pathMap);
         }
         
         PathResolverContainer safePathResolver = new PathResolverContainer(pathResolver);
         
         String[] propathHeader = null;
         
         // walk jars marked as resource jars in the directory
         outer:
         for (String path : StandardServer.getResourceJarPaths())
         {
            String jarFullName = MultiClassLoader.resolveJarName(path, true);
            
            // have the valid jar name
            if (jarFullName != null)
            {
               // get jar file loader from cache
               jarClassLoader = getJarClassLoader(jarFullName, cl.getParent());
               
               // not possible to get jar file, may be the wrong name?
               if (jarClassLoader == null)
               {
                  continue;
               }
               
               JarClassLoader jcl = jarClassLoader;
               resName = getRefinedPathName(filename, JAR_SEP_STR);
               Function<String, String> coreFn = n -> jcl.getResourceName(n, caseSens);
               
               // if no extensions were provided, do a straight search for the resource name
               // in the jar; otherwise, concatenate each extension to the base resource name
               // before searching the jar, and continue until we have a hit, or no more
               // extensions to try
               Function<String, String> searchFn =
                  extensions == null
                  ? coreFn
                  : n -> Utils.tryWhileNull(extensions, e -> coreFn.apply(n.concat(e)));
               
               // first try the search with no PROPATH modifications
               resName = searchFn.apply(resName);
               
               if (resName != null)
               {
                  break;
               }
               
               if (propathHeader == null)
               {
                  propathHeader = EnvironmentOps.getSearchPathList();
               }
               
               // not found in root continue with PROPATH entries
               for (String header : propathHeader)
               {
                  header = safePathResolver.resolvePath(header);
                  
                  // look through PROPATH related locations, clean up possible Windows
                  // separators
                  header = header.replace(WIN_SEP, JAR_SEP);
                  
                  // inside jar we need to use UNIX-style file separator even on Windows
                  String name = getRefinedPathName(
                     pkgRootPath + header + JAR_SEP + filename, JAR_SEP_STR);
                  
                  // try to search new name based on PROPATH header
                  resName = searchFn.apply(name);
                  if (resName != null)
                  {
                     break outer;
                  }
               }
            }
         }
      }
      
      if (jarClassLoader != null && resName != null)
      {
         return new JarResource(jarClassLoader, resName);
      }
      
      return null;
   }
   
   /**
    * Parses a reference to a member of an ABL r-code library and returns the simple member name.
    *
    * @param  filename
    *         Pathname of the file in the r-code library.
    *
    * @return The function parses a character string in the form
    *         <code>path-name&lt;&lt;member-name&gt;&gt;</code>, where <code>path-name</code> is
    *         the pathname of a library and <code>member-name</code> is the name of a file within
    *         the library, and returns <code>member-name</code>.
    */
   public static character member(String filename)
   {
      return member(new character(filename));
   }

   /**
    * Parses a reference to a member of an ABL r-code library and returns the simple member name.
    *
    * @param  filename
    *         Pathname of the file in the r-code library.
    *
    * @return The function parses a character string in the form
    *         <code>path-name&lt;&lt;member-name&gt;&gt;</code>, where <code>path-name</code> is
    *         the pathname of a library and <code>member-name</code> is the name of a file within
    *         the library, and returns <code>member-name</code>.
    */
   public static character member(character filename)
   {
      UnimplementedFeature.unsupported("MEMBER function in not implemented.");
      return new character();
   }

   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Get the current working directory.
    * 
    * @return  Current working directory.
    */
   public static character getWorkingDirectory()
   {
      String dir = work.obtain().fs.getWorkingDirectory();
      
      return new character(dir);
   }
   
   /**
    * Set the current working directory. Does nothing if argument represents unknown value.
    * 
    * @param   workDir
    *          New working directory.
    */
   public static void setWorkingDirectory(character workDir)
   {
      if (workDir.isUnknown())
      {
         return;
      }
      
      work.obtain().fs.setWorkingDirectory(workDir.toStringMessage());
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is used for
    * the Progress compatible environment. This corresponds to the assignment of
    * the "PROPATH" variable in Progress and represents a mechanism to override
    * the default that is available from the directory.
    * 
    * @param    searchPath
    *           The new PROPATH value.
    */
   private static void setSearchPath(String searchPath)
   {
      // split the list
      String[] list = EnvironmentOps.getSearchPathList();
      
      if (list != null)
      {
         // use the resulting propath to set run time value on the client
         work.obtain().fs.setSearchPath(list, EnvironmentOps.getLegacyCaseSensitive());
      }
   }
   
   /**
    * Gets jar file class loader to obtain resources. If not in current cache - the new
    * loader is created and stored in cache.
    * 
    * @param    jarFullName
    *           The fully qualified jar file name to load.  It must be correct refined name.
    * @param    parentCl
    *           The class loader to be used as parent class loaded when new class loader should
    *           be created.
    *
    * @return   The instance of the JarClassLoader to use in resource loading.
    */
   private static JarClassLoader getJarClassLoader(String jarFullName, ClassLoader parentCl)
   {
      // first try to reuse the cache 
      JarClassLoader jclRes = jclCache.get(jarFullName);
      
      // not in cache, create and store for further usage
      if (jclRes == null)
      {
         jclRes = new JarClassLoader(jarFullName, parentCl);
         
         // loader created, initialize and store into the cache
         jclRes.init();
         jclCache.putIfAbsent(jarFullName, jclRes);
      }
      
      return jclRes;
   }
   
   /**
    * Gets image path string to be cleared from invalid sequences and file separator duplication.
    * Handles properly current('.') and uplevel('..') directory entries and empty spaces.  No case
    * letter changes are performing.
    * 
    * @param    nameDirty
    *           The not yet refined image path string.
    * @param    separator
    *           The string representing file separator to use in result.
    *
    * @return   Properly formatted case sensitive image path string.
    */
   private static String getRefinedPathName(String nameDirty, String separator)
   {
      // parse the dirty path into array by separator to clean for multiple separators
      String[] splitPath = nameDirty.split(separator);
      int pathSize = splitPath.length;
      
      // walk through the entries to clean up
      for (int i = 0; i < pathSize; i++)
      {
         // process next path name entry
         String name = splitPath[i];
         
         // skip the current directory, empty directory, directory or files with more than 2 dots
         // in beginning
         if (name.equals(".") || name.isEmpty() || (name.startsWith("..") && !name.equals("..")))
         {
            // go one directory level up
            for (int j = i; j < pathSize - 1; j++)
            {
               splitPath[j] = splitPath[j + 1];
            }
            // one entry has been removed for current dir
            pathSize--;
            i--;
            
            continue;
         }
         
         // try to handle parent directory shift
         if (name.equals(".."))
         {
            // for the first and last entries just ignore this - no parent directory
            if (i > 0 && i < pathSize - 1)
            {
               // go one directory level up
               for (int j = i; j < pathSize - 1; j++)
               {
                  splitPath[j - 1] = splitPath[j + 1];
               }
               i--;
            }
            // two entries have been removed, one for '..' and parent directory
            pathSize -= 2;
            continue;
         }
         
      }
      
      // second pass - build the final buffer
      StringBuilder sb = new StringBuilder();
      // at this time we must have clean array
      for (int i = 0; i < pathSize; i++)
      {
         sb.append(splitPath[i]);
         
         // for all entries except last add the separator
         if (i < pathSize - 1)
         {
            sb.append(separator);
         }
      }
      
      return sb.toString();
   }

   /**
    * Transforms the given legacy path into the valid client system path.
    * 
    * @param    legacyPath
    *           The legacy path name to the target resource
    * 
    * @return   The resolved host path on the client system
    */
   private static String resolveLegacyPath(String legacyPath)
   {
      if (legacyPath == null || legacyPath.isEmpty())
      {
         return legacyPath;
      }
      
      WorkArea wa = work.obtain();
      
      if (wa.clientOS.equalsIgnoreCase("unix"))
      {
         return legacyPath.replaceAll(FileSystem.ANY_PATH_SEPARATOR, "/");
      }
      else
      {
         return legacyPath.replaceAll(FileSystem.ANY_PATH_SEPARATOR, "\\");
      }
   }

   /**
    * Object returned by a search for a resource in a jar file.
    */
   public static class JarResource
   {
      /** Class loader for the jar */
      private final JarClassLoader loader;
      
      /** Fully qualified name of the resource in the jar file */
      private final String name;
      
      /**
       * Constructor.
       * 
       * @param   loader
       *          Class loader for the jar.
       * @param   name
       *          Fully qualified name of the resource in the jar file.
       */
      public JarResource(JarClassLoader loader, String name)
      {
         this.loader = loader;
         this.name = name;
      }
      
      /**
       * Get the jar class loader.
       * 
       * @return  Jar class loader.
       */
      public JarClassLoader getLoader()
      {
         return loader;
      }
      
      /**
       * Get the fully qualified name of the resource in the jar file.
       * 
       * @return  Resource name.
       */
      public String getName()
      {
         return name;
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** The resource's ID. */
      private Long id = null;
      
      /** Stores the Progress error code for the last operation. */ 
      private int lastError = 0; 
      
      /** File information for the current file being examined. */ 
      private FileInfo info = null; 
      
      /** Remote instance of our file system worker. */ 
      private FileSystem fs = null;
      
      /** The legacy OS case sensitive mode */
      private boolean legacyCaseSensitive;
      
      /** The environment variables map */
      private Map<String, String> environmentVariables;
      
      /** The client OS */
      private String clientOS;
      
      /** The client property key interceptors */
      private Map<String, Function<String, String>> keysInterceptors;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal
   { 
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return (WorkArea) this.get();
      }
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized Object initialValue()
      {
         WorkArea wa = new WorkArea();
         
         // obtain the our proxy, server-side code is left-side by convention
         wa.fs = (FileSystem) RemoteObject.obtainInstance(FileSystem.class, true);
         
         // push down our search path to ensure that it is always there
         String[] paths    = EnvironmentOps.getSearchPathList();
         boolean  caseSens = EnvironmentOps.getLegacyCaseSensitive();
         
         wa.legacyCaseSensitive = caseSens;
         
         // listener is added last... we already push the PROPATH above, so no need to notify;
         // plus, it would lead to a double-initialization of this ContextContainer, as the
         // listener (setSearchPath) uses the WorkArea.fs
         EnvironmentOps.addSourcePathListener(FileSystemOps::setSearchPath);
         
         String fileSeparator = EnvironmentOps.getLegacyFileSeparator();
         
         Map<String, String> pathMap = EnvironmentOps.getLegacyAbsolutePathPrefixMap();
         
         String clientOS = EnvironmentOps.getRuntimeOperatingSystem().toJavaType();
         
         wa.clientOS = clientOS;
         
         String legacyOS = EnvironmentOps.getLegacyPlatform();
         
         // if legacy OS platform is not the same as client OS platform, then legacy path names
         // must be resolved on the client OS.
         boolean areLegacyPathsNotResolved = !legacyOS.equalsIgnoreCase(clientOS);
         
         wa.fs.setLegacyFileSystemParameters(fileSeparator,
                                             caseSens,
                                             pathMap,
                                             paths,
                                             areLegacyPathsNotResolved);
         
         wa.environmentVariables = EnvironmentOps.getLegacyEnvironmentVariables(!caseSens);
         
         // add client property keys interceptors
         
         wa.keysInterceptors = new HashMap<>();
         
         Function<String, String> usernameInterceptor = new Function<String, String>()
         {
            /**
             * Returns the OS dependent client user key if it applies to the legacy user key.
             * 
             * @param    key
             *           The given legacy user key
             * 
             * @return   The OS dependent client user key
             */
            @Override
            public String apply(String key)
            {
               if (wa.clientOS.equalsIgnoreCase("unix"))
               {
                  return "USER";
               }
               
               return "USERNAME";
            }
         };
         
         wa.keysInterceptors.put("USER", usernameInterceptor);
         wa.keysInterceptors.put("USERNAME", usernameInterceptor);
         
         return wa;
      }
   }
}