/*
** Module   : LowLevelSocketListener.java
** Abstract : An interface defining server-socket related APIs which execute on P2J client-side.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130902 Created initial version.
** 002 HC  20131031 Changed resource ID data type to Long, see issue #2183.
** 003 OM  20151104 Added support for SSL-connection related errors.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.IOException;
import java.security.*;

/**
 * Interface for Server-Socket APIs which need to be implemented on the P2J client-side.
 */
public interface LowLevelSocketListener
{
   /**
    * Create a server socket and let it listen on the given port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    port
    *           The port number.
    * @param    qsize
    *           The backlog size for this socket.
    * 
    * @return   The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *           the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while starting the server.
    */
   public int enableConnections(long resourceId, int port, int qsize)
   throws IOException;

   /**
    * Create a SSL server socket and let it listen on the given port.
    * 
    * @param   resourceId
    *          The ID of the associated {@link SocketImpl} resource on server side.
    * @param   port
    *          The port number.
    * @param   qsize
    *          The backlog size for this socket.
    * @param   keyalias
    *          The alias for the digital certificate.
    * @param   keyaliaspasswd
    *          The password for the key of the digital certificate.
    * @param   nosessioncache
    *          Flag indicating the SSL client session cache is disabled.
    * @param   sessiontimeout
    *          The number of seconds before the SSL client request is timeout.
    *
    * @return  The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *          the connection was established.
    *
    * @throws  IOException
    *          In case of errors while starting the server socket.
    * @throws  GeneralSecurityException
    *          In case of SSL-related errors while starting the server socket. 
    */
   public int enableSSLConnections(long    resourceId,
                                   int     port,
                                   int     qsize,
                                   String  keyalias,
                                   String  keyaliaspasswd,
                                   boolean nosessioncache,
                                   int     sessiontimeout)
   throws IOException, 
          GeneralSecurityException;

   /**
    * Terminates the listening loop for this server socket. The existing established connections
    * are not terminated.
    * 
    * @param    id
    *           The ID of {@link LowLevelSocketListenerImpl} instance on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   public boolean disableConnections(int id);
}