/*
** Module   : MemoryBuffer.java
** Abstract : Representation of a memptr value, when used from non-P2J code. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130626 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * Provides a byte buffer to be used from non-P2J code, when performing remote appserver calls.
 * Instances of this class will be sent when the remote side expectes a memptr instance.
 * <p>
 * The remote side will be responsible of converting instances of this class to a {@link memptr}
 * instance when it receives a {@link MemoryBuffer} instance and also to send back to the 
 * requester a {@link MemoryBuffer} instance, in case OUTPUT or INPUT-OUTPUT mode is used for the 
 * parameter.
 */
public final class MemoryBuffer
implements Externalizable
{
   /** Stores the state of this instance as an array of bytes. */
   private byte[] value;
   
   /**
    * Default c'tor, needed for deserialization purposes; not for public use.
    */
   public MemoryBuffer()
   {
   }
   
   /**
    * Construct a new instance of this class, by saving a copy of the passed byte array.
    * <p>
    * Any changes made to the passed array will not be reflected in this instance.
    * 
    * @param    value
    *           The array of bytes representing the state of this instance.
    *           May be <code>null</code>.
    */
   public MemoryBuffer(byte[] value)
   {
      this.value = getCopy(value);
   }

   /**
    * Change the {@link #value array of bytes} of this instance by saving a copy of the passed 
    * array.
    * <p>
    * Any changes made to the passed array will not be reflected in this instance.
    * 
    * @param    value
    *           The array of bytes representing the new state of this instance.
    *           May be <code>null</code>.
    */
   public void setValue(byte[] value)
   {
      this.value = getCopy(value);
   }

   /**
    * Get the array of bytes for this instance, as a copy.
    * <p>
    * Any changes made to the returned array will not be reflected in this instance.
    * 
    * @return   See above.
    */
   public byte[] getValue()
   {
      return getCopy(value);
   }
   
   /**
    * Write the state of this instance to the specified stream.
    * 
    * @param    out
    *           The stream where the state of this instance will be sent.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(value != null);
      out.writeObject(value);
   }

   /**
    * Initialize this instance by reading its state from the specified stream.
    * 
    * @param    in
    *           The stream from where the state of this instance will be read.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      boolean hasData = in.readBoolean();
      value = (byte[]) (hasData ? in.readObject() : null);
   }
   
   /**
    * Create a copy of the given array of bytes.
    * 
    * @param    bytes
    *           The array of bytes for which a copy is needed.  May be <code>null</code>.
    *         
    * @return   If the given byte array is <code>null</code>, return <code>null</code>. Else,
    *           return a copy of the array.
    */
   private byte[] getCopy(byte[] bytes)
   {
      byte[] copy = null;
      
      if (bytes != null)
      {
         copy = Arrays.copyOf(bytes, bytes.length);
      }
         
      return copy;
   }
}