/*
** Module   : PDFPageSetup.java
** Abstract : Captures PDF options for PDF printing.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20171024 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.util;

import org.apache.pdfbox.pdmodel.common.*;

/**
 * Captures PDF options for PDF printing.
 */
public class PDFPageSetup
{
   /** The conversion from PDF user-space points to millimeters. */
   private static final float POINTS_PER_MM = 2.8346457F;

   /** Page size in PDF user-space points. */
   public PDRectangle pageSize;

   /** Top page margin in millimeters. */
   public float topMargin;

   /** Left page margin in millimeters. */
   public float leftMargin;

   /** Bottom page margin in millimeters. */
   public float bottomMargin;

   /** Right page margin in millimeters. */
   public float rightMargin;

   /**
    * Constructs new instance from {@link PrintOptions}.
    *
    * @param   options
    *          An instance of {@link PrintOptions}.
    *
    * @return  A new instance of this class.
    */
   public static PDFPageSetup fromPrintOptions(PrintOptions options)
   {
      PDRectangle size;
      switch (options.pageSize)
      {
         case LETTER:
            size = PDRectangle.LETTER;
            break;
         case LEGAL:
            size = PDRectangle.LEGAL;
            break;
         case A0:
            size = PDRectangle.A0;
            break;
         case A1:
            size = PDRectangle.A1;
            break;
         case A2:
            size = PDRectangle.A2;
            break;
         case A3:
            size = PDRectangle.A3;
            break;
         case A5:
            size = PDRectangle.A5;
            break;
         case A6:
            size = PDRectangle.A6;
            break;
         default:
            size = PDRectangle.A4;
      }

      if (options.pageOrientation == PrintOptions.PageOrientation.LANDSCAPE)
      {
         size = new PDRectangle(size.getHeight(), size.getWidth());
      }

      // convert margins from millimeters to user-space points
      float topMargin = options.marginTop * POINTS_PER_MM;
      float leftMargin = options.marginLeft * POINTS_PER_MM;
      float bottomMargin = options.marginBottom * POINTS_PER_MM;
      float rightMargin = options.marginRight * POINTS_PER_MM;

      return new PDFPageSetup(size, topMargin, leftMargin, bottomMargin, rightMargin);
   }

   /**
    * Ctor.
    *
    * @param   pageSize
    *          Page size.
    * @param   topMargin
    *          Top margin.
    * @param   leftMargin
    *          Left margin.
    * @param   bottomMargin
    *          Bottom margin.
    * @param   rightMargin
    *          Right margin.
    */
   public PDFPageSetup(PDRectangle pageSize,
                       float topMargin,
                       float leftMargin,
                       float bottomMargin,
                       float rightMargin)
   {
      this.pageSize = pageSize;
      this.topMargin = topMargin;
      this.leftMargin = leftMargin;
      this.bottomMargin = bottomMargin;
      this.rightMargin = rightMargin;
   }
}
