/*
** Module   : PathResolverContainer.java
** Abstract : PathResolverContainer resolves a legacy system absolute file path into its host
**            system file path by the dedicated path resolved and logs calmly if a given path
**            can't be resolved. 
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SBI 20180522 It catches unresolved path exceptions and logs them at the Level.FINER.
** 002 ECF 20180613 Performance fixes. Avoid constructing log message text when not logging;
**                  minimize overhead of log level check.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.util;

import java.util.logging.*;


/**
 * Delegates the path resolution to the underlined path resolver and logs unresolved exception
 * as warnings.
 */
public class PathResolverContainer
implements PathResolver
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(PathResolverContainer.class.getName());
   
   /** Is logging level FINER enabled? */
   private static boolean logFiner = LOG.isLoggable(Level.FINER);
   
   /** The underlined path resolver */
   private final PathResolver delegate;
   
   /**
    * Creates the nop resolver.
    */
   public PathResolverContainer()
   {
      this(null);
   }
   
   /**
    * Creates this path resolver.
    * 
    * @param    delegate
    *           The underlined path resolver
    */
   public PathResolverContainer(PathResolver delegate)
   {
      this.delegate = delegate;
   }
   
   /**
    * Resolves the given path by the underlined path resolver, otherwise returns the given
    * unresolved value.
    * 
    * @param    path
    *           The given path to be resolved
    */
   @Override
   public String resolvePath(String path)
   {
      try
      {
         if (delegate != null)
         {
            return delegate.resolvePath(path);
         }
      }
      catch (UnresolvedPathException ex)
      {
         // this will be hit A LOT, so we don't want to dynamically check logging level every
         // time
         if (logFiner)
         {
            LOG.log(Level.FINER, ex.toString() + path);
         }
      }
      
      return path;
   }
}