/*
** Module   : PersistentProcedure.java
** Abstract : Defines common APIs used by the PROCEDURE handle.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130221 Added ADM-DATa and UNIQUE-ID support.
** 003 CA  20130222 Moved here procedure-only attributes/methods from CommonProcedure.
** 004 CS  20130222 Added conversion support for CURRENT-WINDOW attribute.
** 005 CA  20130223 Added SET-CALLBACK-PROCEDURE method support.
** 006 CA  20130529 Made this resource Deletable too.
** 007 CA  20130813 Adding async request handle support: added ASYNC-REQUEST-COUNT attribute and
**                  extracted SERVER attribute in a separate interface. 
** 008 CA  20130925 Added support for PUBLISHED-EVENTS attribute.
** 009 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 010 EVL 20131002 Adding legacy attributes and methods annotation.
** 011 CA  20131113 Added INSTANTIATING-PROCEDURE support.
** 012 VIG 20140109 Changed signature of currentWindow() method.
** 013 GES 20141124 Javadoc fix.
** 014 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.ui.*;

/**
 * Exposes the following methods or attributes, common for PROCEDURE handles:
 * <table summary="">
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>INTERNAL-ENTRIES</td>
 *    <td>{@link #internalEntries()}</td>
 * </tr>
 * <tr>
 *    <td>PERSISTENT</td>
 *    <td>{@link #isPersistent()}</td>
 * </tr>
 * <tr>
 *    <td>CURRENT-WINDOW</td>
 *    <td>{@link #currentWindow()} and 
 *        {@link #setCurrentWindow}
 *    </td>
 * </tr>
 * <tr>
 *    <td>GET-SIGNATURE()</td>
 *    <td>{@link #getSignature(character)} and {@link #getSignature(String)}
 *    </td>
 * </tr>
 * <tr>
 *    <td>PROXY</td>
 *    <td>{@link #isProxy()}</td>
 * </tr>
 * <tr>
 *    <td>TRANSACTION</td>
 *    <td>{@link #getTransaction()}</td>
 * </tr>
 * <tr>
 *    <td>SET-CALLBACK-PROCEDURE</td>
 *    <td>{@link #setCallbackProcedure}</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapProcedure} API, which, depending on the 
 * handle's state and type, it either returns an 
 * {@link ExternalProgramWrapper} instance, or a dynamic proxy using 
 * {@link handle#invalidAttrAccessProxy}. 
 */
@LegacyResource(resource = LegacyResource.PROCEDURE)
public interface PersistentProcedure
extends CommonHandleChain,
        CommonProcedure,
        CommonFile,
        ADMData,
        UniqueID,
        Deletable,
        Remotable,
        AsyncRequestCountAttribute,
        InstantiatingProcedure
{
   /**
    * Get the <code>internal-entries</code> attribute of this procedure instance.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "INTERNAL-ENTRIES")
   public character internalEntries();

   /**
    * Get the <code>persistent</code> attribute of this procedure instance.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PERSISTENT")
   public logical isPersistent();

   /**
    * Execute the <code>get-signature</code> method of this procedure instance.
    * 
    * @param    internalEntry
    *           The legacy name of the internal-entry for which the signature
    *           is needed. 
    */
   @LegacyMethod(name = "GET-SIGNATURE")
   public character getSignature(String internalEntry);

   /**
    * Execute the <code>get-signature</code> method of this procedure instance.
    * 
    * @param    internalEntry
    *           The legacy name of the internal-entry for which the signature
    *           is needed. 
    */
   @LegacyMethod(name = "GET-SIGNATURE")
   public character getSignature(character internalEntry);
   
   /**
    * Get CURRENT-WINDOW attribute for this procedure handle.
    * 
    * @return  Reference to current window widget.
    */
   @LegacyAttribute(name = "CURRENT-WINDOW")
   public handle currentWindow();

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * 
    * @param    h
    *           A handle for the window.
    */
   @LegacyAttribute(name = "CURRENT-WINDOW", setter = true)
   public void setCurrentWindow(handle h);

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * 
    * @param    win
    *           Window reference.
    */
   @LegacyAttribute(name = "CURRENT-WINDOW", setter = true)
   public void setCurrentWindow(WindowWidget win);
   
   /**
    * Indicates if the procedure is a proxy procedure.
    * 
    * @return   <code>true</code> if the procedure is a proxy procedure ,
    *           <code>false</code> otherwise.
    */
   @LegacyAttribute(name = "PROXY")
   public logical isProxy(); 
   
   /**
    * Returns a handle to the current opened transaction.
    * 
    * @return   See above
    */
   @LegacyAttribute(name = "TRANSACTION")
   public handle getTransaction();

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(String event, String proc);
   
   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(character event, String proc);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(String event, character proc);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(character event, character proc);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(String event, String proc, handle context);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(character event, String proc, handle context);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(String event, character proc, handle context);

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @LegacyMethod(name = "SET-CALLBACK-PROCEDURE")
   public logical setCallbackProcedure(character event, character proc, handle context);
   
   /**
    * Get the named events published by this procedure.  This is a comma separated list of
    * named events collected from the PUBLISH statements, where the named event is a hard-coded
    * string, not a variable.
    * 
    * @return    The named events published by this procedure, or the empty string if none found.
    */
   @LegacyAttribute(name = "PUBLISHED-EVENTS")
   public character getPublishedEvents();
}