/*
** Module   : RemoteMultiplexer.java
** Abstract : Interface for classes needing remote services multiplexed by an
**            arbitrary key
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20071018 ADD  @35537  Created initial version. Interface for
**                               classes needing remote services multiplexed
**                               by an arbitrary key.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Interface which is implemented by classes which must associate a generic
 * key with an integral identifier, intended to enable the multiplexing of
 * remote service requests.
 * <p>
 * We multiplex on an integral ID primarily for performance reasons, since
 * this interface is intended to support remote object multiplexing:
 * <ul>
 *   <li>The ID is expected to travel across the network with every service
 *       request.  Using an integer is much more efficient than serializing
 *       an arbitrary object.
 *   <li>Using an integer allows for a very efficient lookup strategy to be
 *       employed by implementors.  For instance, the ID can be implemented as
 *       an index in an array-based list, providing a much more efficient (and
 *       simpler) lookup strategy than say, a hash table with an arbitrary
 *       object key.
 * </ul>
 * 
 * @param   <K>
 *          Type of key to be provided, which is to be mapped to an integral
 *          ID.
 * 
 * @author  ECF
 */
public interface RemoteMultiplexer<K>
{
   /**
    * Retrieve a unique ID for the given key.  The ID is guaranteed to be
    * unique within the context of the current server instance and must map
    * consistently to the given key.  The intent is that the ID will be used
    * for subsequent service requests which must be multiplexed by the
    * logical construct represented by the key.
    * 
    * @param   key
    *          An object for which a unique ID is needed.
    *          
    * @return  ID associated with the given key, which is unique for the
    *          server.
    *          
    * @throws  IllegalArgumentException
    *          if <code>key</code> cannot be mapped to an ID on this server.
    */
   public int getMultiplexID(K key);
}