/*
** Module   : UniqueIdGenerator.java
** Abstract : Generator for ids used as UNIQUE-ID attribute values.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20130822 Created initial version.
** 002 CA  20131029 Added QUERY to the list of resources supporting UNIQUE-ID. 
** 003 HC  20131215 Added more to the list of resources supporting UNIQUE-ID. 
** 004 EVK 20131216 Added SAX objects to the list of resources supporting UNIQUE-ID, removed
*                   SAX_ATTRIBUTES and SAX_READER from IdKind enum.
** 005 CA  20140107 UNIQUE-ID for SOAP-HEADER and SOAP-HEADER-ENTRY use the same counter. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.security.*;
import java.util.*;

/**
 * Generator for ids used as UNIQUE-ID attribute values. Generates values that unique across
 * specific kind of entities (temp-tables, buffers, buffer fields etc.) in a specific user
 * context. Note that behavior of this generation may not always match 4GL behavior.
 *
 */
public class UniqueIdGenerator
{
   /**
    * Context-local map of kinds of entities to the next available unique id for this kind.
    */
   private final static ContextLocal<Map<IdKind, integer>> context =
         new ContextLocal<Map<IdKind, integer>>()
         {
            public Map<IdKind, integer> initialValue()
            {
               Map<IdKind, integer> map = new HashMap<IdKind, integer>();
               for (IdKind kind : IdKind.values())
               {
                  map.put(kind, new integer(1));
               }
               return map;
            }
         };

   /**
    * Get new unique id for the specified kind of entities (see {@link IdKind}) in the current
    * user context.
    *
    * @param   kind
    *          Kind of entities for which the id is generated.
    *
    * @return  See above.
    */
   public static integer getUniqueId(IdKind kind)
   {
      Map<IdKind, integer> map = context.get();
      integer id = map.get(kind);
      integer res = new integer(id);

      id.increment();
      return res;
   }

   /**
    * Kind of entities (generates values are unique only across specific kind of entities).
    * <ul>
    * <li>{@link #TEMP_TABLE} -.</li>
    * <li>{@link #BUFFER} - .</li>
    * <li>{@link #BUFFER_FIELD} - .</li>
    * <li>{@link #QUERY} - .</li>
    * <li>{@link #PROCEDURE} - .</li>
    * <li>{@link #SAX_OBJECT} - Any kind of sax objects: sax-reader, sax-writer, sax-attributes.
    * </li>
    * <li>{@link #X_DOCUMENT} - .</li>
    * <li>{@link #X_NODEREF} - .</li>
    * <li>{@link #SOAP_HEADER} - Both SOAP-HEADER and SOAP-HEADER-ENTRY resources.</li>
    * </ul>
    */
   public enum IdKind
   {
      TEMP_TABLE,
      BUFFER,
      BUFFER_FIELD,
      QUERY,
      PROCEDURE,
      SAX_OBJECT,
      X_DOCUMENT,
      X_NODEREF,
      SOAP_HEADER
   }
}