/*
** Module   : WebHelper.java
** Abstract : Helper methods for web integration.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20180123 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.net.*;
import java.io.*;
import java.nio.charset.*;

/**
 * Provides helper methods for web integration.
 */
public class WebHelper
{
   /** UTF-8 character set name. */
   private static final String UTF8 = StandardCharsets.UTF_8.toString();
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.  It is backed by the J2SE
    * <code>URLEncoder</code>.
    *
    * @param    data
    *           The text to encode.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded by <code>URLEncoder.encode()</code>.
    */
   public static character urlEncode(character data)
   {
      if (data == null || data.isUnknown())
      {
         return new character("");
      }
      
      return urlEncode(data.toStringMessage());
   }
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.  It is backed by the J2SE
    * <code>URLEncoder</code>.
    *
    * @param    data
    *           The text to encode.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded by <code>URLEncoder.encode()</code>.
    */
   public static character urlEncode(String data)
   {
      return urlEncode(data, "java");
   }
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.
    *
    * @param    data
    *           The text to encode.
    * @param    type
    *           "java" (for <code>URLEncoder.encode()</code>), "query", "cookie", "default"
    *           (same as "") are the standard types.  Anything else will be interpreted as
    *           a custom set of characters to add to the list of characters needing replacement.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded.
    */
   public static character urlEncode(character data, character type)
   {
      if (data == null || data.isUnknown())
      {
         return new character("");
      }
      
      String typ = (type == null || type.isUnknown()) ? "java" : type.toStringMessage();
      
      return urlEncode(data.toStringMessage(), typ);
   }
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.
    *
    * @param    data
    *           The text to encode.
    * @param    type
    *           "java" (for <code>URLEncoder.encode()</code>), "query", "cookie", "default"
    *           (same as "") are the standard types.  Anything else will be interpreted as
    *           a custom set of characters to add to the list of characters needing replacement.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded.
    */
   public static character urlEncode(String data, character type)
   {
      String typ = (type == null || type.isUnknown()) ? "java" : type.toStringMessage();
      
      return urlEncode(data, typ);
   }
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.
    *
    * @param    data
    *           The text to encode.
    * @param    type
    *           "java" (for <code>URLEncoder.encode()</code>), "query", "cookie", "default"
    *           (same as "") are the standard types.  Anything else will be interpreted as
    *           a custom set of characters to add to the list of characters needing replacement.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded.
    */
   public static character urlEncode(character data, String type)
   {
      if (data == null || data.isUnknown())
      {
         return new character("");
      }
      
      return urlEncode(data.toStringMessage(), (type == null) ? "" : type);
   }
   
   /**
    * Encode the given text to be safe for usage in a URL. This is primarily used for
    * encoding a parameter for safe insertion into a URL.
    *
    * @param    data
    *           The text to encode.
    * @param    type
    *           "java" (for <code>URLEncoder.encode()</code>), "query", "cookie", "default"
    *           (same as "") are the standard types.  Anything else will be interpreted as
    *           a custom set of characters to add to the list of characters needing replacement.
    *
    * @return   Empty string if given the empty string or <code>null</code>, otherwise the text
    *           text as encoded.
    */
   public static character urlEncode(String data, String type)
   {
      if (data == null || data.length() == 0)
      {
         return new character("");
      }
      
      String result = data;
      
      // TODO: add compatible implementation for the non-java forms
      
      // the "java" form doesn't exist in the original WebSpeed implementation
      try
      {
         result = URLEncoder.encode(data, UTF8);
      }
      
      catch (UnsupportedEncodingException uee)
      {
         // this should never happen since the encoding is fixed
      }
      
      return new character(result);
   }
}