/*
** Module   : MethodInfo.java
** Abstract : Information useful to the assembly of a method invocation.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20101119 Created initial version.
** 002 ECF 20150422 Minor format cleanup.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

/**
 * This class encapsulates information necessary to assemble instructions for
 * the invocation of a method, such as the method's name, owner, descriptor,
 * and list of declared exceptions.
 * 
 * @author ECF
 */
final class MethodInfo
{
   /** Method's owner class */
   private final String className;
   
   /** Method name */
   private final String methodName;
   
   /** JVM-internal method descriptor */
   private final String descriptor;
   
   /** List of declared exception names */
   private final String[] exceptions;
   
   /**
    * Convenience constructor for a method which declares no exceptions.
    * 
    * @param   className
    *          Method's owner class.
    * @param   methodName
    *          Method name.
    * @param   descriptor
    *          Internal method descriptor.
    */
   MethodInfo(String className, String methodName, String descriptor)
   {
      this(className, methodName, descriptor, new String[] {});
   }
   
   /**
    * Constructor.
    * 
    * @param   className
    *          Method's owner class.
    * @param   methodName
    *          Method name.
    * @param   descriptor
    *          Internal method descriptor.
    * @param   exceptions
    *          List of declared exception names.
    */
   MethodInfo(String className, String methodName, String descriptor, String[] exceptions)
   {
      this.className = className;
      this.methodName = methodName;
      this.descriptor = descriptor;
      this.exceptions = exceptions;
   }
   
   /**
    * Get owner class name.
    * 
    * @return  Owner class name.
    */
   String getClassName()
   {
      return className;
   }
   
   /**
    * Get method name.
    * 
    * @return  Method name.
    */
   String getMethodName()
   {
      return methodName;
   }
   
   /**
    * Get internal method descriptor.
    * 
    * @return  Method descriptor.
    */
   String getDescriptor()
   {
      return descriptor;
   }
   
   /**
    * Get list of declared exceptions.
    * 
    * @return  Exception names.
    */
   String[] getExceptions()
   {
      return exceptions;
   }
}