/*
** Module   : ProxyAssembler.java
** Abstract : Assembles a dynamic proxy subclass.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20101119 Created initial version.
** 002 ECF 20101216 Pass null to InvocationHandler.invoke() for arguments
**                  parameter, instead of an empty Object array, to match
**                  contract for this method. Generate different code, based
**                  on whether the superclass implements InvocationHandler.
** 003 ECF 20130122 Fix to enable nested interfaces to be used to define the
**                  proxy. Removed unnecessary and incorrect try/catch block
**                  from generated proxy method, which was improperly throwing
**                  UndeclaredThrowableException.
** 004 ECF 20150406 Upgraded to ASM 5.0.3.
** 005 ECF 20160219 Added support for an optional plugin to override the assembly of arbitrary
**                  methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

import java.lang.reflect.*;
import java.util.*;
import org.objectweb.asm.*;
import com.goldencode.asm.*;

/**
 * A class which assembles dynamic proxy classes using the ASM bytecode
 * engineering library, v5.0.3 (http://asm.ow2.org).  As a package private
 * class, it is intended to be invoked only by the {@link ProxyFactory}, which
 * exposes a public API for the definition of new proxy objects.
 * <p>
 * This mechanism is used in preference to the dynamic proxy implementation
 * exposed by <code>java.lang.reflect.Proxy</code> because the latter does not
 * allow one to specify a base class to extend.  In situations where a common
 * interface and implementation combination is used for many dynamically
 * specified classes, this can cause memory problems.  Specifically, the
 * amount of PermGen memory required for <code>java.lang.reflect.Method</code>
 * objects is duplicated across each proxy class which uses the common
 * interface/implementation, because the methods in the common interface are
 * duplicated in every proxy class.
 * <p>
 * The approach used by this assembler is to allow the specification of a base
 * class which implements the common interface, so that these common methods
 * are defined only in a single class, rather than duplicated across every
 * dynamically defined proxy class.  A user would define all common method
 * implementations in this base class, which each proxy class would then
 * inherit "for free".  All unique methods are still specified by an array of
 * interfaces passed to this factory's static <code>getProxy</code> method.
 * <p>
 * The architecture of this dynamic proxy implementation retains the use of
 * <code>java.lang.reflect.InvocationHandler</code>.  All calls to the methods
 * specified in the array of interfaces passed to <code>getProxy</code> are
 * routed to the invocation handler, as with the standard, dynamic proxy
 * implementation.  This requires that an <code>InvocationHandler</code>
 * instance is accessible to the runtime instance of the dynamically generated
 * class.  In the standard implementation, the base class of all proxy classes
 * is <code>java.lang.reflect.Proxy</code>, which holds a reference to the
 * invocation handler.  However, since there are no restrictions or contracts
 * regarding the base class in this implementation, the invocation handler
 * is stored in each proxy instance.  This adds a single, redundant field to
 * each proxy class definition, which is the trade-off for being able to
 * define all common methods in the base class of the user's choice.  So long
 * as there is at least one such common method in the base class to offset the
 * presence of this redundant field, this implementation should be more more
 * memory efficient than the standard proxy implementation.
 * <p>
 * An optional (@link ProxyAssemblerPlugin plugin} can be provided to the
 * constructor. The plugin may override the assembly of any method, and may
 * implement additional methods after the primary proxy methods have been
 * assembled.
 * <p>
 * <strong>Please note that generics and annotations are not supported at this
 * time.</strong>
 * <p>
 * Note:  the CGLib bytecode engineering library was considered as a simpler
 * alternative to this implementation;  however, the closest substitute in
 * that library (the Enhancer class) automatically overrides all non-final
 * methods of the base class in order to instrument callbacks in the
 * overridden methods.  This behavior was considered too heavyweight.  A
 * workaround would have been to require all such methods be marked as final
 * to avoid this override behavior;  however, this requirement was rejected.
 * <p>
 * The following pseudo-code representation is intended to convey the general
 * structure of the generated proxy class.  However, please note that no
 * intermediate source code representation is produced.
 * <p>
 * <pre>
 * public final class [ProxyClassName]
 * implements [Interface List]
 * {
 *    // list of delegate methods
 *    private static Method[] ma;
 *    
 *    // invocation handler
 *    private final InvocationHandler h;
 *    
 *    public [ProxyClassName](InvocationHandler handler)
 *    {
 *       this.h = handler;
 *    }
 *    
 *    public static void __initializeMethods(Method[] methods)
 *    {
 *       if (ma != null)
 *       {
 *          throw new IllegalStateException(
 *             "Proxy methods may not be set more than once");
 *       }
 *       
 *       [ProxyClassName].ma = methods;
 *    }
 *    
 *    ...
 *    
 *    [implemented methods -- see inner class ProxyMethod for description]
 *    
 *    ...
 * }
 * </pre>
 * 
 * @author ECF
 */
public final class ProxyAssembler
implements Opcodes
{
   /** Name of static method to initialize methods upon loading proxy class */
   static final String INIT_METHODS_METHOD = "__initializeMethods";
   
   /** No-op default plugin */
   private static final ProxyAssemblerPlugin defaultPlugin = new ProxyAssemblerPlugin()
   {
      public Set<Method> getImplementedMethods() { return Collections.emptySet(); }
      
      public void assemble(ClassWriter classWriter) { }
      
      public ProxyAssembler.ProxyMethod getProxyMethodAssembler(Method method) { return null; }
   };
   
   /** Name of the proxy class to be assembled */
   private final String proxyName;
   
   /** Name of the proxy class' parent class */
   private final String parentName;
   
   /** Interfaces which the proxy class will implement */
   private final Class<?>[] interfaces;
   
   /** Method objects which will be passed to invocation handler */
   private final Method[] methods;
   
   /** Plugin used to augment the final proxy class */
   private final ProxyAssemblerPlugin plugin;
   
   /** Does parent class implement <code>InvocationHandler</code>? */
   private final boolean parentIsHandler;
   
   /** Object used to create class fields, methods, etc. */
   private final ClassWriter classWriter =
      new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
   
   /**
    * Constructor.
    * 
    * @param   proxyName
    *          Name of the proxy class to be assembled. This is the external
    *          name known to the class loader.
    * @param   parentName
    *          Name of the proxy class' parent class. This is the external
    *          name known to the class loader. This class must not use
    *          generics or annotations.
    * @param   interfaces
    *          Interfaces which the proxy class will implement. These
    *          interfaces must not use generics or annotations.
    * @param   methods
    *          Methods to be proxied. Must not use generics or annotations.
    * @param   plugin
    *          Optional plugin. May be <code>null</code>.
    * @param   parentIsHandler
    *          <code>true</code> if the parent class implements the <code>InvocationHandler</code>
    *          interface, else <code>false</code>.  If <code>true</code>, no separate instance
    *          field will be assembled to store the invocation handler reference, as the instance
    *          of this class itself will act as the handler.
    */
   ProxyAssembler(String proxyName,
                  String parentName,
                  Class<?>[] interfaces,
                  Method[] methods,
                  ProxyAssemblerPlugin plugin,
                  boolean parentIsHandler)
   {
      if (plugin == null)
      {
         plugin = defaultPlugin;
      }
      
      // remove plugin's methods (if any) from methods to be implemented by this object
      Set<Method> pluginMethods = plugin.getImplementedMethods();
      if (!pluginMethods.isEmpty())
      {
         Set<Method> allMethods = new LinkedHashSet<>();
         for (Method m : methods)
         {
            allMethods.add(m);
         }
         allMethods.removeAll(pluginMethods);
         methods = allMethods.toArray(new Method[allMethods.size()]);
      }
      
      this.proxyName = AsmUtils.commonToInternalTypeName(proxyName);
      this.parentName = AsmUtils.commonToInternalTypeName(parentName);
      this.interfaces = interfaces;
      this.methods = methods;
      this.plugin = plugin;
      this.parentIsHandler = parentIsHandler;
   }
   
   /**
    * Get the final list of proxied methods, which may differ from the methods passed to this
    * object's constructor, due to plugin intervention.
    * 
    * @return  Final list of proxied methods.
    */
   Method[] getProxiedMethods()
   {
      // don't bother making a copy; by the time this is invoked, we don't need the list anymore
      return methods;
   }
   
   /**
    * Assemble the class in accordance with the Java class file format as
    * specified in <u>The Java Virtual Machine Specification, Second
    * Edition</u> by Tim Lindholm and Frank Yellin.
    * <p>
    * See the class description above for details regarding the content of
    * the assembled class.
    * 
    * @return  Byte array containing the class definition.
    */
   byte[] assembleClass()
   {
      // gather the list of interfaces being implemented
      String[] ifaceNames = new String[interfaces.length];
      for (int i = 0; i < interfaces.length; i++)
      {
         Class<?> iface = interfaces[i];
         String name = iface.getName();
         ifaceNames[i] = AsmUtils.commonToInternalTypeName(name);
      }
      
      // create the proxy class
      classWriter.visit(V1_7,
                        ACC_PUBLIC | ACC_FINAL,
                        proxyName,
                        null,
                        parentName,
                        ifaceNames);
      
      // add a class member field for the array of delegate Methods
      classWriter.visitField(ACC_PRIVATE | ACC_STATIC | ACC_FINAL,
                             "ma",
                             "[Ljava/lang/reflect/Method;",
                             null,
                             null);
      
      if (!parentIsHandler)
      {
         // add an instance member field for the InvocationHandler
         classWriter.visitField(ACC_PRIVATE | ACC_FINAL,
                                "h",
                                "Ljava/lang/reflect/InvocationHandler;",
                                null,
                                null);
      }
      
      // create the __initializeMethods method
      assembleInitializeMethods();
      
      // create a constructor
      if (parentIsHandler)
      {
         assembleDefaultConstructor();
      }
      else
      {
         assembleHandlerConstructor();
      }
      
      // generate each proxy method
      ProxyMethod defaultProxyMethod = new ProxyMethod();
      for (int i = 0; i < methods.length; i++)
      {
         Method method = methods[i];
         ProxyMethod pluginProxyMethod = plugin.getProxyMethodAssembler(method);
         ProxyMethod proxyMethod = pluginProxyMethod != null
                                   ? pluginProxyMethod
                                   : defaultProxyMethod;
         String desc = ProxyUtils.makeMethodDescriptor(method);
         MethodVisitor mv =
            classWriter.visitMethod(ACC_PUBLIC,
                                    method.getName(),
                                    desc,
                                    null,
                                    ProxyUtils.getCaughtExceptionTypeNames(method));
         
         mv.visitCode();
         proxyMethod.assemble(classWriter, mv, method, i, parentIsHandler, proxyName);
         mv.visitMaxs(0, 0);
         mv.visitEnd();
      }
      
      // let the plugin do its work
      plugin.assemble(classWriter);
      
      // complete the assembly
      classWriter.visitEnd();
      
      // get the byte array that represents the class file
      byte[] code = classWriter.toByteArray();
      
      return code;
   }
   
   /**
    * Assemble the <code>__initializeMethods(Method[])</code> method.  This is
    * a static method which is called once, by the {@link ProxyFactory},
    * immediately after loading the class.  It is assembled such that it will
    * throw an <code>IllegalStateException</code> if it is invoked again.
    */
   private void assembleInitializeMethods()
   {
      MethodVisitor mv =
         classWriter.visitMethod(ACC_PUBLIC | ACC_STATIC,
                                 INIT_METHODS_METHOD,
                                 "([Ljava/lang/reflect/Method;)V",
                                 null,
                                 null);
      
      // get the current value of the "ma" static field
      mv.visitFieldInsn(GETSTATIC,
                        proxyName,
                        "ma",
                        "[Ljava/lang/reflect/Method;");
      
      // if "ma" is still null, skip over the next set of instructions (which
      // otherwise throw IllegalStateException)
      Label jumpTo = new Label();
      mv.visitJumpInsn(IFNULL, jumpTo);
      
      // if we did not skip over this next set of instructions, the "ma" field
      // already was set;  it is invalid to call this method to set it again,
      // so throw an IllegalStateException
      mv.visitTypeInsn(NEW, "java/lang/IllegalStateException");
      mv.visitInsn(DUP);
      mv.visitLdcInsn("Proxy methods may not be set more than once");
      mv.visitMethodInsn(INVOKESPECIAL,
                         "java/lang/IllegalStateException",
                         "<init>",
                         "(Ljava/lang/String;)V",
                         false);
      mv.visitInsn(ATHROW);
      
      // this is where we jump to if all is well...
      mv.visitLabel(jumpTo);
      
      // load the method's single argument
      mv.visitVarInsn(ALOAD, 0);
      
      // store the array of methods into the "ma" static field
      mv.visitFieldInsn(PUTSTATIC,
                        proxyName,
                        "ma",
                        "[Ljava/lang/reflect/Method;");
      
      // return
      mv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      mv.visitMaxs(3, 1);
   }
   
   /**
    * Assemble the proxy class' single, default constructor.
    */
   private void assembleDefaultConstructor()
   {
      MethodVisitor mv =
         classWriter.visitMethod(ACC_PUBLIC,
                                 "<init>",
                                 "()V",
                                 null,
                                 null);
      
      // load implicit "this" variable
      mv.visitVarInsn(ALOAD, 0);
      
      // invoke superclass' default constructor
      mv.visitMethodInsn(INVOKESPECIAL, parentName, "<init>", "()V", false);
      
      // return from c'tor
      mv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      mv.visitMaxs(1, 1);
   }
   
   /**
    * Assemble the proxy class' single constructor, which accepts and stores
    * the <code>InvocationHandler</code> to which method calls will be
    * dispatched.
    */
   private void assembleHandlerConstructor()
   {
      MethodVisitor mv =
         classWriter.visitMethod(ACC_PUBLIC,
                                 "<init>",
                                 "(Ljava/lang/reflect/InvocationHandler;)V",
                                 null,
                                 null);
      
      // load implicit "this" variable
      mv.visitVarInsn(ALOAD, 0);
      
      // invoke superclass' default constructor
      mv.visitMethodInsn(INVOKESPECIAL, parentName, "<init>", "()V", false);
      
      // load implicit "this" variable
      mv.visitVarInsn(ALOAD, 0);
      
      // load argument (InvocationHandler) passed to c'tor
      mv.visitVarInsn(ALOAD, 1);
      
      // store it in the "h" instance variable
      mv.visitFieldInsn(PUTFIELD,
                        proxyName,
                        "h",
                        "Ljava/lang/reflect/InvocationHandler;");
      
      // return from c'tor
      mv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      mv.visitMaxs(2, 2);
   }
   
   /**
    * Helper class to assemble a proxy method.  The proxy method delegates its
    * invocation to an instance of <code>java.lang.reflect.Method</code> in
    * the <code>ma</code> array in the containing class.
    * <p>
    * Pseudo-code for the generated method is:
    * <pre>
    * public ReturnType methodName([ArgumentList])
    * [throws ExceptionList]
    * {
    *    return (ReturnType)
    *       [Handler].invoke(this,
    *                        ma[index],
    *                        [ArgumentList (as Object[])] or null);
    * }
    * </pre>
    * <p>
    * where:
    * <ul>
    * <li><code>index</code> is an index into the enclosing class' array of
    *     delegate <code>Method</code> objects <code>ma</code>;
    * <li><code>Handler</code> is the "this" pointer if the proxy itself
    *     is the invocation handler, or the "h" instance field if a separate
    *     invocation handler is stored within the proxy.
    * </ul>
    * <p>
    * For methods which return <code>void</code>, the <code>return</code>
    * statement and cast to <code>ReturnType</code> are omitted.  In the event
    * of an empty argument list, an empty array of type <code>Object</code>
    * is passed to the <code>InvocationHandler.invoke</code> method.
    */
   public static class ProxyMethod
   {
      /**
       * Default constructor.
       */
      protected ProxyMethod()
      {
      }
      
      /**
       * Assemble the bytecode which comprises the proxy method.
       * 
       * @param   classWriter
       *          Class writer for proxy class.
       * @param   mv
       *          Method visitor for this method.
       * @param   method
       *          Method for which this proxy is being implemented.
       * @param   methodIndex
       *          Index of this method within the array of methods used by the invocation handler.
       * @param   parentIsHandler
       *          <code>true</code> if the parent class implements the
       *          <code>InvocationHandler</code> interface, else <code>false</code>.  If
       *          <code>true</code>, no separate instance field will be assembled to store the
       *          invocation handler reference, as the instance of this class itself will act as
       *          the handler.
       * @param   proxyName
       *          Internal form of the proxy class' name.
       */
      public void assemble(ClassWriter classWriter,
                           MethodVisitor mv,
                           Method method,
                           int methodIndex,
                           boolean parentIsHandler,
                           String proxyName)
      {
         // load invocation handler from field "h" or as "this" if the proxy
         // itself is the invocation handler
         mv.visitVarInsn(ALOAD, 0);
         if (!parentIsHandler)
         {
            mv.visitFieldInsn(GETFIELD,
                              proxyName,
                              "h",
                              "Ljava/lang/reflect/InvocationHandler;");
         }
         
         // load "this" object ref as the first parameter for later call to
         // InvocationHandler.invoke
         mv.visitVarInsn(ALOAD, 0);
         
         // load target method from array of methods in static field "ma"
         mv.visitFieldInsn(GETSTATIC,
                           proxyName,
                           "ma",
                           "[Ljava/lang/reflect/Method;");
         
         // target method is at index "index" in the "ma" array
         AsmUtils.pushInt(mv, methodIndex);
         mv.visitInsn(AALOAD);
         
         // create new array of Objects for method's arguments or push null
         // if there are no arguments
         Class<?>[] parmTypes = method.getParameterTypes();
         Integer numArgs = new Integer(parmTypes.length);
         if (numArgs > 0)
         {
            AsmUtils.pushInt(mv, numArgs);
            mv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
         }
         else
         {
            mv.visitInsn(ACONST_NULL);
         }
         
         // in preparation for marshaling our parameters, start the local
         // variable pool index at 1 (0 is reserved for the "this" reference)
         int localIndex = 1;
         
         // prepare arguments for InvocationHandler.invoke call
         for (int i = 0; i < parmTypes.length; i++)
         {
            TypeInfo parmInfo = ProxyUtils.getTypeInfo(parmTypes[i]);
            int width = parmInfo.getWidth();
            
            // duplicate array reference on top of stack
            mv.visitInsn(DUP);
            
            // push the target index in the array
            AsmUtils.pushInt(mv, i);
            
            // load the value from the local variable pool onto the stack
            mv.visitVarInsn(parmInfo.getLoadInsn(), localIndex);
            
            // if parameter is a primitive, it must be wrapped
            MethodInfo wrapInfo = parmInfo.getWrapInfo();
            if (wrapInfo != null)
            {
               mv.visitMethodInsn(INVOKESTATIC,
                                  wrapInfo.getClassName(),
                                  wrapInfo.getMethodName(),
                                  wrapInfo.getDescriptor(),
                                  false);
            }
            
            // store Object parameter into array
            mv.visitInsn(AASTORE);
            
            // update the current index in the variable pool
            localIndex += width;
         }
         
         // call InvocationHandler.invoke
         mv.visitMethodInsn(INVOKEINTERFACE,
                            "java/lang/reflect/InvocationHandler",
                            "invoke",
                            "(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;",
                            true);
         
         // determine method's return type
         Class<?> returnType = method.getReturnType();
         TypeInfo returnInfo = ProxyUtils.getTypeInfo(returnType);
         
         // cast return type to expected type, unless return type is void
         // (indicated by null cast name) or is java/lang/Object
         String castName = returnInfo.getCastName();
         if (castName != null && !"java/lang/Object".equals(castName))
         {
            mv.visitTypeInsn(CHECKCAST, returnInfo.getCastName());
         }
         
         // extract primitive from java wrapper type, if necessary
         MethodInfo unwrapInfo = returnInfo.getUnwrapInfo();
         if (unwrapInfo != null)
         {
            mv.visitMethodInsn(INVOKEVIRTUAL,
                               unwrapInfo.getClassName(),
                               unwrapInfo.getMethodName(),
                               unwrapInfo.getDescriptor(),
                               false);
         }
         
         // if return type is void, we have to pop the value returned by the
         // invocation handler off the stack before we return
         if (castName == null)
         {
            mv.visitInsn(POP);
         }
         
         // return;  even a void method requires a return instruction
         mv.visitInsn(returnInfo.getReturnInsn());
      }
   }
}