/*
** Module   : ProxyAssemblerPlugin.java
** Abstract : Interface for a ProxyAssembler plugin.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20160219 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

import java.lang.reflect.*;
import java.util.*;
import org.objectweb.asm.*;

/**
 * A plugin which can override the assembly of any proxy method and which can implement
 * additional, arbitrary methods. The plugin must report which methods it is taking the
 * responsibility to implement.
 */
public interface ProxyAssemblerPlugin
{
   /**
    * Report the set of methods which this plugin will implement. This set will not be
    * implemented by the main proxy assembler. This method is invoked before the main proxy
    * assembler begins assembling the proxy class.
    * <p>
    * Note that any method in the returned set will NOT be included in the array of methods the
    * invocation handler will have access to at runtime. Therefore, the returned set should NOT
    * include any method handled specifically by {@link #getProxyMethodAssembler(Method)}.
    * {@link #getProxyMethodAssembler(Method)} will not be invoked for any of the methods in the
    * returned set.
    * 
    * @return  Set of methods which this plugin will implement. May be empty, but may not be
    *          <code>null</code>.
    */
   public Set<Method> getImplementedMethods();
   
   /**
    * Return an object which will assemble the given proxy method, or <code>null</code> if this
    * plugin will allow the default proxy method implementation. If a non-<code>null</code>
    * value is returned, the corresponding method should NOT be included in the set of methods
    * reported by {@link #getImplementedMethods()}.
    * <p>
    * This method is invoked for each proxy method implemented by the main proxy assembler.
    * 
    * @param   method
    *          Method for which a proxy implementation is needed.
    * 
    * @return  A subclass of <code>ProxyAssembler.ProxyMethod</code> or <code>null</code> if the
    *          default proxy method implementation is suitable for the given method.
    * 
    * @see     #getImplementedMethods()
    */
   public ProxyAssembler.ProxyMethod getProxyMethodAssembler(Method method);
   
   /**
    * Entry point into this plugin's arbitrary method assembly process. This will be called after
    * {@link #getImplementedMethods()} and after {@link #getProxyMethodAssembler(Method)} (i.e.,
    * after the main proxy assembler has implemented its methods), but before
    * <code>ClassWriter.visitEnd</code> is invoked to terminate the assembly process.
    * 
    * @param   classWriter
    *          ASM class writer which the plugin will use to do its work.
    */
   public void assemble(ClassWriter classWriter);
}