/*
** Module   : MasterPty.java
** Abstract : an abstract interface with native PTY.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ----------------Description-----------------
** 001 MAG 20140318 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.terminal;

import java.io.*;

/**
 * This class is a simplified interface with PTY native code.
 * It provide buffered read operations.
 * 
 * @author mag
 *
 */
public abstract class MasterPty 
extends NativePty
{
   /** TERM environment variable template */
   private static final String TERM = "TERM=%s";
         
   /** Read buffer */
   private ByteArrayInputStream buffer = null;
   
   /** Used to convert an UTF-16 character to an UTF-8 byte array */
   private char[] utf16 = new char[1];   
         
   /**
    * Constructor.
    * 
    * @param   term
    *          Terminal type.
    * @param   cmdLine
    *          Command line arguments. 
    * @param   overlay
    *          Overlapped mode. 
    */
   public MasterPty(String term, String[] cmdLine, int overlay)
   {
      super();
      this.sh(term, cmdLine, overlay);
   }
   
   /**
    * Run shell.
    * 
    * @param   term
    *          Terminal environment example vt100
    * @param   cmdLine
    *          Command line arguments
    * @param   overlay
    *          Overlapped mode. 
    */
   public int sh(String term, String[] cmdLine, int overlay)
   {
      pid = shell(String.format(MasterPty.TERM, term), cmdLine, overlay);
      return pid;
   }
         
   /**
    * Wait for child process to exit.
    * 
    * @return  Process status or -1 on error.
    */
   public int await()
   {
      return waitPid(pid);
   }
   
   /**
    * Read from master PTY.
    * 
    * @return  A byte array read from master PTY or an empty array if no characters 
    *          are available. 
    */
   public byte[] get()
   {
      return read(mfd_inp);
   }

   /**
    * Write to master PTY.
    * 
    * @param   data
    *          Array of bytes to write.
    *          
    * @return  Number of written bytes. 
    */
   public int put(byte[] data)
   {
      // Write data
      return write(mfd_out, data);
   }   

   /**
    * Write to master PTY.
    * 
    * @param   data
    *          String to write.
    *          
    * @return  Number of written bytes. 
    */
   public int put(String data)
   {
      return put(data.getBytes());
   }   

   /**
    * Convert UTF-16 character to an UTF-8 byte array and write it to master PTY. 
    * 
    * @param   ch
    *          character to write.
    *          
    * @return  Number of written bytes. 
    */
   public int put(char ch)
   {
      int count = -1;

      // store character      
      utf16[0] = ch;
      
      // convert and write to master PTY
      try
      {
         count = put(new String(utf16).getBytes("UTF-8"));
      }
      catch (UnsupportedEncodingException e)
      {
         e.printStackTrace();
      };
      
      // return number of written bytes. 
      return count;
   }      
   
   /**
    * Buffered read next byte from master PTY.
    * 
    * @return  Next byte from master PTY 0r -1 if buffer is empty.
    */
   protected byte nextByte()
   {
      if (buffer == null || buffer.available() == 0)
      {
         byte[] data = this.get();    
         buffer = new ByteArrayInputStream(data);
      }

      return (byte) buffer.read();      
   }
   
   /**
    * Get child process id.
    * 
    * @return  child process id.
    */
   public int getPid()
   {
      return this.pid;
   }
   
   /**
    * Wait for child process to exit.
    */
   public void waitFor()
   {
      this.await();
   }
}