/*
** Module   : VT100AbstractKeyboard.java
** Abstract : VT-100 keyboard implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20140318 First version.
** 002 MAG 20140507 Fix BACKSPACE key on VT100 (assign code 127).
** 003 CA  20140805 Added injectKey.
** 004 GES 20150420 Minor code formatting cleanup.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.terminal;

import java.awt.event.*;

/**
 * VT-100 Swing keyboard implementation. This class must be extended
 * and subclasses must provide code to get keyboard strokes.
 * 
 * @author mag
 *
 */
public abstract class VT100AbstractKeyboard
implements VT100Keyboard
{   
   /** Escape code */
   private static final byte ESC   = (byte) 0x1b;
   
   /** Keyboard mappings */
   private static byte[] TAB       = {(byte) 0x09};

   /** Carriage return */
   private static byte[] ENTER     = {(byte) 0x0d};

   /** Carriage return */
   private static byte[] BACKSPACE = {(byte) 0x7f};
   
   /** ANSI escape sequences */
   private static byte[] INSERT    = {ESC, '[', '2', '~'};
   private static byte[] DELETE    = {ESC, '[', '3', '~'};
   private static byte[] PAGE_UP   = {ESC, '[', '5', '~'};
   private static byte[] PAGE_DOWN = {ESC, '[', '6', '~'};
   private static byte[] HOME      = {ESC, '[', 'H'};
   private static byte[] END       = {ESC, '[', 'F'};
   private static byte[] UP        = {ESC, '[', 'A'};
   private static byte[] DOWN      = {ESC, '[', 'B'};
   private static byte[] RIGHT     = {ESC, '[', 'C'};
   private static byte[] LEFT      = {ESC, '[', 'D'};
   private static byte[] F1        = {ESC, '[', 'P'};
   private static byte[] F2        = {ESC, '[', 'Q'};
   private static byte[] F3        = {ESC, '[', 'R'};
   private static byte[] F4        = {ESC, '[', 'S'};
   private static byte[] F5        = {ESC, '[', 't'};
   private static byte[] F6        = {ESC, '[', 'u'};
   private static byte[] F7        = {ESC, '[', 'v'};
   private static byte[] F8        = {ESC, '[', 'l'};
   private static byte[] F9        = {ESC, '[', 'w'};
   private static byte[] F10       = {ESC, '[', 'x'};
   
   /** Cursor key mode */
   private boolean cursorKey = false;
   
   /** Master PTY interface */
   private MasterPty pty;
   
   /**
    * Handle key inserts.
    * 
    * @param   keyCode
    *          Key code.
    * @param   charCode
    *          Character code
    */
   public abstract void injectKey(int keyCode, char charCode);

   /**
    * Set Master PTY interface.
    * 
    * @param   pty
    *          MasterPty instance.
    */
   public void setMasterPty(MasterPty pty)
   {
      this.pty = pty;
   }
      
   /**
    * Handle keystroke events.
    * 
    * @param   keyCode
    *          Key code.
    * @param   keyChar
    *          Key character.
    */
   public void handleKeystroke(int keyCode, char keyChar)
   {
      if (pty != null)
      {
         switch (keyCode)
         {
            case KeyEvent.VK_CONTROL:
            case KeyEvent.VK_SHIFT:
            case KeyEvent.VK_ALT:
            case KeyEvent.VK_CAPS_LOCK:    
            case KeyEvent.VK_SCROLL_LOCK:
            case KeyEvent.VK_NUM_LOCK:
            case KeyEvent.VK_PAUSE:
            case KeyEvent.VK_WINDOWS:
            case KeyEvent.VK_CONTEXT_MENU:
               break;
            case KeyEvent.VK_HOME:
               pty.put(HOME);
               break;
            case KeyEvent.VK_END:
               pty.put(END);
               break;            
            case KeyEvent.VK_INSERT:
               pty.put(INSERT);
               break;            
            case KeyEvent.VK_BACK_SPACE:
               pty.put(BACKSPACE);
               break;
            case KeyEvent.VK_DELETE:
               pty.put(DELETE);
               break;            
            case KeyEvent.VK_PAGE_UP:
               pty.put(PAGE_UP);
               break;            
            case KeyEvent.VK_PAGE_DOWN:
               pty.put(PAGE_DOWN);
               break;            
            case KeyEvent.VK_TAB:
               pty.put(TAB);
               break;
            case KeyEvent.VK_ENTER:
               pty.put(ENTER);
               break;
            case KeyEvent.VK_F1:
               put(F1);
               break;
            case KeyEvent.VK_F2:
               put(F2);
               break;
            case KeyEvent.VK_F3:
               put(F3);
               break;
            case KeyEvent.VK_F4:
               put(F4);
               break;
            case KeyEvent.VK_F5:
               put(F5);
               break;
            case KeyEvent.VK_F6:
               put(F6);
               break;
            case KeyEvent.VK_F7:
               put(F7);
               break;
            case KeyEvent.VK_F8:
               put(F8);
               break;
            case KeyEvent.VK_F9:
               put(F9);
               break;
            case KeyEvent.VK_F10:
               put(F10);
               break;
            case KeyEvent.VK_UP:
               put(UP);
               break;
            case KeyEvent.VK_DOWN:
               put(DOWN);
               break;
            case KeyEvent.VK_LEFT:
               put(LEFT);
               break;
            case KeyEvent.VK_RIGHT:
               put(RIGHT);
               break;
            default:
               pty.put(keyChar);
               break;
         }
      }      
   }
   
   /**
    * Send data to PTY according to cursorKey parameter.
    * If <code>cursorParameter</code> is true use Application format
    * else use ANSI format.
    * 
    * @param   data
    *          Byte sequence.
    */
   private void put(byte[] data)
   {
      data[1] = cursorKey ? (byte)'O' : (byte)'[';
      pty.put(data);
   }
   
   /**
    * Set cursor key mode.
    * 
    * @param   cursorKey
    *          <code>true</code> Application mode.
    *          <code>false</code> ANSI mode.
    */   
   @Override
   public void setCursorKey(boolean cursorKey)
   {
      this.cursorKey = cursorKey;      
   }   
}