/*
** Module   : XmlHelper.java
** Abstract : Utility methods for parsing and writing XML files
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041123   @18805 Created initial version, including basic
**                           support for parsing existing XML documents
**                           and for creating and writing new XML
**                           documents.
** 002 ECF 20050221   @19953 Added getText method. Extracts text data from
**                           an element.
** 003 GES 20050225   @19954 Add helper methods to encode/decode strings
**                           which can contain invalid XML chars. Such
**                           strings can contain null chars et al, which
**                           are converted on output to the safe NCR
**                           representation. On input any such NCRs are
**                           converted back into the original chars. In
**                           addition, more generic methods to safely
**                           encode invalid chars AND binary data have 
**                           been added.
** 004 ECF 20050303   @20155 Fixed getText method to properly handle CDATA
**                           sections.
** 005 ECF 20050331   @20627 Added methods to set and get an attribute
**                           while handling NCR encoding transparently.
**                           The attribute value is properly encoded or
**                           decoded to/from hex format, respectively, as
**                           needed. A leading underscore character is
**                           prepended to the attribute's name to indicate
**                           an encoded value.
** 006 ECF 20050421   @20803 Moved class to p2j.xml package.
** 007 ECF 20050421   @20841 Added astToDom, astToDomNode, and domToAst
**                           methods. Together, these allow roundtripping
**                           between XML DOM documents and XML ASTs.
** 008 GES 20050428   @20924 Changes to allow getAttribute() to handle
**                           empty string and strings of whitespace.
** 009 ECF 20050830   @22420 Add support for validating parsers. Added
**                           second thread local document builder factory
**                           and a new parse method to support this.
** 010 ECF 20060201   @24282 Added parse() method variants which enable
**                           validation. These are just convenience
**                           wrappers for an existing worker method.
** 011 EVL 20070612   @34019 Migrating to the javax.xml.stream(Java 6)
**                           from obsolete org.apache.crimson.tree from
**                           Java142. Preimplementation of the replacing
**                           setDoctype facility by deferring real
**                           creation of the document to the stage of
**                           setting it's type.
** 012 ECF 20070614   @34056 Completed changes necessary for port to Java
**                           6. Removed references to crimson parser. We
**                           now use only public J2SE APIs to create and
**                           serialize XML documents. Note there is still
**                           one vendor dependency for pretty printed XML
**                           serialized output: we use an XSLT transform
**                           to serialize a Document object to a stream.
**                           In the event this is done with the "pretty"
**                           option enabled, we set an output property
**                           which is specific to the Apache Xalan XSLT
**                           processor (currently the default in Java 6),
**                           to achieve indenting. If a different XSLT
**                           processor becomes the default in a future
**                           version of J2SE, pretty printing may become
**                           disabled or behave differently.
** 013 SIY 20070615   @34060 Fixed NPE if document type is not defined
**                           (which is usual for most manually created
**                           documents).
** 014 SIY 20070615   @34061 Fixed NPE if document public ID is not 
**                           defined and prevented potential similar 
**                           problem with document system ID.
** 015 ECF 20070621   @34193 Fixed comment problem in astToDomNode().
**                           A comment node before the root element in the
**                           AST would appear after the root element in
**                           the resulting DOM.
** 016 ECF 20070621   @34194 Flaw in 'pretty' output when serializing
**                           certain DOMs. When a comment appears directly
**                           after a DOCTYPE node, but before the root
**                           element, 'pretty' serialization inverts these
**                           for some reason, such that the comment is
**                           emitted before the DOCTYPE node in the
**                           serialized output.
** 017 ECF 20080304   @37465 Added variant of parse() which accepts an
**                           EntityResolver.
** 018 GES 20080305   @37671 Inserted commented code that would honor the
**                           default JVM encoding when writing an XML file
**                           out. Switched writing code to use streams
**                           instead of a character oriented writer, since
**                           the DOM layer handles translating the chars
**                           to UTF-8 and using the writer approach causes
**                           additional, incorrect translation.
** 019 GES 20090424   @41976 Import change.
** 020 GES 20090513   @42240 Package change. Added some attribute getters.
**                           Moved AST code from here to XmlAst.
** 021 GES 20110831          Added loadStringMap().
** 022 EVK 20130902          Added namespaceURI and isNamespaceAware support.
** 023 OM  20130924          Make sure the parent directory structure exists when creating files.
** 024 CA  20140106          Changed write API to allow omission of the XML declaration.
** 025 ECF 20150715          Replace StringBuffer with StringBuilder.
** 026 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 027 OM  20160331          Added parse() method that informs the caller whether the returned 
**                           document is valid in case the validation is required. Generified.
** 028 CA  20160216          Fixed convertBytesToHex for negative byte values.
** 029 GES 20160310          Allow forcing map keys to lowercase.
** 030 HC  20160907          Added parseWithPositions(), the method stores originating line
**                           number to every parsed DOM element.
** 031 GES 20170706          Added public constants for map style XML files. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.io.*;
import java.lang.ref.*;
import java.util.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import com.goldencode.p2j.util.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * Collection of useful, static, helper methods for various XML processing
 * tasks, including parsing and writing (serializing) XML data.
 * <p>
 * Generally, all XML-related processing should use this class for document
 * object model creation, parsing, and serialization. This ensures that the
 * same XML support implementation is used consistently, since assumptions are
 * made in several methods herein as to the underlying implementation classes.
 * <p>
 * Methods which rely on a vendor-specific <code>Document</code>
 * implementation are:
 * <ul>
 * <li>{@link #write(Document, String) write(Document, String)}</li>
 * <li>{@link #write(Document, File) write(Document, File)}</li>
 * <li>{@link #write(Document, OutputStream) write(Document, OutputStream)}</li>
 * </ul>
 * These methods may be re-implemented at a later time to support the
 * required capabilities in a vendor-neutral way.
 */
public final class XmlHelper
{
   /**
    * The key to element user data that holds the element's file location when
    * {@link #parseWithPositions(InputStream, boolean, String, boolean)} is used.
    */
   public static final String LINE_NUMBER_KEY = "lineNumber";
   
   /** Element name for the map root. */
   public static final String MAP_ELEM_ROOT = "map";
   
   /** Element name for a map entry. */
   public static final String MAP_ELEM_ENTRY = "entry";

   /** Attribute name for a map entry key. */
   public static final String MAP_ATTR_KEY = "key";
   
   /** Attribute name for a map entry value. */
   public static final String MAP_ATTR_VALUE = "value";
   
   /** Thread-local, validating, document builder factory. */
   private static final ThreadLocal<SoftReference<DocumentBuilderFactory>> validatingFactory = 
         new ThreadLocal<>();
   
   /** Thread-local, non-validating, document builder factory. */
   private static final ThreadLocal<SoftReference<DocumentBuilderFactory>> nonValidatingFactory = 
         new ThreadLocal<>();
   
   /**
    * Private constructor to prevent instances of this class from being
    * created.
    */
   private XmlHelper()
   {
   }

   /**
    * Create an XML document with an optional DOCTYPE tag (if the proper
    * information is supplied).
    *
    * @param   qualifiedName
    *          Name of the document type to be created.  May be
    *          <code>null</code>.
    * @param   publicId
    *          <code>PUBLIC</code> portion of doctype. If <code>null</code>,
    *          the <code>!DOCTYPE</code> element will be empty.
    * @param   systemId
    *          Optional system portion of doctype. If <code>null</code>, only
    *          the <code>publicId</code> will be used.
    * @param   namespaceURI
    *          Namespace URI of root node. If <code>null</code>, then node creates without
    *          namespace URI.
    *
    * @return  A new document.
    *
    * @throws  ParserConfigurationException
    *          if the XML document builder has been configured incorrectly.
    *
    */
   public static Document newDocument(String qualifiedName,
                                      String publicId,
                                      String systemId,
                                      String namespaceURI)
   throws ParserConfigurationException
   {
      DocumentBuilderFactory factory = getDocumentBuilderFactory(false, false);
      DocumentBuilder builder = factory.newDocumentBuilder();
      DOMImplementation domImpl = builder.getDOMImplementation();
      DocumentType docType = null;
      if (qualifiedName != null)
      {
         docType = domImpl.createDocumentType(qualifiedName,
                                              publicId,
                                              systemId);
      }

      return domImpl.createDocument(namespaceURI, qualifiedName, docType);
   }

   /**
    * Create an XML document with an optional DOCTYPE tag (if the proper
    * information is supplied).
    *
    * @param   qualifiedName
    *          Name of the document type to be created.  May be
    *          <code>null</code>.
    * @param   publicId
    *          <code>PUBLIC</code> portion of doctype. If <code>null</code>,
    *          the <code>!DOCTYPE</code> element will be empty.
    * @param   systemId
    *          Optional system portion of doctype. If <code>null</code>, only
    *          the <code>publicId</code> will be used.
    *
    * @return  A new document.
    *
    * @throws  ParserConfigurationException
    *          if the XML document builder has been configured incorrectly.
    *
    */
   public static Document newDocument(String qualifiedName,
                                      String publicId,
                                      String systemId)
   throws ParserConfigurationException
   {
      return newDocument(qualifiedName, publicId, systemId, null);
   }
   
   /**
    * Create and return a new, empty <code>Document</code> XML node.
    *
    * @return  A new document.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    */
   public static Document newDocument()
   throws ParserConfigurationException
   {
      return newDocument(null, null, null);
   }
   
   /**
    * Set the <code>DOCTYPE</code> element for an XML DOM which is built up
    * programmatically (as opposed to one which is created by parsing an
    * existing document). Since editing the <code>DOCTYPE</code> is not
    * supported by the standard <code>Document</code> interface, this method
    * makes the assumption that <code>dom</code> is an object which was
    * created by this helper class (either by parsing an existing document
    * or via the {@link #newDocument() newDocument} method).
    *
    * @param   dom
    *          Document which will have its doctype set.
    * @param   publicId
    *          <code>PUBLIC</code> portion of doctype. If <code>null</code>,
    *          the <code>!DOCTYPE</code> element will be empty.
    * @param   systemId
    *          Optional system portion of doctype. If <code>null</code>, only
    *          the <code>publicId</code> will be used.
    * @param   internalSubset
    *          Optional, internal subset portion of doctype.
    *
    * @throws  UnsupportedOperationException
    *          always;  this method is deprecated because it could not be
    *          implemented in a vendor-neutral manner.
    *
    * @deprecated
    */
   public static void setDoctype(Document dom,
                                 String publicId,
                                 String systemId,
                                 String internalSubset)
   throws UnsupportedOperationException
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Parse the XML file represented by <code>filename</code> into a document
    * object model (DOM) using a non-validating parser.  Delegates to the
    * {@link #parse(InputStream,boolean,String, boolean)
    * parse(InputStream,boolean,String)} method.
    *
    * @param   filename
    *          XML data filename.
    *
    * @return  XML document created by parsing the specified file.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within the specified file.
    */
   public static Document parse(String filename)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      return parse(new FileInputStream(filename), false, null, false);
   }
   
   /**
    * Parse the XML file represented by <code>filename</code> into a document
    * object model (DOM).  Delegates to the {@link
    * #parse(InputStream,boolean,String, boolean) parse(InputStream,boolean,String)}
    * method.
    *
    * @param   filename
    *          XML data filename.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   systemId
    *          A base for resolving relative URIs. Used if validating with
    *          a system DOCTYPE which specifies a relative URI for the DTD.
    * @param   isNamespaceAware
    *          true if the parser produced will provide support for XML namespaces;
    *          false otherwise.
    *
    * @return  XML document created by parsing the specified file.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within the specified file.
    */
   public static Document parse(String filename,
                                boolean validate,
                                String systemId,
                                boolean isNamespaceAware)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      return parse(new FileInputStream(filename), validate, systemId, isNamespaceAware);
   }
   
   /**
    * Parse the specified XML file into a document object model (DOM) using a
    * non-validating parser.  Delegates to the {@link
    * #parse(InputStream,boolean,String, boolean) parse(InputStream,boolean,String)}
    * method.
    *
    * @param   file
    *          A file of XML data.
    *
    * @return  XML document created by parsing the specified file.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>file</code>.
    */
   public static Document parse(File file)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      return parse(new FileInputStream(file), false, null, false);
   }
   
   /**
    * Parse the specified XML file into a document object model (DOM).
    * Delegates to the {@link #parse(InputStream,boolean,String, boolean)
    * parse(InputStream,boolean,String)} method.
    *
    * @param   file
    *          A file of XML data.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   systemId
    *          A base for resolving relative URIs. Used if validating with
    *          a system DOCTYPE which specifies a relative URI for the DTD.
    *
    * @return  XML document created by parsing the specified file.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>file</code>.
    */
   public static Document parse(File file,
                                boolean validate,
                                String systemId)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      return parse(new FileInputStream(file), validate, systemId, false);
   }
   
   /**
    * Parse the specified input stream into a document object model (DOM)
    * using a non-validating parser.  Delegates to the {@link
    * #parse(InputStream,boolean,String, boolean) parse(InputStream,boolean,String)}
    * method.
    *
    * @param   in
    *          An input stream of XML data.
    *
    * @return  XML document created by parsing the input stream data.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>in</code>.
    */
   public static Document parse(InputStream in)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      return parse(in, false, null, false);
   }

   /**
    * Parse the specified input stream into a document object model (DOM),
    * optionally validating the document.
    *
    * @param   in
    *          An input stream of XML data.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   systemId
    *          A base for resolving relative URIs. Used if validating with
    *          a system DOCTYPE which specifies a relative URI for the DTD.
    *
    * @return  XML document created by parsing the input stream data.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>in</code>.
    */
   public static Document parse(InputStream in,
                                boolean validate,
                                String systemId)
           throws ParserConfigurationException,
           IOException,
           SAXException
   {
      return parse(in, validate, systemId, false);
   }
   
   /**
    * Parse the specified input stream into a document object model (DOM),
    * optionally validating the document.
    *
    * @param   in
    *          An input stream of XML data.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   systemId
    *          A base for resolving relative URIs. Used if validating with
    *          a system DOCTYPE which specifies a relative URI for the DTD.
    * @param   isNamespaceAware
    *          true if the parser produced will provide support for XML namespaces; 
    *          false otherwise.
    *
    * @return  XML document created by parsing the input stream data.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>in</code>.
    */
   public static Document parse(InputStream in,
                                boolean validate,
                                String systemId,
                                boolean isNamespaceAware)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      Document dom = null;
      DocumentBuilderFactory factory = getDocumentBuilderFactory(validate, isNamespaceAware);
      BufferedInputStream bis = null;
      try
      {
         // Use a buffered stream.
         bis = new BufferedInputStream(in);
         
         // Create a new document builder and parse the data.
         DocumentBuilder builder = factory.newDocumentBuilder();
         dom = builder.parse(bis, systemId);
      }
      finally
      {
         // Clean up the stream, regardless of success.
         if (bis != null)
         {
            bis.close();
         }
      }
      
      return dom;
   }
   
   /**
    * Parse the specified input stream into a document object model (DOM), optionally validating
    * the document.
    *
    * @param   in
    *          An input stream of XML data.
    * @param   validate
    *          {@code true} to use a validating parser, {@code false} to parse the document
    *          without validation.
    * @param   isNamespaceAware
    *          true if the parser produced will provide support for XML namespaces; 
    *          false otherwise.
    * @param   success
    *          Output parameter. May be {@code null} The method will initialize it to {@code true}
    *          only if the parsing was successful. If xml stream failed validation or any other
    *          fatal error is encountered, the value is set to {@code false}.
    *
    * @return  XML document created by parsing the input stream data. If the 
    *          {@code validate = true} but the document failed validation, it is still returned,
    *          provided there were no fatal errors. However, the {@code success} output parameter
    *          will be set to {@code false}.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within {@code in}.
    */
   public static Document parse(InputStream in,
                                boolean validate,
                                boolean isNamespaceAware, 
                                logical success)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      Document dom = null;
      DocumentBuilderFactory factory = getDocumentBuilderFactory(validate, isNamespaceAware);
      BufferedInputStream bis = null;
      
      // It seems that Progress prints only the first error (fatal or not). 
      // This is the first encountered exception. If a fatal exception is encountered, this is 
      // thrown instead. Otherwise it is stored and eventually print after the parse ends. 
      final SAXParseException nfException[] = { null };
      
      if (success != null)
      {
         success.setUnknown();
      }
      try
      {
         // Use a buffered stream.
         bis = new BufferedInputStream(in);
         
         // Create a new document builder and parse the data.
         DocumentBuilder builder = factory.newDocumentBuilder();
         if (success != null)
         {
            builder.setErrorHandler(new DefaultHandler()
            {
               /**
                * Receive notification of a recoverable parser error. React by setting the
                * {@code success} output parameter of the {@code parse} method to {@code false}.
                *
                * @param   e
                *          SAX exception that contains details on this non-fatal error.
                *
                * @throws  SAXException
                *          Never.
                */
               @Override
               public void error(SAXParseException e)
               throws SAXException
               {
                  if (nfException[0] == null)
                  {
                     // success should be unknown at this moment
                     nfException[0] = e;  
                  }
                  success.assign(false);
               }
               
               /**
                * Receive notification of a non-recoverable parser error. React by setting the
                * {@code success} output parameter of the {@code parse} method to {@code false}.
                *
                * @param   e
                *          SAX exception that contains details on this fatal error.
                *
                * @throws  SAXException
                *          Always.
                */
               @Override
               public void fatalError(SAXParseException e)
               throws SAXException
               {
                  success.assign(false);
                  throw (nfException[0] != null) ? nfException[0] : e; // rethrow to stop parsing
               }
            });
         }
         dom = builder.parse(bis, null);
      }
      finally
      {
         // clean up the stream, regardless of success
         if (bis != null)
         {
            bis.close();
         }
      }
      
      if (success != null && success.isUnknown())
      {
         // there were no parsing errors
         success.assign(true);
      }
      if (nfException[0] != null)
      {
         // if only non-fatal exception were encountered, display it now
         String msg = "X-NODEREF or X-DOCUMENT LOAD got an error: '%s'";
         String err = String.format(msg, nfException[0].getMessage());
         ErrorManager.recordOrShowError(9082, err, false, false, false);
      }
      
      return dom;
   }
   
   /**
    * Parse the specified input stream into a document object model (DOM),
    * optionally validating the document.  This variant accepts an
    * <code>EntityResolver</code>, which may be necessary for certain parsing
    * cases.  For example, a specialized entity resolver may be used to
    * resolve a locally provided DTD which is referred to by a public URL in
    * the XML document's DOCTYPE node, to avoid using the incorrect (public)
    * version of the DTD.
    *
    * @param   in
    *          An input stream of XML data.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   resolver
    *          Object which provides custom resolution for external entities
    *          during SAX parsing.
    * @param   handler
    *          Custom SAX error handler.
    *
    * @return  XML document created by parsing the input stream data.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>in</code>.
    */
   public static Document parse(InputStream in,
                                boolean validate,
                                EntityResolver resolver,
                                ErrorHandler handler)
   throws ParserConfigurationException,
          IOException,
          SAXException
   {
      Document dom = null;
      DocumentBuilderFactory factory = getDocumentBuilderFactory(validate, false);
      BufferedInputStream bis = null;
      try
      {
         // Use a buffered stream.
         bis = new BufferedInputStream(in);
         
         // Create a new document builder and parse the data.
         DocumentBuilder builder = factory.newDocumentBuilder();
         if (resolver != null)
         {
            builder.setEntityResolver(resolver);
         }
         if (handler != null)
         {
            builder.setErrorHandler(handler);
         }
         dom = builder.parse(bis);
      }
      finally
      {
         // Clean up the stream, regardless of success.
         if (bis != null)
         {
            bis.close();
         }
      }
      
      return dom;
   }

   /**
    * Parse the specified input stream into a document object model (DOM),
    * optionally validating the document.
    * <p>
    * This parsing method stores the element's line number in the source stream
    * in the element's user data under the key {@link #LINE_NUMBER_KEY}.
    *
    * @param   in
    *          An input stream of XML data.
    * @param   validate
    *          <code>true</code> to use a validating parser,
    *          <code>false</code> to parse the document without validation.
    * @param   systemId
    *          A base for resolving relative URIs. Used if validating with
    *          a system DOCTYPE which specifies a relative URI for the DTD.
    * @param   isNamespaceAware
    *          true if the parser produced will provide support for XML namespaces;
    *          false otherwise.
    *
    * @return  XML document created by parsing the input stream data.
    *
    * @throws  ParserConfigurationException
    *          if the XML parser has been configured incorrectly.
    * @throws  IOException
    *          if an I/O error occurs during parsing or stream cleanup.
    * @throws  SAXException
    *          if an error occurs parsing the XML within <code>in</code>.
    */
   public static Document parseWithPositions(final InputStream in,
                                             boolean validate,
                                             String systemId,
                                             boolean isNamespaceAware)
   throws IOException,
          SAXException,
          ParserConfigurationException
   {
      SAXParserFactory factory = SAXParserFactory.newInstance();
      SAXParser parser = factory.newSAXParser();

      DocumentBuilderFactory docBuilderFactory =
                 XmlHelper.getDocumentBuilderFactory(validate, isNamespaceAware);

      DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
      final Document doc = docBuilder.newDocument();

      StringBuilder builder = new StringBuilder();
      Stack<Element> elStack = new Stack<Element>();

      DefaultHandler handler = new DefaultHandler()
      {
         private Locator locator;

         @Override
         public void setDocumentLocator(Locator locator)
         {
            this.locator = locator;
         }

         @Override
         public void startElement(String uri,
                                  String localName,
                                  String qName,
                                  Attributes attributes)
         throws SAXException
         {
            flushAccumulatedText();
            Element el = doc.createElement(qName);
            for (int i = 0; i < attributes.getLength(); i++)
            {
               el.setAttribute(attributes.getQName(i), attributes.getValue(i));
            }
            el.setUserData(LINE_NUMBER_KEY, Integer.valueOf(locator.getLineNumber()), null);
            elStack.push(el);
         }

         @Override
         public void endElement(String uri, String localName, String qName)
         {
            flushAccumulatedText();
            Element closedEl = elStack.pop();
            if (elStack.isEmpty())
            {
               // the root element
               doc.appendChild(closedEl);
            }
            else
            {
               Element parentEl = elStack.peek();
               parentEl.appendChild(closedEl);
            }
         }

         @Override
         public void characters(final char ch[], final int start, final int length)
         throws SAXException
         {
            builder.append(ch, start, length);
         }

         private void flushAccumulatedText()
         {
            if (builder.length() > 0)
            {
               final Element el = elStack.peek();
               final Node textNode = doc.createTextNode(builder.toString());
               el.appendChild(textNode);
               builder.delete(0, builder.length());
            }
         }
      };

      parser.parse(in, handler, systemId);
      return doc;
   }

   /**
    * Load the given map with entries read from the given filename.  The file
    * must be in the following format:
    * <p>
    * <pre>
    * &lt;map&gt;
    *    &lt;entry key="a" value="b"/&gt;
    *    &lt;entry key="c" value="d"/&gt;
    * &lt;/map&gt;
    * </pre>
    * <p>
    * Any failure aborts silently.
    *
    * @param    map
    *           The map to load.
    * @param    filename
    *           The file to read from.
    * @param    force
    *           Force map keys to lowercase.
    */
   public static void loadStringMap(Map<String, String> map, String filename, boolean force)
   {
      try
      {
         File file = new File(filename);
         
         if (!file.exists())
            return;
         
         Document doc   = parse(filename);
         Element  root  = doc.getDocumentElement();
         NodeList nodes = root.getElementsByTagName(MAP_ELEM_ENTRY);
         int      len   = nodes.getLength();
                                                                 
         for (int i = 0; i < len; i++)
         {
            Element elem = (Element) nodes.item(i);
            
            String key = elem.getAttribute(MAP_ATTR_KEY);
            String val = elem.getAttribute(MAP_ATTR_VALUE);
            
            if (key != null && val != null)
            {
               if (force)
               {
                  key = key.toLowerCase();
               }
               
               map.put(key, val);
            }
         }
      }
      
      catch (Exception e)
      {
         // TODO: change this
         // ignore everything
      }
   }
   
   /**
    * Get the string data, if any, from a node, whether it is wrapped in a
    * CDATA section or not. This method recursively checks through all
    * children of the specified node and extracts the first non-null text
    * which contains more than just whitespace it encounters (whether in
    * a CDATA section child of a text node or in the text node directly),
    * then stops looking.
    *
    * @param   node
    *          Node which presumably contains string data directly in a text
    *          node child or in a CDATA section grandchild.
    *
    * @return  The string data extracted from the first text node or CDATA
    *          section encountered, or <code>null</code> if none is found.
    */
   public static String getText(Node node)
   {
      String text = null;
      
      NodeList nodes = node.getChildNodes();
      int len = nodes.getLength();
      for (int i = 0; text == null && i < len; i++)
      {
         Node next = nodes.item(i);
         switch (next.getNodeType())
         {
            case Node.TEXT_NODE:
            case Node.CDATA_SECTION_NODE:
               text = StringHelper.safeTrim(((CharacterData) next).getData());
               if (text == null || text.length() == 0)
               {
                  // Try to extract from a child (CDATA) node.
                  text = getText(next);
               }
               break;
         }
      }
      
      return (text != null && text.length() > 0 ? text : null);
   }
   
   /**
    * Serialize the content in the specified XML document in a pretty printed
    * text format to the specified writer.
    *
    * @param   dom
    *          Document object model to be serialized.
    * @param   os
    *          Target character output stream.
    *
    * @throws  IOException
    *          if an error occurs writing to <code>writer</code> or cleaning
    *          up streams.
    */
   public static void write(Document dom, OutputStream os)
   throws IOException
   {
      write(dom, os, true, false);
   }
   
   /**
    * Serialize the content in the specified XML document in text form to the
    * specified output stream, optionally applying pretty print formatting.
    *
    * @param   dom
    *          Document object model to be serialized.
    * @param   os
    *          Target character output stream.
    * @param   pretty
    *          if <code>true</code>, this method will attempt to format the
    *          XML using line breaks and hierarchical indents;  else the
    *          output will contain no special formatting.
    *
    * @throws  IOException
    *          if an error occurs writing to <code>writer</code> or cleaning
    *          up streams.
    */
   public static void write(Document dom, OutputStream os, boolean pretty)
   throws IOException
   {
      write(dom, os, pretty, false);
   }
   
   /**
    * Serialize the content in the specified XML document in text form to the
    * specified output stream, optionally applying pretty print formatting.
    *
    * @param   dom
    *          Document object model to be serialized.
    * @param   os
    *          Target character output stream.
    * @param   pretty
    *          if <code>true</code>, this method will attempt to format the
    *          XML using line breaks and hierarchical indents;  else the
    *          output will contain no special formatting.
    * @param   omitXMLDec
    *          Flag indicating if the XML declaration should be omitted.
    *
    * @throws  IOException
    *          if an error occurs writing to <code>writer</code> or cleaning
    *          up streams.
    */
   public static void write(Document dom, OutputStream os, boolean pretty, boolean omitXMLDec)
   throws IOException
   {
      // Use a buffered stream.
      BufferedOutputStream bo = null;
      
      if (os instanceof BufferedOutputStream)
      {
         bo = (BufferedOutputStream) os;
      }
      else
      {
         bo = new BufferedOutputStream(os);
      }
      
      // Write the XML data to the specified output stream.
      try
      {
         DOMSource source = new DOMSource(dom);
         StreamResult target = new StreamResult(bo);
         TransformerFactory factory = TransformerFactory.newInstance();
         Transformer transformer = factory.newTransformer();
         DocumentType docType = dom.getDoctype();
         
         // DISABLED: the commented code below provides an override for the
         // XML encoding value in the XML header, BUT the built-in JVM parser
         // does not necessarily support each encoding, even when that
         // encoding is also the default encoding for the JVM. For now we need
         // to keep this as UTF-8 and translate into UTF-8 before writing
         // to XML.
         
         // honor the JVM-wide default encoding (instead of the default
         // which otherwise would always be UTF-8)
         // String encoding = Charset.defaultCharset().name();
         // transformer.setOutputProperty(OutputKeys.ENCODING, encoding);
         
         if (docType != null)
         {
            if (docType.getPublicId() != null)
               transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC,
                                             docType.getPublicId());
            
            if (docType.getSystemId() != null)
               transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
                                             docType.getSystemId());
         }
         
         if (pretty)
         {
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            // TODO:  eliminate vendor-specific extension.
            // TODO:  this does not handle comments properly when situated
            // after a DOCTYPE node;  they are output before the DOCTYPE.
            transformer.setOutputProperty(
               "{http://xml.apache.org/xslt}indent-amount", "2");
         }
         
         if (omitXMLDec)
         {
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
         }

         transformer.transform(source, target);
      }
      catch (Exception exc)
      {
         throw new IOException("Error writing XML document to file", exc);
      }
      finally
      {
         bo.close();
      }
   }
   
   /**
    * Serialize the content in the specified XML document in a pretty printed
    * text format to the specified file.
    *
    * @param   dom
    *          Document object model to be serialized.
    * @param   file
    *          Target file.
    *
    * @throws  IOException
    *          if an error occurs writing to <code>file</code> or cleaning
    *          up streams.
    */
   public static void write(Document dom, File file)
   throws IOException
   {
      file.getParentFile().mkdirs(); // make sure the needed directory structure exists
      write(dom, new FileOutputStream(file));
   }
   
   /**
    * Serialize the content in the specified XML document in a pretty printed
    * text format to the file with the specified name.
    *
    * @param   dom
    *          Document object model to be serialized.
    * @param   filename
    *          Target filename.
    *
    * @throws  IOException
    *          if an error occurs writing to <code>filename</code> or cleaning
    *          up streams.
    */
   public static void write(Document dom, String filename)
   throws IOException
   {
      write(dom, new File(filename));
   }
   
   /**
    * Converts the set of ASCII characters which are invalid (to be stored in 
    * the text of an XML attribute) into the associated valid representation
    * using the 'Numeric Character Reference' or NCR. Such references take
    * the form of &amp;#xHHH; where HHH is a hexidecimal number that
    * corresponds with the character being represented.
    * <p>
    * The following ASCII character ranges are escaped as described above
    * (since they cannot otherwise be represented in the text of an XML
    * attribute): 0x00 - 0x08, 0x0B - 0x0C, 0x0E - 0x1F.
    * <p>
    * Note that other invalid XML characters exist outside of the ASCII
    * range.  This method does not handle such characters.
    *
    * @param    text
    *           The string which may contain invalid characters.
    *
    * @return   A converted string which has each invalid character replaced
    *           by the corresponding NCR.
    */
   public static String convertInvalidCharsToNcr(String text)
   {
      StringBuilder sb  = new StringBuilder();
      int           len = text.length();
      
      for (int i = 0; i < len; i++)
      {
         int c = (int)text.charAt(i);
         
         // is this an invalid XML character?
         if ( (c >= 0x00 && c <= 0x08) ||
              (c == 0x0B || c == 0x0C) ||
              (c >= 0x0E && c <= 0x1F) )
         {
            // create the NCR and append it
            sb.append("&#x").append(Integer.toHexString(c)).append(";");
         }
         else
         {
            // safe character, append it
            sb.append((char)c);
         }
      }
      
      return sb.toString();
   }
   
   /**
    * Converts certain 'Numeric Character References' or NCRs into the  
    * into the associated valid ASCII representation.  This is only done for
    * the set of ASCII characters which are invalid (to be stored in the text
    * of an XML attribute). NCRs take the form of &amp;#xHHH; where HHH is a
    * hexidecimal number that corresponds with the character being
    * represented.
    * <p>
    * The following ASCII character ranges are escaped as described above
    * (since they cannot otherwise be represented in the text of an XML
    * attribute): 0x00 - 0x08, 0x0B - 0x0C, 0x0E - 0x1F.
    * <p>
    * Note that other invalid XML characters exist outside of the ASCII
    * range.  This method does not handle such characters.
    * <p>
    * <b>WARNING: This method is intended to be the counterpart to the
    * {@link #convertInvalidCharsToNcr} method used on output to safely
    * store certain characters.  Since it is possible to actually have a 
    * literal '&amp;#xHH;' coding inside the original text passed to the
    * <code>convertInvalidCharsToNcr</code>, there is no way to 
    * symmetrically remove only those NCRs that were added by this class
    * (without storing additional meta-information listing the exact NCRs
    * that were added, which would allow the differentiation of these with
    * any that were in the original string).</b>
    *
    * @param    text
    *           The string which may contain NCRs.
    *
    * @return   A converted string which has each NCR replaced by the
    *           corresponding ASCII character.
    */
   public static String convertNcrToAscii(String text)
   {
      StringBuilder sb   = new StringBuilder();
      int           idx  = text.indexOf("&#x");
      int           len  = text.length();
      int           next = 0;
      
      while (idx != -1)
      {
         // append everything up to our match to the output
         sb.append(text.substring(next, idx));
         next = idx;
         
         // get to the end of the matched string
         idx += 3;
         
         // this is not an NCR AND there is no more string to process
         if (idx >= len)
            break;
         
         int end = text.indexOf(";", idx);
         
         // this is not an NCR AND there are no other NCRs in here since
         // all valid NCRs must end with semicolon
         if (end == -1)
            break;
         
         // this next substring will either be a valid hex number or
         // something unrecognized
         String hex = text.substring(idx, end);
         int    c   = -1;
         
         // account for the semicolon
         end++;
         
         try
         {
            c = Integer.parseInt(hex, 16);
         }
         catch (NumberFormatException nfe)
         {
            // something unrecognized then append the whole lot to our output
            sb.append(text.substring(next, end));
            next = end;
            
            // reset our index and restart our loop
            idx = text.indexOf("&#x", end);
            continue;
         }
         
         // possibly something to convert
         if ( (c >= 0x00 && c <= 0x08) ||
              (c == 0x0B || c == 0x0C) ||
              (c >= 0x0E && c <= 0x1F) )
         {
            // it is an NCR (for a char which we handle), append it
            sb.append((char)c);            
         }
         else
         {
            // not something to convert, just append the lot (up to and
            // including the semicolon
            sb.append(text.substring(next, end)); 
         }
         
         // move our next write position up past semicolon
         next = end;
         
         // reset index for next iteration of loop
         idx = text.indexOf("&#x", end);
      }
      
      // append the rest of the string to our output, if needed
      if (next < len)
         sb.append(text.substring(next));
      
      return sb.toString();
   }
   
   /**
    * Detects the presence of any character from the set of ASCII characters
    * which are invalid for storage as part of a readable string (this
    * is the same as not being able to be directly stored in an XML file).
    * <p>
    * The following ASCII character ranges are detected: 0x00 - 0x08, 0x0B -
    * 0x0C, 0x0E - 0x1F.
    * <p>
    * Note that other invalid XML characters exist outside of the ASCII
    * range.  This method does not handle such characters.
    *
    * @param    bytes
    *           The byte array which may contain invalid characters.
    *
    * @return   <code>true</code> if the array contains any invalid
    *           character, <code>false</code> otherwise
    */
   public static boolean needsConversionToHex(byte[] bytes)
   {      
      for (int i = 0; i < bytes.length; i++)
      {
         int b = (int)bytes[i];
         
         if ( (b >= 0x00 && b <= 0x08) ||
              (b == 0x0B || b == 0x0C) ||
              (b >= 0x0E && b <= 0x1F) )
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * To enable XML to handle binary data or to handle the case where a
    * string including invalid characters must be encoded in XML, this method
    * will convert an entire byte array into a hexidecimal string form where
    * each byte is represented by an equivalent two character hexidecimal
    * encoding.
    * <p>
    * All bytes are encoded using this technique, generally this is
    * only done if {@link #needsConversionToHex} returns <code>true</code>.
    * <p>
    * <b>WARNING: When using this method where the input array is string
    * data, the same character encoding must be used by the encoding side
    * as is used by the decoding side. Otherwise, the results are 
    * undefined.</b>
    *    
    * @param    bytes
    *           The byte array to be encoded.
    *
    * @return   A converted string which has each byte in the input array
    *           replaced by a two character hexidecimal representation.
    */
   public static String convertBytesToHex(byte[] bytes)
   {
      StringBuilder      sb   = new StringBuilder();
      
      for (int i = 0; i < bytes.length; i++)
      {
         // convert to the hexidecimal representation from 0x0 to 0xFF
         String hex = Integer.toHexString((int) bytes[i]);
         
         // prefix with 0 to ensure that things like 0x0 are actually 0x00
         if (hex.length() == 1)
         {
            sb.append("0");
         }
         else if (hex.length() > 2)
         {
            // Integer.toHexString will pad with FFFFFF for negative numbers, so take only the
            // last two characters
            hex = hex.substring(hex.length() - 2);
         }
            
         // write it to the output
         sb.append(hex);
      }      
      
      return sb.toString().toUpperCase();
   }
   
   /**
    * To enable XML to handle binary data or to handle the case where a
    * string including invalid characters must be encoded in XML, this method
    * will convert an encoded hexidecimal string into the corresponding
    * byte array.  Each byte in the output array is represented by an
    * equivalent two character hexidecimal encoding in the input string.
    * <p>
    * <b>WARNING: When using this method where the output array is string
    * data, the same character encoding must be used by the encoding side
    * as is used by the decoding side. Otherwise, the results are 
    * undefined.</b>
    *
    * @param    hex
    *           The string to be decoded.
    *
    * @return   A decoded byte array which has each two character hexidecimal 
    *           representation byte in the input string replaced by the
    *           binary equivalent.
    */
   public static byte[] convertHexToBytes(String hex)
   {
      int len = hex.length();
      
      // the input string MUST have a length that is a multiple of 2 because 
      // we require that every char is represented by a 2 char hexidecimal
      // value
      if ((len % 2) != 0)
      {
         // just return the input string
         return hex.getBytes();
      }
      
      // create our output array
      byte[] bytes = new byte[len/2];
      
      for (int i = 0; i < len; i += 2)
      {
         // read one encoded char
         String ch = hex.substring(i, i + 2);
         byte   b;
         
         // convert to an integer
         try
         {
            b = (byte) Integer.parseInt(ch, 16);
         }
         catch (NumberFormatException nfe)
         {
            // this isn't a valid hexidecimal string, return input string
            return hex.getBytes();
         }
         
         // write it to the output
         bytes[i/2] = b;
      }      
      
      return bytes;
   }
   
   /**
    * Extract the value of an attribute from an XML element, decoding it
    * from hex format first if necessary.
    *
    * @param   elem
    *          The XML element containing the attribute to be extracted.
    * @param   key
    *          The attribute's name. If no attribute is found under this
    *          name, the key is prefixed with a leading underscore
    *          (<code>_</code>). If an attribute is found under this form
    *          of the key, it is assumed the value is encoded in hex
    *          format, and the value is decoded before it is returned.
    *
    * @return  The attribute's value (decoded if necessary), or
    *          <code>null</code> if no attribute match could be found for
    *          either form of <code>key</code>.
    */
   public static String getAttribute(Element elem, String key)
   {  
      String value = null;
      
      if (elem.hasAttribute(key))
      {
         // Attribute value is not encoded.
         value = elem.getAttribute(key);
      }
      else
      {
         // Attempt to get attribute as encoded value.
         String encoded = "_" + key;
         
         if (elem.hasAttribute(encoded))
         {
            value = new String(convertHexToBytes(elem.getAttribute(encoded)));
         }
      }
      
      return value;
   }
   
   /**
    * Set a value into an element's attribute, encoding it into a safe format
    * first if it contains non-printing characters. If encoded, the attribute
    * name will be prefixed with an underscore character.
    *
    * @param   elem
    *          DOM element to be updated.
    * @param   key
    *          Attribute name.
    * @param   value
    *          Value to be written, as a string, into the attribute. If
    *          <code>null</code>, no attribute is added to the element.
    */
   public static void setAttribute(Element elem, String key, String value)
   {
      if (value == null)
      {
         return;
      }
      
      byte[] data = value.getBytes();
      if (XmlHelper.needsConversionToHex(data))
      {
         // Value contains control characters; convert entire data value
         // to a hex string (i.e., two hex values per byte).
         value = XmlHelper.convertBytesToHex(data);
         
         // Prefix the attribute name with an underscore to signal that the data is encoded.
         key = "_" + key;
      }
      
      elem.setAttribute(key, value);
   }
   
   /**
    * Get the specified attribute and convert it to an integer.  If it is
    * not present, return the given default value.
    *
    * @param    elem
    *           The node to search for the attribute.
    * @param    name
    *           The attribute name for which to search.
    * @param    defVal
    *           The default value to return if no attribute of that name is
    *           found OR if there is any error.
    *
    * @return   The parsed value or the given default value.
    */
   public static int getAttributeInt(Element elem, String name, int defVal)
   {
      int num = defVal;
      
      String numTxt = elem.getAttribute(name);
      
      if (numTxt != null && numTxt.length() > 0)
      {
         try
         {
            num = Integer.parseInt(numTxt);
         }
         
         catch (NumberFormatException nfe)
         {
            // ignore and return the default
         }
      }
      
      return num;
   }
   
   /**
    * Get the specified attribute and convert it to a 64-bit integer.  If it is
    * not present, return the given default value.
    *
    * @param    elem
    *           The node to search for the attribute.
    * @param    name
    *           The attribute name for which to search.
    * @param    defVal
    *           The default value to return if no attribute of that name is
    *           found OR if there is any error.
    *
    * @return   The parsed value or the given default value.
    */
   public static long getAttributeLong(Element elem, String name, long defVal)
   {
      long num = defVal;
      
      String numTxt = elem.getAttribute(name);
      
      if (numTxt != null && numTxt.length() > 0)
      {
         try
         {
            num = Long.parseLong(numTxt);
         }
         
         catch (NumberFormatException nfe)
         {
            // ignore and return the default
         }
      }
      
      return num;
   }
   
   /**
    * Get the specified attribute and convert it to a boolean value.  If it is
    * not present, return the given default value.
    *
    * @param    elem
    *           The node to search for the attribute.
    * @param    name
    *           The attribute name for which to search.
    * @param    defVal
    *           The default value to return if no attribute of that name is
    *           found OR if there is any error.
    *
    * @return   The parsed value or the given default value.
    */
   public static boolean getAttributeBoolean(Element elem, String name, boolean defVal)
   {
      boolean result = defVal;
      
      String value = elem.getAttribute(name);
      
      if (value != null && value.length() > 0)
      {
         if ("true".equalsIgnoreCase(value))
         {
            result = true;
         }
      }
      
      return result;
   }
   
   /**
    * Get a document builder factory for the current thread. If one does not
    * already exist, create one and store it in a thread local variable.
    * The thread-local pattern is used because
    * <code>DocumentBuilderFactory</code> is not thread-safe. Since this is
    * not a critical resource to cache, soft references are used to store
    * the factories, allowing the GC to collect them if memory is constrained.
    *
    * @param   validating
    *          <code>true</code> to get a factory for validating document
    *          builders; <code>false</code> to get a factory for
    *          non-validating document builders.
    * @param   isNamespaceAware
    *          true if the parser produced will provide support for XML namespaces; 
    *          false otherwise.
    *
    * @return  Document builder factory for the current thread.
    */
   static DocumentBuilderFactory
   getDocumentBuilderFactory(boolean validating, boolean isNamespaceAware)
   {
      ThreadLocal<SoftReference<DocumentBuilderFactory>> local = (validating
                           ? validatingFactory
                           : nonValidatingFactory);
      SoftReference<DocumentBuilderFactory> ref = local.get();
      
      DocumentBuilderFactory factory = null;
      
      if (ref != null)
      {
         factory = ref.get();
      }
      
      if (factory == null)
      {
         factory = DocumentBuilderFactory.newInstance();
         factory.setValidating(validating);

         // Store factory as referent of a soft reference.
         ref = new SoftReference<>(factory);
         
         // Store reference in appropriate thread-local, static variable.
         local.set(ref);
      }
      
      // make sure the factory is configured properly for this request
      factory.setNamespaceAware(isNamespaceAware);

      return factory;
   }
}