/*
** Module   : library.c
** Abstract : implementation of low level shared library invocation functions called via JNI
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20131023 First version.
** 002 GES 20140212 Added flag to let the native layer conditionally process for loading
**                  persistent libraries.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include "library.h"
#include <ffi.h>

#include "com_goldencode_p2j_library_LibraryManager.h"

// internal helper functions
void*      jlongToPointer(jlong addr);
ffi_type*  convertType(int type);
ffi_abi    calcCallingConvention(jint conv);
                                           
/**
 * Reports on the number of bytes that the native stack uses for push/pop operations (the
 * width or size of the word used).                       
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 *
 * @return   This will return 4 on 32-bit systems and 8 for 64-bit platforms.
 */ 
JNIEXPORT jint JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_wordSize(JNIEnv* env, jclass cls)
{
   #ifdef WORD_SIZE_64
   jint size = 8;
   #else
   jint size = 4;
   #endif
   
   return size;    
}

/**
 * Reports the platform type on which the client is running.                       
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 *
 * @return   A LibraryManager PLATFORM_* constant.
 */ 
JNIEXPORT jint JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_platform(JNIEnv* env, jclass cls)
{
   #ifdef WIN_TARGET
   jint platform = com_goldencode_p2j_library_LibraryManager_PLATFORM_WINDOWS;
   #else
   jint platform = com_goldencode_p2j_library_LibraryManager_PLATFORM_UNIX;
   #endif
   
   return platform;    
}

/**
 * Attempts to load the named library.  This will use the operating-system runtime library
 * loading mechanism.
 *
 * @param    libname
 *           This may be any valid library filename for the current system, including names
 *           that contain path separators (both relative and absolute). This should not be
 *           <code>null</code> or the empty string.
 * @param    persist
 *           If <code>true</code>, the library should not be unloaded at the end of the
 *           function invocation, which means that this method should cache the module
 *           handle if it is successfully loaded.
 *
 * @return   An operating-system specific "handle" to the library. Under Windows this is
 *           the "module handle" of type HMODULE and under Linux it is just referred to as
 *           a "handle" that is returned as a void pointer.  In either case, if
 *           <code>NULL</code> (0) is returned, then the call failed.
 */ 
JNIEXPORT jlong JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_load(JNIEnv*  env,
                                                       jclass   cls,
                                                       jstring  libname,
                                                       jboolean persist)
{
   const char *filename = (*env)->GetStringUTFChars(env, libname, NULL);
   
   if (filename == NULL) 
   {
      return 0;  // OutOfMemoryError already thrown
   }
   
   jlong result = (jlong) load_library_native(filename, persist);
   
   (*env)->ReleaseStringUTFChars(env, libname, filename);
   
   return result;
}

/**
 * Attempts to unload the library identified by the "handle" which was returned from a
 * previous call to {@link #load}.  This will use the operating-system runtime library
 * unloading mechanism. Although this call may fail, Progress seems to always silently ignore
 * any failures, so this method has no return value.
 *
 * @param    modhandle
 *           This must be a valid handle to the library.
 */ 
JNIEXPORT void JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_unload(JNIEnv* env, jclass cls, jlong modhandle)
{
   if (modhandle != 0) 
      unload_library_native((int64_t) modhandle);
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given symbol in the library.  The symbol name must match exactly, including the case
 * and any decorations that exist as exported in the library.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    funcname
 *           This is the exact symbol to lookup in the library.
 *
 * @return   The memory address of the function or <code>NULL</code> (0) if the function cannot
 *           be found by that name in the given library.
 */ 
JNIEXPORT jlong JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_findByName(JNIEnv* env,
                                                             jclass  cls,
                                                             jlong   modhandle,
                                                             jstring funcname)
{
   if (modhandle == 0)
      return (jlong) 0;
   
   const char *name = (*env)->GetStringUTFChars(env, funcname, NULL);
   
   if (name == NULL)
   {
      return 0;  // OutOfMemoryError already thrown
   }
   
   jlong addr = find_by_name_native((int64_t) modhandle, name);
   
   (*env)->ReleaseStringUTFChars(env, funcname, name);
   
   return addr;
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given ordinal in the library.  Only some operating systems allow exporting functions by
 * ordinal number.  Those that do allow it, will typically have a table of contiguous ordinal
 * values in the library, each of which represents a function that may or may not be also
 * exported by name (as a symbol).
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    ordinal
 *           The ordinal to lookup in the library.  Only the lower word is honored.
 *
 * @return   The memory address of the function or <code>NULL</code> (0) if the function cannot
 *           be found by that ordinal in the given library. Operating systems that do not
 *           support ordinal exports will always return 0.
 */ 
JNIEXPORT jlong JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_findByOrdinal(JNIEnv* env,
                                                                jclass  cls,
                                                                jlong   modhandle,
                                                                jint    ordinal)
{
   jlong addr = 0;                                             
   
   if (modhandle != 0) 
      addr = find_by_ordinal_native((int64_t) modhandle, (uint16_t) ordinal);
   
   return addr;
}

/**
 * Call the function with the specified arguments and return value processing. This
 * implementation uses libffi for the "heavy lifting" of handling the actual call.  We just
 * translate between the Java code and the native libffi library.
 *
 * @param    funcAddr
 *           The memory address of the function to call.  Must not be <code>NULL</code>.
 * @param    callConv
 *           A constant representing the calling convention to be used.  This is only honored on
 *           32-bit Windows, where there are multiple calling conventions to choose from.
 * @param    retType
 *           Specifies the data type of the return value. Ignored if the <code>retAddr</code> is
 *           <code>NULL</code>.
 * @param    retAddr
 *           The memory address in which to place the return value. The call is assumed to have
 *           no return value (or a return value that can be ignored) if this pointer is
 *           <code>NULL</code>.
 * @param    numArgs
 *           The number of arguments in the called function's signature. Must be non-negative
 *           (0 is valid).
 * @param    argTypes
 *           If <code>numArgs</code> is greater than 0, this is an array that specifies the data
 *           type of each argment from left to right (the 0 index of the array corresponds to the
 *           leftmost argument).
 * @param    argAddr
 *           If <code>numArgs</code> is greater than 0, this must be the memory address of the
 *           array of <code>void*</code>. Each one points to an argment value from left to right
 *           (the 0 index of the array corresponds to the leftmost argument).
 *
 * @return   A code indicating success or failure of the dispatching.
 */ 
JNIEXPORT jint JNICALL
   Java_com_goldencode_p2j_library_LibraryManager_dispatch(JNIEnv*   env,
                                                           jclass    cls,
                                                           jlong     funcAddr,
                                                           jint      callConv,
                                                           jint      retType,
                                                           jlong     retAddr,
                                                           jint      numArgs,
                                                           jintArray argTypes,
                                                           jlong     argAddr)
{
   ffi_cif   cif;
   ffi_type* argTypeDefs[numArgs];   
   
   // calculate the calling convention to use
   ffi_abi abi = calcCallingConvention(callConv);
   
   // convert the list of p2j argument types to the ffi types
   if (numArgs > 0)
   {
      jint* typeList = (*env)->GetIntArrayElements(env, argTypes, NULL);
      
      if (typeList == NULL)
      {
         // out of memory has been thrown
         return com_goldencode_p2j_library_LibraryManager_ERROR_OUT_OF_MEMORY;
      }
      
      int i;
      
      for (i = 0; i < numArgs; i++)
      {
         argTypeDefs[i] = convertType(typeList[i]);
      }
      
      (*env)->ReleaseIntArrayElements(env, argTypes, typeList, JNI_ABORT);
   }
   
   // calculate the return type as an ffi type
   ffi_type* retTypeDef = convertType(retType);
   
   // prepare the function call definition
   ffi_status rc = ffi_prep_cif(&cif, abi, numArgs, retTypeDef, argTypeDefs);
   
   // there is a possible optimization here where we could store the function call
   // definition for future use, but it is not clear that this has any real positive
   // value AND we would have to manage (and clear) these resources which could be
   // quite messy; for now we aren't going down that path
   
   if (rc != FFI_OK)
   {
      if (rc == FFI_BAD_TYPEDEF)
         return com_goldencode_p2j_library_LibraryManager_ERROR_INVALID_TYPE;
      
      if (rc == FFI_BAD_ABI)
         return com_goldencode_p2j_library_LibraryManager_ERROR_INVALID_CALL_CONV;
      
      return com_goldencode_p2j_library_LibraryManager_ERROR_UNKNOWN_FAILURE;
   }
   
   // invoke the function call via pointer
   ffi_call(&cif, jlongToPointer(funcAddr), jlongToPointer(retAddr), jlongToPointer(argAddr));
      
   return com_goldencode_p2j_library_LibraryManager_ERROR_NONE;
}

/**
 * Converts the given library manager calling convention constant into the proper ffi type.
 *
 * @param    conv
 *           This must be a valid library manager CONVENTION_ constant.
 *
 * @return   The pointer to the corresponding ffi ABI type.
 */
ffi_abi calcCallingConvention(jint conv)
{
   // this could have been put into the platform-specific files, but by embedding it here we
   // ensure that the use of the libffi is encapsulated only in this file
   
   // only windows 32-bit platforms support multiple calling conventions, all other platforms
   // simply use the default
   ffi_abi abi = FFI_DEFAULT_ABI; 
   
   #ifdef WIN_TARGET
      #ifdef WORD_SIZE_32
      // override the default here because ffi uses MS-CDECL as the default which is not right
      if (conv == com_goldencode_p2j_library_LibraryManager_CONVENTION_DEFAULT ||
          conv == com_goldencode_p2j_library_LibraryManager_CONVENTION_STDCALL ||
          conv == com_goldencode_p2j_library_LibraryManager_CONVENTION_PASCAL)
      {
         // TODO: PASCAL may need different support
         abi = FFI_STDCALL;
      }
      else if (conv == com_goldencode_p2j_library_LibraryManager_CONVENTION_CDECL)
      {
         abi = FFI_MS_CDECL;  // or should this be FFI_SYSV?
      }
      #endif
   #endif

   return abi;
}

/**
 * Converts the given library manager type constant into the proper ffi type.
 *
 * @param    type
 *           This must be a valid library manager TYPE_ constant.
 *
 * @return   The pointer to the corresponding ffi type.
 */
ffi_type* convertType(int type)
{
   ffi_type* result = NULL;
   
   switch (type)
   {
      case com_goldencode_p2j_library_LibraryManager_TYPE_SIGNED_INT8:
         result = &ffi_type_sint8;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_UNSIGNED_INT8:
         result = &ffi_type_uint8;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_SIGNED_INT16:
         result = &ffi_type_sint16;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_UNSIGNED_INT16:
         result = &ffi_type_uint16;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_SIGNED_INT32:
         result = &ffi_type_sint32;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_UNSIGNED_INT32:
         result = &ffi_type_uint32;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_SIGNED_INT64:
         result = &ffi_type_sint64;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_FLOAT:
         result = &ffi_type_float;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_DOUBLE:
         result = &ffi_type_double;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_POINTER:
         result = &ffi_type_pointer;
         break;
      case com_goldencode_p2j_library_LibraryManager_TYPE_VOID:
         result = &ffi_type_void;
         break;
   }
   
   return result;
}

/**
 * Convert the given integer value into a void pointer in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    addr
 *           This must be a valid memory address.
 *
 * @return   The address as a void pointer.
 */
void* jlongToPointer(jlong addr)
{
   #ifdef WORD_SIZE_64
   void* ptr = (void*) addr;
   #else
   void* ptr = (void*) ((int32_t) addr);
   #endif
   
   return ptr;
}