/*
** Module   : memory.c
** Abstract : implementation of low level memptr functions called via JNI
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20130429 First version. This code should work on all operating systems and hardware
**                  architectures. The idea is to provide the minimum set of low level memory
**                  functions to enable the full set of memptr features to be implemented in
**                  pure Java. This class has no error processing or boundary checking. That
**                  is expected to be done in the caller.
** 002 GES 20131023 Used conditional preprocessing to emit code for 32-bit systems that will not
**                  trigger compiler warnings.
** 003 EVL 20180228 Fix for memory allocation approach.
** 003 GES 20180322 Added compiledWordSize().
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
#include <inttypes.h>
#include <stdlib.h>
#include <malloc.h>

#include "com_goldencode_p2j_util_MemoryManager.h"
                                           
/**
 * Reports on the native byte-ordering of the platform on which this library is
 * running. 
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 *
 * @return   true if the platform uses little-endian byte-ordering, false if the platform
 *           is big-endian.
 */ 
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_isLittleEndian(JNIEnv* env, jclass cls)
{
   // allocate a 32-bit integer on the stack and set its value such that only the most
   // significant byte is 0x00
   jint num = 0x00FFFFFF;
   
   // obtain the pointer to the lowest addressable byte in the integer
   char* base = (char*) &num;

   // if the contents of the lowest byte are non-zero, then it must not contain the most
   // significant byte and thus the platform is little-endian
   return *base ? JNI_TRUE : JNI_FALSE;   
}

/**
 * Returns the word size (the size of a pointer) for the native code as compiled. This may be
 * different than the word size of the operating system in the case where 32-bit code is run
 * on a 64-bit system.
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 *
 * @return   This value will be either 32 or 64 depending on how the native library was
 *           compiled.
 */ 
JNIEXPORT jint JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_compiledWordSize(JNIEnv* env, jclass cls)
{
   #ifdef WORD_SIZE_64
   jint result = 64;
   #else
   jint result = 32
   #endif
   
   return result;    
}

/**
 * Allocate the given number of bytes as a new memory block and return the 64-bit address. All
 * bytes in the memory region will be initialized to 0x00.
 * <p>
 * To duplicate the permissiveness of the 4GL memptr allocation (where certain operations often
 * can read or write past the end of the allocated buffer without an access violation), the
 * buffer allocated by this function will always be 16 bytes larger than the requested size. 
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    len
 *           The size in bytes of the memory region to allocate.
 *
 * @return   The address of the allocated memory region or 0 if memory could not be allocated.
 */ 
JNIEXPORT jlong JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_allocate(JNIEnv* env, jclass cls, jlong len)
{
   void* addr = calloc(len + 16, sizeof(jbyte));
   
   #ifdef WORD_SIZE_64
   jlong result = (jlong) addr;
   #else
   jlong result = (jlong) ((int32_t) addr);
   #endif
   
   return result;    
}

/**
 * De-allocate the memory block at the given 64-bit address.  Memory in this region MUST NOT
 * be accessed after this call. 
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    addr
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.  DO NOT deallocate the same pointer more than
 *           once.
 */ 
JNIEXPORT void JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_deallocate(JNIEnv* env, jclass cls, jlong addr)
{
   #ifdef WORD_SIZE_64
   free((void*) addr);
   #else
   free((void*) ((int32_t) addr));
   #endif
}

/**
 * Returns the next index position into the array which is the <code>null</code> byte, starting
 * at the given index position.
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    addr
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.
 * @param    idx
 *           The index position at which to start the search.
 * @param    size
 *           The length of the buffer. If the search reaches this byte without finding a
 *           null byte, then this value will be returned. If this is given as -1, then the
 *           caller has no size information and we operate in unsafe mode.  In this mode
 *           the code will read until it finds a <code>null</code> or gets an access
 *           violation.  This is how the 4GL does it too.
 *                                                                        
 * @return   The index of the next byte position which contains the <code>null</code> byte or
 *           the length of the array if no <code>null</code> byte is contained.
 */ 
JNIEXPORT jlong JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_findNextNull(JNIEnv* env,
                                                           jclass  cls,
                                                           jlong   addr,
                                                           jlong   idx,
                                                           jlong   size)
{
   #ifdef WORD_SIZE_64
   char* value = (char*) addr;
   #else
   char* value = (char*) ((int32_t) addr);
   #endif
   
   jlong i     = idx; 
   
   // not all instances will end in null byte, so check length first to
   // prevent overrun
   while ((size == -1 || i < size) && value[i] != '\0')
   {
      i++;
   }
   
   return i;
}

/**                                                       
 * Read the specified range of contents from the memory block at the given 64-bit address.
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    addr
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.
 * @param    offset
 *           The 0-based index position from the beginning of the buffer, from which to start
 *           reading.
 * @param    len
 *           The number of bytes to read. Must be greater than 0 and must NOT extend past the
 *           end of the buffer.
 *
 * @return   The range of bytes read or NULL if there is any problem allocating the array.
 */ 
JNIEXPORT jbyteArray JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_read(JNIEnv* env,
                                                   jclass  cls,
                                                   jlong   addr,
                                                   jlong   offset,
                                                   jlong   len)
{
   #ifdef WORD_SIZE_64
   jlong ptr = addr + offset;
   #else
   int32_t ptr = (int32_t) (addr + offset);
   #endif
   
   jbyteArray data = (*env)->NewByteArray(env, len);
   
   if (data != NULL)
   {
      (*env)->SetByteArrayRegion(env, data, 0, (jsize) len, (jbyte*) ptr);
   }
   
   return data;
}

/**
 * Write the specified contents to the memory block at the given 64-bit address plus offset.
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    addr
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.
 * @param    offset
 *           The 0-based index position from the beginning of the buffer, at which to start
 *           writing.
 * @param    data
 *           The bytes to write.  The entire contents of the array will be written. The length
 *           of the array must not be bigger than the number of bytes between the offset and
 *           the end of the buffer.
 *
 * @return   The range of bytes read.
 */ 
JNIEXPORT void JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_write(JNIEnv*    env,
                                                    jclass     cls,
                                                    jlong      addr,
                                                    jlong      offset,
                                                    jbyteArray data)
{
   #ifdef WORD_SIZE_64
   jlong ptr = addr + offset;
   #else
   int32_t ptr = (int32_t) (addr + offset);
   #endif
   
   jsize len = (*env)->GetArrayLength(env, data);
   (*env)->GetByteArrayRegion(env, data, 0, len, (jbyte*) ptr);
}

/**
 * Copy the specified range of contents from the source memory block to the destination memory
 * block.
 *
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 * @param    cls
 *           Reference to the Java class which called this function (from static context).
 * @param    source
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.
 * @param    dest
 *           The 64-bit pointer to a memory region previously allocated using this module.
 *           DO NOT pass in an invalid pointer.
 * @param    offset
 *           The 0-based index position from the beginning of the buffer, at which to start
 *           writing.
 * @param    len
 *           The number of bytes to write. Must be greater than 0 and must NOT extend past the
 *           end of the destination buffer.
 */ 
JNIEXPORT void JNICALL
   Java_com_goldencode_p2j_util_MemoryManager_copy(JNIEnv* env,
                                                   jclass  cls,
                                                   jlong   source,
                                                   jlong   dest,
                                                   jlong   offset,
                                                   jlong   len)
{
   #ifdef WORD_SIZE_64
   jbyte* src = (jbyte*) source;
   jbyte* dst = (jbyte*) (dest + offset);
   #else
   jbyte* src = (jbyte*) ((int32_t) source);
   jbyte* dst = (jbyte*) ((int32_t) (dest + offset));
   #endif

   jlong i;
   
   for (i = 0; i < len; i++)
   {
      dst[i] = src[i];
   }
}