/*
** Module   : signals_linux.c
** Abstract : Linux implementation for initialize calls to prepare native layer to work properly.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20140117 Created initial version to implement init steps.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include "terminal.h"
#include "signals.h"

#include <pthread.h>
#include <signal.h>

// Internal Linux specific functions and variables

// Semaphores to handle the SINGNAL event
pthread_cond_t signalEvent;
pthread_mutex_t signalMutex;

/**
 * postSigintEvent - posting the SIGING event for blocking threads. Used to unblock the waiting
 *                   threads in regular cycle or shutdown.
 */ 
void postSigintEvent()
{
   struct sigaction sa;
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);

   sa.sa_handler = SIG_IGN;
   // Ignore the signals for other SIGINT's
   sigaction(SIGINT,  &sa, NULL);

   // Posting the event semaphore   
   pthread_mutex_lock(&signalMutex);
   pthread_cond_signal(&signalEvent);
   pthread_mutex_unlock(&signalMutex);
}

/**
 * sigintHandler - The new handler for SIGINT incoming events.
 */ 
void sigintHandler(int sig)
{
   postSigintEvent();
}

// Native calls itself to be called from Java

/**
 * initSignalsProcessing - Set up signal handling to ignore job control signals and redefine
 *                         CTRL-C handlers(reflecteds in the child process as well).
 *                         Linux version.
 */ 
JNIEXPORT void JNICALL initSignalsProcessing(JNIEnv* env)
{
   struct sigaction sa;
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);
   
   // Create event semaphore
   if (pthread_cond_init(&signalEvent, NULL) != 0 ||
       pthread_mutex_init(&signalMutex, NULL) != 0)
   {
      // Unable to create semaphore, terminating
      throwException(env, "java/lang/IllegalStateException", "Event semaphore creation failure");
      return;
   }
   
   // Need to watch for external INT signals to react as for CTRL-C 
   sa.sa_handler = sigintHandler;
   sigaction(SIGINT,  &sa, NULL);
   
   // The rest are ingnoring   
   sa.sa_handler = SIG_IGN;
   sigaction(SIGTSTP, &sa, NULL);
   sigaction(SIGTTIN, &sa, NULL);
   sigaction(SIGTTOU, &sa, NULL);
}

/**
 * waitSigint - Intercepts external SIGING signal to inform the caller the special processing is
 *              required to handle.
 */ 
JNIEXPORT jint JNICALL waitSigint()
{
   jint result = -1; // This is unexpected return code
   
   // Wait for SIGINT to arrive
   pthread_mutex_lock(&signalMutex);
   if (pthread_cond_wait(&signalEvent, &signalMutex) == 0)
   {
      result = 0;
   }
   pthread_mutex_unlock(&signalMutex);
   
   return result;
}

/**
 * unlockSigint - Enable SIGINT handler again.
 */ 
JNIEXPORT void JNICALL unlockSigint(JNIEnv* env)
{
   struct sigaction sa;
   
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);

   // Need to watch for external INT signals to react as for CTRL-C 
   sa.sa_handler = sigintHandler;
   sigaction(SIGINT,  &sa, NULL);
}

/**
 * stopWaitingSigint - Disable block waitings for SIGINT.
 */ 
JNIEXPORT void JNICALL stopWaitingSigint()
{
   postSigintEvent();
}