/*
** Module   : ClassicTheme.java
** Abstract : Windows Classic theme handler. 
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170224 First version.
** 002 SVL 20170418 Fixed drawing of browse sort numbers.
** 003 EVL 20170512 Adding method to get default font aliases.
** 004 OM  20170406 Added Editor, Rectangle and 3DBorder support. Fixed Tooltip support.
** 005 OM  20170625 EditorGuiImpl signature change.
** 006 SVL 20170814 Fixed drawing of title bar for editable browse. Added drawing with
**                  "intermediate" rows. Reflected changes in cell attributes.
** 007 OM  20170705 Fixed Rectangle rendering regression. Added drawSelectionFocus() method.
** 008 SVL 20170901 Added getRowToggleOffset, getRowToggleSize, drawRowToggleImage.
** 009 EVL 20170818 Changing default color constant for application workspace.  Fix for rounded
**                  rectangle issue.
** 010 HC  20170823 Added runtime support of WINDOW:SMALL-TITLE.
** 011 OM  20171027 Replaced concentric drawing of rectangles with a single stroked primitive.
** 012 EVL 20171209 Adding submenu arrow painting fix.  Fixed submenu drawing background when
**                  submenu is not hovered and the the items are not displaying.
** 013 SVL 20180115 Added functions for getting offsets for a text inside browse cells and labels.
**     EVL 20180124 Separating possible theme dependent browse label text drawing method.
** 014 SBI 20180130 Implemented common and specialized versions of getTextLookAndFeel(..) for text
**                  widgets and for fillins respectively, removed unused drawFillInSelection(..),
**                  changed drawEditorContent in order to use GuiDriver.this.drawTextSelection.
** 015 SVL 20180321 Refactoring related to getCellAttr and getCellAttrWithIntermediate signature
**                  change.
** 016 OM  20180321 Added multiline support for tooltips. Added support for file/dir choosers.
**                  Fixed alert box geometry.
** 017 HC  20180509 Fixed drawing of menu items in the root of a MENU-BAR.
** 018 OM  20180503 Added getBinaryResourceLocation().
** 019 CA  20180530 Fixed unknown and empty FRAME:TITLE.
** 020 OM  20180701 Changed to hide a rectangle outline if its EDGE-PIXELS(EDGE-CHARS) is zero.
** 021 OM  20180710 Added support for Slider widgets.
** 022 SVL 20180529 Support for enhanced browse coloring.
** 023 OM  20180808 Fixed browse sub-components drawing on Windows8.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.theme;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Frame;
import com.goldencode.p2j.ui.client.Label;
import com.goldencode.p2j.ui.client.Menu;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.LightweightTextWidget.*;
import com.goldencode.p2j.ui.client.gui.theme.Theme.TextLookAndFeel.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;

import static com.goldencode.p2j.ui.BrowseColumnWidgetType.*;
import static com.goldencode.p2j.ui.client.SysColor.*;
import static com.goldencode.p2j.ui.client.gui.BrowseGuiImpl.*;
import static com.goldencode.p2j.ui.client.gui.BrowseColumnGuiImpl.*;
import static com.goldencode.p2j.ui.client.gui.SliderGuiImpl.*;

/**
 * This {@link Theme} emulates the look of old Windows classic theme. It features very static UI,
 * with little or no sensitivity to mouse hovering. The color palette is low, with few colors
 * outside the line of gray tones. This makes the theme useful when application need to run with
 * lower resources.
 * <p>
 * The theme features multiple widgets with 3D aspect in order to improve overall look.
 */
public class ClassicTheme
implements Theme
{
   /** Standard titlebar height for this theme. */
   private static final int TITLE_BAR_HEIGHT = 18;
   
   /** Standard caption button width for this theme. */
   private static final int CAPTION_BUTTON_WIDTH = 16;
   
   /** Standard caption button height for this theme. */
   private static final int CAPTION_BUTTON_HEIGHT = 14;
   
   /** Small caption button height for this theme. */
   private static final int SMALL_CAPTION_BUTTON_HEIGHT = 11;

   /** Small caption button width for this theme. */
   private static final int SMALL_CAPTION_BUTTON_WIDTH = 13;

   /** Border size of main window border. */
   private static final int BORDER_SIZE = 4;
   
   /** Border size of modal window border. */
   private static final int MODAL_BORDER_SIZE = 3;
   
   /** Prebuilt image size of the arrow style button. */
   private static final int PREBUILT_ARROW_IMAGE_SIZE = 15;
   
   /** Radio Button Control image size. */
   private static final int RADIO_BUTTON_IMAGE_SIZE = 16;
   
   /** Radio Button Radio buttons extra distance. */
   private static final int RADIO_BUTTON_EXTRA_DIST = 3;
   
   /** Radio Button Control image reserved space. */
   private static final int RADIO_BUTTON_IMAGE_SPACE = RADIO_BUTTON_IMAGE_SIZE + RADIO_BUTTON_EXTRA_DIST;
   
   /** The space between checkbox and its label. */
   private static final int CHECKBOX_SPACE = 7;
   
   /** Checkbox width or length */
   private static final int CHECKBOX_SIZE = 11;
   
   /** This menu-item left padding in a sub-menu body or in a MENUBAR in pixels. */
   private static final int MENU_LEFT_PAD = 0;
   
   /** This menu-item right padding in a sub-menu body or in a MENUBAR in pixels. */
   private static final int MENU_RIGHT_PAD = 6;
   
   /** Check-mark width (for toggle-box) in pixels. */
   private static final int MENU_CHECKMARK_WIDTH = 15;
   
   /** Sub-menu arrow width in pixels. */
   private static final int MENU_SM_ARROW_WIDTH = 16;
   
   /** The overall pixel width for vertical scrollbar or height for horizontal scrollbar. */
   public static final int SCROLLBAR_WIDTH = 16;
   
   /** Scrollbar's Thumb button minimum size */
   public static final int SCROLLBAR_THUMB_MIN_SIZE = 8;
   
   /** Padding for alert boxes. */
   private static int ALERT_BOX_PADDING = 20;
   
   /** Additional offset for 3d borders. */
   public static final int THREE_D_BORDER_OFFSET = 2;
   
   /** Size of the drop-down button image. */
   private static final int COMBO_BOX_BUTTON_SIZE = 16;
   
   /** Shift in X and Y to draw entry field for DROP-DOWN and SIMPLE modes. */
   private static final int COMBO_BOX_ENTRY_SHIFT = 2;

   /** Dimension delta to draw entry field for DROP-DOWN and SIMPLE modes. */
   private static final int COMBO_BOX_ENTRY_DELTA = COMBO_BOX_ENTRY_SHIFT + 2;

   /** Color of the outline of the current toggle-box cell. */
   private static final ColorRgb ROW_TOGGLE_OUTLINE_COLOR = new ColorRgb(0, 0, 0);
   
   /**
    * Color which is XORed with background to draw current row highlighting (dotted rectangle) in
    * browse widget.
    */
   private static final ColorRgb BROWSE_ROW_XOR_COLOR = new ColorRgb(192, 192, 192);

   /** Left offset of a text inside a browse label. */
   private static final int BROWSE_LABEL_OFFSET_LEFT = 2;

   /** Left offset of a text inside a browse label, if column is right-aligned. */
   private static final int BROWSE_LABEL_OFFSET_LEFT_RA = 1;

   /** Right offset of a text inside a browse label. */
   private static final int BROWSE_LABEL_OFFSET_RIGHT = 2;

   /** Left offset of a text inside a browse cell. */
   private static final int BROWSE_DATA_OFFSET_LEFT = 2;

   /** Left offset of a text inside a browse cell, if column is right-aligned. */
   private static final int BROWSE_DATA_OFFSET_LEFT_RA = 1;

   /** Default size of an in-browse toggle-box. */
   protected static final int ROW_TOGGLE_SIZE = 13;

   /** The image object for background fill. */
   private ImageWrapper<?> scrollbarImgBg = null;
   
   /** Prohibits multiple attempts to load {@code scrollbarImgBg}. */
   private boolean scrollbarImgBgLoaded = false;
   
   /** Cache with images from caption buttons. */
   private Map<String, ImageWrapper<?>> captionButtonImages = new HashMap<>();
   
   /** Cached map with default system colors. */
   private Map<Integer, ColorRgb> defaultColors = null;
   
   /**
    * Get the name of this {@code Theme}.
    *
    * @return  the name of this theme.
    */
   @Override
   public String getName()
   {
      return "classic";
   }
   
   /**
    * Release any held resources when this {@code Theme} is removed. 
    */
   public void releaseResources()
   {
      captionButtonImages.clear();
      scrollbarImgBg = null;
      scrollbarImgBgLoaded = false;
   }
   
   /**
    * Obtain an image resource specific to this {@code Theme}. The image can be cached for
    * performance reasons. It will be automatically released when {@link #releaseResources()} is
    * called, when a new {@code Theme} is loaded and this one is removed.
    *
    * @param   name
    *          The name of the image as path within jar bundle.
    * @param   width
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@link ImageWrapper} to know the exact value.  
    * @param   height
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@code ImageWrapper} to know the exact value.  
    * @param   antialias
    *          if {@code true} and image is scaled to meet the size hints, an smooth scaling
    *          algorithm will be used, otherwise a faster one will be used but the image quality
    *          will be lower. If image is not scaled the parameter is ignored.
    *
    * @return  an {@code ImageWrapper} containing the requested image resource if it can be found
    *          and loaded and {@code null} otherwise.
    */
   @Override
   public ImageWrapper getImage(String name, int width, int height, boolean antialias)
   {
      return VectorGraphicsHelper.getScalableImage(name, width, height);
   }
   
   /**
    * Obtain the location of binary resources (like pixmaps and sprites) used by this theme. The 
    * location is relative to @Theme.class@ location so that the resources could be loaded easily
    * using @getResourceAsStream()@.
    *
    * @return  the location of binary resources used by this theme.
    */
   @Override
   public String getBinaryResourceLocation()
   {
      return "classic";
   }
   
   /**
    * Check if the theme is mouse hover sensitive. Mouse hover sensitive themes need to repaint
    * the components when the mouse hovers over them (ie. without clicking them).
    *
    * @return  always {@code false}. This theme is NOT mouse hover sensitive.
    */
   @Override
   public boolean isMouseHoverSensitive()
   {
      return false;
   }
   
   /**
    * Check whether this is a classic theme. A classic theme has a low color palette and usually
    * does not have animations which makes it better suited for low UI resources.  Usually classic
    * themes are not mouse hover sensitive. 
    *
    * @return  Always {@code true}. 
    */
   @Override
   public boolean isClassic()
   {
      return true;
   }
   
   /**
    * Obtain the default system palette for this theme.
    *
    * @return  the default system palette for this theme.
    */
   @Override
   public Map<Integer, ColorRgb> getDefaultSystemPalette()
   {
      if (defaultColors == null)
      {
         Map<Integer, ColorRgb> pal = new HashMap<>();
         pal.put(COLOR_SCROLLBAR, new ColorRgb(212, 208, 200));
         pal.put(COLOR_BACKGROUND, new ColorRgb(58, 110, 165));
         pal.put(COLOR_ACTIVECAPTION, new ColorRgb(10, 36, 106));
         pal.put(COLOR_INACTIVECAPTION, new ColorRgb(128, 128, 128));
         pal.put(COLOR_MENU, new ColorRgb(212, 208, 200));
         pal.put(COLOR_WINDOW, new ColorRgb(255, 255, 255));
         pal.put(COLOR_WINDOWFRAME, new ColorRgb(100, 100, 100));
         pal.put(COLOR_MENUTEXT, new ColorRgb(0, 0, 0));
         pal.put(COLOR_WINDOWTEXT, new ColorRgb(0, 0, 0));
         pal.put(COLOR_CAPTIONTEXT, new ColorRgb(255, 255, 255));
         pal.put(COLOR_ACTIVEBORDER, new ColorRgb(212, 208, 200));
         pal.put(COLOR_INACTIVEBORDER, new ColorRgb(212, 208, 200));
         pal.put(COLOR_APPWORKSPACE, new ColorRgb(132, 130, 133));
         pal.put(COLOR_HIGHLIGHT, new ColorRgb(10, 36, 106));
         pal.put(COLOR_HIGHLIGHTTEXT, new ColorRgb(255, 255, 255));
         pal.put(COLOR_3DFACE, new ColorRgb(212, 208, 200));
         pal.put(COLOR_3DSHADOW, new ColorRgb(128, 128, 128));
         pal.put(COLOR_GRAYTEXT, new ColorRgb(128, 128, 128));
         pal.put(COLOR_BTNTEXT, new ColorRgb(0, 0, 0));
         pal.put(COLOR_INACTIVECAPTIONTEXT, new ColorRgb(212, 208, 200));
         pal.put(COLOR_BTNHIGHLIGHT, new ColorRgb(255, 255, 255));
         pal.put(COLOR_3DDKSHADOW, new ColorRgb(64, 64, 64));
         pal.put(COLOR_3DLIGHT, new ColorRgb(212, 208, 200));
         pal.put(COLOR_INFOTEXT, new ColorRgb(0, 0, 0));
         pal.put(COLOR_INFOBK, new ColorRgb(255, 255, 255));
         pal.put(COLOR_BUTTONALTERNATEFACE, new ColorRgb(181, 181, 181));
         pal.put(COLOR_HOTLIGHT, new ColorRgb(0, 0, 128));
         pal.put(COLOR_GRADIENTACTIVECAPTION, new ColorRgb(166, 202, 240));
         pal.put(COLOR_GRADIENTINACTIVECAPTION, new ColorRgb(192, 192, 192));
         pal.put(COLOR_MENUHILIGHT, new ColorRgb(10, 36, 106));
         pal.put(COLOR_MENUBAR, new ColorRgb(212, 208, 200));
         pal.put(COLOR_TOOLTIPBK, new ColorRgb(255, 255, 231));  // shouldn't this be COLOR_INFOBK?
         pal.put(COLOR_FILLIN_BG, new ColorRgb(255, 255, 255));  // this must map to something official in Windows
         pal.put(COLOR_FILLIN_FG, new ColorRgb(0, 0, 0));        // this must map to something official in Windows
         pal.put(COLOR_CARET, new ColorRgb(0, 0, 0));            // this must map to something official in Windows
         
         defaultColors = Collections.unmodifiableMap(pal);
      }
      return defaultColors;
   }
   
   /**
    * Obtain the map of standard fonts specific to current {@code Theme}, to which a key defaults,
    * if not overridden in the directory.
    *
    * @return  the map of standard fonts specific to this {@code Theme}. 
    */
   @Override
   public Map<String, String> getStandardFontTable()
   {
      Map<String, String> fontTable = new HashMap<>();
      
      fontTable.put("default-font",        "MS Sans Serif, size=8");
      fontTable.put("default-fixed-font",  "Courier New, size=8");
      
      // standard fonts
      fontTable.put("font0", "Courier New, size=8");
      fontTable.put("font1", "MS Sans Serif, size=8");
      fontTable.put("font2", "Courier New, size=8");
      fontTable.put("font3", "Courier New, size=8");
      fontTable.put("font4", "MS Sans Serif, size=8");
      fontTable.put("font5", "MS Sans Serif, size=10");
      fontTable.put("font6", "MS Sans Serif, size=8, bold");
      fontTable.put("font7", "MS Sans Serif, size=8");
      
      // system fonts
      fontTable.put(FontTable.WINDOW_TITLE_FONT,       "Tahoma, size=8, bold");
      fontTable.put(FontTable.ALERT_BOX_FONT,          "Tahoma, size=8");
      fontTable.put(FontTable.STACK_TRACE_WINDOW_FONT, "MS Sans Serif, size=8");
      fontTable.put(FontTable.MENU_FONT,               "Tahoma, size=8");
      fontTable.put(FontTable.SORT_NUMBER_FONT,        "MS Sans Serif, size=7");
      
      // freeze the map
      return Collections.unmodifiableMap(fontTable);
   }
   
   /**
    * Obtain the map of default font aliases specific to current {@code Theme}, to which a key
    * defaults, if not overridden in the directory.
    * 
    * @return  the map of standard font aliases specific to this {@code Theme}. 
    */
   @Override
   public Map<String, String> getDefaultFontAliases()
   {
      Map<String, String> fontTable = new HashMap<>();
      
      fontTable.put("default-font",        "Lato, size=8");
      fontTable.put("default-fixed-font",  "Liberation Mono, size=8");
      
      // standard fonts
      fontTable.put("font0", "Liberation Mono, size=8");
      fontTable.put("font1", "DejaVu Sans, size=10, bold");
      fontTable.put("font2", "Liberation Mono, size=8");
      fontTable.put("font3", "Liberation Mono, size=8");
      fontTable.put("font4", "Liberation Sans, size=8");
      fontTable.put("font5", "Liberation Sans, size=10");
      fontTable.put("font6", "Liberation Sans, size=8, bold");
      fontTable.put("font7", "DejaVu Sans, size=10, bold");
      
      // system fonts
      fontTable.put(FontTable.WINDOW_TITLE_FONT,       "Wine Tahoma, size=8, bold");
      fontTable.put(FontTable.ALERT_BOX_FONT,          "Wine Tahoma, size=8");
      fontTable.put(FontTable.STACK_TRACE_WINDOW_FONT, "Liberation Sans, size=8");
      fontTable.put(FontTable.MENU_FONT,               "Wine Tahoma, size=8");
      fontTable.put(FontTable.SORT_NUMBER_FONT,        "Liberation Sans, size=7");
      
      // freeze the map
      return Collections.unmodifiableMap(fontTable);
   }
   
   /**
    * Utility method that returns the size of the prebuilt arrow images.  
    *
    * @return  size of prebuilt arrow size.
    */
   @Override
   public int getPrebuiltArrowImageSize()
   {
      return PREBUILT_ARROW_IMAGE_SIZE;
   }
   
   /**
    * Returns the height of window's titleBar. It DOES NOT include the upper window border. 
    *
    * @param   maximized
    *          On some themes, the maximized window have a different title bar height when 
    *          maximized. 
    *
    * @return  the height of the window's title bar. 
    */
   @Override
   public int getTitleBarHeight(boolean maximized)
   {
      return TITLE_BAR_HEIGHT;
   }
   
   /**
    * Returns the window's border insets. In classic theme, all insets are equals.
    *
    * @param   modal
    *          Usually the modal dialogs/windows have a different width if modal. 
    *
    * @return  the insets of window's border. 
    */
   @Override
   public NativeInsets getWindowBorderInsets(boolean modal)
   {
      int w = modal ? MODAL_BORDER_SIZE : BORDER_SIZE;
      return new NativeInsets(w, w, w, w);
   }
   
   /**
    * Returns the height of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the height of buttons is not the same.
    * @param   maximized
    *          On some themes, the maximized window have a different button height when 
    *          maximized. 
    *
    * @return  the height of the caption button. 
    */
   @Override
   public int getCaptionButtonHeight(CaptionButtonType type, boolean maximized)
   {
      return CAPTION_BUTTON_HEIGHT;
   }
   
   /**
    * Returns the width of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the width of buttons is not the same.
    *
    * @param   dialog
    *          {@code true} for dialogs. The dialogs may have specific width for caption buttons.  
    *
    * @return  the width of the requested caption button. 
    */
   @Override
   public int getCaptionButtonWidth(CaptionButtonType type, boolean dialog)
   {
      return CAPTION_BUTTON_WIDTH;
   }
   
   /**
    * Get the alignment of caption button for this theme. The returned value combines both
    * horizontal and vertical alignment of the buttons.
    *
    * @param   type
    *          On some specific themes, the alignment of buttons is not the same.
    *
    * @return  the alignment of caption button.
    */
   @Override
   public int getCaptionButtonsAlignment(CaptionButtonType type)
   {
      return ALIGN_RIGHT | ALIGN_MIDDLE;
   }
   
   /**
    * Get the title alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of title.
    */
   @Override
   public int getWindowTitleAlignment()
   {
      return ALIGN_LEFT | ALIGN_MIDDLE;
   }
   
   /**
    * Get the icon alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of icon on titlebar.
    */
   @Override
   public int getWindowIconAlignment()
   {
      return ALIGN_LEFT | ALIGN_BOTTOM;
   }
   
   /**
    * Return the amount of space to the right of this button. On some themes, different spaces can
    * be found between the caption buttons. Although unusual, negative values can be returned,
    * when buttons overlap for a few pixels. 
    *
    * @param   type
    *          The caption button type.
    *
    * @return  the amount of space to the right of this button.
    */
   @Override
   public int getSpaceBeforeButton(CaptionButtonType type)
   {
      // in Classic theme, the Close button is separated by 2 pixels on each side
      return (type == CaptionButtonType.CLOSE || type == CaptionButtonType.MAXIMIZE) ? 2 : 0;
   }
   
   /**
    * Obtain the {@link java.awt.Color} used to render the windows title.
    *
    * @param   active
    *          {@code true} if the rendered window is active. The inactive window is usually fade. 
    *
    * @return  the {@code Color} used to render the windows title.
    */
   @Override
   public java.awt.Color getTitleBarTextColor(boolean active)
   {
      return active ? java.awt.Color.WHITE : java.awt.Color.GRAY;
   }
   
   /**
    * Do the layout of a titlebar. Each {@code Theme} can define how its components are arranged. 
    *
    * @param   window
    *          The window whose titlebar is laid out. 
    * @param   icon
    *          The icon widget of the window (if any).
    * @param   title
    *          The {@link Label} that displays the window's title. 
    * @param   buttons
    *          An array with the {@link CaptionButton}-s present on the titlebar of this window. 
    * @param   width
    *          The width of the window (in character units).
    * @param   c
    *          the {@link CoordinatesConversion} used for possible units conversions.
    *
    * @return  the computed titlebar height.
    */
   @Override
   public double layoutTitleBar(TopLevelWindow window,
                                Widget icon,
                                Label title,
                                CaptionButton[] buttons,
                                double width,
                                CoordinatesConversion c)
   {
      NativeInsets iconInsets = new NativeInsets(1, 2, 1, 2);
      NativeInsets titleInsets = new NativeInsets(1, 3, 2, 2);
      NativeInsets buttonInsets = new NativeInsets(2, 2, 2, 2);
      
      int font = window.resolveTitleFont();
      // compute the window title's font height, in pixels.  this will determine the titlebar's
      // height.
      int fheight = c.heightToNative(1, window.getId(), font) + 
                    titleInsets.top + 
                    titleInsets.bottom;

      boolean small = false;
      WidgetConfig cfg = window.config();
      if (cfg instanceof WindowConfig)
      {
         small = ((WindowConfig) cfg).smallTitle;
      }
      
      if (icon != null)
      {
         // - icon on left-side; to ensure the character unit location is properly in sync with
         // the native location, compute the character location first and re-transform to native
         // for the physical location.
         icon.setLocation(c.widthFromNative(iconInsets.left), c.heightFromNative(iconInsets.top));
         NativePoint iconLoc = c.pointToNative(icon.location());
         icon.setPhysicalLocation(iconLoc.x, iconLoc.y);
      }
      
      int btnWidth = small ? SMALL_CAPTION_BUTTON_WIDTH : CAPTION_BUTTON_WIDTH;
      int btnHeight = small ? SMALL_CAPTION_BUTTON_HEIGHT : CAPTION_BUTTON_HEIGHT;
      
      // title height can not be less than the default caption button height, 
      // also give 4 more pixels for margins
      fheight = Math.max(fheight, btnHeight + 4);
      
      // title height: needs to be done early, icon's height depends on it
      title.setHeight(c.heightFromNative(fheight));
      double barHeight = title.height();
      // to return setHeight(barHeight);
      
      // - minimize, maximize, close buttons
      
      double bw = c.widthFromNative(btnWidth);
      double bh = c.heightFromNative(btnHeight);
      
      // add the buttons - all coordinates are relative to this container.
      double bx = width;
      // determine the vertical location so that the button is centered vertically
      double by = c.heightFromNative((c.heightToNative(barHeight) - btnHeight) / 2);
      
      for (int i = buttons.length - 1; i >= 0; i--)
      {
         CaptionButton btn  = buttons[i];
         btn.setDimension(bw, bh);
         
         int buttonSpace = btnWidth + buttonInsets.right;
         
         if (btn.getType() == CaptionButtonType.ICONIFY)
         {
            // TODO: the buttons are touching each other except of the close button, 
            //       that has 2 pixels left-margin. 
            //       make this the button's property or at least make the algorithm more generic.
            buttonSpace -= 2;
         }
         
         bx = bx - c.widthFromNative(buttonSpace);
         
         btn.setLocation(bx, by);
         
         NativePoint btnLoc = c.pointToNative(btn.location());
         btn.setPhysicalLocation(btnLoc.x, btnLoc.y);
      }
      
      double iconColumn = (icon != null && !icon.hidden()) ?
                          icon.bounds().right() :
                          c.widthFromNative(-1);
      
      // - title in the middle, between the icon and the minimize button
      title.setLocation(iconColumn + c.widthFromNative(titleInsets.left), 0);
      NativePoint titleLoc = c.pointToNative(title.location());
      title.setPhysicalLocation(titleLoc.x, titleLoc.y);
      
      // set title width
      double btnsCol = buttons.length == 0 ? width : buttons[0].bounds().left();
      double twidth = btnsCol -
                      title.location().x         - 
                      c.widthFromNative(titleInsets.right);
      title.setWidth(twidth);
      
      return barHeight + c.heightFromNative(1); // 1 pixel width gray line
   }
   
   /**
    * Obtain the {@link Font} used to render the windows title.
    *
    * @return the {@code Font} used to render the windows title.
    */
   @Override
   public Font getTitleBarFont()
   {
      return null;
   }
   
   /**
    * Renders a {@link CaptionButton}.
    *
    * @param   button
    *          The {@code CaptionButton} to be drawn. 
    * @param   gd
    *          The {@link GuiDriver} used for drawing operations. Note that internal state of 
    *          this object can affect the resulting output.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawCaptionButton(CaptionButton button,
                                 GuiDriver gd,
                                 int x,
                                 int y,
                                 int width,
                                 int height)
   {
      boolean enabled = button.isEnabled();
      boolean pressed = button.isPressed() && button.isMouseOver();
      
      // build up the name of the image resource:
      StringBuilder nameBuilder = new StringBuilder("wnd-");
      CaptionButtonType type = button.getType();
      switch (type)
      {
         case CLOSE:
            nameBuilder.append("close-");
            break;
         case CLOSE_SMALL:
            nameBuilder.append("close-small");
            break;
         case HELP:
            nameBuilder.append("help-");
            break;
         case ICONIFY:
            nameBuilder.append("min-");
            break;
         case MAXIMIZE:
            nameBuilder.append("max-");
            break;
         case RESTORE_DOWN:
         case RESTORE_UP:
            nameBuilder.append("restore-");
            break;
      }
      
      if (type != CaptionButtonType.CLOSE_SMALL)
      {
         nameBuilder.append(enabled ? "en" : "dis");
      }
      
      // preload button images from theme's jar on first painting
      String buttonName = nameBuilder.toString();
      String buttonPath = "theme/" + getBinaryResourceLocation() + "/" + buttonName + ".svg";
      
      boolean small = false;
      WidgetConfig cfg = button.topLevelWindow().config();
      if (cfg instanceof WindowConfig)
      {
         small = ((WindowConfig) cfg).smallTitle;
      }
      
      int btnWidth = small ? SMALL_CAPTION_BUTTON_WIDTH : CAPTION_BUTTON_WIDTH;
      int btnHeight = small ? SMALL_CAPTION_BUTTON_HEIGHT : CAPTION_BUTTON_HEIGHT;
      
      ImageWrapper<?> img = captionButtonImages.computeIfAbsent(buttonPath,
                  k -> getImage(buttonName, btnWidth, btnHeight, false));
      
      // the button body
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.fillRect(x, y, width, height);
      // draw 3D border depending on state
      // top-left part
      gd.setColor(gd.getSysColor(pressed ? SysColor.COLOR_3DDKSHADOW : SysColor.COLOR_WINDOW));
      gd.drawLine(x, y, x, height - 2);
      gd.drawLine(x + 1, y, width - 2, y);
      // bottom-right part
      gd.setColor(gd.getSysColor(pressed ? SysColor.COLOR_WINDOW : SysColor.COLOR_3DDKSHADOW));
      gd.drawLine(width - 1, y, width - 1, height - 1);
      gd.drawLine(x, height - 1, width - 1, height - 1);
      // shadow corner
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      if (pressed)
      {
         gd.drawLine(x + 1, y + 1, x + 1, height - 3);
         gd.drawLine(x + 2, y + 1, width - 3, y + 1);
      }
      else
      {
         gd.drawLine(width - 2, y + 1, width - 2, height - 2);
         gd.drawLine(x + 1, height - 2, width - 3, height - 2);
      }
   
      if (img == null)
      {
         return; // nothing more to do
      }
      
      // if enabled, emulate 'press' effect by translating the image by (1, 1) 
      int dx = (enabled & pressed) ? 1 : 0;
      int dy = (enabled & pressed) ? 1 : 0;
      gd.drawImage(img, x + dx, y + dy, width, height, 0, 0, false, false, false, false);
   }
   
   /**
    * Renders the background of the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitleBar(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean active)
   {
      gd.setColor(gd.getSysColor(active ? SysColor.COLOR_ACTIVECAPTION : 
                                          SysColor.COLOR_INACTIVECAPTION));
      gd.fillRect(x, y, width, height - 1);
      
      gd.setColor(gd.getSysColor(active ? SysColor.COLOR_ACTIVEBORDER :
                                          SysColor.COLOR_INACTIVEBORDER));
      gd.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
   }
   
   /**
    * Renders the text on the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@link GuiFontResolver} for selecting the required font.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   title
    *          The title of the window.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitle(GuiDriver gd,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               String title,
                               boolean active)
   {
      int color = active ? SysColor.COLOR_ACTIVECAPTION : SysColor.COLOR_INACTIVECAPTION;
      
      // remove old title
      gd.setColor(gd.getSysColor(color));
      gd.fillRect(x, y, width, height);
      
      if (title.length() > 0)
      {
         FontDetails<?> font = gf.font().font;
         
         // draw text
         gd.setColor(gd.getSysColor(SysColor.COLOR_CAPTIONTEXT));
         gd.setGuiFont(font);
         
         gd.drawStringCentered(title, x, height / 2 - 1);
      }
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border
    *          The width of the border on each side.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowBorder(GuiDriver gd,
                                int x,
                                int y,
                                int width,
                                int height,
                                NativeInsets border,
                                boolean active)
   {
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      
      // clear right and bottom borders
      gd.fillRect(width - border.right, y,                      border.right,  height);
      gd.fillRect(x,                    height - border.bottom, width,         border.bottom);
      
      // clear left and top borders
      gd.fillRect(x, y, border.left, height);
      gd.fillRect(x, y, width,       border.top);
      
      // 3D simulation for the window border
      // light part
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x + 1, y + 1, x + 1,     height - 3);
      gd.drawLine(x + 2, y + 1, width - 3, y + 1);
      // dark part
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawLine(x + 1,     height - 2, width - 2, height - 2);
      gd.drawLine(width - 2, y + 1,      width - 2, height - 2);
      // very dark part
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DDKSHADOW));
      gd.drawLine(x,         height - 1, width - 1, height - 1);
      gd.drawLine(width - 1, y,          width - 1, height - 2);
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   button
    *          The {@link ButtonGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawButton(ButtonGuiImpl button,
                          GuiDriver gd,
                          GuiColorResolver gc,
                          GuiFontResolver gf,
                          int x,
                          int y,
                          int width,
                          int height)
   {
      ButtonConfig bcfg = button.config();
      NativePoint po = new NativePoint(x, y);
      NativeDimension d = new NativeDimension(width, height);
   
      boolean flat =        bcfg.flatButton;
      boolean highlighted = !flat && button.highlighted();
      boolean isDefault =   button.isDefaultButton();
      boolean mouseOver =   button.isMouseOver();
      boolean pressed =     button.isPressed() && mouseOver;
   
      // draw button itself
      if (button.hasImages() && !isDefault)
      {
         // check for built-in imaged
         if (button.hasPrepackagedImages(bcfg))
         {
            drawPrepackagedImages(button, gd, gc, po, d, pressed, flat, mouseOver);
         }
         else
         {
            drawButtonImages(button, gd, gc, po, d, pressed, flat, mouseOver);
         }
      }
      else
      {
         // draw button border
         drawButtonBorder(gd, po, d, highlighted, pressed, flat, mouseOver);
         // and the body
         drawButtonInternal(button, gd, gc, gf, po, d, highlighted, pressed, flat, mouseOver);
      }
   
      // focused button needs special highlight
      if (highlighted)
      {
         gd.setColor(gc.fgColor);
         gd.drawRect(x, y, width - 1, height - 1);         
      }
   }
   
   /**
    * Draw bottom border simulating 3D effect.
    * 
    * @param   gd
    *          The GuiDriver.
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          Is the button flat?
    * @param   mouseIn
    *          {@code true} if mouse hovering the button.
    */
   private void drawButtonBorder(GuiDriver gd,
                                 NativePoint orig,
                                 NativeDimension dim,
                                 boolean focused,
                                 boolean pressed,
                                 boolean flat,
                                 boolean mouseIn)
   {
      ColorRgb lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
   
      // coordinates are dependent from button state
      int x1 = focused ? orig.x + 1 : orig.x;
      int x2 = focused ? orig.x + dim.width - 2 : orig.x + dim.width - 1;
      int y1 = focused ? orig.y + 1 : orig.y;
      int y2 = focused ? orig.y + dim.height - 2 : orig.y + dim.height - 1;
      
      if (pressed)
      {
         if (flat)
         {
            gd.setColor(darkShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);
            
            gd.setColor(lightShadow);
            gd.drawLine(x1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1);
         }
         else
         {
            gd.setColor(darkShadow);
            gd.drawRect(x1, y1, x2 - x1, y2 - y1);
         }
      }
      else
      {
         if (flat)
         {
            if (mouseIn)
            {
               gd.setColor(lightShadow);
               gd.drawLine(x1, y1, x1, y2 - 1);
               gd.drawLine(x1, y1, x2 - 1, y1);
               
               gd.setColor(darkShadow);
               gd.drawLine(x1, y2, x2, y2);
               gd.drawLine(x2, y2, x2, y1);
            }
         }
         else
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);
            
            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 1, x2 - 1, y1 + 1);
            
            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y2, x2, y2);
            gd.drawLine(x2, y2, x2, y1);
         }
      }
   }
   
   /**
    * Draw button internals with different states.
    * @param   button
    * @param   gd
    * @param   gc
    * @param   gf
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   focused
    *          Is the button focused.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    *          Is mouse over the button
    * @param   mouseIn
    *          {@code true} if mouse hovering the button.
    */
   private void drawButtonInternal(ButtonGuiImpl button,
                                   GuiDriver gd,
                                   GuiColorResolver gc,
                                   GuiFontResolver gf,
                                   NativePoint orig,
                                   NativeDimension dim,
                                   boolean focused,
                                   boolean pressed,
                                   boolean flat,
                                   boolean mouseIn)
   {
      MnemonicInfo labelInfo = button.getMnemonic();
      String label = (labelInfo == null) ? button.textLabel() : labelInfo.getPrepLabel();
      ColorRgb lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      
      // coordinates are dependent from button state
      int x = focused ? orig.x + 1 : orig.x;
      int y = focused ? orig.y + 1 : orig.y;
      int w = focused ? dim.width - 3 : dim.width - 1;
      int h = focused ? dim.height - 4 : dim.height - 2;
      
      // draw button body
      gd.setColor(gc.bgColor);
      if (pressed)
      {
         // flat and non-flat modes behave the same way here
         gd.fillRect(x + 1, y + 1, w - 1, h - 1);
      }
      else
      {
         if (flat)
         {
            if (mouseIn)
            {
               gd.fillRect(x + 1, y + 1, w - 1, h);
            }
            else
            {
               gd.fillRect(x, y, w + 1, h + 2);
            }
         }
         else
         {
            gd.fillRect(x + 1, y + 1, w - 2, h - 1);
         }
      }
      
      // draw button text
      FontDetails fnt = gf.font().font;
      gd.setGuiFont(fnt);
      int fontNum = gf.resolveFontNum();
      int fontHeight = FontManager.getFontHeight(WindowManager.resolveWindow(button), fontNum, gd);
      int xt = orig.x + (dim.width - button.getTextWidthNative(label, fontNum)) / 2;
      // do not simplify the line below, this makes the font height to be  up nearest even number 
      int yt = orig.y + (dim.height + fontHeight + 1)/2 - 3;
      gd.setFontStyle(fnt.style);
      // enabled state controls text drawing
      if (button.isEnabled())
      {
         gd.setColor(gc.fgColor);
         if (pressed)
         {
            gd.drawString(label, xt + 1, yt + 1);
         }
         else
         {
            if (labelInfo != null && labelInfo.getMnemIdx() != null)
            {
               NativePoint tp = new NativePoint(xt, yt);
               gd.drawStringWithMnemonic(label, tp, labelInfo, gf, button, false);
            }
            else
            {
               gd.drawString(label, xt, yt);
            }
         }
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(label, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(label, xt, yt);
      }
      
      // draw dotted rectangle around focused button text
      if (button.isEnabled() && focused && !flat && !button.config().noFocus)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         int dx = 2;
         int dy = 2;
         if (pressed)
         {
            gd.drawRect(x + dx + 2, y + dy + 2, w - 2*dx - 2, h - 2*dy - 1);
         }
         else
         {
            gd.drawRect(x + dx + 1, y + dy + 1, w - 2*dx - 2, h - 2*dy - 1);
         }
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Draw button internals with different states using preloaded images.
    * @param   button
    * @param   gd
    * @param   gc
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    * 
    * @param   mouseIn
    *          {@code true} if mouse hovering the button.
    */
   private void drawButtonImages(ButtonGuiImpl button,
                                 GuiDriver gd,
                                 GuiColorResolver gc,
                                 NativePoint orig,
                                 NativeDimension dim,
                                 boolean pressed,
                                 boolean flat,
                                 boolean mouseIn)
   {
      ColorRgb lightShadow     = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow      = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow  = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      ImageGuiImpl imgDown     = button.getImgDown();
      ImageGuiImpl imgUp       = button.getImgUp();
      ImageGuiImpl imgDisabled = button.getImgDisabled();
   
      // NO-FOCUS attribute flag value
      boolean noFocusFlag = button.config().noFocus;
      boolean enabled = button.isEnabled();
      // coordinates are dependent from button state
      int x1 = noFocusFlag ? orig.x : orig.x + 1;
      int x2 = orig.x + dim.width - 1;
      int y1 = noFocusFlag ? orig.y : orig.y + 1;
      int y2 = orig.y + dim.height - 1;
   
      // draw the light frame around image
      gd.setColor(gc.bgColor);
      if (noFocusFlag && imgDown == null)
      {
         gd.fillRect(x1, y1, x2 - x1 + 2, y2 - y1 + 2);
      }
      else if (!noFocusFlag && imgDown != null)
      {
         gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      }
      else if (!noFocusFlag)
      {
         gd.fillRect(x1, y1, x2 - x1 - 1, y2 - y1 - 1);
      }
      
      // draw button body images
      if (pressed)
      {
         // determine what image to draw
         ImageGuiImpl imgPressed = null;
         if (imgDown != null)
         {
            imgPressed = imgDown;
         }
         else if (imgUp != null)
         {
            imgPressed = imgUp;
         }
         // draw pressed state
         if (imgPressed != null)
         {
            // for pressed state need 1 pixel right-down shift 
            NativePoint location = imgPressed.physicalLocation();
            if (imgDown == null)
            {
               imgPressed.setPhysicalLocation(location.x + 1, location.y + 1);
            }
            imgPressed.draw();
            // and restore origin after draw
            imgPressed.setPhysicalLocation(location.x, location.y);
         }
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (enabled && imgUp != null)
         {
            imgReleased = imgUp;
         }
         else if (!enabled && imgDisabled != null)
         {
            imgReleased = imgDisabled;
         }
         // draw released state
         if (imgReleased != null)
         {
            imgReleased.draw();
         }
         else
         {
            // draw the dark filled rectangle instead of image
            gd.setColor(darkShadow);
            gd.fillRect(x1 + 2, y1 + 2, x2 - x1 - 3, y2 - y1 - 3);
            // and lines marks the image as disabled
            gd.setColor(lightShadow);
            gd.drawLine(x1 + 3, y2 - 1, x2 - 2, y2 - 1);
            gd.drawLine(x2 - 1, y1 + 3, x2 - 1, y2 - 1);
         }
      }
      
      // draw 3D shadow for image button
      if (pressed)
      {
         if (flat)
         {
            if (imgDown == null)
            {
               gd.setColor(darkShadow);
               gd.drawLine(x1, y1, x1, y2);
               gd.drawLine(x1, y1, x2, y1);
               
               gd.setColor(lightShadow);
               gd.drawLine(x1, y2 + 1, x2, y2 + 1);
               gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);
            }
            else
            {
               gd.setColor(darkShadow);
               gd.drawLine(x1, y1, x1, y2 - 1);
               gd.drawLine(x1, y1, x2 - 1, y1);
               
               gd.setColor(lightShadow);
               gd.drawLine(x1, y2, x2 - 1, y2);
               gd.drawLine(x2, y2, x2, y1);
            }
         }
         else if (imgDown == null)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y2 - 1, x2 - 2, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 1, x2 - 1, y1);
            
            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2 - 2, x1 + 1, y1 + 1);
            gd.drawLine(x1 + 1, y1 + 1, x2 - 2, y1 + 1);
            
            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y1, x1, y2 - 2);
            gd.drawLine(x1, y1, x2 - 2, y1);
         }
      }
      else
      {
         if (flat && mouseIn && enabled)
         {
            if (imgDown == null)
            {
               gd.setColor(lightShadow);
               gd.drawLine(x1, y1, x1, y2);
               gd.drawLine(x1, y1, x2, y1);
               
               gd.setColor(darkShadow);
               gd.drawLine(x1, y2 + 1, x2 + 1, y2 + 1);
               gd.drawLine(x2 + 1, y2 + 1, x2 + 1, y1);
            }
            else
            {
               gd.setColor(lightShadow);
               gd.drawLine(x1, y1, x1, y2 - 1);
               gd.drawLine(x1, y1, x2 - 1, y1);
               
               gd.setColor(darkShadow);
               gd.drawLine(x1, y2, x2 - 1, y2);
               gd.drawLine(x2, y2, x2, y1);
            }
         }
         else if (!flat && imgDown == null)
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 1);
            gd.drawLine(x1, y1, x2 - 1, y1);
            
            gd.setColor(darkShadow);
            gd.drawLine(x1 + 1, y2 - 1, x2 - 2, y2 - 1);
            gd.drawLine(x2 - 1, y1 + 1, x2 - 1, y2 - 1);
            
            gd.setColor(veryDarkShadow);
            gd.drawLine(x1, y2, x2 - 1, y2);
            gd.drawLine(x2, y1, x2    , y2);
         }
      }
   }
   
   /**
    * Draw button internals with different states using preloaded prepackaged images.
    * @param   button
    * @param   gd
    * @param   gc
    * @param   orig
    *          Button origin.
    * @param   dim
    *          Button size.
    * @param   pressed
    *          Is the button pressed.
    * @param   flat
    * @param   mouseIn
    *          {@code true} if mouse hovering the button.
    */
   private void drawPrepackagedImages(ButtonGuiImpl button,
                                      GuiDriver gd,
                                      GuiColorResolver gc,
                                      NativePoint orig,
                                      NativeDimension dim,
                                      boolean pressed,
                                      boolean flat,
                                      boolean mouseIn)
   {
      ColorRgb lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      ImageGuiImpl imgDown     = button.getImgDown();
      ImageGuiImpl imgUp       = button.getImgUp();
      ImageGuiImpl imgDisabled = button.getImgDisabled();

      // NO-FOCUS attribute flag value
      boolean noFocusFlag = button.config().noFocus;
      // coordinates are dependent from button state
      int x1 = noFocusFlag ? orig.x : orig.x + 1;
      int x2 = noFocusFlag ? orig.x + dim.width : orig.x + dim.width - 1;
      int y1 = noFocusFlag ? orig.y : orig.y + 1;
      int y2 = noFocusFlag ? orig.y + dim.height : orig.y + dim.height - 1;
      
      // draw the light frame around image
      gd.setColor(gc.bgColor);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      
      // draw button body images
      if (pressed)
      {
         // determine what image to draw
         ImageGuiImpl imgPressed = null;
         if (imgDown != null)
         {
            imgPressed = imgDown;
         }
         else if (imgUp != null)
         {
            imgPressed = imgUp;
         }
         // draw pressed state
         if (imgPressed != null)
         {
            // for pressed state need 1 pixel right-down shift 
            NativePoint location = imgPressed.physicalLocation();
            imgPressed.setPhysicalLocation(location.x + 1, location.y + 1);
            imgPressed.draw();
            // draw small button frame
            gd.setColor(darkShadow);
            gd.drawRect(location.x, location.y,
                        PREBUILT_ARROW_IMAGE_SIZE,
                        PREBUILT_ARROW_IMAGE_SIZE);
            // and restore origin after draw
            imgPressed.setPhysicalLocation(location.x, location.y);
         }
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (!button.isEnabled() && imgDisabled != null)
         {
            imgReleased = imgDisabled;
         }
         else if (imgUp != null)
         {
            imgReleased = imgUp;
         }
         // draw released state
         if (imgReleased != null)
         {
            imgReleased.draw();
            // draw small button frame
            NativePoint location = imgReleased.physicalLocation();
            gd.setColor(lightShadow);
            gd.drawLine(location.x + 1, location.y + 1,
                  location.x + 1, location.y + PREBUILT_ARROW_IMAGE_SIZE - 2);
            gd.drawLine(location.x + 1, location.y + 1,
                  location.x + PREBUILT_ARROW_IMAGE_SIZE - 2, location.y + 1);
            gd.setColor(darkShadow);
            gd.drawLine(location.x + 1, location.y + PREBUILT_ARROW_IMAGE_SIZE - 1,
                  location.x + PREBUILT_ARROW_IMAGE_SIZE - 1, location.y + PREBUILT_ARROW_IMAGE_SIZE - 1);
            gd.drawLine(location.x + PREBUILT_ARROW_IMAGE_SIZE - 1, location.y + 1,
                  location.x + PREBUILT_ARROW_IMAGE_SIZE - 1, location.y + PREBUILT_ARROW_IMAGE_SIZE - 2);
            gd.setColor(veryDarkShadow);
            gd.drawLine(location.x, location.y + PREBUILT_ARROW_IMAGE_SIZE,
                  location.x + PREBUILT_ARROW_IMAGE_SIZE, location.y + PREBUILT_ARROW_IMAGE_SIZE);
            gd.drawLine(location.x + PREBUILT_ARROW_IMAGE_SIZE, location.y,
                  location.x + PREBUILT_ARROW_IMAGE_SIZE, location.y + PREBUILT_ARROW_IMAGE_SIZE - 1);
         }
      }
      
      // draw 3D shadow for image button
      if (pressed)
      {
         if (flat)
         {
            gd.setColor(darkShadow);
            gd.drawLine(x1, y1, x1, y2 - 2);
            gd.drawLine(x1 + 1, y1, x2 - 2, y1);
            
            gd.setColor(lightShadow);
            gd.drawLine(x1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 2, x2 - 1, y1);
         }
      }
      else
      {
         if (flat && mouseIn && button.isEnabled())
         {
            gd.setColor(lightShadow);
            gd.drawLine(x1, y1, x1, y2 - 2);
            gd.drawLine(x1, y1, x2 - 2, y1);
            
            gd.setColor(darkShadow);
            gd.drawLine(x1, y2 - 1, x2 - 1, y2 - 1);
            gd.drawLine(x2 - 1, y2 - 2, x2 - 1, y1);
         }
      }
   }
   
   /**
    * Obtain the padding for alert boxes.
    *
    * @return  the padding for alert boxes.
    */
   @Override
   public int getAlertBoxPadding()
   {
      return ALERT_BOX_PADDING;
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   paragraphWidth
    *          The width of the paragraph.
    * @param   icon
    *          The {@code ImageWrapper} containing the image to be shown in the alert dialog. May
    *          be {@code null}.
    * @param   text
    *          The text to be displayed in this alert box.
    * @param   font
    *          The {@link FontDetails} needed for text layout.
    */
   @Override
   public void drawAlertBox(GuiDriver gd,
                            int x,
                            int y,
                            int width,
                            int height,
                            int paragraphWidth,
                            ImageWrapper icon,
                            String text,
                            FontDetails font)
   {
      ColorRgb bgColor = ColorManager.getSysColor(SysColor.COLOR_WINDOW);
      gd.setColor(bgColor);
      gd.fillRect(x, y, width, height);
      
      NativeDimension iconDim = null;
      if (icon != null)
      {
         iconDim = new NativeDimension(icon.getWidth(), icon.getHeight());
         gd.drawImage(icon, x + ALERT_BOX_PADDING, y + ALERT_BOX_PADDING, iconDim.width, iconDim.height, 
                      0, 0, false, false, false, false);
      }
      
      gd.setColor(ColorManager.getSysColor(SysColor.COLOR_WINDOWTEXT));
      
      int px = x + ALERT_BOX_PADDING / 2;
      int py = y + ALERT_BOX_PADDING;
      if (icon != null)
      {
         px += ALERT_BOX_PADDING + iconDim.width;
         
         // TODO: the text seems to be shifted up by half the font height
         int icoOffset = -5;
      }
      
      gd.drawParagraph(text, font, px, py, paragraphWidth);
   }
   
   /**
    * Renders a GUI radio button widget.
    * @param  rb
    *         The {@link RadioButtonGuiImpl} to be rendered.
    * @param  gd
    *         The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *         The {@link GuiColorResolver} to be used for color selection.
    * @param  gf
    *         The {@code GuiFontResolver} to be used.
    * @param  x
    *         The horizontal location (in pixel units) of the widget to be painted/component.
    * @param  y
    *         The vertical location (in pixel units) of the widget to be painted/component.
    * @param  width
    *         The width of the painted widget/component. 
    * @param  height
    *         The height of the painted widget/component.
    * @param  focused
    *         {@code true} if the radio button have the focus
    * @param  hover
    *         {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   @Override
   public void drawRadioButton(RadioButtonGuiImpl rb,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               boolean focused,
                               boolean hover)
   {
      int imgSize = RADIO_BUTTON_IMAGE_SIZE;
      ColorRgb disabledTextColor   = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb disabledTextColorBk = gd.getSysColor(SysColor.COLOR_WINDOW);
   
      // first draw the check mark image
      boolean isEnabled = rb.isEnabled() && rb.parent().isEnabled();
      ImageWrapper<?> imgToDraw = getRadioButtonImage(rb.isSelected(), isEnabled, hover, false); 
   
      // for heights other than 1.0 the Y location should be centered
      gd.drawImage(imgToDraw, x, y + (height - imgSize) / 2,
            imgSize, imgSize, 0, 0, false, false, false, false);
   
      // draw button text
      String label = rb.getText();
      FontDetails<?>  fnt = gf.font().font;
      int fntNum = gf.resolveFontNum();
   
      gd.setGuiFont(fnt);
      int wt = rb.getTextWidthNative(label, fntNum);
      int ht = rb.getTextHeightNative(label, fntNum);
      int xt = imgSize + x + 3;
      int yt = y + (height + fnt.pointSize) / 2;
      gd.setFontStyle(fnt.style);
      // enabled state controls text drawing
      if (!isEnabled)
      {
         gd.setColor(disabledTextColorBk);
         gd.drawString(label, xt + 1, yt + 1);
      }
      gd.setColor(isEnabled ? gc.fgColor : disabledTextColor);
   
      MnemonicInfo info = rb.getMnemonic();
   
      if (info != null && info.getMnemIdx() != null)
      {
         NativePoint tp = new NativePoint(xt, yt);
         gd.drawStringWithMnemonic(label, tp, info, gf, rb, false);
      }
      else
      {
         gd.drawString(label, xt, yt);
      }
   
      if (focused)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         // set dotted rectangle width
         int drWidth = xt + wt <= width ? wt + 1 : width - xt;
         // draw dotted rectangle depending on centered flag and item height
         gd.drawRect(xt - 1,
                     y + (height > ht + 4 ? (height - ht) / 2 - 1 : 0),
                     drWidth, height <= ht + 4 ? height - 2 : ht + 2);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Obtain the extra distance in a radio button. 
    *
    * @return  the extra distance in a radio button.
    */
   @Override
   public int getRadioButtonExtraDist()
   {
      return RADIO_BUTTON_EXTRA_DIST;
   }
   
   /**
    * Obtain the distance before text, including the radio button (text offset).
    *
    * @return  the distance before text, including the radio button.
    */
   @Override
   public int getRadioButtonImageSpace()
   {
      return RADIO_BUTTON_IMAGE_SPACE;
   }
   
   /**
    * Obtain the image from a radio button with specified attributes.
    *
    * @param   on
    *          The button state. {@code true} if button is on.
    * @param   enabled
    *          {@code true} is button enabled. 
    * @param   hover
    *          {@code true} if mouse is hovering over the button. Some styles are mouse hovering
    *          sensitive. 
    * @param   pressed
    *          {@code true} if the radio button is pressed (the mouse button is down on the 
    *          button).
    *
    * @return  The {@code ImageWrapper} containing the image with requested properties. 
    */
   protected ImageWrapper getRadioButtonImage(boolean on,
                                              boolean enabled,
                                              boolean hover,
                                              boolean pressed)
   {
      String sb = "rs-chk-" + (enabled ? "en-" : "dis-") + (on ? "on" : "off");
      return getImage(sb, RADIO_BUTTON_IMAGE_SIZE, RADIO_BUTTON_IMAGE_SIZE, false);
   }
   
   /**
    * Obtains the space needed by a checkbox widget.
    *
    * @return  the space needed by a checkbox widget.
    */
   @Override
   public int getCheckBoxSpace()
   {
      return CHECKBOX_SPACE;
   }
   
   /**
    * Obtain the size of image on a checkbox widget.
    *
    * @return size of image on a checkbox widget.
    */
   @Override
   public int getCheckBoxSize()
   {
      return CHECKBOX_SIZE;
   }

   /**
    * Renders a GUI toggle box widget.
    * @param   cb
    *          The {@link ToggleBoxGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the checkbox is pressed (the mouse button is down on the button).
    * @param   hover
    *          {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   public void drawToggleBox(ToggleBoxGuiImpl cb,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             GuiFontResolver gf,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean pressed,
                             boolean hover)
   {
      NativePoint po = new NativePoint(x, y);
      String text = cb.getMnemonicLabel(false).orElse(cb.config().label);
      GuiFontResolver.FontCache fntCache = gf.font();
      
      // initialize the colors
      ColorRgb chkMarkColor       = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
      ColorRgb lightShadow        = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb labelInactiveColor = gd.getSysColor(SysColor.COLOR_BTNHIGHLIGHT);
      
      // clear the area
      gd.setColor(gc.bgColor);
      gd.fillRect(po.x, po.y, width, height);
      
      // toggle-box state flags
      boolean highlighted = cb.highlighted();
      
      // focused toggle-box needs special highlight
      if (highlighted && text != null && text.length() > 0)
      {
         // draw highlight frame indicating focused button.
         int font = gf.resolveFontNum();
         int textWidth = cb.getTextWidthNative(text, font) + 1;
         int textHeight = cb.getTextHeightNative(text, font) + 3;
         
         int xt = po.x + getCheckBoxSize() + getCheckBoxSpace() - 1;
         int yt = po.y + (height - textHeight) / 2 - 1;
         
         // the color of highlight rectangle: FG XOR BG
         int red = gc.fgColor.getRed() ^ gc.bgColor.getRed();
         int green = gc.fgColor.getGreen() ^ gc.bgColor.getGreen();
         int blue = gc.fgColor.getBlue() ^ gc.bgColor.getBlue();
         
         ColorRgb highlightColor = new ColorRgb(red, green, blue);
         
         gd.setColor(highlightColor);
         gd.drawRect(xt, yt, textWidth, textHeight);
         
         // unconditionally draw black dotted rectangle.
         gd.setColor(chkMarkColor);
         gd.setLineStroke(LineStroke.DOTTED_SMALL_THIN);
         gd.drawRect(xt, yt, textWidth, textHeight);
         gd.setLineStroke(LineStroke.DEFAULT);
      }
      
      drawCheckBox(cb, gd, po.x, po.y, width, height, pressed, false);
      
      if (text != null && text.length() > 0)
      {
         int xt = po.x + getCheckBoxSize() + getCheckBoxSpace(); // shift between checkbox and label
         int yt = po.y + height / 2;
         
         gd.setGuiFont(fntCache.font);
         gd.setFontStyle(fntCache.font.style);
         
         boolean enabled = cb.isEnabled();
         if (!enabled)
         {
            xt++;
            yt++;
         }
         
         // draw the light text first!
         gd.setColor(enabled ? gc.fgColor : labelInactiveColor);
         
         MnemonicInfo info = cb.getMnemonic();
         if (info != null && info.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(xt, yt);
            gd.drawStringWithMnemonic(text, tp, info, gf, cb, true);
         }
         else
         {
            gd.drawStringCentered(text, xt, yt);
         }
      
         if (!enabled)
         {
            gd.setColor(lightShadow);
            gd.drawStringCentered(text, xt - 1, yt - 1);
         }
      }
   }
   
   private void drawCheckBox(ToggleBoxGuiImpl checkbox,
                             GuiDriver gd,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean pressed,
                             boolean hover)
   {
      int w = CHECKBOX_SIZE;
      ColorRgb chkBoxLTShadow     = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      ColorRgb chkBoxLTColor      = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb chkBoxRBShadow     = gd.getSysColor(SysColor.COLOR_BTNHIGHLIGHT);
      ColorRgb chkBoxRBColor      = gd.getSysColor(SysColor.COLOR_3DLIGHT);
      ColorRgb chkBoxBGInactive   = gd.getSysColor(SysColor.COLOR_MENU);
      ColorRgb chkMarkColor       = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
      ColorRgb chkMarkInactiveColor = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      
      NativePoint p = new NativePoint(x, y + (height - w) / 2 - 1);
      
      // when mouse button is pressed and until it will be released background is drawn as
      // inactive. The same with 'SPACE'.
      ColorRgb color = (checkbox.isEnabled() && !pressed) ? chkBoxRBShadow : chkBoxBGInactive;
      
      // draw background
      gd.setColor(color);
      gd.fillRect(p.x, p.y, w + 1, w + 1);
      
      // left  top angle
      gd.setColor(chkBoxLTColor);
      gd.drawLine(p.x, p.y, p.x + w, p.y);
      gd.drawLine(p.x, p.y + 1, p.x, p.y + w);
      gd.setColor(chkBoxLTShadow);
      gd.drawLine(p.x + 1, p.y + 1, p.x + w - 1, p.y + 1);
      gd.drawLine(p.x + 1, p.y + 1, p.x + 1, p.y + w - 1);
      
      // right bottom angle
      gd.setColor(chkBoxRBColor);
      gd.drawLine(p.x + 2, p.y + w, p.x + w, p.y + w);
      gd.drawLine(p.x + w, p.y + 2, p.x + w, p.y + w);
      gd.setColor(chkBoxRBShadow);
      gd.drawLine(p.x, p.y + w + 1, p.x + w + 1, p.y + w + 1);
      gd.drawLine(p.x + w + 1, p.y, p.x + w + 1, p.y + w);
      
      // draw check
      if (checkbox.config().checked)
      {
         drawCheckmark(gd, p, checkbox.isEnabled() ? chkMarkColor : chkMarkInactiveColor);
      }
   }
   
   /**
    * Draws checkmark at specified point using specified driver and using specified color.
    *
    * @param  p
    *         Target point.
    * @param  chkMarkColor
    *         ColorRgb - RGB color for checkmark.
    */
   private void drawCheckmark(GuiDriver gd, NativePoint p, ColorRgb chkMarkColor)
   {
      gd.setColor(chkMarkColor);
      // left half
      gd.drawLine(p.x + 3, p.y + 5, p.x + 5, p.y + 7);
      gd.drawLine(p.x + 3, p.y + 6, p.x + 5, p.y + 8);
      gd.drawLine(p.x + 3, p.y + 7, p.x + 5, p.y + 9);
      // right half
      gd.drawLine(p.x + 9, p.y + 3, p.x + 5, p.y + 7);
      gd.drawLine(p.x + 9, p.y + 4, p.x + 5, p.y + 8);
      gd.drawLine(p.x + 9, p.y + 5, p.x + 5, p.y + 9);
   }
   
   /**
    * Obtain the size of the scroll button on the direction perpendicular to scroll direction.
    *
    * @return  size of the scroll button on the direction perpendicular to scroll direction.
    */
   @Override
   public int getScrollBarButtonWidth()
   {
      return SCROLLBAR_WIDTH;
   }
   
   /**
    * Obtain the thumb/slider width in a scrollbar widget.
    *
    * @return  the thumb/slider width in a scrollbar widget.
    */
   @Override
   public int getScrollBarThumbWidth()
   {
      return SCROLLBAR_WIDTH;
   }
   
   /**
    * Obtain the width of the gauge of a scrollbar widget.
    *
    * @return  width of the gauge of a scrollbar widget.
    */
   @Override
   public int getScrollBarGaugeWidth()
   {
      return SCROLLBAR_WIDTH;
   }
   
   /**
    * Obtain the size of the scroll button on the direction of the scroll.
    *
    * @return  the size of the scroll button on the direction of the scroll.
    */
   @Override
   public int getScrollBarButtonHeight()
   {
      return SCROLLBAR_WIDTH;
   }
   
   /**
    * Obtain the minimum size of the thumb/slider of a scrollbar widget.
    *
    * @return  the minimum size of the thumb/slider of a scrollbar widget.
    */
   @Override
   public int getScrollBarThumbMinSize()
   {
      return SCROLLBAR_THUMB_MIN_SIZE;
   }
   
   /**
    * Renders the gauge (background) of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarGauge(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean horizontal)
   {
      if (!scrollbarImgBgLoaded)
      {
         // load the image to draw the background
         byte[] imgData = ImageHelper.getPrepackagedImageData("sb-bg-fill");
         if (imgData != null)
         {
            //imgBg = gd.createImage(imgData);
            scrollbarImgBg = AbstractGuiDriver.createImage(imgData);
         }
      }
      
      // this is additional safe check if something is going wrong
      if (scrollbarImgBg != null)
      {
         // get image dimensions
         int iw = scrollbarImgBg.getWidth();
         int ih = scrollbarImgBg.getHeight();
         // use tile fill, compute the required rows and columns
         // ceil() ensures filling partial space less than image size
         int colToFill = (int) Math.ceil((double) width / (double) iw);
         int rowToFill = (int) Math.ceil((double) height / (double) ih);
         // fill the area
         for (int i = 0; i < rowToFill; i++)
         {
            for (int j = 0; j < colToFill; j++)
            {
               // next image coordinate
               int ix = x + j * iw;
               int iy = y + i * ih;
               // draw solid image with convert-3d-option flag is on
               gd.drawImage(scrollbarImgBg, ix, iy, iw, ih, 0, 0, false, false, false, true);
            }
         }
      }
      else
      {
         gd.setColor(gc.bgColor);
         gd.fillRect(x, y, width, height);
      }
   }
   
   /**
    * Renders the thumb/slider of the scrollbar widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          If the mouse is dragging the thumb of the scrollbar.
    * @param hover
    *          Is mouse hovering this component ?
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarThumb(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean hover,
                                  boolean horizontal)
   {
      // fill the body
      gd.setColor(gc.thumbColor);
      gd.fillRect(x, y, width, height);
      // 3D border
      // top-left part
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x + 1, y + 1, x + 1, y + height - 3);
      gd.drawLine(x + 2, y + 1, x + width - 3, y + 1);
      // bottom-right part
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DDKSHADOW));
      gd.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
      gd.drawLine(x + width - 1, y, x + width - 1, y + height - 2);
      // shadow corner
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawLine(x + 1, y + height - 2, x + width - 2, y + height - 2);
      gd.drawLine(x + width - 2, y + 1, x + width - 2, y + height - 3);
   }
   
   /**
    * Renders a button of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pos
    *          The position of the button (UP/DOWN, LEFT/RIGHT).
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    * @param   enabled
    *          If the button is enabled.
    * @param   pressed
    *          If the button is down (pressed by mouse).
    * @param   highlighted
    *          If the mouse is hovering over the button.
    */
   @Override
   public void drawScrollBarButton(GuiDriver gd,
                                   GuiColorResolver gc,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   ScrollBar.Position pos,
                                   boolean horizontal,
                                   boolean enabled,
                                   boolean pressed,
                                   boolean highlighted)
   {
      gd.setColor(gc.bgColor);
      gd.fillRect(x, y, width, height);
      
      // 3D border
      if (!pressed)
      {
         // top-left part
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(x + 1, y + 1, x + 1,         y + height - 3);
         gd.drawLine(x + 2, y + 1, x + width - 3, y + 1);
         // bottom-right part
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DDKSHADOW));
         gd.drawLine(x,             y + height - 1, x + width - 1, y + height - 1);
         gd.drawLine(x + width - 1, y,              x + width - 1, y + height - 2);
         // shadow corner
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.drawLine(x + 1,         y + height - 2, x + width - 2, y + height - 2);
         gd.drawLine(x + width - 2, y + 1,          x + width - 2, y + height - 3);
      }
      else
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.drawRect(x, y, width - 1, height - 1);
      }
   
      // draw arrow images loaded before
      StringBuilder sb = new StringBuilder("sb-arr-");
      sb.append(enabled ? "en" : "dis");
      if (horizontal)
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "-left" : "-right");
      }
      else
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "-up" : "-down");
      }
      // calculate image width and height
      int w = width;
      int h = height;
      // check if we need small versions of the images
      int sizeMin = Math.min(w, h);
      if (sizeMin < SCROLLBAR_WIDTH)
      {
         // adjust size
         w = SCROLLBAR_WIDTH / 2;
         h = SCROLLBAR_WIDTH / 2;
         // and name
         sb.append("-small");
      }
      // load enabled and disabled image versions
      ImageWrapper<?> img = VectorGraphicsHelper.getScalableImage(sb.toString(), w, h);
      
      // is anything to draw?
      if (img != null)
      {
         // calc size and centered location
         w = img.getWidth();
         h = img.getHeight();
         x += (width - w) / 2;
         y += (height - h) / 2;
      
         // for enabled buttons need to consider Up/Down state difference
         gd.drawImage(img,
               !pressed || !enabled ? x : x + 1,
               !pressed || !enabled ? y : y + 1,
               w, h, 0, 0, false, false, false, false);
      }
   }
   
   /**
    * Obtain the offset in a 3D bordered widget.
    *
    * @return  the offset in a 3D bordered widget.
    */
   public int get3DBorderOffset()
   {
      return THREE_D_BORDER_OFFSET;
   }
   
   /**
    * Obtain the character used to replace text in password fields. The default is "*" (asterisk)
    * but some themes may use other characters like "bullet" char.
    *
    * @return  the password character for this theme is '*' (asterisk).
    */
   @Override
   public char getPasswordCharacter()
   {
      return '*';
   }
   
   /**
    * Renders the border (and possibly the background) of the fill-in widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border3D
    *          If fill-in widget has a 3d border. Some flat {@code Theme}s may ignore this.
    * @param   hasFocus
    *          The fill-in widget is focused.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawFillInBorder(GuiDriver gd, GuiColorResolver gc,
                                int x, int y, int width, int height, 
                                boolean border3D, boolean hasFocus, boolean hover)
   {
      if (border3D)
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.drawLine(x, y, x,             y + height - 2);
         gd.drawLine(x, y, x + width - 2, y);
      
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DDKSHADOW));
         gd.drawLine(x + 1, y + 1, x + 1,         y + height - 3);
         gd.drawLine(x + 1, y + 1, x + width - 3, y + 1);
      
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DLIGHT));
         gd.drawLine(x + 1,     y + height - 2, x + width - 2, y + height - 2);
         gd.drawLine(width - 2, y + height - 2, x + width - 2, y + 1);
      
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(x,             y + height - 1, x + width - 1, y + height - 1);
         gd.drawLine(x + width - 1, y + height - 1, x + width - 1, y);
      }
      else
      {
         gd.setColor(gc.boxColor);
         gd.drawRect(x, y, width - 1, height - 1);
      }
   }
   
   /**
    * Renders caret of the fill-in widget.
    *
    * @param   fillIn
    *          The {@code FillInGuiImpl} widget whose caret is to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x1
    *          The left extremity of the caret. 
    * @param   y1
    *          The top extremity of the caret.
    * @param   x2
    *          The right extremity of the caret. 
    * @param   y2
    *          The bottom extremity of the caret.
    */
   @Override
   public void drawFillInCaret(FillInGuiImpl fillIn,
                               GuiDriver gd,
                               int x1,
                               int y1,
                               int x2,
                               int y2)
   {
      drawCaret(gd, x1, y1, x2, y2);
   }
   
   
   /**
    * Renders a caret for an enabled widget ({@code FillInGuiImpl} or {@code EditorGuiImpl}). 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x1
    *          The left extremity of the caret. 
    * @param   y1
    *          The top extremity of the caret.
    * @param   x2
    *          The right extremity of the caret. 
    * @param   y2
    */
   protected void drawCaret(GuiDriver gd, int x1, int y1, int x2, int y2)
   {
      gd.setXORComposite();
      setCompositeXorColor(gd);
      gd.drawLine(x1, y1, x2, y2);
      gd.resetComposite();
   }
   
   /**
    * Renders the status bar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used for color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text in the status bar.
    * @param   cache
    *          The font cache used for text rendering. 
    * @param   bounds
    *          The {@link NativePoint} where the text is drawn.
    * @param   txtDim
    *          The {@link NativeDimension} of the text displayed in status bar.
    * @param   resizable
    *          {@code true} window is resizable. In this case a "grip" is drawn in the lower right
    *          corner.
    */
   @Override
   public void drawStatusBar(GuiDriver gd,
                             GuiColorResolver gc,
                             int x,
                             int y,
                             int width,
                             int height,
                             String text,
                             GuiFontResolver.FontCache cache,
                             NativePoint bounds,
                             NativeDimension txtDim,
                             boolean resizable)
   {
      gd.setColor(gc.bgColor);
      // clear the area
      gd.fillRect(x, y, width, height);
   
      // 3D simulation for the status line border
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x, y, x + width - 1, y);
      // dark part
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawLine(bounds.x - 2, bounds.y, bounds.x - 2, bounds.y + txtDim.height - 1);
      gd.drawLine(bounds.x - 1, bounds.y, bounds.x + txtDim.width, bounds.y);
      // light part
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(bounds.x + txtDim.width + 1, bounds.y,
                  bounds.x + txtDim.width + 1, bounds.y + txtDim.height - 1);
      gd.drawLine(bounds.x - 1,            bounds.y + txtDim.height - 1,
                  bounds.x + txtDim.width, bounds.y + txtDim.height - 1);
      
      // draw status bar text
      gd.setColor(gc.fgColor);
      gd.setGuiFont(cache.font);
      gd.setFontStyle(cache.font.style);
      
      gd.draw(bounds, new NativeRectangle(0, 0, txtDim.height - 1, txtDim.width - 1), 
               () -> gd.drawStringCentered(text, -1, txtDim.height / 2 - 1));
      
      // if window is resizable draw a small triangle on bottom right corner
      if (resizable)
      {
         // corner constants
         int W1 = txtDim.height - 1;
         int xRight = bounds.x + txtDim.width + 2;
         int yBottom = bounds.y + txtDim.height;
         
         // draw status bar right-bottom corner
         int[] xPoints = {xRight - W1, xRight,  xRight};
         int[] yPoints = {yBottom,     yBottom, yBottom - W1};
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.fillPolygon(xPoints, yPoints);
      
         int[] xPoints2 = {xRight - W1 + 1, xRight, xRight};
         int[] yPoints2 = {yBottom, yBottom, yBottom - W1 + 1};
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.fillPolygon(xPoints2, yPoints2);
      
         int[] xPoints3 = {xRight - W1 + 3, xRight, xRight};
         int[] yPoints3 = {yBottom, yBottom, yBottom - W1 + 3};
         gd.setColor(gc.bgColor);
         gd.fillPolygon(xPoints3, yPoints3);
      
         int[] xPoints4 = {xRight - W1 + 4, xRight, xRight};
         int[] yPoints4 = {yBottom, yBottom, yBottom - W1 + 4};
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.fillPolygon(xPoints4, yPoints4);
      
         int[] xPoints5 = {xRight - W1 + 5, xRight, xRight};
         int[] yPoints5 = {yBottom, yBottom, yBottom - W1 + 5};
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.fillPolygon(xPoints5, yPoints5);
      
         int[] xPoints6 = {xRight - W1 + 7, xRight, xRight};
         int[] yPoints6 = {yBottom, yBottom, yBottom - W1 + 7};
         gd.setColor(gc.bgColor);
         gd.fillPolygon(xPoints6, yPoints6);
      
         int[] xPoints7 = {xRight - W1 + 8, xRight, xRight};
         int[] yPoints7 = {yBottom, yBottom, yBottom - W1 + 8};
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.fillPolygon(xPoints7, yPoints7);
      
         int[] xPoints8 = {xRight - W1 + 9, xRight, xRight};
         int[] yPoints8 = {yBottom, yBottom, yBottom - W1 + 9};
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.fillPolygon(xPoints8, yPoints8);
      
         int[] xPoints9 = {xRight - W1 + 11, xRight, xRight};
         int[] yPoints9 = {yBottom, yBottom, yBottom - W1 + 11};
         gd.setColor(gc.bgColor);
         gd.fillPolygon(xPoints9, yPoints9);
      
         int[] xPoints10 = {xRight - W1 + 12, xRight, xRight};
         int[] yPoints10 = {yBottom, yBottom, yBottom - W1 + 12};
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.fillPolygon(xPoints10, yPoints10);
      
         int[] xPoints11 = {xRight - W1 + 13, xRight, xRight};
         int[] yPoints11 = {yBottom, yBottom, yBottom - W1 + 13};
         gd.setColor(gc.bgColor);
         gd.fillPolygon(xPoints11, yPoints11);
      }
   
      // this line is commented, as it leaves the JS side in an inconsistent state 
      // (clipping/translated origin) - need to investigate why, when the batch ends, it doesn't
      // restore the clipping/translated origin
      // gd.flush();
   }
   
   /**
    * Obtain the field extra space for combo box layout.
    *
    * @param   horiz
    *          Direction. {@code true} for horizontal space.
    * @param   is3D
    *          Whether this is widget has a 3D border.
    *
    * @return  the field extra space for combo box layout.
    */
   @Override
   public int getComboBoxFieldExtraSpace(boolean horiz, boolean is3D)
   {
      return horiz ? COMBO_BOX_ENTRY_SHIFT :
             /*vertical*/ COMBO_BOX_ENTRY_DELTA + (is3D? 2 : 0);
   }
   
   /**
    * Obtain the shifting distance in a combo box widget.
    *
    * @return  the shifting distance in a combo box widget.
    */
   @Override
   public int getComboBoxEntryShift()
   {
      return COMBO_BOX_ENTRY_SHIFT;
   }
   
   /**
    * Renders the entry field of a combo box. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose entry field is painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   font
    *          A {@code FontDetails} used for text rendering.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxEntryField(ComboBoxGuiImpl combo,
                                      GuiDriver gd,
                                      GuiColorResolver gc,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      FontDetails<?> font,
                                      boolean pressed,
                                      boolean enabled,
                                      boolean hover)
   {
      // text in DROP-DOWN-LIST mode
      if (combo.config().mode != ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         return;
      }
   
      ColorRgb lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb selectionBack  = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      ColorRgb selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                             255 - selectionBack.getGreen(),
                                             255 - selectionBack.getBlue());
      // coordinates of the field area
      int w = width - COMBO_BOX_BUTTON_SIZE - 5;
      int h = height - 5;
      gd.setColor(gc.bgColor);
      gd.drawRect(x + 2, y + 2, w, h);
      
      boolean focused = combo.hasFocus() && combo.isEnabled();
      if (focused)
      {
         gd.setColor(selectionBack);
      }
      gd.fillRect(x + 3, y + 3, w - 1, h - 1);
      
      // get the text to draw with field
      String display = null;
      
      if (combo.isSelected())
      {
         display = combo.model().selected();
      }
      else
      {
         BaseDataType initial = combo.getValue();
         if (initial != null && !initial.isUnknown())
         {
            display = combo.getInitialScreenText(initial);
         }
      }
      
      if (display == null)
      {
         display = "";
      }
      
      // limit the data portion of the field to the maximum allowed size
      int limit = combo.getFormatLength();
      
      if (display.length() > limit && combo.isFormatHonored())
      {
         display = display.substring(0, limit);
      }
      
      // draw entry field text
      gd.setGuiFont(font);
      // left-center alignment
      int xt = x + ComboBoxGuiImpl.TEXT_SHIFT;
      int yt = y + 1 + (height + font.pointSize) / 2;
      gd.setFontStyle(font.style);
      
      final String finalDisplay = display;
      gd.draw(new NativePoint(0, 0), new NativeRectangle(0, 0, height, w),
              () -> {
           // enabled state controls text drawing
           if (combo.isEnabled())
           {
              if (focused)
              {
                 gd.setColor(lightShadow);
              }
              else
              {
                 gd.setColor(gc.fgColor);
              }
              gd.drawString(finalDisplay, xt, yt);
           }
           else
           {
              gd.setColor(lightShadow);
              gd.drawString(finalDisplay, xt + 1, yt + 1);
              gd.setColor(darkShadow);
              gd.drawString(finalDisplay, xt, yt);
           }
      });
      
      // draw dotted rectangle around focused item text
      if (focused)
      {
         // black background for dotted line
         gd.setColor(gc.fgColor);
         gd.drawRect(x + 3, y + 3, w - 2, h - 2);
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         gd.drawRect(x + 3, y + 3, w - 2, h - 2);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Renders the border (and possibly the background) of the combo box widget. Draw 3D border 
    * around different parts of the combo-box widget.
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxFieldBorder(ComboBoxGuiImpl combo,
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height,
                                       boolean pressed,
                                       boolean enabled,
                                       boolean hover)
   {
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);

      int x1 = x;
      int x2 = x + width - 2;
      int y1 = y;
      int y2 = y + height - 1;
   
      // drawing the frame around entry field
      gd.setColor(darkShadow);
      gd.drawLine(x1, y1, x1, y2);
      gd.drawLine(x1 + 1, y1, x2, y1);
   
      gd.setColor(veryDarkShadow);
      gd.drawLine(x1 + 1, y1 + 1, x1 + 1, y2 - 1);
      gd.drawLine(x1 + 2, y1 + 1, x2 - 1, y1 + 1);
   
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.drawLine(x1 + 1, y2 - 1, x2, y2 - 1);
      gd.drawLine(x2, y2 - 1, x2, y1 + 1);
   
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x1, y2, x2 + 1, y2);
      gd.drawLine(x2 + 1, y2, x2 + 1, y1);
   }
   
   /**
    * Obtain the size of the button that opens the combo box.
    *
    * @return  the size of teh combo box button.
    */
   @Override
   public int getComboBoxButtonSize()
   {
      return COMBO_BOX_BUTTON_SIZE;
   }
   
   /**
    * Renders the opening button of the combo box widget. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxButton(ComboBoxGuiImpl combo,
                                  GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean enabled,
                                  boolean hover)
   {
      ColorRgb lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      ColorRgb btnBack        = gd.getSysColor(SysColor.COLOR_3DFACE);

      // simple check to prevent NPE
      ImageWrapper imgButtEn = VectorGraphicsHelper.getScalableImage("sb-arr-en-down",
            COMBO_BOX_BUTTON_SIZE, COMBO_BOX_BUTTON_SIZE);
      ImageWrapper imgButtDis = VectorGraphicsHelper.getScalableImage("sb-arr-dis-down",
            COMBO_BOX_BUTTON_SIZE, COMBO_BOX_BUTTON_SIZE);
      if (imgButtEn == null || imgButtDis == null)
      {
         return;
      }
      
      int arrowImageSize = getPrebuiltArrowImageSize();
   
      // coordinates of the button area
      int x1 = x + width - arrowImageSize - 3;
      int x2 = x + width - 1;
      int y1 = y + 2;
      int y2 = y + height - 2; // border width
   
      // draw the light frame around image
      gd.setColor(btnBack);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
   
      // center the image
      int imageY = y1 + (y2 - y1 - COMBO_BOX_BUTTON_SIZE) / 2;
   
      // disabled state has separate image to draw  
      if (pressed)
      {
         // for pressed state need 1 pixel right-down shift 
         gd.drawImage(imgButtEn, x1 + 1, imageY + 1,
               COMBO_BOX_BUTTON_SIZE, COMBO_BOX_BUTTON_SIZE, 0, 0,
               false, false, false, false);
      
         // draw small button frame
         gd.setColor(darkShadow);
         gd.drawRect(x1, y1, x2 - x1 - 1, y2 - y1 - 1);
      }
      else
      {
         // determine what image to draw
         ImageWrapper<?> imgReleased = null;
         if (enabled)
         {
            imgReleased = imgButtEn;
         }
         else
         {
            imgReleased = imgButtDis;
         }
         gd.drawImage(imgReleased, x1, imageY, COMBO_BOX_BUTTON_SIZE, COMBO_BOX_BUTTON_SIZE, 0, 0,
                      false, false, false, false);
         // draw small button frame
         gd.setColor(lightShadow);
         gd.drawLine(x1 + 1, y1 + 2, x1 + 1, y2 - 3);
         gd.drawLine(x1 + 1, y1 + 1, x1 + arrowImageSize - 2, y1 + 1);
      
         gd.setColor(darkShadow);
         gd.drawLine(x1 + 1, y2 - 2, x1 + arrowImageSize - 1, y2 - 2);
         gd.drawLine(x1 + arrowImageSize - 1, y1 + 1, x1 + arrowImageSize - 1, y2 - 3);
      
         gd.setColor(veryDarkShadow);
         gd.drawLine(x1, y2 - 1, x1 + arrowImageSize, y2 - 1);
         gd.drawLine(x1 + arrowImageSize, y1, x1 + arrowImageSize, y2 - 2);
      }
   }
   
   /**
    * Renders the background of a selection list. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSelectionListBackground(GuiDriver gd, GuiColorResolver gc, int x, int y, int width, int height)
   {
      // fill the background
      gd.setColor(gc.bgColor);
      gd.fillRect(x, y, width, height);
   }
   
   /**
    * Renders a single item in a combo box selection list. 
    *
    * @param   box
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text to be displayed on this item
    * @param   fontSize
    *          The size of teh font to be used.
    * @param   highlighted
    *          The mouse is hovering over the fill-in widget.
    * @param   enabled
    *          Whether the widget has the focus.
    */
   @Override
   public void drawSelectionListItem(ComboBoxGuiImpl box,
                                     GuiDriver gd,
                                     GuiColorResolver gc,
                                     int x,
                                     int y,
                                     int width,
                                     int height,
                                     String text,
                                     int fontSize,
                                     boolean enabled,
                                     boolean highlighted,
                                     boolean selected)
   {
      ColorRgb selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      ColorRgb selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      ColorRgb disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      // top row number will be initialized later in setTopRow()
      
      // the color below is negative to selectionBack
      ColorRgb selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                             255 - selectionBack.getGreen(),
                                             255 - selectionBack.getBlue());
      
      // fill the background of the selected item
      if (selected && highlighted)
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, width, height);
      }
      
      // set the color of the text
      if (enabled && selected && highlighted)
      {
         gd.setColor(selectionFore);
      }
      else if (enabled)
      {
         gd.setColor(gc.fgColor);
      }
      else
      {
         gd.setColor(disabledFore);
      }
      
      // draw item text
      gd.drawString(text,
                    box != null && box.config().mode == ComboBoxConfig.Mode.SIMPLE
                              ? x + getComboBoxEntryShift() : x + ComboBoxGuiImpl.TEXT_SHIFT,
                    y + 1 + (height + fontSize) / 2);
      
      // draw the dotted line for selected item (not for SIMPLE mode of the combo-box)
      if (enabled && selected &&
            !(box == null || box.config().mode != ComboBoxConfig.Mode.DROP_DOWN_LIST))
      {
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         // take into account possible vertical scrollbar width 
         gd.drawRect(x, y, width - 1 - box.getDropDownScrollBarWidth(), height - 1);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Obtain the vertical distance between the combo box and the scrolling selection list.  
    *
    * @return the vertical offset for selection list of the combo box.
    */
   @Override
   public int getComboBoxDropDownOffset()
   {
      return 0;
   }
   
   /**
    * Obtain the minimum size of a titlebar element for this Theme, including any border. 
    *
    * @param   horiz
    *          If {@code true} returns the minimum width, if {@code false} returns minimum height.
    *
    * @return  The minimum size of a titlebar element on specified direction.
    */
   @Override
   public int getMenuBarElementMinSize(boolean horiz)
   {
      return horiz ? 32 : 19;
   }
   
   /**
    * Obtain the left padding of the test in a menu item. 
    *
    * @return  left padding of the test in a menu item. 
    */
   @Override
   public int getMenuLeftPad()
   {
      return MENU_LEFT_PAD;
   }
   
   /**
    * Obtain the right padding of the test in a menu item. 
    *
    * @return  right padding of the test in a menu item. 
    */
   @Override
   public int getMenuRightPad()
   {
      return MENU_RIGHT_PAD;
   }
   
   /**
    * Obtain the width of the check mark to for togglable menus. 
    *
    * @return  the width of the check mark in menu items.
    */
   @Override
   public int getMenuCheckmarkWidth()
   {
      return MENU_CHECKMARK_WIDTH;
   }
   
   /**
    * Obtain the width of the arrow marking menus with submenus (including additional space). 
    *
    * @return  the width of the arrow mark for sub menus.
    */
   @Override
   public int getMenuSmArrowWidth()
   {
      return MENU_SM_ARROW_WIDTH;
   }
   
   /**
    * Obtain the insets of a popup menu. They are used for drawing the box/ 3D border around the
    * popup menu.
    *
    * @return  the insets of a popup menu. 
    */
   @Override
   public NativeInsets getMenuInsets()
   {
      return new NativeInsets(3, 3, 3, 3);
   }
   
   /**
    * Obtain the fixed height of a separator in a popup menu widget.
    *
    * @return  the height of a separator in a popup menu.
    */
   @Override
   public int getMenuSeparatorHeight()
   {
      return 17;
   }
   
   /**
    * Obtain the height of a {@link MenuElement} in a menu widget.
    *
    * @return  the height of a normal item in a menu widget.
    */
   @Override
   public int getMenuItemHeight(MenuElement menuItem)
   {
      return 17; // use font size to compute it 
   }
   
   /**
    * Renders the background of a menu widget. 
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   popup
    *          {@code true} for popup menus.
    */
   @Override
   public void drawMenuBackground(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean popup)
   {
      ColorRgb bgColor    = gd.getSysColor(SysColor.COLOR_MENU);
      ColorRgb darkShadow = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      
      gd.setColor(bgColor);
      if (!popup)
      {
         gd.fillRect(x, y, width, height);
      }
      else
      {
         gd.setColor(bgColor);
         gd.fillRect(x, y, width, height);
         gd.fill3DRect(1, 1, width - 1, height - 1, true);
      
         // shadow
         gd.setColor(gd.getSysColor(SysColor.COLOR_MENUTEXT));
         gd.drawLine(x,         y + height, x + width, y + height);
         gd.drawLine(x + width, y,          x + width, y + height);
         gd.setColor(darkShadow);
         gd.drawLine(x + 1,         y + height - 1, x + width - 1, y + height - 1);
         gd.drawLine(x + width - 1, y + 1,          x + width - 1, y + height - 1);
      }
   }
   
   /**
    * Renders the separator line in a menu widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawMenuSeparator(GuiDriver gd, int x, int y, int width, int height)
   {
      NativeInsets insets = getMenuInsets();
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawLine(x + insets.left, y + height / 2, x + width - insets.right, y + height / 2);
   }
   
   /**
    * Renders the foreground content of menu item in a popup menu. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   textHeight
    *          The height of the text.
    * @param   focused
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuItemContent(AbstractWidget item,
                                   GuiDriver gd,
                                   GuiFontResolver gf,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   int textHeight,
                                   boolean focused)
   {
      ColorRgb chkMarkColor = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
      ColorRgb lightShadow  = gd.getSysColor(SysColor.COLOR_WINDOW);
      ColorRgb darkShadow   = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb fontColor    = gd.getSysColor(focused ? SysColor.COLOR_HIGHLIGHTTEXT : SysColor.COLOR_BTNTEXT);
      
      gd.setColor(fontColor);
      gd.setGuiFont(gf.font().font);
      gd.setFontStyle(gf.font().font.style);
      
      // only for menubar?
      NativePoint tp = new NativePoint(getMenuLeftPad(), textHeight);
      String title;
      MnemonicInfo label;
      boolean disabled = !item.isEnabled();
      
      if (item instanceof MenuItemGuiImpl)
      {
         MenuItemGuiImpl mi = (MenuItemGuiImpl) item;
         MenuItemConfig cfg = mi.config();
         
         // the space for toggle box. Printed for all menu-items in sub-menus or pop-up menus
         int checkmarkWidth = getMenuCheckmarkWidth();
         if (cfg.toggleBox && cfg.checked)
         {
            
            ColorRgb color = (mi.isInitDisabled() || !cfg.enabled) ? darkShadow : chkMarkColor;
            tp.y -= 4 + textHeight / 2;
            tp.x -= checkmarkWidth + 3;
            drawCheckmark(gd, tp, color);
            tp.x += checkmarkWidth + 3;
            tp.y += 4 + textHeight / 2;
            gd.setColor(fontColor);
         }
         tp.x += checkmarkWidth;
         
         title = mi.getPreparedTitle();
         label = mi.getPreparedLabel();
         
         disabled = disabled || mi.isInitDisabled();
      }
      else // sub-menu item 
      {
         SubMenuGuiImpl sm = (SubMenuGuiImpl) item;
         
         if (Menu.isMenubarElement(sm))
         {
            tp.x--;
         }
         else
         {
            tp.x += getMenuCheckmarkWidth();
         }
         
         title = sm.getPreparedTitle();
         label = sm.getPreparedLabel();
         
         tp.y--;
      }
      
      // draw the text:
      MenuElement<?> me = (MenuElement<?>) item;
      if (disabled)
      {
         gd.setColor(lightShadow);
         drawMenuTitle(me, gd, gf, tp, title, label);
         gd.setColor(darkShadow);
         drawMenuTitle(me, gd, gf, new NativePoint(tp.x + 1, tp.y + 1), title, label);
      }
      else
      {
         drawMenuTitle(me, gd, gf, tp, title, label);
      }
      
      int maxTitleWidth = SubMenuGuiImpl.getLongestItemTitleWidth(item.parent(), null, gf, gd);
      tp.x += maxTitleWidth + 3;
      
      // and the optional accelerator:
      if (item instanceof MenuItemGuiImpl)
      {
         String accel = ((MenuItemGuiImpl) item).config().accelerator;
         if (accel != null && !accel.isEmpty())
         {
            gd.drawString(Keyboard.keyName(accel), tp.x, tp.y);
         }
      }
      else
      {
         ImageWrapper<BufferedImage> arrow = VectorGraphicsHelper.getScalableImage(
               !disabled ? focused ? "sb-arr-sel-right" : "sb-arr-en-right" : "sb-arr-dis-right",
               MENU_SM_ARROW_WIDTH, MENU_SM_ARROW_WIDTH);
         if (arrow != null && arrow.getImage() != null)
         {
            gd.drawImage(arrow, width - getMenuSmArrowWidth(), (height - arrow.getHeight()) / 2,
                  arrow.getWidth(), arrow.getHeight(), 0, 0, false, false, true, false);
         }
      }
   }
   
   /**
    * Draws title string at the specified point.
    *
    * @param   menuElement
    *          MenuItem or SubMenu widget.
    * @param   gd
    *          GUI driver.
    * @param   gf
    *          GUI font resolver.
    * @param   tp
    *          Target point.
    * @param   title
    *          The title of menu widget. 
    * @param   label
    *          Prepared title string.
    *
    * @throws  IllegalArgumentException
    *          When menuWdg is not a MenuItem or SubMenu.
    */
   @SuppressWarnings("rawtypes")
   private static void drawMenuTitle(MenuElement<?> menuElement,
                                     GuiDriver gd,
                                     GuiFontResolver gf,
                                     NativePoint tp,
                                     String title,
                                     MnemonicInfo label)
   {
      MenuGuiImpl menu = (MenuGuiImpl) Menu.getMenu(menuElement);
      
      if (menu.isMnemonicsOn() && label != null && label.getMnemIdx() != null)
      {
         gd.drawStringWithMnemonic(title, tp, label, gf, (AbstractWidget) menuElement, false);
      }
      else
      {
         gd.drawString(title, tp.x, tp.y);
      }
   }
   
   /**
    * Draws an element of a menu in a menubar.
    *
    * @param   element
    *          The {@code MenuElement} widget to be painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          GUI font resolver.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuBarElement(MenuElement element,
                                  GuiDriver gd,
                                  GuiFontResolver gf,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean hover)
   {
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      
      // pressed ?
      if (element.hasFocus())
      {
         gd.draw3DRect(x, y, width - 1, height - 1, false);
         
         // down-right border
         gd.setColor(gd.getSysColor(SysColor.COLOR_CAPTIONTEXT));
         gd.drawLine(x + width - 1, y,              x + width - 1, y + height - 1);
         gd.drawLine(x,             y + height - 1, x + width - 1, y + height - 1);
      }
      // convex
      // mouseEntered=true when we used the mouse to select sub-menu. 
      // If we used keyboard than it just will have focus.
      else if (hover && element.hasFocus())
      {
         gd.draw3DRect(x, y, width - 1, height - 1, true);
         
         // left-top border
         gd.setColor(gd.getSysColor(SysColor.COLOR_CAPTIONTEXT));
         gd.drawLine(x, y, x,             y + height - 2);
         gd.drawLine(x, y, x + width - 2, y);
      }
      
      String title = element.getPreparedTitle();
      MnemonicInfo label = element.getPreparedLabel();
      // set up menu bar font and color
      gd.setGuiFont(gf.font().font);
      gd.setFontStyle(gf.font().font.style);
      int fh = FontManager.getFontHeight(WindowManager.resolveWindow(element),
                                         gf.resolveFontNum(), gd);
      NativePoint txtPt = new NativePoint(getMenuLeftPad() + 3, fh);
      // enabled or disabled?
      if (element.isEnabled())
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
         // press effect means +1 pixel to both X and Y coord for text in classic theme
         ClassicTheme.drawMenuTitle(element, gd, gf,
                                    element.isCurrent()
                                       ? new NativePoint(txtPt.x + 1, txtPt.y + 1) : txtPt,
                                    title, label);
      }
      else
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         ClassicTheme.drawMenuTitle(element, gd, gf, new NativePoint(txtPt.x + 1, txtPt.y + 1),
                                    title, label);
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         ClassicTheme.drawMenuTitle(element, gd, gf, txtPt, title, label);
      }
   }
   
   /**
    * Renders the background of an item in a menu widget. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuItemBackground(AbstractWidget item,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      boolean hover)
   {
      if (Menu.isMenubarElement(item))
      {
         System.out.println(
               "WARNING: Use Theme.drawMenuBarElement() instead to render a Menu Bar element");
         return;
      }
      
      if (hover || (!(item instanceof SubMenuGuiImpl) && item.hasFocus()))
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_MENUHILIGHT));
         gd.fillRect(x, y, width, height);
      }
      
      ThemeManager.debug(() -> {
         gd.setColor(java.awt.Color.GREEN);
         gd.drawRect(0, 0, width - 1, height - 1);
      });
   }
   
   /**
    * Renders a label.
    *
    * @param   label
    *          The {@link LabelGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   emptyMode
    *          If it is <code>true</code>, empty string is displayed.
    */
   @Override
   public void drawLabel(LabelGuiImpl label,
                         GuiDriver gd,
                         GuiColorResolver gc,
                         GuiFontResolver gf,
                         int x,
                         int y,
                         int width,
                         int height,
                         boolean emptyMode)
   {
      // clear background first
      gd.setColor(gc.bgColor);
      gd.fillRect(x, y, width, height);
   
      Widget<GuiOutputManager> sideWidget = label.getSideWidget();
      GuiFontResolver.FontCache fc = gf.font();
      final FontDetails<?> font = fc.font;
   
   
      boolean centered = (sideWidget != null);
      if (!emptyMode)
      {
         String txt = label.name();
      
         // needs to be centered vertically
         int txtHeight = label.getTextHeightNative(txt, gf.resolveFontNum());
         // top labels are centered differently
         int ytxt = centered ? height / 2 : txtHeight;
      
         gd.setGuiFont(font);
      
         int xtxt = 0;
         if (label.isRightAligned())
         {
            xtxt = Math.max(1, width - label.getTextWidthNative(txt, gf.resolveFontNum()));
         }
      
         gd.setColor(gc.fgColor);
      
         boolean withMnemonic = false;
         if (sideWidget != null && sideWidget instanceof MnemonicTrait)
         {
            MnemonicTrait w = (MnemonicTrait) sideWidget;
         
            MnemonicInfo info = w.getMnemonic();
            if (info != null && info.getMnemIdx() != null)
            {
               NativePoint tp = new NativePoint(xtxt, ytxt);
               gd.drawStringWithMnemonic(txt, tp, info, gf, label, centered);
            
               withMnemonic = true;
            }
         }
      
         if (!withMnemonic)
         {
            if (centered)
            {
               gd.drawStringCentered(txt, xtxt, ytxt);
            }
            else
            {
               gd.drawString(txt, xtxt, ytxt);
            }
         }
      }
   
      ThemeManager.debug(() ->
      {
         gd.setColor(centered ? java.awt.Color.RED : java.awt.Color.MAGENTA);
         gd.drawRect(x, y, width - 1, height - 1);
      });
   }
   
   /**
    * Draw the horizontal bar under a label if the label is underlined.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   fgColor
    *          The color attribute for the line.
    */
   @Override
   public void drawLabelUnderline(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  ColorRgb fgColor)
   {
      gd.setColor(fgColor);
      gd.drawLine(x, y, x + width - 1 , y + height - 1);
   }
   
   /**
    * Sets the composite color for XOR operations on browse widgets.  
    *
    * @param   gd
    *          the active {@code GuiDriver} used to draw the browse widget.
    */
   protected void setCompositeXorColor(GuiDriver gd)
   {
      gd.setColor(BROWSE_ROW_XOR_COLOR);
   }
   
   /**
    * Highlight the currently active row in a {@link BrowseGuiImpl} widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseRowHighlight(BrowseGuiImpl browse,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height)
   {
      for (int i = 0; i < browse.getCacheSize(); i++)
      {
         if (browse.getCachedRowNumber(i) != browse.getCurrentRow())
         {
            continue; // not the right row
         }
         
         // draw current row marker
         int left = 0;
         if (browse.config().box)
         {
            left += BrowseGuiImpl.BORDER_THICKNESS;
         }
         
         int markersColumnWidth = 0;
         
         // marker column should check the browse widget config for existence
         BrowseColumnGuiImpl markersColumn = browse.getCurrentMarkersColumn();
         if (browse.hasRowMarkers() && markersColumn != null)
         {
            NativeDimension dim = markersColumn.physicalDimension();
            if (dim != null)
            {
               markersColumnWidth = dim.width;
            }
         }
         
         left += markersColumnWidth;
         
         int top = 0;
         
         if (browse.config().box)
         {
            top += BrowseGuiImpl.BORDER_THICKNESS;
         }
         
         if (browse.hasTitle())
         {
            top += browse.getTitleHeight() + BrowseGuiImpl.BORDER_THICKNESS;
         }
         
         top += browse.getTopRowOffset(i);
         
         int viewportWidth = browse.getViewportDimension().width;
         viewportWidth -= markersColumnWidth;
         
         int columnsWidth = browse.getTotalColumnsWidth();
         columnsWidth -= markersColumnWidth;
         
         // honor scrolling
         int visibleWidth = columnsWidth + browse.getScrollableColumnSet().physicalLocation().x;
         int rowWidth = Math.min(visibleWidth, viewportWidth);
         
         gd.setLineStroke(LineStroke.DOTTED_LARGE);
         gd.setXORComposite();
         setCompositeXorColor(gd);
         gd.drawRect(left, top, rowWidth - 1, browse.getRowHeight() - 1);
         gd.resetComposite();
         gd.setLineStroke(LineStroke.DEFAULT);
         
         return;
      }
   }
   
   /**
    * Draw animation of a dragged column in a  {@link BrowseGuiImpl} widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseColumnDrag(BrowseGuiImpl browseGui,
                                    GuiDriver gd,
                                    int x,
                                    int y,
                                    int width,
                                    int height)
   {
      gd.setLineStroke(LineStroke.DOTTED_LARGE);
      setCompositeXorColor(gd);
      gd.setXORComposite();
      gd.drawRect(x, y, width, height);
      gd.resetComposite();
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Draw animation of a dragged column in a  {@link BrowseGuiImpl} widget.
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   left
    *          The left offset of the resized column.
    * @param   top
    *          The top offset where the columns start.
    * @param   right
    *          The right offset of the resized column. 
    * @param   bottom
    *          The bottom offset where the columns end.
    */
   @Override
   public void drawBrowseResizingColumn(BrowseGuiImpl browse,
                                        GuiDriver gd,
                                        int left,
                                        int top,
                                        int right,
                                        int bottom)
   {
      gd.setLineStroke(LineStroke.DOTTED_LARGE);
      setCompositeXorColor(gd);
      gd.setXORComposite();
      gd.drawLine(left, top, left, bottom);
      gd.drawLine(right, top, right, bottom);
      gd.resetComposite();
      gd.setLineStroke(LineStroke.DEFAULT);
   }

   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseTitlePanel(BrowseGuiImpl browse,
                                    GuiDriver gd,
                                    GuiColorResolver gc,
                                    GuiFontResolver gf,
                                    int x,
                                    int y,
                                    int width,
                                    int height)
   {
      boolean focused = browse.isDrawnAsFocused();
      String title = browse.config().title;
      ColorRgb titleInactiveBgColor = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb titleInactiveFgColor = gd.getSysColor(SysColor.COLOR_INACTIVECAPTIONTEXT);
      
      gd.setColor(focused ? gc.titleBgColor : titleInactiveBgColor);
      gd.fillRect(x, y, width, height);
      
      // calculate text position
      FontDetails font = gf.font().font;
      int textWidth = browse.getTextWidth(title, font);
      x += (width - textWidth) / 2;
      
      // draw title string
      gd.setColor(focused ? gc.titleFgColor : titleInactiveFgColor);
      gd.setGuiFont(font);
      gd.drawStringCentered(title, 
                            x, 
                            browse.getTitleHeight() / 2 + BrowseGuiImpl.BORDER_THICKNESS);
   }
   
   /**
    * Draw the label on a {@link BrowseColumnGuiImpl} widget.
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseColumnLabel(BrowseColumnGuiImpl column,
                                     GuiDriver gd,
                                     int x,
                                     int y,
                                     int width,
                                     int height)
   {
      BrowseColumnConfig cfg = column.config();
      BrowseGuiImpl browse = column.getBrowse();
      String[] labelLines = column.getLabelLines();
      
      drawBrowseColumnLabelRect(column, gd, 0, false, column.isMarkersColumn(), false);
      
      if (labelLines.length == 0)
      {
         return;
      }
      
      // draw label text
      NativePoint textClipPoint = column.getLabelTextClipPoint();
      NativeRectangle textClipRectangle = column.getLabelTextClipRectangle(textClipPoint);

      /* DEBUG: show label clip area
      gd.draw(textClipPoint, textClipRectangle, () ->
      {
         gd.setColor(new ColorRgb(230, 255, 179));
         gd.fillRect(0,0,500,500);
      });
      */

      for (int i = 0; i < labelLines.length; i++)
      {
         final int lineNum = i;

         gd.draw(textClipPoint, textClipRectangle, () ->
               drawBrowseCellText(column,
                                  gd,
                                  true,
                                  false,
                                  labelLines[labelLines.length - lineNum - 1],
                                  textClipRectangle,
                                  browse.getLabelLinesNum() - lineNum - 1,
                                  -1));
      }
      
      if (column.config().sortAscending != null)
      {
         int arrowTopOffset = (browse.getLabelsHeight() - SORT_ARROW_HEIGHT) / 2 +
                               BORDER_THICKNESS;
         int textWidth = column.getColumnLabelTextWidth();
         
         int sortAreaWidth = column.getSortAreaWidth(); // sort arrow + gap + sort number
         int labelAreaWidth = column.getLabelAreaWidth();
         
         int leftSideWidth = textClipPoint.x +
                             (Browse.isRightAligned(column) ?
                                textClipRectangle.width() :
                                Math.min(textWidth, textClipRectangle.width()));
         int rightSideWidth = sortAreaWidth + SORT_AREA_RIGHT_CLIP;
         int gap = labelAreaWidth - leftSideWidth - rightSideWidth;
         
         int sortAreaLeftClip = column.isDrawnAsPressed() ?
                                SORT_AREA_LEFT_CLIP - LABEL_PRESSED_OFFSET :
                                SORT_AREA_LEFT_CLIP;
         gap = Math.max(sortAreaLeftClip, Math.min(gap, SORT_ARROW_MAX_LEFT_OFFSET));
         
         NativePoint arrowClipPoint = new NativePoint(
               Math.max(leftSideWidth + gap, BORDER_THICKNESS + SORT_AREA_LEFT_CLIP),
               arrowTopOffset);
         
         // Draw sort arrow
         NativeRectangle arrowClipRect = AbstractWidget.clipRectangle(arrowClipPoint,
            new NativeDimension(
               Math.min(SORT_ARROW_WIDTH,
                  labelAreaWidth - BORDER_THICKNESS - SORT_AREA_LEFT_CLIP - SORT_AREA_RIGHT_CLIP),
               SORT_ARROW_HEIGHT));
         gd.draw(arrowClipPoint, arrowClipRect, () ->
         {
            ImageWrapper<?> arrowImage = browse.getSortArrowImage(cfg.sortAscending);
            gd.drawImage(arrowImage, 0, 0,
                         SORT_ARROW_WIDTH, SORT_ARROW_HEIGHT,
                         0, 0, false, false, false, false);
         });
         
         // Draw sort number
         if (cfg.sortNumber != null && cfg.sortNumber > 0)
         {
            int sortNumTopOffset = arrowTopOffset - (SORT_NUMBER_HEIGHT - SORT_ARROW_HEIGHT) / 2;
            
            NativePoint sortNumClipPoint = new NativePoint(0, 0);
            NativeRectangle sortNumClipRect = AbstractWidget.clipRectangle(sortNumClipPoint,
                  new NativeDimension(column.getLabelAreaWidth() - SORT_AREA_RIGHT_CLIP,
                                      column.physicalDimension().height));
            gd.draw(sortNumClipPoint, sortNumClipRect, () ->
            {
               ImageWrapper<?> numberImage = browse.getSortNumImage(cfg.sortNumber);
               gd.drawImage(numberImage,
                          arrowClipPoint.x + SORT_ARROW_WIDTH + GAP_BETWEEN_SORT_ARROW_AND_NUMBER,
                          sortNumTopOffset,
                          SORT_NUMBER_WIDTH,
                          SORT_NUMBER_HEIGHT,
                          0, 0, false, false, false, false);
            });
         }
      }
   }
   
   /**
    * Draw a browse marker column {@link BrowseColumnGuiImpl} widget (without label).
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseMarkersColumn(BrowseColumnGuiImpl column,
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height)
   {
      BrowseGuiImpl browse = column.getBrowse();
      int down = browse.getDown();
      int rowHeight = browse.getRowHeight();
      
      for (int i = 0; i < down; i++)
      {
         int cachedRowNumber = browse.getCachedRowNumber(i);
         boolean selected = browse.isRowSelected(cachedRowNumber);
         int topOffset = browse.getTopRowOffset(i);
         
         drawBrowseColumnLabelRect(column, gd, topOffset, true, false, selected);
         
         // current row - draw row marker
         if (cachedRowNumber == browse.getCurrentRow())
         {
            NativePoint p = new NativePoint(BrowseGuiImpl.BORDER_THICKNESS, 
                                            topOffset + BrowseGuiImpl.BORDER_THICKNESS);
            NativeDimension d = new NativeDimension(
                  column.physicalDimension().width - 2 * BrowseGuiImpl.BORDER_THICKNESS,
                  rowHeight - BrowseGuiImpl.BORDER_THICKNESS);
            NativeRectangle clip =
                  AbstractWidget.clipRectangle(p, new NativeDimension(d.width - 1, d.height - 1));
            
            gd.draw(p, clip, () ->
            {
               ImageWrapper markerImg = getBrowseMarkerArrow(selected);
               int offsetLeft = (d.width - markerImg.getWidth()) / 2 - 1;
               int offsetTop = (d.height - markerImg.getHeight()) / 2 - 1;
               
               gd.drawImage(markerImg,
                            offsetLeft, offsetTop, markerImg.getWidth(), markerImg.getHeight(),
                            0, 0, false, false, false, false);
            });
         }
      }
   }
   
   /**
    * Draw a normal column {@link BrowseColumnGuiImpl} widget (without label).
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseNormalColumn(BrowseColumnGuiImpl column,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height)
   {
      BrowseGuiImpl browse = column.getBrowse();

      // fill whole column background (except the label that is already drawn)
      gd.setColor(getColumnBackgroundColor(column));
      gd.fillRect(x, y, width, height);
      
      // draw cells
      int down = browse.getDown();
      int cacheSize = browse.getCacheSizeWithIntermediate();
      for (int i = 0; i < down; i++)
      {
         if (i < cacheSize)
         {
            String text = TOGGLE_BOX.equals(column.config().widgetType) ||
                          column.config().hyperlinked ?
                             null :
                             browse.getCellTextWithIntermediate(column, i);
            drawBrowseCell(column, gd, i, text);
         }
         else
         {
            drawBrowseCell(column, gd, i, null);
         }
      }
      
      // vertical separator
      if (browse.config().separators)
      {
         gd.setColor(getBrowseCellSeparatorColor(browse));
         gd.drawLine(x + width - 1, y, x + width - 1, y + height);
         
         //if (column.isLastColumn())
         //{
         //   setBrowseCellSeparatorColor(gd, browse);
         //   gd.drawLine(x + width - BrowseGuiImpl.BORDER_THICKNESS, y, 
         //               width - BrowseGuiImpl.BORDER_THICKNESS, y + height);
         //}
         //else if (column.isLastLockedColumn())
         //{
         //   setBrowseCellSeparatorColor(gd, browse);
         //   gd.drawLine(x + width, y, x + width, y + height);
         //}
      }

      // draw current toggle-box cell outline *over* separators
      if (TOGGLE_BOX.equals(column.config().widgetType) &&
          browse.isCurrentEditingColumn(column) &&
          browse.isCurrentRowInView())
      {
         int viewportRow = browse.getViewportRow();
         gd.setColor(ROW_TOGGLE_OUTLINE_COLOR);
         gd.setLineStroke(LineStroke.DEFAULT);
         gd.drawRect(0,
                     browse.getTopRowOffset(viewportRow),
                     column.physicalDimension().width - BrowseGuiImpl.BORDER_THICKNESS,
                     browse.getRowHeight() - BrowseGuiImpl.BORDER_THICKNESS);
      }

      // draw hyperlinks
      if (column.config().hyperlinked)
      {
         for (Widget widget : column.widgets())
         {
            widget.draw();
         }
      }
   }
   
   /**
    * Get row marker image for a row.
    * 
    * @param   selected
    *          {@code true} if the row is selected.
    *
    * @return row marker image for a row.
    */
   protected ImageWrapper getBrowseMarkerArrow(boolean selected)
   {
      final int ROW_MARKER_IMAGE_SIZE = 16;
      return VectorGraphicsHelper.getScalableImage(
               selected ? "sb-arr-sel-right" : "sb-arr-en-right",
               ROW_MARKER_IMAGE_SIZE, ROW_MARKER_IMAGE_SIZE);
   }

   /**
    * Draw the image of a non-edited in-browse toggle-box.
    *
    * @param  gd
    *         The {@code GuiDriver} used for drawing operations.
    * @param  column
    *         Column which contains the toggle-box.
    * @param  viewportRowIndex
    *         0-based index (in the viewport) of the row which contains the toggle-box.
    * @param  checked
    *         <code>true</code> if the toggle-box is checked, <code>false</code> it is not
    *         checked.
    */
   protected void drawRowToggleImage(GuiDriver gd,
                                     BrowseColumnGuiImpl column,
                                     int viewportRowIndex,
                                     boolean checked)
   {
      ImageWrapper image = getRowToggleImage(checked);
      NativePoint point = getRowToggleOffset(column, false);
      NativeDimension size = getRowToggleSize(column, false);
      gd.drawImage(image,
                   point.x,
                   column.getBrowse().getTopRowOffset(viewportRowIndex) + point.y,
                   size.width,
                   size.height,
                   0, 0, false, true, true, false);
   }

   /**
    * Get the image of a non-edited in-browse toggle-box.
    *
    * @param  checked
    *         <code>true</code> if the toggle-box is checked, <code>false</code> it is not
    *         checked.
    *
    * @return the image of a non-edited in-browse toggle-box.
    */
   protected ImageWrapper getRowToggleImage(boolean checked)
   {
      return VectorGraphicsHelper.getScalableImage(checked ? "row-toggle-on" : "row-toggle-off",
                                                   ROW_TOGGLE_SIZE,
                                                   ROW_TOGGLE_SIZE);
   }

   /**
    * Set the selected cell background color into the {@code gd} {@link GuiDriver}.
    *
    * @param   gd
    *          The {@code GuiDriver} to set the color for separators.
    * @param   unselColor
    *          The color for unselected cells. If {@code null} then the unselected cells' 
    *          background is not drawn. 
    */
   protected void setBrowseCellBackgroundSelectionColor(GuiDriver gd, ColorRgb unselColor)
   {
      gd.setColor(ColorManager.getSysColor(SysColor.COLOR_ACTIVECAPTION));
   }
   
   /**
    * Get the cell separator color.
    *
    * @param   browse
    *          The browse widget.
    *
    * @return  cell separator color.
    */
   @Override
   public ColorRgb getBrowseCellSeparatorColor(BrowseGuiImpl browse)
   {
      return browse.config().ehSeparatorsColor != null ?
                browse.config().ehSeparatorsColor :
                browse.getSeparatorColor();
   }

   /**
    * Get foreground color of the selected cell type.
    * 
    * @param   column
    *          The parent {@code BrowseColumnGuiImpl} of the cell. 
    * @param   label
    *          {@code true} for label cell, {@code false} for normal cell. 
    * @param   selected
    *          {@code true} if the row is selected.
    * @param   viewportRowIndex
    *          0-based number of the target row in the viewport (for cells only). <code>-1</code>
    *          for general cell foreground color.
    *
    * @return  foreground color of the selected cell type.
    */
   public ColorRgb getBrowseCellForegroundColor(BrowseColumnGuiImpl column,
                                                boolean label,
                                                boolean selected,
                                                int viewportRowIndex)
   {
      ColorRgb textColor;
      if (selected)
      {
         textColor = ColorManager.getSysColor(SysColor.COLOR_CAPTIONTEXT);
      }
      else if (label)
      {
         BrowseGuiImpl browse = column.getBrowse();
         textColor = browse.resolveColor(column.config().labelFgColor,
                                         column.config().ehLabelFgColor,
                                         browse.config().labelFgColor,
                                         browse.config().ehLabelFgColor,
                                         true,
                                         true);
      }
      else
      {
         BrowseGuiImpl browse = column.getBrowse();
         GuiCellAttributes cellAttr = null;
         if (viewportRowIndex >= 0)
         {
            cellAttr = browse.getCellAttr(viewportRowIndex, column.config().ordinal);
         }

         if (cellAttr != null && cellAttr.fgColor != null)
         {
            textColor = ColorManager.getFgColor(browse.window(), cellAttr.fgColor);
         }
         else
         {
            textColor = browse.resolveColor(column.config().columnFgColor,
                                            column.config().ehFgColor,
                                            browse.config().columnFgColor,
                                            browse.config().ehFgColor,
                                            true,
                                            false);
         }
      }
      
      return textColor;
   }

   /**
    * Get column background color.
    *
    * @param   column
    *          The browse column.
    *
    * @return  column background color.
    */
   @Override
   public ColorRgb getColumnBackgroundColor(BrowseColumnGuiImpl column)
   {
      BrowseGuiImpl browse = column.getBrowse();
      return browse.resolveColor(column.config().columnBgColor,
                                 column.config().ehBgColor,
                                 browse.config().columnBgColor,
                                 browse.config().ehBgColor,
                                 false,
                                 false);
   }

   /**
    * Get column label background color.
    *
    * @param   column
    *          The browse column.
    *
    * @return  column label background color.
    */
   @Override
   public ColorRgb getColumnLabelBackgroundColor(BrowseColumnGuiImpl column)
   {
      BrowseGuiImpl browse = column.getBrowse();
      return browse.resolveColor(column.config().labelBgColor,
                                 column.config().ehLabelBgColor,
                                 browse.config().labelBgColor,
                                 browse.config().ehLabelBgColor,
                                 false,
                                 true);
   }

   /**
    * Draw label-style rectangle.
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   topOffset
    *          Top offset, in pixels.
    * @param   markersBody
    *          <code>true</code> if it is a cell in the markers column.
    * @param   markersLabel
    *          <code>true</code> if it is the label area in the markers column.
    * @param   selectedRow    
    *          Is the current row selected ?
    */
   protected void drawBrowseColumnLabelRect(BrowseColumnGuiImpl column,
                                            GuiDriver gd,
                                            int topOffset,
                                            boolean markersBody,
                                            boolean markersLabel,
                                            boolean selectedRow)
   {
      NativeDimension d = column.physicalDimension();
      BrowseGuiImpl browse = column.getBrowse();
      int labelsHeight = browse.getLabelsHeight();
      ColorRgb selectedMarkersBgColor = ColorManager.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb markersColumnBgColor = ColorManager.getSysColor(SysColor.COLOR_3DFACE);
      ColorRgb shadowDarkColor = ColorManager.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb shadowLightColor = ColorManager.getSysColor(SysColor.COLOR_BTNHIGHLIGHT);
      ColorRgb labelBgColor = getColumnLabelBackgroundColor(column);
      ColorRgb labelBoxColor = column.getBrowse().getBrowseColorResolver().boxColor;
      
      int width = d.width;
      int height = markersBody ? browse.getRowHeight() + BrowseGuiImpl.BORDER_THICKNESS
                               : labelsHeight + 2 * BrowseGuiImpl.BORDER_THICKNESS;
      
      gd.setLineStroke(LineStroke.DEFAULT);
      gd.setColor(labelBoxColor);
      
      // adjust for proper drawing
      final int rectWidth = width - 1;
      final int rectHeight = height - 1;
      int c1 = column.isFirstColumn() ? 0 : BrowseGuiImpl.BORDER_THICKNESS;
      
      if (isBrowseLabelBorderDrawQuirk(column))
      {
         NativePoint clipPoint = new NativePoint(0, topOffset);
         NativeDimension clipDimension = new NativeDimension(rectWidth, rectHeight + 1);
         NativeRectangle clipRect = AbstractWidget.clipRectangle(clipPoint, clipDimension);
         
         gd.draw(clipPoint, clipRect, () -> gd.drawRect(0, 0, rectWidth, rectHeight));
      }
      else
      {
         gd.drawRect(-c1, topOffset, c1 + rectWidth, rectHeight);
      }
      
      ColorRgb c;
      if (selectedRow)
      {
         c = selectedMarkersBgColor;
      }
      else if (markersBody || markersLabel)
      {
         c = markersColumnBgColor;
      }
      else
      {
         c = labelBgColor;
      }
      gd.setColor(c);
      
      int fillWidth = width - 2 * BrowseGuiImpl.BORDER_THICKNESS;
      int fillHeight = height - 2 * BrowseGuiImpl.BORDER_THICKNESS;
      gd.fillRect(BrowseGuiImpl.BORDER_THICKNESS,
                  topOffset + BrowseGuiImpl.BORDER_THICKNESS,
                  fillWidth,
                  fillHeight);
      
      boolean pressedColumn = column.isDrawnAsPressed();
      gd.setColor(selectedRow || pressedColumn ? shadowDarkColor : shadowLightColor);
      gd.fillRect(BrowseGuiImpl.BORDER_THICKNESS - c1,
                  topOffset + BrowseGuiImpl.BORDER_THICKNESS,
                  fillWidth - BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS + c1,
                  BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS);
      gd.fillRect(BrowseGuiImpl.BORDER_THICKNESS - c1,
                  topOffset + BrowseGuiImpl.BORDER_THICKNESS,
                  BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS,
                  fillHeight);
      
      if (!pressedColumn)
      {
         gd.setColor(selectedRow ? shadowLightColor : shadowDarkColor);
         gd.fillRect(BrowseGuiImpl.BORDER_THICKNESS - c1,
                     topOffset + BrowseGuiImpl.BORDER_THICKNESS + fillHeight - BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS,
                     fillWidth + c1,
                     BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS);
         gd.fillRect(BrowseGuiImpl.BORDER_THICKNESS + fillWidth - BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS,
                     topOffset + BrowseGuiImpl.BORDER_THICKNESS + BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS,
                     BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS,
                     fillHeight - BrowseColumnGuiImpl.LABEL_SHADOW_THICKNESS);
      }
   }
   
   /**
    * Draw the cell.
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   viewportIndex
    *          0-based viewport index of the row the cell belongs to.
    * @param   text
    *          Cell text.
    */
   protected void drawBrowseCell(BrowseColumnGuiImpl column,
                                 GuiDriver gd,
                                 int viewportIndex,
                                 String text)
   {
      BrowseGuiImpl browse = column.getBrowse();
      boolean drawAsSelected = browse.drawRowAsSelected(viewportIndex);
      
      boolean drawBackground = false;
      ColorRgb defBgColor = null; // if null, do not draw background!
      GuiCellAttributes cellAttr =
            browse.getCellAttrWithIntermediate(viewportIndex, column.config().ordinal);
      if (cellAttr != null && cellAttr.bgColor != null)
      {
         defBgColor = ColorManager.getBgColor(browse.window(), cellAttr.bgColor);
      }
      
      if (drawAsSelected)
      {
         if (defBgColor == null)
         {
            defBgColor = getColumnBackgroundColor(column);
         }
         setBrowseCellBackgroundSelectionColor(gd, defBgColor);
         drawBackground = true;
      }
      else
      {
         if (defBgColor != null)
         {
            gd.setColor(defBgColor);
            drawBackground = true;
         }
      }
      
      if (drawBackground)
      {
         gd.fillRect(0, 
                     browse.getTopRowOffset(viewportIndex),
                     column.physicalDimension().width, browse.getRowHeight());
      }
      
      final NativeDimension d = column.physicalDimension();
      
      // cell text
      if (text != null)
      {
         NativePoint clipPoint = column.getCellTextClipPoint(viewportIndex);
         final NativeRectangle clipRectangle = column.getCellTextClipRectangle(clipPoint);
         
         /* DEBUG: show cell clip area
         gd.draw(clipPoint, clipRectangle, () ->
         {
            gd.setColor(new ColorRgb(204, 204, 255));
            gd.fillRect(0,0,500,500);
         });
         */
         
         gd.draw(clipPoint, clipRectangle, () -> drawBrowseCellText(column,
                                                                    gd,
                                                                    false,
                                                                    drawAsSelected,
                                                                    text,
                                                                    clipRectangle,
                                                                    0,
                                                                    viewportIndex));
      }

      if (TOGGLE_BOX.equals(column.config().widgetType))
      {
         int size = browse.getCacheSizeWithIntermediate();
         if (viewportIndex < size)
         {
            logical var = (logical) browse.getCellValueWithIntermediate(column, viewportIndex);
            if (var != null) // row inserted by INSERT-ROW
            {
               drawRowToggleImage(gd, column, viewportIndex, var.booleanValue());
            }
         }
      }
      
      // horizontal separator
      if (browse.config().separators)
      {
         gd.setColor(getBrowseCellSeparatorColor(browse));
         
         int y = browse.getTopRowOffset(viewportIndex) + browse.getRowHeight();
         gd.drawLine(0, y, d.width - BrowseGuiImpl.BORDER_THICKNESS, y);
      }
   }

   /**
    * Draw the text inside a label or cell.
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   label
    *          <code>true</code> to draw the text inside a label. <code>false</code> to draw the
    *          text inside a cell.
    * @param   selectedRow
    *          <code>true</code> if the row the cell belongs to is selected.
    * @param   text
    *          ext to draw.
    * @param   clipRectangle
    *          Clipping rectangle.
    * @param   lineNumber
    *          0-based number of line in the label/cell.
    * @param   viewportRowIndex
    *          0-based number of the target row in the viewport (for cells only).
    */
   protected void drawBrowseCellText(BrowseColumnGuiImpl column,
                                     GuiDriver gd,
                                     boolean label,
                                     boolean selectedRow,
                                     String text,
                                     NativeRectangle clipRectangle,
                                     int lineNumber,
                                     int viewportRowIndex)
   {
      gd.setColor(getBrowseCellForegroundColor(column, label, selectedRow, viewportRowIndex));
      
      FontDetails fd = label ? column.getLabelFont() : column.getCellsFont();
      gd.setGuiFont(fd);
      
      int y = label ? column.getLabelTextTopOffset() : column.getCellTextTopOffset();
      y += fd.legacyHeight * lineNumber;
      
      int x = 0;
      if (Browse.isRightAligned(column))
      {
         x = clipRectangle.width() - column.getTextWidth(text,
               label ? column.getLabelFont() : column.getCellsFont());
      }

      if (label)
      {
         // label drawing style can be theme dependent
         drawBrowseCellLabel(column, gd, text, x, y);
      }
      else
      {
         gd.drawString(text, x, y);
      }
   }
   
   /**
    * Draw the text inside a label or cell.
    *
    * @param   column
    *          The {@code BrowseColumnGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   text
    *          Text to draw.
    * @param   x
    *          Starting X coord of the text.
    * @param   y
    *          Starting Y coord of the text.
    */
   protected void drawBrowseCellLabel(BrowseColumnGuiImpl column,
                                      GuiDriver gd,
                                      String text,
                                      int x,
                                      int y)
   {
      gd.drawString(text, x, y);
   }
   
   /**
    * Draw a {@link ToolTip}.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   fnt
    *          The {@code FontDetails} to be used when drawing the text.
    * @param   text
    *          The actual text to be drawn.
    */
   @Override
   public void drawToolTip(GuiDriver gd,
                           int x,
                           int y,
                           int width,
                           int height,
                           FontDetails fnt,
                           String text)
   {
      // start batch drawing
      gd.startBatch();
      try
      {
         // another version of clipping
         gd.setDrawingArea(new NativeRectangle(new NativePoint(x, y),
                                               new NativeDimension(width, height)));
         
         // tooltip frame
         gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
         gd.drawRect(x, y, width - 1, height - 1);
         // tooltip background
         gd.setColor(gd.getSysColor(SysColor.COLOR_TOOLTIPBK));
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
         // tooltip text
         gd.setGuiFont(fnt);
         gd.setFontStyle(fnt.style);
         gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
         
         String lines[] = text.trim().split("\\r?\\n");
         int tH = (height - 4) / lines.length;
         int tY = y + tH / 2 + 2;
         for (String line : lines)
         {
            gd.drawStringCentered(line, x + 2, tY);
            tY += tH;
         }
      }
      finally
      {
         // draw immediately
         gd.setDrawingArea(null);
         gd.endBatch();
      }
   }
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    *
    * @param   frame
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawFrameTitle(FrameGuiImpl frame,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              GuiFontResolver gf,
                              int x,
                              int y,
                              int width,
                              int height)
   {
      String title = frame.config().title;
      
      // draw the frame title background depending on focused state
      gd.setColor(gc.titleBgColor);
      gd.fillRect(x, y, width, height);

      if (title == null)
      {
         return;
      }

      // calculate text position
      int textWidth = frame.getTextWidthNative(title, gf.resolveFontNum());
      x += (width - textWidth) / 2 - 1;
      
      // draw title string
      gd.setColor(gc.titleFgColor);
      gd.setGuiFont(gf.font().font);
      gd.drawStringCentered(title, x, height / 2);
   }

   /**
    * Get the offset of the in-browse toggle-box relatively to the containing cell.
    *
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the offset of the in-browse toggle-box relatively to the containing cell, in native
    *         units.
    */
   public NativePoint getRowToggleOffset(BrowseColumnGuiImpl column, boolean editor)
   {
      return getRowToggleOffsetCentered(column, editor);
   }

   /**
    * Get the size of the in-browse toggle-box.
    *
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the size of the in-browse toggle-box, in native units.
    */
   public NativeDimension getRowToggleSize(BrowseColumnGuiImpl column, boolean editor)
   {
      // overridden in other themes
      return new NativeDimension(ROW_TOGGLE_SIZE, ROW_TOGGLE_SIZE);
   }
   
   /**
    * Draw the line border {@link LineBorderGuiImpl} widget.
    *
    * @param   border
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawLineBorder(LineBorderGuiImpl border,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              int x,
                              int y,
                              int width,
                              int height)
   {
      NativeInsets ni = border.getNativeInsets();
      
      // setup line color and width
      gd.setColor(gc.boxColor != null ? gc.boxColor : gc.fgColor);
      
      // determine if insets are different; if not, set the line-stroke only once
      boolean equalInsets = ni.top == ni.left && ni.top == ni.bottom && ni.top == ni.right;
      if (equalInsets)
      {
         gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.top);
      }
      
      // top line
      if (ni.top > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.top);
         }
         
         gd.drawLine(x, y, width, y);
      }
      
      // left line
      if (ni.left > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.left);
         }
         
         gd.drawLine(x, y, x, y + height);
      }
      
      // bottom line
      if (ni.bottom > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.bottom);
         }
         
         gd.drawLine(x, y + height - ni.bottom, x + width, y + height - ni.bottom);
      }
      
      // right line
      if (ni.right > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.right);
         }
         
         gd.drawLine(x + width - ni.right, y + height, x + width - ni.right, y);
      }
      
      // restore the line stroke
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Draws a {@link RectangleGuiImpl} widget.
    *
    * @param   rect
    *          The {@code RectangleGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawRectangle(RectangleGuiImpl rect,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             int x,
                             int y,
                             int width,
                             int height)
   {
      RectangleConfig cfg = rect.config();
      
      if (cfg.groupBox)
      {
         // in group box, fill, roundness and line thickness are ignored  
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW)); //dark
         gd.drawRect(x, y, width - 2, height - 2);
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW)); //light 
         gd.drawRect(x + 1, y + 1, width - 2, height - 2);
         return;
      }
      
      int lt = Math.max(cfg.edgePixels, 0); // line thickness
      
      Frame<?> frOwn = UiUtils.locateFrame(rect);
      boolean is3D = cfg.edgePixels > 1 && !cfg.rounded && frOwn != null && frOwn.config().threeD; 
      
      // rounding diameter in pixels
      int ROUNDNESS = cfg.groupBox ? 5 : 11;
      
      if (cfg.filled)
      {
         // fill only the inner part of the rectangle, leaving around space for line thickness 
         gd.setColor(gc.bgColor);
         if (cfg.rounded)
         {
            gd.fillRoundRect(x + lt - 1, y + lt - 1,
                             width - 2 * lt + 1, height - 2 * lt + 1, ROUNDNESS);
         }
         else
         {
            gd.fillRect(x + lt, y + lt, width - 2 * lt, height - 2 * lt);
         }
      }
      
      if (lt == 0)
      {
         return;
      }
      
      if (is3D)
      {
         // if 3D, it is always square
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW)); //dark
         gd.drawLine(x, y, x + width - 2, y);
         gd.drawLine(x, y, x,             y + height - 2);
         gd.drawLine(x + width - lt, y + lt         , x + width - lt, y + height - lt);
         gd.drawLine(x + lt,         y + height - lt, x + width - lt, y + height - lt);
         
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW)); //light 
         gd.drawRect(x + width - 1, y + 1,          x + width - 1, y + height - 1);
         gd.drawRect(x + 1,         y + height - 1, x + width - 1, y + height - 1);
         gd.drawLine(x + lt - 1, y + lt - 1, x + width - lt, y + lt - 1);
         gd.drawLine(x + lt - 1, y + lt - 1, x + lt - 1,     y + height - lt);
      }
      else
      {
         gd.setColor(gc.fgColor);
         gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, lt);
         if (cfg.rounded)
         {
            gd.drawRoundRect(x + lt / 2, y + lt / 2, width - lt, height - lt, ROUNDNESS);
         }
         else
         {
            gd.drawRect(x + lt / 2, y + lt / 2, width - lt, height - lt);
         }
         gd.setLineStroke(LineStroke.DEFAULT, 1);
      }
   }
   
   /**
    * Draws a {@link ThreeDBorderGuiImpl} widget.
    *
    * @param   border
    *          The {@code RectangleGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void draw3DBorder(ThreeDBorderGuiImpl border,
                            GuiDriver gd,
                            int x,
                            int y,
                            int width,
                            int height)
   {
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawLine(x, y, x,             y + height - 2);
      gd.drawLine(x, y, x + width - 2, y);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DDKSHADOW));
      gd.drawLine(x + 1, y + 1, x + 1,         y + height - 3);
      gd.drawLine(x + 1, y + 1, x + width - 3, y + 1);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DLIGHT));
      gd.drawLine(x + 1,         y + height - 2, x + width - 2, y + height - 2);
      gd.drawLine(x + width - 2, y + height - 2, x + width - 2, y + 1);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x,             y + height - 1, x + width - 1, y + height - 1);
      gd.drawLine(x + width - 1, y + height - 1, x + width - 1, y);
      
      // be paranoid, restore the line stroke
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Renders the content of an {@link EditorGuiImpl} widget.
    *
    * @param   editor
    *          The {@code EditorGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   selStart
    *          selection start anchor.
    * @param   selEnd
    *          selection end anchor.
    * @param   dx
    *          the full horizontal delta (left-side text which is scrolled by caret or by mouse).
    * @param   dy
    *          offset when scrolling vertically via the scrollbar (the cursor does not move).
    */
   public void drawEditorContent(EditorGuiImpl editor,
                                 GuiDriver gd,
                                 GuiFontResolver gf,
                                 GuiColorResolver gc,
                                 int x,
                                 int y,
                                 int width,
                                 int height,
                                 NativePoint selStart,
                                 NativePoint selEnd,
                                 int dx,
                                 int dy)
   {
      gd.setGuiFont(gf.font().font);
      
      int fontDescent = 2; // TODO: use font.descent instead of constant 2
      
      int fontHeight = gf.font().fontHeight;
      
      boolean enabled = editor.isEnabled();
      int first = editor.config().scrollY + dy;
      int last = Math.min(first + editor.config().innerLines, editor.getLineCount());
      
      TextLookAndFeel textLookAndFeel = getTextLookAndFeel(gd, gc, enabled, false);
      
      for (int line = first; line < last; line++)
      {
         
         TextLineSelection textLine = editor.getTextSelection(line);
         
         if (textLine != null)
         {
            gd.drawTextSelection(textLine,
                                 textLookAndFeel,
                                 x - dx,
                                 y,
                                 fontHeight,
                                 fontDescent);
         }
         else
         {
            String txt = editor.getLineText(line);
            
            gd.setColor(textLookAndFeel.getTextColor());
            
            // adjust for scrollX
            gd.drawString(txt, x - dx, y + fontHeight - fontDescent);
         }
         
         y += fontHeight;
      }
   }
   
   /**
    * Renders caret of the editor widget. 
    *
    * @param   editor
    *          The {@code EditorGuiImpl} widget whose caret is to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x1
    *          The left extremity of the caret. 
    * @param   y1
    *          The top extremity of the caret.
    * @param   x2
    *          The right extremity of the caret. 
    * @param   y2
    */
   @Override
   public void drawEditorCaret(EditorGuiImpl editor, GuiDriver gd, int x1, int y1, int x2, int y2)
   {
      drawCaret(gd, x1, y1, x2, y2);
   }
   
   /**
    * Utility method for drawing a dotted rectangle using XOR operation.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSelectionFocus(GuiDriver gd, int x, int y, int width, int height)
   {
      // prepare stroke:
      
      gd.setXORComposite();
      setCompositeXorColor(gd);
      gd.setLineStroke(LineStroke.DOTTED_SMALL);
      
      // draw the dotted rectangle 
      gd.drawRect(x, y, width, height);
      
      // restore back the default line style
      gd.setLineStroke(LineStroke.DEFAULT);
      gd.resetComposite();
   }

   /**
    * Get left offset of a text inside a browse label.
    *
    * @param  rightAligned
    *         <code>true</code> if the column is right-aligned. <code>false</code> if the column
    *         is left-aligned.
    *
    * @return see above.
    */
   public int getBrowseLabelOffsetLeft(boolean rightAligned)
   {
      return rightAligned ? BROWSE_LABEL_OFFSET_LEFT_RA : BROWSE_LABEL_OFFSET_LEFT;
   }

   /**
    * Get right offset of a text inside a browse label.
    *
    * @return see above.
    */
   public int getBrowseLabelOffsetRight()
   {
      return BROWSE_LABEL_OFFSET_RIGHT;
   }

   /**
    * Get left offset of a text inside a browse cell.
    *
    * @param  rightAligned
    *         <code>true</code> if the column is right-aligned. <code>false</code> if the column
    *         is left-aligned.
    *
    * @return see above.
    */
   public int getBrowseDataOffsetLeft(boolean rightAligned)
   {
      return rightAligned ? BROWSE_DATA_OFFSET_LEFT_RA : BROWSE_DATA_OFFSET_LEFT;
   }

   /**
    * Returns the text look and feel depending on the requested text properties to be selected or
    * unselected and enabled or disabled.
    * 
    * @param    gd
    *           The {@code GuiDriver} used for drawing operations.
    * @param    gc
    *           The {@code GuiColorResolver} used color selection.
    * @param    enabledTextProperty
    *           The true value defines a request to get the enabled text look and feel, otherwise 
    *           the false value defines a request for the look and feel of disabled text.
    * @param    verticallyCenteredText
    *           The true value defines a request to get the vertically centered text.
    * 
    * @return   The text look and feel.
    */
   @Override
   public TextLookAndFeel getTextLookAndFeel(GuiDriver gd,
                                             GuiColorResolver gc,
                                             boolean enabledTextProperty,
                                             boolean verticallyCenteredText)
   {
      TextBaseline textBaseline;
      
      if (verticallyCenteredText)
      {
         textBaseline = TextBaseline.MIDDLE;
      }
      else
      {
         textBaseline = TextBaseline.ALPHABETIC;
      }
      
      ColorRgb textColor;
      ColorRgb backgroundTextColor;
      ColorRgb selTextColor;
      ColorRgb selBackgroundTextColor;
      
      ColorRgb disabledTextColor;
      
      if (!enabledTextProperty)
      {
         disabledTextColor = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      }
      else
      {
         disabledTextColor = null;
      }
      
      if (enabledTextProperty)
      {
         selTextColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      }
      else
      {
         selTextColor = disabledTextColor;
      }
      
      selBackgroundTextColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      
      if (enabledTextProperty)
      {
         textColor = gc.fgColor;
      }
      else
      {
         textColor = disabledTextColor;
      }
      
      backgroundTextColor = gc.bgColor;
      
      return new TextLookAndFeelImpl(backgroundTextColor,
                                     textColor,
                                     selBackgroundTextColor,
                                     selTextColor,
                                     textBaseline);
   }

   /**
    * Returns text look and feel.
    * 
    * @param    gcd
    *           The gui color resolver used for normal text.
    * @param    gce
    *           The gui color resolver used for selected text.
    * 
    * @return   The Fillin text look and feel.
    */
   @Override
   public TextLookAndFeel getTextLookAndFeel(GuiColorResolver gcd, GuiColorResolver gce)
   {
      return new TextLookAndFeelImpl(gcd.bgColor,
                                     gcd.fgColor,
                                     gce.selectionBgColor,
                                     gce.selectionFgColor,
                                     TextBaseline.MIDDLE);
   }

   /**
    * Get the offset of the in-browse toggle-box, centered in the containing cell, relatively to
    * that cell.
    * .
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the offset of the in-browse toggle-box, centered in the containing cell, relatively
    *         to that cell, in native units.
    */
   protected NativePoint getRowToggleOffsetCentered(BrowseColumnGuiImpl column,
                                                    boolean editor)
   {
      int availWidth = column.physicalDimension().width;
      int availHeight = column.getBrowse().getRowHeight();
      NativeDimension d = getRowToggleSize(column, editor);
      int x = (availWidth - d.width) / 2;
      int y = (availHeight - d.height) / 2;
      return new NativePoint(x, y);
   }
   
   /**
    * There is 4GL quirk: right edge of the label's border is not drawn for the last locked
    * column. That is noticeable while scrolling the scrollable column set with
    * NO-COLUMN-SCROLLING option: you can see 1px of scrollable area under locked area in that
    * place.<p>
    * Because in FWD a browse column contains its right separator, not the left, scrollable
    * area has no leftmost border and we emulate the underlying part of the leftmost border by
    * fully drawing rightmost border of the locked area if the scroll position is 0.
    *
    * @param  column
    *         Column which is tested if it is the last locked column.
    *
    * @return <code>true</code> if the upper part of the column's right separator should NOT be
    *         drawn because of the quirk described above.
    */
   protected boolean isBrowseLabelBorderDrawQuirk(BrowseColumnGuiImpl column)
   {
      return column.isLastLockedColumn() &&
             !column.isLastColumn() &&
             column.getBrowse().getHorizontalScrollingPosition() > 0;
   }
   
   /**
    * Draws a File or Folder Thumbnail of from a file/dir chooser dialog.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    * @param   selected
    *          {@code true} if the thumbnail is currently selected.
    * @param   hover
    *          {@code true} if the mouse is currently hovering the thumbnail.
    * @param   focus
    *          {@code true} if the thumbnail has focus.
    */
   @Override
   public void drawFileChooserThumbnailBackground(GuiDriver gd,
                                                  int x,
                                                  int y,
                                                  int width,
                                                  int height,
                                                  boolean selected,
                                                  boolean hover,
                                                  boolean focus)
   {
      if (focus || (selected && hover))
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_HIGHLIGHT));
      }
      else if (selected && !hover)
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_INACTIVECAPTIONTEXT));
      }
      else
      {
         // leave it transparent
         return;
         // gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      }
      gd.fillRect(x, y, width, height);
      
      if (focus)
      {
         drawSelectionFocus(gd, x + 3, y + 2, width - 7, height - 5);
      }
   }
   
   /**
    * Obtain color information needed to render the foreground of a thumbnail.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   focused
    *          {@code true} if thumbnail is focused.
    * @param   selected
    *          {@code true} if thumbnail is selected.
    * @param   hover
    *          {@code true} if mouse hovers the thumbnail.
    * @param   filename
    *          {@code true} to obtain the color of the filename.
    * @param   infoLabel
    *          {@code true} to obtain the info label (for details view).
    * @param   infodata
    *          {@code true} to obtain the info value (for details view).
    *
    * @return  The foreground color that matches the requested state.
    */
   @Override
   public ColorRgb getFileChooserForegroundColor(GuiDriver gd,
                                                 boolean focused,
                                                 boolean selected,
                                                 boolean hover,
                                                 boolean filename,
                                                 boolean infoLabel,
                                                 boolean infodata)
   {
      if (focused || (selected && hover))
      {
         // return gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
         return ColorRgb.WHITE;
      }
      else if (selected && !hover)
      {
         // return gd.getSysColor(filename ? SysColor.COLOR_INFOTEXT : SysColor.COLOR_WINDOWTEXT); 
         return filename ? ColorRgb.BLACK : ColorRgb.WHITE; 
      }
      else
      {
         // return gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
         return ColorRgb.BLACK;
      }
   }

   /**
    * Get the hyperlink font basing on the regular text font. E.g. hyperlinks may have the
    * same font as the regular text, but underlined.
    *
    * @param  windowId
    *         ID of the currentWindow.
    * @param  sourceFont
    *         Regular font.
    *
    * @return hyperlink font.
    */
   @Override
   public FontDetails getHyperlinkFont(WidgetId windowId, FontDetails sourceFont)
   {
      FontDetails res = sourceFont.copy();
      FontStyle fs = FontStyle.createStyle(res.style.isBold(), res.style.isItalic(), true);

      if (res.fontAlias != null)
      {
         res.fontAlias.key = FontManager.getFontKey(res.fontAlias.fontName,
                                                  res.fontAlias.pointSize,
                                                  res.style.isBold(),
                                                  res.style.isItalic(),
                                                  true);
         res.fontAlias.style = fs;
         res.fontAlias.font = null;
      }


      res.key = FontManager.getFontKey(res.fontName,
                                          res.pointSize,
                                          res.style.isBold(),
                                          res.style.isItalic(),
                                          true);
      res.style = fs;
      res = FontManager.registerFontDetails(windowId.asInt(), res);
      return res;
   }

   /**
    * Get the hyperlink color.
    *
    * @param  selectedRow
    *         <code>true</code> if the hyperlink belongs to the selected row.
    *
    * @return hyperlink color.
    */
   @Override
   public ColorRgb getHyperlinkColor(boolean selectedRow)
   {
      return selectedRow ? ColorManager.getSysColor(SysColor.COLOR_CAPTIONTEXT) : ColorRgb.BLUE;
   }
   
   /**
    * Draws the background of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   w
    *          The width of the painted widget/component. 
    * @param   h
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSliderBackground(SliderGuiImpl slider,
                                    GuiDriver gd,
                                    GuiColorResolver gc,
                                    int x,
                                    int y,
                                    int w,
                                    int h)
   {
      gd.setColor(slider.isThreeD() ? gd.getSysColor(SysColor.COLOR_3DFACE) : gc.bgColor);
      gd.fillRect(0, 0, w, h);
   }
   
   /**
    * Draws the slide track of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   w
    *          The width of the painted widget/component. 
    */
   @Override
   public void drawSliderTrack(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               int x,
                               int y,
                               int w)
   {
      ColorRgb lightShadow = gd.getSysColor(SysColor.COLOR_3DFACE);
      ColorRgb darkShadow  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);

      if (slider.config().horizontal)
      {
         gd.setColor(veryDarkShadow);
         gd.drawLine(x + 1, y + 1, x + w - 3, y + 1);
      
         gd.setColor(darkShadow);
         gd.drawLine(x, y, x + w - 2, y);
         gd.drawLine(x, y + 1, x, y + 2);
      
         gd.setColor(lightShadow);
         gd.drawLine(x + 1, y + 2, x + w - 2, y + 2);
         gd.drawLine(x + w - 2, y + 1, x + w - 2, y + 1);
      
         gd.setColor(gc.bgColor);
         gd.drawLine(x, y + 3, x + w - 1, y + 3);
         gd.drawLine(x + w - 1, y, x + w - 1, y + 2);
      }
      else
      {
         gd.setColor(veryDarkShadow);
         gd.drawLine(x + 1, y + 1, x + 1, y + w - 3);
      
         gd.setColor(darkShadow);
         gd.drawLine(x, y, x, y + w - 1);
         gd.drawLine(x + 1, y, x + 2, y);
      
         gd.setColor(lightShadow);
         gd.drawLine(x + 2, y + 1, x + 2, y + w - 2);
         gd.drawLine(x + 1, y + w - 2, x + 1, y + w - 2);
      
         gd.setColor(gc.bgColor);
         gd.drawLine(x + 3, y, x + 3, y + w - 1);
         gd.drawLine(x, y + w - 1, x + 2, y + w - 1);
      }
   }
   
   /**
    * Draws the slider thumb of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the thumb to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the thumb to be painted/component.
    * @param   w
    *          The width of the thumb. 
    * @param   h
    *          The height of the thumb.
    */
   @Override
   public void drawSliderThumb(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               int x,
                               int y,
                               int w,
                               int h)
   {
      // get slider config
      SliderConfig config = slider.config();
      TicMarks tmMode = slider.getTicmarkMode();
      boolean horiz = config.horizontal;

      // initialize the shadow box colors
      ColorRgb lightShadow = gd.getSysColor(SysColor.COLOR_3DFACE);
      ColorRgb darkShadow  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      ColorRgb veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      
      // the pointer drawing depends on orientation and tic marks mode
      if (tmMode == TicMarks.BOTH)
      {
         int x1 = horiz ? x - w / 2 : x - h / 2 + 2;
         int x2 = horiz ? x1 + w - 1: x1 + h - 1;
         int y1 = horiz ? y - h / 2 + 2 : y - w / 2;
         int y2 = horiz ? y1 + h - 1: y1 + w - 1;
         // 3d frame
         gd.setColor(gc.bgColor);
         gd.drawLine(x1, y1, x1, y2 - 1);
         gd.drawLine(x1 + 1, y1, x2 - 1, y1);
         gd.setColor(darkShadow);
         gd.drawLine(x2 - 1, y1 + 1, x2 - 1, y2 - 1);
         gd.drawLine(x1 + 1, y2 - 1, x2 - 2, y2 - 1);
         gd.setColor(veryDarkShadow);
         gd.drawLine(x2, y1, x2, y2);
         gd.drawLine(x1, y2, x2 - 1, y2);
         // and body
         gd.setColor(lightShadow);
         gd.fillRect(x1 + 1, y1 + 1, x2 - x1 - 2, y2 - y1 - 2);
         
         return;
      }
   
      // calculate image width and height
      int imgW = horiz ? w : h;
      int imgH = horiz ? h : w;
      
      // the name of the image, the base
      StringBuilder sbName = new StringBuilder("slider-tm");
      // different images are storing in different files
      
      switch (tmMode)
      {
         case TOP: 
            sbName.append("-top");
            break;
         case BOTTOM: 
            sbName.append("-bottom");
            break;
         case LEFT: 
            sbName.append("-left");
            break;
         case RIGHT: 
            sbName.append("-right");
            break;
         case NONE:
            sbName.append(horiz ? "-none-bottom" : "-none-right");
            break;
         default: // NOT_SET
            sbName.append(horiz ? "-notset-bottom" : "-notset-right");
            break;
      }
      
      // check the 3d aware version
      if (slider.isThreeD())
      {
         sbName.append("-3d");
      }
      if (this.isMouseHoverSensitive() && slider.highlighted() && slider.isEnabled())
      {
         // load highlighted image
         sbName.append("-hi");
      }
      ImageWrapper img = VectorGraphicsHelper.getScalableImage(sbName.toString(), imgW, imgH);
      
      if (img != null)
      {
         // calculate top-left thumb coordinates depending on orientation and thumb size
         int tlx = horiz ? x - imgW / 2 : x - imgH / 2;
         int tly = horiz ? y - imgH / 2 : y - imgW / 2;
         
         // adding extra pixels depending on the tic marks mode
         if (tmMode == TicMarks.BOTTOM && horiz)
         {
            tly += 3;
         }
         else if (tmMode == TicMarks.RIGHT && !horiz)
         {
            tlx -= 3;
            tly += 6;
         }
         else if (tmMode == TicMarks.TOP && horiz)
         {
            tly += 1;
         }
         else if (tmMode == TicMarks.LEFT && !horiz)
         {
            tlx -= 4;
            tly += 6;
         }
         else if (tmMode == TicMarks.NONE)
         {
            if (horiz)
            {
               tly += 3;
            }
            else
            {
               tlx -= 3;
               tly += 6;
            }
         }
         else if (tmMode == TicMarks.NOT_SET)
         {
            if (horiz)
            {
               tly += 2;
            }
            else
            {
               tly += 2;
            }
         }
         
         gd.drawImage(img, tlx, tly, img.getWidth(), img.getHeight(), 
                      0, 0, false, false, false, false);
      }
   }
   
   /**
    * Draws the slider tic marks of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   length
    *          The trackbar length.
    */
   @Override
   public void drawSliderTicMarks(SliderGuiImpl slider,
                                   GuiDriver gd,
                                   GuiColorResolver gc,
                                   int x,
                                   int y,
                                   int length)
   {
      // get slider config
      SliderConfig config = slider.config();
      TicMarks tmMode = slider.getTicmarkMode();
      
      // the length of the regular marks
      int TM_LENGTH = 2;
      int START_SHIFT = 11;
      
      // using the background color
      gd.setColor(getSliderTicmarkColor(gc));
      
      // distance between rule and tic mark line start, for Y dimension in horizontal slider and
      // for X dimension in vertical slider respectively
      int shift = START_SHIFT + (tmMode == TicMarks.BOTTOM || tmMode == TicMarks.RIGHT ? 4 : 0);
      
      // draw tic marks lines depends on orientation
      if (config.horizontal)
      {
         // tics above ruler 
         if (tmMode == TicMarks.TOP || tmMode == TicMarks.BOTH)
         {
            y += tmMode == TicMarks.TOP ? 1 : 0;
            // extra length marks on edges
            gd.drawLine(x, y - shift, x, y - shift - TM_LENGTH - 1);
            gd.drawLine(x + length, y - shift, x + length, y - shift - TM_LENGTH - 1);
            
            // regular tics between edges, when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int xt = slider.indexToCoord(x, length, true, i);
                  gd.drawLine(xt, y - shift, xt, y - shift - TM_LENGTH);
               }
            }
            // BOTH mode need to change the shift
            if (tmMode == TicMarks.BOTH)
            {
               shift += 3;
            }
         }
         // tics below ruler
         if (tmMode == TicMarks.BOTTOM || tmMode == TicMarks.BOTH)
         {
            y -= tmMode == TicMarks.BOTTOM ? 1 : 0;
            
            // extra length marks on edges
            gd.drawLine(x, y + shift, x, y + shift + TM_LENGTH + 1);
            gd.drawLine(x + length, y + shift, x + length, y + shift + TM_LENGTH + 1);
            
            // regular tics between edges, when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int xt = slider.indexToCoord(x, length, true, i);
                  gd.drawLine(xt, y + shift, xt, y + shift + TM_LENGTH);
               }
            }
         }
      }
      else
      {
         // tics left to the ruler
         if (tmMode == TicMarks.LEFT || tmMode == TicMarks.BOTH)
         {
            x += tmMode == TicMarks.LEFT ? 1 : 0;
            
            // extra length marks on edges
            gd.drawLine(x - shift, y, x - shift - TM_LENGTH - 1, y);
            gd.drawLine(x - shift, y + length, x - shift - TM_LENGTH - 1, y + length);
            
            // regular tics between edges, when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int yt = slider.indexToCoord(y, length, true, i);
                  gd.drawLine(x - shift, yt, x - shift - TM_LENGTH, yt);
               }
            }
            // BOTH mode need to change the shift
            if (tmMode == TicMarks.BOTH)
            {
               shift += 3;
            }
         }
         
         // tics right ruler
         if (tmMode == TicMarks.RIGHT || tmMode == TicMarks.BOTH)
         {
            x -= tmMode == TicMarks.RIGHT ? 1 : 0;
            
            // extra length marks on edges
            gd.drawLine(x + shift, y, x + shift + TM_LENGTH + 1, y);
            gd.drawLine(x + shift, y + length, x + shift + TM_LENGTH + 1, y + length);
            
            // regular tics between edges, when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int yt = slider.indexToCoord(y, length, true, i);
                  gd.drawLine(x + shift, yt, x + shift + TM_LENGTH, yt);
               }
            }
         }
      }
   }
   
   /**
    * Obtain the color used for drawing the tic-marks for this {@code Theme}. 
    *
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    *
    * @return  the color used for drawing the tic-marks. 
    */
   protected ColorRgb getSliderTicmarkColor(GuiColorResolver gc)
   {
      return gc.fgColor;
   }
   
   /**
    * Draws the dotted frame border of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   w
    *          The width of the painted widget/component. 
    * @param   h
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSliderFrame(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               int x,
                               int y,
                               int w,
                               int h)
   {
      // set dotted line style
      gd.setLineStroke(LineStroke.DOTTED_SMALL);
      gd.setColor(gc.fgColor);
      gd.drawRect(x, y, w - 1, h - 1);
      // restore default line style
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Draws the current screen value of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   w
    *          The width of the painted widget/component. 
    * @param   h
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSliderValue(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               GuiFontResolver gf, 
                               int x,
                               int y,
                               int w,
                               int h)
   {
      // obtain the current value
      String current = slider.getScreenValueAsString();
      if (current != null && !current.isEmpty())
      {
         boolean horiz = slider.config().horizontal;
         // compute the text coordinates
         int xt = (horiz ? x : x + w) + 1;
         int yt = (horiz ? y : y + h) - 4;
         // set up font style
         FontDetails fntDetails = gf.font().font;
         gd.setGuiFont(fntDetails);
         gd.setFontStyle(fntDetails.style);
         // draw the current value
         gd.setColor(gc.fgColor);
         gd.drawString(current, xt, yt);
      }
   }
   
   /**
    * Implements text look and feel immutable instances.
    */
   public static class TextLookAndFeelImpl
   implements TextLookAndFeel
   {
      /** The background color */
      private final ColorRgb backgroundColor;
      
      /** The text color */
      private final ColorRgb textColor;
      
      /** The selection background color */
      private final ColorRgb selectionBackgroundColor;
      
      /** The selection text color */
      private final ColorRgb selectionTextColor;
      
      /** The text baseline */
      private final TextBaseline textBaseline;
      
      
      /**
       * Creates this immutable instance. 
       * 
       * @param    backgroundColor
       *           The background color
       * @param    textColor
       *           The text color
       * @param    textBaseline
       *           The text baseline
       */
      public TextLookAndFeelImpl(ColorRgb backgroundColor,
                                 ColorRgb textColor,
                                 ColorRgb selectionBackgroundColor,
                                 ColorRgb selectionTextColor,
                                 TextBaseline textBaseline)
      {
         this.backgroundColor = backgroundColor;
         this.textColor = textColor;
         this.selectionBackgroundColor = selectionBackgroundColor;
         this.selectionTextColor = selectionTextColor;
         this.textBaseline = textBaseline;
      }

      /**
       * Gets the text background color.
       * 
       * @return   The text background color
       */
      @Override
      public ColorRgb getBackgroundColor()
      {
         return backgroundColor;
      }
      
      /**
       * Gets the text color.
       * 
       * @return   The text color
       */
      @Override
      public ColorRgb getTextColor()
      {
         return textColor;
      }
      
      /**
       * Gets the text baseline property.
       * 
       * @return   The text baseline property
       */
      @Override
      public TextBaseline getTextBaseline()
      {
         return textBaseline;
      }

      @Override
      public ColorRgb getSelectionBackgroundColor()
      {
         return selectionBackgroundColor;
      }

      @Override
      public ColorRgb getSelectionTextColor()
      {
         return selectionTextColor;
      }
   }
}
